/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   mhc.c:
 *
 *	Code used to compress images using the Modified Huffman algorithm.
 *	
 *	w_compressRowsMH():			compress one or more rows
 *						of image into Modified
 *						Huffman format.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "ccitt.pub"
#include "mh.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: mhc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");


/* This file contains wrappers for routines used to compress images
 * into Modified Huffman bytestreams.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * The routine compressRowsMH does the real work.  It returns
 * when either it fills the compressed data buffer or has done the right
 * number of lines.  The client tells it whether the image information
 * is new, the compressed data buffer is new and whether the compressed data
 * should be viewed as an extension of the previous block of compressed
 * data.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress rows of a binary image using the Modified Huffman technique.
 *
 * LongDescription:::
 *  Compress some rows of an image using the MH algorithm.  We will
 *  stop when either an error is encountered, we run out of image rows
 *  or we fill the compressed data buffer.
 *
 * Input:::
 *  pIPWState::             Pointer to standard IPWrap state block
 *  pMHCompInfoIn::         Pointer to private MH compression state block
 *  prS::                   Pointer to image to be compressed.
 *  startLine::             First image line to be compressed
 *			    relative to the top of the image, not frame.
 *  numLines::              Number of lines to be compressed.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *  continueImageBlock::    cTrue => we are compressing
 *			    from the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pMHCompInfoIn is
 *			    valid.<BR>
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *			    When this arg is true, startLine and numLines
 *			    should be set to the same value as they were
 *			    on the previous call.  The image should also
 *			    be the same.
 *  continueCompDataStream:: cTrue => The saved compression state
 *			    in pMHCompInfoIn is valid.  If
 *			    the previous call to this routine
 *			    filled the compressed data buffer
 *			    before the strip, tile or image was
 *			    complete, a new buffer should be
 *			    supplied.  This arg is set to 
 *			    cFalse when the block of compressed
 *			    data is unrelated to the block used
 *			    in the previous call to this routine
 *			    as recorded in pMHCompInfoIn.
 *			    SETTING THIS ARG TO cFalse CAUSES
 *			    DATA IN THE PIPELINE TO BE DISCARDED!
 *  flushPipeline::         If this arg is set, this routine
 *			    will flush all the compressed data
 *			    in the pipeline to the compressed
 *			    data buffer when the specified
 *			    number of scan lines have been
 *			    compressed.  The client should not
 *			    set flushPipeline if compressed
 *			    data from subsequent scan lines is
 *			    to be seamlessly concatenated to
 *			    the current compressed data stream.
 *			    When flushPipeline is set and the
 *			    compression task suceeds in
 *			    processing all image lines, all
 *			    saved state is marked invalid.
 *			    Subsequent calls with either
 *			    continueImageBlock or
 *			    continueCompDataStream will return
 *			    cErrInvalidParm.
 *
 * Output:::
 *  pExhaustedImageData::   Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it filled
 *			    the compressed data buffer.  This
 *			    routine may be called again with
 *			    a new buffer.
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been compressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  This routine compresses a horizontal slice of a binary image using the
 *  Modified Huffman technique described in the TIFF 6.0 spec.  There
 *  are several completion conditions:
 *
 *	retCode	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  If flushPipeline
 *		was set, all compressed data has been written to
 *		the compressed data buffer and the saved compressed state
 *		is marked invalid.  If flushPipeline was not set,
 *		some of the compressed data may still be in the
 *		pipeline and the saved compressed data state is valid.
 *		In the latter case, this routine may be called again with
 *		continueCompDataStream = cTrue and the new compressed
 *		data will be seamlessly concatenated with the old
 *		compressed data (on any bit boundary).  In either case, the
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  If flushPipeline
 *		was set, all compressed data has been written to the
 *		compressed data buffer, the saved state is marked
 *		invalid and it is not possible to continue from this state.
 *		Otherwise, some data may remain in the pipeline.
 *		This return state cannot occur of flushPipeline was set
 *		and the pipeline still contains data (see next state).
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine filled the compression data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed, but
 *		flushPipeline was set and some data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.<BR> 
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *<BR><BR>		
 *		The image data is always in PIXR order.
 *		
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_compressRowsMH (
    IPWState           *pIPWState,
    void               *pMHCompInfoIn,
    PIXR               *prS,
    UInt32              startLine,
    UInt32              numLines,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32              continueImageBlock,
    UInt32              continueCompDataStream,
    UInt32              flushPipeline,
    UInt32             *pExhaustedImageData,
    UInt32             *pExhaustedCompData,
    UInt32             *pPercentDone,
    UInt32             *pBytesWritten)
{
    PROLOGUE;
    Int32    w, bpl;
    UInt8   *pS;

    ERR_GEN( pMHCompInfoIn == NULL || prS == NULL || pCompData == NULL ||
	     pExhaustedImageData == NULL || pExhaustedCompData == NULL ||
	     pPercentDone == NULL || pBytesWritten == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

    ERR_PASS( i_compressRowsMH (pIPWState, pMHCompInfoIn,
			        pS,
			        w,
			        startLine + cFrameBits,
			        numLines,
			        bpl,
			        pCompData,
			        compDataBufSize,
			        continueImageBlock,
			        continueCompDataStream,
			        flushPipeline,
			        pExhaustedImageData,
			        pExhaustedCompData,
			        (Int32 *)pPercentDone,
			        (Int32 *)pBytesWritten));

    NORMAL_RETURN;
}


/*

 $Log: mhc.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.4  1998/01/14 00:13:55  ddavies
 mhc.c -> /product/ipcore/ipwrap/src/RCS/mhc.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.3  1996/05/11 01:29:39  ddavies
 mhc.c -> /product/ipcore/ipwrap/src/RCS/mhc.c,v

 Remove the padType arg from w_compressRowsMH.  It is now passed down in the
 private data structure since it doesn't change inside an image.

 Revision 1.2  1996/04/25 19:09:26  ddavies
 mhc.c -> /product/ipcore/ipwrap/src/RCS/mhc.c,v

 Fix comments and white space.  What a perfectionist!

 Revision 1.1  1996/04/10 22:24:02  ddavies
 Initial revision


 */
