/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   mmrd.c:
 *
 *	Code used to decompress images using the Modified Modified Read
 *	(MMR, or CCITT Group 4) algorithm.
 *	
 *	w_decompressRowsMMR():			decompress one or more rows
 *						of image into Modified
 *						Modified Read format.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: mmrd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/* This file contains wrappers for routines used to decompress images
 * from Modified Modified Read, a.k.a. MMR, a.k.a. CCITT Group 4 bytestreams.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * The routine w_decompressRowsMMR does the real work.  It returns
 * when either it fills the compressed data buffer or has done the right
 * number of lines.  The client tells it whether the image information
 * is new, the compressed data buffer is new and whether the compressed data
 * should be viewed as an extension of the previous block of compressed
 * data.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress rows of a binary image using the Modified Modified Read
 *  technique.
 *
 * LongDescription:::
 *  Decompress some rows of an image using the MMR algorithm.  We will
 *  stop when either an error is encountered, we run out of image rows
 *  or we empty the compressed data buffer.
 *
 * Input:::
 *  pIPWState::             Pointer to standard IPWrap state block
 *  pMMRDecompInfoIn::      Pointer to private MMR decompression state block
 *  startLine::             First image line to be decompressed
 *			    relative to the top of the image, not frame.
 *  numLines::              Number of lines to be compressed.
 *  combineResults::	    If set to cTrue, OR the decompressed image
 *			    with the original.  If set to cFalse, the
 *			    image is assumed to already be clear.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    read.  If several calls
 *			    to this routine are needed to read
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *  continueImageBlock::    cTrue => we are decompressing
 *			    into the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pMMRDecompInfoIn is
 *			    valid.<BR>
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *			    When this arg is true, startLine and numLines
 *			    should be set to the same value as they were
 *			    on the previous call.  The image should also
 *			    be the same.
 *  continueCompDataStream:: cTrue => The saved compression state
 *			    in pMMRDecompInfoIn is valid.  If
 *			    the previous call to this routine
 *			    emptied the compressed data buffer
 *			    before the strip, tile or image was
 *			    complete, a new buffer should be
 *			    supplied.  This arg is set to 
 *			    cFalse when the block of compressed
 *			    data is unrelated to the block used
 *			    in the previous call to this routine
 *			    as recorded in pMMRDecompInfoIn.
 *			    SETTING THIS ARG TO cFalse CAUSES
 *			    DATA IN THE PIPELINE TO BE DISCARDED!
 *  flushPipeline::         If this boolean is set, this routine
 *			    will not stop when all data has been
 *			    loaded into its pipeline.  It will attempt
 *			    to continue until all the scan lines have
 *			    been decompressed.  The client should not
 *			    set flushPipeline if compressed
 *			    data from subsequent scan lines is
 *			    to be seamlessly concatenated to
 *			    the current compressed data stream.
 *			    The client *should* set flushPipeline
 *			    when supplying exactly the correct amount
 *			    of compressed data for the scan lines
 *			    requested.
 *			    When flushPipeline is set and the
 *			    decompression task suceeds in
 *			    processing all image lines, all
 *			    saved state is marked invalid.
 *			    Subsequent calls with either
 *			    continueImageBlock or
 *			    continueCompDataStream will return
 *			    cErrInvalidParm.
 *
 * Output:::
 *  prD::                   Pointer to pixr that will contain
 *			    uncompressed image.  If NULL, no image will
 *			    be created.
 *  pExhaustedImageData::   Pointer to returned arg.  When the
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it emptied
 *			    the compressed data buffer.  This
 *			    routine may be called again with
 *			    a new buffer.  This arg will not be 
 *			    set if flushPipeline is set and we 
 *			    failed to decompress the specified 
 *			    number of scan lines (routine returns
 *			    cErrBufferExceeded instead).
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been decompressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesRead::            Number of compressed bytes actually
 *			    read from the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes may still be in the
 *			    pipeline.  These are included in
 *			    this count.
 *			    The client may safely dispose of
 *			    these bytes.
 *
 * Returns:::
 *  Normally, this routine thinks that emptying the compressed data
 *  buffer is ok.  However, if flushPipeline is set to cTrue and we empty
 *  the compressed data buffer, the routine returns cErrBufferExceeded.
 *
 * Notes:::
 *  This routine decompresses a horizontal slice of a binary image using the
 *  Modified Modified Read (MMR, or CCITT Group 4) technique described in
 *  the TIFF 6.0 spec.  There are several completion conditions:
 *
 *	retCode	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.  One possible
 *		catastrophe (cErrBufferExceeded) happens if the client
 *		set flushPipeline and failed to supply enough compressed
 *		data.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we decompressed the desired number of
 *		image lines without emptying the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.
 *		Some of the compressed data is still in the
 *		pipeline and the saved compressed data state is valid.
 *		This routine may be called again with
 *		continueCompDataStream = cTrue and the new compressed
 *		data will be seamlessly concatenated with the old
 *		compressed data (on any bit boundary).  The
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we emptied the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is normal when exactly the correct number of compressed
 *		data bytes were supplied.  It is very unusual for this
 *		condition to occur if flushPipeline was not set.
 *		It is not possible to continue from this state.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine emptied the compressed data buffer before completing
 *		the specified number of lines.  This can only happen if
 *		flushPipeline was not set, so some data remains in the pipeline.
 *		When this condition occurs, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.<BR>
 *		Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.<BR>
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.<BR>
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *<BR><BR>		
 *		The image data is always in PIXR order.
 *		
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how Modified Huffman
 *  really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_decompressRowsMMR (
    IPWState           *pIPWState,
    void               *pMMRDecompInfoIn,
    UInt32              startLine,
    UInt32              numLines,
    Bool		combineResults,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32              continueImageBlock,
    UInt32              continueCompDataStream,
    UInt32              flushPipeline,
    PIXR               *prD,
    UInt32             *pExhaustedImageData,
    UInt32             *pExhaustedCompData,
    UInt32             *pPercentDone,
    UInt32             *pBytesRead)
{
    PROLOGUE;
    Int32    bpl;
    UInt8   *pD;

    ERR_GEN( pMMRDecompInfoIn == NULL || pCompData == NULL ||
	     pExhaustedImageData == NULL || pExhaustedCompData == NULL ||
	     pPercentDone == NULL || pBytesRead == NULL, cErrNullPtr);

    if (prD != NULL)
    {
	ERR_GEN_PIXR(prD, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    
	ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
	ERR_PASS( w_pixrGetBpl(pIPWState, prD, &bpl));
    }
    else /* There's no image to write */
    {
	pD = NULL;
	bpl = 0;
    }

    ERR_PASS( i_decompressRowsMMR (pIPWState, pMMRDecompInfoIn,
				   pD,
				   startLine + cFrameBits,
				   numLines,
				   bpl,
				   combineResults,
				   pCompData,
				   compDataBufSize,
				   continueImageBlock,
				   continueCompDataStream,
				   flushPipeline,
				   pExhaustedImageData,
				   pExhaustedCompData,
				   (Int32 *)pPercentDone,
				   (Int32 *)pBytesRead));

    NORMAL_RETURN;
}


/*

 $Log: mmrd.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.5  1998/01/14 00:14:00  ddavies
 mmrd.c -> /product/ipcore/ipwrap/src/RCS/mmrd.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.4  1996/07/19 00:33:42  ddavies
 mmrd.c -> /product/ipcore/ipwrap/src/RCS/mmrd.c,v

 Change the initImage arg to the combineResults arg.  initImage wasn't much
 use since it was always slower than having the client clear the image.
 When combineResults is set, the decompressed image is ORed into the
 original bitmap.  When combineResults is clear, we assume the original
 bitmap was filled with white pixels.

 Revision 1.3  1996/05/21 19:15:04  ddavies
 mmrd.c -> /product/ipcore/ipwrap/src/RCS/mmrd.c,v

 Don't report an error if prD is null.  This is normal if we only want
 the runs from the encoded image.  Also don't derive parmeters from prD if
 it's NULL.  Finally, don't pass the image width down, it's already in
 the CCITTDecompressionInfo structure.

 Revision 1.2  1996/05/11 01:30:42  ddavies
 mmrd.c -> /product/ipcore/ipwrap/src/RCS/mmrd.c,v

 Fix comments.

 Revision 1.1  1996/05/06 19:04:04  ddavies
 Initial revision


 */

