/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  morph1.c
 *
 *      This file contains the basic morphological library operations of
 *      dilation, erosion, opening and closing.   Our definition of
 *      erosion and opening, however are non-standard, because we all
 *      SEs to be composed of hits, misses and don't-cares for these
 *      two operations.  Thus, our erosion is the operation referred
 *      to in the literature as the Hit-Miss operation, and our opening
 *      is the most useful of the "generalized openings", composed of a
 *      hit-miss followed by dilation with just the hits.
 *
 *      Morphological operations:
 *
 *             Dilation:
 *                   w_dilate()		dilates Hits only
 *
 *             Erosion:
 *                   w_erode()		erodes using both hits and misses
 *
 *             Open/Close:
 *                   w_open()		erosion followed by dilation
 *                   w_close()		dilation followed by erosion
 *
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "morph.pub"
#include "morph.prv"
#include "logicops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: morph1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*--------------------------------------------------------------------------*
 *                                Dilation                                  *
 *--------------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Using binary morphology, dilate an image.
 *
 * LongDescription:::
 *  Binary morphology has a long, complex history.  The term "binary" means
 *  that this version of the technique works only on binary images.
 *  It is used to identify objects of a given shape or texture.  One of the two
 *  primitive operations in morphology is called dilation.  In it,
 *  one starts with a binary input image and a small pattern.  The pattern
 *  has a distinguished location called the "center", though it may not
 *  actually be in the center.  Conceptually, one aligns the pattern's
 *  center point with every pixel in the input image.  At each point where
 *  the pattern's center point is over a black input pixel, the pattern
 *  is ORed into the output image.  This has the effect of extending the
 *  boundaries of objects in the input image in a way specified by the
 *  pattern.
 *
 * PrivateDescription:::
 *  In the accelerated implementations of the libraries, machine code
 *  is generated at run time for each structure element.  The machine code
 *  specifies the minimum set of shifts and ORs (and reads and writes)
 *  required to dilate some section of the image.
 *<BR><BR>
 *  In general, the performance of our version of morphology depends 
 *  primarily on the number of pixels in the input image,
 *  then on the number of columns in a structure element that have a
 *  HIT in them and lastly on the number of HITs in each column.
 *<BR><BR>
 *  On the SPARC platform, there are two classes of strels that are
 *  accelerated.  Purely vertical strels whose height is no more than 19
 *  are accelerated.  Other strels whose height is no more than 17 and whose
 *  distance from center to far left is no more than 31 and whose distance
 *  from center to far right is no more than 31 are also accelerated.
 *  On the x86 platforms (386 and above), strels where the distance from
 *  the center to the top, bottom, left and right sides are all less than
 *  or equal to 31 are accelerated.  Morphology using non-accelerated strels
 *  is done using a series of w_rasterop calls; one for each HIT in the
 *  strel.
 *<BR><BR>
 *  We have attempted to flush the instructure cache (if any) of the
 *  host processor immediately after creating the machine code for each
 *  strel.  Failure to do this properly results in extremely weird bugs.
 *  So far, only some versions of SPARC processor have instruction caches
 *  that must be flushed.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  strel::          Pointer to structure element specifying the dilation
 *		     pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  The other primitive operation is called erosion.  See w_erode for
 *  a description.
 *<BR><BR>
 *  BEWARE!! If the dest image equals the source image, this routine will
 *  make a COPY of the source image!!
 *<BR><BR>
 *  A structure element, like a convolution kernel, is a small 2D array of
 *  locations that are aligned with pixels in an input image.  Each position
 *  of the array may contain a HIT, MISS or DONT_CARE.  Our version of
 *  dilation ignores the MISSes.  When the center of the structure element
 *  is aligned with a black pixel in the input image, structure element points
 *  corresponding to HITs are ORed into the output images.  Structure element
 *  points corresponding to both MISSes and DONT_CAREs are ignored.
 *<BR><BR>
 *  Structure elements are often only 1D, i.e. purely horizontal or
 *  purely vertical.  People write theses on how large, concave, structure
 *  elements can be decomposed into multiple small ones with a general
 *  performance increase.  The only decomposition that is generally used
 *  here is that dilating with a solid n x m structure element (rectangular
 *  and all hits) will produce the same result as dilating with an n x 1
 *  and dilating the result with a 1 x m.  For larger strels, this is often
 *  faster since the number of HITs processed is so much smaller (n + m vs.
 *  n * m).
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  Most image processing books have a chapter on binary morphology and its
 *  uses.  Jean Serra is the grandfather of the field.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_dilate (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*prTemp = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* if the accelflag is set, create and use accelerated code;
	 *     if we can't make the code, use the slow code;
	 * if the accelflag is not set, use the slow code
	 */
    if (strel->accelFlag)
    {
	Int32   codeLength;
	
	if (!strel->dilateOp)    /* make the code */
	{
	    ERR_PASS_CLEANUP( w_genDilateAsm(pIPWState, strel, &codeLength));
	    if (!strel->dilateOp)
	    {  /* don't look again */
		strel->accelFlag = 0;
	    }
	    else
	    {	/* that worked, try making long code */
		ERR_PASS_CLEANUP( w_genLongDilateAsm(pIPWState, strel,
						     &codeLength));
		    /* if that didn't work, well, we won't get back here
		     * anyway since dilateOp is now non-NULL.
		     */
	    }
	}
    }

	/*  if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS_CLEANUP( w_copy(pIPWState, prS, &prTemp));
    else
	prTemp = prD;

	/* Do the dilate, either using the assembly code or using old fashioned
	 * rasterops. */
    ERR_PASS_CLEANUP( wp_fastDilate (pIPWState, prS, strel, prTemp));

	/*  if in-place, copy again and destroy the temporary image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_logicOp(pIPWState, prTemp, COPY, prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prTemp));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    if ((prD == prS) && (prTemp != NULL))
	w_destroyPixr(pIPWState, &prTemp);
    ERR_RETURN;
}


/*--------------------------------------------------------------------------*
 *                                 Erosion                                  *
 *--------------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Using binary morphology, erode an image.
 *
 * LongDescription:::
 *  Binary morphology has a long, complex history.  The term "binary" means
 *  that this version of the technique works only on binary images.
 *  It is used to identify objects of a given shape or texture.  One of the two
 *  primitive operations in morphology is called erosion.  In it,
 *  one starts with a binary input image and a small pattern.  The pattern
 *  has a distinguished location called the "center", though it may not
 *  actually be in the center.  Conceptually, one aligns the pattern's
 *  center point with every pixel in the input image.  At each point where
 *  the pattern matches in the image under it (the pattern contains places
 *  where the image must be black, called HITS, places where the image must
 *  be white, called MISSes, and places where the image is ignored, called
 *  DONT_CAREs), a black pixel is written to the destination image at the
 *  position corresponding to the position of the center pixel of the pattern.
 *
 * PrivateDescription:::
 *  In the accelerated implementations of the libraries, machine code
 *  is generated at run time for each structure element.  The machine code
 *  specifies the minimum set of shifts, inverts and AND operations
 *  (and reads and writes) required to erode some section of the image.
 *<BR><BR>
 *  In general, the performance of our version of morphology depends 
 *  primarily on the number of pixels in the input image,
 *  then on the number of columns in a structure element that have a
 *  HIT or MISS in them and lastly on the number of HITs and MISSes
 *  in each column.
 *<BR><BR>
 *  On the SPARC platform, there are two classes of strels that are
 *  accelerated.  Purely vertical strels whose height is no more than 19
 *  are accelerated.  Other strels whose height is no more than 17 and whose
 *  distance from center to far left is no more than 31 and whose distance
 *  from center to far right is no more than 31 are also accelerated.
 *  On the x86 platforms (386 and above), strels where the distance from
 *  the center to the top, bottom, left and right sides are all less than
 *  or equal to 31 are accelerated.  Morphology using non-accelerated strels
 *  is done using a series of w_rasterop calls; one for each HIT in the
 *  strel.
 *<BR><BR>
 *  We have attempted to flush the instructure cache (if any) of the
 *  host processor immediately after creating the machine code for each
 *  strel.  Failure to do this properly results in extremely weird bugs.
 *  So far, only some versions of SPARC processor have instruction caches
 *  that must be flushed.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  strel::          Pointer to structure element specifying the erosion
 *		     pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  The other primitive operation is called dilation.  See w_dilate for
 *  a description.
 *<BR><BR>
 *  BEWARE!! If the dest image equals the source image, this routine will
 *  make a COPY of the source image!!
 *<BR><BR>
 *  A structure element, like a convolution kernel, is a small 2D array of
 *  locations that are aligned with pixels in an input image.  Each position
 *  of the array may contain a HIT, MISS or DONT_CARE.  At each locatoin
 *  in the input image where the pattern described by the structure element
 *  matches the input image, a black pixel is written to the output image
 *  at the location corresponding to the designated "center" of the
 *  structure element.  Erodes can be used both to generally pull in the
 *  boundaries of objects in an image (structure element specifies a black
 *  blob that doesn't match near the edges of large black things) or to
 *  find specific shapes or textures in an image.
 *<BR><BR>
 *  Erosion is reasonably fast for small structure elements.  However, it
 *  is "brittle" as a matcher since a bad value at any pixel causes the
 *  match to fail.  A more robust, though often slower, method of finding
 *  shapes in an image uses the Hausdorff operators.
 *<BR><BR>
 *  Structure elements are often only 1D, i.e. purely horizontal or
 *  purely vertical.  People write theses on how large, concave, structure
 *  elements can be decomposed into multiple small ones with a general
 *  performance increase.  The only decomposition that is generally used
 *  here is that dilating with a solid n x m structure element (rectangular
 *  and all hits) will produce the same result as dilating with an n x 1
 *  and dilating the result with a 1 x m.  For larger strels, this is often
 *  faster since the number of HITs processed is so much smaller (n + m vs.
 *  n * m).
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  Most image processing books have a chapter on binary morphology and its
 *  uses.  Jean Serra is the grandfather of the field.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_erode (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*prTemp = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* if the accelflag is set, create and use accelerated code;
	 *     if we can't make the code, use the slow code;
	 * if the accelflag is not set, use the slow code
	 */
    if (strel->accelFlag)
    {
	Int32    codeLength;

	if (!strel->erodeOp)    /* make the code */
	{
	    ERR_PASS_CLEANUP( w_genErodeAsm(pIPWState, strel, &codeLength));
	    if (!strel->erodeOp)
	    {  /* don't look again */
		strel->accelFlag = 0;
	    }
	    else
	    {	/* that worked, try making long code */
		ERR_PASS_CLEANUP( w_genLongErodeAsm(pIPWState, strel,
						    &codeLength));
		    /* if that didn't work, well, we won't get back here
		     * anyway since erodeOp is now non-NULL.
		     */
	    }
	}
    }

	/*  if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS_CLEANUP( w_copy(pIPWState, prS, &prTemp));
    else
	prTemp = prD;

	/* Do the erode, either using the assembly code or using old fashioned
	 * rasterops. */
    ERR_PASS_CLEANUP( wp_fastErode (pIPWState, prS, strel, prTemp));

	/*  if in-place, copy again and destroy the temporary image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_logicOp(pIPWState, prTemp, COPY, prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prTemp));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    if ((prD == prS) && (prTemp != NULL))
	w_destroyPixr(pIPWState, &prTemp);
    ERR_RETURN;
}




/*--------------------------------------------------------------------------*
 *                                 Opening                                  *
 *--------------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Erode an image, then dilate the result with the same structure element.
 *
 * LongDescription:::
 *  Eroding an image tends to find things, but changes their shapes.
 *  For example, a complete pattern might turn into a single dot in the
 *  output image.  If we follow the erode op with a dilation using the
 *  same structure element, the shapes of the things that were found
 *  using the erode are restored.  Things that were elimated by the erode
 *  operation stay eliminated.
 *
 * PrivateDescription:::
 *  Since opens are constructed from erosions and dilations, the same
 *  rules about accelerated strels apply here.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  strel::          Pointer to structure element specifying the erosion and
 *		     dilation pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  This operation has two main uses.  First, it is used to simply break
 *  small connections between things without altering their basic shape.
 *  This is the origin of the name, "open."  Second, it is used to find
 *  things without altering their shape.
 *<BR><BR>
 *  To break small connections between things, the strel is normally a
 *  black blob of some shape.  For example, vertical strels break thin
 *  horizontal connections while preserving thin vertical ones.
 *  Everywhere the black blob does not fit entirely inside
 *  an object in the input image, the erode takes away points.  Small
 *  connections tend to disappear entirely while large objects just pull in
 *  a little.  The subsequent dilation restores the boundaries of the large
 *  things, but there's nothing left of the small stuff since there was
 *  nothing left to dilate.
 *<BR><BR>
 *  Finding things is similar.  The strel is the pattern or texture that
 *  should be matched.  Everywhere it's found, a dot is written into the
 *  the output image and everything else is discarded.  The dilation step
 *  restores only the matched objects.
 *<BR><BR>
 *  As was noted in the notes on w_erode, this can be a fairly fast method
 *  for finding stuff, but is fairly brittle in that any pixel can cause
 *  the erode to fail.  To find specific objects in an image, the Hausdorff
 *  operators are more robust, though often slower on small stuff..
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  Most image processing books have a chapter on binary morphology and its
 *  uses.  Jean Serra is the grandfather of the field.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_open (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*prT = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* Create an intermediate image and do the real work */
    ERR_PASS_CLEANUP( w_clonePixrNoInit(pIPWState, prS, &prT));

    ERR_PASS_CLEANUP( w_erode (pIPWState, prS, strel, &prT));
    ERR_PASS_CLEANUP( w_dilate (pIPWState, prT, strel, &prD));
    ERR_PASS_CLEANUP( w_destroyPixr (pIPWState, &prT));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    if (prT != NULL)
	w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*--------------------------------------------------------------------------*
 *                                 Closing                                  *
 *--------------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dilate an image, then erode the result with the same structure element.
 *
 * LongDescription:::
 *  This operation is used less often than w_open.  Its basic purpose is to
 *  close small gaps between objects (hence the name).  The dilation step
 *  expands the boundaries of objects.  If they are near each other, they
 *  merge.  Otherwise, they just spread out.  The erosion step pulls the
 *  boundaries back in.  If two objects have merged, the boundary has vanished,
 *  so is not affected.
 *
 * PrivateDescription:::
 *  Since closes are constructed from erosions and dilations, the same
 *  rules about accelerated strels apply here.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  strel::          Pointer to structure element specifying the erosion and
 *		     dilation pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  This operation has two main uses.  First, it is used to simply break
 *  small connections between things without altering their basic shape.
 *  This is the origin of the name, "open."  Second, it is used to find
 *  things without altering their shape.
 *<BR><BR>
 *  To break small connections between things, the strel is normally a
 *  black blob of some shape.  For example, vertical strels break thin
 *  horizontal connections while preserving thin vertical ones.
 *  Everywhere the black blob does not fit entirely inside
 *  an object in the input image, the erode takes away points.  Small
 *  connections tend to disappear entirely while large objects just pull in
 *  a little.  The subsequent dilation restores the boundaries of the large
 *  things, but there's nothing left of the small stuff since there was
 *  nothing left to dilate.
 *<BR><BR>
 *  Finding things is similar.  The strel is the pattern or texture that
 *  should be matched.  Everywhere it's found, a dot is written into the
 *  the output image and everything else is discarded.  The dilation step
 *  restores only the matched objects.
 *<BR><BR>
 *  As was noted in the notes on w_erode, this can be a fairly fast method
 *  for finding stuff, but is fairly brittle in that any pixel can cause
 *  the erode to fail.  To find specific objects in an image, the Hausdorff
 *  operators are more robust, though often slower on small stuff.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  Most image processing books have a chapter on binary morphology and its
 *  uses.  Jean Serra is the grandfather of the field.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_close (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*prT = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* Create an intermediate image and do the real work */
    ERR_PASS_CLEANUP( w_clonePixrNoInit(pIPWState, prS, &prT));

    ERR_PASS_CLEANUP( w_dilate (pIPWState, prS, strel, &prT));
    ERR_PASS_CLEANUP( w_erode (pIPWState, prT, strel, &prD));
    ERR_PASS_CLEANUP( w_destroyPixr (pIPWState, &prT));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

		/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    if (prT != NULL)
	w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*

 $Log: morph1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.24  1997/04/15 22:14:14  ddavies
 morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.23  1996/12/18 23:18:41  ddavies
 ddavies on Wed Dec 18 15:16:31 PST 1996

 morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v

 Modify w_erode and w_dilate so that they try to create machine code that
 uses long (>= 64 bit) operations along with creating normal machine code.

 Revision 1.22  1996/01/04 20:24:51  ejaquith
 morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.21  1995/12/13  04:10:27  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:13:45  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:03:04  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:27:11  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.17  1995/10/03  01:20:08  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.16  1995/09/21  22:54:37  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/15  18:00:29  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.14  1995/09/11  22:11:12  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.13  1995/09/10  00:33:11  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.12  1995/09/07  01:33:23  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:21:27  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  01:17:11  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.9  1995/08/31  21:15:17  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/25  01:37:25  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.7  1995/08/17  22:02:40  ejaquith
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.6  1995/07/29  04:43:57  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.5  1994/10/07  19:14:14  lperry
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.4  1994/06/08  04:12:57  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.3  1993/12/03  01:17:04  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:14:11  ddavies
 * morph1.c -> /product/ipcore/ipwrap/src/RCS/morph1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

