/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  morph2.c
 *
 *      This file contains the special case of rectangular dilation,
 *      which has been implemented using a different, faster method.
 *
 *
 *      Morphological operations:
 *
 *             Dilation:
 *                   w_rectDilate()     Dilate with HITs only
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "morph.pub"
#include "logicops.pub"
#include "shrpatt.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: morph2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*--------------------------------------------------------------------------*
 *                                Dilation                                  *
 *--------------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Special case, very fast image dilation
 *
 * LongDescription:::
 *  The concept of dilation in binary morphology is explained in the
 *  description of w_dilate().  This routine implements a very fast special
 *  case of dilation.  It uses rectangular structure elements containing
 *  only HITs where the dimensions are powers of two.<BR>
 *
 *  While sizes with width AND height both at least 8 will be faster than
 *  w_dilate(), sizes with both smaller will not.  However, when one
 *  dimension is at least 8 and the other is not, this routine will generally
 *  be faster, but you should check yourself if speed is an issue.<BR>
 *
 *  You can compare w_dilate() with this routine by calling
 *  w_makeBrickStrel() to get your STREL of the desired rectangular size
 *  before calling w_dilate() and calling w_destroyStrel() to clean it up.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to input image.
 *  nx::             Width of strel.  Must be 1, 2, 4, 8, 16, 32, 64 or 128.
 *  ny::             Height of strel.  Must be 1, 2, 4, 8, 16, 32, 64 or 128.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  The center point in these strels is always in the lower right corner.
 *<BR><BR>
 *  The time required to dilate an image is roughly the same for any of these
 *  strels!
 *<BR><BR>
 * BEWARE!! If the dest image equals the source image, this routine
 * makes a COPY of the source image!!
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  W. Rucklidge (of PARC), E. Jaquith
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rectDilate (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               nx,
    Int32               ny,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*prTemp = NULL;
    Int32	 wS, hS;
    UInt8	*pD, *pS;
    Int32	 sBpl, dBpl;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN((nx != 1) && (nx != 2) && (nx != 4) &&
	    (nx != 8) && (nx != 16) && (nx != 32) && (nx != 64) && (nx != 128),
	    cErrInvalidParm);
    ERR_GEN((ny != 1) && (ny != 2) && (ny != 4) &&
	    (ny != 8) && (ny != 16) && (ny != 32) && (ny != 64) && (ny != 128),
	    cErrInvalidParm);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/*  if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS_CLEANUP( w_copy(pIPWState, prS, &prTemp));
    else
	prTemp = prD;

	/* Things seem to be working, do the real work. */
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prTemp, &dBpl));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prTemp, &pD));

    ERR_PASS_CLEANUP( i_rectDilate (pIPWState, pD, pS, wS, hS,
				    dBpl, sBpl, nx, ny));

	/*  if in-place, copy again and destroy the temporary image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prTemp, COPY, prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prTemp));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* Hooray, it worked.  Return the dest pixr */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    if ((prD == prS) && (prTemp != NULL))
	w_destroyPixr(pIPWState, &prTemp);
    ERR_RETURN;
}


/*
 $Log: morph2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.17  1998/01/14 00:14:02  ddavies
 morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.16  1997/04/15 22:14:26  ddavies
 morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.15  1996/01/04 20:24:54  ejaquith
 morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.14  1995/12/13  04:10:31  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.13  1995/11/13  20:13:49  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.12  1995/10/30  18:00:33  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * Updated function block comments who named other IPWrap routines.
 * Added functionality to dilate with rectangular STRELs of size
 * 1, 2, and 4.  (as well as previous ability to do 8,16,32,64,128)
 * Updated comments for really readability/useability for client
 * to use (note the use of other named IPWrap routines that get
 * hyper-linked automagically in auto-doc documentation).
 *
 * Revision 1.11  1995/10/13  22:03:08  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.10  1995/10/05  04:27:15  ddavies
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.9  1995/10/03  01:20:12  ddavies
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.8  1995/09/21  22:54:40  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.7  1995/09/18  23:45:09  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * Updated comments at file's header
 *
 * Revision 1.6  1995/09/15  18:00:33  ddavies
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.5  1995/09/11  22:11:15  ddavies
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.4  1995/09/10  00:33:17  ddavies
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.3  1995/09/07  01:33:25  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.2  1995/09/06  20:21:29  ejaquith
 * morph2.c -> /product/ipcore/ipwrap/src/RCS/morph2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.1  1995/09/05  21:26:00  ejaquith
 * Initial revision
 *
 */

