/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  morph7.c
 *
 *      Composite low-level morphological operations for removing 
 *      various types of image components.  These operations are
 *      useful for higher-level image cleaning applications.
 *
 *	It's an accident of history that this isn't in libappsw.
 *
 *           w_removePepper()
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"
#include "morphg.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "morph.pub"
#include "logicops.pub"
#include "strel.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: morph7.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*-----------------------------------------------------------------------*
 *                              Image Cleaning                           *
 *-----------------------------------------------------------------------*/
/*
 *  removePepper
 *             Input:  StrelInfo state structure ptr;
 *                     source and dest Pixr, which can be the same
 *                     level  (integer:  1, 2, 3, or 4)
 *             Returns: 0 if OK; 1 on error.
 *
 *             Action:  if level = 1, removes 1x1 pepper;
 *                      if level = 2, removes up to 2x2 pepper;
 *                      if level = 3, removes up to 3x3 pepper;
 *                      if level = 4, removes up to 4x4 pepper;
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Remove small, isolated black regions in an image.
 *
 * LongDescription:::
 *  This is a simple application of morphology to identify and remove
 *  small, isolated black spots in an image.  It can remove single dots,
 *  dots contained in a 2x2, 3x3 or 4x4 square depending on the level
 *  argument.  It does not attempt to be clever about the context of the
 *  spots (like avoiding dots on i's).
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  prS::            Pointer to the image to be cleaned
 *  level::          size of spot to be removed: 1 => remove 1x1 spots,
 *		     2 => remove spots that fit inside a 2x2 square,
 *		     3 => remove spots that fit inside a 3x3 square,
 *		     4 => remove spots that fit inside a 4x4 square.
 *		     Other values are illegal.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  This is a pretty primitive function.  Clients probably want to use
 *  w_cleanPepper3 which is much fancier.  In fact, w_cleanPepper3
 *  uses this routine as a primitive.
 *
 * FunctionCategory:::
 *  document drycleaning
 *
 * Keywords:::
 *  document drycleaning, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Bloomberg
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removePepper (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               level,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*prTemp1 = NULL, *prTemp2 = NULL;
    STREL	*strel = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN(level < 1 || level > 4, cErrInvalidParm);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prTemp1));
    ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prTemp2));

    if (prD != prS)
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prS, COPY, prD));

    if (level >= 1) {
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, s_pepper_1, &strel));
	ERR_PASS_CLEANUP( w_erode (pIPWState, prS, strel, &prTemp1));
	strel = NULL;
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prTemp1, XOR, prD));
    }

    if (level >= 2) {
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, s_pepper_2, &strel));
	ERR_PASS_CLEANUP( w_erode (pIPWState, prD, strel, &prTemp1));
	strel = NULL;
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, strel2, &strel));
	ERR_PASS_CLEANUP( w_dilate (pIPWState, prTemp1, strel, &prTemp2));
	strel = NULL;
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prTemp2, SUBTRACT, prD));
    }

    if (level >= 3) {
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, s_pepper_3, &strel));
	ERR_PASS_CLEANUP( w_erode (pIPWState, prD, strel, &prTemp1));
	strel = NULL;
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, strel3, &strel));
	ERR_PASS_CLEANUP( w_dilate (pIPWState, prTemp1, strel, &prTemp2));
	strel = NULL;
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prTemp2, SUBTRACT, prD));
    }

    if (level == 4) {
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, s_pepper_4, &strel));
	ERR_PASS_CLEANUP( w_erode (pIPWState, prD, strel, &prTemp1));
	strel = NULL;
	ERR_PASS_CLEANUP( w_getStrel(pIPWState, strel4, &strel));
	ERR_PASS_CLEANUP( w_dilate (pIPWState, prTemp1, strel, &prTemp2));
	strel = NULL;
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prTemp2, SUBTRACT, prD));
    }

    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prTemp1));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prTemp2));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    w_destroyPixr(pIPWState, &prTemp1);
    w_destroyPixr(pIPWState, &prTemp2);
    ERR_RETURN;
}

/*

 $Log: morph7.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.25  1997/04/15 22:14:49  ddavies
 morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.24  1996/01/04 20:24:57  ejaquith
 morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.23  1995/12/13  04:10:35  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:13:52  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:03:11  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/10/05  04:27:18  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.19  1995/10/03  01:20:15  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.18  1995/09/21  22:54:43  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/15  18:00:36  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.16  1995/09/11  22:11:18  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.15  1995/09/10  00:33:20  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.14  1995/09/07  01:33:28  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:21:32  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/05  21:26:43  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Killed usage of GET_STREL() macro and replaced with getStrel() routine
 * invocations.
 *
 * Revision 1.11  1995/09/01  01:17:16  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:15:19  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/25  01:37:28  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Fixed warnings
 *
 * Revision 1.8  1995/08/17  22:02:43  ejaquith
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1995/07/29  04:44:19  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.6  1995/01/24  22:04:50  lperry
 * lperry on Tue Jan 24 14:04:33 PST 1995
 *
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * Changes interface to removePepper()
 *
 * Revision 1.5  1994/10/11  18:46:47  macleod
 * macleod on Tue Oct 11 11:46:07 PDT 1994
 *
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Changed removePepper to only copy from src to dest initially if they
 * are not the same PIXR.
 *
 * Revision 1.4  1994/10/07  19:14:20  lperry
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.3  1993/12/03  01:17:08  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:14:18  ddavies
 * morph7.c -> /product/ipcore/ipwrap/src/RCS/morph7.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

