/*<HTML><HEAD><TITLE>morphg1.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  morphg1.c
 *
 *       Drivers for morphological operations with arbitrary SEs,
 *       for grayscale images.
*/

/*
 *     Grayscale dilation:
 *            w_dilateGray()
 *
 *     Grayscale erosion:
 *            w_erodeGray()
 *
 *     Grayscale opening and closing:
 *            w_openGray()
 *            w_closeGray()
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "morphg.pub"
#include "morphg.prv"
#include "strel.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: morphg1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*-----------------------------------------------------------------------*
 *                  Grayscale Dilation and Erosion                       *
 *-----------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation::: "w_dilateGray"
 *
 * OneLineDescription:::
 *  Performs a grayscale dilation using a given structure element.
 *
 * LongDescription:::
 *  Binary morphology is useful for finding patterns and filtering binary
 *  images.  Grayscale morphology is similarly used in gray images.  The
 *  general idea is that one finds the maximum (max => white, so this is
 *  erosion) or minimum (min => black, so this is dilation) value in the
 *  set of pixels described by the structure element around every point
 *  in the image.  This value is written into the output image.
 *
 * PrivateDescription:::
 *  This is a fairly slow operation since it has to keep track of a bunch
 *  of gray values.  We normally only use purely horizontal or purely
 *  vertical strels to keep things sort of simple.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  prS::            Pointer to the source image to be dilated.
 *  strel::          Pointer to a structure element.  A structure element
 *		     can be used for either binary or grayscale morphology
 *		     since it only defines a pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  The source and dest images
 *		     may be the same, though this causes an temporary image
 *		     to be allocated and freed.
 *
 *
 * Notes:::
 *  A structure element, like a convolution kernel, is a small 2D array of
 *  locations that are aligned with pixels in an input image.  Each position
 *  of the array may contain a HIT, MISS or DONT_CARE.  Grayscale morphology
 *  ignores the MISSes.  When the center of the structure element
 *  is aligned with a pixel in the input image, structure element points
 *  corresponding to HITs are examined.  If the pixel value under the center
 *  pixel is less than the pixel at at HIT location, the center pixel's value
 *  is written to the HIT's location.  Structure element
 *  points corresponding to both MISSes and DONT_CAREs are ignored.
 *<BR><BR>
 *  Structure elements are often only 1D, i.e. purely horizontal or
 *  purely vertical.  People write theses on how large, concave, structure
 *  elements can be decomposed into multiple small ones with a general
 *  performance increase.  The only decomposition that is generally used
 *  here is that dilating with a solid n x m structure element (rectangular
 *  and all hits) will produce the same result as dilating with an n x 1
 *  and dilating the result with a 1 x m.  For larger strels, this is often
 *  faster since the number of HITs processed is so much smaller (n + m vs.
 *  n * m).
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation::: <A NAME = "w_dilateGray"></A>
 *****************************************************************************/
ErrCode CDECL
w_dilateGray (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL, *prSrcChannel = NULL, *prDestChannel = NULL;
    Int32	 strelType, channels, i;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prS, &channels));

	/* If we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.  Eventually, we should compute the size
	 * of the SE and add a sufficiently large frame to the source image
	 * to handle this.  For now, we assume there is a 32-pixel
	 * frame */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    /* Allow multi-channel images */
    for (i = 0; i < channels; i++)
    {
	/* Now that we have legal inputs, do the real work. */
	prSrcChannel = NULL; prDestChannel = NULL;
	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, prS, i, &prSrcChannel));
	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, prD, i, &prDestChannel));

	/* do special cases if possible */
	ERR_PASS_CLEANUP( w_strelType(pIPWState, strel, &strelType));
	if (strelType == H_1D_STREL || strelType == V_1D_STREL)
	{
	    ERR_PASS_CLEANUP( wp_dilateGray1D (pIPWState, prSrcChannel, 
					       strel, prDestChannel));
	}
	else if (strelType == S_2D_STREL) 
	{
	    ERR_PASS_CLEANUP( wp_dilateGraySep (pIPWState, prSrcChannel, 
						strel, prDestChannel));
	}
	else
	{
	    ERR_PASS_CLEANUP( wp_dilateGrayHisto (pIPWState, prSrcChannel, 
						  strel, prDestChannel));
	    
	    /* This call is commented out, so maybe this routine doesn't exist? */
	    /* ERR_PASS_CLEANUP( wp_dilateGrayBF(pIPWState, prDestChannel, prSrcChannel, strel); */
	}
    }	

    /* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));
	
    /* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::: "w_erodeGray"
 *
 * OneLineDescription:::
 *  Performs a grayscale eorsion using a given structure element.
 *
 * LongDescription:::
 *  Binary morphology is useful for finding patterns and filtering binary
 *  images.  Grayscale morphology is similarly used in gray images.  The
 *  general idea is that one finds the maximum (max => white, so this is
 *  erosion) or minimum (min => black, so this is dilation) value in the
 *  set of pixels described by the structure element around every point
 *  in the image.  This value is written into the output image.
 *
 * PrivateDescription:::
 *  This is a fairly slow operation since it has to keep track of a bunch
 *  of gray values.  We normally only use purely horizontal or purely
 *  vertical strels to keep things sort of simple.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  prS::            Pointer to the source image to be eroded.
 *  strel::          Pointer to a structure element.  A structure element
 *		     can be used for either binary or grayscale morphology
 *		     since it only defines a pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  The source and dest images
 *		     may be the same, though this causes an temporary image
 *		     to be allocated and freed.
 *
 * Notes:::
 *  A structure element, like a convolution kernel, is a small 2D array of
 *  locations that are aligned with pixels in an input image.  Each position
 *  of the array may contain a HIT, MISS or DONT_CARE.  Grayscale morphology
 *  ignores the MISSes.  When the center of the structure element
 *  is aligned with a pixel in the input image, structure element points
 *  corresponding to HITs are examined.  The center pixel is set to the
 *  maximum (whitest) pixel value of found in that set of pixels.
 *<BR><BR>
 *  Structure elements are often only 1D, i.e. purely horizontal or
 *  purely vertical.  People write theses on how large, concave, structure
 *  elements can be decomposed into multiple small ones with a general
 *  performance increase.  The only decomposition that is generally used
 *  here is that dilating with a solid n x m structure element (rectangular
 *  and all hits) will produce the same result as dilating with an n x 1
 *  and dilating the result with a 1 x m.  For larger strels, this is often
 *  faster since the number of HITs processed is so much smaller (n + m vs.
 *  n * m).
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation::: <A NAME = "w_erodeGray"></A>
 *****************************************************************************/
ErrCode CDECL
w_erodeGray (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL, *prSrcChannel = NULL, *prDestChannel = NULL;
    Int32	 strelType, channels, i;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prS, &channels));

	/* If we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.  Eventually, we should compute the size
	 * of the SE and add a sufficiently large frame to the source image
	 * to handle this.  For now, we assume there is a 32-pixel
	 * frame */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    /* Allow multi-channel images */
    for (i = 0; i < channels; i++)
    {
	/* Now that we have legal inputs, do the real work. */
	prSrcChannel = NULL; prDestChannel = NULL;
	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, prS, i, &prSrcChannel));
	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, prD, i, &prDestChannel));

	/* Now that we have legal inputs, do the real work. */

	/* do special cases if possible */
	ERR_PASS( w_strelType(pIPWState, strel, &strelType));
	if (strelType == H_1D_STREL || strelType == V_1D_STREL)
	{
	    ERR_PASS( wp_erodeGray1D (pIPWState, prSrcChannel, strel, 
				      prDestChannel));
	}
	else if (strelType == S_2D_STREL) 
	{
	    ERR_PASS( wp_erodeGraySep (pIPWState, prSrcChannel, strel, 
				       prDestChannel));
	}
	else
	{
	    ERR_PASS_CLEANUP( wp_erodeGrayHisto (pIPWState, prSrcChannel, 
						 strel, prDestChannel));
	/* ERR_PASS_CLEANUP( wp_erodeGrayBF(pIPWState, *pPrD, prSrcChannel, strel); */
	}
	
    }
	
    /* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*-----------------------------------------------------------------------*
 *                   Grayscale Opening and Closing                       *
 *-----------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation::: "w_openGray"
 *
 * OneLineDescription:::
 *  Performs a grayscale open using a given structure element.
 *
 * LongDescription:::
 *  As in binary morphology, one can do an open operation in grayscale
 *  morphology.  As before, it is an erode followed by a dilate, though
 *  the operators are gray in this case.  Openings tend to preserve
 *  stuff we want and destroy stuff we don't want.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  prS::            Pointer to the source image to be opened.
 *  strel::          Pointer to a structure element.  A structure element
 *		     can be used for either binary or grayscale morphology
 *		     since it only defines a pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  This image may be the same as the
 *		     source image.
 *
 * Notes:::
 *  This operation is implemented by calling the w_erodeGray followed by
 *  w_dilateGray.  A temporary image is allocated internally, which may
 *  take up quite a bit of space since it's gray.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation::: <A NAME = "w_openGray"></A>
 *****************************************************************************/
ErrCode CDECL
w_openGray (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL, *prT = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);

	/* If we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.  Eventually, we should compute the size
	 * of the SE and add a sufficiently large frame to the source image
	 * to handle this.  For now, we assume there is a 32-pixel
	 * frame */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* Now that we have legal inputs, do the real work.  Allocate
	 * a temporary pixr to hold the eroded image.
	 */
    ERR_PASS_CLEANUP( w_clonePixrNoInit(pIPWState, prS, &prT));

    ERR_PASS_CLEANUP( w_erodeGray (pIPWState, prS, strel, &prT));
    ERR_PASS_CLEANUP( w_dilateGray (pIPWState, prT, strel, &prD));

    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::: "w_closeGray"
 *
 * OneLineDescription:::
 *  Performs a grayscale close using a given structure element.
 *
 * LongDescription:::
 *  As in binary morphology, one can do an close operation in grayscale
 *  morphology.  As before, it is a dilate followed by an erode, though
 *  the operators are gray in this case.  Close operations tend to consolidate
 *  objects into larger objects.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  prS::            Pointer to the source image to be opened.
 *  strel::          Pointer to a structure element.  A structure element
 *		     can be used for either binary or grayscale morphology
 *		     since it only defines a pattern.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.  This image may be the same as the
 *		     source image.
 *
 * Notes:::
 *  This operation is implemented by calling the w_dilateGray followed by
 *  w_erodeGray.  A temporary image is allocated internally, which may
 *  take up quite a bit of space since it's gray.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation::: <A NAME = "w_closeGray"></A>
 *****************************************************************************/
ErrCode CDECL
w_closeGray (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL, *prT = NULL;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || strel == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);

	/* If we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.  Eventually, we should compute the size
	 * of the SE and add a sufficiently large frame to the source image
	 * to handle this.  For now, we assume there is a 32-pixel
	 * frame */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* Now that we have legal inputs, do the real work.  Allocate
	 * a temporary pixr to hold the eroded image.
	 */
    ERR_PASS_CLEANUP( w_clonePixrNoInit(pIPWState, prS, &prT));

    ERR_PASS_CLEANUP( w_dilateGray (pIPWState, prS, strel, &prT));
    ERR_PASS_CLEANUP( w_erodeGray (pIPWState, prT, strel, &prD));

    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*

$Log: morphg1.c,v $
Revision 1.1.1.1  1999/09/20 18:19:19  jlin
Imported sources

Revision 1.19  1997/04/15 22:16:50  ddavies
ddavies on Tue Apr 15 15:16:14 PDT 1997

morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v

Use non-initialized PIXR creation where possible.

Revision 1.18  1996/01/08 16:10:50  macleod
macleod on Mon Jan 8 08:10:32 PST 1996

morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v

made w_erodeGray and w_dilateGray work on multi-channel images

 * Revision 1.17  1995/12/13  04:10:38  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:13:55  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/10/13  22:03:14  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/10/05  04:27:22  ddavies
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.13  1995/09/21  22:54:46  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/18  23:46:33  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * Bug fixes.
 * Logic in w_erodeGray() and w_dilateGray() was flawed and not returning
 * at the proper time before setting the output pointers and cleaning up.
 *
 * Revision 1.11  1995/09/15  18:00:41  ddavies
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.10  1995/09/11  22:11:22  ddavies
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.9  1995/09/10  00:33:23  ddavies
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.8  1995/09/07  01:33:31  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.7  1995/09/06  20:21:35  ejaquith
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.6  1995/09/01  01:17:21  ddavies
 * morphg1.c -> /product/ipcore/ipwrap/src/RCS/morphg1.c,v
 *
 * Put minimum number of include files back in.
 *

 */

/*</PRE></BODY>*/
