/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  morphg2.c
 *
**
** Description: Private routines for morphological operations with
**		arbitrary structuring elements, for graylevel images
**              Histogram method
**
** Author: Luc Vincent
*/

/*
 *     Grayscale dilation:
 *            wp_dilateGrayHisto()
 *            wp_dilateGrayHistoLV()
 *            wp_dilateGrayHisto_low()
 *
 *     Grayscale erosion:
 *            wp_erodeGrayHisto()
 *            wp_erodeGrayHistoLV()
 *            wp_erodeGrayHisto_low()
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"
#include "logicops.h"

/* PROTOTYPES */
#include "props.pub"
#include "morphg.pub"
#include "morphg.prv"
#include "strel.pub"
#include "logicops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: morphg2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");



/*-----------------------------------------------------------------------*
 *                           Grayscale Dilation                          *
 *-----------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Performs grayscale dilation the hard (slow) way.
 *
 * LongDescription:::
 *  There are various special cases in grayscale morphology, like use
 *  of 1D strels.  This routine sorts out the type of dilation required.
 *  If none of the special cases work, it creates a special grayscale
 *  structure element from a regular structure element and asks
 *  wp_dilateGrayHistoLV to do the general case of gray dilation.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be dilated.
 *  strel::          Pointer to standard structure element.  This type
 *		     can be used for either binary or grayscale
 *		     morphology.
 *
 * Output:::
 *  prD::            Pointer to output image.  The source and dest images
 *		     may be the same, though this causes an temporary image
 *		     to be allocated and freed.
 *
 * Notes:::
 *  Grayscale morphology is slow.  Doing the general case is slower.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGrayHisto (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32         type;
    SE           *se = NULL;

	/* do special cases if possible */
    ERR_PASS( w_strelType(pIPWState, strel, &type));
    if (type == H_1D_STREL || type == V_1D_STREL)
    {
	ERR_PASS( wp_dilateGray1D (pIPWState, prS, strel, prD));
    }
    else if (type == S_2D_STREL)
    {
	ERR_PASS( wp_dilateGraySep (pIPWState, prS, strel, prD));
    }
    else
    {   /* general case */
	ERR_PASS( w_strelToSE(pIPWState, strel, &se));

	ERR_PASS_CLEANUP( wp_dilateGrayHistoLV (pIPWState, prS, se, prD));

	ERR_PASS_CLEANUP( w_destroySE(pIPWState, &se));
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroySE(pIPWState, &se);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform general purpose grayscale dilation.
 *
 * LongDescription:::
 *  This is the routine used when none of the special cases apply.
 *  It allocates a histogram used to keep track of pixel values under
 *  the structure element and asks wp_dilateGrayHistoLow to do the
 *  real work.  If the source and dest images are the same, a temporary
 *  image is allocated and freed here.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be dilated.
 *  se::             Pointer to special grayscale structure element.
 *
 * Output:::
 *  prD::            Pointer to output image.  The source and dest images
 *		     may be the same, though this causes an temporary image
 *		     to be allocated and freed.
 *
 * Notes:::
 *  Grayscale morphology is slow.  Doing the general case is slower.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGrayHistoLV (
    IPWState           *pIPWState,
    PIXR               *prS,
    SE                 *se,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, sFrame;
    Int32        i, j;
    Int32        sBpl, tBpl;
    Int32       *pts = NULL, *sizes = NULL;
    Int32      **sides = NULL;
    UInt8       *datas, *datat;
    PIXR        *prT = NULL;

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &wD));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &hD));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &sFrame));

	/* if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prT));
    else
	prT = prD;

    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prS, &datas));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prT, &datat));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prT, &tBpl));

	/* do this for efficiency later */
    pts = (Int32 *) IPW_CALLOC(se->size, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP(pts == NULL, cErrNomem);

    for (i = 0; i < se->size; i++)
	pts[i] = (se->pts + i)->x + sBpl * (se->pts + i)->y;

    sides = (Int32 **) IPW_CALLOC(4, sizeof(Int32 *), cAllocatorData);
    sizes = (Int32 *) IPW_CALLOC(4, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP(sides == NULL  ||  sizes == NULL, cErrNomem);

    sides[0] = sides[1] = sides[2] = sides[3] = NULL;
    for (i = 0; i < 4; i++)
    {
	sides[i] = (Int32 *) IPW_CALLOC((se->sides + 2 * i)->size,
					sizeof(Int32), cAllocatorData);
	ERR_GEN_CLEANUP(sides[i] == NULL, cErrNomem);
	
	sizes[i] = (se->sides + 2 * i)->size;
	for (j = 0; j < sizes[i]; j++) {
	    sides[i][j] = ((se->sides + 2 * i)->pts + j)->x +
			sBpl * ((se->sides + 2 * i)->pts + j)->y;
	}
    }

    ERR_PASS_CLEANUP( wp_dilateGrayHistoLow (pIPWState, wS, hS, tBpl, 
					 datas, sBpl, sFrame,
					 se->size, pts, sizes, sides, datat));

	/* if in-place, copy again and destroy the temp image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_copy(pIPWState, prT, &prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));
    }

    IPW_FREE(pts, cAllocatorData);
    for (i = 0; i < 4; i++)
	IPW_FREE(sides[i], cAllocatorData);
    IPW_FREE(sides, cAllocatorData);
    IPW_FREE(sizes, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pts, cAllocatorData);
    if (sides != NULL)
    {
	for (i = 0; i < 4; i++)
	    IPW_FREE(sides[i], cAllocatorData);
    }
    IPW_FREE(sides, cAllocatorData);
    IPW_FREE(sizes, cAllocatorData);
    
    if (prD == prS)
	w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Lowest level implementation of general grayscale dilation.
 *
 * LongDescription:::
 *  This is the routine used when none of the special cases apply.
 *  The real dilation work is done here after any images and data
 *  structures have been allocated above.  If something were to move
 *  to ipshared, this would be it.
 *
 * PrivateDescription:::
 *  This routine was ported from Luc Vincent's code at XIS.  It follows
 *  Luc's rules for code format.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  w::              Width of image to be dilated (not including frame)
 *  h::              Height of image to be dilated (not including frame)
 *  dBpl::           Dest image's bytes/line.
 *  pS::             Pointer to word in upper left corner of source image
 *		     inside frame.
 *  sBpl::           Source image's bytes/line.
 *  sFrame::         Number of pixels in source image's frame.
 *  size::           Total number of pixels in structure element
 *  pts::            Array of address offsets from center source locations
 *		     that are under HITs of the structure element.
 *  sizes::          Gosh, golly.  Ask Luc Vincent.
 *  sides::          Gosh, golly.  Ask Luc Vincent.
 *
 * Output:::
 *  pD::             Pointer to word in upper left corner of the dest image
 *		     inside frame.
 *
 * Notes:::
 *  Ask Luc Vincent how this works.  It is very clever.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGrayHistoLow (
    IPWState           *pIPWState,
    Int32               w,
    Int32               h,
    Int32               dBpl,
    UInt8              *pS,
    Int32               sBpl,
    Int32               sFrame,
    Int32               size,
    Int32              *pts,
    Int32              *sizes,
    Int32             **sides,
    UInt8              *pD)
{
    PROLOGUE;
    Int32         fullw, fullh;
    UInt8        *fullimages;
    Int32        *ptsp, *pts_end;
    UInt8        *pixs, *pixd;
    UInt8         smallest_val, largest_val;
    UInt8         val, running_max;
    UInt8        *line, *pix;
    Int32         i, j, x, y;
    UInt16       *histo_array, *histo;

    histo = histo_array = NULL;

        /* Finds largest and smallest values in image */
    largest_val  = SMLVAL;
    smallest_val = BIGVAL8;
    line = pS;
    for (i = h; i > 0; i--)
    {
	for (pix = line, j = w; j > 0; j--, pix++)
	{
	    if (*pix > largest_val)
		largest_val = *pix;
	    if (*pix < smallest_val)
		smallest_val = *pix;
	}
	line += sBpl;
    }

	/* set the frame to the smallest value in the image;
	 * this saves time because the histogram can have a
	 * smaller number of bins */
    fullw = w + 2 * sFrame;
    fullh = h + 2 * sFrame;
    fullimages = pS - sFrame * sBpl - sFrame;

    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)fullimages, fullw, fullh,
				sBpl, sFrame, smallest_val));

        /* Allocate local histogram with value 0 using calloc() */
    histo_array = (UInt16 *)IPW_CALLOC((Int32)(largest_val-smallest_val+1),
				       sizeof(UInt16), cAllocatorData);
    ERR_GEN(histo_array == NULL, cErrNomem);

    histo = histo_array - smallest_val;

        /* Does the placement of structuring element one pixel
	   above the first image pixel */
    running_max = smallest_val;
    pixs = pS - sBpl;
    for (i = 0; i < size; i++) {
	val = *(pixs + pts[i]);
	histo[val]++;
	if (val > running_max)
	    running_max = val;
    }


        /* Starts main loop */
    y   = -1;
    x   = 0;
    pixd = pD - dBpl;   /* point above the first pixel */
    while (1) {
	if (y >= h - 1)
	    break;

	    /* Move the element one step down (direction 6)
	       And writes new running_max value at *pix */
	ptsp    = sides[1];
	pts_end = ptsp + sizes[1];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]--;
	    if (histo[val] == 0 && val == running_max) {
		val--;
		while (val > smallest_val && histo[val] == 0)
		    val--;
		running_max = val;
	    }
	}
	y++;
	pixs += sBpl;
	pixd += dBpl;
	ptsp    = sides[3];
	pts_end = ptsp + sizes[3];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]++;
	    if (val > running_max)
		running_max = val;
	}
	*pixd = running_max;

	    /* Scan line y from left to right */
	while (x < w - 1) {
	    ptsp    = sides[2];
	    pts_end = ptsp + sizes[2];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]--;
		if (histo[val] == 0 && val == running_max) {
		    val--;
		    while (val > smallest_val && histo[val] == 0)
			val--;
		    running_max = val;
		}
	    }
	    x++;
	    pixs++;
	    pixd++;
	    ptsp    = sides[0];
	    pts_end = ptsp + sizes[0];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]++;
		if (val > running_max)
		    running_max = val;
	    }
	    *pixd = running_max;
	}

	if (y >= h - 1)
	    break;

	    /* Move the element one step down (direction 6)
	       And writes new running_max value at *pixd */
	ptsp    = sides[1];
	pts_end = ptsp + sizes[1];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]--;
	    if (histo[val] == 0 && val == running_max) {
		val--;
		while (val > smallest_val && histo[val] == 0)
		    val--;
		running_max = val;
	    }
	}
	y++;
	pixs += sBpl;
	pixd += dBpl;
	ptsp    = sides[3];
	pts_end = ptsp + sizes[3];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]++;
	    if (val > running_max)
		running_max = val;
	}
	*pixd = running_max;

	    /* Scan line y from left to right */
	while (x > 0) {
	    ptsp    = sides[0];
	    pts_end = ptsp + sizes[0];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]--;
		if (histo[val] == 0 && val == running_max) {
		    val--;
		    while (val > smallest_val && histo[val] == 0)
			val--;
		    running_max = val;
		}
	    }
	    x--;
	    pixs--;
	    pixd--;
	    ptsp    = sides[2];
	    pts_end = ptsp + sizes[2];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]++;
		if (val > running_max)
		    running_max = val;
	    }
	    *pixd = running_max;
	}
    }

    IPW_FREE(histo_array, cAllocatorData);

	/* reset frame pixels to 0 */
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)fullimages,
				   fullw, fullh, sBpl, sFrame, 0));

    NORMAL_RETURN;

}  /* grayscale_dilation */




/*-----------------------------------------------------------------------*
 *                           Grayscale Erosion                           *
 *-----------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Performs grayscale erosion the hard (slow) way.
 *
 * LongDescription:::
 *  There are various special cases in grayscale morphology, like use
 *  of 1D strels.  This routine sorts out the type of erosion required.
 *  If none of the special cases work, it creates a special grayscale
 *  structure element from a regular structure element and asks
 *  wp_erodeGrayHistoLV to do the general case of gray erosion.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be eroded.
 *  strel::          Pointer to standard structure element.  This type
 *		     can be used for either binary or grayscale
 *		     morphology.
 *
 * Output:::
 *  prD::            Pointer to output image.  The source and dest images
 *		     may be the same, though this causes an temporary image
 *		     to be allocated and freed.
 *
 * Notes:::
 *  Grayscale morphology is slow.  Doing the general case is slower.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGrayHisto (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32         type;
    SE           *se = NULL;

	/* do special cases if possible */
    ERR_PASS( w_strelType(pIPWState, strel, &type));
    if (type == H_1D_STREL || type == V_1D_STREL)
    {
	ERR_PASS( wp_erodeGray1D (pIPWState, prS, strel, prD));
    }
    else if (type == S_2D_STREL)
    {
	ERR_PASS( wp_erodeGraySep (pIPWState, prS, strel, prD));
    }
    else
    {   /* general case */
	ERR_PASS( w_strelToSE(pIPWState, strel, &se));

	ERR_PASS_CLEANUP( wp_erodeGrayHistoLV (pIPWState, prS, se, prD));

	ERR_PASS_CLEANUP( w_destroySE(pIPWState, &se));
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroySE(pIPWState, &se);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform general purpose grayscale erosion.
 *
 * LongDescription:::
 *  This is the routine used when none of the special cases apply.
 *  It allocates a histogram used to keep track of pixel values under
 *  the structure element and asks wp_erodeGrayHistoLow to do the
 *  real work.  If the source and dest images are the same, a temporary
 *  image is allocated and freed here.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be eroded.
 *  se::             Pointer to special grayscale structure element.
 *
 * Output:::
 *  prD::            Pointer to output image.  The source and dest images
 *		     may be the same, though this causes an temporary image
 *		     to be allocated and freed.
 *
 * Notes:::
 *  Grayscale morphology is slow.  Doing the general case is slower.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGrayHistoLV (
    IPWState           *pIPWState,
    PIXR               *prS,
    SE                 *se,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, sFrame;
    Int32        i, j;
    Int32        sBpl, tBpl;
    Int32       *pts = NULL, *sizes = NULL;
    Int32      **sides = NULL;
    UInt8       *datas, *datat;
    PIXR        *prT = NULL;

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &wD));
    ERR_PASS( w_pixrGetHeight(pIPWState, prD, &hD));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &sFrame));

	/* if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prT));
    else
	prT = prD;

    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prS, &datas));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prT, &datat));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prT, &tBpl));

	/* do this for efficiency later */
    pts = (Int32 *) IPW_CALLOC(se->size, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP(pts == NULL, cErrNomem);

    for (i = 0; i < se->size; i++)
	pts[i] = (se->pts + i)->x + sBpl * (se->pts + i)->y;

    sides = (Int32 **) IPW_CALLOC(4, sizeof(Int32 *), cAllocatorData);
    sizes = (Int32 *) IPW_CALLOC(4, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP(sides == NULL  ||  sizes == NULL, cErrNomem);

    sides[0] = sides[1] = sides[2] = sides[3] = NULL;
    for (i = 0; i < 4; i++)
    {
	sides[i] = (Int32 *) IPW_CALLOC((se->sides + 2 * i)->size,
					sizeof(Int32), cAllocatorData);
	ERR_GEN_CLEANUP(sides[i] == NULL, cErrNomem);
	
	sizes[i] = (se->sides + 2 * i)->size;
	for (j = 0; j < sizes[i]; j++) {
	    sides[i][j] = ((se->sides + 2 * i)->pts + j)->x +
			sBpl * ((se->sides + 2 * i)->pts + j)->y;
	}
    }

    ERR_PASS_CLEANUP( wp_erodeGrayHistoLow (pIPWState, wS, hS, tBpl, 
					datas, sBpl, sFrame,
					se->size, pts, sizes, sides, datat));

	/* if in-place, copy again and destroy the temp image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_copy(pIPWState, prT, &prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));
    }

    IPW_FREE(pts, cAllocatorData);
    for (i = 0; i < 4; i++)
	IPW_FREE(sides[i], cAllocatorData);
    IPW_FREE(sides, cAllocatorData);
    IPW_FREE(sizes, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pts, cAllocatorData);
    if (sides != NULL)
    {
	for (i = 0; i < 4; i++)
	    IPW_FREE(sides[i], cAllocatorData);
    }
    IPW_FREE(sides, cAllocatorData);
    IPW_FREE(sizes, cAllocatorData);
    
    if (prD == prS)
	w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Lowest level implementation of general grayscale erosion.
 *
 * LongDescription:::
 *  This is the routine used when none of the special cases apply.
 *  The real erosion work is done here after any images and data
 *  structures have been allocated above.  If something were to move
 *  to ipshared, this would be it.
 *
 * PrivateDescription:::
 *  This routine was ported from Luc Vincent's code at XIS.  It follows
 *  Luc's rules for code format.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  w::              Width of image to be eroded (not including frame)
 *  h::              Height of image to be eroded (not including frame)
 *  dBpl::           Dest image's bytes/line.
 *  pS::             Pointer to word in upper left corner of source image
 *		     inside frame.
 *  sBpl::           Source image's bytes/line.
 *  sFrame::         Number of pixels in source image's frame.
 *  size::           Total number of pixels in structure element
 *  pts::            Array of address offsets from center source locations
 *		     that are under HITs of the structure element.
 *  sizes::          Gosh, golly.  Ask Luc Vincent.
 *  sides::          Gosh, golly.  Ask Luc Vincent.
 *
 * Output:::
 *  pD::             Pointer to word in upper left corner of the dest image
 *		     inside frame.
 *
 * Notes:::
 *  Ask Luc Vincent how this works.  It is very clever.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGrayHistoLow (
    IPWState           *pIPWState,
    Int32               w,
    Int32               h,
    Int32               dBpl,
    UInt8              *pS,
    Int32               sBpl,
    Int32               sFrame,
    Int32               size,
    Int32              *pts,
    Int32              *sizes,
    Int32             **sides,
    UInt8              *pD)
{
    PROLOGUE;
    Int32         fullw, fullh;
    UInt8        *fullimages;
    Int32        *ptsp, *pts_end;
    UInt8        *pixs, *pixd;
    UInt8         smallest_val, largest_val;
    UInt8         val, running_min;
    UInt8        *line, *pix;
    Int32         i, j, x, y;
    UInt16       *histo_array, *histo;


    histo = histo_array = NULL;

        /* Finds largest and smallest values in image */
    largest_val  = SMLVAL;
    smallest_val = BIGVAL8;
    line = pS;
    for (i = h; i > 0; i--)
    {
	for (pix = line, j = w; j > 0; j--, pix++)
	{
	    if (*pix > largest_val)
		largest_val = *pix;
	    if (*pix < smallest_val)
		smallest_val = *pix;
	}
	line += sBpl;
    }

	/* set the frame to the largest value in the image;
	 * this saves time because the histogram can have a
	 * smaller number of bins */
    fullw = w + 2 * sFrame;
    fullh = h + 2 * sFrame;
    fullimages = pS - sFrame * sBpl - sFrame;

    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)fullimages, fullw, fullh,
				sBpl, sFrame, largest_val));

        /* Allocate local histogram with value 0 using calloc() */
    histo_array = (UInt16 *)IPW_CALLOC((Int32)(largest_val-smallest_val+1),
				       sizeof(UInt16), cAllocatorData);
    ERR_GEN(histo_array == NULL, cErrNomem);

    histo = histo_array - smallest_val;

        /* Does the placement of structuring element one pixel
	   above the first image pixel */
    running_min = largest_val;
    pixs = pS - sBpl;
    for (i = 0; i < size; i++) {
	val = *(pixs + pts[i]);
	histo[val]++;
	if (val < running_min)
	    running_min = val;
    }


        /* Starts main loop */
    y   = -1;
    x   = 0;
    pixd = pD - dBpl;   /* point above the first pixel */
    while (1) {
	if (y >= h - 1)
	    break;

	    /* Move the element one step down (direction 6)
	       And writes new running_min value at *pix */
	ptsp    = sides[1];
	pts_end = ptsp + sizes[1];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]--;
	    if (histo[val] == 0 && val == running_min) {
		val++;
		while (val < largest_val && histo[val] == 0)
		    val++;
		running_min = val;
	    }
	}
	y++;
	pixs += sBpl;
	pixd += dBpl;
	ptsp    = sides[3];
	pts_end = ptsp + sizes[3];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]++;
	    if (val < running_min)
		running_min = val;
	}
	*pixd = running_min;

	    /* Scan line y from left to right */
	while (x < w - 1) {
	    ptsp    = sides[2];
	    pts_end = ptsp + sizes[2];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]--;
		if (histo[val] == 0 && val == running_min) {
		    val++;
		    while (val < largest_val && histo[val] == 0)
			val++;
		    running_min = val;
		}
	    }
	    x++;
	    pixs++;
	    pixd++;
	    ptsp    = sides[0];
	    pts_end = ptsp + sizes[0];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]++;
		if (val < running_min)
		    running_min = val;
	    }
	    *pixd = running_min;
	}

	if (y >= h - 1)
	    break;

	    /* Move the element one step down (direction 6)
	       And writes new running_min value at *pixd */
	ptsp    = sides[1];
	pts_end = ptsp + sizes[1];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]--;
	    if (histo[val] == 0 && val == running_min) {
		val++;
		while (val < largest_val && histo[val] == 0)
		    val++;
		running_min = val;
	    }
	}
	y++;
	pixs += sBpl;
	pixd += dBpl;
	ptsp    = sides[3];
	pts_end = ptsp + sizes[3];
	while (ptsp < pts_end) {
	    val = *(pixs + *ptsp++);
	    histo[val]++;
	    if (val < running_min)
		running_min = val;
	}
	*pixd = running_min;

	    /* Scan line y from left to right */
	while (x > 0) {
	    ptsp    = sides[0];
	    pts_end = ptsp + sizes[0];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]--;
		if (histo[val] == 0 && val == running_min) {
		    val++;
		    while (val < largest_val && histo[val] == 0)
			val++;
		    running_min = val;
		}
	    }
	    x--;
	    pixs--;
	    pixd--;
	    ptsp    = sides[2];
	    pts_end = ptsp + sizes[2];
	    while (ptsp < pts_end) {
		val = *(pixs + *ptsp++);
		histo[val]++;
		if (val < running_min)
		    running_min = val;
	    }
	    *pixd = running_min;
	}
    }

    IPW_FREE(histo_array, cAllocatorData);

	/* reset frame pixels to 0 */
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)fullimages, fullw, fullh,
				sBpl, sFrame, 0));

    NORMAL_RETURN;

}  /* grayscale erosion */



/*

$Log: morphg2.c,v $
Revision 1.1.1.1  1999/09/20 18:19:19  jlin
Imported sources

Revision 1.22  1997/04/15 22:17:15  ddavies
morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v

Use non-initialized PIXR creation where possible.

Revision 1.21  1996/01/04 20:25:00  ejaquith
morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.20  1995/12/13  04:10:42  ejaquith
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.19  1995/11/13  20:13:59  ejaquith
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/11/03  22:48:35  ddavies
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.17  1995/10/25  00:05:18  jcraig
 * jcraig on Tue Oct 24 17:05:02 PDT 1995
 *
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * remove include of ipinc.h
 *
 * Revision 1.16  1995/10/13  22:03:17  ejaquith
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/10/05  04:27:25  ddavies
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.14  1995/09/21  22:54:49  ejaquith
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/15  18:00:45  ddavies
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.12  1995/09/11  22:11:25  ddavies
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.11  1995/09/10  00:33:27  ddavies
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.10  1995/09/07  01:33:34  ejaquith
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:21:38  ejaquith
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  01:17:27  ddavies
 * morphg2.c -> /product/ipcore/ipwrap/src/RCS/morphg2.c,v
 *
 * Put minimum number of include files back in.
 *

 */

