/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  morphg4.c
 *
 *          Grayscale morphological operations that can
 *          use special 1D horizontal and vertical structuring
 *          elements.  This includes separable SEs with all hits.
 */

/*
 *      1D gray dilation:
 *            wp_dilateGray1DPixr()
 *            wp_dilateGray1D()
 *            wp_dilateGrayHLow()
 *            wp_dilateGrayVLow()
 *
 *      1D gray erosion:
 *            wp_erodeGray1DPixr()
 *            wp_erodeGray1D()
 *            wp_erodeGrayHLow()
 *            wp_erodeGrayVLow()
 *
 *      2D separable gray dilation:
 *            wp_dilateGraySepPixr()
 *            wp_dilateGraySep()
 *
 *      2D separable gray erosion:
 *            wp_erodeGraySepPixr()
 *            wp_erodeGraySep()
 *
 */

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"

/* DEFINITIONS */
#include "imageref.pub"
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"
#include "logicops.h"

/* PROTOTYPES */
#include "props.pub"
#include "morphg.pub"
#include "morphg.prv"
#include "strel.pub"
#include "logicops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: morphg4.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/* Suppress all debugging */
#define  DEBUG_PR     0




/*-----------------------------------------------------------------------*
 *                   1 Dimensional Grayscale Dilation                    *
 *-----------------------------------------------------------------------*/

/*
 *  dilateGray1D()
 *
 *        Restrictions:  the input strel must be 1d, composed of solid
 *                       hits, and with the center position along the
 *                       1D line (though not necessarily within the
 *                       set of hits).
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Performs grayscale dilation for 1D structure elements (horiz or vert)
 *
 * LongDescription:::
 *  Grayscale morphology can be done much faster when the structure element
 *  is either purely horizontal or purely vertical.  This routine decides
 *  which of these is true, allocates a temporary image if the source
 *  and dest images are the same, and calls wp_dilateGrayHLow or
 *  wp_dilateGrayVLow do to the real work.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be dilated.
 *  strel::          Pointer to conventional structure element.  This
 *		     type of strel can be used for either binary or
 *		     grayscale morphology.
 *
 * Output:::
 *  prD::            Pointer to output image.  If this is the same as the
 *		     source image, a (possibly quite large) temporary
 *		     image will be allocated and freed here.
 *
 * Notes:::
 *  Grayscale morphology is usually slow.  This special case is faster.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGray1D (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        streltype;
    Int32        ws, hs;
    Int32        frames, framed;
    Int32        bpls, bplt;
    UInt8       *datas, *datat;
    PIXR        *prT = NULL;

    /* find the type of strel */
    ERR_PASS( w_strelType(pIPWState, strel, &streltype));
    
    /* strel must be 1D with solid hit */
    ERR_GEN(streltype != H_1D_STREL && streltype != V_1D_STREL,
	    cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &ws));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hs));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frames));
    ERR_PASS( w_pixrGetFrame(pIPWState, prD, &framed));

	/* if in-place, make a temporary image */
    if (prS == prD)
	ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prT));
    else
	prT = prD;

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &datas));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prT, &datat));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpls));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prT, &bplt));

    if (streltype == H_1D_STREL) /*  horizontal */
	ERR_PASS_CLEANUP( wp_dilateGrayHLow (pIPWState, ws, hs, bplt, framed,
					 datas, bpls, frames,
					 strel->nx, strel->cx, datat));
    else   /* vertical */
	ERR_PASS_CLEANUP( wp_dilateGrayVLow (pIPWState, ws, hs, bplt, framed,
					 datas, bpls, frames,
					 strel->ny, strel->cy, datat));

	/* if in-place, copy again and destroy the temp image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_copy(pIPWState, prT, &prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prD == prS)
	w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Low level routine used to dilate with a horizontal structure element.
 *
 * LongDescription:::
 *  Grayscale morphology can be done much faster when the structure element
 *  is either purely horizontal or purely vertical and filled with solid
 *  HITs.  This routine deals with the horizontal case.  This is a routine that
 *  could be put into ipshared.
 *
 * PrivateDescription:::
 *  This routine uses a clever method that mostly looks only at the
 *  pixels on the ends of the strel.  Only when the pixel having the
 *  minimum value falls off the end of the area of interest as we slide
 *  the strel horizontally do we look at the other pixels to find a new
 *  minimum.  We also look at the new pixels arriving to see if they
 *  are the new minimum.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  w::              Widht of unframed dest image.
 *  h::              Height of unframed dest image.
 *  bpld::           Bytes/line of dest image.
 *  framed::         Number of pixels in dest image's frame.
 *  images::         Pointer to OUTSIDE of frame of source image.
 *  bpls::           Bytes/line of source image.
 *  frames::         Number of pixels in source image's frame.
 *  nx::             Width of structure element.
 *  cx::             Offset of structure element's center position
 *		     relative to left end of structure element.
 *
 * Output:::
 *  imaged::         Pointer to OUTSIDE of frame of dest image.
 *
 * Notes:::
 *  Grayscale morphology is usually slow.  Doing the special case
 *  is often tolerable.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGrayHLow (
    IPWState           *pIPWState,
    register Int32      w,
    Int32               h,
    Int32               bpld,
    Int32               framed,
    UInt8              *images,
    Int32               bpls,
    Int32               frames,
    Int32               nx,
    Int32               cx,
    UInt8              *imaged)
{
    PROLOGUE;
    Int32             fullw, fullh, range;
    UInt8             smallestVal, largestVal;
    register UInt8    val, max;
    register UInt8   *pS, *pD;
    register Int32    j;
    Int32             i /*, x, y */ ;
     Int32             startx, endx;
    register UInt32  *histo;
    UInt32           *histoArray;

    histo = histoArray = NULL;

	/* find smallest and largest values in the image */
    largestVal = BIGVAL8;
    smallestVal = SMLVAL;

    range = (Int32)(largestVal - smallestVal + 1);

	/* Allocate local histogram */
    histoArray = (UInt32 *)IPW_CALLOC(range, sizeof(UInt32), cAllocatorData);
    ERR_GEN(histoArray == NULL, cErrNomem);
    
    histo = histoArray - smallestVal;


	/* Initialize lines to be at the left edge of the frame and
	 * on the first row of the proper image.  
	 * Initialize lined to be the first pixel of the proper image.  */
    pS = (images + frames * bpls);
    pD = (imaged + framed * bpld + framed);

	/* dilate horizontally */
    startx = frames - cx;
    endx = startx + nx - 1;
    for (i = 0; i < h; i++)
    {
	    /* initialize the histogram for this row */
	max = 0;
	MEMSET((void *)histoArray, 0, (range * sizeof(UInt32)));

	    /* get the first dest pixel in the row */
	for (j = startx; j <= endx; j++)
	{
	    val = UCHAR_ACCESS(pS + j);
	    histo[val]++;
	    if (val > max)
		max = val;
	}
	UCHAR_ACCESS (pD) = max;

	    /* do the rest of the row */
	for (j = 1; j < w; j++)
	{
		/* remove the pixel at the left end and update max */
	    val = UCHAR_ACCESS(pS + startx + j - 1);
	    histo[val]--;
	    if (histo[val] == 0 && val == max)
	    {
		++ histo[smallestVal];
		val--;
		while ( histo[val] == 0) val--;
		max = val;
		-- histo[smallestVal];
	    }

		/* add the pixel at the right end and update max */
	    val = UCHAR_ACCESS(pS + endx + j);
	    histo[val]++;
	    if (val > max)
		max = val;
	    UCHAR_ACCESS (pD + j) = max;

	}
	pS += bpls;
	pD += bpld;
    }

    IPW_FREE(histoArray, cAllocatorData);

    /* reset frame pixels to 0 */
    fullw = w + 2 * frames;
    fullh = h + 2 * frames;
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)images, fullw, fullh,
				bpls, frames, 0));


    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Low level routine used to dilate with a vertical structure element.
 *
 * LongDescription:::
 *  Grayscale morphology can be done much faster when the structure element
 *  is either purely horizontal or purely vertical and filled with solid
 *  HITs.  This routine deals with the vertical case.  This is a routine that
 *  could be put into ipshared.
 *
 * PrivateDescription:::
 *  This routine uses a clever method that mostly looks only at the
 *  pixels on the ends of the strel.  Only when the pixel having the
 *  minimum value falls off the end of the area of interest as we slide
 *  the strel horizontally do we look at the other pixels to find a new
 *  minimum.  We also look at the new pixels arriving to see if they
 *  are the new minimum.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  w::              Width of unframed dest image.
 *  h::              Height of unframed dest image.
 *  bpld::           Bytes/line of dest image.
 *  framed::         Number of pixels in dest image's frame.
 *  images::         Pointer to OUTSIDE of frame of source image.
 *  bpls::           Bytes/line of source image.
 *  frames::         Number of pixels in source image's frame.
 *  ny::             Height of structure element.
 *  cy::             Offset of structure element's center position
 *		     relative to top end of structure element.
 *
 * Output:::
 *  imaged::         Pointer to OUTSIDE of frame of dest image.
 *
 * Notes:::
 *  Grayscale morphology is usually slow.  Doing the special case
 *  is often tolerable.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGrayVLow (
    IPWState           *pIPWState,
    Int32               w,
    register Int32      h,
    Int32               bpld,
    Int32               framed,
    UInt8              *images,
    Int32               bpls,
    Int32               frames,
    Int32               ny,
    Int32               cy,
    UInt8              *imaged)
{
    PROLOGUE;
    Int32             fullw, fullh, range;
    UInt8             smallestVal, largestVal;
    register UInt8    val, max;
    register UInt8   *lines1, *lined;
    UInt8            *firstlines, *firstlined;
    register Int32    i;
    Int32             j,/* x, */ y ;
    Int32             starty;
    register UInt32  *histo;
    UInt32           *histoArray;

    histo = histoArray = NULL;

	/* find smallest and largest values in the image */
    largestVal = BIGVAL8;
    smallestVal = SMLVAL;

    range = (Int32)(largestVal - smallestVal + 1);			

	/* Allocate local histogram */
    histoArray = (UInt32 *)IPW_CALLOC(range, sizeof(UInt32), cAllocatorData);
    ERR_GEN(histoArray == NULL, cErrNomem);

    histo = histoArray - smallestVal;

	/* Initialize firstlines to be at the top pixel of the SE when
	 * the center of the SE is on the UL pixel of the proper image.
	 * Initialize firstlined to be the first pixel of the
	 * proper image.     */
    starty = frames - cy;
    firstlines = images + starty * bpls + frames;
    firstlined = imaged + framed * bpld + framed;
    y = ny * bpls;

	/* dilate vertically */
    for (j = 0; j < w; j++)
    {
	    /* initialize the histogram for this column */
	max = 0;
	MEMSET((void *)histoArray, 0, (range * sizeof(UInt32)));

	    /* get the first dest pixel in the column */
	lined = firstlined +j;
	for (i = 0, lines1 = firstlines +j; i < ny; ++i, lines1 += bpls)
	{
	    val = UCHAR_ACCESS(lines1);
	    histo[val]++;
	    if (val > max)
		max = val;
	}
	UCHAR_ACCESS(lined) = max;

	    /* do the rest of the column */
	lines1 = firstlines + j;
	lined += bpld;
	for (i = 1; i < h;  i++, lines1 += bpls, lined += bpld)
	{
		/* remove the pixel at the top and update max  */
	    val = UCHAR_ACCESS(lines1);
	    histo[val]--;
	    if (histo[val] == 0 && val == max)
	    {
		++ histo[smallestVal];
		val--;
		while ( histo[val] == 0) val--;
		max = val;
		-- histo[smallestVal];
	    }

		/* add the pixel at the bottom and update max */
	    val = UCHAR_ACCESS(lines1 + y);
	    histo[val]++;
	    if (val > max)
		max = val;
	    UCHAR_ACCESS(lined) = max;

	}
    }

    IPW_FREE(histoArray, cAllocatorData);

    /* reset frame pixels to 0 */
    fullw = w + 2 * frames;
    fullh = h + 2 * frames;
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)images, fullw, fullh,
				bpls, frames, 0));

    NORMAL_RETURN;
}


/*-----------------------------------------------------------------------*
 *              2 Dimensional Separable Grayscale Dilation               *
 *-----------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform grayscale dilation for separable structure elements.
 *
 * LongDescription:::
 *  A separable structure element is one that can be decomposed into
 *  a purely horizontal strel and a purely vertical strel.  One does
 *  a dilation using the horizontal strel followed by a dilation on
 *  the result with the vertical strel.  This should give the same
 *  result as a dilation with the original strel but uses the optimized
 *  grayscale operators.  However, it does require allocation of
 *  a (possibly quite large) gray temporary image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be dilated.
 *  strel::          Pointer to conventional structure element.  This
 *		     type of strel can be used for either binary or
 *		     grayscale morphology.
 *
 * Output:::
 *  prD::            Pointer to output image.  Since a temporary image
 *		     is always allocated here, there is no extra penalty for
 *		     for setting prS = prD.
 *
 * Notes:::
 *  Grayscale morphology is slow.  This is somewhat faster than the general
 *  case.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dilateGraySep (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        streltype;
    STREL       *strelh = NULL, *strelv = NULL;
    PIXR        *prT = NULL;

    /* find the type of strel */
    ERR_PASS( w_strelType(pIPWState, strel, &streltype));
    ERR_GEN(streltype != S_2D_STREL, cErrInvalidParm);

    ERR_PASS( w_strelSeparation(pIPWState, strel, &strelh, &strelv));

    ERR_PASS_CLEANUP( w_clonePixrNoInit(pIPWState, prS, &prT));

    ERR_PASS_CLEANUP( wp_dilateGray1D (pIPWState, prS, strelh, prT));
    ERR_PASS_CLEANUP( wp_dilateGray1D (pIPWState, prT, strelv, prD));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));
    ERR_PASS_CLEANUP( w_destroyStrel(pIPWState, &strelh));
    ERR_PASS_CLEANUP( w_destroyStrel(pIPWState, &strelv));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prT);
    w_destroyStrel(pIPWState, &strelh);
    w_destroyStrel(pIPWState, &strelv);
    ERR_RETURN;
}



/*-----------------------------------------------------------------------*
 *                   1 Dimensional Grayscale Erosion                     *
 *-----------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Performs grayscale erosion for 1D structure elements (horiz or vert)
 *
 * LongDescription:::
 *  Grayscale morphology can be done much faster when the structure element
 *  is either purely horizontal or purely vertical.  This routine decides
 *  which of these is true, allocates a temporary image if the source
 *  and dest images are the same, and calls wp_erodeGrayHLow or
 *  wp_erodeGrayVLow do to the real work.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be eroded.
 *  strel::          Pointer to conventional structure element.  This
 *		     type of strel can be used for either binary or
 *		     grayscale morphology.
 *
 * Output:::
 *  prD::            Pointer to output image.  If this is the same as the
 *		     source image, a (possibly quite large) temporary
 *		     image will be allocated and freed here.
 *
 * Notes:::
 *  Grayscale morphology is usually slow.  This special case is faster.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGray1D (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        streltype;
    Int32        ws, hs;
    Int32        frames, framed;
    Int32        bpls, bplt;
    UInt8       *datas, *datat;
    PIXR        *prT = NULL;

    /* find the type of strel */
    ERR_PASS( w_strelType(pIPWState, strel, &streltype));
    
    /* strel must be 1D with solid hit */
    ERR_GEN(streltype != H_1D_STREL && streltype != V_1D_STREL,
	    cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &ws));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hs));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frames));
    ERR_PASS( w_pixrGetFrame(pIPWState, prD, &framed));

	/* if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prT));
    else
	prT = prD;

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &datas));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prT, &datat));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpls));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prT, &bplt));

    if (streltype == H_1D_STREL) /*  horizontal */
	ERR_PASS_CLEANUP( wp_erodeGrayHLow (pIPWState, ws, hs, bplt, framed,
					    datas, bpls, frames,
					    strel->nx, strel->cx, datat));
    else   /* vertical */
	ERR_PASS_CLEANUP( wp_erodeGrayVLow (pIPWState, ws, hs, bplt, framed,
					    datas, bpls, frames,
					    strel->ny, strel->cy, datat));

	/* if in-place, copy again and destroy the temp image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_copy(pIPWState, prT, &prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prD == prS)
	w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Low level routine used to erode with a horizontal structure element.
 *
 * LongDescription:::
 *  Grayscale morphology can be done much faster when the structure element
 *  is either purely horizontal or purely vertical and filled with solid
 *  HITs.  This routine deals with the horizontal case.  This is a routine that
 *  could be put into ipshared.
 *
 * PrivateDescription:::
 *  This routine uses a clever method that mostly looks only at the
 *  pixels on the ends of the strel.  Only when the pixel having the
 *  maximum value falls off the end of the area of interest as we slide
 *  the strel horizontally do we look at the other pixels to find a new
 *  minimum.  We also look at the new pixels arriving to see if they
 *  are the new maximum.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  w::              Widht of unframed dest image.
 *  h::              Height of unframed dest image.
 *  bpld::           Bytes/line of dest image.
 *  framed::         Number of pixels in dest image's frame.
 *  images::         Pointer to OUTSIDE of frame of source image.
 *  bpls::           Bytes/line of source image.
 *  frames::         Number of pixels in source image's frame.
 *  nx::             Width of structure element.
 *  cx::             Offset of structure element's center position
 *		     relative to left end of structure element.
 *
 * Output:::
 *  imaged::         Pointer to OUTSIDE of frame of dest image.
 *
 * Notes:::
 *  Grayscale morphology is usually slow.  Doing the special case
 *  is often tolerable.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGrayHLow (
    IPWState           *pIPWState,
    register Int32      w,
    Int32               h,
    Int32               bpld,
    Int32               framed,
    UInt8              *images,
    Int32               bpls,
    Int32               frames,
    Int32               nx,
    Int32               cx,
    UInt8              *imaged)
{
    PROLOGUE;
    Int32            fullw, fullh, range;
    UInt8            smallestVal, largestVal;
    register UInt8   val, min;
    register UInt8  *pS, *pD;
    Int32            i /*, x, y */;
    register Int32   j;
    Int32            startx, endx;
    register UInt32 *histo;
    UInt32          *histoArray;

    histo = histoArray = NULL;

	/* find smallest and largest values in the image */
    largestVal = BIGVAL8;
    smallestVal = SMLVAL;
    range = (Int32)(largestVal - smallestVal + 1);			

	/* Allocate local histogram */
    histoArray = (UInt32 *)IPW_CALLOC(range, sizeof(UInt32), cAllocatorData);
    ERR_GEN(histoArray == NULL, cErrNomem);
					
    histo = histoArray - smallestVal;
    
    /* set fram to max, don't affect real min */
    fullw = w + 2 * frames;
    fullh = h + 2 * frames;
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)images, fullw, fullh,
				bpls, frames, BIGVAL8));

	/* Initialize lines to be at the left edge of the frame and
	 * on the first row of the proper image.  
	 * Initialize lined to be the first pixel of the proper image.  */

    pS = (images + frames * bpls);
    pD = (imaged + framed * bpld + framed);

	/* horizontal erosion */
    startx = frames - cx;
    endx = startx + nx - 1;
    for (i = 0; i < h; i++)
    {
	    /* initialize the histogram for this row */
	min = BIGVAL8;
	MEMSET((void *)histoArray, 0, (range * sizeof(UInt32)));

	    /* get the first dest pixel in the row */
	for (j = startx; j <= endx; j++)
	{
	    val = UCHAR_ACCESS(pS + j);
	    histo[val]++;
	    if (val < min)
		min = val;
	}
	UCHAR_ACCESS(pD) = min;

	    /* do the rest of the row */
	for (j = 1; j < w; j++)
	{
		/* remove the pixel at the left end and update min */
	    val = UCHAR_ACCESS (pS + startx + j - 1);
	    histo[val]--;
	    if (histo[val] == 0 && val == min)
	    {
		++ histo[largestVal];
		val++;
		while (histo[val] == 0)
		    val++;
		min = val;
		-- histo[largestVal];
	    }

		/* add the pixel at the right end and update min */
	    val = UCHAR_ACCESS (pS + endx + j);
	    histo[val]++;
	    if (val < min)
		min = val;
	    UCHAR_ACCESS(pD + j) = min;

	}
	pS += bpls;
	pD += bpld;

    }

    IPW_FREE(histoArray, cAllocatorData);

    /* reset frame pixels to 0 */
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)images, fullw, fullh,
				bpls, frames, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Low level routine used to dilate with a vertical structure element.
 *
 * LongDescription:::
 *  Grayscale morphology can be done much faster when the structure element
 *  is either purely horizontal or purely vertical and filled with solid
 *  HITs.  This routine deals with the vertical case.  This is a routine that
 *  could be put into ipshared.
 *
 * PrivateDescription:::
 *  This routine uses a clever method that mostly looks only at the
 *  pixels on the ends of the strel.  Only when the pixel having the
 *  minimum value falls off the end of the area of interest as we slide
 *  the strel horizontally do we look at the other pixels to find a new
 *  minimum.  We also look at the new pixels arriving to see if they
 *  are the new minimum.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  w::              Width of unframed dest image.
 *  h::              Height of unframed dest image.
 *  bpld::           Bytes/line of dest image.
 *  framed::         Number of pixels in dest image's frame.
 *  images::         Pointer to OUTSIDE of frame of source image.
 *  bpls::           Bytes/line of source image.
 *  frames::         Number of pixels in source image's frame.
 *  ny::             Height of structure element.
 *  cy::             Offset of structure element's center position
 *		     relative to top end of structure element.
 *
 * Output:::
 *  imaged::         Pointer to OUTSIDE of frame of dest image.
 *
 * Notes:::
 *  Grayscale morphology is usually slow.  Doing the special case
 *  is often tolerable.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGrayVLow (
    IPWState           *pIPWState,
    Int32               w,
    register Int32      h,
    Int32               bpld,
    Int32               framed,
    UInt8              *images,
    Int32               bpls,
    Int32               frames,
    Int32               ny,
    Int32               cy,
    UInt8              *imaged)
{
    PROLOGUE;
    Int32            fullw, fullh, range;
    UInt8            smallestVal, largestVal;
    register UInt8   val, min;
    register UInt8  *lines1, *lined;
    UInt8           *firstlines, *firstlined;
    register Int32   i;
    Int32            j, /* x, */ y;
    Int32            starty;
    register UInt32 *histo;
    UInt32          *histoArray;

    histo = histoArray = NULL;

	/* find smallest and largest values in the image */
    largestVal = BIGVAL8;
    smallestVal = SMLVAL;
    range = (Int32)(largestVal - smallestVal + 1);			

	/* Allocate local histogram */
    histoArray = (UInt32 *)IPW_CALLOC(range, sizeof(UInt32), cAllocatorData);
    ERR_GEN(histoArray == NULL, cErrNomem);

    histo = histoArray - smallestVal;

    /* set fram to max, don't affect real min */
    fullw = w + 2 * frames;
    fullh = h + 2 * frames;
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)images, fullw, fullh,
				bpls, frames, BIGVAL8));
    
	/* Initialize firstlines to be at the top pixel of the SE when
	 * the center of the SE is on the UL pixel of the proper image.
	 * Initialize firstlined to be the first pixel of the
	 * proper image.     */
    starty = frames - cy;
    firstlines = images + starty * bpls + frames;
    firstlined = imaged + framed * bpld + framed;
    y = ny * bpls;

	/* vertical erosion */
    for (j = 0; j < w; j++)
    {
	    /* initialize the histogram for this column */
	min = BIGVAL8;
	MEMSET((void *)histoArray, 0, (range * sizeof(UInt32)));

	    /* get the first dest pixel in the column */
	lined = firstlined;
	for (i = 0, lines1 = firstlines+j; i < ny; i++, lines1 += bpls)
	{
	    val = UCHAR_ACCESS(lines1);
	    histo[val]++;
	    if (val < min)
		min = val;
	}
	UCHAR_ACCESS(lined +j) = min;

	    /* do the rest of the column */
	lines1 = firstlines + j;
	lined += bpld;
	for (i = 1; i < h;
		i++, lines1 += bpls, lined +=bpld)
	{

		/* remove the pixel at the top and update min  */
	    val = UCHAR_ACCESS(lines1);
	    histo[val]--;
	    if (histo[val] == 0 && val == min)
	    {
		++ histo[largestVal];
		val++;
		while (histo[val] == 0)
		    val++;
		min = val;
		-- histo[largestVal];
	    }

		/* add the pixel at the bottom and update min */
	    val = UCHAR_ACCESS(lines1 + y);
	    histo[val]++;
	    if (val < min)
		min = val;
	    UCHAR_ACCESS(lined+j) = min;

	}
    }

    IPW_FREE(histoArray, cAllocatorData);

    /* reset frame pixels to 0 */
    ERR_PASS( w_setFramePixelsGray(pIPWState, (UInt32 *)images, fullw, fullh,
				bpls, frames, 0));

    NORMAL_RETURN;
}


/*-----------------------------------------------------------------------*
 *              2 Dimensional Separable Grayscale Erosion                *
 *-----------------------------------------------------------------------*/
/*
 *  erodeGraySep()
 *
 *        Restrictions:  the input strel must be a separable 2D strel,
 *                       composed of solid hits.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform grayscale erosion for separable structure elements.
 *
 * LongDescription:::
 *  A separable structure element is one that can be decomposed into
 *  a purely horizontal strel and a purely vertical strel.  One does
 *  an erosion using the horizontal strel followed by an erosion on
 *  the result with the vertical strel.  This should give the same
 *  result as an erosion with the original strel but uses the optimized
 *  grayscale operators.  However, it does require allocation of
 *  a (possibly quite large) gray temporary image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be eroded.
 *  strel::          Pointer to conventional structure element.  This
 *		     type of strel can be used for either binary or
 *		     grayscale morphology.
 *
 * Output:::
 *  prD::            Pointer to output image.  Since a temporary image
 *		     is always allocated here, there is no extra penalty for
 *		     for setting prS = prD.
 *
 * Notes:::
 *  Grayscale morphology is slow.  This is somewhat faster than the general
 *  case.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Vincent, P. Macleod
 *
 * Citations:::
 *  Lots of books talk about binary morphology.  Grayscale morphology
 *  is newer and slower.  Luc Vincent at XIS knows all there is to know
 *  about this.  Our grayscale morphology code comes from him.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_erodeGraySep (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        streltype;
    STREL       *strelh = NULL, *strelv = NULL;
    PIXR        *prT = NULL;

    /* find the type of strel */
    ERR_PASS( w_strelType(pIPWState, strel, &streltype));
    ERR_GEN(streltype != S_2D_STREL, cErrInvalidParm);

    ERR_PASS( w_strelSeparation(pIPWState, strel, &strelh, &strelv));

    ERR_PASS_CLEANUP( w_clonePixrNoInit(pIPWState, prS, &prT));

    ERR_PASS_CLEANUP( wp_erodeGray1D (pIPWState, prS, strelh, prT));
    ERR_PASS_CLEANUP( wp_erodeGray1D (pIPWState, prT, strelv, prD));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT));
    ERR_PASS_CLEANUP( w_destroyStrel(pIPWState, &strelh));
    ERR_PASS_CLEANUP( w_destroyStrel(pIPWState, &strelv));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prT);
    w_destroyStrel(pIPWState, &strelh);
    w_destroyStrel(pIPWState, &strelv);
    ERR_RETURN;
}


/*

$Log: morphg4.c,v $
Revision 1.1.1.1  1999/09/20 18:19:19  jlin
Imported sources

Revision 1.24  1997/11/21 20:36:22  xzhu
xzhu on Fri Nov 21 12:35:29 PST 1997

morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v

Taked out extra stuff to speed up.

Revision 1.23  1997/04/15 22:17:32  ddavies
morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v

Use non-initialized PIXR creation where possible.

Revision 1.22  1996/01/04 20:25:04  ejaquith
morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.21  1995/12/13  04:10:47  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:14:02  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/25  00:05:30  jcraig
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * remove include of ipinc.h
 *
 * Revision 1.18  1995/10/13  22:03:21  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/05  04:27:30  ddavies
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.16  1995/09/21  22:54:52  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/18  23:47:21  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * Typo/bug.
 * For in-place checking, was checking if (prD == prD) when should have been
 * (prS == prD).
 *
 * Revision 1.14  1995/09/15  18:00:49  ddavies
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.13  1995/09/11  22:11:29  ddavies
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.12  1995/09/10  00:33:30  ddavies
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.11  1995/09/07  01:33:36  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.10  1995/09/06  20:21:41  ejaquith
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  01:17:32  ddavies
 * morphg4.c -> /product/ipcore/ipwrap/src/RCS/morphg4.c,v
 *
 * Put minimum number of include files back in.
 *

 */

