/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   mrc.c:
 *
 *	Code used to compress images using the Modified
 *	Read (CCITT Group 3) algorithm.
 *	
 *	w_compressRowsMR():			compress one or more rows
 *						of image into Modified
 *						Read format.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "ccitt.pub"
#include "mr.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: mrc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/* This file contains wrappers for routines used to compress images
 * into Modified Read, or CCITT Group 3, or CCITT Fascicle
 * VII.3, Recommendation T.4 bytestreams.  Also see section 11 of
 * the TIFF 6.0 specification.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * The routine w_compressRowsMR does the real work.  It returns
 * when either it fills the compressed data buffer or has done the right
 * number of lines.  The client tells it whether the image information
 * is new, the compressed data buffer is new and whether the compressed data
 * should be viewed as an extension of the previous block of compressed
 * data.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress rows of a binary image using the Modified Read technique.
 *
 * LongDescription:::
 *  Compress some rows of an image using the MR, or CCITT Group 3 algorithm.
 *  We will stop when either an error is encountered, we run out of image rows
 *  or we fill the compressed data buffer.
 *
 * Input:::
 *  pIPWState::             Pointer to standard IPWrap state block
 *  pMRCompInfoIn::         Pointer to private MR compression state block
 *  prS::                   Pointer to image to be compressed.
 *  startLine::             First image line to be compressed
 *			    relative to the top of the image, not frame.
 *  numLines::              Number of lines to be compressed.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *  continueImageBlock::    cTrue => we are compressing
 *			    from the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pMRCompInfoIn is
 *			    valid.<BR>
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *			    When this arg is true, startLine and numLines
 *			    should be set to the same value as they were
 *			    on the previous call.  The image should also
 *			    be the same.
 *  continueCompDataStream:: cTrue => The saved compression state
 *			    in pMRCompInfoIn is valid.  If
 *			    the previous call to this routine
 *			    filled the compressed data buffer
 *			    before the strip, tile or image was
 *			    complete, a new buffer should be
 *			    supplied.  This arg is set to 
 *			    cFalse when the block of compressed
 *			    data is unrelated to the block used
 *			    in the previous call to this routine
 *			    as recorded in pMRCompInfoIn.
 *			    SETTING THIS ARG TO cFalse CAUSES
 *			    DATA IN THE PIPELINE TO BE DISCARDED!
 *
 * Output:::
 *  pExhaustedImageData::   Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it filled
 *			    the compressed data buffer.  This
 *			    routine may be called again with
 *			    a new buffer.
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been compressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  Some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *
 * Notes:::
 *  This routine compresses a horizontal slice of a binary image using the
 *  Modified Read (CCITT Group 3) technique described in the
 *  TIFF 6.0 spec and in the CCITT documents cited below.  There
 *  are several completion conditions:
 *
 *	retCode	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  Some of the
 *		compressed data may still be in the
 *		pipeline and the saved compressed data state is valid.
 *		This routine may be called again with
 *		continueCompDataStream = cTrue and the new compressed
 *		data will be seamlessly concatenated with the old
 *		compressed data (on any bit boundary).  The
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  Some data may
 *		remain in the pipeline.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine filled the compressed data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed and some
 *		data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.<BR> 
 *		The image may be moved around as a block after this
 *		condition has been returned, but the internal
 *		composition of the image (and startLine and numLines) 
 *		should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *<BR><BR>		
 *		The image data is always in PIXR order.
 *		
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, MR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  "Standardization of Group 3 facsimile apparatus for document
 *  transmission," Recommendation T.4, Volume VII, Fascicle VII.3,
 *  Terminal Equipment and Protocols for Telematic Services.  The
 *  International Telegraph and Telephone Consultative Committee (CCITT),
 *  Geneva, 1985, pages 16 through 31.  You can write to ANSI,
 *  Attention: Sales, 1430 Broadway, New York, N.Y., 10018.
 *  Or, if this isn't out of date by the time you read this, try
 *  http://icib.igd.fhg.de/icib/telecom/ccitt/rec_t.4/gen.html.  The
 *  version I saw was lame - it didn't have any of the VERY useful figures.
 *  Section 11 of the TIFF 6.0 Specification is also helpful since it
 *  talks about using compression in files instead of in fax machines.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_compressRowsMR (
    IPWState           *pIPWState,
    void               *pMRCompInfoIn,
    PIXR               *prS,
    UInt32              startLine,
    UInt32              numLines,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32              continueImageBlock,
    UInt32              continueCompDataStream,
    UInt32             *pExhaustedImageData,
    UInt32             *pExhaustedCompData,
    UInt32             *pPercentDone,
    UInt32             *pBytesWritten)
{
    PROLOGUE;
    Int32    w, bpl;
    UInt8   *pS;

    ERR_GEN( pMRCompInfoIn == NULL || prS == NULL || pCompData == NULL ||
	     pExhaustedImageData == NULL || pExhaustedCompData == NULL ||
	     pPercentDone == NULL || pBytesWritten == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

    ERR_PASS( i_compressRowsMR (pIPWState, pMRCompInfoIn,
			        pS,
			        w,
			        startLine + cFrameBits,
			        numLines,
			        bpl,
			        pCompData,
			        compDataBufSize,
			        continueImageBlock,
			        continueCompDataStream,
			        pExhaustedImageData,
			        pExhaustedCompData,
			        (Int32 *)pPercentDone,
			        (Int32 *)pBytesWritten));

    NORMAL_RETURN;
}



/*

 $Log: mrc.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.2  1998/01/14 00:14:07  ddavies
 mrc.c -> /product/ipcore/ipwrap/src/RCS/mrc.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.1  1996/05/11 01:23:02  ddavies
 Initial revision


 */


