/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* For malloc */


#include "ipwcore.prv"


#include "nrutil.pub"
#include "fxpmath.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: nrutil.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Matrix x vector multiplication.
 *
 * LongDescription:::
 *  This routine multiplies a matrix by a vector.  The matrix elements
 *  are signed bytes.  The input vector elements are signed 32-bit integers.
 *  At most 15 of the bits in each input vector element are significant,
 *  the rest are sign bits.  The output is another vector of signed
 *  32-bit integers.  This routine is used in neural net-like operations.
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pMatrix::        pointer to input matrix of signed bytes
 *  pInVector::      pointer to input vector
 *  numRows::        number of rows in input matrix.  Also equals
 *		     the length of the output vector.
 *  numCols::        number of columns in input matrix.  Also equals
 *		     the length of the input vector
 *
 * Output:::
 *  pOutVector::     Pointer to output vector
 *
 * Notes:::
 *  The columns of the weight matrix all start on 8-byte boundaries.
 *  The input and output vectors start on 8-byte boundaries.  These help
 *  optimized libraries.
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 * Rich Smith at XIS, Dan Davies
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_matvec (
    IPWState           *pIPWState,
    Int8	       *pMatrix,
    Int32	       *pInVector,
    Int32		numRows,
    Int32		numCols,
    Int32	       *pOutVector)
{
    PROLOGUE;

    ERR_PASS( i_matvec_mult_8BA (pIPWState, pMatrix, pInVector, pOutVector,
				 numRows, numCols));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * allocate a Float64 vector with subscript range v[nl..nh]
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  nl::             	lower vector subscript
 *  nh::             	upper vector subscript
 *
 * Output:::
 *  vector::         	new vector with the given subscript range
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated vector
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_vector (
    IPWState           *pIPWState,
    Int32               nl,
    Int32               nh,
    Float64           **vector)
{
	PROLOGUE;
	Float64 *v;

	v=(Float64 *)IPW_MALLOC((nh-nl+1)*sizeof(Float64), cAllocatorData) - nl;
 	ERR_GEN(v == NULL, cErrNomem);
	*vector = v;
 	NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * allocate an Int32 vector with subscript range v[nl..nh]
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  nl::             	lower vector subscript
 *  nh::             	upper vector subscript
 *
 * Output:::
 *  vector::         	new vector with the given subscript range
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated vector
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_ivector (
    IPWState           *pIPWState,
    Int32               nl,
    Int32               nh,
    Int32             **vector)
{
	PROLOGUE;
	Int32 *v;

	v=(Int32 *)IPW_MALLOC((nh-nl+1)*sizeof(Int32), cAllocatorData) - nl;
 	ERR_GEN(v == NULL, cErrNomem);
	*vector = v;
 	NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * allocate a Float64 matrix with subscript range m[nrl..nrh][ncl..nch]
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  nrl::            	lower range of rows
 *  nrh::            	upper range of rows
 *  ncl::            	lower range of cols
 *  nch::            	upper rangle of cols
 *
 * Output:::
 *  matrix::         	new matrix with the given subscript range
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated vector
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_matrix (
    IPWState           *pIPWState,
    Int32               nrl,
    Int32               nrh,
    Int32               ncl,
    Int32               nch,
    Float64          ***matrix)

{
	PROLOGUE;
	Int32 i;
	Float64 **m;

	/* allocate pointers to rows */
	m=(Float64 **) IPW_MALLOC((nrh-nrl+1)*sizeof(Float64*), cAllocatorData) - nrl;
 	ERR_GEN(m == NULL, cErrNomem);

	/* allocate rows and set pointers to them */
	for(i=nrl;i<=nrh;i++) {
		m[i]=(Float64 *)IPW_MALLOC((nch-ncl+1)*sizeof(Float64),
					   cAllocatorData) - ncl;
		ERR_GEN_CLEANUP(m[i] == NULL, cErrNomem);
	}
	/* return pointer to array of pointers to rows */
	*matrix = m;
 	NORMAL_RETURN;

ERR_CLEANUP:
	for (i = nrl; i <= nrh; i++)
		if (m[i] != NULL) IPW_FREE(m[i], cAllocatorData);
		else break;
	IPW_FREE(m, cAllocatorData);
	ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * allocate an Int32 matrix with subscript range m[nrl..nrh][ncl..nch]
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  nrl::            	lower range of rows
 *  nrh::            	upper range of rows
 *  ncl::            	lower range of cols
 *  nch::            	upper rangle of cols
 *
 * Output:::
 *  matrix::         	new matrix with the given subscript range
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated vector
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_imatrix (
    IPWState           *pIPWState,
    Int32               nrl,
    Int32               nrh,
    Int32               ncl,
    Int32               nch,
    Int32            ***matrix)

{
	PROLOGUE;
	Int32 i,**m;

	/* allocate pointers to rows */
	m = (Int32 **)IPW_MALLOC((nrh-nrl+1)*sizeof(Int32*), cAllocatorData) - nrl;
	ERR_GEN(m == NULL, cErrNomem);

	/* allocate rows and set pointers to them */
	for(i=nrl;i<=nrh;i++) {
		m[i]=(Int32 *)IPW_MALLOC((nch-ncl+1)*sizeof(Int32),
				         cAllocatorData) - ncl;
		ERR_GEN_CLEANUP(m[i] == NULL, cErrNomem);
	}
	/* return pointer to array of pointers to rows */
	*matrix = m;
 	NORMAL_RETURN;

ERR_CLEANUP:
	for (i = nrl; i <= nrh; i++)
		if (m[i] != NULL) IPW_FREE(m[i], cAllocatorData);
		else break;
	IPW_FREE(m, cAllocatorData);
	ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * point a submatrix [newrl..][newcl..] to a[oldrl..oldrh][oldcl..oldch]
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  a::              	original matrix
 *  oldrl::          	original lower range of rows
 *  oldrh::          	original upper range of rows
 *  oldcl::          	original lower range of cols
 *  oldch::          	original upper range of cols
 *  newrl::          	new lower range of rows
 *  newcl::          	new lower range of cols
 *
 * Output:::
 *  submatrix::      	new matrix with the given subscript range
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated vector
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_submatrix (
    IPWState           *pIPWState,
    Float64           **a,
    Int32               oldrl,
    Int32               oldrh,
    Int32               oldcl,
    Int32               oldch,
    Int32               newrl,
    Int32               newcl,
    Float64          ***submatrix)

{
	PROLOGUE;
	Int32 i,j;
	Float64 **m;

	/* keep the compiler happy */
	newcl = newcl;
	oldch = oldch;
	oldcl = oldcl;

	/* allocate array of pointers to rows */
	m=(Float64 **)IPW_MALLOC((oldrh-oldrl+1)*sizeof(Float64*), 
				 cAllocatorData) - newrl;
	ERR_GEN(m == NULL, cErrNomem);

	/* set pointers to rows */
	for(i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i];

	/* return pointer to array of pointers to rows */
	*submatrix = m;
 	NORMAL_RETURN;
}

/* free a Float64 vector allocated with vector() */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * free a Float64 vector allocated with w_nr_vector
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  v::              	vector to free
 *  nl::             	lower range of vector
 *  nh::             	upper range of vector
 *
 * Returns:::
 *  cSuccessful		successfully deallocated memory
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_free_vector (
    IPWState           *pIPWState,
    Float64            *v,
    Int32               nl,
    Int32               nh)
{
	PROLOGUE;
	void		*ptr;

	    /* keep the compiler happy */
	nh = nh;
	ptr = (void *)(v+nl);
	IPW_FREE(ptr, cAllocatorData);
 	NORMAL_RETURN;
}

/* free an int vector allocated with ivector() */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * free an Int32 vector allocated with w_nr_ivector
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  v::              	vector to free
 *  nl::             	lower range of vector
 *  nh::             	upper range of vector
 *
 * Returns:::
 *  cSuccessful		successfully deallocated memory
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_free_ivector (
    IPWState           *pIPWState,
    Int32              *v,
    Int32               nl,
    Int32               nh)
{
	PROLOGUE;
	void		*ptr;

	    /* keep the compiler happy */
	nh = nh;
	ptr = (void *)(v+nl);
	IPW_FREE(ptr, cAllocatorData);
 	NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * free a Float64 matrix allocated by w_nr_matrix
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  m::              	matrix to free
 *  nrl::            	lower range of rows
 *  nrh::            	upper range of rows
 *  ncl::            	lower range of cols
 *  nch::            	upper range of cols
 *
 * Returns:::
 *  cSuccessful		successfully deallocated memory
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_free_matrix (
    IPWState           *pIPWState,
    Float64           **m,
    Int32               nrl,
    Int32               nrh,
    Int32               ncl,
    Int32               nch)
{
	PROLOGUE;
	Int32 i;
	void		*ptr;

	    /* keep the compiler happy */
	nch = nch;
	for(i = nrh; i >= nrl; i--) 
	{
		ptr = (void *)(m[i]+ncl);
		IPW_FREE(ptr, cAllocatorData);
	}
	ptr = (void *)(m+nrl);
	IPW_FREE(ptr, cAllocatorData);
 	NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * free an Int32 matrix allocated by w_nr_imatrix
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  m::              	matrix to free
 *  nrl::            	lower range of rows
 *  nrh::            	upper range of rows
 *  ncl::            	lower range of cols
 *  nch::            	upper range of cols
 *
 * Returns:::
 *  cSuccessful		successfully deallocated memory
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_free_imatrix (
    IPWState           *pIPWState,
    Int32             **m,
    Int32               nrl,
    Int32               nrh,
    Int32               ncl,
    Int32               nch)
{
	PROLOGUE;
	Int32 i;
	void		*ptr;

	    /* keep the compiler happy */
	nch = nch;
	for(i = nrh; i >= nrl; i--) 
	{
		ptr = (void *)(m[i]+ncl);
		IPW_FREE(ptr, cAllocatorData);
	}
	ptr = (void *)(m+nrl);
	IPW_FREE(ptr, cAllocatorData);
 	NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * free a submatrix allocated by w_nr_submatrix
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  b::              	submatrix to free
 *  nrl::            	lower range of rows
 *  nrh::            	upper range of rows
 *  ncl::            	lower range of cols
 *  nch::            	upper range of cols
 *
 * Returns:::
 *  cSuccessful		successfully deallocated memory
 *
 * FunctionCategory:::
 *  array management
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Numerical Recipes in C.  Appendix B.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_nr_free_submatrix (
    IPWState           *pIPWState,
    Float64           **b,
    Int32               nrl,
    Int32               nrh,
    Int32               ncl,
    Int32               nch)
{
	PROLOGUE;
	void		*ptr;

	    /* keep the compiler happy */
	nch = nch;
	ncl = ncl;
	nrh = nrh;
	ptr = (void *)(b+nrl);
	IPW_FREE(ptr, cAllocatorData);
 	NORMAL_RETURN;
}

/*
$Log: nrutil.c,v $
Revision 1.1.1.1  1999/09/20 18:19:19  jlin
Imported sources

Revision 1.17  1998/01/14 00:14:10  ddavies
nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.16  1996/10/30 06:01:01  ddavies
ddavies on Tue Oct 29 23:00:40 PPET 1996

nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v

Add implementation for w_matvec.  This is a wrapper for a matrix x vector
multiplication routine.

Revision 1.15  1996/01/04 20:25:07  ejaquith
nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.14  1995/12/13  04:10:51  ejaquith
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.13  1995/11/13  20:14:06  ejaquith
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.12  1995/10/13  22:03:24  ejaquith
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.11  1995/09/29  00:15:05  jcraig
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: fixup procedure comments
 *
 * Revision 1.10  1995/09/21  22:54:55  ejaquith
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.9  1995/09/06  20:21:44  ejaquith
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  02:15:09  jcraig
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.7  1995/08/31  21:15:30  ejaquith
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/17  22:32:32  jcraig
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.5  1994/09/28  19:06:47  lperry
 * lperry on Wed Sep 28 12:05:48 PDT 1994
 *
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * Update to use nrutil.pub for prototypes rather than
 * utilpro.h. Also get rid of unnecessary includes.
 *
 * Revision 1.4  1994/06/08  20:33:26  macleod
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * Was including "utilpro.h", when it should have been including "utilpros.h".
 *
 * Revision 1.3  1994/05/05  01:44:05  ddavies
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * Change ifdef __GUNC__ to ifdef __GNUC__.
 *
 * Revision 1.2  1994/05/02  20:29:46  jcraig
 * nrutil.c -> /product/ipcore/ipwrap/src/RCS/nrutil.c,v
 *
 * add #ifdef __GNUC__ around include of ansiprot.h
 *
 * Revision 1.1  1994/02/04  19:31:23  macleod
 * Initial revision
 *

*/


