/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *	odither.c:
 *
 *          Dithering code using ordered dither algorithm
 *	    This is color/grayscaledithering code that takes and consumes entire
 *	    images.
 *
 *	    24 bit/pixel ordered dither: table creation
 *		    w_ditherOrderedMakeTables()
 *		    
 *	    24 bit/pixel to 8 bit/pixel ordered dither:
 *		    w_ditherOrdered24To8Pr()
 *		    w_ditherRescale24To8Pr
 *
 *	    24 bit/pixel to 4 bit/pixel ordered dither:
 *		    w_ditherRescale24To4Pr ()
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rectops.pub"
#include "odither.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: odither.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make tables used to perform ordered dithering.
 *
 * LongDescription:::
 *  There are several types of dithering.  Halftoning tesellates a page into
 *  halftone cells.  The more pixels in each cell are black, the darker
 *  the cell looks.  In halftoning, all the dots are clumped together.
 *  Ordered dithering is much the same except that the dots are not clumped
 *  in the cell.  Each cell appears to grow uniformly darker.  The quality
 *  is generally lower than for error diffusion (at least on devices
 *  where every pixel is visible), but ordered dithering is a MUCH faster
 *  operation.  We use it primarily when creating images for color
 *  monitors.
 *<BR><BR>
 *  This routine creates the tables used to decide which dots to turn on
 *  at in what order for given pixel intensities.  These tables are only
 *  used for color dithering.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  reds::           Number of levels of red intensity in the output.
 *  greens::         Number of levels of green intensity in the output.
 *  blues::          Number of levels of blue intensity in the output.
 *  rTRC::           Pointer to red Tone Reproduction Curve.  May be NULL.
 *  gTRC::           Pointer to green Tone Reproduction Curve.  May be NULL.
 *  bTRC::           Pointer to blue Tone Reproduction Curve.  May be NULL.
 *
 * Output:::
 *  ppTable::        Pointer to pointer to the table produced.  *ppTable must
 *		     be NULL on entry.  The table is always allocated here.
 *		     The client can simply free this table using the
 *		     data (as opposed to image) freeing routine that was
 *		     was passed to IPWrap when the IPState structure was
 *		     created.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppTable is not NULL.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, dithering, color
 *
 * Level:::
 *  heuristic, selection of the dither pattern is heuristic.
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  Peter found this in a wonderful paper that he could probably find again.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherOrderedMakeTables (
    IPWState           *pIPWState,
    Int32               reds,
    Int32               greens,
    Int32               blues,
    UInt8              *rTRC,
    UInt8              *gTRC,
    UInt8              *bTRC,
    void              **ppTable)
{
    PROLOGUE;
    Int32    bytes;
    UInt8   *table = NULL;

    ERR_GEN( rTRC == NULL || gTRC == NULL || bTRC == NULL || ppTable == NULL,
	     cErrNullPtr);
    ERR_GEN( *ppTable != NULL, cErrPtrNotNull);

    *ppTable = NULL;
    ERR_PASS( i_orderedRGBToLinearTableSize (pIPWState, reds, greens, blues,
					     &bytes));
    ERR_GEN(bytes == 0, cErrInternal);

    table = IPW_MALLOC (bytes, cAllocatorData);
    ERR_GEN(table == NULL, cErrNomem);

    ERR_PASS_CLEANUP( i_orderedRGBToLinearMakeTable (pIPWState,
						     reds, greens, blues,
						     rTRC, gTRC, bTRC,
						     table));
    *ppTable = table;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(table, cAllocatorData);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform ordered dither from RGB to 8 bpp indexed color using linear palette.
 *
 * LongDescription:::
 *  There are several types of dithering.  Halftoning tesellates a page into
 *  halftone cells.  The more pixels in each cell are black, the darker
 *  the cell looks.  In halftoning, all the dots are clumped together.
 *  Ordered dithering is much the same except that the dots are not clumped
 *  in the cell.  Each cell appears to grow uniformly darker.  The quality
 *  is generally lower than for error diffusion (at least on devices
 *  where every pixel is visible), but ordered dithering is a MUCH faster
 *  operation.  We use it primarily when creating images for color
 *  monitors.
 *<BR><BR>
 *  This routine converts a 24 bit/pixel RGB image into an 8 bit/pixel
 *  indexed colorimage using a linear color map.  Each output pixel is
 *  found by looking it up in the table created by w_ditherOrderedMakeTables.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to source 24 bpp RGB image.
 *  pRect::          Pointer to rectangle of source image to be converted.
 *		     If this is NULL, the entire image is dithered.
 *  xStart::         Starting x-offset, used to determine
 *		     the starting column in the dither matrix.
 *		     This does NOT affect which column is processed.
 *  yStart::         Starting x-offset, used to determine
 *		     the starting line in the dither matrix.
 *		     This does NOT affect which line is processed.
 *  ditherTab::      Pointer to table created by w_ditherOrderedMakeTables.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to output image.  If none is supplied,
 *		     one is created.  If one is supplied, its dimensions,
 *		     depth, channels, etc are checked.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrOutOfBounds
 *  if the rectangle specified lies at least partially outside the
 *  source image.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, dithering, color
 *
 * Level:::
 *  heuristic, selection of the dither pattern is heuristic.
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  Peter found this in a wonderful paper that he could probably find again.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherOrdered24To8 (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    Int32               xStart,
    Int32               yStart,
    void               *ditherTab,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pR, *pG, *pB, *pD;
    PIXR		*subG, *subR, *subB;
    RECTANGL		 lRect;
    Int8		 valid;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL || ditherTab == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 8,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* if we weren't given an rectangle, make one up */
    if (pRect == NULL)
    {
	pRect = &lRect;
	lRect.x = 0;
	lRect.y = 0;
	lRect.w = wS;
	lRect.h = hS;
    }
    else 	/* if we were given a rectangle, is it ok? */
    {
	ERR_PASS_CLEANUP( w_isRectValid (pIPWState, prS, pRect, &valid));
	ERR_GEN_CLEANUP( !valid, cErrOutOfBounds);
    }

    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

    /* Do the call */
    ERR_PASS_CLEANUP( i_orderedRGBToLinear8 (pIPWState, pD,
					     bplD, pR, pG, pB, bplS,
					     pRect->x + cFrameBits,
					     pRect->y + cFrameBits,
					     pRect->w,
					     pRect->h,
					     xStart, yStart, ditherTab));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Performs a non-interpolated rescale followed by an ordered dither.
 *
 * LongDescription:::
 *  As much as we try to avoid combo operations, this one was required
 *  to make the PerfectScan viewer fast enough.  This routine chops a
 *  rectangle out of an image, rescales it to be the right size, then
 *  performs an ordered dither on the result.  The speed up came from
 *  avoiding allocating an extra copy of the rescaled input.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to source 24 bpp RGB image.
 *  pRect::          Pointer to rectangle of source image to be converted.
 *		     If this is NULL, the entire image is dithered.
 *  xStart::         Starting x-offset, used to determine
 *		     the starting column in the dither matrix.
 *		     This does NOT affect which column is processed.
 *  yStart::         Starting x-offset, used to determine
 *		     the starting line in the dither matrix.
 *		     This does NOT affect which line is processed.
 *  ditherTab::      Pointer to table created by w_ditherOrderedMakeTables.
 *
 * Output:::
 *  prD::            Pointer to destination image.  This cannot be NULL.
 *		     the dimensions of the destination image determine
 *		     how the rectangle from the source image is rescaled.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, scaling, dithering, color
 *
 * Level:::
 *  heuristic, selection of the dither pattern is heuristic.
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  Peter found the ordered dither algorithm in a wonderful paper that
 *  he could probably find again.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherRescale24To8 (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    Int32               xStart,
    Int32               yStart,
    void               *ditherTab,
    PIXR               *prD)
{
    PROLOGUE;
    Int32		 wS, hS, wD, hD, bplS, bplD;
    UInt8		*pR, *pG, *pB, *pD;
    PIXR		*subG, *subR, *subB;
    RECTANGL	 lRect;
    Int8		 valid;

    ERR_GEN(prS == NULL || prD == NULL || ditherTab == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth8, cChannels1, cFrame32);

	/* if we weren't given an rectangle, make one up */
    if (pRect == NULL)
    {
	ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
	ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
	pRect = &lRect;
	lRect.x = 0;
	lRect.y = 0;
	lRect.w = wS;
	lRect.h = hS;
    }
    else 	/* if we were given a rectangle, is it ok? */
    {
	ERR_PASS( w_isRectValid (pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrOutOfBounds);
    }

    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS( w_pixrGetBpl(pIPWState, prD, &bplD));

    ERR_PASS( w_pixrGetWidth (pIPWState, prD, &wD));
    ERR_PASS( w_pixrGetHeight (pIPWState, prD, &hD));

    /* Do the call */
    ERR_PASS( i_rescaleRGBToLinear8 (pIPWState, pD, bplD,
				     wD, hD, pR, pG, pB, bplS,
				     pRect->x + cFrameBits,
				     pRect->y + cFrameBits,
				     pRect->w,
				     pRect->h,
				     xStart, yStart, ditherTab));
    NORMAL_RETURN;
}


/***********************************************************************
 * ditherRescale24To4Pr()
 *
 *	Ordered dither from 24 bits/pixel to 4 bits/pixel, using an 
 *	"arbitrary" palette, including a non-interpolated rescale.
 *
 * Arguments:
 *	prD	    -	    Dest image.  Ratio of sizes of source and
 *			    dest images tells us the amount of rescaling
 *			    required.
 *	prS	    -	    Source image, should be RGB
 *	rect	    -	    Rectangle from source image, or NULL if whole
 *			    image
 *	xStart	    -	    Starting x-offset, used to determine
 *			    the starting column in the dither matrix
 *			    Does NOT affect which column is processed.
 *	yStart	    -	    Starting y-offset, used to determine
 *			    the starting line in the dither matrix
 *			    Does NOT affect which line is processed.
 *	ditherTab   -	    Dither table returned by the "makeTable" routine
 *			    above
 *	paletteDecode -     LUT to de-scramble linear palette into
 *			    the destination palette
 *				
 *
 * Returns:
 *	error status
 ***********************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rescale and ordered dither RGB to 4 bpp color using arbitrary palette.
 *
 * LongDescription:::
 *  As much as we'd like to avoid combo functions, sometimes events
 *  conspire against us.  Dumb PC displays only have 16 colors.  This
 *  routine chops a rectangle out of an RGB image, rescales it to fit
 *  the destination image, the performs an ordered dither to a 4 bit/pixel
 *  indexed color image.  An extra table is provided that translates between
 *  the linear color map we really dither to and the arbitrary color
 *  palette used in the display.
 *<BR><BR>
 *  This function exists solely as a method of speeding up the PerfectScan
 *  viewer for an important contract.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to source 24 bpp RGB image.
 *  pRect::          Pointer to rectangle of source image to be converted.
 *		     If this is NULL, the entire image is dithered.
 *  xStart::         Starting x-offset, used to determine
 *		     the starting column in the dither matrix.
 *		     This does NOT affect which column is processed.
 *  yStart::         Starting x-offset, used to determine
 *		     the starting line in the dither matrix.
 *		     This does NOT affect which line is processed.
 *  ditherTab::      Pointer to table created by w_ditherOrderedMakeTables.
 *		     So far as I can tell, this should have two red levels,
 *		     two green levels and two blue levels.
 *  paletteDecode::  A table used to convert 3 bpp indexed color values
 *		     produced by the normal ordered dither process into
 *		     4 bit/pixel values that the output display likes.
 *		     We seem to use the magic VGAPalletteDecode, a table from
 *		     outer space.
 *
 * Output:::
 *  prD::            Pointer to destination image.  This cannot be NULL.
 *		     the dimensions of the destination image determine
 *		     how the rectangle from the source image is rescaled.
 *
 * Notes:::
 *  Possibly only Macduff Hughes really knows how to use this beast.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, scaling, dithering, color
 *
 * Level:::
 *  heuristic, selection of the dither pattern is heuristic.
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * Citations:::
 *  Peter found the ordered dither algorithm in a wonderful paper that
 *  he could probably find again.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ditherRescale24To4 (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    Int32               xStart,
    Int32               yStart,
    void               *ditherTab,
    UInt8              *paletteDecode,
    PIXR               *prD)
{
    PROLOGUE;
    Int32		 wS, hS, wD, hD, bplS, bplD;
    UInt8		*pR, *pG, *pB, *pD;
    PIXR		*subG, *subR, *subB;
    RECTANGL	 lRect;
    Int8		 valid;

    ERR_GEN(prS == NULL || prD == NULL || ditherTab == NULL ||
	    paletteDecode == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth4, cChannels1, cFrame32);

	/* if we weren't given an rectangle, make one up */
    if (pRect == NULL)
    {
	ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
	ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
	pRect = &lRect;
	lRect.x = 0;
	lRect.y = 0;
	lRect.w = wS;
	lRect.h = hS;
    }
    else 	/* if we were given a rectangle, is it ok? */
    {
	ERR_PASS( w_isRectValid (pIPWState, prS, pRect, &valid));
	ERR_GEN( !valid, cErrOutOfBounds);
    }

    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel(pIPWState, prS, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subR, &pR));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subG, &pG));
    ERR_PASS( w_pixrGetImageAll(pIPWState, subB, &pB));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS( w_pixrGetBpl(pIPWState, prD, &bplD));

	/* this is how we know the rescale amount */
    ERR_PASS( w_pixrGetWidth (pIPWState, prD, &wD));
    ERR_PASS( w_pixrGetHeight (pIPWState, prD, &hD));

    /* Do the call */
    ERR_PASS( i_rescaleRGBToPalette4 (pIPWState, pD, bplD,
				      wD, hD, pR, pG, pB, bplS,
				      pRect->x + cFrameBits,
				      pRect->y + cFrameBits,
				      pRect->w,
				      pRect->h,
				      xStart, yStart, ditherTab,
				      paletteDecode));
    NORMAL_RETURN;
}


/*
$Log: odither.c,v $
Revision 1.1.1.1  1999/09/20 18:19:19  jlin
Imported sources

Revision 1.26  1998/01/14 00:14:12  ddavies
odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.25  1997/05/02 01:26:13  ddavies
odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v

Remove wrapper (w_ditherOrdered8To1) for obsolete routines
i_orderedDither8To1.

Revision 1.24  1997/04/15 22:17:45  ddavies
odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v

Use non-initialized PIXR creation where possible.

Revision 1.23  1996/03/21 17:10:30  macleod
odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v

The check for a valid rectangle in w_ditherOrdered24To8 and

 * Revision 1.22  1996/01/04  20:25:09  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.21  1995/12/13  04:10:55  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:14:09  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:03:27  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:30:42  ddavies
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 *
 * Revision 1.17  1995/09/21  22:54:58  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/15  18:02:11  ddavies
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.15  1995/09/11  22:12:42  ddavies
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.14  1995/09/10  00:35:59  ddavies
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.13  1995/09/07  01:33:42  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:21:46  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  01:17:38  ddavies
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:15:33  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/31  19:33:31  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Commented out the "#if ipshared_done" stuff to stop PC warning.
 *
 * Revision 1.8  1995/08/24  02:54:43  ejaquith
 * ejaquith on Wed Aug 23 19:53:40 PDT 1995
 *
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Yanked invocations of isRectValid() that I had added to routines where the
 * rect was referring to the source (larger) image but was checking the
 * dest (reduced) image.
 * Yanked procName() garbage out to kill warnings.
 *
 * Revision 1.7  1995/08/17  22:04:10  ejaquith
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.6  1995/02/07  21:25:35  macleod
 * macleod on Tue Feb 7 13:25:04 PST 1995
 *
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Added ditherOrdered8To1Pr, which does ordered dithering from 8 bit/pixel
 * gray to binary.
 *
 * Revision 1.5  1995/02/01  23:19:29  macleod
 * macleod on Wed Feb 1 15:18:12 PST 1995
 *
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Added 4 bit/pixel ordered dither; removed ditherOrderedOneChannelPr;
 * renamed ditherOrdered24To8MakeTables to ditherOrderedMakeTables,
 * since the same call is used for 24->4 and 24->8 ordered dither.
 * Added xStart to all calls that only took yStart.
 *
 * Revision 1.4  1995/01/24  01:45:38  macleod
 * macleod on Mon Jan 23 17:44:43 PST 1995
 *
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Added ditherOrderedOneChannelPr and ditherRescale24To8Pr, which are
 * wrappers for routines in ipshared. Also changed the parameters
 * of ditherOrdered24To8Pr to NOT take the "reds", "greens", and "blues"
 * parameters, since they were never used.
 *
 * Revision 1.3  1995/01/17  21:33:23  macleod
 * macleod on Tue Jan 17 13:33:22 PST 1995
 *
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * Added "yStart" parameter to ordered dither. This parameter does not alter
 * what gets processed, but does change the line on which we start cycling
 * though the dither matrix. This is used when one is processing in bands
 * which aren't a multiple of 16 in height, allowing one to avoid
 * discontinuities in the dither pattern that would arise if one started
 * at zero. If one is not processing in bands, pass zero for "yStart."
 *
 * Revision 1.2  1995/01/17  18:07:13  macleod
 * macleod on Tue Jan 17 10:06:56 PST 1995
 *
 * odither.c -> /product/ipcore/ipwrap/src/RCS/odither.c,v
 *
 * fixed abortI in function which was returning void *
 *
 * Revision 1.1  1995/01/17  00:20:05  macleod
 * Initial revision
 *
*/
