/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   packc.c:
 *
 *	Code used to compress images using the PackBits algorithm.
 *	
 *		
 *	w_getSizeOfPackBitsCompressionInfo():	returns size of memory block
 *						needed for opaque state block
 *						used by PackBits compression.
 *
 *	w_initializePackBitsCompressionInfo():	initializes opaque PackBits
 *						state block
 *
 *	w_compressRowsPackBits():		compress one or more rows
 *						of image into PackBits format.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "pack.pub"
#include "shrcmprs.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: packc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/* This file contains wrappers for routines used to compress images
 * into PackBits bytestreams.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * To avoid interactions with client memory allocators, we allow the
 * client to allocate and IPW_FREE all memory structures.  The routine
 * getSizeOfPackBitsCompressionInfo is used by the client to get the
 * size of the opaque state block used in PackBits compression.
 *
 * The routine initializePackBitsCompressionInfo is used to
 * initialize this state block.  It sets up the thing that you wouldn't
 * want to calculate on every compression call.
 *
 * The routine compressRowsPackBits does the real work.  It returns
 * when either it fills the compressed data buffer or has done the right
 * number of lines.  The client tells it whether the image information
 * is new, the compressed data buffer is new and whether the compressed data
 * should be viewed as an extension of the previous block of compressed
 * data.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Returns size of private PackBits compression state block.
 *
 * LongDescription:::
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform PackBits
 *	compression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block
 *  w::              Width of image not including frame.
 *  d::              Depth (bits/pixel) of image.
 *
 * Output:::
 *  pSize::          Pointer to returned size in bytes.
 *
 * Notes:::
 *  PackBits compression uses a private state block that is allocated either
 *  locally or by the client.  This procedure returns the size of this
 *  block.
 *<BR><BR>
 *  It would have been more convenient to pass the PIXR to be
 *  compressed rather than the PIXR's width and height explicitly,
 *  but there's no way to do that in the Leffler TIFF code.
 *  Use pixrGetWidth and pixrGetDepth to extract these from
 *  a PIXR.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how PackBits really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfPackBitsCompressionInfo (
    IPWState           *pIPWState,
    UInt32              w,
    UInt32              d,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN( pSize == NULL, cErrNullPtr);

    ERR_PASS( i_getSizeOfPackBitsCompressionInfo (pIPWState,
						  w + 2*cFrameBits, d,
						  (Int32 *) pSize));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize PackBits compression state block.
 *
 * LongDescription:::
 *  This procedure fills in a previously allocated state block with
 *  the information required for PackBits compression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pPBCompInfoIn::  Pointer to private PackBits state block.
 *  w::              Width of image to be compressed not including frame.
 *  d::              Image's bits/pixel.
 *  compInfoBufSize:: Size of the private buffer allocated in bytes.
 *
 * Returns:::
 *  In addition to the normal codes, this routine returns cErrInvalidParm if
 *  the private structure's size (compInfoBufSize) is incorrect.
 *
 * Notes:::
 *  PackBits compression uses a private state block that is allocated either
 *  locally or by the client.  This procedure fills in the contents of this
 *  block.
 *<BR><BR>
 *  It would have been more convenient to pass the PIXR to be
 *  compressed rather than the PIXR's width and height explicitly,
 *  but there's no way to do that in the Leffler TIFF code.
 *  Use pixrGetWidth and pixrGetDepth to extract these from
 *  a PIXR.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how PackBits really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initializePackBitsCompressionInfo (
    IPWState           *pIPWState,
    void               *pPBCompInfoIn,
    UInt32              w,
    UInt32              d,
    UInt32              compInfoBufSize)
{
    PROLOGUE;

    ERR_GEN( pPBCompInfoIn == NULL, cErrNullPtr);

    ERR_PASS( i_initializePackBitsCompressionInfo (pIPWState, pPBCompInfoIn,
						   w + 2*cFrameBits, d,
						   compInfoBufSize));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  PackBits compress rows of an image.
 *
 * LongDescription:::
 *  Compress some rows of an image using the PackBits algorithm.  We will
 *  stop when either an error is encountered, we run out of image rows
 *  or we fill the compressed data buffer.
 *
 * Input:::
 *  pIPWState::             Pointer to standard IPWrap state block
 *  pPBCompInfoIn::         Pointer to private PackBits compression state block
 *  prS::                   Pointer to image to be compressed.
 *  startLine::             First image line to be compressed
 *			    relative to the top of the image, not frame.
 *  numLines::              Number of lines to be compressed.
 *  pCompData::             Pointer to first vacant byte of buffer
 *			    into which compressed data will be
 *			    written.  If several calls
 *			    to this routine are needed to create
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of vacant bytes remaining in
 *			    the compressed data buffer before we begin.
 *  continueImageBlock::    cTrue => we are compressing
 *			    from the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pPBCompInfoIn is
 *			    valid.<BR>
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *  continueCompDataStream:: cTrue => The saved compression state
 *			    in pPBCompInfoIn is valid.  If
 *			    the previous call to this routine
 *			    filled the compressed data buffer
 *			    before the strip, tile or image was
 *			    complete, a new buffer should be
 *			    supplied.  This arg is set to 
 *			    cFalse when the block of compressed
 *			    data is unrelated to the block used
 *			    in the previous call to this routine
 *			    as recorded in pPBCompInfoIn.
 *			    SETTING THIS ARG TO cFalse CAUSES
 *			    DATA IN THE PIPELINE TO BE DISCARDED!
 *  continueLineOffsets::   We compute offsets into the
 *			    compressed data that point to the
 *			    beginning of each scan line.  When
 *			    this arg is true, next offset should
 *			    be computed from the previous one,
 *			    i.e. the current block of compressed
 *			    data is an extension of the block
 *			    produced by the previous call.
 *			    When this is cFalse, the array of
 *			    offsets is restarted from 0.
 *  flushPipeline::         If this arg is set, this routine
 *			    will flush all the compressed data
 *			    in the pipeline to the compressed
 *			    data buffer when the specified
 *			    number of scan lines have been
 *			    compressed.  The client should not
 *			    set flushPipeline if compressed
 *			    data from subsequent scan lines is
 *			    to be seamlessly concatenated to
 *			    the current compressed data stream.
 *			    When flushPipeline is set and the
 *			    compression task suceeds in
 *			    processing all image lines, all
 *			    saved state is marked invalid.
 *			    Subsequent calls with either
 *			    continueImageBlock or
 *			    continueCompDataStream will return
 *			    cErrInvalidParm.
 *
 * Output:::
 *  pExhaustedImageData::   Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it filled
 *			    the compressed data buffer.  This
 *			    routine may be called again with
 *			    a new buffer.
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been compressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesWritten::         Number of compressed bytes actually
 *			    written to the compressed data
 *			    block.  If flushPipeline is false,
 *			    some bytes required to complete
 *			    a scan line may still be in the
 *			    pipeline and are not included in
 *			    this count.
 *  lineArray::             Array of offsets to beginnings
 *			    of compresses scan lines in compressed
 *			    data block.  After line n has been
 *			    successfully compressed,
 *			    lineArray[n+1] = lineArray[n] +
 *			    {number of compressed bytes used
 *			    to represent line n}.  This prepares
 *			    us to compress line n+1.  The client
 *			    should allocated a array that's one
 *			    line too long to account for the
 *			    extra write after the last scan line.
 *			    This pointer is set to NULL if no
 *			    array of offsets is desired.
 *
 * Returns:::
 *  Normally, this routine thinks that running out of the compressed data
 *  buffer is ok.  However, if flushPipeline is set to cTrue and we overflow
 *  the compressed data buffer, the routine returns cErrBufferExceeded.
 *
 * Notes:::
 *  This routine compresses a horizontal slice of a PackBits image.  There
 *  are several completion conditions:
 *
 *	retCode	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we compressed the desired number of
 *		image lines without filling the compressed data buffer.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse.  If flushPipeline
 *		was set, all compressed data has been written to
 *		the compressed data buffer and the saved compressed state
 *		is marked invalid.  If flushPipeline was not set,
 *		some of the compressed data may still be in the
 *		pipeline and the saved compressed data state is valid.
 *		In the latter case, this routine may be called again with
 *		continueCompDataStream = cTrue.  In either case, the
 *		client must supply a new value for pCompData and
 *		compDataSize.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we filled the compressed data buffer at the
 *		same time we completed the correct number of image lines.
 *		This is unusual, but not impossible.  If flushPipeline
 *		was set, all compressed data has been written to the
 *		compressed data buffer and the saved state is marked
 *		invalid.  Otherwise, some data may remain in the pipeline.
 *		This return state cannot occur of flushPipeline was set
 *		and the pipeline still contains data (see next state).
 *		If flushPipeline was set, it is not possible to continue
 *		from this state as all saved state is invalid.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine filled the compression data buffer before completing
 *		the specified number of lines.  One way this can
 *		happen is if all image data has been processed, but
 *		flushPipeline was set and some data remains in the pipeline.
 *		When this condition does occur, the routine may be
 *		restarted by supplying a new compressed data buffer and
 *		setting	BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.<BR> 
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.
 *		When this condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much buffer space is 
 *		needed to complete the slice.
 *<BR><BR>		
 *		The image data is always in PIXR order.  This routine
 *		will byteswap strings of literals in the compressed data
 *		on the way to the image if necessary.  The byte order
 *		of the compressed data was specified in the call to
 *		i_initializePackBitsCompressionInfo that initialized
 *		the opaque compression info pointed to by pPBCompInfoIn.
 *		
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, Packbits
 *
 * Level:::
 *  deterministic, though the method is not obvious.
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how PackBits really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_compressRowsPackBits (
    IPWState           *pIPWState,
    void               *pPBCompInfoIn,
    PIXR               *prS,
    UInt32              startLine,
    UInt32              numLines,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32              continueImageBlock,
    UInt32              continueCompDataStream,
    UInt32              continueLineOffsets,
    UInt32              flushPipeline,
    UInt32             *pExhaustedImageData,
    UInt32             *pExhaustedCompData,
    UInt32             *pPercentDone,
    UInt32             *pBytesWritten,
    UInt32             *lineArray)
{
    PROLOGUE;
    Int32    d, w, bpl;
    UInt8   *pS;

    ERR_GEN( pPBCompInfoIn == NULL || prS == NULL || pCompData == NULL ||
	     pExhaustedImageData == NULL || pExhaustedCompData == NULL ||
	     pPercentDone == NULL || pBytesWritten == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

    ERR_PASS( i_compressRowsPackBits (pIPWState, pPBCompInfoIn,
				      pS,
				      w,
				      startLine + cFrameBits,
				      numLines,
				      d,
				      bpl,
				      pCompData,
				      compDataBufSize,
				      continueImageBlock,
				      continueCompDataStream,
				      continueLineOffsets,
				      flushPipeline,
				      pExhaustedImageData,
				      pExhaustedCompData,
				      (Int32 *)pPercentDone,
				      (Int32 *)pBytesWritten,
				      (Int32 *)lineArray));

    NORMAL_RETURN;
}


/*

 $Log: packc.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.18  1998/01/14 00:14:14  ddavies
 packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.17  1996/01/04 20:25:14  ejaquith
 packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.16  1995/12/13  04:10:58  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.15  1995/11/13  20:14:13  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.14  1995/10/13  22:03:30  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.13  1995/10/05  04:26:48  ddavies
 * ddavies on Wed Oct 4 21:26:47 PDT 1995
 *
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.12  1995/09/28  22:44:02  ddavies
 * ddavies on Thu Sep 28 15:43:38 PDT 1995
 *
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * Change PackBits keyword to Packbits.
 *
 * Revision 1.11  1995/09/28  22:22:12  ddavies
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Fix comments at top of file.
 *
 * Revision 1.10  1995/09/28  19:19:36  ddavies
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Filled in autodoc comments.
 *
 * Revision 1.9  1995/09/21  22:55:01  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.8  1995/09/15  18:00:08  ddavies
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.7  1995/09/11  22:10:50  ddavies
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.6  1995/09/06  20:21:49  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.5  1995/09/01  01:17:44  ddavies
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.4  1995/08/31  21:15:36  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.3  1995/08/17  22:04:35  ejaquith
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.2  1995/07/29  04:46:05  ddavies
 * packc.c -> /product/ipcore/ipwrap/src/RCS/packc.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.1  1995/05/09  03:14:16  ddavies
 * Initial revision
 *

 */


