/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   packd.c:
 *
 *	Code used to decompress PackBits images.
 *	
 *		
 *	w_getSizeOfPackBitsDecompressionInfo():	returns size of memory block
 *						needed for opaque state block
 *						used by PackBits decompression.
 *
 *	w_initializePackBitsDecompressionInfo(): initializes opaque PackBits
 *						state block
 *
 *	w_decompressRowsPackBits():		decompress one or more rows
 *						of image from PackBits.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "pack.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: packd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/* This file contains wrappers for routines used to decompress PackBits
 * streams into images.
 *
 * The compressed data stream is always in the natural byte order of the
 * current platform.
 *
 * To avoid interactions with client memory allocators, we allow the
 * client to allocate and IPW_FREE all memory structures.  The routine
 * getSizeOfPackBitsDecompressionInfo is used by the client to get the
 * size of the opaque state block used in PackBits decompression.
 *
 * The routine initializePackBitsDecompressionInfo is used to
 * initialize this state block.  It sets up the thing that you wouldn't
 * want to calculate on every decompression call.
 *
 * The routine decompressRowsPackBits does the real work.  It returns
 * when either it runs out of compressed data or has done the right
 * number of lines.  The client tells it whether the image information
 * is new, the compressed data is new and whether the compressed data
 * should be viewed as an extension of the previous block of compressed
 * data.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Returns size of private PackBits decompression state block.
 *
 * LongDescription:::
 *	This procedure returns the size of the memory block that will
 *	be needed to hold the state information used to perform PackBits
 *	decompression.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block
 *
 * Output:::
 *  pSize::          Pointer to returned size in bytes.
 *
 * Notes:::
 *  PackBits decompression uses a private state block that is allocated either
 *  locally or by the client.  This procedure returns the size of this
 *  block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how PackBits really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfPackBitsDecompressionInfo (
    IPWState           *pIPWState,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN( pSize == NULL, cErrNullPtr);

    ERR_PASS( i_getSizeOfPackBitsDecompressionInfo (pIPWState, (Int32 *)pSize));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize PackBits decompression state block.
 *
 * LongDescription:::
 *  This procedure fills in a previously allocated state block with
 *  the information required for PackBits decompression.
 *
 * Input:::
 *  pIPWState::         Pointer to standard IPWrap state block.
 *  pPBDecompInfoIn::   Pointer to private PackBits state block.
 *  decompInfoBufSize:: Size of the private buffer allocated in bytes.
 *
 * Returns:::
 *  In addition to the normal codes, this routine returns cErrInvalidParm if
 *  the private structure's size (compInfoBufSize) is incorrect.
 *
 * Notes:::
 *  PackBits decompression uses a private state block that is allocated either
 *  locally or by the client.  This procedure fills in the contents of this
 *  block.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how PackBits really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initializePackBitsDecompressionInfo (
    IPWState           *pIPWState,
    void               *pPBDecompInfoIn,
    UInt32              decompInfoBufSize)
{
    PROLOGUE;

    ERR_GEN( pPBDecompInfoIn == NULL, cErrNullPtr);

    ERR_PASS( i_initializePackBitsDecompressionInfo (pIPWState, pPBDecompInfoIn,
						     decompInfoBufSize));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress one horizontal slice of a PackBits file.
 *
 * LongDescription:::
 *  Decompress some rows of an image using the PackBits algorithm.  We will
 *  stop when either an error is encountered, we run out of image rows
 *  or we empty the compressed data buffer.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::             Pointer to standard IPWrap state block.
 *  pPBDecompInfoIn::       Pointer to PackBits decompression info.
 *  startLine::             First image line to be decompressed
 *			    relative to the top of the image, not frame.
 *  numLines::              Number of lines to be decompressed.
 *  pCompData::             Pointer to next byte of compressed
 *			    data to be read.  If several calls
 *			    to this routine are needed to process
 *			    a block of compressed data, this
 *			    pointer must be updated on each call.
 *  compDataBufSize::       Number of bytes of compressed data
 *			    remaining in the buffer.
 *  continueImageBlock::    cTrue => we are decompressing
 *			    into the same slice of image as was
 *			    specified in the previous call and the
 *			    saved state in pPBDecompInfoIn is
 *			    valid.
 *			    This arg is set to cFalse if either
 *			    we are starting with all new data
 *			    or the previous call returned because it
 *			    ran out of image data.
 *  continueCompDataStream:: cTrue => The saved decompression state
 *			    in pPBDecompInfoIn is valid.  This
 *			    allows the procedure to treat several
 *			    buffers of compressed data as though
 *			    it were one contiguous buffer.
 *			    If set to cFalse, this routine will
 *			    discard all previous compression
 *			    state.
 *  flushPipeline::         If this arg is set, this routine
 *			    will attempt to complete the
 *			    specified number of scan lines
 *			    WITHOUT stopping when it starts to
 *			    run out of data; i.e. it continues
 *			    until its pipeline is completely
 *			    empty.  If this is not possible,
 *			    an error is raised.
 *			    When flushPipeline is set and the
 *			    decompression task suceeds in
 *			    processing all image lines, all
 *			    saved state is marked invalid.
 *			    Subsequent calls with either
 *			    continueImageBlock or
 *			    continueCompDataStream set will return
 *			    cErrInvalidParm.
 *
 * Output:::
 *  prD::                   Pointer to pixr that will contain
 *			    uncompressed image
 *  pExhaustedImageData::   Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because specified
 *			    number of lines have been processed.
 *			    This may require more than one
 *			    call to this routine.
 *  pExhaustedCompData::    Pointer to returned arg.  When
 *			    arg is returned with value cTrue,
 *			    routine stopped because it reached
 *			    the end of the compressed data buffer
 *			    as specified by compDataSize.  This
 *			    routine may be called again with 
 *			    new compression data and the the 
 *			    same image.  This arg will not be 
 *			    set if flushPipeline is set and we 
 *			    failed to decompress the specified 
 *			    number of scan lines (routine returns
 *			    cErrBufferExceeded instead).
 *  pPercentDone::          Pointer to returned arg.  When
 *			    routine returns with *pExhaustedCompData
 *			    set, this arg specifies the rough
 *			    percentage of the image block that
 *			    has been decompressed.  It is
 *			    intended to be used to estimate the
 *			    amount of compressed data needed
 *			    to complete the task.  The result
 *			    ranges from 0 to 100.
 *  pBytesRead::            Pointer to returned arg.  Contains
 *			    number of compressed data bytes
 *			    read by the routine before it returned.
 *			    The client may safely dispose of
 *			    these bytes.
 *
 * Returns:::
 *  Normally, this routine thinks that emptying the compressed data
 *  buffer is ok.  However, if flushPipeline is set to cTrue and we empty
 *  the compressed data buffer, the routine returns cErrBufferExceeded.
 *
 * Notes:::
 *  This function decompresses a horizontal slice of a PackBits file.
 *  There are several termination conditions:
 *
 *	retCode	!= cSuccessful::
 *	*pExhaustedImageData == don't care::
 *	*pExhaustedCompData == don't care::
 *		This means something catastrophic happened.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cFalse::
 *		This means we produced the desired number of uncompressed
 *		image lines without running out of compressed data.
 *		The routine can be restarted with new startLine (and
 *		numLines), continueImageBlock = cFalse and
 *		continueCompDataStream = cTrue.  pCompData and compDataSize
 *		should be adjusted to account for the number of 
 *		compressed data bytes that have already been read (returned
 *		in *pBytesRead).
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cTrue::
 *	*pExhaustedCompData == cTrue::
 *		This means we ran out of compressed data at the same time
 *		we completed the correct number of image lines.  This
 *		is the normal condition at the bottom of a TIFF slice or
 *		the bottom of the image.  It is not possible to restart
 *		from this state, though it is possible to decompress
 *		with new image and compression data.  When the client
 *		expects to run out of compressed data, the flushPipeline
 *		argument should be set.
 *
 *	retCode	== cSuccessful::
 *	*pExhaustedImageData == cFalse::
 *	*pExhaustedCompData == cTrue::
 *		The routine ran out of compression data before completing
 *		the specified number of lines.  This condition will not
 *		occur if flushPipeline was set.  If flushPipeline was set
 *		and we ran out of compressed data before completing the
 *		the task, an cErrBufferExceeded error will be returned.
 *		When this condition does occur, the routine may be
 *		restarted by supplying more compressed data, setting
 *		BOTH continueImageBlock and continueCompDataStream to
 *		cTrue.  Note that the first byte of the new compressed
 *		data must be the one immediately following the last
 *		byte of old compressed data.
 *		The image may be moved around as a block if necessary, but
 *		the internal composition of the image (and startLine
 *		and numLines) should not be changed.  When this
 *		condition occurs, *pPercentDone indicates the rough 
 *		percentage of the slice that's been completed.  This 
 *		can be used to estimate how much compression data is 
 *		needed to complete the slice.
 *<BR><BR>
 *		The image data is always in PIXR order.  This routine
 *		will byteswap strings of literals in the compressed data
 *		on the way to the image if necessary.
 *		
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, run length, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  TIFF Revision 6.0; Leffler's TIFF code to see how PackBits really works.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_decompressRowsPackBits (
    IPWState           *pIPWState,
    void               *pPBDecompInfoIn,
    UInt32              startLine,
    UInt32              numLines,
    UInt8              *pCompData,
    UInt32              compDataBufSize,
    UInt32              continueImageBlock,
    UInt32              continueCompDataStream,
    UInt32              flushPipeline,
    PIXR               *prD,
    UInt32             *pExhaustedImageData,
    UInt32             *pExhaustedCompData,
    UInt32             *pPercentDone,
    UInt32             *pBytesRead)
{
    PROLOGUE;
    Int32    d, w, bpl;
    UInt8   *pD;

    ERR_GEN( pPBDecompInfoIn == NULL || prD == NULL || pCompData == NULL ||
	     pExhaustedImageData == NULL || pExhaustedCompData == NULL ||
	     pPercentDone == NULL || pBytesRead == NULL, cErrNullPtr);

    /* Note we are sometimes expecting a zero frame pixr! */
    ERR_GEN_PIXR(prD, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame0 | cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prD, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prD, &d));
    ERR_PASS( w_pixrGetBpl(pIPWState, prD, &bpl));

    ERR_PASS( i_decompressRowsPackBits (pIPWState, pPBDecompInfoIn,
				        pD,
				        w,
				        startLine + cFrameBits,
				        numLines,
				        d,
				        bpl,
				        pCompData,
				        compDataBufSize,
				        continueImageBlock,
				        continueCompDataStream,
				        flushPipeline,
				        pExhaustedImageData,
				        pExhaustedCompData,
				        (Int32 *)pPercentDone,
				        (Int32 *)pBytesRead));

    NORMAL_RETURN;
}


/*

 $Log: packd.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.20  1998/01/14 00:14:16  ddavies
 packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.19  1996/01/04 20:25:17  ejaquith
 packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.18  1995/12/13  04:11:02  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:14:16  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  22:03:33  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/10/05  04:26:52  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.14  1995/09/28  22:22:18  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Fix comments at top of file.
 *
 * Revision 1.13  1995/09/28  19:19:40  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Filled in autodoc comments.
 *
 * Revision 1.12  1995/09/21  22:55:04  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.11  1995/09/15  18:00:12  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.10  1995/09/13  00:22:02  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * Needed to check for cFrame0 and cFrame32 in decompression routine.
 *
 * Revision 1.9  1995/09/11  22:10:53  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.8  1995/09/10  00:32:56  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.7  1995/09/07  01:33:47  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.6  1995/09/06  20:21:52  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.5  1995/09/01  01:17:49  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.4  1995/08/31  21:15:38  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.3  1995/08/17  22:04:39  ejaquith
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.2  1995/07/29  04:46:40  ddavies
 * packd.c -> /product/ipcore/ipwrap/src/RCS/packd.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.1  1995/05/09  03:15:32  ddavies
 * Initial revision
 *

 */

