/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* 
 * page.c
 *
 *     Utilities for dealing with "Container" and associated structures.
 *     The idea behind this structure is to represent composite pages 
 *     as a possibly hierarchical set of containers, each of which defines
 *     a rectangular bounding box, a type, some data, and a mask.
 *     The data can be an image, a tint box, or another container (though
 *     I don't anticipate making multi-level hierarchies, at least initially).
 *     
 * Functions Defined:
 *
 * Creating Containers:
 *    newPage
 *    newImage
 *    newNullImage
 *    newBox
 *    newSandwich
 *
 * Sticking containers into other containers:
 *    addContainer 
 *
 * Container accessors:
 *    containerGetType
 *    containerGetPixr
 *    containerGetTint 
 *    containerGetMask 
 *    containerGetForeground
 *    containerGetBackground
 *    containerGetBounds 
 *    containerGetNext 
 *    containerGetChild
 *    containerGetChildCount
 *
 * Container Mutators:
 *    containerSetType
 *    containerSetMask
 *    containerSetPixr
 *    containerSetBounds
 *    containerSetForeground
 *    containerSetBackground
 *    containerSetTint
 *    containerSetNext
 *    containerSetChild
 *
 * Tint accessors/mutators:
 *    tintAllocate
 *    tintGetColorCount
 *    tintGetColor
 *    tintSetColor
 *
 * Destruction:
 *    destroyContainer
 *    destroyTint
 *
 * Generic creation:
 *    allocateContainer
 */

#include <stdio.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "defines.pub"
#include "frames.pub"
#include "appsinc.h"
#include "logicops.h"

/* prototypes */
#include "props.pub"
#include "page.pub"
#include "logicops.pub"
#include "intops.pub"
#include "grbinop.pub"
#include "rescale.pub"
#include "reduce.pub"
#include "prxform.pub"
#include "iw_scale.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: page.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Creates an empty "page" container
 *
 * LongDescription:::
 *    Creates an empty "page" container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xSize::          Width, in micas (1/2540 in = 1 mica) of page container
 *  ySize::          Height, in micas, of page container
 *
 * Output:::
 *  ppPageCreated::  The container created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newPage (
    IPWState           *pIPWState,
    Int32               xSize,
    Int32               ySize,
    Container         **ppPageCreated)
{
    PROLOGUE;
    Container *c;
    
    /* allocateContainer sets the container "child" and "sibling"
       field to NULL already */
    ERR_PASS (w_allocateContainer (pIPWState, &c));
    ERR_PASS (w_containerSetType (pIPWState, cContainerPage, c));
    ERR_PASS (w_containerSetBounds (pIPWState, 0, 0, xSize, ySize, c));
    ERR_PASS (w_containerSetMask (pIPWState, NULL, c));

    *ppPageCreated = c;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Creates an "image" container.
 *
 * LongDescription:::
 *    Creates an "image" container.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xOffset::        X-Offset at which image is placed, in micas
 *  yOffset::        Y-Offset at which image is placed, in micas
 *  pImage::         The image to place
 *  pMask::          A mask for the image, NULL if none
 *  type::           type of image, either cContainerText,
 *                   cContainerContone, or cContainerLineArt
 *
 * Output:::
 *  ppNewImage::     Image container created
 *
 * Returns:::
 *  error status
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newImage (
    IPWState           *pIPWState,
    Int32               xOffset,
    Int32               yOffset,
    PIXR               *pImage,
    PIXR               *pMask,
    ContainerType       type,
    Container         **ppNewImage)
{
    PROLOGUE;
    Container *c;
    Int32 xSize, ySize, width, height;
    ScaledInt1616 xRes = 0, yRes = 0, maxRes;
    
    ERR_GEN_PIXR (pImage, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);
    
    /* allocateContainer sets the container "child" and "sibling"
       field to NULL already */
    ERR_PASS (w_allocateContainer (pIPWState, &c));

    ERR_PASS (w_containerSetType (pIPWState, type, c));
    ERR_PASS (w_containerSetMask (pIPWState, pMask, c));
    
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, pImage, &xRes, &yRes));
    maxRes = MAX (xRes, yRes);

    /* Compute the size, in micas */
    ERR_PASS (w_pixrGetWidth (pIPWState, pImage, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, pImage, &height));
    xSize = PIXELS_TO_MICAS(width, maxRes);
    ySize = PIXELS_TO_MICAS(height, maxRes);
    
    /* Set the size in the container */
    ERR_PASS (w_containerSetBounds (pIPWState, xOffset, yOffset, 
				    xSize, ySize, c));
    
    /* Make it point to that image */
    ERR_PASS (w_containerSetPixr (pIPWState, pImage, c));
    
    *ppNewImage = c;

    NORMAL_RETURN;
    
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Creates an " Null image" container.
 *
 * LongDescription:::
 *    Creates an "Null image" container for foreground or background
 *  image when we suppose foreground is black or background is white.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xOffset::        X-Offset at which image is placed, in micas
 *  yOffset::        Y-Offset at which image is placed, in micas
 *  pImage::         The null image to place
 *  pMask::          A mask for the image, NULL if none
 *  type::           type of image, either cContainerText,
 *                   cContainerContone, or cContainerLineArt
 *
 * Output:::
 *  ppNewImage::     Image container created
 *
 * Returns:::
 *  error status
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  x. zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newNullImage (
    IPWState           *pIPWState,
    Int32               xOffset,
    Int32               yOffset,
    PIXR               *pImage,
    PIXR               *pMask,
    ContainerType       type,
    Container         **ppNewImage)
{
    PROLOGUE;
    Container *c;
    
    ERR_GEN (pImage != NULL, cErrNullPtr);
    
    /* allocateContainer sets the container "child" and "sibling"
       field to NULL already */
    ERR_PASS (w_allocateContainer (pIPWState, &c));

    ERR_PASS (w_containerSetType (pIPWState, type, c));
    ERR_PASS (w_containerSetMask (pIPWState, pMask, c));
    
    /* Set the size in the container */
    ERR_PASS (w_containerSetBounds (pIPWState, xOffset, yOffset, 
				    0, 0, c));
    
    /* Make it point to that image */
    ERR_PASS (w_containerSetPixr (pIPWState, pImage, c));
    
    *ppNewImage = c;

    NORMAL_RETURN;
    
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Creates an "image sandwich" container.
 *
 * LongDescription:::
 *    Creates an "image sandwich" container. An "image sandwich" consists
 *    of two images (represented as containers) and a mask, which is
 *    used to toggle between them. Where there is a "1" in the mask,
 *    the foreground image is painted, and where there is a "0" in the
 *    mask, the background image is painted. The mask and images do not
 *    need to be held at the same resolution, and usually won't be.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xOffset::        X-Offset at which image is placed, in micas
 *  yOffset::        Y-Offset at which image is placed, in micas
 *  pMask::          Sandwich's mask
 *  pForeground::    Foreground container, i.e. the image (or possibly
 *                   a composite container) that represents the foreground
 *                   of the image, where corresponding pixels in the mask
 *                   are 1.
 *  pBackground::    Background container, i.e. the image (or possibly
 *                   a composite container) that represents the background
 *                   of the image, where corresponding pixels in the mask
 *                   are 0.
 *
 * Output:::
 *  ppNewSandwich::  Image sandwich container created
 *
 * Returns:::
 *  error status
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newSandwich (
    IPWState           *pIPWState,
    Int32               xOffset,
    Int32               yOffset,
    PIXR               *pMask,
    Container          *pForeground,
    Container          *pBackground,
    Container         **ppNewSandwich)
{
    PROLOGUE;
    Container *c = NULL;
    Int32 xSize, ySize, width, height;
    ScaledInt1616 xRes = 0, yRes = 0, maxRes;
    
    ERR_GEN (!pMask, cErrNullPtr);

    /* allocateContainer sets the container "child" and "sibling"
       field to NULL already */
    ERR_PASS (w_allocateContainer (pIPWState, &c));

    ERR_PASS_CLEANUP (w_containerSetType (pIPWState, cContainerSandwich, c));
    ERR_PASS_CLEANUP (w_containerSetMask (pIPWState, pMask, c));
    
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, pMask, &xRes, &yRes));
    maxRes = MAX (xRes, yRes);

    /* Compute the size, in micas */
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pMask, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pMask, &height));
    xSize = PIXELS_TO_MICAS(width, maxRes);
    ySize = PIXELS_TO_MICAS(height, maxRes);
    
    /* Set the size in the container */
    ERR_PASS_CLEANUP (w_containerSetBounds (pIPWState, xOffset, yOffset, 
				    xSize, ySize, c));
    
    /* Set the fg/bg containers */
    ERR_PASS_CLEANUP (w_containerSetForeground (pIPWState, pForeground, c));
    ERR_PASS_CLEANUP (w_containerSetBackground (pIPWState, pBackground, c));

    *ppNewSandwich = c;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (c)
	w_destroyContainer (pIPWState, cFalse, &c);

    ERR_RETURN;

    
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Create a tint region
 *
 * LongDescription:::
 *    Create a new tint region, which is a rectangle and a mesh of 
 *    color values which are to be interpolated over the region when
 *    attempting to render it.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              X-offset of tint, in micas
 *  y::              Y-offset of tint, in micas
 *  w::              width of tint, in micas
 *  h::              height of tint, in micas
 *  type::           ColorType type of all of the colors on the mesh
 *  xColors::        # of colors in x on the mesh. 1,1 means a tint
 *  yColors::        # of colors in y on the mesh. 1,1 means a tint
 *  colorPoints::    Array of color points, e.g. vector of x*y points,
 *                   laid out as a matrix.
 *
 * Output:::
 *  ppNewTint::      Tint container created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newTint (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ColorType           type,
    Int32               xColors,
    Int32               yColors,
    ColorRecord        *colorPoints,
    Container         **ppNewTint)
{
    PROLOGUE;
    Container *c;
    TintMesh  *tint;
    Int32      i, j;
    

    /* allocateContainer sets the container "child" and "sibling"
       field to NULL already */
    ERR_PASS (w_allocateContainer (pIPWState, &c));
    type = type;
    ERR_PASS (w_containerSetType (pIPWState, cContainerTint, c));
    ERR_PASS (w_containerSetMask (pIPWState, NULL, c));
    ERR_PASS (w_containerSetBounds (pIPWState, x, y, w, h, c));

    /* Allocate a tint */
    ERR_PASS (w_tintAllocate (pIPWState, xColors, yColors, &tint));

    /* Copy the points */
    for (i = 0; i < yColors; i++)
    {
	for (j = 0; j < xColors; j++)
	{
	    ERR_PASS (w_tintSetColor (pIPWState, x, y, 
				      colorPoints[i * yColors + j], tint));
	}
    }

    ERR_PASS (w_containerSetTint (pIPWState, tint, c));
    *ppNewTint = c;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Create a new "tint box", which is a rectangle containing a tint
 *
 * LongDescription:::
 *    Create a new "tint box", which is a rectangle containing a tint
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              X-offset of box, in micas
 *  y::              Y-offset of box, in micas
 *  w::              Width of box, in micas
 *  h::              Height of box, in micas
 *  color::          Color of tint
 *
 * Output:::
 *  ppNewTint::      Container created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page laying
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_newTintBox (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    ColorRecord         color,
    Container         **ppNewTint)
{
    PROLOGUE;
    Container *c;
    TintMesh *tint;

    /* allocateContainer sets the container "child" and "sibling"
       field to NULL already */
    ERR_PASS (w_allocateContainer (pIPWState, &c));
    
    ERR_PASS (w_containerSetType (pIPWState, cContainerTint, c));
    ERR_PASS (w_containerSetMask (pIPWState, NULL, c));
    ERR_PASS (w_containerSetBounds (pIPWState, x, y, w, h, c));
    
    /* Allocate a tint of size (1,1) */
    ERR_PASS (w_tintAllocate (pIPWState, 1, 1, &tint));
    ERR_PASS (w_tintSetColor (pIPWState, 0, 0, color, tint));
    ERR_PASS (w_containerSetTint (pIPWState, tint, c));
    
    *ppNewTint = c;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Add a child container to the child list of this container.
 *
 * LongDescription:::
 *    Add a child container to the child list of this container. The
 *    container must be composite (the "page" container is the only
 *    composite container supported at the moment)
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  parent::         Parent container
 *  child::          Child container
 *  topFlag::        If cTrue, put the child container at the tail
 *                   of the list. If cFalse, put the child container
 *                   at the head of the list.
 *
 * Output:::
 * 
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addContainer (
    IPWState           *pIPWState,
    Container          *parent,
    Container          *child,
    Bool                topFlag)
{
    PROLOGUE;
    Container *p, *n;
    Int32      count;
    ContainerType type;
    
    
    ERR_PASS (w_containerGetType (pIPWState, parent, &type));
    ERR_GEN (type != cContainerPage, cErrInvalidParm);
   
    ERR_PASS (w_containerGetChildCount (pIPWState, parent, &count));
    if (count == 0)
    {
	ERR_PASS (w_containerSetChild (pIPWState, child, parent));
    }
    else
    {
	ERR_PASS (w_containerGetChild (pIPWState, parent, 0, &p));
    
	if (topFlag == cTrue)
	{
	    ERR_PASS (w_containerGetNext (pIPWState, p, &n));
	    while (n != NULL)
	    {
		p = n;
		ERR_PASS (w_containerGetNext (pIPWState, p, &n));
	    }
	
	    ERR_PASS (w_containerSetNext (pIPWState, child, p));
	}
	else
	{
	    if (p)
		ERR_PASS (w_containerSetNext (pIPWState, p, child));
	    ERR_PASS (w_containerSetChild (pIPWState, child, parent));
	}
    }
    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the type of a container
 *
 * LongDescription:::
 *  Get the type of a container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  c::              Container to query
 *
 * Output:::
 *  pType::          Type of container
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetType (
    IPWState           *pIPWState,
    Container          *c,
    ContainerType      *pType)
{
    PROLOGUE;
    pIPWState = pIPWState;
    if (c)
	*pType = c->type;
    else
	*pType = cContainerUnknown;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Get the PIXR for a container
 *
 * LongDescription:::
 *   Get the PIXR for a container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  c::              Container to query
 *
 * Output:::
 *  pPrPixr::        PIXR for container
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetPixr (
    IPWState           *pIPWState,
    Container          *c,
    PIXR              **pPrPixr)
{
    PROLOGUE;
    ContainerType type;
    

    if (c)
    {
	ERR_PASS (w_containerGetType (pIPWState, c, &type));
	ERR_GEN (!((type == cContainerText || type == cContainerLineArt ||
	    type == cContainerContone)), cErrInvalidParm);
	*pPrPixr = c->child.childPixr;
	
    }
    else
	ERR_GEN (1, cErrInvalidParm);

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the "TintMesh" object for a tint box
 *
 * LongDescription:::
 *  Get the "TintMesh" object for a tint box
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  c::              Container to query
 *
 * Output:::
 *  pTint::          Tint mesh
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetTint (
    IPWState           *pIPWState,
    Container          *c,
    TintMesh          **pTint)
{
    PROLOGUE;
    ContainerType type;

    ERR_GEN (!c, cErrInvalidParm);
    ERR_PASS (w_containerGetType (pIPWState, c, &type));
    ERR_GEN (type != cContainerTint, cErrBadType);

    *pTint = c->child.childTint;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the mask for a container.
 *
 * LongDescription:::
 *  Get the mask for a container.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  c::              Container to query
 *
 * Output:::
 *  pPrMask::        The mask PIXR, or NULL if none
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetMask (
    IPWState           *pIPWState,
    Container          *c,
    PIXR              **pPrMask)
{
    PROLOGUE;
   
    ERR_GEN (!c, cErrInvalidParm);
    *pPrMask = c->mask;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the bounding rectangle for a container.
 *
 * LongDescription:::
 *  Get the bounding rectangle for a container. All coordinates are
 *  in micas.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  c::              Container to query
 *
 * Output:::
 *  x::              X-offset of bounding rectangle
 *  y::              Y-offset of bounding rectangle
 *  w::              Width of bounding rectangle
 *  h::              Height of bounding rectangle
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetBounds (
    IPWState           *pIPWState,
    Container          *c,
    Int32              *x,
    Int32              *y,
    Int32              *w,
    Int32              *h)
{
    PROLOGUE;

    ERR_GEN (!c, cErrInvalidParm);
    *x = c->x;
    *y = c->y;
    *w = c->w;
    *h = c->h;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get "foreground" container for a "sandwich" container
 *
 * LongDescription:::
 *  Get "foreground" container for a "sandwich" container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Stat structure
 *  c::              Container to query
 *
 * Output:::
 *  ppForeground::   Container's "foreground" thing
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetForeground (
    IPWState           *pIPWState,
    Container          *c,
    Container         **ppForeground)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    ERR_GEN (c->type != cContainerSandwich, cErrTypeNotSupported);

    *ppForeground = c->child.childSandwich.foregroundContainer;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get "background" container for a "sandwich" container
 *
 * LongDescription:::
 *  Get "background" container for a "sandwich" container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Stat structure
 *  c::              Container to query
 *
 * Output:::
 *  ppBackground::   Container's "background" thing
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetBackground (
    IPWState           *pIPWState,
    Container          *c,
    Container         **ppBackground)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    ERR_GEN (c->type != cContainerSandwich, cErrTypeNotSupported);

    *ppBackground = c->child.childSandwich.backgroundContainer;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the next container in the sibling list
 *
 * LongDescription:::
 *  Get the next container in the sibling list
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Stat structure
 *  c::              Container to query
 *
 * Output:::
 *  ppNext::         Container's sibling
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetNext (
    IPWState           *pIPWState,
    Container          *c,
    Container         **ppNext)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    *ppNext =  c->next;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the nth child for a composite container
 *
 * LongDescription:::
 *  Get the nth child for a composite container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  c::              Container to query
 *  n::              Which child we want
 *
 * Output:::
 *  ppChild::        The nth child
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetChild (
    IPWState           *pIPWState,
    Container          *c,
    Int32               n,
    Container         **ppChild)
{
    PROLOGUE;
    ContainerType type = cContainerUnknown;
    Int32 i = 0;
    Container *p = NULL;
    

    ERR_GEN (!c, cErrInvalidParm);
    ERR_PASS (w_containerGetType (pIPWState, c, &type));

    ERR_GEN (type != cContainerPage, cErrBadType);
    
    p = c->child.childContainer;
    for (i = 0; i < n; i++)
    {
	if (p)
	    p = p->next;
	else
	    break;
    }
	
    /* We don't generate an error when we've gone past the end */
    *ppChild = p;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the number of children in a container.
 *
 * LongDescription:::
 *  Get the number of children in a container.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure 
 *  c::              Container to query
 *
 * Output:::
 *  pCount::         Number of children
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerGetChildCount (
    IPWState           *pIPWState,
    Container          *c,
    Int32              *pCount)
{
    PROLOGUE;
    ContainerType type = cContainerUnknown;
    Int32 n = 0;
    Container *p = NULL;
    

    ERR_GEN (!c, cErrInvalidParm);
    ERR_PASS (w_containerGetType (pIPWState, c, &type));
    ERR_GEN (type != cContainerPage, cErrBadType);
    p = c->child.childContainer;
    for (n = 0; p; p = p->next)
	n++;
    *pCount = n;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the type of a container
 *
 * LongDescription:::
 *  Set the type of a container
 *  
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  t::              Type to use
 *
 * Output:::
 *  c::              Container with type set
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetType (
    IPWState           *pIPWState,
    ContainerType       t,
    Container          *c)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    c->type = t;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PIXR for a container
 *
 * LongDescription:::
 *  Set the PIXR for a container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pixr::           PIXR to use
 *
 * Output:::
 *  c::              Container for which to set PIXR
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetPixr (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Container          *c)
{
    PROLOGUE;
    ContainerType type;
    

    ERR_GEN (!c, cErrNullPtr);
    ERR_PASS (w_containerGetType (pIPWState, c, &type));

    ERR_GEN (type != cContainerText && 
	     type != cContainerLineArt &&
	     type != cContainerContone, 
	     cErrBadType);
    
    c->child.childPixr = pixr;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set "foreground" container for a "sandwich" container
 *
 * LongDescription:::
 *  Set "foreground" container for a "sandwich" container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Stat structure
 *  pForeground::    Foreground to set
 *  c::              Container to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetForeground (
    IPWState           *pIPWState,
    Container          *pForeground,
    Container          *c)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    ERR_GEN (c->type != cContainerSandwich, cErrTypeNotSupported);

    c->child.childSandwich.foregroundContainer = pForeground;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set "background" container for a "sandwich" container
 *
 * LongDescription:::
 *  Set "background" container for a "sandwich" container
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Stat structure
 *  pBackground::    Background to set
 *  c::              Container to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetBackground (
    IPWState           *pIPWState,
    Container          *pBackground,
    Container          *c)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    ERR_GEN (c->type != cContainerSandwich, cErrTypeNotSupported);

    c->child.childSandwich.backgroundContainer = pBackground;

    NORMAL_RETURN;
}
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the "TintMesh" object for a tint box
 *
 * LongDescription:::
 *  Set the "TintMesh" object for a tint box
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  tint::           TintMesh to use
 *
 * Output:::
 *  c::              Container to modify
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetTint (
    IPWState           *pIPWState,
    TintMesh           *tint,
    Container          *c)
{
    PROLOGUE;
    ContainerType type;

    ERR_GEN (!c, cErrNullPtr);
    ERR_PASS (w_containerGetType (pIPWState, c, &type));
    ERR_GEN (type != cContainerTint, cErrBadType);
    c->child.childTint = tint;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Set the mask for a container.
 *
 * LongDescription:::
 *    Set the mask for a container.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pixr::           PIXR to use as the container's mask
 *
 * Output:::
 *  c::              Container to modify
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetMask (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Container          *c)
{
    PROLOGUE;
    ERR_GEN (!c, cErrInvalidParm);
    c->mask = pixr;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Set the bounding rectangle for a container.
 *
 * LongDescription:::
 *    Set the bounding rectangle for a container. All coordinates are
 *    in micas.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              Bounding box x0, in micas
 *  y::              Bounding box y0, in micas
 *  w::              Bounding box width, in micas
 *  h::              Bounding box height, in micas
 *
 * Output:::
 *  c::              Container to modify
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetBounds (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Container          *c)
{
    PROLOGUE;

    ERR_GEN (!c, cErrInvalidParm);
    c->x = x;
    c->y = y;
    c->w = w;
    c->h = h;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the next container in the sibling list
 *
 * LongDescription:::
 *  Set the next container in the sibling list
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  next::           Container to make 'c's sibling
 *
 * Output:::
 *  c::              Container to modify
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetNext (
    IPWState           *pIPWState,
    Container          *next,
    Container          *c)
{
    PROLOGUE;
	
    ERR_GEN (!c, cErrInvalidParm);
    c->next = next;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    If this is a composite container, set the child pointer.
 *
 * LongDescription:::
 *    If this is a composite container, set the child pointer.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  child::          New child pointer
 *
 * Output:::
 *  c::              Container to mutate
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerSetChild (
    IPWState           *pIPWState,
    Container          *child,
    Container          *c)
{
    PROLOGUE;
    ContainerType type = cContainerUnknown;
    
    ERR_GEN (!c, cErrInvalidParm);
    ERR_PASS (w_containerGetType (pIPWState, c, &type));
    ERR_GEN (type != cContainerPage, cErrBadType);

    c->child.childContainer = child;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Allocate a "tint" structure, including colors matrix
 *
 * LongDescription:::
 *    Allocate a "tint" structure, including colors matrix
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xColors::        Number of columns
 *  yColors::        Number of rows
 *
 * Output:::
 *  ppTint::         TintMesh allocated
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_tintAllocate (
    IPWState           *pIPWState,
    Int32               xColors,
    Int32               yColors,
    TintMesh          **ppTint)
{
    PROLOGUE;
    TintMesh *tint;
    
    tint = (TintMesh *) IPW_MALLOC (sizeof(TintMesh), cAllocatorData);
    ERR_GEN (!tint, cErrNomem);
    tint->xPoints = xColors;
    tint->yPoints = yColors;
    
    tint->points = (ColorRecord *) IPW_MALLOC (sizeof (ColorRecord) * xColors *
					       yColors, cAllocatorData); 
    
    ERR_GEN_CLEANUP (!tint->points, cErrNomem);
    *ppTint = tint;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE (tint, cAllocatorData);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Returns the number of colors on the tint mesh
 *
 * LongDescription:::
 *   Returns the number of colors on the tint mesh
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  tint::           The tint in question
 *
 * Output:::
 *  xColors::        set to the number of columns in the tint mesh
 *  yColors::        set to the number of rows in the tint mesh
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_tintGetColorCount (
    IPWState           *pIPWState,
    TintMesh           *tint,
    Int32              *xColors,
    Int32              *yColors)
{
    PROLOGUE;
    
    ERR_GEN (!tint, cErrInvalidParm);
    
    *xColors = tint->xPoints;
    *yColors = tint->yPoints;
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *     Get a color at a particular x,y location
 *
 * LongDescription:::
 *     Get a color at a particular x,y location
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  tint::           Tint in question
 *  x::              x-coordinate of mesh color to get
 *  y::              y-coordinate of mesh color to get
 *
 * Output:::
 *  color::          Place to stuff color record
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_tintGetColor (
    IPWState           *pIPWState,
    TintMesh           *tint,
    Int32               x,
    Int32               y,
    ColorRecord        *color)
{
    PROLOGUE;
    
    ERR_GEN (!tint, cErrInvalidParm);

    ERR_GEN (x >= tint->xPoints || y >= tint->yPoints, cErrOutOfBounds);

    memcpy ((void *)color, (void *)(&(tint->points[y * tint->xPoints + x])), 
	    sizeof (ColorRecord));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *     Set a color at a particular x,y mesh location
 *
 * LongDescription:::
 *     Set a color at a particular x,y mesh location
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  x::              x-coordinate of location of color in mesh
 *  y::              y-coordinate of location of color in mesh
 *  color::          Color to use
 *
 * Output:::
 *  tint::           Tint mesh in which to set
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_tintSetColor (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    ColorRecord         color,
    TintMesh           *tint)
{
    PROLOGUE;
    
    ERR_GEN (!tint, cErrInvalidParm);
    ERR_GEN(x >= tint->xPoints || y >= tint->yPoints, cErrInvalidParm);
	
    memcpy ((void *)(&(tint->points[y * tint->xPoints + x])), (void *) color,
	    sizeof (ColorRecord)); 
    
    NORMAL_RETURN;
    
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate a container and set some stuff up
 *
 * LongDescription:::
 *  Allocate a container and set some reasonable defaults
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *
 * Output:::
 *  ppContainer::    Container allocated
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_allocateContainer (
    IPWState           *pIPWState,
    Container         **ppContainer)
{
    PROLOGUE;
    Container *c;
    
    c = (Container *) IPW_CALLOC (1, sizeof (Container), cAllocatorData);
    ERR_GEN (!c, cErrNomem);
    ERR_PASS (w_containerSetNext (pIPWState, NULL, c));
    ERR_PASS (w_containerSetChild (pIPWState, NULL, c));

    *ppContainer = c;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Count the number of bits in all the PIXRs of a container
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prContainer::     Container to count
 *
 * Output:::
 *  pBits::           Bit count, which must be initialized to zero when
 *                    on entry
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_containerCountBits (
    IPWState           *pIPWState,
    Container          *prContainer,
    Int32              *pBits)
{
    PROLOGUE;
    Container *p, *pNext;
    ContainerType type;
    PIXR      *prThis = NULL, *prMask = NULL;
    Int32      bits;
    
    ERR_GEN (!prContainer, cErrNullPtr);

    ERR_PASS (w_containerGetType (pIPWState, prContainer, &type));

    switch (type)
    {
    case cContainerPage:
	ERR_PASS (w_containerGetChild (pIPWState, prContainer, 0, &p));
	while (p)
	{
	    ERR_PASS (w_containerGetNext(pIPWState, p, &pNext));
	    ERR_PASS (w_containerCountBits (pIPWState, p, pBits));
	    p = pNext;
	}
	break;
	
    case cContainerText:
    case cContainerLineArt:
    case cContainerContone:
	/* A PIXR */
	ERR_PASS (w_containerGetPixr (pIPWState, prContainer, &prThis));
	if (prThis)
	{
	    ERR_PASS (w_numberBits (pIPWState, prThis, &bits));
	    *pBits += bits;
	}
	
	ERR_PASS (w_containerGetMask (pIPWState, prContainer, &prMask));
	if (prMask)
	{
	    ERR_PASS (w_numberBits (pIPWState, prThis, &bits));
	    *pBits += bits;
	}
	break;

    case cContainerTint:
	/* Uh...tbd */
	break;

    case cContainerSandwich:
	ERR_PASS (w_containerGetForeground (pIPWState, prContainer, &p));
	ERR_PASS (w_containerCountBits (pIPWState, p, pBits));
	ERR_PASS (w_containerGetBackground (pIPWState, prContainer, &p));
	ERR_PASS (w_containerCountBits (pIPWState, p, pBits));
	ERR_PASS (w_containerGetMask (pIPWState, prContainer, &prMask));
	if (prMask)
	    ERR_PASS (w_containerCountBits (pIPWState, p, pBits));
	break;

    default:
	ERR_GEN (1, cErrBadType);
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete a container and all of its children
 *
 * LongDescription:::
 *  Deletes a hierarchy of containers. If the "ownOthers" flag is set,
 *  then any non-container child object such as a PIXR is considered
 *  "owned" by the page, and therefore destroyed as the page is destroyed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  ownOthers::      If true, we own PIXRs, etc. attached to the page
 *
 * Output:::
 *  ppContainer::    Container to destroy
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyContainer (
    IPWState           *pIPWState,
    Bool                ownOthers,
    Container         **ppContainer)
{
    PROLOGUE;
    Container *p, *pNext, *container;
    ContainerType type;
    PIXR      *prThis = NULL, *prMask = NULL;
    
    ERR_GEN (!ppContainer, cErrNullPtr);
    ERR_GEN (!*ppContainer, cErrNullPtr);

    container = *ppContainer;

    ERR_PASS (w_containerGetType (pIPWState, container, &type));

    switch (type)
    {
    case cContainerPage:
	ERR_PASS (w_containerGetChild (pIPWState, container, 0, &p));
	while (p)
	{
	    ERR_PASS (w_containerGetNext(pIPWState, p, &pNext));
	    ERR_PASS (w_destroyContainer (pIPWState, ownOthers, &p));
	    p = pNext;
	}
	break;
	
    case cContainerText:
    case cContainerLineArt:
    case cContainerContone:
	/* A PIXR */
	if (ownOthers)
	{
	    ERR_PASS (w_containerGetPixr (pIPWState, container, &prThis));
	    if (prThis)
		ERR_PASS (w_destroyPixr (pIPWState, &prThis));
	    ERR_PASS (w_containerGetMask (pIPWState, container, &prMask));
	    if (prMask)
		ERR_PASS (w_destroyPixr (pIPWState, &prMask));
	}
	break;

    case cContainerTint:
	/* Uh...tbd */
	break;

    case cContainerSandwich:
	ERR_PASS (w_containerGetForeground (pIPWState, container, &p));
	ERR_PASS (w_destroyContainer (pIPWState, ownOthers, &p));
	ERR_PASS (w_containerGetBackground (pIPWState, container, &p));
	ERR_PASS (w_destroyContainer (pIPWState, ownOthers, &p));
	if (ownOthers)
	{
	    ERR_PASS (w_containerGetMask (pIPWState, container, &prMask));
	    if (prMask)
		ERR_PASS (w_destroyPixr (pIPWState, &prMask));
	}
	break;

    default:
	ERR_GEN (1, cErrBadType);
    }

    IPW_FREE (container, cAllocatorData);
    *ppContainer = NULL;

    NORMAL_RETURN;
}

/*
 * $Log: page.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.34  1997/11/15 01:24:52  xzhu
 * xzhu on Fri Nov 14 17:24:52 PST 1997
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Use ScaledInt1616 resolution to calculate Micas.
 *
 * Revision 1.33  1997/05/22 00:25:09  xzhu
 * xzhu on Wed May 21 17:12:16 PDT 1997
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Added a new function w_newNullImage().
 *
 * Revision 1.32  1996/07/22 14:12:27  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Fixed "nag script" warnings--some missing colons, missing ERR_CLEANUP
 * wrappers, misspelled arguments in comment, etc.
 *
 * Revision 1.31  1996/07/19 15:41:45  macleod
 * macleod on Fri Jul 19 08:41:24 PDT 1996
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * added w_containerCountBits, to count the total # of bits in a
 * composite page. This is used mostly for regression testing.
 *
 * Revision 1.30  1996/05/30 22:23:05  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * w_destroyContainer wasn't clearing the pointer passed--fixed.
 *
 * Revision 1.29  1996/05/15 01:06:56  macleod
 * macleod on Tue May 14 18:06:50 PDT 1996
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Fix nag script warnings
 *
 * Revision 1.28  1996/04/29 15:06:19  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Added the notion of an "image sandwich", which consists of two
 * Container objects and a selector plane. This is used to represent
 * "DIR"-esque images. Added the ability to render image sandwiches.
 * Note that the interfaces to a couple of functions have changed.
 *
 * Revision 1.27  1996/03/13 18:32:43  macleod
 * macleod on Wed Mar 13 10:32:24 PST 1996
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * added w_destroyContainer, to free a container object and its children
 *
 * Revision 1.26  1996/02/22 02:24:54  macleod
 * macleod on Wed Feb 21 18:23:57 PST 1996
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * DPI checks were failing when rescaling very small masks. Since the
 * scaling was computed, I just forced the DPI to be the same as the
 * result. This is always going to be a problem when insisting that
 * resolutions always match.
 *
 * Revision 1.25  1996/02/16 02:50:44  ddavies
 * ddavies on Thu Feb 15 18:50:08 PST 1996
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Change Log comment at the end to stop whining from co.  It didn't like
 * having the Log in the first line of the comment.
 *
 * Revision 1.24  1996/02/15 02:01:28  macleod
 * macleod on Wed Feb 14 18:01:04 PST 1996
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * fixed some rendering bugs
 *
 * Revision 1.23  1996/01/04  20:25:20  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.22  1995/12/13  04:11:07  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/12/01  19:32:38  ddavies
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Fix acc compiler warnings.
 *
 * Revision 1.20  1995/11/13  20:14:19  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/11/01  22:20:09  ddavies
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Generally change functions that call w_createPixrComposite so that they
 * deal in arrays of pointers to PIXRs instead of using 3 dedicated pointers.
 * This allows us to process multi-plane images with arbitrary numbers of planes.
 * I think this makes thing easier since we use little loops instead of
 * unrolled assignments.
 *
 * Revision 1.18  1995/10/13  22:03:36  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/04  18:03:32  macleod
 * macleod on Wed Oct 4 11:02:42 PDT 1995
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Changed the order of parameters for all of the "set" routines, where
 * the object having an attribute set needed to be moved to the ned.
 * (end, not ned.)
 *
 * Revision 1.16  1995/10/03  01:59:40  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.15  1995/10/02  19:19:15  jcraig
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * w_invert takes a pointer to a pointer to a destination pixr.
 *
 * Revision 1.14  1995/09/29  20:55:03  ddavies
 * ddavies on Fri Sep 29 13:54:43 PDT 1995
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Change calls to w_opGrayBinIP to equivalent calls to w_opGrayBin.
 *
 * Revision 1.13  1995/09/21  22:55:08  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/19  19:19:46  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.11  1995/09/11  21:54:48  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.10  1995/09/07  01:33:50  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:21:54  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  21:02:59  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.7  1995/08/31  21:15:41  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/25  01:37:38  ejaquith
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Fixed warnings
 *
 * Revision 1.5  1995/08/25  01:08:56  jcraig
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.4  1995/08/17  23:00:13  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Needed to comment out the #include of "shros.pub"
 *
 * Revision 1.3  1995/08/17  21:55:36  macleod
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.2  1995/07/12  20:46:36  macleod
 * macleod on Wed Jul 12 13:46:21 PDT 1995
 *
 * page.c -> /product/ipcore/ipwrap/src/RCS/page.c,v
 *
 * Fixed a bug in page rendering, where masked images weren't getting handled
 * correctly
 *
 * Revision 1.1  1995/06/09  01:03:13  macleod
 * Initial revision
 */

	
