/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* 
 * pageio.c
 *
 *     I/O for "Containers", including TIFF, PDF, XIFF
 *
 * Functions Defined:
 *
 * Rendering:
 *    renderContainerToPixr 
 *    [renderContainerToPDF]
 *    [renderContainerToPS]
 *    saveContainerToPIFF
 *    loadContainerFromPIFF
 *    saveContainerToFiles
 */

#include <stdio.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "defines.pub"
#include "frames.pub"
#include "appsinc.h"
#include "logicops.h"
#include "imagerep.h"

/* prototypes */
#include "props.pub"
#include "page.pub"
#include "logicops.pub"
#include "grbinop.pub"
#include "rescale.pub"
#include "reduce.pub"
#include "prxform.pub"
#include "iw_scale.pub"
#include "io.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: pageio.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    render a container, i.e. a composite page, into a single PIXR.
 *
 * LongDescription:::
 *    Render a container, i.e. a composite page, into a single PIXR.
 *    This routine is being written with "test code" in mind--it's
 *    may be a bit of a pig, especially re memory usage.
 *    <BR>
 *    The idea is to cast all image types up to the resolution and bit
 *    depth of the target PIXR. The rendering will happen in the order
 *    that elements are found in the container, depth-first.
 *    <BR>
 *    If an element has a mask, that mask is used to limit what is copied.
 *    Otherwise, the whole element is painted.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      state structure
 *  container::      Container to render
 *  resultDpi::      DPI of the result image
 *  xOffset::        X-offset, in micas, in result PIXR
 *  yOffset::        Y-offset, in micas, in result PIXR
 *
 * Output:::
 *  prResult::       The resulting PIXR
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_renderContainerToPixr (
    IPWState           *pIPWState,
    Container          *container,
    Int32               resultDpi,
    Int32               xOffset,
    Int32               yOffset,
    PIXR               *prResult)
{
    PROLOGUE;
    Container *p, *pForeground, *pBackground;
    Int32      thisX, thisY, thisW, thisH, thisDepth, pW, pH, resultX,
	       resultY, resultW, resultH, resultDepth, resultChannels, 
	       channels, width, i, binW, binH, scaleW, scaleH;
    Float32    scale;
    PIXR      *prScaled = NULL, *prThis = NULL, *prMask = NULL,
	      *prTmp = NULL, *prTmp1 = NULL, *prColor = NULL,
	      *prForeground = NULL, *prBackground = NULL,
	      *prFGScaled = NULL, *prBGScaled = NULL, *prSelectResult = NULL,
	      *prMaskScaled = NULL;
    PIXR      *planes[cMaxColorPlanes];
    ContainerType type;
    ScaledInt1616 xDPI, yDPI;
    void      *fTab = NULL, *lTab = NULL;

    for (i = 0; i < cMaxColorPlanes; i++)
	planes[i] = NULL;

    ERR_PASS (w_pixrGetDepth (pIPWState, prResult, &resultDepth));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prResult, &resultChannels));
    
    ERR_PASS (w_containerGetType (pIPWState, container, &type));

    switch (type)
    {
    case cContainerPage:
	ERR_PASS (w_containerGetChild (pIPWState, container, 0, &p));
	while (p)
	{
	    ERR_PASS (w_renderContainerToPixr (pIPWState, p, resultDpi,
					       xOffset, yOffset, prResult));
	    ERR_PASS (w_containerGetNext(pIPWState, p, &p));
	}
	break;
	
    case cContainerText:
    case cContainerLineArt:
    case cContainerContone:
	/* A PIXR */
	ERR_PASS (w_containerGetPixr (pIPWState, container, &prThis));
	ERR_PASS (w_containerGetMask (pIPWState, container, &prMask));
	
	ERR_PASS (w_containerGetBounds (pIPWState, container, &thisX, &thisY,
				      &thisW, &thisH));
	thisX += xOffset;
	thisY += yOffset;
	ERR_PASS (w_pixrGetDepth (pIPWState, prThis, &thisDepth));
	
	ERR_PASS (w_pixrGetWidth (pIPWState, prThis, &pW));
	ERR_PASS (w_pixrGetHeight (pIPWState, prThis, &pH));
	resultX = MICAS_TO_PIXELS (thisX, resultDpi);
	resultY = MICAS_TO_PIXELS (thisY, resultDpi);
	resultW = MICAS_TO_PIXELS (thisW, resultDpi);
	resultH = MICAS_TO_PIXELS (thisH, resultDpi);

/*
	fprintf(stderr, "display image at %ld, %ld, %ld, %ld, rdpi %ld..\n",
		resultX, resultY, resultW, resultH, resultDpi);
*/	

	if (resultW == pW && resultH == pH && thisDepth == resultDepth)
	{
	    if (prMask == NULL)
		ERR_PASS (w_rasterop (pIPWState, 0, 0, prThis, PIX_SRC,
				      resultX, resultY, pW, pH, prResult));
	    else
		ERR_PASS (w_copyGrayMasked (
		    pIPWState, 
		    0, 0, prThis, 
		    0, 0, prMask,
		    resultX, resultY, pW, pH, prResult));
	}
	else
	{
	    /* We have to scale and/or depth-convert */
	    prScaled = NULL;
	    if (thisDepth == resultDepth)
	    {
		/* Do a non-interpolated one for now */
		prScaled = NULL;
		if (thisDepth == 4 || thisDepth == 8)
		{
		    ERR_PASS (w_scaleLinearSize (pIPWState, prThis, 
						 resultW, resultH, 
						 &prScaled));
		}
		else
		{
		    ERR_PASS (w_scaleToSize (pIPWState, prThis, 
					     resultW, resultH,
					     &prScaled));
		}
	    }
	    else
	    {
		/* Resize */
		if (thisDepth == 4 || thisDepth == 8)
		{
		    ERR_PASS (w_scaleLinearSize (pIPWState, prThis, 
						 resultW, resultH, 
						 &prTmp));
		}
		else
		{
		    ERR_PASS (w_scaleToSize (pIPWState, prThis, 
					     resultW, resultH,
					     &prTmp));
		}

		if (thisDepth == 1)
		{
		    if (resultDepth == 8)
		    {
			ERR_PASS_CLEANUP (w_depth1To8 (pIPWState, prTmp, 
							 cTrue, &prScaled));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		    }
		    else if (resultDepth == 4)
		    {
			ERR_PASS_CLEANUP (w_depth1To8 (pIPWState, prTmp, 
							 cTrue, &prTmp1));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
			ERR_PASS_CLEANUP (w_scale8To4 (pIPWState, prTmp1, 
						       &prScaled));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
		    }
		    else 
			ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
		}
		else if (thisDepth == 4)
		{
		    if (resultDepth == 8)
		    {
			ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prTmp, 
							 &prScaled));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		    }
		    else 
			ERR_GEN_CLEANUP (1, cErrDepthNotSupported);
		}
		else if (thisDepth == 8)
		{
		    if (resultDepth == 4)
		    {
			ERR_PASS_CLEANUP (w_scale8To4 (pIPWState, prTmp,
						       &prScaled));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		    }
		}
	    }
	    
	    /* And copy that in */
	    if (prMask == NULL)
	    {
		ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prScaled,
							 &channels));
		if (resultChannels > 1 && channels == 1)
		{

		    for (i = 0; i < resultChannels; i++)
			planes[i] = prScaled;

		    ERR_PASS_CLEANUP (w_pixrGetType (pIPWState, prResult, 
						     &type));
		    ERR_PASS_CLEANUP (w_createPixrComposite (pIPWState, 
							     planes,
							     resultChannels,
							     type, &prColor));

		    ERR_PASS_CLEANUP (
			w_rasterop (pIPWState, 0, 0, prColor, PIX_SRC,
				    resultX, resultY, resultW, resultH,
				    prResult));

		    /* Just free prColor, since it doesn't own its children */
		    IPW_FREE (prColor, cAllocatorData);
		}
		else
		{
		    ERR_PASS_CLEANUP (
			w_rasterop (pIPWState, 0, 0, prScaled, PIX_SRC,
				    resultX, resultY, resultW, resultH,
				    prResult));
		}
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prScaled, 
						       &channels));
		if (resultChannels > 1 && channels == 1)
		{
		    for (i = 0; i < resultChannels; i++)
			planes[i] = prScaled;

		    ERR_PASS_CLEANUP (w_pixrGetType (pIPWState,
				      prResult, &type));
		    prColor = NULL;
		    ERR_PASS_CLEANUP (w_createPixrComposite(pIPWState,
							    planes,
							    resultChannels,
							    type, &prColor));
		}
		    
		ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prMask, &width));
		if (width == resultW)
		{
		    ERR_PASS_CLEANUP (w_copyGrayMasked (
			pIPWState,  
			0, 0, prColor ? prColor : prScaled,	
			0, 0, prMask, 
			resultX, resultY, resultW, resultH, prResult));
		}
		else
		{
		    ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, prMask, 
						       resultW, resultH,
						       &prTmp));
		    /* 
		     * Hack: We get inaccuracies in the resolution when
		     * scaling small masks, so just slam the resolution
		     * in from the result so that we don't get in trouble
		     */
		    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (
			pIPWState,
			prResult,
			&xDPI, &yDPI));
		    ERR_PASS_CLEANUP (w_pixrSetResolutionDPI (pIPWState,
							      prTmp, 
							      xDPI, yDPI));
		    ERR_PASS_CLEANUP (w_pixrSetResolutionDPI (pIPWState,
							      prScaled, 
							      xDPI, yDPI));
		    ERR_PASS_CLEANUP (w_copyGrayMasked (
			pIPWState, 
			0, 0, prColor ? prColor : prScaled, 
			0, 0, prTmp,
			resultX, resultY, resultW, resultH, prResult));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		}

		/* prColor doesn't own its' sub PIXRs, so just FREE it 
		   (wouldn't life be better with a garbage collector?) */
		if (prColor)
		    IPW_FREE (prColor, cAllocatorData);

	    }

            ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaled));
        }
    
        break;
	
    case cContainerTint:
	/* Uh...tbd */
	break;

    case cContainerSandwich:
	/* 
	 * Render an "image sandwich". This involves scaling the mask
	 * plane, (to gray, if shrinking) scaling the fg/bg planes, and 
	 * then doing a graySelect between the foreground and background
	 * planes. At some point, we should be able to do this in strips,
	 * so as to not take a huge amount of memory.
	 */
	ERR_PASS_CLEANUP (w_containerGetBounds (pIPWState, container,
						&thisX, &thisY, 
						&thisW, &thisH));
	thisX += xOffset;
	thisY += yOffset;
	
	ERR_PASS_CLEANUP (w_containerGetMask (pIPWState, container, &prMask));
	ERR_GEN_CLEANUP (prMask == NULL, cErrNoImageData);
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prMask, &binW));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prMask, &binH));
	ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prMask, 
						  &xDPI, &yDPI));
	
	resultX = MICAS_TO_PIXELS (thisX, resultDpi);
	resultY = MICAS_TO_PIXELS (thisY, resultDpi);
	resultW = MICAS_TO_PIXELS (thisW, resultDpi);
	resultH = MICAS_TO_PIXELS (thisH, resultDpi);

	/* Compute the scale for the mask */
	scale = resultW * 1.0f / binW;
	scaleW = (Int32) (binW * scale);
	scaleH = (Int32) (binH * scale);

	/* 
	 * This won't support panning at the moment. This needs to 
	 * be fixed, eventually, in order to support a viewer.
	 */

	ERR_PASS_CLEANUP (w_makeScaleToGrayArbFilters (pIPWState, scaleW, 
						       binW, 4, 8, 8, &fTab));
	ERR_PASS_CLEANUP (w_makeScaleToGrayArbLineTab (pIPWState, fTab, 0, 
						       resultW, resultW, 
						       prMask, &lTab));

	ERR_PASS_CLEANUP (w_createPixr (pIPWState, scaleW, scaleH, 4, 
					(xDPI * scaleW) / binW,
					(yDPI * scaleW) / binW, 
					cFrameBits, &prMaskScaled));

	ERR_PASS_CLEANUP (w_scaleToGrayArb (pIPWState, prMask, 0, 
					    fTab, lTab, prMaskScaled));

	/* Get the foreground/background containers */
	ERR_PASS_CLEANUP (w_containerGetForeground (pIPWState, container, 
						    &pForeground));
	ERR_PASS_CLEANUP (w_containerGetBackground (pIPWState, container, 
						    &pBackground));

	/* Get the PIXRs out of the containers */
	ERR_PASS_CLEANUP (w_containerGetPixr (pIPWState, pForeground, 
					      &prForeground));
	ERR_PASS_CLEANUP (w_containerGetPixr (pIPWState, pBackground, 
					      &prBackground));

	/* Scale those */
	ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, prForeground, 
					 scaleW, scaleH, &prFGScaled));
	ERR_PASS_CLEANUP (w_scaleLinearSize (pIPWState, prBackground, 
					     scaleW, scaleH, &prBGScaled));

	/* Do a gray select. If there's an offset, we need to make a 
	   copy. */
	if (xOffset > 0 || yOffset > 0)
	{
	    /* Make a copy */
	    ERR_PASS_CLEANUP (w_graySelect (pIPWState, prFGScaled, prBGScaled,
					    prMaskScaled, &prSelectResult));

	    /* Blast it in */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prSelectResult,
					  PIX_SRC, 
					  xOffset, yOffset, resultW, resultH, 
					  prResult));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSelectResult));
	}
	else
	{
	    ERR_PASS_CLEANUP (w_graySelect (pIPWState, prFGScaled, prBGScaled,
					    prMaskScaled, &prResult));
	}

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMaskScaled));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prFGScaled));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBGScaled));
	break;

    default:
	ERR_GEN (1, cErrBadType);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    if (prTmp1) w_destroyPixr (pIPWState, &prTmp1);
    if (prColor) IPW_FREE (prColor, cAllocatorData);
    if (prScaled) w_destroyPixr (pIPWState, &prScaled);
    if (prFGScaled) w_destroyPixr (pIPWState, &prFGScaled);
    if (prBGScaled) w_destroyPixr (pIPWState, &prBGScaled);
    if (prMaskScaled) w_destroyPixr (pIPWState, &prMaskScaled);
    ERR_RETURN;
}
    
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Copy one gray pixr onto another, through a mask
 *
 * LongDescription:::
 *    This routine is used to copy one gray PIXR into another, using
 *    a mask to determine which bits to copy. ACHTUNG! This uses
 *    opGrayBinPrIP to pick out the bits and to cut a hole in the
 *    dest PIXR, and it is therefore somewhat sub-optimal! If you actually
 *    need to use this in production code, make me re-write it! (It's
 *    not really that hard)
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xS::             Offset into source image
 *  yS::             
 *  prSrc::          Source image
 *  xM::             Offset into mask image
 *  yM::             
 *  prMask::         Mask image 
 *  xD::             Rectangle to copy into in destination image
 *  yD::             
 *  wD::             
 *  hD::             
 *
 * Output:::
 *  prDest::         Destination image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyGrayMasked (
    IPWState           *pIPWState,
    Int32               xS,
    Int32               yS,
    PIXR               *prSrc,
    Int32               xM,
    Int32               yM,
    PIXR               *prMask,
    Int32               xD,
    Int32               yD,
    Int32               wD,
    Int32               hD,
    PIXR               *prDest)
{
    PROLOGUE;
    PIXR     *prDestCrop = NULL, *prSrcCrop = NULL, *prMaskCrop = NULL,
	     *prConv = NULL, *prPlane;
    PIXR     *planes[cMaxColorPlanes];
    Bool      cropDest = cFalse, cropSrc = cFalse, cropMask = cFalse;
    Int32     channels, srcDepth, destDepth, srcChannels, destW, 
	      destH, srcW, srcH, maskW, maskH, i;
    ScaledInt1616     xDPI, yDPI;
    ColorType type;
    ColorMap *map;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1|cChannels3, cFrame32);
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1|cChannels3, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetType (pIPWState, prDest, &type));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prDest, &map));
    ERR_PASS (w_pixrGetDepth (pIPWState, prDest, &destDepth));
    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &srcDepth));
   
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prDest, &channels));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prSrc, &srcChannels));
    ERR_GEN (channels != srcChannels, cErrChannelsNotEqual);
    ERR_PASS (w_pixrGetWidth (pIPWState, prDest, &destW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prDest, &destH));
    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &srcW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &srcH));
    ERR_PASS (w_pixrGetWidth (pIPWState, prMask, &maskW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prMask, &maskH));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prDest, &xDPI, &yDPI));

    /* If we are doing sub-rects, we need to make copies */
    if (xD == 0 && yD == 0 && wD == destW && hD == destH)
    {
	cropDest = cFalse;
	prDestCrop = prDest;
    }
    else
    {
	cropDest = cTrue;

	ERR_PASS (w_createPixrExtended (pIPWState, wD, hD, destDepth, xDPI,
				      yDPI, cFrameBits,
				      channels, type, map, &prDestCrop));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, xD, yD, prDest, PIX_SRC,
				      0, 0, wD, hD, prDestCrop));
    }
    

    
    if (xS == 0 && yS == 0 && wD == srcW && hD == srcH)
    {
	cropSrc = cFalse;
	prSrcCrop = prSrc;
    }
    else
    {
	cropSrc = cTrue;
	ERR_PASS_CLEANUP (w_createPixrExtended (pIPWState, wD, hD, srcDepth,
					      xDPI, yDPI, 
					      cFrameBits, channels, type,
					      map, &prSrcCrop));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, xS, yS, prSrc, PIX_SRC,
				      0, 0, wD, hD, prSrcCrop));
    }

    if (xM == 0 && yM == 0 && wD == maskW && hD == maskH)
    {
	cropMask = cFalse;
	prMaskCrop = prMask;
    }
    else
    {
	cropMask = cTrue;
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, wD, hD, 1, xDPI, yDPI, 
					cFrameBits, &prMaskCrop));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, xM, yM, prMask, PIX_SRC,
				      0, 0, wD, hD, prMaskCrop));
    }
    
    if (channels > 1)
    {

	/* "paint" the dest planes */
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prDestCrop, i, 
						&prPlane));
	    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prPlane, prMaskCrop, 
					   PIX_SRC | PIX_DST, &prPlane));
	}

	/* "paint" the source planes */
	ERR_PASS_CLEANUP (w_invert (pIPWState, prMaskCrop, &prMaskCrop));
	for (i = 0; i < channels; i++)
	{
	    
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prSrcCrop, i, 
					        &prPlane));
	    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prPlane, prMaskCrop, 
					   PIX_SRC | PIX_DST, &prPlane));
	}
    }
    else
    {
	/* "paint" the dest, forcing all in mask to be 0 */
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prDestCrop, prMaskCrop, 
				       PIX_SRC | PIX_DST, &prDestCrop));

	/* "paint" the src, i.e. make all non-src image 0 */
	ERR_PASS_CLEANUP (w_invert (pIPWState, prMaskCrop, &prMaskCrop));
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prSrcCrop, prMaskCrop,
				       PIX_SRC | PIX_DST, &prSrcCrop));
    }

    
    /* Convert depths, if necessary, and copy it in */
    if (srcDepth != destDepth)
    {
	prConv = NULL;
	
	if (channels > 1)
	{
	    for (i = 0; i < channels; i++)
	    {
		ERR_PASS_CLEANUP (w_pixrGetChannel(pIPWState, prSrcCrop, i,
						   &prPlane));
		ERR_PASS_CLEANUP (w_depthConvertGray (pIPWState, prPlane, 
						      destDepth, cFalse, 
						      IP_HI_QUALITY,
						      &(planes[i])));
	    }
	    ERR_PASS_CLEANUP (w_pixrGetType (pIPWState, prSrc, &type));
	    ERR_PASS_CLEANUP (w_createPixrComposite (pIPWState, planes,
						     channels, type, &prConv));

	}
	else
	{
	    ERR_PASS_CLEANUP (w_depthConvertGray (pIPWState, prSrcCrop,
						  destDepth, cFalse, 
						  IP_HI_QUALITY, &prConv));
	}

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prConv, OR, prDestCrop));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prConv));
	
    }
    else
    {
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrcCrop, OR, prDestCrop));
    }
    
    /* Copy back in, if necessary */
    if (cropDest)
    {
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prDestCrop, PIX_SRC,
				      xD, yD, wD, hD, prDest));
	
    }

    /* Free stuff up if needed */
    if (cropDest)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDestCrop));
    if (cropSrc) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSrcCrop));
    if (cropMask)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMaskCrop));
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (cropDest)
	w_destroyPixr (pIPWState, &prDestCrop);
    if (cropSrc)
	w_destroyPixr (pIPWState, &prSrcCrop);
    if (cropMask)
	w_destroyPixr (pIPWState, &prMaskCrop);
    
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Save each element in a container to a file
 *
 * LongDescription:::
 *  This function writes all of the elements of a container to individual
 *  files. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::       State structure
 *  prContainer::     Container to save
 *  pBaseName::       Base file name
 *  saveType::        type of file to save
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  page layout
 *
 * Keywords:::
 *  page layout
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  in process
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_saveContainerToFiles (
    IPWState           *pIPWState,
    Container          *prContainer,
    const char         *pBaseName,
    Int32               saveType)
{
    PROLOGUE;
    Container     *p, *pNext;
    ContainerType  type;
    PIXR          *prThis = NULL, *prMask = NULL;
    Int32          i;
    char           subName[256];
    
    ERR_GEN (!prContainer, cErrNullPtr);

    ERR_PASS (w_containerGetType (pIPWState, prContainer, &type));

    switch (type)
    {
    case cContainerPage:
	ERR_PASS (w_containerGetChild (pIPWState, prContainer, 0, &p));
	for (i = 0; p; i++)
	{
	    sprintf(subName, "%s_%ld", pBaseName, i);
	    ERR_PASS (w_containerGetNext(pIPWState, p, &pNext));
	    ERR_PASS (w_saveContainerToFiles (pIPWState, p, subName, 
					      saveType));
	    p = pNext;
	}
	break;
	
    case cContainerText:
	/* A PIXR */
	ERR_PASS (w_containerGetPixr (pIPWState, prContainer, &prThis));
	if (prThis)
	{
	    sprintf(subName, "%s_text", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prThis, saveType));
	}
	
	ERR_PASS (w_containerGetMask (pIPWState, prContainer, &prMask));
	if (prMask)
	{
	    sprintf(subName, "%s_mask", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prMask, saveType));
	}
	break;

    case cContainerLineArt:
	/* A PIXR */
	ERR_PASS (w_containerGetPixr (pIPWState, prContainer, &prThis));
	if (prThis)
	{
	    sprintf(subName, "%s_line", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prThis, saveType));
	}
	
	ERR_PASS (w_containerGetMask (pIPWState, prContainer, &prMask));
	if (prMask)
	{
	    sprintf(subName, "%s_mask", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prMask, saveType));
	}
	break;

    case cContainerContone:
	/* A PIXR */
	ERR_PASS (w_containerGetPixr (pIPWState, prContainer, &prThis));
	if (prThis)
	{
	    sprintf(subName, "%s_image", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prThis, saveType));
	}
	
	ERR_PASS (w_containerGetMask (pIPWState, prContainer, &prMask));
	if (prMask)
	{
	    sprintf(subName, "%s_mask", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prMask, saveType));
	}
	break;

    case cContainerTint:
	/* Uh...tbd */
	break;

    case cContainerSandwich:
	ERR_PASS (w_containerGetForeground (pIPWState, prContainer, &p));
	sprintf(subName, "%s_fg", pBaseName);
	ERR_PASS (w_saveContainerToFiles (pIPWState, p, subName, saveType));
	ERR_PASS (w_containerGetBackground (pIPWState, prContainer, &p));
	sprintf(subName, "%s_bg", pBaseName);
	ERR_PASS (w_saveContainerToFiles (pIPWState, p, subName, saveType));
	ERR_PASS (w_containerGetMask (pIPWState, prContainer, &prMask));
	if (prMask)
	{
	    sprintf(subName, "%s_sel", pBaseName);
	    ERR_PASS (w_write (pIPWState, subName, prMask, saveType));
	}
	break;

    default:
	ERR_GEN (1, cErrBadType);
    }

    NORMAL_RETURN;
}

/*
 * $Log: pageio.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.3  1996/08/07 23:08:19  macleod
 * pageio.c -> /product/ipcore/ipwrap/src/RCS/pageio.c,v
 *
 * Added w_saveContainerToFiles, which writes all of the images in
 * a container to a bunch of files. This is useful for debugging.
 *
 * Revision 1.2  1996/05/31 17:15:37  macleod
 * pageio.c -> /product/ipcore/ipwrap/src/RCS/pageio.c,v
 *
 * (Er, that was "floating constant" in the previous checkin)
 * Added "f" after floating constant to make VC++ happy
 * Also added cast to Int32 where something was getting
 * coerced to Int32 to make VC++ happy.
 *
 * Revision 1.1  1996/05/03 15:49:31  macleod
 * Initial revision
 *
 */

	
