/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  pcconv.c
 *
 *  Routines:
 *		w_bmpToPixr()
 *		w_pixrToBMP()
 *		w_getDIBPixrLineProc()
 *		w_RGBPixrsToBMP()
 *		w_BMPToRGBPixrs()
 */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"

/* prototypes */
#include "props.pub"
#include "shrpixr.pub"	/* contains types used by pcconv.pub */
#include "pcconv.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: pcconv.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a DIB BMP image to a PIXR.
 *
 * LongDescription:::
 *    Convert the image in the source DIB bitmap into the PIXR format .
 *    Convert (0, 0) to the upper left corner by
 *    starting from the bottom of the bitmap and working our
 *    way upward, byteswapping as we go.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  bmpS::           Ptr to a source bitmap (from a dib so that (0, 0) is in
 *                   the lower left corner.
 *  invertImage::    cTrue to invert the image as we convert; cFalse o.w.
 *  yOffset::        Starting line for conversion (see notes below).
 *  height::         Number of lines to convert (see notes below).
 *
 * Output:::
 *  prD::            Ptr to dest PIXR to convert DIB in to.
 *
 * Returns:::
 *
 * Assumptions:::
 *  The source (and destination) are 32 bit aligned.<BR>
 *  The destination need not have a frame.
 *
 * Notes:::
 *  slice conversion::
 *    The yOffset and height parameters allow us to convert
 *    a Microsoft DIB to a slice of a PIXR.  If height is
 *    greater than 0, we interpret yOffset as the starting
 *    line (0 at the top) and height as the number of lines.
 *    If height == 0, we convert the entire image.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, byteswapping, rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_BMPToPixr (
    IPWState           *pIPWState,
    UInt8 __far        *bmpS,
    UInt32              invertImage,
    Int32               yOffset,
    Int32               height,
    PIXR               *prD)
{
    PROLOGUE;
    Int32	 DIBBpl, pixrBpl, d, w;
    UInt8	*pPixrBits;

    ERR_GEN_PIXR(prD, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN(bmpS == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetImage(pIPWState, prD, &pPixrBits));
    ERR_PASS( w_pixrGetWidth(pIPWState, prD, &w));
    ERR_PASS( w_pixrGetDepth(pIPWState, prD, &d));
    DIBBpl = ((w * d + 31) / 32) * 4;
    ERR_PASS( w_pixrGetBpl(pIPWState, prD, &pixrBpl));

    /* is height == 0?  If so, convert the entire image.  Otherwise,
     * convert height lines starting at yOffset
     */
    if (height == 0)
	ERR_PASS( w_pixrGetHeight(pIPWState, prD, &height));
    else
	pPixrBits += yOffset * pixrBpl;

    /* have i_DIBImageFromToPixr do the work */
    ERR_PASS( i_DIBImageFromToPixr (pIPWState,
				    bmpS,              /* pDIBImage */
				    pPixrBits,         /* pD */
				    height,            /* height */
				    invertImage,       /* invertImage */
				    DIBBpl,            /* DIBBpl */
				    pixrBpl,           /* dBpl */
				    cDIBToPixr));      /* direction */
    
    /* make sure the frame is 0 */
    /* Watch the use of pixr_rop here because it offsets all pixrs
     * past the frame into the upper left corner of the image.
     */
    ERR_PASS( w_setFramePixels(pIPWState, prD, 0));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a PIXR to a DIB BMP image.
 *
 * LongDescription:::
 *   Convert the image in the source PIXR into the DIB format 
 *   Convert (0, 0) to the lower left corner by
 *   writing to the bottom of the bitmap and working our
 *   way upward, byteswapping as we go.  The destination
 *   bitmap will have the frame removed.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prS::            Ptr to source PIXR.
 *  invertImage::    cTrue to invert the image as we convert; cFalse o.w.
 *  yOffset::        Starting line for conversion (see notes below).
 *  height::         Number of lines to convert (see notes below).
 *
 * Output:::
 *  bmpD::           Ptr to a destination bitmap (to a DIB so that (0, 0)
 *                   is in the lower left corner) to convert PIXR in to.
 *
 * Returns:::
 *
 * Assumptions:::
 *   The source (and destination) are 32 bit aligned.
 *
 * Notes:::
 *  slice conversion::
 *   The yOffset and height parameters allow us to convert
 *   a slice of a PIXR to a Microsoft DIB.  If height is
 *   greater than 0, we interpret yOffset as the starting
 *   line (0 at the top) and height as the number of lines.
 *   If height == 0, we convert the entire image.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, byteswapping, rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrToBMP (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt32              invertImage,
    Int32               yOffset,
    Int32               height,
    UInt8 __far        *bmpD)
{
    PROLOGUE;
    Int32	DIBBpl, pixrBpl, w, d;
    UInt8	*pPixrBits;	/* pointer to pixr data */	

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN(bmpD == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetImage(pIPWState, prS, &pPixrBits));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    DIBBpl = ((w * d + 31) / 32) * 4;
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &pixrBpl));

    /* is height == 0?  If so, convert the entire image.  Otherwise,
     * convert height lines starting at yOffset
     */
    if (height == 0)
	ERR_PASS( w_pixrGetHeight(pIPWState, prS, &height));
    else
	pPixrBits += yOffset * pixrBpl;

    /* have i_DIBImageFromToPixr do the work */
    ERR_PASS( i_DIBImageFromToPixr (pIPWState,
				    bmpD,              /* pDIBImage */
				    pPixrBits,         /* pD */
				    height,            /* height */
				    invertImage,       /* invertImage */
				    DIBBpl,            /* DIBBpl */
				    pixrBpl,           /* dBpl */
				    cPixrToDIB));      /* direction */

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get a ptr to a PIXR/DIB line conversion routine.
 *
 * LongDescription:::
 *    The EZScan folks would like to convert scanned images to
 *    PIXRs as quickly, and with as little memory usage, as
 *    possible.  That means they want to send each scan line
 *    to the PIXR quickly.
 *    This routine returns a pointer to a routine that translates
 *    between one DIB scan line and one PIXR scan line.  The
 *    routine is optimized for the type of processor.<BR><BR>
 *  
 *    The conversion routine is called as follows:<BR><BR>
 *
 *      Translation from gray or binary DIB to 1 PIXR:<BR>
 *          (*swapLineProc)((UInt32 __far *)pDIBImage,
 *                          (UInt32 *) pPixrData,
 *                          DIBBpl/4);<BR><BR>
 *
 *      Translation from 1 PIXR to gray or binary DIB:<BR>
 *          (*swapLineProc)((UInt32 *) pPixrData,
 *                          (UInt32 __far *)pDIBImage,
 *                          DIBBpl/4);<BR><BR>
 *
 *      Byteswapping from Pixr to PIXR:<BR>
 *          (*swapLineProc)((UInt32 *) pPixrSource,
 *                          (UInt32 *) pPixrDest,
 *                          bpl/4);<BR><BR>
 *
 *      Translation from RGB DIB to 3 PIXRs or vice versa:<BR>
 *          (*swapLineProc)((UInt8 __far *)pDIBImage,
 *                          (UInt8 *) pRedPixrData,
 *                          (UInt8 *) pGreenPixrData,
 *                          (UInt8 *) pBluePixrData,
 *                          (UInt32) width);<BR><BR>
 *
 *   where pDIBImage is a pointer to the left side of the DIB line to be
 *   translated; pPixrData, pRedPixrData, pGreenPixrData and pBluePixrData
 *   are pointers to the left end of the PIXR line to be translated (inside
 *   the frame).  These are obtained by starting with w_pixrGetImage() and
 *   adding or subtracting the bytes/line (from w_pixrGetBpl()) as
 *   appropriate.<BR><BR>
 *
 *   REMEMBER that the order of scan lines in DIBs is reversed from that in
 *   PIXRs!!!
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  invertImage::       cTrue => bytes should be inverted
 *  direction::         cDIBToPixr => translation from a DIB to a PIXR;
 *                      cPixrToDIB => translation from a PIXR to a DIB.
 *                      cPixrToPixr => translation from a PIXR
 *                      to a PIXR (byteswapping)
 *  isColor::           cTrue => the DIB is an RGB image.  Otherwise,
 *                      the DIB is a gray or binary image.
 *
 * Output:::
 *  pGrayToFarProc::	A procedure that byteswaps a scan line containing
 *			a single channel of binary or gray image data where
 *			the source is a near pointer and the dest is
 *			a far pointer.
 *  pGrayFromFarProc::	A procedure that byteswaps a scan line containing
 *			a single channel of binary or gray image data where
 *			the source is a far pointer and the dest is
 *			a near pointer.
 *  pGrayNearProc::	A procedure that byteswaps a scan line containing
 *			a single channel of binary or gray image data where
 *			both the source and dest pointers are near.
 *  pColorProc::	A procedure that converts between an interleaved
 *			line of three-channel data and three near buffers
 *			each holding one channel.  This form is used both
 *			when converting to an interleaved DIB and when
 *			converting from an interleaved DIB.
 *
 *			The types GraySwapToFarProc, GraySwapFromFarProc,
 *			GraySwapNearProc and ColorSwapProc are defined in
 *			ipshared/include/shrpixr.pub.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, byteswapping, color, multi-channel, rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getDIBPixrLineProc (
    IPWState            *pIPWState,
    UInt32               invertImage,
    UInt32               direction,
    UInt32               isColor,
    GraySwapToFarProc	*pGrayToFarProc,
    GraySwapFromFarProc	*pGrayFromFarProc,
    GraySwapNearProc	*pGrayNearProc,
    ColorSwapProc	*pColorProc)
{
    PROLOGUE;
    GraySwapToFarProc swapGrayToFarProc = NULL;
    GraySwapFromFarProc swapGrayFromFarProc = NULL;
    GraySwapNearProc swapGrayNearProc = NULL;
    ColorSwapProc swapColorProc = NULL;

    ERR_PASS(i_getDIBPixrLineProc (pIPWState, invertImage, direction, isColor,
				   &swapGrayToFarProc, &swapGrayFromFarProc,
				   &swapGrayNearProc, &swapColorProc));

	/* Now decide which of the source pointers to update. */
    if (isColor)
    {
	ERR_GEN(pColorProc == NULL, cErrInvalidParm);
	*pColorProc = swapColorProc;
    }
    else
    {
	switch (direction)
	{
	case cPixrToDIB:
	    ERR_GEN(pGrayToFarProc == NULL, cErrInvalidParm);
	    *pGrayToFarProc = swapGrayToFarProc;
	    break;
	case cDIBToPixr:
	    ERR_GEN(pGrayFromFarProc == NULL, cErrInvalidParm);
	    *pGrayFromFarProc = swapGrayFromFarProc;
	    break;
	default:	/* cPixrToPixr */
	    ERR_GEN(pGrayNearProc == NULL, cErrInvalidParm);
	    *pGrayNearProc = swapGrayNearProc;
	}
    }
    NORMAL_RETURN;


}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an RGB 24-bit PIXR to a DIB BMP image.
 *
 * LongDescription:::
 *   Combine the 3 plane PIXR into the image portion of a 
 *   Microsoft DIB.  One pixel from each image is taken
 *   to form an RGB triple that is put into the DIB. 
 *   Convert (0, 0) to the lower left corner by
 *   writing to the bottom of the bitmap and working our
 *   way upward, byteswapping as we go.  The destination
 *   bitmap will have the frame removed.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prSource::       Ptr to 3-channel, 8 bits per plane RGB PIXR.
 *  yOffset::        Starting line for conversion (see notes below).
 *  height::         Number of lines to convert (see notes below).
 *
 * Output:::
 *  bmpD::           Ptr to a destination bitmap (to a DIB so that (0, 0)
 *                   is in the lower left corner) to convert PIXR in to.
 *
 * Returns:::
 *
 * Assumptions:::
 *   The source (and destination) are 32 bit aligned.
 *   Each component of the RGB triple is 8 bits.
 *
 * Notes:::
 *  slice conversion::
 *   The yOffset and height parameters allow us to convert
 *   a slice of a PIXR-triplet to a Microsoft DIB.  If height is
 *   greater than 0, we interpret yOffset as the starting
 *   line (0 at the top) and height as the number of lines.
 *   If height == 0, we convert the entire image.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, multi-channel, byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_RGBPixrsToBMP (
    IPWState           *pIPWState,
    PIXR               *prSource,
    Int32               yOffset,
    UInt32              height,
    UInt8 __far        *bmpD)
{
    PROLOGUE;
    Int32        w, h;
    Int32        DIBBpl, pixrBpl;
    PIXR        *subR, *subG, *subB;
    UInt8       *pImageR, *pImageG, *pImageB;

    ERR_GEN_PIXR(prSource, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    ERR_GEN(bmpD == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetWidth(pIPWState, prSource, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prSource, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prSource, &pixrBpl));

    /* the depth of all RGB images is 24 bits/pixel, 3 * 8 */
    DIBBpl = (((w * 8 * 3) + 31) / 32) * 4;

    ERR_PASS( w_pixrGetChannel (pIPWState, prSource, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel (pIPWState, prSource, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel (pIPWState, prSource, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImage (pIPWState, subR, &pImageR));
    ERR_PASS( w_pixrGetImage (pIPWState, subG, &pImageG));
    ERR_PASS( w_pixrGetImage (pIPWState, subB, &pImageB));

    /* is height == 0?  If so, convert the entire image.  Otherwise,
     * convert height lines starting at yOffset
     */
    if (height == 0)
    {
        ERR_PASS( i_DIBImageFromTo3Pixrs (pIPWState, bmpD,
					  pImageR, pImageG, pImageB,
				          h, w, DIBBpl, pixrBpl, 
				          cPixrToDIB));
    }
    else
    {
	pImageR += yOffset * pixrBpl;
	pImageG += yOffset * pixrBpl;
	pImageB += yOffset * pixrBpl;
        /* have i_DIBImageFrom3Pixrs do the work */
        ERR_PASS( i_DIBImageFromTo3Pixrs (pIPWState, bmpD,
					  pImageR, pImageG, pImageB,
				          height, w, DIBBpl, pixrBpl, 
				          cPixrToDIB));
    }


    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a RGB 24-bit DIB BMP image to a PIXR.
 *
 * LongDescription:::
 *   Distribute an RGB color Microsoft DIB into a color PIXR.
 *   Each RGB triple is divided into 3 bytes, one going
 *   into each of the 3 sub-PIXRs.
 *   Convert (0, 0) to the lower left corner by
 *   writing to the bottom of the bitmap and working our
 *   way upward, byteswapping as we go.  The destination
 *   bitmap will have the frame removed.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  bmpS::           Ptr to a source bitmap (from a dib so that (0, 0) is in
 *                   the lower left corner.
 *  yOffset::        Starting line for conversion (see notes below).
 *  height::         Number of lines to convert (see notes below).
 *
 * Output:::
 *  prDest::         Ptr to dest 3-channel, 8 bits per plane PIXR to convert
 *                   DIB in to.  Note it must be allocated and valid.
 *
 * Returns:::
 *
 * Assumptions:::
 *   The source (and destination) are 32 bit aligned.
 *   Each component of the RGB triple is 8 bits.
 *
 * Notes:::
 *  slice conversion::
 *   The yOffset and height parameters allow us to convert
 *   a Microsoft DIB into a slice of a PIXR-triplet.  If height is
 *   greater than 0, we interpret yOffset as the starting
 *   line (0 at the top) and height as the number of lines.
 *   If height == 0, we convert the entire image.
 *
 * FunctionCategory:::
 *  format conversion
 *
 * Keywords:::
 *  format conversion, color, multi-channel, byteswapping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_BMPToRGBPixrs (
    IPWState           *pIPWState,
    UInt8 __far        *bmpS,
    Int32               yOffset,
    UInt32              height,
    PIXR               *prDest)
{
    PROLOGUE;
    Int32 	 w, h;
    Int32	 DIBBpl, pixrBpl;
    PIXR        *subR, *subG, *subB;
    UInt8       *pImageR, *pImageG, *pImageB;

    ERR_GEN_PIXR(prDest, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    ERR_GEN(bmpS == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetWidth(pIPWState, prDest, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prDest, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prDest, &pixrBpl));

    /* the depth of all RGB images is 24 bits/pixel, 3 * 8 */
    DIBBpl = (((w * 8 * 3) + 31) / 32) * 4;

    ERR_PASS( w_pixrGetChannel (pIPWState, prDest, cRedPlane, &subR));
    ERR_PASS( w_pixrGetChannel (pIPWState, prDest, cGreenPlane, &subG));
    ERR_PASS( w_pixrGetChannel (pIPWState, prDest, cBluePlane, &subB));
    ERR_PASS( w_pixrGetImage (pIPWState, subR, &pImageR));
    ERR_PASS( w_pixrGetImage (pIPWState, subG, &pImageG));
    ERR_PASS( w_pixrGetImage (pIPWState, subB, &pImageB));

    /* is height == 0?  If so, convert the entire image.  Otherwise,
     * convert height lines starting at yOffset
     */
    if (height == 0)
    {
        /* have i_DIBImageFrom3Pixrs do the work */
        ERR_PASS( i_DIBImageFromTo3Pixrs (pIPWState, bmpS,
					  pImageR, pImageG, pImageB,
				          h, w, DIBBpl, pixrBpl,
				          cDIBToPixr));
    }
    else
    {
	pImageR += yOffset * pixrBpl;
	pImageG += yOffset * pixrBpl;
	pImageB += yOffset * pixrBpl;
        /* have i_DIBImageFrom3Pixrs do the work */
        ERR_PASS( i_DIBImageFromTo3Pixrs (pIPWState, bmpS,
					  pImageR, pImageG, pImageB,
				          height, w, DIBBpl, pixrBpl,
				          cDIBToPixr));
    }
    
    NORMAL_RETURN;
}


/*

 $Log: pcconv.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.30  1998/01/14 00:14:18  ddavies
 pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.29  1996/01/04 20:25:23  ejaquith
 pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.28  1995/12/13  04:11:12  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.27  1995/11/13  23:55:41  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Fixed comment block function reference.
 *
 * Revision 1.26  1995/11/13  20:14:23  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.25  1995/10/18  04:18:20  ddavies
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Add return pointers to w_getDIBPixrLineProc to reflect the four types
 * of conversion procedures is can return (src is __far, dest is __far, both
 * are near, DIB is interleaved color).
 *
 * Revision 1.24  1995/10/13  22:03:40  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.23  1995/09/29  00:36:48  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.22  1995/09/21  22:55:11  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.21  1995/09/12  16:56:03  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.20  1995/09/08  00:15:55  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.19  1995/09/07  01:33:53  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.18  1995/09/06  20:21:58  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.17  1995/09/01  02:21:30  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.16  1995/08/31  21:15:44  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.15  1995/08/29  22:34:19  jcraig
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * in BMPToRGBPixrs and RGBPixrsToBMP, fix the case when the height
 * parameter is not 0.
 *
 * Revision 1.14  1995/08/22  02:21:31  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.13  1995/08/17  22:02:27  ejaquith
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.12  1994/10/10  22:41:04  lperry
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Remove use of pcpros.h
 *
 * Revision 1.11  1994/09/30  15:52:57  lperry
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.10  1994/04/12  01:14:33  macleod
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.9  1994/04/05  00:41:08  macleod
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Changed RGB to BMP and BMP to RGB code to take Color Pixr structures rather
 * than three separate planes
 *
 * Revision 1.8  1994/01/26  21:58:40  jcraig
 * jcraig on Wed Jan 26 13:58:04 PST 1994
 *
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * add comment to getDIBPixrLineProc that indicates direction can
 * also be cPixrToPixr.
 *
 * Revision 1.7  1994/01/22  01:39:51  ddavies
 * ddavies on Fri Jan 21 17:39:03 PST 1994
 *
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Change shrio.pub to shrpixr.pub since ipshared's libiof and libios have
 * been absorbed into libpixrf and libpixrs.
 *
 * Revision 1.6  1993/12/03  01:17:16  ddavies
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.5  1993/11/06  03:14:27  ddavies
 * pcconv.c -> /product/ipcore/ipwrap/src/RCS/pcconv.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

