/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  pixrops1.c--misc. library routines
 *
 *                 Shrinking procs:
 *                     w_rectOfOnPixels()
 *
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"

/* prototypes */
#include "props.pub"
#include "rectops.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: pixrops1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get rectangle that encloses all ON pixels in all or part of a PIXR.
 *
 * LongDescription:::
 *     Returns a new rectangle such that:
 *       (1) if input rect == NULL, new rectangle is of minimum size
 *           to enclose all ON pixels within the PIXR;
 *       (2) if input rect != NULL, new rectangle is of minimum size
 *           to enclose all ON pixels within the input rect
 *           in the PIXR;
 *     If there are no ON pixels or the new rectangle will be smaller
 *     than minSize, it returns a rectangle of size minSize, 
 *     except when minSize == 0, in which case it returns
 *     NULL.<BR><BR>
 *
 *     This procedure only makes sense for binary images.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Input PIXR to search for ON pixels.
 *  rect::           Input rect area in PIXR to search for ON pixels.
 *  minSize::        The new (destination) rectangle's width or height
 *                   must be at least this many pixels.
 *
 * Output:::
 *  ppDRect::        Ptr to ptr to new (destination) rectangle.
 *                   ppDRect is allocated and filled in.
 *                   *ppDRect must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  cropping
 *
 * Keywords:::
 *  cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rectOfOnPixels (
    IPWState           *pIPWState,
    PIXR               *pixr,
    RECTANGL           *rect,
    Int32               minSize,
    RECTANGL          **ppDRect)
{
    PROLOGUE;
    Int32          i, j, y, w, h, width, height, xorig, yorig;
    Int32          top, bottom, left, right;
    UInt8         *data, *byte;
    Int32          Bpl;
    Int32          Wpl, properWpl;
    UInt32        *wptr;
    UInt32         wval;
    PIXR          *prt = NULL;
    RECTANGL      *newRect = NULL;

    ERR_GEN(ppDRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDRect != NULL, cErrPtrNotNull);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    /*  Determine starting rectangle.
     *  Note that xorig and yorig are with respect to the
     *  (0,0) origin of the input image.   */
    if (rect == NULL) {  /* use the entire pixr */
	prt = pixr;
	xorig = 0;
	yorig = 0;
    }
    else
    {    
	/* Take only pixels within rect, but first be sure that
	 * the rect does not exceed the pixr boundaries! */
	ERR_PASS( w_clipRectToPixr(pIPWState, pixr, rect, &newRect));
	ERR_PASS_CLEANUP( w_rectOf(pIPWState, pixr, newRect, &prt));
	xorig = rect->x;
	yorig = rect->y;
	ERR_PASS_CLEANUP( w_destroyRect(pIPWState, &newRect));
    }

    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prt, &w));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prt, &h));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prt, &data));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prt, &Bpl));
    
    properWpl = (w + 31) / 32;
    Wpl = Bpl / 4;

    /*  now shrink to ON pixels */
    for (y = 0; y < h; y++)
    {
	byte = data + Bpl * y;
	wptr = (UInt32 *)byte;

	for ( j = 0; j < properWpl; ++j )
	    if ( *wptr++)
	    {
		top = y;
		goto Bottom;
	    }
    }

    /* only get here if there are zero ON pixels */
    if (minSize == 0)
    {
	if (prt != pixr) ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prt));

	*ppDRect = NULL;
	NORMAL_RETURN;
    }
    else 
	goto Minrect;

    Bottom:
	for (bottom = y = h - 1; y >= 0; y--)
	{
	    byte = data + Bpl * y;
	    wptr = (UInt32 *)byte;

	    for ( j = 0; j < properWpl; ++j )
		if ( *wptr++)
		{
		    bottom = y;
		    goto Left;
		}
	}

    /* OR the bytes columnwise; test; extract left-most bit */
    Left:
	for ( left = j = 0; j < properWpl; ++j) {

	    wptr = (UInt32 *)data + top * Wpl + j;
	    wval = 0;
	    for (y = top; y <= bottom; y++, wptr += Wpl)
		wval |= *wptr;
	    if (wval)
		for (i = 31; i >= 0; i--)
		    if ((wval >>= 1) == 0)
		    {
			left = 32 * j + i;
			goto Right;
		    }
	}
		    
		
	/* OR the bytes columnwise; test; extract right-most bit */
    Right:
	for ( j = properWpl - 1, right = j*32 + 31; j >= 0; --j ) {

	    wptr = (UInt32 *)data + top * Wpl + j;
	    wval = 0;

	    for (y = top; y <= bottom; y++, wptr += Wpl) wval |= *wptr;

	    if (wval)
		for (i = 0; i < 32; i++)
		    if ((wval <<= 1) == 0)
		    {
			right = 32 * j + i;
			goto Finish;
		    }
	}
		    
    Finish:     /* compute rectangle of ON pixels */
	width = right - left + 1;
	height = bottom - top + 1;
	if (width > minSize || height > minSize)
	{
	    ERR_PASS_CLEANUP( w_createRect(pIPWState, xorig + left, 
					   yorig + top,
					   width, height, &newRect));
	    if (prt != pixr) ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prt));
	    *ppDRect = newRect;
	    NORMAL_RETURN;
	}
	else
	    goto Minrect;


    Minrect:  /* if it gets here, it either means there were no ON pixels,
		or no dimension exceeded minSize.  In either case,
		return a minSize square, located at (xorig, yorig).  */
        ERR_PASS_CLEANUP( w_createRect(pIPWState, xorig, yorig,
				       minSize, minSize,
				       &newRect));
        if (prt != pixr) ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prt));
        *ppDRect = newRect;
        NORMAL_RETURN;

ERR_CLEANUP:
    if (prt != pixr) w_destroyPixr(pIPWState, &prt);
    w_destroyRect(pIPWState, &newRect);
    ERR_RETURN;
}

/*

 $Log: pixrops1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.18  1996/01/04 20:25:31  ejaquith
 pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.17  1995/12/13  04:11:23  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:14:30  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/10/13  22:03:49  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/09/29  00:36:29  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.13  1995/09/21  22:55:19  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/12  16:55:43  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.11  1995/09/08  00:15:41  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.10  1995/09/07  01:34:01  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:22:06  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  02:21:17  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.7  1995/08/31  21:15:52  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/17  22:02:12  ejaquith
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.5  1994/10/07  19:13:56  lperry
 * lperry on Fri Oct 7 12:12:42 PDT 1994
 *
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.4  1994/05/05  20:55:48  ddavies
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * Put this back into ipwrap.  It contains rectOfOfPixels, which is used by Paperware.  Also
 * make sure bottom, left and right are (needlessly) initialized to the
 * satisfaction of gcc.
 *
 * Revision 1.3  1993/12/03  01:17:20  ddavies
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:14:34  ddavies
 * pixrops1.c -> /product/ipcore/ipwrap/src/RCS/pixrops1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

