/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  props3.c:
 *      Includes
 *
 *	     Pixr creation/annihilation
 *			w_createPixr()
 *			w_createPixrNoInit()
 *			w_createPixrExtended()
 *			w_createPixrComposite()
 *			w_clonePixr()
 *			w_clonePixrNoInit()
 *			w_destroyPixr()
 *
 *	     Pixr creation utilities
 *			w_allocateImageData()
 *			w_reallocPixrData()
 *			w_createPixrStructOnly()
 *			w_createPixrExtendedStructOnly()
 *			w_clonePixrStructOnly()
 *
 */


#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"

/* prototypes */
#include "props.pub"
#include "shrpixr.pub"
#include "shrrast.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: props3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*--------------------------------------------------------------------*
 *                     Pixr Creation/Annihilation                     *
 *--------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a one-channel, gray-scale or binary PIXR
 *
 * LongDescription:::
 *  Basic routine for making a new Pixr and allocating memory for the image
 *  data.
 *  The image data will be initialized to zero.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value, {1, 2, 4, 8, 16, or 32}, in bits
 *  xdpi::           image dots per inch in X dimen
 *  ydpi::           image dots per inch in Y dimen
 *  frame::          frame value, in pixels
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR.
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  frame must be 0 or 32 pixels
 *  depth must be 1, 2, 4, 8, 16 or 32
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixr (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    PIXR              **ppPixr)
{
    PROLOGUE;
    PIXR          *prd = NULL;
    UInt8         *data = NULL;

    /* frame must be either cFrameBits or 0 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

	/* allocate and initialize the image data block */
    ERR_PASS( w_allocateImageData(pIPWState, width, height, depth,
				  frame, cAllocatorCalloc,
				  &data));

    ERR_PASS_CLEANUP( w_createPixrStructOnly(pIPWState, width, height,
					     depth, xdpi, ydpi, frame,
					     data, &prd));
    *ppPixr = prd;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(data, cAllocatorImage);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a one-channel, gray-scale or binary PIXR with unitialized data.
 *
 * LongDescription:::
 *  Special routine for making a new PIXR and allocating memory for the image,
 *  where the image data is NOT initialized to 0.  (It will be, in fact,
 *  completely unitialized).
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value, {1, 2, 4, 8, 16, or 32}, in bits
 *  xdpi::           image dots per inch in X dimen
 *  ydpi::           image dots per inch in Y dimen
 *  frame::          frame value, in pixels
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR.
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  frame must be 0 or 32 pixels
 *  depth must be 1, 2, 4, 8, 16 or 32
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixrNoInit (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    PIXR              **ppPixr)
{
    PROLOGUE;
    PIXR          *prd = NULL;
    Int32	   bpl;
    UInt8         *data = NULL;

    /* frame must be either cFrameBits or 0 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

	/* allocate and initialize the image data block */
    ERR_PASS( w_allocateImageData(pIPWState, width, height, depth,
				  frame, cAllocatorMalloc,
				  &data));

    ERR_PASS_CLEANUP( w_createPixrStructOnly(pIPWState, width, height,
					     depth, xdpi, ydpi, frame,
					     data, &prd));

	/* while we don't initialize the data, we do need to
	 * initialize the frames, if there is one.  Note that we need to get
	 * the pointer to the beginning of the image again since
	 * we don't know whether there's an offset or not.
	 */
    if (frame == cFrameBits)
    {
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prd, &data));
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prd, &width));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prd, &height));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prd, &bpl));
    
	ERR_PASS_CLEANUP( i_clearAllFrames (pIPWState, data,
					    width, height, bpl, depth));
    }

    *ppPixr = prd;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(data, cAllocatorImage);
    if (prd != NULL)
	IPW_FREE(prd, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create an initialized binary, gray, or multi-plane color PIXR.
 *
 * LongDescription:::
 *  Basic routine for making a new "extended" Pixr and allocating memory for
 *  the image data.  Extended Pixrs can have one or more image planes.
 *  Normal Pixrs have one image plane.<BR><BR>
 *  The image data will be initialized to zero.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value of each plane, {1, 2, 4, 8, 16, or 32}, 
 *                   in bits
 *  xdpi::           image dots per inch in X dimen
 *  ydpi::           image dots per inch in Y dimen
 *  frame::          frame value, in pixels
 *  channels::       Number of image planes
 *  type::           Color type of image, {cColorRGB, cColorCMYK,
 *                   cColorCIELAB, cColorCIELUV, cColorGray, or cColorNone} 
 *  map::            Ptr to color map, or NULL if none
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  depth must be one of 1, 2, 4, 8, 16 or 32, frame must be 0 or cFrameBits
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixrExtended (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Int32               channels,
    ColorType           type,
    ColorMap           *map,
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32          nDone = 0;
    PIXR          *prd = NULL;
    UInt8	  *channelData[cMaxColorPlanes];

    /* frame must be 0 or 32 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

	/* clear all the image pointers we we know what to
	 * free if there's an error later.
	 */
    for (nDone = 0; nDone < cMaxColorPlanes; nDone++)
    {
	channelData[nDone] = NULL;
    }

	/* allocate all the necessary image planes */
    for (nDone = 0; nDone < channels; nDone++)
    {
	ERR_PASS_CLEANUP( w_allocateImageData(pIPWState, width, height,
					      depth, frame, cAllocatorCalloc,
					      &(channelData[nDone])));
    }

	/* go attach structures to all of these */
    ERR_PASS_CLEANUP(
	w_createPixrExtendedStructOnly(pIPWState, width, height, depth,
				       xdpi, ydpi, frame, channels, type,
				       map, channelData, &prd));

    *ppPixr = prd;
    NORMAL_RETURN;
    

ERR_CLEANUP:
    /* Free the sub-channels already allocated (if multi-channel pixr)*/
    for (nDone = 0; nDone < cMaxColorPlanes; nDone++)
    {
	if (channelData[nDone] != NULL)
	    IPW_FREE(channelData[nDone], cAllocatorImage);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create an uninitialized binary, gray, or multi-plane color PIXR.
 *
 * LongDescription:::
 *  Basic routine for making a new "extended" Pixr and allocating memory for
 *  the image data.  Extended Pixrs can have one or more image planes.
 *  Normal Pixrs have one image plane.<BR><BR>
 *  The image data will be uninitialized.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value of each plane, {1, 2, 4, 8, 16, or 32}, 
 *                   in bits
 *  xdpi::           image dots per inch in X dimen
 *  ydpi::           image dots per inch in Y dimen
 *  frame::          frame value, in pixels
 *  channels::       Number of image planes
 *  type::           Color type of image, {cColorRGB, cColorCMYK,
 *                   cColorCIELAB, cColorCIELUV, cColorGray, or cColorNone} 
 *  map::            Ptr to color map, or NULL if none
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  frame must be 0 or cFrameBits, depth must be one of 1, 2, 4, 8, 16 or 32.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixrExtendedNoInit (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Int32               channels,
    ColorType           type,
    ColorMap           *map,
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32          nDone = 0;
    PIXR          *prd = NULL;
    PIXR	  *prChannel = NULL;
    UInt8	  *data;
    Int32	   bpl;
    UInt8	  *channelData[cMaxColorPlanes];

    /* frame must be either cFrameBits or 0 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

	/* clear all the image pointers we we know what to
	 * free if there's an error later.
	 */
    for (nDone = 0; nDone < cMaxColorPlanes; nDone++)
    {
	channelData[nDone] = NULL;
    }

	/* allocate all the necessary image planes */
    for (nDone = 0; nDone < channels; nDone++)
    {
	ERR_PASS_CLEANUP( w_allocateImageData(pIPWState, width, height,
					      depth, frame, cAllocatorMalloc,
					      &(channelData[nDone])));
    }

	/* go attach structures to all of these */
    ERR_PASS_CLEANUP(
	w_createPixrExtendedStructOnly(pIPWState, width, height, depth,
				       xdpi, ydpi, frame, channels, type,
				       map, channelData, &prd));

	/* Clear the frames in the planes, if any */
    if (frame == cFrameBits)
    {
	for (nDone = 0; nDone < channels; nDone++)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prd, nDone,
					       &prChannel));
	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prChannel, &data));
	    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prChannel, &width));
	    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prChannel,
						 &height));
	    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prChannel, &bpl));
    
	    ERR_PASS_CLEANUP( i_clearAllFrames (pIPWState, data, width, height,
					        bpl, depth));
	}
    }

    *ppPixr = prd;
    NORMAL_RETURN;
    

ERR_CLEANUP:
    /* Free the sub-channels already allocated (if multi-channel pixr)*/
    for (nDone = 0; nDone < cMaxColorPlanes; nDone++)
    {
	if (channelData[nDone] != NULL)
	    IPW_FREE(channelData[nDone], cAllocatorImage);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a three channel/plane color PIXR from three gray PIXRs.
 *
 * LongDescription:::
 *  This is used to merge 3 gray PIXRs into a color PIXR.  NOTE that the
 *  PIXRs become channel pixrs of the new pixr, so if the new pixr is
 *  destroyed, the CHANNEL PIXRS WILL ALSO BE DESTROYED. Caveat emptor.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  planes::         An array of pointers to the PIXRS that will be
 *		     assembled into the composite PIXR.
 *  numPlanes::	     number of valid pointers in the planes array.  The
 *		     first numPlanes pointers are valid.
 *  type::           Color type of image, {cColorRGB, cColorCMYK,
 *                   cColorCIELAB, cColorCIELUV, cColorGray, or cColorNone} 
 *                   to be forced upon the planes.
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixrComposite (
    IPWState           *pIPWState,
    PIXR               *planes[cMaxColorPlanes],
    Int32		numPlanes,
    ColorType           type,
    PIXR              **ppPixr)
{
    PROLOGUE;
    PIXR		*result = NULL;
    ScaledInt1616	 xdpi, ydpi;
    Int32		 i;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(*ppPixr != NULL, cErrPtrNotNull);
    ERR_GEN(numPlanes <= 0, cErrInvalidParm);

    ERR_GEN(type != cColorRGB     &&  type != cColorCMYK    &&
	    type != cColorCIELAB  &&  type != cColorCIELUV  &&
	    type != cColorGray    &&  type != cColorNone, cErrInvalidParm);

    result = (PIXR *)IPW_CALLOC(1, sizeof(struct Pixr), cAllocatorData);
    ERR_GEN(result == NULL, cErrNomem);

    ERR_PASS_CLEANUP( w_pixrGetResolutionDPI(pIPWState, planes[0],
					     &xdpi, &ydpi));
    result->channels = (UInt16) numPlanes;
    for (i = 0; i < numPlanes; i++)
    {
	result->pixrData.childPlanes[i] = planes[i];
	(result->pixrData.childPlanes[i])->pixrData.plane.type = type;
	(result->pixrData.childPlanes[i])->pixrData.plane.xRes = xdpi;
	(result->pixrData.childPlanes[i])->pixrData.plane.yRes = ydpi;
    }

    *ppPixr = result;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(result, cAllocatorData);
    ERR_RETURN;
}   


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clone a PIXR -- create initialized PIXR with same dimensions.
 *
 * LongDescription:::
 *  Another basic routine for making a new Pixr and allocating memory for the
 *  image data, by using an existing PIXR as its template.
 *  <BR><BR>
 *  The image data is initialized to 0.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR to clone.  May be multi-channel.
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clonePixr (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32		 w, h, d, frame, channels;
    ColorType		 type;
    ColorMap		*map;
    ScaledInt1616	 xdpi, ydpi;
    PIXR		*pPixr = NULL;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(*ppPixr != NULL, cErrPtrNotNull);    
    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &d));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));
    ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &frame));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, pixr, &channels));
    ERR_PASS( w_pixrGetType(pIPWState, pixr, &type));
    ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &map));

    ERR_PASS( w_createPixrExtended(pIPWState, w, h, d, xdpi, ydpi,
				   frame, channels, type, map, &pPixr));

    *ppPixr = pPixr;
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clone a PIXR -- make a new PIXR of same dimensions, data is uninitialized.
 *
 * LongDescription:::
 *  Another basic routine for making a new Pixr and allocating memory for the
 *  image data, by using an existing PIXR as its template.
 *  <BR><BR>
 *  The image data will be uninitialized.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR to clone.  May be multi-channel.
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   ppPixr is allocated and filled in.
 *                   *ppPixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clonePixrNoInit (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32		 w, h, d, frame, channels;
    ColorType		 type;
    ColorMap		*map;
    ScaledInt1616	 xdpi, ydpi;
    PIXR		*pPixr = NULL;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(*ppPixr != NULL, cErrPtrNotNull);    
    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &d));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));
    ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &frame));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, pixr, &channels));
    ERR_PASS( w_pixrGetType(pIPWState, pixr, &type));
    ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &map));

    ERR_PASS( w_createPixrExtendedNoInit(pIPWState, w, h, d, xdpi, ydpi,
					 frame, channels, type, map, &pPixr));

    *ppPixr = pPixr;
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free PIXR and set ptr to NULL.
 *
 * LongDescription:::
 *  If the PIXR is already free and its ptr points to NULL, do nothing.
 *  O.w. free the image data or unmap the image data file.  Free the PIXR
 *  structure.  If there is a color map, do NOT free it.
 *  Reset the PIXR ptr to point to NULL.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to PIXR to be destroyed.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyPixr (
    IPWState           *pIPWState,
    PIXR              **ppPixr)
{
    PROLOGUE;
    UInt8     *data = NULL;
    Int32      i, channels;
    Plane     *pCurPlane;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);

    if (*ppPixr == NULL)
	/* Already freed.  Just exit and be happy. */
	NORMAL_RETURN;

    ERR_PASS_CLEANUP( w_pixrGetChannelCount(pIPWState, *ppPixr, &channels));

    if (channels == 1)
    {
	/* Get a pointer to the memory buffer containing the image data
	 * so we can free it.
	 */
	ERR_PASS_CLEANUP( wp_pixrGetImageBuffer(pIPWState, *ppPixr, &data));

	/* If this plane was the only member of the ring of owners
	 * of this image data block, free the image data.  Otherwise,
	 * just take this plane out of the ring.
	 */
	pCurPlane = &((*ppPixr)->pixrData.plane);
	if (pCurPlane->pPrevOwner == pCurPlane)
	    IPW_FREE(data, cAllocatorImage);
	else
	{
	    pCurPlane->pPrevOwner->pNextOwner = pCurPlane->pNextOwner;
	    pCurPlane->pNextOwner->pPrevOwner = pCurPlane->pPrevOwner;
	}
	if (pCurPlane->colorMap) {
	    if (pCurPlane->colorMap->map) {
		IPW_FREE(pCurPlane->colorMap->map, cAllocatorData);
	    }
	    IPW_FREE(pCurPlane->colorMap, cAllocatorData);
	}

    }
    else
    {
	PIXR  *ch = NULL;
	/* free the children */
	for (i = 0; i < channels; i++)
	{
	    /* No error return from this, so, we can't tell if it failed */
	    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, *ppPixr, i, &ch));
	    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &ch));
		/* make sure this pointer is null so that none of the
		 * funky checking macros will look at it.  This is especially
		 * important if we've just deleted the primary channel.
		 */
	    ERR_PASS_CLEANUP( w_pixrSetChannel(pIPWState, *ppPixr, i, NULL));
	}
    }

    /* remove the pixr */
    IPW_FREE(*ppPixr, cAllocatorData);
    *ppPixr = NULL; /* Just to be sure */
    NORMAL_RETURN;

ERR_CLEANUP:
    /* We are in some serious trouble if our cleaning routines are dying
     * above.  What can you do?  Well, just free the PIXR struct since
     * we can't get to the data/channels for some reason!
     */
    IPW_FREE(*ppPixr, cAllocatorData);
    *ppPixr = NULL; /* Just to be sure */

    ERR_RETURN;
}



/*--------------------------------------------------------------------*
 *                Pixr Creation/Annihilation Utilities                *
 *--------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate the memory block needed to hold an image.
 *
 * LongDescription:::
 *  A PIXR is a structure that contains, among other things, a pointer
 *  to a block of memory that holds the image pixels.  This routine
 *  allocates a memory block sufficient to hold all the pixels.  The
 *  width, height, depth, frame, horizontal and vertical padding are
 *  taken into account.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value of each plane, {1, 2, 4, 8, 16, or 32}, 
 *                   in bits
 *  frame::          frame value, in pixels
 *  allocatorFlavor:: When set to cAllocatorMalloc, IPW_MALLOC is used to
 *		     allocate the pixel buffer.
 *		     When set to cAllocatorCalloc, IPW_CALLOC is used.
 *		     When set to cAllocatorRealloc, IPW_REALLOC is used.
 *		     When allocatorFlavor is not set to cAllocatorRealloc,
 *		     *ppImage must be NULL on entry.  When allocatorFlavor
 *		     is set to cAllocatorRealloc, *ppImage must NOT be NULL
 *		     on entry.
 *
 * Output:::
 *  ppImage::        Pointer to pointer to pixel data.
 *
 * Returns:::
 *  Returns cErrNomem if insufficient memory is available.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_allocateImageData (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    Int32               frame,
    Int32		allocatorFlavor,
    UInt8             **ppImage)
{
    PROLOGUE;
    UInt8         *data = NULL;
    Int32	   bufSize;
    Int32	   bpl;

    ERR_GEN(ppImage == NULL, cErrNullPtr);
    ERR_GEN((allocatorFlavor != cAllocatorRealloc) &&
	    (*ppImage != NULL), cErrPtrNotNull);
    ERR_GEN((allocatorFlavor == cAllocatorRealloc) &&
	    (*ppImage == NULL), cErrNullPtr);
    ERR_GEN((allocatorFlavor != cAllocatorMalloc) &&
	    (allocatorFlavor != cAllocatorCalloc) &&
	    (allocatorFlavor != cAllocatorRealloc), cErrInvalidParm);

    /* frame must be either cFrameBits or 0 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

	/* calculate the total buf size of the pixr.  This includes the
	 * height padded out so that height*depth is a multiple of
	 * 32 (so we are able to rotation this image by 90 degrees
	 * if we want to) and the frame.
	 */
    ERR_PASS( w_pixrImageProperties(pIPWState, width, height, depth,
				    frame, &bpl, &bufSize));

    switch (allocatorFlavor)
    {
    case cAllocatorMalloc:
	data = (UInt8 *)IPW_MALLOC(bufSize, cAllocatorImage);
	break;
    case cAllocatorCalloc:
	data = (UInt8 *)IPW_CALLOC(1, bufSize, cAllocatorImage);
	break;
    default:	/* cAllocatorRealloc */
	data = (UInt8 *)IPW_REALLOC(*ppImage, bufSize, cAllocatorImage);
	break;
    }

    ERR_GEN(data == NULL, cErrNomem);

	/* it worked, so update the client's pointers and return */
    *ppImage = data;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reallocate the PIXR data to fit to the current size
 *
 * LongDescription:::
 *  Calls a realloc routine to change the amount of data associated with a
 *  pixr.  Uses the pixr's current width, height, frame, and depth to calculate
 *  the amount of data needed.  If called with a multi-plane pixr, reallocates
 *  the image blocks for each plane.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  prS::            Ptr to PIXR whose data gets resized in place
 *
 * Returns:::
 *
 * Notes:::
 *  If this function runs out of memory after reallocating one or more
 *  planes of a multi-plane image, but before successfully allocating
 *  the last plane, the planes that were successfully allocated will
 *  retain the new size.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reallocPixrData (
    IPWState           *pIPWState,
    PIXR               *prS)
{
    PROLOGUE;
    UInt8	*data = NULL;
    Int32	 width, height, depth, bpl, frame, channels, curChannelNum;
    Int32	 offsetTo00;
    PIXR	*curChannel;
    Plane	*pThisPlane, *pNextPlane;

    ERR_GEN_PIXR(prS, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &width));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &height));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frame));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

	/* realloc the data for each channel in a multi-plane image */
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));

    for (curChannelNum = 0; curChannelNum < channels; curChannelNum++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, prS, curChannelNum, &curChannel));
	ERR_PASS( wp_pixrGetImageBuffer(pIPWState, curChannel, &data));

	ERR_PASS( w_allocateImageData(pIPWState, width, height, depth,
				      frame, cAllocatorRealloc,
				      &data));
	ERR_PASS( wp_pixrSetImageBuffer(pIPWState, curChannel, data));

	    /* calculate the offset to 00 for this data block.  It will
	     * be used to set the pointer to the upper left corner of
	     * the frame.
	     */
	ERR_PASS( w_pixrImageOffset(pIPWState, data, bpl, depth, frame,
				    &offsetTo00));
	ERR_PASS( w_pixrSetImageAll(pIPWState, curChannel,
				    data + offsetTo00 -
				    (frame * depth)/8 -
				    frame * bpl));
    
	    /* If any other pixr planes were pointing to this image data,
	     * change their pointers too.
	     */
	for (pThisPlane = &(curChannel->pixrData.plane),
		pNextPlane = curChannel->pixrData.plane.pNextOwner;
	     pNextPlane != pThisPlane;
	     pNextPlane = pNextPlane->pNextOwner)
	{
	    pNextPlane->image = data;
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create the structure for a one-channel, gray-scale or binary PIXR
 *
 * LongDescription:::
 *  Some users wish to create an image buffer somewhere, then convert it
 *  into a PIXR by creating the structure part of a PIXR and pointing to
 *  the existing image data.  This routine does that.
 *
 *  It is assumed that the image data is in the form of an IPWrap image,
 *  i.e. the byte order is correct, it has a frame and so on.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value, {1, 2, 4, 8, 16, or 32}, in bits
 *  xdpi::           image dots per inch in X dimen
 *  ydpi::           image dots per inch in Y dimen
 *  frame::          frame value, in pixels
 *  pImage::	     pointer to the beginning of the buffer containing
 *		     the image.
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   If *ppPixr is NULL, a PIXR is allocated and filled in.<BR>
 *		     If *ppPixr is not NULL, we assume it already points to
 *		     a PIXR structure.  This existing structure is used
 *		     and its values filled in as before.
 *
 * Returns:::
 *
 * Notes:::
 *  frame must be 0 or 32 pixels
 *  depth must be 1, 2, 4, 8, 16 or 32
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixrStructOnly (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    UInt8	       *pImage,
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32          bpl, bufSize;
    Int32	   offsetTo00;
    PIXR          *prd = NULL;

    ERR_GEN(ppPixr == NULL || pImage == NULL, cErrNullPtr);

    /* frame must be either cFrameBits or 0 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

    prd = *ppPixr;
    if (prd == NULL)
	prd = (struct Pixr *)IPW_CALLOC(1, sizeof(struct Pixr),
					cAllocatorData);
    ERR_GEN(prd == NULL, cErrNomem);
    
    /* ALWAYS set the channel count first */
    prd->channels = 1;
    prd->pixrData.plane.w = width;
    prd->pixrData.plane.h = height;
    prd->pixrData.plane.d = depth;
    prd->pixrData.plane.xRes = xdpi;
    prd->pixrData.plane.yRes = ydpi;
    prd->pixrData.plane.frame = frame;
    prd->pixrData.plane.colorMap = NULL;
    prd->pixrData.plane.type = cColorGray;
    prd->pixrData.plane.byteOrder = cBigEndianByteOrder;
    prd->pixrData.plane.imageBuffer = pImage;

	/* get the properties of the image so we can set the
	 * bytes/line and the location of the image's upper left
	 * corner.
	 */
    ERR_PASS_CLEANUP( w_pixrImageProperties(pIPWState, width,
					    height, depth, frame,
					    &bpl, &bufSize));
    prd->pixrData.plane.bpl = bpl;
    ERR_PASS_CLEANUP( w_pixrImageOffset(pIPWState, pImage, bpl, depth, frame,
					&offsetTo00));
    prd->pixrData.plane.image = pImage +
				offsetTo00 -		/* down to (0, 0) */
				(frame * depth)/8 -	/* back to left edge */
				(frame * bpl);		/* up to top */

	/* declare this plane to the be sole owner of this image data */
    prd->pixrData.plane.pPrevOwner = &(prd->pixrData.plane);
    prd->pixrData.plane.pNextOwner = &(prd->pixrData.plane);

    *ppPixr = prd;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(prd, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create the structures for a binary, gray, or multi-plane color PIXR.
 *
 * LongDescription:::
 *  It is sometimes useful to create a PIXR that points to an existing
 *  block of image data.  <w_createPixrStructOnly>() does this for a
 *  single channel (gray or binary) image.  w_createPixrExtendedStructOnly
 *  does the same thing for a multi-channel image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          image width value, in pixels
 *  height::         image height value, in pixels
 *  depth::          depth value of each plane, {1, 2, 4, 8, 16, or 32}, 
 *                   in bits
 *  xdpi::           image dots per inch in X dimen
 *  ydpi::           image dots per inch in Y dimen
 *  frame::          frame value, in pixels
 *  channels::       Number of image planes
 *  type::           Color type of image, {cColorRGB, cColorCMYK,
 *                   cColorCIELAB, cColorCIELUV, cColorGray, or cColorNone} 
 *  map::            Ptr to color map, or NULL if none
 *  channelData::    This is an array of pointers to image blocks.  These are
 *		     the memory buffers containing the images as opposed to
 *		     the upper left corner of the images themselves.  These
 *		     may be different due to alignment constraints.
 *		     The array has cMaxColorPlanes elements.  Only the
 *		     first "channels" (see arg above) elements are accessed.
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   If *ppPixr is NULL, a PIXR is allocated and filled in.<BR>
 *		     If *ppPixr is not NULL, we assume it already points to
 *		     a PIXR structure.  This existing structure is used
 *		     and its values filled in as before.
 *
 * Returns:::
 *
 * Notes:::
 *  frame must be 0 or 32 pixels
 *  depth must be 1, 2, 4, 8, 16 or 32
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createPixrExtendedStructOnly (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Int32               channels,
    ColorType           type,
    ColorMap           *map,
    UInt8	       *channelData[cMaxColorPlanes],
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32          bpl, bufSize, offsetTo00;
    Int32	   nDone = 0, j = 0;
    PIXR          *prd = NULL, *subPixr = NULL;
    UInt8	  *localChannelData[cMaxColorPlanes];

    ERR_GEN(ppPixr == NULL, cErrNullPtr);

    /* frame must be either cFrameBits or 0 */
    ERR_GEN((frame != cFrameBits) && (frame != 0), cErrInvalidParm);

    /* depth must be {1, 2, 4, 8, 16, 32} */
    ERR_GEN(((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8) && 
	     (depth != 16) && (depth != 32)), cErrInvalidParm);

    prd = *ppPixr;
    if (prd == NULL)
	prd = (struct Pixr *)IPW_CALLOC(1, sizeof(struct Pixr),
					cAllocatorData);
    ERR_GEN(prd == NULL, cErrNomem);
    
    ERR_GEN(type != cColorRGB     &&  type != cColorCMYK    &&
	    type != cColorCIELAB  &&  type != cColorCIELUV  &&
	    type != cColorGray    &&  type != cColorNone, cErrInvalidParm);

    /* ALWAYS set the channel count first */
    prd->channels = (UInt16) channels;
    
    /* If this PIXR has more than one channel, make the channels recursivly */
    if (channels > 1)
    {
	/* clear the unused pointers in localChannelData */
	for (nDone = 1; nDone < cMaxColorPlanes; nDone++)
	    localChannelData[nDone] = NULL;

	/* Create the sub-PIXRs */

	/* NOTE: Done with funny 'nDone loop' for easier cleanup on error */
	nDone = 0;
	while (nDone < channels)
	{
	    subPixr = NULL;
	    localChannelData[0] = channelData[nDone];
	    ERR_PASS_CLEANUP(
		w_createPixrExtendedStructOnly(pIPWState, width, height,
					       depth, xdpi, ydpi, frame,
					       1, type, map, 
					       localChannelData,
					       &subPixr));
	    nDone++;

	    prd->pixrData.childPlanes[nDone-1] = subPixr;
	}
    }
    else
    {
	/* Make the image plane for this PIXR */
	prd->pixrData.plane.w = width;
	prd->pixrData.plane.h = height;
	prd->pixrData.plane.d = depth;
	prd->pixrData.plane.xRes = xdpi;
	prd->pixrData.plane.yRes = ydpi;
	prd->pixrData.plane.frame = frame;
	prd->pixrData.plane.colorMap = map;
	prd->pixrData.plane.type = type;
	prd->pixrData.plane.byteOrder = cBigEndianByteOrder;
	prd->pixrData.plane.imageBuffer = channelData[0];

	    /* get the properties of the image so we can set the
	     * bytes/line and the location of the image's upper left
	     * corner.
	     */
	ERR_PASS_CLEANUP( w_pixrImageProperties(pIPWState, width,
						height, depth, frame,
						&bpl, &bufSize));
	prd->pixrData.plane.bpl = bpl;
	ERR_PASS_CLEANUP( w_pixrImageOffset(pIPWState, channelData[0],
					    bpl, depth, frame,
					    &offsetTo00));
	prd->pixrData.plane.image = channelData[0] +
				    offsetTo00 -	/* down to (0, 0) */
				    (frame * depth)/8 -	/* back to left edge */
				    (frame * bpl);	/* up to top */

	    /* declare this plane to the be sole owner of this image data */
	prd->pixrData.plane.pPrevOwner = &(prd->pixrData.plane);
	prd->pixrData.plane.pNextOwner = &(prd->pixrData.plane);
    }

    *ppPixr = prd;
    NORMAL_RETURN;
    

ERR_CLEANUP:
    /* Free the sub-channels already allocated (if multi-channel pixr)*/
    for (j = 0; j < nDone; j++)
    {
        subPixr = NULL;
	w_pixrGetChannel(pIPWState, prd, j, &subPixr);
	w_destroyPixr(pIPWState, &subPixr);
    }
    
    IPW_FREE(prd, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a copy of an existing PIXR structure.
 *
 * LongDescription:::
 *  Very occasionally, clients will want two PIXRs to point to the same image
 *  data.  This function does that.  It adds the Plane of the PIXR to
 *  the ring of owners of that image data block.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR to clone.  May be multi-channel.
 *
 * Output:::
 *  ppPixr::         Ptr to ptr to created PIXR
 *                   If *ppPixr is NULL, a structure is allocated and filled
 *		     in.  Both the new structure and pixr will own the
 *		     image data.<BR>
 *		     If *ppPixr is not NULL, we assume it points to an
 *		     existing PIXR structure.  The contents of this structure
 *		     are overwritten with copies of data from pixr.  As
 *		     before, both the new and old PIXRs will own the image
 *		     data.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clonePixrStructOnly (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppPixr)
{
    PROLOGUE;
    Int32		 i, w, h, d, frame, channels;
    ColorType		 type;
    ColorMap		*map;
    ScaledInt1616	 xdpi, ydpi;
    PIXR		*oldChannel, *newChannel;
    UInt8		*channelData[cMaxColorPlanes];
    PIXR		*pPixr = NULL;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &d));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));
    ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &frame));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, pixr, &channels));
    ERR_PASS( w_pixrGetType(pIPWState, pixr, &type));
    ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &map));

    for (i = 0; i < channels; i++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &oldChannel));
	ERR_PASS( w_pixrGetImageAll(pIPWState, oldChannel, &(channelData[i])));
    }
    for (; i < cMaxColorPlanes; i++)
	channelData[i] = NULL;

    pPixr = *ppPixr;
    ERR_PASS( w_createPixrExtendedStructOnly(pIPWState, w, h, d, xdpi, ydpi,
				             frame, channels, type, map,
					     channelData, &pPixr));

	/* Now we add the planes of the new structures to the rings of
	 * owners of the old image data.  Each new plane is added
	 * immediately after the old plane in the ring.
	 * Note that we can't really fail at this point.
	 * If pPixr exists at all, we'll get the right number of channels.
	 * That's why we don't use ERR_PASS_CLEANUP.
	 * If we did try to clean up before completing this loop,
	 * what would we clean up?  Trying to delete pPixr would cause us
	 * to delete at least some of the image data, and that would be
	 * a disaster.
	 * The image data would be deleted because we haven't yet
	 * marked the plane as being only one of the owners of the
	 * data.
	 */
    for (i = 0; i < channels; i++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &oldChannel));
	ERR_PASS( w_pixrGetChannel(pIPWState, pPixr, i, &newChannel));
	newChannel->pixrData.plane.pNextOwner =
			oldChannel->pixrData.plane.pNextOwner;
	oldChannel->pixrData.plane.pNextOwner =
			&(newChannel->pixrData.plane);
	newChannel->pixrData.plane.pPrevOwner =
			&(oldChannel->pixrData.plane);
	newChannel->pixrData.plane.pNextOwner->pPrevOwner =
			&(newChannel->pixrData.plane);
    }

    *ppPixr = pPixr;
    NORMAL_RETURN;
}



/*

 $Log: props3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.54  1998/01/14 00:14:25  ddavies
 props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.53  1997/11/01 21:02:52  aber
 aber on Sat Nov 1 13:02:42 PST 1997

 props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v

 Free the colormap attached to a PIXR.

 Revision 1.52  1997/04/15 22:23:10  ddavies
 props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v

 Change check on frame and depth on all PIXR-creating procedures.  The
 old test said frame * depth must be multiple of 32.  Well, we really don't
 support that.  The new test says frame must be cFrameBits or 0.  Depth
 must still be a power of 2 and less than or equal to 32.
 Also, initialize frames in the NoInit routines only if there is one.
 Add new routines w_createPixrExtendedNoInit and w_clonePixrNoInit.
 These are like their initialized counterparts except only the frame is
 initialized in these versions.  Also change the name of
 w_clonePixrZero to w_clonePixr to be consistent.

 Revision 1.51  1996/12/19 23:11:54  ddavies
 ddavies on Thu Dec 19 15:09:37 PST 1996

 props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v

 Have w_createPixr, w_createPixrNoInit, w_createPixrExtended and
 w_allocateImageData check the image depth before allocating the
 image memory block.

 Revision 1.50  1996/09/13 23:30:44  ddavies
 props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v

 Modify the routines used to create and destroy PIXRs so that they use
 w_pixrImageProperties and w_pixrImageOffset (new routines).  Modify
 w_allocateImageData and w_reallocPixrData to use the new routines also.
 Set the imageBuffer PIXR entry instead of the "image" entry of the
 PIXR when setting the location of the memory image buffer.
 Calculate the "image" entry by using the offset from the beginning of
 the buffer to location (0, 0), then subtracting off the frame.
 None of these changes should be visible to clients.

 Revision 1.49  1996/01/04 20:25:41  ejaquith
 props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.48  1995/12/13  04:11:36  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.47  1995/11/13  20:14:41  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.46  1995/11/03  22:48:40  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.45  1995/11/03  02:31:02  ddavies
 * ddavies on Thu Nov 2 18:30:22 PST 1995
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add cast to make Visual C++ happy.
 *
 * Revision 1.44  1995/11/01  22:21:56  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Change w_createPixrComposite so that it takes an array of pointers to PIXR
 * planes and the number of valid pointers rather than taking three dedicated
 * pointers.  This allows us to compose images with any number of planes up
 * to cMaxColorPlanes (currently defined to be 4).
 *
 * Revision 1.43  1995/10/28  01:41:31  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Big Change.  All of the raster0p-ish things like w_invert and w_rectOf
 * have been moved to the new file props6.c.  The only stuff left here is
 * used to create and destroy PIXRs.
 * Added the new routines w_allocateImageData (allocates the right size
 * memory block to hold an array of pixels of given width, height, depth, frame
 * and necessary padding), w_createPixrStructOnly (creates a PIXR struct and
 * makes it point to an existing data block), w_createPixrExtendedStructOnly
 * (ditto, except can be used on multi-plane images), w_clonePixrStructOnly
 * (creates a copy of a PIXR structure pointing to the same image data as
 * the original).
 * Also modified w_createPixr, w_createPixrNoInit and w_createPixrExtended
 * to use w_allocateImageDat and the appropriate w_createPixrXXXStructOnly
 * routine instead of duplicating that code.  Same thing for w_clonePixrZero.
 * Finally, modified w_destroyPixr so that it only frees image data if the
 * last plane that owned it is being freed.  This is determined by the
 * fact that the planes owning a block of image data are connected in a
 * ring.
 *
 * Revision 1.42  1995/10/13  22:03:58  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.41  1995/10/12  04:27:44  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.40  1995/10/02  19:20:06  jcraig
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * w_invert now takes a pointer to a pixr pointer. if *dpr ==spr, the
 * operation is done in place.  if *dpr == NULL, a destination pixr is
 * allocated.  otherwise, the destination pixr is used
 *
 * Revision 1.39  1995/09/29  00:36:14  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.38  1995/09/21  22:58:15  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.37  1995/09/19  03:22:15  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * To be truly paranoid in w_destroyPixr, set the pixr's pointer to a deleted
 * channel to be NULL.  This will catch any attempts to reference this channel
 * after it's gone.  Can be helpful if the deleted channel was the primary
 * channel and some helpful sanity check thinks it has to see the primary channel.
 *
 * Revision 1.36  1995/09/14  01:45:22  ddavies
 * ddavies on Wed Sep 13 18:44:42 PDT 1995
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Changed all the resolutions to ScaledInt1616's.
 *
 * Revision 1.35  1995/09/13  00:23:58  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Made PIXR creation routines use PIXR fields directly.
 * Made PIXR accessor public functions use ERR_GEN_PIXR()
 * (and hence wp_pixrGetFlags())
 * Fixed recursive tricky stuff.
 *
 * Revision 1.34  1995/09/12  16:55:34  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.33  1995/09/08  00:15:35  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.32  1995/09/07  01:34:09  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.31  1995/09/06  20:22:16  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.30  1995/09/01  02:21:12  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.29  1995/08/31  21:34:21  ejaquith
 * ejaquith on Thu Aug 31 14:34:20 PDT 1995
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.28  1995/08/30  17:26:17  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * In createPixr() createPixrT() createPixrExtended() createPixrComposite()
 * createPixrNoInit() and rectOfPixr(), these routines now check if the
 * dest ptr to ptr to pixr is NULL, which it is supposed to be.  If it's not,
 * it will return cErrPtrNotNull error code.  This will help in memory leaks
 * and whatnot, and in cleanup logic.  We decided an error was better than
 * just assuming the ptr was valid and over/writing the "pixr" because this
 * could cause weirder effects later than a pure honest to goodness error.
 *
 * Revision 1.27  1995/08/24  02:55:42  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * In createPixrExtended(), fixed bug where was assigning the multiple channels
 * to array indices one more than they should have been.
 *
 * Revision 1.26  1995/08/22  02:21:39  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.25  1995/08/17  22:02:06  ejaquith
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.24  1995/02/03  01:36:37  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Change createPixrComposite, rectOfPixr and createPixrT so they pass on
 * input resolutions to output PIXRs created.
 *
 * Revision 1.23  1994/12/09  22:05:54  macleod
 * macleod on Fri Dec 9 14:04:13 PST 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Added paintPr, which fills a rectangular region of a PIXR with
 * a color passed. If the image is binary, just do a rasterOp
 * directly, otherwise fill the first line (not very efficiently!) and
 * replicate that line using raster ops. Obviously this could be
 * done more efficiently by filling directly without the copy.
 *
 * Revision 1.22  1994/10/27  02:13:51  lperry
 * lperry on Wed Oct 26 19:07:47 PDT 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Fixed a bug in the pixr inversion routine which
 * caused the first bit of the frame to be inverted.
 * This may have been causing some problems for easyscan.
 *
 * Also removed function invertPrGIP, which had handled
 * gray scale inversions. Since there is no difference
 * between how inversion is done at any depth, all
 * the work is now done in invertPrIP.
 *
 * BTW, the function iw_invert (dstPixr, srcPixr)
 * is a more generic way to do this. If the
 * source and dest pixrs are the same, it will
 * call invertPrIP, otherwise it copies the
 * src to the dst and inverts the dst in Place.
 *
 * Revision 1.21  1994/10/25  20:37:18  macleod
 * macleod on Tue Oct 25 13:37:09 PDT 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * createPixrExtended wasn't declared CDECL
 *
 * Revision 1.20  1994/10/11  17:41:00  lperry
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add include of shros.pub to get IP_YIELD definition.
 *
 * Revision 1.19  1994/10/10  22:42:44  lperry
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add include of shrpixr.pub
 *
 * Revision 1.18  1994/09/30  15:52:46  lperry
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.17  1994/09/29  23:13:53  lperry
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Modify to use props.pub instead of pixrpros.h
 *
 * Revision 1.16  1994/09/20  04:29:02  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add reallocPixrData.  This routine calculates the size of the memory buffer
 * needed for a PIXR and calls IPW_REALLOC to set the new size.  This routine is used
 * for weird "in place" routines that change the size and/or depth of a PIXR.
 * Also, free all sub-PIXRs when createPixrExtended encounters an error.  Thanks
 * for finding this Steve!
 *
 * Revision 1.15  1994/09/16  17:39:11  jcraig
 * jcraig on Fri Sep 16 10:38:59 PDT 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.14  1994/06/23  19:04:37  macleod
 * macleod on Thu Jun 23 12:03:32 PDT 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Well, there was a routine which I added, and somehow it didn't get
 * checked in (well, it was write-protected in my directory, but
 * didn't wind up in the revision log. I don't know, maybe I aborted
 * at the wrong time.)
 *
 * Revision 1.13  1994/06/07  02:39:44  macleod
 * macleod on Mon Jun 6 19:39:43 PDT 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.12  1994/05/26  19:25:04  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Change abortP's whose return values are discarded to abortV's.  Enclose the
 * fprintf's that follow them inside #ifndef PRODUCTION clauses.
 *
 * Revision 1.11  1994/05/26  01:06:52  jcraig
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * in createPixrNoInit, clear the frames around the image even tho the
 * image itself is not cleared since frames should always be cleared.
 *
 * Revision 1.10  1994/05/13  18:50:48  dferrell
 * dferrell on Fri May 13 11:49:56 PDT 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Fixed memory leak in createPixr.  When the image data IPW_CALLOC failed,
 * the pixr struct was not freed.
 *
 * Revision 1.9  1994/05/05  01:47:09  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Change lots of %d's to %ld's to make gcc happy.
 *
 * Revision 1.8  1994/04/12  01:15:11  macleod
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.7  1994/03/02  03:58:03  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Added IP_YIELD to give up control during rectOfPixr.
 *
 * Revision 1.6  1994/02/13  01:03:41  ddavies
 * ddavies on Sat Feb 12 17:02:45 PST 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Change createPixr so that it pads the height of each PIXR.  The padding
 * allows each PIXR to be rotated by 90 or 270 degrees in place in that it has
 * enough space to allow the rotated image to fit.
 *
 * Revision 1.5  1994/01/11  20:05:16  lperry
 * lperry on Tue Jan 11 12:04:12 PST 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * replaced call to opPrPr with one to pixr_rop
 * so brain-damaged linker can find it
 *
 * Revision 1.4  1994/01/10  23:44:02  lperry
 * lperry on Mon Jan 10 15:40:58 PST 1994
 *
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Added routine iw_invert to perform functions previously
 * done by invertPrIP and invertPrGIP. These will be retained for
 * a time but should be considered obsolete. The "iw" functions will
 * be used to replace numerous similar calls in the future.
 *
 * Revision 1.3  1993/12/03  01:17:39  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:15:02  ddavies
 * props3.c -> /product/ipcore/ipwrap/src/RCS/props3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

