/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  props4.c
 *
 *      Wrapper for copying a PIXR, for which assembler accelerators exist
 *      for images with 32 pixel frames:
 *
 *               w_copy()
 */




#include "ipwcore.prv"
#include "defines.pub"     /* for cTrue and cFalse */
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"

/* prototypes */
#include "props.pub"
#include "shrpixr.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: props4.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy a PIXR to another PIXR.  If there isn't a second PIXR, create it.
 *
 * LongDescription:::
 *  Returns a copy of the input pixr.
 *  Uses accelerated version only if frame is 32 pixels.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            Ptr to source PIXR to copy.
 *
 * Output:::
 *  prd::            Ptr to ptr to destination PIXR.
 *                   If *prd is NULL on input, we'll create a PIXR and assign
 *                   the ptr to it.  Otherwise, use the dest PIXR.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, rasterOp, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copy (
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR              **prd)
{
    PROLOGUE;
    Int32         w, h, d, frame, channels, i;
    Int32         bytes, sbpl, dbpl;
    Int32         rfc, bfc;
    Bool          destCreated = cFalse;
    ScaledInt1616 xdpi, ydpi;
    ColorType     type;
    ColorMap     *map;
    UInt8        *sdata, *ddata;
    PIXR         *subPixrS, *subPixrD, *pixr = NULL;

    ERR_GEN(prd == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prs, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetWidth(pIPWState, prs, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prs, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, prs, &d));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xdpi, &ydpi));
    ERR_PASS( w_pixrGetFrame(pIPWState, prs, &frame));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prs, &channels));
    ERR_PASS( w_pixrGetType(pIPWState, prs, &type));
    ERR_PASS( w_pixrGetColorMap(pIPWState, prs, &map));

    pixr = *prd;
    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState, 
						      w, h, d, 
						      xdpi, ydpi,
						      cFrameBits,
						      channels, type, map,
						      &pixr, &rfc, &bfc,
						      &destCreated));



    /* if frame is not 32 pixels, just memcpy the bytes */
    if (frame != cFrameBits)
    {
	for (i = 0; i < channels; i++)
	{
	    if (channels == 1)
	    {
		subPixrS = prs;
		subPixrD = pixr;
	    }
	    else
	    {
		ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prs, i,
						   &subPixrS));
		ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, i,
						   &subPixrD));
	    }

	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subPixrS, &sdata));
	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subPixrD, &ddata));
	    ERR_PASS_CLEANUP( w_pixrImageBytes(pIPWState, subPixrS, &bytes));
	    memcpy((char*)ddata, (char*)sdata, bytes);
	}
    }
    else
    {
	/* use accelerator version if frame is 32 pixels */

	for (i = 0; i < channels; i++)
	{
	    if (channels == 1)
	    {
		subPixrS = prs;
		subPixrD = pixr;
	    }
	    else
	    {
		ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prs, i,
						   &subPixrS));
		ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, i,
						   &subPixrD));
	    }

	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subPixrS, &sdata));
	    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, subPixrD, &ddata));
	    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, subPixrS, &w));
	    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, subPixrS, &h));
	    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, subPixrD, &dbpl));
	    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, subPixrS, &sbpl));
	
	    ERR_PASS_CLEANUP( i_copyImage (pIPWState, sdata,
					   w, h, d, sbpl, ddata, dbpl));
	}
    }
    

    *prd = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}

/*

 $Log: props4.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.30  1998/01/14 00:14:28  ddavies
 props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.29  1997/04/15 22:23:25  ddavies
 props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.28  1996/01/04 20:25:44  ejaquith
 props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.27  1995/12/13  04:11:40  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.26  1995/11/13  20:14:45  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.25  1995/11/03  22:48:45  ddavies
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.24  1995/10/13  22:04:02  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.23  1995/09/29  00:36:19  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.22  1995/09/21  22:58:19  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.21  1995/09/18  23:48:31  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Bug in w_copy() where was allocing/initing PIXR twice because the old
 * allocators were accidentally left in when adding
 * wp_createPixrExtendedConditional().
 *
 * Revision 1.20  1995/09/15  22:38:13  ddavies
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.19  1995/09/15  00:01:11  dferrell
 * dferrell on Thu Sep 14 17:00:14 PDT 1995
 *
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Made w_copy () create a dest pixr if a pointer to NULL was passed in
 * (tsk tsk, Eric).
 *
 * Revision 1.18  1995/09/14  18:56:58  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 *  Fixed types to be used with w_pixrGetResolutionDPI().
 *
 * Revision 1.17  1995/09/12  16:55:39  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.16  1995/09/08  00:15:38  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.15  1995/09/07  01:34:13  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.14  1995/09/06  20:22:19  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.13  1995/09/01  02:21:14  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.12  1995/08/31  21:34:24  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/29  21:56:03  jcraig
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * get the correct bytes per line (source and dest were swapped) to
 * i_copyImage
 *
 * Revision 1.10  1995/08/17  22:02:10  ejaquith
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.9  1995/02/03  01:37:40  ddavies
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Make copyPixr pass the resolutions of the input PIXR to the new output PIXR.
 *
 * Revision 1.8  1994/09/30  15:52:51  lperry
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.7  1994/09/29  23:13:58  lperry
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Modify to use props.pub instead of pixrpros.h
 *
 * Revision 1.6  1994/06/07  02:39:53  macleod
 * macleod on Mon Jun 6 19:39:52 PDT 1994
 *
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.5  1994/04/12  01:15:24  macleod
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.4  1994/01/21  21:36:16  ddavies
 * ddavies on Fri Jan 21 13:35:52 PST 1994
 *
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Add destination bytes/line to call to i_copyImage.
 *
 * Revision 1.3  1993/12/03  01:17:43  ddavies
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:15:11  ddavies
 * props4.c -> /product/ipcore/ipwrap/src/RCS/props4.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

