/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  props5.c:
 *
 *	     Checking if PIXR is "appropriate" for use as dest image
 *			w_checkDestPixr()
 */


#include "ipwcore.prv"

#include "frames.pub"
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"

/* prototypes */
#include "props.pub"
#include "props.prv"
#include "rast.pub"
#include "shrpixr.pub"
#include "rectops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: props5.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/*--------------------------------------------------------------------*
 *                     Conditional Pixr Creation                      *
 *--------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create one-channel, gray-scale or binary PIXR or use passed PIXR if it's ok
 *
 * LongDescription:::
 *  Like w_createPixr() except it can take a dest PIXR.
 *  If not given a dest PIXR:<BR>
 *   One is created to desired specs.  It is NOT filled with 0's.<BR><BR>
 *
 *  If a dest PIXR is supplied:<BR>
 *   [Note we refer to a source PIXR here, meaning that which would be created
 *   by creating a PIXR according to the passed in arguments.]<BR><BR>
 *
 *   It is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   The dest PIXR data is NOT touched/initialized!<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  width::                   image width value, in pixels
 *  height::                  image height value, in pixels
 *  depth::                   depth value, {1, 2, 4, 8, 16, or 32}, in bits
 *  xdpi::                    image dots per inch in X dimen
 *  ydpi::                    image dots per inch in Y dimen
 *  frame::                   frame value, in pixels
 *
 * Output:::
 *  ppPixr::                  Ptr to ptr to dest PIXR.
 *                            If *ppPixr points to NULL, create a PIXR and
 *                            set *ppPixr to point to it.  Otherwise use the
 *                            passed in PIXR as the dest PIXR.
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *  pPixrAllocatedHere::      If not NULL, set to cTrue if we allocated a PIXR
 *                            here and cFalse if we did not.
 *
 * Returns:::
 *
 * Notes:::
 *  depth and frame codependence::
 *   This must be true for passed values: (depth * frame) % 32 = 0
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_createPixrConditionalNoInit (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    PIXR              **ppPixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear,
    Bool               *pPixrAllocatedHere)
{
    PROLOGUE;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(pRightFrameColsToClear == NULL || pBottomFrameRowsToClear == NULL,
	    cErrInvalidParm);

	/* If we were not given an image, go create one. */
    if (*ppPixr == NULL)
    {
	ERR_PASS(w_createPixrNoInit(pIPWState, width, height, depth, xdpi, ydpi,
			            frame, ppPixr));
	*pRightFrameColsToClear = 0;
	*pBottomFrameRowsToClear = 0;
	if (pPixrAllocatedHere != NULL)
	    *pPixrAllocatedHere = cTrue;
    }
    else
    {
	    /* At this point, we were given a dest pixr.  See if it's ok. */
	ERR_PASS( w_checkDestPixr(pIPWState, width, height, depth,
				    xdpi, ydpi, frame,
				    1, cColorGray, NULL, *ppPixr,
				    pRightFrameColsToClear,
				    pBottomFrameRowsToClear));
    
	    /* At this point, we know the dest pixr supplied is ok, so
	     * set the arg that says we did NOT allocate one and
	     * set the number of frame rows and cols to clear after
	     * the operation.
	     */
	if (pPixrAllocatedHere != NULL)
		*pPixrAllocatedHere = cFalse;
    }

    NORMAL_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create initialized one-channel PIXR or use passed PIXR if it's ok
 *
 * LongDescription:::
 *  Like w_createPixrConditionalNoInit() except it clears the destination PIXR
 *  if one is not passed in.<BR>.
 *  If not given a dest PIXR:<BR>
 *   One is created to desired specs<BR>
 *   We init the image data in created dest PIXR to zero.<BR><BR>
 *
 *  If a dest PIXR is supplied:<BR>
 *   [Note we refer to a source PIXR here, meaning that which would be created
 *   by creating a PIXR according to the passed in arguments.]<BR><BR>
 *
 *   It is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   If initExistingImage is set, the dest PIXR data is initialized to 0!
 *   Otherewise, the dest PIXR data is left untouched.<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  width::                   image width value, in pixels
 *  height::                  image height value, in pixels
 *  depth::                   depth value, {1, 2, 4, 8, 16, or 32}, in bits
 *  xdpi::                    image dots per inch in X dimen
 *  ydpi::                    image dots per inch in Y dimen
 *  frame::                   frame value, in pixels
 *  initExistingImage::	      set to cTrue if the image passed in should
 *			      be cleared to 0.  Otherwise, it is left
 *			      untouched.
 *
 * Output:::
 *  ppPixr::                  Ptr to ptr to dest PIXR.
 *                            If *ppPixr points to NULL, create a PIXR and
 *                            set *ppPixr to point to it.  Otherwise use the
 *                            passed in PIXR as the dest PIXR.
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *  pPixrAllocatedHere::      If not NULL, set to cTrue if we allocated a PIXR
 *                            here and cFalse if we did not.
 *
 * Returns:::
 *
 * Notes:::
 *  depth and frame codependence::
 *   This must be true for passed values: (depth * frame) % 32 = 0
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_createPixrConditional (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Bool		initExistingImage,
    PIXR              **ppPixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear,
    Bool               *pPixrAllocatedHere)
{
    PROLOGUE;
    Bool	allocatedHere;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(pRightFrameColsToClear == NULL || pBottomFrameRowsToClear == NULL,
	    cErrInvalidParm);

	/* If we were not given an image, go create one. */
    if (*ppPixr == NULL)
    {
	ERR_PASS(w_createPixr(pIPWState, width, height, depth, xdpi, ydpi,
			      frame, ppPixr));
	*pRightFrameColsToClear = 0;
	*pBottomFrameRowsToClear = 0;
	allocatedHere = cTrue;
    }
    else
    {
	    /* At this point, we were given a dest pixr.  See if it's ok. */
	ERR_PASS( w_checkDestPixr(pIPWState, width, height, depth,
				    xdpi, ydpi, frame,
				    1, cColorGray, NULL, *ppPixr,
				    pRightFrameColsToClear,
				    pBottomFrameRowsToClear));

	    /* At this point, we know the dest pixr supplied is ok, so
	     * set the arg that says we did NOT allocate one and
	     * set the number of frame rows and cols to clear after
	     * the operation.
	     */
	allocatedHere = cFalse;

	    /* If we're supposed to clear the existing image, do so */
	if (initExistingImage)
	{
		/* the client wanted a cleared image, so clear it whether or
		 * not it was passed in.
		 */
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, *ppPixr, &width));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, *ppPixr, &height));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
					  0, 0, width, height, *ppPixr));
	}
    }


	/* set the return arg saying whether we allocated the image */
    if (pPixrAllocatedHere != NULL)
	*pPixrAllocatedHere = allocatedHere;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr( pIPWState, ppPixr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clone a PIXR (copy w/o copying actual data) or use passed PIXR if it's ok.
 *
 * LongDescription:::
 *  Like w_clonePixr() except it can take a dest PIXR.
 *  If not given a dest PIXR:<BR>
 *   One is created to desired specs.<BR>
 *   We DO NOT init the image data in created dest PIXR to zero.<BR><BR>
 *
 *  If a dest PIXR is supplied:<BR>
 *   It is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   The dest PIXR data is NOT touched/initialized!<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  pixr::                    Ptr to PIXR to clone.  May be multi-channel.
 *
 * Output:::
 *  ppPixr::                  Ptr to ptr to dest PIXR.
 *                            If *ppPixr points to NULL, create a PIXR and
 *                            set *ppPixr to point to it.  Otherwise use the
 *                            passed in PIXR as the dest PIXR.
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *  pPixrAllocatedHere::      If not NULL, set to cTrue if we allocated a PIXR
 *                            here and cFalse if we did not.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_clonePixrConditionalNoInit (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppPixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear,
    Bool               *pPixrAllocatedHere)
{
    PROLOGUE;
    Int32		 w, h, d, frame, channels;
    ColorType		 type;
    ColorMap		*map;
    ScaledInt1616	 xdpi, ydpi;

    ERR_GEN(ppPixr == NULL || pixr == NULL, cErrNullPtr);
    ERR_GEN(pRightFrameColsToClear == NULL || pBottomFrameRowsToClear == NULL,
	    cErrInvalidParm);

	/* If we were not given an image, go create one. */
    if (*ppPixr == NULL)
    {
	ERR_PASS(w_clonePixrNoInit(pIPWState, pixr, ppPixr));
	*pRightFrameColsToClear = 0;
	*pBottomFrameRowsToClear = 0;
	if (pPixrAllocatedHere != NULL)
	    *pPixrAllocatedHere = cTrue;
    }
    else
    {
	    /* At this point, we were given a dest pixr.  See if it's ok. */
	ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
	ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));
	ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &d));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));
	ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &frame));
	ERR_PASS( w_pixrGetChannelCount (pIPWState, pixr, &channels));
	ERR_PASS( w_pixrGetType(pIPWState, pixr, &type));
	ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &map));
	ERR_PASS( w_checkDestPixr(pIPWState, w, h, d,
				    xdpi, ydpi, frame,
				    channels, type, map, *ppPixr,
				    pRightFrameColsToClear,
				    pBottomFrameRowsToClear));
    
	    /* At this point, we know the dest pixr supplied is ok, so
	     * set the arg that says we did NOT allocate one and
	     * set the number of frame rows and cols to clear after
	     * the operation.
	     */
	if (pPixrAllocatedHere != NULL)
		*pPixrAllocatedHere = cFalse;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clone a PIXR (copy w/o copying actual data) or use passed PIXR if it's ok.
 *
 * LongDescription:::
 *  Like w_clonePixrZero() except it can take a dest PIXR.
 *  If not given a dest PIXR:<BR>
 *   One is created to desired specs.<BR>
 *   We init the image data in created dest PIXR to zero.<BR><BR>
 *
 *  If a dest PIXR is supplied:<BR>
 *   It is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   The dest PIXR data is NOT touched/initialized!<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  pixr::                    Ptr to PIXR to clone.  May be multi-channel.
 *
 * Output:::
 *  ppPixr::                  Ptr to ptr to dest PIXR.
 *                            If *ppPixr points to NULL, create a PIXR and
 *                            set *ppPixr to point to it.  Otherwise use the
 *                            passed in PIXR as the dest PIXR.
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *  pPixrAllocatedHere::      If not NULL, set to cTrue if we allocated a PIXR
 *                            here and cFalse if we did not.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_clonePixrConditional (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppPixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear,
    Bool               *pPixrAllocatedHere)
{
    PROLOGUE;
    Int32		 w, h, d, frame, channels;
    ColorType		 type;
    ColorMap		*map;
    ScaledInt1616	 xdpi, ydpi;

    ERR_GEN(ppPixr == NULL || pixr == NULL, cErrNullPtr);
    ERR_GEN(pRightFrameColsToClear == NULL || pBottomFrameRowsToClear == NULL,
	    cErrInvalidParm);

	/* If we were not given an image, go create one. */
    if (*ppPixr == NULL)
    {
	ERR_PASS(w_clonePixr(pIPWState, pixr, ppPixr));
	*pRightFrameColsToClear = 0;
	*pBottomFrameRowsToClear = 0;
	if (pPixrAllocatedHere != NULL)
	    *pPixrAllocatedHere = cTrue;
    }
    else
    {
	    /* At this point, we were given a dest pixr.  See if it's ok. */
	ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
	ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));
	ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &d));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));
	ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &frame));
	ERR_PASS( w_pixrGetChannelCount (pIPWState, pixr, &channels));
	ERR_PASS( w_pixrGetType(pIPWState, pixr, &type));
	ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &map));
	ERR_PASS( w_checkDestPixr(pIPWState, w, h, d,
				    xdpi, ydpi, frame,
				    channels, type, map, *ppPixr,
				    pRightFrameColsToClear,
				    pBottomFrameRowsToClear));
    
	    /* At this point, we know the dest pixr supplied is ok, so
	     * set the arg that says we did NOT allocate one and
	     * set the number of frame rows and cols to clear after
	     * the operation.
	     */
	if (pPixrAllocatedHere != NULL)
		*pPixrAllocatedHere = cFalse;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create binary, gray, or color PIXR or use passed PIXR if it's ok.
 *
 * LongDescription:::
 *  Like w_createPixrExtended() except it can take a dest PIXR.
 *  If not given a dest PIXR:<BR>
 *   One is created to desired specs.<BR>
 *   We init the image data in created dest PIXR to zero.<BR><BR>
 *
 *  If a dest PIXR is supplied:<BR>
 *   [Note we refer to a source PIXR here, meaning that which would be created
 *   by creating a PIXR according to the passed in arguments.]<BR><BR>
 *
 *   It is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   The dest PIXR data is NOT touched/initialized!<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  width::                   image width value, in pixels
 *  height::                  image height value, in pixels
 *  depth::                   depth value of each plane, {1, 2, 4, 8, 16, 32}, 
 *                            in bits
 *  xdpi::                    image dots per inch in X dimen
 *  ydpi::                    image dots per inch in Y dimen
 *  frame::                   frame value, in pixels
 *  channels::                Number of image planes
 *  type::                    Color type of image, {cColorRGB, cColorCMYK,
 *                            cColorCIELAB, cColorCIELUV, cColorGray, or 
 *                            cColorNone}
 *  map::                     Ptr to color map, or NULL if none
 *
 * Output:::
 *  ppPixr::                  Ptr to ptr to dest PIXR.
 *                            If *ppPixr points to NULL, create a PIXR and
 *                            set *ppPixr to point to it.  Otherwise use the
 *                            passed in PIXR as the dest PIXR.
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *  pPixrAllocatedHere::      If not NULL, set to cTrue if we allocated a PIXR
 *                            here and cFalse if we did not.
 *
 * Returns:::
 *
 * Notes:::
 *
 *  depth and frame codependence::
 *   This must be true for passed values: (depth * frame) % 32 = 0
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_createPixrExtendedConditional (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Int32               channels,
    ColorType           type,
    ColorMap           *map,
    PIXR              **ppPixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear,
    Bool               *pPixrAllocatedHere)
{
    PROLOGUE;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(pRightFrameColsToClear == NULL || pBottomFrameRowsToClear == NULL,
	    cErrInvalidParm);

    ERR_GEN(type != cColorRGB     &&  type != cColorCMYK    &&
	    type != cColorCIELAB  &&  type != cColorCIELUV  &&
	    type != cColorGray    &&  type != cColorNone, cErrInvalidParm);

	/* If we were not given an image, go create one. */
    if (*ppPixr == NULL)
    {
	ERR_PASS(w_createPixrExtended(pIPWState, width, height, depth,
				      xdpi, ydpi, frame, channels, type, map,
				      ppPixr));
	*pRightFrameColsToClear = 0;
	*pBottomFrameRowsToClear = 0;
	if (pPixrAllocatedHere != NULL)
	    *pPixrAllocatedHere = cTrue;
    }
    else
    {
	    /* At this point, we were given a dest pixr.  See if it's ok. */
	ERR_PASS( w_checkDestPixr(pIPWState, width, height, depth,
				    xdpi, ydpi, frame,
				    channels, type, map, *ppPixr,
				    pRightFrameColsToClear,
				    pBottomFrameRowsToClear));
    
	    /* At this point, we know the dest pixr supplied is ok, so
	     * set the arg that says we did NOT allocate one and
	     * set the number of frame rows and cols to clear after
	     * the operation.
	     */
	if (pPixrAllocatedHere != NULL)
		*pPixrAllocatedHere = cFalse;
    }

    NORMAL_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create uninitialized binary, gray, or color PIXR or check input PIXR
 *
 * LongDescription:::
 *  Like w_createPixrExtended() except it can take a dest PIXR.
 *  If not given a dest PIXR:<BR>
 *   One is created to desired specs.<BR>
 *   We DO NOT init the image data in created dest PIXR to zero.<BR><BR>
 *
 *  If a dest PIXR is supplied:<BR>
 *   [Note we refer to a source PIXR here, meaning that which would be created
 *   by creating a PIXR according to the passed in arguments.]<BR><BR>
 *
 *   It is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   The dest PIXR data is NOT touched/initialized!<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   client must then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  width::                   image width value, in pixels
 *  height::                  image height value, in pixels
 *  depth::                   depth value of each plane, {1, 2, 4, 8, 16, 32}, 
 *                            in bits
 *  xdpi::                    image dots per inch in X dimen
 *  ydpi::                    image dots per inch in Y dimen
 *  frame::                   frame value, in pixels
 *  channels::                Number of image planes
 *  type::                    Color type of image, {cColorRGB, cColorCMYK,
 *                            cColorCIELAB, cColorCIELUV, cColorGray, or 
 *                            cColorNone}
 *  map::                     Ptr to color map, or NULL if none
 *
 * Output:::
 *  ppPixr::                  Ptr to ptr to dest PIXR.
 *                            If *ppPixr points to NULL, create a PIXR and
 *                            set *ppPixr to point to it.  Otherwise use the
 *                            passed in PIXR as the dest PIXR.
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *  pPixrAllocatedHere::      If not NULL, set to cTrue if we allocated a PIXR
 *                            here and cFalse if we did not.
 *
 * Returns:::
 *
 * Notes:::
 *
 *  depth and frame codependence::
 *   This must be true for passed values: (depth * frame) % 32 = 0
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_createPixrExtendedConditionalNoInit (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Int32               channels,
    ColorType           type,
    ColorMap           *map,
    PIXR              **ppPixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear,
    Bool               *pPixrAllocatedHere)
{
    PROLOGUE;

    ERR_GEN(ppPixr == NULL, cErrNullPtr);
    ERR_GEN(pRightFrameColsToClear == NULL || pBottomFrameRowsToClear == NULL,
	    cErrInvalidParm);

    ERR_GEN(type != cColorRGB     &&  type != cColorCMYK    &&
	    type != cColorCIELAB  &&  type != cColorCIELUV  &&
	    type != cColorGray    &&  type != cColorNone, cErrInvalidParm);

	/* If we were not given an image, go create one. */
    if (*ppPixr == NULL)
    {
	ERR_PASS(w_createPixrExtendedNoInit(pIPWState, width, height, depth,
					    xdpi, ydpi, frame, channels,
					    type, map, ppPixr));
	*pRightFrameColsToClear = 0;
	*pBottomFrameRowsToClear = 0;
	if (pPixrAllocatedHere != NULL)
	    *pPixrAllocatedHere = cTrue;
    }
    else
    {
	    /* At this point, we were given a dest pixr.  See if it's ok. */
	ERR_PASS( w_checkDestPixr(pIPWState, width, height, depth,
				    xdpi, ydpi, frame,
				    channels, type, map, *ppPixr,
				    pRightFrameColsToClear,
				    pBottomFrameRowsToClear));
    
	    /* At this point, we know the dest pixr supplied is ok, so
	     * set the arg that says we did NOT allocate one and
	     * set the number of frame rows and cols to clear after
	     * the operation.
	     */
	if (pPixrAllocatedHere != NULL)
		*pPixrAllocatedHere = cFalse;
    }

    NORMAL_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check if dest PIXR is close enough to a described PIXR.
 *
 * LongDescription:::
 *  Checks to see if the dest PIXR supplied is close enough to the
 *  described desired PIXR -- returning without error means it is close
 *  enough.<BR><BR>
 *
 *   [Note we refer to a source PIXR here, meaning that which would be created
 *   by creating a PIXR according to the passed in arguments.]<BR><BR>
 *
 *   Dest PIXR is checked for proper number of channels, depth, resolution and
 *   size.<BR><BR>
 *
 *   If the magnitude of the difference between source width and dest width or
 *   if the magnitude of the difference between source height and dest height
 *   is greater than the dest frame, we return an error.<BR><BR>
 *
 *   If source width > dest width, *pRightFrameColsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   the operation will then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into right frame columns in the dest PIXR!<BR><BR>
 *
 *   If source height > dest height, *pBottomFrameRowsToClear will be set to
 *   the (positive) difference, else set to 0.
 *   Thus, after a src and dest PIXR ipcore operation using these PIXRs,
 *   the operation will then call wp_clearRightBottomFrames() to clear out the
 *   possibly written into bottom frame rows in the dest PIXR!<BR><BR>
 *
 *   If source width < dest width, some columns in dest PIXR may not be
 *   touched (warning).<BR><BR>
 *
 *   If source height < dest height, some rows in dest PIXR may not be
 *   touched (warning).
 *
 * Input:::
 *  pIPWState::               Ptr to IP struct for memory alloc/freeing
 *                            routines
 *  width::                   image width value, in pixels
 *  height::                  image height value, in pixels
 *  depth::                   depth value of each plane, {1, 2, 4, 8, 16, 32}, 
 *                            in bits
 *  xdpi::                    image dots per inch in X dimen
 *  ydpi::                    image dots per inch in Y dimen
 *  frame::                   frame value, in pixels
 *  channels::                Number of image planes
 *  type::                    Color type of image, {cColorRGB, cColorCMYK,
 *                            cColorCIELAB, cColorCIELUV, cColorGray,
 *                            or cColorNone} 
 *  map::                     Ptr to color map, or NULL if none
 *  pixr::                    Ptr to dest PIXR to check.
 *
 * Output:::
 *  pRightFrameColsToClear::  Ptr to number of right side frame columns to
 *                            clear later (see description).
 *  pBottomFrameRowsToClear:: Ptr to number of bottom frame rows to clear
 *                            later (see description).
 *
 * Returns:::
 *
 * Notes:::
 *  depth and frame codependence::
 *   This must be true for passed values: (depth * frame) % 32 = 0
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_checkDestPixr (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       xdpi,
    ScaledInt1616       ydpi,
    Int32               frame,
    Int32               channels,
    ColorType           type,
    ColorMap           *map,
    PIXR               *pixr,
    Int32              *pRightFrameColsToClear,
    Int32              *pBottomFrameRowsToClear)
{
    PROLOGUE;
    Int32		 wD, hD, dD, fD, channelsD;
    ScaledInt1616	 xDdpi, yDdpi;
    ColorType		 typeD;
    ColorMap		*mapD;

	/* is the pixr basically ok? */
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrameAny);

	/* right number of channels? */
    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channelsD));
    ERR_GEN(channelsD != channels, cErrDestDimensIncorrect);

	/* Are the ColorType and ColorMap the same?  We check to see
	 * if the color map points to the same block of memory.  This
	 * is not very sophisticated, maybe we'll change it if we
	 * really start using color maps.
	 */
    ERR_PASS( w_pixrGetType(pIPWState, pixr, &typeD));
    ERR_GEN(typeD != type, cErrDestDimensIncorrect);
    ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &mapD));
    ERR_GEN(mapD != map, cErrDestDimensIncorrect);

	/* does it have the correct depth? */
    ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &dD));
    ERR_GEN(dD != depth, cErrDestDimensIncorrect);

	/* the frames should be the same too */
    ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &fD));
    ERR_GEN(fD != frame, cErrDestDimensIncorrect);

	/* the resolution has to be close to the desired resolution.
	 * Since these numbers have 16 bits of fraction, comparing
	 * the difference to 0x10000 is like comparing it to 1.0
	 */
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xDdpi, &yDdpi));
    ERR_GEN((ABS(xDdpi - xdpi) > 0x10000) || (ABS(yDdpi - ydpi) > 0x10000),
	    cErrDestDimensIncorrect);

	/* The difference in widths and heights must be less than or equal
	 * to the dest's frame width so it can absorb the difference.
	 */
    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &wD));
    ERR_GEN( ABS(wD - width) > fD, cErrDestDimensIncorrect);

    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &hD));
    ERR_GEN( ABS(hD - height) > fD, cErrDestDimensIncorrect);

	/* At this point, we know the dest pixr supplied is ok, so
	 * set the arg that says we did NOT allocate one and
	 * set the number of frame rows and cols to clear after
	 * the operation.  We only need to clear frame pixels in
	 * the dest if it was too small so some of the source may
	 * have leaked through.  If the dest was bigger that the
	 * source, we may have some glop around the edge.  People
	 * seem to think that leaving this here is generally better
	 * than trying to set it to something, like white or black
	 * or something else.  Our client are very sure that they
	 * don't want the size of the image changed to reflect the
	 * number of valid pixels.
	 */
    if (width > wD)
	*pRightFrameColsToClear = width - wD;
    else
	*pRightFrameColsToClear = 0;

    if (height > hD)
	*pBottomFrameRowsToClear = height - hD;
    else
	*pBottomFrameRowsToClear = 0;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clears requested parts of right and bottom frame of a pixr.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  inPr::           Pixr to be processed
 *  rightCols::      Number of columns of the right frame to be
 *                   cleared, range [0..32].
 *  bottomRows::     Number of rows of the bottom frame to be
 *                   cleared, range [0..32].
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_clearRightBottomFrames (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               rightCols,
    Int32               bottomRows)
{
    PROLOGUE;
    Int32  w, h, d, channels, bpl, i;
    PIXR  *channel;
    UInt8 *pI;

    /* only do something if we need to */
    if ((rightCols > 0) || (bottomRows > 0))
    {
	ERR_GEN_PIXR (inPr, cByteOrderBig, cDepthMask, cChannelsMask,
		      cFrame32);
    
	ERR_GEN ((rightCols > 32) || (bottomRows > 32), cErrInvalidParm);

	ERR_PASS (w_pixrGetChannelCount (pIPWState, inPr, &channels));
	ERR_PASS (w_pixrFramedWidth (pIPWState, inPr, &w));
	ERR_PASS (w_pixrFramedHeight (pIPWState, inPr, &h));
	ERR_PASS (w_pixrGetDepth (pIPWState, inPr, &d));
    
    
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS (w_pixrGetChannel (pIPWState, inPr, i, &channel));
	    ERR_PASS (w_pixrGetBpl (pIPWState, channel, &bpl));
	    ERR_PASS (w_pixrGetImageAll (pIPWState, channel, &pI));
    
	    ERR_PASS (i_clearRightBottomFrames (pIPWState, pI, w, h,
						bpl, d, rightCols,
						bottomRows));
	}
    }

    NORMAL_RETURN;
}


/*

 $Log: props5.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.17  1998/01/14 00:14:30  ddavies
 props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.16  1997/06/26 20:04:25  ddavies
 props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.15  1997/04/15 22:25:53  ddavies
 props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v

 Add new routines wp_createPixrConditionalNoInit,
 wp_clonePixrConditionalNoInit and wp_createPixrExtendedConditionalNoInit.
 These are like their initialized counterparts except only the frame of
 newly allocated PIXRs is initialized.
 Also, change the name of wp_clonePixrZeroConditional to be
 wp_clonePixrConditional to be consistent.

 Revision 1.14  1997/03/14 23:38:46  ddavies
 ddavies on Fri Mar 14 15:37:54 PST 1997

 props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v

 Use w_pixrFramedWidth and w_pixrFramedHeight in wp_clearRightBottomFrame.
 How did this ever work?  Didn't anyone ever need a frame cleared before?

 Revision 1.13  1996/01/04 20:25:47  ejaquith
 props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.12  1995/12/13  04:11:44  ejaquith
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.11  1995/11/13  20:14:48  ejaquith
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.10  1995/10/13  22:04:05  ejaquith
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.9  1995/09/29  00:36:24  ejaquith
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.8  1995/09/21  22:58:22  ejaquith
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.7  1995/09/14  20:42:06  ejaquith
 * ejaquith on Thu Sep 14 13:41:36 PDT 1995
 *
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * Changed *pPixrAllocatedHere param to routines from Int32 type to bool
 * (as Monsieur Davies had done in the protos file).
 *
 * Revision 1.6  1995/09/14  20:19:46  ejaquith
 * ejaquith on Thu Sep 14 13:19:18 PDT 1995
 *
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * Typo fix: w_pixrGetChannels() to w_pixrGetChannelCount().
 * Warning fix: unused var "
 * frame
 * "
 *
 * Revision 1.5  1995/09/14  19:11:04  dferrell
 * dferrell on Thu Sep 14 12:10:55 PDT 1995
 *
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * Fixed typo/bug.
 *
 * Revision 1.4  1995/09/14  18:26:16  dferrell
 * dferrell on Thu Sep 14 11:25:53 PDT 1995
 *
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * Added new function wp_clearRightBottomFrames () which is the wrapper
 * for i_clearRightBottomFrames ().
 *
 * Revision 1.3  1995/09/14  01:49:54  ddavies
 * ddavies on Wed Sep 13 18:46:51 PDT 1995
 *
 * props5.c -> /product/ipcore/ipwrap/src/RCS/props5.c,v
 *
 * This is a new file that replaces an old file that was also named props5.c.
 * The new file contains implementations for wp_createPixrConditional,
 * wp_clonePixrZeroConditional, wp_createPixrExtendedConditional and
 * w_checkDestPixr.  The first three routines are like their un-conditional
 * counterparts except that they can conditionally accept a dest pixr.  If
 * not supplied with one, they call their un-conditional counterpart to make
 * one.  If supplied with one, they use w_checkDestPixr to see if it's close
 * enough to the desired pixr.  If not, they return with an error.  If so, they
 * return the number of rows and columns of the dest pixrs frame that must
 * be cleared after the operation (if any).
 *

 */

