
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  prxform1.c	procedures that do a variety of threshold and filtering
 *             	operations on 8-bit images.  These are exactly the routines
 *		used by PerfectScan.
 *
 *     Grayscale thresholding:
 *	     w_scale8To4Pr()
 * 
 *     LUT processing:
 *	     w_map
 *
 *     Bit depth up-conversion
 *	     w_depth4To8
 *	     w_depth1To8
 *	     w_depth1To16
 *           w_map1To8
 *
 *     Grayscale histogram:
 *           w_histo
 *	     w_histoSubSampled
 *	     w_histoMask
 *
 */




#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "prxform.pub"
#include "rectops.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: prxform1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Changes pixel values in an image using a table.
 *
 * LongDescription:::
 *  It is often useful to change the brightness or contrast in an image.  One
 *  way to do this is to globally map the original pixels values to a new
 *  set of pixel values using a table.  This routine can be used do perform
 *  this mapping.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to source image to be modified.  This image must
 *		     have a depth of 4 bit/pixel or 8 bits/pixel.  It may
 *		     have multiple channels.  If it has multiple channels,
 *		     the same map will be applied to all the channels.
 *  map::            Pointer to a table used to map the pixels.  If the
 *		     source image has depth 8 bits/pixel, this map should
 *		     contain 256 consecutive UInt8's.  Entry i is the new
 *		     8-bit value for pixels having a value of i.  If the
 *		     source image has depth 4 bit/pixel, the map should
 *		     be comprised of 16 UInt8's.  The new 4 bits should reside
 *		     in the lower four bits of each table byte.  As before,
 *		     entry i contains the new value for all pixels having
 *		     a value of i.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  Remember that the same map will be applied to all channels of a color
 *  image, which is probably not correct.  An operation like
 *  w_aieProcessRGB is used to correct the individual channels.  Alternatively,
 *  a client could use w_pixrGetChannel to extract each channel individually,
 *  calling this routine with a new map for each channel.
 *
 * FunctionCategory:::
 *  pixel mapping
 *
 * Keywords:::
 *  color, pixel mapping, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_map (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt8              *map,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, channels, i;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || map == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannelsAny,
		 cFrame32);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/* The inputs seem to be legal.  If this is a multi-channel
	 * image, map each channel individually.  If it's a one-channel
	 * image, just map that channel.
	 */
    if (channels != 1)
    {
	PIXR   *subS, *subD;

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &subD));
	    
	    ERR_PASS_CLEANUP( w_map (pIPWState, subS, map, &subD));
	}
    }
    else
    {
	UInt8   *pS, *pD;
	Int32    sBpl, dBpl;
	
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
	ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &dS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

	switch (dS)
	{
	    case 4:
		ERR_PASS_CLEANUP( i_map4To4 (pIPWState, pS,
					     wS, hS,
					     sBpl,
					     pD,
					     dBpl,
					     map));
		break;
	    case 8:
		ERR_PASS_CLEANUP( i_map8To8 (pIPWState, pS,
					     wS, hS,
					     sBpl,
					     pD,
					     dBpl,
					     map));
		break;
	    default:
		ERR_GEN_CLEANUP(1, cErrDepthNotSupported);
	}
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts each 8 bit/pixel channel of an image into a 4 bit/pixel channel.
 *
 * LongDescription:::
 *  Color and gray images often arrive with 8 bits/pixel, but these are big.
 *  One can quickly cut the memory requirements in half by only using the
 *  upper four bits of each pixel.  This routine does exactly that.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 8 bit/pixel input image.  This image may
 *		     have more than one channel.  Each channel of
 *		     multi-channel images is converted.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  This routine does not round pixels to the nearest 4 bit value.  It simply
 *  discards the lower 4 bits of each pixel and packs the remaining bits
 *  together.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, gray, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scale8To4 (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, channels, frame, i;
    ScaledInt1616	 xSdpi, ySdpi;
    Bool		 allocatedHere = cFalse;
    ColorType		 type;
    ColorMap		*map;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frame));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, prS, &channels));
    ERR_PASS( w_pixrGetType (pIPWState, prS, &type));
    ERR_PASS( w_pixrGetColorMap (pIPWState, prS, &map));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditionalNoInit(pIPWState, wS, hS, 4,
						     xSdpi, ySdpi, frame,
						     channels, type, map, &prD,
						     &rightFrameColsToClear,
						     &bottomFrameRowsToClear,
						     &allocatedHere));

	/* The inputs seem to be legal.  If this is a multi-channel
	 * image, map each channel individually.  If it's a one-channel
	 * image, just map that channel.
	 */
    if (channels != 1)
    {
	PIXR   *subS, *subD;

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &subD));
	    
	    ERR_PASS_CLEANUP( w_scale8To4 (pIPWState, subS, &subD));
	}
    }
    else
    {
	UInt8   *pS, *pD;
	Int32    sBpl, dBpl;
	
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

	ERR_PASS_CLEANUP( i_threshold8To4bpp (pIPWState, pS,
					      wS, hS,
					      sBpl,
					      pD,
					      dBpl));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts each 4 bit/pixel channel of an image into an 8 bit/pixel channel.
 *
 * LongDescription:::
 *  Occasionlly, devices need 8 bit/pixel images and only 4 bit/pixel images
 *  have been maintained.  This function converts 4 bit/pixel images into
 *  8 bit/pixel images.  This is done by making two copies of each pixel.
 *  Thus, 0x0 => 0x00, 0x1 => 0x11, ..., 0xE => 0xEE, 0xF => 0xFF.
 *<BR><BR>
 *  If a multi-channel image is supplied, all the channels are converted.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 4 bit/pixel input image.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  Duplicating each 4 bit pixel to form the corresponding 8 bit pixel
 *  has the nice property that it completely fills the range from 00 to
 *  0xFF in uniform increments.  Simply filling the bottom 4 bits of each
 *  pixel with a constant (like 0) means we don't fill the complete range
 *  of values.  For example, we want a 4 bit white to turn into a real 8 bit
 *  white, which is different from 0xF0.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, gray, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_depth4To8 (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, sBpl, dBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));
    
	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 8, xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

    ERR_PASS_CLEANUP( i_depth4To8bpp (pIPWState, pS, wS, hS,
				      sBpl, pD, dBpl, NULL));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts each 1 bit/pixel channel of an image into an 8 bit/pixel channel.
 *
 * LongDescription:::
 *  Occasionlly, devices need 8 bit/pixel images and only binary images
 *  have been maintained.  This function converts binary images into
 *  8 bit/pixel images.  This is done by making eight copies of each pixel.
 *  Thus, 0 => 0x00, 1 => 0xFF.
 *<BR><BR>
 *  If a multi-channel image is supplied, all the channels are converted.
 *<BR><BR>
 *  If the invert flag equals cTrue, the result is inverted, otherwise it
 *  is not.  This is often useful since the association of values to black
 *  and white is reversed in gray images vs binary images.  For example,
 *  0 == black in a gray image and 0 == white in a binary image.  Don't
 *  blame us, this is the way the world seems to work!
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  invert::	     if equal to cTrue, the output image is inverted,
 *		     i.e. 0 => 0xFF and 1 => 0x00.
 *		     if equal to cFalse, the output image is not inverted,
 *		     i.e. 0 => 0x00 and 1 => 0xFF.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_depth1To8 (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool                invert,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, sBpl, dBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));
    
	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 8,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

    ERR_PASS_CLEANUP( i_depth1To8bpp (pIPWState, pS, wS, hS, sBpl, pD, dBpl,
				      invert));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts each 1 bit/pixel channel of an image into an 16 bit/pixel channel.
 *
 * LongDescription:::
 *  Occasionlly, devices need 16 bit/pixel images and only binary images
 *  have been maintained.  Normally this happens when the 16-bit pixels are
 *  actually an array of numbers, each of which is closely associated with
 *  a pixel in some other image.  This function converts binary images into
 *  16 bit/pixel images.  This is done by making sixteen copies of each pixel.
 *  Thus, 0 => 0x0000, 1 => 0xFFFF.
 *<BR><BR>
 *  If a multi-channel image is supplied, all the channels are converted.
 *<BR><BR>
 *  If the invert flag equals cTrue, the result is inverted, otherwise it
 *  is not.  This is often useful since the association of values to black
 *  and white is reversed in gray images vs binary images.  For example,
 *  0 == black in a gray image and 0 == white in a binary image.  Don't
 *  blame us, this is the way the world seems to work!
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  invert::	     if equal to cTrue, the output image is inverted,
 *		     i.e. 0 => 0xFFFF and 1 => 0x0000.
 *		     if equal to cFalse, the output image is not inverted,
 *		     i.e. 0 => 0x0000 and 1 => 0xFFFF.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_depth1To16 (
    IPWState           *pIPWState,
    PIXR               *prS,
    Bool                invert,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, sBpl, dBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));
    
	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 16,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

    ERR_PASS_CLEANUP( i_depth1To16bpp (pIPWState, pS, wS, hS, sBpl, pD, dBpl,
				       invert));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert binary image to 8 bpp gray image using specified output vals.
 *
 * LongDescription:::
 *  The routine w_depth1To8 converts a binary image to a gray image by
 *  simply turning each 0 bit into a 0x00 byte and each 1 bit into a
 *  byte filled with 0xFF (unless the invert flag is set).  This routine
 *  turns each 0 bit into a byte filled with the client-supplied value
 *  val0 and each 1 bit into a byte filled with the client-supplied val1.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to binary input image.
 *  val0::           Output gray value that is substituted for each
 *		     0 in the binary image.
 *  val1::           Output gray value that is substituted for each
 *		     1 in the binary image.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  This routine could be useful when merging a binary image with a
 *  color image that already has a color map.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_map1To8 (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt8               val0,
    UInt8               val1,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, sBpl, dBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));
    
	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 8,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

    ERR_PASS_CLEANUP( i_map1To8bpp (pIPWState, pS, wS, hS, sBpl, pD, dBpl,
				    val0, val1));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*-----------------------------------------------------------------------*
 *                         Grayscale Histogram                           *
 *-----------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes histogram of intensity values from 4 bpp or 8 bpp image.
 *
 * LongDescription:::
 *  It is often useful to determine the number of pixels having each
 *  possible intensity value in an image.  One might use this to determine
 *  threshold levels, for example.  This routine computes this histogram
 *  by looking at every pixel in the specified sub-rectangle of an image.
 *  If the pointer to the rectangle is NULL, the entire image is inspected.
 *<BR><BR>
 *  This routine works on either 4 bit/pixel or 8 bit/pixel images.  The
 *  images must be single channel.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 4 bpp or 8 bpp image to be inspected.
 *  pRect::          Pointer to RECTANGLE structure specifying subrect
 *		     of this image.  If this pointer is NULL, the entire
 *		     image is inspected.
 *
 * Output:::
 *  ppHisto::        Pointer to Pointer to returned histogram.  *ppHisto
 *		     must always be NULL since we don't know the size
 *		     of the thing that the user might have allocated.
 *		     We always allocate the histogram here.
 *
 * Returns:::
 *  In addition to the normal errors, we return cErrPtrNotNull if *ppHisto
 *  is not NULL.
 *
 * Notes:::
 *  This function is a combination of the old functions histoPr, histo4Pr,
 *  histoRectPr and histoRect4Pr.  It looks at the image depth instead
 *  of forcing the client to do so.  It also looks at the rectangle pointer
 *  passed in.  If it is NULL, the entire image is histogrammed.  Otherwise,
 *  the rectangle passed in is used.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, adaptive thresholding, pixel mapping, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, P. Macleod, D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_histo (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    Int32             **ppHisto)
{
    PROLOGUE;
    Int32	*pHisto = NULL;
    Int32	 bpl, dS;
    Int8	 valid;
    UInt8	*pS;
    RECTANGL	localRect;

    *ppHisto = NULL;
    ERR_GEN(prS == NULL || ppHisto == NULL, cErrNullPtr);
    ERR_GEN(*ppHisto != NULL, cErrPtrNotNull);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

	/* if no rectangle is passed in, look at the entire image */
    if (pRect == NULL)
    {
	localRect.x = 0;
	localRect.y = 0;
	ERR_PASS( w_pixrGetWidth(pIPWState, prS, &(localRect.w)));
	ERR_PASS( w_pixrGetHeight(pIPWState, prS, &(localRect.h)));
	pRect = &localRect;
    }
    else	/* a rectangle was passed in, so check it */
    {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN(!valid, cErrOutOfBounds);
    }

	/* get the depth so we know how big a histogram to allocate */
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpl));

    if (dS == 4)
    {
	pHisto = (Int32 *) IPW_CALLOC (16, sizeof (Int32), cAllocatorData);
	ERR_GEN(pHisto == NULL, cErrNomem);

	ERR_PASS_CLEANUP( i_histoRect4 (pIPWState, pS, bpl, 
				        pRect->x + cFrameBits,
					pRect->y + cFrameBits,
				        pRect->w, pRect->h, 
				        pHisto));
    }
    else	/* this is an 8 bit/pixel image */
    {
	pHisto = (Int32 *) IPW_CALLOC (256, sizeof (Int32), cAllocatorData);
	ERR_GEN(pHisto == NULL, cErrNomem);

	ERR_PASS_CLEANUP( i_histoRect (pIPWState, pS, bpl, 
				       pRect->x + cFrameBits,
				       pRect->y + cFrameBits,
				       pRect->w, pRect->h, 
				       pHisto));
    }

    *ppHisto = pHisto;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pHisto, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Computes a histogram of intensity values by subsampling an image.
 *
 * LongDescription:::
 *  The regular histogram computation, w_histo, looks at every pixel in
 *  the rectangle given.  In almost all images, you get the same result
 *  by looking at a fairly small fraction of the pixels so long as the
 *  sample is spread out over the entire image.  That's what this
 *  routine does.
 *<BR><BR>
 *  If the rect pointer is NULL, the entire image is examimed.
 *<BR><BR>
 *  This routine works on either 4 bit/pixel or 8 bit/pixel images.  The
 *  images must be single channel.
 *
 * PrivateDescription:::
 *  There's a bit if trick in the subsampling.  You want to avoid sampling
 *  in a nice horizontal/vertical grid since the samples would be skewed
 *  by horizontal and vertical lines that happened to fall on sampling
 *  grid lines.  This routine essentially uses a skewed grid.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 4 bpp or 8 bpp image to be inspected.
 *  pRect::          Pointer to RECTANGLE structure specifying subrect
 *		     of this image.  If this pointer is NULL, the entire
 *		     image is inspected.
 *  factor::         The subsampling factor.  Roughly one pixel in every
 *		     (factor*factor) pixels is sampled.
 *
 * Output:::
 *  ppHisto::        Pointer to Pointer to returned histogram.  *ppHisto
 *		     must always be NULL since we don't know the size
 *		     of the thing that the user might have allocated.
 *		     We always allocate the histogram here.
 *
 * Returns:::
 *  In addition to the normal errors, we return cErrPtrNotNull if *ppHisto
 *  is not NULL.
 *
 * Notes:::
 *  This function is a combination of the old functions histoSubSampledPr and
 *  histoSubSampledRectPr.  It also looks at the rectangle pointer
 *  passed in.  If it is NULL, the entire image is histogrammed.  Otherwise,
 *  the rectangle passed in is used.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, adaptive thresholding, pixel mapping, gray
 *
 * Level:::
 *  heuristic (selection of the sample points)
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_histoSubSampled (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *pRect,
    Int32               factor,
    Int32             **ppHisto)
{
    PROLOGUE;
    Int32	*pHisto = NULL;
    Int32	 w, h, x, y, bpl, depth;
    UInt8	*pS;
    Int8	 valid;
    RECTANGL	 localRect;

    ERR_GEN(prS == NULL || ppHisto == NULL, cErrNullPtr);
    ERR_GEN(*ppHisto != NULL, cErrPtrNotNull);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

	/* if no rectangle is passed in, look at the entire image */
    if (pRect == NULL)
    {
	localRect.x = 0;
	localRect.y = 0;
	ERR_PASS( w_pixrGetWidth(pIPWState, prS, &(localRect.w)));
	ERR_PASS( w_pixrGetHeight(pIPWState, prS, &(localRect.h)));
	pRect = &localRect;
    }
    else	/* a rectangle was passed in, so check it */
    {
	ERR_PASS( w_isRectValid(pIPWState, prS, pRect, &valid));
	ERR_GEN(!valid, cErrOutOfBounds);
    }

    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &depth));

    pHisto = (Int32 *) IPW_CALLOC (1 << depth, sizeof (Int32), 
				   cAllocatorData);    
    ERR_GEN(pHisto == NULL, cErrNomem);

    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prS, &bpl));

	/* This is kind of tricky.  We want to use x and y to adjust
	 * pS, which now points to the upper left corner of the frame.
	 * Both i_histoSubSampled4 and i_histoSubSampled8 assume
	 * that pS points to the upper left corner of the frame.
	 * They will adjust it to point to the real image.  So,
	 * If [x,y] == [0,0], we don't want to move pS at all.
	 * If [x,y] == [0,1], we want to move down one line and so
	 * on.  We do NOT add in the frame bits in at this point
	 * because the i_histoSubSampled routines will do so later.
	 */
    x = pRect->x;
    y = pRect->y;
	/* However, the i_histoSubSampled routines also assume their
	 * widths and heights include the frame, which they immediately
	 * remove.  Therefore, we need to add in the frame width so
	 * it can be removed.
	 */
    w = pRect->w + (2 * cFrameBits); 
    h = pRect->h + (2 * cFrameBits);

    if (depth == 4)
    {
	    /* Always do it on an even nibble */
	if (x & 1)
	{
	    x -= 1;
	    w += 1;
	}

	/* Move to the right place */
	pS += (y * bpl) + (x >> 1);
	
	ERR_PASS_CLEANUP( i_histoSubSampled4 (pIPWState, pS, w, h,
					      bpl, factor, pHisto));
    }
    else
    {
	    /* Move to the right place */
	pS += (y * bpl) + x;

	ERR_PASS_CLEANUP( i_histoSubSampled8 (pIPWState, pS, w, h,
					      bpl, factor, pHisto));
    }

    *ppHisto = pHisto;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pHisto, cAllocatorData);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute histogram of intensity samples of image pixels covered by a mask.
 *
 * LongDescription:::
 *  It is often useful to compute a histogram of the pixels in some object
 *  in an image, like a block of text on tint.  This function takes a mask
 *  that exactly describes the object and computes a histogram of only the
 *  pixels under the mask.  In addition, it may be useful to look at only
 *  one tint region of the many on the page.  If so, specify a rectangle
 *  and only those pixels under the mask in that rectangle will be
 *  histogrammed.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 4 bpp or 8 bpp image to be inspected.
 *  prMask::         Pointer to binary image containing the mask.  This should
 *		     be the same size (roughly) as prS.  Their dimensions
 *		     can differ by no more than 32.  The two images will
 *		     be aligned at the upper left corner.
 *  pRect::          Pointer to RECTANGL structure specifying subrect
 *		     of this image.  If this pointer is NULL, the entire
 *		     image is inspected.
 *
 * Output:::
 *  ppHisto::        Pointer to Pointer to returned histogram.  *ppHisto
 *		     must always be NULL since we don't know the size
 *		     of the thing that the user might have allocated.
 *		     We always allocate the histogram here.
 *
 * Returns:::
 *  In addition to the normal errors, we return cErrPtrNotNull if *ppHisto
 *  is not NULL.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, adaptive thresholding, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_histoMask (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR               *prMask,
    RECTANGL	       *pRect,
    Int32             **ppHisto)
{
    PROLOGUE;
    Int32       depth, wS, hS, sBpl, mBpl, wM, hM;
    Int32      *pHisto = NULL;
    UInt8      *pS, *pM;
    RECTANGL	localRect;
    Int8	valid;

    ERR_GEN(prS == NULL || prMask == NULL || ppHisto == NULL, cErrNullPtr);
    ERR_GEN(*ppHisto != NULL, cErrPtrNotNull);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prMask, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* When doing binary reductions, we can lose a few pixels modulo
       a small power of two, so we allow the image to be up to 32 pixels
       larger, figuring that the frame will be zero, and therefore
       not histogrammed. */
    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetWidth (pIPWState, prMask, &wM));
    ERR_PASS( w_pixrGetHeight (pIPWState, prMask, &hM));
    ERR_PASS( w_pixrGetDepth (pIPWState, prS, &depth));

    /* Ensure image not more than 32 pixels wider or higher than mask */
    ERR_GEN(ABS(wS - wM) > 32 || ABS(hS - hM) > 32, cErrInvalidParm);

    /* if no rectangle is passed in, look at the entire image */
    if (pRect == NULL)
    {
	localRect.x = 0;
	localRect.y = 0;
	ERR_PASS( w_pixrGetWidth(pIPWState, prMask, &(localRect.w)));
	ERR_PASS( w_pixrGetHeight(pIPWState, prMask, &(localRect.h)));
	pRect = &localRect;
    }
    else	/* a rectangle was passed in, so check it */
    {
	ERR_PASS( w_isRectValid(pIPWState, prMask, pRect, &valid));
	ERR_GEN(!valid, cErrOutOfBounds);
    }

    /* Create the histogram */
    pHisto = (Int32 *) IPW_CALLOC (1 << depth, sizeof (Int32), 
				   cAllocatorData);    
    ERR_GEN(pHisto == NULL, cErrNomem);

    ERR_PASS_CLEANUP( w_pixrFramedHeight (pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prMask, &mBpl));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prMask, &pM));

    /* Compute the histogram */
    ERR_PASS_CLEANUP( i_histoMaskRect (pIPWState, pS, pRect->x + cFrameBits, 
				       pRect->y + cFrameBits, pRect->w, 
				       pRect->h, sBpl, depth, pM, mBpl,
				       pHisto));

    *ppHisto = pHisto;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pHisto, cAllocatorData);
    ERR_RETURN;
}
/*

 $Log: prxform1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.48  1998/02/10 19:58:44  ddavies
 prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v

 Add wrapper routine w_depth1To16.  This converts binary images to
 16 bit/pixel images.

 Revision 1.47  1998/01/14 00:14:34  ddavies
 prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.46  1997/07/03 21:49:33  jcraig
 prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v

 w_histoMask now takes a rectangle.  Only those pixels under the mask
 in that rectangle will be histogrammed.  if the rectangle is NULL,
 the entire image will be histogrammed (the previous behavior of
 w_histoMask)

 Revision 1.45  1997/04/15 22:28:03  ddavies
 prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.44  1996/01/04 20:25:51  ejaquith
 prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.43  1995/12/13  04:11:52  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.42  1995/11/13  20:14:55  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.41  1995/11/04  01:36:03  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.40  1995/11/03  22:48:50  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.39  1995/10/13  22:04:09  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.38  1995/10/05  04:27:45  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.37  1995/10/04  20:26:28  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Fix the way frames are handled in i_histoSubSampled.  The ipshared routine
 * called by w_histoSubSampled thinks it's doing an entire image, so the
 * frames must be added to the width and height instead of to the x and y
 * coords.
 *
 * Revision 1.36  1995/10/03  20:54:33  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Removed obsoleted thresholding/subsample keyword.
 *
 * Revision 1.35  1995/10/03  18:13:11  ddavies
 * ddavies on Tue Oct 3 11:12:31 PDT 1995
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Eliminate check for pRect being NULL in w_histoSubSampled.  pRect being
 * NULL simply means the entire image should be histogrammed.
 *
 * Revision 1.34  1995/10/03  01:20:44  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.33  1995/09/21  22:58:26  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.32  1995/09/19  17:45:41  jcraig
 * jcraig on Tue Sep 19 10:44:50 PDT 1995
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * in w_histoSubSampled and w_histoSubSampledRect, depth was used before
 * being intialized.  fix it.
 *
 * Revision 1.31  1995/09/15  18:01:12  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.30  1995/09/11  22:11:47  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.29  1995/09/10  00:33:48  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.28  1995/09/07  01:34:16  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.27  1995/09/06  20:22:21  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.26  1995/09/01  01:18:03  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.25  1995/08/31  21:38:12  ejaquith
 * ejaquith on Thu Aug 31 14:38:11 PDT 1995
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.24  1995/08/25  01:37:56  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.23  1995/08/23  17:52:17  macleod
 * macleod on Wed Aug 23 10:51:43 PDT 1995
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * histoSubSampledPr checked the depth incorrectly, causing an invalid
 * parameter error for all depths
 *
 * Revision 1.22  1995/08/17  22:02:59  ejaquith
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.21  1995/07/29  04:47:51  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.20  1995/04/27  21:30:41  macleod
 * macleod on Thu Apr 27 14:29:58 PDT 1995
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Added map1To8Pixr, which does an upconvert from binary to 8 bits/pixel,
 * with the values for "0" and "1" passed in.
 *
 * Revision 1.19  1994/10/07  19:14:59  lperry
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.18  1994/09/17  18:08:02  ddavies
 * ddavies on Sat Sep 17 11:05:27 PDT 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Add extra NULL tables args to calls to i_depth4To8bpp, i_threshold8To4bpp,
 * i_highpassfiltI4To1bpp, i_highpassfiltI and i_highpassfiltI4To1bpp.
 *
 * Revision 1.17  1994/09/16  17:39:22  jcraig
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.16  1994/06/16  22:41:56  macleod
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Added histoMask, which does a "masked histogram", meaning that it takes
 * an image (4- or 8-bit) and a binary mask which should be the same size,
 * and histograms those points in the image which have a corresponding "1"
 * in the mask. Note that we allow the mask to be "slightly" smaller
 * than the image, due to the fact that the mask is often computed by doing
 * binary reductions and then expansions on the image, so the mask can be
 * a few pixels smaller. It should be ok, since we'll just end up getting
 * zeroes in the frame, and therefore not histogramming those portions. If
 * those portions are important, make an accurate mask.
 *
 * Revision 1.15  1994/06/07  02:40:02  macleod
 * macleod on Mon Jun 6 19:40:00 PDT 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.14  1994/05/24  00:40:42  macleod
 * macleod on Mon May 23 17:40:01 PDT 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * swapped src and dest in mapPr recursive call for composite PIXRs, which
 * was reversed
 *
 * Revision 1.13  1994/05/10  16:22:19  lperry
 * lperry on Tue May 10 09:21:25 PDT 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Removed several unused variables to to eliminate compiler grumbles
 *
 * Revision 1.12  1994/05/06  02:12:42  macleod
 * macleod on Thu May 5 19:12:34 PDT 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Added wrappers for 4to8 and 1to8 upconversions
 *
 * Revision 1.11  1994/04/12  01:15:34  macleod
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.10  1994/03/25  22:47:40  ddavies
 * ddavies on Fri Mar 25 14:45:54 PST 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Add implementations of histo4Pr and histoRect4Pr.  These are wrappers for ipshared
 * routines that calculate histograms of 4 bit images.  Also generally fix
 * up messages and comments.  In particular, make sure the rectangle sent
 * to i_histoRect and i_histoRect4 has the origin at the OUTSIDE of the frame,
 * not the inside.
 *
 * Revision 1.9  1994/03/02  03:59:00  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Added IP_YIELD to give up control after creating dest images during
 * binarization ops.
 *
 * Revision 1.8  1994/02/18  05:23:34  macleod
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Fixed a bug in histoRectPr, where it was passing the image to i_histoRect,
 * when it wanted a pointer to the frame. The result is that some of the
 * frame got histogrammed.
 *
 * Revision 1.7  1994/01/28  04:28:25  ddavies
 * ddavies on Thu Jan 27 20:27:26 PST 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Changed call to i_scale8To4 to call to i_threshold8To4bpp since the
 * ipshared routine was renamed.
 *
 * Revision 1.6  1994/01/06  22:50:02  macleod
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Removed map4To4Pr and map8To8Pr, put mapPixr and mapPr in their place
 *
 * Revision 1.5  1994/01/04  02:11:20  macleod
 * macleod on Mon Jan 3 18:09:51 PST 1994
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * added histoSubSampledRectPr, which is like histoSubSampledPr,
 * but takes a rectangle which defines the region to histogram
 *
 * Revision 1.4  1993/12/03  01:17:51  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/24  20:16:11  macleod
 * macleod on Wed Nov 24 12:16:05 PST 1993
 *
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Added map8To8Pr
 *
 * Revision 1.2  1993/11/06  03:15:17  ddavies
 * prxform1.c -> /product/ipcore/ipwrap/src/RCS/prxform1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

