/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  prxform3.c--The only remaining routine here is the one used to produce
 *		2 bit/pixel images.  These are procedures used
 *	       	by the regression tests and maybe by the DAE.  They're
 *		not used by EZScan or the glyph code.
 *
 *     Grayscale thresholding:
 *	     w_thresholdTo2bpp()
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "prxform.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: prxform3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");



/*-----------------------------------------------------------------------*
 *                       Grayscale Thresholding                          *
 *-----------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Produces a 2 bpp image from an 8 bpp or 4 bpp image.
 *
 * LongDescription:::
 *  Two bit/pixel images aren't good for a lot, but there are some 2-bit
 *  print engines.  This function converts normal gray images into
 *  2 bit/pixel image by simply discarding the low order bits of each pixel.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to 4 bit/pixel or 8 bit/pixel input image.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_thresholdTo2bpp (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    Int32		 wS, hS, dS, sBpl, dBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit(pIPWState, wS, hS, 2,
					     xSdpi, ySdpi,
					     cFrameBits, &prD,
					     &rightFrameColsToClear,
					     &bottomFrameRowsToClear,
					     &allocatedHere));

	/* The inputs seem to be legal.  Do the real work. */
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));

	/* fill it */
    if (dS == 4)
	ERR_PASS_CLEANUP( i_threshold4To2bpp (pIPWState, pS, wS, hS, sBpl,
					      pD, dBpl,
					      NULL));
    else	/* depth of prS must be 8 since we checked above */
	ERR_PASS_CLEANUP( i_thresholdTo2bpp (pIPWState, pS, wS, hS, sBpl,
					     pD, dBpl,
					     NULL, NULL, NULL, NULL));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*

 $Log: prxform3.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.28  1998/01/14 00:14:39  ddavies
 prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.27  1997/04/15 22:28:15  ddavies
 prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.26  1996/01/04 20:25:54  ejaquith
 prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.25  1995/12/13  04:11:55  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.24  1995/11/13  20:14:59  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.23  1995/10/13  22:04:12  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.22  1995/10/05  04:27:49  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.21  1995/10/03  20:54:37  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Removed obsoleted thresholding/subsample keyword.
 *
 * Revision 1.20  1995/10/03  01:20:49  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.19  1995/09/21  22:58:30  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/15  18:01:17  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.17  1995/09/11  22:11:51  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.16  1995/09/10  00:33:52  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.15  1995/09/07  01:34:22  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.14  1995/09/06  20:22:27  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.13  1995/09/01  01:18:14  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.12  1995/08/31  21:38:17  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/17  22:03:06  ejaquith
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.10  1995/07/29  04:48:34  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.9  1994/10/07  19:15:09  lperry
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.8  1994/09/17  18:13:36  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Add extra NULL table args to calls to i_passfilt4To1bpp, i_passfiltI,
 * i_passfilt, i_highpassfilt4To1bpp, i_thresholdI4To1bpp, i_threshold4To2bpp,
 * i_thesholdTo2bpp, i_thresholdI2To1bpp, i_theshold, i_threshold4To1bpp,
 * i_threshold2To1bpp and i_passfiltI4To1bpp.
 *
 * Revision 1.7  1994/09/16  17:39:32  jcraig
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.6  1994/07/30  00:08:47  dferrell
 * dferrell on Fri Jul 29 17:07:16 PDT 1994
 *
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Modified thresholdTo2bppPr () so that it can handle 4->2 depth conversion
 * using the new i_threshold4To2bpp () ipshared function.
 *
 * Revision 1.5  1994/03/24  20:44:16  ddavies
 * ddavies on Thu Mar 24 12:43:59 PST 1994
 *
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Add implementations of bandpass4To1bppPr and bandpassI4To1bppPr.  These are the
 * wrappers for routines used to do bandpass filtering on 4 bit images,
 * producing 1 bit images.
 *
 * Revision 1.4  1994/03/02  03:59:23  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Added IP_YIELD to give up control after creating dest images during
 * binarization ops.
 *
 * Revision 1.3  1993/12/03  01:18:01  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:15:31  ddavies
 * prxform3.c -> /product/ipcore/ipwrap/src/RCS/prxform3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

