/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  rast1.c:
 *      Includes
 *
 *           low-level raster interfaces
 *                      w_rasterop()
 *                      w_pixrGetPixel()
 *                      w_pixrPutPixel()
 *                      w_pixrPutLine()
 *                      w_pixrPutRect()
 */




#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imageref.pub"

/* prototypes */
#include "rast.pub"
#include "props.pub"
#include "shrrast.pub"
#include "rectops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: rast1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*--------------------------------------------------------------------*
 *                     Low-level raster interfaces                    *
 *--------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Perform image raster operation, potentially combining another image
 *
 * LongDescription:::
 *  Do things like image setting, clearing, ANDing, ORing, inverting, etc.
 *  If the image(s) is(are) multi-channeled, do the operation on all channels.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  sx::             x-coordinate in spr of leftmost pel in source rectangle
 *  sy::             y-coordinate in spr of topmost  pel in source rectangle
 *  spr::            source pixr (may be NULL).
 *                   must have same number of channels as dest pixr.
 *  op::             operation code (see notes for examples)
 *  dx::             x-coordinate in dpr of leftmost pel in destination rect
 *  dy::             y-coordinate in dpr of topmost  pel in destination rect
 *  dw::             width of destination rectangle (in pels)
 *  dh::             height of destination rectangle (in pels)
 *
 * Output:::
 *  dpr::            desination pixr (must be non-NULL!)
 *
 * Returns:::
 *
 * Notes:::
 *  op::
 *   Things like: PIX_CLR, PIX_SET, PIX_SRC, PIX_DST,
 *   PIX_DST ^ PIX_SRC, PIX_DST & PIX_SRC, PIX_NOT( PIX_DST ) | PIX_SRC,
 *   PIX_NOT(PIX_DST), PIX_NOT(PIX_DST | PIX_SRC).
 *   May be anded with PIX_DONTCLIP.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, multi-channel, cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rasterop (
    IPWState           *pIPWState,
    Int32               sx,
    Int32               sy,
    PIXR               *spr,
    Int32               op,
    Int32               dx,
    Int32               dy,
    Int32               dw,
    Int32               dh,
    PIXR               *dpr)
{
    PROLOGUE;
    Int32      channels;

    ERR_GEN_PIXR(dpr, cByteOrderBig, cDepthAny, cChannelsAny, 
		 cFrame0 | cFrame32);
    if (spr != NULL)
	ERR_GEN_PIXR(spr, cByteOrderBig, cDepthAny, cChannelsAny,
		     cFrame0 | cFrame32);

    ERR_PASS( w_pixrGetChannelCount (pIPWState, dpr, &channels));

    if (channels == 1)
    {
	Int32   dfh, dfw, df, dd, dbpl;
	UInt8  *pdImage;
	
	ERR_PASS( w_pixrGetImageAll(pIPWState, dpr, &pdImage));
	ERR_PASS( w_pixrFramedWidth(pIPWState, dpr, &dfw));
	ERR_PASS( w_pixrFramedHeight(pIPWState, dpr, &dfh));
	ERR_PASS( w_pixrGetFrame(pIPWState, dpr, &df));
	ERR_PASS( w_pixrGetDepth(pIPWState, dpr, &dd));
	ERR_PASS( w_pixrGetBpl(pIPWState, dpr, &dbpl));

	

	if (spr != NULL)
	{
	    Int32   sfh, sfw, sf, sd, sbpl;
	    UInt8  *psImage;

	    ERR_PASS( w_pixrGetChannelCount (pIPWState, spr, &channels));
	    ERR_GEN(channels != 1, cErrChannelsNotEqual);

	    ERR_PASS( w_pixrGetImageAll(pIPWState, spr, &psImage));
	    ERR_PASS( w_pixrFramedWidth(pIPWState, spr, &sfw));
	    ERR_PASS( w_pixrFramedHeight(pIPWState, spr, &sfh));
	    ERR_PASS( w_pixrGetFrame(pIPWState, spr, &sf));
	    ERR_PASS( w_pixrGetDepth(pIPWState, spr, &sd));
	    ERR_PASS( w_pixrGetBpl(pIPWState, spr, &sbpl));
	    
	    ERR_PASS( i_rasterOpFull (pIPWState, pdImage, dfw, dfh, df, df,
				      dx, dy, dw, dh, dd, dbpl,
				      op, psImage, sfw, sfh, sx, sy,
				      sf, sf, sd, sbpl));
	}
	else
	    ERR_PASS( i_rasterOpFull (pIPWState, pdImage, dfw, dfh, df, df,
				      dx, dy, dw, dh, dd, dbpl,
				      op, NULL, 0, 0, 0, 0, 0, 0, 0, 0));
    }
    else
    {
	Int32   schannels, i;
	
	if (spr != NULL)
	{
	    ERR_PASS( w_pixrGetChannelCount (pIPWState, spr, &schannels));
	    ERR_GEN(schannels != channels, cErrChannelsNotEqual);
	}
	
	for (i = 0; i < channels; i++)
	{
	    PIXR   *dch;

	    ERR_PASS( w_pixrGetChannel(pIPWState, dpr, i, &dch));

	    if (spr != NULL)
	    {
		PIXR   *sch;

		ERR_PASS( w_pixrGetChannel(pIPWState, spr, i, &sch));
		ERR_PASS( w_rasterop (pIPWState, 
				      sx, sy,
				      sch,
				      op,
				      dx, dy, dw, dh,
				      dch));
	    }
	    else
	    {
		ERR_PASS( w_rasterop (pIPWState, 
				      0, 0,
				      NULL,
				      op,
				      dx, dy, dw, dh,
				      dch));
	    }
	}
    }

    NORMAL_RETURN;	
}
    

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the value of a single pixel in a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            ptr to pixr, single channel, depth 1,2,4,8
 *  x::              x-coordinate in prs of pel to read
 *  y::              y-coordinate in prs of pel to read
 *
 * Output:::
 *  pValue::         Ptr to value of pixel at pixr location
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetPixel (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               x,
    Int32               y,
    Int32              *pValue)
{
    PROLOGUE;
    Int32        w, h, d, wpl;
    UInt32      *data, *line;
    UInt8       *pImage, m;

    ERR_GEN(pValue == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prs, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prs, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prs, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, prs, &d));
    ERR_PASS( w_pixrGetImage(pIPWState, prs, &pImage));
    ERR_PASS( w_pixrGetBpl(pIPWState, prs, &wpl));
    wpl >>= 2;

    ERR_GEN(x < 0 || x >= w || y < 0 || y >= h, cErrOutOfBounds);

    data = (UInt32 *)pImage;

    line = data + wpl * y;

    switch (d) {
    case 1:
	*pValue = GET_IMAGE_BIT(line, x);
	break;
    case 8:
	*pValue = GET_IMAGE_BYTE(line, x);
	break;
    case 4:
	*pValue = GET_IMAGE_BYTE(line, x/2);
	if ( (x & 1) == 0 ) *pValue >>= 4;
	*pValue &= 0x0f;
	break;
    default: /* 2 */
	*pValue = GET_IMAGE_BYTE(line, x/4);
	m = x % 4;
	if      (m == 0) *pValue >>= 6;
	else if (m == 1) *pValue >>= 4;
	else if (m == 2) *pValue >>= 2;
	*pValue &= 0x03;
	break;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the value of a single pixel in a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prd::            ptr to pixr, 1 channel, depth 1 or 8 -or- 3 channels,
 *                   each depth 8 (it will write same value at each channel).
 *  x::              x-coordinate in prd of pel to write
 *  y::              y-coordinate in prd of pel to write
 *  val::            value to write
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrPutPixel (
    IPWState           *pIPWState,
    PIXR               *prd,
    Int32               x,
    Int32               y,
    Int32               val)
{
    PROLOGUE;
    Int32        w, h, d, wpl, channels, channel;
    UInt32      *data, *line;
    UInt8       *pImage;
    PIXR        *ch = NULL;

    ERR_GEN_PIXR(prd, cByteOrderBig, cDepth1 | cDepth8,
		 cChannels1 | cChannels3, cFrame32);

    ERR_PASS( w_pixrGetBpl(pIPWState, prd, &wpl));
    wpl >>= 2;
    ERR_PASS( w_pixrGetDepth(pIPWState, prd, &d));
    ERR_PASS( w_pixrGetWidth(pIPWState, prd, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prd, &h));
    ERR_GEN(x < 0 || x >= w || y < 0 || y >= h, cErrOutOfBounds);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prd, &channels));

    for (channel = 0; channel < channels; channel++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, prd, channel, &ch));

	ERR_PASS( w_pixrGetImage(pIPWState, ch, &pImage));
	data = (UInt32 *)pImage;
	
	line = data + wpl * y;
	if (d == 1)
	{
	    if (val)
		SET_IMAGE_BIT(line, x);
	    else
		CLEAR_IMAGE_BIT(line, x);
	}
	else /* d == 8 */
	    SET_IMAGE_BYTE(line, x, (UInt8) val);
    }
    

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the value of pixels in a line in a PIXR
 *
 * LongDescription:::
 *  We'll use Bresenham's line-drawing algorithm here.
 *  This type of line drawing plots a series of points (x,y)
 *  either diagonally or laterally with respect to the previous point.
 *  [Note: This means lines are 8-connected].
 *  Movement from point pixel to point pixel always proceeds in a
 *  dominant lateral direction (parellel to the axis of the larger
 *  dimension).
 *  <BR>
 *  Pixels beyond the image edge are not drawn.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prd::            Ptr to pixr, single channel, depth 1,2,4,8
 *  x0::             First (x,y) pel line in prd to write
 *  y0::             
 *  x1::             Last (x,y) pel line in prd to write
 *  y1::             
 *  val::            value to write
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * Citations:::
 *  Bresenham
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrPutLine (
    IPWState           *pIPWState,
    PIXR               *prd,
    Int32               x0,
    Int32               y0,
    Int32               x1,
    Int32               y1,
    Int32               val)
{
    PROLOGUE;
    Int32        w, h, d, bpl, f;
    UInt8       *pImage;
    Int32	xmax, ymax;

    ERR_GEN_PIXR(prd, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prd, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prd, &h));

    xmax = w - 1;
    ymax = h - 1;

    /* Look for cases where the line extends beyond the image boundary */
    if ((x0 < 0 && x1 < 0) ||	/* Line is to the left of the image */
	(y0 < 0 && y1 < 0) ||	/* Line is above the image */
	(x0 > xmax && x1 > xmax) ||	/* Line is to the right of the image */
	(y0 > ymax && y1 > ymax)) {	/* Line is below the image */
	NORMAL_RETURN;
    }
    if (x0 < 0 || x1 < 0 || y0 < 0 || y1 < 0 ||
	x0 > xmax  || x1 > xmax || y0 > ymax || y1 > ymax) {
	Int32 dx, dy;

	dx = x1 - x0;
	dy = y1 - y0;

	if (x0 < 0) {
	    y0 = y1 - (x1 * dy + dx / 2) / dx;
	    x0 = 0;
	}
	if (x1 < 0) {
	    y1 = y0 - (x0 * dy + dx / 2) / dx;
	    x1 = 0;
	}
	if (x0 > xmax) {
	    y0 = y1 - ((x1 - xmax) * dy + dx / 2) / dx;
	    x0 = xmax;
	}
	if (x1 > xmax) {
	    y1 = y0 - ((x0 - xmax) * dy + dx / 2) / dx;
	    x1 = xmax;
	}
	if (y0 < 0) {
	    x0 = x1 - (y1 * dx + dy / 2) / dy;
	    y0 = 0;
	}
	if (y1 < 0) {
	    x1 = x0 - (y0 * dx + dy / 2) / dy;
	    y1 = 0;
	}
	if (y0 > ymax) {
	    x0 = x1 - ((y1 - ymax) * dx + dy / 2) / dy;
	    y0 = ymax;
	}
	if (y1 > ymax) {
	    x1 = x0 - ((y0 - ymax) * dx + dy / 2) / dy;
	    y1 = ymax;
	}
    }

    ERR_PASS( w_pixrFramedWidth(pIPWState, prd, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prd, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, prd, &d));
    ERR_PASS( w_pixrGetFrame(pIPWState, prd, &f));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prd, &pImage));
    ERR_PASS( w_pixrGetBpl(pIPWState, prd, &bpl));

    ERR_PASS( i_rasterLine (pIPWState, pImage, w, h, d,
			    bpl, f, f, x0, y0, x1, y1, val));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the value of pixels lying ON (not in) a rectangle in a PIXR
 *
 * LongDescription:::
 *  Given rect's x,y,w,h of and a desired thickness, the end result will be
 *  a rect with interior dimensions of (w-2*thickness) x (h-2*thickness)
 *  and exterior dimensions of (w) x (h), drawn in color val.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prd::            ptr to pixr, single channel, depth 1,2,4,8
 *  rect::           Rectangle to draw
 *  thick::          Thickness of rectangle sides, moving interior-wise
 *                   to rect.  If thickness is wider or
 *                   taller than the rectangle, we just fill entire rectangle
 *                   in.
 *  val::            Color/value to draw rectangle with
 *
 * Returns:::
 *
 * Notes:::
 *  clipping::
 *    If rect is not valid, it clips the rect to the Pixr.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrPutRect (
    IPWState           *pIPWState,
    PIXR               *prd,
    RECTANGL           *rect,
    Int32               thick,
    Int32               val)
{
    PROLOGUE;
    RECTANGL     *rectn = NULL;
    Int32         i;
    ColorRecord   color;
    RECTANGL      paintRect;
    Int8         valid;

    ERR_GEN_PIXR(prd, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);
    
    ERR_GEN(rect == NULL, cErrNullPtr);

    rectn = NULL;

    ERR_PASS( w_isRectValid(pIPWState, prd, rect, &valid));
    
    if (!valid)
	/*  rect exists, but extends beyond boundaries of pixr */
	ERR_PASS( w_clipRectToPixr(pIPWState, prd, rect, &rectn));
    else 
	ERR_PASS( w_copyRect(pIPWState, rect, &rectn));

    if (rectn->w == 0 || rectn->h == 0)
    {
	/* No part of the rect overlaps the pixr!  We are done.
	 */
	ERR_PASS( w_destroyRect(pIPWState, &rectn));
	NORMAL_RETURN;
    }

    /* Check for invalid specified thickness! */
    ERR_GEN_CLEANUP(thick < 1, cErrInvalidParm);

    /* Fill the color record (if there's a need the interface
     * could be changed to using a ColorRecord instead of just a
     * single int value)
     */
    for (i = 0; i < cMaxColorPlanes; i++)
	color[i] = (UInt8)val;

    /* Only make it as thick as possible */
    if (thick > (MIN(rectn->w, rectn->h)) / 2)
    {
	/* Fill the whole rectangle in! */
	ERR_PASS_CLEANUP( w_paint(pIPWState, prd, rectn, color));
    }
    else
    {
	/* Draw the top bar */
	paintRect.x = rectn->x;
	paintRect.y = rectn->y;
	paintRect.w = rectn->w;
	paintRect.h = thick;
	ERR_PASS_CLEANUP( w_paint(pIPWState, prd, &paintRect, color));

	/* Draw the bottom bar */
	paintRect.x = rectn->x;
	paintRect.y = rectn->y + rectn->h - thick;
	paintRect.w = rectn->w;
	paintRect.h = thick;
	ERR_PASS_CLEANUP( w_paint(pIPWState, prd, &paintRect, color));

	/* Draw the left bar */
	paintRect.x = rectn->x;
	paintRect.y = rectn->y;
	paintRect.w = thick;
	paintRect.h = rectn->h;
	ERR_PASS_CLEANUP( w_paint(pIPWState, prd, &paintRect, color));

	/* Draw the right bar */
	paintRect.x = rectn->x + rectn->w - thick;
	paintRect.y = rectn->y;
	paintRect.w = thick;
	paintRect.h = rectn->h;
	ERR_PASS_CLEANUP( w_paint(pIPWState, prd, &paintRect, color));
    }

    ERR_PASS_CLEANUP( w_destroyRect(pIPWState, &rectn));
    NORMAL_RETURN;
    

ERR_CLEANUP:
    w_destroyRect(pIPWState, &rectn);
    ERR_RETURN;
}


/*

 $Log: rast1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.3  1998/01/31 01:06:32  aber
 rast1.c -> /product/ipcore/ipwrap/src/RCS/rast1.c,v

 Allow line drawing routine to extend beyond the image oundaries. Calculate
 the edge interception point in the case of crossing the boundary.

 Revision 1.2  1998/01/14 00:14:42  ddavies
 rast1.c -> /product/ipcore/ipwrap/src/RCS/rast1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.1  1997/06/26 19:42:46  ddavies
 Initial revision

 Revision 1.33  1996/05/18 00:36:48  ejaquith
 ejaquith on Fri May 17 17:36:10 PDT 1996

 props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v

 Modified w_pixrPutPixel() to work with 24-bit images.
 It now puts the same value in each of the 3 8-bit planes.

 Revision 1.32  1996/01/04 20:25:34  ejaquith
 props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.31  1995/12/13  04:11:27  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.30  1995/11/16  04:13:02  ejaquith
 * ejaquith on Wed Nov 15 20:11:32 PST 1995
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Changed w_pixrPutRect() to fill in entire rect if desired line thickness
 * so wide it would fill it anyway, or if line thickness was thicker than
 * width or height.  (Previously it returned an error under said conditions).
 *
 * Revision 1.29  1995/11/13  20:14:33  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.28  1995/10/16  22:58:13  dferrell
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.27  1995/10/13  22:03:52  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/10/12  04:26:26  ddavies
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.25  1995/09/29  00:36:00  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.24  1995/09/21  22:58:06  ejaquith
 * ejaquith on Thu Sep 21 15:58:05 PDT 1995
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.23  1995/09/16  02:02:17  ejaquith
 * ejaquith on Fri Sep 15 19:01:46 PDT 1995
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Added allowable to pass frameless PIXR to w_rasterop().
 * (The glyph IPWrap test uses it).
 *
 * Revision 1.22  1995/09/12  16:55:26  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.21  1995/09/08  00:15:30  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.20  1995/09/07  01:34:04  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.19  1995/09/06  20:22:10  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.18  1995/09/01  02:21:06  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.17  1995/08/31  21:15:55  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.16  1995/08/25  01:37:45  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.15  1995/08/17  22:01:57  ejaquith
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.14  1995/07/10  23:10:27  ejaquith
 * ejaquith on Mon Jul 10 16:09:00 PDT 1995
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Added robustness to pixrPutRect().
 * Routine will now gracefully free anything allocated internally
 * when exiting on error.
 * It will also exit (not on error) if the rect desired doesn't
 * overlap the pixr at all.
 *
 * Revision 1.13  1995/04/22  19:27:37  ejaquith
 * ejaquith on Sat Apr 22 12:26:07 PDT 1995
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Updated pixrPutRect() to work with non-binary images by simply
 * making four calls to paintPr(), one for each rule in the rect.
 *
 * Revision 1.12  1995/04/15  01:27:56  ejaquith
 * ejaquith on Fri Apr 14 18:26:48 PDT 1995
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Added two new routines, pixrPutLine() and pixrPutRect().
 * The former is a wrapper to the ipshared routine i_rasterLine().
 * The latter will draw a rect in an image of some given thickness and
 * color using four pixr_rop invocations (one for each side).
 *
 * Revision 1.11  1994/09/30  15:52:36  lperry
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.10  1994/09/29  23:13:31  lperry
 * lperry on Thu Sep 29 16:12:57 PDT 1994
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Modify to use props.pub instead of pixrpros.h
 *
 * Revision 1.9  1994/09/16  17:38:10  jcraig
 * jcraig on Fri Sep 16 10:37:46 PDT 1994
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * include imageref.pub instead of bitacces.h
 *
 * Revision 1.8  1994/06/24  18:43:05  macleod
 * macleod on Fri Jun 24 11:42:39 PDT 1994
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * removed pixrVector, which caused a warning and wasn't being used anyway
 *
 * Revision 1.7  1994/06/23  05:54:03  macleod
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Added some routines from ALPLIB which are used by the grayscale
 * morphology routines.
 *
 * Revision 1.6  1994/06/07  02:39:27  macleod
 * macleod on Mon Jun 6 19:39:26 PDT 1994
 *
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.5  1994/05/05  01:45:02  ddavies
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Change sx,sy from NULL,NULL to 0,0 in call to pixr_rop.
 *
 * Revision 1.4  1994/04/12  01:14:53  macleod
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Added support for color PIXRs
 * If you pass a color PIXR, the op is done on all channels
 *
 * Revision 1.3  1993/12/03  01:17:29  ddavies
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:14:49  ddavies
 * props1.c -> /product/ipcore/ipwrap/src/RCS/props1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

