/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  readfile.c--reads image on file into memory.  Includes subroutines
 *
 *	Generic
 *		    w_read()
 *		    w_readTest()
 *		    w_fread()
 *		    w_ffindFormat()
 *
 *	Who wanted this?
 *		    w_getLine()
 */


#include <stdio.h>

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "imageref.pub"
#include "alplib.h"
#include "defines.pub"
#include "frames.pub"
#include "rastrfil.h"
#include "ptypes.h"
#include "gdefs.h"
#include "pcx.h"
#include "shrcc.pub"

/* prototypes */
#include "props.pub"
#include "io.pub"
#include "iotiff.pub"
#include "io.prv"
#include <ctype.h>
#include "shrpixr.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: readfile.c,v $; $Revision: 1.2 $; $Date: 2001/09/28 23:38:22 $");




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write header info of image file to a stream.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to stream to write header info to
 *  fileheader::     Ptr to file header FH struct.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_fprintHeader (
    IPWState           *pIPWState,
    FILE               *fp,
    FH                 *fileheader)
{
    PROLOGUE;

    ERR_GEN(fp == NULL  ||  fileheader == NULL, cErrNullPtr);

#ifdef _UNIX_
    fprintf(fp, "  Filename: %s\n", fileheader->filename);
    fprintf(fp, "  File type: %ld (see imagerep.h for details)\n",
	    fileheader->type);
    if (fileheader->type != NO_TYPE)
    {
	fprintf(fp, "  Width = %ld;   Height = %ld\n",
		fileheader->width, fileheader->height);
	if (fileheader->Bpl)
	    fprintf(fp, "  Bytes/line = %ld\n", fileheader->Bpl);

	fprintf(fp, "  Total image bytes = %ld\n\n", fileheader->totalBytes);
    }
#endif  /* _UNIX_ */
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find the image format of some file.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to read image from.
 *
 * Output:::
 *  pFormat::        Ptr to format of image, one of values from imagerep.h.
 *
 * Returns:::
 *
 * Notes:::
 *  stream::
 *   This does not assume the stream initially points
 *   to the beginning of the file.  When it exits, the
 *   pointer has been reset to the beginning of the file.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_ffindFormat (
    IPWState           *pIPWState,
    FILE               *fp,
    Int32              *pFormat)
{
    PROLOGUE;
    Int32                type, magic;
    struct rasterfile   *sunHeader;
    UInt16               passwd, intel16;
    char                 pnmHeader[2];

    ERR_GEN(pFormat == NULL  ||  fp == NULL, cErrNullPtr);

    ERR_GEN (!fp, cErrInvalidParm);

	/* if the file is piped through stdin, the only allowed
	 * formats are SUNRAS and SUNRAS_COMP.  The reason is that DJFAX
	 * required seeking to parse the file and page headers.
	 */
    if (fp == stdin)
    {
	*pFormat = SUNRAS;  /* this will be refined in readSunHeaderPr() */
	NORMAL_RETURN;
    }

        /*  check for SUN  */
    rewind(fp);
    sunHeader = (struct rasterfile *)IPW_MALLOC(sizeof(struct rasterfile),
						cAllocatorData);
    ERR_GEN(!sunHeader, cErrNomem);

    if (fread((char *)sunHeader, 1, sizeof(struct rasterfile), fp) ==
	sizeof(struct rasterfile))
    {
	magic = FLIP_INT(sunHeader->ras_magic);
	type = FLIP_INT(sunHeader->ras_type);
	IPW_FREE(sunHeader, cAllocatorData);
	if (magic == RAS_MAGIC)
	{
	    if (type == RT_STANDARD)
		*pFormat = SUNRAS;
	    else if (type == RT_BYTE_ENCODED)
		*pFormat = SUNRAS_COMP;
	    else
		*pFormat = NO_TYPE;

	    rewind(fp);
	    NORMAL_RETURN;
	}
    }
    else 
    {
	IPW_FREE(sunHeader, cAllocatorData);
    }


    /* Get the first 2 characters in the file */
    rewind(fp);
    fread((char *)&passwd, 1, 2, fp);
    rewind(fp);

#if 0  /* R.I.P.  11sep95 -ejaquith */
        /*  check for DJFAX  */
    if ((FLIP_SHORT(passwd) & 0x0000ffff) == DJFAX_MAGIC)
    {
	*pFormat = DJFAX;
	NORMAL_RETURN;
    }
#endif

        /*  check for PS  */
    if ((FLIP_SHORT(passwd) & 0x0000ffff) == PS_MAGIC)
    {
        *pFormat = PS;    /* return the PS format */
	NORMAL_RETURN;
    }

        /*  check for BMP */
    ERR_PASS( w_getIntel16(pIPWState, (UInt8 *)&passwd, &intel16));
    if ((intel16 & 0x0000ffff) == BMP_MAGIC)
    {
	*pFormat = FF_BMP;	/* return the BMP format */
	NORMAL_RETURN;
    }

        /*  check for TIFF */
    if ( passwd == TIFF_BIGENDIAN  ||  passwd == TIFF_LITTLEENDIAN )
    {
	*pFormat = TIFF_DEFAULT;	
	NORMAL_RETURN;
    }

        /*  check for PCX  */
    ERR_PASS( w_getIntel16(pIPWState, (UInt8 *)&passwd, &intel16));
    if ((intel16 & 0x0000ffff) == PCX_MAGICv0 ||
	(intel16 & 0x0000ffff) == PCX_MAGICv2 ||
	(intel16 & 0x0000ffff) == PCX_MAGICv3 ||
	(intel16 & 0x0000ffff) == PCX_MAGICv5)
    {
	*pFormat = PCX;    /* return the PCX format */
	NORMAL_RETURN;
    }

	/* Check for PNM */
    fread (pnmHeader, 1, 2, fp);
    rewind (fp);
    if (pnmHeader[0] == 'P' && isdigit((int)pnmHeader[1]))
    {
	*pFormat = PNM;
	NORMAL_RETURN;
    }

	/* if all else fails, return the NO_TYPE format */
    *pFormat = NO_TYPE;
    NORMAL_RETURN;
}




/*   GET_SHORT_INTS():  sets 16 bits in a short integer.
 *          Note: Needs to be written this way since the
 *                data may not be aligned on Int16 boundaries. */
#define GET_SHORT_INTS(attrPtr, result) \
do \
{ \
    UInt8 *attr = (*(attrPtr)); /* Record the curent address */ \
    (*(attrPtr)) += 2;  /* Update the buffer pointer */ \
    (result) = (Int16)((attr[0] << 8) + attr[1]); \
} while (0)



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  read PS image file header.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to rewound file stream
 *
 * Output:::
 *  FileH::          Ptr to file header FH struct.  Routine will fill in
 *                   fields with read header info.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_readPSHeader (
    IPWState           *pIPWState,
    FILE               *fp,
    FH                 *FileH)
{
    PROLOGUE;
    Int32          width, height;

    ERR_GEN(fp == NULL  ||  FileH == NULL, cErrNullPtr);

    fscanf(fp, " %*s");
    fscanf(fp, " %*s %*d %*d %*d %*d");
    fscanf(fp, " /width %ld %*s %% %*s %*s %*s %*s %*s", &width);
    fscanf(fp, " /height %ld", &height);
    FileH->width = width;
    FileH->height = height;
    FileH->Bpl = (width + 7) >> 2;
    FileH->totalBytes = FileH->Bpl * FileH->height;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a PNM file into a PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream.
 *
 * Output:::
 *  pr::             Ptr to ptr to returned PIXR with image that was read.
 *                   pr is allocated and filled in.
 *                   *pr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *  ejaquith::  Added RGB format capability
 *  ejaquith::  Added depth 2 and 4 PGM capability
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_readPnmFile (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR              **pr)
{
    PROLOGUE;
    Int32        w, h, maxGray, depth = 0, type, bytesPerFileLine = 0;
    Int32        i, lineDelta;
    UInt8       *imageP;
    UInt8       *pLineBuf = NULL;
    PIXR        *prd = NULL;

    ERR_GEN(pr == NULL  ||  fp == NULL, cErrNullPtr);
    ERR_GEN(*pr != NULL, cErrPtrNotNull);

    rewind (fp);
    ERR_GEN(fscanf(fp, "P%ld", &type) != 1, cErrIO_ReadError);

    if (type == 6) {
	/* Color RGB */
	UInt8 *pRedIm, *pGreenIm, *pBlueIm;
	UInt8 *pCol;
	PIXR  *subR, *subG, *subB;
	Int32  x;

	ERR_GEN(fscanf (fp, "%ld%ld%ld%*[\n]", &w, &h, &maxGray) != 3,
		cErrIO_ReadError);
	ERR_GEN(maxGray > 255, cErrDepthNotSupported);

	bytesPerFileLine = w * 3;

	pLineBuf = (UInt8 *)IPW_CALLOC(bytesPerFileLine, sizeof(UInt8),
				       cAllocatorData);
	ERR_GEN(pLineBuf == NULL, cErrNomem);

	ERR_PASS_CLEANUP( w_createPixrExtendedNoInit(pIPWState,
						     w, h, 8, 0, 0, 32, 3,
						     cColorRGB, NULL, &prd));
	
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prd, &lineDelta));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prd, cRedPlane, &subR));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prd, cGreenPlane, &subG));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prd, cBluePlane, &subB));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subR, &pRedIm));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subG, &pGreenIm));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subB, &pBlueIm));
	
	for (i = 0; i < h; i++) {
	    /* Line buffer is stored in consecutive bytes of R,G,B */
	    ERR_GEN_CLEANUP(fread (pLineBuf, 1, bytesPerFileLine, fp)
			    != bytesPerFileLine, cErrIO_ReadError);

	    pCol = pLineBuf;
	    for (x = 0; x < w; x++) {
		UCHAR_ACCESS(&pRedIm[x])   = *(pCol++);
		UCHAR_ACCESS(&pGreenIm[x]) = *(pCol++);
		UCHAR_ACCESS(&pBlueIm[x])  = *(pCol++);
	    }
	    
	    pRedIm   += lineDelta;
	    pGreenIm += lineDelta;
	    pBlueIm  += lineDelta;
	}
    }
    else if (type == 4  ||  type == 5)
    {
	if (type == 4)
	{
	    /* Binary */
	    ERR_GEN(fscanf(fp, "%ld%ld%*[\n]", &w, &h) != 2, cErrIO_ReadError);
	    depth = 1;
	    bytesPerFileLine = w / 8 + (w & 7 ? 1 : 0);
	}
	else if (type == 5)
	{
	    /* Gray */
	    ERR_GEN(fscanf(fp, "%ld%ld%ld%*[\n]", &w, &h, &maxGray) != 3,
		    cErrIO_ReadError);

	    /* We'll do the most flexible and benevolent thing...
	     * since the PGM spec assumes all files will be written in
	     * one pixel per byte, we'll "compress" a file depth-wise down
	     * if we can.  Or more specifically, we'll do so to make I/O of
	     * 4 and 2 bit depth images work for us WHILE obeying the PGM
	     * spec.
	     */
	    if (maxGray <= 3)
		depth = 2;
	    else if (maxGray <= 15)
		depth = 4;
	    else if (maxGray <= 255)
		depth = 8;
	    else
		/* This definitely is NOT legal -- PGM spec says can have
		 * at MOST 255 values! 
		 */
		ERR_GEN(1, cErrBadType);
	}

	ERR_PASS( w_createPixrNoInit(pIPWState, w, h, depth, 0, 0, cFrameBits,
				     &prd));

	ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prd, &lineDelta));
	ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, prd, &imageP));
	
	if (depth == 1  ||  depth == 8)
	{
	    Int32 j;
	    bytesPerFileLine = (depth == 8 ? w : bytesPerFileLine);

	    for (i = 0; i < h; i++)
	    {
		for (j = 0; j < bytesPerFileLine; j++)
		{
		    UCHAR_ACCESS(&imageP[j])  = (UInt8) fgetc(fp);
		}
		imageP += lineDelta;
	    }
	}
	else if (depth == 2  ||  depth == 4)
	{
	    /* These PGMs are stored in one pixel per byte format, a quite
	     * inefficient way.  Our PIXRs are stored packed in the most
	     * efficient way (4 pixels/byte for 2 bit depth, 2 pixels/byte
	     * for 4 bit depth)
	     * so we need an intermediate buffer to read into and then pack
	     * into the PIXR's image.
	     */
	    Int32  j, k;

	    bytesPerFileLine = w;
	    
	    /* Why alloc "+3" here?
	     * Well, because if the file is depth 2, and the image width
	     * mod 4 == 1, then we will have 3 extra (zeroed) line buffer
	     * accesses below when filling the LAST column of the image.
	     */
	    pLineBuf = (UInt8 *)IPW_CALLOC(bytesPerFileLine+3, sizeof(UInt8),
					   cAllocatorData);
	    ERR_GEN_CLEANUP(pLineBuf == NULL, cErrNomem);
	    
	    if (depth == 2)
	    {
		w = (w + 3) / 4;
		for (i = 0; i < h; i++)
		{
		    ERR_GEN_CLEANUP(fread (pLineBuf, 1, bytesPerFileLine, fp)
				    != bytesPerFileLine, cErrIO_ReadError);

		    for (j = k = 0; j < w; j++)
		    {
			UCHAR_ACCESS(&imageP[j])  = pLineBuf[k++] << 6;
			UCHAR_ACCESS(&imageP[j]) |= pLineBuf[k++] << 4;
			UCHAR_ACCESS(&imageP[j]) |= pLineBuf[k++] << 2;
			UCHAR_ACCESS(&imageP[j]) |= pLineBuf[k++];
		    }
		    
		    imageP += lineDelta;
		}
	    }
	    else /* depth == 4 */
	    {
		w = (w + 1) / 2;
		for (i = 0; i < h; i++)
		{
		    ERR_GEN_CLEANUP(fread (pLineBuf, 1, bytesPerFileLine, fp)
				    != bytesPerFileLine, cErrIO_ReadError);
		    
		    for (j = k = 0; j < w; j++)
		    {
			UCHAR_ACCESS(&imageP[j])  = pLineBuf[k++] << 4;
			UCHAR_ACCESS(&imageP[j]) |= pLineBuf[k++];
		    }
		    
		    imageP += lineDelta;
		}
	    }
	}
    }
    
    IPW_FREE(pLineBuf, cAllocatorData);

    *pr = prd;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prd);
    IPW_FREE(pLineBuf, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in an image file into a PIXR from a file stream.
 *
 * LongDescription:::
 *  Can read the following file formats.  
 *    -[X]-Y:  X is the allowable number planes/channels followed by the
 *             list of allowable depths with that number of planes.<BR>
 *    TIFF		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PS		-[1]-1                  <BR>
 *    SUNRAS		-[1]-1,2,4,8            <BR>
 *    BMP		-[1]-1,  4,8  -[3]-8    <BR>
 *    PNM		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PCX		-[1]-1,  4,8            <BR>
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to read image from.
 *                   May be stdin for any file format except TIFF!
 *  page::           0 for extraction of header information only;
 *                   1 for most image reads (there's only 1 "page");
 *  singleLine::     if true, a TIFF file is read one scan line at a time.
 *
 * Output:::
 *  FileH::          Ptr to file header FH struct.  Only the "filename" field
 *                   needs to be filled, and the "header" field must point to
 *                   NULL.  This routine will fill the rest of the fields in.
 *  prs::            Ptr to ptr to returned PIXR with image that was read.
 *                   prs is allocated and filled in.
 *                   *prs must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, cropping, multi-channel, color, compression, Packbits, CCITT
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fread (
    IPWState           *pIPWState,
    FILE               *fp,
    Int32               page,
    Bool		singleLine,
    FH                 *FileH,
    PIXR              **prs)
{
    PROLOGUE;
    PIXR     *pixr = NULL;
    Int32     depth, w, h, bpl;
    UInt8    *pImage;

    ERR_GEN(prs == NULL  ||  fp == NULL  ||  FileH == NULL, cErrNullPtr);
    ERR_GEN(*prs != NULL, cErrPtrNotNull);

    ERR_PASS( w_ffindFormat(pIPWState, fp, &(FileH->type)));

    switch (FileH->type)
    {
	case SUNRAS_COMP:
	case SUNRAS:
	    ERR_PASS( wp_readSunHeader(pIPWState, fp, FileH, &pixr));
	    if (!page)   /* don't read the image data */
		break;
	    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, pixr, &depth));
    
	    if (depth == 1)  /* binary file */
		ERR_PASS_CLEANUP( wp_readBinaryData(pIPWState, fp,
						    FileH, &pixr));
	    else if ((depth == 2) || (depth == 4) || (depth == 8))
		ERR_PASS_CLEANUP( wp_readGrayData(pIPWState, fp,
						  FileH, &pixr));
	    else
		ERR_GEN_CLEANUP(1, cErrDepthNotSupported);
	    break;
	case PS:
	    ERR_GEN(page, cErrNoImpl);
	    ERR_PASS( wp_readPSHeader(pIPWState, fp, FileH));
	    break;
	case FF_BMP:
	    ERR_PASS( wp_readBMP(pIPWState, fp, FileH, &pixr));
	    break;
	case TIFF_DEFAULT:
	    /* Can't read TIFF file from stdin!! */
	    ERR_GEN(fp == stdin, cErrInvalidParm);
	    ERR_PASS( w_readTiff(pIPWState, FileH, page, singleLine, &pixr));
	    break;
	case PNM:
	    ERR_PASS( wp_readPnmFile(pIPWState, fp, &pixr));
	    break;
	case PCX:
	    ERR_PASS( wp_readPcx(pIPWState, fp, &pixr));
	    break;
	case NO_TYPE:
	default:
	    ERR_GEN(1, cErrInvalidParm);
	    break;
    }

	/* just in case, clear the frames of the thing just read.  Some
	 * files apparently have trash in the line padding..
	 */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pixr, &pImage));
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, pixr, &w));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, pixr, &h));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pixr, &bpl));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, pixr, &depth));

    ERR_PASS_CLEANUP( i_clearAllFrames (pIPWState, pImage, w, h, bpl, depth));

    *prs = pixr;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in an image file into a PIXR from a file name.
 *
 * LongDescription:::
 *  Can read the following file formats.  
 *    -[X]-Y:  X is the allowable number planes/channels followed by the
 *             list of allowable depths with that number of planes.<BR>
 *    TIFF		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PS		-[1]-1                  <BR>
 *    SUNRAS		-[1]-1,2,4,8            <BR>
 *    BMP		-[1]-1,  4,8  -[3]-8    <BR>
 *    PNM		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PCX		-[1]-1,  4,8            <BR>
 *
 *   The entire file name, including directory, is assumed to be in fname.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fname::          Ptr to filename string.
 *                   Must contain full pathname.
 *		     If it is "-", and file is NOT in TIFF format,
 *                   stdin is used.  (It's an error if stdin is used and file
 *                   ends up being in TIFF format!)
 *  page::           0 for extraction of header information only;
 *                   1 for most image reads (there's only 1 "page");
 *                   -n for also printing header.
 *
 * Output:::
 *  prs::            Ptr to ptr to returned PIXR with image that was read.
 *                   prs is allocated and filled in.
 *                   *prs must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, cropping, compression, Packbits, CCITT
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_read (
    IPWState           *pIPWState,
    char               *fname,
    Int32               page,
    PIXR              **prs)
{
    PROLOGUE;

    ERR_PASS(w_readTest(pIPWState, fname, page, cFalse, prs));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in an image file into a PIXR from a file name using test args.
 *
 * LongDescription:::
 *  Can read the following file formats.  
 *    -[X]-Y:  X is the allowable number planes/channels followed by the
 *             list of allowable depths with that number of planes.<BR>
 *    TIFF		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PS		-[1]-1                  <BR>
 *    SUNRAS		-[1]-1,2,4,8            <BR>
 *    BMP		-[1]-1,  4,8  -[3]-8    <BR>
 *    PNM		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PCX		-[1]-1,  4,8            <BR>
 *
 *   The entire file name, including directory, is assumed to be in fname.
 *   This procedure differs from w_read only in that it has more weird
 *   arguments used during testing.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fname::          Ptr to filename string.
 *                   Must contain full pathname.
 *		     If it is "-", and file is NOT in TIFF format,
 *                   stdin is used.  (It's an error if stdin is used and file
 *                   ends up being in TIFF format!)
 *  page::           0 for extraction of header information only;
 *                   1 for most image reads (there's only 1 "page");
 *                   -n for also printing header.
 *  singleLine::     if true, a TIFF file is read one scan line at a time.
 *
 * Output:::
 *  prs::            Ptr to ptr to returned PIXR with image that was read.
 *                   prs is allocated and filled in.
 *                   *prs must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, cropping, compression, Packbits, CCITT
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_readTest (
    IPWState           *pIPWState,
    char               *fname,
    Int32               page,
    Bool		singleLine,
    PIXR              **prs)
{
    PROLOGUE;
    FILE         *fp = NULL;
    FH            FileH;  /* Put the file header on the stack */
    Bool          printHeader;
    PIXR         *pixr = NULL;

    ERR_GEN(prs == NULL  ||  fname == NULL, cErrNullPtr);
    ERR_GEN(*prs != NULL, cErrPtrNotNull);

    /* Clear out the file header struct */
    memset((void *) &FileH, 0, sizeof(FH));
    FileH.filename = NULL;
    FileH.header = NULL;
    
    if (page < 0)
    {
	printHeader = cTrue;
	page = -page;
    }
    else
	printHeader = cFalse;

	/* save the filename */
    ERR_PASS( w_strcatSafe(pIPWState, fname, "", &(FileH.filename)));

    if ( !strcmp("-", fname) )
    {
	fp = stdin;
    }
    else
    {
	fp = fopen( FileH.filename, "rb");
    }
	/* read the image */
    ERR_GEN_CLEANUP(!fp, cErrIO_ReadError);
    ERR_PASS_CLEANUP( w_fread(pIPWState, fp, page, singleLine, &FileH, &pixr));

    if (printHeader)
        ERR_PASS_CLEANUP( wp_fprintHeader(pIPWState, stdout, &FileH));

    if ( fp != stdin )
        fclose(fp);

/* Cleanup */
    IPW_FREE(FileH.filename, cAllocatorData);
    IPW_FREE(FileH.header, cAllocatorData);

    *prs = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(FileH.filename, cAllocatorData);
    IPW_FREE(FileH.header, cAllocatorData);
    w_destroyPixr(pIPWState, &pixr);

    ERR_RETURN;
}



/*

 $Log: readfile.c,v $
 Revision 1.2  2001/09/28 23:38:22  dabercro
 w_getLine moved from readfile.c to props2.c (so that Digipaper doesn't need
 the library that readfile.c resides in)

 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.47  1998/11/12 00:25:44  aber
 aber on Wed Nov 11 16:25:11 PST 1998

 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 Synced fixes from ddavies:
 Call w_readTiff with the page number.  This lets us read selected pages
 from multi-page TIFF files.

 Revision 1.46  1998/07/28 21:50:41  aber
 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 CommonCheck: Warning fixes for gcc v2.8x.

 Revision 1.45  1998/01/14 00:14:47  ddavies
 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.44  1997/06/30 23:13:55  ddavies
 ddavies on Mon Jun 30 16:13:32 PDT 1997

 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 Delete inclusion of Sandpiper's internal.h.
 Remove ability to read AIS files.

 Revision 1.43  1997/04/15 22:32:06  ddavies
 ddavies on Tue Apr 15 15:29:13 PDT 1997

 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 Use non-initialized PIXR creation where possible.
 Also generally change the way pixels are stored in wp_readPnmFile.
 In the old system, the entire PIXR was cleared to ensure clearing a
 byte-swapped frame, then the pixels were stored in natural byte order,
 then the entire image was byteswapped.  In the new world, we use
 an uninitialized bitmap, store in PIXR order using UCHAR_ACCESS and
 don't bother with the final byteswap.

 Revision 1.42  1996/05/24 04:45:08  ddavies
 ddavies on Thu May 23 21:44:43 PDT 1996

 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 Add inclusion of shrcc.pub and iotiff.pub

 Revision 1.41  1996/05/16 17:29:48  ddavies
 ddavies on Thu May 16 10:19:04 PDT 1996

 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 w_read now just calls w_readTest with the singleLine arg defaulted to cFalse.
 The new procedure, w_readTest, look a lot like the old w_read except for:
 1. We've eliminated the unused directory and rect parameters.  Instead of the
 lovely search algorithm that w_read used to use to find your file, we assume
 that you passed in the full filename.  If it doesn't work, we just stop instead
 of printing out a message about setenv IMAGES and stuff.
 2. w_readTest now passes the singleLine parameter to w_fread.
 Removed the unused rect parameter from calls to wp_readSunHeader, wp_readBinaryData
 and wp_readGrayData.  Also removed checks to see if rect is legal.
 All that stuff happened in w_fread.  Also modified w_fread to pass the
 singleLine arg to w_readTiff.

 Revision 1.40  1996/01/04 20:25:57  ejaquith
 readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.39  1995/12/13  04:12:00  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.38  1995/11/13  20:02:02  ejaquith
 * ejaquith on Mon Nov 13 12:01:32 PST 1995
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Added needed additional inclusions of "ptypes.h", "gdefs.h", and "pcx.h".
 * Updated PROLOGUE usage.
 *
 * Revision 1.37  1995/11/03  22:48:55  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.36  1995/10/13  22:04:16  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.35  1995/10/13  02:14:26  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Plugged some memory leaks for PNM file reading.
 * Depth 2 and depth 4 PNM reading was wrong in that it was reading either
 * 4 times or 2 times too much of the data buffer.
 * Fixed the depth 2 and depth 4 reading -- before it was assuming the
 * data was packed in the PGM file and it's not.
 * Fixed depth 1 and depth 8 reading bug on the PC that was happening because
 * the PIXR wasn't getting byteswapped before returning.
 * This was happening in both PNM and AIS formats.
 *
 * Revision 1.34  1995/10/10  02:20:18  ejaquith
 * ejaquith on Mon Oct 9 19:16:00 PDT 1995
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Added inclusion of "stdio.h"
 * Bug fixes.
 * PNM reading was not reading depth 2 and depth 4 images correctly.
 * These formats are saved (get this!) at one pixel/byte by the spec, so
 * I had to add a buffer to unpack our PIXRs into line by line and then
 * dump out.  Made routine pick depth of desired image from PGM file
 * based on maximum number of gray colors in file.
 * If it's 1..3, depth 2; 4..15, depth 4; 16..255, depth 8.
 * Added file depth and channel checking to only allow valid PNM files to be
 * read.
 * Added special comments in comment blocks to all public i/o routines
 * to allow user to know with certainty what depths/planes we support!
 * Removed "support" of LZ-compressed files, because there really wasn't
 * any
 *
 * Revision 1.33  1995/09/29  00:35:13  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.32  1995/09/21  22:58:33  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.31  1995/09/13  00:24:19  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * fixed up w_ffindFormat()
 *
 * Revision 1.30  1995/09/12  21:44:07  ejaquith
 * ejaquith on Tue Sep 12 14:43:19 PDT 1995
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Fixed ptr check boo-boo in w_read().  Was checking local var and not
 * (backup) and should not have been.
 *
 * Revision 1.29  1995/09/12  18:29:24  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * vanquished djfax stuff.
 *
 * Revision 1.28  1995/09/12  16:54:53  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.27  1995/09/08  00:14:50  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.26  1995/09/07  01:34:24  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.25  1995/09/06  20:22:30  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.24  1995/09/01  02:20:39  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.23  1995/08/31  21:38:20  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.22  1995/08/29  22:23:59  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.21  1995/08/29  21:56:01  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.20  1995/08/26  01:32:29  ejaquith
 * ejaquith on Fri Aug 25 18:32:12 PDT 1995
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * fixed bug in comparing type against BMP_MAGIC in ffindFormat().
 *
 * Revision 1.19  1995/08/22  02:21:43  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.18  1995/08/17  22:01:21  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.17  1995/07/26  22:45:26  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * The pointer to the block of memory holding the file header was being
 * incremented in readAISHeaderPr to point to each component of the header,
 * then the routine tried to free the block pointed to by the incremented
 * pointer.  Big bug.  Added origAttributes so that we remember the address
 * of the beginning of the block and free that.  This only affects AIS files.
 *
 * Revision 1.16  1995/03/30  21:36:13  ddavies
 * ddavies on Thu Mar 30 13:35:01 PST 1995
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Change readTIFFPr's arg from a file pointer to a FileHandle (FH).  This
 * allows readTIFFPr to use the file's name to open the file instead of
 * trying to derive the file descriptor from the file pointer - an operation
 * only available in POSIX, not ANSI, worlds.
 *
 * Revision 1.15  1994/12/19  23:24:28  ejaquith
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Added capability to read in PPM formatted file in the meta PNM format
 * for reading images to PIXRs.  Allows 24-bit RGB i/o.
 *
 * Revision 1.14  1994/11/22  21:56:01  ejaquith
 * ejaquith on Tue Nov 22 13:54:35 PST 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Fixed bug in PNM format file reading.  Routine was not eating up the
 * final '\n' char in the ASCII header.
 *
 * Revision 1.13  1994/10/21  01:18:38  ejaquith
 * ejaquith on Thu Oct 20 18:17:48 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Added PCX format reading capability.  Fairly untested, suggested to use
 * only for binary images.
 *
 * Revision 1.12  1994/10/10  22:34:45  lperry
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Change to use new header file io.pub instead of iopros.h
 *
 * Revision 1.11  1994/10/10  17:02:08  ejaquith
 * ejaquith on Mon Oct 10 09:56:25 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Changed "NONE" in FiletypeString to "NO_TYPE" to correspond with
 * NONE to NO_TYPE change in imagerep.h.  This is needed (and more correct!)
 * for the new TIFF code that will be introduced soon.
 * Added PCX reading framework, with a PCX format checking in ffindFormat.
 * However, the PCX read routine call is commented out until we get the bug
 * fixed PCX code.
 * Added function wordPaddedBytesPerRow for TIFF and PCX file reading to
 * figure out how many bytes to allocate for row reading.  It pads to a
 * 32 bit word, then returns the 8 bit multiple.
 *
 * Revision 1.10  1994/08/26  16:16:08  lperry
 * lperry on Fri Aug 26 09:12:07 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Fixed a memory leak in readPr. Actually, this had always been there,
 * since the memory allocated for the filename and header commponents
 * of the FileH struct were never freed. Now, with the use of a stack
 * varuable for the FileH struct, these became dangling pointers
 * when readPr exited, and were reported as leaks (vs memory-in-use)
 * by purify.
 *
 * Revision 1.9  1994/08/26  01:42:46  lperry
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Make the FileH struct a stack variable and pass it to
 * called functions to achieve reentrancy. Also needed to
 * change the way fprintHeader functioned, since it used the
 * old global FileH. This function is now private. As far as I
 * can tell, it isn't used anyway. Its functionality is replaced
 * by a (kludgy) mechanism, since its only for debugging anyway.
 * If readPr is called with the page parameter negated, it will
 * also print the header to stdout as a side effect.
 *
 * Revision 1.8  1994/08/24  00:10:55  macleod
 * macleod on Tue Aug 23 17:10:53 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Revision 1.7  1994/08/16  22:53:14  macleod
 * macleod on Tue Aug 16 15:52:44 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Was returning "1" rather than "NULL" as an error
 *
 * Revision 1.6  1994/08/16  21:53:42  macleod
 * macleod on Tue Aug 16 14:53:03 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Added support for PGM ("P5") and PBM ("P4") file types, a la PBMPLUS.
 * Note that we only support the "RAWBITS" formats, not the ascii formats.
 *
 * Revision 1.5  1994/07/11  23:19:11  ddavies
 * ddavies on Mon Jul 11 16:18:32 PDT 1994
 *
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Clear the frame of the image read by freadPr.  Some images apparently
 * have junk in their padding.
 *
 * Revision 1.4  1994/05/05  01:48:47  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Change FiletypeString to be an array of const char * from plain old
 * char *.  Add parens to assignments inside conditional clauses.
 * Change lots of %d's to %ld's to keep gcc happy.
 *
 * Revision 1.3  1993/12/03  01:18:06  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:15:38  ddavies
 * readfile.c -> /product/ipcore/ipwrap/src/RCS/readfile.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

