/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/*
** File:        reconseq.c
**
** Description: Grayscale reconstructions...
**
** Author: Luc Vincent
** Hacked into IPWRAP by P. MacLeod
*/



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imageref.pub"
/* prototypes */
#include "props.pub"
#include "reconseq.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: reconseq.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

#define REGISTER register

/*
**  Exported Routines:
*************************
*
*  SeqRecons()   Sequential grayscale reconstruction in any connectivity
*  SeqDRecons()  Dual sequential grayscale reconstruction
*
*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Sequential grayscale reconstruction 
 * 
 * LongDescription:::
 *  "Sequential grayscale reconstruction" is like "seed filling to a mask"
 *  except in grayscale: the seed or "marker" image is repeatedly 
 *  dilated and then clipped (to a min) with the mask.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSet::          Marker or "seed" image
 *  prMask::         Mask image
 *  connectivity::   4 or 8
 *  iterations::     Positive value or TO_COMPLETION
 *
 * Output:::
 *  pPrDest::        Resulting image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_SeqRecons (
    IPWState           *pIPWState,
    PIXR               *prSet,
    PIXR               *prMask,
    Int32               connectivity,
    Int32               iterations,
    PIXR              **pPrDest)
{
    PROLOGUE;
    UInt8            *pixelsD, *lineD, *pixelsM, *lineM;
    UInt8            *pixM, *pixDh, *pixDm;
    UInt8             val0M, val0D, val2D, val3D, val4D, val5D;
    UInt8             oldval;
    Int32             pplM, pplD;
    Int32             first_scan;
    Int32             image_modified;
    Int32             y, h;
    Int32             x, w;

    
        /*----- Check input parameters, allocate pPrDest if needed -----*/
    
    ERR_GEN (!(iterations == TO_COMPLETION || iterations > 0), cErrInvalidParm);
    ERR_GEN_PIXR (prSet, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN (prSet == prMask, cErrInvalidParm);

    if (!(*pPrDest))
	ERR_PASS( w_copy(pIPWState, prSet, pPrDest));


    /* Set the frame pixels to zero */
    ERR_PASS( w_setFramePixels(pIPWState, prMask, 0));
    ERR_PASS( w_setFramePixels(pIPWState, *pPrDest, 0));

        /*----- Extracts image parameters -----*/

    ERR_PASS( w_pixrGetWidth(pIPWState, *pPrDest, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, *pPrDest, &h));

    ERR_PASS( w_pixrGetImage(pIPWState, prMask, &pixelsM));
    ERR_PASS( w_pixrGetBpl(pIPWState, prMask, &pplM));

    ERR_PASS( w_pixrGetImage(pIPWState, *pPrDest, &pixelsD));
    ERR_PASS( w_pixrGetBpl(pIPWState, *pPrDest, &pplD));

        /*----- Start the succession of raster and anti-raster scans -----*/

    image_modified = 1;
    first_scan     = 1;
    iterations = (iterations == TO_COMPLETION ? 0 : iterations + 1);

        /*-----------------------
	  --- CONNECTIVITY: 8 ---
	  -----------------------*/

    if (connectivity == 8) {

	while (image_modified == 1 && --iterations != 0) {

	    image_modified = 0;

	        /* Raster scan. Neighborhood
		   considered are:    4 3 2  ->  h line
	                              5 0 *  ->  m line
				      * * *             */

	    lineM = pixelsM;
	    lineD = pixelsD - pplD - 1;
	    for (y = 0; y < h; y++, lineM += pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD += pplD;
		pixDm = lineD;
		val4D = UCHAR_ACCESS(pixDh); pixDh++;
		val3D = UCHAR_ACCESS(pixDh); pixDh++;
		val2D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm); pixDm++;
		val0D = UCHAR_ACCESS(pixDm);

		for (x = 0; x < w; x++) {

		    oldval = val0D;
		    if (val0D < val5D)
			val0D = val5D;
		    if (val0D < val4D)
			val0D = val4D;
		    if (val0D < val3D)
			val0D = val3D;
		    if (val0D < val2D)
			val0D = val2D;
		    if (val0D > val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm++) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM++;
		    val0M = UCHAR_ACCESS(pixM);

		    val4D = val3D;   val3D = val2D;
		    pixDh++;
		    val2D = UCHAR_ACCESS(pixDh);
		    val5D = val0D;  
              /* Big giant kludge to convince the
               * Microsoft 5.0 VC/C++ optimizing
               * compiler to generate the right
               * code.  It wasn't loading val0D if do ++pixDm here.
               */ 
		    val0D = UCHAR_ACCESS(pixDm);
		}

	    }

	    if (first_scan == 0 && image_modified == 0)
		break;

	    first_scan = 0;
	    image_modified = 0;

	        /* Anti-raster scan. Neighborhood
	           considered are:    * * *
	                              * 0 5  ->  m line
				      2 3 4  ->  h line   */

	    lineM = pixelsM + (h - 1) * pplM + (w - 1);
	    lineD = pixelsD + pplD * h + w;
	    for (y = h; y > 0; y--, lineM -= pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD -= pplD;
		pixDm = lineD;
		val4D = UCHAR_ACCESS(pixDh); pixDh--;   
		val3D = UCHAR_ACCESS(pixDh); pixDh--;   
		val2D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm); pixDm--;
		val0D = UCHAR_ACCESS(pixDm);

		for (x = w; x > 0; x--) {

		    oldval = val0D;
		    if (val0D < val5D)
			val0D = val5D;
		    if (val0D < val4D)
			val0D = val4D;
		    if (val0D < val3D)
			val0D = val3D;
		    if (val0D < val2D)
			val0D = val2D;
		    if (val0D > val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm--) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM--;
		    val0M = UCHAR_ACCESS(pixM);

		    val4D = val3D;   val3D = val2D;   
		    pixDh--;
		    val2D = UCHAR_ACCESS(pixDh);
		    val5D = val0D;   
	      /*   MS VC++ 5.0  wasn't loading val0D if do --pixDm here.
	       */
		    val0D = UCHAR_ACCESS(pixDm);

		}
	    }
	}
    }

        /*-----------------------
	  --- CONNECTIVITY: 4 ---
	  -----------------------*/

    else {

	while (image_modified == 1 && --iterations != 0) {

	    image_modified = 0;

	        /* Raster scan. Neighborhood
		   considered are:    * 3 *  ->  h line
	                              5 0 *  ->  m line
				      * * *             */

	    lineM = pixelsM;
	    lineD = pixelsD - pplD - 1;
	    for (y = 0; y < h; y++, lineM += pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD += pplD;
		pixDm = lineD;
		pixDh++;
		val3D = UCHAR_ACCESS(pixDh); 
		val5D = UCHAR_ACCESS(pixDm);  
		pixDm++; 
		val0D = UCHAR_ACCESS(pixDm);

		for (x = 0; x < w; x++) {

		    oldval = val0D;
		    if (val0D < val5D)
			val0D = val5D;
		    if (val0D < val3D)
			val0D = val3D;
		    if (val0D > val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm++) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM++;
		    val0M = UCHAR_ACCESS(pixM);

		    pixDh++;
		    val3D = UCHAR_ACCESS(pixDh);
		    val5D = val0D;   
	      /*   MS VC++ 5.0  wasn't loading val0D if do ++pixDm here.
	       */
		    val0D = UCHAR_ACCESS(pixDm);

		}

	    }

	    if (first_scan == 0 && image_modified == 0)
		break;

	    first_scan = 0;
	    image_modified = 0;

	        /* Anti-raster scan. Neighborhood
	           considered are:    * * *
	                              * 0 5  ->  m line
				      * 3 *  ->  h line   */

	    lineM = pixelsM + (h - 1) * pplM + (w - 1);
	    lineD = pixelsD + pplD * h + w;
	    for (y = h; y > 0; y--, lineM -= pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD -= pplD;
		pixDm = lineD;
		pixDh--;
		val3D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm);   
		pixDm--;
		val0D = UCHAR_ACCESS(pixDm);

		for (x = w; x > 0; x--) {

		    oldval = val0D;
		    if (val0D < val5D)
			val0D = val5D;
		    if (val0D < val3D)
			val0D = val3D;
		    if (val0D > val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm--) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM--;
		    val0M = UCHAR_ACCESS (pixM);

		    pixDh--;
		    val3D = UCHAR_ACCESS(pixDh);
		    val5D = val0D;   
	      /*   MS VC++ 5.0  wasn't loading val0D if do --pixDm here.
	       */
		    val0D = UCHAR_ACCESS(pixDm);

		}
	    }
	}
    }

    NORMAL_RETURN;
} /* SeqRecons */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Sequential dual grayscale reconstruction
 *
 * LongDescription:::
 *  "Sequential grayscale reconstruction" is like "seed filling to a mask"
 *  except in grayscale: the seed or "marker" image is repeatedly 
 *  dilated and then clipped (to a min) with the mask. This routine
 *  does the dual, which means it erodes and takes the max with the
 *  mask iteratively
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSet::          Marker image
 *  prMask::         Mask image, supposed to be >= prSet
 *  connectivity::   4 or 8
 *  iterations::     positive value or TO_COMPLETION
 *
 * Output:::
 *  pPrDest::        Result image
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_SeqDRecons (
    IPWState           *pIPWState,
    PIXR               *prSet,
    PIXR               *prMask,
    Int32               connectivity,
    Int32               iterations,
    PIXR              **pPrDest)
{
    PROLOGUE;
    UInt8            *pixelsD, *lineD, *pixelsM, *lineM;
    UInt8            *pixM, *pixDh, *pixDm;
    UInt8             val0M, val0D, val2D, val3D, val4D, val5D;
    UInt8             oldval;
    Int32             pplM, pplD;
    Int32             first_scan;
    Int32             image_modified;
    Int32             y, h;
    Int32             x, w;

        /*----- Check input parameters, allocate pPrDest if needed -----*/

    
    ERR_GEN (!(iterations == TO_COMPLETION || iterations > 0), cErrInvalidParm);
    ERR_GEN_PIXR (prSet, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN (prSet == prMask, cErrInvalidParm);

    if (!*pPrDest)
	ERR_PASS( w_copy(pIPWState, prSet, pPrDest));

    /* Set the frame pixels to all ones (bitwise, so you'll get FF) */
    ERR_PASS( w_setFramePixels(pIPWState, prMask, 1));
    ERR_PASS( w_setFramePixels(pIPWState, *pPrDest, 1));

        /*----- Extracts image parameters -----*/

    ERR_PASS( w_pixrGetWidth(pIPWState, *pPrDest, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, *pPrDest, &h));

    ERR_PASS( w_pixrGetImage(pIPWState, prMask, &pixelsM));
    ERR_PASS( w_pixrGetBpl(pIPWState, prMask, &pplM));

    ERR_PASS( w_pixrGetImage(pIPWState, *pPrDest, &pixelsD));
    ERR_PASS( w_pixrGetBpl(pIPWState, *pPrDest, &pplD));


        /*----- Start the succession of raster and anti-raster scans -----*/

    image_modified = 1;
    first_scan     = 1;
    iterations = (iterations == TO_COMPLETION ? 0 : iterations + 1);

        /*-----------------------
	  --- CONNECTIVITY: 8 ---
	  -----------------------*/

    if (connectivity == 8) {

	while (image_modified == 1 && --iterations != 0) {

	    image_modified = 0;

	        /* Raster scan. Neighborhood
		   considered are:    4 3 2  ->  h line
	                              5 0 *  ->  m line
				      * * *             */

	    lineM = pixelsM;
	    lineD = pixelsD - pplD - 1;
	    for (y = 0; y < h; y++, lineM += pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD += pplD;
		pixDm = lineD;
		val4D = UCHAR_ACCESS(pixDh); pixDh++;
		val3D = UCHAR_ACCESS(pixDh); pixDh++;
		val2D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm); pixDm++;
		val0D = UCHAR_ACCESS(pixDm);

		for (x = 0; x < w; x++) {

		    oldval = val0D;
		    if (val0D > val5D)
			val0D = val5D;
		    if (val0D > val4D)
			val0D = val4D;
		    if (val0D > val3D)
			val0D = val3D;
		    if (val0D > val2D)
			val0D = val2D;
		    if (val0D < val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm++) = val0D;
		    if (val0D != oldval)
			image_modified = 1;
		    
		    pixM++;
		    val0M = UCHAR_ACCESS(pixM);

		    val4D = val3D;   val3D = val2D;   
		    pixDh++;
		    val2D = UCHAR_ACCESS(pixDh);
		    val5D = val0D;   
	      /*   MS VC++ 5.0  wasn't loading val0D if do ++pixDm here.
	       */
		    val0D = UCHAR_ACCESS(pixDm);
		}

	    }

	    if (first_scan == 0 && image_modified == 0)
		break;

	    first_scan = 0;
	    image_modified = 0;

	        /* Anti-raster scan. Neighborhood
	           considered are:    * * *
	                              * 0 5  ->  m line
				      2 3 4  ->  h line   */

	    lineM = pixelsM + (h - 1) * pplM + (w - 1);
	    lineD = pixelsD + pplD * h + w;
	    for (y = h; y > 0; y--, lineM -= pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD -= pplD;
		pixDm = lineD;
		val4D = UCHAR_ACCESS(pixDh); pixDh--;
		val3D = UCHAR_ACCESS(pixDh); pixDh--;
		val2D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm); pixDm--;  
		val0D = UCHAR_ACCESS(pixDm);

		for (x = w; x > 0; x--) {

		    oldval = val0D;
		    if (val0D > val5D)
			val0D = val5D;
		    if (val0D > val4D)
			val0D = val4D;
		    if (val0D > val3D)
			val0D = val3D;
		    if (val0D > val2D)
			val0D = val2D;
		    if (val0D < val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm--) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM--;
		    val0M = UCHAR_ACCESS(pixM);

		    val4D = val3D;   val3D = val2D;
		    pixDh--; val2D = UCHAR_ACCESS(pixDh);
		    val5D = val0D; 
	      /*   MS VC++ 5.0  wasn't loading val0D if do --pixDm here.
	       */  
		    val0D = UCHAR_ACCESS(pixDm);

		}
	    }
	}
    }

        /*-----------------------
	  --- CONNECTIVITY: 4 ---
	  -----------------------*/

    else {

	while (image_modified == 1 && --iterations != 0) {

	    image_modified = 0;

	        /* Raster scan. Neighborhood
		   considered are:    * 3 *  ->  h line
	                              5 0 *  ->  m line
				      * * *             */

	    lineM = pixelsM;
	    lineD = pixelsD - pplD - 1;
	    for (y = 0; y < h; y++, lineM += pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD += pplD;
		pixDm = lineD;
		
		pixDh++; val3D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm);
		pixDm++; val0D = UCHAR_ACCESS(pixDm);

		for (x = 0; x < w; x++) {

		    oldval = val0D;
		    if (val0D > val5D)
			val0D = val5D;
		    if (val0D > val3D)
			val0D = val3D;
		    if (val0D < val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm++) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM++; val0M = UCHAR_ACCESS(pixM);

		    pixDh++; val3D = UCHAR_ACCESS(pixDh);
		    val5D = val0D; 
	      /*   MS VC++ 5.0  wasn't loading val0D if do --pixDm here.
	       */  
		    val0D = UCHAR_ACCESS(pixDm);

		}

	    }

	    if (first_scan == 0 && image_modified == 0)
		break;

	    first_scan = 0;
	    image_modified = 0;

	        /* Anti-raster scan. Neighborhood
	           considered are:    * * *
	                              * 0 5  ->  m line
				      * 3 *  ->  h line   */

	    lineM = pixelsM + (h - 1) * pplM + (w - 1);
	    lineD = pixelsD + pplD * h + w;
	    for (y = h; y > 0; y--, lineM -= pplM) {

		pixM  = lineM;
		val0M = UCHAR_ACCESS(pixM);

		pixDh = lineD;
		lineD -= pplD;
		pixDm = lineD;
		pixDh--; val3D = UCHAR_ACCESS(pixDh);
		val5D = UCHAR_ACCESS(pixDm);
		pixDm--; val0D = UCHAR_ACCESS(pixDm);

		for (x = w; x > 0; x--) {

		    oldval = val0D;
		    if (val0D > val5D)
			val0D = val5D;
		    if (val0D > val3D)
			val0D = val3D;
		    if (val0D < val0M)
			val0D = val0M;

		    UCHAR_ACCESS(pixDm--) = val0D;
		    if (val0D != oldval)
			image_modified = 1;

		    pixM--; val0M = UCHAR_ACCESS(pixM);

		    pixDh--; val3D = UCHAR_ACCESS(pixDh);
		    val5D = val0D;  
	      /*   MS VC++ 5.0  wasn't loading val0D if do --pixDm here.
	       */ 
		    val0D = UCHAR_ACCESS(pixDm);

		}
	    }
	}
    }

    /* Set the frame pixels to zero */
    ERR_PASS( w_setFramePixels(pIPWState, prMask, 0));
    ERR_PASS( w_setFramePixels(pIPWState, *pPrDest, 0));

    NORMAL_RETURN;

} /* SeqDRecons */

/*
 * $Log: reconseq.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.20  1997/12/08 23:28:14  xzhu
 * xzhu on Mon Dec 8 15:24:36 PST 1997
 *
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Dan found this weird problem in Microsoft VC/C++ 5.0 optimizing compiler. It does not generate right code. So we move around our code to let MS compiler do right thing.
 *
 * Revision 1.19  1997/01/13 18:10:46  macleod
 * macleod on Mon Jan 13 10:09:31 PST 1997
 *
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * These functions were doing nothing about endianness, causing things
 * to fail on the PC platform. Added UCHAR_ACCESS all over the place.
 * A better solution is to rewrite the routine to use longword accesses
 * and put it in ipshared where it belongs.
 *
 * Revision 1.18  1996/12/30 15:14:20  macleod
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Fixed a couple of errors in dual reconstruction where the PIXR checking
 * code never let a valid PIXR through
 *
 * Revision 1.17  1996/01/04 20:26:00  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.16  1995/12/13  04:12:04  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.15  1995/11/13  20:15:07  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.14  1995/10/13  22:04:19  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.13  1995/10/03  21:25:19  ejaquith
 * ejaquith on Tue Oct 3 14:25:12 PDT 1995
 *
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Fixed keywords typo
 *
 * Revision 1.12  1995/10/03  01:59:26  macleod
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.11  1995/09/21  22:58:36  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.10  1995/09/07  01:34:27  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:22:33  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  21:02:43  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.7  1995/08/31  21:38:23  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/25  01:38:01  ejaquith
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Fixed warnings
 *
 * Revision 1.5  1995/08/17  21:55:41  macleod
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.4  1994/10/07  19:06:47  lperry
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.3  1994/09/29  18:12:49  lperry
 * lperry on Thu Sep 29 11:12:21 PDT 1994
 *
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Convert function definitions from K&R style to ANSI style.
 * Necessary so that handmade prototype matches function def.
 *
 * Revision 1.2  1994/09/29  00:48:07  lperry
 * reconseq.c -> /product/ipcore/ipwrap/src/RCS/reconseq.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 * Also use reconseq.pub.
 *
 * Revision 1.1  1994/07/27  18:28:52  macleod
 * Initial revision
 *
 */
