/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   rectops.c:  simple operations on RECTANGL structs.  These are the
 *		 ones used by EZScan.
 *      Includes subroutines:
 *
 *	Creation/Destruction, etc.
 *		  w_createRect()
 *		  w_destroyRect()
 *		  w_destroyRectList()
 *		  w_copyRect()
 *		  w_cutRectFromList()
 *	Resizing
 *		  w_resizeRectP()
 *		  w_xformRect()
 *		  w_xformRectIP()
 *		  w_xformRectList()
 *		  w_clipRect()
 *		  w_clipRectToPixr()
 *
 *	Validity test
 *		  w_isRectValid()
 *  
 *	Operators between rectangles
 *		  w_unionRects()
 *
 */




#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"
#include "defines.pub"

/* prototypes */
#include "props.pub"
#include "rectops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: rectops.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

#define  RECT_ERROR_MESSAGE    0      /* error message for rect size = 0 */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a new rectangle.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  x::              x value for rect
 *  y::              y value for rect
 *  width::          width value for rect, >= 0.
 *  height::         height value for rect, >=0.
 *
 * Output:::
 *  ppRect::         Ptr to created rect.
 *                   ppRect is allocated and filled in.
 *                   *ppRect must point to NULL on input.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createRect (
    IPWState           *pIPWState,
    Int32               x,
    Int32               y,
    Int32               width,
    Int32               height,
    RECTANGL          **ppRect)
{
    PROLOGUE;
    RECTANGL  *rect = NULL;

    ERR_GEN(ppRect == NULL, cErrNullPtr);
    ERR_GEN(*ppRect != NULL, cErrPtrNotNull);
    ERR_GEN((width < 0) || (height < 0), cErrInvalidParm);

    rect = (RECTANGL *)IPW_MALLOC(sizeof(RECTANGL), cAllocatorData);
    
    ERR_GEN(rect == NULL, cErrNomem);
    
    rect->next = rect->prev = NULL;
    rect->x = x;
    rect->y = y;
    rect->w = width;
    rect->h = height;
    rect->mass = 0;

    *ppRect = rect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Free rectangle and set ptr to NULL.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppRect::         Ptr to ptr to rectangle to be destroyed.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyRect (
    IPWState           *pIPWState,
    RECTANGL          **ppRect)
{
    PROLOGUE;

    ERR_GEN(ppRect == NULL, cErrNullPtr);

    IPW_FREE(*ppRect, cAllocatorData);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Takes the rectangle out of the rectlist, and then destroys it.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  rect::           Ptr to _actual_ rectangle to remove (ie it is in the list)
 *
 * Output:::
 *  list::           Ptr to head of valid rectlist where rect gets removed
 *
 * Returns:::
 *
 * Notes:::
 *  doubly-linked::
 *   the rectlist must be correctly doubly-linked
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cutRectFromList (
    IPWState           *pIPWState,
    RECTANGL           *rect,
    RECTANGL          **list)
{
    PROLOGUE;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(list == NULL, cErrNullPtr);

    /* Take it out */
    if (rect == *list)
    {
	/* Chopping from the head */
	*list = rect->next;
	rect->next->prev = NULL;
    }
    else
    {
	ERR_GEN(rect->prev == NULL, cErrInvalidParm);

	rect->prev->next = rect->next;
	if (rect->next)
	{
	    rect->next->prev = rect->prev;
	}
    }

    IPW_FREE(rect, cAllocatorData);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroy list of rectangles and set list ptr to NULL.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  rectList::       Ptr to ptr to rectangle list to be destroyed.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyRectList (
    IPWState           *pIPWState,
    RECTANGL          **rectList)
{
    PROLOGUE;
    RECTANGL	*rect, *nextRect; 

    ERR_GEN(rectList == NULL, cErrNullPtr);

    if (*rectList != NULL)
    {
	for (rect = *rectList; rect != NULL; rect = nextRect)
	{
	    nextRect = rect->next;
	    IPW_FREE(rect, cAllocatorData);
	}

	*rectList = NULL;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a new rectangle by copying another.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  rect::           Ptr to rect to copy
 *
 * Output:::
 *  ppDRect::        Ptr to created rect.
 *                   ppDRect is allocated and filled in.
 *                   *ppDRect must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyRect (
    IPWState           *pIPWState,
    RECTANGL           *rect,
    RECTANGL          **ppDRect)
{
    PROLOGUE;
    RECTANGL   *pdrect = NULL;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(ppDRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDRect != NULL, cErrPtrNotNull);
    
    ERR_PASS( w_createRect(pIPWState, rect->x, rect->y, rect->w, rect->h,
			   &pdrect));

    *ppDRect = pdrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Resize a rectangle, leaving its relative center unchanged
 *
 * LongDescription:::
 *    Resizes the input rect by the numbers of pixels
 *    (or template units, or whatever they are) in the
 *    x and y directions, respectively, leaving the
 *    center of the rectangle unchanged.
 *    No constraint is put on the location of the 
 *    resulting rect.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  rect::           Ptr to rect to resize from
 *  xunits::         Amount to increase/decrease overall rectangle width by
 *  yunits::         Amount to increase/decrease overall rectangle height by
 *
 * Output:::
 *  ppDRect::        Ptr to resized rect, minimum width and height of 1.
 *                   ppDRect is allocated and filled in.
 *                   *ppDRect must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_resizeRectP (
    IPWState           *pIPWState,
    RECTANGL           *rect,
    Int32               xunits,
    Int32               yunits,
    RECTANGL          **ppDRect)
{
    PROLOGUE;
    Int32        x0, y0, newX, newY, newW, newH;
    RECTANGL    *pdrect = NULL;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(ppDRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDRect != NULL, cErrPtrNotNull);

    x0 = rect->x + rect->w / 2;          /* old center */
    y0 = rect->y + rect->h / 2;          /* old center */
    newW = MAX(1, rect->w + xunits);     /* must be > 1 */
    newH = MAX(1, rect->h + yunits);     /* must be > 1 */
    newX = x0 - newW / 2;
    newY = y0 - newH / 2;

    ERR_PASS( w_createRect(pIPWState, newX, newY, newW, newH, &pdrect));

    *ppDRect = pdrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Transform rectangles in a list by an X,Y translation and scaling factor
 *
 * LongDescription:::
 *  This first shifts the origin and then scales each
 *  rectangle in the input list.  This procedure is
 *  used when computing the corresponding regions
 *  on a shifted and/or scaled image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  shiftX::         Amount of X translation
 *  shiftY::         Amount of Y translation
 *  scaleX::         Scaling factor in X.  Must be >= 0.
 *  scaleY::         Scaling factor in Y.  Must be >= 0.
 *
 * Output:::
 *  rectList::       Ptr to rectangle list to transform.
 *
 * Returns:::
 *
 * Notes:::
 *  No clipping::
 *   There is no clipping of the transformed rects.
 *
 *  Transformed rects::
 *   The only constraint on the transformed rects is that
 *   the width and height will be greater than 0.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_xformRectList (
    IPWState           *pIPWState,
    Int32               shiftX,
    Int32               shiftY,
    Float32             scaleX,
    Float32             scaleY,
    RECTANGL           *rectList)
{
    PROLOGUE;
    RECTANGL    *pRect;

    ERR_GEN(rectList == NULL, cErrNullPtr);
    ERR_GEN(scaleX < 0.0, cErrInvalidParm);
    ERR_GEN(scaleY < 0.0, cErrInvalidParm);

    /* shift the origin and then scale */
    for (pRect = rectList; pRect != NULL; pRect = pRect->next)
	ERR_PASS( w_xformRectIP (pIPWState,
				 shiftX, shiftY, scaleX, scaleY,
				 pRect));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Transform a rectangle by an X,Y translation and scaling factor
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  rect::           Ptr to rectangle to transform
 *  shiftX::         Amount of X translation
 *  shiftY::         Amount of Y translation
 *  scaleX::         Scaling factor in X.  Must be >= 0.
 *  scaleY::         Scaling factor in Y.  Must be >= 0.
 *
 * Output:::
 *  ppDRect::        Ptr to transformed rect.
 *                   ppDRect is allocated and filled in.
 *                   *ppDRect must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  No clipping::
 *   There is no clipping of the transformed rect.
 *
 *  Transformed rect::
 *   The only constraint on the transformed rect is that
 *   the width and height will be greater than 0.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_xformRect (
    IPWState           *pIPWState,
    RECTANGL           *rect,
    Int32               shiftX,
    Int32               shiftY,
    Float32             scaleX,
    Float32             scaleY,
    RECTANGL          **ppDRect)
{
    PROLOGUE;
    RECTANGL  *pdrect = NULL;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(ppDRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDRect != NULL, cErrPtrNotNull);
    ERR_GEN(scaleX < 0.0, cErrInvalidParm);
    ERR_GEN(scaleY < 0.0, cErrInvalidParm);

    ERR_PASS( w_copyRect(pIPWState, rect, &pdrect));

	/* shift the UL corner first */
    pdrect->x += shiftX;
    pdrect->y += shiftY;

	/* rescale the entire rect */
    pdrect->x = ROUND ((Float32)pdrect->x * scaleX );
    pdrect->y = ROUND ((Float32)pdrect->y * scaleY );
    pdrect->w = MAX(1, ROUND ((Float32)rect->w * scaleX));
    pdrect->h = MAX(1, ROUND ((Float32)rect->h * scaleY));
    
    *ppDRect = pdrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Transform a rectangle in-place by an X,Y translation and scaling factor
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  shiftX::         Amount of X translation
 *  shiftY::         Amount of Y translation
 *  scaleX::         Scaling factor in X.  Must be >= 0.
 *  scaleY::         Scaling factor in Y.  Must be >= 0.
 *
 * Output:::
 *  rect::           Ptr to rect to transform in place.
 *
 * Returns:::
 *
 * Notes:::
 *  No clipping::
 *   There is no clipping of the transformed rect.
 *
 *  Transformed rect::
 *   The only constraint on the transformed rect is that
 *   the width and height will be greater than 0.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_xformRectIP (
    IPWState           *pIPWState,
    Int32               shiftX,
    Int32               shiftY,
    Float32             scaleX,
    Float32             scaleY,
    RECTANGL           *rect)
{
    PROLOGUE;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(scaleX < 0.0, cErrInvalidParm);
    ERR_GEN(scaleY < 0.0, cErrInvalidParm);

    /* shift the UL corner first */
    rect->x += shiftX;
    rect->y += shiftY;

    /* rescale the entire rect */
    rect->x = ROUND ((Float32)rect->x * scaleX );
    rect->y = ROUND ((Float32)rect->y * scaleY );
    rect->w = MAX(1, ROUND ((Float32)rect->w * scaleX));
    rect->h = MAX(1, ROUND ((Float32)rect->h * scaleY));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a rectangle by clipping another rectangle.
 *
 * LongDescription:::
 *  That is, take the input rectangle, and clip it to fall inside region
 *  described by (x,y) in ([0..clipw-1],[0..cliph-1])
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  rect::           Ptr to rectangle to clip
 *  clipw::          clipping width.  Must be > 0.
 *  cliph::          clipping height.  Must be > 0.
 *
 * Output:::
 *  ppDRect::        Ptr to clipped rect.
 *                   ppDRect is allocated and filled in.
 *                   *ppDRect must point to NULL on input.
 *                   If the rect is entirely outside clipping region,
 *                   rect will have 0 width and height.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clipRect (
    IPWState           *pIPWState,
    RECTANGL           *rect,
    Int32               clipw,
    Int32               cliph,
    RECTANGL          **ppDRect)
{
    PROLOGUE;
    Int32        rightedge, bottomedge, xextent, yextent;
    RECTANGL    *pdrect = NULL;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(ppDRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDRect != NULL, cErrPtrNotNull);
    ERR_GEN(clipw <= 0 || cliph <= 0, cErrInvalidParm);

    ERR_PASS( w_copyRect(pIPWState, rect, &pdrect));

	/* clip to the region {0, clipw - 1} */
    pdrect->x = MAX(0, MIN(rect->x, (clipw - 1)));
    pdrect->y = MAX(0, MIN(rect->y, (cliph - 1)));

	/*----------------------------------------------------------------*
	 * test if rect is entirely outside the clipping region
	 * if so, set the width and height to 0 and return
	 *----------------------------------------------------------------*/
    rightedge = rect->x + rect->w - 1;
    bottomedge = rect->y + rect->h - 1;
    if (rightedge < 0 || bottomedge < 0 ||
	rect->x >= clipw || rect->y >= cliph)
    {
	pdrect->w = 0;
	pdrect->h = 0;
	*ppDRect = pdrect;
	NORMAL_RETURN;
    }

	/*----------------------------------------------------------------*
	 * otherwise, part of the rect is within the clipping region, so we
	 * will get a finite width and height
	 *----------------------------------------------------------------*/
	/* right edge of rectangle: clip to {0, clipw - 1} */
    xextent = MAX(0, MIN(rightedge, (clipw - 1)));

	/* bottom edge of rectangle: clip to {0, cliph - 1} */
    yextent = MAX(0, MIN(bottomedge, (cliph - 1)));

    pdrect->w = xextent - pdrect->x + 1;
    pdrect->h = yextent - pdrect->y + 1;

    *ppDRect = pdrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given a rectangle and bounding rect dimensions, crop it down if needed
 *
 * LongDescription:::
 *  Given a rectangle and some cropping dimensions (of say, another rectangle),
 *  place the rectangle inside the cropping rectangle and crop it to not
 *  exceed those dimensions.<BR><BR>
 *
 *  Similar to w_clipRect(), except that the upper left (x,y) location
 *  of the clipping/cropping rect is not 0,0.
 *
 * Input:::
 *  pIPWState::         Ptr to IP struct for memory alloc/freeing routines
 *  x::                 X cropping position
 *  y::                 Y cropping position
 *  w::                 cropping width.  Must be > 0.
 *  h::                 cropping height.  Must be > 0.
 *
 * Output:::
 *  rect::              Ptr to rect on input.  On ouput, it will be cropped
 *                      down to the above dimensions.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cropRect (
    IPWState   *pIPWState,
    Int32       x,
    Int32       y,
    Int32       w,
    Int32       h,
    RECTANGL   *rect)
{
    PROLOGUE;

    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(w <= 0  ||  h <= 0, cErrInvalidParm);

    if (rect->x < x)
    {
        rect->w -= (x - rect->x);
        rect->x = x;
    }
    if (rect->y < y)
    {
        rect->h -= (y - rect->y);
        rect->y = y;
    }
    if (rect->x + rect->w > x + w)
        rect->w = x + w - rect->x;
    if (rect->y + rect->h > y + h)
        rect->h = y + h - rect->y;
    if (rect->w < 0)
	rect->w = 0;
    if (rect->h < 0)
	rect->h = 0;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a rectangle by clipping another rectangle to a PIXR.
 *
 * LongDescription:::
 *    Clips the rect to the input PIXR.
 *    Thus, returns a new rect equal to that part of the 
 *    input rect that is contained within the proper PIXR image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR to use as clipping rectangle.
 *  rect::           Ptr to rectangle to clip.
 *
 * Output:::
 *  ppDRect::        Ptr to clipped rect.
 *                   ppDRect is allocated and filled in.
 *                   *ppDRect must point to NULL on input.
 *                   If the input rect is entirely outside the PIXR
 *                   image, rect will have 0 width and height.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, cropping
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clipRectToPixr (
    IPWState           *pIPWState,
    PIXR               *pixr,
    RECTANGL           *rect,
    RECTANGL          **ppDRect)
{
    PROLOGUE;
    Int32        w,h;
    RECTANGL    *pdrect = NULL;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(rect == NULL, cErrNullPtr);
    ERR_GEN(ppDRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDRect != NULL, cErrPtrNotNull);

    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));

    ERR_PASS( w_clipRect(pIPWState, rect, w, h, &pdrect));

    *ppDRect = pdrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a rectangle from the union of two other rectangles.
 *
 * LongDescription:::
 *	Takes the union of two rectangles, which is the smallest rectangle
 *	that contains both of them. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  rect1::          Ptr to a rectangle.
 *  rect2::          Ptr to yet another rectangle.
 *
 * Output:::
 *  dest::           Ptr to dest rectangle.  Must already exist.  May be
 *                   one of the two input rectangles, for a pseudo in-place
 *                   operation.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_unionRects (
    IPWState           *pIPWState,
    RECTANGL           *rect1,
    RECTANGL           *rect2,
    RECTANGL           *dest)
{
    PROLOGUE;
    Int32 x0, y0, x1, y1;

    ERR_GEN(rect1 == NULL  ||  rect2 == NULL  ||  dest == NULL, cErrNullPtr);

    x0 = MIN (rect1->x, rect2->x);
    y0 = MIN (rect1->y, rect2->y);
    x1 = MAX (rect1->x + rect1->w - 1, rect2->x + rect2->w - 1);
    y1 = MAX (rect1->y + rect1->h - 1, rect2->y + rect2->h - 1);
    dest->x = x0;
    dest->y = y0;
    dest->w = x1 - x0 + 1;
    dest->h = y1 - y0 + 1;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Determine if rect exists and is contained within proper image of PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR to use as clipping area.
 *  rect::           Ptr to rectangle to check.
 *
 * Output:::
 *  pValid::         Ptr to 1 if rect is valid and within PIXR; 0 o.w.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_isRectValid (
    IPWState           *pIPWState,
    PIXR               *pixr,
    RECTANGL           *rect,
    Int8               *pValid)
{
    PROLOGUE;
    Int32        w, h;

    ERR_GEN(rect == NULL  ||  pValid == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));

    *pValid = (rect->x < 0 || rect->y < 0 || (rect->x + rect->w > w) ||
	       (rect->y + rect->h > h)) ? 0 : 1;

    NORMAL_RETURN;
}

/*

 $Log: rectops.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 Imported sources

 Revision 1.24  1996/02/25 01:14:11  ejaquith
 rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v

 Added routine w_cropRect() to crop a rectangle in place given some
 x,y,w,h, params of another rectangle.

 Revision 1.23  1996/01/27 22:27:08  ejaquith
 ejaquith on Sat Jan 27 14:22:30 PST 1996

 rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v

 Modified routines:  w_xformRect()  w_xformRectIP()  w_xformRectList()
 to take both X and Y scaling parameters (instead of just one common param).
 Updated routines to use ROUND() macro (instead of just casting to Int32)
 for increased accuracy.

 * Revision 1.22  1996/01/04  20:26:03  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.21  1995/12/13  04:12:08  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:15:10  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:04:22  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/09/29  00:36:34  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.17  1995/09/21  22:58:40  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/20  21:36:14  macleod
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * if w_clipRectToPixr got a rectangle fully outside the clip rectangle,
 * the destination (created) rectangle was never set.
 *
 * Revision 1.15  1995/09/12  16:55:47  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.14  1995/09/08  00:15:44  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.13  1995/09/07  01:34:30  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:22:36  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  02:21:19  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:38:25  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/22  02:21:49  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.8  1995/08/17  22:02:15  ejaquith
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1994/09/30  16:32:39  lperry
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Conver to use new .pub header files
 *
 * Revision 1.6  1994/09/30  01:51:46  macleod
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * added unionRect, which computes the smallest rectangle which contains
 * the two rectangles passed.
 * Also added cutRectFromList, which takes a rectangle out of a
 * rectangle list and frees it.
 *
 * Revision 1.5  1994/01/07  04:24:14  ddavies
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Have createRect initialize the mass entry in the RECTANGL struct created
 * to 0.
 *
 * Revision 1.4  1993/12/17  22:51:23  ddavies
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Add destroyRectList, xformRectList and xformRectInPlace.  The first proce
 * simply frees the elements of a non-circular linked list of RECTANGLs.
 * The second one calls the third one to move the origins and scale the coordinates
 * of a list of rectangles in place.
 *
 * Revision 1.3  1993/12/03  01:18:10  ddavies
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:15:46  ddavies
 * rectops.c -> /product/ipcore/ipwrap/src/RCS/rectops.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

