/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *   rectops1.c:  simple operations on RECTANGL structs.
 *
 *	Rect list operations:
 *		w_addRectToList()
 *		w_copyRectList()
 *		w_createRectArrayList()
 *		w_removeSmallRects()
 *		w_flatMergeRects()
 *
 *	Rect list sorting routines:
 *		w_rectSort()
 *		w_rectLessorUpperLeftPosition()
 *
 *	Rect Pixels ==> Himetrics conversions:
 *		w_rectPixelsToHimetrics()
 *		w_rectPixelsToHimetricsIP()
 *
 *	Rect Himetric ==> Pixels conversions:
 *		w_rectHimetricsToPixels()
 *		w_rectHimetricsToPixelsIP()
 *
 *	[ 2540 Himetrics == 1 inch ]
 */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"
#include "himetric.h"
#include "defines.pub"

/* prototypes */
#include "props.pub"
#include "rectops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: rectops1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a rectangle from Himetrics to pixels units, in-place
 *
 * Input:::
 *   pIPWState:: Ptr to IP struct for memory alloc/freeing routines
 *   pRect::     Ptr to source rectangle
 *   xres::      Positive image resolution along X axis used to convert to
 *   yres::      Positive image resolution along Y axis used to convert to
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *  Upper left corner assumption::
 *    We assumme that the HI_METRICS have 0,0 in the upper left corner and
 *    that -y moves down the image and +x moves to the right.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
w_rectHimetricsToPixelsIP (
    IPWState *pIPWState,
    Int32     xres,
    Int32     yres,
    RECTANGL *pRect)
{
    PROLOGUE;

    ERR_GEN(pRect == NULL, cErrNullPtr);
    ERR_GEN(xres <= 0  ||  yres <= 0, cErrInvalidParm);

    HIMET_TO_PIX_X_IP(pRect->x, xres);
    HIMET_TO_PIX_Y_IP(pRect->y, yres);
    HIMET_TO_PIX_W_IP(pRect->w, xres);
    HIMET_TO_PIX_H_IP(pRect->h, yres);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a rectangle from Himetrics to pixels units
 *
 * Input:::
 *   pIPWState::  Ptr to IP struct for memory alloc/freeing routines
 *   pRect::      Ptr to source rectangle
 *   xres::       Positive image resolution along X axis used to convert to
 *   yres::       Positive image resolution along Y axis used to convert to
 *
 * Output:::
 *   ppDestRect:: Ptr to ptr to destination rect (which gets created)
 *                Ptr to dest rect must be NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Upper left corner assumption::
 *    We assumme that the HI_METRICS have 0,0 in the upper left corner and
 *    that -y moves down the image and +x moves to the right.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
w_rectHimetricsToPixels (
    IPWState  *pIPWState,
    RECTANGL  *pRect, 
    Int32      xres,
    Int32      yres,
    RECTANGL **ppDestRect)
{
    PROLOGUE;
    RECTANGL *newrect=NULL;

    ERR_GEN(ppDestRect == NULL  ||  pRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDestRect != NULL, cErrPtrNotNull);
    ERR_GEN(xres <= 0  ||  yres <= 0, cErrInvalidParm);

    ERR_PASS(w_createRect (pIPWState,
			   HIMET_TO_PIX_X(pRect->x, xres),
			   HIMET_TO_PIX_Y(pRect->y, yres),
			   HIMET_TO_PIX_W(pRect->w, xres),
			   HIMET_TO_PIX_H(pRect->h, yres),
			   &newrect));
    
    *ppDestRect = newrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a rectangle from pixels to Himetrics units, in-place
 *
 * Input:::
 *   pIPWState:: Ptr to IP struct for memory alloc/freeing routines
 *   pRect::     Ptr to source rectangle
 *   xres::      Positive image resolution along X axis used to convert to
 *   yres::      Positive image resolution along Y axis used to convert to
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *  Upper left corner assumption::
 *    We assumme that the HI_METRICS have 0,0 in the upper left corner and
 *    that -y moves down the image and +x moves to the right.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rectPixelsToHimetricsIP (
    IPWState *pIPWState,
    Int32     xres,
    Int32     yres,
    RECTANGL *pRect)
{
    PROLOGUE;

    ERR_GEN(pRect == NULL, cErrNullPtr);
    ERR_GEN(xres <= 0  ||  yres <= 0, cErrInvalidParm);

    PIX_TO_HIMET_X_IP(pRect->x, xres);
    PIX_TO_HIMET_Y_IP(pRect->y, yres);
    PIX_TO_HIMET_W_IP(pRect->w, xres);
    PIX_TO_HIMET_H_IP(pRect->h, yres);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a rectangle from pixels to Himetrics units
 *
 * Input:::
 *   pIPWState::  Ptr to IP struct for memory alloc/freeing routines
 *   pRect::      Ptr to source rectangle
 *   xres::       Positive image resolution along X axis used to convert to
 *   yres::       Positive image resolution along Y axis used to convert to
 *
 * Output:::
 *   ppDestRect:: Ptr to ptr to destination rect (which gets created)
 *                Ptr to dest rect must be NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Upper left corner assumption::
 *    We assumme that the HI_METRICS have 0,0 in the upper left corner and
 *    that -y moves down the image and +x moves to the right.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
w_rectPixelsToHimetrics (
    IPWState  *pIPWState,
    RECTANGL  *pRect,
    Int32      xres,
    Int32      yres,
    RECTANGL **ppDestRect)
{
    PROLOGUE;
    RECTANGL *newrect=NULL;

    ERR_GEN(ppDestRect == NULL  ||  pRect == NULL, cErrNullPtr);
    ERR_GEN(*ppDestRect != NULL, cErrPtrNotNull);
    ERR_GEN(xres <= 0  ||  yres <= 0, cErrInvalidParm);

    ERR_PASS(w_createRect (pIPWState,
			   PIX_TO_HIMET_X(pRect->x, xres),
			   PIX_TO_HIMET_Y(pRect->y, yres),
			   PIX_TO_HIMET_W(pRect->w, xres),
			   PIX_TO_HIMET_H(pRect->h, yres),
			   &newrect));

    *ppDestRect = newrect;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create an array of rectangles that is also a rect list.
 *
 * LongDescription:::
 *  Sometimes we want to access a linked list of rectangles ALSO as if it
 *  were an array.  This routine allocates an ARRAY of rectangles (note ONE
 *  allocation!) and sets up all the "next" and "prev" fields to point at
 *  the corresponding "intuitive" place.<BR><BR>
 *
 *  <PRE>
 *
 *  That is, referring to the ptr to the returned list as "list":
 *    list[0].prev  points to NULL
 *    list[0].next  points to list[1]
 *    list[1].prev  points to list[0]
 *    ...
 *    list[nRects-1].next points to NULL.
 *
 *  </PRE>
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  nRects::            Nonnegative number of rects for list/array.
 *
 * Output:::
 *  ppRectArrayList::   Ptr to ptr to list of created rectangles on return.
 *                      Ptr must be valid and *ppRectArrayList must be NULL
 *                      on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Destruction warning!!::
 *   Since this creates both an array/list, the ptr to the first rect in the
 *   list is the ONLY thing to be freed to free the entire list.
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createRectArrayList (
    IPWState      *pIPWState,
    Int32          nRects,
    RECTANGL     **ppRectArrayList)
{
    PROLOGUE;
    Int32       i;
    RECTANGL   *pList = NULL;

    ERR_GEN(nRects <= 0, cErrInvalidParm);
    ERR_GEN(ppRectArrayList == NULL, cErrNullPtr);
    ERR_GEN(*ppRectArrayList != NULL, cErrPtrNotNull);
    
    pList = (RECTANGL *)IPW_CALLOC(nRects,
				   sizeof(RECTANGL), cAllocatorData);
    ERR_GEN(pList == NULL, cErrNomem);

    if (nRects == 1)
    {
	/* Take care of 1 rect list cases */
	pList[0].prev = pList[0].next = NULL;
    }
    else 
    {
	pList[0].prev = NULL;
	pList[0].next = &(pList[1]);
	for (i = 1; i < nRects-1; i++)
	{
	    pList[i].prev = &(pList[i-1]);
	    pList[i].next = &(pList[i+1]);
	}
	pList[nRects-1].prev = &(pList[nRects-2]);
	pList[nRects-1].next = NULL;
    }

    *ppRectArrayList = pList;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy a list of rectangles
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rlist::             Ptr to list of rectangles
 *
 * Output:::
 *  clist::             Ptr to ptr to copied list of rectangles on output.
 *                      Ptr must be valid and *clist must be NULL
 *                      on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyRectList (
    IPWState   *pIPWState,
    RECTANGL   *rlist,
    RECTANGL  **clist)
{
    PROLOGUE;
    RECTANGL     *last, *rect, *copyrect = NULL;
    RECTANGL     *copyrlist = NULL;

    ERR_GEN(rlist == NULL  ||  clist == NULL, cErrNullPtr);
    ERR_GEN(*clist != NULL, cErrPtrNotNull);

    /* copy the first rect */
    ERR_PASS( w_copyRect( pIPWState, rlist, &copyrect));
    copyrlist = copyrect;
    last = copyrlist;

    /* now that copyrlist is initialized,
     * copy the rest of the rects.
     */
    for (rect = rlist->next; rect != NULL; rect = rect->next)
    {
	copyrect = NULL;
	ERR_PASS_CLEANUP( w_copyRect( pIPWState, rect, &copyrect));
	last->next = copyrect;
	copyrect->prev = last;
	last = copyrect;
    }
  
    *clist = copyrlist;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRectList( pIPWState, &copyrlist);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a rectangle into a list of rectangles.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rect::              Ptr to rectangle to add.
 *                      If adding to list front, the rect->prev field is
 *                      not modified.
 *                      If adding to list end, the rect->next field is
 *                      not modified.
 *  addToFront::        cTrue to add rect to beginning of list.
 *                      cFalse to add rect to end.
 *                      (And if you don't care, adding to front is more
 *                       efficient).
 *
 * Output:::
 *  ppList::            Ptr to ptr to list of rectangles (or NULL) on input.
 *                      On output, this list will be modified to contain
 *                      the passed rect.
 *
 * Returns:::
 *
 * Notes:::
 *   Merging lists::
 *     Since only one of the "rect" links are modified, this routine can be
 *     used to merge two rect lists...
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addRectToList (
    IPWState   *pIPWState,
    RECTANGL   *rect,
    Bool        addToFront,
    RECTANGL  **ppList)
{
    PROLOGUE;
    RECTANGL   *last;

    ERR_GEN(rect == NULL  ||  ppList == NULL, cErrNullPtr);

    if (*ppList == NULL)
    {
	*ppList = rect;
    }
    else
    {
	if (addToFront)
	{
	    rect->next = *ppList;
	    rect->next->prev = rect;
	    *ppList = rect;
	} 
	else
	{
	    /* add at end */
	    for (last = *ppList; last->next != NULL;  last = last->next);

	    last->next = rect;
	    rect->prev = last;
	}
    }
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compares the upper left positions of two RECTANGLs, finding "lessor" one.
 *
 * LongDescription:::
 *  Based on the approx. distance from upper left corner:
 *   rect1 "<" rect2 if
 *     rect1 X position + ABS(rect1 Y position) is less than that of rect2.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rect1::             Ptr to first rect
 *  rect2::             Ptr to second rect
 *
 * Output:::
 *  p1LessThan2::       Ptr to "cTrue" on return if rect1 "<" rect2.
 *                      Ptr to "cFalse" o.w.
 *
 * Returns:::
 *
 * Notes:::
 *  Suitable for use with w_rectSort()
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rectLessorUpperLeftPosition (
    IPWState  *pIPWState,
    RECTANGL  *rect1,
    RECTANGL  *rect2,
    Bool      *p1LessThan2)
{
    PROLOGUE;
    
    ERR_GEN (rect1 == NULL  ||  rect2 == NULL  ||  p1LessThan2 == NULL,
	     cErrNullPtr);
    
    *p1LessThan2 = ((rect1->x + ABS(rect1->y)) < (rect2->x + ABS(rect2->y)) ?
		    cTrue : cFalse);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Sorts a list of rectangles by a passed in ordering "less than" function.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rectlist::          Ptr to non-NULL list of rectangles
 *  compFn::            Ptr to routine to use for determining when/how a
 *                      rectangle is "less than" another.
 *
 * Output:::
 *  ppSorted::          Ptr to ptr to list of sorted rectangles on return.
 *                      Ptr must be valid and *ppSorted must be NULL
 *                      on input.
 *                      The sorted order will be "lessor" to "greator",
 *                      based on the passed in "rectangle less than" routine,
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rectSort(
    IPWState       *pIPWState,
    RECTANGL       *rectlist,
    ErrCode (CDECL *compFn)(IPWState *pIPWState,
			    RECTANGL *rect1,
			    RECTANGL *rect2,
			    Bool     *p1LessThan2),
    RECTANGL      **ppSorted)
{
    PROLOGUE;
    Int32        gap, i, j, n, nRects;
    RECTANGL     *rj, *rjgap, *rjbefore, *rjafter, *rjgapbefore;
    RECTANGL     *rjgapafter;
    RECTANGL     *sortRectList = NULL, *pRect;
    Bool          lessThan;


    ERR_GEN(rectlist == NULL  ||  ppSorted == NULL  ||  compFn == NULL,
	    cErrNullPtr);
    ERR_GEN(*ppSorted != NULL, cErrPtrNotNull);
  
    /* create copy of the input boxes */
    ERR_PASS(w_copyRectList(pIPWState, rectlist, &sortRectList));


    for (n = 0, pRect = sortRectList; pRect != NULL; pRect = pRect->next, n++);
    nRects = n;

    /* sort:  rect list and index array */
    for (gap = nRects/2; gap > 0; gap >>= 1)
	for (i = gap; i < nRects; i++)
	    for (j = i - gap; j >= 0; j -= gap)
	    {
		/* Get the  "j"th  and  "j+gap"th  elements */
		for (n = 0, rj = sortRectList; 
		     rj != NULL  &&  n < j; 
		     rj = rj->next, n++);

		for (rjgap = rj; 
		     rjgap != NULL  &&  n < j + gap; 
		     rjgap = rjgap->next, n++);

		ERR_PASS_CLEANUP(compFn(pIPWState, rj, rjgap, &lessThan));
		
		if (lessThan)
		    break;

		/* swap rj and rjgap */
		/* first save info */
		rjbefore = rj->prev;
		rjafter = rj->next;
		rjgapbefore = rjgap->prev;
		rjgapafter = rjgap->next;

		/* second put rjgap where rj was */
		if (rjbefore != NULL)
		{
		    if (rjbefore != rjgap)
			rjbefore->next = rjgap;
		    else
			rj->next = rjgap;
		}
		else
		    sortRectList = rjgap; /* rjgap is now in front */
		if (rjafter != NULL)
		{
		    if (rjafter != rjgap)
			rjafter->prev = rjgap;
		    else
			rj->prev = rjgap;
		}
		if (rjbefore != rjgap)
		    rjgap->prev = rjbefore;
		else
		    rjgap->prev = rj;
		if (rjafter != rjgap)
		    rjgap->next = rjafter;
		else
		    rjgap->next = rj;

		/* third put rj where rjgap was */
		if (rjgapbefore != NULL)
		{
		    if (rjgapbefore != rj)
			rjgapbefore->next = rj;
		    else
			rjgap->next = rj;
		}
		else
		    sortRectList = rj; /* rj is now in front */
		if ((rjgapafter != NULL) && (rjgapafter != rj))
		    rjgapafter->prev = rj;
		if (rjgapbefore != rj)
		    rj->prev = rjgapbefore;
		if (rjgapafter != rj)
		    rj->next = rjgapafter;

	    }

    *ppSorted = sortRectList;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRectList(pIPWState, &sortRectList);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compares the horizontal positions of two RECTANGLs, finding "lessor" one.
 *
 * LongDescription:::
 *   rect1 "<" rect2 if:
 *     -rect1 X position is less than that of rect2,
 *     -same X positions, and rect1 width is greater than that of rect2.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rect1::             Ptr to first rect
 *  rect2::             Ptr to second rect
 *
 * Output:::
 *  p1LessThan2::       Ptr to "cTrue" on return if rect1 "<" rect2.
 *                      Ptr to "cFalse" o.w.
 *
 * Returns:::
 *
 * Notes:::
 *  Suitable for use with w_rectSort()
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, comparison
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_rectLessorHorizPosition (
    IPWState  *pIPWState,
    RECTANGL  *rect1,
    RECTANGL  *rect2,
    Bool      *p1LessThan2)
{
    PROLOGUE;
    
    ERR_GEN (rect1 == NULL  ||  rect2 == NULL  ||  p1LessThan2 == NULL,
	     cErrNullPtr);

    *p1LessThan2 = ((rect1->x < rect2->x) ||
		    ((rect1->x == rect2->x) && (rect1->w >= rect2->w)) ?
		    cTrue : cFalse);
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy rect list -- merge input list rects whenever they touch or overlap.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rects::             Ptr to non-NULL list of rectangles
 *
 * Output:::
 *  ppList::            Ptr to ptr to list of rectangles on return.
 *                      Ptr must be valid, and *ppList must be NULL
 *                      on input.
 *
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, filtering
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  hirsch::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_flatMergeRects (
    IPWState   *pIPWState,
    RECTANGL   *rects,
    RECTANGL  **ppList)
{
    PROLOGUE;
    Int32         n, i, j;
    Int32        *taken = NULL;
    RECTANGL     *initrect = NULL, *pRect;
    RECTANGL     *rect = NULL, *rectj = NULL, *beginrect = NULL;
    RECTANGL     *rectlist = NULL;     /* result */
    RECTANGL     *sortrects = NULL;


    ERR_GEN(rects == NULL  ||  ppList == NULL,  cErrNullPtr);
    ERR_GEN(*ppList != NULL, cErrPtrNotNull);
  
  
    for (n = 0, pRect = rects; pRect != NULL; pRect = pRect->next, n++);
    taken = (Int32 *)IPW_CALLOC(n, sizeof(Int32), cAllocatorData);
    ERR_GEN(taken == NULL,  cErrNomem);


    /* init taken array */
    for (i = 0; i < n; i++)
	taken[i] = cFalse;

    /* sort rects by horizontal position */
    ERR_PASS_CLEANUP( w_rectSort( pIPWState, rects, 
				  wp_rectLessorHorizPosition,
				  &sortrects));
    

    i = 0;
    for (initrect = sortrects; initrect != NULL; initrect = initrect->next)
    {
	/*  skip if already included */
	if (taken[i]) {
	    i++;
	    continue;
	}

	ERR_PASS_CLEANUP(w_copyRect(pIPWState, initrect, &rect));

	taken[i] = 1;

	
	/*  expand this rect with any intersecting or touching rects;
	 *  use initrect to decide when to break out of the
	 *  while loop for this rect.  */

	j = i + 1;

	beginrect = initrect->next;
	while (cTrue)
	{
	    for (rectj = beginrect; rectj != NULL; rectj = rectj->next)
	    {
		if (rectj->x > rect->x + rect->w)  /* search no further */
		    break;
	
		if (!((rectj->x > rect ->x + rect ->w) || 
		      (rect ->y > rectj->y + rectj->h) ||
		      (rectj->y > rect ->y + rect ->h) ||
		      (rect ->x > rectj->x + rectj->w)))
		{  
		    /* This means they don't intersect or touch! */

		    /* expand rect */
		    taken[j] = 1;
		    rectj->w = rect->w = 
			MAX(rect->x + rect->w, rectj->x + rectj->w) -
			MIN(rect->x, rectj->x);
		    rectj->h = rect->h =
			MAX(rect->y + rect->h, rectj->y + rectj->h) -
			MIN(rect->y, rectj->y);
		    rectj->x = rect->x = MIN(rect->x, rectj->x);
		    rectj->y = rect->y = MIN(rect->y, rectj->y);
		}
		j++;
	    }

	    if ((rect->x == initrect->x) && (rect->y == initrect->y) &&
		(rect->w == initrect->w) && (rect->h == initrect->h))
		break;
	    else
	    {
		initrect->x = rect->x;
		initrect->y = rect->y;
		initrect->w = rect->w;
		initrect->h = rect->h;
		j = 0;


		/* Get the "j"th element of sortrects */
		for (n = 0, beginrect = sortrects; 
		     beginrect != NULL  &&  n < j; 
		     beginrect = beginrect->next, n++);
	    }
	} /* while */
	i++;
	ERR_PASS_CLEANUP (w_destroyRect (pIPWState, &rect));
    }
    


    /* make up the return list */
    for (initrect = sortrects; initrect != NULL; initrect = initrect->next)
    {
	/* Check if "initrect" is in list "rectlist" already... */
	for (pRect = rectlist; pRect != NULL; pRect = pRect->next)
	    if ((pRect->x == initrect->x) && (pRect->y == initrect->y) &&
		(pRect->w == initrect->w) && (pRect->h == initrect->h))
		break;
	
	if (pRect == NULL)
	{
	    /* Means it isn't already in list, add it */
	    rect = NULL;
	    ERR_PASS_CLEANUP (w_copyRect(pIPWState, initrect, &rect));

	    ERR_PASS_CLEANUP (w_addRectToList (pIPWState, rect, cFalse, 
					       &rectlist));
	    rect = NULL;
	}
    }
    
     
  
    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &sortrects));
    IPW_FREE(taken, cAllocatorData);
  
    *ppList = rectlist;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRect (pIPWState, &rect);
    w_destroyRectList (pIPWState, &sortrects);
    w_destroyRectList (pIPWState, &rectlist);
    IPW_FREE(taken, cAllocatorData);
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Copy a rectangle list, removing small rectangles
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rectList::          Ptr to list of rectangles
 *  minWidth::          Minimum acceptable width in pixels for rectangles
 *  minHeight::         Minimum acceptable height in pixels for rectangles
 *   
 * Output:::
 *  pListNoSmalls::     Ptr to ptr to list of rectangles on return.
 *                      Ptr must be valid, and *pListNoSmalls must be NULL
 *                      on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry, filtering
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_removeSmallRects (
    IPWState  *pIPWState,
    RECTANGL  *rectList, 
    Int32      minWidth,
    Int32      minHeight,
    RECTANGL **pListNoSmalls)
{
    PROLOGUE;
    RECTANGL *newList = NULL, *pRect, *cRect = NULL;

    ERR_GEN(pListNoSmalls == NULL, cErrNullPtr);
    ERR_GEN(*pListNoSmalls != NULL, cErrPtrNotNull);

    for (pRect = rectList; pRect != NULL; pRect = pRect->next)
    {
	if (pRect->w >= minWidth  &&  pRect->h >= minHeight)
	{
	    ERR_PASS_CLEANUP( w_copyRect( pIPWState, pRect, &cRect));
	    
	    ERR_PASS_CLEANUP( w_addRectToList( pIPWState,
					       cRect, cFalse, &newList));
	    cRect = NULL;
	}
    }


    *pListNoSmalls = newList;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRect( pIPWState, &cRect);
    w_destroyRectList( pIPWState, &newList);
    ERR_RETURN;
}

/*
 $Log: rectops1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.4  1998/07/28 21:50:39  aber
 rectops1.c -> /product/ipcore/ipwrap/src/RCS/rectops1.c,v

 CommonCheck: Warning fixes for gcc v2.8x.

 Revision 1.3  1997/07/11 01:59:15  ddavies
 rectops1.c -> /product/ipcore/ipwrap/src/RCS/rectops1.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.2  1996/01/27 22:56:03  ejaquith
 ejaquith on Sat Jan 27 14:55:46 PST 1996

 rectops1.c -> /product/ipcore/ipwrap/src/RCS/rectops1.c,v

 Fixed 2 comment-block typos.

 * Revision 1.1  1996/01/27  22:31:40  ejaquith
 * Initial revision
 *
 */
