/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  reduce1.c: contains top-level calls for threshold reduction
 *
 *          General cascade of threshold reductions:
 *                  w_reduce ()
 *
 *          2x threshold reduction (and cascades):
 *                  w_reduce16 ()
 *                  w_reduce8 ()
 *                  w_reduce4 ()
 *                  w_reduce2 ()
 *
 *
 *          Approx. timings on SS2 (measured in source pixels/second)
 *
 *              With C code (no special cases):
 *                      2x, all thresholds:   50 Mpixel/s
 *
 *              With SPARC assembler:
 *                      2x, all thesholds:   70 Mpixel/s
 *                  Special cases:
 *                      4x, thresh (1,1), (4,4):  100 Mpixel/s
 *                      16x, thresh (1,1,1,1), (4,4,4,4):  160 Mpixel/s
 *
 */



#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"     /* for cTrue and cFalse */
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "reduce.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: reduce1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Power of two thresholded reduction of binary PIXRs.
 *
 * LongDescription:::
 *  Reduces binary PIXRs by 2, 4, 8, or 16 using cascaded 2x2 pixel
 *  neighborhood thresholds to determine the state of the output
 *  pixel at each reduction step.  A pixel is 1 if it's source neighborhood
 *  is greater than the threshold (1, 2, 3, 4).  The first 0 threshold 
 *  determines the reduction (if all are non-zero, the reduction is 16):
 *
 *	level2 = 0  =>	2x reduction
 *	level3 = 0  =>	4x reduction
 *	level4 = 0  =>	8x reduction
 *
 *  The result PIXR is created if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *  level3::         Third neighborhood threshold
 *  level4::         Fourth neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    Int32               level3,
    Int32               level4,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS;
    Bool	 destWasNull;
    ScaledInt1616	     xRes, yRes;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
    {
	destWasNull = cFalse;
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    }
    else
	destWasNull = cTrue;

    /* thesholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);
    ERR_GEN ((level3 < 0 || level3 > 4), cErrInvalidParm);
    ERR_GEN ((level4 < 0 || level4 > 4), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    if (level2 == 0)  /* 2x reduction */
    {
        ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prs, level1, prd));
    }
    else 
	if (level3 == 0)  /* 4x reduction */
	{
	    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prs, level1, level2, prd));
	}
	else 
	    if (level4 == 0)  /* 8x reduction */
	    {
		ERR_PASS_CLEANUP (w_reduce8 (pIPWState, prs, level1, 
		                             level2, level3, prd));
	    }
	    else  /* 16x reduction */
	    {
		ERR_PASS_CLEANUP (w_reduce16 (pIPWState, prs, level1, 
		                              level2, level3, level4, prd));
	    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destWasNull && (*prd != NULL))
        w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduce () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 16
 *
 * LongDescription:::
 *  Reduces a binary image by sixteen using the four 2x2 neighborhood
 *  thresholds to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *  level3::         Third neighborhood threshold
 *  level4::         Fourth neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce16 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    Int32               level3,
    Int32               level4,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, 
                 bplS, bplD;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    UInt8       *iS, *iD;
    ScaledInt1616	     xRes, yRes;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
       ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* thresholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);
    ERR_GEN ((level3 < 0 || level3 > 4), cErrInvalidParm);
    ERR_GEN ((level4 < 0 || level4 > 4), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS / 16, hS / 16, 1, 
					      xRes / 16, yRes / 16, cFrameBits, 
					      prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prs, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prs, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    ERR_PASS_CLEANUP (i_reduce16 (pIPWState, iD, wD, hD, bplD,
		                  iS, wS, hS, bplS,
		                  level1, level2, level3, level4));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduce16 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 8
 *
 * LongDescription:::
 *  Reduces a binary image by eight using the three 2x2 neighborhood
 *  thresholds to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *  level3::         Third neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce8 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    Int32               level3,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, 
                 bplS, bplD;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    UInt8       *iS, *iD;
    ScaledInt1616	     xRes, yRes;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
       ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* thresholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);
    ERR_GEN ((level3 < 0 || level3 > 4), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS / 8, hS / 8, 1, 
					      xRes / 8, yRes / 8, cFrameBits, 
					      prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prs, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prs, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    ERR_PASS_CLEANUP (i_reduce8 (pIPWState, iD, wD, hD, bplD,
		                 iS, wS, hS, bplS,
		                 level1, level2, level3));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduce8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 4
 *
 * LongDescription:::
 *  Reduces a binary image by four using the two 2x2 neighborhood
 *  thresholds to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce4 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, 
                 bplS, bplD;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    UInt8       *iS, *iD;
    ScaledInt1616	     xRes, yRes;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
       ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* thresholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS / 4, hS / 4, 1, 
					      xRes / 4, yRes / 4, cFrameBits, 
					      prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prs, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prs, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    ERR_PASS_CLEANUP (i_reduce4 (pIPWState, iD, wD, hD, bplD,
		                 iS, wS, hS, bplS,
		                 level1, level2));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduce4 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 2
 *
 * LongDescription:::
 *  Reduces a binary image by two using the  2x2 neighborhood
 *  threshold to determine whether a destination pixel is 1 or 0.
 *  If 2x2 neighborhood has at least level 1 pixels, the result will be 1,
 *  Otherwise it will be 0.  Threshold value must be 1, 2, 3, or 4.  The 
 *  result PIXR is created if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level::          Neighborhood (2x2) threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce2 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, 
                 bplS, bplD;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    UInt8       *iS, *iD;
    ScaledInt1616	     xRes, yRes;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
       ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    /* threshold OK? */
    ERR_GEN ((level < 1 || level > 4), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS / 2, hS / 2, 1, 
					      xRes / 2, yRes / 2, cFrameBits, 
					      prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prs, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prs, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    ERR_PASS_CLEANUP (i_reduce2 (pIPWState, iD, wD, hD, bplD,
		                 iS, bplS, level));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduce2 () */


/*
 * $Log: reduce1.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.27  1998/01/14 00:14:51  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.26  1997/04/15 22:32:21  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.25  1996/01/04 20:26:07  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.24  1995/12/13  04:12:12  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/11/13  20:15:13  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.22  1995/10/13  22:04:26  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.21  1995/09/26  21:58:21  dferrell
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.20  1995/09/21  22:58:43  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.19  1995/09/19  21:10:16  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () calls
 *
 * Revision 1.18  1995/09/15  22:34:56  ddavies
 * ddavies on Fri Sep 15 15:34:14 PDT 1995
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Change type of destCreated to Bool.  Also, fix small bug in w_reduce.
 * In case of error, it was never releasing *prd because destCreated was
 * never being set.
 *
 * Revision 1.17  1995/09/15  00:33:08  dferrell
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.16  1995/09/12  22:37:56  macleod
 * macleod on Tue Sep 12 15:36:55 PDT 1995
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * reduce2, reduce4, reduce8 and reduce16 fetched the src. width
 * and height only if the dest. image was NULL, therefore if it wasn't,
 * the call would fail since the width/heights weren't the same
 *
 * Revision 1.15  1995/09/11  21:19:48  dferrell
 * dferrell on Mon Sep 11 14:19:28 PDT 1995
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Removed incorrect testing for dest pixr pointer == NULL (oops).
 *
 * Revision 1.14  1995/09/11  17:51:44  dferrell
 * dferrell on Mon Sep 11 10:50:51 PDT 1995
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.13  1995/09/07  01:34:33  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:22:39  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  18:18:20  dferrell
 * dferrell on Fri Sep 1 11:18:14 PDT 1995
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.10  1995/08/31  21:38:28  ejaquith
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/29  22:24:07  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.8  1995/08/29  21:55:15  jcraig
 * jcraig on Tue Aug 29 14:54:49 PDT 1995
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * correctly set the resolution for each reduction (not always source
 * res / 2).
 *
 * Revision 1.7  1995/08/17  21:53:08  dferrell
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.6  1995/02/03  01:38:08  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.5  1994/10/03  20:34:31  lperry
 * lperry on Mon Oct 3 13:34:02 PDT 1994
 *
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Changed to use reduce.pub instead of scalepros.h
 *
 * Revision 1.4  1994/06/08  04:13:12  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.3  1993/12/03  01:18:15  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:15:55  ddavies
 * reduce1.c -> /product/ipcore/ipwrap/src/RCS/reduce1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *
 */

