/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  reduce3.c: contains top-level calls for I/O versions of threshold reduction
 *
 *          2x threshold reduction (and cascades):
 *                  w_reduceio16 ()
 *                  w_reduceio8 ()
 *                  w_reduceio4 ()
 *                  w_reduceio2 ()
 *
 *
 * These reduction operations are like the ordinary ones except they make
 * no assumptions about whether the PIXRs have frames.  The IPSHARED
 * primitives used to do the real work similarly make no assumptions about
 * frames.  These primitives are intended for use with frameless images
 * that are encountered when doing I/O.
 *
 */



#include "ipwcore.prv"

#include "pixr.h"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "reduce.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: reduce3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 16 on frameless PIXRs
 *
 * LongDescription:::
 *  Reduces a binary image by sixteen using the four 2x2 neighborhood
 *  thresholds to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *  level3::         Third neighborhood threshold
 *  level4::         Fourth neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceio16 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    Int32               level3,
    Int32               level4,
    PIXR               *prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, bplS, bplD;
    UInt8       *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrameAny);
    ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannels1, cFrameAny);

    /* thresholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);
    ERR_GEN ((level3 < 0 || level3 > 4), cErrInvalidParm);
    ERR_GEN ((level4 < 0 || level4 > 4), cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS( w_pixrGetWidth (pIPWState, prd, &wD));
    ERR_PASS( w_pixrGetHeight (pIPWState, prd, &hD));

    /* dest pixr correct size? */
    ERR_GEN ((wD < (wS / 16)), cErrDestDimensIncorrect);
    ERR_GEN ((hD < (hS / 16)), cErrDestDimensIncorrect);

    ERR_PASS( w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prd, &bplD));
    ERR_PASS( w_pixrGetImage (pIPWState, prs, &iS));
    ERR_PASS( w_pixrGetImage (pIPWState, prd, &iD));

    ERR_PASS( i_ioReduce16 (pIPWState, iD, wS / 16, hS / 16, bplD,
			    iS, wS, hS, bplS,
			    level1, level2, level3, level4));

    NORMAL_RETURN;

}  /* w_reduceio16 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 8 for frameless PIXRs
 *
 * LongDescription:::
 *  Reduces a binary image by eight using the three 2x2 neighborhood
 *  thresholds to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *  level3::         Third neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceio8 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    Int32               level3,
    PIXR               *prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, bplS, bplD;
    UInt8       *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrameAny);
    ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannels1, cFrameAny);

    /* thresholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);
    ERR_GEN ((level3 < 0 || level3 > 4), cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS( w_pixrGetWidth (pIPWState, prd, &wD));
    ERR_PASS( w_pixrGetHeight (pIPWState, prd, &hD));

    /* dest pixr correct size? */
    ERR_GEN ((wD < (wS / 8)), cErrDestDimensIncorrect);
    ERR_GEN ((hD < (hS / 8)), cErrDestDimensIncorrect);

    ERR_PASS( w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prd, &bplD));
    ERR_PASS( w_pixrGetImage (pIPWState, prs, &iS));
    ERR_PASS( w_pixrGetImage (pIPWState, prd, &iD));

    ERR_PASS( i_ioReduce8 (pIPWState, iD, wS / 8, hS / 8, bplD,
			   iS, wS, hS, bplS,
			   level1, level2, level3));

    NORMAL_RETURN;

}  /* w_reduceio8 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 4 for frameless PIXRs
 *
 * LongDescription:::
 *  Reduces a binary image by four using the two 2x2 neighborhood
 *  thresholds to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level1::         First neighborhood threshold
 *  level2::         Second neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceio4 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level1,
    Int32               level2,
    PIXR               *prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, bplS, bplD;
    UInt8       *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrameAny);
    ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannels1, cFrameAny);

    /* thresholds OK? */
    ERR_GEN ((level1 < 1 || level1 > 4), cErrInvalidParm);
    ERR_GEN ((level2 < 0 || level2 > 4), cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS( w_pixrGetWidth (pIPWState, prd, &wD));
    ERR_PASS( w_pixrGetHeight (pIPWState, prd, &hD));

    /* dest pixr correct size? */
    ERR_GEN ((wD < (wS / 4)), cErrDestDimensIncorrect);
    ERR_GEN ((hD < (hS / 4)), cErrDestDimensIncorrect);

    ERR_PASS( w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prd, &bplD));
    ERR_PASS( w_pixrGetImage (pIPWState, prs, &iS));
    ERR_PASS( w_pixrGetImage (pIPWState, prd, &iD));

    ERR_PASS( i_ioReduce4 (pIPWState, iD, wS / 4, hS / 4, bplD,
			   iS, wS, hS, bplS,
			   level1, level2));

    NORMAL_RETURN;

}  /* w_reduceio4 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary thresholded reduction by 2 for frameless PIXRs
 *
 * LongDescription:::
 *  Reduces a binary image by two using the  2x2 neighborhood
 *  threshold to determine whether a destination pixel is 1 or 0.
 *  Threshold values must be 1, 2, 3, or 4.  The result PIXR is created 
 *  if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be reduced
 *  level::          Neighborhood threshold
 *
 * Output:::
 *  prd::            Result PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceio2 (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               level,
    PIXR               *prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, bplS, bplD;
    UInt8       *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrameAny);
    ERR_GEN_PIXR (prd, cByteOrderBig, cDepth1, cChannels1, cFrameAny);

    /* threshold OK? */
    ERR_GEN ((level < 1 || level > 4), cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS( w_pixrGetWidth (pIPWState, prd, &wD));
    ERR_PASS( w_pixrGetHeight (pIPWState, prd, &hD));

    /* dest pixr correct size? */
    ERR_GEN ((wD < (wS / 2)), cErrDestDimensIncorrect);
    ERR_GEN ((hD < (hS / 2)), cErrDestDimensIncorrect);

    ERR_PASS( w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS( w_pixrGetBpl (pIPWState, prd, &bplD));
    ERR_PASS( w_pixrGetImage (pIPWState, prs, &iS));
    ERR_PASS( w_pixrGetImage (pIPWState, prd, &iD));

    ERR_PASS( i_ioReduce2 (pIPWState, iD, (wS / 2), (hS / 2), bplD,
			   iS, wS, bplS, level));

    NORMAL_RETURN;

}  /* w_reduceio2 () */


/*
 *
 * $Log: reduce3.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.21  1998/01/14 00:14:53  ddavies
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.20  1996/01/04 20:26:09  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.19  1995/12/13  04:12:16  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/11/13  20:15:17  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.17  1995/10/13  22:04:28  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/09/26  21:58:27  dferrell
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.15  1995/09/21  22:58:46  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.14  1995/09/11  17:51:59  dferrell
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.13  1995/09/07  01:34:36  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:22:41  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  18:18:26  dferrell
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.10  1995/08/31  21:38:31  ejaquith
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/29  22:24:12  ddavies
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.8  1995/08/29  21:56:11  ddavies
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.7  1995/08/23  19:19:27  macleod
 * macleod on Wed Aug 23 12:15:01 PDT 1995
 *
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * reduceio2, reduceio4, reduceio8, and reduceio16 all had a few problems:
 * first, they checked for destW (or H) != srcW / reduction, rather than
 * destW < srcW / reduction, as it used to be, which could cause problems
 * in cases where there were odd numbers of pixels, etc. Second, they called
 * pixrGetImageAll instead of pixrGetImage. Third, they passed wD and hD
 * rather than wS/reduction and hS/reduction, as they used to, which
 * caused roundoff problems, and also caused the test to fail due to
 * it reducing the size of the source image.
 *
 * Revision 1.6  1995/08/17  21:53:15  dferrell
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.5  1994/10/03  20:34:42  lperry
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Changed to use reduce.pub instead of scalepros.
 *
 * Revision 1.4  1994/06/08  04:13:18  ddavies
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.3  1993/12/03  01:18:20  ddavies
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:16:04  ddavies
 * reduce3.c -> /product/ipcore/ipwrap/src/RCS/reduce3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

