/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  reduce4.c: contains top-level calls for scale-to-gray reduction
 *
 *          scale to gray reduction:
 *                  w_scaleToGray4bpp ()
 *
 *
 * These reduction operations convert binary images to gray images by
 * counting the number of active pixels in each area to be reduced,
 * producing a gray value proprotional to the number of active binary
 * pixels.
 *
 *	    arbitrary scale-to-gray:
 *		    w_scaleToGrayArb ()
 *		    w_makeScaleToGrayArbFilters ()
 *		    w_makeScaleToGrayArbLineTab ()
 *		    w_freeScaleToGrayArbFilters ()
 *		    w_freeScaleToGrayArblineTab ()
 *
 * These functions do arbitrary reduction of binary images using a box filter.
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "iaerror.pub"
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "reduce.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: reduce4.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reduces a binary PIXR producing a 4 bit gray result PIXR.
 *
 * LongDescription:::
 *  Reduction of a binary PIXR by 1.5, 2.0, 4.0, or 8.0 resulting in
 *  a four bit gray output PIXR.  The scale factor is passed in
 *  multiplied by 10 (e.g.  1.5 -> 15).  The output PIXR is created
 *  if necessary, otherwise it should be 10 * sourceWidth / factorx10 
 *  wide and 10 * sourceHeight / factorx10 tall.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prS::            Source PIXR to be scaled
 *  factorx10::      Reduction factor x10
 *  pScaleTable::    LUT made by w_makeScaleToGray4bppTable ()
 *
 * Output:::
 *  prD::            4-bit deep result
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray, scaling, isomorphic, depth conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleToGray4bpp (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt32              factorx10,
    UInt32             *pScaleTable,
    PIXR              **prD)
{
    PROLOGUE;
    Int32                wS, hS, wD, hD,
                         bplS, bplD;
    Bool                 destCreated = cFalse;
    Int32                rfc, bfc;
    ScaledInt1616        xRes, yRes;
    UInt8       *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prD != NULL)
	ERR_GEN_PIXR (*prD, cByteOrderBig, cDepth4, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetResolutionDPI(pIPWState, prS, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, (10 * wS) / factorx10,
					      (10 * hS) / factorx10, 4, 
					      (10 * xRes) / factorx10,
					      (10 * yRes) / factorx10,
					      cFrameBits, 
					      prD, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prS, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prD, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prD, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prD, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prD, &iD));

    /* do the real work */
    switch (factorx10)
    {
        case 15:
	    ERR_PASS_CLEANUP (i_scaleToGray4bpp1_5x (pIPWState,
						     iD, wD, hD, bplD,
	                                             iS, bplS,
					             pScaleTable));
	    break;

	case 20:
	    ERR_PASS_CLEANUP (i_scaleToGray4bpp2x (pIPWState, iD, wD, hD, bplD,
	                                           iS, bplS,
					           pScaleTable));
	    break;

	case 40:
	    ERR_PASS_CLEANUP (i_scaleToGray4bpp4x (pIPWState, iD, wD, hD, bplD,
	                                           iS, bplS,
					           pScaleTable));
	    break;

	case 80:
	    ERR_PASS_CLEANUP (i_scaleToGray4bpp8x (pIPWState, iD, wD, hD, bplD,
	                                           iS, bplS,
					           pScaleTable));
	    break;

	default:
	    ERR_GEN_CLEANUP (1, cErrInvalidParm);
	    break;
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prD, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:

    if (destCreated)
        w_destroyPixr (pIPWState, prD);
    ERR_RETURN;

}  /* w_scaleToGray4bpp () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Makes the necessary lookup table for w_scaleToGray4bpp ()
 *
 * LongDescription:::
 *  Given a reduction factor times ten of 15, 20, 40, or 80; creates
 *  the proper lookup table for use by w_scaleToGray4bpp ().
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State poiner
 *  factorx10::      Redection factor x10
 *
 * Output:::
 *  table::          The lookup table for use by w_scaleToGray4bpp ()
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray, scaling, depth conversion, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeScaleToGray4bppTable (
    IPWState           *pIPWState,
    UInt32              factorx10,
    UInt32            **table)
{
    PROLOGUE;
    UInt32	 tableSize;
    UInt32	*localTable;

     switch (factorx10)
    {
	case 15:
	    ERR_PASS (i_getSizeOfScaleToGray4bpp1_5xTab (pIPWState,
							 &tableSize));
	    localTable = IPW_MALLOC(tableSize, cAllocatorData);
	    ERR_GEN (table == NULL, ia_nomem);
	    ERR_PASS_CLEANUP (i_fillScaleToGray4bpp1_5xTab (pIPWState,
							    localTable,
							    tableSize));
	    break;

	case 20:
	    ERR_PASS (i_getSizeOfScaleToGray4bpp2xTab (pIPWState, &tableSize));
	    localTable = IPW_MALLOC(tableSize, cAllocatorData);
	    ERR_GEN (table == NULL, ia_nomem);
	    ERR_PASS_CLEANUP (i_fillScaleToGray4bpp2xTab (pIPWState, localTable,
							  tableSize));
	    break;

	case 40:
	    ERR_PASS (i_getSizeOfScaleToGray4bpp4xTab (pIPWState, &tableSize));
	    localTable = IPW_MALLOC(tableSize, cAllocatorData);
	    ERR_GEN (table == NULL, ia_nomem);
	    ERR_PASS_CLEANUP (i_fillScaleToGray4bpp4xTab (pIPWState, localTable,
							  tableSize));
	    break;

	case 80:
	    ERR_PASS (i_getSizeOfScaleToGray4bpp8xTab (pIPWState, &tableSize));
	    localTable = IPW_MALLOC(tableSize, cAllocatorData);
	    ERR_GEN (table == NULL, ia_nomem);
	    ERR_PASS_CLEANUP (i_fillScaleToGray4bpp8xTab (pIPWState, localTable,
							  tableSize));
	    break;

	default:
	    ERR_GEN (1, cErrInvalidParm);
	    break;
    }

    /* the operation must have worked */
    *table = localTable;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE (table, cAllocatorData);
    ERR_RETURN;

}  /* w_makeScaleToGray4bppTable () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arbitrary scale to gray of a binary PIXR to a 4-bit gray PIXR
 *
 * LongDescription:::
 *  Uses a filter created by w_makeScaleToGrayArbFilters () to perform
 *  arbitrary isomorphic scaling.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prS::            Binary source PIXR
 *  yD::             Destination line, in terms of "virtual" image
 *  filter::         Filter returned by makeScaleToGrayArbFilter
 *  lineTab::        Line table returned by makeScaleToGrayArbLineTab
 *
 * Output:::
 *  prD::            4-bit gray destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray, arbitrary scaling, depth conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleToGrayArb (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt32              yD,
    void               *filter,
    void               *lineTab,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD, bplS, bplD;
    UInt8       *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR (prD, cByteOrderBig, cDepth4, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, prD, &wD));
    ERR_PASS (w_pixrGetHeight (pIPWState, prD, &hD));

    ERR_PASS (w_pixrGetBpl (pIPWState, prS, &bplS));
    ERR_PASS (w_pixrGetBpl (pIPWState, prD, &bplD));
    ERR_PASS (w_pixrGetImageAll (pIPWState, prS, &iS));
    ERR_PASS (w_pixrGetImageAll (pIPWState, prD, &iD));


    ERR_PASS (i_scaleToGray4bppArb (pIPWState, iD, yD, wD, hD, bplD,
				    iS, bplS,
				    filter, lineTab));

    NORMAL_RETURN;

}  /* w_scaleToGrayArb () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generate filters for arbitrary scale-to-gray
 *
 * LongDescription:::
 *  This used to support a "phases" argument, which was the number of
 *  filter phases we'd interpolate. We no longer support that argument,
 *  which means that you should use a box filter if your scale factor
 *  is not an integer.
 *
 *  This also used to support passing in a filter generation function.
 *  Nobody ever used anything but box filters, because making a decent
 *  looking sinc or tent filter that worked on fractional pixel boundaries
 *  and didn't look funny when the alignment of pixels changed required
 *  ridiculously large kernels. The box filter has the advantage of looking
 *  equally fuzzy everywhere.
 *
 *  If want image scaled with one ratio, use wD and wS again, as input of
 *  hD and hS. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  wD::             Width of "virtual" dest image
 *  wS::             Width of "virtual" source image
 *  hD::             height of "virtual" dest image
 *  hS::             height of "virtual" source image
 *  depth::          Bit depth of dest image
 *  maxWidth::       Maximum width of filter
 *  maxHeight::      Maximum height of filter
 *
 * Output:::
 *  filter::         The resulting filter to be used by w_scaleToGrayArb ()
 * 
 * Returns:::
 *
 * Notes:::
 *  (1):: Does not support anamorphic scaling
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeScaleToGrayArbFilters (
    IPWState           *pIPWState,
    UInt32              wD,
    UInt32              wS,
    UInt32              hD,
    UInt32              hS,
    Int32               depth,
    Int32               maxWidth,
    Int32               maxHeight,
    void              **filter)
{
    PROLOGUE;
    Int32    filterBytes;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;
 
    ERR_GEN (*filter != NULL, cErrInvalidParm);

    filterBytes = i_makeScaleToGrayArbFilterBytes (pIPWState, wD, wS, hD, hS, 
						   i_boxFilterInfo,
						   1.0,
						   maxWidth, maxHeight);

    ERR_GEN (!filterBytes, cErrInternal);

    *filter = (void *) IPW_MALLOC (filterBytes, cAllocatorData);
    ERR_GEN ((*filter == NULL), cErrNomem);

    ERR_PASS_CLEANUP (i_makeScaleToGrayArbFilters (pIPWState, wD, wS,
						   hD, hS, depth, 
						   (FilterGen) i_boxFilterGen,
						   (FilterInfo)i_boxFilterInfo,
						   1.0,
				                   maxWidth, maxHeight, 
						   *filter));

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE (filter, cAllocatorData);
    ERR_RETURN;

}  /* w_makeScaleToGrayArbFilters () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create the line table for scale-to-gray
 *
 * LongDescription:::
 *  Create the line table for scale-to-gray. The reason that this
 *  is separate than the filter creation step is that we can make
 *  one filter per reduction factor, but we should make one line table
 *  for every source rectangle. Since the former takes longer, we
 *  allow for making the latter separately (e.g. if doing a "pan",
 *  we might call this several times without making a new filter). 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  filter::         Filter returned by makeScaleToGrayArbFilters
 *  xD::             Offset into "virtual" image for destination
 *  wD::             Width of desired destination image
 *  wVirt::          Width of "virtual" image, i.e. scaled
 *  prSrc::          Source Image
 *
 * Output:::
 *  table::          Line table for use by w_scaleToGrayArb ()
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeScaleToGrayArbLineTab (
    IPWState           *pIPWState,
    void               *filter,
    Int32               xD,
    Int32               wD,
    Int32               wVirt,
    PIXR               *prSrc,
    void              **table)
{
    PROLOGUE;
    Int32   lineTabBytes, wS, bplS;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN (filter == NULL, cErrInvalidParm);
    ERR_GEN (*table != NULL, cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &wS));
    ERR_PASS (w_pixrGetBpl (pIPWState, prSrc, &bplS));

    lineTabBytes = i_makeScaleToGrayArbLineTabBytes (pIPWState, filter,
						     xD,
						     wD,
						     wVirt,
						     wS,
						     bplS);
    ERR_GEN (!lineTabBytes, cErrInternal);

    *table = (void *) IPW_MALLOC (lineTabBytes, cAllocatorData);
    ERR_GEN ((*table == NULL), cErrNomem);

    ERR_PASS_CLEANUP (i_makeScaleToGrayArbLineTab (pIPWState, filter,
				                   xD,
				                   wD,
				                   wVirt,
				                   wS,
				                   bplS,
				                   *table));

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE (*table, cAllocatorData);
    ERR_RETURN;

}  /* w_makeScaleToGrayArbLineTab () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Frees filters allocated by w_makeScaleToGrayArbFilters ()
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  filters::        Filters to be freed.
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyScaleToGrayArbFilters (
    IPWState           *pIPWState,
    void              **filters)
{
    PROLOGUE;
    IPW_FREE (*filters, cAllocatorData);
    *filters = NULL;

    NORMAL_RETURN;

}  /* w_destroyScaleToGrayArbFilters () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free line table allocated by w_makeScaleToGrayArbLineTab ()
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  tab::            Table to be destroyed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scale to gray
 *
 * Keywords:::
 *  scale to gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyScaleToGrayArbLineTab (
    IPWState           *pIPWState,
    void              **tab)
{
    PROLOGUE;
    IPW_FREE (*tab, cAllocatorData);
    *tab = NULL;
    NORMAL_RETURN;

}  /* w_destroyScaleToGrayArbLineTab () */

/*
 $Log: reduce4.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.39  1998/01/14 00:14:55  ddavies
 reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.38  1997/11/12 18:15:17  xzhu
 xzhu on Wed Nov 12 10:10:01 PST 1997

 reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v

 Added input arguments hD and hS to get reductionY. If just want one reduction for the image, can input wD and wS as hD and hS like this: (wD, wS, wD, wS,      ).

 Revision 1.37  1997/04/15 22:32:33  ddavies
 reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.36  1996/03/26 23:30:31  ddavies
 reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v

 Eliminate wp_makeScaleMalloc, which contains a call to the obsolete MALLOC.
 Make each of the scale to gray ops in w_makeScaleToGray4bppTable get
 the size of the table, allocate it using the client's allocator and
 fill it.  This allows the client to free the tables with the client's
 freeing routine instead of having to know that these tables were secretly
 allocated using MALLOC.

 Revision 1.35  1996/02/20 19:02:03  macleod
 macleod on Tue Feb 20 11:01:35 PST 1996

 reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v

 The routines that free filter and line tables were freeing the pointer,
 not the table. This was not caught because the regression test

 Revision 1.34  1996/01/04 20:26:13  ejaquith
 reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.33  1995/12/13  04:12:20  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.32  1995/11/21  01:02:18  macleod
 * macleod on Mon Nov 20 17:01:34 PST 1995
 *
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Got rid of all of the box/tent/sinc2 filter nonsense, since only box
 * filters have worked right for a long time (a tradeoff between
 * functionality and speed--speed won)
 *
 * Revision 1.31  1995/11/16  02:18:25  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * w_makeScaleMalloc() becomes private: wp_makeScaleMalloc().
 * Fixed a typo in Input::: field of a routines comment block.
 *
 * Revision 1.30  1995/11/13  20:15:20  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.29  1995/11/08  22:19:15  dferrell
 * dferrell on Wed Nov 8 14:19:03 PST 1995
 *
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Fixed stupid, stupid typo.
 *
 * Revision 1.28  1995/11/08  19:52:29  dferrell
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Changed logic to replace ERR_CHECKs with ERR_PASS_CLEANUPs even in the
 * most brain-dead cases (I hate gotos even in macros).
 *
 * Revision 1.27  1995/10/13  22:04:31  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/09/26  21:58:33  dferrell
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.25  1995/09/22  00:38:40  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.24  1995/09/21  22:58:50  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.23  1995/09/19  21:10:34  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.22  1995/09/15  22:38:20  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.21  1995/09/15  00:33:36  dferrell
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.20  1995/09/11  22:56:39  dferrell
 * dferrell on Mon Sep 11 15:55:39 PDT 1995
 *
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Cleansed the scale to gray filter functions of all "ipwrapness" with
 * the exception of the w_ name prefix.
 *
 * Revision 1.19  1995/09/11  17:52:09  dferrell
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.18  1995/09/07  01:34:38  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.17  1995/09/06  20:22:44  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.16  1995/09/01  18:18:31  dferrell
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.15  1995/08/31  21:38:33  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.14  1995/08/29  22:24:17  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.13  1995/08/29  21:56:17  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.12  1995/08/25  01:38:06  ejaquith
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Fixed warnings
 *
 * Revision 1.11  1995/08/22  01:36:54  jcraig
 * jcraig on Mon Aug 21 18:27:47 PDT 1995
 *
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * boxFilterInfo and boxFilterGen should not take pIPWState as a
 * parameter and returns void not ErrCode.  (These are callbacks given
 * to ipshared.)
 *
 * Revision 1.10  1995/08/17  21:53:21  dferrell
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.9  1995/04/06  17:04:35  macleod
 * macleod on Thu Apr 6 10:00:31 PDT 1995
 *
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Changed the interfaces to scaleToGrayArbPr and makeScaleToGrayArbLineTab.
 * The original version supported a source window and a destination image.
 * This interface was not useful when scrolling through an image.
 * The interface was changed to take a source image, a "virtual" destination
 * width, which determines the scaling factor, and destination x,y offset
 * into the "virtual" image, allowing one to land on integer pixel boundaries
 * in the scaled image, which may cause pixels to land on fractional
 * boundaries in the source image.
 * Note that the line table must be re-computed every time the
 * x-offset or destination width changes, i.e. you can use the same line
 * table when scrolling vertically, but not when scrolling horizontally.
 *
 * Revision 1.8  1995/02/03  01:38:21  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.7  1995/01/13  01:54:37  macleod
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Changed the interface of abritrary scale-to-gray: eliminated the "phases"
 * parameter to a couple of functions
 *
 * Revision 1.6  1995/01/04  01:29:52  macleod
 * macleod on Tue Jan 3 17:25:56 PST 1995
 *
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Added scaleToGrayArbPr, makeScaleToGrayArbLineTab,
 * makeScaleToGrayArbFilters, tentFilterInfo, tentFilterGen, boxFilterInfo,
 * boxFilterGen, sinc2FilterInfo and sinc2FilterGen.
 * The filter functions are passed as arguments to makeScaleToGrayArbFilters,
 * the result of which is passed, along with the result of
 * makeScaleToGrayArbLineTab, to scaleToGrayArbP{ix}r.
 * The reason for different filter tables and line tables is that the
 * filter tables depend only on the reduction factor, and the line
 * table depends on the horizontal offset, so if one were "panning",
 * one could generate one filter table and then a different line table
 * for each horizontal pan.
 *
 * Revision 1.5  1994/10/03  20:34:57  lperry
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Changed to use reduce.pub instead of scalepros.
 *
 * Revision 1.4  1994/06/08  04:13:23  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.3  1993/12/03  01:18:25  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:16:11  ddavies
 * reduce4.c -> /product/ipcore/ipwrap/src/RCS/reduce4.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

