/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  reduce5.c: Interface layer for grayscale min/max reduction
 *
 *          2x min/max reduction
 *                  w_reduce2MinMax ()
 *                  w_reduceMinMax ()
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "reduce.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: reduce5.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reduce a gray-scale image by 2, taking minimum or maximum of 4 source pixels
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  srcPixr::        PIXR to be reduced
 *  useMax::         If true, use maximum, else use minimum
 *
 * Output:::
 *  destPixr::       Reduced PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce2MinMax (
    IPWState           *pIPWState,
    PIXR               *srcPixr,
    Int32               useMax,
    PIXR              **destPixr)
{
    PROLOGUE;
    PIXR	*thisSChannel, *thisDChannel;
    Int32	 wD, hD, wS, hS, 
		 sChannels, dChannels, i, 
		 sDepth, dDepth, frame;
    Bool	 destCreated = cFalse;
    Int32        rfc, bfc;
    ScaledInt1616		 xRes, yRes;
    ColorType	 type;
    ColorMap	*map;

    /* pixrs OK? */
    ERR_GEN_PIXR (srcPixr, cByteOrderBig, cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);
    if (*destPixr != NULL)
	ERR_GEN_PIXR (*destPixr, cByteOrderBig, cDepth4 | cDepth8, 
		      cChannelsMask, cFrame32);

    ERR_GEN ((useMax != 0 && useMax != 1), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, srcPixr, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, srcPixr, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, srcPixr, &sDepth));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, srcPixr, &sChannels));
    ERR_PASS (w_pixrGetFrame (pIPWState, srcPixr, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, srcPixr, &map));
    ERR_PASS (w_pixrGetType (pIPWState, srcPixr, &type));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, srcPixr, &xRes, &yRes));

    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
						      wS / 2, hS / 2, sDepth,
						      xRes / 2, yRes / 2,
						      frame, sChannels,
						      type, map,
						      destPixr, &rfc, &bfc,
						      &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, *destPixr, &dChannels));
    ERR_GEN_CLEANUP ((sChannels != dChannels), cErrInvalidParm);

    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, *destPixr, &dDepth));
    ERR_GEN_CLEANUP ((sDepth != dDepth), cErrSrcDimensNotEqual);

    if (sChannels > 1)
    {
	for (i = 0; i < sChannels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, srcPixr, 
	                                        i, &thisSChannel));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *destPixr, 
	                                        i, &thisDChannel));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, 
				               thisSChannel,
				               useMax, 
	                                       &thisDChannel));
	}
    }
    else
    {
	Int32        bplS, bplD;
	UInt8       *iS, *iD;

	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *destPixr, &wD));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *destPixr, &hD));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, srcPixr, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *destPixr, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, srcPixr, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *destPixr, &iD));

	if (useMax)
	{
	    if (sDepth == 8)
	        ERR_PASS_CLEANUP (i_reduce2Max8bpp (pIPWState, iD,
						    wD, hD, bplD, 
		                                    iS, bplS));
	    else
	        ERR_PASS_CLEANUP (i_reduce2Max4bpp (pIPWState, iD,
						    wD, hD, bplD, 
		                                    iS, bplS));
	}
	else
	{
	    if (sDepth == 8)
	        ERR_PASS_CLEANUP (i_reduce2Min8bpp (pIPWState, iD,
						    wD, hD, bplD, 
		                                    iS, bplS));
	    else
	        ERR_PASS_CLEANUP (i_reduce2Min4bpp (pIPWState, iD,
						    wD, hD, bplD, 
		                                    iS, bplS));
	}
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *destPixr,
						 rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, destPixr);
    ERR_RETURN;

}  /* w_reduce2MinMax () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Cascaded min/max reduction of gray images.
 *
 * LongDescription:::
 *  Reduces gray PIXR by 2, 4, 8, or 16 using cascaded min or max
 *  reductions.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pSrc::           Source PIXR
 *  reductions::     number of cascades (1, 2, 3, or 4)
 *  f0::             min/max flag for first stage (TRUE = max, else min)
 *  f1::             min/max flag for second stage (TRUE = max, else min)
 *  f2::             min/max flag for third stage (TRUE = max, else min)
 *  f3::             min/max flag for fourth stage (TRUE = max, else min)
 *
 * Output:::
 *  pDst::           Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceMinMax (
    IPWState           *pIPWState,
    PIXR               *pSrc,
    Int32               reductions,
    Int32               f0,
    Int32               f1,
    Int32               f2,
    Int32               f3,
    PIXR              **pDst)
{
    PROLOGUE;
    PIXR *t1 = NULL, 
         *t2 = NULL;

    /* pixrs OK? */
    ERR_GEN_PIXR (pSrc, cByteOrderBig, cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);

    switch (reductions)
    {
	case 1:
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, pSrc, f0, pDst));
	    break;

	case 2:
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, pSrc, f0, &t1));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, t1, f1, pDst));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t1));
	    break;

	case 3:
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, pSrc, f0, &t2));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, t2, f1, &t1));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t2));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, t1, f2, pDst));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t1));
	    break;

	case 4:
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, pSrc, f0, &t1));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, t1, f1, &t2));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t1));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, t2, f2, &t1));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t2));
	    ERR_PASS_CLEANUP (w_reduce2MinMax (pIPWState, t1, f3, pDst));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t1));
	    break;

	default:
	    ERR_GEN (1, cErrInvalidParm);
	    break;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &t1);
    w_destroyPixr (pIPWState, &t2);
    w_destroyPixr (pIPWState, pDst);
    ERR_RETURN;

}  /* w_reduceMinMax () */


/*
 * $Log: reduce5.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.27  1998/01/14 00:14:57  ddavies
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.26  1997/04/15 22:32:45  ddavies
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.25  1997/03/21 00:24:00  dferrell
 * dferrell on Thu Mar 20 16:23:11 PST 1997
 *
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Fixed incorrect comments for reduction type (min/max) parameters in
 * w_reduceMinMax ().
 *
 * Revision 1.24  1996/01/04 20:26:16  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.23  1995/12/13  04:12:24  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:15:23  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:04:35  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/09/26  21:58:39  dferrell
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.19  1995/09/21  22:58:54  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/19  21:10:40  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.17  1995/09/15  22:38:24  ddavies
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.16  1995/09/15  00:33:48  dferrell
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.15  1995/09/11  17:52:18  dferrell
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.14  1995/09/07  01:34:41  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:22:47  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  18:18:37  dferrell
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.11  1995/08/31  21:38:36  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/29  21:56:23  ddavies
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.9  1995/08/25  01:38:12  ejaquith
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * Fixed warnings
 *
 * Revision 1.8  1995/08/17  21:53:28  dferrell
 * reduce5.c -> /product/ipcore/ipwrap/src/RCS/reduce5.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 *
 */

