/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * Region functions
 *
 * This module provides all functions related to the creation and
 * modification of a region. A region is a representation of encoded
 * data within a layout specification, independent of the resolution
 * or rendering method. A region is implemented as a data map, i.e.
 * a bitmap where each bit represents the data bit to be rendered in
 * that position as a glyph.
 *
 * The functions in this module are organized alphabetically for easy
 * reference. Local (to this module) functions are defined first, then
 * internal (to the library) functions, then library entry points. Any
 * functions which are referenced before their definition within this
 * module have prototypes defined under "Forward References" below.
 *
 */

#define	DATAGLYPH_VERSION1

#include <stdio.h>


#include "types.pub"
#include "ipscore.prv"
#include "ipwerrs.pub"
#include "defines.pub"
#include "imageref.pub"
#include "dgencdef.pub"


#include "dgcom.prv"
#include "dgencint.prv"
#include "dgecc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif



IP_RCSINFO(RCSInfo, "$RCSfile: region.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/************************* Forward references *************************/

static void CDECL 
wp_fillOcclFrame (
    IPWState   *pIPWState,
    UInt8      *pData,
    UInt32	nBytes,
    PDGRegion	pRegion,
    UInt32	x,
    UInt32	y);

static ErrCode CDECL 
wp_generateBytes (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8     **genBytes,
    UInt32	nBytes);

static ErrCode CDECL 
wp_setSyncPattern (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8     **pSyncBits,
    UInt32	nSyncLen,
    Bool	anti);

static void CDECL 
wp_setSyncData (
    IPWState   *pIPWState,
    UInt8      *pSyncBits,
    UInt32	nSyncLen,
    UInt32	syncData);

static void CDECL 
wp_setSyncOcclData (
    IPWState   *pIPWState,
    UInt8      *pSyncBits,
    UInt32	nSyncLen);

/************************** Local functions ***************************/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Places the horz and vert sync patterns into the region's data map
 *
 * LongDescription:::
 * Places the horizontal and vertical sync patterns into the region's
 * data map, providing the grid within which the data will be placed.
 * The variable data within the sync pattern represents the distance
 * from the edge of the glyph region, in frames, in each of the horz
 * and vert dimensions, modulo 16, so that, for example, a region 5
 * frames high will have vertical sync data of 0,1,2,2,1,0. Sync is
 * placed only in the area containing valid frames, i.e. there is no
 * sync in the pad areas on the right and bottom of the region.
 * However, if there is padding in either dimension, then "anti-sync"
 * is added to the region. Anti-sync is constructed in such a way as
 * to be recognisable as the inverse of regular sync, regardless of
 * orientation.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::     	region handle   
 *
 * Returns:::
 *
 * Notes:::
 *  data map and mask inside region handle update
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper	initial version
 *  Julie Craig		add occlusions
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL 
wp_createSync (
    IPWState   *pIPWState,
    PDGRegion	pRegion)
{
    PROLOGUE;
    UInt32 nInc, nBit, nRow, nCol, nLen, nCenter;
    UInt32 nWidthNoPad, nHeightNoPad;
    UInt8 *pSync = NULL;
    UInt32 x, y;
    LayoutRect fRect, gRect;
    UInt32 glyphStartOccl, glyphEndOccl;
    Int32 occlSync;
    UInt8 *pDataMap, *pDataMask;

    /* get the address of the data map and mask */
    pDataMap = GET_DATA_MAP(pRegion);
    pDataMask = GET_DATA_MASK(pRegion);

    /* Compute the size for sync coding */
    nWidthNoPad = pRegion->nWidth - pRegion->padHorz;
    nHeightNoPad = pRegion->nHeight - pRegion->padVert;

    /* Create the horizontal sync */
    nLen = pRegion->frameWidth + 2;
    ERR_PASS(wp_setSyncPattern(pIPWState, pRegion, &pSync, nLen, cFalse));

    /* set up occlusion information */
    fRect = pRegion->fOcclusion;
    gRect = pRegion->gOcclusion;
    glyphStartOccl = fRect.x * (pRegion->frameWidth + 1);
    glyphEndOccl = glyphStartOccl + (fRect.w * (pRegion->frameWidth + 1));

    /* Add sync rows */
    nInc = pRegion->frameHeight + 1;
    nCenter = pRegion->frameRows / 2;
    if (pRegion->hasOcclusion)
	/* then for each frame we have to determine if we are in the
	 * occlusion.
	 */
    	for (nRow = 0, y = 0; nRow <= pRegion->frameRows; nRow++, y += nInc)
        {
	    nBit = 0;
	    ERR_IGNORE(wp_setSyncData(pIPWState, pSync, nLen,
	        ((nRow <= nCenter) ? nRow : (pRegion->frameRows - nRow))));
    	    occlSync = cFalse;
	    for (x = 0; x < nWidthNoPad; x++)
	    {
		/* if the fence will even potentially be needed, get it */
		if (((nRow == fRect.y) || (nRow == (fRect.y + fRect.h))) &&
		    ((x >= glyphStartOccl) && (x < glyphEndOccl)))
		{
		    /* only set the sync if necessary */
		    if (occlSync == cFalse)
		    {
		        ERR_IGNORE(wp_setSyncOcclData(pIPWState, pSync, nLen));
			occlSync = cTrue;
		    }
		}
		else
		    if (occlSync == cTrue)
		    {
		        ERR_IGNORE(wp_setSyncData(pIPWState, pSync, nLen,
		            ((nRow <= nCenter) ? nRow : (pRegion->frameRows - nRow))));
		        occlSync = cFalse;
		    }

		/* should we write the sync? */
		if (pRegion->isTransparent ||
		   (x < gRect.x) || (x >= (gRect.x + gRect.w)) ||
		   (y < gRect.y) || (y >= (gRect.y + gRect.h)))
		{
	            if (TEST_BYTEARRAY_BIT(pSync, nBit))
		        SET_BYTEARRAY_BIT(pDataMap, x, y,
				          pRegion->nWidthRound);
		}
		else CLEAR_BYTEARRAY_BIT(pDataMask, x, y,
				         pRegion->nWidthRound);
	        if (++nBit >= nLen)
		    nBit = 1;
	    }
        }
    else
    	for (nRow = 0, y = 0; nRow <= pRegion->frameRows; nRow++, y += nInc)
        {
	    ERR_IGNORE(wp_setSyncData(pIPWState, pSync, nLen,
		((nRow <= nCenter) ? nRow : (pRegion->frameRows - nRow))));
	    nBit = 0;
	    for (x = 0; x < nWidthNoPad; x++)
	    {
	        if (TEST_BYTEARRAY_BIT(pSync, nBit))
		    SET_BYTEARRAY_BIT(pDataMap, x, y, pRegion->nWidthRound);
	        if (++nBit >= nLen)
		    nBit = 1;
	    }
        }

    /* Free the sync pattern */
    GLY_FREE(pRegion->freeProc, pRegion->allocInfo, pSync);

    /* Create the vertical sync */
    nLen = pRegion->frameHeight + 2;
    ERR_PASS(wp_setSyncPattern(pIPWState, pRegion, &pSync, nLen, cFalse));

    /* set up occlusion information */
    glyphStartOccl = fRect.y * (pRegion->frameHeight + 1);
    glyphEndOccl = glyphStartOccl + (fRect.h * (pRegion->frameHeight + 1));
    /* Add sync columns */
    nInc = pRegion->frameWidth + 1;
    nCenter = pRegion->frameCols / 2;
    if (pRegion->hasOcclusion)
        for (nCol = 0, x = 0; nCol <= pRegion->frameCols; nCol++, x += nInc)
        {
	    nBit = 0;
	    ERR_IGNORE(wp_setSyncData(pIPWState, pSync, nLen,
	        ((nCol <= nCenter) ? nCol : (pRegion->frameCols - nCol))));
    	    occlSync = cFalse;
	    for (y = 0; y < nHeightNoPad; y++) 
	    {
		if (((nCol == fRect.x) || (nCol == (fRect.x + fRect.w))) &&
		    ((y >= glyphStartOccl) && (y < glyphEndOccl)))
		{
		    /* only set the sync if necessary */
		    if (occlSync == cFalse)
		    {
		        ERR_IGNORE(wp_setSyncOcclData(pIPWState, pSync, nLen));
		        occlSync = cTrue;
		    }
		}
		else
		    if (occlSync == cTrue)
		    {
		        ERR_IGNORE(wp_setSyncData(pIPWState, pSync, nLen,
		            ((nCol <= nCenter) ? nCol : (pRegion->frameCols - nCol))));
		        occlSync = cFalse;
		    }

		if (pRegion->isTransparent ||
		   (y < gRect.y) || (y >= (gRect.y + gRect.h)) ||
		   (x < gRect.x) || (x >= (gRect.x + gRect.w)))
		{
	            if (TEST_BYTEARRAY_BIT(pSync, nBit))
		        SET_BYTEARRAY_BIT(pDataMap, x, y,
				          pRegion->nWidthRound);
		}
		else CLEAR_BYTEARRAY_BIT(pDataMask, x, y, 
				         pRegion->nWidthRound);
	        if (++nBit >= nLen)
		    nBit = 1;
	    }
         }
    else
        for (nCol = 0, x = 0; nCol <= pRegion->frameCols; nCol++, x += nInc)
        {
	    ERR_IGNORE(wp_setSyncData(pIPWState, pSync, nLen,
		((nCol <= nCenter) ? nCol : (pRegion->frameCols - nCol))));
	    nBit = 0;
	    for (y = 0; y < nHeightNoPad; y++) 
	    {
	        if (TEST_BYTEARRAY_BIT(pSync, nBit))
		    SET_BYTEARRAY_BIT(pDataMap, x, y, pRegion->nWidthRound);
	        if (++nBit >= nLen)
		    nBit = 1;
	    }
         }

    /* Free the sync pattern */
    GLY_FREE(pRegion->freeProc, pRegion->allocInfo, pSync);
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Fills the frame at location (x,y) in the region's data map
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pData::     	encoded data     
 *  nBytes::         	number of bytes of encoded data
 *  pRegion::        	region handle
 *  x::              	frame column in which to place data
 *  y::              	frame row in which to place data
 *
 * Returns:::
 *
 * Notes:::
 *  The bits are placed within the data map from left to right and then
 *  from top to bottom.
 *  dataMap inside region handle is updated.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_fillEntireFrame (
    IPWState   *pIPWState,
    UInt8      *pData,
    UInt32	nBytes,
    PDGRegion	pRegion,
    UInt32	x,
    UInt32	y)
{
    UInt32 xPos, yPos;
    UInt32 nBit;
    UInt8 *pDataMap;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    if (nBytes == 0)
	return;
    nBit = 0;
    pDataMap = GET_DATA_MAP(pRegion);
    for (yPos = y; yPos < y + pRegion->frameHeight; yPos++) 
    {
	for (xPos = x; xPos < x + pRegion->frameWidth; xPos++) 
	{
	    if (nBytes > 0) 
	    {
		if (*pData & (0x80 >> nBit))
		    SET_BYTEARRAY_BIT(pDataMap, xPos, yPos,
				      pRegion->nWidthRound);
		if (++nBit >= 8) 
		{
		    pData++;
		    nBit = 0;
		    nBytes--;
		}
	    }
	}
    }
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Fills the frame at location (x,y) in the region's data mask with 0
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::     	region handle   
 *  x::              	frame column in which to place 0
 *  y::              	frame row in which to place 0
 *
 * Returns:::
 *
 * Notes:::
 *  The bits are placed within the data mask from left to right and then
 *  from top to bottom.
 *  dataMask inside region handle is updated.
 *  0 indicates a glyph which should not be rendered.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_fillEntireMask (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt32	x,
    UInt32	y)
{
    UInt32 xPos, yPos;
    UInt8 *pDataMask;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    pDataMask = (UInt8 *)pRegion + pRegion->nRegionSize + pRegion->nMapSize;
    for (yPos = y; yPos < y + pRegion->frameHeight; yPos++) 
    {
	for (xPos = x; xPos < x + pRegion->frameWidth; xPos++) 
	{
	    CLEAR_BYTEARRAY_BIT(pDataMask, xPos, yPos, pRegion->nWidthRound);
	}
    }
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Places the encoded data into the region's data map, a frame at a time
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::  	      	region handle
 *
 * Notes:::
 *  the frames are placed within the region from left to right and then
 *  from top to bottom.
 *  data map and mask inside region handle is updated.
 *
 * FunctionCategory:::
 *   DataGlyph
 *
 * Keywords:::
 *   DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper		initial version
 *  Julie Craig			occlusions
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_fillFrames (
    IPWState   *pIPWState,
    PDGRegion	pRegion)
{
    UInt32 nBytesPerFrame = pRegion->frameWidth * pRegion->frameHeight / 8;
    UInt32 nBytesRemaining = pRegion->nEncodedBytes;
    UInt8 * pData;
    UInt32 nIncX, nIncY, nRow, nCol, nBytesThisFrame;
    UInt32 x, y;
    LayoutRect fRect;

    nIncX = pRegion->frameWidth + 1;
    nIncY = pRegion->frameHeight + 1;
    fRect = pRegion->fOcclusion;
    pData = GET_ENCODED_DATA(pRegion);
    if (pRegion->hasOcclusion)
        for (nRow = 0, y = 1; nRow < pRegion->frameRows; nRow++, y += nIncY)
        {
	    for (nCol = 0, x = 1; nCol < pRegion->frameCols; nCol++, x += nIncX) 
	    {
	        nBytesThisFrame = MIN(nBytesRemaining, nBytesPerFrame);
		/* are we on one of the frames around the border of the
		 * occlusion? 
		 */
		if ((((nRow == fRect.y) || 
		      (nRow == (fRect.y + fRect.h - 1))) &&
		     ((nCol >= fRect.x) &&
		      (nCol < (fRect.x + fRect.w)))) ||
		    (((nCol == fRect.x) || 
		      (nCol == (fRect.x + fRect.w - 1))) &&
		     ((nRow >= fRect.y) &&
		      (nRow < (fRect.y + fRect.h)))))
		{
		    /* if the occlusion is transparent just fill the entire
		     * frame with the next frame's data.  if the occlusion
		     * is not transparent, only fill the part of the
		     * frame between the sync and the start of the 
		     * occlusion.
		     */
		    if (pRegion->isTransparent)
			ERR_IGNORE(wp_fillEntireFrame(pIPWState, pData,
						      nBytesThisFrame,  
						      pRegion, x, y));
		    else ERR_IGNORE(wp_fillOcclFrame(pIPWState,
						     pData, nBytesThisFrame,
						     pRegion, x, y));
		}
		else if ((nRow > fRect.y) && (nRow < (fRect.y + fRect.h)) &&
		         (nCol > fRect.x) && (nCol < (fRect.x + fRect.w)))
		{
		    /* then we are in the middle of the occlusion */
		    if (pRegion->isTransparent)
			ERR_IGNORE(wp_fillEntireFrame(pIPWState, pData,
						      nBytesThisFrame,
						      pRegion, x, y));
		    else ERR_IGNORE(wp_fillEntireMask(pIPWState,
						      pRegion, x, y));
		}
		else
		{
	            ERR_IGNORE(wp_fillEntireFrame(pIPWState,
						  pData, nBytesThisFrame,
						  pRegion, x, y));
	            pData += nBytesThisFrame;
	            nBytesRemaining -= nBytesThisFrame;
		}
	    }
        }
    else	/* no occlusion */
        for (nRow = 0, y = 1; nRow < pRegion->frameRows; nRow++, y += nIncY)
        {
	    for (nCol = 0, x = 1; nCol < pRegion->frameCols; nCol++, x += nIncX) 
	    {
	        if (nBytesRemaining == 0)
		    break;
	        nBytesThisFrame = MIN(nBytesRemaining, nBytesPerFrame);
	        ERR_IGNORE(wp_fillEntireFrame(pIPWState, pData, nBytesThisFrame,
					      pRegion, x, y));
	        pData += nBytesThisFrame;
	        nBytesRemaining -= nBytesThisFrame;
	    }
        }
}

/*
 * wp_fillOcclFrame()
 *
 * Fills the frame at location (x,y) in the region's data map with
 * data from the byte stream passed as a parameter. The bits are
 * placed within the data map from left to right and then from top
 * to bottom.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Fills the occluded frame at location (x,y) in the data map with data
 *
 * LongDescription:::
 * The occluded frame is filled with data from a valid frame.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pData::			data from a non-occluded frame
 *  nBytes::         		number of bytes to fill
 *  pRegion::        		region handle
 *  x::              		occluded frame column
 *  y::              		occluded frame row
 *
 * Notes:::
 *  The bits are placed within the data map from left to right and then 
 *  from top to bottom.
 *  data map and mask insdide region handle is updated.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper		initial version
 *  Julie Craig			occlusions
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_fillOcclFrame (
    IPWState   *pIPWState,
    UInt8      *pData,
    UInt32	nBytes,
    PDGRegion	pRegion,
    UInt32	x,
    UInt32	y)
{
    UInt32 xPos, yPos;
    UInt32 nBit;
    UInt8 *pDataMap, *pDataMask;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    nBit = 0;
    pDataMap = GET_DATA_MAP(pRegion);
    pDataMask = GET_DATA_MASK(pRegion);
    for (yPos = y; yPos < y + pRegion->frameHeight; yPos++) 
    {
	for (xPos = x; xPos < x + pRegion->frameWidth; xPos++) 
	{
	    if ((xPos < pRegion->gOcclusion.x) ||
		(xPos >= (pRegion->gOcclusion.x + pRegion->gOcclusion.w)) ||
		(yPos < pRegion->gOcclusion.y) ||
		(yPos >= (pRegion->gOcclusion.y + pRegion->gOcclusion.h)))
	    {
	        if (nBytes > 0) 
	        {
		    if (*pData & (0x80 >> nBit))
		        SET_BYTEARRAY_BIT(pDataMap, xPos, yPos, 
				          pRegion->nWidthRound);
		    if (++nBit >= 8) 
		    {
		        pData++;
		        nBit = 0;
		        nBytes--;
		    }
	        }
	    }
	    else CLEAR_BYTEARRAY_BIT(pDataMask, xPos, yPos, 
				     pRegion->nWidthRound);
	}
    }
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Place data into the pad at the right and bottom of the rectangle.
 *
 * LongDescription:::
 * Place something besides all 0's (which we get because the data map
 * was cleared by the calling procedure) into the pad at the right and 
 * bottom of the rectangle.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::        	region handle
 *
 * Returns:::
 *
 * Notes:::
 * data map and mask inside region handle is updated.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL 
wp_fillPad (
    IPWState   *pIPWState,
    PDGRegion	pRegion)
{
    PROLOGUE;
    UInt8		*aSync = NULL;
    UInt32		 dBit;
    LayoutRect	 gRect;
    UInt32		 rectHeight, rectWidth;
    UInt32		 sBit;
    UInt32		 xPos, yPos;
    UInt8 		*pDataMap, *pDataMask;

    gRect = pRegion->gOcclusion;
    rectWidth = pRegion->nWidth - pRegion->padHorz;
    rectHeight = pRegion->nHeight - pRegion->padVert;
    pDataMap =  GET_DATA_MAP(pRegion);
    pDataMask = GET_DATA_MASK(pRegion);

    if (pRegion->padHorz != 0)
    {
	/* set up anti-sync pattern for the rightmost column.  anti-sync
	 * is used so that we will not mistake this side for the
	 * real frame if the image has been rotated.  the chances are 
	 * small that random bits would yield sync but just in case...
	 */
	ERR_PASS_CLEANUP(wp_setSyncPattern(pIPWState, pRegion, &aSync, 
				           pRegion->frameHeight + 2, cTrue));
	/* which bit in the sync pattern should we get next? */
	sBit = dBit = 0;
	for (yPos = 0; yPos < rectHeight; yPos++)
	{
	    /* fill in each row of pad except the last bit which will
	     * be filled with anti-synch at the end of the loop.
	     */
	    for (xPos = rectWidth; xPos < pRegion->nWidth - 1; xPos++, dBit++)
	    {
		if (!pRegion->hasOcclusion || pRegion->isTransparent ||
		    (xPos < gRect.x) || (xPos >= (gRect.x + gRect.w)) ||
		    (yPos < gRect.y) || (yPos >= (gRect.y + gRect.h)))
		{
		    if (TEST_BYTEARRAY_BIT(pRegion->pRandom, (dBit & RTABLE_MASK)))
		        SET_BYTEARRAY_BIT(pDataMap, xPos, yPos,
				          pRegion->nWidthRound);
		}
	 	else CLEAR_BYTEARRAY_BIT(pDataMask, xPos, yPos,
				         pRegion->nWidthRound);
	    }
	    /* now the bit of anti-sync */
	    if (!pRegion->hasOcclusion || pRegion->isTransparent ||
		(xPos < gRect.x) || (xPos >= (gRect.x + gRect.w)) ||
		(yPos < gRect.y) || (yPos >= (gRect.y + gRect.h)))
	    {
		if (TEST_BYTEARRAY_BIT(aSync, sBit))
		    SET_BYTEARRAY_BIT(pDataMap, xPos, yPos,
				      pRegion->nWidthRound);
	    }
	    else CLEAR_BYTEARRAY_BIT(pDataMask, xPos, yPos,
				     pRegion->nWidthRound);
	    if (++sBit >= (pRegion->frameHeight + 2))
		/* the first bit of this frame is the last of the old so
		 * skip a bit.
		 */
		sBit = 1;
	}
	GLY_FREE(pRegion->freeProc, pRegion->allocInfo, aSync);
    }

    /* now fill in the pad at the bottom of the rectangle */
    if (pRegion->padVert != 0)
    {
	ERR_PASS_CLEANUP(wp_setSyncPattern(pIPWState, pRegion, &aSync,
				           pRegion->frameWidth + 2, cTrue));
	/* which bit in the sync pattern should we get next? */
	sBit = dBit = 0;
	for (yPos = rectHeight; 
	     yPos < pRegion->nHeight - 1; 
	     yPos++)
	{
	    /* fill in each row of pad except the last which will be 
	     * filled with anti-synch later.
	     */
	    for (xPos = 0; xPos < pRegion->nWidth; xPos++, dBit++)
	    {
		if (!pRegion->hasOcclusion || pRegion->isTransparent ||
		    (xPos < gRect.x) || (xPos >= (gRect.x + gRect.w)) ||
		    (yPos < gRect.y) || (yPos >= (gRect.y + gRect.h)))
		{
		    if (TEST_BYTEARRAY_BIT(pRegion->pRandom, (dBit & RTABLE_MASK)))
			SET_BYTEARRAY_BIT(pDataMap, xPos, yPos,
				          pRegion->nWidthRound);
		}
	 	else CLEAR_BYTEARRAY_BIT(pDataMask, xPos, yPos,
				         pRegion->nWidthRound);
	    }
	}
	/* fill in the anti-synch that is under the rectangle */
	for (xPos = 0; xPos < pRegion->nWidth; xPos++)
	{
	    if (!pRegion->hasOcclusion || pRegion->isTransparent ||
		(xPos < gRect.x) || (xPos >= (gRect.x + gRect.w)) ||
		(yPos < gRect.y) || (yPos >= (gRect.y + gRect.h)))
	    {
		if (TEST_BYTEARRAY_BIT(aSync, sBit))
		    SET_BYTEARRAY_BIT(pDataMap, xPos, yPos,
				      pRegion->nWidthRound);
	    }
	    else CLEAR_BYTEARRAY_BIT(pDataMask, xPos, yPos,
				     pRegion->nWidthRound);
	    if (++sBit >= (pRegion->frameHeight + 2))
		/* the first bit of this frame is the last of the old so
		 * skip a bit.
		 */
		sBit = 1;
	}
	GLY_FREE(pRegion->freeProc, pRegion->allocInfo, aSync);
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    if (aSync != NULL) 
	GLY_FREE(pRegion->freeProc, pRegion->allocInfo, aSync);
    ERR_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Generate bytes to be used as fill in the pad.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::     	   	region handle
 *  nBytes::         		number of bytes to generate
 *
 * Output:::
 *  genBytes::       		generated fill bytes
 *
 * Returns:::
 *  cErrNomem			unable to allocate memory
 *  cSuccessful			successfully generated table
 *
 * Notes:::
 * the number of entries in the generated table will actually be a
 * multiple of 4.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL 
wp_generateBytes (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8     **genBytes,
    UInt32	nBytes)
{
    PROLOGUE;
    UInt32	 i;
    UInt32	 seed;
    UInt32	*table;

    if (nBytes == 0)
    {
	NORMAL_RETURN;
    }
    /* make nBytes a multiple of 4 bytes */
    nBytes = DIVUP(nBytes, 4) * 4;

    if (*genBytes == NULL)
    {
        table = (UInt32 *)GLY_MALLOC(pRegion->mallocProc, 
				     pRegion->allocInfo, nBytes);
    	ERR_GEN(table == NULL, cErrNomem);
	*genBytes = (UInt8 *)table;
    }
    else table = (UInt32 *)*genBytes;

    seed = DG_RANDOM_SEED;
    for (i = 0; i < nBytes / 4; i++)
    {
	GLY_INT_RAND(seed);
	table[i] = seed;
    }

    NORMAL_RETURN;    
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the variable data bits in the sync pattern.
 *
 * LongDescription:::
 * Sets the variable data bits in the sync pattern. These bits are in
 * the center of the sync pattern, the remainder of the pattern having
 * been created using wp_setSyncPattern(). Each data bit is set to its
 * bit value (i.e. not just set if it is a 1), so that this function
 * can be used to overwrite the data written in a previous call.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nSyncLen::       	length of this sync pattern
 *  syncData::       	data to fill in the sync pattern
 *
 * Output:::
 *  pSyncBits::      	pointer to sync bits
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_setSyncData (
    IPWState   *pIPWState,
    UInt8      *pSyncBits,
    UInt32	nSyncLen,
    UInt32	syncData)
{
    UInt32 nFillBits, bitNum, i;
    static int dataRemap[] =
	{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15};

    /* Keep compiler happy */
    pIPWState = pIPWState;

    /* If the sync is too short, do nothing */
    if (nSyncLen < SYNC_EXT_LEN)
	return;

    /* Set the data bits */
    nFillBits = nSyncLen - (SYNC_START_LEN + SYNC_DATA_LEN + SYNC_END_LEN);
    bitNum = SYNC_START_LEN + nFillBits / 2 + SYNC_DATA_LEN - 1;

    /* transform syncData into the number than goes in the data bits.
     * occlusions took one of the data bit values.
     */
    syncData = dataRemap[syncData % 15];
    for (i = 0; i < SYNC_DATA_LEN; i++, bitNum--, syncData >>= 1)
	PUT_BYTEARRAY_BIT(pSyncBits, bitNum, (syncData & 1));
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the occlusion data bits in the sync pattern.
 *
 * LongDescription:::
 * Sets the occlusion data bits in the sync pattern. These bits are in
 * the center of the sync pattern, the remainder of the pattern having
 * been created using wp_setSyncPattern(). Each data bit is set to its
 * bit value (i.e. not just set if it is a 1), so that this function
 * can be used to overwrite the data written in a previous call.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  nSyncLen::       	size of the sync pattern
 *
 * Output:::
 *  pSyncBits::      	pointer to sync bits
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper	initial version
 *  Julie Craig		occlusions
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL 
wp_setSyncOcclData (
    IPWState   *pIPWState,
    UInt8      *pSyncBits,
    UInt32	nSyncLen)
{
    UInt32 nFillBits, bitNum, i;
    UInt32 syncData;

    /* Keep compiler happy */
    pIPWState = pIPWState;

    /* If the sync is too short, do nothing */
    if (nSyncLen < SYNC_EXT_LEN)
	return;

    /* Set the data bits */
    nFillBits = nSyncLen - (SYNC_START_LEN + SYNC_DATA_LEN + SYNC_END_LEN);
    bitNum = SYNC_START_LEN + nFillBits / 2 + SYNC_DATA_LEN - 1;

    syncData = cOcclusionInfoBits;
    for (i = 0; i < SYNC_DATA_LEN; i++, bitNum--, syncData >>= 1)
	PUT_BYTEARRAY_BIT(pSyncBits, bitNum, (syncData & 1));
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Generate the bits in the sync code.
 *
 * LongDescription:::
 * Sets the bit pattern for a sync code. The actual bit pattern used
 * depends upon the size of the frames to be used in the layout, but
 * is constructed using the pattern ABCDE, where A and E represent
 * fixed length starting and ending sequences, B and D represent some
 * number of fill bits according to the patterns 0101... (for B) and
 * 1010... (for D), and C represents some number of bits of variable
 * data. If the sync length is less than the length of A + C + E,
 * then the pattern becomes ABE (i.e. the variable data is omitted,
 * and there is only one set of fill bits).
 *
 * PrivateDescription:::
 * Anti-synch is just the inverse of the regular sync
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pRegion::        	region handle
 *  nSyncLen::       	length of the sync pattern
 *  anti::           	should anti-sync be generated?
 *
 * Output:::
 *  pSyncBits::      	pointer to sync pattern
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper	initial version
 *  Julie Craig		anti-sync
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL 
wp_setSyncPattern (
    IPWState   *pIPWState,
    PDGRegion	pRegion,
    UInt8     **pSyncBits,
    UInt32	nSyncLen,
    Bool	anti)
{
    PROLOGUE;
    UInt32 nFillBits, nFillStart, fillPattern, mask, bitNum, i;
    UInt8 * syncBits;

    syncBits = GLY_CALLOC(pRegion->callocProc, pRegion->allocInfo,
			  DIVUP(nSyncLen, 8), sizeof(UInt8));
    ERR_GEN(syncBits == NULL, cErrNomem);

    /* Calculate the number of fill bits required */
    if (nSyncLen >= SYNC_EXT_LEN) 
    {
	nFillBits = nSyncLen - SYNC_EXT_LEN;
	nFillStart = nFillBits / 2;
    }
    else
    nFillBits = nFillStart = nSyncLen - SYNC_BASE_LEN;
    if (nFillBits > 0)
    {
	if (anti)
	    ERR_IGNORE(wp_store32(pIPWState, ASYNC_FILL_BITS,
				  (UInt8 *)&fillPattern));
	else ERR_IGNORE(wp_store32(pIPWState, SYNC_FILL_BITS,
				   (UInt8 *)&fillPattern));
    }

    /* Now compose the sync. First the starting sequence. We are
     * guaranteed to get all of these into the byte without crossing
     * any boundaries.
     */
    if (anti)
	*syncBits = ASYNC_START_BITS;
    else *syncBits = SYNC_START_BITS;
    bitNum = SYNC_START_LEN;

    /* Now the start fill bits */
    for (i = 0; i < nFillStart; i++, bitNum++)
	if (TEST_BYTEARRAY_BIT(&fillPattern, i))
	    SET_BYTEARRAY_BIT(syncBits, bitNum, 0, 0);

    /* Leave space for the variable data */
    if (nSyncLen >= SYNC_EXT_LEN)
	bitNum += SYNC_DATA_LEN;

    /* Now the end fill bits */
    for (i = nFillStart; i < nFillBits; i++, bitNum++)
	if (TEST_BYTEARRAY_BIT(&fillPattern, i))
	    SET_BYTEARRAY_BIT(syncBits, bitNum, 0, 0);

    /* Now the ending sequence */
    mask = 0x80;
    if (anti)
    {
        for (i = 0; i < SYNC_END_LEN; i++, bitNum++, mask >>= 1)
            if (ASYNC_END_BITS & mask)
		SET_BYTEARRAY_BIT(syncBits, bitNum, 0, 0);
    }
    else
        for (i = 0; i < SYNC_END_LEN; i++, bitNum++, mask >>= 1)
            if (SYNC_END_BITS & mask)
		SET_BYTEARRAY_BIT(syncBits, bitNum, 0, 0);

    /* Set the sync pattern */
    *pSyncBits = syncBits;

    NORMAL_RETURN;
}

/************************* External functions *************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Creates and initializes the region handle.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  mallocProc::     	memory allocation procedure to use
 *  callocProc::     	memory allocation and clearing procedure to use
 *  freeProc::       	memory deallocation procedure to use
 *  allocInfo::      	pointer to client info passed to memory procs
 *  hLayout::        	layout handle
 *
 * Output:::
 *  phRegion::       	region handle
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated and initialized region handle
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doCreateRegion (
    IPWState   *pIPWState,
    MallocType	mallocProc,
    CallocType	callocProc,
    FreeType	freeProc,
    void       *allocInfo,
    HDGLayout	hLayout,
    HDGRegion  *phRegion)
{
    PROLOGUE;
    PDGLayout pLayout = (PDGLayout)hLayout;
    PDGRegion pRegion;
    UInt32 nMapSize, nRegionSize, nTotalBytes;
    UInt32 handleSize;
    UInt8 *tempPtr;

    ERR_GEN(pLayout == NULL, cErrDG_InvalidLayout);
    ERR_GEN(phRegion == NULL, cErrInvalidParm);
    ERR_GEN(*phRegion != NULL, cErrInvalidParm);

    ERR_IGNORE(wp_doGetRegionHandleSize(pIPWState, pLayout,
					&nMapSize, &nRegionSize, 
			    	        &nTotalBytes, &handleSize));
    /* Create the region */
    pRegion = (PDGRegion)GLY_CALLOC(callocProc, allocInfo, handleSize, 
				    sizeof(UInt8));
    ERR_GEN(pRegion == NULL, cErrNomem);

    /* Set the client's region handle */
    *phRegion = (HDGRegion)pRegion;


    /* Fill in the region fields */
    pRegion->nDataBytes = pLayout->enc.nDataBytes;
    pRegion->nEncodedBytes = nTotalBytes;
    pRegion->nCodewords = pLayout->enc.nCodewords;
    pRegion->nCodewordSize = pLayout->enc.nCodewordSize;
    pRegion->nCodewordParity = pLayout->enc.nCodewordParity;
    pRegion->selfDescribing = pLayout->enc.selfDescribing;

    pRegion->hasOcclusion = pLayout->lyt.hasOcclusion &&
	(pLayout->lyt.fOcclusion.x < pLayout->lyt.frameCols) &&
	(pLayout->lyt.fOcclusion.y < pLayout->lyt.frameRows);
    pRegion->fOcclusion = pLayout->lyt.fOcclusion;
    pRegion->gOcclusion = pLayout->lyt.gOcclusion;
    pRegion->isTransparent = pLayout->lyt.isTransparent;
    pRegion->frameWidth = pLayout->lyt.frameWidth;
    pRegion->frameHeight = pLayout->lyt.frameHeight;
    pRegion->frameRows = pLayout->lyt.frameRows;
    pRegion->frameCols = pLayout->lyt.frameCols;
    pRegion->padHorz = pLayout->lyt.padHorz;
    pRegion->padVert = pLayout->lyt.padVert;
    pRegion->nUnusedFrameBytes = pLayout->lyt.nUnusedFrameBytes;
    pRegion->glyphSize = pLayout->lyt.glyphSize;
    pRegion->nWidth = FRAMES_TO_GLYPHS(pRegion->frameCols, pRegion->frameWidth) +
	     pRegion->padHorz;
    pRegion->nWidthRound = DIVUP(pRegion->nWidth, 8) * 8;
    pRegion->nHeight = FRAMES_TO_GLYPHS(pRegion->frameRows, pRegion->frameHeight) +
	      pRegion->padVert;
    pRegion->nRegionSize = nRegionSize;
    pRegion->nMapSize = nMapSize;
    pRegion->mallocProc = mallocProc;
    pRegion->callocProc = callocProc;
    pRegion->freeProc = freeProc;
    pRegion->allocInfo = allocInfo;

    ERR_IGNORE(wp_initGaloisTables(pIPWState, pRegion->gexp, pRegion->glog));
    ERR_IGNORE(wp_computeGenpoly(pIPWState, KEY_PARITY_SIZE, pRegion->glog, 
				 pRegion->gexp, pRegion->logKeyGenpoly));
    ERR_IGNORE(wp_computeGenpoly(pIPWState,
				 pRegion->nCodewordParity, pRegion->glog,
		      		 pRegion->gexp, 
				 (UInt32 *)GET_LOG_DATA_GENPOLY(pRegion)));
    tempPtr = pRegion->pRandom;
    ERR_PASS(wp_generateBytes(pIPWState, pRegion, &tempPtr, MAX_CODEWORD_SIZE));
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    ERR_PASS(wp_byteSwapArray(pIPWState, pRegion->pRandom,
			      MAX_CODEWORD_SIZE_ALIGNED));
#endif

    NORMAL_RETURN;
}

/*
 * dgiDoDestroyRegion()
 *
 * Releases the structures created by dgCreateRegion for the
 * region and its data map.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Releases the structures created by wp_doCreateRegion
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  hRegion::        	region handle
 *
 * Returns:::
 *  cSuccessful		successful completion
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doDestroyRegion (
    IPWState   *pIPWState,
    HDGRegion  *hRegion)
{
    PROLOGUE;
    PDGRegion pRegion = (PDGRegion)(*hRegion);

    ERR_GEN(pRegion == NULL, cErrDG_InvalidRegion);

    GLY_FREE(pRegion->freeProc, pRegion->allocInfo, pRegion);
    *hRegion = NULL;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get information concerning the frame layout.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  hRegion::              	region handle
 *
 * Output:::
 *  pFrameWidth::          	frame width
 *  pFrameHeight::         	frame height
 *  pNumUnusedFrameBytes:: 	number of unused bytes
 *
 * Returns:::
 *  cSuccessful			successful completion
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doGetFrameInfo (
    IPWState   *pIPWState,
    HDGRegion	hRegion,
    UInt32     *pFrameWidth,
    UInt32     *pFrameHeight,
    UInt32     *pNumUnusedFrameBytes)
{
    PROLOGUE;
    PDGRegion pRegion = (PDGRegion)hRegion;

    ERR_GEN(pRegion == NULL, cErrDG_InvalidRegion);

    if (pFrameWidth != NULL)
	*pFrameWidth = pRegion->frameWidth;
    if (pFrameHeight != NULL)
	*pFrameHeight = pRegion->frameHeight;
    if (pNumUnusedFrameBytes != NULL)
	*pNumUnusedFrameBytes = pRegion->nUnusedFrameBytes;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get size in bytes of the layout handle.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  pLayout::        	layout handle
 *
 * Output:::
 *  nMapSize::       	size of data map
 *  nRegionSize::       size of fixed size region structure
 *  nEncodedBytes::  	number of encoded bytes
 *  handleSize::     	total size required by region handle
 *
 * Returns:::
 *  cSuccessful	successfully calculated size needed
 *
 * Notes:::
 * Called by clients that wish to allocate their own memory.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doGetRegionHandleSize (
    IPWState   *pIPWState,
    PDGLayout	pLayout,
    UInt32     *nMapSize,
    UInt32     *nRegionSize,
    UInt32     *nEncodedBytes,
    UInt32     *handleSize)
{
    PROLOGUE;
    UInt32 nWidth, nHeight;

    /* Create the data map */
    nWidth = FRAMES_TO_GLYPHS(pLayout->lyt.frameCols, pLayout->lyt.frameWidth) +
	     pLayout->lyt.padHorz;
    nWidth = DIVUP(nWidth, 8) * 8;
    nHeight = FRAMES_TO_GLYPHS(pLayout->lyt.frameRows,
			       pLayout->lyt.frameHeight) +
	      pLayout->lyt.padVert;

    /* the data map only really cares about being byte aligned but since
     * we're allocating it within the structure, make sure it is 32-bit
     * word aligned.
     */
    *nMapSize = DIVUP(nWidth * nHeight, 32) * 4;
    /* make sure this is word aligned since the data map and mask will 
     * follow it.
     */
    *nRegionSize = DIVUP(sizeof(DGRegion), 4) * 4;
    /* how much space will it take for the encoded data? */
    *nEncodedBytes = pLayout->lyt.frameCols * pLayout->lyt.frameRows *
		  pLayout->lyt.frameWidth * pLayout->lyt.frameHeight / 8;
    if ((pLayout->lyt.hasOcclusion) && 
	(pLayout->lyt.fOcclusion.x < pLayout->lyt.frameCols) &&
	(pLayout->lyt.fOcclusion.y < pLayout->lyt.frameRows))
	*nEncodedBytes -= pLayout->lyt.fOcclusion.w * 
		          pLayout->lyt.fOcclusion.h *
		          pLayout->lyt.frameWidth * pLayout->lyt.frameHeight / 8;

	/* The first three bytes of the generator polynomial are copied
	 * beyond the last one, so we need to increase the size of this
	 * buffer by 3.
	 */
    *handleSize = *nRegionSize + (*nMapSize * 2) + *nEncodedBytes +
		  ((pLayout->enc.nCodewordParity + 4) * sizeof(UInt32));
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get information about the occlusion (described in glyphs).
 *
 * LongDescription::: 
 * The occlusion information that the caller originally gave the layout
 * routines was based on the unit size.  Return information about the
 * occlusion in terms of glyphs.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  hRegion::			region handle
 *
 * Output:::
 *  hasOcclusion::		does this rectangle have an occlusion?
 *  x::				x offset of the upper left occluded glyph
 *  y::				y offset of the upper left occluded glyph
 *  width::			width (in glyphs) of the occluded area
 *  height::			height (in glyphs) of the occluded area
 *  isTransparent::		can glyphs be seen in the occluded area?
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doGetOcclusionInfo (
    IPWState   *pIPWState,
    HDGRegion	hRegion,
    Bool       *hasOcclusion,
    UInt32     *x,
    UInt32     *y,
    UInt32     *width,
    UInt32     *height,
    Bool       *isTransparent)
{
    PROLOGUE;
    PDGRegion pRegion = (PDGRegion)hRegion;

    ERR_GEN(pRegion == NULL, cErrDG_InvalidRegion);

    *hasOcclusion = pRegion->hasOcclusion;
    *x = pRegion->gOcclusion.x;
    *y = pRegion->gOcclusion.y;
    *width = pRegion->gOcclusion.w;
    *height = pRegion->gOcclusion.h;
    *isTransparent = pRegion->isTransparent;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get info about the region data map
 *
 * LongDescription:::
 * Returns a pointer to the data as a bit map, the mask of glyphs that 
 * should be written, and the width and height of the region.  
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  hRegion::        		region handle
 *
 * Output:::
 *  pDataMap::       		data map
 *  pDataMask::      		data mask
 *  pWidth::         		width (in glyphs) of region
 *  pHeight::        		height (in glyphs) of region
 *
 * Returns:::
 *  cErrDG_InvalidRegion	region handle is invalid
 *  cSuccessful			successful completion
 *
 * Notes:::
 * The map (and mask) are byte aligned arrays.
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Julie Craig
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doGetRegionMap (
    IPWState   *pIPWState,
    HDGRegion	hRegion,
    UInt8     **pDataMap,
    UInt8     **pDataMask,
    UInt32     *pWidth,
    UInt32     *pHeight)
{
    PROLOGUE;
    PDGRegion pRegion = (PDGRegion)hRegion;

    ERR_GEN(pRegion == NULL, cErrDG_InvalidRegion);

    *pDataMap = GET_DATA_MAP(pRegion);
    *pDataMask = GET_DATA_MASK(pRegion);
    *pWidth = pRegion->nWidth;
    *pHeight = pRegion->nHeight;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Encode the user data and generate data map and mask.
 *
 * LongDescription:::
 * Creates an encoding for the byte stream passed as a parameter, and
 * lays it out in the data map corresponding to the specified region,
 * along with the horizontal and vertical sync patterns.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  hRegion::        		region handle
 *  pData::          		data to be encoded
 *  nBytes::         		number of byte of data
 *
 * Returns:::
 *  cErrDG_InvalidRegion	invalid region handle
 *  cSuccessful			successful completion
 *
 * Notes:::
 *  data map and mask in region handle is updated
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Martin Cooper		initial version
 *  Julie Craig			occlusions
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
wp_doSetData (
    IPWState   *pIPWState,
    HDGRegion	hRegion,
    UInt8      *pData,
    UInt32	nBytes)
{
    PROLOGUE;
    PDGRegion pRegion = (PDGRegion)hRegion;

    ERR_GEN(pRegion == NULL, cErrDG_InvalidRegion);
    ERR_GEN(nBytes == 0, cErrDG_InvalidByteCount);
    /* note that instead of updating pRegion->nDataBytes to the number
     * of bytes passed into this procedure, we use it to indicate the
     * maximum number of bytes that the client could specify.  since
     * a client can call w_setData multiple times with the same region
     * handle, we can never reset nDataBytes.
     */
    ERR_GEN(nBytes > pRegion->nDataBytes, cErrDG_InvalidByteCount);

    /* Encode the data */
    ERR_PASS(wp_encodeData(pIPWState, pRegion, pData, nBytes));

    /* Zap the data map and mask */
    memset((void *)GET_DATA_MAP(pRegion), 0, (int)pRegion->nMapSize);
    /* assume that all glyhs in the rectangle should be rendered */
    memset((void *)GET_DATA_MASK(pRegion), 0xFF, (int)pRegion->nMapSize);

    /* Add the sync */
    ERR_PASS(wp_createSync(pIPWState, pRegion));
    
    /* Add the data */
    ERR_IGNORE(wp_fillFrames(pIPWState, pRegion));

    /* Fill in the pad at the right and bottom of the rectangle */
    ERR_PASS(wp_fillPad(pIPWState, pRegion));

    NORMAL_RETURN;
}

/**************************** End of file *****************************/

/*
 $Log: region.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.51  1998/07/28 21:50:46  aber
 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 CommonCheck: Warning fixes for gcc v2.8x.

 Revision 1.50  1998/01/14 00:14:59  ddavies
 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.49  1997/01/20 21:14:30  jcraig
 jcraig on Mon Jan 20 13:13:55 PST 1997

 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 use number of bytes from client's call to w_setData when recording
 number of bytes in the key codeword.
 in wp_doSetData, check number of bytes for > 0 and less than the
 (or equal) to the amount that the client indicated when making the layout.

 Revision 1.48  1996/03/21 21:53:29  ddavies
 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 Change size of generator polynomial from Int32 to UInt16's.  Also reduce
 the number of units of coeffieicnt by 2.

 Revision 1.47  1996/03/01 01:25:38  ddavies
 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 Changed all the wp_o_ names to the original wp_
 names.  This leaves the old routine names associated with the old
 routines in the grand glyph toolkit conversion.  New routine names
 will have V2 appended to them.

 Revision 1.46  1996/02/13 23:49:56  ddavies
 ddavies on Tue Feb 13 15:49:04 PST 1996

 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 Change name wp_byteSwapArray back to wp_byteSwapArray since this is not
 technically an encode routine and probably won't change in the new world
 anyway.

 Revision 1.45  1996/02/13 21:13:50  ddavies
 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 Change names of all existing encoding routines by adding o_ to their name
 prefixes.  The corresponding new routines will not have the o_.

 Revision 1.44  1996/01/04 20:26:19  ejaquith
 region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.43  1995/12/20  01:54:24  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * when filling the pad from the array of generated bytes, access them
 * as byte quantities.  when making the array of generated bytes, byte
 * swap them if this is a pc.  these 2 changes should allow the same
 * images to be generated on the pc and on the sun so that my regression
 * tests are happy campers.
 *
 * Revision 1.42  1995/12/18  23:02:24  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * instead of generating bytes each time, wp_fillPad should use the
 * 256 byte generated table in the region handle (also used by
 * wp_generateCodewords).
 * generate the bytes to put in the region handle.
 * in wp_generateBytes, if the table pointer is not NULL, use it instead
 * of allocating more memory.
 *
 * Revision 1.41  1995/12/15  00:51:51  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * wp_doGetOcclusionInfo should return the occlusion info in units of
 * glyphs not frames
 *
 * Revision 1.40  1995/12/13  04:12:27  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.39  1995/11/23  02:08:08  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: make Eric's scripts quit whining when I forgot to put :: after
 * parameter names in the comments
 *
 * Revision 1.38  1995/11/13  20:15:27  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.37  1995/11/10  00:46:39  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * add wp_doGetOcclusionInfo which does all the work for w_getOcclusionInfo
 *
 * Revision 1.36  1995/11/09  01:19:43  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: wrap ERR_IGNORE around all private procs which do not return an
 * ErrCode
 *
 * Revision 1.35  1995/11/08  22:40:56  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * wp_intRand => GLY_INT_RAND
 * change proc names from wp_ upper case letter to wp_ lower case letter
 *
 * Revision 1.34  1995/11/07  21:54:08  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * init ecc tables when region handle is init'd
 *
 * Revision 1.33  1995/10/13  22:04:38  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.32  1995/09/30  00:19:36  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Converted private-ish glyph routines to be of "wp_" form and 100% private.
 *
 * Revision 1.31  1995/09/29  23:17:10  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * PDGLAYOUT => PDGLayout.  PDGREGION => PDGRegion.
 * region handle must be NULL on entry to wp_doGetRegionHandle
 *
 * Revision 1.30  1995/09/27  23:44:47  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * new procedure comments
 *
 * Revision 1.29  1995/09/23  00:51:07  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * Removed local static routine wp_store32() to instead use same one in
 * encode.c
 *
 * Revision 1.28  1995/09/22  00:38:43  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Fixed up wp_ issues.
 *
 * Revision 1.27  1995/09/21  22:58:58  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.26  1995/09/19  00:48:14  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * in wp_doDestroyRegion, hRegion is a pointer to a handle not the handle
 * itself
 *
 * Revision 1.25  1995/09/07  01:02:33  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * HDGLAYOUT => HDGLayout. HDGREGION => HDGRegion.
 * dgi prefix becomes w_
 *
 * Revision 1.24  1995/09/06  20:22:50  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.23  1995/09/01  02:15:40  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.22  1995/08/31  21:38:39  ejaquith
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.21  1995/08/17  22:32:37  jcraig
 * region.c -> /product/ipcore/ipwrap/src/RCS/region.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 */
