/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1996, 1997 Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  regseg.c--Region segmentation: Given a background or foreground
 *  map, segment that map (which is an image with flat regions of color
 *  or gray) into regions. 
 *
 *           w_segmentRegions ()
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "arithops.h"
#include "thrshold.h"
#include "morphg.h"
#include "logicops.h"

/* PROTOTYPES */
#include "shrip.pub"
#include "shrcc.pub"
#include "shrrast.pub"
#include "morph.pub"
#include "thrshold.pub"
#include "wshed.pub"
#include "reconseq.pub"
#include "intops.pub"
#include "props.pub"
#include "strel.pub"
#include "gradient.pub"
#include "arithops.pub"
#include "regseg.pub"
#include "morphg.pub"
#include "frame.pub"
#include "logicops.pub"
#include "prxform.pub"
#include "curves.pub"
#include "fill.pub"
#include "rast.pub"
#include "rectops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: regseg.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

#define cRegionSegBorderSize	1

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the exterior edges of a connected component(s).
 *
 * LongDescription:::
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  prs::		Image containing connected component(s)
 *
 * Output:::
 *  pPrd::		Binary image containing exterior edges.  This image
 *			will contain a border.
 *
 * Notes:::
 *  The lines in pPrD are 8 connected and may be a couple of pixels wide.
 *  pPrd may be non-NULL but it is ASSUMED that it will already have a
 *  border around it if so.
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  in progress
 *
 * Authors/Modifications:::
 *  J. Craig
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL
wp_getRegionEdges(
    IPWState           *pIPWState,
    PIXR               *prs,
    PIXR              **pPrd)
{
PROLOGUE;
Int32	    	 i;
UInt8		*seedImage, *maskImage;
Int32		 seedPitch, maskPitch;	/* in bytes */
Int32	   	 imageWidth, imageHeight;
Int32	    	 fullImageWidth, fullImageHeight;
PIXR        	*prSeed = NULL, *prB = NULL;
Int32		 writeCount;
ScaledInt1616	 xDPI, yDPI;
STREL	    	*pStrel3 = NULL;

    /* first fill the area outside the connected component, up to but
     * not including the edge.  to do this construct a seed.  this is
     * an image with a 1 pixel line around the inside.
     * note that this image should have a border around it or the
     * interior will get filled if the edge of the connected component
     * touches the outside edge of the image.
     */
    ERR_PASS(w_addBorder(pIPWState, prs, cRegionSegBorderSize, 0, &prB));
    ERR_PASS_CLEANUP(w_clonePixrZero(pIPWState, prB, &prSeed));
    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, prB, &imageWidth));
    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, prB, &imageHeight));
    ERR_PASS_CLEANUP(w_pixrGetResolutionDPI (pIPWState, prB, &xDPI, &yDPI));

    /* draw the border, here's the line across the top. */
    ERR_PASS_CLEANUP(w_rasterop (pIPWState, 0, 0, NULL, /* no src */
			         PIX_SET,		/* operation */
			  	 0, 0,	        	/* x, y offsets */
			  	 imageWidth, 1,		/* width, height */
			  	 prSeed));


    /* line across the bottom */
    ERR_PASS_CLEANUP(w_rasterop (pIPWState, 0, 0, NULL, /* no src */
			  	 PIX_SET,		/* operation */
			 	 0, imageHeight-1,	/* x, y offsets */
			  	imageWidth, 1,	   	/* width, height */
			 	 prSeed));

    /* line down left side */
    ERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, NULL,  /* no src */
			  	PIX_SET,		/* operation */
			  	0, 0,	        	/* x, y offsets */
			  	1, imageHeight,	  	/* width, height */
			 	prSeed));

    /* line down right side */
    ERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, NULL,  /* no src */
			 	PIX_SET,		/* operation */
			  	imageWidth-1, 0,	/* x, y offsets */
			  	1, imageHeight,	   	/* width, height */
			  	prSeed));

    /* generate the mask image we'll actually use.  This is the inverse
     * of the mask we're given. */
    ERR_PASS_CLEANUP(w_invert(pIPWState, prB, &prB));

    /* prepare args for i_fillClipSeedTwoPass */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prSeed, &seedImage));
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prB, &maskImage));
    ERR_PASS_CLEANUP(w_pixrFramedWidth(pIPWState, prB, &fullImageWidth));
    ERR_PASS_CLEANUP(w_pixrFramedHeight(pIPWState, prB, &fullImageHeight));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prSeed, &seedPitch));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prB, &maskPitch));

    /* do the first pass, initially clipping the seed with the mask */
    ERR_PASS_CLEANUP(i_fillClipSeedTwoPass (pIPWState, seedImage,
					    seedImage, maskImage,
				            seedPitch, seedPitch, maskPitch,
				            fullImageWidth, fullImageHeight, 
				            cTrue, &writeCount));

    /* do remaining cycles with no initial clip */
    for (i = 1; (i < 25) && (writeCount > 0); i++)	
    {
	ERR_PASS_CLEANUP(i_fillClipSeedTwoPass (pIPWState, seedImage,
						seedImage, 
						maskImage, seedPitch,
						seedPitch, maskPitch,
						fullImageWidth,
						fullImageHeight, cFalse,
						&writeCount));
    }

    /* restore original */
    ERR_PASS_CLEANUP(w_invert(pIPWState, prB, &prB));

    /* now extend the mask to include the edge of the connected component */
    ERR_PASS_CLEANUP(w_initStrelByName(pIPWState, strel3));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel3, &pStrel3));
    ERR_PASS_CLEANUP(w_dilate(pIPWState, prSeed, pStrel3, pPrd));

    ERR_PASS_CLEANUP(w_logicOp(pIPWState, prB, AND, *pPrd));

    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prSeed));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prB));
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prSeed);
    w_destroyPixr(pIPWState, &prB);
    ERR_RETURN;

}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Given the watershed and gradient, re-segment into fewer regions.
 *
 * LongDescription:::
 * The watershed frequently results in an over-segmented image.  This
 * routine takes the watershed and the gradient image and attempts to 
 * re-segment the image into fewer regions.
 *  color.
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  prG::		Gradient image
 *  prW::		Watershed regions
 *
 * Output:::
 *  pprd::		Binary image containing black dividing lines
 *			between regions
 *
 * Notes:::
 *  The lines in pPrD are 8 connected and may be a couple of pixels wide.
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  in progress
 *
 * Authors/Modifications:::
 *  J. Craig
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL
wp_regionsByThreshold(
    IPWState		*pIPWState,
    PIXR		*prG,
    PIXR		*prW,
    PIXR		**pprd)
{
PROLOGUE;
Int32		 allMax, edgeMax, threshValue;
Int32		 frame, imageWidth, imageHeight, sBpl, dBpl;
Int32		*pHisto = NULL;
PIXR		*prEdge = NULL, *prF = NULL, *prT = NULL;
PIXR		*prR = NULL, *prC = NULL, *prI = NULL, *prCI = NULL;
RECTANGL	*rects = NULL, *currR, *iRect = NULL, *currI;
STREL		*pStrel5 = NULL;
ScaledInt1616	 xdpi, ydpi;
UInt8		*sPtr, *dPtr;

    /* this code will attempt to reduce the number of segments by choosing
     * only those with "strong" edges.  we'll do this by thresholding
     * the gradient image and removing any edges which are not connected.
     * first, determine a threshold value by looking at the histogram
     * of the gradient using edges from the watershed.  next threshold.
     * and finally, remove unconnected edges.
     */

    /* find the outer edge of each of the connected components in the
     * watershed.
     */
    ERR_PASS_CLEANUP(wp_getRegionEdges(pIPWState, prW, &prEdge));
    /* "remove" the border without actually creating yet another pixr */
    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, prEdge, &imageWidth));
    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, prEdge, &imageHeight));
    ERR_PASS_CLEANUP(w_rasterop(pIPWState, cRegionSegBorderSize,  
	       			cRegionSegBorderSize, prEdge, PIX_SRC, 0, 
				0, imageWidth - 2*cRegionSegBorderSize,
	       			imageHeight - 2*cRegionSegBorderSize,
				prEdge));
    imageHeight -= 2*cRegionSegBorderSize;
    imageWidth -= 2*cRegionSegBorderSize;
    ERR_PASS_CLEANUP(w_pixrSetHeight(pIPWState, prEdge, imageHeight));
    ERR_PASS_CLEANUP(w_pixrSetWidth(pIPWState, prEdge, imageWidth));

    /* look at the values of the gradient which correspond to the edges
     * of the connected components.  these majority of these edges 
     * (especially if the component is large enough) should remain.
     * there are frequently many extra edges in the interior of these
     * regions.  we cannot just eliminate the interior edges since many of
     * them are actually needed because we have adjacent regions.
     */
    ERR_PASS_CLEANUP(w_histoMask(pIPWState, prG, prEdge, NULL, &pHisto));

    /* find the most frequently occurring value */
    ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, pHisto, 0, 255,
					 &edgeMax));
    IPW_FREE(pHisto, cAllocatorData);

    /* now look at all the edges in the watershed */
    ERR_PASS_CLEANUP(w_histoMask(pIPWState, prG, prW, NULL, &pHisto));
    /* and find the maximum between 0 and the edge-only maximum */
    ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, pHisto, 0, edgeMax-1, 
					 &allMax));

    /* starting at the all-edges maxiumum, look for a minimum.  using
     * the all-edges maximum will probably get too many edges since 
     * watershed was over-segmented.  so look for a minimum between
     * the 2 maximums.  hopefully, this will give a reasonable 
     * threshold value.
     */
    ERR_PASS_CLEANUP(w_findLocalMinimum(pIPWState, pHisto, 255, allMax, 
		           		edgeMax, 1, 2, &threshValue));

    IPW_FREE(pHisto, cAllocatorData);

    /* get the edges which are above the threshold */
    ERR_PASS_CLEANUP(w_threshold(pIPWState, prG, 0, threshValue,  
			         THRESH_INVERSE, &prT));
    ERR_PASS_CLEANUP(w_logicOp(pIPWState, prEdge, OR, prT));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prEdge));

    /* this images will have problems.  it may have disconnected edges.
     * regions may be adjacent.  there may be some black blobs that
     * have no interior.
     */

    /* NOTE.  this code below removes unconnected edges from the inside
     * and the outside of a connected component.  if what we are
     * trying to do is use the contour finding code to get polygon
     * points, we probably don't really need to remove the interior
     * edges.  this will, of course, speed up the code somewhat.
     */

    /* add a one pixel wide black border around the image.  this will 
     * allow us to find things like platen noise around the edge as
     * an enclosed connected component.
     */
    ERR_PASS_CLEANUP(w_addBorder(pIPWState, prT, cRegionSegBorderSize, 1,
				 &prEdge));
    /* TODO.  this and a rabb below use a minWidth and minHeight of 3.
     * unfortunately, this means that the width AND the height must both
     * be at least 3.  not what I had in mind since a long thin 1 pixel
     * tall line will not be noticed and this is probably incorrect.
     */
    ERR_PASS_CLEANUP(w_rabb(pIPWState, prEdge, 3, 3, &rects));

    ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, prEdge, &xdpi,
					    &ydpi));
    ERR_PASS_CLEANUP(w_pixrGetFrame(pIPWState, prEdge, &frame));
    if (*pprd == NULL)
    {
        ERR_PASS_CLEANUP(w_clonePixrZero(pIPWState, prT, &prF));
    }
    else
    {
	prF = *pprd;
	ERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, NULL, 
			     	    PIX_CLR, 0, 0, 0, 0, prF));
    }
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, prF, &dPtr));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prF, &dBpl));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prT));

    ERR_PASS_CLEANUP(w_initStrelByName(pIPWState, strel5));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel5, &pStrel5));

    for (currR = rects; currR != NULL; currR = currR->next)
    {
	/* TODO.  this rectangle needs to be filtered against the
	 * other rects on the list.  if this rectangle is contained 
	 * within another rectangle because it is actually inside the
	 * other connected component it will be processed as part of the
	 * encompassing rectangle.  however, if the rectangles are
	 * contained but the connected component is not, we still need
	 * to process the rectangle.  if the rectangles overlap, we
	 * still need to process the rectangle.  
	 * this is a speed issue that I'm not sure how to solve at
	 * this point.
	 * actually, as long as the rectangle don't connect, they 
	 * won't be in this list.  if they are connected (and hence
	 * cause adjacent regions), they will be processed in the
	 * code below.
	 * I don't understand the last paragraph.  we are still double
	 * processing some rectangles.
	 */


	/* we want to make a copy of this connected component and
	 * invert it.  unfortunately, we cannot just do the obvious
	 * rasterop.  the connected component is described as a 
 	 * rectangle and so may include other components either
	 * partial or full inside it.  as a result, we must copy
	 * the component, seed and fill it before inverting it.
	 */
	ERR_PASS_CLEANUP(w_createPixr(pIPWState, currR->w, currR->h, 1,
			              xdpi, ydpi, frame, &prC));
	ERR_PASS_CLEANUP(w_createPixrNoInit(pIPWState, currR->w, currR->h,
				            1, xdpi, ydpi, frame, &prR));
	ERR_PASS_CLEANUP(w_rasterop(pIPWState, currR->x, currR->y, prEdge, 
			     	    PIX_SRC, 0, 0, currR->w, currR->h,
			    	    prR));
	/* adjust the point now that we've clipped the rectangle. */
	ERR_PASS_CLEANUP(w_pixrPutPixel(pIPWState, prC, 
					currR->px - currR->x,
			         	currR->py - currR->y, 1));
	ERR_PASS_CLEANUP(w_fillClipSweep8(pIPWState, prC, prR, 250, &prC));
	ERR_PASS_CLEANUP(w_invert(pIPWState, prC, &prCI));
	
	/* now determine if the interior belonged to one region or
	 * if there were multiple, adjacent regions.
	 */
	ERR_PASS_CLEANUP(w_rabb4Connected(pIPWState, prCI, 3, 3, &iRect));

	/* if there was more than one region in the connected component,
         * these regions may overlap with each other, may contained
	 * within another (to an arbitrary depth) or may just be
	 * adjacent.  we need the polygon points for this region
	 * (and it may share an edge with an adjacent region)
	 * and we need the mask to use to get the color of the
	 * region.
	 * the one thing that these regions share with each
	 * other is that their edges are connected at some point
	 * (maybe for only a point).  thus these regions are 
	 * adjacent.  regions which are contained within another
	 * (to an arbitrary depth) will be found by the original
	 * connected component analysis.
	 * this code should also work if there are no components
	 * or only one.
	 */

	for (currI = iRect; currI != NULL; currI = currI->next)
	{
	/* check to make sure that this is an interior region.
	 * since stray edge may protrude from the connected
	 * component, inverting the image gets no only the 
	 * interior of the connected component but also the
	 * region outside the component since the component 
	 * doesn't touch the edges.  we only want to look at
	 * the interior so we'll try to filter the rectangles
	 * based on the coordinates.  since the original
	 * rectangle contained the connected component, we
	 * are guaranteed that an interior of the component
	 * does not hit one of the edges.  so, check to see
	 * if this subrectangle touches any of the edges.  if
	 * it does, ignore it since it isn't a rectangle that
	 * we wish to process.
	 */
	    if ((currI->x == 0) || (currI->y == 0) ||
		((currI->x + currI->w) >= currR->w) ||
		((currI->y + currI->h) >= currR->h))
		continue;

	    /* first get the interior of only the current component. */
	    ERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, prR, PIX_CLR, 0, 
			                0, currR->w, currR->h, prR));
	    ERR_PASS_CLEANUP(w_pixrPutPixel(pIPWState, prR, currI->px,
					    currI->py, 1));
	    ERR_PASS_CLEANUP(w_fillClipSweep4(pIPWState, prR, prCI, 250, 
					      &prR));

	    /* now dilate this slightly to get the edges (which we 
	     * assume are not more than 2 pixels wide).  this may also 
	     * leave small bumps along the edge which come from other
	     * unconnected edges.  these could probably be filtered 
	     * later by looking at the polygon points.
	     */
	    ERR_PASS_CLEANUP(w_dilate(pIPWState, prR, pStrel5, &prT));
	    /* get the edges of the region */
	    ERR_PASS_CLEANUP(w_logicOp(pIPWState, prC, AND, prT));

	    /* now get edges to get rid of unconnected edges on the
	     * inside.
	     */
	    ERR_PASS_CLEANUP(wp_getRegionEdges(pIPWState, prT, &prI));

	    /* and place into a composite image.  
	     * note that we are relying on w_rasterop to clip if
	     * necessary.  prF does not have a border.  however, 
	     * currR was generated from an image that had a border.
	     * as a result, we need to adjust the (x, y) coordinates.
	     * if this was an interior component the width and height
	     * are correct, just the upper left corner is incorrect.
	     * if this connect component was the result of the border
	     * that we placed around the image, it will be at (0, 0)
	     * and the width and height will be incorrect.  so let 
	     * w_rasterop do clipping.  note also that prI had a
	     * border added by wp_getRegionEdges so we have to take
	     * that into account as well.
	     * why are we calling i_rasterOpFull directly instead of
	     * using w_rasterop?  well, since w_rasterop calls into
	     * i_rasterOpFull pointing to the upper left corner of the
	     * frame, clipping does not happen on the right and 
	     * bottom until the rectangles goes past the frame,  
	     * causing this code to write into the frame.  so, call
	     * i_rasterOpFull directly, pointing at the upper left
	     * corner of the image.
	     */
	    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, prI, &sPtr));
	    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prI, &sBpl));
	    ERR_PASS_CLEANUP(
		i_rasterOpFull (pIPWState, dPtr, imageWidth, imageHeight, 0, 0,
			        currR->x - cRegionSegBorderSize,
			        currR->y - cRegionSegBorderSize,
			        currR->w, currR->h, 1, dBpl, 
			        PIX_SRC | PIX_DST, sPtr, 
			        currR->w + 2*cRegionSegBorderSize,
			        currR->h + 2*cRegionSegBorderSize,
			        cRegionSegBorderSize, cRegionSegBorderSize,
			        0, 0, 1, sBpl));

	}
	ERR_PASS_CLEANUP(w_destroyRectList(pIPWState, &iRect));
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prC));
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prR));
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prCI));
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prT));
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prI));
    }

    ERR_PASS_CLEANUP(w_destroyRectList(pIPWState, &rects));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prEdge));
    *pprd = prF;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (*pprd == NULL) w_destroyPixr(pIPWState, &prF);
    w_destroyRectList(pIPWState, &iRect);
    w_destroyRectList(pIPWState, &rects);
    w_destroyPixr(pIPWState, &prEdge);
    w_destroyPixr(pIPWState, &prC);
    w_destroyPixr(pIPWState, &prR);
    w_destroyPixr(pIPWState, &prCI);
    w_destroyPixr(pIPWState, &prT);
    w_destroyPixr(pIPWState, &prI);
    ERR_RETURN;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Segment an image with flat colored regions
 *
 * LongDescription:::
 *  The "map ink and page" option of the autoscan pipeline produces
 *  a "background" or "foreground" map, which is a map of what color
 *  to put down at a given pixel, controlled by a high-resolution selector
 *  plane. This routine segments one of those maps, so that one can
 *  represent flat regions by some geometric description and a single 
 *  color.
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  prS::		Image to segment
 *
 * Output:::
 *  pPrD::		Binary image containing black dividing lines
 *			between regions
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  in progress
 *
 * Authors/Modifications:::
 *  J. Craig
 *  P. MacLeod
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_segmentRegions (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
PROLOGUE;
PIXR 		*prGrad = NULL, *prTmp = NULL, *pWShed = NULL;
PIXR		*prEH = NULL, *prEV = NULL, *prT = NULL;
STREL 		*pStrel10h = NULL, *pStrel10v = NULL, *pStrel2 = NULL;
    
    /* Compute the gradient fo that image using an L-inf distance */
    ERR_PASS(w_findMagnitudeOfGradient(pIPWState, prS, 
				       cGradOpMaxLInfFromCenter, 
				       cGradNeighborhoodDownRight, 
				       &prGrad));			       
    /* This code used to add some constant and perform dual 
     * reconstruction.  Then, it added one, performed another dual
     * reconstruction and found the difference in order to extract the
     * minima.  
     * However, this was kinda time consuming so we'll attempt to find
     * the minima by doing a grayscale erosion and then pixel by pixel
     * thresholding with the results of the erosion.
     */
    ERR_PASS_CLEANUP(w_copy(pIPWState, prGrad, &prTmp));
    ERR_PASS_CLEANUP(w_arithOpConst(pIPWState, cAdd, 10.0, prTmp));

    ERR_PASS_CLEANUP(w_initStrelByName (pIPWState, strel10h));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel10h, &pStrel10h));
    ERR_PASS_CLEANUP(w_initStrelByName (pIPWState, strel10v));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel10v, &pStrel10v));

    ERR_PASS_CLEANUP(w_erodeGray(pIPWState, prTmp, pStrel10h, &prEH));
    ERR_PASS_CLEANUP(w_erodeGray(pIPWState, prTmp, pStrel10v, &prEV));
    ERR_PASS_CLEANUP(w_min(pIPWState, prEH, prEV, &prTmp));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prEH));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prEV));

    ERR_PASS_CLEANUP(w_thresholdByPixel(pIPWState, prGrad, prTmp,  
					THRESH_NORM, &prT));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prTmp));

    /* get rid of small black spots that will give us very small 
     * regions.
     */
    ERR_PASS_CLEANUP(w_initStrelByName(pIPWState, strel2));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel2, &pStrel2));
    ERR_PASS_CLEANUP(w_open(pIPWState, prT, pStrel2, &prT));

    /* 
     * Do the watershed transform, using prGrad as the image, and
     * prT as the marker set. This gives us a catchment basin
     * around each minimum, where the crest lines divide the watershed
     * regions.
     */
    ERR_PASS_CLEANUP(w_watershedSegment(pIPWState, prGrad, prT, 0, 
					&pWShed));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prT));

    ERR_PASS_CLEANUP(wp_regionsByThreshold(pIPWState, prGrad, pWShed,
					   pPrD));

    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prGrad));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pWShed));
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prGrad);
    w_destroyPixr(pIPWState, &pWShed);
    w_destroyPixr(pIPWState, &prTmp);
    w_destroyPixr(pIPWState, &prEH);
    w_destroyPixr(pIPWState, &prEV);
    w_destroyPixr(pIPWState, &prT);
    ERR_RETURN;
}


/*

 $Log: regseg.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.4  1998/01/14 00:15:02  ddavies
 regseg.c -> /product/ipcore/ipwrap/src/RCS/regseg.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.3  1997/07/21 17:56:41  jcraig
 jcraig on Mon Jul 21 10:56:05 PDT 1997

 regseg.c -> /product/ipcore/ipwrap/src/RCS/regseg.c,v

 fix nag script warnings

 Revision 1.2  1997/07/19 01:31:16  jcraig
 jcraig on Fri Jul 18 18:28:17 PDT 1997

 regseg.c -> /product/ipcore/ipwrap/src/RCS/regseg.c,v

 snapshot of w_segmentRegions (to segment background plane into
 regions).  this code is a work-in-progress but I needed a stable base
 before continuing development.

 Revision 1.1  1997/01/08 22:24:52  macleod
 Initial revision


 */

