/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  rescale1.c--procedures that rescale an image (and shift data)
 *
 *     Arbitrary scaling:
 *           w_scaleToSize ()
 *	     w_scaleLinearSize ()
 *           w_scaleMaskedArea()
 */

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "cmask.pub"
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrcc.pub"
#include "imageref.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "rescale.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: rescale1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arbitrary rescaling of single or multi-channel, 1, 4, or 8 bit PIXRs.
 *
 * LongDescription:::
 *  Rescales the source PIXR to newW wide and newH high destination
 *  PIXR.  Destination PIXR is created if necessary.  Uses subsampling
 *  and/or supersampling for speed, sacrificing some quality.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be scaled
 *  newW::           Destination width
 *  newH::           Destination height
 *
 * Output:::
 *  prd::            Scaled destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If newW = 0 and newH = 0 a 1x1 PIXR will result.
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleToSize (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               newW,
    Int32               newH,
    PIXR              **prd)
{
    PROLOGUE;
    Int32             wS, hS, frame;
    Int32             depthS;
    Bool              destCreated = cFalse;
    Int32	      channels, channel;
    Int32        rfc, bfc;
    ScaledInt1616	      xRes, yRes;
    ColorMap	     *map;
    ColorType	      type;
    PIXR             *subS, *subD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
		      cChannelsMask, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prs, &depthS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prs, &channels));
    ERR_PASS (w_pixrGetFrame (pIPWState, prs, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prs, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prs, &type));

    /* Rescale the DPI fields */
    xRes = Int32_TO_ScaledInt1616(ScaledInt1616_TO_Int32(xRes) * newW / wS);
    yRes = Int32_TO_ScaledInt1616(ScaledInt1616_TO_Int32(yRes) * newH / hS);
    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
                                               newW, newH, depthS,
				               xRes, yRes,
				               frame, channels, type, map,
					       prd, &rfc, &bfc, &destCreated));

    /* vanishingly small */
    if ((newW < 1) || (newH < 1))
	NORMAL_RETURN;

    /* Same size */
    if ((newW == wS) && (newH == hS))
    {
	ERR_PASS_CLEANUP (w_copy (pIPWState, prs, prd));
    }
    else
    {
	for (channel = 0; channel < channels; channel++)
	{
	    Int32        wD, hD, bplS, bplD;
	    UInt8       *iS, *iD;
    
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, channel,
						&subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, channel,
						&subD));
    
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subD, &wD));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wS));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subD, &hD));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));
    
	    ERR_PASS_CLEANUP (i_rescaleNoInterp (pIPWState, iS, wS, hS,
						 bplS, depthS,
						 iD, wD, hD, bplD, depthS));
	}
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_scaleToSize () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Creates and initializes arbitrary sub/super sampling tables
 *
 * LongDescription:::
 *	The tables can then be passed in to w_scaleToSizeRect() multiple
 *	times for strip at a time operation.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		State pointer
 *  prS::		Source PIXR to be rescaled
 *  destW::		The width the entire source image would have if it were
 *			scaled by the desired scale factor.
 *  destH::		The width the entire source image would have if it were
 *			scaled by the desired scale factor.
 *
 * Output:::
 *  ppTable::		Pointer to pointer to returned table.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeScaleToSizeTables (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               destW,
    Int32               destH,
    void              **ppTable)
{
    PROLOGUE;
    Int32         srcW, srcH, d, bpl,
		  tableSize;
    Bool	  allocatedHere;

    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &srcW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &srcH));
    ERR_PASS (w_pixrGetDepth (pIPWState, prS, &d));
    ERR_PASS (w_pixrGetBpl (pIPWState, prS, &bpl));

    ERR_PASS (i_getScaleToSizeTableSize (pIPWState, d, destW, destH, 
                                         &tableSize));

    /* Allocate the index tables */
    /* each entry in the array indicates which source pixel (offset
     * from the left side of the line) to get that destination
     * pixel.
     */
    allocatedHere = cFalse;
    if ((*ppTable = (Int32 *)IPW_MALLOC(tableSize, 
				        cAllocatorData)) == NULL)
	ERR_GEN_CLEANUP (cTrue, cErrNomem);
    allocatedHere = cTrue;

    ERR_PASS_CLEANUP (i_initScaleToSizeTables (pIPWState, srcW, srcH, bpl,
                                               d, destW, destH, 
					       tableSize, *ppTable));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
        IPW_FREE (*ppTable, cAllocatorData);

    ERR_RETURN;

}  /* w_makeScaleToSizeTables () */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Destroys tables made by w_makeScaleToSizeTables ()
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		State pointer
 *
 * Output:::
 *  ppTable::		Pointer to pointer to composite scaling table.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyScaleToSizeTables (
    IPWState           *pIPWState,
    void              **ppTable)
{
    PROLOGUE;

    if ((*ppTable) != NULL)
        IPW_FREE (*ppTable, cAllocatorData);

    *ppTable = NULL;

    NORMAL_RETURN;

}  /* w_destroyScaleToSizeTables () */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arbitrary rescaling of single or multi-channel, 1, 4, or 8 bit PIXRs.
 *
 * LongDescription:::
 *  Rescales a portion of the source PIXR to newW wide and newH
 *  high destination PIXR.  Destination PIXR is created if necessary.
 *  Uses subsampling and/or supersampling for speed, sacrificing some quality.
 *  <BR><BR>
 *  This routine takes pointers to an entire source image.  A rectangle
 *  of this source image will be scaled.  One might like to specify
 *  the coordinates of this rectangle relative to the source image,
 *  but the client can't know the exact coordinates in the source
 *  space since they depend on the vagaries of the scaling function.
 *  That is, the choice of exactly which source pixel will be chosen for
 *  a given dest pixel is not public information.
 *  <BR><BR>
 *  Therefore, we adapt the strategy used in arbitrary scaling to gray.
 *  The client supplies an entire source image and a dest image that's
 *  just large enough to hold the scaled rectangle.  The coordinates
 *  of the rectangle are given relative to a virtual destination image,
 *  i.e., the image one would get if one scaled the entire source
 *  image.
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prS::            Source PIXR to be scaled
 *  roi::            Subrect of virtual dest image which which contains prd.
 *		     Must be non-NULL.  The width and height of this rectangle
 *		     must equal the width and height of *pPrd, if it's provided.
 *  pTable::         Pointer to scaling tables make with
 *		     w_makeScaleToSizeTables.  Must be non-NULL.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to scaled destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If newW = 0 and newH = 0 a 1x1 PIXR will result.
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, anamorphic, isomorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleToSizeRect (
    IPWState           *pIPWState,
    PIXR               *prS,
    RECTANGL           *roi,
    void               *pTable,
    PIXR              **pPrD)
{
    PROLOGUE;
    Int32             wS, hS, frame;
    Int32             depthS;
    Int32	      wVirtD, hVirtD, wD, hD;
    Bool              destCreated = cFalse;
    Int32	      channels, channel;
    Int32             rfc, bfc;
    ScaledInt1616     xRes, yRes;
    ColorMap	     *map;
    ColorType	      type;
    PIXR             *subS, *subD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);

    ERR_GEN ((pTable == NULL) || (roi == NULL), cErrNullPtr);

    /* is dest vanishingly small? */
    if ((roi->w < 1) || (roi->h < 1))
	NORMAL_RETURN;

    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prS, &depthS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prS, &xRes, &yRes));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prS, &channels));
    ERR_PASS (w_pixrGetFrame (pIPWState, prS, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prS, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prS, &type));
    ERR_PASS (i_getDimensionsFromScalingTables (pIPWState, pTable,
						&wVirtD, &hVirtD));

    /* make sure the rect fits inside the virtual dest image */
    ERR_GEN (roi->x < 0, cErrInvalidParm);
    ERR_GEN (roi->y < 0, cErrInvalidParm);
    ERR_GEN ((roi->x + roi->w) > wVirtD, cErrInvalidParm);
    ERR_GEN ((roi->y + roi->h) > hVirtD, cErrInvalidParm);

    /* Rescale the DPI fields */
    xRes = Int32_TO_ScaledInt1616(ScaledInt1616_TO_Int32(xRes) * wVirtD / wS);
    yRes = Int32_TO_ScaledInt1616(ScaledInt1616_TO_Int32(yRes) * hVirtD / hS);

    /* make the dest image if necessary */
    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
						      roi->w, roi->h, depthS,
						      xRes, yRes,
						      frame, channels,
						      type, map,
						      pPrD, &rfc, &bfc,
						      &destCreated));

	/* return an error if the rectangle's width and height
	 * don't match the dest images's width and height.
	 * We do this here because we're guaranteed to have a
	 * dest image here.
	 */
    ERR_PASS (w_pixrGetWidth (pIPWState, *pPrD, &wD));
    ERR_PASS (w_pixrGetHeight (pIPWState, *pPrD, &hD));
    ERR_GEN_CLEANUP ((wD != roi->w) || (hD != roi->h), cErrInvalidParm);


    /* scale each channel */
    for (channel = 0; channel < channels; channel++)
    {
	Int32        bplS, bplD;
	UInt8       *iS, *iD;

	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prS, channel, &subS));
        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *pPrD, channel, &subD));

	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wS));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

        ERR_PASS_CLEANUP (i_rescaleNoInterpSubRegion (pIPWState, iS, wS, hS, 
	                                              bplS, depthS,
						      pTable,
	                                              iD, roi->x, roi->y,
						      roi->w + 2*frame,
						      roi->h + 2*frame, 
						      bplD, depthS));
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *pPrD, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, pPrD);
    ERR_RETURN;

}  /* w_scaleToSizeRect () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arbitrary, interpolated scaling of color or gray, 4, or 8 bit PIXRs.
 *
 * LongDescription:::
 *  Rescales the source PIXR to newW wide and newH high destination
 *  PIXR.  Destination PIXR is created if necessary.  Uses linear
 *  weighted interpolation for quality, scarificing speed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be scaled
 *  newW::           Destination width
 *  newH::           Destination height
 *
 * Output:::
 *  prd::            Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, isomorphic, anamorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleLinearSize (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               newW,
    Int32               newH,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, wD, hS, hD, 
                 bplS, bplD, frame;
    Int32        depthS;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    ScaledInt1616 xRes, yRes;
    UInt8       *iS, *iD;
    PIXR        *subS, *subD;
    Int32        channels, channel;
    ColorMap	*map;
    ColorType	 type;


    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth4 | cDepth8, 
		      cChannelsMask, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prs, &depthS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prs, &channels));
    ERR_PASS (w_pixrGetFrame (pIPWState, prs, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prs, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prs, &type));

    /* Rescale the DPI fields */
    xRes = Int32_TO_ScaledInt1616(ScaledInt1616_TO_Int32(xRes) * newW / wS);
    yRes = Int32_TO_ScaledInt1616(ScaledInt1616_TO_Int32(yRes) * newH / hS);
    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
                                               newW, newH, depthS,
				               xRes, yRes,
				               frame, channels, type, map,
					       prd, &rfc, &bfc, &destCreated));

    /* vanishingly small */
    if ((newW < 1) || (newH < 1))
	NORMAL_RETURN;

    if ((newW == wS) && (newH == hS))
    {
	ERR_PASS_CLEANUP (w_copy (pIPWState, prs, prd));
    }
    else
    if (channels > 1)
    {
	for (channel = 0; channel < channels; channel++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, 
	                                        channel, &subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, 
	                                        channel, &subD));
    
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, subD, &wD));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, subD, &hD));
    
	    ERR_PASS_CLEANUP (w_scaleLinearSize (pIPWState, subS, 
	                                         wD, hD, &subD));
	}
    }
    else
    {
	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prs, &wS));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prs, &hS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

	ERR_PASS_CLEANUP (i_rescaleArb (pIPWState, iS, wS, hS, bplS, depthS,
                                        iD, wD, hD, bplD, depthS));
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_scaleLinearSize () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Scale masked area of gray scale or color image. 
 *
 * LongDescription:::
 *	Given a binary mask image and a color or gray scale image, 
 * scale the portion of the color image that's covered by the mask.
 * The pixels outside the mask will not be used in scale  processing and
 * set zero value in scaled image.  The binary and color images must
 * have the same resolution and size.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure 
 *  pBinImage::      Binary image ( thresholded from color image )
 *  pColorImage::    Gray scale or color image, depth is 8 
 *  scaleX::         Horizontal scale fraction
 *  scaleY::         Vertical scale fration
 *  yOffset::        Output image y - offset
 *
 * Output:::
 *  prScaled::       Scaled color or gray image ( must be initialized )
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  colored text
 *
 * Keywords:::
 *  colored text
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleMaskedArea (
    IPWState               *pIPWState,
    PIXR                   *pBinImage,
    PIXR                   *pColorImage,
    Float32                scaleX,
    Float32                scaleY,
    Int32                  yOffset,
    PIXR                   *prScaled)
{
    PROLOGUE;
    Run32        *runBuffer = NULL, *dummyRun, *prun;
    Int32        runBuffSize, width, height, depth;
    UInt8	 *pScaled[3], *pBin, *pColor[3], *srcImage;
    UInt8	 *pScaledPlane, *pColorPlane;
    PIXR         *prPlane;
    Int32        channels, bpl, cbpl, cha, dbpl;
    Int32        row, col, y, scaledOffset, len;
    Int32        scaledH, scaledW;
    Int32        *yIdx, *xIdx = NULL;
    UInt8        value, minValue;

    ERR_GEN_PIXR (pBinImage, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR (pColorImage, cByteOrderBig, cDepth8, cChannels1|cChannels3,
		  cFrame32);
    ERR_GEN (prScaled == NULL, cErrNullPtr);
    ERR_PASS (w_pixrGetWidth (pIPWState, pBinImage, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, pBinImage, &height));
    ERR_PASS (w_pixrGetDepth(pIPWState, pBinImage, &depth));
    ERR_PASS (w_pixrGetImage (pIPWState, pBinImage, &pBin));
    ERR_PASS (w_pixrGetBpl (pIPWState, pBinImage, &bpl));

    runBuffSize = width + 3;
    runBuffer = (Run32 *)IPW_MALLOC (runBuffSize * sizeof (Run32), 
				     cAllocatorData);
    ERR_GEN (runBuffer == NULL, cErrNomem);
    xIdx = (Int32 *)IPW_MALLOC ((width + height) * sizeof (Int32), 
				cAllocatorData);
    ERR_GEN (xIdx == NULL, cErrNomem);
    yIdx = xIdx + width;

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prScaled, &scaledW));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prScaled, &scaledH));

    for( row=0; row<height; ++row)
    {
	yIdx[row] = (Int32)((Float32)(row + yOffset) * scaleY + 0.5);
	if( yIdx[row] >= scaledH )
	    yIdx[row] = scaledH -1;
    }
    for( col=0; col<width; ++col)
    {
	    /* If this is a big endian machine, compute the scaled
	     * offset in the straightforward way.
	     * If this is a little endian machine, invert the lower
	     * 2 bits of the index.  We know that the pointers to the
	     * beginnings of lines always point to a 4-byte boundary.
	     * Inverting the LS two bits here means we can avoid using
	     * UCHAR_ACCESS (which slows down a poor PC).
	     */
	xIdx[col] = (Int32)((Float32)col * scaleX);
	if( xIdx[col] >= scaledW )
	    xIdx[col] = scaledW -1;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	xIdx[col] ^= 0x3;
#endif
    }
    
    scaledOffset = yIdx[0];
    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, pColorImage, 
					     &channels));
    for(cha=0; cha<channels; ++cha)
    {
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, pColorImage, cha, 
					    &prPlane));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pColor[cha]));
	if(cha == 0)
	{
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prPlane, &cbpl));
	}
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prScaled, cha, 
					    &prPlane));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pScaled[cha]));
	if(cha == 0)
	{
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prPlane, &dbpl));
	}
	pScaled[cha] += scaledOffset * dbpl;
    }

    srcImage = pBin;
    len  = width *3 *sizeof(UInt8);
    for(row=0; row<height; )
    {
	y = yIdx[row];
	for( ; row < height && yIdx[row] == y; ++row)
	{
	    ERR_PASS_CLEANUP (i_findAllRunsInLineArray (pIPWState, srcImage,
							width,
							0,
							runBuffer,
							runBuffSize,
							0,  
							&dummyRun));
	    for(cha=0; cha<channels; ++cha)
	    {
		prun = runBuffer;
		pScaledPlane = pScaled[cha];
		pColorPlane = pColor[cha];
		while(prun != dummyRun)
		{
		    for(col= prun->runLeft; col<= prun->runRight; ++col )
		    {
			value = MAX(UCHAR_ACCESS(pColorPlane+col), 2);
			minValue = *(pScaledPlane+xIdx[col]);
			if( minValue > value || minValue < 2)
			    *(pScaledPlane+xIdx[col]) = value;
		    }
		    ++ prun;
		}
	    }
	    srcImage += bpl;
	    for(cha=0; cha<channels; ++cha)
		pColor[cha] += cbpl;
	}
	for(cha=0; cha<channels; ++cha)
	{
	    pScaled[cha] += dbpl;
	}
    }


    IPW_FREE(runBuffer, cAllocatorData);
    IPW_FREE(xIdx, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:

    if(runBuffer != NULL) IPW_FREE(runBuffer, cAllocatorData);
    if(xIdx != NULL) IPW_FREE(xIdx, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Scale masked area of color image. 
 *
 * LongDescription:::
 *	Given a binary mask image and a color image, scale the portion
 * of the color image that's covered by the mask.
 * The pixels outside the mask will not be used in scale  processing and
 * set zero value in scaled image.  The binary and color images must
 * have the same resolution and size.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure 
 *  pBinImage::      Binary image ( thresholded from color image )
 *  pIndex::         Color index array, depth is 16 
 *  scaleX::         Horizontal scale fraction
 *  scaleY::         Vertical scale fration
 *  yOffset::        Output image y - offset
 *  pColorTab::      Color index table
 *
 * Output:::
 *  prScaled::       Scaled color or gray image ( must be initialized )
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  colored text
 *
 * Keywords:::
 *  colored text
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleMaskedIndexArea (
    IPWState               *pIPWState,
    PIXR                   *pBinImage,
    UInt16                 *pIndex,
    Float32                scaleX,
    Float32                scaleY,
    Int32                  yOffset,
    void                   *pColorTab,
    PIXR                   *prScaled)
{
    PROLOGUE;
    Run32        *runBuffer = NULL, *dummyRun, *prun;
    Int32        runBuffSize, width, height, depth;
    UInt8	 *pred, *pgreen, *pblue, *pBin, *srcImage;
    PIXR         *prPlane;
    Int32        channels, bpl, dbpl;
    Int32        row, col, y, x, scaledOffset, len;
    Int32        scaledH, scaledW;
    Int32        *yIdx, *xIdx = NULL;
    UInt8        value, r, g, b;
    UInt16       *pidx;
    CCMAPTABLE   *pCTable;

    ERR_GEN_PIXR (pBinImage, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN (pIndex == NULL, cErrNullPtr);
    ERR_GEN (prScaled == NULL, cErrNullPtr);

    ERR_PASS (w_pixrGetWidth (pIPWState, pBinImage, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, pBinImage, &height));
    ERR_PASS (w_pixrGetDepth(pIPWState, pBinImage, &depth));
    ERR_PASS (w_pixrGetImage (pIPWState, pBinImage, &pBin));
    ERR_PASS (w_pixrGetBpl (pIPWState, pBinImage, &bpl));

    pCTable = (CCMAPTABLE *)pColorTab;

    runBuffSize = width + 3;
    runBuffer = (Run32 *)IPW_MALLOC (runBuffSize * sizeof (Run32), 
				     cAllocatorData);
    ERR_GEN (runBuffer == NULL, cErrNomem);
    xIdx = (Int32 *)IPW_MALLOC ((width + height) * sizeof (Int32), 
				cAllocatorData);
    ERR_GEN (xIdx == NULL, cErrNomem);
    yIdx = xIdx + width;

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prScaled, &scaledW));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prScaled, &scaledH));

    for( row=0; row<height; ++row)
    {
	yIdx[row] = (Int32)((Float32)(row + yOffset) * scaleY + 0.5);
	if( yIdx[row] >= scaledH )
	    yIdx[row] = scaledH -1;
    }
    for( col=0; col<width; ++col)
    {
	    /* If this is a big endian machine, compute the scaled
	     * offset in the straightforward way.
	     * If this is a little endian machine, invert the lower
	     * 2 bits of the index.  We know that the pointers to the
	     * beginnings of lines always point to a 4-byte boundary.
	     * Inverting the LS two bits here means we can avoid using
	     * UCHAR_ACCESS (which slows down a poor PC).
	     */
	xIdx[col] = (Int32)((Float32)col * scaleX);
	if( xIdx[col] >= scaledW )
	    xIdx[col] = scaledW -1;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	xIdx[col] ^= 0x3;
#endif
    }
    
    scaledOffset = yIdx[0];
    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prScaled, 
					     &channels));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prScaled, &dbpl));

    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prScaled, cRedPlane, 
					&prPlane));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pred));
    pred += scaledOffset * dbpl;
    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prScaled, cGreenPlane, 
					&prPlane));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pgreen));
    pgreen += scaledOffset * dbpl;
    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prScaled, cBluePlane, 
					&prPlane));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pblue));
    pblue += scaledOffset * dbpl;

    srcImage = pBin;
    pidx = pIndex;
    for(row=0; row<height; )
    {
	y = yIdx[row];
	for( ; row < height && yIdx[row] == y; ++row)
	{
	    ERR_PASS_CLEANUP (i_findAllRunsInLineArray (pIPWState, srcImage,
							width,
							0,
							runBuffer,
							runBuffSize,
							0,  
							&dummyRun));
	    

	    prun = runBuffer;
	    while(prun != dummyRun)
	    {
		for(col= prun->runLeft; col<= prun->runRight; ++col )
		{
		    x = xIdx[col];
		    r = (pCTable + pidx[col])->red;
		    g = (pCTable + pidx[col])->green;
		    b = (pCTable + pidx[col])->blue;
		    value = *(pgreen+ x);
		    if(value == 0)
		    {
			*(pred+ x) = r;
			*(pgreen+ x) = g;
			*(pblue+ x) = b;
		    }
		    else
		    {
			len =0;
			if( value > g ) ++len;
			if( *(pred+ x) > r) ++len;
			if( *(pblue+ x) > b) ++len;
			if( len > 1)
			{
			    *(pred+ x) = r;
			    *(pgreen+ x) = g;
			    *(pblue+ x) = b;
			}
		    }
		}
		++ prun;
	    }	    
	    srcImage += bpl;
	    pidx += width;
	}
	pred += dbpl;
	pgreen += dbpl;
	pblue += dbpl;
    }


    IPW_FREE(runBuffer, cAllocatorData);
    IPW_FREE(xIdx, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:

    if(runBuffer != NULL) IPW_FREE(runBuffer, cAllocatorData);
    if(xIdx != NULL) IPW_FREE(xIdx, cAllocatorData);

    ERR_RETURN;
}


/*
 *
 * $Log: rescale1.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.44  1998/08/26 17:34:58  xzhu
 * xzhu on Wed Aug 26 10:34:10 PDT 1998
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Added new scaling function w_scaleMaskedIndexArea.
 *
 * Revision 1.43  1998/03/04 01:23:20  xzhu
 * xzhu on Tue Mar 3 17:21:47 PST 1998
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Set min-value 2 in order to get mask from w_thresh.
 *
 * Revision 1.42  1998/02/17 19:10:30  ddavies
 * ddavies on Tue Feb 17 11:09:48 PST 1998
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Fix coreCheckC bug, had inserted and taken out a parameter but not
 * removed parameter's specification in auto-doc comments.
 *
 * Revision 1.41  1998/02/17 03:06:00  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Minor performance tweaks for w_scaleMaskedArea.  Process each scan line
 * separately for each channel instead of processing all channels as we go.
 *
 * Revision 1.40  1998/01/14 00:15:06  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.39  1997/11/15 01:28:05  xzhu
 * xzhu on Fri Nov 14 17:26:37 PST 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Deleted offset 0.5 in generating X-Index.
 *
 * Revision 1.38  1997/07/08 21:08:17  xzhu
 * xzhu on Tue Jul 8 14:07:44 PDT 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Oh! last one is orignal one, no any changes.
 *
 * Revision 1.37  1997/05/16 18:19:02  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * w_scaleMapedArea renamed to w_scaleMaskedArea.
 * Both w_scaleToSize and w_scaleLinearSize now just copy the input to the
 * output if they're the same size.
 * The sizeX,sizeY args to w_scaleLinearSize and w_scaleAreaMapSize have
 * been changed to newW and newH.
 * Some of the comments have been clarified in w_scaleMaskedArea.
 *
 * Revision 1.36  1997/05/08 00:32:10  ddavies
 * ddavies on Wed May 7 17:12:32 PDT 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Big rework of w_scaleToSizeRect functionality.  In the old world, you
 * made a new set of tables for every different window in the source image.
 * In the new world, there's one set of tables for each source image and scale
 * factor pair.  The tables have also been consolidated into one opaque
 * thingie instead of three explicit tables.
 * So, in w_makeScaleToSizeTables, we don't need a rectangle anymore.
 * We allocate one table thingie and ask i_initScaleToSizeTables to
 * initialize it.
 * w_destroyScaleToSizeTables now frees one thing instead of three tables.
 * w_scaleToSizeRect now takes one thingie instead of three tables.
 * It never creates tables, since it would need the scale factor to do
 * so and it doesn't have that either.  The rectangle passed in is now
 * relative to the virtual destination image instead of relative to the
 * source image.  This makes is MUCH easier to scale abutting pieces of
 * the dest image correctly.
 * We now check to see if the rectangle's width and height match the
 * dest PIXR's width and height (error if not).
 *
 * Revision 1.35  1997/04/17 22:49:02  dferrell
 * dferrell on Thu Apr 17 15:48:35 PDT 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Removed unused parameters from calls to i_rescaleNoInterpSubRegion ().
 *
 * Revision 1.34  1997/04/16 18:27:32  dferrell
 * dferrell on Wed Apr 16 11:25:23 PDT 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Added new ipwrap interface for subrect arbitrary noninterpolated
 * scaling.  Modified the internals of the original function to reduce
 * code size given the new underlying functions.
 *
 * Revision 1.33  1997/02/28 00:56:17  xzhu
 * xzhu on Thu Feb 27 16:56:16 PST 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Fixed memory leak.
 *
 * Revision 1.32  1997/02/26 23:12:03  xzhu
 * xzhu on Wed Feb 26 15:11:09 PST 1997
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Added a new function w_scaleMapedArea() .
 *
 * Revision 1.31  1996/01/26 21:31:41  macleod
 * macleod on Fri Jan 26 13:30:36 PST 1996
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * fixed a bug in w_scaleLinearSize where it was unable to rescale multi-channel
 * images.
 *
 * Revision 1.30  1996/01/04  20:26:22  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.29  1995/12/13  04:12:31  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.28  1995/11/13  20:15:30  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.27  1995/10/13  22:04:44  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/09/26  22:08:54  dferrell
 * dferrell on Tue Sep 26 14:59:21 PDT 1995
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.25  1995/09/21  22:59:05  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.24  1995/09/19  21:11:43  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 * Fixed transfer of PIXR DPI to IPShared params by using macro
 * ScaledInt1616_TO_Int32().
 *
 * Revision 1.23  1995/09/15  22:38:28  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.22  1995/09/15  00:34:53  dferrell
 * dferrell on Thu Sep 14 17:34:44 PDT 1995
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.21  1995/09/11  17:53:53  dferrell
 * dferrell on Mon Sep 11 10:53:46 PDT 1995
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.20  1995/09/07  01:34:49  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.19  1995/09/06  20:22:57  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.18  1995/09/01  18:20:29  dferrell
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.17  1995/08/31  21:38:44  ejaquith
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.16  1995/08/17  21:54:40  dferrell
 * dferrell on Thu Aug 17 14:54:24 PDT 1995
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.15  1995/03/01  20:05:11  dferrell
 * dferrell on Wed Mar 1 11:56:41 PST 1995
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Moved binary rescale code from scaleToSizePixr () in ipwrap to
 * i_rescaleNoInterp () in ipshared.  scaleToSizePixr () is now purely
 * a wrapper for the underlying ipshared function.
 *
 * Revision 1.14  1995/02/03  01:38:47  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 *  any input resolutions through the scaling process.
 *
 * Revision 1.13  1994/11/22  20:10:32  dferrell
 * dferrell on Tue Nov 22 12:08:47 PST 1994
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Fixed mod 32 bug in scaleToSizePixr which caused binary rescale to a
 * destination width which was an integer multiple of 32 to not have the
 * rightmost 32 pixels set.
 *
 * Revision 1.12  1994/10/10  22:43:16  lperry
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Remove use of pcpros.h
 *
 * Revision 1.11  1994/10/03  20:37:27  lperry
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Changed to use rescale.pub instead of scalepros.h
 *
 * Revision 1.10  1994/09/16  17:39:39  jcraig
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.9  1994/06/15  23:30:16  dferrell
 * dferrell on Wed Jun 15 16:28:07 PDT 1994
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Optimized binary scaleToSizePixr performance by reading and writing
 * a word at a time.  Performance is still not great as a table lookup
 * occurs for each output pixel.  Speed is 10-20% faster, depending on
 * scale factor.
 *
 * Revision 1.8  1994/05/19  23:27:53  macleod
 * macleod on Thu May 19 16:27:28 PDT 1994
 *
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * scaleToSizePixr now also works on 4 bit/pixel images
 *
 * Revision 1.7  1994/05/05  01:49:18  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Add parens to assignments inside conditional clauses.
 *
 * Revision 1.6  1994/04/12  01:15:45  macleod
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.5  1994/02/09  22:29:16  macleod
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * scaleLinear now handles 4-bit images
 *
 * Revision 1.4  1994/01/14  23:23:34  dferrell
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Fixed #ifdef/ifndef checking to work with Mac as well as Suns & PCs.
 *
 * Revision 1.3  1993/12/03  01:18:30  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:16:17  ddavies
 * rescale1.c -> /product/ipcore/ipwrap/src/RCS/rescale1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

