/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  rescale2.c--procedures that rescale an image (and shift data)
 *
 *     Arbitrary scaling:
 *	     w_scaleAreaMapSize ()
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "rescale.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: rescale2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rescales an 8-bit single or multi-channel PIXR by a factor between 1.0 and 2.0
 * LongDescription:::
 *  Rescales and 8-bit PIXR using area mapping which is faster than linear
 *  interpolation.  While it will work using factors between 1.0 and 2.0,
 *  only factors between 1.0 and 1.5 are recommended.  The destination
 *  PIXR is created if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be scaled
 *  newW::           width of destination image (wS <= newW <= 2.0 * wS)
 *  newH::           height of destination image (hS <= newH <= 2.0 * hS)
 *
 * Output:::
 *  prd::            Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  arbitrary scaling
 *
 * Keywords:::
 *  arbitrary scaling, scaling, isomorphic, anamorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_scaleAreaMapSize (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               newW,
    Int32               newH,
    PIXR              **prd)
{
    PROLOGUE;
    Int32		 wS, hS,
                         i, sChannels, dChannels, 
			 frame, bplS, bplD;
    Bool		 destCreated = cFalse;
    Int32                rfc, bfc;
    ScaledInt1616        xRes, yRes;
    Float32		 scaleX, scaleY;
    ColorMap	        *map;
    ColorType	         type;
    UInt8               *iS, *iD;
    PIXR                *subS, *subD;

    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth8, cChannelsMask, cFrame32);

    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth8, cChannelsMask, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prs, &sChannels));
    ERR_PASS (w_pixrGetFrame (pIPWState, prs, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prs, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prs, &type));

    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
						      newW, newH, 8,
						      (xRes / wS) * newW,
						      (yRes / hS) * newH,
						      frame, sChannels,
						      type, map, prd,
						      &rfc, &bfc,
						      &destCreated));

    scaleX = (Float32)newW / (Float32)wS;
    scaleY = (Float32)newH / (Float32)hS;

    ERR_GEN_CLEANUP (((scaleX < 1.0) || (scaleX >= 2.0)), cErrInvalidParm);
    ERR_GEN_CLEANUP (((scaleY < 1.0) || (scaleY >= 2.0)), cErrInvalidParm);

    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, *prd, &dChannels));
    ERR_GEN_CLEANUP ((sChannels != dChannels), cErrInvalidParm);

    if ((newW == wS) && (newH == hS))
    {
        ERR_PASS_CLEANUP (w_copy (pIPWState, prs, prd));
	/* if not return, it will continue scaling image */
	NORMAL_RETURN;
    }

    for (i = 0; i < sChannels; i++)
    {
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));

	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wS));
	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subD, &newW));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hS));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subD, &newH));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

	ERR_PASS (i_rescaleAreaMap (pIPWState, iS, wS, hS, bplS, 8,
				    iD, newW, newH, bplD, 8));	
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_scaleAreaMapSize () */



/*
 *
 * $Log: rescale2.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.29  1998/01/14 00:15:08  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.28  1997/08/06 23:22:53  xzhu
 * xzhu on Wed Aug 6 16:21:58 PDT 1997
 *
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Added return after copy. If not, it will continue scaling image.
 *
 * Revision 1.27  1997/05/22 00:00:02  ddavies
 * ddavies on Wed May 21 16:59:19 PDT 1997
 *
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Fix bug in w_scaleAreaMapSize in which the calculation that produced
 * the resolution of the scaled image had overflow and got the wrong answer.
 *
 * Revision 1.26  1997/05/16 18:19:44  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Change sizeX, sizeY args to newW and newH in w_scaleAreaMapSize.
 *
 * Revision 1.25  1997/04/15 22:33:34  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.24  1996/01/04 20:26:25  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.23  1995/12/13  04:12:35  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:15:33  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:04:47  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/09/26  22:09:00  dferrell
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.19  1995/09/21  22:59:09  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/19  21:11:59  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.17  1995/09/15  22:38:33  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.16  1995/09/15  00:35:05  dferrell
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.15  1995/09/11  17:54:03  dferrell
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.14  1995/09/07  01:34:52  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:22:59  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  18:20:35  dferrell
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.11  1995/08/31  21:38:47  ejaquith
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/17  21:54:48  dferrell
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.9  1995/02/03  01:38:59  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.8  1994/10/10  22:43:29  lperry
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Remove use of pcpros.h
 *
 * Revision 1.7  1994/10/03  20:37:34  lperry
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Changed to use rescale.pub instead of scalepros.h
 *
 * Revision 1.6  1994/09/16  17:39:46  jcraig
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.5  1994/04/12  01:15:54  macleod
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.4  1994/01/14  23:23:52  dferrell
 * dferrell on Fri Jan 14 15:23:48 PST 1994
 *
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Fixed #ifdef/ifndef checking to work with Mac as well as Suns & PCs.
 *
 * Revision 1.3  1993/12/03  01:18:35  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:16:24  ddavies
 * rescale2.c -> /product/ipcore/ipwrap/src/RCS/rescale2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

