/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  resconv.c--procedures that perform resolution conversion
 *
 * 	General resolution conversion procedure:
 *		w_resConv ()
 *              w_resConvAndBinarize ()
 */


#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"
#include "shrip.pub"
#include "shrpixr.pub"

#include "resconv.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: resconv.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts between common resolutions for 4 and 8-bit PIXRs
 *
 * LongDescription:::
 *  The conversions currently supported are:
 *	200x200x4 -> 300x300x4	(or any 1.5x expansion)
 *	200x200x8 -> 300x300x8	(or any 1.5x expansion)
 *	100x100x4 -> 200x200x4	(or any 2x expansion)
 *	100x100x8 -> 200x200x8	(or any 2x expansion)
 *	200x200x4 -> 100x100x4	(or any 2x reduction)
 *	200x200x8 -> 100x100x8	(or any 2x reduction)
 *	300x300x4 -> 100x100x4	(or any 3x reduction)
 *	300x300x8 -> 100x100x8	(or any 3x reduction)
 *  
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR to be converted
 *  sRes::           Resolution of source PIXR
 *  dRes::           Resolution of destination PIXR
 *
 * Output:::
 *  prd::            Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  resolution conversion
 *
 * Keywords:::
 *  resolution conversion, scaling, isomorphic, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_resConv (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               sRes,
    Int32               dRes,
    PIXR              **prd)
{
    PROLOGUE;
    PIXR	    *subD, *subS;
    Int32	     wS, hS,
                     wSF, hSF, wDF, hDF,
                     bplS, bplD,
                     sDepth, i, 
		     channels, frame;
    Bool	     destCreated = cFalse;
    Int32            rfc, bfc;
    ScaledInt1616    xRes, yRes;
    ColorType        type;
    ColorMap        *map;
    UInt8           *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth4 | cDepth8, 
                  cChannelsMask, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prs, &sDepth));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prs, &channels));
    ERR_PASS (w_pixrGetFrame (pIPWState, prs, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prs, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prs, &type));


    if (dRes == (sRes + sRes / 2))	/* 1.5x expansion */
    {
	ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
							  wS * 3 / 2,
							  hS * 3 / 2,
							  sDepth,
							  xRes * 3 / 2,
							  yRes * 3 / 2,
							  frame, channels, type, map,
					       prd, &rfc, &bfc, &destCreated));

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subD, &wDF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subD, &hDF));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wSF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hSF));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

            ERR_PASS_CLEANUP (i_rescon2To3 (pIPWState, iS, wSF, hSF,
					    bplS, sDepth,
	                                    iD, wDF, hDF, bplD, sDepth, NULL));
	}
    }
    else
    if (dRes == 2 * sRes)		/* 2x expansion */
    {
	ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
							  wS * 2, hS * 2,
							  sDepth,
							  xRes * 2, yRes * 2,
							  frame, channels,
							  type, map, prd,
							  &rfc, &bfc,
							  &destCreated));

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wSF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hSF));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subD, &wDF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subD, &hDF));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

            ERR_PASS_CLEANUP (i_rescon1To2 (pIPWState, iS, wSF, hSF,
					    bplS, sDepth,
	                                    iD, wDF, hDF, bplD, sDepth, NULL));
	}
    }
    else if (sRes == 2 * dRes)		/* 2:1 reduction */
    {
	ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
							  wS / 2, hS / 2,
							  sDepth,
							  xRes / 2, yRes / 2,
							  frame, channels,
							  type, map, prd,
							  &rfc, &bfc,
							  &destCreated));

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wSF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hSF));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subD, &wDF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subD, &hDF));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

            ERR_PASS_CLEANUP (i_rescon2To1 (pIPWState, iS, wSF, hSF,
					    bplS, sDepth,
	                                    iD, wDF, hDF, bplD, sDepth));
	}
    }
    else if (sRes == 3 * dRes)		/* 3:1 reduction */
    {
	ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState,
							  wS / 3, hS / 3,
							  sDepth,
							  xRes / 3, yRes / 3,
							  frame, channels,
							  type, map, prd,
							  &rfc, &bfc,
							  &destCreated));

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subS, &wSF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subS, &hSF));
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, subD, &wDF));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, subD, &hDF));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subS, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, subD, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subS, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, subD, &iD));

            ERR_PASS_CLEANUP (i_rescon3To1 (pIPWState, iS, wSF, hSF,
					    bplS, sDepth,
	                                    iD, wDF, hDF, bplD, sDepth));
	}
    }
    else /* the expansion factor is illegal */
        ERR_GEN (1, cErrInvalidParm);

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_resConv () */


/* the next proc does transfers in strips.  This is the height of an
 * image used to hold the strips of expanded data before it's thresholded.
 * We really use the top and bottom frames too, so we expand into
 * 3*cFrameBits or 4*cFrameBits lines on each pass through the loop.
 * We add the extra two lines to the image because the resolution
 * conversion code assumes it can copy an extra frame width's bytes to the
 * end of the last line of the image.  Without this pad (1 line would
 * probably do), we get an extra 16 bytes written off the end of the
 * buffer image, smashing the heap.
 */
#define c2To1SrcSliceHeight	(6*cFrameBits)
#define c2To1DestSliceHeight	(3*cFrameBits)
#define	c2To1ImgSliceHeight	cFrameBits + 2	/* plus 2*cFrameBits */

#define c3To1SrcSliceHeight	(9*cFrameBits)
#define c3To1DestSliceHeight	(3*cFrameBits)
#define	c3To1ImgSliceHeight	cFrameBits + 2	/* plus 2*cFrameBits */

#define c2To3SrcSliceHeight	(2*cFrameBits)
#define c2To3DestSliceHeight	(3*cFrameBits)
#define	c2To3ImgSliceHeight	cFrameBits + 2	/* plus 2*cFrameBits */

#define c1To2SrcSliceHeight	(2*cFrameBits)
#define	c1To2DestSliceHeight	(4*cFrameBits)
#define	c1To2ImgSliceHeight	(2*cFrameBits) + 2	/* plus 2*cFrameBits */

/* conversion types */
#define c2xReduction	0	/* e.g. 200x200 -> 100x100 */
#define c3xReduction	1	/* e.g. 300x300 -> 100x100 */
#define	c1_5xExpansion	2	/* e.g. 200x200 -> 300x300 */
#define	c2xExpansion	3	/* e.g. 100x100 -> 200x200 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts resolution of 4 or 8 bit PIXR and creates binary result
 *
 * LongDescription:::
 *  Changes the resolution and binarizes an image.  For example,
 *  200x200x4 -> 300x300x1.  Can do this "in place" converting
 *  the input image into the output image.  Only operates on gray
 *  PIXRs.
 *  
 *  The conversions currently supported are:
 *	200x200x4 -> 100x100x1	(or any 2x reduction)
 *	200x200x8 -> 100x100x1	(or any 2x reduction)
 *	300x300x4 -> 100x100x1	(or any 3x reduction)
 *	300x300x8 -> 100x100x1	(or any 3x reduction)
 *	200x200x4 -> 300x300x1	(or any 1.5x expansion)
 *	200x200x8 -> 300x300x1	(or any 1.5x expansion)
 *	100x100x4 -> 200x200x1	(or any 2x expansion)
 *	100x100x8 -> 200x200x1	(or any 2x expansion)
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prS::            Source PIXR
 *  sRes::           resolution of source PIXR
 *  dRes::           destination resolution
 *  thresh::         threshold used during binarization
 *
 * Output:::
 *  prD::            destination PIXR (may be same as source)
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the source and dest images point to the same PIXR,
 *  the result is returned in that PIXR and it is modified to
 *  have the new depth, bytes/line.  The amount of memory allocated
 *  for it's image bits is also modified.
 *  WARNING: if prS == prD and something goes wrong after we've
 *  starting scribbling dest bits on the source pixr, we restore the source
 *  PIXR dimensions, but whatever image data has been overwritten
 *  is lost.  We check for all the common errors before this point.
 *
 *  When prD != prS, the width and height of prD must be set as follows:
 *		2x reduction:	wD = wS/2
 *				hD = hD/2
 *		3x reduction:	wD = wS/3
 *				hD = hS/3
 *		1.5x expansion:	wD = (wS & ~0x1) + (wS/2)
 *				hD = (hS & ~0x1) + (hS/2)
 *		2x expansion: wD = 2 * wS
 *				hD = 2 * hS
 *
 * FunctionCategory:::
 *  resolution conversion
 *
 * Keywords:::
 *  resolution conversion, depth conversion, isomorphic, scaling, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_resConvAndBinarize (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               sRes,
    Int32               dRes,
    Int32               thresh,
    PIXR              **prD)
{
    PROLOGUE;
    PIXR	    *prBuf;
    UInt8	    *pS, *pBuf, *pD;
    Int32	     wS, hS, wD, hD, sDepth, 
                     sBpl, dBpl, bufBpl, 
		     frame,
		     w, h;
    Bool	     destCreated = cFalse;
    Int32	     linesRemaining;	/* in src when expanding, in dest when
				         * reducing */
    Int32            rfc, bfc;
    UInt32	     conversionType;
    UInt32	    *pResconTable;
    UInt32	    *pThreshTabByte0;
    UInt32	    *pThreshTabByte1;
    UInt32	    *pThreshTabByte2;
    UInt32	    *pThreshTabByte3;
    Int32	     srcAddrLoopInc, destAddrLoopInc;
    ScaledInt1616    xSRes, ySRes, xDRes, yDRes;


    /* pixrs OK? */
    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth4 | cDepth8, cChannels1, cFrame32);
    if (*prD != prS)
	ERR_GEN_PIXR (*prD, cByteOrderBig, cDepth1, cChannels1, cFrame32);


    /* initialize these so we can know whether to free them in
     * case of error
     */
    prBuf = NULL;
    pResconTable = NULL;
    pThreshTabByte0 = NULL;
    pThreshTabByte1 = NULL;
    pThreshTabByte2 = NULL;
    pThreshTabByte3 = NULL;

    /* get the source image parms before they're smashed.  This can
     * happen if the source and dest are the same.
     */
    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prS, &sDepth));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prS, &xSRes, &ySRes));
    ERR_PASS (w_pixrGetFrame (pIPWState, prS, &frame));
    ERR_PASS (w_pixrGetBpl (pIPWState, prS, &sBpl));

    /* see if the threshold is legal */
    ERR_GEN ((((sDepth == 4) && (thresh > 15)) ||
	     ((sDepth == 8) && (thresh > 255)) ||
	      (thresh < 0)), cErrInvalidParm);

    /* make the threshold tables, we'll need these in any case */
    if (sDepth == 4)
	ERR_PASS (i_make4To1InvBinarizationTables (pIPWState,
						   &pThreshTabByte0,
						   &pThreshTabByte1,
						   &pThreshTabByte2,
						   &pThreshTabByte3));
    else
	ERR_PASS_CLEANUP (i_makeInvBinarizationTables (pIPWState,
						       &pThreshTabByte0,
					               &pThreshTabByte1,
					               &pThreshTabByte2,
					               &pThreshTabByte3));


    /* Calc the resolution conversion and compute dest width and height.
     * Create dest pixr if necessary.
     */
    if (dRes == sRes/2)			/* 2x reduction */
    {
	conversionType = c2xReduction;
	wD = wS/2;
	hD = hS/2;
	xDRes = xSRes/2;
	yDRes = ySRes/2;
    }
    else
    if (dRes == sRes/3)			/* 3x reduction */
    {
	conversionType = c3xReduction;
	wD = wS/3;
	hD = hS/3;
	xDRes = xSRes/3;
	yDRes = ySRes/3;
    }
    else
    if (dRes == (sRes + sRes/2))	/* 1.5x expansion */
    {
	conversionType = c1_5xExpansion;
	wD = wS + wS/2;
	hD = hS + hS/2;
	xDRes = xSRes + xSRes/2;
	yDRes = ySRes + ySRes/2;
    }
    else
    if (dRes == 2*sRes)			/* 2x expansion */
    {
	conversionType = c2xExpansion;
	wD = 2*wS;
	hD = 2*hS;
	xDRes = 2*xSRes;
	yDRes = 2*ySRes;
    }
    else
        ERR_GEN_CLEANUP (1, cErrInvalidParm);

    if (prS != *prD)
	ERR_PASS_CLEANUP (wp_createPixrConditionalNoInit (pIPWState,
							  wD, hD, 1,
							  xDRes, yDRes, 
							  frame, 
							  prD, &rfc, &bfc, 
							  &destCreated));

    /* If the source and dest pixrs are the same, convert the source
     * pixr now.  We'll need some of this info when computing where
     * the destination bits go.
     */
    if (prS == *prD)
    {
	Int32	offsetTo00;
	ERR_PASS_CLEANUP (w_pixrSetWidth (pIPWState, *prD, wD));
	ERR_PASS_CLEANUP (w_pixrSetHeight (pIPWState, *prD, hD));
	ERR_PASS_CLEANUP (w_pixrSetDepth (pIPWState, *prD, 1));
	/* Recalc the bpl since we changed the other parameters */
	ERR_PASS_CLEANUP (w_pixrBpl (pIPWState, *prD, &dBpl));
	/* Set the new bpl */
	ERR_PASS_CLEANUP (w_pixrSetBpl (pIPWState, *prD, dBpl));
	/* set the new resolution calced above */
	ERR_PASS_CLEANUP (w_pixrSetResolutionDPI (pIPWState, *prD,
						  xDRes, yDRes));
	/* Here's a tricky part.  The source and destination images'
	 * offsets from the beginning of the image buffer might
	 * not be the same.  Calculate the image offset of the
	 * dest image here.
	 */
	ERR_PASS_CLEANUP (wp_pixrGetImageBuffer(pIPWState, *prD, &pD));
	ERR_PASS_CLEANUP (w_pixrImageOffset(pIPWState, pD, dBpl, 1, frame,
					    &offsetTo00));
	pD = pD + offsetTo00 - (frame/8 + frame * dBpl);
    }
    else	/* make sure the client's prD has right dimensions */
    {
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, *prD, &w));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, *prD, &h));
	ERR_GEN_CLEANUP (((w != wD) || (h != hD)), cErrInvalidParm);
        ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prD, &dBpl));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prD, &pD));
    }

	/* MAIN BRANCH ON CONVERSION TYPE */

    switch (conversionType)
    {

	/* ***************** 2x REDUCTION  ************************* */
    case c2xReduction:
        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      wD, c2To1ImgSliceHeight,
					      sDepth, xDRes, yDRes,
					      cFrameBits, &prBuf));

	/* set the buffer's address pointer cFrameBits too high so we
	 * can use the buffer image's top and bottom frame bits as
	 * more buffer.  This is a cheat, but why let that memory
	 * go to waste?
	 * Note that we use pixrGetImageAll here so that i_rescon2To1
	 * can "remove the frame" from the addresses.
	 */
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prBuf, &bufBpl));
        ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prBuf, &pBuf));
	pBuf -= (bufBpl << cLogFrameBits);
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &pS));

	/* set the amount by which we increment the source and
	 * dest addresses each time through the loop.
	 */
	srcAddrLoopInc = c2To1SrcSliceHeight * sBpl;
	destAddrLoopInc = c2To1DestSliceHeight * dBpl;

	/* here's the loop where we do the work.  We do all the
	 * whole slices in this loop and the last, partial slice
	 * in the loop below.
	 */
	for (linesRemaining = hD;
	     linesRemaining > c2To1DestSliceHeight;
	     linesRemaining -= c2To1DestSliceHeight,
		pS += srcAddrLoopInc, pD += destAddrLoopInc)
	{
	    /* do the resize first */
	    ERR_PASS_CLEANUP (i_rescon2To1 (pIPWState, pS, wS + 2*cFrameBits,
				            c2To1SrcSliceHeight + 2*cFrameBits,
				            sBpl, sDepth,
				            pBuf, wD + 2*cFrameBits,
				            c2To1DestSliceHeight + 2*cFrameBits,
				            bufBpl, sDepth));

            /* now binarize */
	    if (sDepth == 4)
	    {
		if (thresh == 8)
		    ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState, pBuf, 
						      wD + 2 * cFrameBits,
						      c2To1DestSliceHeight +
							2 * cFrameBits,
						      bufBpl, pD, dBpl,
						      pThreshTabByte0,
						      pThreshTabByte1,
						      pThreshTabByte2,
						      pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState, pBuf, 
				           wD + 2 * cFrameBits,
					   c2To1DestSliceHeight +
					     2 * cFrameBits,
					   bufBpl, thresh,
					   pD, dBpl,
					   pThreshTabByte0,
					   pThreshTabByte1,
					   pThreshTabByte2,
					   pThreshTabByte3));
	    }
	    else
	    {
		if (thresh == 128)
		    ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						    pBuf, wD + 2 * cFrameBits,
					            c2To1DestSliceHeight + 
						      2 * cFrameBits,
					            bufBpl, pD, dBpl,
					            pThreshTabByte0, 
						    pThreshTabByte1,
					            pThreshTabByte2, 
						    pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						       pBuf, wD + 2*cFrameBits,
					               c2To1DestSliceHeight + 
						       2 * cFrameBits,
						       bufBpl, thresh,
						       pD, dBpl,
						       pThreshTabByte0, 
						       pThreshTabByte1,
						       pThreshTabByte2, 
						       pThreshTabByte3));
	    }
	}   /* for each full slice */

	/* there's always one more slice to do.  It's height is
	 * less than or equal to c2To1DestSliceHeight at this
	 * point.
	 */
	ERR_PASS_CLEANUP (i_rescon2To1 (pIPWState, pS, wS + 2 * cFrameBits,
			                2 * linesRemaining + 2 * cFrameBits,
			                sBpl, sDepth,
			                pBuf, wD + 2 * cFrameBits,
			                linesRemaining + 2 * cFrameBits,
			                bufBpl, sDepth));

	if (sDepth == 4)
	{
	    if (thresh == 8)
		ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState, pBuf, 
		                                       wD + 2 * cFrameBits,
					               linesRemaining +
						         2 * cFrameBits,
					               bufBpl, pD, dBpl,
					               pThreshTabByte0,
					               pThreshTabByte1,
					               pThreshTabByte2,
					               pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState, pBuf, 
		                                          wD + 2 * cFrameBits,
					                  linesRemaining +
						            2 * cFrameBits,
							    bufBpl, thresh,
							    pD, dBpl,
							    pThreshTabByte0,
							    pThreshTabByte1,
							    pThreshTabByte2,
							    pThreshTabByte3));
	}
	else
	{
	    if (thresh == 128)
		ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						pBuf, wD + 2 * cFrameBits,
				                linesRemaining + 2 * cFrameBits,
				                bufBpl, pD, dBpl,
				                pThreshTabByte0, 
						pThreshTabByte1,
				                pThreshTabByte2, 
						pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						   pBuf, wD + 2 * cFrameBits,
				                   linesRemaining + 
						     2 * cFrameBits,
				                   bufBpl, thresh,
				                   pD, dBpl,
				                   pThreshTabByte0, 
						   pThreshTabByte1,
				                   pThreshTabByte2, 
						   pThreshTabByte3));
	}
	break;	/* end of 2 to 1 resolution conversion step */


	/* ***************** 3x REDUCTION  ************************* */
    case c3xReduction:
        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      wD, c3To1ImgSliceHeight,
					      sDepth, xDRes, yDRes,
					      cFrameBits, &prBuf));

	/* set the buffer's address pointer cFrameBits too high so we
	 * can use the buffer image's top and bottom frame bits as
	 * more buffer.  This is a cheat, but why let that memory
	 * go to waste?
	 * Note that we use pixrGetImageAll here so that i_rescon2To1
	 * can "remove the frame" from the addresses.
	 */
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prBuf, &bufBpl));
        ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prBuf, &pBuf));
	pBuf -= (bufBpl << cLogFrameBits);
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &pS));

	/* set the amount by which we increment the source and
	 * dest addresses each time through the loop.
	 */
	srcAddrLoopInc = c3To1SrcSliceHeight * sBpl;
	destAddrLoopInc = c3To1DestSliceHeight * dBpl;

	/* here's the loop where we do the work.  We do all the
	 * whole slices in this loop an the last, partial slice
	 * in the loop below.
	 */
	for (linesRemaining = hD;
	     linesRemaining > c3To1DestSliceHeight;
	     linesRemaining -= c3To1DestSliceHeight,
		pS += srcAddrLoopInc, pD += destAddrLoopInc)
	{
	    /* do the resize first */
	    ERR_PASS_CLEANUP (i_rescon3To1 (pIPWState, pS, wS + 2*cFrameBits,
					    c3To1SrcSliceHeight + 2*cFrameBits,
					    sBpl, sDepth,
					    pBuf, wD + 2*cFrameBits,
					    c3To1DestSliceHeight + 2*cFrameBits,
					    bufBpl, sDepth));

            /* now binarize */
	    if (sDepth == 4)
	    {
		if (thresh == 8)
		    ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState, pBuf, 
		                                  wD + 2*cFrameBits,
						  c3To1DestSliceHeight +
							2*cFrameBits,
						  bufBpl, pD, dBpl,
						  pThreshTabByte0,
						  pThreshTabByte1,
						  pThreshTabByte2,
						  pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState, pBuf, 
		                                  wD + 2*cFrameBits,
						  c3To1DestSliceHeight +
							2*cFrameBits,
						  bufBpl, thresh,
						  pD, dBpl,
						  pThreshTabByte0,
						  pThreshTabByte1,
						  pThreshTabByte2,
						  pThreshTabByte3));
	    }
	    else
	    {
		if (thresh == 128)
		    ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
					   pBuf, wD + 2*cFrameBits,
					   c3To1DestSliceHeight + 2*cFrameBits,
					   bufBpl, pD, dBpl,
					   pThreshTabByte0, pThreshTabByte1,
					   pThreshTabByte2, pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
					   pBuf, wD + 2*cFrameBits,
					   c3To1DestSliceHeight + 2*cFrameBits,
					   bufBpl, thresh,
					   pD, dBpl,
					   pThreshTabByte0, pThreshTabByte1,
					   pThreshTabByte2, pThreshTabByte3));
	    }
	}

	/* there's always one more slice to do.  It's height is
	 * less than or equal to c3To1SrcSliceHeight at this
	 * point.
	 */
	ERR_PASS_CLEANUP (i_rescon3To1 (pIPWState, pS, wS + 2*cFrameBits,
			       3*linesRemaining + 2*cFrameBits,
			       sBpl, sDepth,
			       pBuf, wD + 2*cFrameBits,
			       linesRemaining + 2*cFrameBits,
			       bufBpl, sDepth));

	if (sDepth == 4)
	{
	    if (thresh == 8)
		ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState,
						       pBuf, wD + 2*cFrameBits,
						       linesRemaining +
								2*cFrameBits,
						       bufBpl, pD, dBpl,
						       pThreshTabByte0,
						       pThreshTabByte1,
						       pThreshTabByte2,
						       pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState,
							  pBuf, 
							  wD + 2*cFrameBits,
							  linesRemaining +
								2*cFrameBits,
							  bufBpl, thresh,
							  pD, dBpl,
							  pThreshTabByte0,
							  pThreshTabByte1,
							  pThreshTabByte2,
							  pThreshTabByte3));
	}
	else
	{
	    if (thresh == 128)
		ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						pBuf, wD + 2*cFrameBits,
						linesRemaining + 2*cFrameBits,
						bufBpl, pD, dBpl,
						pThreshTabByte0,
						pThreshTabByte1,
						pThreshTabByte2,
						pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						   pBuf, wD + 2*cFrameBits,
						   linesRemaining +
								2*cFrameBits,
						   bufBpl, thresh,
						   pD, dBpl,
						   pThreshTabByte0,
						   pThreshTabByte1,
						   pThreshTabByte2,
						   pThreshTabByte3));
	}
	break;	/* end of 3 to 1 resolution conversion step */

    /* ***************** 1.5x EXPANSION  ************************* */
    case c1_5xExpansion:	/* 1.5x expansion */
        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      wD, c2To3ImgSliceHeight,
					      sDepth, xDRes, yDRes,
					      cFrameBits, &prBuf));

	/* set the buffer's address pointer cFrameBits too high so we
	 * can use the buffer image's top and bottom frame bits as
	 * more buffer.  This is a cheat, but why let that memory
	 * go to waste?
	 * Note that we use pixrGetImageAll here so that i_rescon2To1
	 * can "remove the frame" from the addresses.
	 */
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prBuf, &bufBpl));
        ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prBuf, &pBuf));
	pBuf -= (bufBpl << cLogFrameBits);
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &pS));


	/* allocate the tables used to do the resolution conversion.
	 * It's more efficient to do it
	 * here than to do it on every pass through the loop.
	 */
	ERR_PASS_CLEANUP (i_make2To3ResconTable (pIPWState,
						 &pResconTable, sDepth));

	/* set the amount by which we increment the source and
	 * dest addresses each time through the loop.
	 */
	srcAddrLoopInc = c2To3SrcSliceHeight * sBpl;
	destAddrLoopInc = c2To3DestSliceHeight * dBpl;

	/* here's the loop where we do the work.  We do all the
	 * whole slices in this loop an the last, partial slice
	 * in the loop below.
	 */
	for (linesRemaining = (hS & ~0x1);
	     linesRemaining > c2To3SrcSliceHeight;
	     linesRemaining -= c2To3SrcSliceHeight,
		pS += srcAddrLoopInc, pD += destAddrLoopInc)
	{
	    /* do the resize first */
	    ERR_PASS_CLEANUP (i_rescon2To3 (pIPWState, pS, wS + 2*cFrameBits,
				   c2To3SrcSliceHeight + 2*cFrameBits,
				   sBpl, sDepth,
				   pBuf, wD + 2*cFrameBits,
				   c2To3DestSliceHeight + 2*cFrameBits,
				   bufBpl, sDepth,
				   pResconTable));

            /* now binarize */
	    if (sDepth == 4)
	    {
		if (thresh == 8)
		    ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState, pBuf, 
		                                  wD + 2*cFrameBits,
						  c2To3DestSliceHeight +
							2*cFrameBits,
						  bufBpl, pD, dBpl,
						  pThreshTabByte0,
						  pThreshTabByte1,
						  pThreshTabByte2,
						  pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState, pBuf, 
		                                  wD + 2*cFrameBits,
						  c2To3DestSliceHeight +
							2*cFrameBits,
						  bufBpl, thresh,
						  pD, dBpl,
						  pThreshTabByte0,
						  pThreshTabByte1,
						  pThreshTabByte2,
						  pThreshTabByte3));
	    }
	    else
	    {
		if (thresh == 128)
		    ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						    pBuf, wD + 2*cFrameBits,
						    c2To3DestSliceHeight +
								2*cFrameBits,
						    bufBpl, pD, dBpl,
						    pThreshTabByte0,
						    pThreshTabByte1,
						    pThreshTabByte2,
						    pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						       pBuf, wD + 2*cFrameBits,
						       c2To3DestSliceHeight +
								  2*cFrameBits,
						       bufBpl, thresh,
						       pD, dBpl,
						       pThreshTabByte0,
						       pThreshTabByte1,
						       pThreshTabByte2,
						       pThreshTabByte3));
	    }

	}

	/* there's always one more slice to do.  It's height is
	 * less than or equal to c2To3SrcSliceHeight at this
	 * point.
	 */
	ERR_PASS_CLEANUP (i_rescon2To3 (pIPWState, pS, wS + 2*cFrameBits,
			       linesRemaining + 2*cFrameBits,
			       sBpl, sDepth,
			       pBuf, wD + 2*cFrameBits,
			       linesRemaining + linesRemaining/2 + 2*cFrameBits,
			       bufBpl, sDepth,
			       pResconTable));

	if (sDepth == 4)
	{
	    if (thresh == 8)
		ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState,
						       pBuf, wD + 2*cFrameBits,
						       linesRemaining +
							linesRemaining/2 +
							2*cFrameBits,
						       bufBpl, pD, dBpl,
						       pThreshTabByte0,
						       pThreshTabByte1,
						       pThreshTabByte2,
						       pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState, pBuf, 
							  wD + 2*cFrameBits,
							  linesRemaining +
							    linesRemaining/2 +
							    2*cFrameBits,
							  bufBpl, thresh,
							  pD, dBpl,
							  pThreshTabByte0,
							  pThreshTabByte1,
							  pThreshTabByte2,
							  pThreshTabByte3));
	}
	else
	{
	    if (thresh == 128)
		ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						pBuf, wD + 2*cFrameBits,
						linesRemaining +
						  linesRemaining/2 +
						  2*cFrameBits,
						bufBpl, pD, dBpl,
						pThreshTabByte0,
						pThreshTabByte1,
						pThreshTabByte2,
						pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						   pBuf, wD + 2*cFrameBits,
						   linesRemaining + 
						     linesRemaining/2 +
						     2*cFrameBits,
						   bufBpl, thresh,
						   pD, dBpl,
						   pThreshTabByte0,
						   pThreshTabByte1,
						   pThreshTabByte2,
						   pThreshTabByte3));
	}
	break;  /* end of 2 to 3 resolution conversion step */



	/* ***************** 2x EXPANSION  ************************* */
    default:	/* case c2xExpansion: */
        ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      wD, c1To2ImgSliceHeight,
					      sDepth, xDRes, yDRes,
					      cFrameBits, &prBuf));

	/* set the buffer's address pointer cFrameBits too high so we
	 * can use the buffer image's top and bottom frame bits as
	 * more buffer.  This is a cheat, but why let that memory
	 * go to waste?
	 * Note that we use pixrGetImageAll here so that i_rescon2To1
	 * can "remove the frame" from the addresses.
	 */
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prBuf, &bufBpl));
        ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prBuf, &pBuf));
	pBuf -= (bufBpl << cLogFrameBits);
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &pS));

	/* allocate the tables used to do the resolution conversion.
	 * It's more efficient to do it
	 * here than to do it on every pass through the loop.
	 */
	ERR_PASS_CLEANUP (i_make1To2ResconTable (pIPWState,
						 &pResconTable, sDepth));

	/* set the amount by which we increment the source and
	 * dest addresses each time through the loop.
	 */
	srcAddrLoopInc = c1To2SrcSliceHeight * sBpl;
	destAddrLoopInc = c1To2DestSliceHeight * dBpl;

	/* here's the loop where we do the work.  We do all the
	 * whole slices in this loop an the last, partial slice
	 * in the loop below.
	 */
	for (linesRemaining = hS;
	     linesRemaining > c1To2SrcSliceHeight;
	     linesRemaining -= c1To2SrcSliceHeight,
		pS += srcAddrLoopInc, pD += destAddrLoopInc)
	{
	    /* do the resize first */
	    ERR_PASS_CLEANUP (i_rescon1To2 (pIPWState, pS, wS + 2*cFrameBits,
					    c1To2SrcSliceHeight + 2*cFrameBits,
					    sBpl, sDepth,
					    pBuf, wD + 2*cFrameBits,
					    c1To2DestSliceHeight + 2*cFrameBits,
					    bufBpl, sDepth,
					    pResconTable));

            /* now binarize */
	    if (sDepth == 4)
	    {
		if (thresh == 8)
		    ERR_PASS_CLEANUP (
				i_thresholdI4To1bpp (pIPWState, pBuf, 
						     wD + 2*cFrameBits,
						     c1To2DestSliceHeight +
							2*cFrameBits,
						     bufBpl, pD, dBpl,
						     pThreshTabByte0,
						     pThreshTabByte1,
						     pThreshTabByte2,
						     pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (
				i_highpassfiltI4To1bpp (pIPWState, pBuf, 
							wD + 2*cFrameBits,
							c1To2DestSliceHeight +
							  2*cFrameBits,
							bufBpl, thresh,
							pD, dBpl,
							pThreshTabByte0,
							pThreshTabByte1,
							pThreshTabByte2,
							pThreshTabByte3));
	    }
	    else
	    {
		if (thresh == 128)
		    ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						    pBuf, wD + 2*cFrameBits,
						    c1To2DestSliceHeight +
							2*cFrameBits,
						    bufBpl, pD, dBpl,
						    pThreshTabByte0,
						    pThreshTabByte1,
						    pThreshTabByte2,
						    pThreshTabByte3));
		else
		    ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						       pBuf, wD + 2*cFrameBits,
						       c1To2DestSliceHeight +
						           2*cFrameBits,
						       bufBpl, thresh,
						       pD, dBpl,
						       pThreshTabByte0,
						       pThreshTabByte1,
						       pThreshTabByte2,
						       pThreshTabByte3));
	    }
	}

	/* there's always one more slice to do.  It's height is
	 * less than or equal to c1To2SrcSliceHeight at this
	 * point.
	 */
	ERR_PASS_CLEANUP (i_rescon1To2 (pIPWState, pS, wS + 2*cFrameBits,
					linesRemaining + 2*cFrameBits,
					sBpl, sDepth,
					pBuf, wD + 2*cFrameBits,
					2*linesRemaining + 2*cFrameBits,
					bufBpl, sDepth,
					pResconTable));

	if (sDepth == 4)
	{
	    if (thresh == 8)
		ERR_PASS_CLEANUP (i_thresholdI4To1bpp (pIPWState,
						       pBuf, wD + 2*cFrameBits,
						       2*linesRemaining +
							 2*cFrameBits,
						       bufBpl, pD, dBpl,
						       pThreshTabByte0,
						       pThreshTabByte1,
						       pThreshTabByte2,
						       pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI4To1bpp (pIPWState, pBuf, 
							  wD + 2*cFrameBits,
							  2*linesRemaining +
							    2*cFrameBits,
							  bufBpl, thresh,
							  pD, dBpl,
							  pThreshTabByte0,
							  pThreshTabByte1,
							  pThreshTabByte2,
							  pThreshTabByte3));
	}
	else
	{
	    if (thresh == 128)
		ERR_PASS_CLEANUP (i_thresholdI (pIPWState,
						pBuf, wD + 2*cFrameBits,
						2*linesRemaining + 2*cFrameBits,
						bufBpl, pD, dBpl,
						pThreshTabByte0,
						pThreshTabByte1,
						pThreshTabByte2,
						pThreshTabByte3));
	    else
		ERR_PASS_CLEANUP (i_highpassfiltI (pIPWState,
						   pBuf, wD + 2*cFrameBits,
						   2*linesRemaining +
								2*cFrameBits,
						   bufBpl, thresh,
						   pD, dBpl,
						   pThreshTabByte0,
						   pThreshTabByte1,
						   pThreshTabByte2,
						   pThreshTabByte3));
	}
	break;	/* end of 1 to 2 resolution conversion step */
    }

    /* At this point, prD has all the correct bits in it.  If prD is
     * really prS, fix the amount of memory allocated for the
     * image bits and clear the frames.
     */
    if (prS == *prD)
    {
	/* this should always work because
	 * we're always making the image buffer
	 * smaller */
	ERR_PASS_CLEANUP (w_reallocPixrData (pIPWState, *prD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prD, &pD));
	ERR_PASS_CLEANUP (i_clearAllFrames (pIPWState, pD,
				            wD + 2*cFrameBits,
				            hD + 2*cFrameBits,
				            dBpl, 1));
    }
    else
	/* Clear the dest frame if necessary */
	ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prD,
						     rfc, bfc));

    /* free the temp slice image buffer */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBuf));

    /* free the tables.  Note that since they were allocated by an
     * ipshared function which did not take an allocator pointer,
     * they can only safely be freed using the old FREE macro.
     */
    IPW_FREE (pResconTable, cAllocatorData);
    IPW_FREE (pThreshTabByte0, cAllocatorData);
    IPW_FREE (pThreshTabByte1, cAllocatorData);
    IPW_FREE (pThreshTabByte2, cAllocatorData);
    IPW_FREE (pThreshTabByte3, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:

    /* free the temp slice image buffer */
    w_destroyPixr (pIPWState, &prBuf);

    if (destCreated)
	w_destroyPixr (pIPWState, prD);


    /* free the tables.  Note that since they were allocated by an
     * ipshared function which did not take an allocator pointer,
     * they can only safely be freed using the old FREE macro.
     */
    IPW_FREE (pResconTable, cAllocatorData);
    IPW_FREE (pThreshTabByte0, cAllocatorData);
    IPW_FREE (pThreshTabByte1, cAllocatorData);
    IPW_FREE (pThreshTabByte2, cAllocatorData);
    IPW_FREE (pThreshTabByte3, cAllocatorData);

    /* attempt to restore prS's dimensions.  Unfortunately, the image data
     * may or may not be trashed at this point.
     */
    w_pixrSetWidth (pIPWState, prS, wS);
    w_pixrSetHeight (pIPWState, prS, hS);
    w_pixrSetDepth (pIPWState, prS, sDepth);
    w_pixrSetBpl (pIPWState, prS, sBpl);
    w_pixrSetResolutionDPI (pIPWState, prS, xSRes, ySRes);
    ERR_RETURN;

}  /* resConvAndBinarizePr () */


/*
 *
 * $Log: resconv.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.37  1998/01/14 00:15:10  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.36  1997/04/15 22:35:15  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Use non-initialized PIXR creation where possible.
 * Also use wp_createPixrConditionalNoInit instead of
 * wp_createPixrExtendedConditionalNoInit in w_resConfAndBinarize since
 * the dest is guaranteed to have exactly one plane.
 *
 * Revision 1.35  1996/09/24 02:50:24  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Use w_pixrImageOffset and wp_pixrGetImageBuffer to calculate the new
 * offset from the beginning of the destination buffer in w_resConvAndBinarize.
 * Images do not necessarily start at the beginning of the image buffer that
 * holds the image.
 *
 * Revision 1.34  1996/01/04 20:26:28  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.33  1995/12/13  04:12:39  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.32  1995/11/13  20:15:36  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.31  1995/10/13  22:04:50  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.30  1995/09/26  22:10:05  dferrell
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Updated function comment blocks.  Removed dest depth parameter from
 * w_resConv ().
 *
 * Revision 1.29  1995/09/21  22:59:12  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.28  1995/09/19  21:12:09  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () calls
 *
 * Revision 1.27  1995/09/15  22:38:38  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.26  1995/09/15  00:35:24  dferrell
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.25  1995/09/11  17:54:13  dferrell
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.24  1995/09/07  01:34:54  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.23  1995/09/06  20:23:02  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.22  1995/09/01  18:20:22  dferrell
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.21  1995/08/31  21:38:50  ejaquith
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.20  1995/08/24  18:10:52  macleod
 * macleod on Thu Aug 24 11:09:13 PDT 1995
 *
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Fixed a bunch of problems--passing image width without frame when a
 * routine expected framed with, passing pointer to image rather than
 * pointer to frame, etc. Also fixed a bug in resConvAndBinarize where
 * we would always complain about wrong depth when called with an
 *
 * Revision 1.19  1995/08/17  21:54:55  dferrell
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.18  1995/03/08  04:30:47  ddavies
 * ddavies on Tue Mar 7 20:29:04 PST 1995
 *
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Change the way resConvAndBinarize calculates the output width and height
 * on 1.5x expansion from (wS & ~0x1) + wS/2 to wS + wS/2.  This fixed allows
 * us to touch the entire output image.  Update comments on calculation of
 * image addresses used by resConvAndBinarize.
 *
 * Revision 1.17  1995/02/03  01:39:10  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.16  1994/10/20  02:24:45  ddavies
 * ddavies on Wed Oct 19 19:22:17 PDT 1994
 *
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Change resConvAndBinarize so that the client can specify the threshold
 * used for binarization.  When the threshold is half of the range, we sue
 * the standard i_threshold functions.  Otherwise, use the i_highpassfiltI
 * functions.
 *
 * Revision 1.15  1994/10/10  22:44:06  lperry
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Remove use of pcpros.h
 * Add include of shrpixr.pub
 *
 * Revision 1.14  1994/10/03  20:36:50  lperry
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Changed to use resconv.pub instead of scalepros.h
 *
 * Revision 1.13  1994/09/20  04:33:52  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Add the new procedure resConvAndBinarizePr.  This is one of the first examples
 * of a routine that does two separate operations using an image strip as the
 * intermediate buffer.  This routine does resolution conversion and
 * binarization of the output.  Resolution conversions supported are 2:1, 1:2,
 * 2:3 and 3:1.  Input images may be 4bpp or 8bpp.  EXCEPT, someone supplied
 * only dummy versions of the 2:1 and 3:1 reduction routines for 4 bpp iamges.
 * Nasty.  Maybe no one has ever wanted to do this?  Also, the source and
 * dest images may be the same.  If this happens, the source pixr has its
 * depth and bytes/line changed.  The amount of memory associated with this
 * image is adjusted after the conversion.  Wow.
 *
 * Revision 1.12  1994/09/17  18:15:49  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Add extra NULL table arg to calls to i_rescon1To2 and i_rescon2To3.
 *
 * Revision 1.11  1994/07/01  05:13:39  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Replaced inclusion of ippros.h with inclusion of scalpros.h since this
 * is in scallib, not iplib.
 *
 * Revision 1.10  1994/06/07  02:40:26  macleod
 * macleod on Mon Jun 6 19:40:25 PDT 1994
 *
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.9  1994/05/26  00:16:26  macleod
 * macleod on Wed May 25 17:15:31 PDT 1994
 *
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Added 2:1 and 3:1 down-conversion. It takes either 4- or 8-bit
 * images, though *note* that the 4-bit inner loops are stubbed out at
 * the moment.
 *
 * Revision 1.8  1994/04/12  01:16:02  macleod
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.7  1994/01/14  23:23:19  dferrell
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Fixed #ifdef/ifndef checking to work with Mac as well as Suns & PCs.
 *
 * Revision 1.6  1993/12/03  01:18:39  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.5  1993/11/06  03:16:31  ddavies
 * resconv.c -> /product/ipcore/ipwrap/src/RCS/resconv.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

