/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  rotatop1.c:  includes procedures
 *
 *         w_rotate180():	rotates by 180 degrees
 *	   w_rotate180InPlace()
 *
 *         w_rotate90():	rotates by 90 degrees clockwise
 *	   w_rotate90InPlace()
 *
 *         w_rotate270():	rotates by 90 degrees counterclockwise
 *	   w_rotate270InPlace()
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "rotatop.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: rotatop1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

#define  TIMING                 0


/*---------------------------------------------------------------------*
 *                          180 Degree Rotation                        *
 *---------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by 180 degrees.
 *
 * LongDescription:::
 *  This function rotates binary, gray and color images by 180 degrees.
 *  That is, it turns them upside down.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to input image.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *<BR><BR>
 *		     If the dest image is the same as the source image, the
 *		     rotation is done in place.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, orthogonal
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotate180 (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    Int32	 wS, hS, dS, sBpl, dBpl, channels, i;
    UInt8	*pS, *pD;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));

	/* Do the actual rotation, one channel at a time.
	 * If there's only one channel, just rotate it.
	 */
    if (channels != 1)
    {
	PIXR   *subD, *subS;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prS, i, &subS));
	    ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, prD, i, &subD));
	    
	    ERR_PASS_CLEANUP( w_rotate180 (pIPWState, subS, &subD));
	}
    }
    else
    {
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
	
	ERR_PASS_CLEANUP( i_fastRot180 (pIPWState, pD, pS, wS, hS,
					sBpl, dBpl, dS));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by 180 degrees in place.
 *
 * LongDescription:::
 *  An image is rotated by 180 degrees.  The result is always returned
 *  in the same image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  prS::            Pointer to source/dest image.
 *
 * Notes:::
 *  This is the same as calling w_rotate180 with *pPrD == prS.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, orthogonal
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotate180InPlace (
    IPWState           *pIPWState,
    PIXR               *prS)
{
    PROLOGUE;
    Int32	 wS, hS, dS, sBpl, channels, i;
    UInt8	*pS;

    ERR_GEN(prS == NULL, cErrPixr_null);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));

    if (channels != 1)
    {
	PIXR   *subS;
	
	for (i = 0 ; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS( w_rotate180InPlace (pIPWState, subS));
	}
    }
    else
    {
	ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
	ERR_PASS( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
	
	ERR_PASS( i_fastRot180 (pIPWState, pS, pS, wS, hS, sBpl, sBpl, dS));
    }

    NORMAL_RETURN;
}

/*---------------------------------------------------------------------*
 *                           90 Degree Rotation                        *
 *---------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by 90 degrees clockwise (270 degrees counterclockwise).
 *
 * LongDescription:::
 *  This function rotates binary, gray and color images by 90 degrees
 *  clockwise.  It is an exact operation in that there is a 1 to 1 mapping
 *  from pixels in the input to pixels in the output.  No jaggies are
 *  introduced.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be rotated.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if the source and dest images are the same.
 *
 * Notes:::
 *  It is not possible to rotate an image in place using this function.
 *  See w_rotate90InPlace for that capability.
 *<BR><BR>
 *  This is a pretty fast function.  Normally quicker than small angle
 *  rotates.  People who want to do both horizontal and vertical processing
 *  sometimes find it's faster to write good horizontal processing code
 *  and rotate the image to get the vertical data.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, orthogonal
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotate90 (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    ScaledInt1616	 xSRes, ySRes;
    Int32		 hS, wS, dS, sBpl, dBpl, channels, i;
    ColorType		 sType;
    ColorMap		*map;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSRes, &ySRes));
    ERR_PASS( w_pixrGetType(pIPWState, prS, &sType));
    ERR_PASS( w_pixrGetColorMap(pIPWState, prS, &map));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrExtendedConditionalNoInit(pIPWState, hS, wS, dS,
						     ySRes, xSRes, cFrameBits,
						     channels, sType, map, &prD,
						     &rightFrameColsToClear,
						     &bottomFrameRowsToClear,
						     &allocatedHere));

	/* Do the actual rotation, one channel at a time.
	 * If there's only one channel, just rotate it.
	 */
    if (channels != 1)
    {
	PIXR  *subS, *subD;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &subD));
	    
	    ERR_PASS_CLEANUP( w_rotate90 (pIPWState, subS, &subD));
	}
    }
    else
    {
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));

	ERR_PASS_CLEANUP( i_fastRot90 (pIPWState, pD, pS, wS, hS,
				       sBpl, dBpl, dS));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by 90 degrees clockwise, leaving result in source image.
 *
 * LongDescription:::
 *  This function rotates binary, gray and color images by 270 degrees
 *  clockwise (90 degrees counterclockwise).  It is an exact operation
 *  in that there is a 1 to 1 mapping from pixels in the input to pixels 
 *  in the output.  No jaggies are introduced.
 *<BR><BR>
 *  The difference between this operation and w_rotate90 is that the result
 *  is left in the source image here.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  prS::            Pointer to source/dest image.
 *  prTransfer::     Pointer to image that is used to buffer blocks of
 *		     rotated image during the transfer.  prTransfer must
 *		     be framed, just be the same height as the rotated
 *		     image (same width as the source image) and must have
 *		     a width that's a multiple of 32 pixels.  The wider the
 *		     transfer image, the faster the rotate op (modulo
 *		     swapping interference).
 *<BR><BR>
 *		     The left and right frames of prTransfer are trashed
 *		     by this operation (they're used for data).  The caller
 *		     is expected to discard prTransfer immediately on returning
 *		     from this function.
 *
 * Notes:::
 *  This is a pretty tricky operation when you think about it.  We rotate an
 *  image by 90 degrees clockwise and leave the result in the same memory
 *  block.  The trick is that the bytes/line of the source image seldom
 *  matches the bytes/line of the dest image.  One's naive picture of simply
 *  turning an image on its side doesn't work very well when two images
 *  having different numbers of bytes/line try to share the same block of
 *  memory.  To make this work at all, the user must supply a separate block
 *  of memory, prTransfer.  This block holds strips of rotated image
 *  while we juggle things around in the source image to make room.
 *<BR><BR>
 *  The in-place version of rotation is normally slower that the normal
 *  version.  It is useful for very large images and/or smaller memory
 *  environments where much time would be spent swapping between a source
 *  image and a destination image.
 *<BR><BR>
 *  When they are allocated, all images reserve enough padding at the bottom
 *  so that they can be rotated by 90 or 270 degrees in place.  In practice,
 *  this isn't a lot of memory.  It's the difference between the amount of
 *  right hand pad (pad to the nearest word boundary on the right side of each
 *  scan line) required in one orientation vs the amount needed in the other.
 *  Since frames are allocated in pixels (not bits), the amount of frame
 *  needed in both orientations is the same.  Ditto for the number of bytes
 *  of raw image data.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, orthogonal
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  The algorithm used here is based on some clever work done at WRC,
 *  but I can't put my finger on it.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotate90InPlace (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR               *prTransfer)
{
    PROLOGUE;
    Int32		 wS, hS, dS, wT, i, channels;
    Int32		 bpls, tBpl;
    UInt8		*pS, *pT;
    ScaledInt1616	 xRes, yRes;

    ERR_GEN(prS == NULL || prTransfer == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetChannelCount (pIPWState, prS, &channels));

    if (channels != 1)
    {
	PIXR  *subS;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS( w_rotate90InPlace (pIPWState, subS, prTransfer));
	}
    }
    else
    {
	Int32  bpltmp;
	
	ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS( w_pixrGetImageAll(pIPWState, prTransfer, &pT));
	ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS( w_pixrFramedWidth(pIPWState, prTransfer, &wT));
	ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
	ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpls));
	ERR_PASS( w_pixrGetBpl(pIPWState, prTransfer, &tBpl));
	
	ERR_PASS( i_fastRot90InPlace (pIPWState, pT, pS, wS, hS,
				      wT, bpls, tBpl, dS));

	/* exchange the width and height parameters and set the new
	 * bytes/line of the rotated image.
	 */
	ERR_PASS( w_pixrSetWidth(pIPWState, prS, hS - 2*cFrameBits));
	ERR_PASS( w_pixrSetHeight(pIPWState, prS, wS - 2*cFrameBits));
	ERR_PASS( w_pixrBpl(pIPWState, prS, &bpltmp));
	ERR_PASS( w_pixrSetBpl(pIPWState, prS, bpltmp));

	    /* fix the resolution of the dest image */
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xRes, &yRes));
	ERR_PASS( w_pixrSetResolutionDPI(pIPWState, prS, yRes, xRes));
    }

    NORMAL_RETURN;
}



/*---------------------------------------------------------------------*
 *                          270 Degree Rotation                        *
 *---------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by 270 degrees clockwise (90 degrees counterclockwise).
 *
 * LongDescription:::
 *  This function rotates binary, gray and color images by 270 degrees
 *  clockwise.  It is an exact operation in that there is a 1 to 1 mapping
 *  from pixels in the input to pixels in the output.  No jaggies are
 *  introduced.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be rotated.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if the source and dest images are the same.
 *
 * Notes:::
 *  It is not possible to rotate an image in place using this function.
 *  See w_rotate270InPlace for that capability.
 *<BR><BR>
 *  This is a pretty fast function.  Normally quicker than small angle
 *  rotates.  People who want to do both horizontal and vertical processing
 *  sometimes find it's faster to write good horizontal processing code
 *  and rotate the image to get the vertical data.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, orthogonal
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotate270 (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    ScaledInt1616	 xSRes, ySRes;
    Int32		 hS, wS, dS, sBpl, dBpl, channels, i;
    ColorType		 sType;
    ColorMap		*map;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrPixr_null);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSRes, &ySRes));
    ERR_PASS( w_pixrGetType(pIPWState, prS, &sType));
    ERR_PASS( w_pixrGetColorMap(pIPWState, prS, &map));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_createPixrExtendedConditionalNoInit(pIPWState, hS, wS, dS,
						     ySRes, xSRes, cFrameBits,
						     channels, sType, map, &prD,
						     &rightFrameColsToClear,
						     &bottomFrameRowsToClear,
						     &allocatedHere));

	/* Do the actual rotation, one channel at a time.
	 * If there's only one channel, just rotate it.
	 */
    if (channels != 1)
    {
	PIXR  *subS, *subD;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &subD));
	    
	    ERR_PASS_CLEANUP( w_rotate270 (pIPWState, subS, &subD));
	}
    }
    else
    {
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));

	ERR_PASS_CLEANUP( i_fastRot270 (pIPWState, pD, pS, wS, hS,
					sBpl, dBpl, dS));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by 270 degrees clockwise, leaving result in source image.
 *
 * LongDescription:::
 *  This function rotates binary, gray and color images by 270 degrees
 *  clockwise (90 degrees counterclockwise).  It is an exact operation
 *  in that there is a 1 to 1 mapping from pixels in the input to pixels 
 *  in the output.  No jaggies are introduced.
 *<BR><BR>
 *  The difference between this operation and w_rotate270 is that the result
 *  is left in the source image here.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  prS::            Pointer to source/dest image.
 *  prTransfer::     Pointer to image that is used to buffer blocks of
 *		     rotated image during the transfer.  prTransfer must
 *		     be framed, just be the same height as the rotated
 *		     image (same width as the source image) and must have
 *		     a width that's a multiple of 32 pixels.  The wider the
 *		     transfer image, the faster the rotate op (modulo
 *		     swapping interference).
 *<BR><BR>
 *		     The left and right frames of prTransfer are trashed
 *		     by this operation (they're used for data).  The caller
 *		     is expected to discard prTransfer immediately on returning
 *		     from this function.
 *
 * Notes:::
 *  This is a pretty tricky operation when you think about it.  We rotate an
 *  image by 270 degrees clockwise and leave the result in the same memory
 *  block.  The trick is that the bytes/line of the source image seldom
 *  matches the bytes/line of the dest image.  One's naive picture of simply
 *  turning an image on its side doesn't work very well when two images
 *  having different numbers of bytes/line try to share the same block of
 *  memory.  To make this work at all, the user must supply a separate block
 *  of memory, prTransfer.  This block holds strips of rotated image
 *  while we juggle things around in the source image to make room.
 *<BR><BR>
 *  The in-place version of rotation is normally slower that the normal
 *  version.  It is useful for very large images and/or smaller memory
 *  environments where much time would be spent swapping between a source
 *  image and a destination image.
 *<BR><BR>
 *  When they are allocated, all images reserve enough padding at the bottom
 *  so that they can be rotated by 90 or 270 degrees in place.  In practice,
 *  this isn't a lot of memory.  It's the difference between the amount of
 *  right hand pad (pad to the nearest word boundary on the right side of each
 *  scan line) required in one orientation vs the amount needed in the other.
 *  Since frames are allocated in pixels (not bits), the amount of frame
 *  needed in both orientations is the same.  Ditto for the number of bytes
 *  of raw image data.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, orthogonal
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  The algorithm used here is based on some clever work done at WRC,
 *  but I can't put my finger on it.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotate270InPlace (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR               *prTransfer)
{
PROLOGUE;
Int32		 wS, hS, dS, wT, i, channels;
Int32		 bpls, tBpl;
UInt8		*pS, *pT;
ScaledInt1616	 xRes, yRes;

    ERR_GEN(prS == NULL || prTransfer == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetChannelCount (pIPWState, prS, &channels));

    if (channels != 1)
    {
	PIXR   *subS;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel (pIPWState, prS, i, &subS));
	    ERR_PASS( w_rotate270InPlace (pIPWState, subS, prTransfer));
	}
    }
    else
    {
	Int32  bpltmp;
	
	ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS( w_pixrGetImageAll(pIPWState, prTransfer, &pT));
	ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS( w_pixrFramedWidth(pIPWState, prTransfer, &wT));
	ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
	ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpls));
	ERR_PASS( w_pixrGetBpl(pIPWState, prTransfer, &tBpl));

	ERR_PASS( i_fastRot270InPlace (pIPWState, pT, pS, wS, hS, wT,
				       bpls, tBpl, dS));

	/* exchange the width and height parameters and set the new
	 * bytes/line of the rotated image.
	 */
	ERR_PASS( w_pixrSetWidth(pIPWState, prS, hS - 2*cFrameBits));
	ERR_PASS( w_pixrSetHeight(pIPWState, prS, wS - 2*cFrameBits));
	ERR_PASS( w_pixrBpl(pIPWState, prS, &bpltmp));
	ERR_PASS( w_pixrSetBpl(pIPWState, prS, bpltmp));

	    /* fix the resolution of the dest image */
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xRes, &yRes));
	ERR_PASS( w_pixrSetResolutionDPI(pIPWState, prS, yRes, xRes));
    }

    NORMAL_RETURN;
}


/*

 $Log: rotatop1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.34  1998/01/14 00:15:12  ddavies
 rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.33  1997/04/15 22:35:30  ddavies
 rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.32  1996/01/04 20:26:31  ejaquith
 rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.31  1995/12/13  04:12:43  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.30  1995/11/13  20:15:40  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.29  1995/10/13  22:04:53  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.28  1995/10/05  04:27:41  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.27  1995/10/03  01:20:39  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.26  1995/09/21  22:59:16  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.25  1995/09/16  02:03:09  ejaquith
 * ejaquith on Fri Sep 15 19:02:34 PDT 1995
 *
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Bug fix.
 * Calls to wp_createPixrExtendedConditional() were using cFrame32 (4096!)
 * for the frame instead of cFrameBits (32).
 *
 * Revision 1.24  1995/09/15  18:01:03  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.23  1995/09/11  22:11:40  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.22  1995/09/10  00:33:41  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.21  1995/09/07  01:34:57  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.20  1995/09/06  20:23:05  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.19  1995/09/01  01:18:28  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.18  1995/08/31  21:38:53  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.17  1995/08/25  01:38:18  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.16  1995/08/23  01:18:05  jcraig
 * jcraig on Tue Aug 22 18:16:19 PDT 1995
 *
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * use createPixrExtended instead of createPixrT in rotate90Pixr and
 * rotate270Pixr so that the width and height will get correctly
 * swapped.
 *
 * Revision 1.15  1995/08/21  17:31:00  ejaquith
 * ejaquith on Mon Aug 21 10:29:37 PDT 1995
 *
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Had introduced bug with first pass of changes where I wasn't correctly
 * re-setting the DPI resolutions after a 90/270 degree rotation.
 * Found 2 other additional routines that weren't setting the DPI as they
 * should, so all 4 instances fixed now.
 *
 * Revision 1.14  1995/08/17  22:02:54  ejaquith
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.13  1995/07/29  04:49:26  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.12  1995/07/14  18:17:19  ddavies
 * ddavies on Fri Jul 14 11:16:41 PDT 1995
 *
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Set x and y resolutions in output images in rotate90Pixr, rotate270Pixr,
 * rotate90InPlace and rotate270InPlace.
 *
 * Revision 1.11  1994/10/07  19:14:50  lperry
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.10  1994/09/16  17:39:51  jcraig
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.9  1994/06/07  02:40:34  macleod
 * macleod on Mon Jun 6 19:40:34 PDT 1994
 *
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.8  1994/05/13  18:20:42  ddavies
 * ddavies on Fri May 13 11:19:46 PDT 1994
 *
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Fix bug in rotate270InPlace.  When called with a composite color pixr,
 * it called rotate270 instead of rotate270InPlace on each plane.
 *
 * Revision 1.7  1994/05/07  23:43:08  jcraig
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * in rotate270Pixr call only createPixrExtended to get the destination
 * pixr - not createPixrExtended followed by a call to CREATE_PIXR.
 * fixes a purify complaint about memory leakage.
 *
 * Revision 1.6  1994/04/12  01:16:12  macleod
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.5  1994/03/02  04:00:05  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Added IP_YIELD to give up control after creating dest images during
 * rotateArb and rotateSmallAngleIP.
 *
 * Revision 1.4  1994/02/13  01:05:24  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Add rotate90InPlace, rotate180InPlace and rotate270InPlace.  These routines
 * rotate an image by the specified angle and return the result in the
 * same image buffer.  READ THE PROTOTYPES CAREFULLY before using.  Also
 * ripped 8 bit rotates out of rotate90, rotate180 and rotate270.  These are
 * now implemented in ipshared.
 *
 * Revision 1.3  1993/12/03  01:18:43  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:16:39  ddavies
 * rotatop1.c -> /product/ipcore/ipwrap/src/RCS/rotatop1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

