/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  runfilt.c -- Filtering of runs by length.
 *			
 *			w_initRunFilter ()
 *			w_finishRunFilter ()
 *			w_filterHorizontalRuns ()
 *			w_filterVerticalRuns ()
 *
 */

#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"     /* for cTrue and cFalse */
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "shrcc.pub"

#include "runfilt.prv"
#include "runfilt.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: runfilt.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates and initializes a state struct for use in run filtering an image.
 *
 * LongDescription:::
 *  If multiple run filters are used on the same source PIXR, use this
 *  function and w_finishRunFilter () to create and destroy the run filter
 *  state information.  This will save time and reduce fragmentation.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  inPr::           Source PIXR to be filtered
 *  filters::	     What type of filtering will be done to the
 *		     image.  Can be one of cVerticalRuns, cHorizontalRuns,
 *		     or cBothRuns
 *
 * Output:::
 *  runFilterState:: doubly indirect pointer which contains the run filter 
 *                   state stuff upon successful completion
 *
 * Returns:::
 *	cSuccessful
 *	cErrNomem
 *	cErrInvalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  filtering, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_initRunFilter (
                 IPWState  *pIPWState,
                 PIXR      *inPr,
		 Int32      filters,
		 void     **runFilterState)
{
    PROLOGUE;
    Int32 w, h;
    rfState *realState;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrFramedHeight (pIPWState, inPr, &h));
    ERR_PASS (w_pixrFramedWidth (pIPWState, inPr, &w));

    realState = (rfState *)IPW_MALLOC (sizeof (rfState), cAllocatorData);

    ERR_GEN ((realState == NULL), cErrNomem);

    ERR_PASS_CLEANUP (i_getRunFilterBufferSizes(pIPWState, w, h, filters, 
						&(realState->runArrayLength),
						&(realState->rotateBufferSize),
						&(realState->rotateBufferBpl)));

    ERR_GEN_CLEANUP ((realState->runArrayLength == 0), cErrInternal);

    /* allocate the run array */
    realState->runArray = (UInt8 *)IPW_MALLOC (realState->runArrayLength, 
                                               cAllocatorData);
    ERR_GEN_CLEANUP ((realState->runArray == NULL), cErrNomem);

    /* allocate the rotation buffer if necessary */
    if (realState->rotateBufferSize > 0)
    {
        realState->rotateBuffer = 
	              (UInt8 *)IPW_CALLOC (realState->rotateBufferSize, 1,
		                           cAllocatorData);
	ERR_GEN_CLEANUP ((realState->rotateBuffer == NULL), cErrNomem);
    }

    *runFilterState = realState;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE (realState->runArray, cAllocatorData);
    IPW_FREE (realState, cAllocatorData);
    ERR_RETURN;

}  /* w_initRunFilter () */


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Cleans up the run filter state info created by w_initRunFilter ().
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *
 * Output:::
 *  runFilterState:: doubly indirect pointer to state
 *		stuff to be freed.  Set to NULL.
 *
 * Returns:::
 *	cSuccessful
 *	cErrNomem
 *	cErrInvalidParm
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  filtering, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_finishRunFilter (
                   IPWState  *pIPWState,
		   void     **runFilterState)
{
    PROLOGUE;
    rfState *realState;

    realState = *runFilterState;

    /* or should it just return? */
    ERR_GEN ((*runFilterState == NULL), cErrInvalidParm);

    IPW_FREE (realState->runArray, cAllocatorData);
    if (realState->rotateBufferSize != 0)
	IPW_FREE (realState->rotateBuffer, cAllocatorData);
    IPW_FREE (realState, cAllocatorData);

    NORMAL_RETURN;

}  /* w_finishRunFilter () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Filter horizontal black or white runs by their length.
 *
 * LongDescription:::
 *  Provides the canonical set of filter operations: high pass, low
 *  pass, band pass, and band reject (notch); based on run length.  
 *  Creates the result PIXR if necessary.  Multi-modal filters can be 
 *  implemented by combining results of multiple runs with different 
 *  pass regions.  Can run "in-place" if source and destination are the 
 *  same PIXR, otherwise the source and destination PIXRs must be the 
 *  same dimensions.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  runFilterState:: Run filter state from w_initRunFilter ()
 *  inPr::           Source PIXR to be filtered
 *  runValue::       "Color" of runs to filter (0 = white, 1 = black)
 *  borderValue::    "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is filtered appropriately).
 *  minRunLength::   Minimum run length in filter band (inclusive)
 *  maxRunLength::   Maximum run length in filter band (inclusive)
 *  polarity::       Filter polarity can be
 *			cPassRuns - Runs which are within the band
 *			are kept, others are discarded, or
 *			cRejectRuns - Runs which are within the band
 *			are discarded, others are kept.
 *
 * Output:::
 *  outPr::          Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be combined with the results
 *	produced by this function (black runs are ORed into the
 *	destination, white runs are ANDed).  Be sure to set the
 *	output image (to white for black runs or to black for
 *	white runs) before calling i_filterHorizontalRuns if you 
 *	only want the result of a single filter operation.
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  filtering, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_filterHorizontalRuns (
    IPWState           *pIPWState,
    void               *runFilterState,
    PIXR               *inPr,
    Int32               runValue,
    Int32               borderValue,
    Int32               minRunLength,
    Int32               maxRunLength,
    Int32               polarity,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
						      * of images */
    Bool		 destCreated = cFalse,
                         stateCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;
    rfState		*realState = NULL;


    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

    if (runValue == 1)
        /* For filtering black runs we need a white destination PIXR */
	ERR_PASS (wp_createPixrConditional (pIPWState, wS, hS, 1,
					    xRes, yRes, cFrameBits,
					    cFalse, outPr, &rfc, &bfc, 
					    &destCreated));
    else
    {
        /* For filtering white runs, we need a black destination PIXR */
        ERR_PASS (wp_clonePixrConditionalNoInit (pIPWState, inPr, outPr,
	                                         &rfc, &bfc, &destCreated));

        if (destCreated)
	{
	    /* I want to paint it black */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET,
					  0, 0, wS, hS, *outPr));
    
	    /* Set the frame pixels to zero */
	    ERR_PASS_CLEANUP (w_setFramePixels (pIPWState, *outPr, 0));
	}
    }

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cHorizontalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_filterHorizontalRuns (pIPWState, realState->runArray,
                                              realState->runArrayLength,
                                              iS, 
			                      wS, hS,
			                      bplS, bplD,
				              runValue,
				              borderValue,
			                      minRunLength, maxRunLength,
			                      polarity,
				              iD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr, rfc, bfc));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);
    ERR_RETURN;

}  /* w_filterHorizontalRuns () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Filter vertical black or white runs by their length.
 *
 * LongDescription:::
 *  Provides the canonical set of filter operations: high pass, low
 *  pass, band pass, and band reject (notch); based on run length.  
 *  Creates the result PIXR if necessary.  Multi-modal filters can be 
 *  implemented by combining results of multiple runs with different 
 *  pass regions.  Can run "in-place" if sourceand destination are the 
 *  same PIXR, otherwise the source and destination PIXRs must be the 
 *  same dimensions.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  runFilterState:: Run filter state from w_initRunFilter ()
 *  inPr::           Source PIXR to be filtered
 *  runValue::       "Color" of runs to filter (0 = white, 1 = black)
 *  borderValue::    "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is filtered appropriately).
 *  minRunLength::   Minimum run length in band
 *  maxRunLength::   Maximum run length in band
 *  polarity::       Filter polarity can be
 *		     cPassRuns - Runs which are within the band
 *		     are kept, others are discarded, or
 *		     cRejectRuns - Runs which are within the band
 *		     are discarded, others are kept.
 *
 * Output:::
 *  outPr::          Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the output pixr is different from the input pixr,
 *        its original contents will be replaced with the results
 *        produced by this function.  This differs from
 *        w_filterHorizontalRuns ().
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  filtering, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_filterVerticalRuns (
    IPWState           *pIPWState,
    void               *runFilterState,
    PIXR               *inPr,
    Int32               runValue,
    Int32               borderValue,
    Int32               minRunLength,
    Int32               maxRunLength,
    Int32               polarity,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
						      * of images */
    Bool		 destCreated = cFalse,
                         stateCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;                    /* image pointers */
    rfState		*realState = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

	/* Since the output of vertical run finding doesn't depend
	 * on the original value of the dest, just create it
	 * uninitialized.
	 */
    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS, 1,
					      xRes, yRes, cFrameBits,
					      outPr, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cVerticalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_filterVerticalRuns (pIPWState, realState->runArray,
                                            realState->runArrayLength,
					    realState->rotateBuffer,
					    realState->rotateBufferBpl,
                                            iS, 
			                    wS, hS,
			                    bplS, bplD,
				            runValue,
				            borderValue,
			                    minRunLength, maxRunLength,
			                    polarity, 
				            iD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr, rfc, bfc));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);
    ERR_RETURN;

}  /* w_filterVerticalRuns () */



/*
 * $Log: runfilt.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.28  1998/01/14 00:15:14  ddavies
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.27  1997/06/26 20:04:39  ddavies
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Add inclusion of new file rast.pub.
 *
 * Revision 1.26  1997/04/15 22:37:20  ddavies
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Use non-initialized PIXR creation where possible.
 * Remove excess initialization of bitmap in w_filterVerticalRuns, which
 * doesn't care what value the PIXR was initialized to (looks like codes
 * was copied from w_filterHorizontalRuns, which does care).
 *
 * Revision 1.25  1996/05/29 20:18:44  dferrell
 * dferrell on Wed May 29 13:18:02 PDT 1996
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Use framed width and height when calling ipshared run filtering
 * functions.
 *
 * Revision 1.24  1996/05/22 22:16:49  dferrell
 * dferrell on Wed May 22 15:16:27 PDT 1996
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Fixed incorrect documentation comments.
 *
 * Revision 1.23  1996/03/21 23:22:07  dferrell
 * dferrell on Thu Mar 21 15:21:25 PST 1996
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Changed old memory allocation macros (MALLOC, CALLOC, FREE) to new
 * state-enabled macros (IPW_MALLOC...)
 *
 * Revision 1.22  1996/02/08 23:01:11  dferrell
 * dferrell on Thu Feb 8 14:59:14 PST 1996
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Use framed width and height in call to i_getRunFilterBufferSizes ().
 * Fix error checking logic in w_initRunFilter ().
 *
 * Revision 1.21  1996/02/08 00:39:07  dferrell
 * dferrell on Wed Feb 7 16:38:02 PST 1996
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Modified internals to deal with changes to IPShared.  Note that the
 * IPWrap function interfaces remain unchanged.
 *
 * Revision 1.20  1996/01/04  20:26:34  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.19  1995/12/13  04:12:47  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/12/12  20:22:25  dferrell
 * dferrell on Tue Dec 12 12:21:10 PST 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Changed interface to wrappers to reflect IPShared level rewrite of run filtering
 * code.  Added new state creation/destruction functions.
 *
 * Revision 1.17  1995/11/13  20:15:43  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  22:04:56  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/09/28  21:40:25  dferrell
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.14  1995/09/21  22:59:22  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/19  22:26:44  dferrell
 * dferrell on Tue Sep 19 15:25:10 PDT 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Updated functions to use new conditional pixr creation and right/
 * bottom frame clearing functionality.
 *
 * Revision 1.12  1995/09/15  22:38:42  ddavies
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.11  1995/09/11  21:21:11  dferrell
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.10  1995/09/07  01:35:03  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:23:11  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  17:07:12  dferrell
 * dferrell on Fri Sep 1 10:06:41 PDT 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.7  1995/08/31  21:38:58  ejaquith
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/23  22:09:24  jcraig
 * jcraig on Wed Aug 23 15:08:57 PDT 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * filterVerticalRuns should call i_filterVerticalRuns not
 * i_filterHorizontalRuns.
 *
 * Revision 1.5  1995/08/17  21:55:02  dferrell
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.4  1995/04/14  16:51:54  dferrell
 * dferrell on Fri Apr 14 09:51:38 PDT 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Added IP_RCSINFO stuff at top of file.
 *
 * Revision 1.3  1995/01/12  23:46:42  dferrell
 * dferrell on Thu Jan 12 15:45:39 PST 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Corrected comment in filterVerticalRuns.
 *
 * Revision 1.2  1995/01/12  19:37:00  dferrell
 * dferrell on Thu Jan 12 11:36:30 PST 1995
 *
 * runfilt.c -> /product/ipcore/ipwrap/src/RCS/runfilt.c,v
 *
 * Added more comments and RCS log at bottom.
 *
 * Revision 1.1  1995/01/11  14:44:58  dferrell
 * dferrell on Wed Jan 11 14:44:58 PST 1995
 *
 * i_runflt.c -> /product/ipcore/ipshared/src/RCS/i_runflt.c,v
 *
 * New file containing ipwrap level interface to horizontal and vertical run
 * filtering operations.
 *
 */

