/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * runhist1.c
 *
 * Functions defined:
 *
 *	w_horizRunLengthHistogram ()
 */




#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"

#include "pixr.h"
#include "geomadt.h"
#include "props.pub"
#include "shrcc.pub"

#include "runhist.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: runhist1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a histogram of horizontal runs (1 and 0) or cycles in a binary PIXR
 *
 * LongDescription:::
 *  Can create a black/white run length histogram set or a cycle length
 *  histogram.  The histogram array is neither allocated nor cleared by 
 *  this function.  This version has extra parameters which allow for: runs 
 *  at the beginning and end of lines to be handled differently than 
 *  "internal" runs, runs which exceed the maximum histogram size to be 
 *  either ignored or truncated to the maximum length, and runs to be 
 *  weighted by their length in the resulting histogram.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  inPixr::         Binary input PIXR
 *  lineSkip::       Number of lines to skip after each line, 0=do every line
 *  rect::           Region to process. If NULL, do the whole image
 *  runType::        if equal to cRunOfZeros, count runs of 0's.
 *                   if equal to cRunOfOnes, count runs of 1's
 *                   if equal to cRunOfCycles, count cycles (adjacent
 *                   pairs of runs of 0's and runs of 1's).
 *  numHistoBins::   number of bins in the histogram pHistos
 *  useEndRuns::     If true, count the runs that abut the ends of the
 *                   rectangle.  If false, ignore them.  When counting
 *                   cycles, this parameter must always be set to false
 *  countLongRuns::  If true, put the counts for all the runs that
 *                   don't fit into the histogram into the last histogram
 *                   bin.  If false, ignore these runs.
 *  weightByLength:: If true, each run add the length of the run to
 *                   its histogram bin.  If false, each run adds
 *                   1 to the corresponding bin.
 *
 * Output:::
 *  pHistos::        The resulting histogram(s)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, run length, cycle, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_horizRunLengthHistogram (
    IPWState           *pIPWState,
    PIXR               *inPixr,
    Int32               lineSkip,
    RECTANGL           *rect,
    Int32               runType,
    Int32               numHistoBins,
    Int32               useEndRuns,
    Int32               countLongRuns,
    Int32               weightByLength,
    Int32              *pHistos)
{
    PROLOGUE;
    Int32   x, y, w, h, bpl, fr, wP, hP;
    UInt8  *iS;                         /* image pointer */

    ERR_GEN_PIXR (inPixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN (pHistos == NULL, cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, inPixr, &wP));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPixr, &hP));

    if (rect == NULL)
    {
	x = 0;
	y = 0;
	w = wP;
	h = hP;
    }
    else
    {
	x = rect->x;
	y = rect->y;
	w = rect->w;
	h = rect->h;
    }

    /* make sure the rectangle has area and is on the image */
    ERR_GEN (((x < 0) || (y < 0)), cErrInvalidParm);
    ERR_GEN (((w <= 0) || (h <= 0)), cErrInvalidParm);
    ERR_GEN ((((x+w) > wP) ||
	      ((y+h) > hP)), cErrInvalidParm);

    ERR_PASS (w_pixrGetBpl (pIPWState, inPixr, &bpl));
    ERR_PASS (w_pixrGetFrame (pIPWState, inPixr, &fr));
    ERR_PASS (w_pixrGetImageAll (pIPWState, inPixr, &iS));

    /* Clear the histogram */
    MEMSET ((void *)pHistos, 0, numHistoBins * sizeof(Int32));

    /* Call the ipshared function */
    ERR_PASS (i_horizRunHistogram (pIPWState, iS,
                                   bpl,
				   x + fr,
				   y + fr,
				   w, h,
				   pHistos, numHistoBins,
				   lineSkip + 1,
				   runType, useEndRuns,
				   countLongRuns, weightByLength));

    NORMAL_RETURN;

}  /* w_horizRunLengthHistogram () */


/*

 $Log: runhist1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.18  1998/01/14 00:15:16  ddavies
 runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.17  1996/01/04 20:26:40  ejaquith
 runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.16  1995/12/13  04:12:54  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.15  1995/11/14  22:18:44  ddavies
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Fix documentation bugs found by Eric's scripts.
 *
 * Revision 1.14  1995/11/13  20:15:49  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.13  1995/10/13  22:05:02  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.12  1995/09/28  21:40:39  dferrell
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.11  1995/09/21  22:59:28  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.10  1995/09/11  21:21:30  dferrell
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.9  1995/09/07  01:35:09  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.8  1995/09/06  20:23:16  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.7  1995/09/01  17:07:32  dferrell
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.6  1995/08/31  21:39:03  ejaquith
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/17  21:55:15  dferrell
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.4  1995/05/09  03:19:03  ddavies
 * runhist1.c -> /product/ipcore/ipwrap/src/RCS/runhist1.c,v
 *
 * Added RCSINFO and Log macros.
 *

 */

