/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  runfilt.c -- Run-based morphology
 *
 *	w_dilateHorizontalRuns ()
 *	w_dilateVerticalRuns ()
 *	w_erodeHorizontalRuns ()
 *	w_erodeVerticalRuns ()
 */

#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"     /* for cTrue and cFalse */
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrcc.pub"

#include "runfilt.prv"     /* for state */
#include "runfilt.pub"
#include "runmorph.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: runmorph.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dilate horizontal black runs.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  runFilterState:: Run filter state from w_initRunFilter ()
 *  inPr::           Source PIXR to be filtered
 *  strelWidth::     Width of dilation "strel."  The strel is a
 *		     solid horizontal line of hits.
 *  strelCenter::    Center location of "strel."  0 -> left end
 *		     and strelWidth - 1 -> right end.
 *
 * Output:::
 *  outPr::          Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be combined (ORed) with the results
 *	produced by this function.  Be sure to set the
 *	output image (to white for black runs or to black for
 *	white runs) before calling dilateHorizontalRuns if you 
 *	only want the result of a single morphological operation.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_dilateHorizontalRuns (
    IPWState           *pIPWState,
    void               *runFilterState,
    PIXR               *inPr,
    Int32      		strelWidth,
    Int32      		strelCenter,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
						      * of images */
    Bool		 destCreated = cFalse,
                         stateCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;
    rfState		*realState = NULL;


    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

    /* For dilating black runs we need a white destination PIXR */
    ERR_PASS (wp_createPixrConditional (pIPWState, wS, hS, 1, xRes, yRes,
				        cFrameBits, cFalse, outPr,
					&rfc, &bfc, &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cHorizontalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_dilateHorizontalRuns (pIPWState, realState->runArray,
                                              realState->runArrayLength,
                                              iS, 
			                      wS, hS,
			                      bplS, bplD,
					      strelWidth,
					      strelCenter,
				              iD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr, rfc, bfc));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);
    ERR_RETURN;

}  /* w_dilateHorizontalRuns () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dilate vertical black runs.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  runFilterState:: Run filter state from w_initRunFilter ()
 *  inPr::           Source PIXR to be filtered
 *  strelHeight::    Height of dilation "strel."  The strel is a
 *		     solid vertical line of hits.
 *  strelCenter::    Center location of "strel."  0 -> top end
 *		     and strelWidth - 1 -> bottom end.
 *
 * Output:::
 *  outPr::          Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the output pixr is different from the input pixr,
 *        its original contents will be replaced with the results
 *        produced by this function.  This differs from
 *        w_dilateHorizontalRuns ().
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_dilateVerticalRuns (
    IPWState           *pIPWState,
    void               *runFilterState,
    PIXR               *inPr,
    Int32      		strelHeight,
    Int32      		strelCenter,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
						      * of images */
    Bool		 destCreated = cFalse,
                         stateCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;                    /* image pointers */
    rfState		*realState = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

    /* For dilating black runs we need a white destination PIXR */
    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS, 1, xRes, yRes,
					      cFrameBits, outPr, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cVerticalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_dilateVerticalRuns (pIPWState, realState->runArray,
                                            realState->runArrayLength,
					    realState->rotateBuffer,
					    realState->rotateBufferBpl,
                                            iS, 
			                    wS, hS,
			                    bplS, bplD,
					    strelHeight,
					    strelCenter,
				            iD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr, rfc, bfc));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);
    ERR_RETURN;

}  /* w_dilateVerticalRuns () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Erode horizontal black runs.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  runFilterState:: Run filter state from w_initRunFilter ()
 *  inPr::           Source PIXR to be filtered
 *  borderValue::    "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is filtered appropriately).
 *  strelWidth::     Width of erosion "strel."  The strel is a
 *		     solid horizontal line of hits.
 *  strelCenter::    Center location of "strel."  0 -> left end
 *		     and strelWidth - 1 -> right end.
 *
 * Output:::
 *  outPr::          Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the output image pointer is different from the 
 *	input image pointer, the original contents of the
 *	output image will be combined (ORed) with the results
 *	produced by this function.  Be sure to set the
 *	output image (to white for black runs or to black for
 *	white runs) before calling erodeHorizontalRuns if you 
 *	only want the result of a single morphological operation.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_erodeHorizontalRuns (
    IPWState           *pIPWState,
    void               *runFilterState,
    PIXR               *inPr,
    Int32		borderValue,
    Int32      		strelWidth,
    Int32      		strelCenter,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
						      * of images */
    Bool		 destCreated = cFalse,
                         stateCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;
    rfState		*realState = NULL;


    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

    /* For eroding black runs we need a white destination PIXR */
    ERR_PASS (wp_createPixrConditional (pIPWState, wS, hS, 1, xRes, yRes,
					cFrameBits, cFalse, outPr,
					&rfc, &bfc, &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cHorizontalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_erodeHorizontalRuns (pIPWState, realState->runArray,
                                             realState->runArrayLength,
                                             iS, 
			                     wS, hS,
			                     bplS, bplD,
					     borderValue,
					     strelWidth,
					     strelCenter,
				             iD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr, rfc, bfc));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);
    ERR_RETURN;

}  /* w_erodeHorizontalRuns () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Erode vertical black runs.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  runFilterState:: Run filter state from w_initRunFilter ()
 *  inPr::           Source PIXR to be filtered
 *  borderValue::    "Color" of border beyond the image (0 = white,
 *                      1 = black).  The border is assumed to be infinite
 *                      for run length calculations (e.g. a black run which
 *			abuts a black border is assumed to be infinitely
 *			long and is filtered appropriately).
 *  strelHeight::    Height of erosion "strel."  The strel is a
 *		     solid vertical line of hits.
 *  strelCenter::    Center location of "strel."  0 -> top end
 *		     and strelWidth - 1 -> bottom end.
 *
 * Output:::
 *  outPr::          Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  (1):: If the output pixr is different from the input pixr,
 *        its original contents will be replaced with the results
 *        produced by this function.  This differs from
 *        w_erodeHorizontalRuns ().
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_erodeVerticalRuns (
    IPWState           *pIPWState,
    void               *runFilterState,
    PIXR               *inPr,
    Int32		borderValue,
    Int32      		strelHeight,
    Int32      		strelCenter,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32		 wS, hS, bplS, bplD; /* width, height, bpl
						      * of images */
    Bool		 destCreated = cFalse,
                         stateCreated = cFalse;
    ScaledInt1616	 xRes, yRes;
    Int32                rfc, bfc;
    UInt8		*iS, *iD;                    /* image pointers */
    rfState		*realState = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*outPr != NULL)
        ERR_GEN_PIXR (*outPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hS));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xRes, &yRes));

    /* For eroding black runs we need a white destination PIXR */
    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wS, hS, 1, xRes, yRes,
					      cFrameBits, outPr, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cVerticalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_erodeVerticalRuns (pIPWState, realState->runArray,
                                           realState->runArrayLength,
					   realState->rotateBuffer,
					   realState->rotateBufferBpl,
                                           iS, 
			                   wS, hS,
			                   bplS, bplD,
				           borderValue,
					   strelHeight,
					   strelCenter,
				           iD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *outPr, rfc, bfc));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, outPr);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);
    ERR_RETURN;

}  /* w_erodeVerticalRuns () */


/*
 * $Log: runmorph.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.4  1998/01/14 00:15:19  ddavies
 * runmorph.c -> /product/ipcore/ipwrap/src/RCS/runmorph.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.3  1997/04/15 22:37:46  ddavies
 * runmorph.c -> /product/ipcore/ipwrap/src/RCS/runmorph.c,v
 *
 * Fix indentation.
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.2  1996/05/29 20:19:18  dferrell
 * runmorph.c -> /product/ipcore/ipwrap/src/RCS/runmorph.c,v
 *
 * Use framed width and height when calling ipshared run morphology
 * functions.
 *
 * Revision 1.1  1996/02/20 23:57:16  dferrell
 * Initial revision
 *
 *
 */

