/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  runproj.c -- Orthogonal run projection routines.
 *
 *			PUBLIC
 *				w_projectHorizontalRuns ()
 *				w_projectVerticalRuns ()
 *
 */

#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"     /* for cTrue and cFalse */
#include "geomadt.h"
#include "props.pub"
#include "shrcc.pub"

#include "runfilt.prv"
#include "runfilt.pub"
#include "runproj.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: runproj.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Create a horizontal projection of black or white runs.
 *
 * LongDescription:::
 *	Includes a run pre-filter which allows selection of a band of
 *	runs to be projected.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		IPWrap state pointer
 *  runFilterState::    Run filter state from w_initRunFilter (), or
 *			NULL.
 *  inPr::		Binary input image.
 *  runValue::		1 = sum of blackness, 0 = sum of whiteness.
 *  includeEndRuns::    If cTrue, project all runs, if cFalse, ignore
 *                      runs which begin or end at the left or right 
 *			edge of the image.
 *  minRunLength::      Minimum run length in filter band (inclusive)
 *  maxRunLength::      Maximum run length in filter band (inclusive)
 *  polarity::          Filter polarity can be
 *		        cPassRuns - Runs which are within the band
 *		        are kept, others are discarded, or
 *		        cRejectRuns - Runs which are within the band
 *		        are discarded, others are kept.
 *
 * Output:::
 *  projection::	Array of projection data, created if NULL.
 *
 * Returns:::
 *	cSuccessful
 *	cErrNomem
 *	cErrInvalidParm
 *
 * Notes:::
 *			Projection results are cumulative between calls,
 *			so for individual use, be sure to clear the
 *			projection array between calls.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  filtering, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_projectHorizontalRuns (
		IPWState  *pIPWState,
                void      *runFilterState,
		PIXR      *inPr, 
		Int32      runValue,
		Int32      includeEndRuns,
		Int32      minRunLength,
		Int32      maxRunLength,
		Int32      polarity,
		Int32    **projection)
{
    PROLOGUE;
    Int32       w, h, bpl;
    Bool	stateCreated = cFalse,
                projectionCreated = cFalse;
    UInt8	*iS;
    rfState	*realState = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrFramedHeight (pIPWState, inPr, &h));
    ERR_PASS (w_pixrFramedWidth (pIPWState, inPr, &w));
    ERR_PASS (w_pixrGetBpl (pIPWState, inPr, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));

    if ((*projection) == NULL)
    {
        *projection = (Int32 *)IPW_CALLOC (sizeof (Int32), h, cAllocatorData);
	ERR_GEN ((*projection) == NULL, cErrNomem);
	projectionCreated = cTrue;
    }

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cHorizontalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_projectHorizontalRuns (pIPWState, realState->runArray,
                                               realState->runArrayLength,
                                               iS, 
                                               w, h,
				               1, bpl,
				               runValue,
				               includeEndRuns,
					       minRunLength,
					       maxRunLength,
					       polarity, 0,
				               *projection));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:

    if (projectionCreated)
        IPW_FREE (*projection, cAllocatorData);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);

    ERR_RETURN;

}  /* w_projectHorizontalRuns () */


/**************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Create a vertical projection of black or white runs.
 *
 * LongDescription:::
 *	Includes a run pre-filter which allows selection of a band of
 *	runs to be projected.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		IPWrap state pointer
 *  runFilterState::    Run filter state from w_initRunFilter (), or
 *			NULL.
 *  inPr::		Binary input image.
 *  runValue::		1 = sum of blackness, 0 = sum of whiteness.
 *  includeEndRuns::    If cTrue, project all runs, if cFalse, ignore
 *                      runs which begin or end at the top or bottom edge
 *			of the image.
 *  minRunLength::      Minimum run length in filter band (inclusive)
 *  maxRunLength::      Maximum run length in filter band (inclusive)
 *  polarity::          Filter polarity can be
 *		        cPassRuns - Runs which are within the band
 *		        are kept, others are discarded, or
 *		        cRejectRuns - Runs which are within the band
 *		        are discarded, others are kept.
 *
 * Output:::
 *  projection::	Array of projection data, created if NULL.
 *
 * Returns:::
 *	cSuccessful
 *	cErrNomem
 *	cErrInvalidParm
 *
 * Notes:::
 *			Projection results are cumulative between calls,
 *			so for individual use, be sure to clear the
 *			projection array between calls.
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  filtering, run length, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_projectVerticalRuns (
		IPWState  *pIPWState,
                void      *runFilterState,
		PIXR      *inPr, 
		Int32      runValue,
		Int32      includeEndRuns,
		Int32      minRunLength,
		Int32      maxRunLength,
		Int32      polarity,
		Int32    **projection)
{
    PROLOGUE;
    Int32       w, h, bpl;
    Bool	stateCreated = cFalse,
                projectionCreated = cFalse;
    UInt8	*iS;
    rfState	*realState = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrFramedHeight (pIPWState, inPr, &h));
    ERR_PASS (w_pixrFramedWidth (pIPWState, inPr, &w));
    ERR_PASS (w_pixrGetBpl (pIPWState, inPr, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));

    if ((*projection) == NULL)
    {
        *projection = (Int32 *)IPW_CALLOC (sizeof (Int32), w, cAllocatorData);
	ERR_GEN ((*projection) == NULL, cErrNomem);
	projectionCreated = cTrue;
    }

    if (runFilterState == NULL)
    {
        ERR_PASS_CLEANUP (w_initRunFilter (pIPWState, inPr, cVerticalRuns,
	                                   &runFilterState));
	stateCreated = cTrue;
    }

    realState = (rfState *)runFilterState;

    ERR_PASS_CLEANUP (i_projectVerticalRuns (pIPWState, realState->runArray,
                                             realState->runArrayLength,
					     realState->rotateBuffer,
					     realState->rotateBufferBpl,
                                             iS, 
                                             w, h,
				             1, bpl,
				             runValue,
				             includeEndRuns,
					     minRunLength,
					     maxRunLength,
					     polarity, 0,
				             *projection));

    if (stateCreated)
        ERR_PASS_CLEANUP (w_finishRunFilter (pIPWState, &runFilterState));

    NORMAL_RETURN;

ERR_CLEANUP:

    if (projectionCreated)
        IPW_FREE (*projection, cAllocatorData);
    if (stateCreated)
        w_finishRunFilter (pIPWState, &runFilterState);

    ERR_RETURN;

}  /* w_projectVerticalRuns () */


/*
 * $Log: runproj.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 * Imported sources
 *
 * Revision 1.3  1998/01/14 00:15:21  ddavies
 * runproj.c -> /product/ipcore/ipwrap/src/RCS/runproj.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.2  1996/05/30 17:10:52  dferrell
 * dferrell on Thu May 30 10:10:45 PDT 1996
 *
 * runproj.c -> /product/ipcore/ipwrap/src/RCS/runproj.c,v
 *
 * Changed return types from Int32 to ErrCode.
 *
 * Revision 1.1  1996/05/29 20:19:40  dferrell
 * Initial revision
 *
 */

