/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * sharpen.c
 *
 *	Image sharpening, smoothing, etc.
 *
 * Exported:
 *
 *	w_sharpen () - Sharpen an image 
 *	w_smooth () - Blur (smooth) an image 
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "defines.pub"
#include "props.pub"
#include "props.prv"
#include "sharpen.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: sharpen.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Enhance edges in gray and color images.
 *
 * LongDescription:::
 *  Gray and color images often look better when the edges of objects are
 *  made more distinct.  This is called "sharpening" an image.  This
 *  routine performs this function by adjusting intensities locally so
 *  that edges stand out more.
 *<BR><BR>
 *  Unfortunately, this also enhances noise.  Caveat Emptor.
 *
 * PrivateDescription:::
 *  This is done using a simple convolution with a 3x3 kernel.  The values
 *  in the kernel are chosen as a function of the sharpness parameter.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to gray or color image to be sharpened.
 *  sharpness::      The sharpness value.  This number must be in the
 *		     range [0,3].  0 => not much sharpness; 3 => lots
 *		     of sharpness.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  If none is
 *		     provided, one will be created.  If a dest image is
 *		     provided, its dimensions, depth, etc will be checked.
 *		     The dest image should be a clone of the source image.
 *		     However, it is not possible for the dest image and the
 *		     source image to be the same image.  Doing this would
 *		     overwrite source pixels before we would be finished
 *		     with them.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if sharpness is not in the range [0,3].  It also returns cErrInvalidParm
 *  if prS == *pPrD.
 *
 * Notes:::
 *  Do NOT perform this operation on an indexed color image (each pixel
 *  is really an index into a color map).  The results are often entertaining,
 *  but seldom useful.
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  filtering, gray, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_sharpen (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               sharpness,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    PIXR		*subS, *subD;
    Int32		 wS, hS, channels, i;
    Int32		 bplS, bplD;
    UInt8		*pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrPixr_null);
    ERR_GEN((sharpness < 0) || (sharpness > 3), cErrInvalidParm);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1 | cChannels3,
		 cFrame32);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    for (i = 0; i < channels; i++)
    {
	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &subS));
	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &subD));

	ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subD, &pD));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, subS, &bplS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, subD, &bplD));
	ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, subS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, subS, &hS));

	ERR_PASS_CLEANUP( i_sharpen (pIPWState, pS, bplS,
				     pD, bplD,
				     wS, hS, sharpness));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Blur edges in gray and color images.
 *
 * LongDescription:::
 *  Gray and color images sometimes look better when the edges of objects are
 *  blurred.  This is often true when the edges are the result of random
 *  noise or moire effects (scaling of images containing strong periodic
 *  components often produces ripples).  Performing the blurring operation
 *  is called "smoothing" an image.  This routine performs this function
 *  by adjusting intensities locally so that local intensity differences
 *  are diminished..
 *<BR><BR>
 *  Unfortunately, this also makes things look fuzzy.  Caveat Emptor.
 *
 * PrivateDescription:::
 *  This is done using a simple convolution with a 3x3 kernel.  The values
 *  in the kernel are chosen as a function of the smoothness parameter.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to gray or color image to be smoothed.
 *  smoothness::     The smoothness value.  This number must be in the
 *		     range [0,3].  0 => not smoothed much; 3 => smoothed a lot.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  If none is
 *		     provided, one will be created.  If a dest image is
 *		     provided, its dimensions, depth, etc will be checked.
 *		     The dest image should be a clone of the source image.
 *		     However, it is not possible for the dest image and the
 *		     source image to be the same image.  Doing this would
 *		     overwrite source pixels before we would be finished
 *		     with them.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if sharpness is not in the range [0,3].  It also returns cErrInvalidParm
 *  if prS == *pPrD.
 *
 * Notes:::
 *  Do NOT perform this operation on an indexed color image (each pixel
 *  is really an index into a color map).  The results are often entertaining,
 *  but seldom useful.
 *
 * FunctionCategory:::
 *  filtering
 *
 * Keywords:::
 *  filtering, gray, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_smooth (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               smoothness,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    PIXR	*subS, *subD;
    Int32	 wS, hS, channels, i;
    Int32	 bplS, bplD;
    UInt8	*pS, *pD;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrPixr_null);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1 | cChannels3,
		 cFrame32);

    ERR_PASS_CLEANUP( w_pixrGetChannelCount (pIPWState, prS, &channels));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

    for (i = 0; i < channels; i++)
    {
	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, i, &subS));
	ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, i, &subD));

	ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subD, &pD));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, subS, &bplS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, subD, &bplD));
	ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, subS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, subS, &hS));

	ERR_PASS_CLEANUP( i_smooth (pIPWState, pS, bplS,
				    pD, bplD,
				    wS, hS, smoothness));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}

/*

 $Log: sharpen.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.25  1998/01/14 00:15:22  ddavies
 sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.24  1997/04/15 22:41:16  ddavies
 ddavies on Tue Apr 15 15:41:15 PDT 1997

 sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.23  1996/07/22 18:53:58  macleod
 macleod on Mon Jul 22 11:53:54 PDT 1996

 sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v

 w_sharpen and w_smooth were forcing images to be single-channel,
 even though they both support multi-channel images.

 Revision 1.22  1996/01/04 20:26:44  ejaquith
 sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.21  1995/12/13  04:13:01  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:15:55  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:05:09  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:30:48  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 *
 * Revision 1.17  1995/09/21  22:59:34  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/15  18:02:03  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.15  1995/09/11  22:12:35  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.14  1995/09/10  00:35:50  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.13  1995/09/07  01:35:15  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:23:22  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  01:18:41  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:39:09  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/17  22:03:51  ejaquith
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.8  1995/07/29  04:50:16  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.7  1994/10/07  19:19:27  lperry
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.6  1994/06/08  04:13:43  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.5  1994/06/07  02:40:44  macleod
 * macleod on Mon Jun 6 19:40:43 PDT 1994
 *
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.4  1994/04/12  01:16:21  macleod
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.3  1993/12/03  01:18:53  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:16:54  ddavies
 * sharpen.c -> /product/ipcore/ipwrap/src/RCS/sharpen.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

