/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  shearops.c: contains subroutines for shear and non-interpolated rotation
 *
 *     Procedures that allocate image memory:
 *        w_rotateArb()
 *
 *     In-place rotation:
 *        wp_rotateSmallAngleIP()
 *        wp_rotateLargeAngleIP()
 *
 *     Shearing routines:
 *	  wp_rotate2C
 *	  wp_rotate3C
 *	  wp_shearC
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include <math.h>
#include "defines.pub"
#include "frames.pub"
#include "alplib.h"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "rast.pub"
#include "rotatop.pub"
#include "shrrast.pub"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: shearops.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

#define  SMALL_ANGLE        0.12    /* radians; about 7 degrees */
#define  MIN_ANGLE	    0.002   /* radians; about 0.1 degrees */

    /* shrink options */
#define  NO_SHRINK           1   /* fastest for small angles; slower for very
				 * large angles; leaves white space
				 * around image */

/* This file contains procedures used to perform non-interpolated rotations.
 * w_rotateArb rotates by any angle between -PI/4 and PI/4 (-45, +45).
 * The result image may be the same as the source image (in place rotation) or
 * different.  If it's different, it may be expanded to retain all the
 * pixels of the original or may be kept the same size as the original.
 *
 * As mentioned above, w_rotateArb has two output options, resizing
 * the output image to be the same size as the source image and making
 * a bigger output image so that none of the corners of the input image
 * are lost.  The problem with the second option is that it can use a
 * very large amount of memory.
 *
 * The large angle rotation requires three shears.  For largish angles (around
 * 45 degrees), the output image for each shear is about twice as big
 * (required twice as many pixels) as the input image.  Say we start with
 * an 8 MB image.  The result of the first stage is a 16 MB image.  The
 * result of the second stage requires 32 MB and the third stage takes
 * 64 MB.  Each memory block must be contiguous in virtual space.
 * When I tried this on my SPARC, the machine thrashed its disk into
 * a froth for a minute or so, then was unable to allocate
 * the third image.  Remember that the original 8 MB image, the final
 * 8 MB image and the 32MB image already exist (at least).
 *
 * Furthermore, the way we implemented the option that returned an image
 * that was the same size as the input image was to cut a section out
 * of the center of the last image and return that.  This means that even
 * if we were planning on discarding all the other stuff from the beginning,
 * we still compute all those bits and consume the memory required to hold
 * them.
 *
 * We can probably do better in general, but I'll just do something for the
 * special case in which the input and output images are the same.
 * This is the most common usage of this routine anyway.  Well, ok,
 * there's also an optimization for rotating by small angles where
 * the separate output image is the same size as the input - just call
 * i_rotateNoInterpolation.
 *
 * There are two sub-cases of in place rotation: small and large angles.
 * For small angles, we always call wp_rotateSmallAngleIP - no shearing
 * involved.  For large angles, we call wp_rotateLargeAngleIP, which uses
 * three shears.
 *
 * What are the three shears used in the original code?  This is the
 * version that makes the image larger so it retains all bits that were
 * in the original.  Let's illustrate with a clockwise rotation.
 *
 * The first shear is horizontal.  It makes the image wider.
 *
 *	_________			_______________
 *	|	|			|  |	   |  |
 *	|	|			|  | --->  |  |
 *	|	|	step 1 ->	|  |       |  |
 *	|	|			|  | <---  |  |
 *	|	|			|  |	   |  |
 *	---------			---------------
 *
 * The second shear is vertical.  It makes the wider image taller.
 *
 *					- - - - - - - -
 *					|	      |
 *	_______________			_______________
 *	|	      |			|	      |
 *	|	      |			|  ^	      |
 *	|	      |	  step 2 ->	|  |   	  |   |
 *	|	      |			|     	  v   |
 *	|	      |			|	      |
 *	---------------			---------------
 *					|	      |
 *					_______________
 *
 * The third shear is horizontal again.  It makes the taller, wider
 * image wider yet again.  The extra width is a function of the width
 * of the step 2 image, not the width of the original.  That's because
 * the routine doesn't know whether it's being called the first time
 * or the third time.
 *
 *	_______________		____________________________
 *	|	      |		|     |             |      |
 *	|	      |		|     |             |      |
 *	|	      |		|     |             |      |
 *	|	      |		|     |   ----->    |      |
 *	|	      |	step 3	|     |             |      |
 *	|	      |	---->	|     |   <-----    |      |
 *	|	      |		|     |             |      |
 *	|	      |		|     |             |      |
 *	|	      |		|     |             |      |
 *	---------------		----------------------------
 *
 * If we wanted the output and input images to be the same size in the
 * original code, we would cut the middle out of the final image and return
 * it.
 *
 * At first glance, this is sort of silly if we really just want an image
 * that's the same size as the original.  Can't we just discard the bits
 * that are shifted off the original image at each step?
 *
 * Mostly, but not completely.  Look at the image bits that were shifted
 * into the new region in the upper right corner in step 1.  Some of those
 * bits will move down to the lower right quadrant in step 2 and back onto
 * the original image in step 3.  The same thing happens to bits that are
 * shifted off to the lower left in step 1, move to the upper left in step
 * 2 and move back on the original image in step 3.  All the other bits
 * (roughly 7/8 of the space in the final image) are tossed.
 *
 * Our solution to this is to allocate two "wings" to the original image.
 * These just catch the bits falling off the edge that we'll need later.
 * I spent time coming up with all sorts of really clever ways of saving
 * space in the wings, like using ones that were half height and cleverly
 * shuffling all the bits around.  They got weird and had lots of extra
 * setting and clearing.  We'll just be dumb and use full height
 * wings.  Here's the new scheme:
 *	_________			_________________
 *	|	|			|   |	    |   |
 *	|	|			|   | ------->  |
 *	|	|	step 1 ->	|   |       |   |
 *	|	|			|   |	    |   |
 *	|	|			|  <------  |   |
 *	|_______|			|___|_______|___|
 *
 * Note that in the shearing steps, we save bits that move off the image
 * into the wings, but don't bother writing bits from the wings onto
 * the image.  We know that all wing bits are white initially, so can
 * just set the cleared portions of the image to white.
 *
 * In the new step 2, we do the vertical shear inside the original image
 * as before.  Bits shifted off the top and bottom are discarded.  Bits
 * shifted in from the top and bottom are set to white.  We continue the
 * shears out into the wings with the same rules.
 *
 *	_________________
 *	| ^ |	    |   |
 *	| | | ^	 |  | | | step 2
 *	| | | |	 |  | | |
 *	| | | |	 V  | | |
 *	| | |	    | V |
 *	|___|_______|___|
 *
 * In step 3, we perform a shear like the one performed in step 1 except
 * that we import bits onto the image from the wings and discard any bits
 * that would have fallen off the image.
 *
 * There are a few more points to cover.  First, our rendition of w_rasterop
 * always moves through the image from left to right, top to bottom.
 * That means that if we want to move a block of memory to the left and/or
 * down AND the block of destination memory locations overlaps the
 * block of source memory locations (as normally happens in a shear)
 * source bits from the upper left will overwrite source bits on the
 * lower right before those source bits are read.  This means the results
 * of the operation will be incorrect.
 *
 * We really only move either purely horizontally or purely vertically.
 * When moving to the left, we copy each block into
 * the frame first, then blit them into the image as required.
 * Will the blocks fit?  Well, the smallest large angle is 7 degrees.
 * tan(7 degrees/2) = 0.00349.  The number of lines in a slice is
 * 1/0.00349 = 29.  This is less thatn 32 (the number of lines in the
 * frame, so we're safe.  We do have to remember to clear the frame
 * where we're done.
 *
 * When we move vertically, we reset the image pointer to point to the
 * bottom of the image and negate the image's bytes/line.  This lets
 * us pretend that we're moving "up" instead of down so w_rasterop works
 * correctly.
 *
 * A smaller point has to do with setting portions of the image to
 * white when they correspond to bits that were shifted in from off the
 * page.  We first notice that we don't need to bother in the wings.
 * In step 1, they're receiving bits from the image, so we don't care.
 * In step 2, we're writing bits from the upper half of a wing to the
 * lower, or vice versa, and don't care what the old half of the wing
 * looks like anyway.  In the third step, we don't care what the wing
 * looks like after the shear, so don't need to bother fixing anything.
 *
 * So what happens in the proper image?  Filling in after horizontal
 * shears is easy.  We just do a blit from the edge of the new block
 * to the edge of the image.  Conceptually, it's easy in
 * vertical shears too, except that thin vertical blits are very slow.
 * We'll use horizontal blits in this case too.  Say we work from left
 * to right across the image and we're rotating clockwise so the blocks
 * on the left are moved up and those on the right go down. After moving
 * the first block up, we do nothing.  After moving the second block up
 * (ends one scan line below the first block), we clear all the bits to
 * the left of the left end of the lowest line of the second block.
 * The same thing happens for the third block and so on.  So long as
 * we're on the left side of the image, this works because we're writing
 * over bits that have already been moved.
 *
 * So what happens when we get to the right side of the image?  It's
 * tempting to start from the right edge and work inwards.  The only
 * tiny problem is knowing where to put the vertical shear lines.
 * We base each shear position on the previous position.  If we worked
 * in from both the left and right, things would look strange in the
 * middle of the image.  So, we have a choice.  We can either just
 * run a loop that steps out to the right side of the image, then steps
 * back doing the shears and clears on the way or we can do the shears
 * while we're stepping out and the clears while we're stepping back.
 * Just to make the code look sort of coherent, we'll choose the first
 * option.
 *
 * Note also that vertical shear block will probably span the border
 * between the regular image and the wings.  We'll just have to remember
 * how much is on each image.
 *
 * Finally, one relatively small point.  When doing the vertical shears
 * in the wings, we really only care about bits that move across the
 * horizontal axis.  Therefore, we only do that portion of the blit.  This
 * also ensures that we never write over source bits before needing them. 
 */

/* Prototypes for private routines. */

static ErrCode CDECL
wp_rotate2C(
    IPWState            *pIPWState,
    PIXR               *prS,
    Float32             angle,
    Int32               shrinkflag,
    PIXR	      **pPrD);

static ErrCode CDECL
wp_rotate3C(
    IPWState            *pIPWState,
    PIXR               *prS,
    Float32             angle,
    Int32               shrinkflag,
    PIXR	      **pPrD);

static ErrCode CDECL
wp_rotateLargeAngleIP(
    IPWState            *pIPWState,
    Float32             angle,
    PIXR               *prS);

static ErrCode CDECL
wp_horizontalShearIP(
    IPWState            *pIPWState,
    Float32             shearSlope,
    Bool		wingIsDest,
    PIXR               *prS,
    PIXR	       *prLeftWing,
    PIXR	       *prRightWing);

static ErrCode CDECL
wp_horizontalShearLeftIP(
    IPWState           *pIPWState,
    Int32	       *pY1,
    ScaledInt1616      *pY2,
    Int32		width,
    Int32		height,
    Int32		depth,
    ScaledInt1616	delta,
    Int32		shiftDist,
    Int32		shiftDistDelta,
    Bool		wingIsDest,
    Int32		wingW,
    PIXR               *prS,
    PIXR	       *prLeftWing,
    PIXR	       *prRightWing);

static ErrCode CDECL
wp_horizontalShearRightIP(
    IPWState           *pIPWState,
    Int32	       *pY1,
    ScaledInt1616      *pY2,
    Int32		height,
    Int32		depth,
    ScaledInt1616	delta,
    Int32		shiftDist,
    Int32		shiftDistDelta,
    Bool		wingIsDest,
    Int32		wingW,
    PIXR               *prS,
    UInt8	       *pImage,
    Int32		width,
    Int32		bpl,
    PIXR	       *prLeftWing,
    UInt8	       *pRightWing,
    Int32		wingBpl);

static ErrCode CDECL
wp_verticalShearIP(
    IPWState            *pIPWState,
    Float32             shearSlope,
    PIXR               *prS,
    PIXR	       *prLeftWing,
    PIXR	       *prRightWing);

static ErrCode CDECL
wp_rotateSmallAngleIP(
    IPWState            *pIPWState,
    Float32             angleRad,
    Int32		nice,
    PIXR               *prS);

static ErrCode CDECL
wp_shearC(
    IPWState            *pIPWState,
    PIXR               *prS,
    Int32               direction,
    Float32             angle,
    Int32               shrinkflag,
    PIXR	      **pPrD);


/***********************************************************************
 **********             Full Image Shear/Rotation             **********
 **********************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotates an image by an arbitrary angle.
 * LongDescription:::
 *  This routine rotates an image by an arbitrary angle (less than 2 pi).
 *  The effect of the rotation is to move all the existing pixel around;
 *  the result pixels are not interpolated from the original pixels.
 *  Interpolation makes images that look nices, but it's significantly
 *  slower.
 *<BR><BR>
 *  If the resizeToSource flag is set, the destination image has the same
 *  dimensions as the source image, so portions of the original source image
 *  are rotated off the page.  If the resizeToSource flag is not set,
 *  the result image is larger than the source, but all source pixels are
 *  retained.
 *<BR><BR>
 *  When the "nice" flag is set for small angles, an attempt is made to
 *  avoid long horizontal shear lines.  This improves the appearance of
 *  deskewed text somewhat at a performance cost.  Setting the "nice" flag
 *  roughly halves the rotation speed.
 *  
 * Input:::
 *  pIPWState::         Pointer to standard IPWrap state block.
 *  prS::               Pointer to binary input image.
 *  angleRad::          The rotation angle in radians.  The following
 *			condition should be met: -pi/4 <= angleRad <= pi/4.
 *			This means that the client should rotate the
 *			image to the nearest multiple of 90 degrees
 *			(pi/2) before calling this routine.  Not only
 *			will that preserve the aspect ratio, it will
 *			keep this routine from doing dumb things.
 *  resizeToSourceFlag:: When set to cTrue, the result image will be the
 *			same size as the source image.  This results in the
 *			loss of some source pixels since they are rotated off
 *			the edge of the image.  When set to cFalse, the
 *			result image is just large enough to contain all
 *			original source pixels.  This makes it larger than
 *			the source.<BR>
 *			This flag must be set to cTrue when the source
 *			and destination images are the same image (rotation
 *			in place).
 *  nice::              When the rotation angle is small (see below) and
 *			the resizeToSource flag is set, setting this flag
 *			causes the code to attempt to reduce the number of
 *			horizontal shears in rotated objects.  This often
 *			improves the appearance of text.
 *
 * Output:::
 *  pPrD::		Pointer to a pointer to a destination image.  As usual,
 *			if a destination image is supplied, its dimensions,
 *			depth, etc are checked.  If no dest image is supplied,
 *			one is created.  If *pPrD == prS, the rotation
 *			is done in place..
 *
 * Returns:::
 *  In addition to the normal checks, this routine returns cErrInvalidParm
 *  if the source and dest images are equal.  It also returns cErrInvalidParm
 *  if angleRad > pi/4 or angleRad < -pi/4.  It also returns cErrInvalidParm
 *  if prS == *pPrD (rotation in place) and resizeToSourceFlag is not cTrue.
 *<BR><BR>
 *  Rotation of anisotropic images (x resolution != y resolution) is
 *  a pain.  For small angle rotations, we can get away with introducing
 *  a shear (sort of) and get something which looks correct when converted
 *  to isotropic resolution.  For large angle rotations, things get truly
 *  weird.  So, in sloth, we will return cErrInvalidParm if a client
 *  tries a large angle rotation of an anisotropic image.
 *
 * Notes:::
 *  This function invokes two quite different types of rotation.  If the
 *  angle is small, less than 0.12 radians (~7 degrees), and the
 *  resizeToSource flag is set, an optimized rotation routine is used.
 *  Portions of this routine may be written in assembly language.  If the
 *  angle is large or the resizeToSource flag equal cFalse, the rotation
 *  is done using the standard shearing method.
 *<BR><BR>
 *  When resizeToSourceFlag is cTrue, the rotated image is the same size
 *  as the source.  Determining the region of ON pixels and possibly
 *  centering this region within the destination has been considered.
 *  An ever-present problem with this approach is image noise; noise 
 *  pixels should be removed before finding the region of ON pixels.
 *  Unfortunately, noise is not always pepper-like, and what constitues noise
 *  depends on the system.  For example, some scanners put heavy black 
 *  borders around the image.
 *<BR><BR>
 *  Since the rotated image is always bigger than the source, we need to
 *  select the portion of the rotated image that will be preserved.  If
 *  we happened to know that the reason we were rotating was that the
 *  client had put a piece of paper on a platen so that only one corner
 *  was touching the left boundary of the scanner, we could decide to
 *  throw away all the rotated pixel on the left since they would have
 *  been pictures of platen anyway.  However, we really don't have any
 *  knowledge of why a client wants the image rotated.  In the absence
 *  of knowledge, our best guess is to rotate around the center and preserve
 *  the center pixels of the document.  That's where the important stuff
 *  normally is.
 *<BR><BR>
 * Determining the exact size of the output image can be hard for the client.
 * If the resizeToSource flag is set, it's easy.  This is the common case
 * in which a client might supply a destination image.  When the
 * resizeToSource flag is set, we check the supplied dest image in the
 * normal way before using it.  However, if resizeToSource equals cFalse,
 * we go ahead and do the rotation using an extra image buffer.  If the
 * rotated image is exactly the same size as the destination image, we
 * copy the rotated image into the destination image buffer.  Otherwise,
 * an error is returned.  Clearly, this is not very efficient since the
 * high water mark for memory usage is twice the size of the dest image.
 * However, we assume that this happens very seldom.  Since it is so hard
 * for clients to calculate the size of a rotated image (when resizeToSource
 * equal cFalse), we assume they will ask this routine to both calculate the
 * size and allocate the image.  In this case, we would not have two
 * copies of the final destination image.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, arbitrary, small angle, non-interpolated, gray, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateArb (
    IPWState           *pIPWState,
    PIXR               *prS,
    Float32             angleRad,
    Bool                resizeToSourceFlag,
    Int8                nice,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prRot;         /* not H. Ross */
    PIXR		*prD = NULL;
    PIXR		*subS;
    PIXR		*planes[cMaxColorPlanes];
    UInt8		*pS, *pD;      	/* pointers to source and dest bits */
    Int32		 sBpl, dBpl;   	/* source and dest bytes/line */
    Int32		 wS, hS, wD, hD; /* width and height of source, dest */
    Int32		 depth, depthD; /* bits/pixel in source, dest */
    ColorType		 type, typeD;
    ColorMap		*map, *mapD;
    Int32		 i, channels, channelsD;
    ScaledInt1616	 xSdpi, ySdpi, xDdpi, yDdpi;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear = 0, bottomFrameRowsToClear = 0;


    prRot = NULL;
    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    ERR_GEN((prS == *pPrD) && (!resizeToSourceFlag), cErrInvalidParm);

/* Original code
    ERR_GEN((angleRad < -3.1415926F/4.0F) || (angleRad > 3.1415926F/4.0F),
	    cErrInvalidParm);
*/

    /* Due to differences in the 7th decimal place and below on
     * certain platforms (Mac with CodeWarrior 7.0), this range
     * check has been truncated to 5 decimal places.
     */
    ERR_GEN(
        (((Float32)((Int32)(angleRad * 10000))) / 10000.0F < -0.785340F) || 
	(((Float32)((Int32)(angleRad * 10000))) / 10000.0F > 0.785340F),
	cErrInvalidParm);

	/* clear all the pointers so we know whether to free
	 * them in case of error.
	 */
    for (i = 0; i < cMaxColorPlanes; i++)
	planes[i] = NULL;

	/* the source image REALLY should have a frame.  We cheat
	 * and use one line of it during some of the horizontal
	 * shears since our rasterOp doesn't go backwards..
	 */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		cChannelsAny, cFrame32);

    if (angleRad == 0.0)
    {
	ERR_PASS( w_copy(pIPWState, prS, pPrD));
	NORMAL_RETURN;
    }

	/* In the first version of this code, we used too much memory
	 * when rotating color images.  We expanded each plane horizontally,
	 * then expanded each of those vertically, then expanded each
	 * of those horizontally again.  That's silly.  In this version,
	 * we simply call this routine recursively for each channel
	 * of the source.  If there's a destination image, we pass it along.
	 * Otherwise, we let the called version of this routine create
	 * a new plane and use w_createPixrComposite to assemble them
	 * into a new color image.
	 */
    prD = *pPrD;
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));
    if (prD != NULL)
    {
	ERR_PASS (w_pixrGetChannelCount(pIPWState, prD, &channelsD));
	ERR_GEN (channels != channelsD, cErrChannelsNotEqual);
    }

    if (channels != 1)
    {
	    /* If we were given a dest image, it should have the
	     * right number of channels.
	     */

	    /* go rotate the planes one at a time */
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel(pIPWState, prS, i, &subS));

		/* if there's a dest plane, point to it.  Otherwise,
		 * use the existing NULL pointer in planes[i].
		 */
	    if (prD != NULL)
		ERR_PASS_CLEANUP (w_pixrGetChannel(pIPWState, prD, i,
						   &(planes[i])));

	    ERR_PASS_CLEANUP (w_rotateArb(pIPWState, subS, angleRad,
				  resizeToSourceFlag, nice, &(planes[i])));
	}

	    /* if the source is multi-channel and there was no dest
	     * passed in, create one from the rotated planes.  If
	     * a dest was passed in, its planes now contain the rotated
	     * images.
	     */
	if (prD == NULL)
	{
	    ERR_PASS_CLEANUP (w_pixrGetType (pIPWState, prS, &type));
	    ERR_PASS_CLEANUP (w_createPixrComposite(pIPWState, planes,
						    channels, type, &prD));
	}

	    /* things seemed to work, so return the dest pixr */
	*pPrD = prD;
	NORMAL_RETURN;
    }

	/* This is what we do for single-channel sources. */

	/* get size of source image, we'll need it later */
    ERR_PASS (w_pixrGetWidth( pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight( pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetDepth( pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));
    ERR_PASS( w_pixrGetType(pIPWState, prS, &type));
    ERR_PASS( w_pixrGetColorMap(pIPWState, prS, &map));

	/* If we were given a dest image, it must meet some basic
	 * criteria.  Check it now since we'll have to check it
	 * in any case.
	 */
    if (prD != NULL)
    {
	    /* if it's supposed to be the same size as the source image,
	     * see if it is that size.
	     */
	if ((resizeToSourceFlag) && (prD != prS))
	{
		/* the dest image REALLY should have a frame.  We cheat
		 * and use one line of it during some of the horizontal
		 * shears since our rasterOp doesn't go backwards.
		 */
	    ERR_PASS (w_checkDestPixr(pIPWState, wS, hS, depth, xSdpi,
				       ySdpi, cFrame32, channels,
				       type, map, prD,
				       &rightFrameColsToClear,
				       &bottomFrameRowsToClear));
	}
	else
	if (prD != prS)
	{
	    /* If the resizeToSourceFlag is false and 
	     * purity were strictly maintained, we'd check the
	     * the dest image to see if it was the size of the rotated
	     * image.  This is a yucky calculation.  It's hard to convince
	     * the routines below to use the dest pixr anyway. 
	     * It's also unlikely that a client will happen to supply
	     * an image of exactly the right size (unless they rotated
	     * a bunch of source images having the same dimensions by
	     * the same angle).  So, we won't bother.  However, we will
	     * check the other stuff.
	     */
		/* is the pixr basically ok? */
	    ERR_GEN_PIXR(prD, cByteOrderBig, cDepthAny, cChannelsAny,
			  cFrame32);
	
		/* already checked the number of channels */

		/* Are the ColorType and ColorMap the same?  We check to see
		 * if the color map points to the same block of memory.  This
		 * is not very sophisticated, maybe we'll change it if we
		 * really start using color maps.
		 */
	    ERR_PASS( w_pixrGetType(pIPWState, prD, &typeD));
	    ERR_GEN(typeD != type, cErrDestDimensIncorrect);
	    ERR_PASS( w_pixrGetColorMap(pIPWState, prD, &mapD));
	    ERR_GEN(mapD != map, cErrDestDimensIncorrect);
	
		/* does it have the correct depth? */
	    ERR_PASS( w_pixrGetDepth(pIPWState, prD, &depthD));
	    ERR_GEN(depthD != depth, cErrDestDimensIncorrect);
	
		/* the resolution has to be close to the desired resolution.
		 * Since these numbers have 16 bits of fraction, comparing
		 * the difference to 0x10000 is like comparing it to 1.0
		 */
	    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prD, &xDdpi, &yDdpi));
	    ERR_GEN((ABS(xDdpi - xSdpi) > 0x10000) ||
		    (ABS(yDdpi - ySdpi) > 0x10000), cErrDestDimensIncorrect);
	}
    }
	/* if no dest image was passed in and resizeToSourceFlag is set,
	 * create one now.  We'll need it for either the small or
	 * large angle cases.
	 */
    else	/* prD == NULL */
    if (resizeToSourceFlag)
    {
	ERR_PASS( w_clonePixrNoInit(pIPWState, prS, &prD));
	allocatedHere = cTrue;
	rightFrameColsToClear = 0;
	bottomFrameRowsToClear = 0;
    }

	/* At this point, if the resizeToSourceFlag was set, we have
	 * a dest image.  If it wasn't set, we might have a dest
	 * image.  Here's where we start the real rotation.
	 */

	/* if we're supposed to do rotation in place, check for
	 * large or small angle.
	 */
    if (prS == prD)
    {
	if (ABS(angleRad) < SMALL_ANGLE)
	    ERR_PASS_CLEANUP( wp_rotateSmallAngleIP(pIPWState, angleRad,
						   nice, prS));
	else
	    ERR_PASS_CLEANUP( wp_rotateLargeAngleIP(pIPWState, angleRad,
						    prS));
    }

	/* If we're going to do the small angle rotation and
	 * the resizeToSourceFlag is set, do it with optimized
	 * code.
	 */
    else
    if ((ABS(angleRad) < SMALL_ANGLE) && (resizeToSourceFlag))
    {
	    /* go do the real rotation of this plane */
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &depth));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));

	ERR_PASS_CLEANUP( i_rotateNoInterpolation (
					pIPWState,
					pS, wS, hS, sBpl,
					angleRad,
					pD, dBpl, depth, nice,
					(Int16)ScaledInt1616_TO_Int32(xSdpi),
					(Int16)ScaledInt1616_TO_Int32(ySdpi)));

	    /* If we need to clear any parts of the frame, do so */
	if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	    ERR_PASS_CLEANUP(wp_clearRightBottomFrames(pIPWState, prD,
						       rightFrameColsToClear,
						       bottomFrameRowsToClear));
    }
	/* If this is a large angle rotation and the resizeToSourceFlag
	 * is set, use the special version of rotation that tosses
	 * all the bits that fall off the edge.
	 */
    else
    if (resizeToSourceFlag)
    {
	    /* copy the source image to the dest before we call
	     * w_rotateLargeAngleIP.  This isn't quite as efficient as
	     * having the large angle rotation routine be able to
	     * handle a source and dest that may or may not
	     * be the same (though that's no picnic either) but
	     * we assume that the vast majority of rotations are
	     * in place anyway, so this case is seldom encountered.
	     */
	ERR_PASS_CLEANUP( w_copy(pIPWState, prS, &prD));

	    /* Do the rotation */
	ERR_PASS_CLEANUP( wp_rotateLargeAngleIP(pIPWState, angleRad, prD));
    }

	/* Otherwise, we're doing a rotate in which we want to
	 * keep all the original pixels.
	 * Even if we were given an output image, we still have to
	 * create a new output image to rotate into initially
	 * since we don't (yet) know how to create just the center
	 * portion of the output.
	 * Chicken out if this is an anisotropic image since I
	 * don't understand the shearing method well enough to
	 * use it on anisotropic images.
	 */
    else
    {
	    /* Quit if image is anistropic. */
	ERR_GEN_CLEANUP(ABS(xSdpi - ySdpi) > 0x10000, cErrInvalidParm);

	if (ABS(angleRad) < SMALL_ANGLE)
	{
	    ERR_PASS( wp_rotate2C(pIPWState, prS, angleRad, NO_SHRINK, &prRot));
	}
	else
	{
	    ERR_PASS( wp_rotate3C(pIPWState, prS, angleRad, NO_SHRINK, &prRot));
	}

	    /* If the output image should be different size than the AND
	     * we were given a dest
	     * image, see if the dest image is the right size.  If so, use it.
	     * If not, return an error and give up.
	     */
	    /* were we given a dest image? */
	if (*pPrD == NULL)
	{
		/* nope, just use the one we've calculated. */
	    *pPrD = prRot;
	    NORMAL_RETURN;
	}
	else 	/* we were given a dest image. Is it the right size? */
	{
	    ERR_PASS_CLEANUP (w_pixrGetWidth( pIPWState, prRot, &wS));
	    ERR_PASS_CLEANUP (w_pixrGetHeight( pIPWState, prRot, &hS));
	    ERR_PASS_CLEANUP (w_pixrGetWidth( pIPWState, prD, &wD));
	    ERR_PASS_CLEANUP (w_pixrGetHeight( pIPWState, prD, &hD));
	    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI(pIPWState, prD,
						     &xDdpi, &yDdpi));
	    ERR_GEN_CLEANUP ((wD != wS) || (hD != hS) ||
		     (xDdpi != xSdpi) || (yDdpi != ySdpi),
		     cErrDestDimensIncorrect);

		/* it's the right size, so use it. */
	    ERR_PASS_CLEANUP( w_copy (pIPWState, prRot, &prD));
	}
    }

	/* things seemed to work, so return the dest pixr */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere == cTrue)
	w_destroyPixr(pIPWState, &prD);
    w_destroyPixr(pIPWState, &prRot);

	/* if no dest was passed in, delete any planes that were allocated.
	 */
    if (*pPrD == NULL)
    {
	for (i = 0; i < cMaxColorPlanes; i++)
	    w_destroyPixr(pIPWState, &(planes[i]));
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotates an image using two shears.
 *
 * LongDescription:::
 *  It is possible to rotate an image using three shear operations.
 *  For small angles, the distortion created by eliminating one of the
 *  shears is negligible.  This routine is used for small angle rotation.
 *  The result is an image where the input pixels have just been shuffled
 *  around.  A prettier rotation would compute each output pixel from
 *  the weighted sum of four input pixels.  That also takes longer.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be rotated.
 *  angle::          Rotation angle in radians.
 *  shrinkflag::     An unused flag.  Could later be used to shrink
 *		     the result image to the region occupied by black
 *		     pixels.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  A destination
 *		     image is always allocated here.
 *
 * Notes:::
 *  This is a 2-shear rotation about the image center.
 *  For small angles, this gives a close approximation to
 *  an actual rotation.  For large angles,
 *  the length/width ratio is noticably changed,
 *  because the change is proportional to the square of the angle.
 *<BR><BR>
 *  If angle > 0, rotation is clockwise; if angle < 0, rotation is 
 *  counterclockwise.
 *  Rotation is by shearing image about the origin by an amount
 *  equal to the variable tanAngle where tanAngle is the tangent of
 *  the rotation angle:<BR>
 *  x' = x + tanAngle * y      for x-shear<BR>
 *  y' = y + tanAngle * x      for y-shear<BR>
 *<BR>
 *  For small angles, two such shears can be used together to
 *  approximate a small rotation.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, small angle, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_rotate2C (
    IPWState           *pIPWState,
    PIXR               *prS,
    Float32             angle,
    Int32               shrinkflag,
    PIXR              **pPrD)
{
    PROLOGUE;
    Float32        tanAngle;
    PIXR          *prT1, *prT2;

    *pPrD = prT1 = prT2 = NULL;

    tanAngle = (Float32) tan(angle);
    ERR_PASS( wp_shearC(pIPWState, prS, VERT, tanAngle, NO_SHRINK, &prT1));
    IP_YIELD;

    ERR_PASS_CLEANUP( wp_shearC(pIPWState, prT1, HORIZ, -tanAngle, NO_SHRINK,
			      &prT2));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT1));
    IP_YIELD;

/* The only calls to this routine so far are listed above.  In those
 * calls, shrinkflag is set to NO_SHRINK.  To avoid linking in
 * useless stuff, we'll just comment this out.
 */
    shrinkflag = shrinkflag;	/* keep the compiler happy */

    *pPrD = prT2;
    NORMAL_RETURN;


ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prT1);
    w_destroyPixr(pIPWState, &prT2);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotates an image using three shears.
 *
 * LongDescription:::
 *  It is possible to rotate an image using three shear operations.
 *  The result is an image where the input pixels have just been shuffled
 *  around.  A prettier rotation would compute each output pixel from
 *  the weighted sum of four input pixels.  That also takes longer.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be rotated.
 *  angle::          Rotation angle in radians.
 *  shrinkflag::     An unused flag.  Could later be used to shrink
 *		     the result image to the region occupied by black
 *		     pixels.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  A destination
 *		     image is always allocated here.
 *
 * Notes:::
 *  This is a 3-shear rotation about the image center.
 *  Modulo the insertion of jaggies, this rotation should be accurate for
 *  all angles.
 *<BR><BR>
 *  If angle > 0, rotation is clockwise; if angle < 0, rotation is 
 *  counterclockwise.
 *  Rotation is by shearing image about the origin three times.  The
 *  first shear is horizontal, the second is vertical and the third
 *  is horizontal again:<BR>
 *  x' = x + tan(angle/2) * y      for first x-shear<BR>
 *  y' = y + sin(angle) * x        for y-shear<BR>
 *  x' = x + tan(angle/2) * y      for second x-shear<BR>
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, small angle, arbitrary, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_rotate3C (
    IPWState           *pIPWState,
    PIXR               *prS,
    Float32             angle,
    Int32               shrinkflag,
    PIXR              **pPrD)
{
    PROLOGUE;
    Float32        shearAngle1, shearAngle2;
    PIXR          *prT1, *prT2;

    *pPrD = prT1 = prT2 = NULL;

    shearAngle1 = (Float32) tan(angle / 2.);
    shearAngle2 = (Float32) sin(angle);
    ERR_PASS( wp_shearC(pIPWState, prS, HORIZ, -shearAngle1, NO_SHRINK, &prT1));
    IP_YIELD;

    ERR_PASS_CLEANUP( wp_shearC(pIPWState, prT1, VERT, shearAngle2, NO_SHRINK,
			      &prT2));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT1));
    IP_YIELD;

    ERR_PASS_CLEANUP( wp_shearC(pIPWState, prT2, HORIZ, -shearAngle1, NO_SHRINK,
			      &prT1));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prT2));
    IP_YIELD;


/* The only calls to this routine so far are listed above.  In those
 * calls, shrinkflag is set to NO_SHRINK.  To avoid linking in
 * useless stuff, we'll just comment this out.
 */
    shrinkflag = shrinkflag;	/* keep the compiler happy */

    *pPrD = prT1;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prT1);
    w_destroyPixr(pIPWState, &prT2);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Shear either horizontally or vertically about the center of an image.
 *
 * LongDescription:::
 *  This routine shears images.  If the direction arg equals HORIZ, it
 *  shears horizontally, otherwise it shears vertically.  What's a shear?
 *  You divide the image into a bunch of strips and slide the strips
 *  relative to each other.  In a horizontal shear, the strips are horizontal,
 *  in a vertical shear, they're vertical.  The width of the strips depends
 *  on the shear angle.  Ideally, we'd be moving every pixel some sub-pixel
 *  distance, but we can't do that.  The strips in the middle don't move
 *  very much, the strips at the edges move a lot.  This routine is used
 *  to rotate an image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be sheared.
 *  direction::      Shear direction, either HORIZ or VERT
 *  angle::          A misnomer.  When doing 2-shear rotations, this is
 *		     really the tangent of the shear angle.  When doing
 *		     3-shear rotations, this is tan(angle/2) for the
 *		     two horizontal shears and sin(angle) for the vertical
 *		     shear.
 *  shrinkflag::     An unused flag.  Could later be used to shrink
 *		     the result image to the region occupied by black
 *		     pixels.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  A destination
 *		     image is always allocated here.
 *
 * Notes:::
 *  If I were a good person, I'd read all about rotation using shears and
 *  explain it in wonderfully lucid prose.  Instead, I'll ask you to
 *  read about it.
 *<BR><BR>
 *  FEATURE ADDITION: The PerfectScan guys complained about the black stuff
 *  at the edges of the image where "new" pixels were created in gray images.
 *  To fix this, we set all the pixels that would otherwise be untouched
 *  at the sides (horiz shear) or top and bottom (vertical shear to white
 *  before doing the shear.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, arbitrary, small angle, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_shearC (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               direction,
    Float32             angle,
    Int32               shrinkflag,
    PIXR              **pPrD)
{
    PROLOGUE;
    Int32		 i, nx, ny, sign, width, height;
    Int32		 depth;
    Int32		 newW, newH, destW, destH, transX, transY;
    Int32		 x1, y1, x2, y2, incr;
    Int32		 frame;
    Float32		 delta;
    PIXR		*prT, *prShear;
    ScaledInt1616	 xdpi, ydpi;

    *pPrD = prT = prShear = NULL;

    if (angle == 0.0)
    {
	ERR_PASS( w_copy(pIPWState, prS, pPrD));
	NORMAL_RETURN;
    }

	/* set up */
    sign = SIGN(angle);
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &width));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &height));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frame));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xdpi, &ydpi));

	/* calculate half the shifts in the horiz and vert directions */
    nx = (Int32)ABS(angle * (Float32)(width / 2));
    ny = (Int32)ABS(angle * (Float32)(height / 2));
    delta = ABS(1.0F / angle);  /* Float32 of # pixels per one bit shift */

	/* Tiny bug fix (8/29/97): Round the new width and height up,
	 * not down.  This makes sure we include any portion of a new
	 * row or column.  However, we like the shifts that are being
	 * done, so don't change those (newW and newH).
	 */
    if (direction == HORIZ)
    {
	newW = width + (Int32)(height * ABS(angle));
	destW = width + (Int32)(height * ABS(angle) + 0.9999);
	newH = height;
	destH = newH;
    }
    else  /* VERT */
    {
	newW = width;
	destW = newW;
	newH = height + (Int32)(width * ABS(angle));
	destH = height + (Int32)(width * ABS(angle) + 0.9999);
    }

	/* Since we're only called from w_rotateArb and it's already
	 * broken multi-channel images into individual planes before
	 * calling down here, we can create a single image here.
	 */
    ERR_PASS( w_createPixr(pIPWState, destW, destH, depth, xdpi, ydpi,
				   frame, &prT));

    transX = (newW - width) / 2;
    transY = (newH - height) / 2;

    x1 = 0;
    y1 = 0;

	/* Do rasterops in horizontal segments */
    if (direction == HORIZ)   /* horizontal shear */
    {
	    /* If this is a gray pixr, set the right and left
	     * sides to white before starting.
	     */
	if (depth != 1)
	{
	    ERR_PASS_CLEANUP( w_rasterop (
		pIPWState,
		0, 0, NULL,		/* src doesn't matter */
		PIX_SET,		/* set to white */
		0, 0,			/* x, y */
		newW - width, newH,	/* w, h */
		prT));			/* dest */
	    ERR_PASS_CLEANUP( w_rasterop (
		pIPWState,
		0, 0, NULL,		/* src doesn't matter */
		PIX_SET,		/* set to white */
		newW - (newW - width),	/* x */
		0,			/* y */
		newW - width, newH,	/* w, h */
		prT));			/* dest */
	}

	for (i = 1; y1 < height; i++)
	{
	    y2 = (Int32) ((Float32)i * delta);
	    if (y2 >= height)  /* last one */
		incr = height - y1;
	      else
		incr = y2 - y1;
	    ERR_PASS_CLEANUP( w_rasterop (
		pIPWState,
		0, y1,			    /* srcX, srcY */
		prS,			    /* src */
		PIX_SRC,		    /* just copy */
		transX - sign*(ny - i), y1, /* destX, destY */
		width, incr,		    /* width, height */
		prT));	 	 	    /* dest */
	    y1 = y2;
	}
    }
    
	/* Do rasterops in vertical segments */
    if (direction == VERT)   /* vertical shear */
    {
	    /* If this is a gray pixr, set the top and bottom
	     * to white before starting.
	     */
	if (depth != 1)
	{
	    ERR_PASS_CLEANUP( w_rasterop (
		pIPWState,
		0, 0, NULL,		/* src doesn't matter */
		PIX_SET,		/* set to white */
		0, 0,			/* x, y */
		newW, newH - height,	/* w, h */
		prT));			/* dest */
	    ERR_PASS_CLEANUP( w_rasterop (
		pIPWState,
		0, 0, NULL,		/* src doesn't matter */
		PIX_SET,		/* set to white */
		0,			/* x */
		newH - (newH - height),	/* y */
		newW, newH - height,	/* w, h */
		prT));			/* dest */
	}

	for (i = 1; x1 < width; i++)
	{
	    x2 = (Int32) ((Float32)i * delta);
	    if (x2 >= width)  /* last one */
		incr = width - x1;
	      else
		incr = x2 - x1;
	    ERR_PASS_CLEANUP( w_rasterop (
		pIPWState,
		x1, 0,			    /* srcX, srcY */
		prS,			    /* src */
		PIX_SRC,		    /* just copy */
		x1, transY - sign*(nx - i), /* destX, destY */
		incr, height,		    /* width, height */
		prT));			    /* dest */
	    x1 = x2;
	}
    }

/* Since the only calls to this routine are from rotate2CPr and rotate3CPr
 * above with the NO_SHRINK flag set, we will comment this stuff out
 * to avoid linking in useless stuff.
 */
    shrinkflag = shrinkflag;	/* keep the compiler happy */

    *pPrD = prT;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prT);
    w_destroyPixr(pIPWState, &prShear);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotates an image in place using three shears.
 *
 * LongDescription:::
 *  It is possible to rotate an image using three shear operations.
 *  The result is an image where the input pixels have just been shuffled
 *  around.  A prettier rotation would compute each output pixel from
 *  the weighted sum of four input pixels.  That also takes longer.
 *
 * PrivateDescription:::
 *  The difference between this method and the one implemented in wp_shear3C
 *  is that much smaller images are used.  In the first rotate, small "wing"
 *  images are created on the left and right sides just to catch the bits
 *  that are shifted off the edges of the original image. 
 *  In the second, vertical, shear, the image is sheared as is normal,
 *  bits shifted off the top and bottom being discarded.  The wings
 *  are also sheared vertically, though only the data that crosses the
 *  central horizontal axis will be used in the third shear, so we only
 *  move that data.  The wings now act as data sources for the final shear.
 *  Bits shifted off the image at this step are discarded.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  angle::          Rotation angle in radians.
 *  prS::            Pointer to image to be rotated.
 *
 * Notes:::
 *  This is a 3-shear rotation about the image center.
 *  Modulo the insertion of jaggies, this rotation should be accurate for
 *  all legal angles (-pi/4 <= angle <= pi/4).
 *<BR><BR>
 *  If angle > 0, rotation is clockwise; if angle < 0, rotation is 
 *  counterclockwise.
 *  Rotation is by shearing image about the origin three times.  The
 *  first shear is horizontal, the second is vertical and the third
 *  is horizontal again:<BR>
 *  x' = x + tan(angle/2) * y      for first x-shear<BR>
 *  y' = y + sin(angle) * x        for y-shear<BR>
 *  x' = x + tan(angle/2) * y      for second x-shear<BR>
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, small angle, arbitrary, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_rotateLargeAngleIP (
    IPWState           *pIPWState,
    Float32             angle,
    PIXR               *prS)
{
    PROLOGUE;
    Float32		 shearSlope1, shearSlope2;
    PIXR		*prLeftWing = NULL, *prRightWing = NULL;
    Int32		 height, depth;
    Int32		 wingW;
    ScaledInt1616	 xdpi, ydpi;

	/* shearSlope1 is used for the first and third (horizontal) shears.
	 * shearSlope2 is used for the second (vertical) shear.
	 */
    shearSlope1 = (Float32) tan(angle / 2.0);
    shearSlope2 = (Float32) sin(angle);

	/* set up */
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &height));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xdpi, &ydpi));

	/* calculate width and height of the two wings.  We'll
	 * allocate them as two blocks on the theory that it's
	 * no harder, and generally easier, to allocate two
	 * buffers of size X/2 than to allocate a single
	 * contiguous buffer of size X.
	 * NOTE that the wings are frameless to save space
	 * and pLeftWing and pRightWing point to the
	 * upper left corners of the real image.
	 * We add 2 to the calculated width to account for
	 * truncation and for good measure.  We did get a
	 * general protection fault when we added 0.
	 */
    wingW = (Int32) (((height + 1)/2) * ABS(shearSlope1) + 2);

    ERR_PASS( w_createPixr(pIPWState, wingW, height, depth, xdpi, ydpi,
			   0, &prLeftWing));
    ERR_PASS_CLEANUP( w_createPixr(pIPWState, wingW, height, depth, xdpi, ydpi,
				   0, &prRightWing));

	/* If this is a binary image, the wings are already filled
	 * with white (0).  If this is gray image, the same 0 is
	 * black, so fill them with white.
	 */
    if (depth != 1)
    {
	ERR_PASS_CLEANUP( w_rasterop(pIPWState, 0, 0, NULL,
				     PIX_SET, 0, 0, wingW, height,
				     prLeftWing));
	ERR_PASS_CLEANUP( w_rasterop(pIPWState, 0, 0, NULL,
				     PIX_SET, 0, 0, wingW, height,
				     prRightWing));
    }

	/* Begin the first horizontal shear.  We shear the image on
	 * the way from the source image to the dest image.
	 * Any bits that fall off the dest image are written to the
	 * wings.
	 * If we're rotating clockwise (angle >= 0), the right wing
	 * gets stuff that falls off the upper right and the left
	 * wing gets the stuff that falls off the lower left.
	 * For counterclockwise rotation, the right wing gets stuff
	 * that falls off the lower right and the left wing gets
	 * stuff that falls off the upper left.
	 * Logically, we align wings with the top of the image.
	 */
    ERR_PASS_CLEANUP( wp_horizontalShearIP(pIPWState, shearSlope1,
					   cTrue, prS,
					   prLeftWing, prRightWing));

	/* Do the vertical shear */
    ERR_PASS_CLEANUP( wp_verticalShearIP(pIPWState, shearSlope2,
					 prS, prLeftWing, prRightWing));

	/* Do the final horizontal shear */
    ERR_PASS_CLEANUP( wp_horizontalShearIP(pIPWState, shearSlope1,
					   cFalse, prS,
					   prLeftWing, prRightWing));

	/* That's it.  Free the wings and return */
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prLeftWing));
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prRightWing));
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &prLeftWing);
    w_destroyPixr(pIPWState, &prRightWing);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Shears an image horizontally in place using two side images.
 *
 * LongDescription:::
 *  When rotating an image in place using shear operations, two of the
 *  operations are horizontal shears.  This routine is used to implement
 *  them.<BR>
 *
 * PrivateDescription:::
 *  In addition to the original image, this routine is supplied with
 *  two wing images, one on either side.  In the first horizontal shear,
 *  the wing serve to catch bits that fell off the edges of the main
 *  image.  In the second, and final, horizontal shear (after the vertical
 *  shear in the middle), the wings are sources of bits for the
 *  main image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  shearSlope::     The slope of the shear expressed as a ratio
 *		     of x units / y unit.  This number has 16 bits of
 *		     fraction and 16 bits of integer.  This number
 *		     is positive or negative.
 *  wingIsDest::     A Boolean that is set to cTrue when the wings
 *		     receive data from the image.  When set to cFalse,
 *		     the wings provide data to the image.
 *  prS::            Pointer to image to be rotated.
 *  prLeftWing::     Pointer to wing image on the left of the real image.
 *  prRightWing::    Pointer to wing image on the right of the real image.
 *
 * Notes:::
 *  We assume that the rotation angle is between (-45 <= angle <= 45).
 *  It's also assumed that the shearSlope is tan(angle/2).
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, arbitrary, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_horizontalShearIP(
    IPWState           *pIPWState,
    Float32             shearSlope,
    Bool		wingIsDest,
    PIXR               *prS,
    PIXR	       *prLeftWing,
    PIXR	       *prRightWing)
{
    PROLOGUE;
    Int32		 shiftDist;
    Int32		 width, height, depth;
    Int32		 wingW;
    Int32		 y1;
    ScaledInt1616	 delta, y2;
    UInt8		*pImage, *pRightWing;
    Int32		 bpl, wingBpl;

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &width));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &height));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pImage));
    ERR_PASS( w_pixrGetWidth(pIPWState, prLeftWing, &wingW));
    ERR_PASS( w_pixrGetBpl(pIPWState, prLeftWing, &wingBpl));
    ERR_PASS( w_pixrGetImage(pIPWState, prRightWing, &pRightWing));

	/* If this is a binary image, the wings are already filled
	 * with white (0).  If this is gray image, the same 0 is
	 * black, so fill them with white.
	 */
	/* calculate # of pixels per one bit shift.  This number
	 * has 16 bits of integer and 16 bits of fraction.
	 */
    delta = (ScaledInt1616)(ABS((Float32)0x10000 / shearSlope));

	/* Do clockwise rotations */
    if (shearSlope >= 0)
    {
	    /* Since we're rotating clockwise, bits shift off
	     * to the upper right and the lower left.
	     */
	y1 = 0;		/* start at top of the image */
	y2 = 0;
	    /* calc distance topmost line will be shifted to the right */
	shiftDist = (Int32)(ABS(shearSlope) * (Float32)(height/2));

	ERR_PASS(wp_horizontalShearRightIP(pIPWState, &y1, &y2, height,
					   depth, delta, shiftDist,
					   -1, wingIsDest, wingW,
					   prS, pImage, width,
					   bpl, prLeftWing,
					   pRightWing, wingBpl));

	    /* The lower half of the image is shifted left.
	     * Since w_rasterop moves strictly from left to
	     * right, we can use it without the sillyness
	     * of copying the bits to the frame first.
	     * Copy the leftover bits to the left wing.
	     */
	shiftDist = 1;
	ERR_PASS(wp_horizontalShearLeftIP(pIPWState, &y1, &y2, width,
					  height, depth, delta, shiftDist,
					  +1, wingIsDest, wingW,
					  prS, prLeftWing,
					  prRightWing));

    }		/* end of doing horiz shift clockwise */
    else	/* do horiz shift counterclockwise */
    {
	    /* Since we're rotating counterclockwise, bits shift off
	     * to the upper left and the lower right.
	     */
	y1 = 0;		/* start at top of the image */
	y2 = 0;
	    /* calc distance topmost line will be shifted to the left */
	shiftDist = (Int32)(ABS(shearSlope) * (Float32)(height/2));

	ERR_PASS(wp_horizontalShearLeftIP(pIPWState, &y1, &y2, width,
					  height, depth, delta, shiftDist,
					  -1, wingIsDest, wingW,
					  prS, prLeftWing,
					  prRightWing));

	    /* the lower half of the image is shifted right.  Copy
	     * the leftover bits to the right wing.
	     */
	shiftDist = 1;
	ERR_PASS(wp_horizontalShearRightIP(pIPWState, &y1, &y2, height,
					   depth, delta, shiftDist,
					   +1, wingIsDest, wingW,
					   prS, pImage, width,
					   bpl, prLeftWing,
					   pRightWing, wingBpl));
    }		/* end of doing horiz shift counterclockwise */

	/* Clear the line of frame in the source that was used for
	 * right shifts.
	 */
    ERR_PASS( i_rasterOp (pIPWState, pImage,	/* ptr to dest */
			  0, 0,			/* destX, destY */
			  width, cFrameBits,	/* width, height */
			  depth,		/* dest depth */
			  bpl,			/* dest Bytes/line */
			  PIX_CLR,		/* always set to 0 */
			  NULL,			/* source ptr */
			  0, 0,			/* srcX, srcY */
			  0,			/* src depth */
			  0));			/* src bytes/line */
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Shears a portion of an image horizontally left in place using
 *  two side images.
 *
 * LongDescription:::
 *  When rotating an image in place using shear operations, two of the
 *  operations are horizontal shears.  This routine is used to perform
 *  shears where all blocks are moving to the left.
 *
 * PrivateDescription:::
 *  In addition to the original image, this routine is supplied with
 *  two wing images, one on either side.  In the first horizontal shear,
 *  the wing serve to catch bits that fell off the edges of the main
 *  image.  In the second, and final, horizontal shear (after the vertical
 *  shear in the middle), the wings are sources of bits for the
 *  main image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pY1::	     Ptr to Y coord of top line of first block to be shifted.
 *  pY2::	     Ptr to Y coord of last line + 1 of first block to be
 *		     shifted.  This is in ScaledInt1616 units.
 *  width::	     Width of the image in pixels.
 *  height::	     Height of the image in pixels.
 *  depth::	     Bits/pixel in the image.
 *  delta::	     The height of each block in ScaleInt1616 units.
 *  shiftDist::	     Distance (left) the first block should be shifted.
 *  shiftDistDelta:: This number is added to shiftDist after each block
 *		     is shifted.  It should be +1 or -1.
 *  wingIsDest::     A Boolean that is set to cTrue when the wings
 *		     receive data from the image.  When set to cFalse,
 *		     the wings provide data to the image.
 *  wingW::	     Width of wings in pixels.
 *  prS::            Pointer to image to be rotated.
 *  prLeftWing::     Pointer to wing image on the left of the real image.
 *  prRightWing::    Pointer to wing image on the right of the real image.
 *
 * Notes:::
 *  We assume that the rotation angle is between (-45 <= angle <= 45).
 *  It's also assumed that the shearSlope is tan(angle/2).
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, arbitrary, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_horizontalShearLeftIP(
    IPWState           *pIPWState,
    Int32	       *pY1,
    ScaledInt1616      *pY2,
    Int32		width,
    Int32		height,
    Int32		depth,
    ScaledInt1616	delta,
    Int32		shiftDist,
    Int32		shiftDistDelta,
    Bool		wingIsDest,
    Int32		wingW,
    PIXR               *prS,
    PIXR	       *prLeftWing,
    PIXR	       *prRightWing)
{
    PROLOGUE;
    Int32		setToWhiteOp;
    Int32		y1 = *pY1;
    ScaledInt1616	y2 = *pY2;
    Int32		incr;

    if (depth != 1)
	setToWhiteOp = PIX_SET;
    else
	setToWhiteOp = PIX_CLR;

    while ((shiftDist >= 0) && (y1 < height))
    {
	y2 += delta;			/* calculate address of bottom of
					 * current horiz shear with
					 * 16 bits of fraction.
					 */
	if (ScaledInt1616_TO_Int32(y2) >= height)
	    incr = height - y1;
	else
	    incr = ScaledInt1616_TO_Int32(y2) - y1;

	    /* If leftover bits are supposed to be copied to the
	     * wings, copy them to the left wing now before they are
	     * overwritten.
	     */
	if (wingIsDest)
	    ERR_PASS( w_rasterop(pIPWState,
				 0, y1,			/* srcX, srcY */
				 prS,			/* src image */
				 PIX_SRC,		/* just copy */
				 wingW - shiftDist,	/* destX */
				 y1,			/* destY */
				 MIN(width, shiftDist),	/* width */
				 incr,			/* height */
				 prLeftWing));		/* dest image */

	    /* do the shift inside the image.  Don't do this
	     * if there's nothing left inside the image
	     * (width <= shiftDist)
	     */
	if (width > shiftDist)
	{
	    ERR_PASS( w_rasterop(pIPWState,
				 shiftDist, y1,		/* srcX, srcY */
				 prS,			/* src image */
				 PIX_SRC,		/* just copy */
				 0,			/* destX */
				 y1,			/* destY */
				 width - shiftDist,	/* width */
				 incr,			/* height */
				 prS));			/* dest image */
	}

	    /* If we're copying bits to the wings, clear
	     * the hole we just made.  Otherwise, fill it with
	     * data from the right wing.
	     */
	if (wingIsDest)
	    ERR_PASS( w_rasterop(pIPWState,
				 0, 0,			/* srcX, srcY */
				 NULL,			/* src image */
				 setToWhiteOp,		/* clear it */
				 MAX(0, width - shiftDist), /* destX */
				 y1,			/* destY */
				 MIN(width, shiftDist),	/* width */
				 incr,			/* height */
				 prS));			/* dest image */
	else
	if (width > shiftDist)
		/* We need to fill the right side of the image from
		 * data on the left side of the right wing.  This
		 * is the normal case.
		 */
	    ERR_PASS( w_rasterop(pIPWState,
				 0, y1,			/* srcX, srcY */
				 prRightWing,		/* src image */
				 PIX_SRC,		/* copy */
				 width - shiftDist,	/* destX */
				 y1,			/* destY */
				 shiftDist,		/* width */
				 incr,			/* height */
				 prS));			/* dest image */
	else
		/* We need to fill the image with data taken from farther
		 * over in the right wing.  This happens when the image
		 * is very tall and narrow.
		 */
	    ERR_PASS( w_rasterop(pIPWState,
				 shiftDist - width,	/* srcX */
				 y1,			/* srcY */
				 prRightWing,		/* src image */
				 PIX_SRC,		/* copy */
				 0,			/* destX */
				 y1,			/* destY */
				 MIN(width,
				     wingW -
				     (shiftDist - width)), /* width */
				 incr,			/* height */
				 prS));			/* dest image */

	y1 = ScaledInt1616_TO_Int32(y2);	/* prepare for next shift */
	shiftDist += shiftDistDelta;
    }
	/* update the client's y1 and y2 ptrs */
    *pY1 = y1;
    *pY2 = y2;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Shears a portion of an image horizontally right in place using
 *  two side images.
 *
 * LongDescription:::
 *  When rotating an image in place using shear operations, two of the
 *  operations are horizontal shears.  This routine is used to perform
 *  shears where all blocks are moving to the right.
 *
 * PrivateDescription:::
 *  In addition to the original image, this routine is supplied with
 *  two wing images, one on either side.  In the first horizontal shear,
 *  the wing serve to catch bits that fell off the edges of the main
 *  image.  In the second, and final, horizontal shear (after the vertical
 *  shear in the middle), the wings are sources of bits for the
 *  main image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pY1::	     Ptr to Y coord of top line of first block to be shifted.
 *  pY2::	     Ptr to Y coord of last line + 1 of first block to be
 *		     shifted.  This is in ScaledInt1616 units.
 *  height::	     Height of the image in pixels.
 *  delta::	     The height of each block in ScaleInt1616 units.
 *  shiftDist::	     Distance (left) the first block should be shifted.
 *  shiftDistDelta:: This number is added to shiftDist after each block
 *		     is shifted.  It should be +1 or -1.
 *  wingIsDest::     A Boolean that is set to cTrue when the wings
 *		     receive data from the image.  When set to cFalse,
 *		     the wings provide data to the image.
 *  wingW::	     Width of wings in pixels.
 *  prS::            Pointer to image to be rotated.
 *  pImage::	     Pointer to outside of frame of image.
 *  width::	     Width of image in pixels.
 *  depth::	     Image's bits/pixel.
 *  bpl::            Image/s bytes per scan line.
 *  prLeftWing::     Pointer to wing image on the left of the real image.
 *  pRightWing::     Pointer to first image byte, outside of frame.
 *  wingBpl::	     Bytes/line for both wings.
 *
 * Notes:::
 *  We assume that the rotation angle is between (-45 <= angle <= 45).
 *  It's also assumed that the shearSlope is tan(angle/2).
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, arbitrary, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_horizontalShearRightIP(
    IPWState           *pIPWState,
    Int32	       *pY1,
    ScaledInt1616      *pY2,
    Int32		height,
    Int32		depth,
    ScaledInt1616	delta,
    Int32		shiftDist,
    Int32		shiftDistDelta,
    Bool		wingIsDest,
    Int32		wingW,
    PIXR               *prS,
    UInt8	       *pImage,
    Int32		width,
    Int32		bpl,
    PIXR	       *prLeftWing,
    UInt8	       *pRightWing,
    Int32		wingBpl)
{
    PROLOGUE;
    Int32		setToWhiteOp;
    Int32		y1 = *pY1;
    ScaledInt1616	y2 = *pY2;
    Int32		incr;

    if (depth != 1)
	setToWhiteOp = PIX_SET;
    else
	setToWhiteOp = PIX_CLR;

    while ((shiftDist >= 0) && (y1 < height))
    {
	y2 += delta;		/* calculate address of bottom of
				     * current horiz shear with
				     * 16 bits of fraction.
				     */
	if (ScaledInt1616_TO_Int32(y2) >= height)
	    incr = height - y1;
	else
	    incr = ScaledInt1616_TO_Int32(y2) - y1;

	    /* Write the block of lines to the frame so we
	     * can do a non-overlapped rasterOp.
	     */
	ERR_PASS( i_rasterOp (pIPWState, pImage,	/* ptr to dest */
			      0, 0,			/* destX, destY */
			      width, incr,		/* width, height */
			      depth,			/* dest depth */
			      bpl,			/* dest Bytes/line */
			      PIX_SRC,			/* just copy */
			      pImage,			/* source ptr */
			      cFrameBits,		/* srcX */
			      y1 + cFrameBits,		/* srcY */
			      depth,			/* src depth */
			      bpl));			/* src bytes/line */

	    /* Copy the relevant portion of the block to the
	     * correct portion of the image.  If the relevant portion
	     * is null (width <= shiftDist), skip this step.
	     */
	if (width > shiftDist)
	{
	ERR_PASS( i_rasterOp (pIPWState, pImage,	/* ptr to dest */
			      shiftDist + cFrameBits,	/* destX */
			      y1 + cFrameBits,		/* destY */
			      width - shiftDist,	/* width */
			      incr,			/* height */
			      depth,			/* dest depth */
			      bpl,			/* dest Bytes/line */
			      PIX_SRC,			/* just copy */
			      pImage,			/* source ptr */
			      0,			/* srcX */
			      0,			/* srcY */
			      depth,			/* src depth */
			      bpl));			/* src bytes/line */
	}

	    /* If the wings are supposed to recieve data from the
	     * image, Set the hole just vacated to white and
	     * copy leftover bits from buffer in top frame
	     * to right wing.
	     * Otherwise, copy bits from the left wing to
	     * the image to fill the hole.
	     */
	if (wingIsDest)
	{
	    ERR_PASS( w_rasterop(pIPWState,
				 0, 0,			/* srcX, srcY */
				 NULL,			/* src image */
				 setToWhiteOp,		/* clear it */
				 0,			/* destX */
				 y1,			/* destY */
				 MIN(width, shiftDist),	/* width */
				 incr,			/* height */
				 prS));			/* dest image */

	    if (width > shiftDist)
	    {
		ERR_PASS( i_rasterOp (pIPWState,
				      pRightWing,	/* dest ptr */
				      0,		/* destX */
				      y1,		/* destY */
				      shiftDist,	/* width */
				      incr,		/* height */
				      depth,		/* dest depth */
				      wingBpl,		/* dest bpl */
				      PIX_SRC,		/* just copy */
				      pImage,		/* source ptr */
				      width-shiftDist,	/* srcX */
				      0,		/* srcY */
				      depth,		/* src depth */
				      bpl));		/* src bpl */
	    }
	    else	/* copy a fragment from the image to the right wing */
	    {
		ERR_PASS( i_rasterOp (pIPWState,
				      pRightWing,	/* dest ptr */
				      shiftDist - width,	/* destX */
				      y1,		/* destY */
				      MIN(width,
					 wingW -
					 (shiftDist - width)), /* width */
				      incr,		/* height */
				      depth,		/* dest depth */
				      wingBpl,		/* dest bpl */
				      PIX_SRC,		/* just copy */
				      pImage,		/* source ptr */
				      0,			/* srcX */
				      0,			/* srcY */
				      depth,		/* src depth */
				      bpl));		/* src bpl */
	    }
	}
	else	/* fill hole with data from left wing */
	{
	    ERR_PASS( w_rasterop(pIPWState,
				 wingW - shiftDist,	/* srcX */
				 y1,			/* srcY */
				 prLeftWing,		/* src image */
				 PIX_SRC,		/* clear it */
				 0,			/* destX */
				 y1,			/* destY */
				 MIN(width, shiftDist),	/* width */
				 incr,			/* height */
				 prS));			/* dest image */

	}
	y1 = ScaledInt1616_TO_Int32(y2);	/* prepare for next shift */
	shiftDist += shiftDistDelta;
    }
	/* update the client's y1 and y2 ptrs */
    *pY1 = y1;
    *pY2 = y2;
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Shears an image and two side images vertically in place.
 *
 * LongDescription:::
 *  When rotating an image in place using shear operations, two of the
 *  operations are horizontal shears.  This routine is used to implement
 *  them.<BR>
 *
 * PrivateDescription:::
 *  In addition to the original image, this routine is supplied with
 *  two wing images, one on either side.  In the first horizontal shear,
 *  the wing serve to catch bits that fell off the edges of the main
 *  image.  In the second, and final, horizontal shear (after the vertical
 *  shear in the middle), the wings are sources of bits for the
 *  main image..
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  shearSlope::     The slope of the shear expressed as a ratio
 *		     of x units / y unit.  This number has 16 bits of
 *		     fraction and 16 bits of integer.  This number
 *		     is positive or negative.
 *  prS::            Pointer to image to be rotated.
 *  prLeftWing::     Pointer to wing image on the left of the real image.
 *  prRightWing::    Pointer to wing image on the right of the real image.
 *
 * Notes:::
 *  We assume that the rotation angle is between (-45 <= angle <= 45).
 *  It's also assumed that the shearSlope is tan(angle/2).
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, arbitrary, non-interpolated
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  To read about the shearing method of rotation, see "A Fast Algorithm for
 *  General Raster Rotation," pp 179-195 by Alan Paeth in Graphics Gems
 * (edited by Andrew Glassner), Academic Press, Inc., 1990.
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_verticalShearIP(
    IPWState           *pIPWState,
    Float32             shearSlope,
    PIXR               *prS,
    PIXR	       *prLeftWing,
    PIXR	       *prRightWing)
{
    PROLOGUE;
    Int32		 shiftDist;
    Int32		 width, height, depth, fullHeight;
    Int32		 wingW;
    Int32		 x1, incr;
    Int32		 setToWhiteOp;
    ScaledInt1616	 delta, x2;
    UInt8		*pImage, *pRightWing, *pLeftWing;
    UInt8		*pImageTemp, *pRightWingTemp;
    Int32		 bpl, wingBpl;
    Int32		 bplTemp, wingBplTemp;

	/* Prepare for the vertical shear.  If this is clockwise
	 * rotation, no preparation is needed.  If this is a
	 * counterclockwise shear, we invert the bytes/line of all
	 * images and start them at their bottoms.  This turns a
	 * shear that goes up on the left and down on the right
	 * into one that goes down on the left and up on the right.
	 */
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &width));
    ERR_PASS( w_pixrGetWidth(pIPWState, prLeftWing, &wingW));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &height));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetBpl(pIPWState, prLeftWing, &wingBpl));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS,
					 &fullHeight));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pImage));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prLeftWing,
					&pLeftWing));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prRightWing,
					&pRightWing));

	/* initialize the operation used to set portions of the image
	 * to white.
	 */
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
    if (depth > 1)
	setToWhiteOp = PIX_SET;
    else
	setToWhiteOp = PIX_CLR;

    if (shearSlope < 0)
    {
	   /* point to the bottom of all three images. */
	ERR_PASS( w_pixrSetImageAll(pIPWState, prS,
				    pImage + (fullHeight-1)*bpl));
	ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prLeftWing,
					    pLeftWing + (height-1)*wingBpl));
	ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prRightWing,
					    pRightWing + (height-1)*wingBpl));
	    /* negate the bytes/line */
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prS, -bpl));
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prLeftWing, -wingBpl));
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prRightWing, -wingBpl));
    }


	/* If the rotation direction is clockwise, we want to shift
	 * the left side of the image up and the right side down.
	 * We want to shift the left wing up and the right wing
	 * down too.
	 * There is a sort of cute trick here.  Since we really only care
	 * about the wing data that crosses the horizontal centerline
	 * of the wing (that's the only data that will be shifted into
	 * the image in the next horizontal shear), we limit the
	 * shears to just move that block.  This has the other nice
	 * side effect of guaranteeing that we don't overwrite wing
	 * source data before moving it to the top as would happen if
	 * we just did a downward shear naively.
	 *
	 * Also note that the half (upper or lower) of each wing
	 * receiving data starts out white, so we don't need to clear
	 * any portion of it after the blits.  The portion of each wing
	 * that's sourcing data will be ignored in the final horizontal
	 * shear anyway, so we don't have to bother clearing it either.
	 *
	 * There's still the issue of clearing newly vacant parts of
	 * the real image.  We want to do this with horizontal rasterOps
	 * since they're faster.  How?
	 *.
	 * Say we move bits up, leaving space at the bottom of the image
	 * or wing.  Say the first shift is large and subsequent shifts
	 * are smaller.  This is what happens if we work from the outside
	 * of the image towards the center.  We know that the distance
	 * from the bottom edge of one sheared block to the bottom edge
	 * of the next one is 1 row.  Each time we finish shifting a row
	 * up, we simply clear all the bits in the bottom row where the block
	 * now ends that are between the outside edge of the row and
	 * the outside of the image.   If we're moving down, we do the
	 * same thing from the top row of the moved block.
	 * Which edge of the image to we clear towards?  There are
	 * two constraints.  You always want to clear towards the nearest
	 * outside edge (so you don't go through the rest of the dest
	 * image) and you always want to clear backwards from the direction
	 * you're processing (so you don't overwrite source data that
	 * hasn't been moved yet).  So, we want to do all shears from
	 * the outside towards the center.
	 *
	 * Ooops, that means we can process the image from left to right
	 * from the left edge to the center, but we have to run out
	 * to the right edge before we can begin working back in.  We
	 * have to run out to the right edge so the vertical shear
	 * boundaries will be consistent on the way back in.  Duriung
	 * the trip out to the right edge, we update the x coords where
	 * shear boundaries lie so we can run the process backwards
	 * on the way back in.
	 */

	/* calculate # of pixels per one bit shift */
    delta = (ScaledInt1616)(ABS((Float32)0x10000 / shearSlope));

	/* do the vertical shear */

    x1 = 0;		/* start at outside of left wing */
    x2 = delta;
	/* calc distance cols will be shifted to the "up" */
    shiftDist = (Int32)(ABS(shearSlope) * (Float32)((width + 2*wingW)/2));

	/* this loop works across the left wing. */
    while (ScaledInt1616_TO_Int32(x2) <= wingW)
    {
	incr = ScaledInt1616_TO_Int32(x2) - x1;

	    /* We really want the top line of the block to be moved
	     * to be at coord height/2.  It will move up by shiftDist.
	     * We only care about the portion of the block that fits
	     * into the upper half of the left wing.  Therefore,
	     * we could get away with moving a block of height
	     * shiftDist.  What happens if shiftDist is very large,
	     * say larger than height/2?  Well, then we would try
	     * to read below the bottom of the image and that
	     * wouldn't be a good thing.  Besides we want to
	     * make sure that any bits below the bottom of the
	     * shifted block stay white, so we don't want to
	     * have w_rasterop putting garbage there.
	     */
	if (height/2 >= shiftDist)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 x1,		/* srcX */
					 height/2,	/* srcY */
					 prLeftWing,	/* src image */
					 PIX_SRC,	/* just copy */
					 x1,		/* destX */
					 height/2 - shiftDist, /* destY */
					 incr,		/* width */
					 shiftDist,	/* height */
					 prLeftWing));	/* dest image */
	else  /* we're trying to shift above the top of the image.
	       * Just insert the portion that fits.
	       */
	if (shiftDist < height)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 x1,		/* srcX */
					 shiftDist,	/* srcY */
					 prLeftWing,	/* src image */
					 PIX_SRC,	/* just copy */
					 x1,		/* destX */
					 0,		/* destY */
					 incr,		/* width */
					 height - shiftDist, /* height */
					 prLeftWing));	/* dest image */


	x1 = ScaledInt1616_TO_Int32(x2);
	x2 += delta;
	shiftDist--;
    }

	/* At this point, we may either be completely done with
	 * the left wing or may be trying to do a slice that
	 * exists on both the left wing and the image.  Look
	 * for that case.
	 */
    if (x1 < wingW)		/* next slice bridges the gap */
    {
	incr = wingW - x1;
	    /* transfer the section of the slice that's on the
	     * left wing.
	     */
	if (height/2 >= shiftDist)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 x1, height/2,	/* srcX, srcY */
					 prLeftWing,	/* src image */
					 PIX_SRC,	/* just copy */
					 x1,		/* destX */
					 height/2 - shiftDist, /* destY */
					 incr,		/* width */
					 shiftDist,	/* height */
					 prLeftWing));	/* dest image */
	else
	if (shiftDist < height)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 x1,		/* srcX */
					 shiftDist,	/* srcY */
					 prLeftWing,	/* src image */
					 PIX_SRC,	/* just copy */
					 x1,		/* destX */
					 0,		/* destY */
					 incr,		/* width */
					 height - shiftDist, /* height */
					 prLeftWing));	/* dest image */
    }
	/* set x1 and x2 to be correct for the real image
	 * which restarts the x coord at 0.
	 */
    x1 = 0;
    x2 -= Int32_TO_ScaledInt1616(wingW);

	/* this is the loop for the left side of the real image.
	 * Here, we move the entire column up, then fill in the
	 * space to the left of the last row with white to clear
	 * out the space left over.
	 */
    while (shiftDist >= 0)
    {
	incr = ScaledInt1616_TO_Int32(x2) - x1;
	ERR_PASS_CLEANUP( w_rasterop(pIPWState,
				     x1, shiftDist,	/* srcX, srcY */
				     prS,		/* src image */
				     PIX_SRC,		/* just copy */
				     x1, 0,		/* destX, destY */
				     incr,		/* width */
				     MAX(0, height-shiftDist), /* height */
				     prS));		/* dest image */
	    /* clear all the stuff to the left of the bottom row.
	     * Eventually, this clears all the stuff below
	     * the bottom of the shifted stuff.
	     */
	if (height - shiftDist - 1 >= 0)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 0, 0,		/* srcX, srcY */
					 NULL,		/* src image */
					 setToWhiteOp,	/* just copy */
					 0,		/* dest X */
					 height-shiftDist - 1, /* destY */
					 x1,		/* width */
					 1,		/* height */
					 prS));		/* dest image */

	    /* get coords of the new column */
	x1 = ScaledInt1616_TO_Int32(x2);
	x2 += delta;
	shiftDist--;
    }

	/* Here we are ready to shift stuff DOWN.  Scary.
	 * First, walk out to the edge of the right wing.
	 * This lets us maintain x1 and x2 properly and keep
	 * the shear boundaries in the right spot.
	 */
    x1 = Int32_TO_ScaledInt1616(width + wingW);
    shiftDist = 1;	/* amount to shift the next block */
    while (x2 < x1)
    {
	x2 += delta;
	shiftDist++;
    }

	/* Convert these values do they're relative to the
	 * the left edge of the right wing instead of the
	 * left edge of the image.  That's because we'll
	 * shift the right wing first.
	 */
    x2 -= Int32_TO_ScaledInt1616(width);

	/* Now we have the final value for x2.  Start working
	 * in towards the center, shifting columns down on the
	 * way.  But first, set the starting
	 * address and bytes/line of prRightWing so that it
	 * looks like we're shifting "up towards the bottom."
	 */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prRightWing,
					&pRightWingTemp));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prRightWing, &wingBplTemp));
    ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prRightWing,
					pRightWingTemp +
					  (height-1) * wingBplTemp));
    ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prRightWing, -wingBplTemp));

    x1 = ScaledInt1616_TO_Int32(x2 - delta);	/* point to left
						     * end of col */
    while (x1 >= 0)
    {
	incr = ScaledInt1616_TO_Int32(x2) - x1;

	    /* As when we were vertically shearing the left wing,
	     * we have to decide whether this stuff is on the
	     * image before moving it.
	     */
	if (height/2 >= shiftDist)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 x1, height/2,		/* srcX, srcY */
					 prRightWing,		/* src image */
					 PIX_SRC,		/* just copy */
					 x1,			/* destX */
					 height/2 - shiftDist,	/* destY */
					 incr,			/* width */
					 shiftDist,		/* height */
					 prRightWing));		/* dest image */
        else	/* just shift the portion that fits */
	if (shiftDist < height)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 x1, shiftDist,		/* srcX, srcY */
					 prRightWing,		/* src image */
					 PIX_SRC,		/* just copy */
					 x1,			/* destX */
					 0,			/* destY */
					 incr,			/* width */
					 height - shiftDist,	/* height */
					 prRightWing));		/* dest image */

	x2 -= delta;
	x1 = ScaledInt1616_TO_Int32(x2 - delta);
	shiftDist--;
    }

	/* At this point, we're either entirely done with the right
	 * wing or the next block straddles it.  If we have a
	 * partial block, shift the part that's inside the right
	 * wing.
	 */
    if (x2 >= 0)		/* next slice bridges the gap */
    {
	incr = ScaledInt1616_TO_Int32(x2) + 1;

	    /* transfer the section of the slice that's on the
	     * right wing.
	     */
	if (height/2 >= shiftDist)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 0, height/2,		/* srcX, srcY */
					 prRightWing,		/* src image */
					 PIX_SRC,		/* just copy */
					 0,			/* destX */
					 height/2 - shiftDist, /* destY */
					 incr,			/* width */
					 shiftDist,		/* height */
					 prRightWing));		/* dest image */
	else
	if (shiftDist < height)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 0, shiftDist,		/* srcX, srcY */
					 prRightWing,		/* src image */
					 PIX_SRC,		/* just copy */
					 0,			/* destX */
					 0,			/* destY */
					 incr,			/* width */
					 height - shiftDist,	/* height */
					 prRightWing));		/* dest img */
    }

	/* Restore the right wing address and bytes/line */
    ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prRightWing,
					pRightWingTemp));
    ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prRightWing, wingBplTemp));

	/* Perform the same trick for the real image. */
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS,
					&pImageTemp));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplTemp));
    ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prS,
					pImageTemp +
					    (fullHeight-1) * bplTemp));
    ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prS, -bplTemp));

	/* Set x1 and x2 so they're relative to the left side of
	 * the image.
	 */
    x1 += width;
    x2 += Int32_TO_ScaledInt1616(width);

	/* Set incr out here because the first block always
	 * ends at the right side of the image.  Inside the
	 * loop, the right side of the block is at x2.
	 */
    incr = width - x1;

	/* this is the loop for the right side of the real image.
	 * Here, we move the entire column "up towards the bottom,"
	 * then fill in the space to the right of the last row with
	 * white to clear out the space left over.
	 */
    while (shiftDist > 0)
    {
	ERR_PASS_CLEANUP( w_rasterop(pIPWState,
				     x1, shiftDist,	/* srcX, srcY */
				     prS,		/* src image */
				     PIX_SRC,		/* just copy */
				     x1, 0,		/* destX, destY */
				     incr,		/* width */
				     MAX(0, height-shiftDist), /* height */
				     prS));		/* dest image */
	    /* clear all the stuff to the right of the top row.
	     * Eventually, this clears all the stuff above
	     * the top of the shifted stuff.
	     */
	if (height-shiftDist-1 >= 0)
	    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
					 0, 0,		/* srcX, srcY */
					 NULL,		/* src image */
					 setToWhiteOp,	/* just copy */
					 x1+incr,	/* dest X */
					 height-shiftDist - 1,	/* destY */
					 MAX(0, width-(x1+incr)),/* width */
					 1,		/* height */
					 prS));		/* dest image */

	    /* get coords of the new column */
	shiftDist--;
	x2 -= delta;
	x1 = ScaledInt1616_TO_Int32(x2 - delta);
	incr = ScaledInt1616_TO_Int32(x2) - x1;
    }

	/* Clear the top line of the image on the right side.  Note that
	 * we didn't do the shiftDist == 0 case, so this one didn't
	 * get cleared.  Note that x1 and incr are now set to do
	 * the shiftDist == 0 vertical block, so we just use them.
	 */
    ERR_PASS_CLEANUP( w_rasterop(pIPWState,
				 0, 0,				/* srcX, srcY */
				 NULL,				/* src image */
				 setToWhiteOp,			/* just copy */
				 x1+incr,			/* dest X */
				 height-shiftDist - 1,		/* destY */
				 MAX(0, width-(x1+incr)),	/* width */
				 1,				/* height */
				 prS));				/* dest image */

	/* restore the image address and bytes/line */
    ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prS, pImageTemp));
    ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prS, bplTemp));

	/* end of vertical shear */

    if (shearSlope < 0)
    {
	    /* Restore the image pointers and bytes/line */
	ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prS, pImage));
	ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prLeftWing, pLeftWing));
	ERR_PASS_CLEANUP( w_pixrSetImageAll(pIPWState, prRightWing,
					    pRightWing));
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prS, bpl));
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prLeftWing, wingBpl));
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prRightWing, wingBpl));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (shearSlope < 0)
    {
	    /* Restore the image pointers and bytes/line */
	w_pixrSetImageAll(pIPWState, prS, pImage);
	w_pixrSetImageAll(pIPWState, prLeftWing, pLeftWing);
	w_pixrSetImageAll(pIPWState, prRightWing, pRightWing);
	w_pixrSetBpl(pIPWState, prS, bpl);
	w_pixrSetBpl(pIPWState, prLeftWing, wingBpl);
	w_pixrSetBpl(pIPWState, prRightWing, wingBpl);
    }

    ERR_RETURN;
}





/**************************************************************************
 ***                        In-place rotation                           ***
 *************************************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate an image by a small angle in place.
 *
 * LongDescription:::
 *  Both w_rotateArb and wp_rotateSmallAngleIP perform non-interpolated
 *  rotations.  "Non-interpolated" means that the input pixels are just
 *  shuffled around to form output pixels, there is no averaging of inputs
 *  to form outputs.  Both w_rotateArb and wp_rotateSmallAngleIP can
 *  perform small angle rotations.  The difference is that the source and
 *  destination images for w_rotateArb are always different and the
 *  source and dest images for wp_rotateSmallAngleIP are always the same
 *  image.
 *<BR><BR>
 *  If the "nice" flag is set, the rotation routine attempts to minimize
 *  horizontal jaggies in the rotated image.
 *
 * Private Description:::
 *  A smallish memory buffer is allocated and freed internally.  This buffer
 *  holds portions of rotated image before they are put back into the
 *  input image.
 *
 * Input:::
 *  pIPWState::      Pointer to the standard IPWrap state block.
 *  angleRad::       Rotation angle in radians.  A positive angle causes
 *		     clockwise rotation, a negative angle causes
 *		     counterclockwise rotation.
 *  nice::           If set to cTrue, the rotation routine attempts to
 *		     eliminate jaggies in the output.  Setting this arg
 *		     to cTrue roughly doubles the rotation time.
 *
 * Output:::
 *  prS::            Pointer to the image and serves as both source and
 *		     destination.  This image may have more than one channel.
 *		     Each channel may be binary or gray.  Even though
 *		     the input image may be gray, no interpolation is
 *		     ever done.
 *
 * Returns:::
 *  In addition to the normal error codes, this routine returns
 *  cErrInvalidParm if the rotation angle is larger than 0.12 radians
 *  (about 7 degrees).  The rotation technique used produces unacceptable
 *  distortion for larger rotations.
 *
 * Notes:::
 *  If the rotation angle is less than 0.002 radians (about 0.1 degrees),
 *  the routine just returns without rotating.  This is because any rotation
 *  introduces jaggies.  The harm done by introducing the jaggies outweighs
 *  the miniscule rotation applied.
 *<BR><BR>
 *  This routine invokes an optimized rotation algorithm that is very fast,
 *  but is approximate.  For small angles, the distortion should not be
 *  noticable.
 *<BR><BR>
 *  Clients wishing to use a somewhat slower rotation routine that produces
 *  nicer results on gray images should look at w_rotateGray.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, small angle, non-interpolated
 *
 * Level:::
 *  heuristic, rotation is approximate
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_rotateSmallAngleIP (
    IPWState           *pIPWState,
    Float32             angleRad,
    Int32               nice,
    PIXR               *prS)
{
    PROLOGUE;
    UInt8		*pS;            /* pointer to source bits */
    Int32		 sBpl;          /* source bytes/line */
    Int32		 wS, hS;        /* width and height of source */
    Int32		 depth;         /* bits/pixel in source */
    Int32		 channels, i;
    ScaledInt1616	 xdpi, ydpi;

    ERR_GEN(prS == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		cChannelsAny, cFrame32);

    ERR_GEN(angleRad > SMALL_ANGLE, cErrInvalidParm);

	/* if the angle is really small, a rotation would introduce
	 * shear lines into the image without doing any noticable
	 * rotation.  Avoid this by just returning.
	 */
    if (ABS(angleRad) < MIN_ANGLE)
	NORMAL_RETURN;

    ERR_PASS( w_pixrGetChannelCount (pIPWState, prS, &channels));
    if (channels != 1)
    {
	PIXR   *sub;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel (pIPWState, prS, i, &sub));
	    ERR_PASS( wp_rotateSmallAngleIP (pIPWState, angleRad, nice, sub));
	}
	NORMAL_RETURN;
    }
    else
    {
	ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
	ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &wS));
	ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &hS));
	ERR_PASS( w_pixrGetBpl(pIPWState, prS, &sBpl));
	ERR_PASS( w_pixrGetDepth(pIPWState, prS, &depth));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xdpi, &ydpi));
	
	ERR_PASS(i_rotateNoInterpolationIP (
				pIPWState, pS, wS, hS,
				sBpl, angleRad, depth,
				(Int8) nice,
				(Int16)ScaledInt1616_TO_Int32(xdpi),
				(Int16)ScaledInt1616_TO_Int32(ydpi)));
	NORMAL_RETURN;
    }
}

/*

 $Log: shearops.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.38  1998/01/14 00:15:24  ddavies
 shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.37  1997/09/03 22:43:11  ddavies
 ddavies on Wed Sep 3 15:41:30 PDT 1997

 shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v

 Fix tiny bug in wp_shearC.  We now round the new width or height of the
 dest image up instead of down.  This should prevent loss of an extra
 (extra?) row or column.

 Revision 1.36  1997/06/26 20:04:43  ddavies
 shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.35  1997/04/15 22:41:53  ddavies
 ddavies on Tue Apr 15 15:41:40 PDT 1997

 shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.34  1996/10/05 02:38:21  ddavies
 shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v

 Fix bugs in large angle rotate in place that caused accesses outside the image
 for very wide images and very tall images.  We basically now check all the
 parameters to rasterOp, doing something sensible when the might either be
 negative or excessively positive.  There are LOTS of calls to rasterOp.

 Revision 1.33  1996/02/06 18:46:56  dferrell
 dferrell on Tue Feb 6 10:45:38 PST 1996

 shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v

 Truncated floating point range check on angle parameter so that the
 Mac under CodeWarrior 7.0 doesn't fail at the boundary of the range.

 * Revision 1.32  1996/01/04  20:26:47  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.31  1995/12/21  04:46:41  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Modify w_rotateArb so that it can do rotations in place when *prD == prS.
 * Convert w_smallAngleRotateIP to the private function
 * (whoops, I meant w_rotateSmallAngleIP) wp_rotateSmallAngleIP.
 * Create the new private function w_rotateLargeAngleIP, which calls
 * wp_horizontalShearIP (which calls wp_horizontalShearLeftIP and
 * wp_horizontalShearRightIP) and wp_verticalShearIP.  When asked to
 * rotate an image with the resizeToSourceFlag set to cTrue AND where
 * the input and output images are different, w_rotateArb copies the
 * input to the output and calls the in-place rotation functions.  This is
 * sort of groty, but the actual performance effect should be pretty
 * small.  Besides, I think rotation in place is by far the most common
 * case.
 *
 * Revision 1.30  1995/12/13  04:13:05  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/12/13  02:06:17  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Add checks in rotate
 * Arb (ooops) so that it will not attempt a large angle rotate on an anisotropic
 * image (x resolution != y resolution).  cErrInvalidParm is returned in
 * this case.
 *
 * Revision 1.28  1995/11/13  20:16:00  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.27  1995/11/03  22:49:00  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.26  1995/11/01  22:27:37  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Change the way rotateArb operates on color (multi-plane) images.  In the
 * old world (before this change), it went down to the lowest levels before
 * looking at the number of channels.  When using the shear method of rotation,
 * this resulted in allocating n (number of channels) copies of the wider
 * image, then n copies of the wider&taller image, then n copies of the
 * wider&taller&wider image.  The new version of rotateArb simply calls
 * itself for each plane of the source image.  This means that one plane
 * goes all the way through the source -> wider -> wider&taller ->
 * wider&taller&wider -> (image same size as source chopped out of the middle)
 * sequence at a time.  This should dramatically reduce the amount of memory
 * required for this operation when using multi-plane images.
 *
 * Revision 1.25  1995/10/13  22:05:12  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.24  1995/10/12  04:28:16  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.23  1995/10/05  04:27:53  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.22  1995/10/03  01:20:53  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.21  1995/09/21  22:59:37  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.20  1995/09/19  21:12:38  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Fixed transfer of PIXR DPI to IPShared params by using macro
 * ScaledInt1616_TO_Int32()
 *
 * Revision 1.19  1995/09/15  18:01:20  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.18  1995/09/11  22:11:55  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.17  1995/09/10  00:33:56  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.16  1995/09/07  01:35:18  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.15  1995/09/06  20:23:25  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.14  1995/09/01  01:18:46  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.13  1995/08/31  21:39:11  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.12  1995/08/25  01:38:24  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Fixed warnings
 *
 * Revision 1.11  1995/08/22  02:21:54  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.10  1995/08/17  22:03:12  ejaquith
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.9  1995/07/29  04:50:32  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.8  1995/02/03  01:44:40  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Have rotateArb pass the x and y resolutions down to i_rotateNoInterpolation
 * so it can shear anisotropic images.  The shear is needed to account for
 * the shear naturally present in those images.  Also have rotateArb set the
 * resolution in all PIXRs is creates and returns.
 *
 * Revision 1.7  1995/01/07  02:59:52  ddavies
 * ddavies on Fri Jan 6 18:58:51 PST 1995
 *
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * In shearCPr, set the edges of gray images to white before shearing into
 * them.  This eliminates the ugly black triangles inserted into images that
 * are rotated by shearing.
 *
 * Revision 1.6  1994/10/07  19:18:06  lperry
 * lperry on Fri Oct 7 12:18:01 PDT 1994
 *
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.5  1994/04/12  01:16:42  macleod
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.4  1994/03/02  04:01:03  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Ooops, the message for rotatop1 should have said "during 90, 180 and
 * 270 degree rotates."  Added IP_YIELD to give up control after creating dest
 * images during rotateArb and rotateSmallAngleIP.
 *
 * Revision 1.3  1993/12/03  01:18:57  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:17:02  ddavies
 * shearops.c -> /product/ipcore/ipwrap/src/RCS/shearops.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

