/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  skew2.c:  skew detection by the DDavies sampling methods.
 *
 *          Contains subroutines
 *
 *                w_findSkewSample()
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "skew.pub"
#include "shrskew.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: skew2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/* This version of skew code is a variant of the original skew.c.  The
 * job is to detect and correct image skew.  It depends on
 * images having a lot of horizontal stuff (lines, text).
 *
 * The original version, skewShear() in skew1.c, accepted binary
 * images as input and produced them as output.
 * We would like to improve the speed and quality of skew
 * detection  [who wouldn't??].   We would also like to explore
 * gray scale rotation for skew correction.  We'll work on skew detection first.
 *
 * The detection algorithm in findskewShear() had three major steps:
 *
 *	1. Reduce image to a tractable size.
 *	2. Coarsely rotate the reduced image over a range of angles (say
 *	   -10 degrees to +10 degrees in 1 degree increments) to try to
 *	   find the position in which the skew is minimized.
 *	3. Use a bifurcation algorithm around the best angle found in the
 *	   sweep to fine tune the results.  Stop looking when the amount by
 *	   which the image will be rotated for the next step reaches a
 *	   value passed in as a parameter.
 *
 * The method used to obtain a "skew score" for a particular angle is to
 * vertically shear the reduced image by the angle and count the number of
 * pixels in each horizontal line.  The score equals the sum of the squares
 * of the differences in pixel counts across all adjacent lines.  In effect,
 * we're looking for skew angles at which we find the steepest horizontal
 * edges.  The squaring operation sharpens our sensitivity to steep edges.
 *
 * We'd like to explore improving this algorithm on two fronts.  First,
 * information is lost when the image is reduced.  This degrades our ability
 * to get good answers even with an ideal detection algorithm.  We'd like
 * to avoid reducing the image.
 *
 * Second, the process of continually shearing the image is relatively slow,
 * especially for a full scale image.  We'd like to avoid shearing.
 *
 * Why is the image sheared?  Because it's easier to count pixels
 * horizontally than diagonally.  However, if we could count diagonally,
 * we could dispense with the shearing, hence with the reduction.  We will
 * do so.
 *
 * We also note that it's not necessary to count every pixel in the image for
 * every potential shear angle.  Skew detection is a matter of statistics.
 * The more image samples we make, the better the statistics.  However, there
 * is probably some point of diminishing returns, but we don't know where
 * that is now.  We'll start by sampling roughly every other long word
 * horizontally and a pair of lines out of every 2*reductionFactor lines
 * (see below) vertically.  For a full scale 2300 x 3300 image, that's
 * about 32 horizontal samples and (reduction factor of 8), 400 samples
 * vertically, for a total of 12,800 samples.  One of the interesting jobs
 * will be to see how much this can be reduced while retaining sufficient
 * accuracy.
 *
 * Not only is it not necessary to count every pixel in a line, it's really
 * not necessary to count every pixel in a word.  If we think of looking for
 * horizontal rivers of white space, we can get by with saying the value of
 * a particular 32 bit sample is "1" if there are any bits on and 0 otherwise.
 * This improves counting speed.
 *
 * The original deskewShear() had three levels of code hierarchy.
 * The top level, deskewShear(), used findSkewShear() to find the skew
 * angle and rotateArb() to correct the image.  In turn,
 * findSkewShear() used findSkewShearSweep() and findSkewShearBif() to
 * do the (coarse) sweep and (fine) bifurcation steps respectively.
 * Each of those reduced the image and used shearC() and
 * findDiffPowerSum() to compute the scores.
 *
 * The sweep stage is used because the bifurcate step can otherwise choose a
 * local, but not global, maximum in skew score.  The number of false peaks
 * and amount of general noise of a hypothethical skew score vs angle curve
 * appears to be strong function of reduction factor.  This is because when
 * the program effectively looks at every 4th or 8th line, it's resolution
 * for edge discrimination is reduced.  Naturally, you say.  Edges can only
 * be determined to within 4 or 8 lines.  However, in addition to this
 * effect, the algorithm starts to lose its ability to distinguish between
 * one long edge and two separate horizontal edges which are slightly
 * displaced relative to each other.  For example, take a page with three
 * columns of text.  Say the inter-line spacing in all columns is identical,
 * but the columns are verticaly shifted relative to each other.  This is
 * reasonably common.
 *
 * Say the skew detection algorithm samples along a slanted line that goes
 * through all the black pixels of every column even though it is misaligned
 * with each column individually.  If the next sample line is far enough down
 * (due to reduction), it can go through all the white pixels of every column
 * even though it is also misaligned with each column.  The difference in
 * pixel counts between these two lines will be high, giving a misleading
 * high score.  I have an image where this happens - the "deskewed" image is
 * consistently skewed by about 0.5 degrees.
 *
 * One way to fix this is to sample every line of the image.  That's too
 * expensive.  A cheaper way is to sample adjacent pairs of lines of the
 * original image.  That is, instead of sampling every 4th or 8th line, as
 * is effectively done by reduction, sample 2 lines, then skip 6 or 14.
 * The skew score is calculated only between these pairs.  This is what we'll
 * do.
 *
 * Sampling adjacent pairs of lines has the effect of giving the skew score
 * curve a VERY thin, spikey peak.  This gives wonderful final resolution if
 * the algorithm can find it.  If the curve has no slope except in a very
 * thin region around the peak, the peak becomes a needle in the haystack.
 *
 * The spike will be widened by the fact that text lines, a prominent
 * horizontal feature do not present perfect horizontal surfaces.  They're
 * a bit noisy, so will contribute non-zero skew scores when they are close
 * to aligned.
 *
 * Unfortunately, even with this wonderful technique, we cannot eliminate
 * false peaks in the curve.  Even if the skew evaluation were perfect, pages
 * with slanted areas (almost any page with a picture on it) will have
 * blips in their skew curves.  These blips can potentially fool a pure
 * bifurcation algorithm.
 * 
 * The brute force way to identify the real peak is to sweep across the entire
 * band of angles.  The sweep can be a bit coarse, using bifurcation to home
 * in on the exact value.  But it's boring to be brutes.  Slow too.
 *
 * Another way to use two methods of generating skew scores.  One isn't very
 * exact, but produces a wide curve with enough amplitude to swamp out most
 * of the noise.  The other produces a thin, spikey curve for good resolution.
 * The second method is the one described above.  What function would produce
 * a big, wide curve?  Simply performing the previous algorithm with fewer
 * samples flattens the peak of the curve, but also decreases the signal to
 * noise ratio.
 *
 * A property of text lines is that they are very long, thin, horizontal bars.
 * We could construct a set of slanted paths across the image, counting
 * ON pixels encountered by these paths.  Higher counts would tend to
 * indicate that the paths are stay inside text lines.  However, this measure
 * is sensitve to both local orientation of text lines (which is good) and
 * global placement of lines and graphics across the page (which is bad).
 * A line's score can be high simply because it goes through many
 * independent text lines or through two or three text lines in different
 * columns that are displaced relative to each other.
 *
 * Instead, we sample pixels along linear paths in the image, remembering for
 * each path the *maximum number of consecutive pixels* encountered.  A path
 * that stays inside one text line will get a high score.  A path that goes
 * through a large number of text lines will get a low score since it will not
 * see a large section of any one of the lines.
 *
 * This scoring scheme doesn't give very good resolution since it can't
 * distinguish any two paths that follow a text line across the page.  Its
 * minimum resolution is the height of a text line.  However, it does produce
 * a nice broad curve that's reasonably insensitive to noise.  As the angle
 * of the paths relative to the text lines increases, the skew scores fall off
 * almost linearly.
 *
 * We could use bifurcation with the MaxPixels curve to find the skew angle
 * to within a degree, then use the HorizontalEdge method to fine tune. 
 * This is probably what we'll end up doing.  However, just to be adventurous,
 * we'll try another scheme first.  We assume the broad curve is pretty
 * linear.  We find a couple of points on each slope (four samples in all),
 * compute the angle at which the slopes intersect and use the bifurcation
 * method starting from there.  If that doesn't give a sufficiently accurate
 * first guess, we'll go back to bifurcation with the wide curve.
 *
 * How do we do bifurcation?  The original skew.c used 5 points, choosing one
 * of the interior three as the new center, then halving the angular distance
 * between them.  The number of samples required for this method is
 * 5 + 2 * number of steps.  Initially, we'll try a method that relies on the
 * the curves being fairly smooth.  We pick three inital points separated by
 * the maximum initial separation.  While the middle sample is not the highest,
 * take one more sample at the maximum separation on the other side of the
 * highest endpoint.  Once the middle is the highest, take a sample half way
 * between the middle and the next highest endpoint.  If everything were
 * smooth and symmetric, we'd know that the highest point was somewhere
 * in the range between the two highest samples, so we could just keep taking
 * a new sample between the two highest values and find it.  This gives a
 * total number of samples of 3 + number of samples.  If the curves for real
 * images are too noisy, we'll just repeat the first part of the algorithm -
 * taking samples on the far side of the highest endpoint until the middle
 * sample is the highest, then halving the separation and taking a sample
 * between the two highest samples.
 *
 * Efficiently tracing diagonal paths is more complex than I originally
 * thought.  Moving horizontally is easy, figuring out when to move
 * vertically and by how much is tricky.  To get sufficient accuracy,
 * we need to move vertically by fractional pixels.  Well, at least we need
 * to approximate fractional pixel movement.  This can be done pretty well
 * by moving up or down in units of whole lines at selected intervals.
 *
 * The job of plotting a diagonal path through the image is reduced then to
 * determining the size and frequency of vertical movement as a function of
 * angle.  One way to approach the problem is to pick an angle,
 * compute its tangent, multiply that by the width of the image (giving the
 * total rise at the end of the path) and divide the result by the number of
 * horizontal samples to be taken.  Yikes!  A trig function, a multiply and
 * a divide for every sample line!  Simple for scientists, but scary
 * for engineers.
 *
 * Instead of using angles as the space through which we search or sweep,
 * say we use vertical path displacement.  For small angles, the vertical
 * displacement is very nearly linear with angle anyway.  We can compute the
 * maximum positive and negative vertical displacments of the initial
 * search space as well as the vertical displacement associated with the
 * initial deltaAngle.  We then search or sweep through these displacements
 * skipping the number of lines associated with deltaAngle at each step.
 * To do bifurcation, we can just divide the incremental number of displacment
 * lines in half.  All this eliminates the trig function and the multiply
 * in the original algorithm.
 *
 * This is still a little messy because the width of the image still figures
 * into the equations.  We note that the angle of a path is (obviously)
 * independent of the width of the image.  In fact, we could use a hypothetical
 * image of any convenient width to compute vertical displacements so long as
 * the code that's traversing the real image checks to see when it's reached
 * an edge.  Convenient image widths are powers of 2 so divides and multiplies
 * turn into shifts.  We will follow this approach - a hypothetical image
 * used to compute vertical displacements that are used in turn to traverse
 * the real image.
 *
 * We discussed simulating fractional vertical line movement by occasional
 * full line movement.  For example, if the path rises 10 lines in
 * 20 samples, we move up one line every other sample.  Rather than have the
 * sampling code compute some marvelous error function, we'll just define
 * when vertical movements should be taken in a set of bit vectors.  We will
 * not produce a set of bit vectors for every possible image width.  Instead,
 * we will produce sets of fairly short bit vectors (16 bits) that can be
 * reused if the code takes more than 16 horizontal samples in a line.
 *
 * Nice theory, but what really happens?  Here are the steps in calculating
 * the necessary vertical displacement and choosing a bit vector corresponding
 * to the angle of the path.
 * 	1. Decide on the spacing between horizontal samples.  One algorithm
 *	   might want a 4 byte sample every 4 bytes.  Another might take a
 *	   4 byte sample every 12 or 16 bytes.
 *	2. Since our traversal algorithm will be periodic every 16 samples,
 *	   multiply the sample distance chosen in step 1 by 16 (shift left 4).
 *	   This gives the width of the hypothetical image.
 *	3. To determine the vertical offset in the hypothetical image for
 *	   maxAngle, multiply tan(maxAngle) by the width of the hypothetical
 *	   image computed in step 2.  Repeat for minAngle and delAngle.
 *	4. Sweep or search through the space of vertical offsets between
 *	   maxAngle's offset and minAngle's offset using an (initial)
 *	   increment of delAngle's offset.
 *	5. The intra-sample vertical offset for a particular path comes in
 * 	   two parts, integer and fraction.  The integer part is added to
 *	   the address on every sample.  This is simply equal to the
 *	   hypothetical vertical offset (the things we're searching through)
 *	   divided by 16 (shifted right 4).  The fractional part equals the
 *	   hypothetical vertical offset MOD 16 (AND with 0xF).  This is used
 *	   to choose a bit vector.  The bit vector is shifted left one place
 *	   after taking each sample.  If the bit that fell off the end was a
 *	   1, the bytes/line is added to (subtracted from) the next sample
 *	   address.
 *
 ***************************************************************************
 *
 * That worked, but we can do better in the coarse search.  Cycling through
 * (and reloading) the bit vectors is harder than we'd like.  We still end
 * up computing too many trig functions, doing too many multiplies and
 * divides, etc.  There are also too many cache refills.
 *
 * It turns out that a rise of 0.5 bits across 32 bits is 0.9 degrees.  That's
 * pretty close to 1 degree, which is the traditional value of deltaAngle
 * for coarse skew detection.
 * Anyway, it's good enough for a coarse search.  We'll divide the range from
 * minAngle to maxAngle into vertical displacements that change in units of
 * 8 bits.
 *
 * The assembly code that actually computes scores does so in a slightly
 * unrolled loop.  On every other read, it adds rSrcPitch to the source
 * address.  The code passes 0, +sourcePitch or -sourcePitch in this argument
 * depending on whether the current vertical displacment/32 horiz bits
 * is an integer (0), a positive non-integer (+sourcePitch) or a
 * negative non-integer (-sourcePitch).  Things become simpler, hence faster.
 *
 * We can improve the cache hit ratio on sweep detection algorithms by
 * dividing the image into blocks and doing the entire sweep on each block,
 * then moving on to the next block.  After a few passes, most of the bits
 * in a block will already be in the cache, so the remaining passes will go
 * faster.  However, if the blocks aren't wide enough, we end up adding lots
 * of little maximums to the skew score instead of discarding them in favor
 * of fewer, bigger maximums.  This flattens the score curves.
 *
 * Since the slant calculation for the coarse algorithm is independent of
 * the width of the block or image, we might as well make blocks cover the
 * entire width of the image.  This will cause more cache misses than we'd
 * like, but will improve the accuracy of the result.
 *
 ***************************************************************************
 *
 * That was successful and still pretty fast.  Now, we change from a sweep
 * to a binary search.  In the search algorithm, we sample both extreme grid
 * points and the middle.  We then continue to sample half way between the
 * two highest grid points until we reach a sample interval of 1 degree.
 *
 * Well, almost.  Say the image is almost centered, the skew angle is just
 * a bit positive (for example).  Noise in the skew function can cause
 * minAngle's score to be a little bigger than maxAngle's score.  This has us
 * eliminating all the positive angles in the first step, a mistake.  Instead
 * we don't adjust the trial angle quite so far in the first couple of steps.
 * For example, instead of curMaxAngle = (curMaxAngle+curMinAngle)/2, we say
 * curMaxAngle = (curMaxAngle+curMinAngle)/2 + (curMaxAngle-curMinAngle)/8.
 * The second term shortens the first couple of steps.
 *
 ***************************************************************************
 *
 * The fine skew detection algorithm will also use a binary search.  Instead
 * of using angles, we deal in terms of vertical displacment of slanted lines
 * over a hypothetical 4k bit horizontal span.  Yeah, but real images aren't
 * 4K bits wide, right?  This just means that if the slanted line would have
 * risen n bits over a 4Kb stride, it rises n*(imageWidth/4K) bits over a
 * imageWidth stride.
 *
 * Since everything is proportional, why not choose a smaller hypothetical
 * stride?  If we choose a stride smaller than the width of a real image, we
 * get insufficient angular resolution.  In an extreme case, say the
 * hypothetical image were 32 bits wide and the smallest angle we could
 * handle had 1 bit rise in 32 bit stride.  This corresponds to a rise of
 * 80 lines in a 2560 bit real image.  We'd like a resolution of +- 1 line
 * in 2560 bits, so need to choose a hypothetical image wider than the real
 * one.  The next larger power of 2 is 4K.
 *
 * The fine skew detection algorithms is based on adding up differences
 * vertically adjacant samples.  It's not necessary to sample every point
 * along each grid line, so we'll spread the samples out a bit.  Initially,
 * we'll try sampling 32 bits (8 bits?) out of every 128.  That would give
 * 4K/128 = 32 samples in the hypothetical image.  Real images are less wide,
 * so we have to stop when we reach the edge.
 *
 * Since we'll need good accuracy, we can't get away with the 1 degree trick
 * used by the coarse routine.  In fact, we need to be able to rise or fall
 * 0 to 31 lines smoothly over the course of 32 samples.  There is a
 * cute way to do this using one counter setting the interval at which
 * the pitch is added to the address and another counter setting the interval
 * for subtracting the pitch (error correction).  However, since we know
 * there will never be more than 32 samples, we'll just use a bit vector
 * to decide when to add the (positive or negative) pitch to the current
 * address when moving "horizontally".
 *
 *********************
 * The original version of this just got noise, no peak to the curve.  More
 * samples are probably needed.  But that brings up the question, "Why are
 * so many samples needed?"  Because very few of them actually find edges
 * to detect.  We need to take lots of samples so that some interesting
 * number actually find edges to measure.
 *
 * That sounds awfully inefficient.  What if we looked for edges instead of
 * simply hoping that they'd show up sometimes?  This is the idea behind
 * findSkewSampleRobot.
 *
 * Say we view an image as a maze of text and data also having white and
 * black deserts.  The text and data (especially text) have nice horizontal
 * edges we'd like to find and whose angle we'd like to record.  The deserts
 * are big vacant areas.
 *
 * We'll pick a bunch of points (not at random, that takes too long).  An easy
 * set of points forms a big X across the image.  At each image point, we put
 * down a little "robot."  If the little fellow lands in a white space
 * (all 32 bits are white), it looks vertically down for some black.  It
 * samples every 16th line.  If no black is found in a few samples, the
 * robot concludes it's in a desert and gives up (returns).
 *
 * If the robot finds some black, or lands in black to begin with, it
 * looks below for white.  It samples every 8th line (white space being
 * generally thin between text lines).  If no white is found in a few samples
 * the robot concludes it's in a black desert and gives up.
 *
 * If white space is found, we have a black/white edge.  These tend to be
 * the best kind since text is usually darker than the background (not
 * always, but tough) and the underside of a text line isn't as jagged as
 * the top side.  We do a short binary search until we have found the edge
 * exactly.
 *
 * We now attempt to follow the edge to the right.  Starting on the white
 * line under the black line, we sample the next word to the right.  If it's
 * also white, we look up to see if the black stuff is still there.  If
 * so, we record that one horizontal step was taken and no vertical steps
 * were taken.  If not, we keep looking up for 5 lines.  If the black stuff
 * is found, we record one horizontal step and as many vertical steps
 * (a negative number) as were required to find the black stuff.  We also
 * move up to the line immediately below the black stuff.  If no black
 * stuff was found, we conclude that the edge has ended (or was too steep
 * to consider) and quit.
 *
 * If the step to the right found black stuff, we look down for up to 5 lines.
 * If white stuff is found again, we record one horizontal step and as many
 * the number of lines (this time a positive number) as required to find white.
 * We also move down.  If no white is found, we just quit.
 *
 * The previous two steps are repeated so long as the edge is seen.
 *
 * The purpose of all this is to record the slope of the edges.  We pick some
 * number of total horizontal steps to record and keep putting down robots
 * until that number is found.  We then get the slope by dividing the total
 * number of vertical steps by the total number of horizontal steps.
 *
 * Clearly, we need to traverse the image several times to get rid of noise.
 * The number of horizontal steps required will be determined by
 * experimentation.
 *
 * A more complex version of this code could maintain several vertical
 * counters for different slopes.  We'd then pick the biggest one as the
 * dominant slope.  As it is now, slanted lines in the image contribute to
 * the average slope calculation.
 *
 **************************************************************
 *
 * We had some problems.  First, there weren`t enough samples.  We now
 * take sample points in columns.  The sample points within a column are
 * close enough to catch the majority of text lines.
 *
 * Second, the samples were so wide that we could bridge across slanted
 * text rows.  The sample width is now 16 bits instead of 32.
 *
 */ 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Find skew angle of an image.
 *
 * LongDescription:::
 *  Find the skew angle, in radians, of an image.  This function depends
 *  primarily on being able to find lots of horizontal stuff in the
 *  image.  Really only works for skew angles less than 10 degrees.
 *
 * PrivateDescription:::
 *  Two methods are used.  The first tries to trace almost-horizontal lines
 *  starting at various points on the image.  This gives a rough idea of
 *  the skew.  The second method tries to find angles which line up with
 *  long horizontal black/white edges.  The second method does a binary search
 *  through angles centered around the best angle found by the first
 *  method.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to skewed image.
 *  redFactor::      A factor used to indicate how hard to try.
 *		     Allowable "reduction factors" are 1, 2, 4, 8, 16, and 32.
 *		     These are not really reduction factors, but numbers that
 *		     generally indicate how hard the skew routine should try,
 *		     hence how long it will take.  The lower the numbers, the
 *		     harder it will try and the longer the routine will take.
 *		     Typical reduction factors for 300 dpi images range from
 *		     4 to 8.
 * Output:::
 *  pConfidence::    A way of measuring how confident this routine is that
 *		     it has really found a skew angle.  The value of the
 *		     confidence ranges between 0 and 100.  While the
 *		     interpretation of the confidence factor is still
 *		     evoloving, confidence factors above 5 or so usually
 *		     indicate a reasonable angle.
 *  pSkewAngle::     Skew angle of image in radians.  If no skew angle is
 *		     found, returns 0 with confidence of 0
 *
 * Returns:::
 *  In addition to checking the normal stuff, this routine returns
 *  cErrInvalidParm if the x and y resolutions of the image are not
 *  reasonable.  That is, (xRes != yRes) and either (xRes == 0) or (yRes == 0)
 *  
 * Notes:::
 *  Every now and then, less often recently, someone provides an image
 *  whose skew cannot be found.  Usually, the images need more black stuff
 *  on them, particularly more black horizontal lines or lines of text.
 *  Long lines are generally better than lots of short lines.
 *
 * FunctionCategory:::
 *  skew detection
 *
 * Keywords:::
 *  skew detection, small angle, orientation, rotation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_findSkewSample (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               redFactor,
    Int32              *pConfidence,
    Float32            *pSkewAngle)
{
    PROLOGUE;
    Float64		 skewAngle64;
    Int32                w, h, bpl;
    UInt8               *pS;
    ScaledInt1616	 xdpi, ydpi;

    ERR_GEN(prS == NULL || pConfidence == NULL || pSkewAngle == NULL,
	    cErrNullPtr);

    /* If we can't find the skew angle, return this */
    *pSkewAngle = 0.0F;
    
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* make sure redFactor is valid */
    ERR_GEN(redFactor != 1 && redFactor != 2 && redFactor != 4 && 
	    redFactor != 8 && redFactor != 16 && redFactor != 32,
	    cErrInvalidParm);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xdpi, &ydpi));

	/* quit if the resolutions are not resonable.  This is
	 * done to avoid an arithmetic exception later on.
	 */
    ERR_GEN((xdpi != ydpi) && ((xdpi == 0) || (ydpi == 0)), cErrInvalidParm);

    ERR_PASS( i_detectSkew (pIPWState, pS, w, h, bpl,
			    redFactor, &skewAngle64, (UInt32 *) pConfidence,
			    ScaledInt1616_TO_Int32(xdpi),
			    ScaledInt1616_TO_Int32(ydpi)));

    *pSkewAngle = (Float32) skewAngle64; /* force conversion from Float64
					  * to Float32 without any whining
					  * from Visual C++
					  */
    NORMAL_RETURN;
}

/*

 $Log: skew2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.26  1998/01/14 00:15:28  ddavies
 skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.25  1996/05/22 22:01:18  ddavies
 ddavies on Wed May 22 15:00:59 PDT 1996

 skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v

 Cast 0.0 to be a float to keep Visual C/C++ from whining.

 Revision 1.24  1996/05/22 01:24:42  ddavies
 ddavies on Tue May 21 18:23:49 PDT 1996

 skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v

 Remove extraneous procedure comment.  Return angle of 0 with confidence of
 0 when no skew angle is found.
 This is instead of returning BEAST_ANGLE.

 Revision 1.23  1996/01/04 20:26:51  ejaquith
 skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.22  1995/12/13  04:13:10  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/12/13  02:07:36  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Calculate the y and x resolutions of the input image and pass them
 * down to the ipshared code that does the real work.  This allows the
 * ipshared code to remove the angle's native dependence on the ratio of
 * the resolutions.
 *
 * Revision 1.20  1995/11/13  20:16:04  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/11/01  20:21:06  ejaquith
 * ejaquith on Wed Nov 1 12:20:18 PST 1995
 *
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Ripped out macro definitions BEAST_ANGLE and BEAST_OFFSET, because the
 * former should be accessible to clients (since it IS returned as possible
 * value), and the latter was used nowhere (ie useless).
 *
 * Revision 1.18  1995/10/13  22:05:16  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/12  04:28:26  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.16  1995/10/05  04:26:56  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.15  1995/09/28  22:23:33  ddavies
 * ddavies on Thu Sep 28 15:23:25 PDT 1995
 *
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Fix comments at top of file.
 *
 * Revision 1.14  1995/09/28  19:21:18  ddavies
 * ddavies on Thu Sep 28 12:21:11 PDT 1995
 *
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Filled in autodoc comments.
 *
 * Revision 1.13  1995/09/21  22:59:41  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/15  18:00:16  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.11  1995/09/11  22:10:57  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.10  1995/09/06  20:23:28  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  01:18:52  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.8  1995/08/31  21:39:14  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/17  22:04:43  ejaquith
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.6  1995/07/29  04:50:55  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.5  1994/10/01  00:08:15  lperry
 * lperry on Fri Sep 30 17:07:36 PDT 1994
 *
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Change to use skew.pub instead of skewpros.h.
 * Eliminate some unused include files.
 *
 * Revision 1.4  1994/04/19  22:22:47  ddavies
 * ddavies on Tue Apr 19 15:22:04 PDT 1994
 *
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Add to the comments at the top of findSkewSample, explaining what the
 * reduction factor is and how it affects performance and accuracy.
 *
 * Revision 1.3  1993/12/03  01:19:03  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:17:11  ddavies
 * skew2.c -> /product/ipcore/ipwrap/src/RCS/skew2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

