/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * spigot.c -- "Image Spigot"
 *	       
 *
 * Exports:
 *
 *	imageSpigot ()
 */



#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "thrshold.h"

/* prototypes */
#include "props.pub"
#include "thrshold.pub"
#include "shrscal.pub"
#include "spigot.pub"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: spigot.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  The so-called "image spigot" (pat. pend.)
 *
 * LongDescription:::
 *	The so-called "image spigot", which is used to take
 *	bands from a scanner and peel off both a reduced-resolution
 *	gray or color image and a full-resolution binary or gray
 *	image. This is to be used when one has a single-pass-only
 *	scanner, such as a hand scanner or ADF, and also in cases
 *	where a single high-resolution color pass is faster than
 *	a low-resolution color pass followed by a high-resolution
 *	binary pass, which depends on the scanner.
 *<BR>
 *	It is indended that this function will be called on bands.
 *	The bands *must* be a multiple of the reduction factor (2 or 3)
 *	in height. The bands should also be a multiple of the reduction
 *	factor in width, though if they aren't, the consequence will be
 *	a dark line along the right edge. If the image being scanned is
 *	not the right width, pad the extra pixels with copies of the last
 *	pixel, and the image shouldn't be too far off. [A more "correct"
 *	thing to do is to extrapolate the last pixels based on the slope
 *	coming into the edge, but it is probably not worth the effort.]
 *	The same thing applies for height -- if the source image does not
 *	fit evenly into bands, pad out the bands. Note that this routine
 *	does not care whether each band passed is the same height,
 *	only that each band is a multiple of the reduction factor.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  srcPixr::        The source band, multiple of reduction factor in both
 *                   width and height
 *  destReduction::  The destination reduction factor, either 2 or 3
 *  destThreshold::  If the grayPixr image is binary, use this threshold
 *  destInvert::     If the grayPixr image is binary, invert the dest.
 *
 * Output:::
 *  prReduced::      The resolution-downcoverted result. Should be allocated
 *                   by the caller.
 *  prGray::         The full resolution binary or gray image. Should be
 *                   allocated by the caller.
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  resolution conversion
 *
 * Keywords:::
 *  resolution conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_imageSpigot (
    IPWState           *pIPWState,
    PIXR               *srcPixr,
    Int32               destReduction,
    Int32               destThreshold,
    Int32               destInvert,
    PIXR               *prReduced,
    PIXR               *prGray)
{
    PROLOGUE;
    Int32    srcWidth, srcHeight, redWidth, redHeight, grayWidth, grayHeight,
	     i, channels, srcDepth, grayDepth, grayBpl, srcFramedWidth, 
	     srcFramedHeight, srcBpl, destFramedWidth, destFramedHeight, 
	     destBpl, destDepth, rChannels;
    PIXR    *subS, *subD;
    UInt8   *pSrcImgAll, *pDestImgAll, *pGrayImgAll;

    /* Check for errors */
    ERR_GEN_PIXR (srcPixr, cByteOrderBig, cDepth4|cDepth8, cChannelsAny,
		  cFrame32);
    ERR_GEN_PIXR (prReduced, cByteOrderBig, cDepth4|cDepth8, 
		  cChannelsAny, cFrame32);
    ERR_GEN_PIXR (prGray, cByteOrderBig, cDepth1|cDepth4, cChannelsAny,
		  cFrame32);

    ERR_GEN (destReduction != 2 && destReduction != 3, cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, srcPixr, &srcWidth));
    ERR_PASS (w_pixrGetHeight (pIPWState, srcPixr, &srcHeight));
    ERR_PASS (w_pixrGetWidth (pIPWState, prReduced, &redWidth));
    ERR_PASS (w_pixrGetHeight (pIPWState, prReduced, &redHeight));
    ERR_PASS (w_pixrGetWidth (pIPWState, prGray, &grayWidth));
    ERR_PASS (w_pixrGetHeight (pIPWState, prGray, &grayHeight));
    ERR_PASS (w_pixrGetDepth (pIPWState, srcPixr, &srcDepth));
    ERR_PASS (w_pixrGetDepth (pIPWState, prGray, &grayDepth));

    ERR_GEN (srcHeight % destReduction, cErrBadRatio);

    ERR_GEN ((redHeight * destReduction != srcHeight) ||
	(redWidth * destReduction != srcWidth), cErrBadRatio);

    ERR_GEN ((srcWidth != grayWidth) || (srcHeight != grayHeight),
	     cErrBadRatio);

    ERR_PASS( w_pixrGetChannelCount (pIPWState, srcPixr, &channels));
    ERR_PASS( w_pixrGetChannelCount (pIPWState, prReduced, &rChannels));
    ERR_GEN (channels != rChannels, cErrChannelsNotEqual);

    /* Do the resolution down-conversion */
    for (i = 0; i < channels; i++)
    {
	subS = NULL;
	ERR_PASS( w_pixrGetChannel (pIPWState, srcPixr, i, &subS));

	subD = NULL;
	ERR_PASS( w_pixrGetChannel (pIPWState, prReduced, i, &subD));
	ERR_PASS( w_pixrGetImageAll (pIPWState, subS, &pSrcImgAll));
        ERR_PASS( w_pixrFramedWidth (pIPWState, subS, &srcFramedWidth));
	ERR_PASS( w_pixrFramedHeight(pIPWState, subS, &srcFramedHeight));
	ERR_PASS( w_pixrGetBpl(pIPWState, subS, &srcBpl));
	ERR_PASS( w_pixrGetImageAll(pIPWState, subD, &pDestImgAll));
	ERR_PASS( w_pixrFramedWidth(pIPWState, subD, &destFramedWidth));
	ERR_PASS( w_pixrFramedHeight(pIPWState, subD, &destFramedHeight));
	ERR_PASS( w_pixrGetBpl (pIPWState, subD, &destBpl));
	ERR_PASS( w_pixrGetDepth (pIPWState, subD, &destDepth));

	if (destReduction == 2)
	{
	    ERR_PASS( i_rescon2To1 (pIPWState, pSrcImgAll,
				    srcFramedWidth,
				    srcFramedHeight,
				    srcBpl, 
				    srcDepth,
				    pDestImgAll,
				    destFramedWidth,
				    destFramedHeight,
				    destBpl, 
				    destDepth));
	}
	else
	{
	    ERR_PASS( i_rescon3To1 (pIPWState, pSrcImgAll,
				    srcFramedWidth,
				    srcFramedHeight,
				    srcBpl, 
				    srcDepth,
				    pDestImgAll,
				    destFramedWidth,
				    destFramedHeight,
				    destBpl, 
				    destDepth));
	}
    }
    
    /* Now, do the thresholding or 8->4 conversion */
    if (grayDepth == 1)
    {
	if (destInvert)
	    ERR_PASS( w_threshold (pIPWState, srcPixr, 0, 
				    destThreshold, THRESH_INVERSE, &prGray));
	else
	    ERR_PASS( w_threshold (pIPWState, srcPixr, 
				    destThreshold, (1 << srcDepth) - 1, 
				    THRESH_NORM, &prGray));
    }
    else
    {
	
	ERR_PASS( w_pixrGetImageAll (pIPWState, srcPixr, &pSrcImgAll));
        ERR_PASS( w_pixrFramedWidth (pIPWState, srcPixr, &srcFramedWidth));
	ERR_PASS( w_pixrFramedHeight(pIPWState, srcPixr, &srcFramedHeight));
	ERR_PASS( w_pixrGetBpl (pIPWState, srcPixr, &srcBpl));
	ERR_PASS( w_pixrGetImageAll (pIPWState, prGray, &pGrayImgAll));
	ERR_PASS( w_pixrGetBpl (pIPWState, prGray, &grayBpl));
    
	ERR_PASS( i_threshold8To4bpp (pIPWState, pSrcImgAll,
				      srcFramedWidth,
				      srcFramedHeight,
				      srcBpl,
				      pGrayImgAll,
				      grayBpl));
    }

    NORMAL_RETURN;
}
			  
/*
 * $Log: spigot.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.20  1998/01/14 00:15:29  ddavies
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.19  1996/01/04 20:26:53  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.18  1995/12/13  04:13:14  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:16:09  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  22:05:19  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/10/03  01:59:16  macleod
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.14  1995/09/21  22:59:44  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/11  21:54:11  macleod
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.12  1995/09/07  01:35:23  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Ran script to add tag "$" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:23:31  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  21:02:32  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.9  1995/08/31  21:39:16  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/29  21:56:41  ddavies
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.7  1995/08/25  01:38:28  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Fixed warnings
 *
 * Revision 1.6  1995/08/17  21:55:44  macleod
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.5  1994/10/07  19:06:28  lperry
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.4  1994/09/16  17:40:00  jcraig
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.3  1994/06/07  02:41:00  macleod
 * macleod on Mon Jun 6 19:40:58 PDT 1994
 *
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.2  1994/06/01  01:15:45  macleod
 * macleod on Tue May 31 18:15:44 PDT 1994
 *
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Forgot the
 * Revision 1.12  1995/09/07  01:35:23  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Ran script to add tag "" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:23:31  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  21:02:32  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.9  1995/08/31  21:39:16  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/29  21:56:41  ddavies
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.7  1995/08/25  01:38:28  ejaquith
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Fixed warnings
 *
 * Revision 1.6  1995/08/17  21:55:44  macleod
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.5  1994/10/07  19:06:28  lperry
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.4  1994/09/16  17:40:00  jcraig
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * remove include of bitacces.h which was not needed
 *
 * Revision 1.3  1994/06/07  02:41:00  macleod
 * macleod on Mon Jun 6 19:40:58 PDT 1994
 *
 * spigot.c -> /product/ipcore/ipwrap/src/RCS/spigot.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 * at the bottom
 * 
 */
