/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Spline interpolation */


#include "ipwcore.prv"

/* Prototypes */
#include "nrutil.pub"
#include "spline.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: spline.c,v $; $Revision: 1.5 $; Date$");
/*
 * This code was translated from the ALGOL in ACM TOMS 514.
 * It replaces some non-public-domain code.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Compute the gradients at knots for cubic spline interpolation
 *
 * LongDescription:::
 *   
 *   grad is an array having the same subscripts as x
 *   and y into which will be placed the calculated
 *   gradients of the required cubics at each of the
 *   data points.
 *   lbx and ubx are integers giving the minimum and
 *   maximum values respectively of the subscripts of
 *   the arrays x, y, and grad;
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState:: IPWrap state structure
 *  x::
 *  y::         x and y are arrays with subscripts from lbx
 *              to ubx giving the abscissae and ordinates respectively 
 *              of the data points.  These should be in ascending order 
 *              of abscissae, with no two abscissae equal.  
 *              The arrays must contain at least four points.
 *  lbx::       Lower bound of index into arrays x[] and y[]
 *  ubx::       Upper bound of index into arrays x[] and y[]
 *
 * Output:::
 *  ppGrad::    The gradients array, allocated here if *ppGrad==NULL
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod:: Translated from Algol
 *
 * Citations::: (optional tag)
 *  From ACM TOMS #514
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
w_cubicSplineGradients (
    IPWState *pIPWState,
    Float64  *x,
    Float64  *y,
    Int32     lbx,
    Int32     ubx,
    Float64 **ppGrad)
{
    PROLOGUE;
    Int32   i, iless2, iless1, iplus1, iplus2;
    Float64 x0, x1, x2, x3, x4, y2, prod1, prod2, num, denom, g,
            coeff2, xdiff, xprod, weight, *grad;
    
    ERR_GEN (ppGrad == NULL, cErrNullPtr);
    if (*ppGrad == NULL)
    {
	grad = IPW_CALLOC (ubx + 1, sizeof (Float64), cAllocatorData);
	ERR_GEN (!grad, cErrNomem);
    }
    else
    {
	grad = *ppGrad;
    }

    for (i = lbx; i <= ubx; i++)
    {
	/* Special treatment is needed at end point */
	iless1 = (i > lbx) ?  i - 1 : i + 3;
	iplus1 = (i < ubx) ?  i + 1 : i - 3;
	x2 = x[i];
	y2 = y[i];
	x1 = x[iless1] - x2;
	x3 = x[iplus1] - x2;

	/* First fit a quadratic through x1, x2, x3 */
	prod1 = (y[iless1] - y2) * x3;
	prod2 = (y[iplus1] - y2) * x1;
	denom = x1 * x3 * (x[iless1] - x[iplus1]);
	g = (x1 * prod2 - x3 * prod1) / denom;
	coeff2 = (prod1 - prod2) / denom;
	
	/* If x0 exists, find its contribution to the cubic adjustment */
	if (i <= lbx + 1) 
	    num = denom = 0.0;
	else
	{
	    iless2 = i - 2;
	    x0 = x[iless2] - x2;
	    xdiff = x[iless2] - x[iless1];
	    xprod = x0 * xdiff * (x[iless2] - x[iplus1]);
	    weight = xprod / (xdiff * xdiff);
	    num = weight * (y[iless2] - y2 - x0 * (g + x0 * coeff2));
	    denom = weight * xprod;
	}
	
	/*  If x4 exists, find its contribution to the cubic adjustment */
	if (i < ubx - 1)
	{
	    iplus2 = i + 2;
	    x4 = x[iplus2] - x2;
	    xdiff = x[iplus2] - x[iplus1];
	    xprod = x4 * xdiff * (x[iplus2] - x[iless1]);
	    weight = xprod / (xdiff * xdiff);
	    num = num + weight * (y[iplus2] - y2 - x4 * (g + x4 * coeff2));
	    denom = denom + weight * xprod;
	}
	grad[i] = g + num * x1 * x3 / denom;
    }

    *ppGrad = grad;
    NORMAL_RETURN;

} 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Compute coefficients of the cubic to interpolate between x[i] and x[i + 1]
 *
 * LongDescription:::
 *  this procedure calculates the coefficients
 *  of the cubic (in the range x[i] to x[i + 1]) which
 *  has values y[i], y[i + 1] and gradients grad[i],
 *  grad[i + 1] at x[i], x[i + 1] respectively.
 *  the cubic takes the form <BR><BR>
 *      c0 + c1*(x - x[i]) + c2*(x - x[i])**2 + c3*(x - x[i])**3
 *  <BR>
 *
 * PrivateDescription::: 
 *
 * Input:::
 *  pIPWState:: IPWrap state structure
 *  x::
 *  y::         x and y are arrays with subscripts from lbx
 *              to ubx giving the abscissae and ordinates respectively 
 *              of the data points.  These should be in ascending order 
 *              of abscissae, with no two abscissae equal.  
 *              The arrays must contain at least four points.
 *  grad::      The gradient array.
 *  i::         Index--we're interpolating between x[i] and x[i + 1]
 *
 * Output:::
 *  pC0::       Pointer to c0, the 0th coefficient of the cubic
 *  pC1::       Pointer to c1, the 1st coefficient of the cubic
 *  pC2::       Pointer to c2, the 2nd coefficient of the cubic
 *  pC3::       Pointer to c3, the 3rd coefficient of the cubic
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work - in-progress
 *
 * Authors/Modifications:::
 *  macleod:: translated from Algol
 *
 * Citations::: (optional tag)
 *  From ACM TOMS #514
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cubicSplineCoeffs (
    IPWState *pIPWState,
    Float64 *x,
    Float64 *y,
    Float64 *grad,
    Int32 i,
    Float64 *pC0,
    Float64 *pC1,
    Float64 *pC2,
    Float64 *pC3)
{
    PROLOGUE;
    Float64 h, dy, c0, c1, c2, c3;
    pIPWState = pIPWState;	/* keep compiler happy */
    c0 = y[i];
    h = x[i + 1] - x[i];
    dy = y[i + 1] - c0;
    c1 = grad[i];
    c2 = (3.0 * dy - h * (2.0 * c1 + grad[i + 1])) / (h * h);
    c3 = (h * (c1 + grad[i + 1]) - 2.0 * dy) / (h * h * h);

    *pC0 = c0;
    *pC1 = c1;
    *pC2 = c2;
    *pC3 = c3;
    NORMAL_RETURN;

} 
 
/****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Piecewise cubic spline interpolation, given gradients and nodes
 *
 * LongDescription:::
 *  Given abscissae x[] and ordinates y[], in ascending order of abscissae,
 *  and given gradients[] computed by w_cubicSplineGradients, interpolate
 *  an ordinate y given an abscissa x. 
 *<BR><BR>
 *  The reason for having a separate fuction to compute gradients is
 *  that one can compute them once for a given data set, and then
 *  use those pre-computed gradients many times to perform interpolations
 *  within that data set.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState:: IPWrap state structure
 *  x::
 *  y::         x and y are arrays with subscripts from lbx
 *              to ubx giving the abscissae and ordinates respectively 
 *              of the data points.  These should be in ascending order 
 *              of abscissae, with no two abscissae equal.  
 *              The arrays must contain at least four points.
 *  grad::      The gradient at each point in x[] and y[], as computed
 *              by w_cubicSplineGradients. 
 *  lbx::       lower bound of index into x[], y[], and grad[]
 *  ubx::       upper bound of index into x[], y[], and grad[]
 *  xValue::    value of X for which we want to interpolate a Y
 *
 * Output:::
 *  pYValue::   Interpolated y-value
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  math
 *
 * Keywords:::
 *  math
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_cubicSplineInterpolate (
    IPWState *pIPWState,
    Float64  *x,
    Float64  *y,
    Float64  *grad,
    Int32     lbx,
    Int32     ubx,
    Float64   xValue,
    Float64  *pYValue)
{
    PROLOGUE;
    Int32 i;
    Float64 c0, c1, c2, c3, yValue, xd;

    /* Find the knot */
    for (i = lbx; i <= ubx; i++)
    {
	if (xValue <= x[i])
	{
	    i--;
	    break;
	}
    }

    ERR_GEN_CLEANUP (i < 0, cErrOutOfBounds);

    if (i < lbx)
	i = lbx;
    /* Clamp it so that we don't interpolate off the end */
    if (i > ubx)
	i = ubx - 1;

    /* Compute the cubic coefficients */
    ERR_PASS (w_cubicSplineCoeffs (pIPWState, x, y, grad, i, 
				   &c0, &c1, &c2, &c3));

    /* Expand the cubic */
    xd = xValue - x[i];
    yValue = c0 + xd * (c1 + xd * (c2 + xd * c3));
    *pYValue = yValue;

    NORMAL_RETURN;

ERR_CLEANUP:
    *pYValue = 0.0;
    ERR_RETURN;
}

/*

 $Log: spline.c,v $
 Revision 1.5  2000/10/19 23:26:09  jlin
 minor fix to avoid build warning

 Revision 1.4  2000/07/08  00:38:13  dtracy
 x86_linux added to supported os's.

 Revision 1.3  2000/06/27 23:11:54  jlin
 minor change to avoid build warning on solaris



 Revision 1.2  2000/06/07 00:28:27  jlin
 change wp_jcWriteCallback

 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.21  1996/01/04 20:26:56  ejaquith
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 Revision 1.20  1995/12/13  04:13:17  ejaquith
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 CommonCheck: Added ";" after macro IP_RCSINFO()

 Revision 1.19  1995/11/16  20:29:24  ejaquith
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 Fixed parameter typo in a function's comment block.

 Revision 1.18  1995/11/13  20:16:13  ejaquith
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 (Previously it took the string of the name of the routine).

 Revision 1.17  1995/11/04  01:36:07  ejaquith
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 style wrappers, etc..

 Revision 1.16  1995/11/03  22:49:04  ddavies
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 Eric's scripts.

 Revision 1.15  1995/10/13  22:05:22  ejaquith
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 CommonCheck: Some minor changes, like adding routine name after
 BeginDocumentation tag, some keyword condensing.

 Revision 1.14  1995/10/10  00:16:15  ddavies
 ddavies on Mon Oct 9 17:15:47 PDT 1995

 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 Use pIPWState in w_cubicSplineCoeffs to keep compilers from complaining
 that it's unused.

 Revision 1.13  1995/10/05  21:51:27  macleod
 macleod on Thu Oct 5 14:50:13 PDT 1995

 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 We weren't handling cases on the first and last point
 (sample) correctly. Also added "pIPWState" parameter to
 w_cubicSplineCoeffs.

 Revision 1.12  1995/10/04  18:31:00  macleod
 spline.c -> /product/ipcore/ipwrap/src/RCS/spline.c,v

 Changed w_cubicSplineGradients so that it takes a pointer to a pointer
 and allocates the gradient table if the pointer is NULL, like it's
 supposed to.

 Revision 1.11  1995/10/03  21:10:02  macleod
 spline.c -<> /product/ipcore/ipwrap/src/RCS/spline.c,v

 Threw out cubic spline interpolator (it was copyrighted,
 so we shouldn't use it) and replaced it with a public-domain
 one.
 (Actually, translated it from Algol 60!)

 */

