/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  strel12.c-- initializes strels used by the morphology testing routines
 *                   w_strel12Inits ()
 *                   w_strel12Term ()
 *
 *              strelF40h, strelF40v, s_4_2_t, strelItal1,
 *		strelHTR, strelVLT, s_v21_right, emptyStrel;
 */



#include "ipwcore.prv"

#include "defines.pub"
#include "pixr.h"
#include "morphg.h"

#include "strel.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: strel12.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

#undef CHECKERR
#define CHECKERR(statement) \
if ((statement)) return abortI("No memory for strel", procName, cErrNomem);


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initializes an set of specialized structuring elements.
 *
 * LongDescription:::
 *  These STRELs are created and stored in the strel manager instance
 *  associated with the client state structure passed in and may
 *  be destroyed by calling w_strel12Term ():
 *
 *    strelF40h, strelF40v, s_4_2_t, strelItal1, strelHTR, strelVLT, 
 *    s_v21_right, emptyStrel
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel management, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strel12Inits (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    if (pIPWState->strelInfo->strel12InitsDone)
	NORMAL_RETURN;


    /* FAST horizontal strel of length 40: [6:(100) 0 1c 0 6:(100) 1] */
    ERR_PASS( w_initStrelByName (pIPWState, strelF40h));

    /* FAST vertical strel of length 40: v:[6:(100) 0 1c 0 6:(100) 1] */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strelF40v)); 

    /* pattern match for s_4_2_t operator:
	0   2   0
	2   0c  0
	0   1   1      */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_4_2_t)); 


    /* pattern:  0  0  0  2  0  1
	     (2) 0  0  0  0  0  0
		 0  0  2  0  1  0
		 0  0  0  0  0  0
		 0  0  0  0c 0  0
		 0  2  0  1  0  0
	     (2) 0  0  0  0  0  0
		 2  0  1  0  0  0  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strelItal1)); 

    /* anti-alias pattern match for horizontal, top, right:
	    2   2   2   2   2
	    1   1   1c  1   2    */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strelHTR)); 

    /* anti-alias pattern match for vertical, left, top:
	    2   2
	    2   1c
	    2   1
	    2   1
	    2   1  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strelVLT)); 

    /* registration for left edge of line or bar */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_v21_right));

    /* strel containing only don't cares */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, emptyStrel));

    pIPWState->strelInfo->strel12InitsDone = cTrue;
    NORMAL_RETURN;


ERR_CLEANUP:
    /* we purposefully ignore the return value here since
     * 1) it will mask the original error condition which brought us here, and
     * 2) strel1Term () will return an error since it can't de-init all the
     * strels (at least one of them has not been initted if we got to the
     * error condition).
     */
    w_strel12Term (pIPWState);
    pIPWState->strelInfo->strel12InitsDone = cFalse;
    ERR_RETURN;

}  /* strel12Inits () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroys the set structuring elements created by w_strel12Inits ()
 *
 * LongDescription:::
 *  If the strels are no longer used by any other instances of the strel
 *  manager, they are freed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel management, binary, morphology
 * 
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strel12Term (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    /* Note that we don't check to see if the strel12InitsDone
     * flag is set as a precondition as this function may be
     * called as a cleanup operation if an error occurred in
     * strel12Init () in which case some strels may have been
     * created, but the flag has not been set.
     */

    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, strelF40h));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, strelF40v));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_4_2_t));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, strelItal1));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, strelHTR));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, strelVLT));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_v21_right));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, emptyStrel));

    /* Mark this group of strels as uninitialized */
    pIPWState->strelInfo->strel12InitsDone = cFalse;
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;

}  /* strel12Term () */


/*
 *
 * $Log: strel12.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.24  1996/01/04 20:27:05  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.23  1995/12/13  04:13:29  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:16:25  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:05:31  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/09/29  21:11:03  dferrell
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.19  1995/09/21  22:59:56  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/07  22:09:13  jcraig
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * strel12Term => w_strel12Term
 *
 * Revision 1.17  1995/09/06  20:23:42  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.16  1995/09/01  18:49:58  dferrell
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.15  1995/08/31  21:39:26  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.14  1995/08/30  17:27:40  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Made routines accessing pIPWState first check that it is valid (not NULL).
 * If it's null, return error: cErrNullIPWState.
 *
 * Revision 1.13  1995/08/25  01:38:44  ejaquith
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Fixed warnings
 *
 * Revision 1.12  1995/08/23  18:26:44  macleod
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Checked the flag for "strel1InitsDone" instead of "strel12InitsDone",
 * causing strel12 inits to never happen if strel1 inits were done
 *
 * Revision 1.11  1995/08/22  23:46:33  dferrell
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Added new parameter to all calls to releaseStrelByName ().
 *
 * Revision 1.10  1995/08/17  21:56:31  dferrell
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.9  1995/01/24  22:08:15  lperry
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Incorporate changes to use new strel manager interface.
 *
 * Revision 1.8  1994/12/08  04:42:37  ddavies
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Add initialization for the emptyStrel.  This is used to test our ability
 * to handle a strel containing only don't-care terms.
 *
 * Revision 1.7  1994/09/30  17:08:39  lperry
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Add include of utils.pub to get _abortX defs.
 *
 * Revision 1.6  1994/09/29  21:18:47  lperry
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Change to use strel.pub instead of strlpros.h
 *
 * Revision 1.5  1994/08/10  22:05:00  lperry
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Change to use the new strel manager functions to
 * allocate and track strels. There will only be a
 * single copy of each unique strel maintained, even
 * if multiple init calls are made.
 *
 * Revision 1.4  1993/12/09  19:59:46  macleod
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Changed INT32 to Int32
 *
 * Revision 1.3  1993/12/03  01:19:18  ddavies
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:17:35  ddavies
 * strel12.c -> /product/ipcore/ipwrap/src/RCS/strel12.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

