/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  streldyn.c
 *
 *       Dynamic creation of Strels:
 *            w_makeBrickStrel ()
 *            w_makeParityStrel ()
 *
 *       Finding maximum extension of Strel from center
 *            w_maxExtentStrel ()
 *
 *       Interconversion between Strel and Pixr
 *            w_strelToPixr ()
 *            w_pixrToStrel ()
 *
 *       Disk and Ellipse Strels
 *            w_makeDiskStrel ()
 *            w_makeEllipseStrel ()
 *
 *       Other Strel operations:
 *            w_strelType ()
 *            w_strelSeparation ()
 *
 */


#include "defines.pub"  /* for cTrue and cFalse */
#include "ipwcore.prv"

/*
#include "alplib.h"
*/
#include "morphg.h"
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"

#include "strel.pub"



#define   MAX_WIDTH          100
#define   MAX_HEIGHT         100


IP_RCSINFO(RCSInfo, "$RCSfile: streldyn.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Makes a rectangular STREL which is filled with hits
 *
 * LongDescription:::
 *  The caller is responsible for keeping track of the resulting strel--it
 *  is not stored in the strel manager.  Use w_destroyStrel () to free it.
 *  The result pointer should be set to NULL before its address is passed 
 *  in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  w::              Width of STREL
 *  h::              Height of STREL
 *  centerX::        Center x location of STREL relative to (0,0) in UL
 *  centerY::        Center y location of STREL relative to (0,0) in UL
 *
 * Output:::
 *  strel::          Resulting STREL (created)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeBrickStrel (
    IPWState           *pIPWState,
    Int32               w,
    Int32               h,
    Int32               centerX,
    Int32               centerY,
    STREL             **strel)
{
    PROLOGUE;
    Int32        i, j;

    ERR_GEN (*strel != NULL, cErrInvalidParm);

    ERR_GEN ((w <= 0) || (h <= 0), cErrInvalidParm);

/* I'm not quite sure why we're limiting the size, and it screws up stuff,
   so I've disabled it -- psm 11/6/95 */
#if 0
    ERR_GEN (ABS (centerX) > MAX_WIDTH, cErrInvalidParm);
    ERR_GEN (ABS (centerY) > MAX_HEIGHT, cErrInvalidParm);
#endif
    ERR_PASS( w_createStrel (pIPWState, w, h, strel));

    (*strel)->cx = centerX;
    (*strel)->cy = centerY;
    for (i = 0; i < h; i++)
	for (j = 0; j < w; j++)
	    (*strel)->data[i][j] = 1;
    (*strel)->accelFlag = 1;

    NORMAL_RETURN;

}  /* makeBrickStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given a reference STREL, makes a 2D mirrored STREL.
 *
 * LongDescription:::
 *  The new STREL has the contents of the reference STREL as if it were
 *  flipped upside down (NOT rotated) and flipped side to side (NOT
 *  rotated).  The caller is responsible for keeping track of the resulting 
 *  strel--it is not stored in the strel manager.  Use w_destroyStrel () to 
 *  free it.  The result pointer should be set to NULL before its address
 *  is passed in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  sStrel::         Reference strel (not altered)
 *
 * Output:::
 *  dStrel::         Resulting parity strel (created)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeParityStrel (
    IPWState           *pIPWState,
    STREL              *sStrel,
    STREL             **dStrel)
{
    PROLOGUE;
    Int32        w, h, i, j;

    ERR_GEN (sStrel == NULL, cErrInvalidParm);
    ERR_GEN (*dStrel != NULL, cErrInvalidParm);


    w = sStrel->nx;
    h = sStrel->ny;

    ERR_PASS( w_createStrel (pIPWState, w, h, dStrel));

    /* assign the new center */
    (*dStrel)->cx = w - sStrel->cx;
    (*dStrel)->cy = h - sStrel->cy;

    /* assign the new data */
    for (i = 0; i < h; i++)
	for (j = 0; j < w; j++)
	    (*dStrel)->data[i][j] = sStrel->data[h - i - 1][w - j - 1];

    (*dStrel)->accelFlag = 1;

    NORMAL_RETURN;

}  /* makeParityStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculates the maximum center to edge distances (in x or y) of a strel.
 *
 * LongDescription:::
 *  This is useful in calculating how large a boundary is needed in an image
 *  to be processed by this strel via morphological operation(s).
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          Strel to be analyzed
 *
 * Output:::
 *  extent::         Largest distance from center to edge
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_maxExtentStrel (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *extent)
{
    PROLOGUE;
    Int32          dx, dy;

    ERR_GEN (strel == NULL, cErrInvalidParm);

    dx = MAX (ABS (strel->nx - strel->cx - 1), ABS (strel->cx));
    dy = MAX (ABS (strel->ny - strel->cy - 1), ABS (strel->cy));
    *extent = MAX (dx, dy);

    NORMAL_RETURN;

}  /* maxExtentStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates an 8-bit PIXR with pixel values equal to the input strel values
 *
 * LongDescription:::
 *  The PIXR pointer should be set to NULL before its address is passed in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          Input strel
 *
 * Output:::
 *  prD::            Resulting 8-bit deep PIXR (created)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, depth conversion, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelToPixr (
    IPWState           *pIPWState,
    STREL              *strel,
    PIXR              **prD)
{
    PROLOGUE;
    Int32        i, j;

    ERR_GEN (strel == NULL, cErrInvalidParm);
    ERR_GEN ((*prD) != NULL, cErrInvalidParm);

    ERR_PASS( w_createPixr (pIPWState, strel->nx, strel->ny, 
                          8, cFrameBits, 0, 0, prD));

    for (i = 0; i < strel->ny; i++)
	for (j = 0; j < strel->nx; j++)
	    ERR_PASS_CLEANUP( w_pixrPutPixel (pIPWState, *prD, j, i, 
	                                   strel->data[i][j]));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, prD);
    ERR_RETURN;

}  /* strelToPixr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts a PIXR to a strel of the same dimension.
 *
 * LongDescription:::
 *  Sets the strel data to the corresponding pixel values in the PIXR.
 *  The center is set using cx and cy.  The caller is responsible for 
 *  keeping track of the resulting strel--it is not stored in the strel
 *  manager.  Use w_destroyStrel () to free it.  The result pointer should 
 *  be set to NULL before its address is passed in.

 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pixr::           PIXR to be converted
 *  cx::             The strel center x ordinate
 *  cy::             The strel center y ordinate
 *
 * Output:::
 *  strelD::         The resulting strel (created)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrToStrel (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               cx,
    Int32               cy,
    STREL             **strelD)
{
    PROLOGUE;
    Int32        w, h, i, j, d;

    ERR_GEN (pixr == NULL, cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth (pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight (pIPWState, pixr, &h));

    ERR_PASS( w_createStrel (pIPWState, w, h, strelD));

    (*strelD)->cx = cx;
    (*strelD)->cy = cy;
    for (i = 0; i < h; i++)
	for (j = 0; j < w; j++)
	{
            ERR_PASS_CLEANUP( w_pixrGetPixel (pIPWState, pixr, j, i, &d));
	    (*strelD)->data[i][j] = (Int16)d;
        }

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyStrel (pIPWState, strelD);
    ERR_RETURN;

}  /* pixrToStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Analyzes the input strel and returns a code for the type.
 *
 * LongDescription:::
 *  Types are:
 *                  H_1D_STREL     horizontal, 1D, hits only   
 *                  V_1D_STREL     vertical, 1D, hits only   
 *                  HMD_1D_STREL   1D strel with misses and/or don't cares
 *                  S_2D_STREL     separable 2D strel, hits only
 *                  HMD_S_2D_STREL separable 2D strel, hits with misses
 *				      and/or don't cares
 *                  NS_2D_STREL    non-separable 2D strel
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          Strel to be analyzed
 *
 * Output:::
 *  type::           Resulting type
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelType (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *type)
{
    PROLOGUE;
    Int32        i, j;
    Int32        separable_hitsonly;

    ERR_GEN (strel == NULL, cErrInvalidParm);

    if (strel->ny == 1)  /* 1D horizontal */
    {
	for (j = 0; j < strel->nx; j++) 
	{
	    if (strel->data[0][j] != 1)
	    {
		*type = HMD_1D_STREL;
		NORMAL_RETURN;
	    }
	}
	*type = H_1D_STREL;
	NORMAL_RETURN;
    }

    if (strel->nx == 1)  /* 1D vertical */
    {
	for (i = 0; i < strel->ny; i++) 
	{
	    if (strel->data[i][0] != 1)
	    {
		*type = HMD_1D_STREL;
		NORMAL_RETURN;
	    }
	}
	*type = V_1D_STREL;
	NORMAL_RETURN;
    }

    /* it's 2D */
    separable_hitsonly = cTrue;
    for (i = 0; i < strel->ny; i++) 
    {
	for (j = 0; j < strel->nx; j++) 
	{
	    if (strel->data[i][j] != 1)
		separable_hitsonly = cFalse;
	}
    }

    if (separable_hitsonly == cTrue)
    {
	*type = S_2D_STREL;
	NORMAL_RETURN;
    }

    /* it still may be separable with hits and/or misses, but at
     * this point we don't care */

    *type = NS_2D_STREL;
    NORMAL_RETURN;

}  /* strelType () */
    

/*
 *  strelSeparation()
 *
 *        Input:  strel and addresses into which two constructed strels
 *                are placed.
 *        Action:  Checks to be sure the input SE is a separable 2D
 *                 strel composed entirely of hits, and then generates
 *                 two linear SEs to which it is equivalent.
 *        Return:  0 if OK; 1 on error.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Seperates a strel into two (smaller) component strels
 *
 * LongDescription:::
 *  If the strel is 2D seperable, two smaller linear strels are created 
 *  such that morphological operations performed using the combined results 
 *  of the component strels is the same as if the larger strel had been used.
 *  This can be used to increase computation speed when large strels are
 *  needed.  The caller is responsible for keeping track of the resulting 
 *  strels--they are not stored in the strel manager.  Use w_destroyStrel () 
 *  to free them.  The result pointers should be set to NULL before their
 *  addresses are passed in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          Source strel to be seperated
 *
 * Output:::
 *  strelh::         The horizontal linear component strel (created)
 *  strelv::         The vertical linear component strel (created)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelSeparation (
    IPWState           *pIPWState,
    STREL              *strel,
    STREL             **strelh,
    STREL             **strelv)
{
    PROLOGUE;
    Int32        streltype;

    ERR_GEN (strel == NULL, cErrInvalidParm);

    ERR_GEN (*strelh != NULL, cErrInvalidParm);
    ERR_GEN (*strelv != NULL, cErrInvalidParm);

    ERR_PASS (w_strelType (pIPWState, strel, &streltype));
    ERR_GEN (streltype != S_2D_STREL, cErrStrel_Not2DSeparable);

    ERR_PASS (w_makeBrickStrel (pIPWState, strel->nx, 1, strel->cx, 1, strelh));

    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1,
					strel->ny, 1, strel->cy, strelv));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyStrel (pIPWState, strelh);
    ERR_RETURN;
}  /* strelSeparation () */


/*
 * $Log: streldyn.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.20  1997/09/25 01:29:32  ddavies
 * ddavies on Wed Sep 24 18:29:22 PDT 1997
 *
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * Spelling correction: change cErrStrel_Not2DSeperable to
 * cErrStrel_Not2DSeparable.
 *
 * Revision 1.19  1997/06/26 20:04:50  ddavies
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Add inclusion of new file rast.pub.
 *
 * Revision 1.18  1996/01/04 20:27:07  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.17  1995/12/13  04:13:33  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:16:29  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/11/06  23:08:19  macleod
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * w_makeBrickStrel had the annoying feature of limiting the size of
 * STREL created to 200 x 200 pixels.
 * I disabled this test.
 *
 * Revision 1.14  1995/11/04  01:36:09  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.13  1995/11/03  22:49:08  ddavies
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.12  1995/10/16  22:59:17  dferrell
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.11  1995/10/13  22:05:33  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.10  1995/09/29  21:12:02  dferrell
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * Updated function comment blocks.  Replaced TRUE and FALSE with
 * cTrue and cFalse, respectively.
 *
 * Revision 1.9  1995/09/21  22:59:59  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.8  1995/09/07  22:10:41  jcraig
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * fill in missing w_'s
 *
 * Revision 1.7  1995/09/06  20:23:45  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.6  1995/09/01  18:50:09  dferrell
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.5  1995/08/31  21:39:29  ejaquith
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.4  1995/08/17  21:56:37  dferrell
 * streldyn.c -> /product/ipcore/ipwrap/src/RCS/streldyn.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 *
 */

