/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
** Author: Luc Vincent
*/

/*
 *  strelg.c:  Utility routines for Strels used for grayscale morphology
 *
 *
 *         SE Operations:
 *
 *              Basic SE Operations
 *                  w_createSE()
 *                  w_destroySE()
 *                  w_printSE()
 *                  w_copySE()
 *                  w_transposeSE()
 *
 *              Conversions from Strel and Pixr to SE
 *                  w_strelToSE()
 *                  w_pixrToSE()
 *
 */


#include "ipwcore.prv"

#include "morphg.h"
#include "pixr.h"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"

#include "strel.pub"
#include "shrutils.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: strelg.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*----------------------------------------------------------------------*
 *                             Globals                                  *
 *----------------------------------------------------------------------*/
    /* shift vectors for directions 0 through 7 */
Int32       deltax[] = {1, 1, 0, -1, -1, -1, 0, 1};
Int32       deltay[] = {0, -1, -1, -1, 0, 1, 1, 1};

Int32       delta6x[] = {1, 0, -1, -1, -1, 0};
Int32       delta6y[] = {0, -1, -1, 0, 1, 1};




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates a grayscale morphological structuring element
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  nb_cc::          Number of connected components in the strel
 *  size::           Total number of pixels in SE
 *  side_sizes::     Array of sizes of side of element
 *
 * Output:::
 *  se::             The newly created SE
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray, strel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createSE (
    IPWState           *pIPWState,
    Int32               nb_cc,
    Int32               size,
    Int32              *side_sizes,
    SE                **se)
{
    PROLOGUE;
    Int32        total_sides_size;
    Int32        i;

    ERR_GEN (*se != NULL, cErrInvalidParm);

    *se = (SE *)IPW_MALLOC (sizeof (SE), cAllocatorData);
    ERR_GEN (*se == NULL, cErrNomem);

    (*se)->nb_cc = nb_cc;
    (*se)->origins = (struct point *)IPW_MALLOC (sizeof (struct point) * nb_cc,
                                                 cAllocatorData);
    if ((*se)->origins == NULL)
    {
	IPW_FREE (*se, cAllocatorData);
	ERR_GEN (1, cErrNomem);
    }

    (*se)->size = size;
    (*se)->pts = (struct point *)IPW_MALLOC (sizeof (struct point) * size,
                                             cAllocatorData);
    if ((*se)->pts == NULL)
    {
	IPW_FREE ((*se)->origins, cAllocatorData);
	IPW_FREE (*se, cAllocatorData);
	ERR_GEN (1, cErrNomem);
    }

    (*se)->sides = (struct side_se *)IPW_MALLOC (sizeof (struct side_se) * 8,
                                                 cAllocatorData);
    if ((*se)->sides == NULL)
    {
	IPW_FREE ((*se)->pts, cAllocatorData);
	IPW_FREE ((*se)->origins, cAllocatorData);
	IPW_FREE (*se, cAllocatorData);
	ERR_GEN (1, cErrNomem);
    }

    total_sides_size = 0;
    for (i = 0; i < 8; i++)
	total_sides_size += *(side_sizes + i);

    (*se)->sides->pts = (struct point *)IPW_MALLOC (sizeof (struct point) * 
                                                   total_sides_size,
                                                   cAllocatorData);
    if ((*se)->sides->pts == NULL) 
    {
	IPW_FREE ((*se)->sides, cAllocatorData);
	IPW_FREE ((*se)->pts, cAllocatorData);
	IPW_FREE ((*se)->origins, cAllocatorData);
	IPW_FREE (*se, cAllocatorData);
	ERR_GEN (1, cErrNomem);
    }

    (*se)->sides->size = *side_sizes;
    for (i = 1; i < 8; i++) 
    {
	((*se)->sides + i)->size = *(side_sizes + i);
	((*se)->sides + i)->pts = ((*se)->sides + i - 1)->pts +
	                          ((*se)->sides + i - 1)->size;
    }

    NORMAL_RETURN;

}  /* w_createSE () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroys a given SE
 *
 * LongDescription:::
 *  Frees all memory associated with the SE and sets the pointer to NULL.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *  se::             The SE to be destroyed
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray, strel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroySE (
    IPWState           *pIPWState,
    SE                **se)
{
    PROLOGUE;

    ERR_GEN (*se == NULL, cErrInvalidParm);

    IPW_FREE ((*se)->sides->pts, cAllocatorData);
    IPW_FREE ((*se)->sides, cAllocatorData);
    IPW_FREE ((*se)->pts, cAllocatorData);
    IPW_FREE ((*se)->origins, cAllocatorData);
    IPW_FREE (*se, cAllocatorData);

    NORMAL_RETURN;

}  /* w_destroySE () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy an SE to another SE
 *
 * LongDescription:::
 *  The destination SE is not created here (use w_createSE ()).  Only
 *  the source strel's contents are copied.  The destination must have
 *  the same dimensions as the source.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  se1::            Source strel
 *
 * Output:::
 *  se2::            Destination strel
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copySE (
    IPWState           *pIPWState,
    SE                 *se1,
    SE                 *se2)
{
    PROLOGUE;
    Int32         i;
    struct point *pt1, *pt2, *pt1_end;

    ERR_GEN ((se1 == NULL) || (se2 == NULL), cErrInvalidParm);

    se2->center_location = se1->center_location;
    se2->parity          = se1->parity;
    se2->nb_cc           = se1->nb_cc;
    se2->size            = se1->size;
    se2->minx            = se1->minx;
    se2->miny            = se1->miny;
    se2->maxx            = se1->maxx;
    se2->maxy            = se1->maxy;

    for (i = 0; i < se1->nb_cc; i++) 
    {
	(se2->origins + i)->x = (se1->origins + i)->x;
	(se2->origins + i)->y = (se1->origins + i)->y;
    }

    for (i = 0; i < se1->size; i++) 
    {
	(se2->pts + i)->x = (se1->pts + i)->x;
	(se2->pts + i)->y = (se1->pts + i)->y;
    }

    for (i = 0; i < 8; i++) 
    {
	(se2->sides + i)->size = (se1->sides + i)->size;
	pt1 = (se1->sides + i)->pts;
	pt2 = (se2->sides + i)->pts;
	pt1_end = pt1 + (se1->sides + i)->size;
	for (; pt1 < pt1_end; pt1++, pt2++) 
	{
	    pt2->x = pt1->x;
	    pt2->y = pt1->y;
	}
    }

    NORMAL_RETURN;

}  /* w_copySE () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Swap x and y dimensions of an SE in place.
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *  se::             Strel to be transposed
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray, strel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lucv
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_transposeSE (
    IPWState           *pIPWState,
    SE                 *se)
{
    PROLOGUE;
    Int32         i;
    struct point *pt1, *pt1_end, *pt2;
    Int16         temp;

    ERR_GEN (se == NULL, cErrInvalidParm);

    temp = se->minx;
    se->minx  = - se->maxx;
    se->maxx  = - temp;
    temp = se->miny;
    se->miny  = - se->maxy;
    se->maxy  = - temp;

    for (i = 0; i < se->nb_cc; i++) 
    {
	(se->origins + i)->x = - (se->origins + i)->x;
	(se->origins + i)->y = - (se->origins + i)->y;
    }

    for (i = 0; i < se->size; i++) 
    {
	(se->pts + i)->x = - (se->pts + i)->x;
	(se->pts + i)->y = - (se->pts + i)->y;
    }

    for (i = 0; i < 4; i++) 
    {
	pt1 = (se->sides + i)->pts;
	pt2 = (se->sides + i + 4)->pts;
	pt1_end = pt1 + (se->sides + i)->size;
	for (; pt1 < pt1_end; pt1++, pt2++) 
	{
	    temp = pt2->x;
	    pt2->x = - pt1->x;
	    pt1->x = - temp;
	    temp = pt2->y;
	    pt2->y = - pt1->y;
	    pt1->y = - temp;
	}
    }

    NORMAL_RETURN;

}  /* w_transposeSE () */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a binary STREL to a gray SE
 *
 * LongDescription:::
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          Binary source strel
 *
 * Output:::
 *  se::             Gray destination strel
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray, binary, strel, depth conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelToSE (
    IPWState           *pIPWState,
    STREL              *strel,
    SE                **se)
{
    PROLOGUE;
    PIXR        *pixr = NULL;

    ERR_GEN (strel == NULL, cErrInvalidParm);

    ERR_PASS (w_strelToPixr (pIPWState, strel, &pixr));

    ERR_PASS_CLEANUP (w_pixrToSE (pIPWState, pixr, strel->cx, strel->cy, se));

    ERR_PASS(w_destroyPixr (pIPWState, &pixr));
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}  /* w_strelToSE () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts a binary PIXR to an SE
 *
 * LongDescription:::
 *  The destination SE is created here, so the pointer should be set to
 *  NULL before its address is passed in.  ON pixels in the PIXR are
 *  converted to hits in the SE.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  pixr::           Binary source PIXR
 *  cx::             Destination SE's center x location
 *  cy::             Destination SE's center y location
 *
 * Output:::
 *  se::             Destination SE (created)
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, gray, strel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrToSE (
    IPWState           *pIPWState,
    PIXR*               pixr,
    Int32               cx,
    Int32               cy,
    SE                **se)
{
    PROLOGUE;
    Int32         countSide[8];        /* array of sizes for createSE()*/
    Int32         i, countPix, sizx, sizy, p;
    Int16         x, y, x1, y1;
    struct point *pts_ptr;
    struct point *side_pts_ptr[8];
    Int16         current_minx, current_maxx;

    ERR_GEN_PIXR (pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN (*se != NULL, cErrInvalidParm);

    ERR_PASS( w_pixrGetWidth (pIPWState, pixr, &sizx));
    ERR_PASS( w_pixrGetHeight (pIPWState, pixr, &sizy));

    /* we fill enough fields of SE for grayscale dilation, erosion.
     * if the SE has more than 1 c.c., we don't care..
     */
    
    for (i = 0; i < 8; i++)
	countSide[i] = 0;

    /* find the parameters for SE allocation */
    countPix = 0;
    current_minx = BIGVAL16;
    current_maxx = SMLVAL;
    for (y = 0; y < sizy; y++) 
    {
	for (x = 0; x < sizx; x++) 
	{
	    ERR_PASS( w_pixrGetPixel (pIPWState, pixr, x, y, &p));
	    if (p != 0)
	    {
		countPix++;
		if (x < current_minx)
		    current_minx = x;
		if (x > current_maxx)
		    current_maxx = x;
		
		for (i = 7; i >= 0; i--)
		{
		    x1 = (Int16)(x + deltax[i]);
		    y1 = (Int16)(y + deltay[i]);

		    if (x1 < 0 || x1 >= sizx || y1 < 0 || y1 >= sizy)
			countSide[i]++;
		    else 
		    {
			ERR_PASS( w_pixrGetPixel (pIPWState, pixr, x1, y1, &p));
		        if (p == 0)
			    countSide[i]++;
		    }
		}
	    }
	}
    }

    /* allocates structuring element: this isn't correct in general,
     * but we don't need it to be correct for erosion and dilation */
    ERR_PASS( w_createSE (pIPWState, 1, countPix, countSide, se));

    /* now fills in the origins array, the pts array
       and the sides arrays */
    pts_ptr = (*se)->pts;

    for (i = 7; i >= 0; i--)
	side_pts_ptr[i] = ((*se)->sides + i)->pts;

    for (y = 0; y < sizy; y++) 
    {
	for (x = 0; x < sizx; x++) 
	{
	    ERR_PASS( w_pixrGetPixel (pIPWState, pixr, x, y, &p));
	    if (p != 0)
	    {
		pts_ptr->x = (Int16)(x - cx);
		pts_ptr->y = (Int16)(y - cy);
		pts_ptr++;
		for (i = 7; i >= 0; i--) 
		{
		    x1 = (Int16)(x + deltax[i]);
		    y1 = (Int16)(y + deltay[i]);

		    if (x1 < 0 || x1 >= sizx || y1 < 0 || y1 >= sizy) 
		    {
			side_pts_ptr[i]->x = (Int16)(x - cx);
			side_pts_ptr[i]->y = (Int16)(y - cy);
			side_pts_ptr[i]++;
		    }
		    else 
		    {
			ERR_PASS( w_pixrGetPixel (pIPWState, pixr, x1, y1, &p));
		        if (p == 0)
			{
			    side_pts_ptr[i]->x = (Int16)(x - cx);
			    side_pts_ptr[i]->y = (Int16)(y - cy);
			    side_pts_ptr[i]++;
			}
		    }
		}
	    }
	}
    }

    /* can now fill the miny and maxy, as well as minx and maxx
       fields of structuring element */
    if (pts_ptr > (*se)->pts) 
    {
	(*se)->miny = (*se)->pts->y;
	(*se)->maxy = (pts_ptr - 1)->y;
	(*se)->minx = (Int16)(current_minx - cx);
	(*se)->maxx = (Int16)(current_maxx - cx);
    }

    NORMAL_RETURN;

}  /* w_pixrToSE () */


/*
 * $Log: strelg.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.21  1997/06/26 20:04:53  ddavies
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Add inclusion of new file rast.pub.
 *
 * Revision 1.20  1996/01/04 20:27:10  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.19  1995/12/13  04:13:36  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/11/13  20:16:33  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.17  1995/11/04  01:36:12  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.16  1995/10/16  22:59:29  dferrell
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.15  1995/10/13  22:05:37  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/10/12  14:20:56  macleod
 * macleod on Thu Oct 12 07:20:47 PDT 1995
 *
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * Got rid of w_printSE, which printed a strel to the error stream.
 * The reason was that it was never being used, and was causing a link
 * failure due to its use of i_printf and its location in the link.
 *
 * Revision 1.13  1995/10/07  03:14:30  ddavies
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * Delete inclusion of ansiprot.h since fprintf is gone.  Add shrutils.pub
 * since i_printf replaced it.  Change all references to stderr to
 * pIPWState->errorProc, pIPWState->errorInfo in the calls to i_printf.
 * This reflects the way printing works now.
 *
 * Revision 1.12  1995/10/02  00:19:07  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * Fixed keywords typo
 *
 * Revision 1.11  1995/09/29  21:12:19  dferrell
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.10  1995/09/21  23:00:02  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.9  1995/09/07  22:10:00  jcraig
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * pixrToSE => w_pixrToSE.  destroyPixr => w_destroyPixr.
 *
 * Revision 1.8  1995/09/07  01:35:39  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.7  1995/09/06  20:23:48  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.6  1995/09/01  18:50:04  dferrell
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.5  1995/08/31  21:39:31  ejaquith
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.4  1995/08/17  21:56:44  dferrell
 * strelg.c -> /product/ipcore/ipwrap/src/RCS/strelg.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 *
 */

