/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  strelio.c
 *
 *         Creation/annihilation:
 *            w_createStrel ()
 *            w_destroyStrel ()
 *
 *         Set/Get strel pixels
 *            w_strelSetPoint ()
 *            w_strelGetPoint ()
 *
 *           the ubiquitous matrix
 *               w_matrix()
 *
 */



#include "defines.pub"  /* for cTrue and cFalse */
#include "ipwcore.prv"

#include "pixr.h"
#include "morphg.h"

#include "strel.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: strelio.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");
#define  MORPH_ACCELERATOR     cTrue


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates a specified STREL
 *
 * LongDescription:::
 *  Given width and height, creates an empty (miss filled) STREL.  The caller
 *  must set the matrix values and center location before use.  The caller is 
 *  responsible for keeping track of the resulting strel--it is not stored in 
 *  the strel manager.  Use w_destroyStrel () to free it.  The result pointer
 *  should be set to NULL before its address is passed in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  width::          Desired width of STREL
 *  height::         Desired height of STREL
 *
 * Output:::
 *  strel::          The resulting miss filled STREL
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createStrel (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    STREL             **strel)
{
    PROLOGUE;

    *strel = (STREL *)IPW_CALLOC (1, sizeof (STREL), cAllocatorData);
    ERR_GEN (*strel == NULL, cErrNomem);
    (*strel)->nx = width;
    (*strel)->ny = height;

    /* 
     * Init strel manager fields. Set the strelID field to a
     * value of unknownStrel. If not overwritten by one of the
     * strel manager initialization functions, this will indicate 
     * that the strel is a custom strel, not one of the named strels.
     */
    (*strel)->ID = unknownStrel;

    ERR_CHECK(w_matrix(pIPWState, width, height, &((*strel)->data)))
    {
        IPW_FREE (*strel, cAllocatorData);
	ERR_RETURN;
    }

    if (MORPH_ACCELERATOR == cTrue)
	(*strel)->accelFlag = 1;

    NORMAL_RETURN;

}  /* w_createStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Frees memory associated with a given STREL.
 *
 * LongDescription:::
 *  Sets STREL pointer to NULL.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          STREL to be destroyed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyStrel (
    IPWState           *pIPWState,
    STREL             **strel)
{
    PROLOGUE;
    Int32          i;

    if (*strel == NULL)
	NORMAL_RETURN;

    for (i = 0; i < (*strel)->ny; i++)
	IPW_FREE ((*strel)->data[i], cAllocatorData);

    IPW_FREE ((*strel)->data, cAllocatorData);

    /* Gets rid of cached stuff */
    ERR_PASS (w_strelCacheFree (pIPWState, *strel));

    IPW_FREE (*strel, cAllocatorData);

    NORMAL_RETURN;

}  /* w_destroyStrel () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Sets the value of a particular (x,y) position in the STREL matrix.
 *
 * LongDescription:::
 *  Sets the value of a particular (x,y) position in the STREL matrix.
 *  The purpose of this abstract interface is to allow one to manipulate
 *  STRELs from interpreters and the like without having to export the data
 *  structure.
 *
 * PrivateDescription:::
 *  If the STREL had any code generated for it, the code is invalid
 *  after the change, so we must free the code.
 *
 * Input:::
 *  pIPWState::      State pointer
 *  x::              column of strel value to set
 *  y::              row of strel value to set
 *  value::          value, HIT, MISS, or DONT_CARE
 *  strel::          STREL to be modified
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelSetPoint (
    IPWState           *pIPWState,
    Int32               x, 
    Int32               y,
    Int32               value,
    STREL              *strel)
{
    PROLOGUE;

    ERR_GEN (strel == NULL, cErrNullPtr);
    ERR_GEN (x < 0 || x >= strel->nx, cErrOutOfBounds);
    ERR_GEN (y < 0 || y >= strel->ny, cErrOutOfBounds);
    ERR_GEN ((value != HIT) && (value != MISS) && (value != DONT_CARE),
	     cErrInvalidParm);

    strel->data[y][x] = (Int16) value;

    /*
     * If there was generated code, it must be destroyed, since we
     * changed the contents of the strel, and the generated code
     * therefore doesn't match the strel anymore 
     */
    ERR_PASS (w_strelCacheFree (pIPWState, strel));

    NORMAL_RETURN;

}  /* w_strelSetPoint */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Gets the value of a particular (x,y) position in the STREL matrix.
 *
 * LongDescription:::
 *  Get the value of a particular (x,y) position in the STREL matrix.
 *  The purpose of this abstract interface is to allow one to manipulate
 *  STRELs from interpreters and the like without having to export the data
 *  structure.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  x::              column of strel value to set
 *  y::              row of strel value to set
 *  strel::          STREL from which to get value
 *
 * Output:::
 *  value::          value at the point requested , HIT, MISS, or DONT_CARE
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelGetPoint (
    IPWState           *pIPWState,
    Int32               x, 
    Int32               y,
    STREL              *strel,
    Int32              *value)
{
    PROLOGUE;

    ERR_GEN (strel == NULL, cErrNullPtr);
    ERR_GEN (x < 0 || x >= strel->nx, cErrOutOfBounds);
    ERR_GEN (y < 0 || y >= strel->ny, cErrOutOfBounds);

    *value = strel->data[y][x];

    NORMAL_RETURN;
}  /* w_strelGetPoint */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Sets the value of the center point for a STERL
 *
 * LongDescription:::
 *  Sets the center point of the STREL to a particular (x,y) position.
 *  The purpose of this abstract interface is to allow one to manipulate
 *  STRELs from interpreters and the like without having to export the data
 *  structure.
 *
 * PrivateDescription:::
 *  If the STREL had any code generated for it, the code is invalid
 *  after the change, so we must free the code.
 *
 * Input:::
 *  pIPWState::      State pointer
 *  x::              Center X
 *  y::              Center Y
 *  strel::          STREL to be modified
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelSetCenter (
    IPWState           *pIPWState,
    Int32               x, 
    Int32               y,
    STREL              *strel)
{
    PROLOGUE;

    ERR_GEN (strel == NULL, cErrNullPtr);
    ERR_GEN (x < 0 || x >= strel->nx, cErrOutOfBounds);
    ERR_GEN (y < 0 || y >= strel->ny, cErrOutOfBounds);

    strel->cx = x;
    strel->cy = y;

    /*
     * If there was generated code, it must be destroyed, since we
     * changed the contents of the strel, and the generated code
     * therefore doesn't match the strel anymore 
     */
    ERR_PASS (w_strelCacheFree (pIPWState, strel));

    NORMAL_RETURN;

}  /* w_strelSetCenter */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Gets the value of the center point for a STREL
 *
 * LongDescription:::
 *  Get the (x,y) position of the STREL's center point.
 *  The purpose of this abstract interface is to allow one to manipulate
 *  STRELs from interpreters and the like without having to export the data
 *  structure.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          STREL from which to get value
 *
 * Output:::
 *  x::              column of center point
 *  y::              row of of center point
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelGetCenter (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *x, 
    Int32              *y)
{
    PROLOGUE;

    ERR_GEN (strel == NULL, cErrNullPtr);
    ERR_GEN (x == NULL, cErrNullPtr);
    ERR_GEN (y == NULL, cErrNullPtr);

    *x = strel->cx;
    *y = strel->cy;

    NORMAL_RETURN;
}  /* w_strelGetCenter */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  If there is any information cached about the STREL, get rid of it
 *
 * LongDescription:::
 *  If there is any information cached about the STREL, get rid of it;
 *  this is usually used when the STREL is modified or when the STREL
 *  is being destroyed
 *  
 * PrivateDescription:::
 *  This flushes any generated code for the STREL, which becomes
 *  invalid if the strel data are modified.
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strel::          STREL for which the cache is to be Freed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strelCacheFree (
    IPWState           *pIPWState,
    STREL              *strel)
{
    PROLOGUE;
    void	  *proc;

    ERR_GEN (strel == NULL, cErrNullPtr);

    if (strel->erodeOp)
         /* We have to do this weird thing with the pointer so that
	  * ALL the compilers will be happy.  The Visual C++ compiler
	  * gets upset if we just try to supply (*strel)->erodeOp as
	  * and arg to IPW_FREE since it wants to free a pointer to
	  * void, not a pointer to a fancy procedure.  On the other
	  * hand, WATCOM gets upset if we cast the pointer in the
	  * call since the macro sets the first arg equal to NULL
	  * and we can't have a cast on the left side of an expression,
	  * now can we?  So, turn off the warning if we're compiling
	  * with Visual C++.
	  */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    {
	proc = (void *)(strel->erodeOp);
	IPW_FREE (proc, cAllocatorData);
	strel->erodeOp = NULL;
    }

    if (strel->longErodeOp)
    {
	proc = (void *)(strel->longErodeOp);
	IPW_FREE (proc, cAllocatorData);
	strel->longErodeOp = NULL;
    }


    if (strel->dilateOp)
    {
	proc = (void *)(strel->dilateOp);
	IPW_FREE (proc, cAllocatorData);
	strel->dilateOp = NULL;
    }

    if (strel->longDilateOp)
    {
	proc = (void *)(strel->longDilateOp);
	IPW_FREE (proc, cAllocatorData);
	strel->longDilateOp = NULL;
    }
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif


    NORMAL_RETURN;

}  /* w_strelCacheFree () */


/*--------------------------------------------------------------------*
 *                          Matrix Constructor                        *
 *--------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Construct a matrix as an array of pointers to arrays.
 *
 * LongDescription:::
 * Returns a "matrix" that has ny rows of nx Int16 arrays in each row.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  nx::             	number of elements in each row
 *  ny::             	number of rows
 *
 * Output:::
 *  pppMatrix::      	matrix
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully allocated matrix
 *
 * FunctionCategory:::
 *  array management 
 *
 * Keywords:::
 *  array management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *  Adapted from Numerical Recipes.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_matrix (
    IPWState           *pIPWState,
    Int32               nx,
    Int32               ny,
    Int16            ***pppMatrix)

{
PROLOGUE;
Int16   **m;
Int32     j;

        /* m is a pointer to an array of ny pointers */
    m = (Int16 **)IPW_CALLOC(ny, sizeof(Int16 *), cAllocatorData);
    ERR_GEN(m == NULL, cErrNomem);
        /* each of which points to an array of nx short ints */
    for (j = 0; j < ny; j++)
    {
        m[j] = (Int16 *)IPW_CALLOC (nx, sizeof(Int16), cAllocatorData);
	ERR_GEN_CLEANUP(m[j] == NULL, cErrNomem);
    }
    *pppMatrix = m;
    NORMAL_RETURN;

ERR_CLEANUP:
    for (j = 0; j < ny; j++)
    	if (m[j] != NULL) IPW_FREE(m[j], cAllocatorData);
 	else break;
    IPW_FREE(m, cAllocatorData);
    ERR_RETURN;
}


/*
 *
 * $Log: strelio.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.27  1998/01/21 21:56:53  ddavies
 * ddavies on Wed Jan 21 13:56:12 PST 1998
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Moved w_matrix from utils1.c to here.  It's only used by the strel code
 * and breaks a circular dependency between utils and strel.
 *
 * Revision 1.26  1998/01/14 00:15:33  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.25  1996/12/18 23:19:24  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Add code to free machine code pointed to by longErodeOp and longDilateOp in
 * the operation that frees STRELs.
 *
 * Revision 1.24  1996/01/04 20:27:14  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.23  1995/12/13  04:13:40  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/12/08  20:02:20  macleod
 * macleod on Fri Dec 8 12:02:01 PST 1995
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Added ERR_PASS() around a call which didn't have one, so was ignoring
 * the error code
 *
 * Revision 1.21  1995/12/08  19:50:11  macleod
 * macleod on Fri Dec 8 11:49:48 PST 1995
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 *  Fix to shut up the NT compiler's whining
 *
 * Revision 1.20  1995/12/07  20:26:24  macleod
 * macleod on Thu Dec 7 12:25:21 PST 1995
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Added accessor/mutator functions:
 * w_strelGetPoint(), w_strelSetPoint(), wStrelGetCenter(), wStrelSetCenter()
 * Also added w_strelCacheFree(), for freeing cached code, since the code
 * needs to be freed in the mutator functions as well as the destroy function.
 *
 * Revision 1.19  1995/11/13  20:16:37  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/18  04:18:48  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Add casts to make Visual C++ happy.
 *
 * Revision 1.17  1995/10/13  22:05:40  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/10/12  04:28:42  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.15  1995/09/29  21:12:38  dferrell
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Updated function comment blocks.  Replaced TRUE and FALSE with
 * cTrue and cFalse, respectively.
 *
 * Revision 1.14  1995/09/21  23:00:05  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_v" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/07  22:08:21  jcraig
 * jcraig on Thu Sep 7 15:07:54 PDT 1995
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * matrix => w_matrix
 *
 * Revision 1.12  1995/09/06  20:23:51  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  18:49:42  dferrell
 * dferrell on Fri Sep 1 11:49:40 PDT 1995
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.10  1995/08/31  21:39:34  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/29  00:16:41  ejaquith
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * fixed routine destroyStrel() which was returning "cSuccessful" via the
 * error returning macros when the passed in strel was null.
 * Now we just return cSuccessful if it's null because that's not an error
 * (and all destroy..() routines should do the following:
 *  check if ptr is null.  if so, great, exit successfully.
 *  ow, free what ptr points to completely.
 *  set ptr to NULL
 *  exit sucessfully).
 * (yeah yeah it's mr. more-anal-than-though-coder)
 *
 * Revision 1.8  1995/08/17  21:56:52  dferrell
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.7  1995/01/26  22:28:57  lperry
 * lperry on Thu Jan 26 14:24:55 PST 1995
 *
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Initialize the ID field of every strel with the value
 * unknownStrel. This is later overwritten by the strel
 * manager with the ID of any NAMED strels. This will
 * allow detection of any strels not allocated and
 * initialized by the strel manager.
 *
 * Also remove inits of obsolete fields in the strel
 * itself (crc and refCount). Reference counting is now
 * maintained in the strel list not within the strel
 * struct itself.
 *
 * Revision 1.6  1994/09/29  21:19:07  lperry
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Change to use strel.pub instead of strlpros.h
 *
 * Revision 1.5  1994/08/10  22:08:45  lperry
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Add code to createStrel to initialize three new fields --
 * a crc, and ID number and a reference count.
 *
 * Revision 1.4  1994/06/08  04:13:54  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.3  1993/12/03  01:19:26  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:17:49  ddavies
 * strelio.c -> /product/ipcore/ipwrap/src/RCS/strelio.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

