/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *        w_initStrelMgr ()
 *        w_termStrelMgr ()
 *        w_releaseStrelByName ()
 *        w_releaseAllStrels ()
 *        w_getStrel ()
 *
 *        wp_createStrelMgrInstance ()
 *        wp_destroyStrelMgrInstance ()
 *
 */

#include "ipwcore.prv"

#include "defines.pub"
#include "morphg.h"
#include "pixr.h"

#include "strel.pub"


/* local (static) prototypes */
static ErrCode CDECL
wp_createStrelMgrInstance (IPWState *pIPWState, StrelInfo **sinfo);

static ErrCode CDECL
wp_destroyStrelMgrInstance (IPWState *pIPWState, StrelInfo **sinfo);


IP_RCSINFO(RCSInfo, "$RCSfile: strelmgr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/***************************************************************
*                 STREL MANAGER 
*
*   This file contains code to implement a manager
*   function for strels. Previously, a typical application would
*   allocate all the strels it needed at the beginning of 
*   execution. This was done using a set of three high level
*   routines, each of which allocated a large group of strels.
*
*   While the strel data isn't all that large,
*   the assembly code generated to accelerate morphological ops
*   using strels is about 1500 to 2K bytes per strel.
*   As we deliver DLLs using strel code, there is the possibility
*   of multiple applications calling the strel init routines.
*   Since the strel data is constant, there isn't really a
*   reentrancy problem, but there is a potential for serious
*   errors as strels are released (freed). If one application
*   frees a strel that another application is using -- major
*   catastrophic failure.
*
*   There is also the problem of two applications both allocating 
*   the same strel. There are then two identical 
*   copies around, both pointed to by the same global pointer.
*   Only the last one allocated can ever be freed, creating a 
*   potential memory leak. If the first user allocated all his
*   strels at one time to avoid memory fragmentation, then the
*   second user's allocation would undo these efforts.
*
*   To prevent this kind of occurrence, this file implements a
*   strel manager which creates a structure listing all
*   of the "named" strels. Names are just elements of an
*   enumerated type which act as indices into an array holding
*   all of the strel pointers. This array is initialized to
*   NULLs on creation. Later non-null values indicate that the
*   strel has been allocated. Each call to w_initStrelMgr ()
*   creates a new instance of the StrelInfo structure, but
*   all strel instances (at least those allocated using
*   strel manager services) share a single set of strels,
*   which are tracked by a separate "global" StrelInfo struct.
*
*****************************************************************/

/* 
  This is the only piece of global state involved
  in the strel manager. This points to the REAL
  list of strels. All user instances of the sinfo structure
  contain the same strel pointers which are on the gSinfo structure.
*/

static StrelInfo *gSinfo = NULL; 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocates and initializes a data structure for tracking a single 
 *  "user's" strels.
 *
 * LongDescription:::
 *  Also allocates the "standard" strel table, but does not initialize
 *  any strels.  This function is used to create both the global strel
 *  structure (gSinfo) as well as client strel manager instances.  If
 *  the created structure is not the global one, the caller must keep
 *  track of it until it is destroyed using wp_destroyStrelMgrInstance ().
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *  sinfo::          The strel info structure to be created
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 * 
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_createStrelMgrInstance (
    IPWState           *pIPWState,
    StrelInfo         **sinfo)
{
    PROLOGUE;

    ERR_GEN (*sinfo != NULL, cErrInvalidParm);

    /* Create and initialize user StrelInfo structure
     * Supports accounting for this user's strels.
     * Use CALLOC to insure that all strel ptrs are
     * initialized to zero.
     */

    *sinfo = (StrelInfo *)IPW_CALLOC (1, sizeof (StrelInfo), cAllocatorData);
    ERR_GEN (*sinfo == NULL, cErrNomem);

    /* Store a reference to the global strel pointer in every
     * instance of the StrelInfo structure. This is sort of a
     * hack, but it lets code dealing with StrelInfo instances
     * get at the global gSinfo pointer. Otherwise all of the
     * code would have to be in THIS file. This would be much
     * cleaner in C++ with its public and private members.
     */
    if (gSinfo == NULL)
        (*sinfo)->globalStrelInfo = *sinfo;
    else
        (*sinfo)->globalStrelInfo = gSinfo;

    /* Create the array which holds all of the strel pointers.
     * Use CALLOC to insure that all strel pointers are 
     * initialized to NULL and the reference count is zero..
     * The array size is determined by two dummy elements in the
     * enumeration list which bracket all the strel names.
     */
    (*sinfo)->strelArray = (StrelRef *)IPW_CALLOC (strelListEnd - 
                                                   strelListStart, 
                                                   sizeof (StrelRef), 
						   cAllocatorData);
    if ((*sinfo)->strelArray == NULL)
    {
        IPW_FREE (*sinfo, cAllocatorData);
	ERR_GEN (1, cErrNomem);
    }

    /* Set these flags to indicate that none of the 
     * "group" initializations have been performed.
     * This will eliminate the need to check individual
     * strels in later calls.
     */
    (*sinfo)->strel1InitsDone = cFalse;
    (*sinfo)->strel11InitsDone = cFalse;
    (*sinfo)->strel12InitsDone = cFalse;

    NORMAL_RETURN;

}  /* wp_createStrelMgrInstance () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroys an instance of the strel info structure
 *
 * LongDescription:::
 *  If any strels are currently sotred in this instance, they are
 *  also destroyed.  The strel info pointer is set to NULL.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  sinfo::          Strel info to be destroyed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_destroyStrelMgrInstance (
    IPWState           *pIPWState,
    StrelInfo         **sinfo)
{
    PROLOGUE;

    /* Check arguments */
    if (*sinfo == NULL)
        NORMAL_RETURN;

    /* Release all of the strels */
    ERR_PASS( w_releaseAllStrels (pIPWState, *sinfo));

    /* Release the strel array */
    if ((*sinfo)->strelArray != NULL)
        IPW_FREE (((*sinfo)->strelArray), cAllocatorData);

    /* Free the StrelInfo structure itself */
    IPW_FREE (*sinfo, cAllocatorData);

    NORMAL_RETURN;

}  /* wp_destroyStrelMgrInstance () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Public initialization function for the strel manager.
 *
 * LongDescription:::
 *  If the global strel manager instance has not been created, it
 *  is allocated first.  The user instance of the strel manager
 *  associated with the state structure passed in is then initialized.
 *  No strels are actually created.  Use w_strel1Inits (), 
 *  w_strel11Inits (), w_strel12Inits (), or w_initStrelByName () to
 *  create and store strels in the strel manager.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Client state structure whose strel manager
 *                   instance is to be created.
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initStrelMgr (
    IPWState           *pIPWState)
{
    PROLOGUE;
    StrelInfo         *sinfo = NULL;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);

    /* If the gSinfo variable is NULL. it
     * indicates that the global strel list has
     * never been allocated. This only happens
     * on the first call to initStrelMgr(),
     * or when all strel managers have terminated.
     * We want to have a complete global
     * instance as well as complete user
     * instances of the strel manager
     * data structure.
     */

    if (gSinfo == NULL)
    {
        ERR_PASS( wp_createStrelMgrInstance (pIPWState, &gSinfo));
        gSinfo->userCount = 0;
    }

    /* Create and initialize user StrelInfo structure
     * Supports accounting for this user's strels.
     */
    gSinfo->userCount++;
    ERR_PASS( wp_createStrelMgrInstance (pIPWState, &sinfo));
    ERR_PASS( w_IPWStateSetStrelInfo (pIPWState, sinfo));

    NORMAL_RETURN;

}  /* w_initStrelMgr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Releases all strel manager memory for a given client state instance.
 *
 * LongDescription:::
 *  Any strels managed by the state structure's associated strel manager
 *  instance which are not used by other instances are destroyed.  If no
 *  other client strel manager instances exist, the global strel manager
 *  instance is also destroyed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      Client's state structure whose strel manager
 *                   instance is to be destroyed
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_termStrelMgr (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);

    /* Destroy the resources for this instance of the strel mgr */
    ERR_PASS( wp_destroyStrelMgrInstance (pIPWState, &(pIPWState->strelInfo)));

    /* Decrement the global user count */
    gSinfo->userCount--;

    /* If this is the last user, destroy the resources of the global strelmgr */
    if (gSinfo->userCount == 0)
    {
        ERR_PASS( wp_destroyStrelMgrInstance(pIPWState, &gSinfo));
        gSinfo = NULL;
    }

    NORMAL_RETURN;

}  /* w_termStrelMgr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Removes registration info for a specific strel under the specified strel manager
 * LongDescription:::
 *  To release a strel under the client's state info, pass a NULL pointer
 *  for the sinfo parameter.  If sinfo is non-NULL, this is the strel manager
 *  info which will be accessed.  If there are no other strel manager 
 *  instances which use this strel, the strel's memory is actually released.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  sinfo::          The specific strel manager instance (or NULL)
 *  strelID::        The strel ID (constant) to be released
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_releaseStrelByName (
    IPWState           *pIPWState,
    StrelInfo          *sinfo,
    StrelID             strelID)
{
    PROLOGUE;

    /* Check arguments */
    ERR_GEN (pIPWState == NULL, cErrNullIPWState);

    if (pIPWState->strelInfo == NULL)
	ERR_GEN (sinfo == NULL, cErrStrel_MgrNotInitted);
    else
        if (sinfo == NULL)
            sinfo = pIPWState->strelInfo;

    /* If all references to this strel in this instance have been
     * released already, don't let this go below zero. Otherwise,
     * we'd be releasing someone elses strel references.
     */
    ERR_GEN ((sinfo->strelArray[strelID].refCount <= 0), 
              cErrInternal);

    /* OK, its safe to decrement reference counts now */
    gSinfo->strelArray[strelID].refCount--;
    /* BUG 1/22/97:  w_termStrelMgr destroys the strels associated with
     * a particular IPState.  If that was the last IPState, it actually
     * destroys the strels.  However, in that case, gSinfo == sinfo.
     * Since w_releaseStrelsByName reads the initial refCount and
     * calls this routine that many times, AND this routine returns
     * an error if refCount goes negative, do not decrement refCount
     * twice here.
     */
    if (gSinfo != sinfo)
	sinfo->strelArray[strelID].refCount--;
     
    /* If this is the last reference, destroy strel and reset the entries */
    if (gSinfo->strelArray[strelID].refCount <= 0)
    {
	/* Reset the user entries */
        sinfo->strelArray[strelID].strel = NULL; 
        sinfo->strelArray[strelID].refCount = 0; 

	/* Reset the global entries */
        gSinfo->strelArray[strelID].refCount = 0; 
        ERR_PASS( w_destroyStrel (pIPWState, &(gSinfo->strelArray[strelID].strel)));
    }

    NORMAL_RETURN;

}  /* w_releaseStrelByName () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Release registration of all strels in a particular strel manager instance
 *
 * LongDescription:::
 *  If any of the strels released from the specific strel manager isntance
 *  are not in use by any other instance, they are actually destroyed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  sinfo::          Strel manager instance which is to have its strels released
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_releaseAllStrels (
    IPWState           *pIPWState,
    StrelInfo          *sinfo)
{
    PROLOGUE;
    Int32 i, j;
    StrelID strelID;
    Int32 instanceRefCount;

    /* Check arguments */
    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (sinfo == NULL, cErrStrel_MgrNotInitted);

    /* Step through all the possible strels and release each.
     * To make sure that extra strel init ops are accounted for,
     * release the strel the same number of times as its ref count
     * in THIS instance.
     */
    for (i = strelListStart + 1; i < strelListEnd; i++)
    {
        strelID = (StrelID) i;
        if (strelID != unknownStrel)  /* Ignore this placeholder */
        {
            instanceRefCount = sinfo->strelArray[strelID].refCount;
            if (instanceRefCount > 0)
            {
                for (j = 0; j < instanceRefCount; j++)
                    ERR_PASS( w_releaseStrelByName (pIPWState, sinfo, strelID));
            }
        }
    }

    NORMAL_RETURN;

}  /* w_releaseAllStrels () */


/**************************************************************
 * getStrel ()
 *
 *    Given a strel ID, set strelPtr to the strel manager
 *    instance for that strel.
 *
 *  Returns: 
 *    cSuccessful:  On success;
 *    cErrNullIPWState: no IPWState
 *    cErrStrel_MgrNotInitted: strel manager not initialized
 *    cErrInvalidParm: *strelPtr != NULL
 *    cErrStrel_InvalidID: not a valid strel ID
 *    cErrStrel_IDNotInitted: strel ID has not yet been initialized
 **************************************************************/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Given a strel ID, set the strel pointer to the corresponding strel.
 *
 * LongDescription:::
 *  Uses the strel manager associated with the IPWState instance.
 *  
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strelID::        Strel ID (constant) of the strel to be gotten
 *
 * Output:::
 *  strelPtr::       The resulting strel
 *
 * Returns:::
 *    cSuccessful::  Successful completion
 *    cErrNullIPWState:: no IPWState
 *    cErrStrel_MgrNotInitted:: strel manager not initialized
 *    cErrInvalidParm:: *strelPtr != NULL
 *    cErrStrel_InvalidID:: not a valid strel ID
 *    cErrStrel_IDNotInitted:: strel ID has not yet been initialized
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  strel management
 *
 * Keywords:::
 *  strel management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getStrel (
    IPWState           *pIPWState,
    StrelID             strelID,
    STREL             **strelPtr)
{
    PROLOGUE;

    /* Check arguments */
    ERR_GEN (pIPWState == NULL, cErrNullIPWState);

    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    ERR_GEN (*strelPtr != NULL, cErrInvalidParm);

    if ((strelID < strelListStart + 1) || (strelID > strelListEnd - 1))
        ERR_GEN (1, cErrStrel_InvalidID);

    *strelPtr = pIPWState->strelInfo->strelArray[strelID].strel;

    if (*strelPtr == NULL)
        ERR_GEN (1, cErrStrel_IDNotInitted);

    NORMAL_RETURN;

}  /* w_getStrel () */



/*
 * $Log: strelmgr.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.28  1998/01/21 22:07:52  ddavies
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Use accessor to set pIPWState->strelInfo instead of doing it directly.
 *
 * Revision 1.27  1997/03/14 17:15:30  dferrell
 * dferrell on Fri Mar 14 09:12:43 PST 1997
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Fixed egregious bug (use of single index variable for two nested loops)
 * which caused memory leaks in the strel manager in certain fairly common
 * conditions.
 *
 * Revision 1.26  1997/01/31 00:56:21  ddavies
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Fix a bug encountered when releasing the last strels.  In that case, the
 * global strel info pointer is handed to w_releaseStrelByName.  When a
 * strel is released, w_releaseStrelByName decrements the instance counts
 * in both the arg StrelInfo and in the global manager, effectively
 * decrementing the count by 2 in this case.  This eventually causes a
 * error when we check to see if the count is <= 0.  We now only
 * decrement the count by 1 in this case.
 * Why didn't we see this until now???
 *
 * Revision 1.25  1996/01/04 20:27:16  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.24  1995/12/13  04:13:44  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/11/13  20:16:40  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.22  1995/10/16  23:27:07  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Removed obsolete #def and #undef of _STRELMGR_C_.
 *
 * Revision 1.21  1995/10/13  22:05:43  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/09/29  21:12:48  dferrell
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.19  1995/09/21  23:00:09  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/06  20:23:53  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.17  1995/09/05  17:40:20  dferrell
 * dferrell on Tue Sep 5 10:39:43 PDT 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Added new strel manager accessor function, getStrel (), which replaces
 * the GET_STREL macro.
 *
 * Revision 1.16  1995/09/01  21:43:38  dferrell
 * dferrell on Fri Sep 1 14:43:15 PDT 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Removed duplicate IP_RCSINFO stuff (whoops).
 *
 * Revision 1.15  1995/09/01  18:50:15  dferrell
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.14  1995/08/31  21:39:37  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.13  1995/08/30  17:28:07  ejaquith
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Made routines accessing pIPWState first check that it is valid (not NULL).
 * If it's null, return error: cErrNullIPWState.
 *
 * Revision 1.12  1995/08/22  23:45:40  dferrell
 * dferrell on Tue Aug 22 16:43:30 PDT 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Modified logic, references, and added new parameters to releaseAllStrels ()
 * and releaseStrelByName in order to fix problems associated with cleaning
 * up both local and global strel manager structures.  Yuck.
 *
 * Revision 1.11  1995/08/22  17:52:56  dferrell
 * dferrell on Tue Aug 22 10:52:26 PDT 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Fixed bad parameter checking logic in createStrelMgrInstance ().
 *
 * Revision 1.10  1995/08/22  17:17:32  dferrell
 * dferrell on Tue Aug 22 10:17:11 PDT 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Set sinfo to NULL at start of initStrelMgr ().
 *
 * Revision 1.9  1995/08/17  21:56:59  dferrell
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.8  1995/01/26  22:31:03  lperry
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Avoid taking any action on the named strel "unknownStrel".
 * in the releaseAllStrels fucntion. This is just a
 * placeholder name to indicate a strel not under
 * strel manager control.
 *
 * Revision 1.7  1995/01/25  22:34:19  lperry
 * lperry on Wed Jan 25 14:33:15 PST 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Add typecast to CALLOC. Pre-emptive strike before
 * macintosh compile bombs. How come gcc ignores this ?
 *
 * Revision 1.6  1995/01/24  22:17:21  lperry
 * lperry on Tue Jan 24 14:08:43 PST 1995
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Major change to the strel manager. Previously,
 * a strel was actually created every time an initStrel
 * call was made. It was then tested against the existing strels
 * to see if it was the same. This was meant to allow
 * a user to create any old strel and the manager would determine
 * if a matching one already existed.
 *
 * This was major overkill, and was also consuming a good deal
 * of time on the PC. The new implementation only creates
 * single instances of strels which are members of a know set,
 * i.e. those initialized by the strel1Inits, etc calls.
 * Individual strels can also be initialized. Reference counting
 * is used to control access, as before. However, since the
 * list of possible strels is now restricted, an array is used
 * rather than a linked list to do the bookeeping.
 *
 * Global variables are no longer used for strels. Instead, the
 * strel manager must be used to get a strel state variable.
 * A macro GET_STREL(strel5h), for example is used to access
 * the strel from the array. so that the old familiar names
 * for strels can still be used.
 *
 * Revision 1.5  1994/09/29  21:19:29  lperry
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Change to use strel.pub instead of strlpros.h
 * Also include strel.prv for static fucntions
 *
 * Revision 1.4  1994/09/29  00:48:14  lperry
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 *
 * Revision 1.3  1994/08/18  00:35:43  lperry
 * lperry on Wed Aug 17 17:32:18 PDT 1994
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Change to support reference count tracking per user
 * for each strel. Previous version considered it an error
 * for one user to allocate the same strel more than once.
 * This change allows multiple modules in a program to
 * init the same strel, but they'll still be sharing a
 * single copy.
 *
 * Revision 1.2  1994/08/11  22:49:27  lperry
 * lperry on Thu Aug 11 15:48:00 PDT 1994
 *
 * strelmgr.c -> /product/ipcore/ipwrap/src/RCS/strelmgr.c,v
 *
 * Modify termStrelMgr to free the StrelInfo structure (sinfo).
 * Callers should not reference sinfo after the call to termStrelMgr
 * since it will then be an invalid pointer.
 *
 * Revision 1.1  1994/08/10  22:16:14  lperry
 * Initial revision
 *
*/

