/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *       w_createHorizRuleStrel ()
 *       w_createVertRuleStrel ()
 *       w_makeSolidCenterStrel ()
 *       w_initStrelByName ()
 */

#include "ipwcore.prv"

#include "morphg.h"
#include "pixr.h"
#include "strel.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: strlinit.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/******************************************************* 
 *  Utility functions to create horizontal and vertical
 *  rule-removing strels. The alternate dimension is
 *  always 5 in these strels
 *******************************************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates a STREL which matches some horizontal rules
 *
 * LongDescription:::
 *  The STREL will be 5 x (2 * ruleWidth) in size.  The caller is 
 *  responsible for keeping track of the resulting strel--it is not stored 
 *  in the strel manager.  Use w_destroyStrel () to free it.  The result 
 *  pointer should be set to NULL before its address is passed in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  ruleWidth::      Size (height) of rules to match
 *
 * Output:::
 *  strel::          Resulting STREL
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createHorizRuleStrel (
    IPWState           *pIPWState,
    UInt32              ruleWidth,
    STREL             **strel)
{
    PROLOGUE;

    ERR_GEN (*strel != NULL, cErrInvalidParm);

    ERR_PASS( w_createStrel(pIPWState, 5, 2 * ruleWidth + 1, strel));

    (*strel)->cx = 2;
    (*strel)->cy = ruleWidth;

    (*strel)->data[0][0] = 2;  /* Misses on top row */
    (*strel)->data[0][2] = 2;
    (*strel)->data[0][4] = 2;

    (*strel)->data[2*ruleWidth][0] = 2;  /* Misses on bottom row */
    (*strel)->data[2*ruleWidth][2] = 2;
    (*strel)->data[2*ruleWidth][4] = 2;

    (*strel)->data[ruleWidth][0] = 1;  /* Hits across the middle */
    (*strel)->data[ruleWidth][2] = 1;
    (*strel)->data[ruleWidth][4] = 1;

    NORMAL_RETURN;

}  /* w_createHorizRuleStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates a STREL which matches some vertical rules
 *
 * LongDescription:::
 *  The STREL will be (2 * ruleWidth) x 5 in size.  The caller is 
 *  responsible for keeping track of the resulting strel--it is not stored 
 *  in the strel manager.  Use w_destroyStrel () to free it.  The result 
 *  pointer should be set to NULL before its address is passed in.
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  ruleWidth::      Size (width) of rules to be matched
 *
 * Output:::
 *  strel::          Resulting STREL
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createVertRuleStrel (
    IPWState           *pIPWState,
    UInt32              ruleWidth,
    STREL             **strel)
{
    PROLOGUE;

    ERR_GEN (*strel != NULL, cErrInvalidParm);

    ERR_PASS( w_createStrel(pIPWState, 2 * ruleWidth + 1, 5, strel));

    (*strel)->cx = ruleWidth;
    (*strel)->cy = 2;

    (*strel)->data[0][0] = 2;  /* Misses on left edge */
    (*strel)->data[2][0] = 2;
    (*strel)->data[4][0] = 2;

    (*strel)->data[0][2*ruleWidth] = 2;  /* Misses on right edge */
    (*strel)->data[2][2*ruleWidth] = 2;
    (*strel)->data[4][2*ruleWidth] = 2;

    (*strel)->data[0][ruleWidth] = 1;  /* Hits down the middle */
    (*strel)->data[2][ruleWidth] = 1;
    (*strel)->data[4][ruleWidth] = 1;

    NORMAL_RETURN;

}  /* w_createVertRuleStrel () */


/**************************************************
 *  Utility function to create "solid" strels, 
 *  i.e. strels with all elements set to 1 (HIT).
 *  The origin of the strel is the center of the matrix.
 *  For strels with an even number of elements, this
 *  is always the next lower integral position.
 **************************************************/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Creates a rectangular STREL which is filled with hits and centered
 *
 * LongDescription:::
 *  The caller is responsible for keeping track of the resulting strel--it
 *  is not stored in the strel manager.  Use w_destroyStrel () to free it.
 *  The result pointer should be set to NULL before its address is passed 
 *  in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  width::          Width of strel
 *  height::         Height of strel
 *
 * Output:::
 *  strel::          Resulting strel
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_makeSolidCenterStrel (
    IPWState           *pIPWState,
    UInt32              width,
    UInt32              height,
    STREL             **strel)
{
    PROLOGUE;
    UInt32 i, j;

    ERR_GEN (*strel != NULL, cErrInvalidParm);

    ERR_PASS( w_createStrel(pIPWState, width, height, strel));

    (*strel)->cx = (width - 1) / 2;
    (*strel)->cy = (height - 1) / 2;
    for (i = 0; i < height; i++)
        for (j = 0; j < width; j++)
            (*strel)->data[i][j] = 1;

    NORMAL_RETURN;

}  /* makeSolidCenterStrel () */


/**************************************************************
 *
 *    Int32 * CDECL
 *    initStrelByName( StrelInfo sinfo,  StrelID strelID )
 *
 *  The initStrel function is used to allocate and
 *  "register" any named strel. All of the common
 *  strels have IDs.The IDs are members of the enumerated
 *  type StrelID.  In most cases, the StrelID is very
 *  similar to the global variable. For example, if
 *  the global variable was strel2h, the StrelID
 *  would be namedStrel2h. The rule is a little 
 *  different for strels with hypheneated names.
 *  The strel s_pepper_1 would have a StrelID of
 *  namedStrel_pepper_1.
 *
 *  The initStrel function allocates and initializes
 *  a single strel. After creating the strel, a CRC
 *  is calculated from the strel contents. This is then
 *  hashed to a table address. If the hash table contains
 *  the identical strel, as verified by a complete comparison 
 *  of the strel parameters and data contents, the strel
 *  just created is destroyed and the existing strel's
 *  address is returned. This allows a single copy of the
 *  strel to serve multiple applications.
 *
 *  This process of using the strel contents to verify
 *  sameness was used rather than just using the StrelID
 *  in order to allow this same code to be used with
 *  "custom" strels, which might be generated on a more
 *  adhoc basis and would never get added to the set of
 *  named strels. A very large set of named strels was
 *  created in hopes that the need for custom strel
 *  generation would be minimized. If a strel is not
 *  on the "named" list, and it get used regularly,
 *  it should be added to the list.
 *
 *   Returns:
 *      ptr to the strel structure
 *          or 
 *      NULL on error
 *
 **************************************************************/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate and "register" any named strel in the strel manager
 *
 * LongDescription:::
 *  After creating the strel, a CRC is calculated from the strel 
 *  contents. This is then hashed to a table address. If the hash 
 *  table contains the identical strel, as verified by a complete
 *  comparison of the strel parameters and data contents, the strel
 *  just created is destroyed and the existing strel's address is 
 *  returned. This allows a single copy of the strel to serve multiple 
 *  applications.
 *
 *  This process of using the strel contents to verify sameness was 
 *  used rather than just using the StrelID in order to allow this
 *  same code to be used with "custom" strels, which might be generated 
 *  on a more adhoc basis and would never get added to the set of named 
 *  strels. A very large set of named strels was created in hopes that 
 *  the need for custom strel generation would be minimized. If a strel 
 *  is not on the "named" list, and it get used regularly, it should be 
 *  added to the list.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  strelID::        Strel "name" constant
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, strel management, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_initStrelByName (
    IPWState           *pIPWState,
    StrelID             strelID)
{
    PROLOGUE;
    Int32        i;
    STREL        *strel = NULL;
    StrelInfo    *gSinfo, *sinfo;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    sinfo = pIPWState->strelInfo;
    ERR_GEN (sinfo == NULL, cErrStrel_MgrNotInitted);

    /* We need to be able to get access to the global strel list.
     * So get the pointer to it which is store in each StrelInfo
     * instance. This would be much cleaner in C++.
     */
    gSinfo = sinfo->globalStrelInfo;
    ERR_GEN (gSinfo == NULL, cErrStrel_MgrNotInitted);

    /* If already allocated by THIS strel mgr instance, 
     * increment reference counts and return.
     */
    if (sinfo->strelArray[strelID].strel != NULL)
    {
        gSinfo->strelArray[strelID].refCount++;
        sinfo->strelArray[strelID].refCount++;

        NORMAL_RETURN;
    }

    /* If strel has already been allocated previously by 
     * ANOTHER instance of the strel manager,
     * increment its global reference count and return the
     * global pointer to it on the user's strel list
     */
    if (gSinfo->strelArray[strelID].refCount > 0)
    {
        gSinfo->strelArray[strelID].refCount++;
        sinfo->strelArray[strelID].refCount++;
        sinfo->strelArray[strelID].strel = 
                gSinfo->strelArray[strelID].strel;

        NORMAL_RETURN;
    }

    /* OTHERWISE, allocate it for real and add it to the global
     * list and the user list.
     */
    switch (strelID)
    {

	case strel2h:
	    /* horizontal strel of length 2: (1c 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 2, 1, &strel));
	    break;
    
	case strel2v:
	    /* vertical strel of length 2: v(1c 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 2, &strel));
	    break;
    
	case strel3h:
	    /* horizontal strel of length 3: (1 1c 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 3, 1, &strel));
	    break;
    
	case strel3v:
	    /* vertical strel of length 3: v(1 1c 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 3, &strel));
	    break;
    
	case strel4h:
	    /* horizontal strel of length 4: (1 1c 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 4, 1, &strel));
	    break;
    
	case strel4v:
	    /* vertical strel of length 4: v(1 1c 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 4, &strel));
	    break;
    
	case strel5h:
	    /* horizontal strel of length 5: (1 1 1c 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 5, 1, &strel));
	    break;
    
	case strel5v:
	    /* vertical strel of length 5: v(1 1 1c 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 5, &strel));
	    break;
    
	case strel6h:
	    /* horizontal strel of length 6: (1 1 1c 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 6, 1, &strel));
	    break;
    
	case strel6v:
	    /* vertical strel of length 6: v(1 1 1c 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 6, &strel));
	    break;
    
	case strel7h:
	    /* horizontal strel of length 7: (1 1 1 1c 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 7, 1, &strel));
	    break;
    
	case strel7v:
	    /* vertical strel of length 7: v(1 1 1 1c 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 7, &strel));
	    break;
    
	case strel8h:
	    /* horizontal strel of length 8: (1 1 1 1c 1 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 8, 1, &strel));
	    break;
    
	case strel8v:
	    /* vertical strel of length 8: v(1 1 1 1c 1 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 8, &strel));
	    break;
    
	case strel9h:
	    /* horizontal strel of length 9: (1 1 1 1 1c 1 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 9, 1, &strel));
	    break;
    
	case strel9v:
	    /* vertical strel of length 9: v(1 1 1 1 1c 1 1 1 1) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 9, &strel));
	    break;
    
	case strel10h:
	    /* horizontal strel of length 10: ((4:1) 1c (5:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 10, 1, &strel));
	    break;
    
	case strel10v:
	    /* vertical strel of length 10: v((4:1) 1c (5:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 10, &strel));
	    break;
    
	case strel11h:
	    /* horizontal strel of length 11: ((5:1) 1c (5:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 11, 1, &strel));
	    break;
    
	case strel11v:
	    /* vertical strel of length 11: v((5:1) 1c (5:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 11, &strel));
	    break;
    
	case strel15h:
	    /* horizontal strel of length 15: ((7:1) 1c (7:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 15, 1, &strel));
	    break;
    
	case strel15v:
	    /* vertical strel of length 15: v((7:1) 1c (7:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 15, &strel));
	    break;
    
	case strel19h:
	    /* horizontal strel of length 19: ((9:1) 1c (9:1)) */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 19, 1, &strel));
	    break;
    
	case strel19v:
	    /* vertical strel of length 19: v((10:1) 1c (8:1)) */
	    /* Note the off-center origin.
	     * This is the only one of the 1-D strels
	     * with an origin not in the center.
	     * We manually correct this below.
	     */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 19, &strel));
	    strel->cy = 10; /* Offset the origin */
	    break;
    
	case strel20h:
	    /* horizontal strel of length 20: ((9:1) 1c (10:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 20, 1, &strel));
	    break;
    
	case strel20v:
	    /* vertical strel of length 20: v((9:1) 1c (10:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 20, &strel));
	    break;
    
	case strel25h:
	    /* horizontal strel of length 25: ((12:1) 1c (12:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 25, 1, &strel));
	    break;
    
	case strel25v:
	    /* vertical strel of length 25: v((12:1) 1c (12:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 25, &strel));
	    break;
    
	case strel30h:
	    /* horizontal strel of length 30: ((14:1) 1c (15:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 30, 1, &strel));
	    break;
    
	case strel30v:
	    /* vertical strel of length 30: v((14:1) 1c (15:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 30, &strel));
	    break;
    
	case strel35h:
	    /* horizontal strel of length 35: ((17:1) 1c (17:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 35, 1, &strel));
	    break;
    
	case strel35v:
	    /* vertical strel of length 35: v((17:1) 1c (17:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 35, &strel));
	    break;
    
	case strel40h:
	    /* horizontal strel of length 40: ((19:1) 1c (20:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 40, 1, &strel));
	    break;
    
	case strel40v:
	    /* vertical strel of length 40: v((19:1) 1c (20:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 40, &strel));
	    break;
    
	case strel45h:
	    /* horizontal strel of length 45: ((22:1) 1c (22:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 45, 1, &strel));
	    break;
    
	case strel45v:
	    /* vertical strel of length 45: v((22:1) 1c (22:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 45, &strel));
	    break;
    
	case strel50h:
	    /* horizontal strel of length 50: ((24:1) 1c (25:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 50, 1, &strel));
	    break;
    
	case strel50v:
	    /* vertical strel of length 50: v((24:1) 1c (25:1))  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 50, &strel));
	    break;
    
	case strel1:
	    /* pattern:  1c  */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 1, 1, &strel));
	    break;
    
	case strel2:
	    /* pattern:  1c  1 
		     1   1      */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 2, 2, &strel));
	    break;
    
	case strel3:
	    /* pattern:  1   1   1 
		     1   1c  1
		     1   1   1   */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 3, 3, &strel));
	    break;
    
	case strel4:
	    /* pattern:  1   1   1   1 
		     1   1c  1   1
		     1   1   1   1
		     1   1   1   1    */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 4, 4, &strel));
	    break;
    
	case strel5:
	    /* pattern:  1   1   1   1   1
		     1   1   1   1   1
		     1   1   1c  1   1
		     1   1   1   1   1
		     1   1   1   1   1     */
	    ERR_PASS( w_makeSolidCenterStrel (pIPWState, 5, 5, &strel));
	    break;
    
	case strel2dp:
	    /* Diagonal positive strel - 1 pix wide 2 pix long in 2x2 array */
	    /* pattern:  0c  1
			 1   0      */

	    ERR_PASS( w_createStrel (pIPWState, 2, 2, &strel));
	    strel->cx = strel->cy = 0;
	    strel->data[0][1] = 1;
	    strel->data[1][0] = 1;
	    break;
    
	case strel2dm:
	    /* Diagonal negative strel - 1 pix wide 2 pix long in 2x2 array */
	    /* pattern:  1c  0
			 0   1      */ 
	    ERR_PASS( w_createStrel (pIPWState, 2, 2, &strel));
	    strel->cx = strel->cy = 0;
	    strel->data[0][0] = 1;
	    strel->data[1][1] = 1;
	    break;
    
	case s_pepper_1:
	    /* 1x1 pepper */
	    ERR_PASS( w_createStrel (pIPWState, 3, 3, &strel));
	    strel->cx = 1;
	    strel->cy = 1;
	    strel->data[0][0] = 2;
	    strel->data[0][1] = 2;
	    strel->data[0][2] = 2;
	    strel->data[1][0] = 2;
	    strel->data[1][1] = 1;
	    strel->data[1][2] = 2;
	    strel->data[2][0] = 2;
	    strel->data[2][1] = 2;
	    strel->data[2][2] = 2;
	    break;
    
	case s_pepper_2:
	    /* 2x2 pepper */
	    ERR_PASS( w_createStrel (pIPWState, 4, 4, &strel));
	    strel->cx = 1;
	    strel->cy = 1;
	    strel->data[0][0] = 2;
	    strel->data[0][1] = 2;
	    strel->data[0][2] = 2;
	    strel->data[0][3] = 2;
	    strel->data[1][0] = 2;
	    strel->data[1][3] = 2;
	    strel->data[2][0] = 2;
	    strel->data[2][3] = 2;
	    strel->data[3][0] = 2;
	    strel->data[3][1] = 2;
	    strel->data[3][2] = 2;
	    strel->data[3][3] = 2;
	    break;
    
	case s_pepper_3:
	    /* 3x3 pepper */
	    ERR_PASS( w_createStrel (pIPWState, 5, 5, &strel));
	    strel->cx = 2;
	    strel->cy = 2;
	    strel->data[0][0] = 2;
	    strel->data[0][1] = 2;
	    strel->data[0][2] = 2;
	    strel->data[0][3] = 2;
	    strel->data[0][4] = 2;
	    strel->data[1][0] = 2;
	    strel->data[1][4] = 2;
	    strel->data[2][0] = 2;
	    strel->data[2][4] = 2;
	    strel->data[3][0] = 2;
	    strel->data[3][4] = 2;
	    strel->data[4][0] = 2;
	    strel->data[4][1] = 2;
	    strel->data[4][2] = 2;
	    strel->data[4][3] = 2;
	    strel->data[4][4] = 2;
	    break;
    
	case s_pepper_4:
	    /* 4x4 pepper */
	    ERR_PASS( w_createStrel (pIPWState, 6, 6, &strel));
	    strel->cx = 2;
	    strel->cy = 2;
	    strel->data[0][0] = 2;
	    strel->data[0][1] = 2;
	    strel->data[0][2] = 2;
	    strel->data[0][3] = 2;
	    strel->data[0][4] = 2;
	    strel->data[0][5] = 2;
	    strel->data[1][0] = 2;
	    strel->data[1][5] = 2;
	    strel->data[2][0] = 2;
	    strel->data[2][5] = 2;
	    strel->data[3][0] = 2;
	    strel->data[3][5] = 2;
	    strel->data[4][0] = 2;
	    strel->data[4][5] = 2;
	    strel->data[5][0] = 2;
	    strel->data[5][1] = 2;
	    strel->data[5][2] = 2;
	    strel->data[5][3] = 2;
	    strel->data[5][4] = 2;
	    strel->data[5][5] = 2;
	    break;
    
	case s_salt_1:
	    /* 1x1 salt */
	    ERR_PASS( w_createStrel (pIPWState, 3, 3, &strel));
	    strel->cx = 1;
	    strel->cy = 1;
	    strel->data[0][0] = 1;
	    strel->data[0][1] = 1;
	    strel->data[0][2] = 1;
	    strel->data[1][0] = 1;
	    strel->data[1][1] = 2;
	    strel->data[1][2] = 1;
	    strel->data[2][0] = 1;
	    strel->data[2][1] = 1;
	    strel->data[2][2] = 1;
	    break;
    
	case s_salt_2:
	    /* 2x2 salt */
	    ERR_PASS( w_createStrel (pIPWState, 4, 4, &strel));
	    strel->cx = 1;
	    strel->cy = 1;
	    strel->data[0][0] = 1;
	    strel->data[0][1] = 1;
	    strel->data[0][2] = 1;
	    strel->data[0][3] = 1;
	    strel->data[1][0] = 1;
	    strel->data[1][3] = 1;
	    strel->data[2][0] = 1;
	    strel->data[2][3] = 1;
	    strel->data[3][0] = 1;
	    strel->data[3][1] = 1;
	    strel->data[3][2] = 1;
	    strel->data[3][3] = 1;
	    break;
    
	case s_salt_3:
	    /* 3x3 salt */
	    ERR_PASS( w_createStrel (pIPWState, 5, 5, &strel));
	    strel->cx = 2;
	    strel->cy = 2;
	    strel->data[0][0] = 1;
	    strel->data[0][1] = 1;
	    strel->data[0][2] = 1;
	    strel->data[0][3] = 1;
	    strel->data[0][4] = 1;
	    strel->data[1][0] = 1;
	    strel->data[1][4] = 1;
	    strel->data[2][0] = 1;
	    strel->data[2][4] = 1;
	    strel->data[3][0] = 1;
	    strel->data[3][4] = 1;
	    strel->data[4][0] = 1;
	    strel->data[4][1] = 1;
	    strel->data[4][2] = 1;
	    strel->data[4][3] = 1;
	    strel->data[4][4] = 1;
	    break;
    
    
    
    
    /*********************************************************************
    *****************          Spread Replication        *****************
    *********************************************************************/
	case s_repl3:
	    /* Spread 3 */
	    ERR_PASS( w_createStrel (pIPWState, 5, 5, &strel));
	    strel->cx = strel->cy = 2;
	    strel->data[0][0] = 1;
	    strel->data[0][2] = 1;
	    strel->data[0][4] = 1;
	    strel->data[2][0] = 1;
	    strel->data[2][2] = 1;
	    strel->data[2][4] = 1;
	    strel->data[4][0] = 1;
	    strel->data[4][2] = 1;
	    strel->data[4][4] = 1;
	    break;
    
	case s_repl4:
	    /* Spread 4 */
	    ERR_PASS( w_createStrel (pIPWState, 7, 7, &strel));
	    strel->cx = strel->cy = 3;
	    strel->data[0][3] = 1;
	    strel->data[1][1] = 1;
	    strel->data[1][2] = 1;
	    strel->data[1][3] = 1;
	    strel->data[1][4] = 1;
	    strel->data[2][1] = 1;
	    strel->data[2][6] = 1;
	    strel->data[3][0] = 1;
	    strel->data[3][1] = 1;
	    strel->data[3][3] = 1;
	    strel->data[3][6] = 1;
	    strel->data[4][0] = 1;
	    strel->data[4][1] = 1;
	    strel->data[4][6] = 1;
	    strel->data[6][3] = 1;
	    strel->data[6][4] = 1;
	    break;
    
	case s_rulew9h7:
    /********************************************************************
    *****************          Rule Detection           *****************
    *********************************************************************/
    
	/*    strel:  matches horizontal rules of width at least 9
	 *             and height (thickness) up to 7.
	 *         Used for removal of rules at 75/inch
	 *
	 *    strel
	 *    9 9 4 4 2
	 *    2 0 0 0 2 0 0 0 2
	 *    0 0 0 0 0 0 0 0 0
	 *    0 0 0 0 0 0 0 0 0
	 *    0 0 0 0 0 0 0 0 0
	 *    1 0 1 0 1 0 1 0 1
	 *    0 0 0 0 0 0 0 0 0
	 *    0 0 0 0 0 0 0 0 0
	 *    0 0 0 0 0 0 0 0 0
	 *    2 0 0 0 2 0 0 0 2
	 */
	    ERR_PASS( w_createStrel (pIPWState, 9, 9, &strel));
	    strel->cx = strel->cy = 4;
	    strel->data[0][0] = 2;
	    strel->data[0][4] = 2;
	    strel->data[0][8] = 2;
	    strel->data[4][0] = 1;
	    strel->data[4][2] = 1;
	    strel->data[4][4] = 1;
	    strel->data[4][6] = 1;
	    strel->data[4][8] = 1;
	    strel->data[8][0] = 2;
	    strel->data[8][4] = 2;
	    strel->data[8][8] = 2;
	    break;
    
    
	case s_ruleh9w7:
	/*    strel:  matches vertical rules of height at least 10
	 *         and width (thickness) up to 7.
	 *         Used for removal of rules at 75/inch
	 *
	 *    strel
	 *    9 9 4 4 2
	 *    2 0 0 0 1 0 0 0 2
	 *    0 0 0 0 0 0 0 0 0
	 *    0 0 0 0 1 0 0 0 0
	 *    0 0 0 0 0 0 0 0 0
	 *    2 0 0 0 1 0 0 0 2
	 *    0 0 0 0 0 0 0 0 0
	 *    0 0 0 0 1 0 0 0 0
	 *    0 0 0 0 0 0 0 0 0
	 *    2 0 0 0 1 0 0 0 2
	 */
	    ERR_PASS( w_createStrel (pIPWState, 9, 9, &strel));
	    strel->cx = strel->cy = 4;
	    strel->data[0][0] = 2;
	    strel->data[0][4] = 1;
	    strel->data[0][8] = 2;
	    strel->data[2][4] = 1;
	    strel->data[4][0] = 2;
	    strel->data[4][4] = 1;
	    strel->data[4][8] = 2; 
	    strel->data[6][4] = 1; 
	    strel->data[8][0] = 2;
	    strel->data[8][4] = 1;
	    strel->data[8][8] = 2;
	    break;
    
	case strelF40h:
	/* FAST horizontal strel of length 40: [6:(100) 0 1c 0 6:(100) 1] */
	    ERR_PASS( w_createStrel (pIPWState, 40, 1, &strel));
	    strel->cx = 19;  /* near center pixel */
	    strel->cy = 0;
	    for (i = 0; i < 40; i = i + 3)
		strel->data[0][i] = 1;
	    break;
    
	case strelF40v:
	/* FAST vertical strel of length 40: v:[6:(100) 0 1c 0 6:(100) 1] */
	    ERR_PASS( w_createStrel (pIPWState, 1, 40, &strel));
	    strel->cx = 0;
	    strel->cy = 19;  /* near center pixel */
	    for (i = 0; i < 40; i = i + 3)
		strel->data[i][0] = 1;
	    break;
    
    
	case s_4_2_t:
	    /* pattern match for strel operator:
		0   2   0
		2   0c  0
		0   1   1      */
	    ERR_PASS( w_createStrel (pIPWState, 3, 3, &strel));
	    strel->cx = 1;  /* center pixel in center */
	    strel->cy = 1;
	    strel->data[0][1] = 2;
	    strel->data[1][0] = 2;
	    strel->data[2][1] = 1;
	    strel->data[2][2] = 1;
	    break;
    
    
	case strelItal1:
	    /* pattern:  0  0  0  2  0  1
		 (2) 0  0  0  0  0  0
		     0  0  2  0  1  0
		     0  0  0  0  0  0
		     0  0  0  0c 0  0
		     0  2  0  1  0  0
		 (2) 0  0  0  0  0  0
		     2  0  1  0  0  0  */
	    ERR_PASS( w_createStrel (pIPWState, 6, 10, &strel));
	    strel->cx = 3;
	    strel->cy = 5;
	    strel->data[0][3] = 2;
	    strel->data[0][5] = 1;
	    strel->data[3][2] = 2;
	    strel->data[3][4] = 1;
	    strel->data[6][1] = 2;
	    strel->data[6][3] = 1;
	    strel->data[9][0] = 2;
	    strel->data[9][2] = 1;
	    break;
    
    
	case strelHTR:
	    /* anti-alias pattern match for horizontal, top, right:
		2   2   2   2   2
		1   1   1c  1   2    */
	    ERR_PASS( w_createStrel (pIPWState, 5, 2, &strel));
	    strel->cx = 2;
	    strel->cy = 1;
	    strel->data[0][0] = 2;
	    strel->data[0][1] = 2;
	    strel->data[0][2] = 2;
	    strel->data[0][3] = 2;
	    strel->data[0][4] = 2;
	    strel->data[1][0] = 1;
	    strel->data[1][1] = 1;
	    strel->data[1][2] = 1;
	    strel->data[1][3] = 1;
	    strel->data[1][4] = 2;
	    break;
    
    
	case strelVLT:
	    /* anti-alias pattern match for vertical, left, top:
		2   2
		2   1c
		2   1
		2   1
		2   1  */
	    ERR_PASS( w_createStrel (pIPWState, 2, 5, &strel));
	    strel->cx = 1;
	    strel->cy = 1;
	    strel->data[0][0] = 2;
	    strel->data[0][1] = 2;
	    strel->data[1][0] = 2;
	    strel->data[1][1] = 1;
	    strel->data[2][0] = 2;
	    strel->data[2][1] = 1;
	    strel->data[3][0] = 2;
	    strel->data[3][1] = 1;
	    strel->data[4][0] = 2;
	    strel->data[4][1] = 1;
	    break;
    
    
	case s_v21_right:
	    /* registration for left edge of line or bar */
	    ERR_PASS( w_createStrel (pIPWState, 5, 21, &strel));
	    strel->cx = 4;
	    strel->cy = 10;
	    strel->data[0][0] = 1;
	    strel->data[0][4] = 2;
	    strel->data[2][0] = 1;
	    strel->data[4][0] = 1;
	    strel->data[5][4] = 2;
	    strel->data[6][0] = 1;
	    strel->data[8][0] = 1;
	    strel->data[10][0] = 1;
	    strel->data[10][4] = 2;
	    strel->data[12][0] = 1;
	    strel->data[14][0] = 1;
	    strel->data[15][4] = 2;
	    strel->data[16][0] = 1;
	    strel->data[18][0] = 1;
	    strel->data[20][0] = 1;
	    strel->data[20][4] = 2;
	    break;
    
	case emptyStrel:
	    /* tests to see if we handle the case of a strel filled with
	     * don't cares correctly.
	     */
	    ERR_PASS( w_createStrel (pIPWState, 2, 2, &strel));
	    strel->cx = 0;
	    strel->cy = 0;
	    /* leave the data filled with don't-cares. */
	    break;
    
	default:
	    ERR_GEN (1, cErrInvalidParm);
	    break;

    }    /* switch (strelID) */

    /*
     * Test if the strel create failed. This is redundant (but
     * harmless) for some of the strels, but it avoids some duplicate
     * code by having all the switch statements break to here.  
     */
    if (strel == NULL) 
    {
        sinfo->strelArray[strelID].strel = NULL;
	ERR_GEN (1, cErrNomem);
    }

    /* Load the name (ID) of the strel into the strel's ID field */
    strel->ID = strelID;

    /*
     * Add the strel to the global list and the user list
     * and increment the reference count on the global list.
     */
    gSinfo->strelArray[strelID].refCount++;
    sinfo->strelArray[strelID].refCount++;
    gSinfo->strelArray[strelID].strel = strel; 
    sinfo->strelArray[strelID].strel = strel; 

    NORMAL_RETURN;

}  /* w_initStrelByName () */


/*
 * $Log: strlinit.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.18  1996/01/04 20:27:19  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.17  1995/12/13  04:13:48  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.16  1995/11/13  20:16:45  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.15  1995/10/13  22:05:46  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/09/29  21:12:55  dferrell
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.13  1995/09/21  23:00:12  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/06  20:23:57  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  18:50:21  dferrell
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.10  1995/08/31  21:39:39  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/30  17:28:28  ejaquith
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Made routines accessing pIPWState first check that it is valid (not NULL).
 * If it's null, return error: cErrNullIPWState.
 *
 * Revision 1.8  1995/08/22  18:02:07  dferrell
 * dferrell on Tue Aug 22 11:01:50 PDT 1995
 *
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Set strel pointer to NULL at start of initStrelByName ()
 *
 * Revision 1.7  1995/08/17  21:57:06  dferrell
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.6  1995/01/26  22:37:18  lperry
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Overwrite the ID field of each strel initialized
 * by initStrelByName with its strelID. This will
 * allow functions using a strel to know what strel
 * (specifically) it is.
 *
 * A potential use of this ID is to allow an ipwrap
 * morphological routine to implement the operation
 * on a special case basis. For example, large
 * strel dilates could be decomposed into a sequence
 * of two or more much smaller dilates. Square strels
 * could be implemented as separable H and V strels, etc.
 *
 * Revision 1.5  1995/01/24  22:22:08  lperry
 * lperry on Tue Jan 24 14:18:11 PST 1995
 *
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * The API is basically the same, but a new call,
 * releaseStrelByName has also been added.
 * Named strels are no longer referred to as namedStrel5h,
 * but as simply strel5h, which is a member of an enumerated
 * type used to index the array of strels.
 *
 * Revision 1.4  1994/12/08  04:43:15  ddavies
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 *  Add initialization for the emptyStrel.  This is used to test our ability
 * to handle a strel containing only don't-care terms.
 *
 * Revision 1.3  1994/09/29  21:19:36  lperry
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Change to use strel.pub instead of strlpros.h
 *
 * Revision 1.2  1994/09/29  00:48:21  lperry
 * strlinit.c -> /product/ipcore/ipwrap/src/RCS/strlinit.c,v
 *
 * Eliminate use of utilpros. Use utils.pub instead.
 *
 * Revision 1.1  1994/08/10  22:22:17  lperry
 * Initial revision
 *
*/

