/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  subsmpl1.c: contains subroutines for fast horizontally
 *              pre-filtered subsampled reduction.
 *
 *                  w_reduceH ()
 *
 *     Present timings:  on 4/330 (-O optimizer):
 *                 0.07 sec/Mbit for reduction (level 1,4).
 *                 0.06 sec/target-Mbit for 2x expansion.
 *                 0.05 sec/Mbit for 2x subsampled reduction (reduceH).
 *                 0.02 sec/Mbit for 4x subsampled reduction (reduceH).
 *                 0.01 sec/Mbit for 8x subsampled reduction (reduceH).
 *                 0.002 sec/Mbit for 16x subsampled reduction
 */




#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "logicops.h"
#include "logicops.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "subsmpl.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: subsmpl1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Horizontally filtered, thresholded, subsampled reduction by 2, 4, 8, 16, or 32.
 * LongDescription:::
 *  Horizontally reduction of binary PIXR using logical OR or AND to
 *  determine output pixel value.  The destination PIXR is created if
 *  necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Binary source PIXR
 *  factor::         Reduction factor (2, 4, 8, 16, or 32)
 *  type::           AND or OR
 *
 * Output:::
 *  prd::            Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, anamorphic
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceH (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    Int32               type,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD,
                 bplS, bplD;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    ScaledInt1616	 xRes, yRes;
    UInt8        *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor !=8 &&
              factor != 16 && factor != 32), cErrInvalidParm);
    ERR_GEN ((type != OR && type != AND), cErrInvalidParm);

    /* make the dest image. */
    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    wD = wS / factor;
    hD = hS / factor;

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wD, hD, 1, 
					      xRes / factor, yRes / factor,
					      cFrameBits, prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    if (type == OR)
        ERR_PASS_CLEANUP (i_reduceHOr (pIPWState, iD, wD, hD,
				       bplD, iS, bplS, factor));
    else
        ERR_PASS_CLEANUP (i_reduceHAnd (pIPWState, iD, wD, hD,
				       bplD, iS, bplS, factor));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduceH () */



/*
 *
 * $Log: subsmpl1.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.24  1998/01/14 00:15:36  ddavies
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.23  1997/04/15 22:42:07  ddavies
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.22  1996/01/04 20:27:22  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.21  1995/12/13  04:13:51  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/13  20:16:50  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:05:50  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/09/26  22:09:07  dferrell
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.17  1995/09/21  23:00:15  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/19  21:12:50  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.15  1995/09/18  21:43:48  dferrell
 * dferrell on Mon Sep 18 14:42:33 PDT 1995
 *
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Made the pixr creation/checking code in w_reduceH reduce BOTH
 * dimensions of size and resolution.
 *
 * Revision 1.14  1995/09/15  22:38:46  ddavies
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.13  1995/09/15  00:35:39  dferrell
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.12  1995/09/11  17:54:24  dferrell
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.11  1995/09/07  01:35:49  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.10  1995/09/06  20:24:00  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  18:19:58  dferrell
 * dferrell on Fri Sep 1 11:19:52 PDT 1995
 *
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.8  1995/08/31  21:39:42  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/24  02:57:07  ejaquith
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * reduceH() and reduceHPixr() reduce the image in BOTH dimensions by the
 * passed in factor, and they were only reducing the width and x DPI and not
 * the height and y DPI.
 *
 * Revision 1.6  1995/08/17  21:57:47  dferrell
 * dferrell on Thu Aug 17 14:57:40 PDT 1995
 *
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.5  1995/02/03  01:44:55  ddavies
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.4  1994/10/03  20:36:03  lperry
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Changed to use subsmpl.pub instead of scalepros.h
 *
 * Revision 1.3  1993/12/03  01:19:30  ddavies
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:17:56  ddavies
 * subsmpl1.c -> /product/ipcore/ipwrap/src/RCS/subsmpl1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

