/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  subsmpl2.c: contains subroutines for fast vertically
 *              pre-filtered subsampled reduction.
 *
 *                  w_reduceV () 
 *                  w_reduce2D () 
 *
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "logicops.h"
#include "logicops.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"
#include "reduce.pub"

#include "subsmpl.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: subsmpl2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Vertically filtered, thresholded, subsampled reduction by 2, 4, 8, 16, or 32
 *
 * LongDescription:::
 *  Vertically (OR or AND) filtered, thresholded, subsampled reduction by
 *  2, 4, 8, 16, or 32.  The destination PIXR is created if necessary.

 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Binary source PIXR
 *  factor::         Vertical reduction factor
 *  type::           AND or OR logical filtering
 *
 * Output:::
 *  prd::            Binary destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, anamorphic, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceV (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    Int32               type,
    PIXR              **prd)
{
    PROLOGUE;
    Int32        wS, hS, wD, hD,
                 bplS, bplD;
    Bool         destCreated = cFalse;
    Int32        rfc, bfc;
    ScaledInt1616	 xRes, yRes;
    UInt8        *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor !=8 &&
              factor != 16 && factor != 32), cErrInvalidParm);
    ERR_GEN ((type != OR && type != AND), cErrInvalidParm);

    /* make the dest image. */
    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    wD = wS / factor;
    hD = hS / factor;

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wD, hD, 1, 
					      xRes / factor, yRes / factor,
					      cFrameBits, prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    if (type == OR)
        ERR_PASS_CLEANUP (i_reduceVOr (pIPWState, iD, wD, hD,
				       bplD, iS, bplS, factor));
    else
        ERR_PASS_CLEANUP (i_reduceVAnd (pIPWState, iD, wD, hD,
					bplD, iS, bplS, factor));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* reduceVPixr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Binary isomorphic reduction with independent AND/OR filtering.
 *
 * LongDescription:::
 *  Binary isomorphic reduction with independent vertical and
 *  horizontal logical filtering (AND/OR).  Reduction factor must
 *  be 2, 4, 8, 16, or 32.  Destination PIXR created if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Binary source PIXR
 *  factor::         Reduction factor (2, 4, 8, 16, or 32)
 *  verticalType::   Vertical reduction filter (AND/OR)
 *  horizontalType:: Horizontal reduction filter (AND/OR)
 *
 * Output:::
 *  prd::            Binary destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  jcraig
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduce2D (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    Int32               verticalType,
    Int32               horizontalType,
    PIXR              **prd)
{
    PROLOGUE;
    Int32         wS, hS, wD, hD;
    Int32         bplS, bplD;
    Bool          destCreated = cFalse;
    Int32         rfc, bfc;
    ScaledInt1616	  xRes, yRes;
    UInt8        *iS, *iD;
    PIXR	 *prt = NULL;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor !=8 &&
              factor != 16 && factor != 32), cErrInvalidParm);
    ERR_GEN (((horizontalType != OR) && (horizontalType != AND)), 
               cErrInvalidParm);
    ERR_GEN (((verticalType != OR) && (verticalType != AND)), 
               cErrInvalidParm);

    /* make the dest image. */
    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    wD = wS / factor;
    hD = hS / factor;

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wD, hD, 1, 
					      xRes / factor, yRes / factor,
					      cFrameBits, prd, &rfc, &bfc, 
					      &destCreated));

    /* Do the correct reduction sequence */
    if (verticalType == OR)
    {
	if (horizontalType == OR)
	{
	    switch (factor)
	    {
		case 2:
		    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prs, 1, prd));
		    break;

		case 4:
		    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prs, 1, 1, prd));
		    break;

		case 8:
		    ERR_PASS_CLEANUP (w_reduce8 (pIPWState, prs, 1, 1, 1, prd));
		    break;

		case 16:
		    ERR_PASS_CLEANUP (w_reduce16 (pIPWState, prs, 1, 1, 1, 1, 
		                                  prd));
		    break;

		case 32:
		    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prs, 
		                                              &xRes, &yRes));
		    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, wS / 16, 
							  hS / 16, 1,
							  xRes / 16, yRes / 16,
							  cFrameBits, &prt));
		    ERR_PASS_CLEANUP (w_reduce16 (pIPWState, prs, 1, 1, 1, 1, 
		                                  &prt));
		    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prt, 1, prd));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prt));
		    break;

		default:
		    ERR_GEN_CLEANUP (1, cErrInvalidParm);
		    break;
	    }
	}
	else	/* horizontalType == AND */
	{
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

            ERR_PASS_CLEANUP (i_reduceVOrHAnd (pIPWState, iD, wD, hD, bplD, 
	                                       iS, bplS, factor));
	}
    }
    else	/* verticalType == AND */
    {
	if (horizontalType == OR)
	{
	    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
	    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
	    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
	    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

            ERR_PASS_CLEANUP (i_reduceVAndHOr (pIPWState, iD, wD, hD, bplD, 
	                                       iS, bplS, factor));
	}
	else	/* horizontalType == AND */
	{
 	    switch (factor)
	    {
		case 2:
		    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prs, 4, prd));
		    break;

		case 4:
		    ERR_PASS_CLEANUP (w_reduce4 (pIPWState, prs, 4, 4, prd));
		    break;

		case 8:
		    ERR_PASS_CLEANUP (w_reduce8 (pIPWState, prs, 4, 4, 4, prd));
		    break;

		case 16:
		    ERR_PASS_CLEANUP (w_reduce16 (pIPWState, prs, 4, 4, 4, 4, 
		                                  prd));
		    break;

		case 32:
		    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prs, 
		                                              &xRes, &yRes));
		    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, wS / 16, 
							  hS / 16, 1,
							  xRes / 16, yRes / 16,
							  cFrameBits, &prt));
		    ERR_PASS_CLEANUP (w_reduce16 (pIPWState, prs, 4, 4, 4, 4, 
		                                  &prt));

		    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, prt, 1, prd));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prt));
		    break;

		default:
		    ERR_GEN_CLEANUP (1, cErrInvalidParm);
		    break;
	    }
	}
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &prt);
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduce2D () */



/*
 *
 * $Log: subsmpl2.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.25  1998/01/14 00:15:38  ddavies
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.24  1997/04/15 22:42:24  ddavies
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.23  1996/01/04 20:27:25  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.22  1995/12/13  04:13:54  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/11/13  20:16:54  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.20  1995/11/10  02:04:02  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.19  1995/10/13  22:05:52  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/09/26  22:09:14  dferrell
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.17  1995/09/21  23:00:18  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/19  21:12:59  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () calls
 *
 * Revision 1.15  1995/09/18  21:44:16  dferrell
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Made the pixr creation/checking code in w_reduceV reduceBOTH
 * dimensions of size and resolution.
 *
 * Revision 1.14  1995/09/15  22:38:52  ddavies
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.13  1995/09/15  00:35:52  dferrell
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 *
 * Revision 1.12  1995/09/11  17:54:33  dferrell
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.11  1995/09/07  01:35:52  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.10  1995/09/06  20:24:02  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  18:20:04  dferrell
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.8  1995/08/31  21:39:45  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/24  02:58:55  ejaquith
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * reduceV() and reduceVPixr() reduce the image in BOTH dimensions by the
 * passed in factor, and they were only reducing the reducing the r
 * (oh great, i love this "editor") height and y DPI and not the width and
 * x DPI.
 *
 * Revision 1.6  1995/08/17  21:57:54  dferrell
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.5  1995/02/03  01:45:46  ddavies
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.4  1994/10/03  20:36:12  lperry
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Changed to use subsmpl.pub instead of scalepros.h
 *
 * Revision 1.3  1993/12/03  01:19:35  ddavies
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:18:02  ddavies
 * subsmpl2.c -> /product/ipcore/ipwrap/src/RCS/subsmpl2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

