/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  subsmpl3.c: contains subroutines for fast subsampled reduction,
 *              with no pre-filtering
 *
 *                  w_reduceS ()
 *
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "subsmpl.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: subsmpl3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/*********************************************************************
 ************       Unfiltered Subsampled Reduction     **************
 ********************************************************************/
/*
 *  reduceSPixr():  Simple subsampled reduction by 1.5 (represented by "1")
 *			2, 4, 8, 16, or 32.
 *                  Takes an input Pixr and a reduction factor n.
 *                  Samples one pixel in the nxn region
 *		    NOTE!! 1.5x reduction only works for 4bpp and 8bpp
 *		    images.  IT DOES NOT WORK ON BINARY IMAGES!
 *		    Use scaleToSizePixr for non-power-of-2 scaling
 *		    of binary images.
 *
 *                  For 4bpp and 8bpp gray, extracts appropriate pixel.
 *		    For color, operates on each plane independently.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Subsampled reduction of 1, 4, or, 8-bit PIXRs by 1.5, 2, 4, 8, 16 or 32.
 *
 * LongDescription:::
 *  Simple subsampling reduction.  For binary PIXRs only the powers of
 *  two factors are valid.  A factor of 1 indicates a reduction by 1.5
 *  for 4 and 8 bit gray PIXRs only.  The destination PIXR is created if
 *  necessary.  Multi-channel PIXRs are handled correctly.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Source PIXR
 *  factor::         reduction factor (1=1.5, 2, 4, 8, 16, 32)
 *
 * Output:::
 *  prd::            Destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, isomorphic, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceS (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    PIXR              **prd)
{
    PROLOGUE;
    Int32         wS, hS, wD, hD, frame;
    Int32         bplS, bplD,
                  dS, dD, i, 
		  channelsS, channelsD;
    Bool	  destCreated = cFalse;
    Int32         rfc, bfc;
    UInt8        *iS, *iD;
    PIXR         *subS, *subD;
    ColorType     type;
    ColorMap     *map;
    ScaledInt1616	 xRes, yRes;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
                  cChannels1 | cChannels3, cFrame32);
    if (*prd != NULL)
    ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1 | cDepth4 | cDepth8, 
                  cChannels1 | cChannels3, cFrame32);

    ERR_GEN ((factor != 1 && factor != 2 && factor != 4 && factor !=8 &&
              factor != 16 && factor != 32), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetDepth (pIPWState, prs, &dS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prs, &channelsS));
    ERR_PASS (w_pixrGetFrame (pIPWState, prs, &frame));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prs, &map));
    ERR_PASS (w_pixrGetType (pIPWState, prs, &type));

    ERR_GEN (((factor == 1) && (dS == 1)), cErrInvalidParm);

    if (factor == 1)
    {
	/* really means 1.5 */
	wD = (wS * 2) / 3;
	hD = (hS * 2) / 3;
	xRes = (xRes * 2) / 3;
	yRes = (yRes * 2) / 3;
    }
    else
    {
	wD = wS / factor;
	hD = hS / factor;
	xRes = xRes / factor;
	yRes = yRes / factor;
    }

    dD = dS;

    ERR_PASS (wp_createPixrExtendedConditionalNoInit (pIPWState, wD, hD, dD, 
						      xRes, yRes, cFrameBits,
						      channelsS, type, map,
						      prd, &rfc, &bfc,
						      &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prs, &channelsS));
    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, *prd, &channelsD));

    ERR_GEN_CLEANUP ((channelsS != channelsD), cErrInvalidParm);

    if (channelsS != 1)
    {
	for (i = 0; i < channelsS; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prs, i, &subS));
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, *prd, i, &subD));
	    ERR_PASS_CLEANUP (w_reduceS (pIPWState, subS, factor, &subD));
	}
    }
    else
    {
	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prs, &wS));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prs, &hS));
	ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
	ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

        switch (dS)
	{
	    case 1:
		ERR_PASS_CLEANUP (i_subsample1bpp (pIPWState, iD, wD, hD, bplD, 
		                                   iS, bplS, factor));
	        break;

	    case 4:
		ERR_PASS_CLEANUP (i_subsample4bpp (pIPWState, iD, wD, hD, bplD, 
		                                   iS, bplS, factor));
	        break;

	    case 8:
		ERR_PASS_CLEANUP (i_subsample8bpp (pIPWState, iD, wD, hD, bplD, 
		                                   iS, bplS, factor));
	        break;

	    default:
	        ERR_GEN_CLEANUP (1, cErrInvalidParm);
		break;
	}
    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* w_reduceS () */



/*
 *
 * $Log: subsmpl3.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.26  1998/01/14 00:15:39  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.25  1997/04/15 22:42:38  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Use non-initialized PIXR creation where possible.
 *
 * Revision 1.24  1996/01/04 20:27:27  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.23  1995/12/13  04:13:57  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:16:59  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/13  22:05:55  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.20  1995/09/26  22:09:20  dferrell
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.19  1995/09/21  23:00:21  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.18  1995/09/19  21:13:08  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () call
 *
 * Revision 1.17  1995/09/15  22:38:57  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.16  1995/09/15  00:36:09  dferrell
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.15  1995/09/11  17:54:43  dferrell
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.14  1995/09/07  01:35:54  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:24:06  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  18:20:10  dferrell
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.11  1995/08/31  21:39:47  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/24  03:00:26  ejaquith
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * reduceSPixr(), reduceS(), and reduceSGray() are supposed to handle
 * 3 channel images, so added that as allowable type in ERR_GEN_PIXR().
 *
 * Revision 1.9  1995/08/23  01:19:08  jcraig
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * in reduceSPixr, use the newly calculated width and height when
 * creating the new destination image (not the source width and height).
 *
 * Revision 1.8  1995/08/17  21:58:02  dferrell
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.7  1995/04/01  01:45:07  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Change comments of reduceSPixr to say that this routine does not handle
 * 1.5x reduction of binary images.  Add a check to detect this case and
 * return an error if it appears.
 *
 * Revision 1.6  1995/02/03  01:45:57  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.5  1994/10/03  20:36:17  lperry
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Changed to use subsmpl.pub instead of scalepros.h
 *
 * Revision 1.4  1994/04/12  01:16:53  macleod
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.3  1993/12/03  01:19:40  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:18:09  ddavies
 * subsmpl3.c -> /product/ipcore/ipwrap/src/RCS/subsmpl3.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *
 *
 */

