/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  subsmpl5.c: contains subroutines for fast vertical or horizontal
 *              reduction.
 *
 *                  w_reduceRows ()
 *                  w_reduceCols ()
 *
 */



#include "ipwcore.prv"

#include "defines.pub"     /* for cTrue and cFalse */
#include "pixr.h"
#include "frames.pub"
#include "geomadt.h"
#include "props.pub"
#include "logicops.h"
#include "logicops.pub"
#include "shrscgr.pub"     /* for FilterGen in shrscal.pub */
#include "shrscal.pub"

#include "subsmpl.pub"



IP_RCSINFO(RCSInfo, "$RCSfile: subsmpl5.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Horizontal binary reduction using AND, OR, or subsample filtering.
 *
 * LongDescription:::
 *  Reduces rows by 2, 4, 8, 16, or 32 using the specified fitler type.
 *  Creates the destination PIXR if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Binary source PIXR
 *  factor::         Horizontal reduction factor (2, 4, 8, 16, 32)
 *  type::           Filter type (AND, OR, SAMPLE)
 *
 * Output:::
 *  prd::            Binary destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, anamorphic, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceRows (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    Int32               type,
    PIXR              **prd)
{
    PROLOGUE;
    Int32	  wS, hS, wD, hD;
    Bool          destCreated = cFalse;
    Int32         bplS, bplD;
    Int32        rfc, bfc;
    ScaledInt1616	 xRes, yRes;
    UInt8        *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor != 8 &&
              factor != 16 && factor != 32), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    wD = wS;
    hD = hS / factor;

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wD, hD, 1, 
					      xRes, yRes / factor,
					      cFrameBits, prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    switch (type)
    {
        case OR:
            ERR_PASS_CLEANUP (i_reduceRowsOr (pIPWState, iD, wD, hD, bplD, 
	                                      iS, bplS, factor));
	    break;

        case AND:
            ERR_PASS_CLEANUP (i_reduceRowsAnd (pIPWState, iD, wD, hD, bplD, 
	                                       iS, bplS, factor));
	    break;

        case SAMPLE:
            ERR_PASS_CLEANUP (i_reduceRowsSample (pIPWState, iD, wD, hD, bplD, 
	                                          iS, bplS, factor));
	    break;

	default:
	    ERR_GEN_CLEANUP (1, cErrInvalidParm);
	    break;

    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* reduceRowsPixr () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Vertical binary reduction using AND, OR, or subsample filtering.
 *
 * LongDescription:::
 *  Reduces columns by 2, 4, 8, 16, or 32 using the specified fitler type.
 *  Creates the destination PIXR if necessary.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prs::            Binary source PIXR
 *  factor::         Vertical reduction factor (2, 4, 8, 16, 32)
 *  type::           Filter type (AND, OR, SAMPLE)
 *
 * Output:::
 *  prd::            Binary destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  scaling
 *
 * Keywords:::
 *  scaling, anamorphic, binary
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_reduceCols (
    IPWState           *pIPWState,
    PIXR               *prs,
    Int32               factor,
    Int32               type,
    PIXR              **prd)
{
    PROLOGUE;
    Int32	  wS, hS, wD, hD;
    Bool          destCreated = cFalse;
    Int32         bplS, bplD;
    Int32        rfc, bfc;
    ScaledInt1616	 xRes, yRes;
    UInt8        *iS, *iD;

    /* pixrs OK? */
    ERR_GEN_PIXR (prs, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prd != NULL)
	ERR_GEN_PIXR (*prd, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_GEN ((factor != 2 && factor != 4 && factor != 8 &&
              factor != 16 && factor != 32), cErrInvalidParm);

    ERR_PASS (w_pixrGetWidth (pIPWState, prs, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prs, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prs, &xRes, &yRes));

    wD = wS / factor;
    hD = hS;

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState, wD, hD, 1, 
					      xRes / factor, yRes,
					      cFrameBits, prd, &rfc, &bfc, 
					      &destCreated));

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *prd, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *prd, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prs, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prd, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prs, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prd, &iD));

    switch (type)
    {
        case OR:
            ERR_PASS_CLEANUP (i_reduceColsOr (pIPWState, iD, wD, hD, bplD, 
	                                      iS, bplS, factor));
	    break;

        case AND:
            ERR_PASS_CLEANUP (i_reduceColsAnd (pIPWState, iD, wD, hD, bplD, 
	                                       iS, bplS, factor));
	    break;

        case SAMPLE:
            ERR_PASS_CLEANUP (i_reduceColsSample (pIPWState, iD, wD, hD, bplD, 
	                                          iS, bplS, factor));
	    break;

	default:
	    ERR_GEN_CLEANUP (1, cErrInvalidParm);
	    break;

    }

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prd, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
	w_destroyPixr (pIPWState, prd);
    ERR_RETURN;

}  /* reduceColsPixr () */



/*

 $Log: subsmpl5.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.25  1998/01/14 00:15:41  ddavies
 subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.24  1997/04/15 22:42:51  ddavies
 subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.23  1996/01/04 20:27:30  ejaquith
 subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.22  1995/12/15  00:53:45  jcraig
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * in w_reduceCols, only the number of columns is being reduced do only
 * the x resolution should be reduced not both the x and y resolution.
 *
 * Revision 1.21  1995/12/13  04:13:59  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.20  1995/11/15  20:10:46  dferrell
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Fixed function comment blocks.
 *
 * Revision 1.19  1995/11/13  20:17:03  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/13  22:05:58  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/09/26  22:09:29  dferrell
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.16  1995/09/21  23:00:25  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/19  21:13:18  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Added miss ERR_PASS_CLEANUP() wrapping to wp_clearRightBottomFrames () calls
 *
 * Revision 1.14  1995/09/18  18:45:27  ejaquith
 * ejaquith on Mon Sep 18 11:43:21 PDT 1995
 *
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * The calls to wp_createPixrConditional() in w_reduceCols() and w_reduceRows()
 * were passing in unitialized height and width parameters respectively,
 * and were causing seg faults.
 *
 * Revision 1.13  1995/09/15  22:39:01  ddavies
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.12  1995/09/15  00:44:59  dferrell
 * dferrell on Thu Sep 14 17:44:48 PDT 1995
 *
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Changed resolutions to scaled ints, updated functions to use new
 * conditional pixr creation and right/bottom frame clearing functionality.
 *
 * Revision 1.11  1995/09/11  17:54:52  dferrell
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.10  1995/09/07  01:35:57  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:24:09  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  18:20:15  dferrell
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.7  1995/08/31  21:39:49  ejaquith
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/17  21:58:10  dferrell
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.5  1995/02/03  01:46:08  ddavies
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Set the scaled x and y resolutions of PIXRs created here.  This preserves
 * any input resolutions through the scaling process.
 *
 * Revision 1.4  1994/10/03  20:36:30  lperry
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Changed to use subsmpl.pub instead of scalepros.h
 *
 * Revision 1.3  1993/12/03  01:19:44  ddavies
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:18:17  ddavies
 * subsmpl5.c -> /product/ipcore/ipwrap/src/RCS/subsmpl5.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

