/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  sunio.c--io between sun rasterfile and pixr
 *
 *  Note:  all routines are private (thus, pointers and PIXRs are not checked
 *  at this level)
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "pixr.h"
#include "imagerep.h"
#include "alplib.h"
#include "rastrfil.h"
#include "frames.pub"
#include "wmask.h"
#include "defines.pub"
#include "ptypes.h"
#include "gdefs.h"
#include "pcx.h"

/*prototypes*/
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "io.pub"
#include "io.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif



IP_RCSINFO(RCSInfo, "$RCSfile: sunio.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/* suppress all debugging; these should always be 0 */
#define  DEBUG_STUFF         0

    /*  globals  */
static Int32       totalBytes; /* bytes in file */

Int32		   cmprlen = 0; /* used by pr_dump */

#define  BLOCKSIZE           4096
#define  RAST_RUNFLAG        128
#define  BLOCKING            (32*1024)
#define  CMS_SIZE            128

    /* fseek parameters */
#define L_SET       0   /* absolute offset */
#define L_INCR      1   /* relative to current offset */
#define L_XTND      2   /* relative to end of file */

    /* static globals for rast2ScanLine() and byteFBCount() */
static UInt8            rast2SLValue = 0;     /* value from previous run */
static Int32            rast2SLCount = 0;     /* if > 0, last run not
					       *         finished yet */
static UInt8           *getRastBufPtr;
static UInt8            getRastBuf[BLOCKING + 1];



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read some SUNRAS bytes into a buffer.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream at location to read from.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_readBytesIntoBuffer (
    IPWState           *pIPWState,
    FILE               *fp)
{
    PROLOGUE;

    ERR_GEN(totalBytes <= 0, cErrIO_EarlyEndEncountered);

    if (totalBytes < BLOCKING) 
        fread( getRastBuf, totalBytes, 1, fp); 
    else
        fread( getRastBuf, BLOCKING, 1, fp);

    totalBytes -= BLOCKING;
    getRastBufPtr = &getRastBuf[0];

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write some SUNRAS bytes out of a buffer.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream at location to write to.
 *  buffer::         Ptr to output buffer.
 *  length::         Number of bytes to write.
 *  flush::          cTrue => flush buffer by writing out what remains.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::  Modified from sandpiper
 *
 * Citations:::
 *  taken from /import/sandpiper/src/bin/prim/writesunras/writesunras.c
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_writeByteFromBuffer (
    IPWState           *pIPWState,
    FILE               *fp,
    UInt8              *buffer,
    Int32               length,
    Int32               flush)
{
    PROLOGUE;
    static UInt8    outbuf[BLOCKSIZE+1];
    static Int32    count = -1;
    Int32           i;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    for (i = 0; i < length; i++)
    {
        outbuf[++count] = buffer[i];
        if (count == BLOCKSIZE)
        {
            fwrite( outbuf, BLOCKSIZE+1, 1, fp);
            count = -1;
        }
    }

    if ((flush == 1) && (count != -1))
    {
        fwrite( outbuf, count + 1, 1, fp);
        count = -1;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  makes SUN rasterfile header from PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pr::             Ptr to PIXR
 *
 * Output:::
 *  ppHeader::       Ptr to ptr to created header.
 *                   ppHeader will get allocated and filled here.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrMakeHeaderSUN (
    IPWState             *pIPWState,
    PIXR                 *pr,
    struct rasterfile   **ppHeader)
{
    PROLOGUE;
    Int32               depth, height, width;
    struct rasterfile  *pHeader = NULL;

    ERR_PASS( w_pixrGetDepth(pIPWState, pr, &depth));
    ERR_PASS( w_pixrGetHeight(pIPWState, pr, &height));
    ERR_PASS( w_pixrGetWidth(pIPWState, pr, &width));

    pHeader = (struct rasterfile *) IPW_CALLOC(1, sizeof(struct rasterfile),
						cAllocatorData);
    ERR_GEN (pHeader == NULL, cErrNomem);

    pHeader->ras_magic  = FLIP_INT( RAS_MAGIC);
    pHeader->ras_width  = FLIP_INT( width );
    pHeader->ras_height = FLIP_INT( height );
    pHeader->ras_depth  = FLIP_INT( depth);
    pHeader->ras_length = FLIP_INT( height * (((width * depth + 15) >> 4)
						 << 1 ));
    pHeader->ras_type   = FLIP_INT( RT_STANDARD);
        /* no color map */
    pHeader->ras_maptype   = FLIP_INT( RMT_NONE);
    pHeader->ras_maplength = FLIP_INT( 0);

    *ppHeader = pHeader;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  reads Sun rasterfile header from file stream.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to read from
 *  FileH::          Ptr to file header struct
 *
 * Output:::
 *  pr::             Ptr to ptr to returned PIXR with image that was read.
 *                   pr is allocated and filled in.
 *                   *pr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_readSunHeader (
    IPWState           *pIPWState,
    FILE               *fp,
    FH                 *FileH,
    PIXR              **pr)
{
    PROLOGUE;
    struct rasterfile  *sunHeader = NULL;
    Int32               ret, type, w, h, bpp;
    PIXR               *pixr = NULL;

    sunHeader = (struct rasterfile *) IPW_MALLOC(sizeof(struct rasterfile),
						 cAllocatorData);
    ERR_GEN (sunHeader == NULL, cErrNomem);

    ret = fread((char *)sunHeader, 1, sizeof(struct rasterfile), fp);

    ERR_GEN_CLEANUP (ret != sizeof(struct rasterfile), cErrIO_ReadError);

	/* check magic number and assign the type */
    ERR_GEN_CLEANUP(FLIP_INT(sunHeader->ras_magic) != RAS_MAGIC,
		    cErrBadType);

    type = FLIP_INT(sunHeader->ras_type);
    if (type == RT_STANDARD)
	FileH->type = SUNRAS;
    else if (type == RT_BYTE_ENCODED)
	FileH->type = SUNRAS_COMP;
    else
	ERR_GEN_CLEANUP(1, cErrBadType);


    IPW_FREE(FileH->header, cAllocatorData);
    FileH->header = (void *)sunHeader;

#if DEBUG_STUFF
    fprintf(stderr, " pixels/scanline =      %ld\n", sunHeader->ras_width);
    fprintf(stderr, " scanlines =            %ld\n", sunHeader->ras_height);
    fprintf(stderr, " bits/Pixel =           %ld\n", sunHeader->ras_depth);
    fprintf(stderr, " data type:             %ld\n", sunHeader->ras_type);
    fprintf(stderr, " maplength offset =     %ld\n", sunHeader->ras_maplength);
    fprintf(stderr, " databytes in file =    %ld\n", sunHeader->ras_length);
#endif  /* DEBUG_STUFF */

	/* get the FileHeader data */
    w             = FLIP_INT(sunHeader->ras_width);
    h             = FLIP_INT(sunHeader->ras_height);
    bpp           = FLIP_INT(sunHeader->ras_depth);
    FileH->width  = w;
    FileH->height = h;
    if (type == RT_STANDARD)
    {
	if (bpp == 1)
	    FileH->Wpl = (w + 15) >> 4;
	else if (bpp == 2)
	    FileH->Wpl = (w + 7) >> 3;
	else if (bpp == 4)
	    FileH->Wpl = (w + 3) >> 2;
	else if (bpp == 8)
	    FileH->Wpl = (w + 1) >> 1;
	else  /* illegal bpp */
	    ERR_GEN_CLEANUP(1, cErrDepthNotSupported);
	FileH->Bpl = FileH->Wpl << 1;
    }
    else  /* BYTE_ENCODED */
	FileH->Bpl = FileH->Wpl = 0;

    FileH->totalBytes = FLIP_INT(sunHeader->ras_length);

#if DEBUG_STUFF
    fprintf(stderr, "width = %ld (0x%x), height = %ld (0x%x), bits = %ld\n",
		w, w, h, h, bpp);
    fprintf(stderr, "ras_length = %ld (0x%x)\n",
		FileH->totalBytes, FileH->totalBytes);
#endif  /* DEBUG_STUFF */

    ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, w, h, bpp, 0, 0, cFrameBits,
					 &pixr));

    *pr = pixr;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (FileH->header == (void *)sunHeader)
	/* Means we already freed the old header and reassigned it.. */
	FileH->header = NULL;
    IPW_FREE(sunHeader, cAllocatorData);
    w_destroyPixr(pIPWState, &pixr);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  read a single scan line from the input file
 *
 * LongDescription:::
 *  For RT_BYTE_ENCODED binary files.
 *  Uses a buffer of size BLOCKSIZE bytes.
 *  
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  infile::         Ptr to file stream to read from
 *  scanLine::       duh? I dunno
 *  byteCount::      duh? I dunno
 *
 * Returns:::
 *
 * Notes:::
 *  differences from Sandpiper::
 *   This differs from the Sandpiper version by skipping most calls to
 *   wp_readBytesIntoBuffer(), by calling memset replicate the byte
 *   run values, and by using pointers to access both source and
 *   destination buffers.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::  Modified from sandpiper
 *
 * Citations:::
 *  taken from /import/sandpiper/src/bin/prim/readsunras
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_getRast2ScanLine (
    IPWState           *pIPWState,
    register FILE      *infile,
    register UInt8     *scanLine,
    register Int32      byteCount)
{
    PROLOGUE;
    register Int32      index;
    register Int32      fillcount;
    register UInt8     *rastBufPtr;

    index = 0;
    rastBufPtr = getRastBufPtr;

    while (index < byteCount) 
    {
        fillcount = (rast2SLCount <= (byteCount - index)) ? rast2SLCount
                                                          : byteCount - index;
        if ( fillcount < 10 ) 
	{
            index        += fillcount;
            rast2SLCount -= fillcount;
                    /* use most recent run */
            while (fillcount-- > 0) 
	        *scanLine++  = rast2SLValue;
        }
        else 
	{
            memset( (char*)scanLine, rast2SLValue, fillcount );
            scanLine     += fillcount;
            index        += fillcount;
            rast2SLCount -= fillcount;
        }

        if ( index >= byteCount ) 
	{
            getRastBufPtr = rastBufPtr;
            NORMAL_RETURN;
        }

        if ( rastBufPtr < &getRastBuf[ BLOCKING - 3 ] )
        {
            rast2SLValue = *rastBufPtr++;

            if (rast2SLValue != RAST_RUNFLAG) 
	    {
                *scanLine++ = rast2SLValue;   /* normal byte */
                index++; 
            }
            else 
	    {                                      /* Run flag byte */
                rast2SLCount = 1 + *rastBufPtr++;
                                        /* note: <flag><0> ==> <flag> */

                if (rast2SLCount > 1) 
		{ /* get value to be repeated */
                    rast2SLValue = *rastBufPtr++;
                }
            }
        }
        else 
	{
            if ( rastBufPtr == &getRastBuf[ BLOCKING ] )
            {
                ERR_PASS( wp_readBytesIntoBuffer(pIPWState, infile));
                rastBufPtr = getRastBufPtr;
            }

            rast2SLValue = *rastBufPtr++;

            if (rast2SLValue != RAST_RUNFLAG) 
	    {
                *scanLine++ = rast2SLValue;   /* normal byte */
                index++; 
            }
            else 
	    {                                      /* Run flag byte */
                if ( rastBufPtr == &getRastBuf[ BLOCKING ] )
                {
                    ERR_PASS( wp_readBytesIntoBuffer(pIPWState, infile ));
                    rastBufPtr = getRastBufPtr;
                }
                                        /* note: <flag><0> ==> <flag> */
                rast2SLCount = 1 + *rastBufPtr++;

                if (rast2SLCount > 1)
                {
                    if (rastBufPtr == &getRastBuf[ BLOCKING ] )
                    {
                        ERR_PASS( wp_readBytesIntoBuffer(pIPWState, infile));
                        rastBufPtr = getRastBufPtr;
                    }
                    rast2SLValue = *rastBufPtr++;
                                /* get value to be repeated */
                }
            }
        }
    }
    getRastBufPtr = rastBufPtr;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the binary image data from file in to a PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to read from.
 *  FileH::          Ptr to file header struct.
 *
 * Output:::
 *  pixr::           Ptr to binary PIXR whose data will be filled here.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_readBinaryData (
    IPWState           *pIPWState,
    FILE               *fp,
    FH                 *FileH,
    PIXR              **pixr)
{
    PROLOGUE; 
    Int16              *lastPtr, mask;
    Int32               h, fileBpl, byteDiff, Bpl;
    register Int32      y, iBytes;
    UInt8              *bdata;
    struct rasterfile  *header;

    /* Must be binary */
    ERR_GEN_PIXR(*pixr, cByteOrderBig, cDepth1, cChannels1,
		 cFrame32 | cFrame0);

        /*  internal data */
    ERR_PASS( w_pixrGetHeight(pIPWState, *pixr, &h));

    header = (struct rasterfile *)FileH->header;
    if (FileH->type == SUNRAS_COMP)
    {
        /**************************************************
         *  do this section if the data is byte encoded   *
         **************************************************/
        rast2SLValue  = 0;
        rast2SLCount  = 0;
        getRastBufPtr = &getRastBuf[ BLOCKING ];

        iBytes = ((FileH->width + 15) / 16) * 2;  /* input padded to
                                                  * 16 bit boundary */
	ERR_PASS( w_pixrGetBpl(pIPWState, *pixr, &Bpl));
        totalBytes = FileH->totalBytes;

#if DEBUG_STUFF
	fprintf(stderr, "totalBytes = %ld\n", totalBytes );
#endif  /* DEBUG_STUFF */

		/* point to 1st byte of the "real" image -- inside
		 * the pad bits
		 */
	ERR_PASS( w_pixrGetImage(pIPWState,  *pixr , &bdata));
        mask = wlmask[FileH->width & 15];

        for (y = 0; y < h; y++)
            ERR_PASS( wp_getRast2ScanLine(pIPWState, fp, bdata + y * Bpl, iBytes));
    

        ERR_PASS( w_flip(pIPWState, *pixr));

    
        if ((FileH->width & 15) != 0)
        {
                /* lastPtr points to last 16-bit word of first line in
                 * which there are pixels in the image */
 
            lastPtr = (Int16 *)(bdata) + iBytes/2 - 1;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* if the last word has an odd index, decrement the
		 * pointer on 80386 to point to the less significant
		 * 16-bit word; if it has an even index, increment
		 * pointer on 80386 to point to the more significant
		 * 16-bit word. 
		 */ 
	    if ( FileH->width & 16 ) 
	        lastPtr -= 1;
	    else                     
	        lastPtr += 1;
#endif
            for (y = 0; y < h; y++, lastPtr += Bpl / 2)
                *lastPtr &= mask;
        }
    }
    else
    {
        /**************************************************
         *  do this section if the data is packed binary  *
         **************************************************/
	fileBpl = FileH->Bpl;

	ERR_PASS( w_pixrGetBpl(pIPWState, *pixr, &Bpl));
	byteDiff = Bpl - fileBpl;  
				
	/* 0 if file and pixr Bpl are equal */
	/* point to 1st byte of the "real" image -- inside
	 * the pad bits
	 */
	ERR_PASS( w_pixrGetImage(pIPWState, *pixr, &bdata));
#if  DEBUG_STUFF
	fprintf(stderr, " Bpl = %ld, fileBpl = %ld\n", Bpl, fileBpl);
#endif  /* DEBUG_STUFF */

	/* happens on some ais files; abort
	 * fileBpl > Bpl
	 */
	ERR_GEN(byteDiff < 0, cErrIO_ReadError);

	
	for (y = 0; y < h; y++) 
	{
	  ERR_GEN(fread( bdata, 1, fileBpl, fp) != fileBpl,
		  cErrIO_ReadError);
	  bdata += Bpl;
	}
    }


    if (FileH->type != SUNRAS_COMP)
	ERR_PASS( w_flip(pIPWState, *pixr));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the gray image data from file in to a PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to read from
 *  FileH::          Ptr to file header struct
 *
 * Output:::
 *  pixr::           Ptr to gray PIXR whose data will be filled here.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_readGrayData (
    IPWState           *pIPWState,
    FILE               *fp,
    FH                 *FileH,
    PIXR              **pixr)
{
    PROLOGUE; 
    UInt16             *lastPtr;
    register Int32      y, iBytes, mask;
    UInt8              *bdata;
    Int32               h, w, sww, Wpl, wordDiff, mapSize, depth, Bpl;
    UInt16             *data, buff[2];
    struct rasterfile  *header;

    ERR_GEN_PIXR(*pixr, cByteOrderBig, cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32 | cFrame0);


        /*  internal data */
    ERR_PASS( w_pixrGetDepth(pIPWState, *pixr, &depth));
    ERR_PASS( w_pixrGetHeight(pIPWState, *pixr, &h));
    ERR_PASS( w_pixrGetWidth(pIPWState, *pixr, &w));
    ERR_PASS( w_pixrGetBpl(pIPWState, *pixr, &Bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, *pixr, &bdata));

    data  = (UInt16 *)bdata;

        /* read colormap info if present and image is a Sun raster file */
    mapSize = 0;
    header = (struct rasterfile *)FileH->header;  /* may be needed */
    if ((FileH->type == SUNRAS || FileH->type == SUNRAS_COMP) &&
        (FLIP_INT(header->ras_maptype) == RMT_EQUAL_RGB) &&
        (FLIP_INT(header->ras_maplength) != 0))
    {
        mapSize = FLIP_INT(header->ras_maplength);
        ERR_GEN(fseek( fp, mapSize, SEEK_CUR), cErrIO_SeekFailure);
    }


        /**************************************************
         *  do this section if the data is byte encoded   *
         **************************************************/
    header = (struct rasterfile *)FileH->header;
    if (FileH->type == SUNRAS_COMP)
    {
        rast2SLValue  = 0;
        rast2SLCount  = 0;
        getRastBufPtr = &getRastBuf[ BLOCKING ];

	switch (depth)
	{
	    case 2:
		iBytes = ((FileH->width + 7) >> 3) << 1;
		break;
	    case 4:
		iBytes = ((FileH->width + 3) >> 2) << 1;
		break;
	    default:		/* 8 bpp, use default so gcc won't whine */
		iBytes = ((FileH->width + 1) >> 1) << 1;  /* input padded to
						      * 16 bit boundary */
	}
        totalBytes = FileH->totalBytes;

#if DEBUG_STUFF
	fprintf(stderr, "totalBytes = %ld\n", totalBytes );
#endif  /* DEBUG_STUFF */

        for (y = 0; y < h; y++)
            ERR_PASS( wp_getRast2ScanLine(pIPWState, fp, bdata + y * Bpl, iBytes));


        ERR_PASS( w_flip(pIPWState, *pixr));

    
	    /* zero last pixels in each line */
        if ((FileH->width % (16 / depth)) != 0)
        {
	    mask = wlmask[((FileH->width)*depth) & 0xF];

                /* lastPtr initally points to last 16-bit word in first line
		 * holding real data. */
            lastPtr = (UInt16 *)(bdata) + iBytes/2 - 1;
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
		/* if the last word has an odd index, decrement the
			pointer on 80386 to point to the less significant
			word.   If the last word has an even index, 
			increment the pointer on 80386 to point to the
			more significant word.  Since the data is read in 
			Sun format, the rightmost word is really at the 
			LS address and the leftmost word is at the MS address.
		*/ 
	    if ( ((FileH->width) * depth) & 16 ) 
	        lastPtr -= 1;
	    else				 
	        lastPtr += 1;
#endif
            for (y = 0; y < h; y++, lastPtr += Bpl/2)
                *lastPtr &= mask;
        }
    }
    else
    {
        /****************************************************
         *  do this section if the data is unencoded bytes  *
         ****************************************************/
	Wpl = Bpl >> 1;
	switch (depth)		/* sww is number of 16-bit ints needed to
				 * hold the scan line's data */
	{
	case 2:
	    sww = (w + 7) >> 3;
	    break;
	case 4:
	    sww = (w + 3) >> 2;
	    break;
	default:		/* 8 bpp, use default so gcc won't whine */
	    sww = (w + 1) >> 1;
	}
	wordDiff = sww - FileH->Wpl;
	
	if ( wordDiff >= 0 ) 
	{
	    for (y = 0; y < h; y++) 
	    {         
		fread( data + y * Wpl, 1, FileH->Bpl, fp);
	    }
	}
	else 
	{   /* wordDiff == -1  */
	    for (y = 0; y < h; y++) 
	    {         
		fread( data + y * Wpl, 1, w, fp);
		fread( (char *)buff, 1, 2, fp); /* read and discard extra word
						   from the file */
	    }
	}
    }
    
    
    if (FileH->type != SUNRAS_COMP)
        ERR_PASS( w_flip(pIPWState, *pixr));


    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write out a run in SUNRAS compressed format
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  runchar::        Character that the run is made of.
 *  runlength::      Length of this character run.
 *  fp::             Ptr to file stream to write to.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::  Modified from sandpiper
 *
 * Citations:::
 *  taken from /import/sandpiper/src/bin/prim/writesunras/writesunras.c
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_endRun (
    IPWState           *pIPWState,
    UInt8               runchar,
    Int32               runlength,
    FILE               *fp)
{
    PROLOGUE;
    UInt8    runbuf[3]; 

    if (runlength > 2)
    {
        runbuf[0] = (UInt8) RAST_RUNFLAG;
        runbuf[1] = (UInt8) (runlength - 1);
        runbuf[2] = runchar;
        ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, runbuf, 3, 0)); 
        cmprlen += 3;
    }
    else /* runlength is less then or equal to 2 (<= 2) */
    {
        if (runchar == RAST_RUNFLAG)
        {
            if (runlength == 1)
            {
                runbuf[0] = (UInt8) RAST_RUNFLAG;
                runbuf[1] = (UInt8) 0;
                ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, runbuf, 2, 0));
                cmprlen += 2;
            }
            else
            {
                if (runlength == 2)
                {
                    runbuf[0] = runbuf[2] = (UInt8) RAST_RUNFLAG;
                    runbuf[1] = (UInt8) 1;
                    ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, runbuf, 3, 0));
                    cmprlen += 3;
                }
                else
                    fprintf(stderr,"Error, with runlength\n");
            }
        }
        else
        {
            if (runlength == 2)
            {
                ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, &runchar, 1, 0));
                ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, &runchar, 1, 0));
                cmprlen += 2;
            }
            else 
            {
                ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, &runchar, 1, 0));
                cmprlen++;
            }
        }
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write out some SUNRAS data compressed.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream at location to write to.
 *  buffer::         Ptr to output buffer.
 *  length::         Number of bytes to write.
 *  flush::          cTrue => first flush buffer by writing out what remains.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::  Modified from sandpiper
 *
 * Citations:::
 *  taken from /import/sandpiper/src/bin/prim/writesunras/writesunras.c
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_writecmpr (
    IPWState           *pIPWState,
    FILE               *fp,
    UInt8               buffer[],
    Int32               length,
    Int32               flush)
{
    PROLOGUE;
    Int32          i = -1;
    static Int32   runlength = 0;
    static UInt8   runchar = 0;
    static UInt8   runbuf[3];

    if (flush)  /* Then we need to write out what remains. */
    {
        ERR_PASS( wp_endRun(pIPWState, runchar, runlength, fp));

	/* flush the buffer */
        ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, buffer, 0, 1));
        runlength = 0;
    }


    while (++i < length)
    {
        if (runlength == 0)
        {
            runchar =  buffer[i];
            runlength++;
        }
        else /* we have a run of characters going. */
        {
            if (runchar != buffer[i]) /* then we need to start a new run */
            {
                ERR_PASS( wp_endRun(pIPWState, runchar, runlength, fp));
                runlength = 1;
                runchar = buffer[i];
            }
            else
            {
                runlength++;
                if (runlength == 256) 
                {
                    runbuf[0] = (UInt8) RAST_RUNFLAG;
                    runbuf[1] = (UInt8) (runlength - 1) ;
                    runbuf[2] = (UInt8) runchar;
                    ERR_PASS( wp_writeByteFromBuffer(pIPWState, fp, runbuf, 3, 0));
                    cmprlen += 3;
                    runlength = 0;
                }
            }
        }
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write out a PIXR to a file stream in SUNRAS image format.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  input_pr::       Ptr to 1,2,4 or 8-bit PIXR to write out.
 *  output::         Ptr to file stream to write to.
 *                   Note:  assumes stream buffer ptr is at correct place,
 *                   we simply begin writing from where it is.
 *  type::           RT_STANDARD or RT_BYTE_ENCODED
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_pixr_dump (
    IPWState           *pIPWState,
    PIXR               *input_pr,
    FILE               *output,
    Int32               type)
{
    PROLOGUE;
    struct rasterfile  *hdrSUN = NULL;
    PIXR               *temp_pr = NULL;
    Int32               row;
    UInt8              *imgptr;
    Int32               write_size;
    Int32               width, depth, height, tBpl, needToStripFrame;

    ERR_GEN (!input_pr, cErrPixr_null);


    /* first construct the header and write it to the file */
    ERR_PASS( wp_pixrMakeHeaderSUN( pIPWState, input_pr, &hdrSUN ));

    if ((type != RT_STANDARD) && (type != RT_BYTE_ENCODED)) {
#ifndef  PRODUCTION
	fprintf(stderr, "pixr_dump called with unexpected output type (%ld)\n",
		type);
#endif  /* ! PRODUCTION */
	ERR_GEN(1, cErrBadType);
    }
    hdrSUN->ras_type = FLIP_INT(type);

    ERR_GEN_CLEANUP(fwrite( (char *)hdrSUN, sizeof( struct rasterfile), 1,
			    output ) != 1, cErrIO_WriteError);

    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, input_pr, &width));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, input_pr, &height));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, input_pr, &depth));

        /* If this is a PC, we need to copy the pixr so we can flip the
	 * bytes in each 32-bit word (OUCH!!!).  If this is a Sun, don't
	 * bother. */
#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	/*---------------------------------------------------------------
	 * This is a PC, so create the pixr which will hold the stripped,
	 * byte-swapped image. 
	 * Then strip the frame and flip all the pixr bytes 
	 *---------------------------------------------------------------*/
    ERR_PASS_CLEANUP( w_createPixrNoInit(pIPWState, width, height, depth,
					 0, 0, 0, &temp_pr));
    ERR_PASS_CLEANUP( w_rasterop (pIPWState, 
				  0, 0,
				  input_pr, 
				  PIX_SRC,
				  0, 0, width, height,
				  temp_pr));

    ERR_PASS_CLEANUP( w_flip(pIPWState, temp_pr));
    needToStripFrame = cFalse;

#else
	/* This is NOT a PC, so the pixr we write from is the input pixr 
	 * Also remember to strip the frame as we write the image */
    temp_pr 	     = input_pr;
    needToStripFrame = cTrue;

#endif	/* _ALPACA_IMAGE_FMT_ */

    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState,  temp_pr , &tBpl));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, temp_pr, &imgptr));

    write_size = ((width*depth + 15) >> 4) << 1;

    if ( type == RT_STANDARD ) 
    {
                
	    /* if image width / 16 = 2 * ( image width / 32) */
	if ((tBpl == write_size) && !needToStripFrame)
	{
	    ERR_GEN_CLEANUP(fwrite(imgptr, tBpl, height, output) != height,
			    cErrIO_WriteError);
	}
	else /* either lines aren't padded to 32 bit boundaries or we need
	      * to strip the frame */
	{
	    for (row = 0; row < height; ++row, imgptr += tBpl)
		ERR_GEN_CLEANUP(fwrite( imgptr, write_size, 1, output) != 1,
				cErrIO_WriteError);
	}
    }
    else if (type == RT_BYTE_ENCODED)
    {
	cmprlen = 0;

	if ((tBpl == write_size) && !needToStripFrame)
	{   /* if image width / 16 = 2 * ( image width / 32) */
	    ERR_PASS_CLEANUP( wp_writecmpr(pIPWState, output, imgptr,
				       tBpl * height, 0));
	}
	else
	{   /* not packed to 32-bit boundary or need to strip frame off
	     * each line. */
	    for (row = 0; row < height; ++row, imgptr += tBpl)
		ERR_PASS_CLEANUP( wp_writecmpr(pIPWState, output, imgptr, 
					   write_size, 0));
	}

	    /* flush the compression routine */
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, temp_pr, &imgptr));
	ERR_PASS_CLEANUP( wp_writecmpr(pIPWState, output, imgptr, 0, 1));

	     /* seek to position of ras_length */
	fseek(output, 0L, L_SET); /* seek to start of file */
	hdrSUN->ras_length = FLIP_INT( cmprlen);
	ERR_GEN_CLEANUP(fwrite((char *)hdrSUN, sizeof(struct rasterfile), 1,
			       output) != 1, cErrIO_WriteError);
    }

    /* Clean up */
    IPW_FREE(hdrSUN, cAllocatorData);

    /* this is why we don't bother restoring temp Pr's width and depth */
    if (temp_pr != input_pr)
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &temp_pr));

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(hdrSUN, cAllocatorData);
    if (temp_pr != input_pr)
	w_destroyPixr(pIPWState, &temp_pr);

    ERR_RETURN;
}






/*

 $Log: sunio.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.39  1997/06/30 23:11:36  ddavies
 sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v

 Delete inclusion of Sandpiper's internal.h

 Revision 1.38  1997/06/26 20:04:58  ddavies
 sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.37  1997/04/15 22:43:07  ddavies
 sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.36  1996/05/16 17:30:39  ddavies
 sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v

 Removed unused rect parameter from wp_readSunHeader, wp_readBinaryData and
 wp_readGrayData.

 Revision 1.35  1996/01/04 20:27:33  ejaquith
 sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.34  1995/12/13  04:14:02  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.33  1995/11/13  20:02:21  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Added needed additional inclusions of "ptypes.h", "gdefs.h", and "pcx.h".
 * Updated PROLOGUE usage.
 *
 * Revision 1.32  1995/10/30  18:04:43  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Updated function block comments who named other IPWrap routines.
 *
 * Revision 1.31  1995/10/13  22:06:01  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.30  1995/10/13  02:16:06  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Fixed AIS depth 1 and depth 8 bug on the PC where the image was not
 * getting byteswapped after the data had been read.
 * Condensed (and made better!) the code for cutting a rect out of the
 * AIS/SUN images by converting to calling w_rectOf().
 *
 * Revision 1.29  1995/10/07  01:22:06  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.28  1995/09/29  00:35:18  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.27  1995/09/21  23:00:28  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.26  1995/09/14  18:57:06  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 *  Fixed types to be used with w_pixrGetResolutionDPI().
 *
 * Revision 1.25  1995/09/12  16:54:56  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.24  1995/09/08  00:14:54  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.23  1995/09/07  01:36:00  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.22  1995/09/06  20:24:13  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.21  1995/09/01  02:20:41  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.20  1995/08/31  21:39:52  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.19  1995/08/29  22:24:46  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.18  1995/08/29  21:56:55  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.17  1995/08/22  02:21:58  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.16  1995/08/17  22:01:25  ejaquith
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.15  1994/10/10  22:35:30  lperry
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Change to use new header file io.pub instead of iopros.h
 * Remove function flip_pixr and move to ioutils.c
 *
 * Revision 1.14  1994/10/04  22:19:04  lperry
 * lperry on Tue Oct 4 15:17:22 PDT 1994
 *
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Add mention of argument to flip_pixr(PIXR) to
 * prevent compiler whining on big-endian machines,
 * where the routine just returns w/o referencing PIXR.
 *
 * Revision 1.13  1994/09/20  04:34:54  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Fix to the fix to the fix to the fix.  flip_pixr now references procName
 * if and only if it's defined.  Got the sense of one of the tests wrong last time.
 *
 * Revision 1.12  1994/09/18  01:40:54  ddavies
 * ddavies on Sat Sep 17 18:40:06 PDT 1994
 *
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Fix to the Fix to the Fix:  flip_pixr now references procName when it's
 * supposed to on the right platforms. It now returns 0 like it's supposed to.
 *
 * Revision 1.11  1994/09/17  20:21:42  ddavies
 * ddavies on Sat Sep 17 13:15:04 PDT 1994
 *
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Correct previous fix by putting ifndef PRODUCTION around use of procName
 * on big-endian platforms.
 *
 * Revision 1.10  1994/09/17  18:16:18  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Put ifndef PRODUCTION around declaration of procName in flip_pixr.
 *
 * Revision 1.9  1994/08/26  01:44:45  lperry
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Make the FileH file header struct a parameter to achieve reentrancy.
 * Creates interface changes to readSunHeaderPr, readBinaryDataPr,
 * and readGrayDataPr, but these are only called from readPr/freadPr.
 *
 * Revision 1.8  1994/05/05  01:53:08  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Include ansiprot.h when compiling with gcc.  Change lots of %d's to
 * %ld's.  Change last case of switches to "default" so gcc will know that
 * vars are initialized.  Initialize "status" whereever it's used.
 * Use procName inside flip_pixr to keep gcc happy.
 *
 * Revision 1.7  1994/01/22  01:42:09  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Change shrio.pub to shrpixr.pub since ipshared's libiof and libios have
 * been absorbed into libpixrf and libpixrs.
 *
 * Revision 1.6  1994/01/14  04:40:16  ddavies
 * ddavies on Thu Jan 13 20:39:15 PST 1994
 *
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Supply return values (0) for flip_pixr and redBytesIntoBuffer so WATCOM
 * 9.5 won't complain.
 * Yeah, I probably should have changed the prototypes to return void, but
 * I was too lazy.
 *
 * Revision 1.5  1994/01/12  23:46:43  dferrell
 * dferrell on Wed Jan 12 15:43:01 PST 1994
 *
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Removed call to setmode () for PCs since we have fixed writePR and
 * fwritePR to open files in binary mode when necessary.
 * Removed calls to readport () for PCs.  This was needed when we
 * employed the MetaWare compiler which had a 64K limit on file I/O.
 * Cleaned up code format and replaced old style image format #defines
 * (e.g. _ALPACA_PC_FMT_) with the new style (e.g. cAlpacaPCFmt).
 *
 * Revision 1.4  1994/01/12  01:16:53  dferrell
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * #if 0'ed out calls to setmode for the PC.
 *
 * Revision 1.3  1993/12/03  01:19:49  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:18:25  ddavies
 * sunio.c -> /product/ipcore/ipwrap/src/RCS/sunio.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

