/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tagrun.c -- procedures that tag every pixel in a run of pixels with the
 *		length of that run.
 *
 *              w_tagHorizontalRuns ()
 *              w_tagVerticalRuns ()
 *
 */



#include "ipwcore.prv"

#include "pixr.h"
#include "defines.pub"     /* for cTrue and cFalse */
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "shrcc.pub"

#include "tagrun.pub"


/* prototypes */

IP_RCSINFO(RCSInfo, "$RCSfile: tagrun.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Tag each pixel in the 8-bit output with the length of the run it's in in the input
 *
 * LongDescription:::
 *  For instance, all the pixels in an n-bit horizontal run in the binary 
 *  input PIXR will be tagged with MIN (n, 255) in the output PIXR.  If the 
 *  destination PIXR has not been created, this function will attempt to 
 *  create it.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prS::            Binary source PIXR
 *
 * Output:::
 *  prD::            8-bit gray destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  run length
 *
 * Keywords:::
 *  run length
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_tagHorizontalRuns (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **prD)
{
    PROLOGUE;
    Int32	 wS, hS,     /* width, height and depth of images */
                 bplS, bplD;         /* bpl */
    Bool	 destCreated = cFalse;
    ScaledInt1616  xRes, yRes;         /* Resolution of images */
    Int32        rfc, bfc;
    UInt8       *iS, *iD;            /* image pointers */

    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prD != NULL)
	ERR_GEN_PIXR (*prD, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prS, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState,
					      wS, hS, 8,
					      xRes, yRes, 
					      cFrameBits, prD, 
					      &rfc, &bfc, &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prD, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prS, &hS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prD, &iD));

    ERR_PASS_CLEANUP (i_tagHorizontalRuns (pIPWState, iS, bplS,
					   wS, hS, iD, bplD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prD, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prD);
    ERR_RETURN;

}  /* tagHorizontalRunsPixr () */


/*************************************************************
 **************        Tag Vertical Runs       *************
 ************************************************************/
/*
 *  tagVerticalRunsPixr():   
 *	Input:	binary image.
 *	Output: Gray image in which each pixel is tagged with
 *		the length of the vertical run it's in.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Tag each pixel in the 8-bit output with the length of the run it's in in the input
 *
 * LongDescription:::
 *  For instance, all the pixels in an n-bit vertical run in the binary 
 *  input PIXR will be tagged with MIN (n, 255) in the output PIXR.  If the 
 *  destination PIXR has not been created, this function will attempt to 
 *  create it.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  prS::            Binary source PIXR
 *
 * Output:::
 *  prD::            8-bit gray destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  run length
 *
 * Keywords:::
 *  run length
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_tagVerticalRuns (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **prD)
{
    PROLOGUE;
    Int32	 wS, hS,     /* width, height and depth of images */
                 bplS, bplD;         /* bpl */
    Bool	 destCreated = cFalse;
    ScaledInt1616  xRes, yRes;         /* Resolution of images */
    Int32        rfc, bfc;
    UInt8       *iS, *iD;            /* image pointers */

    ERR_GEN_PIXR (prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*prD != NULL)
	ERR_GEN_PIXR (*prD, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prS, &xRes, &yRes));

    ERR_PASS (wp_createPixrConditionalNoInit (pIPWState,
					      wS, hS, 8,
					      xRes, yRes, 
					      cFrameBits, prD, 
					      &rfc, &bfc, &destCreated));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *prD, &bplD));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, prS, &hS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &iS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *prD, &iD));

    ERR_PASS_CLEANUP (i_tagVerticalRuns (pIPWState, iS, bplS,
					 wS, hS, iD, bplD));

    /* Clear the dest frame if necessary */
    ERR_PASS_CLEANUP (wp_clearRightBottomFrames (pIPWState, *prD, rfc, bfc));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (destCreated)
        w_destroyPixr (pIPWState, prD);
    ERR_RETURN;


}  /* tagVerticalRunsPixr () */



/*

 $Log: tagrun.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.26  1998/01/14 00:15:43  ddavies
 tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.25  1997/04/15 22:43:22  ddavies
 tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.24  1996/01/04 20:27:36  ejaquith
 tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.23  1995/12/13  04:14:06  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/13  20:17:11  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/10/30  23:22:31  dferrell
 * dferrell on Mon Oct 30 15:22:04 PST 1995
 *
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Added error checking macros around wp_clearRightBottomFrames () calls.
 *
 * Revision 1.20  1995/10/13  22:06:05  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/10/13  02:18:11  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Fixed OneLineDescription typo in function -- the tag was missing.
 *
 * Revision 1.18  1995/09/28  21:40:52  dferrell
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.17  1995/09/21  23:00:31  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/19  18:55:44  dferrell
 * dferrell on Tue Sep 19 11:55:00 PDT 1995
 *
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Pass framed width/height into ipshared run tagging functions instead
 * of unframed width/height.
 *
 * Revision 1.15  1995/09/15  22:39:05  ddavies
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Changed type of destCreated to Bool to keep WATCOM happy.  Changed type of
 * resolutions to ScaledInt1616 where necessary.
 *
 * Revision 1.14  1995/09/11  21:21:47  dferrell
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Updated function names, parameter order, and consolidated xxxPr and xxxPixr
 * functions.
 *
 * Revision 1.13  1995/09/07  01:36:03  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:24:16  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  17:07:21  dferrell
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.10  1995/08/31  21:39:55  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/25  01:38:52  ejaquith
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Fixed warnings
 *
 * Revision 1.8  1995/08/24  00:32:31  jcraig
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * switch source and destination in parameter list when calling
 * tagHorizontalRunsPr and tagVerticalRunsPr
 *
 * Revision 1.7  1995/08/17  21:58:16  dferrell
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.6  1995/01/11  22:44:55  dferrell
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Changed include of shrip.pub to shrcc.pub.
 *
 * Revision 1.5  1994/10/07  19:19:23  lperry
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.4  1994/06/08  04:14:02  ddavies
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.3  1993/12/03  01:19:53  ddavies
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:18:32  ddavies
 * tagrun.c -> /product/ipcore/ipwrap/src/RCS/tagrun.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

