/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tcommon.c:
 *
 *	Code used for both compression and decomprssion of images
 *	containing text.  The image is encoded as repeated instances
 *	of shapes.  A compressed image has two parts: position
 *	information and shape information.
 *	A document may be composed of multiple blocks of position
 *	information and a single block of shape information.
 *	
 *		
 *
 *	w_allocateTextCompressionInfo(): Allocate the structure used
 *					to hold information about
 *					a compressed document including
 *					pointers to shape info and
 *					all allocated memory.
 *
 *	w_freeTextCompressionInfo():	frees all information and
 *					memory associated with a
 *					compressed image.
 *
 */

#include "ipwcore.prv"

/* DEFINITIONS */
#include "pixr.h"

/* PROTOTYPES */
#include "tcompr.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: tccommon.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* This file contains wrappers for utility routines that are used for
 * both text compression and decompression.
 *
 * The theory and use of text compression is documented in
 * ipshared/src/i_tcompr.c.
 *
 * The basic idea behind our text compression is that a page of text
 * contains many repetitions of a relatively small set of characters.
 * We decide each new blob is either another instance of an
 * old blob or is a new type of blob.  In either case, we remember
 * where we saw it.  The compressed image consists of a set of positions
 * and a description of each type of blobs.
 *
 * The code may be used to compress several different documents,
 * each consisting of several images, concurrently.  We use a single
 * data structure to hold the information relating to each image.  The
 * data structure contains pointers to allocated memory that holds
 * shape information, the largest symbol number used so far and so on.
 *
 * allocateTextCompressoinInfo is used to allocate this structure.
 * Client's must call this function to get the structure before doing any
 * compression or decompression.  The structure is passed down to the
 * compression and/or decompression code.  The client does not access
 * it directly.
 *
 * freeTextCompressionInfo is used to free all memory allocated by the
 * text compression code in the process of compressing or decompressing
 * an image.  Clients should call this routine after completing the
 * compression or decompression of a document.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Allocates block used by text compression and decompression.
 *	
 * LongDescription:::
 *	Allocates and initializes the private structure, returning
 *	a pointer to it in *ppTextCompInfo.
 *
 * Input:::
 *  pIPWState::       Standard IPWrap state structure
 *
 * Output:::
 *  ppTextCompInfo::  Pointer to Pointer to returned private text
 *		      compression data block
 *
 * Returns:::
 *
 * Notes:::
 *	It's pretty hard to for this to fail unless we run out of memory.
 *      The private structure will survive garbage collection if the clients
 *	follow the rules.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	symbol-based, garbage collection
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *	D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_allocateTextCompressionInfo (
    IPWState           *pIPWState,
    void              **ppTextCompInfo)
{
    PROLOGUE;

    ERR_GEN( ppTextCompInfo == NULL, cErrInvalidParm);

	/* just ask the ipshared routine to do it's job */
    ERR_PASS( i_allocateTextCompressionInfo (pIPWState, ppTextCompInfo,
					     i_compressAlloc,
					     i_compressFree));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Frees block used by text compression and decompression.
 * LongDescription:::
 *	Frees the private structure and all other memory allocated by
 *	the text compression/decompression activity associated with this
 *	document.  Sets *ppTextCompInfo to NULL on return.
 *
 * Input:::
 *  pIPWState::       Standard IPWrap state structure
 *  ppTextCompInfo::  Pointer to pointer to private block used by text
 *		      compression and decompression.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	symbol-based, garbage collection
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *	D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_freeTextCompressionInfo (
    IPWState           *pIPWState,
    void              **ppTextCompInfo)
{
    PROLOGUE;

    ERR_GEN( ppTextCompInfo == NULL, cErrInvalidParm);

	/* If it's already been freed, just return */
    if (*ppTextCompInfo == NULL)
	goto fTCFinish;

	/* just ask the ipshared routine to do it's job */
    ERR_PASS( i_freeTextCompressionInfo (pIPWState, *ppTextCompInfo));

	/* i_freeTextCompressionInfo has already freed everything,
	 * so set the pointer to NULL.
	 */
    *ppTextCompInfo = NULL;

fTCFinish:
    NORMAL_RETURN;
}



/*

 $Log: tccommon.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.20  1998/01/14 00:15:44  ddavies
 tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.19  1996/01/04 20:27:38  ejaquith
 tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.18  1995/12/13  04:14:09  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:17:16  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  22:06:09  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/09/28  19:19:22  ddavies
 * ddavies on Thu Sep 28 12:19:21 PDT 1995
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Filled in autodoc comments.
 *
 * Revision 1.14  1995/09/21  23:00:34  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/15  17:59:49  ddavies
 * ddavies on Fri Sep 15 10:59:48 PDT 1995
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.12  1995/09/11  22:10:35  ddavies
 * ddavies on Mon Sep 11 15:10:35 PDT 1995
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.11  1995/09/10  00:32:39  ddavies
 * ddavies on Sat Sep 9 17:32:38 PDT 1995
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.10  1995/09/07  01:36:05  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.9  1995/09/06  20:24:20  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.8  1995/09/01  01:19:10  ddavies
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.7  1995/08/31  21:39:57  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/17  22:04:17  ejaquith
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.5  1995/07/29  04:52:00  ddavies
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.4  1994/10/11  01:07:46  lperry
 * lperry on Mon Oct 10 18:07:17 PDT 1994
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * Changes to use .pub interface.
 *
 * Revision 1.3  1994/06/06  19:42:17  ddavies
 * ddavies on Mon Jun 6 12:41:10 PDT 1994
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * Change call to i_allocateTextCompInfo to take i_compressAlloc and
 * i_compressFree instead of i
 * (ouch) tcAllocator and tcFreer.  This centralizes the type of
 * allocation and freeing proc in one place.
 *
 * Revision 1.2  1994/05/31  17:57:12  ddavies
 * ddavies on Tue May 31 10:54:29 PDT 1994
 *
 * tccommon.c -> /product/ipcore/ipwrap/src/RCS/tccommon.c,v
 *
 * Change calls to i_allocateTextCompressionInfo so we pass both the allocation
 * and freeing routines.  These are stored with the compression info.
 *
 * Revision 1.1  1994/04/26  02:36:17  ddavies
 * Initial revision
 *

 */

