/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tcdbug.c:
 *
 *	Code used to debug text compression.
 *	
 *		
 *	w_createDictionaryPositions():		produces a position block
 *						that puts the symbols found
 *						into a grid.
 *	w_decodePositionInfo:			Converts position/shapeID
 *						info into ASCII.
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"

/* PROTOTYPES */
#include "tcompr.pub"
#include "shrcmprs.pub"
#include "shrcmprs.prv"



IP_RCSINFO(RCSInfo, "$RCSfile: tcdbug.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* This file contains wrappers for routines used to debug text image
 * compression routines.
 *
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Debugging routine used to view unique symbols on a page
 *
 * LongDescription:::
 *  This procedure produces a block of position data that, when
 *  decoded, causes the dictionary symbols to be printed in a grid
 *  in an image.  This is used to inspect the dictionary symbols
 *  for debugging purposes.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pTextCompInfo::  Pointer to private symbol-based compression state block.
 *  firstSymbol::    Symbol ID of symbol that will be placed
 *		     in the upper left box of the output image.
 *  imgWidth::       Width of the dest image.
 *  imgHeight::      Height of the dest image.
 *  boxWidth::       Width of box assigned to each symbol in the image.
 *  boxHeight::      Height of box assigned to each symbol in the image.
 *  positionBufSize:: Number of bytes in buffer to be filled in.  If
 *		     no position buffer is passed in, this is ignored.
 *
 * Output:::
 *  ppPositionBuf::  Pointer to Pointer to resulting buffer filled with
 *		     symbol IDs and positions.  If *ppPositionBuf is NULL,
 *		     a position buffer is created.  If it is created and
 *		     there's an error, it is freed here too.
 *  pResultSize::    Number of bytes of position buf actually used.
 *
 * Returns:::
 *  In addition to the normal code, this routine returns cErrBufferExceeded.
 *  This code is set when the position buffer was too small.  Since this is
 *  only a debugging function (not of use to real clients), we don't give the
 *  client an easy way to figure out how big a buffer to allocate.
 *
 * Notes:::
 *  This function creates a page that's divided into a grid.  We start through
 *  the list of symbols starting at firstSymbol.  If a symbol fits into the
 *  grid space (boxWidth x boxHeight pixels), it's ID and position are recorded
 *  in the output position buffer block.  If the symbol doesn't fit, it's
 *  ignored.  When the position buffer is decompressed, we can see the
 *  set of unique symbols.  We can also derive the number of each symbol,
 *  which is also useful for debugging.  See ipshared/testsrc/s_symcmp.c
 *  for an example of the use of this function.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  debug, compression, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createDictionaryPositions (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    Int32               firstSymbol,
    Int32               imgWidth,
    Int32               imgHeight,
    Int32               boxWidth,
    Int32               boxHeight,
    UInt32              positionBufSize,
    UInt8             **ppPositionBuf,
    UInt32             *pResultSize)
{
    PROLOGUE;
    UInt32  bufAllocatedHere;

    ERR_GEN(pTextCompInfo == NULL || ppPositionBuf == NULL ||
	   pResultSize == NULL, cErrInvalidParm);

	/* if no buffer was passed in, allocate one */
    if (*ppPositionBuf == NULL)
    {
	ERR_PASS(w_getNumberOfSymbols(pIPWState, pTextCompInfo,
				    cSymbolsAll, 0, &positionBufSize));

	    /* at the outside, assume 5 bytes/symbol */
	positionBufSize *= 5;
	positionBufSize += 16;	/* for boilerplat on the ends */
	*ppPositionBuf = (UInt8 *)IPW_MALLOC(positionBufSize, cAllocatorData);
	ERR_GEN(*ppPositionBuf == NULL, cErrNomem);

	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

	    /* just ask the ipshared routine to do its job */
    ERR_PASS_CLEANUP( i_createDictionaryPositions (pIPWState, pTextCompInfo,
						   firstSymbol,
						   imgWidth + 2*cFrameBits,
						   imgHeight + 2*cFrameBits,
						   boxWidth, boxHeight,
						   *ppPositionBuf,
						   positionBufSize,
						   (Int32 *)pResultSize));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (bufAllocatedHere)
	IPW_FREE(*ppPositionBuf, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Debugging routine used to view symbol-based compressed files.
 *
 * LongDescription:::
 *	This procedure converts a block of encoded position information
 *	into a block of ASCII text.  It is used only for debugging
 *	symbol-based compression.  Each scan line is converted to
 *	to the form:
 *	yyyy: xxxx/ssss xxxx/ssss ... xxxx/ssss\n
 *	where yyyy = 4 digit y position, xxxx = 4 digit x position and
 *	ssss = 4 digit symbol ID.
 *
 * Input:::
 *  pIPWState::       Pointer to standard IPWrap state block.
 *  pTextCompInfoIn:: Pointer to private symbol-based compression state block.
 *  pPositionBuf::    Pointer ot poistion buffer to be decoded.
 *  charBufSize::     Size of buffer supplied to hold decoded position info.
 *		      If no character buffer is supplied, this is ignored.
 *
 * Output:::
 *  ppCharBuf::       Pointer to Pointer to char buffer that will contain
 *		      the decoded output.  If *ppCharBuf is NULL, a buffer
 *		      will be allocated.  If allocated here and there is
 *		      an error, it will be freed.
 *  pResultSize::     Pointer to word that will contain the number of chars
 *		      actually written into *ppCharBuf.
 *
 * Returns:::
 *  If it allocates its own position buffer, this routine tries not to
 *  return cErrBufferExceeded in that it keeps trying to allocate larger
 *  buffers until the decoded info fits.  It will not do this if a buffer
 *  is passed in..
 *
 * Notes:::
 *  This function is really only used by people who are trying to debug
 *  the symbol-based compression code.  It's used to figure how which
 *  symbol was found where.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, debug, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_decodePositionInfo (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8              *pPositionBuf,
    UInt32              charBufSize,
    char              **ppCharBuf,
    UInt32             *pResultSize)
{
    PROLOGUE;
    TCCompressionInfoPtr     pTextCompInfo;
    char                    *pNewCharBuf;
    UInt32                   bufAllocatedHere;

    ERR_GEN(pTextCompInfoIn == NULL || pPositionBuf == NULL ||
	   ppCharBuf == NULL || pResultSize == NULL, cErrInvalidParm);

	/* do the typecasting */
    pTextCompInfo = (TCCompressionInfoPtr) pTextCompInfoIn;

	/* if no buffer was passed in, allocate one */
    if (*ppCharBuf == NULL)
    {
	charBufSize = pTextCompInfo->totalSymbols;
	    /* at the outside, assume 12 bytes/symbol */
	charBufSize *= 12;

	*ppCharBuf = pNewCharBuf = (char *)IPW_MALLOC(charBufSize, 
						      cAllocatorData);
	ERR_GEN(pNewCharBuf == NULL, cErrNomem);

	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

	/* try to decode the position info.  Keep trying with larger
	 * buffers until it works.
	 */
    while (cTrue)
    {
		/* just ask the ipshared routine to do its job */
	ERR_CHECK_NO_MSG(i_decodePositionInfo (pIPWState, pPositionBuf,
					       *ppCharBuf,
					       charBufSize,
					       pResultSize))
	    {
		if (__error__ == cErrBufferExceeded)
		{
		    if (bufAllocatedHere)
		    {
			/* try again with a larger character buffer */
			charBufSize += charBufSize/2;
			pNewCharBuf = IPW_REALLOC(*ppCharBuf, charBufSize, cAllocatorData);
			ERR_GEN_CLEANUP(pNewCharBuf == NULL, cErrNomem);

			*ppCharBuf = pNewCharBuf;

			/* and loop around to try again */
			continue;
		    }
		    else	
			/* the client allocated the buffer, so just fall
			 * through with the offending retCode.
			 */
			ERR_GEN(cTrue, __error__);
			ERR_RETURN;
		}
		else
		{
		    /* something else bad happened.  Deallocate the buffer if
		     * it was allocated here.  retCode will contain the
		     * offending return code.
		     */
		    ERR_GEN_CLEANUP(cTrue, __error__);
		    goto ERR_CLEANUP;
		}
	    }

	/* Will only get here if routine completed successfully above. */
	break;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (bufAllocatedHere)
	IPW_FREE(*ppCharBuf, cAllocatorData);

    ERR_RETURN;
}



/*

 $Log: tcdbug.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.23  1998/08/04 00:21:26  aber
 aber on Mon Aug 3 17:21:25 PDT 1998

 tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v

 CommonCheck: Support multiple dictionaries when token compressing, used to retain
 single-use tokens in the same dictionary -- to be used when color
 tokenizing.

 Revision 1.22  1998/01/14 00:15:46  ddavies
 tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.21  1996/07/04 03:09:45  ddavies
 ddavies on Wed Jul 3 20:08:30 PDT 1996

 tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v

 Call i_decodePositionInfo with the new ERR_CHECK_NO_MSG macro so we can
 handle buffer overflow (alloc a bigger buffer) silently.

 * Revision 1.20  1996/01/04  20:27:42  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.19  1995/12/13  04:14:11  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.18  1995/11/13  20:17:20  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.17  1995/10/13  22:06:12  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.16  1995/09/28  22:22:07  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Fix comments at top of file.
 *
 * Revision 1.15  1995/09/28  19:19:33  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Filled in autodoc comments.
 *
 * Revision 1.14  1995/09/21  23:00:37  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.13  1995/09/15  17:59:55  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.12  1995/09/12  18:28:47  ejaquith
 * ejaquith on Tue Sep 12 11:28:01 PDT 1995
 *
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * fixed boo-boo in w_decodePositionInfo().  was ptr checking local var
 * instead of similarly named param.
 *
 * Revision 1.11  1995/09/11  22:10:39  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.10  1995/09/10  00:32:43  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.9  1995/09/07  01:36:08  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.8  1995/09/06  20:24:23  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.7  1995/09/01  01:19:16  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.6  1995/08/31  21:40:00  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/17  22:04:31  ejaquith
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.4  1995/07/29  04:52:20  ddavies
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.3  1995/03/15  18:55:55  ddavies
 * ddavies on Wed Mar 15 10:55:34 PST 1995
 *
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * Fix cast of char * so that acc won't complain.
 *
 * Revision 1.2  1995/03/15  03:19:55  ddavies
 * ddavies on Tue Mar 14 19:19:36 PST 1995
 *
 * tcdbug.c -> /product/ipcore/ipwrap/src/RCS/tcdbug.c,v
 *
 * Add decodePositionInfo.  This is a wrapper for i_decodePositionInfo.  It
 * converts a block of position information into an ASCII block containing
 * yyyy: xxxx/ssss xxxx/ssss ....
 * for each scan line containing the bottom of a symbol.  yyyy = the y coord,
 * xxxx = the x coord of the lower right corner of the bounding box, ssss =
 * the symbol ID.
 *
 * Revision 1.1  1994/10/26  01:09:19  ddavies
 * Initial revision
 *

 */

