/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tcdcomp.c:
 *
 *	Code used to decompress images containing text.  The image is
 *	encoded as repeated instances of shapes.  A compressed
 *	image has two parts: position information and shape information.
 *	A document may be composed of multiple blocks of position
 *	information and a single block of shape information.
 *	
 *
 *	w_deserializeCompressionInfo():		Convert compression info
 *						from serial bytestream to
 *						internal format
 *	w_convertSymbolsToImagePixr():		Convert the internal
 *						representation of a page
 *						into the page's bitmap.
 *
 *	w_getNumberOfSymbols():			returns number of symbols in
 *						current shape info
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "frames.pub"
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "tcompr.pub"
#include "shrcmprs.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: tcdcomp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* This file contains wrappers for routines used to decompress images
 * containing text.
 *
 * A compressed file contains two parts.  There are one or more sections
 * of position information.  Each section specifies the position of each symbol
 * in an image.  There is a final section specifying the shapes of all
 * the symbols used.
 *
 * getNumberOfSymbols returns the number of symbols represented in the
 * shape information.
 *
 * convertSymbolsToImagePixr creates a new PIXR and fills it with
 * decompressed image data.
 *
 * convertSymbolsToImage does the same with an already existing PIXR.
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress bytestream containing symbol shapes.
 *
 * LongDescription:::
 *  A symbol-based compressed image contains two basic components, one or
 *  more blocks containing symbol IDs with positions and a block of
 *  compressed symbol shapes.  This routine decompresses the symbol shaped,
 *  putting the result into the internal data structures.  From there,
 *  the symbols may be used to either compress more pages or display
 *  decompressed pages.
 *
 * PrivateDescription:::
 *  See description of w_serializeCompressionInfo for description of
 *  compressed symbols.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pTextCompInfo::  Pointer to private symbol-based compression data block.
 *		     This will point to the decompressed symbols when this
 *		     routine completes.
 *  pShapeInfoBuf::  Pointer to block of memory containing compressed
 *		     shapes.  This block of memory was produced by
 *		     w_serializeCompressionInfo.
 *
 * Returns:::
 *  In addition to the normal return values, this routine can return
 *  cErrDecodeFailure.  This means that there was an internal error in
 *  decoding the compressed symbols.  The most likely cause of this is
 *  an attempt to decode shapes encoded with an incompatible version of
 *  w_serializeCompressionInfo.
 *
 * Notes:::
 *  See description for w_convertImageToSymbols to see how this routine is
 *  used to decompress images.  See description for w_convertImageToSymbols
 *  to see how this routine is used during compression of images with an
 *  old dictionary of symbols.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable, unless we change the format of symbol-based compressed files.
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_deserializeCompressionInfo (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    UInt8              *pShapeInfoBuf)
{
    PROLOGUE;

	/* see if the pointers are valid */
    ERR_GEN((pTextCompInfo == NULL) || (pShapeInfoBuf == NULL),
	    cErrInvalidParm);

	/* do the real work */
    ERR_PASS( i_deserializeCompressionInfo (pIPWState, pTextCompInfo,
					    pShapeInfoBuf));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress symbol-based compressed image
 *
 * LongDescription:::
 *  A symbol-based compressed page consists of a memory block containing
 *  the positions and IDs of symbols appearing on that page plus internal
 *  information about the shapes of the symbols.  The internal information
 *  was generated by either compressing the page or by
 *  w_deserializeCompressionInfo from a bytestream of compressed shapes.
 *  This routine converts one block of symbol IDs and positions into one
 *  image.
 *
 * PrivateDescription:::
 *  This routine causes pictures of the small symbols to be cached so that
 *  it can simply blit them to the image as needed.  The pictures are
 *  created the first time this procedure is invoked, further invocations
 *  use the same pictures.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pTextCompInfo::  Pointer to private symbol-based compression data block.
 *		     This contains the shape information.
 *  pPositionBuf::   Pointer to the buffer containing symbol IDs and
 *		     positions.
 *
 * Output:::
 *  pPrD::	     Pointer to Pointer to destination pixr.  If *pPrD is NULL,
 *		     an output image will be created.  If *pPrD is not NULL,
 *		     the output image supplied will be used if it has the
 *		     proper dimensions.  If an error is encountered and
 *		     the dest image was allocted here, it is freed.
 *
 * Notes:::
 *  Here is some pseudo code showing a typical use of this routine to
 *  decompress an image.
 *<BR>
 *<BR>		* get a private data block *
 *<BR>	w_allocateTextCompressionInfo(..., &pTextCompInfo, ...);
 *<BR>
 *<BR>		* decompress the symbol shapes, keeping the result in
 *<BR>		* in the private data block.
 *<BR>	w_deserializeCompressionInfo(
 *<BR>				<ptr to bytes containing compression info>,
 *<BR>				pTextCompInfo, ...);
 *<BR>
 *<BR>	for (p = 0; p < maxPageNum; p++)
 *<BR>	{
 *<BR>	    <page p info> = yourRoutineForGettingImageSpace(...);
 *<BR>	    w_convertSymbolsToImage(..., <pointer to buffer containing
 *<BR>				    symbol IDs & positions for page p>,
 *<BR>				    pTextCompInfo, ...);
 *<BR>	    <do something with image of page p>;
 *<BR>	}
 *<BR>
 *<BR>	w_freeTextCompressionInfo(..., pTextCompInfo);
 *<BR>
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable, unless we change the format of compressed images.
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_convertSymbolsToImage (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    UInt8              *pPositionBuf,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    UInt32	 width, height;		/* width, height of compressed pixr */
    Int32	 dWidth, dHeight, dBpl;	/* dimensions of pixr passed in */
    UInt8	*pD;
    Bool	 allocatedHere = cFalse;
    Int32	 rightFrameColsToClear, bottomFrameRowsToClear;

	/* see if the pointers are valid */
    ERR_GEN((pTextCompInfo == NULL) || (pPositionBuf == NULL) || (pPrD == NULL),
	    cErrInvalidParm);

	/* get the width and height of the new pixr from the position info. */
    ERR_PASS (i_getCompressedImageDimensions (pIPWState, pPositionBuf,
					      &width, &height));


    prD = *pPrD;
    ERR_PASS (wp_createPixrConditional(pIPWState, width, height, 1, 0, 0,
				       cFrameBits, cTrue, &prD,
				       &rightFrameColsToClear,
				       &bottomFrameRowsToClear,
				       &allocatedHere));

	/* Things appear to be legal, try decoding the image. */
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prD, &dWidth));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prD, &dHeight));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &dBpl));
    
    ERR_PASS_CLEANUP( i_convertSymbolsToImage (pIPWState, pD,
					       dWidth, dHeight, dBpl,
					       pTextCompInfo, pPositionBuf));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return number of unique symbols in symbol-based compressed dictionary.
 *
 * LongDescription:::
 *  The essence of symbol-based compression is to convert a large number
 *  of blobs found in an image into a smaller number of unique symbols.
 *  These unique symbols can be used to reconstruct the page.  This routine
 *  returns the current number of unique symbols.  It is particularly
 *  useful when preparing to consolidate symbols so that the client can
 *  allocate the table used to map old symbol IDs to new ones.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pTextCompInfo::  Pointer to private symbol-based compression data block.
 *  symbolSet::	     Set of symbols to find count of: cSymbolsAll: all in
 *		     dictionary; cSymbolsMoreThanOnce: counts > 1;
 *		     cSymbolsOnceOnPage: counts = 1, same page ID.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  pNumOfSymbols::  Pointer to word into which number of symbols will be
 *		     written.
 *
 * Notes:::
 *  See description for i_convertImageToSymbols to see typical sequence
 *  of calls using this function to consolidate symbols.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getNumberOfSymbols (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    Int32		symbolSet,
    UInt32		pageRequest,
    UInt32             *pNumOfSymbols)
{
    PROLOGUE;

    ERR_GEN((pTextCompInfo == NULL) || (pNumOfSymbols == NULL),
	    cErrInvalidParm);

    ERR_PASS( i_getNumberOfSymbols (pIPWState, pTextCompInfo, symbolSet,
				    pageRequest, (Int32 *)pNumOfSymbols));

    NORMAL_RETURN;
}


/*

 $Log: tcdcomp.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.23  1998/08/05 23:35:08  aber
 aber on Wed Aug 5 16:35:01 PDT 1998

 tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v

 Fixed coreCheckC complaints.

 Revision 1.22  1998/08/04 00:21:27  aber
 tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v

 CommonCheck: Support multiple dictionaries when token compressing, used to retain
 single-use tokens in the same dictionary -- to be used when color
 tokenizing.

 Revision 1.21  1998/01/14 00:15:47  ddavies
 tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.20  1997/04/15 22:43:47  ddavies
 tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v

 Fix indentation.

 Revision 1.19  1996/01/04 20:27:44  ejaquith
 tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.18  1995/12/13  04:14:14  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:17:24  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/13  22:06:15  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/09/28  22:22:01  ddavies
 * ddavies on Thu Sep 28 15:22:01 PDT 1995
 *
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Fix comments at top of file.
 *
 * Revision 1.14  1995/09/28  19:19:29  ddavies
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Filled in autodoc comments.
 *
 * Revision 1.13  1995/09/21  23:00:40  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/15  17:59:59  ddavies
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.11  1995/09/11  22:10:43  ddavies
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.10  1995/09/10  00:32:46  ddavies
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.9  1995/09/07  01:36:10  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.8  1995/09/06  20:24:27  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.7  1995/09/01  01:19:20  ddavies
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.6  1995/08/31  21:40:03  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/22  02:22:03  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.4  1995/08/17  22:04:28  ejaquith
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.3  1995/07/29  04:52:34  ddavies
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.2  1994/11/23  05:58:23  ddavies
 * ddavies on Tue Nov 22 21:57:22 PST 1994
 *
 * tcdcomp.c -> /product/ipcore/ipwrap/src/RCS/tcdcomp.c,v
 *
 * Add image width and height to call to i_convertSymbolsToImage.  This
 * allows the frames to be cleaned up.
 *
 * Revision 1.1  1994/10/14  01:41:16  ddavies
 * Initial revision
 *

 */


