/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tcompr.c:
 *
 *	Code used to compress images containing text.  The image is
 *	encoded as repeated instances of shapes.  A compressed
 *	image has two parts: position information and shape information.
 *	A document may be composed of multiple blocks of position
 *	information and a single block of shape information.
 *	
 *		
 *
 *	w_estimateSizeOfPositionInfo():		estimates size of memory block
 *						needed for position information
 *						portion of compressed image.
 *
 *	w_estimateSizeOfSerializedCompressionInfo():
 *						returns size of memory block
 *						needed to hold size of
 *						compression info
 *
 *	w_convertImageToSymbols():		converts an image containing
 *						text into a set of encoded
 *						positions.  If given an
 *						existing set of symbol shapes,
 *						the set is enlarged as
 *						necessary.  If no dictionary
 *						is supplied, one is created.
 *
 *	w_serializeCompressionInfo():		converts internal compression
 *						info into a serial bytestream.
 *
 *	w_getSizeOfSymbolRemapArray():		returns size of array used
 *						to modify position buffer(s)
 *						after symbol consolidation.
 *
 *	w_consolidateShapeInfo():		merges similar symbols,
 *						pruning the shape tree as
 *						needed.  Also produces a
 *						map used to turn old symbol ID's
 *						into new ones.
 *
 *	w_remapSymbolIDs:			Uses the table produced by
 *						consolidateShapeInfo to
 *						convert the symbol IDs in a
 *						block of position info.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "pixr.h"
#include "geomadt.h"

/* PROTOTYPES */
#include "tcompr.pub"
#include "shrcmprs.pub"
#include "props.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: tcompr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* This file contains wrappers for routines used to compress images
 * containing text.
 *
 * A compressed file contains two parts.  There are one or more sections
 * of position information.  Each section specifies the position of each symbol
 * in an image.  There is a final section specifying the shapes of all
 * the symbols used.
 *
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of memory block to hold position info for compressed image.
 *
 * LongDescription:::
 *  When an image is compressed using symbol-based compression, the output
 *  comes in two parts: the shapes of the symbols and the positions of
 *  each symbol on a page.  This function estimates the size of the memory
 *  block needed to hold just the position information.
 *
 * PrivateDescription:::
 *  We use a weird heuristic to do this.  It's based on the rough observation
 *  that the number of bits in a Group4 file is roughly equal to the number
 *  of black bits in that file.  Since our output shouldn't be bigger than
 *  a Group 4 compressed version, that's our first guess.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  prS::            Pointer to image to be compressed.
 *
 * Output:::
 *  pSize::          Pointer to size to be returned.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateSizeOfPositionInfo (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt32             *pSize)
{
    PROLOGUE;
    UInt8    *pS;
    Int32     w, h, d, bpl;

    ERR_GEN(prS == NULL || pSize == NULL, cErrInvalidParm);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &d));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));

	/* just ask the ipshared routine to do its job */
    ERR_PASS( i_estimateSizeOfPositionInfo (pIPWState, pS, w, h,
					    d, bpl, pSize));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of memory block to hold shape info for compressed image.
 *
 * LongDescription:::
 *  When an image is compressed using symbol-based compression, the output
 *  comes in two parts: the shapes of the symbols and the positions of
 *  each symbol on a page.  This function estimates the size of the memory
 *  block needed to hold just the encoded shapes.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfo::  Pointer to private state block used by symbol based
 *		     compression and decompression.
 *
 * Output:::
 *  pSize::          Pointer to size to be returned.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateSizeOfSerializedCompressionInfo (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    UInt32             *pSize)
{
    PROLOGUE;

    ERR_GEN(pTextCompInfo == NULL || pSize == NULL, cErrInvalidParm);

	/* just ask the ipshared routine to do its job */
    ERR_PASS( i_estimateSizeOfSerializedCompressionInfo (pIPWState,
							 pTextCompInfo,
						         pSize));

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compresses single image using symbol-based compression.
 *
 * LongDescription:::
 *  This routine attempts to find a match between each connected component
 *  in an image and a connected component in a dictionary.  If a match is
 *  found, the identity of the dictionary token and the position of the
 *  image token are recorded in an output file.  If the connected component
 *  from the image is not found in the dictionary, the connected component
 *  is entered into the dictionary.
 *
 * PrivateDescription:::
 *  This routine uses novel comparison techniques based on run lengths
 *  instead of using bitmaps.  There is an enormous comment describing
 *  the approach in detail inside ipshared/src/i_tcompr.c.
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  prS::		Pointer to image to be compressed.
 *  pTextCompInfo::	Pointer to private state block used by symbol based
 *			compression and decompression.
 *  positionBufSize::	Size of output buffer used to hold positions and
 *			symbol ID in bytes.  If *ppPositionBuf is NULL,
 *			positionBufSize is ignored.
 *  exactMatch::	Set to cTrue if symbols should only be put in the
 *			same equivalence class if they match exactly.
 *
 * Output:::
 *  ppPositionBuf::	Pointer to Pointer to memory block that will contain
 *			the encoded shape ID's and position info.
 *			If *ppPositionBuf is NULL, a position buffer is
 *			allocated.
 *  pResultSize::	Pointer to returned number of bytes that have actually
 *			been used in the position buffer.
 *
 * Returns:::
 *  In addition to the normal return codes, cErrBufferExceeded is returned
 *  when the routine attempted to overrun the position buffer.  The client
 *  should allocate a larger position buffer and try again.  If the position
 *  buffer was allocated here and this error was encountered, it will be
 *  freed before returning.  Clients can use w_estimateSizeOfPositionInfo to
 *  know how big the original buffer was.
 *
 * Notes:::
 *  This is one of a set of routines used to compress images containing text.
 *  The general sequence of calls is:
 *<BR>
 *<BR>	w_allocateTextCompressionInfo(..., &pTextCompInfo, ...)
 *<BR>
 *<BR>	    * If you want to use an existing set of symbols do this step.
 *<BR>	w_deserializeCompressionInfo(pTextCompInfo, ...);
 *<BR>
 *<BR>	for (p = 0; p < maxPageNum; p++)
 *<BR>	{
 *<BR>	    w_estimateSizeOfPositionInfo(<page p image info>, &positionBufSize);
 *<BR>	    positionBuf = yourMemoryAllocator(positionBufSize);
 *<BR>	    w_convertImageToSymbols(<page p image info>, pTextCompInfo,
 *<BR>				    positionBuf, positionBufSize, ...);
 *<BR>	    <shrink positionBuf down to size used if desired>
 *<BR>	}
 *<BR>
 *<BR>	     * Now that we've seen all versions of each symbol,
 *<BR>	     * consolidate the versions.  Note that even if we started
 *<BR>	     * with a set of symbols, the classification process
 *<BR>	     * probably added some, so we need to write out the
 *<BR>	     * shape info again.  First, get the number of symbols.
 *<BR>	w_getSizeOfSymbolRemapArray(pTextCompInfo, &sizeOfRemapArray);
 *<BR>  <allocate remapTable>
 *<BR>	w_consolidateShapeInfo(..., pTextCompInfo, &remapTable, ...);
 *<BR>	for (p = 0; p < maxPageNum; p++)
 *<BR>	{
 *<BR>		* this step is optional, you can do the remapping in place *
 *<BR>	    newPositionBuf = yourMemoryAllocator(size of old position Buf p +
 *<BR>						 cTCPositionBufferSlop);
 *<BR>	    w_remapSymbolIDs(pTextCompInfo, <page p positionBuf>,
 *<BR>				newPositionBuf, remapTable, ...);
 *<BR>	    <free page p's old position buf>
 *<BR>	    <write new position buf data to file?>
 *<BR>	}
 *<BR>
 *<BR>  <free remapTable>
 *<BR>
 *<BR>		* Now we're ready to write out all the symbol shapes.  Get
 *<BR>		* the size of the buffer required to hold the encoded shapes.
 *<BR>	w_estimateSizeOfSerializedCompressionInfo(pTextCompInfo,
 *<BR>						  &compInfoBufSize);
 *<BR>      compInfoBuf = yourMemoryAllocator(compInfoSize);
 *<BR>
 *<BR>		* And serialize all the internally stored shapes into
 *<BR>		* this buffer.
 *<BR>	w_serializeCompressionInfo(..., pTextCompInfo, compInfoBuf,
 *<BR>				   compInfoBufSize, ...);
 *<BR>	<write serialized compression info to file?>
 *<BR>
 *<BR>	w_freeTextCompressionInfo(pTextCompInfo);
 *
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress, this will always be true.
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  Dan Huttenlocher's Hausdorf based compression stuff.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_convertImageToSymbols (
    IPWState           *pIPWState,
    PIXR               *prS,
    void               *pTextCompInfo,
    UInt32              positionBufSize,
    Bool		exactMatch,
    UInt8             **ppPositionBuf,
    UInt32             *pResultSize)
{
    PROLOGUE;
    UInt32           bufAllocatedHere;
    UInt8           *pS;
    Int32            w, h, bpl;

    ERR_GEN(prS == NULL || pTextCompInfo == NULL || ppPositionBuf == NULL ||
	    pResultSize == NULL, cErrInvalidParm);
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if no buffer was passed in, allocate one */
    if (*ppPositionBuf == NULL)
    {
	ERR_PASS( w_estimateSizeOfPositionInfo(pIPWState, prS, &positionBufSize));

	*ppPositionBuf = (UInt8 *)IPW_MALLOC(positionBufSize, cAllocatorData);
	ERR_GEN(*ppPositionBuf == NULL, cErrNomem);

	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &h));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpl));

	/* just ask the correct ipshared routine to do its job */
    if (exactMatch)
	ERR_PASS_CLEANUP( i_convertImageToSymbolsExactly (pIPWState,
							  pS, w, h, bpl,
							  pTextCompInfo,
							  0,
							  *ppPositionBuf,
							  positionBufSize,
							  pResultSize));
    else
	ERR_PASS_CLEANUP( i_convertImageToSymbols (pIPWState, pS, w, h, bpl,
						   pTextCompInfo,
						   0,
						   *ppPositionBuf,
						   positionBufSize,
						   pResultSize));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (bufAllocatedHere)
	IPW_FREE(*ppPositionBuf, cAllocatorData);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert text shapes from internal representation to a bytestream.
 *
 * LongDescription:::
 *  After using symbol based compression on one or more pages of text, we
 *  are left with a bunch of memory blocks containing position info and
 *  an internal description of symbol shapes.  This internal description
 *  is suited for searching by w_convertImageToSymbols.  This routine
 *  converts it into a serialized bytestream, compressing each shape,
 *  suitable for inclusion in a compressed data file.
 *  w_deserializeCompressionInfo converts the bytestream back into the
 *  internal format for further compression or for decompression.
 *
 * PrivateDescription:::
 *  The symbols are encoded using a private variant of group 4 encoding.
 *  In this form, we encode from right to left, bottom to top (because that's
 *  the way the symbol's runs are stored) and we wrap around from one line
 *  to the next instead of encoding the distance from the last run in a line
 *  to the end of the line and the distance from the beginning of a line
 *  to the first run.
 *
 * Input:::
 *  pIPWState::       Pointer to standard IPWrap state block
 *  pTextCompInfo::   Pointer to private state block used by symbol based
 *		      compression and decompression.
 *  bufSize::         Size in bytes of buffer into which encoded symbol
 *		      shapes will be written.  If *ppShapeInfoBuf == NULL,
 *		      bufSize is ignored.
 *
 * Output:::
 *  ppShapeInfoBuf::  Pointer to Pointer to buffer that will hold encoded
 *		      shape information.  If *ppShapeInfoBuf == NULL, a
 *		      shape buffer will be allocated here.  If an error
 *		      occurs and the buffer is allocated here, it will be
 *		      freed here.  Client's are responsible for freeing
 *		      all buffers.
 *  pBytesUsed::      Pointer to the number of bytes actually used in
 *		      the shape buffer.  This is returned by this routine.
 *
 * Returns:::
 *  In addition to the normal return codes, cErrBufferExceeded is returned
 *  when the routine attempted to overrun the shape buffer.  The client
 *  should allocate a larger shape buffer and try again.  If the shape
 *  buffer was allocated here and this error was encountered, it will be
 *  freed before returning.  Clients can use
 *  w_estimateSizeOfSerializedCompressionInfo to know how big the
 *  original buffer was.
 *
 * Notes:::
 *  See the description of w_convertSymbolsToImage to see a typical
 *  sequence of calls used to compress one or more pages of text.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable, well, except we might decide to encode the symbol shapes
 *  some other way entirely.  The correct format for a symbol-based
 *  compressed image is still under debate.
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_serializeCompressionInfo (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    UInt32              bufSize,
    UInt8             **ppShapeInfoBuf,
    UInt32             *pBytesUsed)
{
    PROLOGUE;
    Bool  bufAllocatedHere = cFalse;
    Bool  success = cFalse;

    ERR_GEN(pTextCompInfo == NULL || ppShapeInfoBuf == NULL ||
	    pBytesUsed == NULL, cErrInvalidParm);

	/* if no buffer was passed in, allocate one */
    if (*ppShapeInfoBuf == NULL)
    {
	ERR_PASS( w_estimateSizeOfSerializedCompressionInfo(pIPWState,
							 pTextCompInfo,
							 &bufSize));

	*ppShapeInfoBuf = (UInt8 *)IPW_MALLOC(bufSize, cAllocatorData);
	ERR_GEN(*ppShapeInfoBuf == NULL, cErrNomem);

	bufAllocatedHere = cTrue;
    }

	/* try to serialize the compression info.  If the shape
	 * buffer isn't big enough, make it bigger and
	 * try again.
	 */
    while (!success)
    {
	success = cTrue;	/* assume the best */
	    /* go do the serialization, do NOT delete the
	     * residual symbols (those that only occur once.
	     */
	ERR_CHECK_NO_MSG( i_serializeCompressionInfo (pIPWState, pTextCompInfo,
						      *ppShapeInfoBuf,
						      bufSize,
						      cFalse,
						      pBytesUsed))
	    /* We get here if i_serializeCompressionInfo failed
	     * for some reason.  If it failed because the buffer
	     * was too small AND that buffer was allocated here,
	     * try making it bigger.
	     */
	{
	    if ((__error__ == cErrBufferExceeded) && bufAllocatedHere)
	    {
		    /* the buffer was too small, try allocating a
		     * larger one and try again.
		     */
		bufSize += bufSize/2;
		*ppShapeInfoBuf = (UInt8 *)IPW_REALLOC(*ppShapeInfoBuf,
							bufSize,
							cAllocatorData);
		ERR_GEN_CLEANUP(*ppShapeInfoBuf == NULL, cErrNomem);
		success = cFalse;
		__error__ = cSuccessful;	/* try again */
	    }
	    else	/* it was some other error, so quit now */
	    {
	        ERR_GEN_CLEANUP(1, __error__);
	    }
	}
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    if (bufAllocatedHere)
	IPW_FREE(*ppShapeInfoBuf, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get size of table used by w_remapSymbolIDs
 *
 * LongDescription:::
 *  There are two steps used to compress an image using symbol-based
 *  compression.  w_convertSymbolsToImage does a first pass through the
 *  image.  It tries to quickly convert the large number of symbols found
 *  on a page into a much smaller number of tokens.  w_consolidateShapeInfo
 *  then tries to consolidate these tokens into a smaller number of tokens.
 *  w_remapSymbolIDs is used to modify the internal position buffers,
 *  replacing the unconsolidated symbols with the consolidated ones.
 *  This routine returns the size of an array that's allocated by the
 *  client, filled in by w_consolidateShapeInfo, used by
 *  w_remapSymbolIDs and freed by the client.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state.
 *  pTextCompInfo::  Pointer to private state block used by symbol based
 *		     compression and decompression.
 *
 * Output:::
 *  pSymIDMapSize::  Pointer to returned size of the remapping array.
 *		     It is the client's responsibility to allocate this
 *		     array.  The number of old symbols is found using 
 *		     w_getNumberOfSymbols.  The client should free this
 *		     table after using it to call w_remapSymbolIDs.
 *
 * Notes:::
 *  See the description of w_convertSymbolsToImage to see a typical
 *  sequence of calls used to compress one or more pages of text.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress, like w_convertSymbolsToImage, this can always be improved
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getSizeOfSymbolRemapArray (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    UInt32             *pSymIDMapSize)
{
    PROLOGUE;

    ERR_GEN((pTextCompInfo == NULL) ||
	    (pSymIDMapSize == NULL), cErrInvalidParm);

    ERR_PASS( i_getSizeOfSymbolRemapArray (pIPWState, pTextCompInfo,
					   (Int32 *)pSymIDMapSize));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Consolidate symbols produced by w_convertImageToSymbols.
 *
 * LongDescription:::
 *  There are two steps used to compress an image using symbol-based
 *  compression.  w_convertSymbolsToImage does a first pass through the
 *  image.  It tries to quickly convert the large number of symbols found
 *  on a page into a much smaller number of tokens.  This routine then
 *  tries to consolidate these tokens into a smaller number of tokens,
 *  spending more time on each token than w_convertSymbolsToImage can
 *  afford to do.
 *
 * PrivateDescription:::
 *  The basic idea is to shift each token a little inside its bounding
 *  box and see if any other symbols now match it.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state.
 *  pTextCompInfo::  Pointer to private state block used by symbol based
 *		     compression and decompression.
 *
 * Output:::
 *  pSymIDMap::      Pointer to map that can be used to convert an old symbol
 *		     number into a new symbol number.  It is the client's
 *		     responsibility to allocated this map.  The size of the
 *		     map is found using w_getSizeOfSymbolRemapArray.
 *		     The client should free this table after using it
 *		     to call w_remapSymbolIDs.
 *
 * Notes:::
 *  See the description of w_convertSymbolsToImage to see a typical
 *  sequence of calls used to compress one or more pages of text.
 *<BR><BR>
 *  As part of consolidating the list of symbols, the shape trees inside
 *  the private data structures are pruned here.
 *<BR><BR>
 *  One would probably NOT want to consolidate the shape information after
 *  compressing each page of text since that would cause the algorithm to
 *  keep rediscovering the common variants of the basic shapes on each page.
 *  Consolidating once after all pages have been seen by w_convertImageToSymbols
 *  makes more sense.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress, like w_convertSymbolsToImage, this can always be improved
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_consolidateShapeInfo (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    void               *pSymIDMap)
{
    PROLOGUE;

    ERR_GEN((pTextCompInfo == NULL) || (pSymIDMap == NULL), cErrInvalidParm);

    ERR_PASS( i_consolidateShapeInfo (pIPWState, pTextCompInfo, pSymIDMap));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Changes symbol IDs in compressed position info using remap table.
 *
 * LongDescription:::
 *  After w_convertImageToSymbols has seen some group of pages, further
 *  compression can be achieved by consolidating many small variants of
 *  symbols into a smaller set of symbols.  w_consolidateShapeInfo does
 *  this.  As part of its job, it re-assigns the symbol IDs so that the
 *  new set is contiguous from 0 again.  This saves some space in the
 *  output.  w_consolidateShapeInfo fills in a table that can be used
 *  to convert old symbol numbers to new ones.  That table is used by
 *  this routine to convert the old symbol IDs encoded in bodies of
 *  position information by w_convertImageToSymbols into the new symbol IDs.
 *
 * PrivateDescription:::
 *  It is important that in every case, the new symbol ID has the same number
 *  by bytes or fewer.  This allows us to always make the new position buffers
 *  shorter, which means we can avoid writing over the top of some position
 *  info we'll need to read later.  This is important if the position buffer
 *  is being converted in place.
 *
 * Input:::
 *  pIPWState::           Pointer to standard IPWrap state.
 *  pTextCompInfo::	  Pointer to private state block used by symbol based
 *		   	  compression and decompression.
 *  pSymIDMap::           Pointer to map used to convert old symbol IDs
 *			  into new ones.  This is allocated by the client
 *			  using the size from w_getSizeOfSymbolRemapArray
 *			  and filled in by w_consolidateShapeInfo.
 *  pOldPositionBuf::     Pointer to old position buffer.
 *  oldPositionBufSize::  Number of bytes in old position buffer.
 *  pNewPositionBuf::     Pointer to new position buffer.  If this is equal
 *			  to pNewPositionBuf, the conversion will be done
 *			  in place.
 *  newPositionBufSize::  Number of bytes in new position buffer.
 *
 * Output:::
 *  pBytesUsed::	  Pointer to a word that will contain the number
 *			  of bytes in the new position buffer that were
 *			  actually used.
 *
 * Notes:::
 *  See the description of w_convertSymbolsToImage to see a typical
 *  sequence of calls used to compress one or more pages of text.
 *<BR><BR>
 *  We don't need the TextCompInfo block in this operation because it
 *  doesn't contain anything we need here.  This operation figures out
 *  how long the old position block is by parsing it.
 *<BR><BR>
 *  If the input and output position blocks are the same, lower level
 *  code allocates a temporary buffer, puts the new position block info
 *  into it and copies the new info into the input/output buffer.  This
 *  is done because we cannot guarantee that the new symbol ID for every
 *  symbol is no longer than the old symbol ID for that symbol.  This
 *  was true in an earlier version of the code, but not anymore.  The
 *  new code assigns symbol IDs by usage so the heavily used symbols
 *  always get shorter IDs.  Therefore, a symbol that was discovered
 *  early (got a short ID), but is seldom used, will get a longer ID
 *  in the new world.
 *<BR><BR>
 *  In the event that the new position buffer cannot hold the remapped
 *  position data, this routine will return cErrBufferExceeded.  If this
 *  happens, the client should provide a larger output buffer and try
 *  again.  If the new and old position buffers are the same, the contents
 *  of the old position buffer will not have been changed.
 *
 * FunctionCategory:::
 *  compression
 *
 * Keywords:::
 *  compression, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable, unless we decide on some entirely different format for encoded
 *  output.
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_remapSymbolIDs (
    IPWState           *pIPWState,
    void               *pTextCompInfo,
    UInt8              *pOldPositionBuf,
    UInt32		oldPositionBufSize,
    void               *pSymIDMap,
    UInt8              *pNewPositionBuf,
    UInt32		newPositionBufSize,
    UInt32             *pBytesUsed)
{
    PROLOGUE;

    ERR_GEN((pOldPositionBuf == NULL) || (pSymIDMap == NULL) ||
	    (pNewPositionBuf == NULL || pBytesUsed == NULL),
	    cErrInvalidParm);

    ERR_CHECK_NO_MSG(i_remapSymbolIDs (pIPWState, pTextCompInfo,
				       pSymIDMap, pOldPositionBuf,
				       oldPositionBufSize, pNewPositionBuf,
				       newPositionBufSize, pBytesUsed))
    {
	    /* cErrBufferExceeded isn't a disaster, it's a signal to
	     * to try again with a larger buffer.
	     */
	ERR_GEN(__error__ != cErrBufferExceeded, __error__);
    }

	/* return the error code in any case */
    NORMAL_RETURN;
}


/*

 $Log: tcompr.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.31  1998/08/04 00:21:28  aber
 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 CommonCheck: Support multiple dictionaries when token compressing, used to retain
 single-use tokens in the same dictionary -- to be used when color
 tokenizing.

 Revision 1.30  1998/01/14 00:15:49  ddavies
 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.29  1997/04/29 02:52:09  ddavies
 ddavies on Mon Apr 28 19:52:01 PDT 1997

 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 Add exactMatch parameter to w_convertImageToSymbols.  If set, lossless
 matching is used.  Otherwise, lossy matching is used.

 Revision 1.28  1997/02/12 03:17:03  ddavies
 ddavies on Tue Feb 11 19:15:01 PST 1997

 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 Add implementation of w_getSizeOfSymbolRemapArray.  This is used by clients
 to get the size of the data structure used that's filled in by
 w_consolidateShapeInfo and used by w_remapSymbolIDs.
 Also change the type of pSymIDMap from pointer to UInt32 to pointer
 to void whereever it is found.  Update comments appropriately.

 Revision 1.27  1997/01/08 04:31:30  ddavies
 ddavies on Tue Jan 7 20:30:32 PST 1997

 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 Add newPositionBufSize parameter to w_remapSymbolIDs.  The position buffer
 doesn't always get smaller, so we have to check to see if it overflows.
 Also don't print out an error message if i_remapSymbolIDs returns
 cErrBufferExceeded since this just means the client should try again
 with a larger buffer.

 Revision 1.26  1996/07/19 00:36:03  ddavies
 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 Set the new deleteResidual arg to cFalse when serializing ALL the
 Text compression information.

 Revision 1.25  1996/07/04 03:10:55  ddavies
 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 Call i_serializeCompressionInfo with the new ERR_CHECK_NO_MSG macro so we can
 handle buffer overflow (alloc a bigger buffer) silently.
 Add oldPositionSize arg to w_remapSymbolIDs.  This is the size of the
 input buffer.  It should have been there all along anyway.

 Revision 1.24  1996/01/04 20:27:47  ejaquith
 tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.23  1995/12/13  04:14:16  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.22  1995/11/14  22:18:50  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Fix documentation bugs found by Eric's scripts.
 *
 * Revision 1.21  1995/11/13  20:17:27  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.20  1995/10/13  22:06:18  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/10/06  18:32:17  ddavies
 * ddavies on Fri Oct 6 11:30:09 PDT 1995
 *
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Fix comments for remapSymbolIDs - pTextCompInfo isn't an argument for
 * that routine.  Also have w_serializeCompressionInfo keep trying with
 * larger buffers if i_serializeCompressionInfo says the buffer for the
 * serialized compression info wasn't large enough.
 *
 * Revision 1.18  1995/09/28  19:19:26  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Filled in autodoc comments.
 *
 * Revision 1.17  1995/09/21  23:00:42  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/15  18:00:03  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.15  1995/09/11  22:10:46  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.14  1995/09/10  00:32:50  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.13  1995/09/07  01:36:13  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:24:31  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  01:19:26  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:40:06  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/17  22:04:21  ejaquith
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.8  1995/07/29  04:53:09  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.7  1995/05/09  03:19:56  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Remove inclusion of io.pub since it was superfluous.
 *
 * Revision 1.6  1994/11/10  21:56:24  ddavies
 * ddavies on Thu Nov 10 13:54:40 PST 1994
 *
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Try a method for removing noise.  It was slow and didn't work very well, but the
 * commented-out code is still here in case we want to think of something similar
 * in the future.
 *
 * Revision 1.5  1994/10/14  01:39:38  ddavies
 * ddavies on Thu Oct 13 18:36:07 PDT 1994
 *
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Add inclusion of tcompr.pub, props.pub, utils.pub.  Remove tcpros.h.
 * Remove convertSymbolsToImage, this will go to tcdcomp.c.  Add routines
 * serializeCompressionInfo.  Allow convertImageToSymbols to allocate the
 * position buffer if one is not passed in, otherwise use the one passed in.
 *
 * Revision 1.4  1994/09/07  23:38:40  ddavies
 * ddavies on Wed Sep 7 16:35:30 PDT 1994
 *
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Added implementations of convertSymbolsToImagePixr and convertSymbolsToImage.
 * The former calls the latter.  These are used to convert the internal
 * representation of a compressed image into an image.  Also added implementation
 * of getNumberOfSymbols, which returns the number of symbols in a compressed
 * image.  This is used to allocate buffers.  Another new procedure,
 * consolidateShapeInfo is used to reduce the symbol dictionary by getting
 * rid of redundant symbols.  Finally, remapSymbolIDs takes a pointer to
 * the old position info and a table used to make old symbol IDs to new ones.
 * It produces a new block of position info.
 *
 * Revision 1.3  1994/07/14  21:09:07  ddavies
 * ddavies on Thu Jul 14 14:08:30 PDT 1994
 *
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Snapshot of wrapper to text compression code.  Just made changes to keep
 * up with interface changes.
 *
 * Revision 1.2  1994/05/31  17:57:49  ddavies
 * tcompr.c -> /product/ipcore/ipwrap/src/RCS/tcompr.c,v
 *
 * Change call to i_freeTextCompressionInfo to eliminate address of memory
 * IPW_FREEing routine.
 *
 * Revision 1.1  1994/04/26  02:36:52  ddavies
 * Initial revision
 *

 */

