/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tctnd.c:
 *
 *	Code used to read symbol-based compression info in Tundra
 *	format.  These routines parse memory blocks containing
 *	Tundra data.  It's the client's responsibility to do file I/O.
 *
 *	w_readTundraHeader
 *		parse a Tundra file header.
 *
 *	w_readTundraDictHeader
 *		Reads a Tundra dictionary header.
 *
 *	w_readTundraDictionaryData
 *		convert the data portion of a Tundra symbol dictionary
 *		into internal format.
 *
 *	w_readTundraDictionaryBlock
 *		convert an entire Tundra symbol dictionary block (header +
 *		data) into internal format.
 *
 *	w_readTundraNumberOfPagesField
 *		read the Tundra  "number of pages" field.
 *
 *	w_readTundraPageHeaders
 *		reads the page and residual headers in a Tundra page block.
 *
 *	w_decompressTundraPositionBlock
 *		parse the info contained in a single
 *		Tundra symbol position block.  Convert this into a bitmap.
 *
 *	w_decompressTundraResidualBlock
 *		parse the info contained in a single
 *		Tundra residual image block.  Convert this into a bitmap.
 *
 *	w_decompressTundraPageBlock
 *		parse the info contained in a single
 *		Tundra page (symbol position + residual image).
 *		Convert this into a bitmap.
 *
 *	w_readTundraDocument
 *		A convenience routine.  Uses the routines above to parse
 *		a Tundra formatted block containing a header, dictionary
 *		and one or more pages.  Loads the dictionary and returns
 *		an array of pointers to the starts of the pages.  These
 *		can be decompressed using w_decompressTundraPageBlock
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "geomadt.h"

/* PROTOTYPES */
#include "tcompr.pub"
#include "shrcmprs.pub"
#include "props.pub"
#include "props.prv"


IP_RCSINFO(RCSInfo, "$RCSfile: tctnd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* This file contains wrappers for routines used to parse Tundra format
 * data.  The data is contained in memory, the client is responsible for
 * doing file I/O.
 *
 * A compressed file contains four main parts.  It begins with a file
 * header.  Following that is a dictionary.  It is theoretically possible
 * to have multiple dictionaries in a file, but we do not support that
 * at this time.  A dictionary contains the pictures of the symbols.
 * The dictionary is followed by a few bytes containing the number of
 * pages to follow.  Finally, each page's individual info (symbol positions
 * and residual image) are at the end.
 *
 * The first routines listed above each parse one section of the
 * Tundra block.  Clients may use these individually, especially when
 * reading files with a large number of pages.  The last routine,
 * w_readTundraDocument is a convenience routine that reads the header
 * dictionary and number of pages field.  It then finds offsets relative
 * to the beginning of the file for the starting position of each
 * page block.  Clients may use w_decompressTundraPageBlock to decompress
 * each page.  This seems preferable to asking the client to supply
 * bitmaps for all the pages at once.
 */


#define cDefaultPageNameLength	256	/* characters */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Parse a Tundra file header.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header.  This routine parses an existing header.  If the
 *  block of memory pointed to is not a Tundra header, an error
 *  code is returned.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pHeader::	     Pointer to Tundra header block.
 *  bufSize::	     Number of bytes remaining in buffer containing header.
 *
 * Output:::
 *  pBytesRead::     Pointer to returned number of bytes read.
 *
 * Returns:::
 *	If the buffer is not big enough to contain the header, returns
 *	cErrBufferExceeded.  If the block does not begin with a header,
 *	returns cErrDecodeFailure.  If the version of the file is
 *	not supported, returns cErrBadVersion.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_readTundraHeader (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pHeader,
    UInt32              bufSize,
    UInt32	       *pBytesRead)
{
    PROLOGUE;
    UInt8	version;
    Bool	versionIsOk;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pHeader == NULL) ||
	    (pBytesRead == NULL), cErrNullPtr);

	/* Go read the header */
    ERR_PASS(i_readTundraHeader (pIPWState, pTextCompInfoIn,
				 pHeader, bufSize, &version,
				 pBytesRead));

	/* See if the version is legal */
    ERR_PASS(i_checkTundraVersion (pIPWState, pTextCompInfoIn,
				   version, &versionIsOk));

	/* Die if it's a bad version */
    ERR_GEN(versionIsOk != cTrue, cErrBadVersion);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in a Tundra dictionary header.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  The dictionary is itself composed of a header and data.
 *  This routine reads the dictionary header, returning the
 *  data contained in it.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pDictBuf::	     Pointer to buffer containing Tundra dictionary header.
 *  bufSize::	     the number of bytes remaining in the dictionary buffer.
 *
 * Output:::
 *  pDictLength::    Pointer to returned number of bytes in dictionary data
 *		     buffer.
 *  pBytesRead::     The number of Tundra bytes read
 *
 * Returns:::
 *	If an insufficient dictionary buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_readTundraDictionaryHeader(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pDictBuf,
    UInt32		bufSize,
    UInt32	       *pDictLength,
    UInt32	       *pBytesRead)
{
    PROLOGUE;
    UInt8	*pLocalBuf = pDictBuf;	/* current buf pointer */
    UInt32	 bytesRead;

	/* Test the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pDictBuf == NULL) ||
	    (pDictLength == NULL) || (pBytesRead == NULL), cErrNullPtr);

	/* go read the dictionary length in bytes */
    ERR_PASS( i_readTundraDictionaryBlockLengthField (pIPWState,
						      pTextCompInfoIn,
						      pLocalBuf,
						      bufSize,
						      pDictLength,
						      &bytesRead) );

	/* Ducky, it's been read.  Compute the number of bytes read
	 * and return */
    *pBytesRead = bytesRead;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in a Tundra dictionary data.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  The dictionary is itself composed of a header and data.
 *  Since this routine does not read any headers, it can be used
 *  for non-Tundra files, like XIFF.  When reading a real Tundra
 *  dictionary block, the client will probably find
 *  w_readTundraDictionaryBlock more convenient.
 *  This routine reads the dictionary data converting it into the
 *  internal data structures used for compression or decompression.
 *  The client must specify whether the symbols will subsequently
 *  be used for compression or decompression as the internal data
 *  structures are different.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  dictNumber::     The dictionary number to be associated with all
 *		     symbols in this dictionary.
 *  forDecompression:: Boolean that should be set if the symbols will
 *		     subsequently be used ONLY for decompression.  If
 *		     They will NOT be used for decompression (perhaps
 *		     for further compression), set this to cFalse.
 *  pDictBuf::	     Pointer to buffer containing Tundra dictionary data.
 *  bufSize::	     the number of bytes remaining in the dictionary buffer.
 *
 * Returns:::
 *	If an insufficient dictionary buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_readTundraDictionaryData(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32		dictNumber,
    Bool                forDecompression,
    UInt8	       *pDictBuf,
    UInt32		bufSize)
{
    PROLOGUE;
    UInt8	*pLocalBuf = pDictBuf;	/* current buf pointer */

	/* Test the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pDictBuf == NULL), cErrNullPtr);

    ERR_PASS( i_readTundraDictionaryBlock (pIPWState, pTextCompInfoIn,
					   pLocalBuf,
					   bufSize,
					   dictNumber,
					   forDecompression) );

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in a Tundra dictionary block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  A dictionary is composed of a header and data.
 *  This routine reads the entire dictionary block (header and data)
 *  into internal data structures.  Sioce the header is Tundra
 *  depenent (it doesn't appear in a XIFF file, for example), this
 *  routine should only be used when reading Tundra files.
 *  The client must specify whether the symbols will subsequently
 *  be used for compression or decompression as the internal data
 *  structures are different.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  dictNumber::     The dictionary number to be associated with all
 *		     symbols in this dictionary.
 *  forDecompression:: Boolean that should be set if the symbols will
 *		     subsequently be used ONLY for decompression.  If
 *		     They will NOT be used for decompression (perhaps
 *		     for further compression), set this to cFalse.
 *  pDictBuf::	     Pointer to memory block holding Tundra dictionary bytes.
 *  bufSize::	     the number of bytes remaining in the dictionary buffer.
 *
 * Output:::
 *  pBytesRead::     The number of Tundra bytes read
 *
 * Returns:::
 *	If an insufficient dictionary buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_readTundraDictionaryBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32		dictNumber,
    Bool                forDecompression,
    UInt8	       *pDictBuf,
    UInt32		bufSize,
    UInt32	       *pBytesRead)
{
    PROLOGUE;
    UInt32	 dictionaryBlockLength;
    UInt8	*pLocalBuf = pDictBuf;	/* current buf pointer */
    UInt32	 bytesRead;

	/* Test the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pDictBuf == NULL) ||
	    (pBytesRead == NULL), cErrNullPtr);

	/* go read the dictionary length in bytes */
    ERR_PASS( i_readTundraDictionaryBlockLengthField (pIPWState,
						      pTextCompInfoIn,
						      pLocalBuf,
						      bufSize,
						      &dictionaryBlockLength,
						      &bytesRead) );

	/* Fine, adjust the pointers and read the dictionary */
    bufSize -= bytesRead;
    pLocalBuf += bytesRead;

    ERR_PASS( i_readTundraDictionaryBlock (pIPWState, pTextCompInfoIn,
					   pLocalBuf,
					   dictionaryBlockLength,
					   dictNumber,
					   forDecompression) );

	/* Ducky, it's been read.  Compute the number of bytes read
	 * and return */
    *pBytesRead = bytesRead + dictionaryBlockLength;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a Tundra "number of pages" field.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine reads in the "number of pages" field.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pNumOfPagesField:: Pointer to the Tundra field.
 *  bufSize::	     Number of bytes remaining in buffer containing "number
 *		     of pages" field.
 *
 * Output:::
 *  pNumOfPages::    Pointer to returned contents of "number of pages" field.
 *  pBytesRead::     Pointer to returned number of bytes actually read.
 *
 * Returns:::
 *	If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_readTundraNumberOfPagesField (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pNumOfPagesField,
    UInt32              bufSize,
    UInt32	       *pNumOfPages,
    UInt32	       *pBytesRead)
{
    PROLOGUE;

	/* Test the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pNumOfPagesField == NULL) ||
	    (pNumOfPages == NULL) || (pBytesRead == NULL), cErrNullPtr);

	/* Go read the field */
    ERR_PASS( i_readTundraNumberOfPagesField (pIPWState, pTextCompInfoIn,
					      pNumOfPagesField,
					      bufSize,
					      pNumOfPages,
					      pBytesRead));

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read Tundra Page and Residual Headers.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  A page is in turn composed of a page header, a position block,
 *  a residual header and a residual block.  This routine reads the
 *  page and residual headers contained in a page.  This allows
 *  the client to know what there is to know about the page including
 *  the positions of the component parts.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pPageBuf::	     Pointer to memory block holding Tundra Page header.
 *  bufSize::	     the number of bytes remaining in the dictionary buffer.
 *  pageNameBufLen:: if *ppPageName is not NULL, this is the length of
 *		     the buffer that will hold the page name.
 *
 * Output:::
 *  ppPageName::     Pointer to pointer to string containing page name.
 *		     if *ppPageName is not NULL, a string allocated by
 *		     the client is used.  Otherwise, a string of the
 *		     correct length is allocated.
 *  pPageNameLen::   Pointer to real length of the page name (not including
 *		     the nul at the end of the string, compatible with strlen).
 *  pPageWidth::     Pointer to returned page width in pixels (not including
 *		     frame).
 *  pPageHeight::    Pointer to returned page height in pixels (not including
 *		     frame).
 *  pPageHeaderBytes:: Ptr to returned number of bytes in the encoded
 *		     Tundra page header.
 *  pPosBlockBytes:: Ptr to returned number of bytes needed to encode
 *		     the Tundra position block.
 *  pResX::	     Ptr to returned X coord (relative to dest image) of
 *		     residual block.
 *  pResY::	     Ptr to returned Y coord (relative to dest image) of
 *		     residual block.
 *  pResW::	     Ptr to returned width of residual block.
 *  pResH::	     Ptr to returned height of residual block.
 *  pResHeaderBytes:: Ptr to returned number of bytes in the encoded
 *		     Tundra residual header.
 *  pResBlockBytes:: Ptr to returned number or bytes needed to encode
 *		     the Tundra residual block.
 *
 * Returns:::
 *	If the page buffer doesn't hold at least the page header, the
 *	position info and the residual header, returns cErrBufferExceeded.
 *	Also returns cErrBufferExceeded if the client passes in a buffer
 *	for the page name that's too short to hold the name encoded in
 *	the page header.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_readTundraPageHeaders(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pPageBuf,
    UInt32		bufSize,
    UInt32		pageNameBufLen,
    char	      **ppPageName,
    UInt32	       *pPageNameLen,
    UInt32	       *pPageWidth,
    UInt32	       *pPageHeight,
    UInt32	       *pPageHeaderBytes,
    UInt32	       *pPosBlockBytes,
    UInt32	       *pResX,
    UInt32	       *pResY,
    UInt32	       *pResW,
    UInt32	       *pResH,
    UInt32	       *pResHeaderBytes,
    UInt32	       *pResBlockBytes)
{
    PROLOGUE;
    UInt32	realPageNameLen;
    UInt32	posBufLength;
    char       *pLocalPage = NULL;
    Bool	bufAllocatedHere = cFalse;
    UInt32	pageHeaderSize, resHeaderSize;
    UInt32	resLength;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pPageBuf == NULL) ||
	    (ppPageName == NULL) || (pPageNameLen == NULL) ||
	    (pPageWidth == NULL) || (pPageHeight == NULL) ||
	    (pPageHeaderBytes == NULL) || (pPosBlockBytes == NULL) ||
	    (pResX == NULL) || (pResY == NULL) ||
	    (pResW == NULL) || (pResH == NULL) ||
	    (pResHeaderBytes == NULL) || (pResBlockBytes == NULL), cErrNullPtr);

	/* Go get the length of the page name.  Remember that
	 * this length does NOT include the NULL at the end.
	 * It's compatible with strlen in this way.
	 */
    ERR_PASS( i_getExistingTundraPageNameSize (pIPWState, pTextCompInfoIn,
					       pPageBuf,
					       bufSize,
					       &realPageNameLen));

	/* See if client passed in a page name buffer */
    if (*ppPageName == NULL)
    {
	    /* Client didn't pass in a buffer, so allocate one.
	     * Remember to make room for the NULL at the end!
	     */
	pLocalPage = (char *)IPW_MALLOC(realPageNameLen + 1, cAllocatorData);
	ERR_GEN(pLocalPage == NULL, cErrNomem);

	bufAllocatedHere = cTrue;
    }
    else	/* client passed in buffer.  Is it big enough? */
    {
	ERR_GEN(pageNameBufLen < realPageNameLen + 1, cErrBufferExceeded);
	pLocalPage = *ppPageName;
    }

	/* Go read the header */
    ERR_PASS_CLEANUP( i_readTundraPageHeader (pIPWState, pTextCompInfoIn,
					      pPageBuf,
					      realPageNameLen + 1,
					      bufSize,
					      pLocalPage,
					      pPageWidth,
					      pPageHeight,
					      &posBufLength,
					      &pageHeaderSize));

	/* Go find the residual header.  If we don't have room,
	 * complain.  To calculate how much room we need, we
	 * need the length of the residual header.
	 */
    ERR_PASS_CLEANUP( i_getTundraResidualHeaderSize (pIPWState,
						     pTextCompInfoIn,
						     &resHeaderSize));
    ERR_GEN_CLEANUP(pageHeaderSize + posBufLength + resHeaderSize > bufSize,
		    cErrBufferExceeded);

	/* Go read the residual header */
    ERR_PASS_CLEANUP( i_readTundraResidualHeader (pIPWState, pTextCompInfoIn,
						  pPageBuf +
						   pageHeaderSize +
						   posBufLength,
						  resHeaderSize,
						  pResX, pResY,
						  pResW, pResH,
						  &resLength,
						  &resHeaderSize));

	/* Swell, that worked.  Return set up the page name pointer
	 * and lengths and return.
	 */
    *ppPageName = (char *) pLocalPage;
    *pPageNameLen = realPageNameLen;
    *pPageHeaderBytes = pageHeaderSize;
    *pPosBlockBytes = posBufLength;
    *pResHeaderBytes = resHeaderSize;
    *pResBlockBytes = resLength;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (bufAllocatedHere)
	IPW_FREE(pLocalPage, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a Tundra Position block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  An official Tundra page block contains a page header, a position
 *  block, a residual header and a residual block.  Other methods of
 *  storing symbol-based compressed pages may use other encapsulations
 *  for the position and residual blocks.<BR>
 *  This routine converts a position block into a bitmap.  This
 *  routine exists for the times when the position block is 
 *  encapsulated in something other than a Tundra file, like a XIFF file.<BR>
 *  Symbols for this page must already have been read using
 *  w_readTundraDictionaryBlock with the forDecompression flag set.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pageWidth::	     Width of dest image in pixels.
 *  pageHeight::     Height of dest image in pixels.
 *  pPositionBuf::   Pointer to memory block holding Tundra position bytes.
 *  bufSize::	     the number of bytes remaining in the position buffer.
 *
 * Output:::
 *  pPrD::	     Pointer to ptr to dest image.  If *pPrD is NULL, a
 *		     destination PIXR is allocated.  Otherwise, we check
 *		     the PIXR passed in.  The dimensions of the target
 *		     image may be found using w_readTundraPageHeaders.
 *
 * Returns:::
 *	If an insufficient page buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_decompressTundraPositionBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32		pageWidth,
    UInt32		pageHeight,
    UInt8	       *pPositionBuf,
    UInt32		bufSize,
    PIXR	      **pPrD)
{
    PROLOGUE;
    PIXR		*prD;
    ScaledInt1616	 xdpi, ydpi;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;
    UInt8		*pImage;
    UInt32		 bpl;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pPositionBuf == NULL) ||
	    (pPrD == NULL), cErrNullPtr);

	/* If there's an incoming image, check it.  Otherwise, go create one */
    prD = *pPrD;

	/* We really don't care about the dest image's resolution since
	 * that's not held in the file anyway.  So compare it with itself.
	 */
    if (prD != NULL)
    {
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prD, &xdpi, &ydpi));
    }
    else
    {
	xdpi = ydpi = 0;
    }

	ERR_PASS (wp_createPixrConditional(pIPWState, pageWidth, pageHeight,
					   1, xdpi, ydpi,
					   cFrameBits, cFalse, &prD,
					   &rightFrameColsToClear,
					   &bottomFrameRowsToClear,
					   &allocatedHere));
	/* That seemed to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  (Int32 *)(&bpl) ));

    ERR_PASS_CLEANUP( i_readTundraPositionBlock (pIPWState, pTextCompInfoIn,
						 pPositionBuf,
					         bufSize, pImage,
					         pageWidth + 2*cFrameBits,
						 pageHeight + 2*cFrameBits,
						 bpl));

	/* Keen!  If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a Tundra Residual block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  An official Tundra page block contains a page header, a position
 *  block, a residual header and a residual block.  Other methods of
 *  storing symbol-based compressed pages may use other encapsulations
 *  for the position and residual blocks.<BR>
 *  This routine converts a residual block into a bitmap.  This
 *  routine exists for the times when the residual block is 
 *  encapsulated in something other than a Tundra file, like a XIFF file.<BR>
 *  The residual image may be smaller than the destination image.
 *  The client must specify the dimensions and position of the residual
 *  image relative to the destination image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  destWidth::	     width of dest image in pixels
 *  destHeight::     height of dest image in pixels
 *  resX::	     X coord of left edge of residual image relative to
 *		     dest image
 *  resY::	     Y coord of top edge of residual image relative to
 *		     dest image
 *  resWidth::	     Width of residual image in pixels.
 *  resHeight::	     Height of residual image in pixels.
 *  combineResults:: If set to cTrue, the residual pixels are ORed into
 *		     an existing image.  If set to cFalse, the original
 *		     image is assumed to be clear.
 *  pResidualBuf::   Pointer to memory block holding Tundra position bytes.
 *  bufSize::	     the number of bytes in the residual block. If this
 *		     is 0, the block is uncompressed and the real size
 *		     is calculated as ((resWidth + 7)/8) * resHeight bytes.
 *
 * Output:::
 *  pPrD::	     Pointer to ptr to dest image.  If *pPrD is NULL, a
 *		     destination PIXR is allocated.  Otherwise, we check
 *		     the PIXR passed in.  The dimensions of the target
 *		     image may be found using w_readTundraPageHeaders.
 *
 * Returns:::
 *	If an insufficient page buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * Notes:::
 *	If bufSize is passed in as 0, the buffer is uncompressed.  Its
 *	size is calculated as ((resWidth + 7) / 8) * resHeight bytes.
 *	This follows the residual header semantics specified in the
 *	Tundra spec.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_decompressTundraResidualBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32		destWidth,
    UInt32		destHeight,
    UInt32		resX,
    UInt32		resY,
    UInt32		resWidth,
    UInt32		resHeight,
    UInt32		combineResults,
    UInt8	       *pResidualBuf,
    UInt32		bufSize,
    PIXR	      **pPrD)
{
    PROLOGUE;
    PIXR		*prD;
    ScaledInt1616	 xdpi, ydpi;
    Int32		 realDestWidth, realDestHeight;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;
    UInt8		*pImage;
    UInt32		 bpl;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pResidualBuf == NULL) ||
	    (pPrD == NULL), cErrNullPtr);

	/* If there's an incoming image, check it.  Otherwise, go create one */
    prD = *pPrD;

	/* We really don't care about the dest image's resolution since
	 * that's not held in the file anyway.  So compare it with itself.
	 */
    if (prD != NULL)
    {
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prD, &xdpi, &ydpi));
    }
    else
    {
	xdpi = ydpi = 0;
    }

    ERR_PASS (wp_createPixrConditional(pIPWState, destWidth, destHeight,
				       1, xdpi, ydpi,
				       cFrameBits, cFalse, &prD,
				       &rightFrameColsToClear,
				       &bottomFrameRowsToClear,
				       &allocatedHere));

	/* That seemed to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD,  (Int32 *)(&bpl) ));
    ERR_PASS_CLEANUP(w_pixrFramedWidth(pIPWState, prD,  &realDestWidth ));
    ERR_PASS_CLEANUP(w_pixrFramedHeight(pIPWState, prD,  &realDestHeight ));

    ERR_PASS_CLEANUP( i_readTundraResidualBlock (pIPWState, pTextCompInfoIn,
						 pResidualBuf,
					         bufSize, combineResults,
						 pImage,
					         realDestWidth, realDestHeight,
						 bpl,
						 resX + cFrameBits,
						 resY + cFrameBits,
						 resWidth, resHeight));

	/* Keen!  If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a Tundra Page block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine converts a specific page block into a bitmap.  This
 *  routine is only used to decompose Tundra page blocks.  It should
 *  not be used to create pages for XIFF pages since their
 *  encapsulation is different.<BR>
 *  The symbols for this page must already have been read using
 *  w_readTundraDictionaryBlock with the forDecompression flag set.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pPageBuf::	     Pointer to memory block holding Tundra page bytes.
 *  bufSize::	     the number of bytes remaining in the page buffer.
 *
 * Output:::
 *  pPrD::	     Pointer to ptr to dest image.  If *pPrD is NULL, a
 *		     destination PIXR is allocated.  Otherwise, we check
 *		     the PIXR passed in.  The dimensions of the target
 *		     image may be found using w_readTundraPageHeaders.
 *
 * Returns:::
 *	If an insufficient page buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_decompressTundraPageBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pPageBuf,
    UInt32		bufSize,
    PIXR	      **pPrD)
{
    PROLOGUE;
    UInt32		 pageWidth, pageHeight;
    UInt32		 resX, resY, resW, resH;
    char		 pageName[cDefaultPageNameLength];
    char		*pPageName = &(pageName[0]);
    UInt32		 realPageNameLen;
    UInt32		 pageHeaderLen, posBlockLen;
    UInt32		 resHeaderLen, resBlockLen;
    PIXR		*prD;
    ScaledInt1616	 xdpi, ydpi;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;
    UInt8		*pImage;
    UInt32		 bpl;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pPageBuf == NULL) ||
	    (pPrD == NULL), cErrNullPtr);

	/* get the page dimensions */
    ERR_PASS( w_readTundraPageHeaders(pIPWState, pTextCompInfoIn,
				     pPageBuf, bufSize,
				     cDefaultPageNameLength,
				     &pPageName,
				     &realPageNameLen,
				     &pageWidth,
				     &pageHeight,
				     &pageHeaderLen,
				     &posBlockLen,
				     &resX, &resY, &resW, &resH,
				     &resHeaderLen, &resBlockLen));

	/* Complain if the input buffer doesn't hold the whole page.
	 * Note that if resBlockLen is 0, the residual block is
	 * uncompressed, so we have to calculate a different value
	 * for its length from the width and height.
	 */
    if (resBlockLen == 0)
    ERR_GEN(pageHeaderLen + posBlockLen + resHeaderLen +
	    (resH * ((resW + 7)/8)) > bufSize, cErrBufferExceeded);
    else
    ERR_GEN(pageHeaderLen + posBlockLen + resHeaderLen + resBlockLen > bufSize,
	    cErrBufferExceeded);

	/* If there's an incoming image, check it.  Otherwise, go create one */
    prD = *pPrD;
	/* We really don't care about the dest image's resolution since
	 * that's not held in the file anyway.  So compare it with itself.
	 */
    if (prD != NULL)
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prD, &xdpi, &ydpi));
    else
	xdpi = ydpi = 0;

    ERR_PASS (wp_createPixrConditional(pIPWState, pageWidth, pageHeight,
				       1, xdpi, ydpi,
				       cFrameBits, cTrue, &prD,
				       &rightFrameColsToClear,
				       &bottomFrameRowsToClear,
				       &allocatedHere));

	/* That seemed to work, so do the real work */
    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prD, &pImage));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, prD, (Int32 *)(&bpl)));
    ERR_PASS_CLEANUP(w_pixrFramedWidth(pIPWState, prD,
					(Int32 *)(&pageWidth)));
    ERR_PASS_CLEANUP(w_pixrFramedHeight(pIPWState, prD,
					(Int32 *)(&pageHeight)));

	/* First, read the residual block so we don't have to
	 * OR it on top of the position data.  Note that passing
	 * a resBlockLen of 0 says that the block is uncompressed.
	 */
    ERR_PASS_CLEANUP( i_readTundraResidualBlock (pIPWState, pTextCompInfoIn,
						 pPageBuf +
						    pageHeaderLen +
						    posBlockLen +
						    resHeaderLen,
					         resBlockLen,
						 cFalse,
						 pImage,
					         pageWidth, pageHeight, bpl,
						 resX + cFrameBits,
						 resY + cFrameBits,
						 resW, resH));

	/* Now, read the position block.  All of its sybols get ORed
	 * into the image anyway.
	 */
    ERR_PASS_CLEANUP( i_readTundraPositionBlock (pIPWState, pTextCompInfoIn,
						 pPageBuf +
						    pageHeaderLen,
					         bufSize -
						    pageHeaderLen,
						 pImage,
					         pageWidth, pageHeight, bpl));

	/* Keen!  If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a Tundra Document.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine reads the header, dictionary and number of pages
 *  field.  It then allocates an array of page offsets and fills
 *  in the offset, relative to the beginning of the file, of the
 *  start of each page.  Individual pages may be read by the client
 *  using w_decompressTundraPage.<BR>
 *  This is a convenience routine in that it handles only Tundra files
 *  and only those with one dictionary.  It insists that the entire
 *  file be in memory.  Clients should handle longer and/or more complex
 *  files (like XIFF) with the individual block reading routines.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pDocBuf::	     Pointer to memory block holding Tundra document.
 *		     This originally points to the document header.
 *  bufSize::	     the number of bytes remaining in the document buffer.
 *
 * Output:::
 *  ppPageArray::     Pointer to pointer to returned array of page offsets.
 *		      This array is always allocated here and must be
 *		      freed by the client.  The offsets are in units of
 *		      bytes from the beginning of the document header.
 *  pNumOfPages::     Pointer to returned number of pages
 *
 * Returns:::
 *	If an insufficient document buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_readTundraDocument(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pDocBuf,
    UInt32		bufSize,
    UInt32	      **ppPageArray,
    UInt32	       *pNumOfPages)
{
    PROLOGUE;
    UInt8	       *pLocalBuf = pDocBuf;
    UInt32		localBufSize = bufSize;
    UInt32		headerSize;
    UInt32		dictionarySize;
    UInt32		numOfPagesFieldSize;
    UInt32		numOfPages;
    UInt32		curPageNum;
    UInt32	       *pPageArray = NULL;
	/* we'll toss the page names, but we need a place to put them */
    char	        pageName[cDefaultPageNameLength];
    char		*pPageName = &(pageName[0]);
    UInt32		realPageNameLength;
    UInt32		pageWidth;
    UInt32		pageHeight;
    UInt32		pageHeaderSize, posBlkSize, resHeaderSize, resBlkSize;
    UInt32		resX, resY, resW, resH;

	/* See if the pointers are ok */
    ERR_GEN((pTextCompInfoIn == NULL) || (pDocBuf == NULL) ||
	    (ppPageArray == NULL) || (pNumOfPages == NULL), cErrNullPtr);

    ERR_GEN((*ppPageArray != NULL), cErrPtrNotNull);

	/* Go read the file header */
    ERR_PASS( w_readTundraHeader( pIPWState, pTextCompInfoIn,
				 pLocalBuf, localBufSize,
				 &headerSize));

	/* If that was ok, go read the dictionary. */
    pLocalBuf += headerSize;
    localBufSize -= headerSize;

    ERR_PASS( w_readTundraDictionaryBlock( pIPWState, pTextCompInfoIn,
					  1,		/* dict Number */
					  cTrue,	/* for decompression */
					  pLocalBuf,
					  localBufSize,
					  &dictionarySize));

	/* If that worked, go read the number of pages */
    pLocalBuf += dictionarySize;
    localBufSize -= dictionarySize;

    ERR_PASS( w_readTundraNumberOfPagesField( pIPWState, pTextCompInfoIn,
					     pLocalBuf,
					     localBufSize,
					     &numOfPages,
					     &numOfPagesFieldSize));

	/* Boy, we're on a roll now!  Allocate an array of page offset
	 * pointers and fill it in.
	 */
    pPageArray = (UInt32 *) IPW_MALLOC( numOfPages * sizeof(UInt32),
					cAllocatorData);
    ERR_GEN( pPageArray == NULL, cErrNomem);

    pLocalBuf += numOfPagesFieldSize;
    localBufSize -= numOfPagesFieldSize;

    for (curPageNum = 0; curPageNum < numOfPages; curPageNum++)
    {
	    /* fill in offset of the current page */
	pPageArray[curPageNum] = (UInt32) (pLocalBuf - pDocBuf);

	    /* and read it's header.  This tells us how many
	     * bytes in the page.
	     */
	ERR_PASS_CLEANUP( w_readTundraPageHeaders( pIPWState,
						  pTextCompInfoIn,
						  pLocalBuf,
						  localBufSize,
						  cDefaultPageNameLength,
						  &pPageName,
						  &realPageNameLength,
						  &pageWidth,
						  &pageHeight,
						  &pageHeaderSize,
						  &posBlkSize,
						  &resX, &resY,
						  &resW, &resH,
						  &resHeaderSize,
						  &resBlkSize));
	    /* Skip over that page */
	pLocalBuf += pageHeaderSize + posBlkSize +
		     resHeaderSize + resBlkSize;
	localBufSize -= pageHeaderSize + posBlkSize +
			resHeaderSize + resBlkSize;
    }

	/* We're all set.  Set the values to be returned and return */
    *ppPageArray = pPageArray;
    *pNumOfPages = numOfPages;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pPageArray != NULL)
	IPW_FREE(pPageArray, cAllocatorData);

    ERR_RETURN;
}


/*

 $Log: tctnd.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.7  1998/01/14 00:15:51  ddavies
 tctnd.c -> /product/ipcore/ipwrap/src/RCS/tctnd.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.6  1997/04/15 22:45:22  ddavies
 tctnd.c -> /product/ipcore/ipwrap/src/RCS/tctnd.c,v

 Make things prettier.
 Use non-initialized PIXR creation where possible.
 (really means add new arg to wp_createPixrConditional).

 Revision 1.5  1996/09/06 04:51:40  ddavies
 ddavies on Thu Sep 5 21:49:30 PDT 1996

 tctnd.c -> /product/ipcore/ipwrap/src/RCS/tctnd.c,v

 Account for uncompressed residual images.  Now if bufSize is passed as
 0 to w_decompressTundraResidualBlock or is found to be 0 inside
 w_decompressTundraPageBlock, we assume the residual block is uncompressed.
 The size of the uncompressed block is calculated from the width and height
 of the residual image.

 Revision 1.4  1996/08/30 23:50:24  ddavies
 ddavies on Fri Aug 30 16:49:48 PDT 1996

 tctnd.c -> /product/ipcore/ipwrap/src/RCS/tctnd.c,v

 Add frame to width and height being passed to i_readTundrayPositionBlock.
 The absence of the frame was a bug.

 Revision 1.3  1996/08/30 04:08:08  ddavies
 ddavies on Thu Aug 29 21:07:29 PDT 1996

 tctnd.c -> /product/ipcore/ipwrap/src/RCS/tctnd.c,v

 Add image width and height args to w_decompressTundraResidualBlock so that
 it can create/check the dest image using its real dimensions, not the
 dimensions of the residual image inside it.

 Revision 1.2  1996/07/03 21:13:47  ddavies
 ddavies on Wed Jul 3 14:13:30 PDT 1996

 tctnd.c -> /product/ipcore/ipwrap/src/RCS/tctnd.c,v

 Fix comments and nag script warning.

 Revision 1.1  1996/07/02 23:09:30  ddavies
 Initial revision


 */

