/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tctne.c:
 *
 *	Code used to emit symbol-based compression info in Tundra
 *	format.  These routines create and fill in memory blocks.
 *	It's the client's responsibility to do file I/O.
 *
 *	w_writeTundraHeader
 *		Create and fill in a Tundra file header.
 *
 *	w_estimateTundraDictionaryDataSize
 *		Estimate the amount of memory in bytes needed to
 *		hold the data portion of a Tundra dictionary block.
 *
 *	w_estimateTundraDictionaryBlockSize
 *		Estimate the amount of memory in bytes needed to
 *		hold an entire (header + data) Tundra dictionary block.
 *
 *	w_writeTundraDictionaryData
 *		Create and fill in the data portion of a Tundra
 *		symbol dictionary.
 *
 *	w_writeTundraDictionaryBlock
 *		Create and fill in an entire Tundra symbol dictionary,
 *		header + data.
 *
 *	w_writeTundraNumberOfPagesField
 *		Create and fill in the Tundra  "number of pages" field.
 *
 *	w_estimateTundraPositionBlockSize
 *		Estimate the amount of memory in bytes needed to
 *		hold a Tundra position block.
 *
 *	w_estimateTundraResidualBlockSize
 *		Estimate the amount of memory in bytes needed to
 *		hold a Tundra residual block.
 *
 *	w_estimateTundraPageBlockSize
 *		Estimate the amount of memory in bytes needed to
 *		hold a Tundra page (position + residual) block.
 *
 *	w_writeTundraPositionBlock
 *		Create and fill in the info that goes with a single
 *		Tundra symbol position block.
 *
 *	w_writeTundraResidualBlock
 *		Create and fill in the info that goes with a single
 *		Tundra residual image block.
 *
 *	w_writeTundraPageBlock
 *		Create and fill in the info that goes with a single
 *		Tundra page (symbol position + residual image).
 *
 *	w_writeTundraDocument
 *		A convenience routine.  Uses the routines above to create
 *		a Tundra formatted block containing a header, dictionary
 *		and one or more pages.
 *
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "pixr.h"
#include "geomadt.h"

/* PROTOTYPES */
#include "tcompr.pub"
#include "shrcmprs.pub"
#include "props.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: tctne.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* This file contains wrappers for routines used to convert our internal
 * format for symbols and their positions on pages into a Tundra formatted
 * memory block that can be written or transmitted by the client.
 *
 * A compressed file contains four main parts.  It begins with a file
 * header.  Following that is a dictionary.  It is theoretically possible
 * to have multiple dictionarys in a file, but we do not support that
 * at this time.  A dictionary contains the pictures of the symbols.
 * The dictionary is followed by a few bytes containing the number of
 * pages to follow.  Finally, each page's individual info (symbol positions
 * and residual image) are at the end.
 *
 * The first routines listed above each estinate the size of a Tundra section
 * or produce a Tundra section of the output block.  Clients may use
 * these individually, especially when writing files with a large number
 * of pages and/or dictionaries.  The last routine, w_writeTundraDocument
 * is a convenience routine that bundles all the previous routines together.
 * It might be used to create a memory block holding the info for a
 * small number of pages.  It's also instructive.
 *
 */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a Tundra file header.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header.  This routine fills in a header.  If memory for a
 *  header is passed in, it is used.  If not, a memory block is
 *  allocated.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  bufSize::	     If *ppHeader is not NULL, this is the number
 *		     of bytes remaining in the buffer where the
 *		     header could be written.
 *
 * Output:::
 *  ppHeader::       Pointer to returned header pointer.  If *ppHeader
 *		     is non-null, existing memory at this site is used.
 *		     Otherwise, a new block is allocated.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeTundraHeader (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32              bufSize,
    UInt8	      **ppHeader,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 realHeaderSize;
    UInt8	*pNewHeader;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (ppHeader == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* Figure out how big the header should be */
    ERR_PASS(w_getTundraHeaderSize(pIPWState, pTextCompInfoIn, &realHeaderSize));

	/* If no memory was passed in, allocate a block big enough for
	 * the header.  Otherwise, see if the block passed in is
	 * big enough.
	 */
    if (*ppHeader == NULL)	/* no memory passed in */
    {
	pNewHeader = (UInt8 *) IPW_MALLOC(realHeaderSize, cAllocatorData);
	ERR_GEN(pNewHeader == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in */
    {
	ERR_GEN(bufSize < realHeaderSize, cErrBufferExceeded);
	pNewHeader = *ppHeader;
    }

	/* fill in the header */
    ERR_PASS_CLEANUP(i_writeTundraHeader (pIPWState, pTextCompInfoIn,
					  pNewHeader,
					  realHeaderSize, pBytesUsed));

	/* It must have really worked, so update the returned pointer
	 * and return */
    *ppHeader = pNewHeader;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pNewHeader, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of data portion of a Tundra dictionary block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  A dictionary is composed of a header and some data.  The data
 *  portion alone is used in non-Tundra files, like XIFF.
 *  This routine estimates the size in bytes of the data portion
 *  of a dictionary block so that a client can allocate one.  It takes
 *  as input an array of booleans, one entry per possible input dictionary.
 *  Each entry of the array indicates whether the corresponding dictionary
 *  should be included in the output.  For example, if dictArray[i] is
 *  true, symbols from input dictionary i should be included in the output
 *  dictionary block.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  dictArray::      Array of booleans, one per possible input dictionary.
 *		     if dictArray[i] is set, symbols from input dictionary
 *		     i will be included in the output dictionary block.
 *  dictArrayLen::   number of elements in dictArray.
 *
 * Output:::
 *  pDictDataSize:: Pointer to returned estimated dictionary data size
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateTundraDictionaryDataSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    Bool	       *dictArray,
    UInt32              dictArrayLen,
    UInt32	       *pDictDataSize)
{
    PROLOGUE;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pDictDataSize == NULL),
	    cErrNullPtr);

	/* This is easy! */
    ERR_PASS(i_estimateTundraDictionaryBlockSize (pIPWState, pTextCompInfoIn,
						  dictArray, dictArrayLen,
						  0, pDictDataSize));

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of a Tundra dictionary block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  A dictionary is composed of a header and some data.
 *  This routine estimates the size in bytes of the entire dictionary
 *  block (header + data) so that a client can allocate one.  It
 *  takes as input an array of booleans, one entry per possible
 *  input dictionary.  Each entry of the array indicates whether
 *  the corresponding dictionary should be included in the output.
 *  For example, if dictArray[i] is true, symbols from input dictionary
 *  i should be included in the output dictionary block.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  dictArray::      Array of booleans, one per possible input dictionary.
 *		     if dictArray[i] is set, symbols from input dictionary
 *		     i will be included in the output dictionary block.
 *  dictArrayLen::   number of elements in dictArray.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  pDictBlockSize:: Pointer to returned estimated dictionary block size
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies, aber
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateTundraDictionaryBlockSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    Bool	       *dictArray,
    UInt32              dictArrayLen,
    UInt32		pageRequest,
    UInt32	       *pDictBlockSize)
{
    PROLOGUE;
    UInt32		dictBlockLengthFieldSize;
    UInt32		dictBlockSize;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pDictBlockSize == NULL),
	    cErrNullPtr);

	/* A dictionary block consists of two parts from our point
	 * of view, a length spec and the dictionary itself.  Get
	 * the sizes of these two components and return the sum.
	 */
    ERR_PASS(i_getTundraDictBlockLengthSize (pIPWState, pTextCompInfoIn,
					     &dictBlockLengthFieldSize));
    ERR_PASS(i_estimateTundraDictionaryBlockSize (pIPWState, pTextCompInfoIn,
						  dictArray, dictArrayLen,
						  pageRequest,
						  &dictBlockSize));

	/* Those worked fine, return the sum */
    *pDictBlockSize = dictBlockLengthFieldSize + dictBlockSize;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in data portion of a Tundra dictionary block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  A dictionary is composed of a header and some data.
 *  This routine fills in the data portion alone.  This data portion
 *  is also used in other file formats, like XIFF, which do not
 *  use the Tundra header.<BR>
 *  If memory for this field is passed in, it is used.  If not,
 *  a memory block is allocated.  This routine takes as input an
 *  array of booleans,  one entry per possible input dictionary.
 *  Each entry of the array indicates whether the corresponding dictionary
 *  should be included in the output.  For example, if dictArray[i] is
 *  true, symbols from input dictionary i should be included in the output
 *  dictionary block.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  dictArray::      Array of booleans, one per possible input dictionary.
 *		     if dictArray[i] is set, symbols from input dictionary
 *		     i will be included in the output dictionary block.
 *  dictArrayLen::   number of elements in dictArray.
 *  bufSize::	     If *ppDictBuf is not NULL, this is the number
 *		     of bytes remaining in the buffer where the
 *		     field could be written.
 *
 * Output:::
 *  ppDictBuf::	     Pointer to returned dictionary block pointer.
 *		     If *ppDictBuf is non-null, existing memory at this
 *		     site is used.  Otherwise, a new block is allocated.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * Notes:::
 *	If the dictionary buffer is allocated internally, this routine will
 *	keep trying with ever-larger buffers if the dictionary doesn't fit.
 *	If the dictionary buffer is passed in by the client, we can't change
 *	the size of the client's buffer, so we return immediately if the
 *	dictionary doesn't fit.
 *	If the dictionary buffer is allocated internally and there is no
 *	dictionary data, the buffer will be freed and *ppDictBuf will be
 *	returned with a value of NULL.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_writeTundraDictionaryData(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    Bool	       *dictArray,
    UInt32              dictArrayLen,
    UInt32		bufSize,
    UInt8	      **ppDictBuf,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 estDictDataSize;
    UInt32	 realDictDataSize;
    UInt8	*pNewDictData;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (ppDictBuf == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* See if the the client passed in a buffer for the dictionary */
    if (*ppDictBuf == NULL)
    {
	    /* allocate a new buffer, estimating it's size first */
	ERR_PASS(w_estimateTundraDictionaryDataSize(pIPWState,
						   pTextCompInfoIn,
						   dictArray,
						   dictArrayLen,
						   &estDictDataSize));
	pNewDictData = (UInt8 *) IPW_MALLOC(estDictDataSize, cAllocatorData);
	ERR_GEN(pNewDictData == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in.  Since
				 * the new dictionary data size is only
				 * an estimate, we don't check to see
				 * if the client has supplied something
				 * smaller here.
				 */
    {
	pNewDictData = *ppDictBuf;
	estDictDataSize = bufSize;
    }

	/* If we allocated the buffer, keep increasing its size until
	 * the dictionary data fits.  Otherwise, just quit if the client
	 * supplied the buffer.
	 */
    if (allocatedBufferHere)
    {
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(i_writeTundraDictionaryData (pIPWState,
							  pTextCompInfoIn,
							  dictArray,
							  dictArrayLen,
							  pNewDictData,
							  estDictDataSize,
							  0,
							  &realDictDataSize))
	    {
		    /* Something went wrong.  Was it a buffer size
		     * problem or something else?
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		    /* The dictionary didn't fit.  Increase the size of
		     * the buffer by 1.5x */
		estDictDataSize = estDictDataSize + estDictDataSize/2;
		IPW_FREE(pNewDictData, cAllocatorData);
		pNewDictData = (UInt8 *) IPW_MALLOC(estDictDataSize,
						    cAllocatorData);
		    /* We can return directly if pNewDictData is now NULL
		     * since it's the only thing we allocate */
		ERR_GEN(pNewDictData == NULL, cErrNomem);
	    }
	    else
	    {
		    /* it worked!  Trim the unused bytes off the end
		     * of the buffer.  If the buffer size is 0, perhaps
		     * because the page was blank or there were only
		     * residual symbols, just free the buffer.
		     */
		if (realDictDataSize != 0)
		{
		    pNewDictData = (UInt8 *)IPW_REALLOC(pNewDictData,
							realDictDataSize,
							cAllocatorData);
		    ERR_GEN(pNewDictData == NULL, cErrNomem);
		}
		else	/* there's no dictionary data */
		    IPW_FREE(pNewDictData, cAllocatorData);
		break;
	    }
	}
    }
    else	/* the client allocated the buffer so we can't change it.
		 * If the dictionary doesn't fit, tough.
		 */
    {
	    /* Fill in the dictionary block */
	ERR_PASS_CLEANUP(i_writeTundraDictionaryData (pIPWState,
						      pTextCompInfoIn,
						      dictArray,
						      dictArrayLen,
						      pNewDictData,
						      estDictDataSize,
						      0,
						      &realDictDataSize));
    }

	/* Return a pointer to the final dictionary block and the number
	 * of bytes used.
	 */
    *pBytesUsed = realDictDataSize;
    *ppDictBuf = pNewDictData;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pNewDictData, cAllocatorData);

    ERR_RETURN;
}




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a Tundra dictionary block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine fills in the dictionary.  If memory for this
 *  field is passed in, it is used.  If not, a memory block is
 *  allocated.  This routine takes as input an array of booleans, 
 *  one entry per possible input dictionary.
 *  Each entry of the array indicates whether the corresponding dictionary
 *  should be included in the output.  For example, if dictArray[i] is
 *  true, symbols from input dictionary i should be included in the output
 *  dictionary block.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  dictArray::      Array of booleans, one per possible input dictionary.
 *		     if dictArray[i] is set, symbols from input dictionary
 *		     i will be included in the output dictionary block.
 *  dictArrayLen::   number of elements in dictArray.
 *  bufSize::	     If *ppDictBuf is not NULL, this is the number
 *		     of bytes remaining in the buffer where the
 *		     field could be written.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  ppDictBuf::	     Pointer to returned dictionary block pointer.
 *		     If *ppDictBuf is non-null, existing memory at this
 *		     site is used.  Otherwise, a new block is allocated.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * Notes:::
 *	If the dictionary buffer is allocated internally, this routine will
 *	keep trying with ever-larger buffers if the dictionary doesn't fit.
 *	If the dictionary buffer is passed in by the client, we can't change
 *	the size of the client's buffer, so we return immediately if the
 *	dictionary doesn't fit.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_writeTundraDictionaryBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    Bool	       *dictArray,
    UInt32              dictArrayLen,
    UInt32		bufSize,
    UInt8	      **ppDictBuf,
    UInt32		pageRequest,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 estDictSize;
    UInt32	 realDictSize;
    UInt32	 dictLengthFieldSize;
    UInt8	*pNewDict;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (ppDictBuf == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* See if the the client passed in a buffer for the dictionary */
    if (*ppDictBuf == NULL)
    {
	    /* allocate a new buffer, estimating it's size first */
	ERR_PASS(w_estimateTundraDictionaryBlockSize(pIPWState,
						     pTextCompInfoIn,
						     dictArray,
						     dictArrayLen,
						     pageRequest,
						     &estDictSize));
	pNewDict = (UInt8 *) IPW_MALLOC(estDictSize, cAllocatorData);
	ERR_GEN(pNewDict == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in.  Since
				 * the new dictionary block size is only
				 * an estimate, we don't check to see
				 * if the client has supplied something
				 * smaller here.
				 */
    {
	pNewDict = *ppDictBuf;
	estDictSize = bufSize;
    }

	/* Get the length of the dictionary length field.  We skip over
	 * that to start filling in the dictionary.
	 */
    ERR_PASS_CLEANUP(i_getTundraDictBlockLengthSize (pIPWState, pTextCompInfoIn,
						     &dictLengthFieldSize));

	/* If we allocated the buffer, keep increasing its size until
	 * the dictionary fits.  Otherwise, just quit if the client
	 * supplied the buffer.
	 */
    if (allocatedBufferHere)
    {
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(i_writeTundraDictionaryData (pIPWState,
							  pTextCompInfoIn,
							  dictArray,
							  dictArrayLen,
							  (pNewDict +
							   dictLengthFieldSize),
							  (estDictSize -
							   dictLengthFieldSize),
							  0,
							  &realDictSize))
	    {
		    /* Something went wrong.  Was it a buffer size
		     * problem or something else?
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		    /* The dictionary didn't fit.  Increase the size of
		     * the buffer by 1.5x */
		estDictSize = estDictSize + estDictSize/2;
		IPW_FREE(pNewDict, cAllocatorData);
		pNewDict = (UInt8 *) IPW_MALLOC(estDictSize, cAllocatorData);
		    /* We can return directly if pNewDict is now NULL
		     * since it's the only thing we allocate */
		ERR_GEN(pNewDict == NULL, cErrNomem);
	    }
	    else
	    {
		    /* it worked!  Trim the unused bytes off the end
		     * of the buffer.  There will always be some data
		     * here since the header takes up space even if
		     * there is no dictionary data.
		     */
		pNewDict = (UInt8 *)IPW_REALLOC(pNewDict,
						realDictSize +
						    dictLengthFieldSize,
						cAllocatorData);
		ERR_GEN(pNewDict == NULL, cErrNomem);
		break;
	    }
	}
    }
    else	/* the client allocated the buffer so we can't change it.
		 * If the dictionary doesn't fit, tough.
		 */
    {
	    /* Fill in the dictionary block */
	ERR_PASS_CLEANUP(i_writeTundraDictionaryData (pIPWState,
						      pTextCompInfoIn,
						      dictArray,
						      dictArrayLen,
						      (pNewDict +
							dictLengthFieldSize),
						      (estDictSize -
							dictLengthFieldSize),
						      0,
						      &realDictSize));
    }

	/* if that worked, fill in the length of the dictionary block */
    ERR_PASS_CLEANUP(i_writeTundraDictBlockLengthField (pIPWState,
							pTextCompInfoIn,
							realDictSize,
							pNewDict,
							dictLengthFieldSize,
							&dictLengthFieldSize));

	/* Return a pointer to the final dictionary block and the number
	 * of bytes used.
	 */
    *pBytesUsed = realDictSize + dictLengthFieldSize;
    *ppDictBuf = pNewDict;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pNewDict, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a Tundra "number of pages" field.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine fills in the "number of pages" field.  If memory for this
 *  field is passed in, it is used.  If not, a memory block is
 *  allocated.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  numOfPages::     The value to be written.
 *  bufSize::	     If *ppNumOfPagesField is not NULL, this is the number
 *		     of bytes remaining in the buffer where the
 *		     field could be written.
 *
 * Output:::
 *  ppNumOfPagesField:: Pointer to returned "number of pages" field pointer.
 *		     If *ppNumOfPagesField is non-null,
 *		     existing memory at this site is used.  Otherwise,
 *		     a new block is allocated.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeTundraNumberOfPagesField (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32		numOfPages,
    UInt32              bufSize,
    UInt8	      **ppNumOfPagesField,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 realFieldSize;
    UInt8	*pNewField;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (ppNumOfPagesField == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* Figure out how big the field should be */
    ERR_PASS(w_getTundraNumberOfPagesSize(pIPWState, pTextCompInfoIn,
					 &realFieldSize));

	/* If no memory was passed in, allocate a block big enough for
	 * the field.  Otherwise, see if the block passed in is
	 * big enough.
	 */
    if (*ppNumOfPagesField == NULL)	/* no memory passed in */
    {
	pNewField = (UInt8 *) IPW_MALLOC(realFieldSize, cAllocatorData);
	ERR_GEN(pNewField == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in */
    {
	ERR_GEN(bufSize < realFieldSize, cErrBufferExceeded);
	pNewField = *ppNumOfPagesField;
    }

	/* fill in the header */
    ERR_PASS_CLEANUP(i_writeTundraNumberOfPagesField (pIPWState,
						      pTextCompInfoIn,
						      numOfPages,
						      pNewField,
						      realFieldSize,
						      pBytesUsed));

	/* It must have really worked, so update the returned pointer
	 * and return */
    *ppNumOfPagesField = pNewField;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pNewField, cAllocatorData);

    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of a Tundra position block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine estimates the size in bytes of the position block
 *  component of a page block so that client can allocate one.
 *  It takes as input a position block in our internal format.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pInputBuf::	     Pointer to first byte of a position block formatted
 *		     in our internal format, not in Tundra format.  This
 *		     is the block produced by w_convertImageToSymbols and
 *		     possibly modified by i_remapSymbolIDs.
 *  inputBufLen::    number of bytes in the input position buffer.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  pPositionBlockSize:: Pointer to returned estimated position block size
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateTundraPositionBlockSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pInputBuf,
    UInt32              inputBufLen,
    UInt32		pageRequest,
    UInt32	       *pPositionBlockSize)
{
    PROLOGUE;
    UInt32		positionSize;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) ||
	    (pInputBuf == NULL) || (pPositionBlockSize == NULL),
	    cErrNullPtr);

	/* Just do it */
    ERR_PASS(i_estimateTundraPositionBlockSize (pIPWState, pTextCompInfoIn,
					        pInputBuf, inputBufLen,
						pageRequest,
					        &positionSize));

	/* Those worked fine, return the sum */
    *pPositionBlockSize = positionSize;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of a Tundra residual block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine estimates the size in bytes of the residual block
 *  component of a page so that client can allocate one.
 *  It takes as input a position block in our internal format.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pInputBuf::	     Pointer to first byte of a position block formatted
 *		     in our internal format, not in Tundra format.  This
 *		     is the block produced by w_convertImageToSymbols and
 *		     possibly modified by i_remapSymbolIDs.
 *  inputBufLen::    number of bytes in the input position buffer.
 *
 * Output:::
 *  pResidualBlockSize:: Pointer to returned estimated residual block size
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateTundraResidualBlockSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pInputBuf,
    UInt32              inputBufLen,
    UInt32	       *pResidualBlockSize)
{
    PROLOGUE;
    UInt32		residualSize;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) ||
	    (pInputBuf == NULL) || (pResidualBlockSize == NULL),
	    cErrNullPtr);

	/* Just do it */
    ERR_PASS(i_estimateTundraResidualBlockSize (pIPWState, pTextCompInfoIn,
					        pInputBuf, inputBufLen,
					        &residualSize));

	/* Those worked fine, return the sum */
    *pResidualBlockSize = residualSize;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Estimate size of a Tundra page block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.
 *  This routine estimates the size in bytes of the page
 *  block so that client can allocate one.  It takes as input an
 *  position block in our internal format.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pPageName::	     Ptr to null-terminated string holding name of
 *		     the page.  This string becomes part of the page
 *		     block, so we need it to guess the header's length.
 *  pInputBuf::	     Pointer to first byte of a position block formatted
 *		     in our internal format, not in Tundra format.  This
 *		     is the block produced by w_convertImageToSymbols and
 *		     possibly modified by i_remapSymbolIDs.
 *  inputBufLen::    Number of bytes in the input position buffer.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  pPageBlockSize:: Pointer to returned estimated page block size
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_estimateTundraPageBlockSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    char	       *pPageName,
    UInt8	       *pInputBuf,
    UInt32              inputBufLen,
    UInt32		pageRequest,
    UInt32	       *pPageBlockSize)
{
    PROLOGUE;
    UInt32		pageHeaderSize;
    UInt32		positionSize, residualHeaderSize, residualSize;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pPageName == NULL) ||
	    (pInputBuf == NULL) || (pPageBlockSize == NULL),
	    cErrNullPtr);

	/* A page block consists of four parts from our point
	 * of view, a header, a position block, a residual header
	 * and the residual block.
	 * Get the sizes of these four components and return the sum.
	 */
    ERR_PASS(i_getTundraPageHeaderSize (pIPWState, pTextCompInfoIn,
				        pPageName,
				        &pageHeaderSize));

    ERR_PASS(w_estimateTundraPositionBlockSize(pIPWState, pTextCompInfoIn,
					      pInputBuf, inputBufLen,
					      pageRequest, &positionSize));

    ERR_PASS(i_getTundraResidualHeaderSize (pIPWState, pTextCompInfoIn,
					    &residualHeaderSize));

    ERR_PASS(w_estimateTundraResidualBlockSize(pIPWState, pTextCompInfoIn,
					      pInputBuf, inputBufLen,
					      &residualSize));

	/* Those worked fine, return the sum */
    *pPageBlockSize = pageHeaderSize +
		      positionSize +
		      residualHeaderSize +
		      residualSize;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a Tundra position block.
 *
 * LongDescription:::
 *  There may be several formats for symbol-based compressed data.
 *  The official Tundra format is one, XIFF is another.  Both of
 *  these formats share a basic position block.  This routine
 *  writes naked position blocks.  If the client wants to write
 *  an official Tundra page (contains header, position and residual
 *  blocks), w_writeTundraPageBlock should be used.
 *  If the client wants to insert position info as an IFD in a
 *  XIFF file, this is the routine to use.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pInputBuf::	     Pointer to buffer containing symbol positions in
 *		     internal format.
 *  inputBufLength:: Number of bytes in input position buffer.
 *  outputBufLength::  If *ppOutputBuf is not NULL, this is the number
 *		     of bytes remaining in the output buffer.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  ppOutputBuf::    Pointer to returned output position pointer.
 *		     If *ppOutputBuf is non-null, existing memory at this
 *		     site is used.  Otherwise, a new block is allocated.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * Notes:::
 *	If the output buffer is allocated internally, this routine will
 *	keep trying with ever-larger buffers if the page doesn't fit.
 *	If the output buffer is passed in by the client, we can't change
 *	the size of the client's buffer, so we return immediately if the
 *	page doesn't fit.
 *	If the output buffer is allocated internally and there is no
 *	position data for this page, it is freed and *ppOutputBuf will be
 *	set to NULL on return.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_writeTundraPositionBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pInputBuf,
    UInt32		inputBufLength,
    UInt32		outputBufLength,
    UInt8	      **ppOutputBuf,
    UInt32		pageRequest,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 estPosBlockSize;
    UInt32	 encodedPositionSize;
    UInt8	*pOutputBuf;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) ||
	    (pInputBuf == NULL) || (ppOutputBuf == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* See if the the client passed in a buffer for the page */
    if (*ppOutputBuf == NULL)
    {
	    /* allocate a new buffer, estimating it's size first */
	ERR_PASS(w_estimateTundraPositionBlockSize(pIPWState,
						   pTextCompInfoIn,
						   pInputBuf,
						   inputBufLength,
						   pageRequest,
						   &estPosBlockSize));

	pOutputBuf = (UInt8 *) IPW_MALLOC(estPosBlockSize, cAllocatorData);
	ERR_GEN(pOutputBuf == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in.  Since
				 * the new page block size is only
				 * an estimate, we don't check to see
				 * if the client has supplied something
				 * smaller here.
				 */
    {
	pOutputBuf = *ppOutputBuf;
	estPosBlockSize = outputBufLength;
    }

	/* If we allocated the buffer, keep increasing its size until
	 * the position block fits.  Otherwise, just quit if the client
	 * supplied the buffer.
	 */
    if (allocatedBufferHere)
    {
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(i_writeTundraPositionBlock (pIPWState,
							 pTextCompInfoIn,
							 pInputBuf,
							 inputBufLength,
							 pOutputBuf,
							 estPosBlockSize,
							 pageRequest,
							 &encodedPositionSize))
	    {
		    /* Something went wrong.  Was it a buffer size
		     * problem or something else?
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		    /* The page didn't fit.  Increase the size of
		     * the buffer by 1.5x */
		estPosBlockSize = estPosBlockSize + estPosBlockSize/2;
		IPW_FREE(pOutputBuf, cAllocatorData);
		pOutputBuf = (UInt8 *) IPW_MALLOC(estPosBlockSize,
						  cAllocatorData);
		    /* We can return directly if pOutputBuf is now NULL
		     * since it's the only thing we allocate */
		ERR_GEN(pOutputBuf == NULL, cErrNomem);
	    }
	    else
	    {
		    /* it worked!  Trim the unused bytes off the end
		     * of the buffer.  If there is no position info,
		     * perhaps because the image is blank or contains
		     * only residual images, just free the position
		     * buffer.
		     */
		if (encodedPositionSize != 0)
		{
		    pOutputBuf = (UInt8 *)IPW_REALLOC(pOutputBuf,
						      encodedPositionSize,
						      cAllocatorData);
		    ERR_GEN(pOutputBuf == NULL, cErrNomem);
		}
		else	/* position buffer was empty */
		    IPW_FREE(pOutputBuf, cAllocatorData);
		break;
	    }
	}
    }
    else	/* the client allocated the buffer so we can't change it.
		 * If the page doesn't fit, tough.
		 */
    {
	    /* Fill in the page block */
	ERR_PASS_CLEANUP(i_writeTundraPositionBlock (pIPWState,
						     pTextCompInfoIn,
						     pInputBuf,
					             inputBufLength,
						     pOutputBuf,
						     estPosBlockSize,
						     pageRequest,
						     &encodedPositionSize));
    }

	/* Return a pointer to the final position block and the number
	 * of bytes used.
	 */
    *pBytesUsed = encodedPositionSize;
    *ppOutputBuf = pOutputBuf;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pOutputBuf, cAllocatorData);

    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a Tundra residual block.
 *
 * LongDescription:::
 *  There may be several formats for symbol-based compressed data.
 *  The official Tundra format is one, XIFF is another.  Both of
 *  these formats share a basic residual block.  This routine
 *  writes naked residual blocks.  If the client wants to write
 *  an official Tundra page (contains header, position and residual
 *  blocks), w_writeTundraPageBlock should be used.
 *  If the client wants to insert residual info as a subIFD in a
 *  XIFF file, this is the routine to use.
 *  In addition to writing the residual block, this routine returns
 *  the residual image's bounding box.  If there is no residual,
 *  the x, y, w, h and bytesUsed values are all set to 0.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pInputBuf::	     Pointer to buffer containing symbol positions in
 *		     internal format.
 *  inputBufLength:: Number of bytes in input position buffer.
 *  outputBufLength::  If *ppOutputBuf is not NULL, this is the number
 *		     of bytes remaining in the output buffer.
 *
 * Output:::
 *  ppOutputBuf::    Pointer to returned output residual buffer pointer.
 *		     If *ppOutputBuf is non-null, existing memory at this
 *		     site is used.  Otherwise, a new block is allocated.
 *  pResX::	     Pointer to returned X coord of left edge of residual img.
 *  pResY::	     Pointer to returned Y coord of top edge of residual img.
 *  pResW::	     Pointer to returned width of residual img.
 *  pResH::	     Pointer to returned X height of residual img.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * Notes:::
 *	If the output buffer is allocated internally, this routine will
 *	keep trying with ever-larger buffers if the page doesn't fit.
 *	If the output buffer is passed in by the client, we can't change
 *	the size of the client's buffer, so we return immediately if the
 *	page doesn't fit.
 *	If the output buffer is allocated internally and there is no
 *	residual data for this page, it is freed and *ppOutputBuf will be
 *	set to NULL on return.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_writeTundraResidualBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt8	       *pInputBuf,
    UInt32		inputBufLength,
    UInt32		outputBufLength,
    UInt8	      **ppOutputBuf,
    UInt32	       *pResX,
    UInt32	       *pResY,
    UInt32	       *pResW,
    UInt32	       *pResH,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 estResBlockSize;
    UInt32	 encodedResidualSize;
    UInt8	*pOutputBuf;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) ||
	    (pInputBuf == NULL) || (ppOutputBuf == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* See if the the client passed in a buffer for the page */
    if (*ppOutputBuf == NULL)
    {
	    /* allocate a new buffer, estimating it's size first */
	ERR_PASS(w_estimateTundraResidualBlockSize(pIPWState,
						  pTextCompInfoIn,
					          pInputBuf,
					          inputBufLength,
					          &estResBlockSize));

	pOutputBuf = (UInt8 *) IPW_MALLOC(estResBlockSize, cAllocatorData);
	ERR_GEN(pOutputBuf == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in.  Since
				 * the new page block size is only
				 * an estimate, we don't check to see
				 * if the client has supplied something
				 * smaller here.
				 */
    {
	pOutputBuf = *ppOutputBuf;
	estResBlockSize = outputBufLength;
    }

	/* If we allocated the buffer, keep increasing its size until
	 * the residual block fits.  Otherwise, just quit if the client
	 * supplied the buffer.
	 */
    if (allocatedBufferHere)
    {
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(i_writeTundraResidualBlock (pIPWState,
							 pTextCompInfoIn,
							 pInputBuf,
							 inputBufLength,
							 pOutputBuf,
							 estResBlockSize,
							 pResX, pResY,
							 pResW, pResH,
							 &encodedResidualSize))
	    {
		    /* Something went wrong.  Was it a buffer size
		     * problem or something else?
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		    /* The page didn't fit.  Increase the size of
		     * the buffer by 1.5x */
		estResBlockSize = estResBlockSize + estResBlockSize/2;
		IPW_FREE(pOutputBuf, cAllocatorData);
		pOutputBuf = (UInt8 *) IPW_MALLOC(estResBlockSize,
						  cAllocatorData);
		    /* We can return directly if pOutputBuf is now NULL
		     * since it's the only thing we allocate */
		ERR_GEN(pOutputBuf == NULL, cErrNomem);
	    }
	    else
	    {
		    /* it worked!  Trim the unused bytes off the end
		     * of the buffer.  If the number of bytes written
		     * is 0 (?), just free the buffer.
		     */
		if (encodedResidualSize > 0)
		{
		    pOutputBuf = (UInt8 *)IPW_REALLOC(pOutputBuf,
						      encodedResidualSize,
						      cAllocatorData);
		    ERR_GEN(pOutputBuf == NULL, cErrNomem);
		}
		else	/* buf size is 0 */
		    IPW_FREE(pOutputBuf, cAllocatorData);
		break;
	    }
	}
    }
    else	/* the client allocated the buffer so we can't change it.
		 * If the page doesn't fit, tough.
		 */
    {
	    /* Fill in the page block */
	ERR_PASS_CLEANUP(i_writeTundraResidualBlock (pIPWState, pTextCompInfoIn,
						     pInputBuf,
					             inputBufLength,
						     pOutputBuf,
						     estResBlockSize,
						     pResX, pResY,
						     pResW, pResH,
						     &encodedResidualSize));
    }

	/* Return a pointer to the final residual block and the number
	 * of bytes used.
	 */
    *pBytesUsed = encodedResidualSize;
    *ppOutputBuf = pOutputBuf;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pOutputBuf, cAllocatorData);

    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a Tundra page block.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page blocks.
 *  This routine fills in one page blocks.  If memory for this
 *  field is passed in, it is used.  If not, a memory block is
 *  allocated.<BR>
 *  A page block is composed of a page header, a position block (itself
 *  composed of a position header, a bunch of strips and an extensions
 *  section), a residual header and a residual block.  This routine
 *  writes these four components individually.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  pPageName::	     Pointer to null-terminated string containg name
 *		     of the page.
 *  pInputBuf::      Pointer to buffer containing symbol positions in
 *		     internal format.
 *  inputBufLength:: Number of bytes in input position buffer.
 *  pageBufLength::  If *ppPageBuf is not NULL, this is the number
 *		     of bytes remaining in the input page buffer.
 *  pageRequest::    Reqested page ID: 0 = all pages, counts > 1;
 *			>0 = page ID, counts = 1
 *
 * Output:::
 *  ppPageBuf::	     Pointer to returned page block pointer.
 *		     If *ppPageBuf is non-null, existing memory at this
 *		     site is used.  Otherwise, a new block is allocated.
 *  pBytesUsed::     Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.
 *
 * Notes:::
 *	If the page buffer is allocated internally, this routine will
 *	keep trying with ever-larger buffers if the page doesn't fit.
 *	If the page buffer is passed in by the client, we can't change
 *	the size of the client's buffer, so we return immediately if the
 *	page doesn't fit.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_writeTundraPageBlock(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    char	       *pPageName,
    UInt8	       *pInputBuf,
    UInt32		inputBufLength,
    UInt32		pageBufLength,
    UInt8	      **ppPageBuf,
    UInt32		pageRequest,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 estPageSize;
    UInt32	 encodedPositionSize, encodedResidualSize;
    UInt32	 pageHeaderSize, resHeaderSize, remainingBufSize;
    UInt8	*pPageStart, *pPageCurrent;
    UInt8	*pResHeader;
    UInt32	 pageWidth, pageHeight;
    UInt32	 resX, resY, resW, resH;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pPageName == NULL) ||
	    (pInputBuf == NULL) || (ppPageBuf == NULL) ||
	    (pBytesUsed == NULL), cErrNullPtr);

	/* See if the the client passed in a buffer for the page */
    if (*ppPageBuf == NULL)
    {
	    /* allocate a new buffer, estimating it's size first */
	ERR_PASS(w_estimateTundraPageBlockSize(pIPWState,
					       pTextCompInfoIn,
					       pPageName,
					       pInputBuf,
					       inputBufLength,
					       pageRequest,
					       &estPageSize));

	pPageStart = (UInt8 *) IPW_MALLOC(estPageSize, cAllocatorData);
	ERR_GEN(pPageStart == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in.  Since
				 * the new page block size is only
				 * an estimate, we don't check to see
				 * if the client has supplied something
				 * smaller here.
				 */
    {
	pPageStart = *ppPageBuf;
	estPageSize = pageBufLength;
    }

	/* Get the length of the page header.  We skip over
	 * that to start filling in the page.
	 */
    ERR_PASS_CLEANUP(i_getTundraPageHeaderSize (pIPWState, pTextCompInfoIn,
					        pPageName,
					        &pageHeaderSize));

	/* Get the length of the residual header.  We skip over
	 * that before filling in the residual.
	 */
    ERR_PASS_CLEANUP(i_getTundraResidualHeaderSize (pIPWState, pTextCompInfoIn,
						    &resHeaderSize));

	/* We'll also need the page dimensions when writing the
	 * page header.
	 */
    ERR_PASS_CLEANUP(i_getCompressedImageDimensions (pIPWState, pInputBuf,
						     &pageWidth,
						     &pageHeight));

	/* If we allocated the buffer, keep increasing its size until
	 * the page fits.  We do this by reallocating with ever
	 * larger buffer sizes.  Otherwise, just quit if the client
	 * supplied the buffer.
	 */
    if (allocatedBufferHere)
    {
	while (cTrue)
	{
		/* adjust the current page pointers to point past the
		 * header that we'll write later.
		 */
	    pPageCurrent = pPageStart + pageHeaderSize;
	    remainingBufSize = estPageSize - pageHeaderSize;
    
	    ERR_CHECK_NO_MSG(i_writeTundraPositionBlock (pIPWState,
							 pTextCompInfoIn,
							 pInputBuf,
							 inputBufLength,
							 pPageCurrent,
							 remainingBufSize,
							 pageRequest,
							 &encodedPositionSize))
	    {
		    /* Something went wrong.  Was it a buffer size
		     * problem or something else?
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		    /* The page didn't fit.  Increase the size of
		     * the buffer by 1.5x */
		estPageSize = estPageSize + estPageSize/2;
		IPW_FREE(pPageStart, cAllocatorData);
		pPageStart =
			(UInt8 *) IPW_MALLOC(estPageSize, cAllocatorData);
		    /* We can return directly if pPage is now NULL
		     * since it's the only thing we allocate */
		ERR_GEN(pPageStart == NULL, cErrNomem);
	    }
	    else
	    {
		    /* it worked.  Update the pointers and break out
		     * of the position-writing loop.
		     */
		pPageCurrent += encodedPositionSize;
		remainingBufSize -= encodedPositionSize;
		break;
	    }
	}

	    /* if that worked, write the page header, which includes the
	     * position block length.
	     */
	ERR_PASS_CLEANUP( i_writeTundraPageHeader (pIPWState, pTextCompInfoIn,
						   pPageName,
						   pageWidth,
						   pageHeight,
						   encodedPositionSize,
						   pPageStart,
						   pageHeaderSize,
						   &pageHeaderSize));

	    /* Remember the location of the residual header and
	     * skip over it.  We'll write it after writing the
	     * residual itself.
	     */
	pResHeader = pPageCurrent;
	pPageCurrent += resHeaderSize;
	remainingBufSize -= resHeaderSize;


	    /* Write the residual block and get its bounding box */
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(i_writeTundraResidualBlock (pIPWState,
							 pTextCompInfoIn,
							 pInputBuf,
							 inputBufLength,
							 pPageCurrent,
							 remainingBufSize,
							 &resX, &resY,
							 &resW, &resH,
							 &encodedResidualSize))
	    {
		    /* Something went wrong.  Was it a buffer size
		     * problem or something else?
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		    /* The page didn't fit.  Increase the size of
		     * the buffer by 1.5x */
		remainingBufSize += estPageSize/2;
		estPageSize += estPageSize/2;
		pPageStart = (UInt8 *) IPW_REALLOC(pPageStart,
						   estPageSize,
						   cAllocatorData);
		    /* We can return directly if pPage is now NULL
		     * since it's the only thing we allocate */
		ERR_GEN(pPageStart == NULL, cErrNomem);
		pPageCurrent = pPageStart + (estPageSize - remainingBufSize);
	    }
	    else
	    {
		    /* it worked!  Trim the unused bytes off the end
		     * of the buffer.  Note that the number of bytes
		     * written can't be 0 since the headers take up
		     * space even if there is no data.
		     */
		pPageStart = (UInt8 *)IPW_REALLOC(pPageStart,
						  pageHeaderSize +
						    encodedPositionSize +
						    resHeaderSize +
						    encodedResidualSize,
						  cAllocatorData);
		ERR_GEN(pPageStart == NULL, cErrNomem);
		break;
	    }
	}


    }
    else	/* the client allocated the buffer so we can't change it.
		 * If the page doesn't fit, tough.  Write the position
		 * block, then the page header in front of it since
		 * that's when we know the length of the position block.
		 * Then skip over the residual header and write the
		 * residual block.  We write the residual header
		 * at the end of the routine.
		 */
    {
	pPageCurrent = pPageStart + pageHeaderSize;
	remainingBufSize = estPageSize - pageHeaderSize;

	ERR_PASS_CLEANUP(i_writeTundraPositionBlock (pIPWState,
						     pTextCompInfoIn,
						     pInputBuf,
						     inputBufLength,
						     pPageCurrent,
						     remainingBufSize,
						     pageRequest,
						     &encodedPositionSize));

	pPageCurrent += encodedPositionSize;
	remainingBufSize -= encodedPositionSize;

	    /* if that worked, write the page header, which includes the
	     * position block length.
	     */
	ERR_PASS_CLEANUP( i_writeTundraPageHeader (pIPWState, pTextCompInfoIn,
						   pPageName,
						   pageWidth,
						   pageHeight,
						   encodedPositionSize,
						   pPageStart,
						   pageHeaderSize,
						   &pageHeaderSize));

	    /* Remember the location of the residual header and
	     * skip over it.  We'll write it after writing the
	     * residual itself.
	     */
	pResHeader = pPageCurrent;
	pPageCurrent += resHeaderSize;
	remainingBufSize -= resHeaderSize;

	    /* write the residual and get its dimensions */
	ERR_PASS_CLEANUP(i_writeTundraResidualBlock (pIPWState, pTextCompInfoIn,
						     pInputBuf,
						     inputBufLength,
						     pPageCurrent,
						     remainingBufSize,
						     &resX, &resY,
						     &resW, &resH,
						     &encodedResidualSize));
    }

	/* Write the residual header now that we know all the
	 * dimensions and the length of the residual.
	 */
    ERR_PASS_CLEANUP( i_writeTundraResidualHeader (pIPWState, pTextCompInfoIn,
						   resX, resY,
						   resW, resH,
						   encodedResidualSize,
						   pResHeader,
						   resHeaderSize,
							&resHeaderSize));

	/* Return a pointer to the final page block and the number
	 * of bytes used.
	 */
    *pBytesUsed = pageHeaderSize + encodedPositionSize +
		  resHeaderSize +  encodedResidualSize;
    *ppPageBuf = pPageStart;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedBufferHere)
	IPW_FREE(pPageStart, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create and fill in a simple Tundra document.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a page and some number of page descriptions.
 *  There are other encapsulations for symbol-based compressed
 *  data, like XIFF and Digipaper's format.
 *  This routine uses other IPWrap routines to create
 *  a simple Tundra formatted document in memory.<BR>
 *  A simple document contains a header, one dictionary and some number,
 *  probably a small number, of page blocks.  To encode more complex,
 *  and/or longer, documents, clients should use the individual
 *  routines that create headers, dictionaries, position, residual and
 *  page blocks.<BR>
 *  This routine is not particularly memory efficient in that
 *  it creates one large memory block that contains the entire file
 *  image.<BR>
 *  The dictionary written out will contain the merger of all
 *  input dictionaries (if any) plus all symbols that were found during
 *  compression and remain after symbol consolidation.
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn::	Pointer to text compression info.
 *  numOfPages::	The number of pages to be written.
 *  pageNameArray::	Pointer to array of numOfPages pointers to
 *			null-terminated strings.  pageNameArray[i] is a pointer
 *			to the name of page i.
 *  posBufArray::	Pointer to array of numOfPages pointers to
 *			position buffers.  Each of the position buffers are
 *			in our internal format.
 *  posBufLengthArray::	Array of numOfPages lengths of the position buffers.
 *  docBufLength::	If *ppDocBuf is not NULL, this is the number
 *			of bytes remaining in the input document buffer.
 *
 * Output:::
 *  ppDocBuf::		Pointer to returned document block pointer.
 *			This memory block will contain a memory image
 *			of the entire Tundra file in one contiguous chunk.
 *			If *ppDocBuf is non-null, existing memory at this
 *			site is used.  Otherwise, a new block is allocated.
 *  pBytesUsed::	Pointer to returned number of bytes actually written.
 *
 * Returns:::
 *	If no buffer is passed in and one cannot be allocated, returns
 *	cErrNomem.  If an insufficient buffer is passed in, returns
 *	cErrBufferExceeded.<BR>
 *	This routine also returns ia_invalidParm if there are more than
 *	64 input dictionaries.  If your document has more than 64
 *	input dictionaries, it's probably too big to be efficiently
 *	written with this routine.
 *
 * Notes:::
 *	If the document buffer is allocated internally, this routine will
 *	keep trying with ever-larger buffers if the document doesn't fit.
 *	If the document buffer is passed in by the client, we can't change
 *	the size of the client's buffer, so we return immediately if the
 *	document doesn't fit.<BR>
 *	
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode  CDECL
w_writeTundraDocument(
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32		numOfPages,
    char	      **pageNameArray,
    UInt8	      **posBufArray,
    UInt32	       *posBufLengthArray,
    UInt32		docBufLength,
    UInt8	      **ppDocBuf,
    UInt32	       *pBytesUsed)
{
    PROLOGUE;
    UInt32	 estPageSize, totalEstPageSize;
    UInt32	 estDictSize;
    UInt32	 docHeaderSize, numOfPagesSize;
    UInt32	 docBufSize, remainingBufSize;
    UInt32	 blockBytesUsed;
    UInt8	*pNewDocStart = NULL;
    UInt8	*pNewDocCurrent;
    Int32	 currentDictNum, maxDictNum;
    UInt32	 curPage;
    Bool	*dictArray = NULL;
    Bool	 allocatedBufferHere = cFalse;

	/* Check the pointers */
    ERR_GEN((pTextCompInfoIn == NULL) || (pageNameArray == NULL) ||
	    (posBufArray == NULL) || (posBufLengthArray == NULL) ||
	    (ppDocBuf == NULL) || (pBytesUsed == NULL), cErrNullPtr);

	/* get number of dictionaries in the current internal
	 * symbol blob.  We'll need it whether or not the client
	 * supplied an output buffer.
	 */
    ERR_PASS(w_getDictionaryNumbers(pIPWState, pTextCompInfoIn,
				    &currentDictNum, &maxDictNum));
    dictArray = (Bool *) IPW_MALLOC((maxDictNum + 1) * sizeof(Bool), cAllocatorData);
    ERR_GEN(dictArray == NULL, cErrNomem);

    for (currentDictNum = 0; currentDictNum <= maxDictNum; currentDictNum++)
	dictArray[currentDictNum] = cTrue;

	/* See if the the client passed in a buffer for the page */
    if (*ppDocBuf == NULL)
    {
	    /* Estimate the size of the buffer needed to hold
	     * the entire file.
	     */
	ERR_PASS_CLEANUP(w_getTundraHeaderSize(pIPWState, pTextCompInfoIn,
				              &docHeaderSize));

	ERR_PASS_CLEANUP(w_estimateTundraDictionaryBlockSize(pIPWState,
							     pTextCompInfoIn,
							     dictArray,
							     maxDictNum + 1,
							     0, /* all pages */
							     &estDictSize));

	ERR_PASS_CLEANUP(w_getTundraNumberOfPagesSize(pIPWState,
						     pTextCompInfoIn,
						     &numOfPagesSize));

	    /* go through all the pages, estimating the size of
	     * each one and computing the sum.
	     */
	totalEstPageSize = 0;
	for (curPage = 0; curPage < numOfPages; curPage++)
	{
	    ERR_PASS_CLEANUP(w_estimateTundraPageBlockSize(
						pIPWState,
						pTextCompInfoIn,
						pageNameArray[curPage],
						posBufArray[curPage],
						posBufLengthArray[curPage],
						0,  /* All pages, count > 1 */
						&estPageSize));
	    totalEstPageSize += estPageSize;
	}

	    /* We've got all the components, add them up */
	docBufSize = docHeaderSize +
		     estDictSize +
		     numOfPagesSize +
		     totalEstPageSize;

	pNewDocStart = (UInt8 *) IPW_MALLOC(docBufSize, cAllocatorData);

	ERR_GEN_CLEANUP(pNewDocStart == NULL, cErrNomem);
	allocatedBufferHere = cTrue;
    }
    else			/* a memory block was passed in.  Since
				 * the new document size is only
				 * an estimate, we don't check to see
				 * if the client has supplied something
				 * smaller here.
				 */
    {
	pNewDocStart = *ppDocBuf;
	docBufSize = docBufLength;
    }

	/* So, get started on the real work.  Initialize
	 * the current buffer pointers.
	 */
    remainingBufSize = docBufSize;
    pNewDocCurrent = pNewDocStart;


	/* If we allocated the buffer, keep increasing its size until
	 * the page fits.  Otherwise, just quit if the client
	 * supplied the buffer.
	 */
    if (allocatedBufferHere)
    {
	ERR_CHECK_NO_MSG(w_writeTundraHeader(pIPWState,
					     pTextCompInfoIn,
					     remainingBufSize,
					     &pNewDocCurrent,
					     &blockBytesUsed))
	{
		/* Something went wrong.  If it's a regular
		 * error, return it.  If the buffer is supposed
		 * to be too small, something is very weird.
		 * Since we allocated
		 * the buffer and it is at least as big as
		 * the header, this is an internal error.
		 */
	    ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);
		/* Buffer was too small.  Weird */
	    ERR_GEN_CLEANUP(__error__ == cErrBufferExceeded, cErrInternal);
	}

	remainingBufSize -= blockBytesUsed;
	pNewDocCurrent += blockBytesUsed;

	    /* Having successfully written the header, write the
	     * dictionary block */
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(w_writeTundraDictionaryBlock(pIPWState,
							  pTextCompInfoIn,
							  dictArray,
							  maxDictNum + 1,
							  remainingBufSize,
							  &pNewDocCurrent,
							  0,  /* all pages */
							  &blockBytesUsed))
	    {
		    /* as before, if this is something other than
		     * running out of space, stop now.  Otherwise,
		     * try again with a larger buffer.
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		remainingBufSize += docBufSize/2;
		docBufSize += docBufSize/2;
		pNewDocStart = (UInt8 *)IPW_REALLOC(pNewDocStart,
						    docBufSize,
						    cAllocatorData);
		ERR_GEN_CLEANUP(pNewDocStart == NULL, cErrNomem);
		    /* fix up pNewDocCurrent since the buffer has
		     * moved (probably).
		     */
		pNewDocCurrent = pNewDocStart + (docBufSize - remainingBufSize);
	    }
	    else	/* the dictionary fit, so update the pointers and
			 * break out of the loop.
			 */
	    {
		remainingBufSize -= blockBytesUsed;
		pNewDocCurrent += blockBytesUsed;
	        break;
	    }
	}

	    /* Now write the number of pages field */
	while (cTrue)
	{
	    ERR_CHECK_NO_MSG(w_writeTundraNumberOfPagesField(pIPWState,
							     pTextCompInfoIn,
							     numOfPages,
							     remainingBufSize,
							     &pNewDocCurrent,
							     &blockBytesUsed))
	    {
		    /* as before, if this is something other than
		     * running out of space, stop now.  Otherwise,
		     * try again with a larger buffer.
		     */
		ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);

		remainingBufSize += docBufSize/2;
		docBufSize += docBufSize/2;
		pNewDocStart = (UInt8 *)IPW_REALLOC(pNewDocStart,
						    docBufSize,
						    cAllocatorData);
		ERR_GEN_CLEANUP(pNewDocStart == NULL, cErrNomem);
		    /* fix up pNewDocCurrent since the buffer has
		     * moved (probably).
		     */
		pNewDocCurrent = pNewDocStart + (docBufSize - remainingBufSize);
	    }
	    else	/* the dictionary fit, so update the pointers and
			 * break out of the loop.
			 */
	    {
		remainingBufSize -= blockBytesUsed;
		pNewDocCurrent += blockBytesUsed;
	        break;
	    }
	}

	    /* Go through the pages, writing each one out */
	for (curPage = 0; curPage < numOfPages; curPage++)
	{
	    while (cTrue)
	    {
		ERR_CHECK_NO_MSG(w_writeTundraPageBlock(
						pIPWState,
						pTextCompInfoIn,
						pageNameArray[curPage],
						posBufArray[curPage],
						posBufLengthArray[curPage],
						remainingBufSize,
						&pNewDocCurrent,
						0, /* all pages */
						&blockBytesUsed))
		{
			/* as before, if this is something other than
			 * running out of space, stop now.  Otherwise,
			 * try again with a larger buffer.
			 */
		    ERR_GEN_CLEANUP(__error__ != cErrBufferExceeded, __error__);
    
		    remainingBufSize += docBufSize/2;
		    docBufSize += docBufSize/2;
		    pNewDocStart = (UInt8 *)IPW_REALLOC(pNewDocStart,
							docBufSize,
							cAllocatorData);
		    ERR_GEN_CLEANUP(pNewDocStart == NULL, cErrNomem);
			/* fix up pNewDocCurrent since the buffer has
			 * moved (probably).
			 */
		    pNewDocCurrent = pNewDocStart + (docBufSize -
						     remainingBufSize);
		}
		else	/* the dictionary fit, so update the pointers and
			 * break out of the inner loop.
			 */
		{
		    remainingBufSize -= blockBytesUsed;
		    pNewDocCurrent += blockBytesUsed;
		    break;
		}
	    }	/* end of while(true) loop that writes on page */
	}	/* end of loop that writes pages */


	    /* trim off the unused portion of our buffer. */
	docBufSize -= remainingBufSize;
	pNewDocStart = (UInt8 *)IPW_REALLOC(pNewDocStart, docBufSize,
					    cAllocatorData);
	ERR_GEN_CLEANUP(pNewDocStart == NULL, cErrNomem);
	remainingBufSize = 0;  /* need to do this so the bytesUsed
				* computation at the end works */

    }		/* end of case where buffer was allocated internally */

    else	/* the client allocated the buffer so we can't change it.
		 * If the entire document doesn't fit, tough.
		 */
    {
	    /* Write the header */
	ERR_PASS_CLEANUP(w_writeTundraHeader(pIPWState,
					    pTextCompInfoIn,
					    remainingBufSize,
					    &pNewDocCurrent,
					    &blockBytesUsed));

	remainingBufSize -= blockBytesUsed;
	pNewDocCurrent += blockBytesUsed;

	    /* Write the dictionary */
	ERR_PASS_CLEANUP(w_writeTundraDictionaryBlock(pIPWState,
						      pTextCompInfoIn,
						      dictArray,
						      maxDictNum + 1,
						      remainingBufSize,
						      &pNewDocCurrent,
						      0, /* all pages */
						      &blockBytesUsed));

	remainingBufSize -= blockBytesUsed;
	pNewDocCurrent += blockBytesUsed;

	    /* Write the number of pages field */
	ERR_PASS_CLEANUP(w_writeTundraNumberOfPagesField(pIPWState,
						        pTextCompInfoIn,
						        numOfPages,
						        remainingBufSize,
						        &pNewDocCurrent,
						        &blockBytesUsed));

	remainingBufSize -= blockBytesUsed;
	pNewDocCurrent += blockBytesUsed;

	    /* write each of the pages */
	for (curPage = 0; curPage < numOfPages; curPage++)
	{
	    ERR_PASS_CLEANUP(w_writeTundraPageBlock(pIPWState,
						    pTextCompInfoIn,
						    pageNameArray[curPage],
						    posBufArray[curPage],
						    posBufLengthArray[curPage],
						    remainingBufSize,
						    &pNewDocCurrent,
						    0, /* all pages */
						    &blockBytesUsed));

	    remainingBufSize -= blockBytesUsed;
	    pNewDocCurrent += blockBytesUsed;
	}
    }	/* end of case where client supplied entire output buffer */


	/* Return a pointer to the final page block and the number
	 * of bytes used.
	 */
    IPW_FREE(dictArray, cAllocatorData);
    *pBytesUsed = docBufSize - remainingBufSize;
    *ppDocBuf = pNewDocStart;

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(dictArray, cAllocatorData);
    if (allocatedBufferHere)
	IPW_FREE(pNewDocStart, cAllocatorData);

    ERR_RETURN;
}




/*

 $Log: tctne.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.5  1998/08/04 00:21:30  aber
 tctne.c -> /product/ipcore/ipwrap/src/RCS/tctne.c,v

 CommonCheck: Support multiple dictionaries when token compressing, used to retain
 single-use tokens in the same dictionary -- to be used when color
 tokenizing.

 Revision 1.4  1998/01/14 00:15:52  ddavies
 tctne.c -> /product/ipcore/ipwrap/src/RCS/tctne.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.3  1996/07/19 00:37:20  ddavies
 tctne.c -> /product/ipcore/ipwrap/src/RCS/tctne.c,v

 Fix all the places where we trim the size of an allocated buffer to fit
 the real size.  If the real size is 0, free the buffer instead of reallocating
 it.

 Revision 1.2  1996/07/04 03:12:21  ddavies
 tctne.c -> /product/ipcore/ipwrap/src/RCS/tctne.c,v

 Call any routine that might return a buffer overflow condition that could
 be corrected by allocating a larger buffer with the new ERR_CHECK_NO_MSG
 macro instead of ERR_CHECK.  This lets us handle this condition silently.

 Revision 1.1  1996/07/02 23:07:47  ddavies
 Initial revision


 */

