/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *   tctnut.c:
 *
 *	Common routines for reading and writing Tundra formatted
 *	images.
 *
 *	w_getTundraHeaderSize
 *		Returns size of Tundra header in bytes.
 *
 *	w_getTundraNumberOfPagesSize
 *		Returns size of Tundra "Number of pages" field in bytes.
 *
 *	w_setDictionaryNumber
 *		Sets the current dictionary number in the internal
 *		text compression data structure.  All subsequent symbols
 *		will belong to this dictionary.
 *
 *	w_getDictionaryNumbers
 *		Returns current dictionary number and maximum dictionary
 *		number seen..
 *
 */

#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"

/* PROTOTYPES */
#include "tcompr.pub"
#include "shrcmprs.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: tctnut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/* These are routines that are used both to produce Tundra file images in
 * memory and to decode such images.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return the size of a Tundra header in bytes.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header.  This routine returns the size of the header so a client
 *  can allocate it.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *
 * Output:::
 *  pHeaderSize::    Pointer to returned header size in bytes.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getTundraHeaderSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32             *pHeaderSize)
{
    PROLOGUE;

    ERR_PASS(i_getTundraHeaderSize (pIPWState, pTextCompInfoIn, pHeaderSize));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Return the size of a Tundra "number of pages" field in bytes.
 *
 * LongDescription:::
 *  A Tundra format file consists of several parts, beginning with
 *  a header, a dictionary and some number of page descriptions.  This
 *  routine returns the size of the field that holds the number of page
 *  descriptions.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *
 * Output:::
 *  pNumOfPagesSize:: Pointer to returned "number of pages" field size in bytes.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getTundraNumberOfPagesSize (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    UInt32             *pNumOfPagesSize)
{
    PROLOGUE;

    ERR_PASS(i_getTundraNumberOfPagesSize (pIPWState, pTextCompInfoIn,
					   pNumOfPagesSize));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of the current dictionary.
 *
 * LongDescription:::
 *  A Tundra format file may contain several dictionaries.  Each has a
 *  unique number, starting from 0.  This routine is used to inform the
 *  internal compression data structures of the number of the dictionary
 *  to which new symbols should be assigned.  New symbols may arrive
 *  from a new dictionary being read in or from the classification process.
 *
 * Input:::
 *  pIPWState::       Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *  newDictionaryNum:: number of the current dictionary.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_setDictionaryNumber (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    Int32               newDictionaryNum)
{
    PROLOGUE;

    ERR_GEN(pTextCompInfoIn == NULL, cErrNullPtr);

	/* Let the ipshared routine do the real work */
    ERR_PASS(i_setDictionaryNumber (pIPWState, pTextCompInfoIn,
				    newDictionaryNum));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the current and maximum dictionary numbers.
 *
 * LongDescription:::
 *  A Tundra format file may contain several dictionaries.  Each has a
 *  unique number, starting from 0.  This routine retrieves the number
 *  of the current dictionary as well as the maximum dictionary number
 *  seen so far.  There is no guarantee that all dictionaries from
 *  0 through maxDictionaryNum have been seen.
 *
 * Input:::
 *  pIPWState::       Pointer to standard IPWrap state structure.
 *  pTextCompInfoIn:: Pointer to text compression info.
 *
 * Output:::
 *  pCurDictionaryNum:: Pointer to returned number of the current dictionary.
 *  pMaxDictionaryNum:: Pointer to returned maximum dictionary number seen
 *			so far.
 *
 * FunctionCategory:::
 *	compression
 *
 * Keywords:::
 *	compression, symbol-based
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getDictionaryNumbers (
    IPWState           *pIPWState,
    void               *pTextCompInfoIn,
    Int32              *pCurDictionaryNum,
    Int32	       *pMaxDictionaryNum)
{
    PROLOGUE;

    ERR_GEN(pTextCompInfoIn == NULL, cErrNullPtr);

	/* Let the ipshared routine do the real work */
    ERR_PASS(i_getDictionaryNumbers (pIPWState, pTextCompInfoIn,
				     pCurDictionaryNum,
				     pMaxDictionaryNum));

    NORMAL_RETURN;
}


/*

 $Log: tctnut.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 Imported sources

 Revision 1.2  1998/01/14 00:15:54  ddavies
 tctnut.c -> /product/ipcore/ipwrap/src/RCS/tctnut.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.1  1996/07/02 23:05:26  ddavies
 Initial revision


 */



