/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* 
 * thin.c
 * 
 *    Iterative thinning.
 *
 * Exports:
 *    thin8 - thin, preserving 8-connectedness
 *
 */


#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "morphg.h"
#include "logicops.h"

/* prototypes */
#include "props.pub"
#include "strel.pub"
#include "morph.pub"
#include "logicops.pub"
#include "thin.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: thin.c,v $; $Revision: 1.3 $; $Date: 2000/06/07 00:31:52 $");
static ErrCode CDECL
wp_rotateStrel90 (IPWState *pIPWState, STREL *pStrelIn, STREL **ppStrelOut);

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Topology-preserving erosion, one iteration
 *
 * LongDescription:::
 *  Does a topology-preserving erosion, one iteration.
 *  This preserves 8-connectedness in the
 *  foreground (which preserves 4-connectedness in the background)
 *  using the "s8,5" strel from Dan Bloomberg's paper.
 *  NOTE that this is intended to be used to get rid of stuff
 *  around the edge of a connected component--if you want to 
 *  do nice skeletonization, there are better schemes, like using
 *  sets of composite strels, or some more recent non-iterative 
 *  schemes.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Source image
 *
 * Output:::
 *  pPrDest::        Thinned image
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_thin8 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    PIXR              **pPrDest)
{
    PROLOGUE;
    STREL *pThinStrel1 = NULL, *pThinStrel2 = NULL, *pThinStrel3 = NULL, 
	  *pThinStrel4 = NULL;
    PIXR *prTmp = NULL, *prDest = NULL;


    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    if (*pPrDest == NULL)
    {
	ERR_PASS (w_copy (pIPWState, prSrc, pPrDest));
    }
    else
    {
	ERR_GEN_PIXR (*pPrDest, cByteOrderBig, cDepth1, cChannels1, cFrame32);
	if (*pPrDest != prSrc)
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prDest));
    }

    prDest = *pPrDest;

    ERR_PASS (w_createStrel(pIPWState, 3, 3, &pThinStrel1));
    pThinStrel1->cx = 1;
    pThinStrel1->cy = 1;
    pThinStrel1->data[0][0] = MISS;
    pThinStrel1->data[1][0] = MISS;
    pThinStrel1->data[2][0] = MISS;
    pThinStrel1->data[0][1] = DONT_CARE;
    pThinStrel1->data[1][1] = DONT_CARE;
    pThinStrel1->data[2][1] = DONT_CARE;
    pThinStrel1->data[0][2] = HIT;
    pThinStrel1->data[1][2] = HIT;
    pThinStrel1->data[2][2] = DONT_CARE;

    ERR_PASS_CLEANUP (wp_rotateStrel90 (pIPWState, pThinStrel1, &pThinStrel2));
    ERR_PASS_CLEANUP (wp_rotateStrel90 (pIPWState, pThinStrel2, &pThinStrel3));
    ERR_PASS_CLEANUP (wp_rotateStrel90 (pIPWState, pThinStrel3, &pThinStrel4));

    ERR_PASS_CLEANUP (w_erode (pIPWState, prSrc, pThinStrel1, &prTmp));
	
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prDest));
    ERR_PASS_CLEANUP (w_erode (pIPWState, prDest, pThinStrel2, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prDest));
    ERR_PASS_CLEANUP (w_erode (pIPWState, prDest, pThinStrel3, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prDest));
    ERR_PASS_CLEANUP (w_erode (pIPWState, prDest, pThinStrel4, &prTmp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, prDest));

    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &pThinStrel1));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &pThinStrel2));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &pThinStrel3));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &pThinStrel4));
    if (prTmp) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pThinStrel1) w_destroyStrel (pIPWState, &pThinStrel1);
    if (pThinStrel2) w_destroyStrel (pIPWState, &pThinStrel2);
    if (pThinStrel3) w_destroyStrel (pIPWState, &pThinStrel3);
    if (pThinStrel4) w_destroyStrel (pIPWState, &pThinStrel4);
    if (prTmp) 
	w_destroyPixr (pIPWState, &prTmp);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Make a copy of a strel which is rotated 90 degrees from the strel passed.
 *
 * LongDescription:::
 *    Make a copy of a strel which is rotated 90 degrees from the strel passed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pStrelIn::       Strel to rotate
 *
 * Output:::
 *  ppStrelOut::     Rotated strel
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_rotateStrel90 (
    IPWState           *pIPWState,
    STREL              *pStrelIn,
    STREL             **ppStrelOut)
{
    PROLOGUE;
    Int32 i, j, w, h;
    STREL *pStrelOut;

    w = pStrelIn->nx;
    h = pStrelIn->ny;
    ERR_PASS (w_createStrel (pIPWState, h, w, ppStrelOut));
    pStrelOut = *ppStrelOut;
    pStrelOut->cy = pStrelIn->cx;
    pStrelOut->cx = pStrelIn->cy;

    for (i = 0; i < h; i++)
    {
	for (j = 0; j < w; j++)
	{
	    pStrelOut->data[i][j] = pStrelIn->data[h - j - 1][i];
	}
    }

    NORMAL_RETURN;
}
/* 
 *
 * $Log: thin.c,v $
 * Revision 1.3  2000/06/07 00:31:52  jlin
 *  minor format change
 *
 * Revision 1.2  2000/06/07 00:28:27  jlin
 * change wp_jcWriteCallback
 *
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.14  1996/01/04 20:27:49  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.13  1995/12/13  04:14:19  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.12  1995/11/13  20:17:31  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.11  1995/10/13  22:06:22  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.10  1995/10/03  01:59:43  macleod
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.9  1995/09/21  23:00:46  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.8  1995/09/11  21:54:52  macleod
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.7  1995/09/07  01:36:16  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.6  1995/09/06  20:24:36  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.5  1995/09/01  21:03:02  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.4  1995/08/31  21:40:09  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.3  1995/08/25  01:38:55  ejaquith
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * Fixed warnings
 *
 * Revision 1.2  1995/08/17  21:55:50  macleod
 * thin.c -> /product/ipcore/ipwrap/src/RCS/thin.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.1  1995/07/13  17:48:05  macleod
 * Initial revision
 */
