/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	thrshold.c:
 *
 *          Routines to convert gray images to binary images.
 *		
 *
 *          8, 4 or 2 bit/pixel to 1 bit/pixel by either discarding
 *	    low order bits or comparing with 1 or 2 thresholds:
 *                  w_threshold()
 *
 *	    8 bits/pixel to 1 bit/pixel using different threshold
 *	    for every pixel (a threshold image)
 *		    w_thresholdByPixel
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "thrshold.h"
#include "frames.pub"
/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "morph.pub"
#include "thrshold.pub"
#include "shrip.pub"
#include "shrpixr.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: thrshold.c,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:20 $ */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert gray images to binary images by thresholding pixel values.
 *
 * LongDescription:::
 *  There are several ways to convert gray images to binary images.  This
 *  procedure does so by (logically) comparing each pixel value in the
 *  source to two thresholds.  If threshSense == THRESH_NORM and the 
 *  pixel value in the source is inside the range described by the 
 *  two thresholds (lowThresh <= pixelValue <= highThresh), the
 *  pixel becomes a 1 in the output.  Otherwise, it's set to 0.
 *  If threshSense == THRESH_INVERT, the sense of the output is inverted.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to input gray image.
 *  lowThresh::      Lower threshold value.  This number must be in the
 *		     range of legal pixel values for images of the depth
 *		     of prS.  For example, if prS is an 8 bpp image,
 *		     0 <= lowThresh <= 255.  If prS is a 4 bpp image,
 *		     0 <= highThresh <= 15.
 *  highThresh::     Upper threshold value.  The same restrictions on value
 *		     relative to image depth apply.
 *  threshSense::    If equal to THRESH_NORM, all pixels in the range
 *		     [lowThresh, highThresh] turn into 1's.  If threshSense
 *		     equals THRESH_INVERT, these pixels turn into 0's.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 *  The client may treat this function as a bandpass filter with an optional
 *  inversion of the output.  The inversion is usually needed since the black/
 *  white sense of gray images is backward from that of binary images 
 *  (0 = black in gray; 0 = white in binary).
 *<BR><BR>
 *  This function attempts to choose the most efficient primitive for
 *  performing the desired operation.  Let's define two quantities,
 *  midGray and maxGray.  midGray is the lowest value in the upper half of
 *  the range of possible pixel values ( (1 << (depth-1)) ).  For example,
 *  in 8 bpp images, midGray == 128.  maxGray is the maximum legal pixel 
 *  value for a given image depth ( (1 << depth) - 1 ).  For example, 
 *  in 8 bpp images, maxGray equals 255.
 *<BR><BR>
 *  In general, functions in which either (lowThresh == 0) or
 *  (highThresh == maxGray) will be quicker than thresholds in which 
 *  neither of these are true.  The fastest thresholding will be achieved 
 *  when either ((lowThresh == 0) && (highThresh == midGray - 1)) or 
 *  ((lowThresh == minGray) && (highThresh == maxGray)).  The setting of
 *  threshSense has no performance impact.
 *<BR><BR>
 *  There are often two ways to acheive a given result.  For example,
 *  thresholding an 8 bpp gray image with a range [0, 127] and threshSense
 *  equal to THRESH_NORM gives the same result as thresholding with a range
 *  of [128, 255] and threshSense equal to THRESH_INVERSE.
 *<BR><BR>
 *  Note that not all functions are defined on all depths.  In particular,
 *  if the source image has depth 2, we can only do midpoint thresholding
 *  (range must be either [0,1] or [2,3]).
 *<BR><BR>
 *  The destination image can be the same as the source image.  When this
 *  happens, no additional temporary memory is used.  The depth and bytes/
 *  line of the destination memory and the amount of memory it uses will 
 *  be reduced appropriately.  However, if a failure occurs during an in-place
 *  thresholding operation, the source data is probably irretrievably
 *  mangled.
 *
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, comparison, pixel mapping, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  L. Perry, D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_threshold (
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               lowThresh,
    Int32               highThresh,
    Int32               threshSense,
    PIXR              **pPrD)
{
    PROLOGUE;
    Int32		 wS, hS, dS, frame;
    Int32		 midGray, maxGray;
    UInt8		*srcImage, *dstImage;
    Int32		 srcWidth, srcHeight;
    Int32		 srcBpl, dstBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;
    PIXR		*prT;
    enum {inv_highpass, highpass, 
	  midpoint, inv_midpoint, 
	  bandpass, inv_bandpass} threshType;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrameAny);

/* Get all of the image parameters */
    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS( w_pixrGetFrame(pIPWState, prS, &frame));
    ERR_PASS( w_pixrGetImageAll(pIPWState, prS, &srcImage));
    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &srcBpl));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &srcWidth));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &srcHeight));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

/* Calculate the max gray level and the midpoint threshold value */
    maxGray = (1 << dS) - 1;
    midGray = (1 << (dS-1));

/* Validate parameters */
    ERR_GEN((lowThresh < 0) || (lowThresh > maxGray), cErrInvalidParm );
    
    ERR_GEN((highThresh < 0) || (highThresh > maxGray), cErrInvalidParm );
    
    ERR_GEN(! ((threshSense==THRESH_NORM) || (threshSense==THRESH_INVERSE)),
	    cErrInvalidParm );
    
/* 
   Determine what type of thresholding operation we have from the
   low and high thresholding args and the normal/inverse setting
*/
    if (((lowThresh == midGray) && (highThresh == maxGray) &&
	 (threshSense == THRESH_NORM))  ||
	((lowThresh == 0) && (highThresh == (midGray-1)) &&
	 (threshSense == THRESH_INVERSE)))
        threshType = midpoint;

    else
    if (((lowThresh == 0) && (highThresh == (midGray-1)) &&
	 (threshSense == THRESH_NORM))  ||
	((lowThresh == midGray) && (highThresh == maxGray) &&
	 (threshSense == THRESH_INVERSE)))
        threshType = inv_midpoint;

    else
    if (lowThresh == 0)	/* and highThresh != (midGray-1) */
    {
	if (threshSense == THRESH_NORM)
	{
		/* note that i_highpassfiltI turns all pixels >= highThresh
		 * to 0 and those < highThresh to 1.  this is not what we
		 * want.  we want everything <= highThresh to be 1 so we
		 * must increase the upper threshold before we call
		 * i_highpassfiltI.  this is only a problem if highThresh
		 * was already equal to maxGray (although why anyone would
		 * do this is beyond me).
		 */
	    if (highThresh == maxGray)
		threshType = highpass;
	    else 
	    {
		highThresh++;
		threshType = inv_highpass;
	    }
	}
	else	/* threshSense == THRESH_INVERSE */
	{
		/* Here, we want the exact opposite of the actions above.
		 * Note that highpass uses lowThresh and inv_highpass uses
		 * highThresh.
		 */
	    if (highThresh == maxGray)
		threshType = inv_highpass;
	    else 
	    {
		lowThresh = highThresh + 1;	/* because highpass uses
						 * lowThresh. */
		threshType = highpass;
	    }
	}
    }
    else
    if (highThresh == maxGray)
    {
	if (threshSense == THRESH_NORM)
	{
		/* i_highpass does the right thing, so leave the thresholds
		 * alone.
		 */
	    threshType = highpass;
	}
	else /* threshSense == THRESH_INVERSE */
	{
	    highThresh = lowThresh;		/* because inv_highpass uses
						 * highThresh. */
	    threshType = inv_highpass;
	}
    }

    else	/* We've exhausted the midpoint, lowpass and highpass cases,
		 * only the bandpass case is left.
		 */
    if (threshSense==THRESH_NORM)
        threshType = bandpass;
    else 
        threshType = inv_bandpass;

    /* 2 bpp unsupported for inv_highpass, highpass */
    ERR_GEN(((threshType == inv_highpass) || (threshType == highpass)) &&
	    (dS == 2), cErrInvalidParm);

    /* bandpass thresholding not supported for 2 bpp */
    ERR_GEN((threshType==bandpass || threshType==inv_bandpass) &&
	    (dS == 2), cErrInvalidParm);

    prT = *pPrD;
	/* if a non-null destination is provided AND it points
	 * to the source image (threshold in place), change
	 * the dest image depth to 1 (and fix the bytes/line
	 * so we can check the dest image.
	 */
    if (prT == prS)
    {
	Int32    bplT;
	Int32	 offsetTo00;
	    /* Set the new depth now so we can get the new
	     * bytes/line.  If the conversion fails, we'll
	     * restore the original values (dpeth held in dS).
	     */
	ERR_PASS_CLEANUP( w_pixrSetDepth(pIPWState, prT, 1));
	ERR_PASS_CLEANUP( w_pixrBpl(pIPWState, prT, &bplT));
	ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, prT, bplT));

	    /* Here's a tricky part.  The offset from the beginning
	     * of the memory buffer containing the image to the
	     * beginning of the image might not be the same in
	     * the source and dest images.  Calculate the
	     * dest image pointer relative to the beginning of
	     * the image buffer here.  We do this by finding the
	     * the offset from the beginning of the image buffer
	     * to location (0, 0) (inside the frame), then subtracting
	     * off the frame.
	     */
	ERR_PASS_CLEANUP (wp_pixrGetImageBuffer(pIPWState, prT, &dstImage));
	ERR_PASS_CLEANUP (w_pixrImageOffset(pIPWState, dstImage, bplT, 1, frame,
					    &offsetTo00));
	dstImage = dstImage + offsetTo00 - (frame/8 + frame * bplT);

    }
    else
    {
	    /* Check the dest pixr to see if it's the right size.
	     * If there is no dest pixr, allocate one.
	     */
	ERR_PASS_CLEANUP(wp_createPixrConditionalNoInit(pIPWState, wS, hS, 1,
							xSdpi, ySdpi,
							frame, &prT,
							&rightFrameColsToClear,
							&bottomFrameRowsToClear,
							&allocatedHere));
	ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, prT, &dstImage));
    }

	/* Get the dst bytes/line */
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prT, &dstBpl));

    switch (threshType)
    {
    case midpoint:
    
        switch(dS)
        {
        case 2:  /* equiv to threshold2To1BppPr() */
	    ERR_PASS_CLEANUP( i_threshold2To1bpp (pIPWState, srcImage,
						  srcWidth, srcHeight,
						  srcBpl, dstImage, dstBpl,
						  NULL, NULL, NULL, NULL));
            break;

        case 4:  /* equiv to threshold4To1BppPr() */
	    ERR_PASS_CLEANUP( i_threshold4To1bpp (pIPWState, srcImage,
						  srcWidth, srcHeight,
						  srcBpl, dstImage, dstBpl,
						  NULL, NULL, NULL, NULL));
            break;

        case 8:  /* equiv to thresholdPr() */
	    ERR_PASS_CLEANUP( i_threshold (pIPWState, srcImage,
					   srcWidth, srcHeight,
					   srcBpl, dstImage, dstBpl, 
					   NULL, NULL, NULL, NULL));
            break;

        default:
	    ERR_GEN_CLEANUP(1, cErrInvalidParm);

        }

        break;

    case inv_midpoint:

        switch(dS)
        {
        case 2:  /* equiv to thresholdI2To1BppPr() */
            ERR_PASS_CLEANUP( i_thresholdI2To1bpp (pIPWState, srcImage,
						   srcWidth, srcHeight,
						   srcBpl, dstImage, dstBpl,
						   NULL, NULL, NULL, NULL));
            break;

        case 4:  /* equiv to thresholdI4To1BppPr() */
            ERR_PASS_CLEANUP( i_thresholdI4To1bpp (pIPWState, srcImage,
						   srcWidth, srcHeight,
						   srcBpl, dstImage, dstBpl,
						   NULL, NULL, NULL, NULL));
            break;

        case 8:  /* equiv to thresholdIPr() */
            ERR_PASS_CLEANUP( i_thresholdI (pIPWState, srcImage,
					    srcWidth, srcHeight,
					    srcBpl, dstImage, dstBpl,
					    NULL, NULL, NULL, NULL));
            break;

        default:
	    ERR_GEN_CLEANUP(1, cErrInvalidParm);
        }

       break;

    
    case inv_highpass:  /* inv_highpass threshold at arbitrary gray level  */

        switch(dS)
        {
        case 4:  /* equiv to inv_highpass4To1BppPr() */
            ERR_PASS_CLEANUP( i_highpassfiltI4To1bpp (pIPWState, srcImage,
						      srcWidth,
						      srcHeight,
						      srcBpl, highThresh, 
						      dstImage, dstBpl,
						      NULL, NULL, NULL, NULL));
            break;

        case 8:  /* equiv to inv_highpassPr() */
            ERR_PASS_CLEANUP( i_highpassfiltI (pIPWState, srcImage,
					       srcWidth, srcHeight,
					       srcBpl, highThresh,
					       dstImage, dstBpl,
					       NULL, NULL, NULL, NULL));
            break;

        default:
	    ERR_GEN_CLEANUP(1, cErrInvalidParm);
        }

       break;

    case highpass: /* highpass threshold at arbitrary gray level  */
    
        switch(dS)
        {
        case 4:  /* equiv to highpass4To1BppPr() */
            ERR_PASS_CLEANUP( i_highpassfilt4To1bpp (pIPWState, srcImage,
						     srcWidth,
						     srcHeight,
						     srcBpl, lowThresh,
						     dstImage, dstBpl,
						     NULL, NULL, NULL, NULL));
            break;

        case 8:  /* equiv to highpassPr() */
            ERR_PASS_CLEANUP( i_highpassfilt (pIPWState, srcImage,
					      srcWidth, srcHeight,
					      srcBpl, lowThresh, 
					      dstImage, dstBpl,
					      NULL, NULL, NULL, NULL));
            break;

        default:
	    ERR_GEN_CLEANUP(1, cErrInvalidParm);
        }

        break;

    case bandpass:  /* pass levels within a gray level window */
    
        switch(dS)
        {
        case 4:  /* equiv to bandpass4To1bppPr() */
            ERR_PASS_CLEANUP( i_passfilt4To1bpp (pIPWState, srcImage,
						 srcWidth, srcHeight,
						 srcBpl, lowThresh, 
						 highThresh, dstImage, dstBpl,
						 NULL, NULL, NULL, NULL));
                           
            break;

        case 8:  /* equiv to bandpassPr() */
            ERR_PASS_CLEANUP( i_passfilt (pIPWState, srcImage,
					  srcWidth, srcHeight,
					  srcBpl, lowThresh, highThresh, 
					  dstImage, dstBpl,
					  NULL, NULL, NULL, NULL));
                           
            break;

        default:
	    ERR_GEN_CLEANUP(1, cErrInvalidParm);
        }

        break;

    case inv_bandpass:  /* pass levels outside a gray level window */

        switch(dS)
        {
        case 4:  /* equiv to bandpassI4To1bppPr() */
            ERR_PASS_CLEANUP( i_passfiltI4To1bpp (pIPWState, srcImage,
						  srcWidth, srcHeight,
						  srcBpl, lowThresh, 
						  highThresh, dstImage, dstBpl,
						  NULL, NULL, NULL, NULL));
                           
            break;

        case 8:  /* equiv to bandpassIPr() */
            ERR_PASS_CLEANUP( i_passfiltI (pIPWState, srcImage,
					   srcWidth, srcHeight,
					   srcBpl, lowThresh, highThresh, 
					   dstImage, dstBpl,
					   NULL, NULL, NULL, NULL));
                           
            break;

        default:
	    ERR_GEN_CLEANUP(1, cErrInvalidParm);
        }

        break;

    default:
	ERR_GEN_CLEANUP(1, cErrInvalidParm);

    } /* end of switch(threshType) */

    

    /* it worked, so change the memory for the data if
     * we were doing the threshold in place.
     */
    if (prS == prT)
    {
	UInt8     *pImage;
	Int32      fw, fh;
	
	ERR_PASS_CLEANUP( w_reallocPixrData(pIPWState, prT));
	    /* Note that when we clear all the frames, we have to get
	     * the pointer to the dest pixr again.  There is no
	     * guarantee that realloc will just the unused part
	     * of the block off the end.  At least Purify doesn't
	     * think so.
	     */
	ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prT, &fw));
	ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prT, &fh));
	ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prT, &pImage));

	/* On error, don't bother trying to recover, we've already lost
	 * both the content and the memory buffer associated
	 * with the original image.  Besides, what could
	 * go wrong?
	 */
	ERR_PASS( i_clearAllFrames (pIPWState, pImage, fw, fh, dstBpl, 1));
    }
    else
    {
	    /* if the dest is possibly old, but the transfer was not
	     * done in place, clear any dest frame that needs to be
	     * cleared.
	     */
	if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	    ERR_PASS_CLEANUP(wp_clearRightBottomFrames(pIPWState, prT,
						       rightFrameColsToClear,
						       bottomFrameRowsToClear));
    }

	/* Return the thresholded pixr as the destination argument */
    *pPrD = prT;
    NORMAL_RETURN;


ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prT);
    else
	if (prS == prT)		/* if threshold in place, restore prS's
				 * parameters */
	{
	    w_pixrSetDepth(pIPWState, prS, dS);
	    w_pixrSetBpl(pIPWState, prS, srcBpl);
	}
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert gray images to binary images by thresholding pixel values.
 *
 * LongDescription:::
 *  Pixel-by-pixel thresholding. The thresholding image is a pixel-by-pixel
 *  map of the threshold to use. If the threshold map image is smaller
 *  than the source image, the thresholds will be computed by doing a
 *  nearest-neighbor resize of the threshold map image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to input gray image.
 *  prThresh::       Pointer to threshold map image.
 *  threshSense::    If equal to THRESH_NORM, all pixels in the range
 *		     [0, thresholdPixel] turn into 1's.  If threshSense
 *		     equals THRESH_INVERT, these pixels turn into 0's.
 *
 * Output:::
 *  pPrD::           Pointer to a pointer to a destination image.  As usual,
 *		     if a destination image is supplied, its dimensions, depth,
 *		     etc are checked.  If no dest image is supplied, one
 *		     is created.
 *
 * Notes:::
 * FunctionCategory:::
 *  depth conversion
 *
 * Keywords:::
 *  depth conversion, comparison, pixel mapping, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  P. MacLeod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_thresholdByPixel (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR               *prThresh,
    Int32               threshSense,
    PIXR              **pPrD)
{
    PROLOGUE;
    Int32		 wS, hS, wThresh, hThresh;
    UInt8		*pSrc, *pDst, *pThresh;
    Int32		 srcBpl, dstBpl, threshBpl;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*prDest;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

	/* check the input pointers */
    ERR_GEN(prS == NULL || pPrD == NULL || prThresh == NULL, cErrNullPtr);

	/* Check the input images */
    ERR_GEN_PIXR(prS, cByteOrderAny, cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR(prThresh, cByteOrderAny, cDepth8, cChannels1, cFrame32);

    /* Get all of the image parameters */
    ERR_PASS (w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS (w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS (w_pixrGetBpl (pIPWState, prS, &srcBpl));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prS, &xSdpi, &ySdpi));

    ERR_PASS (w_pixrGetWidth (pIPWState, prThresh, &wThresh));
    ERR_PASS (w_pixrGetHeight (pIPWState, prThresh, &hThresh));
    ERR_PASS (w_pixrGetBpl (pIPWState, prThresh, &threshBpl));

    /* Make sure the threshold map isn't too big.  Also discard
     * any extra threshold values.
     */
    if (wThresh > wS)
    {
	ERR_GEN ((wThresh - wS > 32), cErrSrcDimensNotEqual);
	wThresh = wS;
    }
    if (hThresh > hS)
    {
	ERR_GEN ((hThresh - hS > 32), cErrSrcDimensNotEqual);
	hThresh = hS;
    }

    /* Check the dest pixr to see if it's the right size.
     * If there is no dest pixr, allocate one.
     */
    prDest = *pPrD;
    ERR_PASS_CLEANUP (wp_createPixrConditionalNoInit(pIPWState, wS, hS, 1,
						     xSdpi, ySdpi,
						     cFrameBits, &prDest,
						     &rightFrameColsToClear,
						     &bottomFrameRowsToClear,
						     &allocatedHere));

    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prDest, &dstBpl));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prS, &pSrc));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prDest, &pDst));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prThresh, &pThresh));

	/* Add 2 * frame to width and height.  This accounts
	 * for any changes we made to wThresh and hThresh above.
	 * Note that we've already checked to see if the frame
	 * was cFrameBits wide, which it was.
	 */
    wS += 2*cFrameBits;
    hS += 2*cFrameBits;
    wThresh += 2*cFrameBits;
    hThresh += 2*cFrameBits;

	/* Do the real work.  This function will also scale the input
	 * threshold image if necessary.
	 */
    ERR_PASS_CLEANUP (i_thresholdByPixel (pIPWState,
					  pSrc, wS, hS, srcBpl,
					  pThresh, wThresh, hThresh, threshBpl,
					  threshSense == THRESH_NORM ?
							cFalse : cTrue,
					  pDst, dstBpl,
					  NULL, NULL, NULL, NULL));

    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP(wp_clearRightBottomFrames(pIPWState, prDest,
						   rightFrameColsToClear,
						   bottomFrameRowsToClear));
    *pPrD = prDest;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere && prDest)
	w_destroyPixr (pIPWState, &prDest);

    ERR_RETURN;
}

/*
$Log: thrshold.c,v $
Revision 1.1.1.1  1999/09/20 18:19:20  jlin
Imported sources

Revision 1.34  1998/02/24 03:49:45  ddavies
ddavies on Mon Feb 23 19:47:11 PST 1998

thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v

Convert w_thresholdByPixel into a wrapper for the new ipshared function
i_thresholdByPixel.  The ipshared implementation is significantly faster
(2x?).  It also eliminates a bug.  The old implementation failed to
stop borrows at pixel boundaries, so that it subtracted iethe
ooops, that's "either", (thresholdVal) or (thresholdVal+1) from a source
pixel depending on the result of the subtraction on the pixel to the right.

Revision 1.33  1998/01/14 00:15:56  ddavies
thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.32  1997/06/12 00:49:08  xzhu
xzhu on Wed Jun 11 17:46:37 PDT 1997

thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v

Fixed bug in w_thresholdByPixr. The last bytes ( less than 32 ) was not to assign to output image and caused radom image in right edge.

Revision 1.31  1997/04/15 22:45:47  ddavies
thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v

Use non-initialized PIXR creation where possible.

Revision 1.30  1996/09/24 02:52:56  ddavies
thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v

Use w_pixrImageOffset and wp_pixrGetImageBuffer to calculate the new
offset from the beginning of the destination buffer in w_threshold when
the output binary image is supposed to be put into the same image buffer
as the source gray image.  This is because images of different depths
do not necessarily start at the same offset relative to the beginning of
the image buffer.

Revision 1.29  1996/01/04 20:27:52  ejaquith
thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.28  1995/12/13  04:14:22  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.27  1995/12/04  16:11:59  macleod
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Fixed a bug in w_thresholdByPixel() where if the image width wasn't a
 * multiple of 32, the right edge pixels could end up in the wrong place
 *
 * Revision 1.26  1995/12/03  00:50:21  macleod
 * macleod on Sat Dec 2 16:49:32 PST 1995
 *
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Added w_thresholdByPixel, which does a thresholding (8bit to 1 bit/pixel)
 * against a threshold map, which is an 8 bit/pixel map of threshold values.
 * If the threshold map is smaller than the source image, the nearest-neighbor
 * thresholds are used.
 *
 * Revision 1.25  1995/11/13  20:17:35  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.24  1995/10/13  22:06:25  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.23  1995/10/05  04:27:37  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Return an error code of cErrNullPtr for null pointer args instead of
 * cErrInvalidParm.  Also return cErrPtrNotNull instead of cErrInvalidParm
 * when a proc is supposed to allocate something and the client handed in
 * a pointer to something that already exists.
 *
 * Revision 1.22  1995/10/03  20:54:40  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Removed obsoleted thresholding/subsample keyword.
 *
 * Revision 1.21  1995/10/03  01:20:34  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 * Sometimes, parmeter order is fixed, especially in in-place ops, so output
 * (or input/output) parameters are last.
 *
 * Revision 1.20  1995/09/21  23:00:48  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.19  1995/09/19  02:55:04  ddavies
 * ddavies on Mon Sep 18 19:53:28 PDT 1995
 *
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Make the threshSense parameter (THRESH_NORM vs THRESH_INVERSE) always work.
 * It had been ignored except in bandpass filtering.  This way, you can always
 * get an inverted output if you want it.  This is useful if you have some
 * program that sets lowThresh and highThresh and doesn't want to blow up
 * whenever lowThresh == 0 or highThresh == maxGray.
 *
 * Revision 1.18  1995/09/18  23:57:53  jcraig
 * jcraig on Mon Sep 18 16:56:21 PDT 1995
 *
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * continuing on with Dan's previous change to w_threshold
 * all pixels that fall in the range [lowThresh..highThresh] will be
 * threshold'd to 1 (unless this is a bandpass and the threshSense is
 * THRESH_INVERSE).
 *
 * Revision 1.17  1995/09/18  16:55:14  ddavies
 * ddavies on Mon Sep 18 09:53:58 PDT 1995
 *
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Get rid of the _POCHANGE's in the comments.  Change the value that signals
 * inv_mid_threshold from [0, midGray] to [0, midGray-1], which is actually
 * correct.
 *
 * Revision 1.16  1995/09/15  18:00:58  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.15  1995/09/11  22:11:36  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.14  1995/09/10  00:33:37  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.13  1995/09/07  01:36:19  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:24:41  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  01:19:32  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:40:11  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/17  22:02:51  ejaquith
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.8  1995/07/29  04:54:25  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.7  1994/10/22  01:28:17  ddavies
 * ddavies on Fri Oct 21 18:27:19 PDT 1994
 *
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Fix bug in call to i_clearAllFrames.  It was using the pointer to a
 * memory buffer that had been realloced.  Even though the buffer got smaller,
 * it was moved and Purify complained.
 *
 * Revision 1.6  1994/10/20  02:26:37  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Modify so that when *pPrD == prS, the dest image is returned in the memory
 * that held the source image.  The source image is destroyed.  The extra
 * memory in the source PIXR is deallocated and the pixr depth and bytes/line
 * are updated.  This works for 8:1, 4:1 and 2:1 thresholds, highpass, lowpass,
 * and bandpass filters.
 *
 * Revision 1.5  1994/10/07  19:14:45  lperry
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.4  1994/09/17  18:19:27  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Add extra NULL table args to calls to i_threshold2To1bpp, i_threshold4To1bpp,
 * i_threshold, i_thresholdI2To1bpp, i_thresholdI4To1bpp, i_thresholdI,
 * i_highpassfiltI4to1bpp, i_highpassfiltI, i_highpassfilt4To1bpp,
 * i_highpassfilt, i_passfil4to1bpp, i_passfilt, i_passfiltI4To1bpp,
 * i_passfiltI.
 *
 * Revision 1.3  1994/09/08  16:19:04  lperry
 * lperry on Thu Sep 8 09:15:32 PDT 1994
 *
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Change to use local variables to hold image size, depth, etc.,
 * instead of using pixrGetXX calls in function arg list. Since this
 * routine made a large number of function calls in a big nested
 * switch statement, this significantly reduced the code size.
 *
 * Revision 1.2  1994/03/24  20:44:49  ddavies
 * thrshold.c -> /product/ipcore/ipwrap/src/RCS/thrshold.c,v
 *
 * Modify iw_threshold so that it will do bandpass filtering of 4 bit/pixel images.
 *
 * Revision 1.1  1994/01/12  23:27:14  lperry
 * Initial revision
 *
*/
