/* $Header: /project/icons/cvsroot/ipcore/src/ipwrap/src/tif_unix.c,v 1.1.1.1 1999/09/20 18:19:21 jlin Exp $ */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994 Sam Leffler
 * Copyright (c) 1991, 1992, 1993, 1994 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
 * TIFF Library UNIX-specific Routines.
 */
#include "tiffiop.h"
#include <unistd.h>
#include <stdlib.h>
#include "ipwcore.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


static tsize_t
_tiffReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return ((tsize_t) read((int) fd, buf, (size_t) size));
}

static tsize_t
_tiffWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return ((tsize_t) write((int) fd, buf, (size_t) size));
}

static toff_t
_tiffSeekProc(thandle_t fd, toff_t off, int whence)
{
	return ((toff_t) lseek((int) fd, (off_t) off, whence));
}

static int
_tiffCloseProc(thandle_t fd)
{
	return (close((int) fd));
}

#include <sys/stat.h>

static toff_t
_tiffSizeProc(TIFF *tif, thandle_t fd)
{
#ifdef _AM29K
	long fsize;
	tif = tif;
	return ((fsize = lseek((int) fd, 0, SEEK_END)) < 0 ? 0 : fsize);
#else
	struct stat sb;
	tif = tif;
	return (toff_t) (fstat((int) fd, &sb) < 0 ? 0 : sb.st_size);
#endif
}

#ifdef MMAP_SUPPORT
#include <sys/mman.h>

static int
_tiffMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
	toff_t size = _tiffSizeProc(fd);
	if (size != (toff_t) -1) {
		*pbase = (tdata_t)
		    mmap(0, size, PROT_READ, MAP_SHARED, (int) fd, 0);
		if (*pbase != (tdata_t) -1) {
			*psize = size;
			return (1);
		}
	}
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
	(void) munmap(base, (off_t) size);
}
#else /* !MMAP_SUPPORT */
static int
_tiffMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
}
#endif /* !MMAP_SUPPORT */

/*
 * Open a TIFF file descriptor for read/writing.
 */
TIFF*
TIFFFdOpen(void *pState, int fd, const char* name, const char* mode)
{
	TIFF* tif;

	tif = TIFFClientOpen(pState, name, mode,
	    (thandle_t) fd,
	    _tiffReadProc, _tiffWriteProc,
	    _tiffSeekProc, _tiffCloseProc, _tiffSizeProc,
	    _tiffMapProc, _tiffUnmapProc);
	if (tif)
		tif->tif_fd = fd;
	return (tif);
}

/*
 * Open a TIFF file for read/writing.
 */
TIFF*
TIFFOpen(void *pState, const char* name, const char* mode)
{
	static const char module[] = "TIFFOpen";
	int m, fd;

	m = _TIFFgetMode(pState, mode, module);
	if (m == -1)
		return ((TIFF*)0);
#ifdef _AM29K
	fd = open(name, m);
#else
	fd = open(name, m, 0666);
#endif
	if (fd < 0) {
		TIFFError(pState, module, "%s: Cannot open", name);
		return ((TIFF *)0);
	}
	return (TIFFFdOpen(pState, fd, name, mode));
}

void *
_TIFFmalloc(TIFF *tif, size_t s)
{
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    return (IPW_MALLOC(s, cAllocatorData));
}

void
_TIFFfree(TIFF *tif, void* p)
{
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    IPW_FREE(p, cAllocatorData);
}

void *
_TIFFrealloc(TIFF *tif, void* p, size_t s)
{
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    return (IPW_REALLOC(p, s, cAllocatorData));
}

void
_TIFFmemset(void* p, int v, size_t c)
{
	memset(p, v, c);
}

void
_TIFFmemcpy(void* d, const void* s, size_t c)
{
	memcpy(d, s, c);
}

int
_TIFFmemcmp(const void* p1, const void* p2, size_t c)
{
	return (memcmp(p1, p2, c));
}

static void
unixWarningHandler(void *pState, const char* module,
		   const char* fmt, va_list ap)
{
    char	errorString[512];
    char	outputString[512];
    IPWState   *pIPWState = (IPWState *)pState;

	/* DDavies, 11/18/96: Only print stuff is there's a printing routine */
    if ((pIPWState != NULL) && (pIPWState->errorProc != NULL))
    {
	    /* print using the IPWrap error handler.  That thing takes
	     * a string, so put the strings together.
	     * First, assemble "Warning, " and the module name.
	     */
	errorString[0] = '\0';		/* null strings so far */
	outputString[0] = '\0';
	if (module != NULL)
	{
	    strcat(&(outputString[0]), "Minor Problem in ");
	    strcat(&(outputString[0]), module);
		/* Add a colon and space after module */
	    strcat(&(outputString[0]), ": ");
	}
	else
	{
	    strcat(&(outputString[0]), "Minor Problem: ");
	}
    
	    /* create formatted error string */
	vsprintf(&(errorString[0]), fmt, ap);

	    /* Combine with warning/module string */
	strcat(&(outputString[0]), &(errorString[0]));
	    /* end with period and CR */
	strcat(&(outputString[0]), ".\n");

	    /* send the error string wherever the client wants */
	pIPWState->errorProc(pIPWState->errorInfo, &(outputString[0]));
    }
}
TIFFErrorHandler _TIFFwarningHandler = unixWarningHandler;

static void
unixErrorHandler(void *pState, const char* module, const char* fmt, va_list ap)
{
    char	errorString[512];
    char	outputString[512];
    IPWState   *pIPWState = (IPWState *)pState;

	/* DDavies, 11/18/96: Only print stuff is there's a printing routine */
    if ((pIPWState != NULL) && (pIPWState->errorProc != NULL))
    {
	    /* print using the IPWrap error handler.  That thing takes
	     * a string, so put the strings together.
	     * First, assemble "Error, " and the module name.
	     */
	errorString[0] = '\0';		/* null strings so far */
	outputString[0] = '\0';
	if (module != NULL)
	{
	    strcat(&(outputString[0]), "Disaster in ");
	    strcat(&(outputString[0]), module);
		/* Add a colon and space after module */
	    strcat(&(outputString[0]), ": ");
	}
	else
	{
	    strcat(&(outputString[0]), "Disaster: ");
	}

	    /* create formatted error string */
	vsprintf(&(errorString[0]), fmt, ap);
    
	    /* Combine with warning/module string */
	strcat(&(outputString[0]), &(errorString[0]));
	    /* end with period and CR */
	strcat(&(outputString[0]), ".\n");

	    /* send the error string wherever the client wants */
	pIPWState->errorProc(pIPWState->errorInfo, &(outputString[0]));
    }
}
TIFFErrorHandler _TIFFerrorHandler = unixErrorHandler;
