/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tiffccit.c: routines that link ipwrap's Modified Huffman encoding and
 *		decoding routines to Leffler's TIFF libraries.
 *
 *                     
 */



#include <string.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "defines.pub"

/* prototypes */
#include "tiffiop.h"
#include "tiffio.h"
#include "props.pub"
#include "mh.pub"
#include "shrcmprs.pub"	/* for cPadToByte, cPadToWord */
#include "ccitt.pub"
#include "tiffccut.h"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tiffccit.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");


/*********************************************************************/
/************************** ENCODING CODE ****************************/
/*********************************************************************/

/*
 * NB: tidata is the type representing *(tidata_t);
 *     if tidata_t is made signed then this type must
 *     be adjusted accordingly.
 */
typedef unsigned char tidata;


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Encode rows of a PIXR into a stream of Modified Huffman data.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *  prS::            pointer to PIXR that holds uncompressed
 *                   data.  This is sort of a cheat in that
 *                   the Leffler convention is that this is a
 *                   pointer to a data buffer.  But, hey, this
 *                   is C, and all pointers are cousins, right?
 *  uncompBytes::    number of uncompressed bytes this routine
 *                   is expected to consume.  We assume this
 *                   is a multiple of the number of bytes in
 *                   an image row.
 *  sampleNum::      number of current pixel sample.
 *                   This is apparently used when different
 *                   samples are stored in different image planes.
 *                   It is not used here.
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
mhEncode (
    TIFF               *tif,
    tidata_t            prS,
    tsize_t             uncompBytes,
    tsample_t           sampleNum)
{
    UInt32	exhaustedImageData;	/* if set on return from
					 * compressRowsMH, we've processed the
					 * correct number of lines.
					 */
    UInt32	exhaustedCompData;	/* if set on return from
					 * compressRowsMH, we've reached the
					 * end of the compressed data buffer
					 */
    UInt32	percentDone;		/* if we ran out of compressed data
					 * before completing the image, this
					 * tells how much of the image
					 * has been compressed.
					 */
    UInt32	bytesWritten;		/* number of compressed bytes read on
					 * each call. */
    UInt32	numLines;		/* number of lines to compress */
    Int32	retCode;
    UInt32	remainingBytesInBuf;	/* number of vacant bytes in the
					 * compressed data buffer. */
    IPWState *pIPWState = NULL;

	/* Initialize the IPWState from the spot where it was saved
	 * by the nice IPWrap tiffio code.
	 */
    pIPWState = (IPWState *)tif->tif_ipwstate;

    sampleNum = sampleNum;	/* keep the compilers happy */

	/* try to compress all the rows.  We figure out the number of lines
	 * to compress by dividing uncompBytes by the TIFF package's
	 * idea of the number of bytes/line.  The information in the
	 * PIXR gives the real bytes/line.
	 */
    remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
    numLines = uncompBytes / TIFFScanlineSize(tif);
    bytesWritten = 0;		/* just in case compressRowsMH doesn't
				 * do anything for some reason */


    retCode = w_compressRowsMH(
	pIPWState,
	(void *)(tif->tif_data), /* private decode data */
	(PIXR *) prS,		/* source pixr */
	TIFFCurrentRow(tif),	/* first dest line of pixr */
	numLines,		/* num lines to compress */
	tif->tif_rawcp,		/* pointer to comp data */
	remainingBytesInBuf,	/* number of comp data
				 * bytes remaining */
	cFalse,			/* this is a new image block */
	cFalse,			/* this is new comp data */
	cTrue,			/* hope to get to the end
				 * of the comp data in this
				 * call */
	&exhaustedImageData,
	&exhaustedCompData,
	&percentDone,
	&bytesWritten);		/* no array of line offsets */

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* Increment the number of bytes now in the buffer and byte
	 * pointer.
	 */
    tif->tif_rawcp += bytesWritten;
    tif->tif_rawcc += bytesWritten;

	/* Now execute a loop that keeps flushing compressed data and
	 * doing more compressing until the proper number of lines has
	 * been processed.
	 */
    while (!exhaustedImageData)
    {
	/* TIFFFlushData1 flushes the buffer of compressed data
	 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
	 * -1 is weird, but that's what the original code did.
	 */
	if (!TIFFFlushData1(tif))
	{
	    return (-1);
	}

	remainingBytesInBuf = tif->tif_rawdatasize;
	bytesWritten = 0;	/* just in case compressRowsMH doesn't
				 * do anything for some reason */

	retCode = w_compressRowsMH(
			    pIPWState,
			    (void *)(tif->tif_data),
			    (PIXR *) prS,
			    TIFFCurrentRow(tif),
			    numLines,
			    tif->tif_rawcp,
			    remainingBytesInBuf,
			    cTrue,		/* this is an old image block */
			    cTrue,		/* this is old comp data */
			    cTrue,		/* hope to get to the end
						 * of the comp data in this
						 * call */
			    &exhaustedImageData,
			    &exhaustedCompData,
			    &percentDone,
			    &bytesWritten);

	if (retCode != cSuccessful)
	{
	    return (0);	/* something failed */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;
    }

 
	/* if it appears to have worked so return happily.  We don`t
	 * check to see if we ran out of image lines since we wouldn`t
	 * have exited the loop above if there were more image lines.
	 * We don't check to see if we've run out of room on the
	 * comp data buffer since there's no reason to think that this
	 * would have happened on compression.
	 */
    return (1);	/* worked */
}


/*********************************************************************/
/************************** DECODING CODE ****************************/
/*********************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Decode rows of a PIXR from a stream of Modified Huffman data.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *  uncompBytes::    number of uncompressed bytes this routine
 *                   is expected to produce.  We assume this
 *                   is a multiple of the number of bytes in
 *                   an image row.
 *  sampleNum::      number of current pixel sample.
 *                   This is apparently used when different
 *                   samples are stored in different image planes.
 *                   It is not used here.
 *
 * Output:::
 *  prD::            pointer to PIXR that will hold decompressed
 *                   data.  This is sort of a cheat in that
 *                   the Leffler convention is that this is a
 *                   pointer to a data buffer.  But, hey, this
 *                   is C, and all pointers are cousins, right?
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 *  Non-standard parameter order!:: 
 *       The output parameter is not the last in the param order
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
mhDecode (
    TIFF               *tif,
    tidata_t            prD,
    tsize_t             uncompBytes,
    tsample_t           sampleNum)
{
    Int32	firstBlock;		/* for testing only */
    UInt32	exhaustedImageData;	/* if set on return from
					 * decompressRowsMH, we've processed the
					 * correct number of lines.
					 */
    UInt32	exhaustedCompData;	/* if set on return from
					 * decompressRowsMH, we've reached the
					 * end of the compressed data
					 */
    UInt32	percentDone;		/* if we ran out of compressed data
					 * before completing the image,
					 * this tells how much of the image
					 * has been decompressed.
					 */
    UInt32	bytesRead;		/* number of compressed bytes read on
					 * each call. */
    Int32	numRowsToRead;		/* number of rows we're to read */
    Bool	continueCompBlock;	/* set if we've already decompressed
					 * some lines of this strip or tile. */
    Int32	retCode;
    IPWState   *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

    sampleNum = sampleNum;	/* keep the compilers happy */

	/* If a pixr is not provided, return with an error.  We
	 * like PIXRs a lot.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* If we're going to read the first line of a strip, then
	 * we'll tell the callee that the compression block does
	 * not continue.
	 */
    if (tif->tif_rawcp == tif->tif_rawdata)
	continueCompBlock = cFalse;
    else
	continueCompBlock = cTrue;

    numRowsToRead = uncompBytes / TIFFScanlineSize(tif);
    if (TIFFIsTiled(tif))	/* can be no bigger than the tile height */
    {
	if (numRowsToRead > tif->tif_dir.td_tilelength)
	    numRowsToRead = tif->tif_dir.td_tilelength;
    }
    else			/* can be no bigger than the strip height */
    {
	if (numRowsToRead > tif->tif_dir.td_rowsperstrip)
	    numRowsToRead = tif->tif_dir.td_rowsperstrip;
    }


#if 0
	/* TESTING ONLY
	 * do this in two parts to see if that works correctly */
    firstBlock = tif->tif_rawcc / 2;

    retCode = w_decompressRowsMH(
			pIPWState,
			(void *)(tif->tif_data), /* private decode data */
			TIFFCurrentRow(tif),	/* first dest line of pixr */
			numRowsToRead,		/* num lines to decompress */
			cFalse,			/* the image was already
						 * cleared, so don't clear
						 * it again. */
			tif->tif_rawcp,		/* pointer to comp data */
			firstBlock,		/* number of comp data bytes */
			cFalse,			/* this is a new image block */
			continueCompBlock,	/* cTrue if is new comp data */
			cFalse,			/* don't expect to get to
						 * the end of the comp data
						 * in this call */
			(PIXR *) prD,		/* dest pixr */
			&exhaustedImageData,
			&exhaustedCompData,
			&percentDone,
			&bytesRead);

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* if it appears to have worked, make sure we did all of
	 * image and ran out of compressed data too.
	 */
    if ((exhaustedImageData) || (!exhaustedCompData) ||
	(bytesRead != firstBlock))
    {
	return (0);	/* failed */
    }
#else
    firstBlock = 0;
#endif
	/* go decompress all the rows.  We figure out the number of lines
	 * to decompress by dividing uncompBytes by the TIFF package's
	 * idea of the number of bytes/line.  The information in the
	 * PIXR gives the real bytes/line.
	 */
    retCode = w_decompressRowsMH(
			pIPWState,
			(void *)(tif->tif_data), /* private decode data */
			TIFFCurrentRow(tif),	/* first dest line of pixr */
			numRowsToRead,		/* num lines to decompress */
			cFalse,			/* the image was already
						 * cleared, so don't clear
						 * it again. */
			tif->tif_rawcp + firstBlock, /* pointer to comp data */
			tif->tif_rawcc - firstBlock, /* number of comp data
						 * bytes remaining */
#if 0
			cTrue,			/* this is an old image block */
			cTrue,			/* this is a continuation of
						 * the previous comp data */
#else
			cFalse,			/* this is a new image block */
			continueCompBlock,	/* cTrue if is new comp data */
#endif
			cTrue,			/* expect to get to the end
						 * of the comp data in this
						 * call */
			(PIXR *) prD,		/* dest pixr */
			&exhaustedImageData,
			&exhaustedCompData,
			&percentDone,
			&bytesRead);

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* Update the input pointers.  This is used if we're not
	 * reading the entire strip or tile in one shot.
	 */
    tif->tif_rawcc -= bytesRead;
    tif->tif_rawcp += bytesRead;

	/* if it appears to have worked, make sure we did all of
	 * image.  We do not check to see if we've exhausted the
	 * compressed data in case we were supposed to stop
	 * before reading the entire strip or tile.
	 */
    if (!exhaustedImageData)
	return (0);	/* failed */
    else
	return (1);	/* worked */
}


/*********************************************************************/
/************************** UTILITY CODE ****************************/
/*********************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Set up a bunch of procedure ptrs for doing compression and decompression.
 *
 * LongDescription:::
 *	This procedure sets up a bunch of procedure pointers so that
 *	Modified Huffman compression and decompression is done.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		1	it worked
 *		0	it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
TIFFInitCCITTRLE (
    TIFF*               tif)
{
    tif->tif_predecode = ccittPreDecode;
    tif->tif_decoderow = mhDecode;
    tif->tif_decodestrip = mhDecode;
    tif->tif_decodetile = mhDecode;
    tif->tif_preencode   = ccittPreEncode;
    tif->tif_encoderow   = mhEncode;
    tif->tif_encodestrip = mhEncode;
    tif->tif_encodetile  = mhEncode;
    tif->tif_postencode  = NULL;	/* we've already flushed the last
					 * of the line, strip or tile to
					 * the compressed data file */
    tif->tif_close       = NULL;	/* No need to write RTC codes either */
    tif->tif_cleanup = ccittCleanup;

	/* tell users that we're using decoding code that deals
	 * directly with PIXRs.  Also tell us that we're
	 * supposed to skip the EOL marker and should flush
	 * data to a byte boundary at the end of each line.
	 */
    tif->tif_options |= FAX3_NOEOL|FAX3_BYTEALIGN;
    tif->tif_flags |= TIFF_PIXRDECODE | TIFF_PIXRENCODE;
    tif->tif_flags &= ~TIFF_NOBITREV;	/* we do not handle bit reversal */
    return (1);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Set up a bunch of procedure ptrs for doing compression and decompression.
 *
 * LongDescription:::
 *	This procedure sets up a bunch of procedure pointers so that
 *	Modified Huffman compression and decompression is done.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		1	it worked
 *		0	it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *  The only difference between this and TIFFInitCCITTRLE is that we
 *  set the flag that says scan lines are padded to word (2-byte) boundaries
 *  instead of byte boundaries.  This is a weird Leffler-ism that's not
 *  in the TIFF spec.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
TIFFInitCCITTRLEW (
    TIFF*               tif)
{
	/* This is almost the same and the normal RLE initialization */
    TIFFInitCCITTRLE(tif);

	/* Turn off the byte-alignment flag and turn on the word-alignment
	 * flag.
	 */
    tif->tif_options &= ~FAX3_BYTEALIGN;
    tif->tif_options |= FAX3_WORDALIGN;
    return (1);
}


/*

 $Log: tiffccit.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.8  1996/05/24 04:45:47  ddavies
 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Don't check to see if we've exhausted the compressed data after reading
 some.  We might only be reading on scan line.

 Revision 1.7  1996/05/23 00:10:36  ddavies
 ddavies on Wed May 22 17:09:40 PDT 1996

 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Bound the number of rows to read by the tile height or strip height.
 This is needed to be able to read FX's wacky TIFF files.

 Revision 1.6  1996/05/16 17:34:30  ddavies
 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Modify mhDecode so that it calculates the number of rows to read once.
 Also fix a bug that prevented us from reading single scan lines from a
 large block of data.  We now set the new continueCompBlock variable based
 on whether we're at the top of a strip or tile (tif_rawcp == tif_rawdata)
 and pass that down to w_decompressRowsMH.  In the old world, we assumed
 we were always at the top of a strip or tile.

 Revision 1.5  1996/05/11 01:31:53  ddavies
 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Quit calculating and passing the padType to w_compressRowsMH and
 w_decompressRowsMH.  It's now passed inside the private data structure.

 Revision 1.4  1996/05/02 00:15:13  ddavies
 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Changed include of tiffccut.pub to tiffccut.h.  File was renamed so that
 coreCheckProtos would be happy.

 Revision 1.3  1996/05/01 23:51:06  ddavies
 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Move (and rename) mhPreEncode, mhPreDecode and mhCleanup to the new file
 tiffccut.c.  These become the utility routines ccittPreEncode, ccittPreDecode
 and ccittCleanup.  They will be used by our MH, MR and MMR code.

 Revision 1.2  1996/04/25 19:13:21  ddavies
 tiffccit.c -> /product/ipcore/ipwrap/src/RCS/tiffccit.c,v

 Add mhPreDecode and mhDecode.  These routines are the interfaces from the
 Leffler TIFF libraries to IPCore routines to do MH decompression.
 Replacing the Leffler routines with IPCore routines results in roughly a
 1.8x performance improvment on the SPARC 5 and 2x-3x performance
 improvement on the PC (486-66).  Also change TIFFInitCCITTRLEW so that
 it calls TIFFInitCCITTRLE and just changes the byte-align flag to the
 word-align flag.  Eliminated the calls to the Leffler initialization in
 these routines since we don't use it anymore.

 Revision 1.1  1996/04/10 23:54:12  ddavies
 Initial revision


 */

