/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tiffccut.c: utility routines that link ipwrap's CCITT encoding and
 *		decoding routines to Leffler's TIFF libraries.
 *
 *                     
 */



#include <string.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "defines.pub"
#include "iaerror.pub"

/* prototypes */
#include "tiffiop.h"
#include "props.pub"
#include "ccitt.pub"
#include "tiffccut.h"
#include "shrcmprs.pub"		/* cNoPad, cEOL1D, cPlainEOL */
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tiffccut.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Prepare the state used to encode Modified Huffman blocks.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
ccittPreEncode (
    TIFF*               tif)
{
    void    *pCCITTCompInfo;        /* pointer to opaque data type */
    UInt32   stateBlockSize;        /* size of opaque data block */
    UInt32   imWidth;               /* width and depth are passed to lower */
    Int32    retCode;
    UInt32   padType;		    /* type of scanline padding */
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

	/* See if the state block already exists.  The tif_data
	 * entry is cleared when the TIFF structure is originally
	 * allocated.
	 */
    pCCITTCompInfo = (void *)tif->tif_data;
    if (pCCITTCompInfo != NULL)
	return (1);		/* nothing else to do */

	/* If we're not going to be given a PIXR to encode,
	 * complain.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* We need a new state block.  Start by determining the
	 * image width */
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);

    retCode = w_getSizeOfCCITTCompressionInfo(pIPWState,
					      imWidth, &stateBlockSize);

    if (retCode != cSuccessful)
    {
	return (0);		/* failed */
    }

	/* figure out the pad type. */
    if (tif->tif_dir.td_compression == COMPRESSION_CCITTRLE)
    {
	if (tif->tif_options & FAX3_WORDALIGN)
	    padType = cPadToWord;
	else
	    padType = cPadToByte;
    }
    else
    if (tif->tif_dir.td_compression == COMPRESSION_CCITTFAX3)
    {
	if (tif->tif_dir.td_group3options & GROUP3OPT_FILLBITS)
	    padType = cPadToByte;
	else
	    padType = cNoPad;
    }
    else
	padType = cNoPad;


	/* Allocate the state block */
    pCCITTCompInfo = (void *)_TIFFmalloc(tif, stateBlockSize);

    if (pCCITTCompInfo == NULL)
    {
	return (0);		/* no memory for this block */
    }

	/* Initalize the state block */
    if ((tif->tif_dir.td_compression == COMPRESSION_CCITTFAX3) &&
	(tif->tif_dir.td_group3options & GROUP3OPT_2DENCODING))
    {
	float res = tif->tif_dir.td_yresolution;
	    /*
	     * The CCITT spec says that when doing 2d encoding, you
	     * should only do it on K consecutive scanlines, where K
	     * depends on the resolution of the image being encoded
	     * (2 for <= 200 lpi, 4 for > 200 lpi).  Since the directory
	     * code initializes td_yresolution to 0, this code will
	     * select a K of 2 unless the YResolution tag is set
	     * appropriately.  (Note also that we fudge a little here
	     * and use 150 lpi to avoid problems with units conversion.)
	     * Note that Leffler's code at this point had the following
	     * completely weird code:
	     * if (tif->tif_dir.td_resolutionunit == RESUNIT_CENTIMETER)
	     *      res = (res * .3937f) / 2.54f;
	     * Note that 0.3937 = 1/2.54, so we're dividing res by
	     * (2.54)^2.  Huh?
	     */
	if (tif->tif_dir.td_resolutionunit == RESUNIT_CENTIMETER)
		    res = res * 2.54f;		/* convert to inches */

	retCode = w_initializeCCITTCompressionInfo(pIPWState,
						   pCCITTCompInfo,
						   imWidth,
						   cTrue,
						   (res > 150 ? 4 : 2),
						   padType,
						   stateBlockSize);
    }
    else	/* this is not Group 3 or is pure Group 3, 1D */
    {
	retCode = w_initializeCCITTCompressionInfo(pIPWState,
						   pCCITTCompInfo,
						   imWidth,
						   cFalse, 0,
						   padType,
						   stateBlockSize);
    }


    if (retCode != cSuccessful)
    {
	_TIFFfree(tif, pCCITTCompInfo);
	return (0);		/* didn't work */
    }

	/* it worked, so put a pointer to the state block into the
	 * tif space reserved for it.
	 */
    tif->tif_data = (tidata_t) pCCITTCompInfo;

    return (1);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Prepare the state used to decode Modified Huffman blocks.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits, MH, MR, MMR
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
ccittPreDecode (
    TIFF               *tif)
{
    void     *pCCITTDecompInfo;      /* pointer to opaque data type */
    UInt32    stateBlockSize;        /* size of opaque data block */
    Int32     retCode;
    UInt32    imWidth;
    UInt32    padType;		    /* type of scanline padding */
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

	/* See if the state block already exists.  The tif_data
	 * entry is cleared when the TIFF structure is originally
	 * allocated.
	 */
    pCCITTDecompInfo = (void *)tif->tif_data;
    if (pCCITTDecompInfo != NULL)
	return (1);		/* nothing else to do */

	/* If we're not going to be given a PIXR to decode into,
	 * just return unsuccessfully.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* We need a new state block.  Start by determining the
	 * width of the thing we'll be decoding.  If it's tiled,
	 * it's the tile width.  Otherwise, it's the image width.
	 */
    if (!TIFFIsTiled(tif))
	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);
    else
	TIFFGetField(tif, TIFFTAG_TILEWIDTH, &imWidth);

	/* We need a new state block.  Start by determining the
	 * size of the state block */
    retCode = w_getSizeOfCCITTDecompressionInfo(pIPWState, imWidth,
						&stateBlockSize);

    if (retCode != cSuccessful)
    {
	return (0);		/* failed */
    }

	/* figure out the pad type. */
    if (tif->tif_dir.td_compression == COMPRESSION_CCITTRLE)
    {
	if (tif->tif_options & FAX3_WORDALIGN)
	    padType = cPadToWord;
	else
	    padType = cPadToByte;
    }
    else
    if (tif->tif_dir.td_compression == COMPRESSION_CCITTFAX3)
    {
	if (tif->tif_dir.td_group3options & GROUP3OPT_FILLBITS)
	    padType = cPadToByte;
	else
	    padType = cNoPad;
    }
    else
	padType = cNoPad;


	/* Allocate the state block */
    pCCITTDecompInfo = (void *)_TIFFmalloc(tif, stateBlockSize);

    if (pCCITTDecompInfo == NULL)
    {
	return (0);		/* no memory for this block */
    }

	/* initialize the state block */
    if ((tif->tif_dir.td_compression == COMPRESSION_CCITTFAX3) &&
	(tif->tif_dir.td_group3options & GROUP3OPT_2DENCODING))
    {
	retCode = w_initializeCCITTDecompressionInfo(pIPWState,
						     pCCITTDecompInfo,
						     imWidth,
						     cTrue,
						     padType,
						     stateBlockSize);
    }
    else	/* this is not Group 3 or is pure Group 3, 1D */
    {
	retCode = w_initializeCCITTDecompressionInfo(pIPWState,
						     pCCITTDecompInfo,
						     imWidth,
						     cFalse,
						     padType,
						     stateBlockSize);
    }

    if (retCode != cSuccessful)
    {
	_TIFFfree(tif, pCCITTDecompInfo);
	return (0);		/* didn't work */
    }

	/* it worked, so put a pointer to the state block into the
	 * tif space reserved for it.
	 */
    tif->tif_data = (tidata_t) pCCITTDecompInfo;


    return (1);			/* it worked */
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Tidy up after Modified Huffman compression or decompression.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		void
 *
 * Notes:::
 *  Non-standard error return codes!::
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
void
ccittCleanup (
    TIFF               *tif)
{
    if (tif->tif_data != NULL)
    {
	_TIFFfree(tif, tif->tif_data);
	tif->tif_data = NULL;
    }
	/* Turn off the flags that say the encoder or decoder knows
	 * about PIXRs.
	 */
    tif->tif_flags &= ~(TIFF_PIXRDECODE | TIFF_PIXRENCODE);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Write EOL codes and flush pipeline after MR or MMR compression
 *  or decompression.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *  numEOLs	     number of EOLs to write.
 *  EOLType	     type of EOLs to write.
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		0 - failed
 *		1 - it worked
 *
 * Notes:::
 *  Non-standard error return codes!::
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
ccittWriteEOLsAndFlush (
    TIFF               *tif,
    UInt32		numEOLs,
    UInt32		EOLType)
{
    UInt32	bytesWritten;		/* number of compressed bytes read on
					 * each call. */
    UInt32	remainingBytesInBuf;	/* number of vacant bytes in the
					 * compressed data buffer. */
	/* Initialize the IPWState from the spot where it was saved
	 * by the nice IPWrap tiffio code.
	 */
    IPWState *pIPWState = (IPWState *)tif->tif_ipwstate;
    Int32	retCode;

	/* try to write out six EOL codes to form one RTC.
	 * We need to detect the case of running out of buffer space
	 * when trying to do this.
	 */
    while (numEOLs > 0)
    {
	remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
	bytesWritten = 0;		/* just in case w_writeCCITTEOL doesn't
					 * do anything for some reason */
	retCode = w_writeCCITTEOL(
	    pIPWState,
	    (void *)(tif->tif_data),	/* private decode data */
	    cNoPad,			/* no pad before the EOL */
	    EOLType,			/* no indicator bit after the EOL */
	    tif->tif_rawcp,		/* pointer to comp data */
	    remainingBytesInBuf,	/* number of comp data
					 * bytes remaining */
	    &bytesWritten);		/* number of bytes written */
    
	if ((retCode != cSuccessful) && (retCode != ia_bufferExceeded))
	{
	    return (0);			/* disaster */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;

	    /* If we failed to write the EOL because we filled
	     * the buffer first, flush it */
	if (retCode == ia_bufferExceeded)
	{
		/* TIFFFlushData1 flushes the buffer of compressed data
		 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
		 * -1 is weird, but that's what the original code did.
		 */
	    if (!TIFFFlushData1(tif))
	    {
		return (-1);
	    }
	}
	else
	    numEOLs--;	/* we succeeded in writing an EOL */
    }

	/* Now try to flush the pipeline to the bytestream.
	 * As usual, we might run out of data.
	 */
    while (cTrue)
    {
	remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
	bytesWritten = 0;		/* just in case w_writeCCITTEOL doesn't
					 * do anything for some reason */
	retCode = w_flushCCITT(
	    pIPWState,
	    (void *)(tif->tif_data),	/* private decode data */
	    tif->tif_rawcp,		/* pointer to comp data */
	    remainingBytesInBuf,	/* number of comp data
					 * bytes remaining */
	    &bytesWritten);		/* number of bytes written */
    
	if ((retCode != cSuccessful) && (retCode != ia_bufferExceeded))
	{
	    return (0);			/* disaster */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;

	    /* If we failed to flush the pipeline because we filled
	     * the buffer first, flush the buffer */
	if (retCode == ia_bufferExceeded)
	{
		/* TIFFFlushData1 flushes the buffer of compressed data
		 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
		 * -1 is weird, but that's what the original code did.
		 */
	    if (!TIFFFlushData1(tif))
	    {
		return (-1);
	    }
	}
	else
	    break;	/* we succeeded in flushing the pipeline */
    }

	/* we're happy, so return */
    return (1);
}



/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Conditionally write RTC after MR or MMR compression or decompression.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		void
 *
 * Notes:::
 *  Non-standard error return codes!::
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, MH
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
void
ccittClose (
    TIFF               *tif)
{
    UInt32	EOLType;		/* type of EOL to write */

    if ((tif->tif_options & FAX3_CLASSF) == 0)		/* append RTC */
    {
	    /* An RTC code is composed of 6 EOL codes.  These may be
	     * followed by indicator bits.  Figure out what type of
	     * EOL to write.
	     */
	if (tif->tif_dir.td_group3options & GROUP3OPT_2DENCODING)
	    EOLType = cEOL1D;
	else
	    EOLType = cPlainEOL;

	    /* Write out six EOL codes to form one RTC.
	     * ccittWriteEOLsAndFlush will flush the compression
	     * buffer if it fills while doing this.
	     * Theoretically, this routine can return errors,
	     * but Leffler's code wants to ignore them so we
	     * will too.
	     */
	(void) ccittWriteEOLsAndFlush(tif, 6, EOLType);
    }
}



/*

 $Log: tiffccut.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.5  1998/01/27 01:57:16  ddavies
 tiffccut.c -> /product/ipcore/ipwrap/src/RCS/tiffccut.c,v

 CommonCheck: Add tif arg to all invocations of _TIFFmalloc, _TIFFfree and _TIFFrealloc.
 These routines now use tif->ipwstate to get the IPState memory allocators
 instead of using malloc and free directly.

 Revision 1.4  1996/05/23 00:11:20  ddavies
 tiffccut.c -> /product/ipcore/ipwrap/src/RCS/tiffccut.c,v

 Set the image with to be the tile width if the image is tiled.  Otherwise,
 set it to be the real image width.

 Revision 1.3  1996/05/11 01:37:18  ddavies
 tiffccut.c -> /product/ipcore/ipwrap/src/RCS/tiffccut.c,v

 Figure out the padType to pass to w_initializeCCITTCompressionInfo and
 w_initializeCCITTDecompressionInfo (and pass it).  The padType is passed
 around in the private state block since it doesn't change inside an image.
 Also figure out the number of lines to be compressed using 2D codes
 (1 or 3, for a K value of 2 or 4) when using the 2D option in Group 3.
 Replace Leffler's totally weird lines/centimeter -> lines/inch conversion
 (was (res * 0.3973)/2.54) with res * 2.54.  0.3973 = 1/2.54.  So what?
 Replace "width" with "imWidth" for readability.
 Calculate the useMR2D flag when initialize the state block for
 Group 3 compresssion and the (different) block for Group 3 decompression.

 Revision 1.2  1996/05/02 00:15:23  ddavies
 tiffccut.c -> /product/ipcore/ipwrap/src/RCS/tiffccut.c,v

 Changed include of tiffccut.pub to tiffccut.h.  File was renamed so that
 coreCheckProtos would be happy.

 Revision 1.1  1996/05/01 23:55:45  ddavies
 Initial revision


 */

