/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tiffio.c: routines for reading TIFF files into PIXRs and for
 *	      writing PIXRs to TIFF files.
 *
 *		       
 *	      Includes procedures for reading:
 *		      w_readTiff()
 *
 *	      and for writing:
 *		      w_writeTiff()
 *		      w_writeTiffHeader()
 *		      w_writeTiffDirectory()
 *		      w_cleanupTiff()
 */


#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#ifdef __WATCOMC__
#include <io.h>
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "imageref.pub"
#include "tiff.h"
#include "tiffio.h"
#include "tiffiop.h"
#include "alplib.h"
#include "frames.pub"
#include "defines.pub"
#include "shrpixr.pub"
#include "shrcc.pub"	/* for TCScanLineRun in iotiff.pub */

/* prototypes */
#include "cut.pub"
#include "pcconv.pub"
#include "props.pub"
#include "io.pub"
#include "iotiff.pub"
#include "frame.pub"
#include "color.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: tiffio.c,v $; $Revision: 1.8 $; $Date: 2002/11/20 22:22:16 $");


#define CENTIMETERS_PER_INCH   2.54F



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read in an TIFF formatted image file into a PIXR from a file name.
 *
 * LongDescription:::
 *   May be single channel, depth 1, 2, 4, or 8 image.
 *   May be 3 channel, depth 8 image.
 *
 * Input:::
 *  pIPWState::      
 *  FileH::          Ptr to file header FH struct.  Only the "filename" field
 *                   needs to be filled
 *  page::	     the page of the multi-page TIFF file to be read.
 *  singleLine::     A testing parameter.  If set to cTrue, CCITT files are
 *		     read using TIFFReadScanline instead of using
 *		     TIFFReadEncodedStrip.  This is just for testing.
 *
 * Output:::
 *  pixr::           Ptr to ptr to returned PIXR with image that was read.
 *                   pixr is allocated and filled in.
 *                   *pixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, compression, Packbits, CCITT, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *  ejaquith::  conversion to use new Leffler TIFF port.
 *              added RGB, resolution stuff.
 *  ddavies::  added tiles, strips stuff.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_readTiff (
    IPWState           *pIPWState,
    FH                 *FileH,
    Int32		page,
    Bool		singleLine,
    PIXR              **pixr)
{
    PROLOGUE;
    UInt32       imHeight, imWidth;  
    UInt16       photometric;
    Int32	 bytesRead, bpl;
    Int32	 depth;
    UInt16       samplesperpixel;
    UInt16       bitspersample;
    Int32        row;
    Int32	 strip, totalStrips;
    Float32      xdpi, ydpi;
    UInt16       resUnit;
    Int32	 tileWidth, tileLength;
    UInt8       *pImage;
    GraySwapNearProc swapLineProc;	/* ptr to proc used to swap bytes
					 * for little-endian processors */
    TIFF	*tif = NULL;		/* TIFF image */
    UInt32      *pic24 = NULL;
    PIXR        *pr = NULL;
    Int32	 retCode;
    ColorMap	*colormap = NULL;

    ERR_GEN(pixr == NULL  ||  FileH == NULL, cErrNullPtr);
    ERR_GEN(*pixr != NULL, cErrPtrNotNull);

    /* --ejaquith added 28sep95--
     * --DDavies moved 11/18/96--
     * We need this to pass our little IPWState ptr around THROUGH the Leffler
     * code.  This happens when you are Packbits or CCITT compressing
     * because we (eventually) call IPWrap routines (outside of Leffler)
     * which need our IPWState ptr.  We also use the IPWState to
     * print error messages.
     * So the solution is to set this up in the TIFFOpen() call!
     */
    tif = TIFFOpen((void *)pIPWState, FileH->filename, "r");
    ERR_GEN_CLEANUP(tif == NULL, cErrIO_ReadError);

    tif->tif_ipwstate = (void *)pIPWState;

    /* Seek down to the right page.  Since you can't know the number
     * of pages, don't return an error if you ask for a page that isn't
     * there.
     */

    while (page > 1)
    {
	int result;
        result = TIFFReadDirectory(tif);
	if (result == 0){
	    /* return normally, but there's no image here. */
		if(tif != NULL) TIFFClose(tif);
                tif = NULL;
	    	NORMAL_RETURN;
        }
	page--;
    }


    TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
    TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);

    ERR_GEN_CLEANUP(!((samplesperpixel == 1  &&  (bitspersample == 1 ||
						  bitspersample == 2 ||
						  bitspersample == 4 ||
						  bitspersample == 8)) ||
		      (samplesperpixel == 3  &&   bitspersample == 8)),
		    cErrTypeNotSupported);

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &imHeight);
    TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric);

    /* Get the dpi resolution in x and y */
    resUnit = RESUNIT_NONE;
    xdpi = ydpi = -1.0F;
    TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xdpi);
    TIFFGetField(tif, TIFFTAG_YRESOLUTION, &ydpi);
    TIFFGetField(tif, TIFFTAG_RESOLUTIONUNIT, &resUnit);
    switch (resUnit)
    {
    case RESUNIT_CENTIMETER:
	xdpi *= CENTIMETERS_PER_INCH;
	ydpi *= CENTIMETERS_PER_INCH;
    case RESUNIT_INCH:
	break;
    case RESUNIT_NONE:
    default:
	/* The resolution unit is nonstandard.  Therefore we cannot convert
	 * to dpi, so don't set the pixr resolution. */
	break;
    }
    if (xdpi < 0. || ydpi < 0.)
	/* Resolution wasn't set above -- make it 0 in the pixr */
	xdpi = ydpi = 0.0F;
    

    if (samplesperpixel == 3)
    {
	/* Special case for full color 24-bit RGB images */
	PIXR         *subR, *subG, *subB;
	UInt32       *ptrPic24, packPixel;
	UInt8        *pRedIm, *pGreenIm, *pBlueIm;
	Int32         x;
	
	ERR_GEN_CLEANUP(photometric != PHOTOMETRIC_RGB || bitspersample != 8,
			cErrTypeNotSupported);
	
	/* Now this is really weird.  Looking at other calls of other programs
	 * (I'm looking at source code for "xv") to earlier versions of the
	 * TIFF library, they seem to work just fine with the
	 * returned packed Int32 image below.  However, our images seem to be
	 * flipped vertically, so we divide it into the color planes keeping
	 * this in mind below.
	 */
	
	ERR_PASS_CLEANUP( w_createPixrExtendedNoInit(
						pIPWState,
						imWidth,
						imHeight,
						bitspersample,  /*depth*/
						Float_TO_ScaledInt1616(xdpi),
						Float_TO_ScaledInt1616(ydpi),
						cFrameBits,     /*frame*/
						samplesperpixel,/*channels*/
						cColorRGB,      /*type*/
						NULL,           /*no colormap*/
						&pr ));

	/* --ejaquith 28sep95--
	 * moved PIXR creation to be the first thing done to help keep
	 * risks of memory fragmentation lower (because the pic24 is something
	 * local to this routine, but the PIXR is persistent!)
	 */
	
	pic24 = (UInt32 *)IPW_CALLOC(imWidth * imHeight, sizeof(UInt32),
				 cAllocatorImage);
	ERR_GEN_CLEANUP(pic24 == NULL, cErrNomem);

	/* This routine returns a raster bitmap that is the original image
	 * width x height, in units of UInt32.  The 8-bit R,G,and B components
	 * have been packed into the 32 bit word for each pixel.
	 * The ordering is:
	 *  	xxxxxxxxBBBBBBBBGGGGGGGGRRRRRRRR
	 * which we must unpack when we split this into 8-bit color planes.
	 * If this routine is encoded using a compression scheme that
	 * knows about PIXRs, tell that routine that TIFFReadRGBAImage will
	 * NOT supply a PIXR.  That routine is too weird for words.
	 * It fusses at great length with orientation (ends up mirroring
	 * images top to bottom?), so I'm afraid to gut it.
	 */
	tif->tif_flags &= ~TIFF_PIXRPROVIDED;

	ERR_GEN_CLEANUP(!TIFFReadRGBAImage(tif, imWidth, imHeight, pic24, 0),
			cErrIO_ReadError);

	/* Now take the 24 bit per pixel image we read and convert it to
	 * 3 8-bit image planes */
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pr, 
					   cRedPlane , &subR));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pr, 
					   cGreenPlane , &subG));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pr, 
					   cBluePlane , &subB));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subR , &pRedIm));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subG , &pGreenIm));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subB , &pBlueIm));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pr , &bpl));

	/* Go to the last rows of the image planes */
	pRedIm   += bpl*(imHeight-1);
	pGreenIm += bpl*(imHeight-1);
	pBlueIm  += bpl*(imHeight-1);
	
	ptrPic24 = pic24;
	for (row = 0; row < imHeight; row++)
	{
	    for (x = 0; x < imWidth; x++)
	    {
		packPixel = *ptrPic24;
		ptrPic24++;
		UCHAR_ACCESS(&pRedIm[x])   = (UInt8) (0xff &  packPixel);
		UCHAR_ACCESS(&pGreenIm[x]) = (UInt8) (0xff & (packPixel >> 8));
		UCHAR_ACCESS(&pBlueIm[x])  = (UInt8) (0xff & (packPixel >> 16));
	    }

	    /* Go to previous row of image planes */
	    pRedIm   -= bpl;
	    pGreenIm -= bpl;
	    pBlueIm  -= bpl;
	}
	
	IPW_FREE(pic24, cAllocatorImage);
    }
    else
    {

	/* Get the address of the proc used to swap all the bytes,
	 * We do not invert the image, this is a PixrToPixr byteswapping
	 * (no FAR pointers) and the image is not color.
	 * This routine is only needed if the decode routine doesn`t
	 * deal with PIXRs directly.
	 */
	if (!decodeUsesPixrs(tif))
	    ERR_PASS_CLEANUP( w_getDIBPixrLineProc (pIPWState, 
						    cFalse, cPixrToPixr,
						    cFalse, NULL, NULL,
						    &swapLineProc, NULL));

	/* Now we have to think about strips vs tiles.  The original
	 * code only read strips, but the DAE is giving us (weird) tiled
	 * images.  If the image is in strips, execute the strip code.
	 */
	if (!TIFFIsTiled(tif))
	{
	    /* In the old code, we read each scan line individually and
	     * moved each one into place in the image.  In the new code,
	     * we read all the strips into the image, then swap all
	     * the bytes read if this a little endian processor.  Finally,
	     * we add the frame.  We do things in this order because
	     * the frame adding procedure expects the bytes to already
	     * in the correct order.
	     * HOWEVER, if we're using the new decoding routines that
	     * just put the bits into a PIXR directly, just pass in the
	     * PIXR on each call and forget about the byteswaps and
	     * frame-adding.
	     */
	    ERR_PASS_CLEANUP( w_createPixr( pIPWState, imWidth, imHeight,
					    bitspersample,
					    Float_TO_ScaledInt1616(xdpi),
					    Float_TO_ScaledInt1616(ydpi),
					    cFrameBits,
					    &pr ));

	    totalStrips = TIFFNumberOfStrips(tif);
	    if (decodeUsesPixrs(tif))
	    {
		/* Read in all the strips in the image. 
		 * Put them in contiguous,
		 * framed rows.  The current row number is calculated from
		 * the strip number.  The number of row read is calculated
		 * from the height of the strip.  The last parameter (-1)
		 * means that the TIFF package should maintain the buffer
		 * used to hold compressed data.  Note that we pass a
		 * pointer to the PIXR in the parameter normally used to
		 * point to the uncompressed data buffer.  All's fair
		 * in C, right?
		 * But first, tell that routine that it's getting a
		 * pointer to a PIXR instead of a pointer to a buffer.
		 * ALSO, if this is a weird test case, read the image
		 * one scan line at a time.  We only test the PIXR version
		 * of the code since that's the only stuff we wrote.
		 */
		tif->tif_flags |= TIFF_PIXRPROVIDED;

		if (!singleLine)
		{
		    for (strip = 0; strip < totalStrips; strip++)
		    {
			bytesRead = TIFFReadEncodedStrip(tif,
							 strip,
							 (tdata_t) pr,
							 -1);
			ERR_GEN_CLEANUP(bytesRead < 0, cErrIO_ReadError);
		    }
		}
		else	/* the weird test case */
		{
		    for (row = 0; row < imHeight; row++)
		    {
			retCode = TIFFReadScanline(tif, pr, row, 0);
    
			ERR_GEN_CLEANUP(retCode == -1, cErrIO_ReadError);
		    }
		}
	    }
	    else	/* the decoder doesn't know about PIXRs, so
			 * we have to mess around with byteswapping
			 * and adding frames.
			 */
	    {
		UInt8	*bufOut;

		    /* Tell the lower level routines that they're not
		     * getting PIXRs.
		     */
		tif->tif_flags &= ~TIFF_PIXRPROVIDED;

		/* We needed to use cFrameBits above to get enough memory to
		 * later add the frame in place.  However, this pixr is/will be
		 * frameless, so reset the frame in the struct.  (Not to 
		 * mention, later call to w_addFrameIP() requires a zero
		 * frame!)
		 */
		ERR_PASS_CLEANUP( w_pixrSetFrame( pIPWState, pr, 0));

		/* set the byte/line to be correct for a frameless image */
		ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, pr,
					    TIFFScanlineSize(tif)));
		ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pr, &bufOut));
	
		/* Read in all the strips in the image.  Put them in 
		 * contiguous, frameless rows.
		 */
		for (strip = 0; strip < totalStrips; strip++)
		{
		    bytesRead = TIFFReadEncodedStrip(tif, strip,
						     bufOut, -1);

		    ERR_GEN_CLEANUP(bytesRead < 0, cErrIO_ReadError);

		    bufOut += bytesRead;
		}
		/* At this point, we should have a frameless PIXR.  It might
		 * be byteswapped.  We have to fix the byte order before
		 * adding the frame because addFrameInPlace expects the
		 * Pixr byte order.
		 */
		if (swapLineProc != NULL)
		{
		    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pr,
						     &pImage));
		    (*swapLineProc)(pIPWState,
				    (UInt32 *)pImage, /* source */
				    (UInt32 *)pImage, /* dest */
				    (bufOut + 3 - pImage)/4);
		}
		
		ERR_PASS_CLEANUP( w_addFrameIP(pIPWState, pr));
	    }
	}
	else
	{	/* the image is tiled.  If it's one big tile (the weird
		 * DAE format), allocate a PIXR big enough for the tile
		 * plus frames, read in the tile, add the frame and
		 * cut off the padding.
		 */
	    TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tileWidth);
	    TIFFGetField(tif, TIFFTAG_TILELENGTH, &tileLength);
	    if (TIFFNumberOfTiles(tif) == 1)
	    {
		/* It's one big tile */
		/* WORKAROUND FOR BROKEN FX FILES!!  They didn't
		 * pad the image to the tile length.
		 */
		tif->tif_dir.td_tilelength = tileLength = imHeight;
		tif->tif_tilesize = TIFFTileRowSize(tif) * tileLength;

		ERR_PASS_CLEANUP( w_createPixr(pIPWState,
					       tileWidth, tileLength,
					       bitspersample,
					       Float_TO_ScaledInt1616(xdpi),
					       Float_TO_ScaledInt1616(ydpi),
					       cFrameBits,
					       &pr));

		if (decodeUsesPixrs(tif))
		{
			/* But first, tell that routine that it's getting a
			 * pointer to a PIXR instead of a pointer to a buffer.
			 */
		    tif->tif_flags |= TIFF_PIXRPROVIDED;

			/* Read the tile.  It's at x=0, y=0, z=0, sample=0 */
		    bytesRead = TIFFReadTile(tif, (tdata_t)pr,
					     0, 0, 0, 0);
		    ERR_GEN_CLEANUP(bytesRead < 0, cErrIO_ReadError);
		}
		else	/* the decoder doesn't know about tiles, so we
			 * have to do the byteswapping and frame-adding
			 * number.
			 */
		{
			/* Tell the lower level routines that they're not
			 * getting PIXRs.
			 */
		    tif->tif_flags &= ~TIFF_PIXRPROVIDED;

		    /* set the byte/line to be correct for a frameless image */
		    ERR_PASS_CLEANUP( w_pixrSetBpl(pIPWState, pr,
						TIFFTileRowSize(tif)));

		    /* Tell the image that it has no frame */
		    ERR_PASS_CLEANUP( w_pixrSetFrame(pIPWState, pr, 0));
    
		    /* Read the tile.  It's at x=0, y=0, z=0, sample=0 */
		    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pr,
						     &pImage));
		    bytesRead = TIFFReadTile(tif, pImage, 0, 0, 0, 0);
		    ERR_GEN_CLEANUP(bytesRead < 0, cErrIO_ReadError);
    
		    /* At this point, we should have a frameless PIXR. It might
		     * be byteswapped.  We have to fix the byte order before
		     * adding the frame because addFrameInPlace expects the
		     * Pixr byte order.
		     */
		    if (swapLineProc != NULL)
		    {
			ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pr,
							 &pImage));
			(*swapLineProc)(pIPWState,
					(UInt32 *)pImage, /* source */
					(UInt32 *)pImage, /* dest */
					(bytesRead + 3)/4);
		    }

		    /* add the frame to the image.  This gives the image a
		     * legal number of bytes/line.
		     */
		    ERR_PASS_CLEANUP( w_addFrameIP(pIPWState, pr));
		}

		/* chop off the image pad on the right and bottom */
		ERR_PASS_CLEANUP( w_cutRectIP (pIPWState,
					       0, 0, imWidth, imHeight,
					       pr));
	    }
	    else
		/* read an image with more than 1 tile */
		ERR_GEN_CLEANUP(1, cErrIO_NTilesNotSupported);
	}


	/*---------------------------------------------------------------
	 * Ensure the pixr has the expected photometric interpretation:
	 *   for binary images, 0 is white and 1 is black
	 *   for 8-bit images,  0 is black and 255 is white.
	 *-------------------------------------------------------------*/
	
	ERR_PASS( w_pixrGetDepth(pIPWState, pr , &depth));
	if (depth == 1)
	{
	    if ( photometric == PHOTOMETRIC_MINISBLACK )
	    {
		UInt16 compression;
		
		/* If CCITT compression scheme, then image has _already_ been
		 * inverted!
		 */
		TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression);
		if (compression != COMPRESSION_CCITTFAX3 &&
		    compression != COMPRESSION_CCITTFAX4 &&
		    compression != COMPRESSION_CCITTRLE  &&
		    compression != COMPRESSION_CCITTRLEW &&
		    compression != COMPRESSION_LOSSYDITHER)
		    ERR_PASS_CLEANUP( w_invert(pIPWState, pr, &pr));
	    }
	}
	else if ( photometric == PHOTOMETRIC_MINISWHITE ) 
	    ERR_PASS_CLEANUP( w_invert(pIPWState, pr, &pr));
	else if ( photometric == PHOTOMETRIC_PALETTE ) 
	{
	    UInt32 colors;

	    colors = 1 << depth;

	    ERR_PASS_CLEANUP( w_createColorMap(pIPWState, colors, &colormap));

	    /* Convert colormap from the format of TIFF */
	    {
            UInt16 *red = NULL, *blue = NULL, *green = NULL;
		    UInt8 *color;
		    UInt32 index;
		    ColorRecord *cRec = colormap->map;

            /* By pass TIFFGetField here, it seems to give wrong result*/
            /* for nodebug version.  --- Why???*/
            /*TIFFGetField(tif, TIFFTAG_COLORMAP, &red, &green, &blue)*/;
            red = tif->tif_dir.td_colormap[0];
            green = tif->tif_dir.td_colormap[1];
            blue = tif->tif_dir.td_colormap[2];
		        for (index = 0; index < colors; index++) {
		            color = (UInt8 *)cRec++;
		            color[cRedPlane] = *red++ >> 8;
		            color[cGreenPlane] = *green++ >> 8;
		            color[cBluePlane] = *blue++ >> 8;
		        }
	    }

	    /* Add it to the PIXR */
	    ERR_PASS_CLEANUP( w_pixrSetColorMap(pIPWState, pr, colormap));
	}
    }

    
    TIFFClose(tif); tif = NULL;
    
    *pixr = pr;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (tif != NULL)
	TIFFClose(tif);
    w_destroyPixr(pIPWState, &pr);
    IPW_FREE(pic24, cAllocatorImage);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to a TIFF formatted image file from a file name.
 *
 * Input:::
 *  pIPWState::      
 *  pixr::           Ptr to PIXR with standard padding to write out.
 *                   May be single channel, depth 1, 2, 4, or 8 image.
 *                   May be 3 channel, depth 8 image.
 *  fname::          Ptr to string of actual filename.  (No way to use
 *                   stdin for reading in TIFF files!)
 *  compression::    One of these:<BR>
 *                   COMPRESSION_NONE	 none<BR>
 *                   COMPRESSION_LZW        Lempel-Ziv & Welch<BR>
 *                   COMPRESSION_CCITTFAX3  CCITT Group 3 fax encoding<BR>
 *                   COMPRESSION_CCITTFAX4  CCITT Group 4 fax encoding<BR>
 *                   COMPRESSION_PACKBITS   Macintosh Run Length Encoding<BR>
 *                   COMPRESSION_CCITTRLE   Hodified Huffmann Run Length Encoding<BR>
 *                   COMPRESSION_LOSSYDITHER   Lossy Dither encoding
 *  singleStrip::    cTrue => write out entire image in one strip
 *  options::	     holds option flags used for CCITT Group 3 fax encoding.
 *
 * Returns:::
 *
 * Notes:::
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *   [There happens to be no persistent memory allocated in this routine].
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, compression, Packbits, CCITT, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *  ejaquith::  conversion to use new Leffler TIFF port.
 *              added RGB, resolution stuff.
 *  ddavies::  added strips stuff.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeTiff (
    IPWState           *pIPWState,
    PIXR               *pixr,
    const char         *fname,
    const UInt16        compression,
    Bool                singleStrip,
    const UInt16	options)
{
    PROLOGUE;
    TIFF         *tif = NULL;
    ERR_PASS_CLEANUP(w_writeTiffHeader ( pIPWState, fname, &tif));
    ERR_PASS_CLEANUP(w_writeTiffDirectory(pIPWState, 
                                  pixr, 
                                  compression, 
                                  singleStrip, 
                                  options, 
                                  tif));
    (void) TIFFCloseFile(tif);
    _TIFFfree(tif, (char *)tif);

    NORMAL_RETURN;

ERR_CLEANUP:
    w_cleanupTiff(tif);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a TIFF formatted image file header to a file.
 *  Obtain a tif pointer.
 *
 * Input:::
 *  pIPWState::      
 *  fname::          Ptr to string of actual filename.  (No way to use
 *                   stdin for reading in TIFF files!)
 * Returns:::
 *  *tif             Ptr to a TIFF structure
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, compression, Packbits, CCITT, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 * Jing Lin
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeTiffHeader (
    IPWState           *pIPWState,
    const char         *fname,
    TIFF		**tif)
{
    PROLOGUE;

    ERR_GEN(fname == NULL, cErrNullPtr);

    /* --ejaquith added 28sep95--
     * --DDavies moved 11/18/96--
     * We need this to pass our little IPWState ptr around THROUGH the Leffler
     * code.  This happens when you are Packbits or CCITT compressing
     * because we (eventually) call IPWrap routines (outside of Leffler)
     * which need our IPWState ptr.  We also use the IPWState to
     * print error messages.
     * So the solution is to set this up in the TIFFOpen() call!
     */
    *tif = TIFFOpen((void *)pIPWState, fname, "w");
    ERR_GEN(tif == NULL, cErrIO_WriteError);
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to one page  TIFF formatted image file from a tif structure.
 *
 * Input:::
 *  pIPWState::      
 *  pixr::           Ptr to PIXR with standard padding to write out.
 *                   May be single channel, depth 1, 2, 4, or 8 image.
 *                   May be 3 channel, depth 8 image.
 *  compression::    One of these:<BR>
 *	COMPRESSION_NONE	 none<BR>
 *	COMPRESSION_LZW		 Lempel-Ziv & Welch<BR>
 *	COMPRESSION_CCITTFAX3	 CCITT Group 3 fax encoding<BR>
 *	COMPRESSION_CCITTFAX4	 CCITT Group 4 fax encoding<BR>
 *	COMPRESSION_PACKBITS	 Macintosh Run Length Encoding<BR>
 *	COMPRESSION_CCITTRLE	 Hodified Huffmann Run Length Encoding<BR>
 *	COMPRESSION_LOSSYDITHER	 Lossy Dither encoding
 *  singleStrip::    cTrue => write out entire image in one strip
 *  options::	     holds option flags used for CCITT Group 3 fax encoding.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, compression, Packbits, CCITT, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 **************************************************************************/
ErrCode CDECL
w_writeTiffDirectory(
    IPWState           *pIPWState,
    PIXR               *pixr,
    const UInt16        compression,
    Bool                singleStrip,
    const UInt16	options,
    TIFF		*tif)
{
    PROLOGUE;
    Int32         row, bpl, tiffBytes;
    Int32         rowsPerStrip, bytesPerStrip;
    Int32         stripNum, depth, channels;
    ScaledInt1616 xdpi, ydpi;
    Int32         imHeight, imWidth;
    ColorType     type;
    UInt8        *buf = NULL;
    ColorMap	 *colormap = NULL;

    ERR_GEN(tif == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, pixr, &imHeight));
    ERR_PASS_CLEANUP( w_pixrGetWidth (pIPWState, pixr, &imWidth));
    ERR_PASS_CLEANUP( w_pixrGetBpl   (pIPWState, pixr, &bpl));
    ERR_PASS_CLEANUP( w_pixrGetDepth (pIPWState, pixr, &depth));
    ERR_PASS_CLEANUP( w_pixrGetType  (pIPWState,  pixr, &type));
    ERR_PASS_CLEANUP( w_pixrGetChannelCount(pIPWState, pixr, &channels));
    ERR_PASS_CLEANUP( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));
    ERR_PASS_CLEANUP( w_pixrGetColorMap(pIPWState, pixr, &colormap));

    ERR_GEN_CLEANUP(!((channels == 1  &&  (depth == 1 || 
					   depth == 2 || 
					   depth == 4 ||
					   depth == 8)) ||
		      (channels == 3  &&  depth == 8)), cErrTypeNotSupported);
    ERR_GEN_CLEANUP(colormap != NULL && (depth != 4 && depth != 8) &&
		    channels != 1, cErrTypeNotSupported);
    
    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, imWidth );
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH, imHeight );
    TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, (UInt16)depth);
    TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (UInt16)channels);
    
    if (ScaledInt1616_TO_Int32(xdpi) > 0 && 
	ScaledInt1616_TO_Int32(ydpi) > 0)
    {
	/* dpi has been set.  Note that we have to pass the ADDRESS
	 * of the floating pt value of the resolution to TIFFSetField
	 * instead of the float itself.  That's because all floats
	 * passed as variable args are converted to doubles and
	 * acc tries to read them as double without bothering to
	 * align them on 64-bit boundaries first, so gets a Bus Error.
	 * We changed TIFFSetField so that when setting the tags
	 * that expect a float, they expect the address of a float
	 * instead.
	 */
	float xdpiFloat, ydpiFloat;
	xdpiFloat = (float)(ScaledInt1616_TO_Int32(xdpi));
	ydpiFloat = (float)(ScaledInt1616_TO_Int32(ydpi));
	TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
	TIFFSetField(tif, TIFFTAG_XRESOLUTION, &xdpiFloat);
	TIFFSetField(tif, TIFFTAG_YRESOLUTION, &ydpiFloat);
    }
    
    if (type == cColorRGB)
    {
	/* Special case for full 24-bit color images */
	UInt8  *pRedIm, *pGreenIm, *pBlueIm;
	PIXR   *subR, *subG, *subB;
	Int32   byteDiff, x;
	
	ERR_GEN_CLEANUP(channels != 3 || depth != 8, cErrTypeNotSupported);
	
	TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);

	/* Set up the buffer to hold 1 row of the image at a time, that is,
	 * 3 bytes of R,G,B data per pixel in the row */
	tiffBytes = imWidth * 3;

	/* check for TIFF Scanline size discrepancy */
	ERR_GEN_CLEANUP(TIFFScanlineSize(tif) != tiffBytes, cErrInternal);

	buf = (UInt8 *)IPW_CALLOC(1, tiffBytes, cAllocatorData);
	ERR_GEN_CLEANUP(buf == NULL, cErrNomem);

	rowsPerStrip = (singleStrip ? imHeight : MAX(8196 / tiffBytes, 1));
	TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);

	/* Take the 3 8-bit planes and pack them together line by line,
	 * with byte values for a pixel going left to right R,G,B */
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, 
					   cRedPlane , &subR));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr,
					   cGreenPlane , &subG));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, 
					   cBluePlane , &subB));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subR, &pRedIm));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subG, &pGreenIm));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subB, &pBlueIm));
	
	byteDiff = bpl - imWidth;
	for (row = 0; row < imHeight; row++)
	{ 
	    /* Fill in contiguous byte buffer */
	    for (x = 0; x < tiffBytes; )
	    {
		buf[x++] = UCHAR_ACCESS(pRedIm);
		pRedIm++;
		buf[x++] = UCHAR_ACCESS(pGreenIm);
		pGreenIm++;
		buf[x++] = UCHAR_ACCESS(pBlueIm);
		pBlueIm++;
	    }
	    
	    ERR_GEN_CLEANUP(TIFFWriteScanline(tif, buf, row, 0) < 0,
			    cErrIO_WriteError);

	    /* Move to next rows */
	    pRedIm   += byteDiff;
	    pGreenIm += byteDiff;
	    pBlueIm  += byteDiff;
	}

	IPW_FREE(buf, cAllocatorData);
    }
    else	/* this is a binary or gray image */
    {
        if (colormap != NULL)
	{
	    TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
	    /* Convert colormap to the format of TIFF */
	    {
		UInt16 cmap[768], *red, *green, *blue;
		UInt8 *color;
		UInt32 index, colors;
		ColorRecord *cRec = colormap->map;

		colors = 1 << depth;
		red = cmap; green = red + colors; blue = green + colors;
		for (index = 0; index < colors; index++) {
		    color = (UInt8 *)cRec++;
		    *red++ = (color[cRedPlane] << 8) | color[cRedPlane];
		    *green++ = (color[cGreenPlane] << 8) | color[cGreenPlane];
		    *blue++ = (color[cBluePlane] << 8) | color[cBluePlane];
		}
		TIFFSetField(tif, TIFFTAG_COLORMAP, cmap, cmap + colors,
			     cmap + colors * 2);
	    }
	}
	else
	{
	    if ( depth == 1 )
		TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,
			     PHOTOMETRIC_MINISWHITE);
	    else
		TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,
			     PHOTOMETRIC_MINISBLACK);
	}

	switch (compression)
	{
	case COMPRESSION_NONE:
	case COMPRESSION_LZW:
	case COMPRESSION_CCITTFAX4:
	case COMPRESSION_PACKBITS:
	case COMPRESSION_CCITTRLE:
	case COMPRESSION_LOSSYDITHER:
	    TIFFSetField(tif, TIFFTAG_COMPRESSION, compression);
	    break;
	case COMPRESSION_CCITTFAX3:
	    TIFFSetField(tif, TIFFTAG_COMPRESSION, compression);
	    TIFFSetField(tif, TIFFTAG_GROUP3OPTIONS, options);
	    break;
	default:
	    ERR_GEN_CLEANUP(1, cErrCompressionNotSupported);
	}

	    /* set the number of rows per strip.  We use this number
	     * to decide how many lines are encoded in each strip,
	     * so we can tell when to stop asking for lines to
	     * be encoded.
	     */
	rowsPerStrip = (singleStrip ?
		imHeight : MAX(8196 / TIFFScanlineSize(tif), 1));
	TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);
	
	    /* if the encoder knows about PIXRs, we don't have to bother
	     * byteswapping the image and writing out one line at a
	     * time (to strip the frame).  Otherwise, we have to do
	     * all that tiresome stuff.
	     * It would be lovely to just index through the strips
	     * when encoding strips.  However, the maximum strip
	     * number isn't calculated until TIFFWriteEncodedStrip
	     * is called the first time.  This number is initialized
	     * to 0.  We are supposed to stop with the current
	     * strip number is >= the maximum strip number.  Since
	     * any (unsigned) number is >= 0, we'd never get through
	     * the first pass of the loop without some thought.
	     * Therefore, we always do the first pass, and check
	     * only after each call to see if there's another strip.
	     */
	if (encodeUsesPixrs(tif))
	{
		/* Tell lower level routines that they're getting
		 * PIXRs.
		 */
	    tif->tif_flags |= TIFF_PIXRPROVIDED;

	    bytesPerStrip = rowsPerStrip * TIFFScanlineSize(tif);
	    stripNum = 0;
	    tif->tif_row = 0;

	    while (cTrue)
	    {
		    /* For reasons which escape me ENTIRELY, the
		     * TIFF package does not set tif->tif_row from
		     * the strip number when writing strips.  Sets up
		     * tif_row when writing scan lines or tiles.  It sets up
		     * tif_row when reading strips, scanlines or tiles.
		     * Gag.
		     * So anyway, we add rowsPerStrip to tif_row after
		     * each write.
		     * Ooops, and before doing the write, check to
		     * see if this is the last strip.  The TIFF code
		     * checks on strips reads, but not on strip writes.
		     * If this is the last strip, reduce the bytecount
		     * so it includes only the correct number of lines.
		     */
		if (stripNum == tif->tif_dir.td_nstrips - 1 ||
		    imHeight <= rowsPerStrip)
		{
		    bytesPerStrip = (imHeight % rowsPerStrip);
		    if (bytesPerStrip == 0)
			bytesPerStrip = rowsPerStrip * TIFFScanlineSize(tif);
		    else
			bytesPerStrip *= TIFFScanlineSize(tif);
		}

		ERR_GEN_CLEANUP(TIFFWriteEncodedStrip(
		    tif,
		    stripNum,	/* # of current strip */
		    (tdata_t)pixr, /* ptr to uncompressed
				    * data (sort of) */
		    bytesPerStrip) < 0, cErrIO_WriteError);

		tif->tif_row += rowsPerStrip;
		stripNum++;

		    /* quit if we just did the last strip */
		if (stripNum >= tif->tif_dir.td_nstrips)
		    break;
	    }
	}
	else	/* too bad, the encoder doesn't know about PIXRs, so
		 * we have to fiddle around with each line.
		 */
	{
	    UInt8	*pIm;
	    GraySwapNearProc swapLineProc;
	    UInt32	 wordsPerLine;

		/* Tell lower level routines that they're NOT getting
		 * PIXRs.
		 */
	    tif->tif_flags &= ~TIFF_PIXRPROVIDED;

		/* Get a pointer to the proc used to swap the bytes
		 * in each scan line as we copy it into a buffer
		 * on the little-endian platforms.  If this is
		 * a big endian platform, this pointer will be
		 * NULL.
		 */
	    ERR_PASS_CLEANUP( w_getDIBPixrLineProc (
		pIPWState,
		cFalse,		/* don't invert image */
		cPixrToPixr,	/* just swap bytes */
		cFalse,		/* is not a color image */
		NULL,		/* dest is not __far ptr */
		NULL,		/* src is not __far ptr */
		&swapLineProc,	/* both source and dest are near */
		NULL));		/* not an interleaved color image */

	    ERR_PASS( w_pixrGetImage(pIPWState, pixr, &pIm));
	    if (swapLineProc == NULL)
	    {
		    /* we're on a big-endian platform, so just
		     * write the data directly from the image,
		     * i.e. skip the copy into the buffer.
		     */
		for (row = 0 ; row < imHeight ; row++, pIm += bpl)
		    ERR_GEN_CLEANUP(TIFFWriteScanline(tif, pIm, row, 0) < 0,
				    cErrIO_WriteError);
	    }
	    else
	    {
		    /* We're on a little endian platform, so swap
		     * the bytes into a buffer before writing them
		     * out.
		     */
		wordsPerLine = (TIFFScanlineSize(tif) + 3)/4;

		buf = (UInt8 *)IPW_CALLOC(1, bpl, cAllocatorData);
		ERR_GEN_CLEANUP(buf == NULL, cErrNomem);

		for (row = 0 ; row < imHeight ; row++, pIm += bpl)
		{
		    (*swapLineProc)(pIPWState,
				    (UInt32 *)pIm,
				    (UInt32 *)buf, wordsPerLine);
		
		    ERR_GEN_CLEANUP(TIFFWriteScanline(tif, buf, row, 0) < 0,
				    cErrIO_WriteError);
		}
		IPW_FREE(buf, cAllocatorData);
	    }
	}
    }
    TIFFCloseStruct(tif); 	/* still keep the tif structure*/
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(buf, cAllocatorData);
    ERR_RETURN;
}

void
w_cleanupTiff( TIFF *tif)
{
        if (tif->tif_cleanup)
                (*tif->tif_cleanup)(tif);
        TIFFFreeDirectory(tif);
        if (tif->tif_rawdata && (tif->tif_flags&TIFF_MYBUFFER))
                _TIFFfree(tif, tif->tif_rawdata);
        if (isMapped(tif))
                TIFFUnmapFileContents(tif, tif->tif_base, tif->tif_size);
        (void) TIFFCloseFile(tif);
        _TIFFfree(tif, (char *)tif);
}

/*

 $Log: tiffio.c,v $
 Revision 1.8  2002/11/20 22:22:16  jlin
 comment change

 Revision 1.7  2002/11/14 02:26:38  jlin
 minor change

 Revision 1.6  2002/11/14 01:00:07  jlin
 by pass TiffGetFile for colorMap info. use tif struct directly.

 Revision 1.5  2002/01/03 22:58:14  jlin
 Bug fix. Read color map correcly

 Revision 1.4  2000/07/08 01:19:25  jlin
 fix memory leak for multi-page tiff read in w_readTiff

 Revision 1.2  2000/06/23 23:24:22  jlin
 Add function to save multi-page tif image

 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.59  1998/11/12 00:26:30  aber
 aber on Wed Nov 11 16:26:12 PST 1998

 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Synced fixes from ddavies:
 Have w_readTiff pay attention to the new page number parameter.  It does the
 dumb thing, simply skipping TIFF directories until it finds the right one.

 Revision 1.58  1998/02/19 21:27:40  aber
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Read the color palette and attach the colormap to the PIXR.

 Revision 1.57  1998/01/14 00:16:11  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.56  1997/06/30 23:11:54  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Delete inclusion of Sandpiper's internal.h

 Revision 1.55  1997/04/21 23:28:15  aber
 aber on Mon Apr 21 16:26:08 PDT 1997

 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Attach the colormap to the TIFF if there's one in the PIXR, call it
 paletted. This update really enables some of the palette code already
 in place.

 Revision 1.54  1997/04/17 02:44:10  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Fix bug found when reading single Tile, non-pixr images.  A normal PIXR
 was allocated to hold the result, then the bytes/line was adjusted to
 eliminate the frame and pad, the bits were read in and the frame and pad
 were put back on.  Unfortunately, when the bytes/line was made small, no
 one set the frame size in the PIXR to 0.  This caused the routine that
 added the frame back on to complain.
 The reason this wasn't caught months (years?) ago is because this only
 happen when we use the original LefflerTI
 TIFF code, which only happens in when we make the IPWrap regression
 test with the slow libs, which isn't done in the nightly build.

 Revision 1.53  1997/04/15 22:48:45  ddavies
 ddavies on Tue Apr 15 15:46:41 PDT 1997

 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Use non-initialized PIXR creation where possible.
 Change way in which RGB pixels are inserted into PIXRs.  In the old method,
 the image was completely cleared (to make sure byte-swapped frame was cleared),
 then the pixels were written in natural byte order, then the bytes were
 swapped.  The new method uses an uninitialized PIXR, uses UCHAR_ACCESS to
 write the pixels in PIXR order and skips the final byteswap.

 Revision 1.52  1997/03/17 20:19:16  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Add Lossy Dither compression as a compression type.  Also fix a bug in
 which rowsPerStrip was being calculated as 8K/bpl instead of 8K / TIFFScanlineSize.
 bpl changes if you go on an MMX machine, but the strip size should not change.

 Revision 1.51  1996/11/20 01:19:29  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 CommonCheck: Pass the TIFF structure containing a pointer to the IPWrap state structure
 to all routines that call TIFFWarning and TIFFError so that the TIFF code
 can use the normal error reporting mechanism.

 Revision 1.50  1996/05/24 04:46:31  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Include shrcc.pub and iotiff.pub.  Also fix return code checking after
 calling TIFFReadScanline.

 Revision 1.49  1996/05/16 17:32:07  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Added singleLine parameter to w_readTiff.  When we're doing a non-tiled,
 PIXR-enabled read, we use the singleLine parameter to decide whether to read
 it in strips or in units of one scan line.  This is for testing only.  It
 does slow things down by about 25%.

 Revision 1.48  1996/05/11 01:39:13  ddavies
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 Add the options parameter to w_writeTiff.  This allows one to pass down
 the 2D and Pad options when compressing using the Group 3 method.

 Revision 1.47  1996/01/04 20:29:09  ejaquith
 tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.46  1995/12/13  04:15:44  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.45  1995/12/06  03:35:02  ddavies
 * ddavies on Tue Dec 5 19:34:00 PST 1995
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Since TIFFSetField can't deal with a float when compiled with acc,
 * change the calls to TIFFSetField that set the TIFFTAG_XRESOLUTION and
 * TIFFTAG_YRESOLUTION so they pass pointers to floats instead of floats.
 *
 * Revision 1.44  1995/11/13  20:19:27  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.43  1995/11/04  01:36:16  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.42  1995/10/18  04:19:52  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Add casts to make Visual C++ happy.  Add NULL pointers to call to
 * w_getDIBPixrLineProc since it can really return 4 types of routines and
 * we now have to specify which one we really use.
 *
 * Revision 1.41  1995/10/13  22:07:55  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.40  1995/10/12  04:29:21  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.39  1995/10/11  19:41:58  ddavies
 * ddavies on Wed Oct 11 12:40:11 PDT 1995
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Make the tiff routines that think they're calling lower level routines that
 * know about PIXRs set the new TIFF_PIXRPROVIDED flag.  When they're calling
 * routines that either don't know about PIXRs or go through baroque Leffler
 * code that makes up secret buffers, clear the TIFF_PIXRPROVIDED flag.
 *
 * Revision 1.38  1995/10/10  03:12:58  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Added file depth and channel checking to only allow valid TIFF files
 * to be read/written.
 * Added special comments in comment blocks to all public i/o routines
 * to allow user to know with certainty what depths/planes we support!
 *
 * Revision 1.37  1995/10/02  19:26:50  jcraig
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * w_invert takes a pointer to a pointer to a destination pixr.
 *
 * Revision 1.36  1995/09/29  00:35:49  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.35  1995/09/21  23:00:51  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.34  1995/09/19  21:14:43  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Fixed transfer of image DPI from PIXR to TIFF and from TIFF to PIXR by
 * using macros ScaledInt1616_TO_Int32() and Float_TO_ScaledInt1616().
 *
 * Revision 1.33  1995/09/14  18:57:16  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 *  Fixed types to be used with w_pixrGetResolutionDPI().
 *
 * Revision 1.32  1995/09/13  00:25:28  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Made routine call w_pixrSetFrame() to set frame bits to zero in instance
 * where we created a frame of 32 for space, but then call w_addFrameIP().
 * to add frame of 32 in place later.
 *
 * Revision 1.31  1995/09/12  16:55:19  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.30  1995/09/08  00:15:24  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.29  1995/09/07  01:37:36  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.28  1995/09/06  20:26:25  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.27  1995/09/01  02:21:00  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.26  1995/08/31  21:40:34  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.25  1995/08/29  21:57:39  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.24  1995/08/22  02:23:00  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Added calls to IPWStateTerm() in all routines that were using
 * IPWStateInit() to clean memory.
 *
 * Revision 1.23  1995/08/17  22:01:50  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.22  1995/05/09  03:22:29  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * This is ipwrap`s interface to the Leffler TIFF library.  Both readPrTiff
 * and writePrTiff have been changed to allow them to use the new ipwrap
 * routines that operate directly on PIXRs.  When presence of these routines
 * is detected using the new TIFF_PIXRENCODE and TIFF_PIXRDECODE flags,
 * we can skip the steps that swap bytes and add/remove borders.
 * Currently this is only done for gray and binary images.
 *
 * Revision 1.21  1995/03/30  21:37:11  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Change readTIFFPr's arg from a file pointer to a FileHandle (FH).  This
 * trying to derive the file descriptor from the file pointer - an operation
 * allows readTIFFPr to use the file's name to open the file instead of
 * Ooops, reverse the order of those last two lines --
 * only available in POSIX, not ANSI, worlds.
 *
 * Revision 1.20  1995/03/23  22:55:37  ddavies
 * ddavies on Thu Mar 23 14:54:30 PST 1995
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Fix byte-order bug in yesterday's change.  We have to read in all the
 * strips or tiles, then byteswap the entire block, then add the frame.  The
 * frame-adding code expects the byte order to be correct.  The old code
 * added the frame, then did the byteswapping.
 *
 * Revision 1.19  1995/03/23  03:16:51  ddavies
 * ddavies on Wed Mar 22 19:14:40 PST 1995
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Modify so that stripped images are read in units of strips instead of
 * scan lines.  Frames are added after all strips are read.
 * Also modify to read the DAE's wacky 1-tile format.  Incorporate workaround
 * for apparent bug in DAE's files.  The TIFF 6.0 spec says that images
 * containing tiles must be padded in both x and y to an integral number
 * of tiles.  If there is only one tile ("not recommended"), the image must
 * be padded out to the size of this tile.  The files are apparently padded
 * in x, but seem to stop at the bottom of the image (not the tile) in y.
 * The tile-reading procedure is hacked to say "if there is only one tile,
 * change it's height to match the height of the image and change the
 * total number of uncompressed bytes to produce to be the tile's bytes/row
 * times the reduced height.  Code now reads one tile, adds a frame, then
 * cuts off the padding.
 *
 * Revision 1.18  1995/03/21  04:40:04  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Added CCITTRLE format so we can read and write files in Modified Huffman
 * format.
 *
 * Revision 1.17  1995/02/08  18:36:51  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Bug fix.
 * In Group3 and Group4 compression, Leffler's routines were reading and
 * writing one byte past the end of the correctly allocated scanline buffer.
 * To fix, now I allocate one more byte for the line buffer.
 *
 * Revision 1.16  1994/12/19  23:25:33  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Fixed i/o bug where the PIXRs weren't getting byte swapped after i/o from
 * disk on the PC.
 *
 * Revision 1.15  1994/12/08  22:21:38  ejaquith
 * ejaquith on Thu Dec 8 14:19:24 PST 1994
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Added TIFF i/o of 24-bit color RGB image capability.
 * This will read 3 8-bit planes, or 24-bit "interlaced", and will
 * always write out 24-bit "interlaced".
 * It generates 3 8-bit plane PIXRs.
 * Made exiting on error from TIFF i/o do proper cleanup first.
 *
 * Revision 1.14  1994/11/02  01:38:57  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Fixed a bug that was either always present or just in the new TIFF library
 * when we are reading TIFF files to PIXRs.
 * If the depth is 1, the photometric was MINISBLACK, and the file was
 * compressed with a CCITT compression scheme, then the image would get
 * inverted.  This is because the CCITT decompress automatically inverts
 * the image.  Thus we had an extra invert, removed when CCITT compressed.
 * Changed format for calling writePrTiff() routine.  Now it takes a
 * 16 bit integer for the desired compression, which is what gets directly
 * written into the compression TAG.  Previously, we used strings and
 * strcmp() calls, and didn't use the TIFF TAG values nicely defined for
 * us.
 * writePrTiff() now is also smart about the compression type desired.
 * It only allows one of the 5 supported types as the others still have
 * some bugs when writing.
 *
 * Revision 1.13  1994/11/01  02:46:14  ejaquith
 * ejaquith on Mon Oct 31 18:45:47 PST 1994
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Removed inclusion of "aodsport.h" as this file is obsolete.
 *
 * Revision 1.12  1994/10/31  22:54:38  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Made TIFF fetching of BITSPERSAMPLE TAG fetch the default TAG value (1)
 * for when it is not included in the TIFF header.
 * Added JPEG compression TAG writing ability for when JPEG compression
 * is completed.
 * Made TIFF writing return abort() error on writing errors.
 *
 * Revision 1.11  1994/10/27  17:08:42  ejaquith
 * ejaquith on Thu Oct 27 10:06:53 PDT 1994
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Fixed purify warning in TIFF resolution reading.
 * (TIFFGetField() will not touch any desired parameter that has not
 * been explicitly set so purify found uninitialized values getting ref'd!)
 *
 * Revision 1.10  1994/10/26  18:44:04  ejaquith
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Added dpi reading and writing in readTiffPr() and writePrTiff().
 * Will read and write resolutions in x and y now.  In writing, if the
 * pixr has its resolution fields set to 0 (default), it won't write
 * the resolution fields out.
 *
 * Revision 1.9  1994/10/10  22:36:01  lperry
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Change to use new header file io.pub instead of iopros.h
 *
 * Revision 1.8  1994/05/26  19:26:41  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Enca
 * Ooops.  Encapsulate writePrTiff's and readPrTiff's declarations of
 * procName inside #ifndef PRODUCTION clauses.
 *
 * Revision 1.7  1994/05/05  01:58:19  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Add "const" to "char *" parameters of getMode and TIFFOpenStream to make gcc
 * happy.  Ditto for writePrTiff.
 *
 * Revision 1.6  1994/04/13  23:49:34  macleod
 * macleod on Wed Apr 13 16:49:06 PDT 1994
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Took out "reading tiff file" and "writing tiff file" messages,
 * which were always printed, which caused problems for the
 * image database testbed.
 *
 * Revision 1.5  1994/01/22  01:43:38  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Include stdlib.h so calloc and free have prototypes (WATCOM complained).
 *
 * Revision 1.4  1994/01/07  22:26:58  dferrell
 * dferrell on Fri Jan 7 14:25:11 PST 1994
 *
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Removed unnecessary #include <malloc.h> line.
 *
 * Revision 1.3  1993/12/03  01:20:21  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:18:57  ddavies
 * tiffio.c -> /product/ipcore/ipwrap/src/RCS/tiffio.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

