/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tiffld.c:   routines that link ipwrap's Lossy Dither encoding and
 *		decoding routines to Leffler's TIFF libraries.
 *
 *                     
 */



#include <string.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "defines.pub"
#include "iaerror.pub"

/* prototypes */
#include "tiffiop.h"
#include "props.pub"
#include "ld.pub"
#include "tiffccut.h"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tiffld.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");


/*********************************************************************/
/**************************** PROTOTYPES *****************************/
/*********************************************************************/

static int
ldWriteEOFBAndFlush (TIFF *tif);

/*********************************************************************/
/************************** ENCODING CODE ****************************/
/*********************************************************************/

/*
 * NB: tidata is the type representing *(tidata_t);
 *     if tidata_t is made signed then this type must
 *     be adjusted accordingly.
 */
typedef unsigned char tidata;

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Prepare the state used to encode Lossy Dither blocks.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
ldPreEncode (
    TIFF*               tif)
{
    void    *pLDCompInfo;        /* pointer to opaque data type */
    UInt32   stateBlockSize;        /* size of opaque data block */
    UInt32   imWidth;               /* width and depth are passed to lower */
    Int32    retCode;
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

	/* See if the state block already exists.  The tif_data
	 * entry is cleared when the TIFF structure is originally
	 * allocated.
	 */
    pLDCompInfo = (void *)tif->tif_data;
    if (pLDCompInfo != NULL)
	return (1);		/* nothing else to do */

	/* If we're not going to be given a PIXR to encode,
	 * complain.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* We need a new state block.  Start by determining the
	 * image width */
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);

    retCode = w_getSizeOfLDCompressionInfo(pIPWState,
					   imWidth, &stateBlockSize);

    if (retCode != cSuccessful)
    {
	return (0);		/* failed */
    }

	/* Allocate the state block */
    pLDCompInfo = (void *)_TIFFmalloc(tif, stateBlockSize);

    if (pLDCompInfo == NULL)
    {
	return (0);		/* no memory for this block */
    }

	/* Initalize the state block */
    retCode = w_initializeLDCompressionInfo(pIPWState,
					    pLDCompInfo,
					    imWidth,
					    stateBlockSize);

    if (retCode != cSuccessful)
    {
	_TIFFfree(tif, pLDCompInfo);
	return (0);		/* didn't work */
    }

	/* it worked, so put a pointer to the state block into the
	 * tif space reserved for it.
	 */
    tif->tif_data = (tidata_t) pLDCompInfo;

    return (1);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Encode a PIXR strip into Lossy Dither encoding.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *  prS::            pointer to PIXR that holds uncompressed
 *                   data.  This is sort of a cheat in that
 *                   the Leffler convention is that this is a
 *                   pointer to a data buffer.  But, hey, this
 *                   is C, and all pointers are cousins, right?
 *  uncompBytes::    number of uncompressed bytes this routine
 *                   is expected to consume.  We assume this
 *                   is a multiple of the number of bytes in
 *                   an image row.
 *  sampleNum::      number of current pixel sample.
 *                   This is apparently used when different
 *                   samples are stored in different image planes.
 *                   It is not used here.
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
ldEncode (
    TIFF               *tif,
    tidata_t            prS,
    tsize_t             uncompBytes,
    tsample_t           sampleNum)
{
    UInt32	exhaustedImageData;	/* if set on return from
					 * compressRowsMH, we've processed the
					 * correct number of lines.
					 */
    UInt32	exhaustedCompData;	/* if set on return from
					 * compressRowsMH, we've reached the
					 * end of the compressed data buffer
					 */
    UInt32	percentDone;		/* if we ran out of compressed data
					 * before completing the image, this
					 * tells how much of the image
					 * has been compressed.
					 */
    UInt32	bytesWritten;		/* number of compressed bytes read on
					 * each call. */
    UInt32	numLines;		/* number of lines to compress */
    Int32	retCode;
    UInt32	remainingBytesInBuf;	/* number of vacant bytes in the
					 * compressed data buffer. */
    IPWState *pIPWState = NULL;

	/* Initialize the IPWState from the spot where it was saved
	 * by the nice IPWrap tiffio code.
	 */
    pIPWState = (IPWState *)tif->tif_ipwstate;

    sampleNum = sampleNum;	/* keep the compilers happy */

	/* try to compress all the rows.  We figure out the number of lines
	 * to compress by dividing uncompBytes by the TIFF package's
	 * idea of the number of bytes/line.  The information in the
	 * PIXR gives the real bytes/line.
	 */
    remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
    numLines = uncompBytes / TIFFScanlineSize(tif);
    bytesWritten = 0;		/* just in case compressRowsLD doesn't
				 * do anything for some reason */

    retCode = w_compressRowsLD(
	pIPWState,
	(void *)(tif->tif_data), /* private decode data */
	(PIXR *) prS,		/* source pixr */
	TIFFCurrentRow(tif),	/* first dest line of pixr */
	numLines,		/* num lines to compress */
	tif->tif_rawcp,		/* pointer to comp data */
	remainingBytesInBuf,	/* number of comp data
				 * bytes remaining */
	cFalse,			/* this is a new image block */
	cFalse,			/* this is new comp data */
	&exhaustedImageData,
	&exhaustedCompData,
	&percentDone,
	&bytesWritten);		/* no array of line offsets */

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* Increment the number of bytes now in the buffer and byte
	 * pointer.
	 */
    tif->tif_rawcp += bytesWritten;
    tif->tif_rawcc += bytesWritten;

	/* Now execute a loop that keeps flushing compressed data and
	 * doing more compressing until the proper number of lines has
	 * been processed.
	 */
    while (!exhaustedImageData)
    {
	/* TIFFFlushData1 flushes the buffer of compressed data
	 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
	 * -1 is weird, but that's what the original code did.
	 */
	if (!TIFFFlushData1(tif))
	{
	    return (-1);
	}

	remainingBytesInBuf = tif->tif_rawdatasize;
	bytesWritten = 0;	/* just in case compressRowsLD doesn't
				 * do anything for some reason */

	retCode = w_compressRowsLD(
			    pIPWState,
			    (void *)(tif->tif_data),
			    (PIXR *) prS,
			    TIFFCurrentRow(tif),
			    numLines,
			    tif->tif_rawcp,
			    remainingBytesInBuf,
			    cTrue,		/* this is an old image block */
			    cTrue,		/* this is old comp data */
			    &exhaustedImageData,
			    &exhaustedCompData,
			    &percentDone,
			    &bytesWritten);

	if (retCode != cSuccessful)
	{
	    return (0);	/* something failed */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;
    }

 
	/* if it appears to have worked so return happily.  We don`t
	 * check to see if we ran out of image lines since we wouldn`t
	 * have exited the loop above if there were more image lines.
	 * We don't check to see if we've run out of room on the
	 * comp data buffer since there's no reason to think that this
	 * would have happened on compression.
	 */
    return (1);	/* worked */
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Clean up after doing LD compression.
 *
 * LongDescription:::
 *	This procedure copies the action of Leffler's Fax4PostEncode
 *	procedure in tif_fax4.c.  Basically, it writes two EOL
 *	characters, then flushes the pipeline.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		1	it worked
 *		0	it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LD
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
ldPostEncode (
    TIFF*               tif)
{
	/* To end a Lossy Dither strip or tile, write two EOLs */
    return ldWriteEOFBAndFlush(tif);
}




/*********************************************************************/
/************************** DECODING CODE ****************************/
/*********************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Prepare the state used to decode Lossy Dither blocks.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
ldPreDecode (
    TIFF               *tif)
{
    void     *pLDDecompInfo;      /* pointer to opaque data type */
    UInt32    stateBlockSize;        /* size of opaque data block */
    Int32     retCode;
    UInt32    imWidth;
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

	/* See if the state block already exists.  The tif_data
	 * entry is cleared when the TIFF structure is originally
	 * allocated.
	 */
    pLDDecompInfo = (void *)tif->tif_data;
    if (pLDDecompInfo != NULL)
	return (1);		/* nothing else to do */

	/* If we're not going to be given a PIXR to decode into,
	 * just return unsuccessfully.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* We need a new state block.  Start by determining the
	 * width of the thing we'll be decoding.  If it's tiled,
	 * it's the tile width.  Otherwise, it's the image width.
	 */
    if (!TIFFIsTiled(tif))
	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);
    else
	TIFFGetField(tif, TIFFTAG_TILEWIDTH, &imWidth);

	/* We need a new state block.  Start by determining the
	 * size of the state block */
    retCode = w_getSizeOfLDDecompressionInfo(pIPWState, imWidth,
					     &stateBlockSize);

    if (retCode != cSuccessful)
    {
	return (0);		/* failed */
    }

	/* Allocate the state block */
    pLDDecompInfo = (void *)_TIFFmalloc(tif, stateBlockSize);

    if (pLDDecompInfo == NULL)
    {
	return (0);		/* no memory for this block */
    }

	/* initialize the state block */
    retCode = w_initializeLDDecompressionInfo(pIPWState,
						 pLDDecompInfo,
						 imWidth,
						 stateBlockSize);

    if (retCode != cSuccessful)
    {
	_TIFFfree(tif, pLDDecompInfo);
	return (0);		/* didn't work */
    }

	/* it worked, so put a pointer to the state block into the
	 * tif space reserved for it.
	 */
    tif->tif_data = (tidata_t) pLDDecompInfo;


    return (1);			/* it worked */
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Decode rows of a PIXR from a stream of Lossy Dither data.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *  uncompBytes::    number of uncompressed bytes this routine
 *                   is expected to produce.  We assume this
 *                   is a multiple of the number of bytes in
 *                   an image row.
 *  sampleNum::      number of current pixel sample.
 *                   This is apparently used when different
 *                   samples are stored in different image planes.
 *                   It is not used here.
 *
 * Output:::
 *  prD::            pointer to PIXR that will hold decompressed
 *                   data.  This is sort of a cheat in that
 *                   the Leffler convention is that this is a
 *                   pointer to a data buffer.  But, hey, this
 *                   is C, and all pointers are cousins, right?
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 *  Non-standard parameter order!:: 
 *       The output parameter is not the last in the param order
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
ldDecode (
    TIFF               *tif,
    tidata_t            prD,
    tsize_t             uncompBytes,
    tsample_t           sampleNum)
{
    Int32	firstBlock;		/* for testing only */
    UInt32	exhaustedImageData;	/* if set on return from
					 * decompressRowsLD, we've
					 * processed the correct number
					 * of lines.
					 */
    UInt32	exhaustedCompData;	/* if set on return from
					 * decompressRowsLD, we've reached the
					 * end of the compressed data
					 */
    UInt32	percentDone;		/* if we ran out of compressed data
					 * before completing the image,
					 * this tells how much of the image
					 * has been decompressed.
					 */
    UInt32	bytesRead;		/* number of compressed bytes read on
					 * each call. */
    Int32	numRowsToRead;		/* number of rows we're to read */
    Bool	continueCompBlock;	/* set if we've already decompressed
					 * some lines of this strip or tile. */
    Int32	retCode;
    IPWState   *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

    sampleNum = sampleNum;	/* keep the compilers happy */

	/* If a pixr is not provided, return with an error.  We
	 * like PIXRs a lot.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* If we're going to read the first line of a strip, then
	 * we'll tell the callee that the compression block does
	 * not continue.
	 */
    if (tif->tif_rawcp == tif->tif_rawdata)
	continueCompBlock = cFalse;
    else
	continueCompBlock = cTrue;

    numRowsToRead = uncompBytes / TIFFScanlineSize(tif);
    if (TIFFIsTiled(tif))	/* can be no bigger than the tile height */
    {
	if (numRowsToRead > tif->tif_dir.td_tilelength)
	    numRowsToRead = tif->tif_dir.td_tilelength;
    }
    else			/* can be no bigger than the strip height */
    {
	if (numRowsToRead > tif->tif_dir.td_rowsperstrip)
	    numRowsToRead = tif->tif_dir.td_rowsperstrip;
    }

/* This dreck is here so we can test our ability to stop in the middle of
 * a strip, then restart.
 */
#if 0
	/* TESTING ONLY
	 * do this in two parts to see if that works correctly */
    firstBlock = tif->tif_rawcc / 2;

    retCode = w_decompressRowsLD(
			pIPWState,
			(void *)(tif->tif_data), /* private decode data */
			TIFFCurrentRow(tif),	/* first dest line of pixr */
			numRowsToRead,		/* num lines to decompress */
			tif->tif_rawcp,		/* pointer to comp data */
			firstBlock,		/* number of comp data bytes */
			cFalse,			/* this is a new image block */
			continueCompBlock,	/* cTrue if is new comp data */
			cFalse,			/* don't expect to get to
						 * the end of the comp data
						 * in this call */
			(PIXR *) prD,		/* dest pixr */
			&exhaustedImageData,
			&exhaustedCompData,
			&percentDone,
			&bytesRead);

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* if it appears to have worked, make sure we did all of
	 * image and did NOT run out of compressed data.
	 */
    if ((exhaustedImageData) || (!exhaustedCompData) ||
	(bytesRead != firstBlock))
    {
	return (0);	/* failed */
    }
#else
    firstBlock = 0;
#endif
	/* go decompress all the rows.  We figure out the number of lines
	 * to decompress by dividing uncompBytes by the TIFF package's
	 * idea of the number of bytes/line.  The information in the
	 * PIXR gives the real bytes/line.
	 */
    retCode = w_decompressRowsLD(
			pIPWState,
			(void *)(tif->tif_data), /* private decode data */
			TIFFCurrentRow(tif),	/* first dest line of pixr */
			numRowsToRead,		/* num lines
						 * to decompress */
			tif->tif_rawcp + firstBlock, /* pointer to comp data */
			tif->tif_rawcc - firstBlock, /* number of comp data
						 * bytes remaining */
#if 0
			cTrue,			/* this is an old image block */
			cTrue,			/* this is a continuation of
						 * the previous comp data */
#else
			cFalse,			/* this is a new image block */
			continueCompBlock,	/* cTrue if is new comp data */
#endif
			cTrue,			/* expect to get to the end
						 * of the comp data in this
						 * call */
			(PIXR *) prD,		/* dest pixr */
			&exhaustedImageData,
			&exhaustedCompData,
			&percentDone,
			&bytesRead);

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* Update the input pointers.  This is used if we're not
	 * reading the entire strip or tile in one shot.
	 */
    tif->tif_rawcc -= bytesRead;
    tif->tif_rawcp += bytesRead;

	/* if it appears to have worked, make sure we did all of
	 * image.  We do NOT check to see if we ran out of image
	 * data because we probably didn't.  There's an
	 * EndOfFacsimile code at the end of each strip that we
	 * don't read.
	 */
    if (!exhaustedImageData)
	return (0);	/* failed */
    else
	return (1);	/* worked */
}


/*****************************************************************************
 *************************** UTILITY ROUTINES ********************************
 *****************************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Tidy up after Lossy Dither compression or decompression.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		void
 *
 * Notes:::
 *  Non-standard error return codes!::
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static void
ldCleanup (
    TIFF               *tif)
{
    if (tif->tif_data != NULL)
    {
	_TIFFfree(tif, tif->tif_data);
	tif->tif_data = NULL;
    }
	/* Turn off the flags that say the encoder or decoder knows
	 * about PIXRs.
	 */
    tif->tif_flags &= ~(TIFF_PIXRDECODE | TIFF_PIXRENCODE);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Write EOFB code and flush pipeline after Lossy Dither comp/decomp
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		0 - failed
 *		1 - it worked
 *
 * Notes:::
 *  Non-standard error return codes!::
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
ldWriteEOFBAndFlush (
    TIFF               *tif)
{
    UInt32	bytesWritten;		/* number of compressed bytes read on
					 * each call. */
    UInt32	remainingBytesInBuf;	/* number of vacant bytes in the
					 * compressed data buffer. */
    UInt32	numEOLs;		/* number of EOLs in an EOFB. */

	/* Initialize the IPWState from the spot where it was saved
	 * by the nice IPWrap tiffio code.
	 */
    IPWState *pIPWState = (IPWState *)tif->tif_ipwstate;
    Int32	retCode;

	/* try to write out two EOL codes to form one End Of Facsimile Block.
	 * We need to detect the case of running out of buffer space
	 * when trying to do this.
	 */
    for (numEOLs = 2; numEOLs > 0; numEOLs--)
    {
	remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
	bytesWritten = 0;		/* just in case w_writeLDEOL doesn't
					 * do anything for some reason */
	retCode = w_writeLDEOL(
	    pIPWState,
	    (void *)(tif->tif_data),	/* private decode data */
	    tif->tif_rawcp,		/* pointer to comp data */
	    remainingBytesInBuf,	/* number of comp data
					 * bytes remaining */
	    &bytesWritten);		/* number of bytes written */
    
	if ((retCode != cSuccessful) && (retCode != ia_bufferExceeded))
	{
	    return (0);			/* disaster */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;

	    /* If we failed to write the EOL because we filled
	     * the buffer first, flush it */
	if (retCode == ia_bufferExceeded)
	{
		/* TIFFFlushData1 flushes the buffer of compressed data
		 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
		 * -1 is weird, but that's what the original code did.
		 */
	    if (!TIFFFlushData1(tif))
	    {
		return (-1);
	    }
	}
    }

	/* Now try to flush the pipeline to the bytestream.
	 * As usual, we might run out of data.
	 */
    while (cTrue)
    {
	remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
	bytesWritten = 0;		/* just in case w_flushLD doesn't
					 * do anything for some reason */
	retCode = w_flushLD(
	    pIPWState,
	    (void *)(tif->tif_data),	/* private decode data */
	    tif->tif_rawcp,		/* pointer to comp data */
	    remainingBytesInBuf,	/* number of comp data
					 * bytes remaining */
	    &bytesWritten);		/* number of bytes written */
    
	if ((retCode != cSuccessful) && (retCode != ia_bufferExceeded))
	{
	    return (0);			/* disaster */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;

	    /* If we failed to flush the pipeline because we filled
	     * the buffer first, flush the buffer */
	if (retCode == ia_bufferExceeded)
	{
		/* TIFFFlushData1 flushes the buffer of compressed data
		 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
		 * -1 is weird, but that's what the original code did.
		 */
	    if (!TIFFFlushData1(tif))
	    {
		return (-1);
	    }
	}
	else
	    break;	/* we succeeded in flushing the pipeline */
    }

	/* we're happy, so return */
    return (1);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Set up a bunch of procedure ptrs for doing compression and decompression.
 *
 * LongDescription:::
 *	This procedure sets up a bunch of procedure pointers so that
 *	Lossy Dither compression and decompression is done.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		1	it worked
 *		0	it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, LDC
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
TIFFInitLossyDither (
    TIFF*               tif)
{
	/* Use our own decoding routines */
    tif->tif_predecode = ldPreDecode;
    tif->tif_decoderow = ldDecode;
    tif->tif_decodestrip = ldDecode;
    tif->tif_decodetile = ldDecode;

	/* Use our own encoding routines. */
    tif->tif_preencode   = ldPreEncode;
    tif->tif_encoderow   = ldEncode;
    tif->tif_encodestrip = ldEncode;
    tif->tif_encodetile  = ldEncode;
    tif->tif_postencode  = ldPostEncode;
    tif->tif_cleanup	 = ldCleanup;

	/* tell users that we're using decoding code that deals
	 * directly with PIXRs.
	 */
    tif->tif_flags |= TIFF_PIXRDECODE | TIFF_PIXRENCODE;

    tif->tif_flags &= ~TIFF_NOBITREV;	/* we do not handle bit reversal */
    return (1);
}



/*

 $Log: tiffld.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.3  1998/01/27 01:57:19  ddavies
 tiffld.c -> /product/ipcore/ipwrap/src/RCS/tiffld.c,v

 CommonCheck: Add tif arg to all invocations of _TIFFmalloc, _TIFFfree and _TIFFrealloc.
 These routines now use tif->ipwstate to get the IPState memory allocators
 instead of using malloc and free directly.

 Revision 1.2  1997/03/18 20:14:09  ddavies
 tiffld.c -> /product/ipcore/ipwrap/src/RCS/tiffld.c,v

 Remove combineResults parameter from call to w_decompressLDRows.

 Revision 1.1  1997/03/17 20:16:05  ddavies
 Initial revision


 */

