/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tiffpack.c: routines that link ipwrap's PackBits encoding and
 *		decoding routines to Leffler's TIFF libraries.
 *
 *                     
 */



#include <string.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "defines.pub"

/* prototypes */
#include "tiffiop.h"
#include "props.pub"
#include "pack.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tiffpack.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");



/*********************************************************************/
/************************** ENCODING CODE ****************************/
/*********************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Prepare the state used to encode PackBits blocks.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
packBitsPreEncode (
    TIFF*               tif)
{
    void    *pPBCompInfo;           /* pointer to opaque data type */
    UInt32   stateBlockSize;        /* size of opaque data block */
    UInt32   imWidth;               /* width and depth are passed to lower */
    UInt16   imDepth;               /* level routines.  imDepth is a UInt16
				     * because that's how TIFFGetField wants
				     * to return the depth. */
    Int32    retCode;
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

	/* If we're not going to be given a PIXR to encode,
	 * complain.
	 */
    if (!pixrProvided(tif))
	return (0);

	/* See if the state block already exists.  The tif_data
	 * entry is cleared when the TIFF structure is originally
	 * allocated.
	 */
    pPBCompInfo = (void *)tif->tif_data;
    if (pPBCompInfo != NULL)
	return (1);		/* nothing else to do */

	/* We need a new state block.  Start by determining the
	 * image width and height */
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);
    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &imDepth);


    retCode = w_getSizeOfPackBitsCompressionInfo(pIPWState,
						 imWidth, (UInt32) imDepth,
						 &stateBlockSize);

    if (retCode != cSuccessful)
    {
	return (0);		/* failed */
    }

	/* Allocate the state block */
    pPBCompInfo = (void *)_TIFFmalloc(tif, stateBlockSize);

    if (pPBCompInfo == NULL)
    {
	return (0);		/* no memory for this block */
    }

	/* Initalize the state block */
    retCode = w_initializePackBitsCompressionInfo(pIPWState,
						  pPBCompInfo,
						  imWidth,
						  (UInt32) imDepth,
						  stateBlockSize);

    if (retCode != cSuccessful)
    {
	_TIFFfree(tif, pPBCompInfo);
	return (0);		/* didn't work */
    }

	/* it worked, so put a pointer to the state block into the
	 * tif space reserved for it.
	 */
    tif->tif_data = (tidata_t) pPBCompInfo;

    return (1);
}



/*
 * NB: tidata is the type representing *(tidata_t);
 *     if tidata_t is made signed then this type must
 *     be adjusted accordingly.
 */
typedef unsigned char tidata;

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Encode rows of a PIXR into a stream of PackBits data.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *  prS::            pointer to PIXR that holds uncompressed
 *                   data.  This is sort of a cheat in that
 *                   the Leffler convention is that this is a
 *                   pointer to a data buffer.  But, hey, this
 *                   is C, and all pointers are cousins, right?
 *  uncompBytes::    number of uncompressed bytes this routine
 *                   is expected to consume.  We assume this
 *                   is a multiple of the number of bytes in
 *                   an image row.
 *  sampleNum::      number of current pixel sample.
 *                   This is apparently used when different
 *                   samples are stored in different image planes.
 *                   It is not used here.
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
packBitsEncode (
    TIFF               *tif,
    tidata_t            prS,
    tsize_t             uncompBytes,
    tsample_t           sampleNum)
{
UInt32	exhaustedImageData;	/* if set on return from
				 * compressRowsPackBits, we've processed the
				 * correct number of lines.
				 */
UInt32	exhaustedCompData;	/* if set on return from
				 * compressRowsPackBits, we've reached the
				 * end of the compressed data buffer
				 */
UInt32	percentDone;		/* if we ran out of compressed data before
				 * completing the image, this tells how
				 * much of the image has been decompressed.
				 */
UInt32	bytesWritten;		/* number of compressed bytes read on
				 * each call. */
UInt32	numLines;		/* number of lines to compress */
Int32	retCode;
UInt32	remainingBytesInBuf;	/* number of vacant bytes in the compressed
				 * data buffer. */
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

    sampleNum = sampleNum;	/* keep the compilers happy */

	/* try to compress all the rows.  We figure out the number of lines
	 * to compress by dividing uncompBytes by the TIFF package's
	 * idea of the number of bytes/line.  The information in the
	 * PIXR gives the real bytes/line.
	 */
    remainingBytesInBuf = tif->tif_rawdatasize - tif->tif_rawcc;
    numLines = uncompBytes / TIFFScanlineSize(tif);
    bytesWritten = 0;		/* just in case compressRowsPackBits doesn't
				 * do anything for some reason */


    retCode = w_compressRowsPackBits(
	pIPWState,
	(void *)(tif->tif_data), /* private decode data */
	(PIXR *) prS,		/* source pixr */
	TIFFCurrentRow(tif),	/* first dest line of pixr */
	numLines,		/* num lines to compress */
	tif->tif_rawcp,		/* pointer to comp data */
	remainingBytesInBuf,	/* number of comp data
				 * bytes remaining */
	cFalse,			/* this is a new image block */
	cFalse,			/* this is new comp data */
	cFalse,			/* start line offsets over */
	cFalse,			/* hope to get to the end
				 * of the comp data in this
				 * call */
	&exhaustedImageData,
	&exhaustedCompData,
	&percentDone,
	&bytesWritten,
	NULL);			/* no array of line offsets */

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* Increment the number of bytes now in the buffer and byte
	 * pointer.
	 */
    tif->tif_rawcp += bytesWritten;
    tif->tif_rawcc += bytesWritten;

	/* Now execute a loop that keeps flushing compressed data and
	 * doing more compressing until the proper number of lines has
	 * been processed.
	 */
    while (!exhaustedImageData)
    {
	/* TIFFFlushData1 flushes the buffer of compressed data
	 * and resets tif->tif_rawcp and tif->tif_rawcc.  The
	 * -1 is weird, but that's what the original code did.
	 */
	if (!TIFFFlushData1(tif))
	{
	    return (-1);
	}

	remainingBytesInBuf = tif->tif_rawdatasize;
	bytesWritten = 0;	/* just in case compressRowsPackBits doesn't
				 * do anything for some reason */

	retCode = w_compressRowsPackBits(
			    pIPWState,
			    (void *)(tif->tif_data),
			    (PIXR *) prS,
			    TIFFCurrentRow(tif),
			    numLines,
			    tif->tif_rawcp,
			    remainingBytesInBuf,
			    cTrue,		/* this is an old image block */
			    cTrue,		/* this is old comp data */
			    cFalse,		/* don't care about
						 * line offsets */
			    cFalse,		/* hope to get to the end
						 * of the comp data in this
						 * call */
			    &exhaustedImageData,
			    &exhaustedCompData,
			    &percentDone,
			    &bytesWritten,
			    NULL);

	if (retCode != cSuccessful)
	{
	    return (0);	/* something failed */
	}
    
	    /* Increment the number of bytes now in the buffer and byte
	     * pointer.
	     */
	tif->tif_rawcp += bytesWritten;
	tif->tif_rawcc += bytesWritten;
    }

 
	/* if it appears to have worked so return happily.  We don`t
	 * check to see if we ran out of image lines since we wouldn`t
	 * have exited the loop above if there were more image lines.
	 * We don't check to see if we've run out of room on the
	 * comp data buffer since there's no reason to think that this
	 * would have happened on compression.
	 */
    return (1);	/* worked */
}


/*********************************************************************/
/************************** DECODING CODE ****************************/
/*********************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Prepare the state used to decode PackBits blocks.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
packBitsPreDecode (
    TIFF               *tif)
{
    void    *pPBDecompInfo;         /* pointer to opaque data type */
    UInt32   stateBlockSize;        /* size of opaque data block */
    Int32    retCode;
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;

	/* If we're not going to be given a PIXR to encode,
	 * just return successfully since there's nothing to
	 * set up.
	 */
    if (!pixrProvided(tif))
	return (1);

	/* See if the state block already exists.  The tif_data
	 * entry is cleared when the TIFF structure is originally
	 * allocated.
	 */
    pPBDecompInfo = (void *)tif->tif_data;
    if (pPBDecompInfo != NULL)
	return (1);		/* nothing else to do */


	/* We need a new state block.  Start by determining the
	 * size of the state block */
    retCode = w_getSizeOfPackBitsDecompressionInfo(pIPWState, &stateBlockSize);

    if (retCode != cSuccessful)
    {
	return (0);		/* failed */
    }

	/* Allocate the state block */
    pPBDecompInfo = (void *)_TIFFmalloc(tif, stateBlockSize);

    if (pPBDecompInfo == NULL)
    {
	return (0);		/* no memory for this block */
    }

	/* initialize the state block */
    retCode = w_initializePackBitsDecompressionInfo(pIPWState, pPBDecompInfo,
						  stateBlockSize);

    if (retCode != cSuccessful)
    {
	_TIFFfree(tif, pPBDecompInfo);
	return (0);		/* didn't work */
    }

	/* it worked, so put a pointer to the state block into the
	 * tif space reserved for it.
	 */
    tif->tif_data = (tidata_t) pPBDecompInfo;


    return (1);			/* it worked */
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Decode rows of a PIXR from a stream of PackBits data.
 *
 * Input:::
 *  tif::            pointer to TIFF compression state block
 *  uncompBytes::    number of uncompressed bytes this routine
 *                   is expected to produce.  We assume this
 *                   is a multiple of the number of bytes in
 *                   an image row.
 *  sampleNum::      number of current pixel sample.
 *                   This is apparently used when different
 *                   samples are stored in different image planes.
 *                   It is not used here.
 *
 * Output:::
 *  prD::            pointer to PIXR that will hold decompressed
 *                   data.  This is sort of a cheat in that
 *                   the Leffler convention is that this is a
 *                   pointer to a data buffer.  But, hey, this
 *                   is C, and all pointers are cousins, right?
 *
 * Returns:::
 *      Returns (following Leffler conventions!):
 *              1       it worked
 *              0       it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 *  Non-standard parameter order!:: 
 *       The output parameter is not the last in the param order
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static int
packBitsDecode (
    TIFF               *tif,
    tidata_t            prD,
    tsize_t             uncompBytes,
    tsample_t           sampleNum)
{
Int32	firstBlock;		/* for testing only */
UInt32	exhaustedImageData;	/* if set on return from
				 * decompressRowsPackBits, we've processed the
				 * correct number of lines.
				 */
UInt32	exhaustedCompData;	/* if set on return from
				 * decompressRowsPackBits, we've reached the
				 * end of the compressed data
				 */
UInt32	percentDone;		/* if we ran out of compressed data before
				 * completing the image, this tells how
				 * much of the image has been decompressed.
				 */
UInt32	bytesRead;		/* number of compressed bytes read on
				 * each call. */
Int32	retCode;
    IPWState *pIPWState = NULL;

    pIPWState = (IPWState *)tif->tif_ipwstate;


    sampleNum = sampleNum;	/* keep the compilers happy */

	/* If no PIXR was provided, we must be decompressing
	 * into a straight buffer.  The totally weird RGB code
	 * does this.  When it happens, just do the dumbest
	 * decoding possible.
	 * Note that all the scan lines abut in the buffer (there
	 * is no line padding), so we can just keep decoding until
	 * we decompress uncomp bytes.
	 */
    if (!pixrProvided(tif))
    {
	UInt8	*pD;			/* pointer to destination */
	UInt8	*pDEnd;			/* pointer to end of buffer + 1 */
	Int8	*pS;			/* pointer to source bytes */
	Int32	 count;			/* the count byte */

	    /* initialize the source and dest pointers */
	pD = (UInt8 *) prD;		/* ptr to buffer */
	pS = (Int8 *)tif->tif_rawcp;	/* ptr to source data */
	pDEnd = pD + uncompBytes;	/* point just beyond end of buffer */
	while (pD < pDEnd)
	{
	    count = *pS;		/* get the replica or literal count. */
	    pS++;			/* point to the first data byte
					 * following the count */
	    if (count >= 0)		/* is 0 <= count <= 127 */
	    {
		count++;		/* want to copy the next count+1
					 * bytes literally.
					 */
		memcpy((void *)pD, (void *)pS, count);
		pS += count;		/* skip those bytes in the source */
		pD += count;		/* skip same bytes in dest */
	    }
	    else	/* copy the next source bytes -count + 1 times */
	    if (count != -128)		/* -128 is a noop */
	    {
		count = -count + 1;	/* number of copies to make */
		memset((void *)pD, (UInt8)(*pS), count);
		pS++;			/* skip the replica byte in source */
		pD += count;		/* skip the expanded bytes in dest */
	    }
	}
	tif->tif_rawcc -= ((Int32)pS - (Int32)(tif->tif_rawcp));
	tif->tif_rawcp = (UInt8 *)pS;
	return (1);	/* we're done, it worked */
    }


	/* the following code executes in the normal case, i.e. when
	 * a PIXR is provided as the destination image instead of
	 * a weird buffer whose contents would have to be copied
	 * (and byteswapped on a little endian platform) to a PIXR.
	 */

#if 0
	/* TESTING ONLY
	 * do this in two parts to see if that works correctly */
    firstBlock = tif->tif_rawcc / 2;

    retCode = w_decompressRowsPackBits(
			pIPWState,
			(void *)(tif->tif_data), /* private decode data */
			TIFFCurrentRow(tif),	/* first dest line of pixr */
			uncompBytes / TIFFScanlineSize(tif), /* num lines
						 * to decompress */
			tif->tif_rawcp,		/* pointer to comp data */
			firstBlock,		/* number of comp data bytes */
			cFalse,			/* this is a new image block */
			cFalse,			/* this is new comp data */
			cFalse,			/* don't expect to get to
						 * the end of the comp data
						 * in this call */
			(PIXR *) prD,		/* dest pixr */
			&exhaustedImageData,
			&exhaustedCompData,
			&percentDone,
			&bytesRead);

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* if it appears to have worked, make sure we did all of
	 * image and ran out of compressed data too.
	 */
    if ((exhaustedImageData) || (!exhaustedCompData) ||
	(bytesRead != firstBlock))
    {
	return (0);	/* failed */
    }
#else
    firstBlock = 0;
#endif
	/* go decompress all the rows.  We figure out the number of lines
	 * to decompress by dividing uncompBytes by the TIFF package's
	 * idea of the number of bytes/line.  The information in the
	 * PIXR gives the real bytes/line.
	 */
    retCode = w_decompressRowsPackBits(
			pIPWState,
			(void *)(tif->tif_data), /* private decode data */
			TIFFCurrentRow(tif),	/* first dest line of pixr */
			uncompBytes / TIFFScanlineSize(tif), /* num lines
						 * to decompress */
			tif->tif_rawcp + firstBlock, /* pointer to comp data */
			tif->tif_rawcc - firstBlock, /* number of comp data
						 * bytes remaining */
#if 0
			cTrue,			/* this is an old image block */
			cTrue,			/* this is a continuation of
						 * the previous comp data */
#else
			cFalse,			/* this is a new image block */
			cFalse,			/* this is new comp data */
#endif
			cTrue,			/* expect to get to the end
						 * of the comp data in this
						 * call */
			(PIXR *) prD,		/* dest pixr */
			&exhaustedImageData,
			&exhaustedCompData,
			&percentDone,
			&bytesRead);

    if (retCode != cSuccessful)
    {
	return (0);	/* something failed */
    }

	/* Update the input pointers.  This is used if we're not
	 * reading the entire strip or tile in one shot.
	 */
    tif->tif_rawcc -= bytesRead;
    tif->tif_rawcp += bytesRead;

	/* if it appears to have worked, make sure we did all of
	 * image and ran out of compressed data too.
	 */
    if ((!exhaustedImageData) || (!exhaustedCompData))
	return (0);	/* failed */
    else
	return (1);	/* worked */
}



/*********************************************************************/
/************************** UTILITY CODE ****************************/
/*********************************************************************/

/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Tidy up after PackBits compression or decompression.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		void
 *
 * Notes:::
 *  Non-standard error return codes!::
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
static void
packBitsCleanup (
    TIFF               *tif)
{
    if (tif->tif_data != NULL)
    {
	_TIFFfree(tif, tif->tif_data);
	tif->tif_data = NULL;
    }
	/* Turn off the flags that say the encoder or decoder knows
	 * about PIXRs.
	 */
    tif->tif_flags &= ~(TIFF_PIXRDECODE | TIFF_PIXRENCODE);
}


/*****************************************************************************
 * This routine will not be recognized by the automatic documentation code.
 *
 * OneLineDescription:::
 *  Set up a bunch of procedure ptrs for doing compression and decompression.
 *
 * LongDescription:::
 *	This procedure sets up a bunch of procedure pointers so that
 *	PackBits compression and decompression is done.
 *
 * Output:::
 *  tif::            pointer to TIFF compression state block
 *
 * Returns:::
 *	Returns (following Leffler conventions!):
 *		1	it worked
 *		0	it failed somehow
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *       because routine is called from inside Leffler's TIFF code
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, Packbits
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 *****************************************************************************/
int
TIFFInitPackBits (
    TIFF*               tif)
{
    tif->tif_predecode   = packBitsPreDecode;
    tif->tif_decoderow   = packBitsDecode;
    tif->tif_decodestrip = packBitsDecode;
    tif->tif_decodetile  = packBitsDecode;

    tif->tif_preencode   = packBitsPreEncode;
    tif->tif_encoderow   = packBitsEncode;
    tif->tif_encodestrip = packBitsEncode;
    tif->tif_encodetile  = packBitsEncode;

    tif->tif_cleanup = packBitsCleanup;

	/* tell users that we're using decoding code that deals
	 * directly with PIXRs.
	 */
    tif->tif_flags |= TIFF_PIXRDECODE | TIFF_PIXRENCODE;
    return (1);
}


/*

 $Log: tiffpack.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.19  1998/01/27 01:57:20  ddavies
 tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v

 CommonCheck: Add tif arg to all invocations of _TIFFmalloc, _TIFFfree and _TIFFrealloc.
 These routines now use tif->ipwstate to get the IPState memory allocators
 instead of using malloc and free directly.

 Revision 1.18  1996/04/09 02:33:02  ddavies
 ddavies on Mon Apr 8 19:32:50 PDT 1996

 tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v

 Add cast to make MSVC4.0 happy.

 Revision 1.17  1996/04/09 01:33:34  ddavies
 ddavies on Mon Apr 8 18:31:34 PDT 1996

 tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v

 Fix bug in which the packbits decoder wasn't updating tif->tif_rawcp and
 tif->tif_rawcc after reading blocks.  This didn't matter if we were reading
 entire strips, but it mattered if we were reading one scan line at a time.
 Ron Danisewicz had a test program that duplicated the top line of each
 strip for all lines in the strip.

 Revision 1.16  1996/01/04 20:29:13  ejaquith
 tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.15  1995/11/15  02:26:54  ddavies
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Turn off all the autodocumentation of routines in this file because they're
 * in Leffler style, not IPWrap style.  Also rename the routines, removing the
 * wp_ prefix for the same reason.
 *
 * Revision 1.14  1995/10/18  04:20:04  ddavies
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Add casts to make Visual C++ happy.
 *
 * Revision 1.13  1995/10/16  23:00:59  dferrell
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.12  1995/10/13  22:07:58  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.11  1995/10/11  19:43:56  ddavies
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Check the new PIXR_PROVIDED flag.  When encoding, return an error if it's
 * not set.  When decoding, decode using the simple, brute force method if
 * we're decoding into a buffer instead of a PIXR.  This should only happen
 * with RGB images anyway.  Anyone who encodes an RGB image in a single plane
 * using a run length scheme like PackBits is doomed anyway.  The only way
 * you get a run is in large areas of uniform gray (NOT COLOR).
 *
 * Revision 1.10  1995/09/29  00:35:55  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.9  1995/09/12  16:55:22  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.8  1995/09/08  00:15:27  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.7  1995/09/01  02:21:03  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.6  1995/08/31  21:40:37  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/30  17:18:51  ejaquith
 * ejaquith on Wed Aug 30 10:05:08 PDT 1995
 *
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Bug Fix for target "file" in ipwtest.
 * Had accidentally (someone hit me w/ h20 balloon) overlooked that where
 * I added w_IPWStateInit() to routine packBitsDecode() was actually stuck
 * in an #if 0 so it never got inited, and the destruction routine didn't
 * check to make sure pIPWState wasn't NULL, and destroyed memory.
 * This is now fixed.  Further, I've added checks into all routines that
 * modify/access pIPWState to check if it's NULL in other routines/files.
 * Fixed ipTIFFCalloc() routine, which was using a call to _TIFFmalloc()
 * (there is no _TIFFcalloc() routine in the TIFF library), to afterwards
 * do a memset in the allocated block to 0s.
 * Fixed ipTIFFMalloc() routine to actually b
 * Fixed ipTIFFMalloc() routine forward declaration to be ipTIFFMalloc()
 * and not ipDefaultMalloc()  (previous copy/paste thinko).
 *
 * Revision 1.4  1995/08/22  02:23:26  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Added calls to w_IPWStateTerm() in all routines that were using
 * w_IPWStateInit() to clean memory.
 * q
 *
 * Revision 1.3  1995/08/21  17:33:12  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * If 0ed out tiffIPWStateInit() routine in favor of using newer more flexible
 * version of w_IPWStateInit() and just passing in the special ipTIFF** memory
 * management routine pointers.  Updated file to accomodate these changes.
 * Now allocation of the IP structure itself is taken care of in IPWStateIni()
 * using our passed in allocators.
 *
 * Revision 1.2  1995/08/17  22:01:54  ejaquith
 * tiffpack.c -> /product/ipcore/ipwrap/src/RCS/tiffpack.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.1  1995/05/09  03:18:40  ddavies
 * Initial revision
 *

 */
