/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  tiffio.c: routines for reading TIFF files, producing connected
 *		components directly.
 *
 *		       
 *	Includes procedures:
 *	w_readTiffToCC()	Convert TIFF file to connected components
 *	w_runsFromTiffRow()	Decode one row of MMR rows into runs
 *	wp_startTiffToCCStrip()	Initialize at start of strip and decode
 *				one row.
 *		      
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#ifdef __WATCOMC__
#include <io.h>
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "tiff.h"
#include "tiffio.h"
#include "tiffiop.h"
#include "alplib.h"
#include "defines.pub"
#include "shrcc.pub"
#include "shrcc.prv"
#include "shrpixr.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* prototypes */
#include "props.pub"
#include "shrcmprs.pub"
#include "shrcmprs.prv"
#include "ccitt.pub"
#include "pcconv.pub"
#include "fill.pub"
#include "io.pub"
#include "iotiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: tifftocc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");

    /* This is the structure passed to the glue routine w_runsFromTiffRow.
     * It contains a pointer to the Leffler TIFF data structure,
     * parameters we used to figure out when to fill the strip with
     * compressed data and pointers to TCScanLineRuns for the
     * reference lines.  We maintain two reference lines, one
     * for a hypothetical blank line that's used at the top of each
     * strip and one for the last line decoded.  That's the
     * one we return.
     */
typedef struct TiffToCCStateDecl
{
    TIFF		*tif;
    Int32		 width;
    Int32		 curStrip;		/* number of current strip */
    Int32		 rowsLeftInStrip;	/* number of rows left in
						 * current strip */
    Int32		 rowsInLastStrip;	/* number of rows in last
						 * strip */
    TCScanLineRun	*pBlankRefLine;		/* ptr to blank reference
						 * line */
    TCScanLineRun	*pRefLine;		/* ptr to last real reference
						 * line */
    TCScanLineRun	*pRefLineEnd;		/* ptr to first dummy run at
						 * end of reference line */
} TiffToCCState;


/*****************************************************************************
 * BeginDocumentation::
 *
 * OneLineDescription:::
 *  Start a new strip when converting TIFF G4 to connected components.
 *
 * LongDescription:::
 *   When converting a TIFF file directly to connected components, we
 *   need special code to initialize things at the beginning of each
 *   TIFF strip..
 *
 * Input:::
 *  pIPWState::	      pointer to IPWState.
 *  tif::             ptr to Leffler state struct.
 *  pCCITTDecompInfo  ptr to CCITTDecompressionInfo struct
 *  ppCompData::      Ptr to ptr to next compressed byte.
 *  pRunHeap::	      Ptr to heap from which TCScanLineRuns are allocated.
 *  ppDummy::	      Ptr to ptr to TCScanLineRun after the last valid
 *		      TCScanLineRun in the list of runs found.
 *
 * Notes:::
 *  This is called both to set the first strip and all subsequent strips.
 *  The first strip is setup in w_readTiffToCC.  The other strips are
 *  set up from w_runsFromTiffRow.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, CCITT, connected component analysis,
 *  run length, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation::
 *****************************************************************************/
static Int32 CDECL
wp_startTiffToCCStrip (
    IPWState	       *pIPWState,
    TiffToCCState      *pTiffToCCState,
    TIFF	       *tif,
    CCITTDecompressionInfo *pCCITTDecompInfo,
    UInt8	      **ppCompData,
    TCBlockHeader      *pRunHeap,
    TCScanLineRun     **ppDummy)
{
    PROLOGUE;
    UInt8	*pCurCompData;

	/* get a new strip */
    pTiffToCCState->curStrip++;

	/* Fill the strip */
    __error__ = TIFFFillStrip(tif, pTiffToCCState->curStrip);
    ERR_GEN(__error__ == 0, cErrIO_StripNotFilled);

	/* point to the new strip data */
    pCurCompData = tif->tif_rawcp;

	/* calculate the number of rows in this strip */
    if (pTiffToCCState->curStrip != tif->tif_dir.td_nstrips-1)
	pTiffToCCState->rowsLeftInStrip = tif->tif_dir.td_rowsperstrip;
    else
	pTiffToCCState->rowsLeftInStrip =
				pTiffToCCState->rowsInLastStrip;

	/* Initialize the coding and reference run pointers.
	 * Decode the new line using the blank reference line.
	 */
    pCCITTDecompInfo->decompState = cCCITTDecompNewLine;
    pCCITTDecompInfo->run2ArrayIsCoding = cFalse;
    pCCITTDecompInfo->currentRun = 0;
    pCCITTDecompInfo->refRun = 0;
    pCCITTDecompInfo->accumBits = 0;
    pCCITTDecompInfo->vacantBit = 0;
    pCCITTDecompInfo->remainingLength = 0;

	/* Get a new line using blank reference line */
    ERR_PASS(ip_decompMMRRowList (pIPWState, pCCITTDecompInfo,
				  pTiffToCCState->width,
				  &pCurCompData,
				  pCurCompData + tif->tif_rawcc,
				  pRunHeap,
				  pTiffToCCState->pRefLine, /* coding */
				  pTiffToCCState->pBlankRefLine,
				  ppDummy));

    *ppCompData = pCurCompData;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::
 *
 * OneLineDescription:::
 *  Glue between rabbRunsLite and TIFFReadScanline.
 *
 * LongDescription:::
 *   When converting a TIFF file directly to connected components, we
 *   use rabbRunsLite to assemble the connected components.  It calls
 *   This procedure to get the runs for each scan line.  This procedure
 *   in turn calls TIFFReadScanline to decode one scan line, producing
 *   an array of runs.  i_getMostRecentRuns returns a pointer to the
 *   first of these runs.  We convert the array of runs to a linked
 *   list of runs and return the result to rabbRunsLite.
 *
 * Input:::
 *  pIPWState	      ptr to IPCore state
 *  pStateIn::        ptr to Leffler state struct.
 *  pFirst::          Ptr to first TCScanLineRun to be filled in.
 *  pRunHeap::	      Ptr to heap from which TCScanLineRuns are allocated.
 *
 * Notes:::
 *  This routine follows the style of i_findAllRunsInLine in the way it
 *  fills in the linked list of TCScanLineRuns.
 *
 *  This callback routine doesn't follow the IPWrap conventions.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, CCITT, connected component analysis,
 *  run length, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation::
 *****************************************************************************/
static Int32 CDECL
w_runsFromTiffRow (
    IPWState	       *pIPWState,
    void               *pStateIn,
    TCScanLineRun      *pFirst,
    TCBlockHeader      *pRunHeap)
{
    PROLOGUE;
    TiffToCCState      *pTiffToCCState;
    TIFF	       *tif;
    CCITTDecompressionInfo *pCCITTDecompInfo;
    TCScanLineRun      *pDummy;
    UInt8	       *pCurCompData;	/* current compressed data
					 * pointer */


	/* put the pointer to Leffler's struct into a useful form. */
    pTiffToCCState = (TiffToCCState *)pStateIn;
    tif = pTiffToCCState->tif;
    pCCITTDecompInfo = (CCITTDecompressionInfo *) (tif->tif_data);

	/* We have to return a list of runs each time we're called.
	 * The calling routine will take the list apart, so we
	 * can't use the list after returning it.  Therefore, we can
	 * safely only return the runs for the previous reference
	 * line.  We do this now by setting the contents of
	 * *pFirst to the contents of the first reference run.
	 * Since we'll need a new coding run, we use the first
	 * reference run for that, after setting its pNext pointer
	 * to the list that used to follow *pFirst.
	 */
    pDummy = pFirst->pNext;			/* remember the rest of
						 * pFirst's list */
    *pFirst = *(pTiffToCCState->pRefLine);	/* copy contents of first
						 * run in old reference line.
						 */
    pTiffToCCState->pRefLine->pNext = pDummy;	/* use the rest of the
						 * incoming list for the
						 * new coding runs */

	/* Only read a new line if we're not already at the bottom of
	 * the image.
	 */
    if (tif->tif_row < tif->tif_dir.td_imagelength)
    {
	    /* If we're at the top of a strip, go get a new
	     * strip.  In any case, read decode the runs in a
	     * strip and convert them from an array to a linked
	     * list of runs.
	     */
	pCurCompData = tif->tif_rawcp;
	if (pTiffToCCState->rowsLeftInStrip == 0)
	{
		/* start a new strip */
	    ERR_PASS(wp_startTiffToCCStrip(pIPWState, pTiffToCCState,
					   tif, pCCITTDecompInfo,
					   &pCurCompData, pRunHeap,
					   &pDummy));
	}
	else
	{
		/* Get a new line using the real reference line */
	    ERR_PASS(ip_decompMMRRowList (pIPWState, pCCITTDecompInfo,
					  pTiffToCCState->width,
					  &pCurCompData,
					  pCurCompData + tif->tif_rawcc,
					  pRunHeap,
					  pTiffToCCState->pRefLine, /* coding */
					  pFirst,		 /* reference */
					  &pDummy));
	}

	    /* Fix up the tif data pointers and byte count.
	     */
	tif->tif_rawcc -= (pCurCompData - tif->tif_rawcp);
	tif->tif_rawcp = pCurCompData;
	tif->tif_row++;
	pTiffToCCState->rowsLeftInStrip--;

	    /* Change the markings of the dummy runs that we're about
	     * to return so they look like the markings that
	     * the caller wants.  If the reference line was blank,
	     * mark the runs in *pFirst instead since the values
	     * have been copied there.  In that case, we particularly
	     * do not want to touch pTiffToCCState->pRefLineEnd since
	     * it equals pTiffToCCState->pRefLine and we just put the first
	     * run for the new coding line there.
	     */
	if (pTiffToCCState->pRefLine != pTiffToCCState->pRefLineEnd)
	{
	    pTiffToCCState->pRefLineEnd->runLeft =
		pTiffToCCState->pRefLineEnd->runRight = cVeryLargeNum;
	    pTiffToCCState->pRefLineEnd->pClient = NULL;
	}
	else	/* reference line was blank */
	{
	    pFirst->runLeft = pFirst->runRight = cVeryLargeNum;
	    pFirst->pClient = NULL;
	}

	    /* Mark the two runs beyond the last run in the same
	     * way they were marked during encoding.  This guarantees
	     * that we'll get the same answer when we're searching
	     * through these runs when they become the reference row.
	     */
	pDummy->runLeft = (Int16) pTiffToCCState->width;
	pDummy->runRight = (Int16) pTiffToCCState->width-1;
	    /* If there is no next dummy run, make one */
	if (pDummy->pNext == NULL)
	{
	    ALLOCATE_STRUCT(pRunHeap, TCScanLineRun, pDummy->pNext);
	    ERR_GEN(__error__ != cSuccessful, cErrNomem);
	    pDummy->pNext->pNext = NULL;
	}
	pTiffToCCState->pRefLineEnd = pDummy;	/* remember where this is */
	pDummy = pDummy->pNext;
	pDummy->runLeft = (Int16) pTiffToCCState->width;
	pDummy->runRight = cVeryLargeNum;
    }
    else	/* still have to fix up the last dummy runs on last line */
    {
	pTiffToCCState->pRefLineEnd->runLeft =
	    pTiffToCCState->pRefLineEnd->runRight = cVeryLargeNum;
    }

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a run-length encoded TIFF image file into connected components.
 *
 * LongDescription:::
 *   Must be single channel, depth 1.
 *
 * Input:::
 *  pIPWState::      standard state structure.
 *  fname::          Ptr to character string containing name of file.
 *  connectType::    Connectivity type, 0 = 4-connected, 1 = 8-connected
 *  minWidth::       Minimum width of connected component to consider
 *  minHeight::      Minimum height of connected component to consider
 *  callbackProc::   Ptr to procedure that will be called for each connected
 *		     component found.
 *  pCallbackData::  Ptr to whatever state data is needed by callbackProc.
 *
 * Output:::
 *  pCallbackReturnCode:: Ptr to return code returned by callbackProc.
 *		     *pCallbackReturnCode is only set if callbackProc fails.
 *
 * Returns:::
 *   If the type of file is not a CCITT Group 4, this routine
 *   returns cErrIO_NotGroup4.
 *
 * Notes:::
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression, CCITT, connected component analysis,
 *  run length, symbol-based
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::  adapted from w_readTiff.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_readTiffToCC (
    IPWState           *pIPWState,
    char               *fname,
    UInt32              connectType,
    UInt32              minWidth,
    UInt32              minHeight,
    void               *pCallbackData,
    Int32	(CDECL *callbackProc)(IPWState *pIPWState,
				      void *, Int32 x, Int32 y,
				      Int32 w, Int32 h, Int32 m,
				      Int32 numRuns, TCScanLineRun *pRuns),
    Int32              *pCallbackReturnCode)
{
    PROLOGUE;
    FILE		*fp = NULL;
    FH			 FileH;  /* Put the file header on the stack */
    TIFF		*tif = NULL;		/* TIFF image */
    UInt16		 samplesperpixel;
    UInt16		 bitspersample;
    UInt16		 compression;
    CCState		*pRabbRunsState = NULL;
    Int32		 imWidth, imHeight;
    Int32		 runRetCode = 0;
    UInt32		 stateBlockSize;        /* size of opaque data block */
    TiffToCCState	*pTiffToCCState = NULL;
    CCITTDecompressionInfo *pCCITTDecompInfo = NULL;
    TCScanLineRun	*pRefRun;
    UInt8	        *pCurCompData;	/* current compressed data
					 * pointer */

    /* Clear out the file header struct */
    memset((void *) &FileH, 0, sizeof(FH));
    FileH.filename = NULL;
    FileH.header = NULL;

	/* save the filename */
    ERR_PASS( w_strcatSafe(pIPWState, fname, "", &(FileH.filename)));

	/* open the file */
    fp = fopen( FileH.filename, "rb");
    ERR_GEN_CLEANUP(!fp, cErrIO_ReadError);

	/* Get the file type, complain if it's not a TIFF file */
    ERR_PASS_CLEANUP( w_ffindFormat(pIPWState, fp, &(FileH.type)));
    ERR_GEN_CLEANUP(FileH.type != TIFF_DEFAULT, cErrIO_NotGroup4);

	/* Create the tif data structure */
    tif = TIFFOpen((void *)pIPWState, FileH.filename, "r");
    ERR_GEN_CLEANUP(tif == NULL, cErrIO_ReadError);

	/* Lower level routines need the IPState, so pass it inside the
	 * new tif data structure.
	 */
    tif->tif_ipwstate = (void *)pIPWState;

	/* Make sure this is a binary file with one plane */
    TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
    TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);

    ERR_GEN_CLEANUP((samplesperpixel != 1)  || (bitspersample != 1),
		    cErrTypeNotSupported);

	/* Bail out on tiled images.  Yeah, images with one tile would
	 * work - fix this if it becomes important.
	 */
    ERR_GEN_CLEANUP(TIFFIsTiled(tif), cErrTypeNotSupported);

	/* Quit if this isn't the right type of compression */
    TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression);
    ERR_GEN_CLEANUP((compression != COMPRESSION_CCITTFAX4),
		    cErrIO_NotGroup4);

	/* We need the image dimensions */
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &imWidth);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &imHeight);

	/* Don't have TIFFFillStrip mess with predecode.
	 * However, we do want the new-style decompression structure to
	 * be created no matter what TIFF library is linked,
	 * so do it here.  This structure is used by
	 * ip_decompMMRRow.
	 */
    tif->tif_predecode = NULL;
    ERR_PASS_CLEANUP(w_getSizeOfCCITTDecompressionInfo(pIPWState, imWidth,
						       &stateBlockSize));
    pCCITTDecompInfo = (void *)IPW_MALLOC(stateBlockSize, cAllocatorData);
    ERR_GEN_CLEANUP(pCCITTDecompInfo == NULL, cErrNomem);

    ERR_PASS_CLEANUP(w_initializeCCITTDecompressionInfo(pIPWState,
							pCCITTDecompInfo,
							imWidth,
							cFalse,	/* useMR2D */
							cNoPad, /* padType */
							stateBlockSize));
	/* put a pointer to the decompression block into the TIFF
	 * structure so w_runsFromTiffRow can find it.
	 */
    tif->tif_data = (tidata_t) pCCITTDecompInfo;
    tif->tif_flags &= ~TIFF_NOBITREV;	/* we do not handle bit reversal */

	/* Now we're almost ready to really start reading the scan lines.
	 * We want to do this via w_rabbRunsLite, which will call
	 * w_runsFromTiffRow for each scan line.  That procedure
	 * will call ip_decompMMRRow to translate the encoded runs
	 * into a linked list of runs.  This list will then be
	 * returned to w_rabbRunsLite.  This process continues
	 * until we're done.
	 * But first, allocate the state structure needed by
	 * w_rabbRunsLite and w_runsFromTiffRow.
	 */
	/* Create the state used by rabbRunsLite */
    ERR_PASS_CLEANUP(w_createRabbState(pIPWState, (void *)&pRabbRunsState));

	/* Initialize the private data structure that
	 * w_runsFromTiffRow uses.
	 */
    pTiffToCCState = (TiffToCCState *)
		      IPW_MALLOC(sizeof(TiffToCCState), cAllocatorData);
    ERR_GEN_CLEANUP(pTiffToCCState == NULL, cErrNomem);

    pTiffToCCState->tif = tif;
    pTiffToCCState->width = imWidth;
    pTiffToCCState->curStrip = -1;		/* will be increased to 0 */
    pTiffToCCState->rowsLeftInStrip = 0;	/* start with a new strip */
    pTiffToCCState->rowsInLastStrip =
	tif->tif_dir.td_imagelength % tif->tif_dir.td_rowsperstrip;
    if (pTiffToCCState->rowsInLastStrip == 0)
	pTiffToCCState->rowsInLastStrip = tif->tif_dir.td_rowsperstrip;

	/* There is no reference line.  However, we'll need a starter
	 * to create the first reference line with wp_startTiffToCCStrip
	 * below.
	 */
    ALLOCATE_STRUCT(pRabbRunsState->pRunHeap, TCScanLineRun,
			pTiffToCCState->pRefLine);
    ERR_GEN_CLEANUP(__error__ != cSuccessful, cErrNomem);
    pTiffToCCState->pRefLine->pNext = NULL;
    pTiffToCCState->pRefLine->pClient = NULL;

	/* Allocate the first dummy structure in the special
	 * blank reference line */
    ALLOCATE_STRUCT(pRabbRunsState->pRunHeap, TCScanLineRun, pRefRun);
    ERR_GEN_CLEANUP(__error__ != cSuccessful, cErrNomem);

    pRefRun->runLeft = (Int16) imWidth;
    pRefRun->runRight = (Int16) imWidth - 1;
    pRefRun->pClient = NULL;

	/* Allocate the second dummy structure in the blank reference line */
    ALLOCATE_STRUCT(pRabbRunsState->pRunHeap, TCScanLineRun, pRefRun->pNext);
    ERR_GEN_CLEANUP(__error__ != cSuccessful, cErrNomem);

    pRefRun->pNext->runLeft = (Int16) imWidth;
    pRefRun->pNext->runRight = cVeryLargeNum;
    pRefRun->pNext->pClient = NULL;
    pRefRun->pNext->pNext = NULL;
	/* Make this the blank reference run. */
    pTiffToCCState->pBlankRefLine = pRefRun;

	/* Fill the reference line with the runs from the
	 * first image line so that w_runsFromTiffRow
	 * can always assume that there's a valid reference line.
	 */
    ERR_PASS_CLEANUP(wp_startTiffToCCStrip(pIPWState, pTiffToCCState,
					   tif, pCCITTDecompInfo,
					   &pCurCompData,
					   pRabbRunsState->pRunHeap,
					   &pRefRun));

	/* Fix up the tif data pointers and byte count.
	 */
    tif->tif_rawcc -= (pCurCompData - tif->tif_rawcp);
    tif->tif_rawcp = pCurCompData;
    tif->tif_row++;
    pTiffToCCState->rowsLeftInStrip--;

	/* Mark the two runs beyond the last run in the same
	 * way they were marked during encoding.  This guarantees
	 * that we'll get the same answer when we're searching
	 * through these runs when they become the reference row.
	 */
    pRefRun->runLeft = (Int16) imWidth;
    pRefRun->runRight = (Int16) imWidth-1;
	/* If there is no next dummy run, make one */
    if (pRefRun->pNext == NULL)
    {
	ALLOCATE_STRUCT(pRabbRunsState->pRunHeap, TCScanLineRun,
			pRefRun->pNext);
	ERR_GEN(__error__ != cSuccessful, cErrNomem);
	pRefRun->pNext->pNext = NULL;
    }
    pTiffToCCState->pRefLineEnd = pRefRun;	/* remember where this is */
    pRefRun = pRefRun->pNext;
    pRefRun->runLeft = (Int16) imWidth;
    pRefRun->runRight = cVeryLargeNum;


	/* FINALLY, all the preparation is over.  We're ready
	 * to get the connected components.  Ask w_rabbRunsLite
	 * to get all the connected components and call
	 * the callback proc for each one.
	 */
    ERR_PASS_CLEANUP(w_rabbRunsLite(pIPWState,
				    NULL,		/* ptr to image */
				    (void *)pRabbRunsState,
				    connectType,
				    0, 0,		/* clipX, clipY */
				    imWidth, imHeight,	/* clipW, clipH */
				    minWidth,		/* minWidth */
				    minHeight,		/* minHeight */
				    (void *)pTiffToCCState, /* pRunData */
				    &(w_runsFromTiffRow), /* runFindingProc */
				    pCallbackData,
				    callbackProc,
				    &runRetCode,
				    pCallbackReturnCode));

	/* Yea, it worked!  Just free everything and return */
    IPW_FREE(FileH.filename, cAllocatorData);
    IPW_FREE(FileH.header, cAllocatorData);
    IPW_FREE(pCCITTDecompInfo, cAllocatorData);
	/* don't try to free CCITTDecompressionInfo again */
    tif->tif_data = NULL;	
    IPW_FREE(pTiffToCCState, cAllocatorData);
	/* This cleans up all the runs allocated with ALLOCATE_STRUCT */
    ERR_PASS_CLEANUP(w_destroyRabbState(pIPWState, (void *)&pRabbRunsState));
    TIFFClose(tif);
    fclose(fp);

    NORMAL_RETURN;

	/* Error cleanup */
ERR_CLEANUP:
    IPW_FREE(FileH.filename, cAllocatorData);
    IPW_FREE(FileH.header, cAllocatorData);
    IPW_FREE(pCCITTDecompInfo, cAllocatorData);
    IPW_FREE(pTiffToCCState, cAllocatorData);
    if (pRabbRunsState != NULL)
	w_destroyRabbState(pIPWState, (void *)&pRabbRunsState);
    if (tif != NULL)
    {
	tif->tif_data = NULL;	/* don't free pCCITTDecompInfo again */
	TIFFClose(tif);
    }
    if (fp != NULL)
	fclose(fp);

    ERR_RETURN;
}


/*

 $Log: tifftocc.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.5  1998/01/14 00:16:15  ddavies
 tifftocc.c -> /product/ipcore/ipwrap/src/RCS/tifftocc.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.4  1997/07/11 01:59:19  ddavies
 tifftocc.c -> /product/ipcore/ipwrap/src/RCS/tifftocc.c,v

 CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
 type CDECL (*procPtrName) -> type (CDECL *procPtrName).

 Revision 1.3  1997/06/30 23:12:07  ddavies
 tifftocc.c -> /product/ipcore/ipwrap/src/RCS/tifftocc.c,v

 Delete inclusion of Sandpiper's internal.h

 Revision 1.2  1996/11/20 01:19:33  ddavies
 tifftocc.c -> /product/ipcore/ipwrap/src/RCS/tifftocc.c,v

 CommonCheck: Pass the TIFF structure containing a pointer to the IPWrap state structure
 to all routines that call TIFFWarning and TIFFError so that the TIFF code
 can use the normal error reporting mechanism.

 Revision 1.1  1996/05/31 20:56:05  ddavies
 Initial revision


 */


