/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  	tpdilate.c:
 *
 *  	    Topology Preserving Dilation
 *			w_topologyPreserveDilate()
 */

#include "ipwcore.prv"
#include "pixr.h"
#include "geomadt.h"
#include "logicops.h"
#include "defines.pub"
#include "hauscomp.h"

/* prototypes */
#include "hauscomp.pub"
#include "props.pub"
#include "shrpatt.pub"
#include "logicops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: tpdilate.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dilate a binary image, preserving its topology
 *
 * LongDescription:::
 *  Take a source image and a destination image and dilate the source into
 *  the dest.  Dilate every source pixel by a centered circle of radius 1.0
 *  (in other words, all four neighbors).
 *  Be conservative in dilating all four neighbors, minding the local topology
 *  of the image at each pixel.  Do not turn on neighbors that will "connect"
 *  pieces of the image together.
 *  <BR><BR>
 *  Typically this would find uses in Hausdorff-related applications as a
 *  substitute for normal 3x3 brick dilation when an image is of low
 *  resolution, poor quality, or a bad fax, and it is desirable to dilate
 *  tiny features as best as possible without filling in informative holes
 *  and gaps.  This routine uses a heuristic of rules to determine whether
 *  or not an "off" bit adjacent to an "on" bit should be turned on or not.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prS::            Ptr to source PIXR
 *
 * Output:::
 *  pPrD::           Ptr to ptr to dest PIXR.
 *                   If ptr is NULL, gets allocated and filled in here;
 *                   o.w. it uses the ptr to the PIXR, and fills it in.
 *
 * Returns:::
 *
 * Notes:::
 *  The source and dest PIXR may be the same.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, Hausdorff
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::  The underlying code was developed at PARC and ported here.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_topologyPreserveDilate (
    IPWState           *pIPWState,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    UInt8       *pImageD, *pImageS;
    Int32        fw, fh, bplD, bplS;
    Bool         allocatedHere = cFalse;
    Int32        rightFrameColsToClear, bottomFrameRowsToClear;
    PIXR        *prD = NULL;
    PIXR        *prTemp = NULL;

    ERR_GEN(pPrD == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1, cChannels1, cFrame32);

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS (wp_clonePixrConditionalNoInit(pIPWState, prS, &prD,
					    &rightFrameColsToClear,
					    &bottomFrameRowsToClear,
					    &allocatedHere));

	/*  if in-place, make a temporary image */
    if (prD == prS)
	ERR_PASS_CLEANUP( w_copy(pIPWState, prS, &prTemp));
    else
	prTemp = prD;


    ERR_PASS_CLEANUP( w_pixrFramedWidth(pIPWState, prS, &fw));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &fh));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pImageS));

    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prTemp, &bplD));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prTemp, &pImageD));

    ERR_PASS_CLEANUP( i_topologyPreserveDilate (pIPWState, pImageD, pImageS, 
					        fw, fh,
					        bplD, bplS));

	/*  if in-place, copy again and destroy the temporary image */
    if (prD == prS)
    {
	ERR_PASS_CLEANUP( w_logicOp (pIPWState, prTemp, COPY, prD));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prTemp));
    }

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* Hooray, it worked.  Return the dest pixr */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    if ((prD == prS) && (prTemp != NULL))
	w_destroyPixr(pIPWState, &prTemp);
    ERR_RETURN;
}


/*

 $Log: tpdilate.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.21  1998/01/14 00:16:16  ddavies
 tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.20  1997/04/15 22:49:01  ddavies
 tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v

 Use non-initialized PIXR creation where possible.

 Revision 1.19  1996/01/04 20:29:16  ejaquith
 tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.18  1995/12/13  04:15:56  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.17  1995/11/13  20:19:34  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.16  1995/10/30  18:06:18  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * Added inclusion of "hauscomp.h", which isn't needed, but it's too
 * late to abort checkin with our silly current scripts.
 *
 * Revision 1.15  1995/10/13  22:08:01  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.14  1995/09/29  00:35:09  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.13  1995/09/21  23:00:58  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.12  1995/09/18  23:50:14  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * Updated w_topologyPreserveDilate() to use the more standard convention
 * where client passes in ptr to ptr to dest image, which can be set (and
 * then checked in the routine), null (and created in the routine), or be
 * the same as the source image (an in-place operation).  It's the 90s.
 *
 * Revision 1.11  1995/09/12  16:54:50  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.10  1995/09/08  00:14:46  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.9  1995/09/07  01:37:42  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.8  1995/09/06  20:26:32  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.7  1995/09/01  02:20:36  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.6  1995/08/31  21:40:40  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/25  01:38:58  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * Fixed warnings
 *
 * Revision 1.4  1995/08/17  22:01:19  ejaquith
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.3  1995/03/29  02:07:28  ejaquith
 * ejaquith on Tue Mar 28 18:07:24 PST 1995
 *
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * Changed calls to use standard convention of FRAMED width and height of
 * images
 *
 * Revision 1.2  1995/02/02  01:14:32  ejaquith
 * ejaquith on Wed Feb 1 17:13:57 PST 1995
 *
 * tpdilate.c -> /product/ipcore/ipwrap/src/RCS/tpdilate.c,v
 *
 * Changed order of PIXRs in topologyPreserveDilate()
 * to more conventional order of dest,src
 *
 * Revision 1.1  1995/02/01  21:55:24  ejaquith
 * Initial revision
 *

 */
