/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  utils1.c:
 *      Includes
 *
 *           IPWState init/accessor/mutator functions
 *               w_IPWStateInit ()
 *               w_IPWStateSetAllocator ()
 *               w_IPWStateSetErrorProc ()
 *               w_IPWStateGetErrorProc ()
 *               w_IPWStateSetStrelInfo ()
 *               w_IPWStateGetStrelInfo ()
 *		 wp_IPWStateSetLongAccessFlag ()
 *		 wp_IPWStateGetLongAccessFlag ()
 *		 wp_IPWStateSetPurifyRunning ()
 *		 wp_IPWStateGetPurifyRunning ()
 *
 *		 w_printErrorInfo ()
 *		 w_printf ()
 *
 */

#ifdef VXWORKS
// Require definition of cpu type for stdarg to work... / dhtesq.
#include <stdlib.h>
#endif

#include <stdarg.h>
#include <stdio.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"
#include "defines.pub"
#include "pixr.h"
#include "morphg.h"

#include "strel.pub"
#include "shrutils.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif





IP_RCSINFO(RCSInfo, "$RCSfile: utils1.c,v $; $Revision: 1.2 $; $Date: 2000/03/24 22:09:55 $");

/*-------------- Default allocators: just call malloc, etc. --------------*/


/*-------------------------IPWState functions------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Allocate and initialize an IPWState structure.
 *
 * Input:::
 *  imageMalloc::    	proc called when allocating images.
 *                      If NULL, default to system call via ip_defaultMalloc().
 *  imageCalloc::    	proc called when allocating and clearing images
 *                      If NULL, default to system call via ip_defaultCalloc().
 *  imageRealloc::   	proc called when reallocating images
 *                      If NULL, default to system call via ip_defaultRealloc().
 *  imageFree::      	proc called when deallocating images
 *                      If NULL, default to system call via ip_defaultFree().
 *  imageInfo::      	client ptr given to memory routines when allocating
 *			images
 *  dataMalloc::     	proc called when allocating data
 *                      If NULL, default to system call via ip_defaultMalloc().
 *  dataCalloc::     	proc called when allocating and clearing data
 *                      If NULL, default to system call via ip_defaultCalloc().
 *  dataRealloc::    	proc called when reallocating data
 *                      If NULL, default to system call via ip_defaultRealloc()
 *  dataFree::       	proc called when deallocating for data
 *                      If NULL, default to system call via ip_defaultFree().
 *  dataInfo::       	client ptr given to memory routines when allocating
 *			data
 *  errorProc::    	client procedure called to print error messages.
 *			The type of this procedure is defined in
 *			ipshared/include/types.pub.
 *  errorInfo::		client ptr that's included in every call
 *			to errorProc.
 *  maxFilter::		A value in the IPWState structure is decremented each
 *			time ERR_PASS or ERR_PASS_CLEANUP is executed.  When
 *			the value reaches 0, it is reset to maxFilter and
 *			yieldProc is called.  This allows the client to
 *			abort processes without knowing about the OS.
 *  yieldProc::		The procedure called to yield control to the client.
 *			This takes pointer to void as its parameter.  The
 *			pointer is set to yieldInfo (below)
 *			as a means of identifying the executing process.
 *			If yieldProc == NULL, it will never be called.
 *  yieldInfo::		This client-supplied data is passed back by the
 *			yieldProc.
 *
 * Output:::
 *  ppIPWState::     	ipwrap state structure
 *
 * Returns:::
 *  cErrNomem		unable to allocate memroy
 *  cSuccessful		successfully initialized IPWState
 *
 * Notes:::
 * If any of the allocation/deallocation routines are null,
 * sets them to a hopefully reasonable default.
 * However,  errorStream or *Info the are left alone if null.
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, jcraig
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateInit (
    MallocType          imageMalloc,
    CallocType          imageCalloc,
    ReallocType         imageRealloc,
    FreeType            imageFree,
    void               *imageInfo,
    MallocType          dataMalloc,
    CallocType          dataCalloc,
    ReallocType         dataRealloc,
    FreeType            dataFree,
    void               *dataInfo,
    ClientErrorProc     errorProc,
    void	       *errorInfo,
    UInt32		maxFilter,
    ClientYieldProc	yieldProc,
    void	       *yieldInfo,
    IPWState          **ppIPWState)
{
    PROLOGUE;
    IPWState	*pIPWState = NULL;
    Int32	 stateStructSize;
    ErrCode	 retCode;

	/* NOTE: We can't start off using the procedure calling
	 * macros since they use pIPWState and that's not
	 * initialized yet!
	 */

    /* make sure that we're not causing a memory leak */
    if (*ppIPWState != NULL)
	return (cErrInvalidParm);

    /* allocate the space for the ipstate structure */
    retCode = i_getSizeOfIPState(&stateStructSize);
    if (retCode != cSuccessful)
	return retCode;

    if (dataCalloc != NULL)
	pIPWState =
		(IPWState *)dataCalloc(1, stateStructSize, cFalse, dataInfo);
    else pIPWState = (IPWState *)calloc(1, stateStructSize);
    if (pIPWState == NULL)
	return (cErrNomem);

	/* go fill in most of the state structure */
    retCode = i_IPStateInit (pIPWState, imageMalloc, imageCalloc,
			     imageRealloc, imageFree, imageInfo,
			     dataMalloc, dataCalloc, dataRealloc,
			     dataFree, dataInfo,
			     errorProc, errorInfo,
			     maxFilter, yieldProc, yieldInfo,
			     stateStructSize);
    if (retCode != cSuccessful)
	goto ERR_CLEANUP;

	/* At this point, enough of the IPWState is initialized to
	 * print error info, so we can use the procedure calling
	 * macros.
	 */

	/* The IPShared level doesn't know anything about strels, so
	 * initialize that here.
	 */
    ERR_PASS_CLEANUP( w_initStrelMgr(pIPWState));

    *ppIPWState = pIPWState;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pIPWState != NULL) dataFree((void *)pIPWState, 0, dataInfo);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Free an IPWState structure
 *
 * Input:::
 *  ppIPWState::     	pointer to pointer to ipwrap state structure
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateTerm (
    IPWState          **ppIPWState)
{
    PROLOGUE;
    IPWState 	*pIPWState = *ppIPWState;

    ERR_PASS( w_termStrelMgr(pIPWState));
    IPW_FREE(pIPWState, cAllocatorData);
    *ppIPWState = NULL;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets allocation procedures in pIPWState
 *
 * LongDescription:::
 * Sets the malloc, calloc, realloc, and free functions for
 * an IPWSTATE structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  allocator::      	are these allocators for data or images?
 *  pMalloc::        	proc called when allocating memory
 *  pCalloc::         	proc called when allocating and clearing memory
 *  pRealloc::       	proc called when reallocating memory
 *  pFree::          	proc called when deallocating memory
 *  allocInfo::      	client ptr given to memory routines
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateSetAllocator (
    IPWState           *pIPWState,
    Int32               allocator,
    MallocType          pMalloc,
    CallocType          pCalloc,
    ReallocType         pRealloc,
    FreeType            pFree,
    void               *allocInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN((allocator >= cAllocatorCount) || (allocator < 0),
	    cErrInvalidParm);

    ERR_PASS( i_IPStateSetAllocator( pIPWState, allocator, pMalloc,
				     pCalloc, pRealloc, pFree, allocInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the error stream in an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  errorProc::		Pointer to procedure used to print error messages.
 *  errorInfo::		Pointer to client data included in every call
 *			to errorProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateSetErrorProc (
    IPWState           *pIPWState,
    ClientErrorProc     errorProc,
    void	       *errorInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_PASS( i_IPStateSetErrorProc( pIPWState, errorProc, errorInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the error stream from an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *
 * Output:::
 *  pErrorProc::	Pointer to pointer to procedure used to print
 *			error messages.
 *  pErrorInfo::	Pointer to pointer to client data included in
 *			every call to errorProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateGetErrorProc (
    IPWState           *pIPWState,
    ClientErrorProc    *pErrorProc,
    void	      **pErrorInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pErrorProc == NULL || pErrorInfo == NULL,
	     cErrNullPtr);
    ERR_PASS( i_IPStateGetErrorProc( pIPWState, pErrorProc, pErrorInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the yield proc and filter in an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  maxFilter::		Value specifying how often yieldProc is called.
 *			yieldProc is called once for every maxFilter
 *			times ERR_PASS or ERR_PASS_CLEANUP is executed.
 *  yieldProc::		Pointer to procedure used to yield control back to
 *			the client.
 *  yieldInfo::		pointer to client info that is passed back as an
 *			argument to yieldProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateSetYieldProc (
    IPWState           *pIPWState,
    UInt32		maxFilter,
    ClientYieldProc	yieldProc,
    void	       *yieldInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_PASS( i_IPStateSetYieldProc(pIPWState, (Int32)maxFilter, yieldProc,
				    yieldInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the yield procedure and max filter value from an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *
 * Output:::
 *  pMaxFilter::	Pointer to returned maximum filter value.
 *			The yieldProc is called every maxFilter times
 *			ERR_PASS or ERR_PASS_CLEANUP is invoked.
 *  pYieldProc::	Pointer to pointer to yield procedure.
 *  pYieldInfo::	Pointer to client info that is passed back as
 *			an argument to yieldProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateGetYieldProc (
    IPWState           *pIPWState,
    UInt32	       *pMaxFilter,
    ClientYieldProc    *pYieldProc,
    void	      **pYieldInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pYieldProc == NULL || pMaxFilter == NULL,
	     cErrNullPtr);
    ERR_PASS( i_IPStateGetYieldProc(pIPWState, (Int32 *)pMaxFilter, pYieldProc,
				    pYieldInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set the StrelInfo structure for the IPWState structure.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  pStrelInfo::     	StrelInfo structure
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateSetStrelInfo (
    IPWState           *pIPWState,
    StrelInfo          *pStrelInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_PASS (i_IPStateSetStrelInfo (pIPWState, pStrelInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the StrelInfo structure for the IPWState structure.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *
 * Output:::
 *  pStrelInfo::     	StrelInfo structure
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_IPWStateGetStrelInfo (
    IPWState           *pIPWState,
    StrelInfo         **pStrelInfo)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_PASS (i_IPStateGetStrelInfo (pIPWState, pStrelInfo));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the enableLongAccesses flag in an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  enableLongAccesses:: Boolean set to cTrue if long (>= 64 bit) accesses
 *			 should be enabled.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IPWStateSetLongAccessFlag (
    IPWState           *pIPWState,
    Bool		enableLongAccesses)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_PASS (ip_IPStateSetLongAccessFlag (pIPWState, enableLongAccesses));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the purifyIsRunning flag from an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *
 * Output:::
 *  pEnableLongAccesses:: Pointer to returned Boolean that will hold the
 *			  contents of the enableLongAccesses flag.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IPWStateGetLongAccessFlag (
    IPWState           *pIPWState,
    Bool	       *pEnableLongAccesses)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pEnableLongAccesses == NULL,
	     cErrNullPtr);
    ERR_PASS (ip_IPStateGetLongAccessFlag (pIPWState, pEnableLongAccesses));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the purifyRunning flag in an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  purifyIsRunning::	Boolean set to cTrue if Purify is running, cFalse
 *			otherwise.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IPWStateSetPurifyIsRunning (
    IPWState           *pIPWState,
    Bool		purifyIsRunning)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_PASS (ip_IPStateSetPurifyIsRunning (pIPWState, purifyIsRunning));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the purifyIsRunning flag from an IPWState structure
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *
 * Output:::
 *  pPurifyIsRunning::	Pointer to returned Boolean that will hold the
 *			contents of the purifyIsRunning flag.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IPWStateGetPurifyIsRunning (
    IPWState           *pIPWState,
    Bool	       *pPurifyIsRunning)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pPurifyIsRunning == NULL,
	     cErrNullPtr);
    ERR_PASS (ip_IPStateGetPurifyIsRunning (pIPWState, pPurifyIsRunning));

    NORMAL_RETURN;
}




/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Print standard error message using client's error printing procedure
 *
 * LongDescription:::
 * Print an error message that contains the file name, line number and
 * error number.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  filename::       	name of file that generated/caught error
 *  line_num::       	line number in filename
 *  error::          	error number
 *
 * Returns:::
 *  cSuccessful		successfully printed error message
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_printErrorInfo (
    IPWState           *pIPWState,
    const char         *filename,
    Int32               line_num,
    ErrCode             error)
{
    PROLOGUE;
    char  errorString[100];

    if (pIPWState != NULL && pIPWState->errorProc != NULL)
    {
	sprintf(&(errorString[0]), "--IPWrap Error-- File:%s Line:%ld Error: %ld\n", filename, line_num, error);
	pIPWState->errorProc(pIPWState->errorInfo, &(errorString[0]));
    }

    NORMAL_RETURN;
}



/*
 $Log: utils1.c,v $
 Revision 1.2  2000/03/24 22:09:55  dtracy
 VxWorks-friendly modifications.  New targets 'sparc_vxworks' and 'debugsparc_vxworks' added to src Makefile.

 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.33  1998/01/21 22:10:26  ddavies
 utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v

 Moved w_matrix from utils1.c to strelio.c It's only used by the strel code
 and breaks a circular dependency between utils and strel.
 Convered the IPWState accessor routines into wrappers for the real
 (new) accessor routines in IPShared.  Added wp_IPWStateSetLongAccessFlag
 and wp_IPWStateGetLongAccessFlag.  Moved w_printErrorInfo so it's
 below the IPWState accessor routines instead of in the middle of them.

 Revision 1.32  1997/04/17 02:45:57  ddavies
 utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v

 Add yieldInfo arg to autodoc comments in w_IPWStateSetYieldProc and
 w_IPWStateGetYieldProc.

 Revision 1.31  1997/04/15 22:51:38  ddavies
 utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v

 Add yieldInfo parameter to w_IPWStateInit.  This is passed in by the client
 and supplied to yieldProc as a parameter.  Also make sure the purifyIsRunning
 flag is reset in w_IPWStateInit.
 Add new procedures, wp_IPWStateSetPurifyIsRunning and wp_IPWStateGetPurifyIsRunning.
 These are used to access the new purifyIsRunning flag in the ipstate.
 See comments in ipstate.pub.

 Revision 1.30  1997/02/26 03:53:43  ddavies
 utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v

 Add maxFilter and yieldProc parameters to w_IPWStateInit.  These are used to
 give the client periodic opportunities to abort processing.
 Also add new routines w_IPWStateSetYieldProc and w_IPWStateGetYieldProc.

 Revision 1.29  1996/05/21 03:38:45  ejaquith
 utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v

 wp_defaultCalloc(), wp_defaultFree(), and wp_defaultRealloc()
 did not have anything entered for their OneLineDescription:::
 tags.  Added a few words of wisdom.  Now cron auto-doc stuff is happy.

 Revision 1.28  1996/01/04 20:29:19  ejaquith
 utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.27  1995/12/13  04:16:00  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.26  1995/11/16  02:19:45  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * w_defaultMalloc() w_defaultCalloc() w_defaultFree() and w_defaultRealloc()
 * all become static private routines now (client's can use them as their
 * memory routines by passing in NULLs to w_IPWStateInit().)
 *
 * Revision 1.25  1995/11/13  20:19:37  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.24  1995/11/03  22:49:15  ddavies
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.23  1995/10/13  22:08:09  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.22  1995/10/07  03:22:20  ddavies
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Change w_IPWStateGetErrorStream and w_IPWStateSetErrorStream to
 * w_IPWStateGetErrorProc and w_IPWStateSetErrorProc respectively.  Debug and
 * error messages are now printed using a routine (and private data for that
 * routine) that's passed in by the client.  w_IPStateInit now takes the
 * errorProc and its errorInfo instead of an errorStream as well.
 * printErrorInfo now formulates a string from the formatted info passed
 * Ooops, start over.  It formats an error string from some boilerplat
 * text (-- IPWrap Error --) and the file and line number where the
 * error occurred, then asks pIPWState->errorProc to print that string.
 * Also added inclusions of stdarg.h, stdio.h and ansiprot.h.  All this
 * nonsense allows us to print error/debug info on advanced systems like
 * Win95 and NT where fprintf(stderr, ...) doesn't work.
 *
 * Revision 1.21  1995/10/02  00:19:14  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Fixed keywords typos
 *
 * Revision 1.20  1995/09/29  00:15:14  jcraig
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: fixup procedure comments
 *
 * Revision 1.19  1995/09/23  00:52:00  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Removed double comment blocks and re-organized allocation/freeing routines
 * to be useable and sensible with auto-doc system.
 *
 * Revision 1.18  1995/09/21  23:01:04  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/07  01:05:12  jcraig
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * change ipDefault allocation routines to w_default
 *
 * Revision 1.16  1995/09/06  20:26:38  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.15  1995/09/01  02:16:58  jcraig
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.14  1995/08/31  21:40:45  ejaquith
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.13  1995/08/30  17:36:18  ejaquith
 * ejaquith on Wed Aug 30 10:34:49 PDT 1995
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Made routines accessing pIPWState to first check to ensure it's not NULL.
 * If it is, they return cErrNullIPWState.
 * Added error handling/passing wrappers to routine invocations that were
 * previously missing them.
 *
 * Revision 1.12  1995/08/22  18:26:57  jcraig
 * jcraig on Tue Aug 22 11:26:45 PDT 1995
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * calloc the ipstate structure instead of malloc'ing
 *
 * Revision 1.11  1995/08/18  21:44:38  jcraig
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * IPWStateInit now takes all info and allocates the ipstate structure
 * using the data allocator.  also inits strel manager.
 * added IPWStateTerm which frees the ipstate structure and terminates
 * the strel manager.
 *
 * Revision 1.10  1995/08/17  22:33:02  jcraig
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.9  1995/08/16  18:21:57  ejaquith
 * ejaquith on Wed Aug 16 11:20:27 PDT 1995
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Added printErrorInfo() routine for use with Error macros.
 * This will first check if pIPWState and then if the errorStream field
 * are valid, and if so, dump a warning message to the file stream.
 *
 * Revision 1.8  1995/08/15  01:50:12  jcraig
 * jcraig on Mon Aug 14 18:49:12 PDT 1995
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * first pass of macro changes
 *
 * Revision 1.7  1995/08/10  21:52:42  macleod
 * macleod on Thu Aug 10 14:52:13 PDT 1995
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Added routines to deal with the IPWState structure,
 * also added some default allocators which use malloc, calloc, realloc,
 * and free.
 *
 * Revision 1.6  1995/07/19  19:57:36  ddavies
 * ddavies on Wed Jul 19 12:56:52 PDT 1995
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Add const to _fatal's string pointer arg to keep gcc happy.
 *
 * Revision 1.5  1994/09/28  18:31:45  lperry
 * lperry on Wed Sep 28 11:31:06 PDT 1994
 *
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Update to use utils.pub header file. Remove extraneous
 * includes.
 *
 * Revision 1.4  1994/03/29  01:40:05  ddavies
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Changed declarations for _abortI, _abortP and _abortF so they take a
 * "const char * message" instead of "char * message".  This allows them to
 * use the output of iw_jerrorString.
 *
 * Revision 1.3  1993/12/03  01:20:26  ddavies
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:19:05  ddavies
 * utils1.c -> /product/ipcore/ipwrap/src/RCS/utils1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

