/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  utils2.c:
 *      Includes
 *
 *           string manipulation procs
 *               char   *fixDirectoryName()
 *               char   *strcatSafe()
 *
 *	     data plotting
 *		  void	 plotPoints()
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "iodef.h"
#include "defines.pub"
#include "plotdefs.h"

#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: utils2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");

/*
 *  fixDirectoryName:
 *        if slashFlag == True, append a
 *            trailing "/" if not already there;
 *        if slashFlag == False, remove
 *            any existing trailing "/".
 *        N.B.  Frees the input string and returns a new string in all cases.
 */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Remove or append a trailing directory separator 
 *
 * LongDescription:::
 * if slashFlag == True, append a trailing "/" or "\" (as appropriate) if
 * not already there.
 * if slashFlag == False, remove any existing trailing "/" or "\" (as
 * appropriate).
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure 
 *  directory::       	directory string
 *  slashFlag::       	should separator be appended or removed?
 *
 * Output:::
 *  fixedDirectory::  	new directory name
 *
 * Returns:::
 *  cErrInvalidParm	directory name is null
 *  cSuccessful		successful completion
 *
 * Notes:::
 * Frees the input string and returns a new string in all cases
 *
 * FunctionCategory:::
 *  string manipulation
 *
 * Keywords:::
 *  string manipulation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_fixDirectoryName (
    IPWState           *pIPWState,
    char               *directory,
    Int32               slashFlag,
    char              **fixedDirectory)
{
PROLOGUE;
Int32         n;
char          lastChar;
char         *newDirectory;

    ERR_GEN(directory == NULL, cErrInvalidParm);

    n = strlen(directory);
    lastChar = directory[n - 1];

    if ((lastChar == DIR_SEP_CHAR) && (slashFlag == cFalse))
    {
        ERR_PASS( w_strcatSafe(pIPWState, directory, "", &newDirectory));
        newDirectory[n - 1] = '\0';  /* remove it */
    }
    else if ((lastChar != DIR_SEP_CHAR) && (slashFlag == cTrue))
        ERR_PASS( w_strcatSafe(pIPWState, directory, DIR_SEP_STRING,
			    &newDirectory));   /* add it */
    else   /* don't do anything to the directory name */
        ERR_PASS( w_strcatSafe(pIPWState, directory, "", &newDirectory));

    *fixedDirectory = newDirectory;
    NORMAL_RETURN;
}


#define  EXTRA     8

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Concatenate two strings.
 *
 * LongDescription:::
 * takes char pointers for the two source strings.  creates space for the
 * strcat of the sources. does th strcat and returns the destination string
 * pointer adds EXTRA bytes to the string allocation.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  strS1::          	source string
 *  strS2::          	source string
 *
 * Output:::
 *  strDst::         	concatenation of strS1,,strS2
 *
 * Returns:::
 *  cErrNomem		unable to allocate memory
 *  cSuccessful		successfully concatenated strings
 *
 * FunctionCategory:::
 *  string manipulation
 *
 * Keywords:::
 *  string manipulation
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strcatSafe (
    IPWState           *pIPWState,
    const char         *strS1,
    const char         *strS2,
    char              **strDst)
{
PROLOGUE;
char       *strD;
Int32       len1, len2, len;

    if (strS1 == NULL)
	len1 = 0;
    else
	len1 = strlen(strS1);

    if (strS2 == NULL)
	len2 = 0;
    else
	len2 = strlen(strS2);

    len = len1 + len2 + EXTRA;

    strD = (char *)IPW_CALLOC(len, sizeof(char), cAllocatorData);
    ERR_GEN(strD == NULL, cErrNomem);

	/* assemble the final string */
    if (strS1 != NULL)
	strcpy(strD, strS1);
    if (strS2 != NULL)
	strcat(strD, strS2);
    *strDst = strD;
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Plot the given points.
 *
 * LongDescription:::
 *  This function does a graph of some points. There are two types
 *  of graphs supported, line graphs and impulse graphs. The 
 *  way this function works is to write a gnuplot script and then
 *  call gnuplot. The gnuplot script pauses for a newline.
 *
 * Input:::
 *  pIPWState::      	ipwrap state structure
 *  yPoints::		Ptr to the Y-axis points
 *  xPoints::        	If NULL, plot Y against (0:(count-1)),
 *			otherwise plot against these points
 *  dataType::       	If cPlotInt32, data points are Int32,
 *			if cPlotFloat64, they are Float64
 *  count::          	Number of data points
 *  plotType::       	cPlotLine or cPlotImpulse
 *
 * Returns:::
 *
 * Notes:::
 *  this won't work on a PC unless someone puts gnuplot on it.
 *
 * FunctionCategory:::
 *  debug
 *
 * Keywords:::
 *  debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_plotPoints (
    IPWState           *pIPWState,
    Int32              *yPoints,
    Int32              *xPoints,
    Int32               dataType,
    Int32               count,
    Int32               plotType)
{
    PROLOGUE;
    FILE *tmp;
    Float64 *yF, *xF;
    Int32 i;
    char dataFileName[L_tmpnam], scriptFileName[L_tmpnam], 
	 command[L_tmpnam + 10];


	/* This is only needed if datatype == cPlotFloat64, but
	 * knuckleheaded gcc complains that they might not be
	 * initialized if you put this inside a conditional.
	 */
    yF = (Float64 *) yPoints;
    xF = (Float64 *) xPoints;

    /* Write a data file */
    tmpnam (dataFileName);
    tmp = fopen (dataFileName, "w");
    ERR_GEN(tmp == NULL, cErrIO_WriteError);

    for (i = 0; i < count; i++)
    {
	if (dataType == cPlotFloat64)
	{
	    if (xPoints)
		fprintf(tmp, "%f %f\n", xF[i], yF[i]);
	    else
		fprintf(tmp, "%ld %f\n", i, yF[i]);
	}
	else
	{
	    fprintf(tmp, "%ld %ld\n", xPoints ? xPoints[i] : i, yPoints[i]);
	}
    }

    fclose (tmp);

    tmpnam(scriptFileName);
    tmp = fopen (scriptFileName, "w");
    ERR_GEN(tmp == NULL, cErrIO_WriteError);

    fprintf (tmp, "plot [0:%ld] '%s' with %s\n", count, dataFileName,
	     plotType == cPlotLine ? "lines" : "impulses");
    fprintf (tmp, "pause -1 \"Hit return when done...\"\n");
    fclose (tmp);

    sprintf (command, "gnuplot %s", scriptFileName);
    system (command);

    remove ((char *)scriptFileName);
    remove ((char *)dataFileName);

    NORMAL_RETURN;
}

/*

 $Log: utils2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.23  1996/01/04 20:29:22  ejaquith
 utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.22  1995/12/13  04:16:05  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/12/01  19:35:11  ddavies
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Bulletproof w_strcatSafe so that it treats NULL string pointers the same
 * as strings of length 0.
 *
 * Revision 1.20  1995/11/13  20:19:40  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:08:12  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/07  01:22:11  ddavies
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.17  1995/09/29  00:23:59  jcraig
 * jcraig on Thu Sep 28 17:23:39 PDT 1995
 *
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * change keyword and function category of w_plotPoints from plot to debug
 *
 * Revision 1.16  1995/09/29  00:15:16  jcraig
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: fixup procedure comments
 *
 * Revision 1.15  1995/09/21  23:01:08  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.14  1995/09/06  20:26:41  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.13  1995/09/01  02:17:06  jcraig
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * added back the minimum number of include files
 *
 * Revision 1.12  1995/08/31  21:40:47  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/25  01:39:02  ejaquith
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.10  1995/08/17  22:33:08  jcraig
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.9  1994/09/28  18:32:04  lperry
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Update to use utils.pub header file. Remove extraneous
 * includes.
 *
 * Revision 1.8  1994/05/05  02:00:30  ddavies
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Add "const" to "char *" parms of strcatSafe to make gcc happy.  Set xF and yF
 * unconditionally inside plotPoints even though the original code worked just
 * fine to keep gcc happy.  Change lots of %d's to %ld's.
 *
 * Revision 1.7  1994/02/12  01:28:03  dferrell
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Removed #defines of DIRSLASH and DIRSTRING.  Replaced use of these with
 * DIR_SEP_STRING and DIR_SEP_CHAR which are defined in iodef.h.
 * Added #include of iopros.h to get new constants above.
 * Removed unused #defines for home directory and strel directory.
 *
 * Revision 1.6  1994/01/22  01:44:18  ddavies
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * change "unlink" to "remove".  WATCOM doesn't know about unlink.
 *
 * Revision 1.5  1994/01/20  22:15:22  macleod
 * macleod on Thu Jan 20 14:14:41 PST 1994
 *
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * added "plotPoints", a curve-plotting utility which calls "gnuplot".
 * You have to have "gnu" enabled to use it.
 *
 * Revision 1.4  1994/01/14  23:22:48  dferrell
 * dferrell on Fri Jan 14 15:22:06 PST 1994
 *
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Fixed #ifdef/ifndef checking to work with Mac as well as Suns & PCs.
 *
 * Revision 1.3  1993/12/03  01:20:31  ddavies
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:19:13  ddavies
 * utils2.c -> /product/ipcore/ipwrap/src/RCS/utils2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

