/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * warp.c
 *
 *	Functions for arbitrary warping of images
 *
 * Functions defined:
 *
 *	w_rotateMag ()		- rotate and magnify an image
 *	w_quadTransform ()	- map a quadrilateral in the source
 *				  to a rect in the dest.
 *	w_warp ()		- arbitrarily warp an image.
 */


#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "warp.pub"
#include "shrwarp.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: warp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rotate and magnify a gray or binary image.
 *
 * LongDescription:::
 *  This routine allows a client to rotate and magnify an image.
 *  These are often useful functions and both can be accomplished by
 *  resampling and interpolating.  They were originally developed to
 *  help stitch image scans together.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be rotated and magnified.
 *  angle::          Rotation angle in radians.
 *  Mx::             Magnification factor along X axis
 *  My::             Magnification factor along Y axis
 *  srcXc::          X coord of displayed image center in source
 *  srcYc::          Y coord of displayed image center in source
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  If none
 *		     is supplied, one will be created.  If one is created,
 *		     its dimensions will be checked.  It should be
 *		     a clone of the source image.  However, it cannot
 *		     be the same as the source image.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if prS == *pPrD.  Also returns cErrInvalidParm if the X and Y resolutions
 *  of the input image are not equal.  That's because the transformation
 *  performed to an anisotropic image will probably look very strange when
 *  the image is scaled to be isotropic.
 *
 * Notes:::
 *  This is pretty general, but not blindingly fast.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, scaling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Perry
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rotateMag (
    IPWState           *pIPWState,
    PIXR               *prS,
    Float64             angle,
    Float64             Mx,
    Float64             My,
    Float64             srcXc,
    Float64             srcYc,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    Int32	 wS, hS, bplS, dS;  /* Width, height, bytesperline, depth */
    Int32	 wD, hD, bplD, dD;
    ScaledInt1616 xdpi, ydpi;
    UInt8	*pS, *pD;
    Bool	 allocatedHere = cFalse;
    PIXR *prSub;
    Int32 srcChannels, dstChannels;	/* Channel counts */
    Int32 channel;			/* Individual channel */

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1 | cChannels3, cFrame32);

    ERR_PASS( w_pixrGetResolutionDPI( pIPWState, prS, &xdpi, &ydpi));
    ERR_GEN(ABS(xdpi - ydpi) > 0x10000, cErrInvalidParm);

	/* If we already have a dest pixr, use its dimensions and
	 * depth and everything.  These determine the type of
	 * warping that's to be done.
	 * Otherwise, create one having the same dimensions as the
	 * source.
	 */
    prD = *pPrD;
    if (prD != NULL)
	ERR_GEN_PIXR( prD, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		      cChannels1 | cChannels3, cFrame32);
    else
    {
	ERR_PASS (w_clonePixrNoInit(pIPWState, prS, &prD));
	allocatedHere = cTrue;
    }

    ERR_PASS(w_pixrGetChannelCount(pIPWState, prS, &srcChannels));
    ERR_PASS(w_pixrGetChannelCount(pIPWState, prD, &dstChannels));
    ERR_GEN(srcChannels != dstChannels, cErrChannelsNotEqual);

    if (srcChannels > 1) {
	for (channel = 0; channel < srcChannels; channel++) {
	    prSub = NULL;
	    ERR_PASS(w_pixrGetChannel(pIPWState, prS, channel, &prSub));
	    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prSub, &wS));
	    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prSub, &hS));
	    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prSub, &dS));
	    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prSub, &pS));
	    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prSub, &bplS));
	    prSub = NULL;
	    ERR_PASS(w_pixrGetChannel(pIPWState, prD, channel, &prSub));
	    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prSub, &pD));
	    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prSub, &wD));
	    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prSub, &hD));
	    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prSub, &dD));
	    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prSub, &bplD));

	    /* The rectangle describing the image to be produced
	     * must be at least 2x2 pixels
	     */
	    ERR_GEN_CLEANUP(wD < 2  ||  hD < 2, cErrInvalidParm);

	    /* Do the recursive subdivision version of the resampler,
	     * with rotate/mag parameters.
	     */
	    ERR_PASS_CLEANUP( i_resample (pIPWState, pS, wS, hS, dS, bplS,
					  pD, wD, hD, dD, bplD,
					  angle, Mx, My, srcXc, srcYc,
					  (Float64)((wD - 1.0)/2.0),
					  (Float64)((hD - 1.0)/2.0)));
	} /* each channel */
    } else {	/* single channel: */

	ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prS, &wS));
	ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prS, &hS));
	ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &dS));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prS, &pS));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prD, &pD));
	ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prD, &wD));
	ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prD, &hD));
	ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prD, &dD));
	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

	/* The rectangle describing the image to be produced
	 * must be at least 2x2 pixels
	 */
	ERR_GEN_CLEANUP(wD < 2  ||  hD < 2, cErrInvalidParm);

	/* Do the recursive subdivision version of the resampler,
	 * with rotate/mag parameters.
	 */
	ERR_PASS_CLEANUP( i_resample (pIPWState, pS, wS, hS, dS, bplS,
				      pD, wD, hD, dD, bplD,
				      angle, Mx, My, srcXc, srcYc,
				      (Float64)((wD - 1.0)/2.0),
				      (Float64)((hD - 1.0)/2.0)));
    }

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Map an arbitrary source quadrilateral to a rectangular destination.
 *
 * LongDescription:::
 *  This is even more wonderful than w_rotateMag.  It is used when the
 *  client has clipped out a quadrilateral that represents a page in the
 *  source and wishes to make it into the real page in the destination.
 *  This function maps the pixels in the quadrilateral to the pixels
 *  in the destination rectangle, reducing or expanding as nececessary.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to source image.
 *  u0::             X coord of upper left corner of rectangle in dest.
 *  v0::             Y coord of upper left corner of rectangle in dest.
 *  nu::             Width of dest rectangle.
 *  nv::             Height of dest rectangle.
 *  xul::            X coord of upper left corner of source quadrilateral.
 *  yul::            Y coord of upper left corner of source quadrilateral.
 *  xur::            X coord of upper right corner of source quadrilateral.
 *  yur::            Y coord of upper right corner of source quadrilateral.
 *  xll::            X coord of lower left corner of source quadrilateral.
 *  yll::            Y coord of lower left corner of source quadrilateral.
 *  xlr::            X coord of lower right corner of source quadrilateral.
 *  ylr::            Y coord of lower right corner of source quadrilateral.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  If none
 *		     is supplied, one will be created.  If one is created,
 *		     its dimensions will be checked.  It should be
 *		     a clone of the source image.  However, it cannot
 *		     be the same as the source image.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if prS == *pPrD.  Also returns cErrInvalidParm if the X and Y resolutions
 *  of the input image are not equal.  That's because the transformation
 *  performed to an anisotropic image will probably look very strange when
 *  the image is scaled to be isotropic.
 *
 * Notes:::
 *  This is pretty general, but not blindingly fast.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, scaling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Perry
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_quadTransform (
    IPWState           *pIPWState,
    PIXR               *prS,
    UInt32              u0,
    UInt32              v0,
    UInt32              nu,
    UInt32              nv,
    Float64             xul,
    Float64             yul,
    Float64             xur,
    Float64             yur,
    Float64             xll,
    Float64             yll,
    Float64             xlr,
    Float64             ylr,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    Int32	 wS, hS, bplS, dS;  /* Width, height, bytesperline, depth */
    Int32	 wD, hD, bplD, dD;
    UInt8	*pS, *pD;
    ScaledInt1616 xdpi, ydpi;
    Bool	 allocatedHere = cFalse;

    ERR_GEN(prS == NULL || pPrD == NULL, cErrNullPtr);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetResolutionDPI( pIPWState, prS, &xdpi, &ydpi));
    ERR_GEN(ABS(xdpi - ydpi) > 0x10000, cErrInvalidParm);

	/* If we already have a dest pixr, use its dimensions and
	 * depth and everything.  These determine the type of
	 * warping that's to be done.
	 * Otherwise, create one having the same dimensions as the
	 * source.
	 */
    prD = *pPrD;
    if (prD != NULL)
	ERR_GEN_PIXR( prD, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		      cChannels1, cFrame32);
    else
    {
	ERR_PASS (w_clonePixrNoInit(pIPWState, prS, &prD));
	allocatedHere = cTrue;
    }

    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prD, &wD));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prD, &hD));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prD, &dD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

    /* The rectangle describing the image to be produced
     * must fall within destination pixr and be at least 2x2
     */
    ERR_GEN_CLEANUP(nu<2 || nv<2, cErrInvalidParm);
    ERR_GEN_CLEANUP((u0>(wD-2)) || (v0>(hD-2)) || ((u0+nu)>wD) || ((v0+nv)>hD),
	    cErrOutOfBounds);
			
    /* Do the quadrilateral version of the resampler */
    ERR_PASS_CLEANUP( i_resampleQuad (pIPWState, pS, wS, hS, dS, bplS,
				      pD, wD, hD, dD, bplD,
				      u0, v0, nu, nv,
				      xul, yul, xur, yur, xll, yll, xlr, ylr));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arbitrarily warps an input image.
 *
 * LongDescription:::
 *  When attempting to fit two image scans together, it may be necessary
 *  to warp them in many subtle ways.  This function arbitrarily warps
 *  an input image.  The warping function is supplied by the client.
 *  The argument userTransform is a pointer to this function.
 *
 * Input:::
 *  pIPWState::       Pointer to standard IPWrap state block.
 *  prS::             Pointer to source image to be warped.
 *  userTransform::   Pointer to client function that converts destination
 *		      coordinates (u = x coord, v = y coord) into
 *		      source image coords.  This function takes u, v,
 *		      a pointer to the returned source x coord and a pointer
 *		      to the source y coord.  The first argument is a
 *		      dummy pointer.  It exists because this procedure will
 *		      be called in exactly the same way as internal
 *		      transform procedures used in w_quadTransform and
 *		      w_rotateMag.  Those internal procedures have a
 *		      useful pointer there.
 *  blksiz::          The output image is computed in square blocks of
 *		      size blksiz x blksiz.  16 is a nice value.  Smaller
 *		      numbers give better looking results and take longer.
 *
 * Output:::
 *  pPrD::            Pointer to pointer to destination image.  If none
 *		      is supplied, one will be created.  If one is created,
 *		      its dimensions will be checked.  It should be
 *		      a clone of the source image.  However, it cannot
 *		      be the same as the source image.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrInvalidParm
 *  if prS == *pPrD.  Also returns cErrInvalidParm if the X and Y resolutions
 *  of the input image are not equal.  That's because the transformation
 *  performed to an anisotropic image will probably look very strange when
 *  the image is scaled to be isotropic.
 *
 * Notes:::
 *  This is pretty general, but not blindingly fast.
 *
 * FunctionCategory:::
 *  rotation
 *
 * Keywords:::
 *  rotation, scaling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  L. Perry
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_warp (
    IPWState           *pIPWState,
    PIXR               *prS,
    void	(CDECL *userTransform) (IPWState *pIPWState,
					void *dummy,
					Int32 u,
					Int32 v,
					Float64 *x,
					Float64 *y ),
    UInt32              blksiz,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR	*prD = NULL;
    Int32	 wS, hS, bplS, dS;  /* Width, height, bytesperline, depth */
    Int32	 wD, hD, bplD, dD;
    UInt8	*pS, *pD;
    ScaledInt1616 xdpi, ydpi;
    Bool	 allocatedHere = cFalse;

    ERR_GEN(prS == NULL || pPrD == NULL || userTransform == NULL,
	    cErrNullPtr);
    ERR_GEN(prS == *pPrD, cErrInvalidParm);

    ERR_GEN(blksiz < 2, cErrInvalidParm);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetResolutionDPI( pIPWState, prS, &xdpi, &ydpi));
    ERR_GEN(ABS(xdpi - ydpi) > 0x10000, cErrInvalidParm);

	/* If we already have a dest pixr, use its dimensions and
	 * depth and everything.  These determine the type of
	 * warping that's to be done.
	 * Otherwise, create one having the same dimensions as the
	 * source.
	 */
    prD = *pPrD;
    if (prD != NULL)
	ERR_GEN_PIXR( prD, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		      cChannels1, cFrame32);
    else
    {
	ERR_PASS (w_clonePixrNoInit(pIPWState, prS, &prD));
	allocatedHere = cTrue;
    }

    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prS, &dS));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prD, &wD));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prD, &hD));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prD, &dD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));

    /* Do the tesselated version of the resampler,
     * in anticipation that some significant curvatures
     * may be encountered, requiring the piecewise approach
     */
    ERR_PASS_CLEANUP( i_resampleUser (pIPWState, pS, wS, hS, dS, bplS,
				      pD, wD, hD, dD, bplD,
				      userTransform, blksiz));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*
$Log: warp.c,v $
Revision 1.1.1.1  1999/09/20 18:19:21  jlin
Imported sources

Revision 1.27  1998/01/16 23:17:48  aber
aber on Fri Jan 16 15:17:07 PST 1998

warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v

Expand to rotation and magnification routine to work with multi-channel
data.

Revision 1.26  1998/01/14 00:16:19  ddavies
warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.25  1997/07/11 01:59:22  ddavies
warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.24  1997/04/15 22:51:51  ddavies
warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v

Use non-initialized PIXR creation where possible.

Revision 1.23  1996/01/04 20:29:25  ejaquith
warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.22  1995/12/13  04:16:12  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/12/13  02:09:04  ddavies
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Return an error if any of the three routines defined here is given
 * an anisotropic image.  That's because we don't know what transform to
 * apply so that the operation as specified will look right on an
 * anisotropic image.
 *
 * Revision 1.20  1995/11/13  20:19:44  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.19  1995/10/13  22:08:15  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.18  1995/10/05  04:24:07  ddavies
 * ddavies on Wed Oct 4 21:21:25 PDT 1995
 *
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Fill in all the comment blocks.  Change the angle passed to i_rotateMag
 * from degrees to radians to be consistent with the other rotation functions.
 * Fill in all the parameters of the userTransform callback proc declared
 * in i_warp so C++ doesn't complain about the empty parens (means 0 args in
 * C++ instead of unspecified args in C).
 *
 * Revision 1.17  1995/09/21  23:01:11  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/16  00:16:22  ddavies
 * ddavies on Fri Sep 15 17:15:55 PDT 1995
 *
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Let dest image be any size, any depth in these routines.  Converting between
 * images of arbitrary size is what they do.
 *
 * Revision 1.15  1995/09/15  18:02:07  ddavies
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.14  1995/09/11  22:12:38  ddavies
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.13  1995/09/10  00:35:53  ddavies
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.12  1995/09/07  01:37:52  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:26:44  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  01:20:05  ddavies
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.9  1995/08/31  21:40:50  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/17  22:03:54  ejaquith
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.7  1995/07/29  04:56:13  ddavies
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.6  1994/10/07  19:19:45  lperry
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.5  1994/10/04  23:41:07  lperry
 * lperry on Tue Oct 4 16:39:53 PDT 1994
 *
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Temporarily delete the argument list from the
 * function ptr argument of iw_warpPixr. This is necessary
 * so that protos can be made w/o needing the def of
 * WarpInfoPtr. This will be fixed when all the xxpro.h
 * files are repaced with the pub/prv headers
 *
 * Revision 1.4  1994/09/29  21:52:55  lperry
 * lperry on Thu Sep 29 14:52:32 PDT 1994
 *
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Change to use warp.pub header file instead if ippros
 *
 * Revision 1.3  1994/05/12  21:52:21  lperry
 * lperry on Thu May 12 14:51:19 PDT 1994
 *
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Altered wrapper code to use new ipshared warp interface
 * The ipwrap interface itself is unchanged.
 *
 * Revision 1.2  1994/03/23  01:58:53  lperry
 * lperry on Tue Mar 22 17:54:44 PST 1994
 *
 * warp.c -> /product/ipcore/ipwrap/src/RCS/warp.c,v
 *
 * Needed to redeclare a function pointer argument as CDECL
 * to keep the PC happy. Also corrected testing of a UInt32 variable
 * for negative values. The PC compiler correctly labelled it as
 * a "meaningless expression". How come the PC compiler is so much
 * smarter than the Sun's
 *
 * Revision 1.1  1994/03/22  02:44:25  lperry
 * Initial revision
 *
*/
