/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  wmorph.c
 *
 *    Wrappers for fast binary morphology
 *      wp_fastDilate():   fast dilation operation based on assembly routines
 *      wp_fastErode():    fast erosion operation based on assembly routines
 *
 *    Wrappers for generation of machine code
 *      w_genDilateAsm():   generates assembly routines that implement the
 *                          dilation operation for a specified
 *                          structuring element
 *      w_genErodeAsm():    generates assembly routines that implement the
 *                          erosion operation for a specified
 *                          structuring element
 *      w_genLongDilateAsm(): generates assembly routines that implement the
 *                          dilation operation for a specified
 *                          structuring element using long (>= 64 bit)
 *			    operations.
 *      w_genLongErodeAsm(): generates assembly routines that implement the
 *                          erosion operation for a specified
 *                          structuring element using long (>= 64 bit)
 *			    operations.
 *
 */



#include "ipwcore.prv"

/* DEFINITIONS */
#include "alplib.h"
#include "geomadt.h"
#include "pixr.h"

/* PROTOTYPES */
#include "props.pub"
#include "morph.pub"
#include "morph.prv"
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: wmorph.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");


/*----------------------------------------------------------------------*
 *                 Wrappers for fast binary morphology                  *
 *----------------------------------------------------------------------*/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Dilates an image using a strel.  Uses optimized code if available.
 *
 * LongDescription:::
 *  Binary morphology is a useful image processing method.  One of the
 *  two morphology primitives is dilation.  See the explanation for
 *  w_dilate for a description of this.
 *<BR><BR>
 *  This is a private routine is called by w_dilate.  It assumes it
 *  is given either a pointer to a piece of optimized machine code or
 *  enough information about the structure element to do the operation as
 *  a series of bit-blits.
 *
 * PrivateDescription:::
 *  As described in w_dilate, the optimized code is generated at runtime
 *  for the individual platform.  This routine looks at the pointer to
 *  the optimized code.  If it's NULL, the routine grinds through a bunch
 *  of rasterops to shift and OR the image with itself.  If the optimized
 *  code is present, it is called on each subblock (size is also passed
 *  in) of the image.  The process of proceeding through the blocks
 *  and subblocks is also platform dependent, so portions of this routine
 *  are controlled by ifdefs.
 *<BR><BR>
 *  Note that the PC code processes the entire image as one block instead
 *  of using multiple subblocks.  Also note that the PC version operates
 *  a cache of shifted input words to avoid the expensive shifting operation.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be dilated.
 *  strel::          Pointer to structure element used in dilation.
 *
 * Output:::
 *  prD::            Pointer to output image.  This can never be NULL.  It
 *		     MUST be different than prS.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  Lots of image processing texts talk about binary morphology.  The
 *  father of the field is Jean Serra.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_fastDilate (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        imageWidth, h;            /* includes pads */
    Int32        imageHeight, w;           /* includes pads */
    Int32        blockWidthInBytes;     /* width of 1 block */
    Int32        r, c;
    Int32        nx, ny;
    Int32        bplS, bplD;
    UInt8       *strelData = NULL, *fill, *pS, *pD;
    Bool	 purifyIsRunning;

    ERR_PASS( w_pixrFramedHeight(pIPWState, prD, &h));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &imageHeight));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prD, &imageWidth));

    imageHeight = MIN(imageHeight, h);
        /*  the space allocated for the image is padded up
	 *  to the next multiple of 32      */
    imageWidth = MIN(imageWidth, w);

	/* This will be passed to accelerated routines. */
    blockWidthInBytes = strel->blockWidth;

	/*  i_fastDilate believes that the strel data is an UInt8.
	 *  Allocate space for this and tranfer strel->Data to it. */
    nx = strel->nx;
    ny = strel->ny;
    strelData = (UInt8 *)IPW_MALLOC((nx * ny) * sizeof(UInt8), cAllocatorData);
    ERR_GEN(strelData == NULL, cErrNomem);

    fill = strelData;
    for (r = 0; r < ny; r++)
    {
	for (c = 0; c < nx; c++)
	{
	    *fill = (UInt8)strel->data[r][c];
	    fill++;
	}
    }

	/* If we're running purify now and there is fast code, clear the dest
	 * image to initialize it.  Purify doesn't notice when the
	 * dest image is touched by runtime-generated code.
	 * Further Fix 1/7/98: One of the great Purify mysteries is
	 * when it decides that uninitialized memory is being read by
	 * rasterOp.  At the beginning and end of every scan line,
	 * rasterOp ALWAYS combines a source word with a dest word,
	 * even during copy.  The first thing the rasterOp (slow)
	 * implementation of i_fastDilate does is copy one version of
	 * the source to the dest.  Erode does this too.  However,
	 * for mysterious reasons, Purify 4.0.1 seems to complain when
	 * dilate does it.  So, whenever Purify is running, always
	 * clear, i.e. touch, the entire output image first.  Weird.
	 */
    ERR_PASS_CLEANUP( wp_IPWStateGetPurifyIsRunning(pIPWState,
						    &purifyIsRunning));
    if (purifyIsRunning)
	ERR_PASS_CLEANUP (w_clear (pIPWState, prD));

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    
    ERR_PASS_CLEANUP( i_fastDilate (pIPWState, pD, pS, strel->dilateOp,
				    strel->longDilateOp,
				    imageWidth, imageHeight, 
				    bplS, bplD,
				    strel->blockWidth,
				    strel->longBlockWidth,
				    nx, ny, strel->cx, strel->cy,
				    strelData));
    IPW_FREE(strelData, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(strelData, cAllocatorData);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Erodes an image using a strel.  Uses optimized code if available.
 *
 * LongDescription:::
 *  Binary morphology is a useful image processing method.  One of the
 *  two morphology primitives is erosion.  See the explanation for
 *  w_erode for a description of this.
 *<BR><BR>
 *  This is a private routine is called by w_erode.  It assumes it
 *  is given either a pointer to a piece of optimized machine code or
 *  enough information about the structure element to do the operation as
 *  a series of bit-blits.
 *
 * PrivateDescription:::
 *  As described in w_erode, the optimized code is generated at runtime
 *  for the individual platform.  This routine looks at the pointer to
 *  the optimized code.  If it's NULL, the routine grinds through a bunch
 *  of rasterops to shift and AND the image with itself.  If the optimized
 *  code is present, it is called on each subblock (size is also passed
 *  in) of the image.  The process of proceeding through the blocks
 *  and subblocks is also platform dependent, so portions of this routine
 *  are controlled by ifdefs.
 *<BR><BR>
 *  Note that the PC code processes the entire image as one block instead
 *  of using multiple subblocks.  Also note that the PC version operates
 *  a cache of shifted input words to avoid the expensive shifting operation.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to image to be eroded.
 *  strel::          Pointer to structure element used in dilation.
 *
 * Output:::
 *  prD::            Pointer to output image.  This can never be NULL.  It
 *		     MUST be different than prS.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Bloomberg, D. Davies
 *
 * Citations:::
 *  Lots of image processing texts talk about binary morphology.  The
 *  father of the field is Jean Serra.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_fastErode (
    IPWState           *pIPWState,
    PIXR               *prS,
    STREL              *strel,
    PIXR               *prD)
{
    PROLOGUE;
    Int32        imageWidth, h;            /* includes pads */
    Int32        imageHeight, w;           /* includes pads */
    Int32        blockWidthInBytes;     /* width of 1 block */
    Int32        r, c;
    Int32        nx, ny;
    Int32        bplS, bplD;
    UInt8       *strelData = NULL, *fill, *pS, *pD;
    Bool	 purifyIsRunning;

    ERR_PASS( w_pixrFramedHeight(pIPWState, prD, &h));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &imageHeight));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedWidth(pIPWState, prD, &imageWidth));

    imageHeight = MIN(imageHeight, h);
        /*  the space allocated for the image is padded up
	 *  to the next multiple of 32      */
    imageWidth = MIN(imageWidth, w);

	/* This will be passed to accelerated routines. */
    blockWidthInBytes = strel->blockWidth;

	/*  i_fastErode believes that the strel data is an UInt8.
	 *  Allocate space for this and tranfer strel->Data to it. */
    nx = strel->nx;
    ny = strel->ny;
    strelData = (UInt8 *)IPW_MALLOC((nx * ny) * sizeof(UInt8), cAllocatorData);
    ERR_GEN(strelData == NULL, cErrNomem);

    fill = strelData;
    for (r = 0; r < ny; r++)
    {
	for (c = 0; c < nx; c++)
	{
	    *fill = (UInt8)strel->data[r][c];
	    fill++;
	}
    }

	/* If we're running purify now and there is fast code, clear the dest
	 * image to initialize it.  Purify doesn't notice when the
	 * dest image is touched by runtime-generated code.
	 * Further Fix 1/7/98: One of the great Purify mysteries is
	 * when it decides that uninitialized memory is being read by
	 * rasterOp.  At the beginning and end of every scan line,
	 * rasterOp ALWAYS combines a source word with a dest word,
	 * even during copy.  The first thing the rasterOp (slow)
	 * implementation of i_fastDilate does is copy one version of
	 * the source to the dest.  Erode does this too.  However,
	 * for mysterious reasons, Purify 4.0.1 seems to complain when
	 * dilate does it.  Maybe it will start complaining about
	 * erode sometime.  So, whenever Purify is running, always
	 * clear, i.e. touch, the entire output image first.  Weird.
	 */
    ERR_PASS_CLEANUP( wp_IPWStateGetPurifyIsRunning(pIPWState,
						    &purifyIsRunning));
    if (purifyIsRunning)
	ERR_PASS_CLEANUP (w_clear (pIPWState, prD));

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prD, &pD));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    
    ERR_PASS_CLEANUP( i_fastErode (pIPWState, pD, pS, strel->erodeOp,
				   strel->longErodeOp,
				   imageWidth, imageHeight, 
				   bplS, bplD,
				   strel->blockWidth,
				   strel->longBlockWidth,
				   nx, ny, strel->cx, strel->cy,
				   strelData));
    IPW_FREE(strelData, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(strelData, cAllocatorData);
    ERR_RETURN;
}



/*----------------------------------------------------------------------*
 *                Wrappers for generation of machine code              *
 *----------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generates optimized code to dilate with a particular structure element.
 *
 * LongDescription:::
 *  Binary morphology is a useful image processing method.  One of the
 *  two morphology primitives is dilation.  See the explanation for
 *  w_dilate for a description of this.  This routine generates code used
 *  in image dilation.
 *
 * PrivateDescription:::
 *  There are many ways to implement dilation.  This routine generates
 *  machine code to perform the dilation operation specified by a particular
 *  structure element.  The code generated attempts to take advantage
 *  of any special features of the platform while avoiding as many
 *  conditionals as possible.
 *<BR><BR>
 *  The resulting code is attached to the structure element in the
 *  dilateOp slot of the structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  strel::          Pointer to structure element for which the code
 *		     will be generated.
 *
 * Output:::
 *  pCodeLength::    Pointer to length of generated code.  This can be
 *		     used by the client when moving the code around.
 *		     The length is given in units of 4-byte words.
 *
 * Returns:::
 *  In addition to the normal returns, the lower level routines check to
 *  see if 1) the strel is too big to be accelerated, or 2) it is not
 *  possible to generate code for the current platform.  These two conditions
 *  are filtered out before being returned here since they are not
 *  really errors.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  heuristic 
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_genDilateAsm (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *pCodeLength)
{
    PROLOGUE;
    Int32        r, c;
    Int32        nx, ny, result;
    Int32	 bufSize;
    UInt8       *strelData = NULL, *fill;

    ERR_GEN(strel == NULL || pCodeLength == NULL, cErrNullPtr);

	/* Make sure we know what to free */
    strel->dilateOp = NULL;

	/* i_genDilateAsm believes that the strel data is an UInt8.
	 * Allocate space for this and tranfer strel->Data to it. */
    nx = strel->nx;
    ny = strel->ny;
    strelData = (UInt8 *)IPW_MALLOC((nx * ny) * sizeof(UInt8), cAllocatorData);
    ERR_GEN(strelData == NULL, cErrNomem);

    fill = strelData;
    for (r = 0; r < ny; r++)
    {
	for (c = 0; c < nx; c++)
	{
	    *fill = (UInt8)strel->data[r][c];
	    fill++;
	}
    }

	/* Generate the machine code.  Figure out the size of the
	 * buffer needed to hold the machine code.
	 */
    result = i_getDilateAsmSize (pIPWState, nx, ny,
				 strel->cx, strel->cy, &bufSize);

	/* If it failed in a bad way, quit now */
    ERR_GEN_CLEANUP (((result != cSuccessful) &&
		      (result != cErrSETooBig) &&
		      (result != cErrNoImpl)), result);

	/* If that wasn't successful, don't try to allocate a buffer
	 * and fill it with machine code.  However, we will return
	 * with a cSuccessful error code since this is not
	 * unexpected behavior.
	 */
    if (result == cSuccessful)
    {
	    /* In Visual C++, turn off the warning that says we're casting
	     * a pointer to data to a pointer to a function.  We're going
	     * to do exactly that.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	strel->dilateOp = (DilateProc)IPW_CALLOC(bufSize, 1, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif
	ERR_GEN_CLEANUP((strel->dilateOp == NULL), cErrNomem);
    
	    /* In the current scheme of things, failing to generate
	     * code isn't really an "error", so don't generate an
	     * error code in that case.  There are legal two reasons why we
	     * might fail to generate code.  First, we might be linked
	     * to the slow code, in which case no machine code can be
	     * generated.  Second, the structure element may be too
	     * big.  If any other error appears, complain.
	     */
	result = i_genDilateAsm (pIPWState, nx, ny, strel->cx, strel->cy,
				 strelData, &strel->blockWidth,
				 &strel->dilateOp, 
				 pCodeLength);
    
	    /* If it failed in a bad way, quit now */
	ERR_GEN_CLEANUP (((result != cSuccessful) &&
			  (result != cErrSETooBig) &&
			  (result != cErrNoImpl)), result);
    
	    /* If it failed in any way, free the allocated memory. */
	    /* In Visual C++, turn off the warning that says we're about to
	     * cast a pointer to a function to a pointer to data, since
	     * that's exactly what we intend to do.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
	if (result != cSuccessful)
	    IPW_FREE(strel->dilateOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif
    }

    IPW_FREE(strelData, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:

#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    if (strel->dilateOp != NULL)
	IPW_FREE(strel->dilateOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    IPW_FREE(strelData, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generates optimized code to erode with a particular structure element.
 *
 * LongDescription:::
 *  Binary morphology is a useful image processing method.  One of the
 *  two morphology primitives is erosion.  See the explanation for
 *  w_erode for a description of this.  This routine generates code used
 *  in image erosion.
 *
 * PrivateDescription:::
 *  There are many ways to implement erosion.  This routine generates
 *  machine code to perform the erosion operation specified by a particular
 *  structure element.  The code generated attempts to take advantage
 *  of any special features of the platform while avoiding as many
 *  conditionals as possible.
 *<BR><BR>
 *  The resulting code is attached to the structure element in the
 *  erodeOp slot of the structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  strel::          Pointer to structure element for which the code
 *		     will be generated.
 *
 * Output:::
 *  pCodeLength::    Pointer to length of generated code.  This can be
 *		     used by the client when moving the code around.
 *		     The length is given in units of 4-byte words.
 *
 * Returns:::
 *  In addition to the normal returns, the lower level routines check to
 *  see if 1) the strel is too big to be accelerated, or 2) it is not
 *  possible to generate code for the current platform.  These two conditions
 *  are filtered out before being returned here since they are not
 *  really errors.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  heuristic 
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_genErodeAsm (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *pCodeLength)
{
    PROLOGUE;
    Int32        r, c;
    Int32        nx, ny, result;
    Int32	 bufSize;
    UInt8       *strelData = NULL, *fill;

    ERR_GEN(strel == NULL || pCodeLength == NULL, cErrNullPtr);

	/* Make sure we know what to free */
    strel->erodeOp = NULL;

	/* i_genErodeAsm believes that the strel data is an UInt8.
	 * Allocate space for this and tranfer strel->Data to it. */
    nx = strel->nx;
    ny = strel->ny;
    strelData = (UInt8 *)IPW_MALLOC((nx * ny) * sizeof(UInt8), cAllocatorData);
    ERR_GEN(strelData == NULL, cErrNomem);

    fill = strelData;
    for (r = 0; r < ny; r++)
    {
	for (c = 0; c < nx; c++)
	{
	    *fill = (UInt8)strel->data[r][c];
	    fill++;
	}
    }

	/* Generate the machine code.  Figure out the size of the
	 * buffer needed to hold the machine code.
	 */
    result = i_getErodeAsmSize (pIPWState, nx, ny,
				strel->cx, strel->cy, &bufSize);

	/* If it failed in a bad way, quit now */
    ERR_GEN_CLEANUP (((result != cSuccessful) &&
		      (result != cErrSETooBig) &&
		      (result != cErrNoImpl)), result);

	/* If that wasn't successful, don't try to allocate a buffer
	 * and fill it with machine code.  However, we will return
	 * with a cSuccessful error code since this is not
	 * unexpected behavior.
	 */
    if (result == cSuccessful)
    {
	    /* In Visual C++, turn off the warning that says we're casting
	     * a pointer to data to a pointer to a function.  We're going
	     * to do exactly that.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	strel->erodeOp = (ErodeProc)IPW_CALLOC(bufSize, 1, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif
	ERR_GEN_CLEANUP((strel->erodeOp == NULL), cErrNomem);
    
	    /* In the current scheme of things, failing to generate
	     * code isn't really an "error", so don't generate an
	     * error code in that case.  There are two legal reasons why we
	     * might fail to generate code.  First, we might be linked
	     * to the slow code, in which case no machine code can be
	     * generated.  Second, the structure element may be too
	     * big.  If any other error appears, complain. 
	     */
	result = i_genErodeAsm (pIPWState, nx, ny, strel->cx, strel->cy,
			        strelData, &strel->blockWidth,
			        &strel->erodeOp, 
			        pCodeLength);
    
	    /* If it failed in a bad way, quit now */
	ERR_GEN_CLEANUP (((result != cSuccessful) &&
			  (result != cErrSETooBig) &&
			  (result != cErrNoImpl)), result);
    
	    /* If it failed in any way, free the allocated memory. */
	    /* In Visual C++, turn off the warning that says we're about to
	     * cast a pointer to a function to a pointer to data, since
	     * that's exactly what we intend to do.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
	if (result != cSuccessful)
	    IPW_FREE(strel->erodeOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif
    }

    IPW_FREE(strelData, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:

#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    if (strel->erodeOp != NULL)
	IPW_FREE(strel->erodeOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    IPW_FREE(strelData, cAllocatorData);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generates code that uses long operations to do dilation with an SE.
 *
 * LongDescription:::
 *  Binary morphology is a useful image processing method.  One of the
 *  two morphology primitives is dilation.  See the explanation for
 *  w_dilate for a description of this.  This routine generates code used
 *  in image dilation.  In particular, this routine generates code that uses
 *  long (>= 64 bit) memory accesses and operations.
 *
 * PrivateDescription:::
 *  There are many ways to implement dilation.  This routine generates
 *  machine code to perform the dilation operation specified by a particular
 *  structure element.  The code generated attempts to take advantage
 *  of any special features of the platform while avoiding as many
 *  conditionals as possible.  Among the special features used here
 *  are long ( >= 64 bit ) memory and logic operations.
 *<BR><BR>
 *  The resulting code is attached to the structure element in the
 *  longDilateOp slot of the structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  strel::          Pointer to structure element for which the code
 *		     will be generated.
 *
 * Output:::
 *  pCodeLength::    Pointer to length of generated code.  This can be
 *		     used by the client when moving the code around.
 *		     The length is given in units of 4-byte words.
 *
 * Returns:::
 *  In addition to the normal returns, the lower level routines check to
 *  see if 1) the strel is too big to be accelerated, 2) the platform
 *  supports long operations, or 3) it is not possible to generate
 *  code for the current platform.  These three conditions
 *  are filtered out before being returned here since they are not
 *  really errors.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  heuristic 
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_genLongDilateAsm (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *pCodeLength)
{
    PROLOGUE;
    Int32        r, c;
    Int32        nx, ny, result;
    Int32	 bufSize;
    UInt8       *strelData = NULL, *fill;

    ERR_GEN(strel == NULL || pCodeLength == NULL, cErrNullPtr);

	/* Make sure we know what to free */
    strel->longDilateOp = NULL;

	/* i_genLongDilateAsm believes that the strel data is an UInt8.
	 * Allocate space for this and tranfer strel->Data to it. */
    nx = strel->nx;
    ny = strel->ny;
    strelData = (UInt8 *)IPW_MALLOC((nx * ny) * sizeof(UInt8), cAllocatorData);
    ERR_GEN(strelData == NULL, cErrNomem);

    fill = strelData;
    for (r = 0; r < ny; r++)
    {
	for (c = 0; c < nx; c++)
	{
	    *fill = (UInt8)strel->data[r][c];
	    fill++;
	}
    }

	/* Generate the machine code.  Figure out the size of the
	 * buffer needed to hold the machine code.
	 */
    result = i_getLongDilateAsmSize (pIPWState, nx, ny,
				     strel->cx, strel->cy, &bufSize);

	/* If it failed in a bad way, quit now */
    ERR_GEN_CLEANUP (((result != cSuccessful) &&
		      (result != cErrSETooBig) &&
		      (result != cErrNoImpl)), result);

	/* If that wasn't successful, don't try to allocate a buffer
	 * and fill it with machine code.  However, we will return
	 * with a cSuccessful error code since this is not
	 * unexpected behavior.
	 */
    if (result == cSuccessful)
    {
	    /* In Visual C++, turn off the warning that says we're casting
	     * a pointer to data to a pointer to a function.  We're going
	     * to do exactly that.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	strel->longDilateOp =
		(LongDilateProc)IPW_CALLOC(bufSize, 1, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif
	ERR_GEN_CLEANUP((strel->longDilateOp == NULL), cErrNomem);
    
	    /* In the current scheme of things, failing to generate
	     * code isn't really an "error", so don't generate an
	     * error code in that case.  There are two legal reasons why we
	     * might fail to generate code.  First, we might be linked
	     * to the slow code, in which case no machine code can be
	     * generated.  Second, the structure element may be too
	     * big.  If any other error appears, complain. 
	     */
	result = i_genLongDilateAsm (pIPWState, nx, ny, strel->cx, strel->cy,
				     strelData, &strel->longBlockWidth,
				     &(strel->longDilateOp), 
				     pCodeLength);
    
	    /* If it failed in a bad way, quit now */
	ERR_GEN_CLEANUP (((result != cSuccessful) &&
			  (result != cErrSETooBig) &&
			  (result != cErrNoImpl)), result);
    
	    /* If it failed in any way, free the allocated memory. */
	    /* In Visual C++, turn off the warning that says we're about to
	     * cast a pointer to a function to a pointer to data, since
	     * that's exactly what we intend to do.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
	if (result != cSuccessful)
	    IPW_FREE(strel->longDilateOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif
    }

    IPW_FREE(strelData, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:

#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    if (strel->longDilateOp != NULL)
	IPW_FREE(strel->longDilateOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    IPW_FREE(strelData, cAllocatorData);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generates code that uses long operations to do erosion with an SE.
 *
 * LongDescription:::
 *  Binary morphology is a useful image processing method.  One of the
 *  two morphology primitives is erosion.  See the explanation for
 *  w_erode for a description of this.  This routine generates code used
 *  in image erosion.  In particular, this routine generates code that uses
 *  long (>= 64 bit) memory accesses and operations.
 *
 * PrivateDescription:::
 *  There are many ways to implement erosion.  This routine generates
 *  machine code to perform the erosion operation specified by a particular
 *  structure element.  The code generated attempts to take advantage
 *  of any special features of the platform while avoiding as many
 *  conditionals as possible.  Among the special features used here
 *  are long ( >= 64 bit ) memory and logic operations.
 *<BR><BR>
 *  The resulting code is attached to the structure element in the
 *  longErodeOp slot of the structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  strel::          Pointer to structure element for which the code
 *		     will be generated.
 *
 * Output:::
 *  pCodeLength::    Pointer to length of generated code.  This can be
 *		     used by the client when moving the code around.
 *		     The length is given in units of 4-byte words.
 *
 * Returns:::
 *  In addition to the normal returns, the lower level routines check to
 *  see if 1) the strel is too big to be accelerated, 2) the platform
 *  supports long operations, or 3) it is not possible to generate
 *  code for the current platform.  These three conditions
 *  are filtered out before being returned here since they are not
 *  really errors.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology
 *
 * Level:::
 *  heuristic 
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_genLongErodeAsm (
    IPWState           *pIPWState,
    STREL              *strel,
    Int32              *pCodeLength)
{
    PROLOGUE;
    Int32        r, c;
    Int32        nx, ny, result;
    Int32	 bufSize;
    UInt8       *strelData = NULL, *fill;

    ERR_GEN(strel == NULL || pCodeLength == NULL, cErrNullPtr);

	/* Make sure we know what to free */
    strel->longErodeOp = NULL;

	/* i_genLongErodeAsm believes that the strel data is an UInt8.
	 * Allocate space for this and tranfer strel->Data to it. */
    nx = strel->nx;
    ny = strel->ny;
    strelData = (UInt8 *)IPW_MALLOC((nx * ny) * sizeof(UInt8), cAllocatorData);
    ERR_GEN(strelData == NULL, cErrNomem);

    fill = strelData;
    for (r = 0; r < ny; r++)
    {
	for (c = 0; c < nx; c++)
	{
	    *fill = (UInt8)strel->data[r][c];
	    fill++;
	}
    }

	/* Generate the machine code.  Figure out the size of the
	 * buffer needed to hold the machine code.
	 */
    result = i_getLongErodeAsmSize (pIPWState, nx, ny,
				    strel->cx, strel->cy, &bufSize);

	/* If it failed in a bad way, quit now */
    ERR_GEN_CLEANUP (((result != cSuccessful) &&
		      (result != cErrSETooBig) &&
		      (result != cErrNoImpl)), result);

	/* If that wasn't successful, don't try to allocate a buffer
	 * and fill it with machine code.  However, we will return
	 * with a cSuccessful error code since this is not
	 * unexpected behavior.
	 */
    if (result == cSuccessful)
    {
	    /* In Visual C++, turn off the warning that says we're casting
	     * a pointer to data to a pointer to a function.  We're going
	     * to do exactly that.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	strel->longErodeOp =
			(LongErodeProc)IPW_CALLOC(bufSize, 1, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif
	ERR_GEN_CLEANUP((strel->longErodeOp == NULL), cErrNomem);
    
	    /* In the current scheme of things, failing to generate
	     * code isn't really an "error", so don't generate an
	     * error code in that case.  There are two legal reasons why we
	     * might fail to generate code.  First, we might be linked
	     * to the slow code, in which case no machine code can be
	     * generated.  Second, the structure element may be too
	     * big.  If any other error appears, complain. 
	     */
	result = i_genLongErodeAsm (pIPWState, nx, ny, strel->cx, strel->cy,
				    strelData, &strel->longBlockWidth,
				    &(strel->longErodeOp), 
				    pCodeLength);
    
	    /* If it failed in a bad way, quit now */
	ERR_GEN_CLEANUP (((result != cSuccessful) &&
			  (result != cErrSETooBig) &&
			  (result != cErrNoImpl)), result);
    
	    /* If it failed in any way, free the allocated memory. */
	    /* In Visual C++, turn off the warning that says we're about to
	     * cast a pointer to a function to a pointer to data, since
	     * that's exactly what we intend to do.
	     */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
	if (result != cSuccessful)
	    IPW_FREE(strel->longErodeOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif
    }

    IPW_FREE(strelData, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:

#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    if (strel->longErodeOp != NULL)
	IPW_FREE(strel->longErodeOp, cAllocatorData);
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    IPW_FREE(strelData, cAllocatorData);
    ERR_RETURN;
}



/*

 $Log: wmorph.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.29  1998/01/14 00:16:20  ddavies
 wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.28  1997/04/15 22:55:09  ddavies
 wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v

 Modify w_fastErode and w_fastDilate so that they check the new
 purifyIsRunning flag in the ipstate.  If it's set, they touch (clear) the
 destination bitmap.  This tells purify that the dest bitmap is initialized.
 The erode and dilate operations may be implemented in code that's generated
 at runtime (they also check to see if there's such code before doing this)
 so Purify won't notice when that code initializes the image.

 Revision 1.27  1996/12/19 23:14:29  ddavies
 wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v

 Have the routines that call i_getErodeAsmSize, i_getDilateAsmSize,
 i_getLongErodeAsmSize and i_getLongDilateAsmSize check the return code
 before allocating the memory buffer.  On platforms that don't support
 machine code, don't allocate a buffer of size 0.  Purify doesn't like to
 do that.  It refuses to free such a buffer and complains that it was
 leaked.  The WATCOM package also returns NULL when asked to allocate
 a buffer of size 0, so the checks complain that we ran out of memory.

 Revision 1.26  1996/12/18 23:23:59  ddavies
 wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v

 Modify w_genDilateAsm and w_genErodeAsm so that they preallocate the buffer
 that will hold machine code (using IPW_CALLOC) and pass it to i_genErodeAsm
 and i_genDilateAsm.  In the old code, we let the ipshared routines allocate
 the buffer (using MALLOC) and freed it in ipwrap using IPW_FREE.  This is
 generally a bad thing unless the client happens to pass our allocator.
 Also add implementations for new routines w_genLongDilateAsm and
 w_genLongErodeAsm.  These routines are like w_genDilateAsm and w_genErodeAsm
 except that they have ipshared routines generate machine code that uses
 long (>= 64 bit) operations.

 Revision 1.25  1996/01/04 20:29:28  ejaquith
 wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.24  1995/12/13  04:16:17  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.23  1995/12/01  19:34:27  ddavies
 * ddavies on Fri Dec 1 11:33:14 PST 1995
 *
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Fix two error cleanup clauses that said ERR_RETURN; IPW_FREE instead of
 * IPW_FREE; ERR_RETURN;.  acc found these with the bizarre warning "loop
 * not entered at top".
 *
 * Revision 1.22  1995/11/13  20:19:48  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.21  1995/11/03  22:50:21  ddavies
 * ddavies on Fri Nov 3 14:49:59 PST 1995
 *
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Fix problems with PROLOGUES, lack of ERR_PASS[_CLEANUP], etc found by
 * Eric's scripts.
 *
 * Revision 1.20  1995/10/13  22:08:18  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/10/04  20:25:29  ddavies
 * ddavies on Wed Oct 4 13:25:19 PDT 1995
 *
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Add in ErodeProc and DilateProc to keep compiler happy.
 *
 * Revision 1.18  1995/09/21  23:01:14  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/15  18:01:52  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.16  1995/09/11  22:12:23  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.15  1995/09/10  00:34:23  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.14  1995/09/07  01:37:54  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.13  1995/09/06  20:26:47  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.12  1995/09/01  01:20:10  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.11  1995/08/31  21:40:52  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/08/30  00:49:48  ddavies
 * ddavies on Tue Aug 29 17:48:24 PDT 1995
 *
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Dont't return an error when i_genErodeAsm and i_genDilateAsm are really dummy
 * routines because we're linked to slow code.  To avoid this, do not generate
 * an error when these routines return ia_noImpl.
 *
 * Revision 1.9  1995/08/24  20:23:30  ejaquith
 * ejaquith on Thu Aug 24 13:22:59 PDT 1995
 *
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Added missing calls to free strelData before return to fix some memory
 * leaks.
 *
 * Revision 1.8  1995/08/23  18:39:57  macleod
 * macleod on Wed Aug 23 11:37:23 PDT 1995
 *
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * When i_genDilateAsm and i_genErodeAsm couldn't generate code because
 * the strel was too high, they generated ia_SETooBig. The wrapper code
 * was generating an error in this case, causing strels above 19 to never
 * work. The right thing to do, given how everything works, is to not
 * consider this an error to the caller, since the non-accelerated code
 * gets cut in when the code isn't generated.
 * (When I said "19" above, I mean "19 high")
 *
 * Revision 1.7  1995/08/17  22:03:42  ejaquith
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.6  1995/07/29  04:56:20  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.5  1994/10/07  19:19:09  lperry
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.4  1994/09/20  04:36:33  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Guaranteed that strel->dilateOp is NULL before calling i_genErodeAsm and
 * i_genDilateAsm.  These routines use this as a cue that they should allocate
 * space for the assembly code using IPW_MALLOC.  Also deleted erodeMalloc,
 * dilateMalloc and allocShrink since they're no longer used.
 *
 * Revision 1.3  1993/12/03  01:20:36  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:19:20  ddavies
 * wmorph.c -> /product/ipcore/ipwrap/src/RCS/wmorph.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

