/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * writfile.c--includes procedures
 *
 *                 w_write()          standard interface: removes padding
 *                 w_writeEx()          standard interface: removes padding
 *                 w_writeImage()       standard interface: optional clipping
 */


#include <stdio.h>

#include "tiffiop.h"
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "tiff.h"
#include "rastrfil.h"
#include "defines.pub"
#include "ptypes.h"
#include "gdefs.h"
#include "pcx.h"
#include "shrcc.pub"

/* prototypes */
#include "io.pub"
#include "iotiff.pub"
#include "io.prv"
#include "props.pub"
#include "rectops.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: writfile.c,v $; $Revision: 1.5 $; $Date: 2000/07/08 02:14:15 $");

#define  EXTRA_BITS          100
#define  MAX_WIDTH           (2550 + EXTRA_BITS)
#define  MAX_HEIGHT          (3300 + EXTRA_BITS)

#define  NOT_COMPRESSED      0
#define  COMPRESSED          1

#define  ATTLENGTH           2048


#define PUT_SHORT(num, attributes) \
    (*(attributes))[0] = ((num) >> 8) & 0377;\
    (*(attributes))[1] = (num) & 0377;\
    (*(attributes)) += 2;

#define PUT_NYBBLE_CHAR(fptr, nibble) \
    if ((nybble) < 0xa) \
        putc('0' + (nybble), (fp)); \
    else    /* if (nybble > 9 ) */ \
        putc('a' - 10 + (nybble), (fp));


void finishTiff( TIFF *tif);

void
finishTiff( TIFF *tif)
{
    (void) TIFFCloseFile(tif);
    _TIFFfree(tif, (char *)tif);
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to file stream in SUNRAS format.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to write to.
 *  pixr::           Ptr to PIXR
 *  compFlag::       NOT_COMPRESSED or COMPRESSED.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_fwriteSun (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR               *pixr,
    Int32               compFlag)
{
    PROLOGUE;
    Int32  channels;

    ERR_GEN(fp == NULL  ||  pixr == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));
    ERR_GEN(channels != 1, cErrDepthNotSupported);

    if (compFlag == NOT_COMPRESSED)
        ERR_PASS( wp_pixr_dump(pIPWState, pixr, fp, RT_STANDARD));
    else if (compFlag == COMPRESSED)
        ERR_PASS( wp_pixr_dump(pIPWState, pixr, fp, RT_BYTE_ENCODED));
    else
        ERR_GEN(1, cErrInvalidParm);

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Writes PIXR to stream in PS format.
 *
 * LongDescription:::
 *  Write to properly encapsulated postscript: 1 pixel per bit
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to write to.
 *  pixr::           Ptr to single channel, 1-bit depth PIXR.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 * 
 * Citations:::
 *   with the help of Emil Rainero and Rob Tow
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_fwritePS (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR               *pixr)
{
    PROLOGUE;
    UInt8         *imagePtr, *bytePtr;
    Int32          c,  nybble;
    Int32          width, height, Bpl, psBpl, i, j;
    Int32          bbllx, bblly, bburx, bbury;
    Float32        winch, hinch;

    ERR_GEN(fp == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);

/* PARANOID??  The code below contains character strings
 * which could be used with a disassembler to figure out our glorious
 * PIXR structure.  Do we care??
 */

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    ERR_PASS( w_flip(pIPWState, pixr));
#endif  /* _ALPACA_IMAGE_FMT_ */

	/*
	 * find width and height without frame;
	 * get imagePtr to part of image within frame;
	 * Bpl is full width of image (in bytes).
	 */
    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &width));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &height));
    ERR_GEN (width > MAX_WIDTH, cErrImageTooWide);

    if (height > MAX_HEIGHT)
        height = MAX_HEIGHT;

    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &Bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, pixr, &imagePtr));
    psBpl = (width + 7) >> 3;

        /* compute BB comment information */
    winch = width / 300.0F;
    hinch = height / 300.0F;
    bbllx = (Int32)(36. * (8.5 - winch));
    bblly = (Int32)(36. * (11.0 - hinch));
    bburx = bbllx + (Int32)(72. * winch);
    bbury = bblly + (Int32)(72. * hinch);

        /* write the encapsulation header information */
    fprintf(fp, "%%!\n");
    fprintf(fp, "%%%%BoundingBox: %ld %ld %ld %ld\n", bbllx, bblly, bburx, bbury);
    fprintf(fp, "/width %ld def        %% width of image in pixels\n", width);
    fprintf(fp, "/height %ld def       %% height of image in pixels\n", height);
    fprintf(fp, "/negh height neg def      %% negative of height\n");
    fprintf(fp, "/bits_per_pixel 1 def          %% bits per pixel\n\n");

    fprintf(fp, "/spi 300 def                   %% spots per inch\n\n");

    fprintf(fp, "/page_height 11.0 spi mul def  %% in pixels\n");
    fprintf(fp, "/page_width 8.5 spi mul def    %% in pixels\n\n");

    fprintf(fp, "        %% x origin of image in pixels\n");
    fprintf(fp, "/xorigin page_width 2 div width 2 div sub def\n\n");
    fprintf(fp, "        %% y origin of image in pixels\n");
    fprintf(fp, "/yorigin page_height 2 div height 2 div sub def\n\n");

    fprintf(fp, "/xscale width def              %% x scale factor\n");
    fprintf(fp, "/yscale height def             %% y scale factor\n\n");

    fprintf(fp, "        %% number of bytes per line of image\n");
    fprintf(fp, "/str width 7 add 8 idiv cvi string def\n\n");

    fprintf(fp, "72 spi div dup scale       %% 1 unit = 1 bit @ 300 spi\n\n");

    fprintf(fp, "xorigin yorigin translate      %% set the image origin\n");
    fprintf(fp, "xscale yscale scale            %% set the image scale\n\n");

    fprintf(fp, "{1 exch sub} settransfer       %% black pixels!\n\n");

    fprintf(fp, "width height bits_per_pixel [width 0 0 negh 0 height]\n");
    fprintf(fp, "{currentfile str readhexstring pop} image\n\n");

        /* convert the data to hex ascii and write it out */
    for (i = 0; i < height; i++)
    {
	bytePtr = imagePtr + i * Bpl;
        for (j = 0; j < psBpl; j++)
        {
            c = *bytePtr++;
            nybble = (c & 0xf0) >> 4;
            PUT_NYBBLE_CHAR(fp, nybble);
            nybble = c & 0xf;
            PUT_NYBBLE_CHAR(fp, nybble);
        }
        fprintf(fp, "\n");
    }

        /* finish */
    fprintf(fp, "\n\nshowpage\n");

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
    ERR_PASS( w_flip(pIPWState, pixr));
#endif  /* _ALPACA_IMAGE_FMT_ */

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to file stream in PNM format.
 *
 * LongDescription:::
 *  We support PBM (Portable BitMap), PGM (Portable GrayMap), and
 *  PPM (Portable PixMap) of the superclass PNM (Portable AnyMap) format, for
 *  binary, gray, and RGB PIXRs, respectively.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to write to.
 *  pixr::           Ptr to PIXR, depth 1,2,4, or 8.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *  ejaquith::  Added RGB format capability
 *  ejaquith::  Added depth 2 and 4 PGM capability
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_fwritePnm (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR               *pixr)
{
    PROLOGUE;
    UInt8	*imageP, *pLineBuffer = NULL;
    Int32	 bytesPerFileLine, lineDelta, depth, w, h, i, channels;

    ERR_PASS( w_pixrGetDepth (pIPWState, pixr, &depth));
    ERR_PASS( w_pixrGetWidth (pIPWState, pixr, &w));
    ERR_PASS( w_pixrGetHeight (pIPWState, pixr, &h));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));
    ERR_PASS( w_pixrGetBpl (pIPWState, pixr, &lineDelta));

    /* Check for all allowable types */
    ERR_GEN(!(depth == 1  &&  channels == 1)  &&
	    !(depth == 2  &&  channels == 1)  &&
	    !(depth == 4  &&  channels == 1)  &&
	    !(depth == 8  &&  (channels == 1  ||  channels == 3)),
	    cErrTypeNotSupported);


    ERR_PASS( w_flip(pIPWState, pixr));


    /* Write the header */
    if (depth == 1  ||  (depth == 8  &&  channels == 1))
    {
	if (depth == 1)
	{
	    /* binary */
	    fprintf(fp, "P4\n%ld %ld\n", w, h);
	    bytesPerFileLine = w / 8 + (w & 7 ? 1 : 0);
	}
	else 
	{
	    /* 8-bit gray */
	    fprintf(fp, "P5\n%ld %ld\n255\n", w, h);
	    bytesPerFileLine = w;
	}


	/*
	 * Dump the bits. It's not clear from the doc. if the binary version
	 * requires bit/line or byte/line alignment. I'll assume the latter,
	 * but will have to fix it if necessary. (The former could potentially
	 * cause one to need a shift/mask on EVERY BYTE.)
	 *
	 * We're counting on stdio to do buffering, which is probably insane.
	 */
	ERR_PASS( w_pixrGetImage (pIPWState, pixr, &imageP));
	
	for (i = 0; i < h; i++)
	{
	    ERR_GEN(fwrite (imageP, 1, bytesPerFileLine, fp)
		    != bytesPerFileLine,
		    cErrIO_WriteError);
	    
	    imageP += lineDelta;
	}
    }
    /* for depth 2 and 4, we will still have to write out one byte per pixel
     * to go along with the PGM spec...
     */
    else if (depth == 2  ||  depth == 4)
    {
	Int32   j, k;

	ERR_PASS( w_pixrGetImage (pIPWState, pixr, &imageP));
	bytesPerFileLine = w;

	/* Set up the buffer to hold 1 row of the image at a time, that is,
	 * 1 byte per pixel in the row.
	 * We'll add three extra to the buffer in case our image width
	 * ends on the beginning of a byte in a depth 2 case.
	 * (It would mean there are 3 pixels of padding in the rest of the
	 *  byte that will get written into the line buffer array, which
	 *  wouldn't actually be written out to fp, but will save an extra
	 *  check below when filling the line buffer.
	 */
	pLineBuffer = (UInt8 *)IPW_CALLOC(bytesPerFileLine + 3,
					  sizeof(UInt8), cAllocatorData);
	ERR_GEN(pLineBuffer == NULL, cErrNomem);

	if (depth == 2)
	{
	    fprintf(fp, "P5\n%ld %ld\n3\n", w, h);

	    for (i = 0; i < h; i++)
	    {
		for (j = k = 0; j < w; k++)
		{
		    pLineBuffer[j++] =  imageP[k] >> 6;
		    pLineBuffer[j++] = (imageP[k] >> 4) & 0x03;
		    pLineBuffer[j++] = (imageP[k] >> 2) & 0x03;
		    pLineBuffer[j++] =  imageP[k]       & 0x03;
		}
		
		ERR_GEN_CLEANUP(fwrite (pLineBuffer, 1, bytesPerFileLine, fp)
				!= bytesPerFileLine,
				cErrIO_WriteError);
		
		imageP += lineDelta;
	    }
	}
	else /* depth == 4 */
	{
	    fprintf(fp, "P5\n%ld %ld\n15\n", w, h);

	    for (i = 0; i < h; i++)
	    {
		for (j = k = 0; j < w; k++)
		{
		    pLineBuffer[j++] = imageP[k] >> 4;
		    pLineBuffer[j++] = imageP[k] & 0x0F;
		}
		
		ERR_GEN_CLEANUP(fwrite (pLineBuffer, 1, bytesPerFileLine, fp)
				!= bytesPerFileLine,
				cErrIO_WriteError);
		
		imageP += lineDelta;
	    }
	}

	IPW_FREE(pLineBuffer, cAllocatorData);
    }
    else if (depth == 8  &&  channels == 3)
    {
	/* 24-bit color RGB PIXR to output to PPM */
	UInt8 *pR, *pG, *pB;
	Int32  byteDiff, x;
	PIXR  *subR, *subG, *subB;
	    
	fprintf(fp, "P6\n%ld %ld\n255\n", w, h);
	bytesPerFileLine = 3 * w;

	/* Set up the buffer to hold 1 row of the image at a time, that is,
	 * 3 bytes of R,G,B data per pixel in the row */
	pLineBuffer = (UInt8 *)IPW_CALLOC(bytesPerFileLine,
					  sizeof(UInt8), cAllocatorData);
	ERR_GEN(pLineBuffer == NULL, cErrNomem);

	/* Take the 3 8-bit planes and pack them together line by line,
	 * with byte values for a pixel going left to right R,G,B */
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr,
					   cRedPlane, &subR));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, 
					   cGreenPlane, &subG));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, 
					   cBluePlane, &subB));
	    
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subR, &pR));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subG, &pG));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subB, &pB));
	
	byteDiff = lineDelta - w;

	for (i = 0; i < h; i++) { 
	    /* Fill in contiguous byte buffer */
	    for (x = 0; x < bytesPerFileLine; ) {
		pLineBuffer[x++] = *(pR++);
		pLineBuffer[x++] = *(pG++);
		pLineBuffer[x++] = *(pB++);
	    }

	    ERR_GEN_CLEANUP(fwrite (pLineBuffer, 1, bytesPerFileLine, fp)
			    != bytesPerFileLine, cErrIO_WriteError);

	    /* Move to next rows */
	    pR += byteDiff;
	    pG += byteDiff;
	    pB += byteDiff;
	}
    }

    ERR_PASS_CLEANUP( w_flip(pIPWState, pixr));
	
    IPW_FREE(pLineBuffer, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pLineBuffer, cAllocatorData);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write rectangular portion of PIXR to an image file from a file name.
 *
 * LongDescription:::
 *  Can write the following file formats.  
 *    -[X]-Y:  X is the allowable number planes/channels followed by the
 *             list of allowable depths with that number of planes.<BR>
 *    TIFF		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PS		-[1]-1                  <BR>
 *    SUNRAS		-[1]-1,2,4,8            <BR>
 *    BMP		-[1]-1,  4,8  -[3]-8    <BR>
 *    PNM		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PCX		-[1]-1,  4,8            <BR>
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  filename::       Ptr to filename string.
 *                   If it is "-", write to stdout.
 *  prs::            Ptr to PIXR with standard padding to write out.
 *  rect::           Ptr to rectangle telling which portion of
 *                   PIXR to write.  Will crop out-of-bound rectangles.
 *                   If entire rectangle is out-of-bounds, will write nothing.
 *                   If rectangle ptr is NULL, will write entire image.
 *  format::         Desired type of file format that image should be written
 *                   to.
 *
 * Returns:::
 *
 * Notes:::
 *  formats::
 *   Not all PIXR types and depths can be written using each file format.
 *   For complete list of possibilities, see imagerep.h.
 *
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *   [There happens to be no persistent memory allocated in this routine].
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, cropping, compression, Packbits, CCITT
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeRect (
    IPWState           *pIPWState,
    char               *filename,
    PIXR               *prs,
    RECTANGL           *rect,
    Int32               format)
{
    PROLOGUE;
    PIXR          *prt = NULL;
    RECTANGL      *clipRect = NULL;
    Int32          depth, channels;

    ERR_GEN(filename == NULL, cErrNullPtr);
    ERR_GEN_PIXR(prs, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, prs, &depth));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prs, &channels));
    ERR_GEN(!((channels == 1  &&  (depth == 1 || depth == 2 || depth == 4
				   || depth == 8)) ||
	      (channels == 3  &&  depth == 8)), cErrTypeNotSupported);


    if (rect != NULL)  /* use the rect */
    {
	Int32   w, h;
	
	ERR_PASS( w_createRect(pIPWState, rect->x, rect->y, rect->w, rect->h,
			       &clipRect));
	ERR_PASS_CLEANUP( w_rectOf(pIPWState, prs, clipRect, &prt));

	ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState, prt, &w));
	ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState, prt, &h));

	if (w > 0  &&  h > 0) 
	{
	    ERR_PASS_CLEANUP( w_write(pIPWState, filename, prt, format));
	}
	
	ERR_PASS_CLEANUP( w_destroyRect(pIPWState, &clipRect));
	ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prt));
    }
    else  /* dump the whole image */
	ERR_PASS( w_write(pIPWState, filename, prs, format));

    NORMAL_RETURN;
    
ERR_CLEANUP:
    w_destroyRect(pIPWState, &clipRect);
    w_destroyPixr(pIPWState, &prt);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to an image file from a file name.
 *
 * LongDescription:::
 *  Can write the following file formats.  
 *    -[X]-Y:  X is the allowable number planes/channels followed by the
 *             list of allowable depths with that number of planes.<BR>
 *    TIFF		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PS		-[1]-1                  <BR>
 *    SUNRAS		-[1]-1,2,4,8            <BR>
 *    BMP		-[1]-1,  4,8  -[3]-8    <BR>
 *    PNM		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PCX		-[1]-1  ,4,8            <BR>
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  filename::       Ptr to filename string.
 *                   If it is "-" and NOT using TIFF format, write to stdout.
 *  pixr::           Ptr to PIXR with standard padding to write out.
 *  format::         Desired type of file format that image should be written
 *                   to.
 *
 * Returns:::
 *
 * Notes:::
 *  formats::
 *   Not all PIXR types and depths can be written using each file format.
 *   For complete list of possibilities, see imagerep.h.
 *
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *   [There happens to be no persistent memory allocated in this routine].
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, compression, Packbits, CCITT
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *  ejaquith::  major rework
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_write (
    IPWState           *pIPWState,
    char               *filename,
    PIXR               *pixr,
    Int32               format)
{
    PROLOGUE;
    Int32          tmp, depth, channels;
    FILE          *fp;
    Bool           singleStrip = cFalse;
    UInt16	   options = 0;	/* used in TIFF Group 3 */
    
    ERR_GEN(filename == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, pixr, &depth));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));
    ERR_GEN(!((channels == 1  &&  (depth == 1 || depth == 2 || depth == 4
				   || depth == 8)) ||
	      (channels == 3  &&  depth == 8)), cErrTypeNotSupported);

    if ( !strcmp("-", filename) ) 
    {
	switch (format)
	{
	case TIFF_DEFAULT:
	case TIFF_UNCOMP:
	case TIFF_LZW:
	case TIFF_G3:
	case TIFF_G3_1LINE:
	case TIFF_G4:
	case TIFF_G4_1LINE:
	case TIFF_PACKBITS:
	case TIFF_RLE:
	case TIFF_RLE_1LINE:
	case TIFF_JPEG:
	case TIFF_LOSSYDITHER:

	case TIFF_G3_2D:
	case TIFF_G3_PAD:
	case TIFF_G3_2D_PAD:

	case TIFF_DEFAULT_SS:
	case TIFF_UNCOMP_SS:
	case TIFF_LZW_SS:
	case TIFF_G3_SS:
	case TIFF_G4_SS:
	case TIFF_PACKBITS_SS:
	case TIFF_RLE_SS:
	case TIFF_JPEG_SS:
	case TIFF_LOSSYDITHER_SS:
	    /* Can't write to stdout in TIFF format! */
	    ERR_GEN(1, cErrInvalidParm);
	    break;
	default:
	    fp = stdout;
	    break;
	}
    }
    else
    {
	/* decide whether to open file as binary or ascii.
	*/
	switch (format)
	{
	    case SUNRAS:
	    case SUNRAS_COMP:
	    case FF_BMP:
	    case PNM:
	    case PCX:
		fp = fopen(filename, "w+b");
		ERR_GEN_CLEANUP(fp == NULL, cErrIO_WriteError);
		break;

	    case TIFF_DEFAULT:
	    case TIFF_UNCOMP:
	    case TIFF_LZW:
	    case TIFF_G3:
	    case TIFF_G3_1LINE:
	    case TIFF_G4:
	    case TIFF_G4_1LINE:
	    case TIFF_PACKBITS:
	    case TIFF_RLE:
	    case TIFF_RLE_1LINE:
	    case TIFF_LOSSYDITHER:
		/* JPEG compression TAGs done, but the actual routines
		 * for de/compression are for us to write.
		 * Thus, this format is temporarily not supported.
		 * --ejaquith--oct94
	    case TIFF_JPEG:
	    */
		fp = stdout;  /* Done to allow correct calls below */
		break;

	    case TIFF_G3_2D:
		fp = stdout;  /* Done to allow correct calls below */
		options = GROUP3OPT_2DENCODING;
		break;

	    case TIFF_G3_PAD:
		    fp = stdout;  /* Done to allow correct calls below */
		    options = GROUP3OPT_FILLBITS;
		    break;
    
	    case TIFF_G3_2D_PAD:
		    fp = stdout;  /* Done to allow correct calls below */
		    options = GROUP3OPT_2DENCODING | GROUP3OPT_FILLBITS;
		    break;

	    case TIFF_DEFAULT_SS:
	    case TIFF_UNCOMP_SS:
	    case TIFF_LZW_SS:
	    case TIFF_G3_SS:
	    case TIFF_G4_SS:
	    case TIFF_PACKBITS_SS:
	    case TIFF_RLE_SS:
	    case TIFF_LOSSYDITHER_SS:
		/* JPEG compression TAGs done, but the actual routines
		 * for de/compression are for us to write.
		 * Thus, this format is temporarily not supported.
		 * --ejaquith--oct94
	    case TIFF_JPEG_SS:
	    */
		singleStrip = cTrue;
		fp = stdout;  /* Done to allow correct calls below */
		break;
		
	    case PS:  /* PostScript -- ascii file type */
		fp = fopen(filename, "w+");
		ERR_GEN_CLEANUP(fp == NULL, cErrIO_WriteError);
		break;
    
	    default:
		fp = NULL;
		ERR_GEN_CLEANUP(1, cErrBadType);
		break;
	}
    }	/* end of case where filename != "-" */

    if (fp != NULL)  /* if we opened a stream, write to it */
    {
	switch(format)
	{
	    case SUNRAS:
		ERR_PASS_CLEANUP( wp_fwriteSun(pIPWState, fp, pixr,
					       NOT_COMPRESSED));
		break;
	
	    case SUNRAS_COMP:
		ERR_PASS_CLEANUP( wp_fwriteSun(pIPWState, fp, pixr,
					       COMPRESSED));
		break;
	
	    case FF_BMP:
		ERR_PASS_CLEANUP( wp_fwriteBMP(pIPWState, fp, pixr));
		break;
	    case PS:
		ERR_PASS_CLEANUP( wp_fwritePS(pIPWState, fp, pixr));
		break;

	    case PNM:
		ERR_PASS_CLEANUP( wp_fwritePnm(pIPWState, fp, pixr));
		break;

	    case PCX:
		ERR_PASS_CLEANUP( wp_fwritePcx(pIPWState, fp, pixr));
		break;
		
	    case TIFF_DEFAULT:
	    case TIFF_DEFAULT_SS:
		ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &tmp));
		if ( tmp == 1 )
		    ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename,
						   COMPRESSION_CCITTFAX3,
						   singleStrip,
						   0));
		else
		    ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename,
						   COMPRESSION_LZW,
						   singleStrip,
						   0));
		fp = NULL;
		break;

	    case TIFF_UNCOMP:
	    case TIFF_UNCOMP_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename,
					       COMPRESSION_NONE,
					       singleStrip,
					       0));
		fp = NULL;
		break;

	    case TIFF_LZW:
	    case TIFF_LZW_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename, 
					       COMPRESSION_LZW,
					       singleStrip,
					       0));
		fp = NULL;
		break;

	    case TIFF_G3:
	    case TIFF_G3_1LINE:
	    case TIFF_G3_2D:
	    case TIFF_G3_PAD:
	    case TIFF_G3_2D_PAD:
	    case TIFF_G3_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename, 
					       COMPRESSION_CCITTFAX3,
					       singleStrip,
					       options));
		fp = NULL;
		break;

	    case TIFF_G4:
	    case TIFF_G4_1LINE:
	    case TIFF_G4_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename, 
					       COMPRESSION_CCITTFAX4,
					       singleStrip,
					       0));
		fp = NULL;
		break;
		
	    case TIFF_LOSSYDITHER:
	    case TIFF_LOSSYDITHER_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename, 
					       COMPRESSION_LOSSYDITHER,
					       singleStrip,
					       0));
		fp = NULL;
		break;
		
	    case TIFF_PACKBITS:
	    case TIFF_PACKBITS_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename, 
					       COMPRESSION_PACKBITS,
					       singleStrip,
					       0));
		fp = NULL;
		break;

	    case TIFF_RLE:
	    case TIFF_RLE_1LINE:
	    case TIFF_RLE_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename, 
					       COMPRESSION_CCITTRLE,
					       singleStrip,
					       0));
		fp = NULL;
		break;

		/* JPEG compression TAGs done, but the actual routines
		 * for de/compression are for us to write.
		 * Thus, this format is temporarily not supported.
		 * --ejaquith--oct94
	    case TIFF_JPEG:
	    case TIFF_JPEG_SS:
		ERR_PASS_CLEANUP( w_writeTiff( pIPWState, pixr, filename,
		                              COMPRESSION_JPEG, singleStrip,
					      0);
		fp = NULL;
		break;
		*/

	    default:
		ERR_GEN_CLEANUP(1, cErrBadType);
		break;
	}

	/* close the stream if it's a file */
	if (fp != NULL && fp != stdout)
	    fclose(fp);
    
    }  /* if we had a valid stream */

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (fp != NULL && fp != stdout)
	fclose(fp);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to an image file from a file name. Allow muti-page tiff output
 *
 * LongDescription:::
 *  Can write the following file formats.  
 *    -[X]-Y:  X is the allowable number planes/channels followed by the
 *             list of allowable depths with that number of planes.<BR>
 *    TIFF		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PS		-[1]-1                  <BR>
 *    SUNRAS		-[1]-1,2,4,8            <BR>
 *    BMP		-[1]-1,  4,8  -[3]-8    <BR>
 *    PNM		-[1]-1,2,4,8  -[3]-8    <BR>
 *    PCX		-[1]-1  ,4,8            <BR>
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  filename::       Ptr to filename string.
 *                   If it is "-" and NOT using TIFF format, write to stdout.
 *  pixr::           Ptr to PIXR with standard padding to write out.
 *  format::         Desired type of file format that image should be written
 *                   to.
 *  tiff::           Ptr to TIFF structure
 *  firstPage::      True if this is first page
 *  lastPage::       True if this is last Page

 * Returns:::
 *
 * Notes:::
 * firstPage::
 * lastPage::
 *  For pages in between first and last page These two values should all be
 *  CFalse. For single page image, these two values should all be CTrue.
 *  formats::
 *   Not all PIXR types and depths can be written using each file format.
 *   For complete list of possibilities, see imagerep.h.
 *
 *  Warning!  TIFF temporary allocation/freeing::
 *   This routine uses Leffler's TIFF package for TIFF i/o.
 *   It does not allow for us to use pIPWState for _temporary_ 
 *   allocation/freeing, and uses the standard system calls.  
 *   However, any _persistent_ memory does use pIPWState allocation.
 *   [There happens to be no persistent memory allocated in this routine].
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color, multi-channel, compression, Packbits, CCITT
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 * jlin
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_writeEx (
    IPWState           *pIPWState,
    char               *filename,
    PIXR               *pixr,
    Int32               format,
    TIFF               **tif,
    Bool               firstPage,
    Bool               lastPage)
{
    PROLOGUE;
    Int32          tmp, depth, channels;
    FILE          *fp;
    Bool           singleStrip = cFalse;
    UInt16	   options = 0;	/* used in TIFF Group 3 */
    
    ERR_GEN(filename == NULL, cErrNullPtr);
    ERR_GEN(tif == NULL, cErrNullPtr);
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, pixr, &depth));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));
    ERR_GEN(!((channels == 1  &&  (depth == 1 || depth == 2 || depth == 4
				   || depth == 8)) ||
	      (channels == 3  &&  depth == 8)), cErrTypeNotSupported);

    if ( !strcmp("-", filename) ) 
    {
	switch (format)
	{
	case TIFF_DEFAULT:
	case TIFF_UNCOMP:
	case TIFF_LZW:
	case TIFF_G3:
	case TIFF_G3_1LINE:
	case TIFF_G4:
	case TIFF_G4_1LINE:
	case TIFF_PACKBITS:
	case TIFF_RLE:
	case TIFF_RLE_1LINE:
	case TIFF_JPEG:
	case TIFF_LOSSYDITHER:

	case TIFF_G3_2D:
	case TIFF_G3_PAD:
	case TIFF_G3_2D_PAD:

	case TIFF_DEFAULT_SS:
	case TIFF_UNCOMP_SS:
	case TIFF_LZW_SS:
	case TIFF_G3_SS:
	case TIFF_G4_SS:
	case TIFF_PACKBITS_SS:
	case TIFF_RLE_SS:
	case TIFF_JPEG_SS:
	case TIFF_LOSSYDITHER_SS:
	    /* Can't write to stdout in TIFF format! */
	    ERR_GEN(1, cErrInvalidParm);
	    break;
	default:
	    fp = stdout;
	    break;
	}
    }
    else
    {
	/* decide whether to open file as binary or ascii.
	*/
	switch (format)
	{
	    case SUNRAS:
	    case SUNRAS_COMP:
	    case FF_BMP:
	    case PNM:
	    case PCX:
		fp = fopen(filename, "w+b");
		ERR_GEN_CLEANUP(fp == NULL, cErrIO_WriteError);
		break;

	    case TIFF_DEFAULT:
	    case TIFF_UNCOMP:
	    case TIFF_LZW:
	    case TIFF_G3:
	    case TIFF_G3_1LINE:
	    case TIFF_G4:
	    case TIFF_G4_1LINE:
	    case TIFF_PACKBITS:
	    case TIFF_RLE:
	    case TIFF_RLE_1LINE:
	    case TIFF_LOSSYDITHER:
		/* JPEG compression TAGs done, but the actual routines
		 * for de/compression are for us to write.
		 * Thus, this format is temporarily not supported.
		 * --ejaquith--oct94
	    case TIFF_JPEG:
	    */
		fp = stdout;  /* Done to allow correct calls below */
		break;

	    case TIFF_G3_2D:
		fp = stdout;  /* Done to allow correct calls below */
		options = GROUP3OPT_2DENCODING;
		break;

	    case TIFF_G3_PAD:
		    fp = stdout;  /* Done to allow correct calls below */
		    options = GROUP3OPT_FILLBITS;
		    break;
    
	    case TIFF_G3_2D_PAD:
		    fp = stdout;  /* Done to allow correct calls below */
		    options = GROUP3OPT_2DENCODING | GROUP3OPT_FILLBITS;
		    break;

	    case TIFF_DEFAULT_SS:
	    case TIFF_UNCOMP_SS:
	    case TIFF_LZW_SS:
	    case TIFF_G3_SS:
	    case TIFF_G4_SS:
	    case TIFF_PACKBITS_SS:
	    case TIFF_RLE_SS:
	    case TIFF_LOSSYDITHER_SS:
		/* JPEG compression TAGs done, but the actual routines
		 * for de/compression are for us to write.
		 * Thus, this format is temporarily not supported.
		 * --ejaquith--oct94
	    case TIFF_JPEG_SS:
	    */
		singleStrip = cTrue;
		fp = stdout;  /* Done to allow correct calls below */
		break;
		
	    case PS:  /* PostScript -- ascii file type */
		fp = fopen(filename, "w+");
		ERR_GEN_CLEANUP(fp == NULL, cErrIO_WriteError);
		break;
    
	    default:
		fp = NULL;
		ERR_GEN_CLEANUP(1, cErrBadType);
		break;
	}
    }	/* end of case where filename != "-" */

    if (fp != NULL)  /* if we opened a stream, write to it */
    {
	switch(format)
	{
	    case SUNRAS:
		ERR_PASS_CLEANUP( wp_fwriteSun(pIPWState, fp, pixr,
					       NOT_COMPRESSED));
		break;
	
	    case SUNRAS_COMP:
		ERR_PASS_CLEANUP( wp_fwriteSun(pIPWState, fp, pixr,
					       COMPRESSED));
		break;
	
	    case FF_BMP:
		ERR_PASS_CLEANUP( wp_fwriteBMP(pIPWState, fp, pixr));
		break;
	    case PS:
		ERR_PASS_CLEANUP( wp_fwritePS(pIPWState, fp, pixr));
		break;

	    case PNM:
		ERR_PASS_CLEANUP( wp_fwritePnm(pIPWState, fp, pixr));
		break;

	    case PCX:
		ERR_PASS_CLEANUP( wp_fwritePcx(pIPWState, fp, pixr));
		break;
		
	    case TIFF_DEFAULT:
	    case TIFF_DEFAULT_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &tmp));
		if ( tmp == 1 )
		    ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
						   COMPRESSION_CCITTFAX3,
						   singleStrip,
						   0, *tif));
		else
		    ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
						   COMPRESSION_LZW,
						   singleStrip,
						   0, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		break;

	    case TIFF_UNCOMP:
	    case TIFF_UNCOMP_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
					       COMPRESSION_NONE,
					       singleStrip,
					       0, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;

	    case TIFF_LZW:
	    case TIFF_LZW_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr, 
                                              COMPRESSION_LZW,
					       singleStrip,
					       0, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;

	    case TIFF_G3:
	    case TIFF_G3_1LINE:
	    case TIFF_G3_2D:
	    case TIFF_G3_PAD:
	    case TIFF_G3_2D_PAD:
	    case TIFF_G3_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
					       COMPRESSION_CCITTFAX3,
					       singleStrip, options, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;

	    case TIFF_G4:
	    case TIFF_G4_1LINE:
	    case TIFF_G4_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
					       COMPRESSION_CCITTFAX4,
					       singleStrip,
					       0, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;
		
	    case TIFF_LOSSYDITHER:
	    case TIFF_LOSSYDITHER_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,
                                                        filename,tif)); 
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
					       COMPRESSION_LOSSYDITHER,
					       singleStrip,
					       0, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;
		
	    case TIFF_PACKBITS:
	    case TIFF_PACKBITS_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
					       COMPRESSION_PACKBITS,
					       singleStrip,
					       0, *tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;

	    case TIFF_RLE:
	    case TIFF_RLE_1LINE:
	    case TIFF_RLE_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
					       COMPRESSION_CCITTRLE,
					       singleStrip,
					       0,*tif));
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;

		/* JPEG compression TAGs done, but the actual routines
		 * for de/compression are for us to write.
		 * Thus, this format is temporarily not supported.
		 * --ejaquith--oct94
	    case TIFF_JPEG:
	    case TIFF_JPEG_SS:
                if(firstPage){
                     ERR_PASS_CLEANUP(w_writeTiffHeader(pIPWState,filename,tif));
                }
		ERR_PASS_CLEANUP( w_writeTiffDirectory( pIPWState, pixr,
		                              COMPRESSION_JPEG, singleStrip,
					      0, *tif);
                if(lastPage){
		   finishTiff(*tif);
                }
		fp = NULL;
		fp = NULL;
		break;
		*/

	    default:
		ERR_GEN_CLEANUP(1, cErrBadType);
		break;
	}

	/* close the stream if it's a file */
	if (fp != NULL && fp != stdout)
	    fclose(fp);
    
    }  /* if we had a valid stream */

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (fp != NULL && fp != stdout)
	fclose(fp);

    ERR_RETURN;
}

/*

 $Log: writfile.c,v $
 Revision 1.5  2000/07/08 02:14:15  jlin
 comment change

 Revision 1.4  2000/06/28 00:14:25  jlin
 lastPage and firstPage can both be true for one page document

 Revision 1.2  2000/06/23 23:24:22  jlin
 Add function to save multi-page tif image

 Revision 1.1.1.1  1999/09/20 18:19:21  jlin
 Imported sources

 Revision 1.41  1997/06/30 23:13:01  ddavies
 writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v

 Delete inclusion of Sandpiper's internal.h
 Remove ability to write AIS files.  That used the Sandpiper file header
 from internal.h.

 Revision 1.40  1997/03/17 20:20:17  ddavies
 writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v

 Add TIFF_LOSSYDITHER and TIFF_LOSSYDITHER_SS as compression types.

 Revision 1.39  1996/05/24 04:47:00  ddavies
 writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v

 Include shrcc.pub and iotiff.pub.

 Revision 1.38  1996/05/16 17:37:28  ddavies
 writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v

 Add TIFF_G3_1LINE, TIFF_G4_1LINE and TIFF_RLE_1LINE to file types that
 are equivalent to TIFF_G3, TIFF_G4 and TIFF_RLE respectively.  These new
 file types are used for testing reads only.

 Revision 1.37  1996/05/11 01:40:41  ddavies
 writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v

  Calculate the Group 3 options from the new TIFF_G3_2D, TIFF_G3_PAD and
 TIFF_G3_2D_PAD file types.  Add the options parameter to all calls to
 w_writeTiff.  Set it to 0 unless we're really doing Group 3.

 Revision 1.36  1996/01/04 20:29:31  ejaquith
 writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.35  1995/12/13  04:16:22  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.34  1995/11/13  20:02:44  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added needed additional inclusions of "ptypes.h", "gdefs.h", and "pcx.h".
 * Updated PROLOGUE usage.
 *
 * Revision 1.33  1995/10/13  22:08:21  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.32  1995/10/13  02:17:48  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Plugged some memory leaks that were occuring when writing out PNM files.
 * Fixed PNM depth 1 and depth 8 reading bug on PC where the image data
 * was not getting byteswapped before (and of course then after) writing
 * out the image data.
 *
 * Revision 1.31  1995/10/12  04:29:39  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.30  1995/10/10  02:30:03  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added inclusion of "stdio.h"
 * Bug fixes.
 * PNM writing was not writing depth 2 and depth 4 images correctly.
 * These formats are saved (get this!) at one pixel/byte by the spec, so
 * I had to add a buffer to unpack our PIXRs into line by line and then
 * dump out.  Set maximum number of gray colors in file according to
 * depth -- depth 2 ==> 3, depth 4 ==> 15, depth 8 ==> 255.
 * This will allow us to read PNM files of depth 2 and 4 seamlessly while
 * still obeying the PNM spec.
 * Added file depth and channel checking to only allow valid PNM files to be
 * written.
 * Added special comments in comment blocks to all public i/o routines
 * to allow user to know with certainty what depths/planes we support!
 * Fixed PNM writing of color RGB 24 PIXRs.  Before we were only writing
 * out the red plane three times (typo where the blue and green planes
 * were switched with the red plane!), and since our IPWrap i/o testing
 * was bogus all along (more on that later) we never detected this.
 * Fixed bug in writing AIS files -- we were accepting RGB pixrs
 * and attempting to "write" them.  what we were in fact doing was writing
 * out the primary (green) channel as 8-bit gray!  now AIS does not
 * support RGB i/o.
 *
 * Revision 1.29  1995/09/29  00:35:23  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.28  1995/09/21  23:01:17  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.27  1995/09/12  18:29:13  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * vanquished djfax stuff.
 *
 * Revision 1.26  1995/09/12  16:55:00  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.25  1995/09/08  00:14:58  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.24  1995/09/07  01:37:57  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.23  1995/09/06  20:26:51  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.22  1995/09/01  02:20:44  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.21  1995/08/31  21:40:55  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.20  1995/08/29  21:57:51  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.19  1995/08/22  02:23:33  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Nailed some PC warnings
 *
 * Revision 1.18  1995/08/17  22:01:28  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.17  1995/03/21  04:40:25  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added CCITTRLE format so we can write files in Modified Huffman format.
 *
 * Revision 1.16  1994/12/19  23:23:40  ejaquith
 * ejaquith on Mon Dec 19 15:22:31 PST 1994
 *
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added capability to write out PPM format file in the meta PNM format
 * for outputting PIXRs.  This will allow 24-bit RGB images to be written
 * out.
 *
 * Revision 1.15  1994/11/02  01:42:21  ejaquith
 * ejaquith on Tue Nov 1 17:39:59 PST 1994
 *
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Removed the Perfect Scan new TIFF library anticipated remarks and
 * #ifdefs.  For now, the new Leffler stuff seems very nice, bug free,
 * smaller than the Perfect Scan routines, and more robust.
 * Changed calls to writePrTiff() to use the compression desired field
 * as macroed integers instead of strings.
 *
 * Revision 1.14  1994/10/31  22:56:56  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Changed the fwritePr() routine to always return the status from the
 * writing procedure itself, in case of errors.  Previously, we always
 * returned no errors as long as the file stream opened successfully,
 * but completely ignored the results of the actual writing routines!
 *
 * Revision 1.13  1994/10/21  01:19:07  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added PCX format writing capability.  Fairly untested, suggested to use
 * only for binary images
 *
 * Revision 1.12  1994/10/10  22:36:08  lperry
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Change to use new header file io.pub instead of iopros.h
 *
 * Revision 1.11  1994/10/10  17:20:50  ejaquith
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added framework for new TIFF library writing routines.  ThWe are
 * (argh!) the original Leffler TIFF code for now.  Added framework for PCX
 * writing format.  The actual write routine is not called yet as bug fixes are
 * needed first.
 * Changed the TIFF current TIFF write calls to invoke fwritePr() and then do
 * the appropriate write calls from there.
 * When the new TIFF library is in place, JPEG compression will be a TIFF
 * compress format.
 * Expect to add bug fixed new TIFF library 11/94.
 *
 * Revision 1.10  1994/08/26  01:45:39  lperry
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Got rid of some globals in the AIE write routines
 * to achieve reentrancy.
 *
 * Revision 1.9  1994/08/24  00:11:11  macleod
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Changed "<" to "!=" to shut up Sun "C"
 *
 * Revision 1.8  1994/08/16  21:54:13  macleod
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Added support for PGM ("P5") and PBM ("P4") file types, a la PBMPLUS.
 * Note that we only support the "RAWBITS" formats, not the ascii formats.
 *
 * Revision 1.7  1994/05/26  19:29:04  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Delete the #ifndef PRODUCTION clauses around the body of fwritePS.  These
 * were here because of fear that the ASCII strings written to a Postscript file
 * would allow someone with a disassembler to deduce the structure of our prosaic
 * PIXR data structure.  If anyone decides to ship this code in a product
 * and they care, we'll revisit the issue.
 *
 * Revision 1.6  1994/05/05  02:01:41  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Initialize status in fwritePr so we can tell the difference between
 * success and failure.  Cahnge lots of %d's to %ld's to keep gcc happy.
 *
 * Revision 1.5  1994/01/12  04:08:22  ddavies
 * ddavies on Tue Jan 11 20:05:00 PST 1994
 *
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Fix several bugs: 1. toss the check to see if fp is defined at the top of
 * fwritePr.  fp won't be defined until the file is opened below.
 * 2. Make the glob of code that opens files in different formats into the
 * "else" clause of the code that sets fp to stdout if filename == "-".
 * 3. Change file modes from wb+ and rb+ to w+b and r+b because "w+" and "r+"
 * are atomic units.
 *
 * Revision 1.4  1994/01/12  01:18:31  dferrell
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Pushed file opening mode (ascii/binary) stuff down from writePR into
 * fwritePR for simplicity in dealing with PC weirdness with these two
 * file types.
 *
 * Revision 1.3  1993/12/03  01:20:42  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:19:28  ddavies
 * writfile.c -> /product/ipcore/ipwrap/src/RCS/writfile.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

