/*<HTML><HEAD><TITLE>wshed.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  wshed.c:  Watershed "flooding" algorithm
 *
 *	w_watershedSegment ()
 *
 */

#define DEBUG_F 0
#define DEBUG_T 0
#define DEBUG_P 0

#if DEBUG_F || DEBUG_T || DEBUG_P
#include "stdio.h"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#include "morphg.h"
#include "morph.pub"
#include "strel.pub"
#include "logicops.h"
#include "logicops.pub"
#include "wshed.pub"
#include "shrapp.pub"
#include "imageref.pub"

#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "shrcc.pub"

#if DEBUG_F || DEBUG_T
#include "wshed.prv"
#include "imagerep.h"
#include "io.pub"
#endif

#if DEBUG_P
#include "shrpixr.pub"
#endif



IP_RCSINFO(RCSInfo, "$RCSfile: wshed.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Wrapper for 8-bit watershed boundary finding algorithm
 *
 * LongDescription:::
 *  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state pointer
 *  inPr::           Input image to be processed (must be 8 bit)
 *  minimaPr::       Minima marker image (must be binary)
 *  connectType::    0 = 4 neighbor watershed search, 
 *                   1 = 8 neighbor watershed search
 *
 * Output:::
 *  outPr::          Resulting binary watershed boundary image where
 *                   watershed boundaries are black (1) and all else
 *                   is white (0).
 *
 * Returns:::
 *  cSuccessful      if it worked
 *  cErrInvalidParm  if connectType specifier is invalid
 *	
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_watershedSegment (
    IPWState *pIPWState, 
    PIXR     *inPr,
    PIXR     *minimaPr,
    Int32     connectType,
    PIXR    **outPr)
{
    PROLOGUE;
    Int32     wS, hS, bplS,
              wE, hE, bplE,
	      wM, hM, bplM,
	      wD, hD, bplD,
	      wT, hT, bplT,
	      tagPlaneBytes,
              rightColsToClear,
	      bottomRowsToClear;
    Bool      allocatedHere = cFalse;
    UInt8    *iS, *iM, *iD, *iE;
    UInt16   *iT = NULL;
    PIXR     *edgePr = NULL;
    STREL    *eStrel = NULL;

    ERR_GEN (inPr == NULL, cErrNullPtr);
    ERR_GEN (minimaPr == NULL, cErrNullPtr);
    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    ERR_GEN_PIXR (minimaPr, cByteOrderBig, cDepth1, cChannels1, cFrame32);


    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &wT));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &hT));
    ERR_PASS (w_pixrGetWidth (pIPWState, minimaPr, &wM));
    ERR_PASS (w_pixrGetHeight (pIPWState, minimaPr, &hM));

    ERR_GEN ((wT != wM) || (hT != hM), cErrInvalidParm);

    /* make the result pixr if necessary */
    ERR_PASS (wp_clonePixrConditional (pIPWState, minimaPr, outPr,
				       &rightColsToClear, 
				       &bottomRowsToClear,
				       &allocatedHere));

    /*** these steps need to somehow get moved into ipshared */

    /* make the edge pixr */
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, minimaPr, &edgePr));

    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 3, 3, 1, 1, &eStrel));
    if (connectType == 0)
    {
        /* make it a disc strel for 4-connected */
	eStrel->data[0][0] = DONT_CARE;
	eStrel->data[0][2] = DONT_CARE;
	eStrel->data[2][0] = DONT_CARE;
	eStrel->data[2][2] = DONT_CARE;
    }

    ERR_PASS_CLEANUP (w_dilate (pIPWState, minimaPr, eStrel, &edgePr));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, minimaPr, XOR, edgePr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &eStrel));

    /*** end of stuff that should be in ipshared */

#if DEBUG_P
    fprintf (stdout, "Find minima boundaries: %f\n", StopTime ());
#endif

#if DEBUG_F
    ERR_IGNORE (w_write (pIPWState, "edge", edgePr, TIFF_G4));
#endif

    /* find out how big the tag plane is */
    ERR_PASS_CLEANUP (i_getWatershedTagPlaneSize (pIPWState, &wT, &hT, 
                                                  &bplT, &tagPlaneBytes));

    /* allocate the tag plane */
    iT = IPW_CALLOC (1, tagPlaneBytes, cAllocatorImage);
    ERR_GEN_CLEANUP (iT == NULL, cErrNomem);

    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, inPr, &wS));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, inPr, &hS));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, inPr, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, inPr, &iS));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, edgePr, &wE));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, edgePr, &hE));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, edgePr, &bplE));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, edgePr, &iE));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, minimaPr, &wM));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, minimaPr, &hM));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, minimaPr, &bplM));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, minimaPr, &iM));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, *outPr, &wD));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, *outPr, &hD));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, *outPr, &bplD));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, *outPr, &iD));

    /* produce the watershed boundaries */
#if DEBUG_T
    ERR_IGNORE (i_watershedSegment (pIPWState, connectType,
                                    iS, wS, hS, bplS,
                                    iE, wE, hE, bplE,
                                    iM, wM, hM, bplM,
				    iT, wT, hT, bplT,
				    iD, wD, hD, bplD));
#else
    ERR_PASS_CLEANUP (i_watershedSegment (pIPWState, connectType,
                                          iS, wS, hS, bplS,
                                          iE, wE, hE, bplE,
                                          iM, wM, hM, bplM,
					  iT, wT, hT, bplT,
					  iD, wD, hD, bplD));
#endif

#if DEBUG_T
    {
        Int32 r, c;
	UInt16 *ip;

	iS += (bplS << cLogFrameBits) + (8 * cFrameBytes);
	wS -= 2 * cFrameBits;
	hS -= 2 * cFrameBits;
	bplT = bplT / 2;
	ip = iT + (bplT * cWSTagPlaneFramePix) + cWSTagPlaneFramePix;

        for (r = 0; r < hS; r++)
	{
	    for (c = 0; c < wS; c++)
	    {
	        UCHAR_ACCESS (iS + (r * bplS) + c) = 
		          (UInt8)(USHORT_ACCESS (ip + (r * bplT) + c));
	    }
	}

    }

    ERR_IGNORE (w_write (pIPWState, "tags", inPr, TIFF_LZW));
#endif

    /* destroy the tag plane */
    IPW_FREE (iT, cAllocatorImage);
    iT = NULL;

    /* destroy the edge pixr */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &edgePr));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
        w_destroyPixr (pIPWState, outPr);

    /* destroy the edge pixr and the nasty strel used to create it */
    w_destroyPixr (pIPWState, &edgePr);
    w_destroyStrel (pIPWState, &eStrel);

    /* destroy the tag plane */
    if (iT != NULL)
        IPW_FREE (iT, cAllocatorImage);
    ERR_RETURN;

}  /* w_watershedSegment () */




/*
 * $Log: wshed.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:22  jlin
 * Imported sources
 *
 * Revision 1.6  1998/01/14 00:16:24  ddavies
 * wshed.c -> /product/ipcore/ipwrap/src/RCS/wshed.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.5  1997/04/15 22:55:49  ddavies
 * wshed.c -> /product/ipcore/ipwrap/src/RCS/wshed.c,v
 *
 * Use non-initialized PIXR creation where possible.
 * Change name of wp_clonePixrZeroConditional to wp_clonePixrConditional.
 *
 * Revision 1.4  1997/04/03 00:53:32  dferrell
 * dferrell on Wed Apr 2 16:53:01 PST 1997
 *
 * wshed.c -> /product/ipcore/ipwrap/src/RCS/wshed.c,v
 *
 * Removed unnecessary setting of resolution of input pixr.
 *
 * Revision 1.3  1996/07/26 00:49:16  dferrell
 * dferrell on Thu Jul 25 17:49:00 PDT 1996
 *
 * wshed.c -> /product/ipcore/ipwrap/src/RCS/wshed.c,v
 *
 * Added ERR_IGNORE macros to debug only calls.
 *
 * Revision 1.2  1996/07/24 01:36:56  dferrell
 * dferrell on Tue Jul 23 18:36:51 PDT 1996
 *
 * wshed.c -> /product/ipcore/ipwrap/src/RCS/wshed.c,v
 *
 * Fixed compiler warnings from NT build.
 *
 * Revision 1.1  1996/07/22 23:11:16  dferrell
 * Initial revision
 *
 *
 */

/*</PRE></BODY>*/
