/*************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * i_depth.c -- upper levels of procedures that change the depth of images
 *
 *     Grayscale thresholding:
 *           Int32     i_threshold()		optimized with assembler
 *           Int32     i_thresholdI()		optimized with assembler
 *	     Int32     i_threshold2To1bpp()	optimized with assembler
 *	     Int32     i_thresholdI2To1bpp()	optimized with assembler
 *	     Int32     i_threshold4To1bpp()	optimized with assembler
 *	     Int32     i_thresholdI4To1bpp()	optimized with assembler
 *           Int32     i_thresholdTo2bpp()	optimized with assembler
 *           Int32     i_threshold4To2bpp()
 *           Int32     i_highpassfilt()		optimized with assembler
 *	     Int32     i_highpassfilt4To1bpp()
 *           Int32     i_highpassfiltI()	optimized with assembler
 *	     Int32     i_highpassfiltI4To1bpp()
 *           Int32     i_passfilt()		optimized with assembler
 *           Int32     i_passfiltI()		optimized with assembler
 *	     Int32     i_passfilt4To1bpp()
 *	     Int32     i_passfiltI4To1bpp()
 *	     Int32     i_map4To4 ()
 *	     Int32     i_map8To8 ()
 *	     Int32     i_threshold8To4bpp ()
 *	     Int32     i_thresholdByPixel ()
 *
 *     Depth Upconverting:
 *	     Int32     i_depth4To8bpp ()
 *	     Int32     i_depth1To8bpp ()
 *	     Int32     i_depth1To16bpp ()
 *           Int32     i_map1To8bpp ()
 */



#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "frames.pub"
#include "memory.pub"
#include "except.prv"
#include "imageref.pub"
#include "defines.pub"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shros.pub"
#include "shrscal.pub"
#include "shrscal.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_depth.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

#define	cMSBitMask	0x80808080
#define cMSBitMask4bpp	0x88888888
#define cLSBitMask4bpp	0x11111111
#define	cLSBitCarry4bpp	0x22222222

/*-----------------------------------------------------------------------*
 *             Grayscale Thresholding -- Assembler Version               *
 *-----------------------------------------------------------------------*/

/* PUBLIC */
/*************************************************************************** 
 *  i_threshold()  
 *  		Takes an 8-bit grayscale image and returns
 *             	a thresholded binary image, using the most significant
 *             	bit in each 8-bit pixel as the corresponding binary pixel. *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		pTabBinarizeByte0,
 *		pTabBinarizeByte1,
 *		pTabBinarizeByte2,
 *		pTabBinarizeByte3 = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_threshold (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0,
    UInt32     *pTabBinarizeByte1,
    UInt32     *pTabBinarizeByte2,
    UInt32     *pTabBinarizeByte3)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    
    UInt32	tableShouldBeFreed;


    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_makeBinarizationTables (pIPState, &pTabBinarizeByte0,
						 &pTabBinarizeByte1,
						 &pTabBinarizeByte2,
						 &pTabBinarizeByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/4) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toBinary128Line (pIPState, pS, pD,
			    dstInitOffset, srcInitOffset, srcLimit,
			    pTabBinarizeByte0, pTabBinarizeByte1,
			    pTabBinarizeByte2, pTabBinarizeByte3);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0);
	FREE(pTabBinarizeByte1);
	FREE(pTabBinarizeByte2);
	FREE(pTabBinarizeByte3);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************** 
 *  i_thresholdI()
 *  		Takes an 8-bit grayscale image and returns
 *              a thresholded binary image, using the *inverse* of the
 *              most significant bit in each 8-bit pixel as the
 *              corresponding binary pixel.
 *              Returned image is inverse (0 <-> 1) of that returned by
 *              i_threshold. *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *     		sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTabBinarizeByte0Inv,
 *		pTabBinarizeByte1Inv,
 *		pTabBinarizeByte2Inv,
 *		pTabBinarizeByte3Inv = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeInvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_thresholdI (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0Inv,
    UInt32     *pTabBinarizeByte1Inv,
    UInt32     *pTabBinarizeByte2Inv,
    UInt32     *pTabBinarizeByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	  hOrig;		/* ditto */
    UInt32	 tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* remember for ip_clearRightFrame */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_makeInvBinarizationTables (pIPState,
						    &pTabBinarizeByte0Inv,
						    &pTabBinarizeByte1Inv,
						    &pTabBinarizeByte2Inv,
						    &pTabBinarizeByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/4) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toBinary128Line (pIPState, pS, pD,
			    dstInitOffset, srcInitOffset, srcLimit,
			    pTabBinarizeByte0Inv, pTabBinarizeByte1Inv,
			    pTabBinarizeByte2Inv, pTabBinarizeByte3Inv);
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0Inv);
	FREE(pTabBinarizeByte1Inv);
	FREE(pTabBinarizeByte2Inv);
	FREE(pTabBinarizeByte3Inv);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***************************************************************************  
 *  i_threshold2To1bpp()  
 *		Takes an 2-bit grayscale image and returns
 *              a thresholded 1 bit/pixel image.  The value of
 *              each pixel in the result image is the Most
 *              Significant bit of the corresponding pixel in
 *              the source image.
 *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination 
 *			image's frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTab2To1bppByte0,
 *		pTab2To1bppByte1,
 *		pTab2To1bppByte2,
 *		pTab2To1bppByte3 = four tables used for the operation.
 *				   If these args are NULL, NULL, NULL and NULL,
 *				   the tables will be allocated and freed
 *				   internally.  These tables can be obtained
 *				   from i_make2To1BinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes frame.
 **************************************************************************/
Int32  CDECL 
i_threshold2To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab2To1bppByte0,
    UInt32     *pTab2To1bppByte1,
    UInt32     *pTab2To1bppByte2,
    UInt32     *pTab2To1bppByte3)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    
    UInt32	tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 2*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab2To1bppByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
       if ((__error__ = i_make2To1BinarizationTables (pIPState,
						    &pTab2To1bppByte0,
						    &pTab2To1bppByte1,
						    &pTab2To1bppByte2,
						    &pTab2To1bppByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word.  The dest has 1-bit pixels. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 2 source words to make one destination words */
    srcInitOffset = dstInitOffset * 2;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.  We use 16 since there are
	 * 16 2-bit pixels per 32-bit word.
	 */
    srcLimit = ((w - 1)/16) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toBinaryFrom2bppLine (pIPState, pS, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTab2To1bppByte0, pTab2To1bppByte1,
				 pTab2To1bppByte2, pTab2To1bppByte3);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTab2To1bppByte0);
	FREE(pTab2To1bppByte1);
	FREE(pTab2To1bppByte2);
	FREE(pTab2To1bppByte3);
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/***************************************************************************
 * i_thresholdI2To1bpp()  
 *		Takes an 2-bit grayscale image and returns
 *              a thresholded 1 bit/pixel image.  The value of
 *              each pixel in the result image is the inverse of the Most
 *              Significant bit of the corresponding pixel in
 *              the source image.
 *              The result images are the complemented versions of
 *              those produced by i_threshold2To1bpp.  The
 *              complemented versions are more likely to match
 *              definitions of binary images (0 = white).
 *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination 
 *			image's frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTab2To1bppByte0Inv,
 *		pTab2To1bppByte1Inv,
 *		pTab2To1bppByte2Inv,
 *		pTab2To1bppByte3Inv = four tables used for the operation.
 *				   If these args are NULL, NULL, NULL and NULL,
 *				   the tables will be allocated and freed
 *				   internally.  These tables can be obtained
 *				   from i_make2To1InvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_thresholdI2To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab2To1bppByte0Inv,
    UInt32     *pTab2To1bppByte1Inv,
    UInt32     *pTab2To1bppByte2Inv,
    UInt32     *pTab2To1bppByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	 hOrig;			/* ditto */
    UInt32	 tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 2*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* remember for ip_clearRightFrame */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab2To1bppByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make2To1InvBinarizationTables (pIPState,
							&pTab2To1bppByte0Inv,
							&pTab2To1bppByte1Inv,
							&pTab2To1bppByte2Inv,
							&pTab2To1bppByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word.  The test has 1 bit pixels. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 2 source words to make one destination words */
    srcInitOffset = dstInitOffset * 2;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.  We use 16 since there are
	 * 16 2-bit pixels per 32-bit word.
	 */
    srcLimit = ((w - 1)/16) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toBinaryFrom2bppLine (pIPState, pS, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTab2To1bppByte0Inv, pTab2To1bppByte1Inv,
				 pTab2To1bppByte2Inv, pTab2To1bppByte3Inv);
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTab2To1bppByte0Inv);
	FREE(pTab2To1bppByte1Inv);
	FREE(pTab2To1bppByte2Inv);
	FREE(pTab2To1bppByte3Inv);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***************************************************************************  
 *  i_threshold4To1bpp()  
 *		Takes an 4-bit grayscale image and returns
 *              a thresholded 1 bit/pixel image.  The value of
 *              each pixel in the result image is the Most
 *              Significant bit of the corresponding pixel in
 *              the source image.
 *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination 
 *			image's frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTab4To1bppByte0,
 *		pTab4To1bppByte1,
 *		pTab4To1bppByte2,
 *		pTab4To1bppByte3 = four tables used for the operation.
 *				   If these args are NULL, NULL, NULL and NULL,
 *				   the tables will be allocated and freed
 *				   internally.  These tables can be obtained
 *				   from i_make4To1BinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes frame.
 **************************************************************************/
Int32  CDECL 
i_threshold4To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To1bppByte0,
    UInt32     *pTab4To1bppByte1,
    UInt32     *pTab4To1bppByte2,
    UInt32     *pTab4To1bppByte3)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    
    UInt32	tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To1bppByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To1BinarizationTables (pIPState,
						     &pTab4To1bppByte0,
						     &pTab4To1bppByte1,
						     &pTab4To1bppByte2,
						     &pTab4To1bppByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word.  The dest has 1 bpp pixels. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 4 source words to make one destination words */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.  We use 8 since there are
	 * 8 4-bit pixels per 32-bit word.
	 */
    srcLimit = ((w - 1)/8) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	/* This is the primitive routine that does the real work */
    ip_toBinaryFrom4bppLine (pIPState, pS, pD,
			     dstInitOffset, srcInitOffset, srcLimit,
			     pTab4To1bppByte0, pTab4To1bppByte1,
			     pTab4To1bppByte2, pTab4To1bppByte3);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTab4To1bppByte0);
	FREE(pTab4To1bppByte1);
	FREE(pTab4To1bppByte2);
	FREE(pTab4To1bppByte3);
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/***************************************************************************
 * i_thresholdI4To1bpp()  
 *		Takes an 4-bit grayscale image and returns
 *              a thresholded 1 bit/pixel image.  The value of
 *              each pixel in the result image is the inverse of the Most
 *              Significant bit of the corresponding pixel in
 *              the source image.
 *              The result images are the complemented versions of
 *              those produced by i_threshold4To1bpp.  The
 *              complemented versions are more likely to match
 *              definitions of binary images (0 = white).
 *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination 
 *			image's frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTab4To1bppByte0Inv,
 *		pTab4To1bppByte1Inv,
 *		pTab4To1bppByte2Inv,
 *		pTab4To1bppByte3Inv = four tables used for the operation.
 *				   If these args are NULL, NULL, NULL and NULL,
 *				   the tables will be allocated and freed
 *				   internally.  These tables can be obtained
 *				   from i_make4To1InvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_thresholdI4To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To1bppByte0Inv,
    UInt32     *pTab4To1bppByte1Inv,
    UInt32     *pTab4To1bppByte2Inv,
    UInt32     *pTab4To1bppByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	 hOrig;			/* ditto */
    UInt32	 tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* remember for ip_clearRightFrame */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To1bppByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To1InvBinarizationTables (pIPState,
							&pTab4To1bppByte0Inv,
							&pTab4To1bppByte1Inv,
							&pTab4To1bppByte2Inv,
							&pTab4To1bppByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word.  The dest has 1 bpp pixels. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 4 source words to make one destination words */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.  We use 8 since there are
	 * 8 4-bit pixels per 32-bit word.
	 */
    srcLimit = ((w - 1)/8) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	/* This is the primitive routine that does the real work */
    ip_toBinaryFrom4bppLine (pIPState, pS, pD,
			     dstInitOffset, srcInitOffset, srcLimit,
			     pTab4To1bppByte0Inv, pTab4To1bppByte1Inv,
			     pTab4To1bppByte2Inv, pTab4To1bppByte3Inv);
    }

    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTab4To1bppByte0Inv);
	FREE(pTab4To1bppByte1Inv);
	FREE(pTab4To1bppByte2Inv);
	FREE(pTab4To1bppByte3Inv);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************** 
 *  i_thresholdTo2bpp()
 *		Takes an 8-bit grayscale image and returns
 *	 	a thresholded 2 bit/pixel image.  The value of
 *	 	each pixel in the result image is the two Most
 *	 	Significant bits of the corresponding pixel in
 *	 	the source image.
 *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination 
 *			image's frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTabTo2bppByte0,
 *		pTabTo2bppByte1,
 *		pTabTo2bppByte2,
 *		pTabTo2bppByte3 = four tables used for the operation.
 *				  If these args are NULL, NULL, NULL and NULL,
 *				  the tables will be allocated and freed
 *				  internally.  These tables can be obtained
 *				  from i_make2bppBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_thresholdTo2bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabTo2bppByte0,
    UInt32     *pTabTo2bppByte1,
    UInt32     *pTabTo2bppByte2,
    UInt32     *pTabTo2bppByte3)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    
    UInt32	tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 2*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabTo2bppByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make2bppBinarizationTables (pIPState,
						     &pTabTo2bppByte0,
						     &pTabTo2bppByte1,
						     &pTabTo2bppByte2,
						     &pTabTo2bppByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word.  The dest has 2 bpp pixels. */
    dstInitOffset = ((w - 1)/16) * 4 ;
	/* it takes 4 source words to make one destination words */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.  We use 4 since there are
	 * 4 8-bit pixels per 32-bit word.
	 */
    srcLimit = ((w - 1)/4) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toTwoBitsPerPixelLine (pIPState, pS, pD,
				  dstInitOffset, srcInitOffset, srcLimit,
				  pTabTo2bppByte0, pTabTo2bppByte1,
				  pTabTo2bppByte2, pTabTo2bppByte3);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTabTo2bppByte0);
	FREE(pTabTo2bppByte1);
	FREE(pTabTo2bppByte2);
	FREE(pTabTo2bppByte3);
    }

    NORMAL_RETURN;

}  /* i_thresholdTo2bpp() */

/* PUBLIC */
/*************************************************************************** 
 *  i_threshold4To2bpp()
 *		Takes an 4-bit grayscale image and returns
 *	 	a thresholded 2 bit/pixel image.  The value of
 *	 	each pixel in the result image is the two Most
 *	 	Significant bits of the corresponding pixel in
 *	 	the source image.
 *
 *	Input:
 *      	pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	pD =    Pointer to upper left corner of destination 
 *			image's frame
 *      	dBpl =  bytes/line of destination including frame. 
 *		pTab4To2bpp =	table used for the operation.
 *				If passed as NULL the table will be
 *				allocated and freed internally.  These
 *				tables can be obtained from
 *				i_make4To2bppBinarizationTable.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_threshold4To2bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To2bpp)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    
    UInt32  tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 2*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

    /* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To2bpp == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To2bppBinarizationTable (pIPState, &pTab4To2bpp))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

    /* this points to the last dest word.  The dest has 2 bpp pixels. */
    dstInitOffset = ((w - 1) / 16) * 4; 

    /* it takes 2 source words to make one destination words */
    srcInitOffset = dstInitOffset * 2; 

    /* we want a pointer to the last word in the source line since
     * this might not be a multiple of 4.  We use 8 since there are
     * 8 4-bit pixels per 32-bit word.
     */
    srcLimit = ((w - 1) / 8) * 4;

    /* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_fourToTwoBitsPerPixelLine (pIPState, pS, pD,
				      dstInitOffset, srcInitOffset, srcLimit,
				      pTab4To2bpp);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTab4To2bpp);
    }

    NORMAL_RETURN;

}  /* i_threshold4To2bpp () */


/* PUBLIC */
/***************************************************************************  
 *  i_highpassfilt()
 *             	Returns the highpass filtered thresholded binary image
 *		of the input grayscale image.
 *              If GrayPix(x,y) >= thresh then BinaryPix(x,y) = 1 (roughly)
 *              The algorithm disregards the LS bits of both the gray data
 *              and the threshold, using this bit to hold the results of
 *              four comparisons done in parallel.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	thresh = high pass filter threshold in [1,,255]. 
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		pTabBinarizeByte0,
 *		pTabBinarizeByte1,
 *		pTabBinarizeByte2,
 *		pTabBinarizeByte3 = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32 CDECL 
i_highpassfilt (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	thresh,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0,
    UInt32     *pTabBinarizeByte1,
    UInt32     *pTabBinarizeByte2,
    UInt32     *pTabBinarizeByte3)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    
    UInt32	tableShouldBeFreed;

	/* skip the pad */;
    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_makeBinarizationTables (pIPState,
						 &pTabBinarizeByte0,
						 &pTabBinarizeByte1,
						 &pTabBinarizeByte2,
						 &pTabBinarizeByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/4) * 4;

    if (thresh == 128)	/* do it the fastest way */
    {
	    /* for each line in the destination image */
	for (; h > 0; h--, pS += sBpl, pD += dBpl)
	{
		/* This is the primitive routine that does the real work */
	    ip_toBinary128Line (pIPState, pS, pD,
			        dstInitOffset, srcInitOffset, srcLimit,
			        pTabBinarizeByte0, pTabBinarizeByte1,
			        pTabBinarizeByte2, pTabBinarizeByte3);
	}
    }
    else
    {
	    /* repeat the threshold in all 4 bytes so that we can do the
	     * compares for one 32 bit word in parallel.  The least 
	     * significant bit of each pixel is ignored in these comparisons.
	     * Why do we need to ignore the least significant bit?  We
	     * need that bit to contain the carry-out for the 7 most significant
	     * bits.  We preset those bits to 1 so that the borrows don't
	     * propagate between the bytes; then subtract the shifted 
	     * threshold.  The carry-out is 1 if the data was greater than 
	     * the threshold.
	     */
	thresh = thresh | (thresh << 8);
	thresh = thresh | (thresh << 16);
	thresh = thresh >> 1;
	thresh = thresh & ~cMSBitMask;

	    /* for each line in the destination image */
	for (; h > 0; h--, pS += sBpl, pD += dBpl)
	{
	    ip_toBinaryThreshLine (pIPState, pS, pD,
				   dstInitOffset, srcInitOffset, srcLimit,
			           pTabBinarizeByte0, pTabBinarizeByte1,
			           pTabBinarizeByte2, pTabBinarizeByte3,
			           thresh);
	}
    }

    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0);
	FREE(pTabBinarizeByte1);
	FREE(pTabBinarizeByte2);
	FREE(pTabBinarizeByte3);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***************************************************************************  
 *  i_highpassfiltI()
 *             	Returns the inverted highpass filtered thresholded binary 
 *		image of the input grayscale image.
 *              If GrayPix(x,y) >= thresh, BinaryPix(x,y) = 1 (roughly)
 *              The algorithm disregards the LS bits of both the gray data
 *              and the threshold, using this bit to hold the results of
 *              four comparisons done in parallel.
 *              Returns inverse (0 <-> 1) of image returned by highpassfilt
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *      	thresh = high pass filter threshold in [1,,255]. 
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		pTabBinarizeByte0Inv,
 *		pTabBinarizeByte1Inv,
 *		pTabBinarizeByte2Inv,
 *		pTabBinarizeByte3Inv = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeInvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_highpassfiltI (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	thresh,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0Inv,
    UInt32     *pTabBinarizeByte1Inv,
    UInt32     *pTabBinarizeByte2Inv,
    UInt32     *pTabBinarizeByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	 hOrig;			/* ditto */
    UInt32	 tableShouldBeFreed;

	/* skip the pad */;
    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* for  below */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_makeInvBinarizationTables (pIPState,
						    &pTabBinarizeByte0Inv,
						    &pTabBinarizeByte1Inv,
						    &pTabBinarizeByte2Inv,
						    &pTabBinarizeByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/4) * 4;

    if (thresh == 128)	/* do it the fastest way */
    {
	    /* for each line in the destination image */
	for (; h > 0; h--, pS += sBpl, pD += dBpl)
	{
		/* This is the primitive routine that does the real work */
	    ip_toBinary128Line (pIPState, pS, pD,
			        dstInitOffset, srcInitOffset, srcLimit,
			        pTabBinarizeByte0Inv, pTabBinarizeByte1Inv,
			        pTabBinarizeByte2Inv, pTabBinarizeByte3Inv);
	}
    }
    else
    {
	    /* repeat the threshold in all 4 bytes so that we can do the
	     * compares for one 32 bit word in parallel.  The least 
	     * significant bit of each pixel is ignored in these comparisons.
	     * Why do we need to ignore the least significant bit?  We
	     * need that bit to contain the carry-out for the 7 most significant
	     * bits.  We preset those bits to 1 so that the borrows don't
	     * propagate between the bytes; then subtract the shifted 
	     * threshold.  The carry-out is 1 if the data was greater than 
	     * the threshold.
	     */
	thresh = thresh | (thresh << 8);
	thresh = thresh | (thresh << 16);
	thresh = thresh >> 1;
	thresh = thresh & ~cMSBitMask;

	    /* for each line in the destination image */
	for (; h > 0; h--, pS += sBpl, pD += dBpl)
	{
	    ip_toBinaryThreshLine (pIPState, pS, pD,
				   dstInitOffset, srcInitOffset, srcLimit,
			           pTabBinarizeByte0Inv, pTabBinarizeByte1Inv,
			           pTabBinarizeByte2Inv, pTabBinarizeByte3Inv,
			           thresh);
	}
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0Inv);
	FREE(pTabBinarizeByte1Inv);
	FREE(pTabBinarizeByte2Inv);
	FREE(pTabBinarizeByte3Inv);
    }

    NORMAL_RETURN;
}



/* This is a public routine */
/* PUBLIC */
/***************************************************************************  
 *  i_highpassfilt4To1bpp():  takes an 4-bit grayscale image and a threshold
 *		value between 1 and 15 and a pointer to the destination
 *		image.
 *		If GrayPix(x,y) >= thresh, BinaryPix(x,y) = 1 (roughly)
 *
 *	pS =	 Pointer to upper left corner of source image's frame
 *	w =	 width of image in pixels including frame
 *	h =	 height of image in pixels including frame
 *	sBpl =	 bytes/line of source including frame.
 *	thresh = high pass filter threshold in [1,,15].
 *	pD =	 Pointer to upper left corner of destination image's frame
 *	dBpl =	 bytes/line of destination including frame.
 *	pTab4To1bppByte0,
 *	pTab4To1bppByte1,
 *	pTab4To1bppByte2,
 *	pTab4To1bppByte3 = four tables used for the operation.
 *			   If these args are NULL, NULL, NULL and NULL,
 *			   the tables will be allocated and freed
 *			   internally.  These tables can be obtained
 *			   from i_make4To1BinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32 CDECL 
i_highpassfilt4To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	thresh,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To1bppByte0,
    UInt32     *pTab4To1bppByte1,
    UInt32     *pTab4To1bppByte2,
    UInt32     *pTab4To1bppByte3)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */
    UInt32	threshLow;		/* The LSB of the threshold
					 * replicated in each nibble */
    
    UInt32	tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To1bppByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To1BinarizationTables (pIPState,
						     &pTab4To1bppByte0,
						     &pTab4To1bppByte1,
						     &pTab4To1bppByte2,
						     &pTab4To1bppByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 4 source words to make one destination word */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/8) * 4;

	/* repeat the threshold in all 4 bytes so that we can do the
	 * compares for one 32 bit word in parallel.  Unlike the
	 * 8 bpp case, we cannot afford to ignore the least significant
	 * bit of the input pixel but we do still need a bit to contain
	 * the carry-out.  So we split the pixel into two words.  One
	 * word contains the 3 most significant bits of the pixel
	 * and the second word contains the least significant bit.
	 * We only need the information from the least significant bit
	 * of the pixel if the most significant 3 bits were the same. 
	 */
    thresh = thresh | (thresh << 4);
    thresh = thresh | (thresh << 8);
    thresh = thresh | (thresh << 16);
    threshLow = thresh & cLSBitMask4bpp;
    thresh  = thresh >> 1;
    thresh = thresh & ~cMSBitMask4bpp;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	ip_toBinaryFrom4bppThreshLine (pIPState, pS, pD,
				       dstInitOffset, srcInitOffset, srcLimit,
				       pTab4To1bppByte0, pTab4To1bppByte1,
				       pTab4To1bppByte2, pTab4To1bppByte3,
				       thresh, threshLow);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTab4To1bppByte0);
	FREE(pTab4To1bppByte1);
	FREE(pTab4To1bppByte2);
	FREE(pTab4To1bppByte3);
    }

    NORMAL_RETURN;
}


/* This is a public routine */
/* PUBLIC */
/***************************************************************************  
 *  i_highpassfiltI4To1bpp():  takes an 4-bit grayscale image and a threshold
 *		value between 1 and 15 and a pointer to the destination
 *		image.
 *		If GrayPix(x,y) >= thresh, BinaryPix(x,y) = 0 (roughly)
 *
 *	pS =	 Pointer to upper left corner of source image's frame
 *	w =	 width of image in pixels including frame
 *	h =	 height of image in pixels including frame
 *	sBpl =	 bytes/line of source including frame.
 *	thresh = high pass filter threshold in [1,,15].
 *	pD =	 Pointer to upper left corner of destination image's frame
 *	dBpl =	 bytes/line of destination including frame.
 *	pTab4To1bppByte0Inv,
 *	pTab4To1bppByte1Inv,
 *	pTab4To1bppByte2Inv,
 *	pTab4To1bppByte3Inv = four tables used for the operation.
 *			   If these args are NULL, NULL, NULL and NULL,
 *			   the tables will be allocated and freed
 *			   internally.  These tables can be obtained
 *			   from i_make4To1InvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32 CDECL 
i_highpassfiltI4To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	thresh,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To1bppByte0Inv,
    UInt32     *pTab4To1bppByte1Inv,
    UInt32     *pTab4To1bppByte2Inv,
    UInt32     *pTab4To1bppByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    UInt32	 threshLow;		/* The LSB of the threshold
					 * replicated in each nibble */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	 hOrig;			/* ditto */
    UInt32	 tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* for ip_clearRightFrame below */
    hOrig = h;				/* ditto */

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To1bppByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To1InvBinarizationTables (pIPState,
							&pTab4To1bppByte0Inv,
							&pTab4To1bppByte1Inv,
							&pTab4To1bppByte2Inv,
							&pTab4To1bppByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 4 source words to make one destination word */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/8) * 4;

	/* repeat the threshold in all 4 bytes so that we can do the
	 * compares for one 32 bit word in parallel.  Unlike the
	 * 8 bpp case, we cannot afford to ignore the least significant
	 * bit of the input pixel but we do still need a bit to contain
	 * the carry-out.  So we split the pixel into two words.  One
	 * word contains the 3 most significant bits of the pixel
	 * and the second word contains the least significant bit.
	 * We only need the information from the least significant bit
	 * of the pixel if the most significant 3 bits were the same. 
	 */
    thresh = thresh | (thresh << 4);
    thresh = thresh | (thresh << 8);
    thresh = thresh | (thresh << 16);
    threshLow = thresh & cLSBitMask4bpp;
    thresh  = thresh >> 1;
    thresh = thresh & ~cMSBitMask4bpp;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	ip_toBinaryFrom4bppThreshLine (pIPState, pS, pD,
				       dstInitOffset, srcInitOffset, srcLimit,
				       pTab4To1bppByte0Inv, pTab4To1bppByte1Inv,
				       pTab4To1bppByte2Inv, pTab4To1bppByte3Inv,
				       thresh, threshLow);
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTab4To1bppByte0Inv);
	FREE(pTab4To1bppByte1Inv);
	FREE(pTab4To1bppByte2Inv);
	FREE(pTab4To1bppByte3Inv);
    }

    NORMAL_RETURN;
}



/* PUBLIC */
/***************************************************************************  
 *  i_passfilt()
 *  		Takes an 8-bit grayscale image and two threshold values, 
 *              lower and upper (1 <= lower <= upper <= 255).
 *              {lower <= GrayPix(x,y) <= upper} => BinaryPix(x,y) = 1 (roughly)
 *              The algorithm disregards the LS bits of both the gray data
 *              and the thresholds, using this bit to hold the results of
 *              four comparisons done in parallel. 
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		lower =	lower pass filter threshold in [0,,255]
 *		upper =	upper pass filter threshold in [0,,255]
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		pTabBinarizeByte0,
 *		pTabBinarizeByte1,
 *		pTabBinarizeByte2,
 *		pTabBinarizeByte3 = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_passfilt (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	lower,
    Int32	upper,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0,
    UInt32     *pTabBinarizeByte1,
    UInt32     *pTabBinarizeByte2,
    UInt32     *pTabBinarizeByte3)
{
    PROLOGUE;
    Int32	srcLimit;			/* max bytes in source line */
    Int32	srcInitOffset;			/* initial source line offset */
    Int32	dstInitOffset;			/* initial dst line offset */
    
    UInt32	tableShouldBeFreed;

	/* skip the pad */;
    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_makeBinarizationTables (pIPState,
						 &pTabBinarizeByte0,
						 &pTabBinarizeByte1,
						 &pTabBinarizeByte2,
						 &pTabBinarizeByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/4) * 4;

	/* Repeat the thresholds in all 4 bytes so that we can do the
	 * compares for one 32 bit word in parallel.  The least 
	 * significant bit of each pixel is ignored in these comparisons.
	 * Why do we need to ignore the least significant bit?  We
	 * need that bit to contain the carry-out for the 7 most significant
	 * bits.  We preset those bits to 1 so that the borrows don't
	 * propagate between the bytes.
	 * But we really don't use the upper and lower thresholds as is.
	 * First, we compute upper-lower threshold and replicate that
	 * with each copy shifted right by 1 and the MS bits turned
	 * on.  To form each output word inside ip_toBinaryBandPass,
	 * we subtract a right-shifted and replicated copy of the lower
	 * threshold from the raw input data.  This transforms
	 * values in the range [lower, upper] to [0, upper-lower].
	 * Now, set the MS bit of each pixel to 0.  Subtract this value
	 * from (upper-lower) (with carry-out bits set) duplicated in all 
	 * 4 bytes.  The resulting carry-out will be set if the original
	 * data was greater than or equal to lower and less than or equal
	 * to upper.
	 */
    upper = upper - lower;
    upper = upper | (upper << 8);
    upper = upper | (upper << 16);
    upper = upper >> 1;
    upper = upper | cMSBitMask;

    lower = lower | (lower << 8);
    lower = lower | (lower << 16);
    lower = lower >> 1;
    lower = lower & ~cMSBitMask;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toBinaryBandPassLine (pIPState, pS, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTabBinarizeByte0, pTabBinarizeByte1,
				 pTabBinarizeByte2, pTabBinarizeByte3,
				 lower, upper);
    }

    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0);
	FREE(pTabBinarizeByte1);
	FREE(pTabBinarizeByte2);
	FREE(pTabBinarizeByte3);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************** 
 * i_passfiltI()
 * 		Takes an 8-bit grayscale image and two threshold values, 
 *              lower and upper (1 <= lower <= upper <= 255).
 *              {lower <= GrayPix(x,y) <= upper} => BinaryPix(x,y) = 0 (roughly)
 *              The algorithm disregards the LS bits of both the gray data
 *              and the thresholds, using this bit to hold the results of
 *              four comparisons done in parallel.
 *
 *              Returned image is inverse (0 <-> 1) of that returned
 *              by i_passfilt. 
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		lower =	lower pass filter threshold in [0,,255]
 *		upper =	upper pass filter threshold in [0,,255]
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		pTabBinarizeByte0Inv,
 *		pTabBinarizeByte1Inv,
 *		pTabBinarizeByte2Inv,
 *		pTabBinarizeByte3Inv = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeInvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_passfiltI (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	lower,
    Int32	upper,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0Inv,
    UInt32     *pTabBinarizeByte1Inv,
    UInt32     *pTabBinarizeByte2Inv,
    UInt32     *pTabBinarizeByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;			/* max bytes in source line */
    Int32	 srcInitOffset;			/* initial source line offset */
    Int32	 dstInitOffset;			/* initial dst line offset */
    
    UInt8	*pDOrig;			/* used to clear right frame */
    Int32	 hOrig;				/* ditto */
    UInt32	 tableShouldBeFreed;

	/* skip the pad */;
    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD +=  (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* for ip_clearRightFrame below */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_makeInvBinarizationTables (pIPState,
						    &pTabBinarizeByte0Inv,
						    &pTabBinarizeByte1Inv,
						    &pTabBinarizeByte2Inv,
						    &pTabBinarizeByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word. */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/4) * 4;

	/* Repeat the thresholds in all 4 bytes so that we can do the
	 * compares for one 32 bit word in parallel.  The least 
	 * significant bit of each pixel is ignored in these comparisons.
	 * Why do we need to ignore the least significant bit?  We
	 * need that bit to contain the carry-out for the 7 most significant
	 * bits.  We preset those bits to 1 so that the borrows don't
	 * propagate between the bytes.
	 * But we really don't use the upper and lower thresholds as is.
	 * First, we compute upper-lower threshold and replicate that
	 * with each copy shifted right by 1 and the MS bits turned
	 * on.  To form each output word inside ip_toBinaryBandPass,
	 * we subtract a right-shifted and replicated copy of the lower
	 * threshold from the raw input data.  This transforms
	 * values in the range [lower, upper] to [0, upper-lower].
	 * Now, set the MS bit of each pixel to 0.  Subtract this value
	 * from (upper-lower) (with carry-out bits set) duplicated in all 
	 * 4 bytes.  The resulting carry-out will be set if the original
	 * data was greater than or equal to lower and less than or equal
	 * to upper.  The inverting tables are used to invert the final
	 * output.
	 */
    upper = upper - lower;
    upper = upper | (upper << 8);
    upper = upper | (upper << 16);
    upper = upper >> 1;
    upper = upper | cMSBitMask;

    lower = lower | (lower << 8);
    lower = lower | (lower << 16);
    lower = lower >> 1;
    lower = lower & ~cMSBitMask;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_toBinaryBandPassLine (pIPState, pS, pD,
				 dstInitOffset, srcInitOffset, srcLimit,
				 pTabBinarizeByte0Inv, pTabBinarizeByte1Inv,
				 pTabBinarizeByte2Inv, pTabBinarizeByte3Inv,
				 lower, upper);
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0Inv);
	FREE(pTabBinarizeByte1Inv);
	FREE(pTabBinarizeByte2Inv);
	FREE(pTabBinarizeByte3Inv);
    }

    NORMAL_RETURN;
}

/* This is a public routine */
/* PUBLIC */
/***************************************************************************  
 *  i_passfilt4To1bpp():  takes an 4-bit grayscale image and two threshold
 *		values between 0 and 15 and a pointer to the destination
 *		image.
 *              {lower <= GrayPix(x,y) <= upper} => BinaryPix(x,y) = 1
 *
 *	pS =	Pointer to upper left corner of source image's frame
 *	w =	width of image in pixels including frame
 *	h =	height of image in pixels including frame
 *	sBpl =	bytes/line of source including frame.
 *	lower =	lower pass filter threshold in [0,,15]
 *	upper =	upper pass filter threshold in [0,,15]
 *	pD =	Pointer to upper left corner of destination image's frame
 *	dBpl =	bytes/line of destination including frame.
 *	pTab4To1bppByte0,
 *	pTab4To1bppByte1,
 *	pTab4To1bppByte2,
 *	pTab4To1bppByte3 = four tables used for the operation.
 *			   If these args are NULL, NULL, NULL and NULL,
 *			   the tables will be allocated and freed
 *			   internally.  These tables can be obtained
 *			   from i_make4To1BinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32 CDECL 
i_passfilt4To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	lower,
    Int32	upper,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To1bppByte0,
    UInt32     *pTab4To1bppByte1,
    UInt32     *pTab4To1bppByte2,
    UInt32     *pTab4To1bppByte3)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    Int32	 lsBits;		/* contains LS bit of (upper - lower) in
				         * bit 1 and LS bit of lower in bit 0 */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	 hOrig;			/* ditto */
    UInt32	 tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;		/* for ip_clearRightFrame below */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To1bppByte0 == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To1BinarizationTables (pIPState,
						     &pTab4To1bppByte0,
						     &pTab4To1bppByte1,
						     &pTab4To1bppByte2,
						     &pTab4To1bppByte3))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 4 source words to make one destination word */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/8) * 4;

	/* repeat the threshold in all 4 bytes so that we can do the
	 * compares for one 32 bit word in parallel.  Unlike the
	 * 8 bpp case, we cannot afford to ignore the least significant
	 * bit of the input pixel but we do still need a bit to contain
	 * the carry-out.  So we split the pixel into two words.  One
	 * word contains the 3 most significant bits of the pixel
	 * and the second word contains the least significant bit.
	 * We only need the information from the least significant bit
	 * of the pixel if the most significant 3 bits were the same. 
	 * But we really don't use the upper and lower thresholds as is.
	 * First, we compute upper-lower threshold and replicate that
	 * with each copy shifted right by 1 and the MS bits turned
	 * on.  To form each output word inside ip_toBinaryFrom4bppBandpassLine,
	 * we subtract a right-shifted and replicated copy of the lower
	 * threshold from the raw input data.  This transforms
	 * values in the range [lower, upper] to [0, upper-lower].
	 * Now, set the MS bit of each pixel to 0.  Subtract this value
	 * from (upper-lower) (with carry-out bits set) duplicated in all 
	 * 4 bytes.  The resulting carry-out will be set if the original
	 * data was greater than or equal to lower and less than or equal
	 * to upper.
	 */
    upper = upper - lower;
    lsBits = ((upper << 1) | (lower & 1)) & 0x3;
    upper = upper | (upper << 4);
    upper = upper | (upper << 8);
    upper = upper | (upper << 16);
    upper  = upper >> 1;
    upper = upper | cMSBitMask4bpp;

    lower = lower | (lower << 4);
    lower = lower | (lower << 8);
    lower = lower | (lower << 16);
    lower  = lower >> 1;
    lower = lower & ~cMSBitMask4bpp;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	ip_toBinaryFrom4bppBandpassLine (pIPState, pS, pD,
					 dstInitOffset, srcInitOffset, srcLimit,
					 pTab4To1bppByte0, pTab4To1bppByte1,
					 pTab4To1bppByte2, pTab4To1bppByte3,
					 upper, lower, lsBits);
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTab4To1bppByte0);
	FREE(pTab4To1bppByte1);
	FREE(pTab4To1bppByte2);
	FREE(pTab4To1bppByte3);
    }

    NORMAL_RETURN;
}


/* This is a public routine */
/* PUBLIC */
/***************************************************************************  
 *  i_passfiltI4To1bpp():  takes an 4-bit grayscale image and two threshold
 *		values between 0 and 15 and a pointer to the destination
 *		image.
 *              {lower <= GrayPix(x,y) <= upper} => BinaryPix(x,y) = 0
 *
 *              Returned image is inverse (0 <-> 1) of that returned
 *              by i_passfilt4To1bpp. 
 *
 *	pS =	Pointer to upper left corner of source image's frame
 *	w =	width of image in pixels including frame
 *	h =	height of image in pixels including frame
 *	sBpl =	bytes/line of source including frame.
 *	lower =	lower pass filter threshold in [0,,15]
 *	upper =	upper pass filter threshold in [0,,15]
 *	pD =	Pointer to upper left corner of destination image's frame
 *	dBpl =	bytes/line of destination including frame.
 *	pTab4To1bppByte0Inv,
 *	pTab4To1bppByte1Inv,
 *	pTab4To1bppByte2Inv,
 *	pTab4To1bppByte3Inv = four tables used for the operation.
 *			   If these args are NULL, NULL, NULL and NULL,
 *			   the tables will be allocated and freed
 *			   internally.  These tables can be obtained
 *			   from i_make4To1InvBinarizationTables.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32 CDECL 
i_passfiltI4To1bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	lower,
    Int32	upper,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTab4To1bppByte0Inv,
    UInt32     *pTab4To1bppByte1Inv,
    UInt32     *pTab4To1bppByte2Inv,
    UInt32     *pTab4To1bppByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;	/* max bytes in source line */
    Int32	 srcInitOffset;	/* initial source line offset */
    Int32	 dstInitOffset;	/* initial dst line offset */
    Int32	 lsBits;	/* contains LS bit of (upper - lower) in
				 * bit 1 and LS bit of lower in bit 0 */
    
    UInt8	*pDOrig;	/* used to clear right frame */
    Int32	 hOrig;		/* ditto */
    UInt32	 tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;			/* for ip_clearRightFrame below */
    hOrig = h;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate binarization table if necessary */
    if (pTab4To1bppByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if ((__error__ = i_make4To1InvBinarizationTables (pIPState,
							&pTab4To1bppByte0Inv,
							&pTab4To1bppByte1Inv,
							&pTab4To1bppByte2Inv,
							&pTab4To1bppByte3Inv))
	    != ia_successful)
	    NORMAL_RETURN;
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((w - 1)/32) * 4 ;
	/* it takes 4 source words to make one destination word */
    srcInitOffset = dstInitOffset * 4;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((w - 1)/8) * 4;

	/* repeat the threshold in all 4 bytes so that we can do the
	 * compares for one 32 bit word in parallel.  Unlike the
	 * 8 bpp case, we cannot afford to ignore the least significant
	 * bit of the input pixel but we do still need a bit to contain
	 * the carry-out.  So we split the pixel into two words.  One
	 * word contains the 3 most significant bits of the pixel
	 * and the second word contains the least significant bit.
	 * We only need the information from the least significant bit
	 * of the pixel if the most significant 3 bits were the same. 
	 * But we really don't use the upper and lower thresholds as is.
	 * First, we compute upper-lower threshold and replicate that
	 * with each copy shifted right by 1 and the MS bits turned
	 * on.  To form each output word inside ip_toBinaryFrom4bppBandpassLine,
	 * we subtract a right-shifted and replicated copy of the lower
	 * threshold from the raw input data.  This transforms
	 * values in the range [lower, upper] to [0, upper-lower].
	 * Now, set the MS bit of each pixel to 0.  Subtract this value
	 * from (upper-lower) (with carry-out bits set) duplicated in all 
	 * 4 bytes.  The resulting carry-out will be set if the original
	 * data was greater than or equal to lower and less than or equal
	 * to upper.  The inverting tables are used to invert the final
	 * output.
	 */
    upper = upper - lower;
    lsBits = ((upper << 1) | (lower & 1)) & 0x3;
    upper = upper | (upper << 4);
    upper = upper | (upper << 8);
    upper = upper | (upper << 16);
    upper  = upper >> 1;
    upper = upper | cMSBitMask4bpp;

    lower = lower | (lower << 4);
    lower = lower | (lower << 8);
    lower = lower | (lower << 16);
    lower  = lower >> 1;
    lower = lower & ~cMSBitMask4bpp;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	ip_toBinaryFrom4bppBandpassLine (pIPState, pS, pD,
					 dstInitOffset, srcInitOffset, srcLimit,
					 pTab4To1bppByte0Inv,
					 pTab4To1bppByte1Inv,
					 pTab4To1bppByte2Inv,
					 pTab4To1bppByte3Inv,
					 upper, lower, lsBits);
    }
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 1);

    if (tableShouldBeFreed)
    {
	FREE(pTab4To1bppByte0Inv);
	FREE(pTab4To1bppByte1Inv);
	FREE(pTab4To1bppByte2Inv);
	FREE(pTab4To1bppByte3Inv);
    }

    NORMAL_RETURN;
}



/* PUBLIC */
/*************************************************************************** 
 * i_map4To4
 *		Maps a 4 bit to another one through a look-up table.
 *		This is used, for example, to map 4-bit gray images
 *		into a color mapped 4-bit space where there are, say,
 *		4 grays (including B & W), but they are not linear.
 *		This could, of course, be used for any other purpose.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		map =	Look-up table mapping input.
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_map4To4 (
    IPState    *pIPState,
    register UInt8 *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    register UInt8 *pD,
    Int32	dBpl,
    UInt8      *map)
{
    /* What we do is make an 8-bit map so that we can map 2 pixels
       at once. */
    PROLOGUE;
    UInt8	*fullMap;
    Int32	 widthInWords, j;
    UInt8	*pDOrig;
    Int32	 hOrig;

	/* allocate the table that maps 2 pixels at a time */
    ExceptRaise ((fullMap = (UInt8 *) MALLOC (256)) == NULL, ia_nomem);

    for (j = 0; j < 256; j++)
    {
	fullMap[j] = map[j & 0xf] | (map[j >> 4] << 4);
    }

    /* Move past the frame */
    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD +=  (dBpl << cLogFrameBits) + 4*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;
    hOrig = h;

    /* Grind through the pixels */
    /* Calculate the width in bytes of the sequence of longwords that
     * just covers each line.  There are 8 pixels/longword.
     */
    widthInWords = ((w + 7) / 8);

    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	ip_map8To8Line (pIPState, pS, pD, widthInWords, fullMap);
    }

    /* Just assume that we trashed the frame */
    ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 4);

    FREE(fullMap);
    NORMAL_RETURN;
}
/* PUBLIC */
/*************************************************************************** 
 * i_map8To8
 *		Maps a 8 bit image to another one through a look-up table.
 *		This is used, for example, to correct the tone-response curve of
 *		an image
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		map =	Look-up table mapping input.
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_map8To8 (
    IPState    *pIPState,
    register UInt8 *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    register UInt8 *pD,
    Int32	dBpl,
    UInt8      *map)
{
    PROLOGUE;
    Int32	widthInWords;
    UInt8      *pDOrig;
    Int32	hOrig;

    /* Move past the frame */
    pS += (sBpl << cLogFrameBits) + cFrameBits;
    pD +=  (dBpl << cLogFrameBits) + cFrameBits;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
    pDOrig = pD;
    hOrig = h;

    /* Grind through the pixels */
    widthInWords = ((w + 3) / 4 );

    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	ip_map8To8Line (pIPState, pS, pD, widthInWords, map);
        
        if ((UInt32)h % 1024 == 0)
            IP_YIELD(funcExit);
    }

    /* Just assume that we trashed the frame */
    ERR_PASS (ip_clearRightFrame (pIPState, pDOrig, w, hOrig, dBpl, 8));

funcExit:
    NORMAL_RETURN;
}

/* PUBLIC */
/*************************************************************************** 
 * i_threshold8To4bpp
 *		Scale an 8-bit gray image to a 4-bit gray image.
 *		Does a simple shift (or equivalent), so contouring
 *		will occur on contone images.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_threshold8To4bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl)
{
    PROLOGUE;
    Int32	srcLimit;		/* max bytes in source line */
    Int32	srcInitOffset;		/* initial source line offset */
    Int32	dstInitOffset;		/* initial dst line offset */

    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 4*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* this points to the last dest word.  The dest has 4-bit pixels. */
    dstInitOffset = ((w - 1)/8) * 4 ;
	/* it takes 2 source words to make one destination words */
    srcInitOffset = dstInitOffset * 2;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.  We use 4 since there are
	 * 4 8-bit pixels per 32-bit word.
	 */
    srcLimit = ((w - 1)/4) * 4;

	/* for each line in the destination image */
    for (; h > 0; h--, pS += sBpl, pD += dBpl)
    {
	    /* This is the primitive routine that does the real work */
	ip_threshold8To4Line (pIPState, pS, pD,
			      dstInitOffset, srcInitOffset, srcLimit);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/***************************************************************************  
 *  i_thresholdByPixel()
 *             	Returns the inverted result of comparing an 8bpp gray
 *		image with an 8bpp threshold image.
 *              If GrayPix(x,y) >= ThreshPix(x,y), BinaryPix(x,y) = 0 (roughly).
 *		This means that if the gray value is whiter than the threshold
 *		value, the result is white in the binary value.
 *              The algorithm disregards the LS bits of both the gray data
 *              and the threshold, using this bit to hold the results of
 *              four comparisons done in parallel.
 *		In a disgusting act of creeping featurism, this operation
 *		also scales the threshold image to match the source image
 *		if necessary.  This was done because the binarization routines
 *		that call it like to produce highly reduced threshold images.
 *		Scaling them up to full scale before doing the threshold
 *		takes an appreciable amount of time and memory.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	wS =    width of source & dest image in pixels including frame
 *      	hS =    height of source & dest image in pixels including frame
 *      	sBpl =  bytes/line of source image including frame.
 *		pT =    Pointer to upper left corner of threshold image's frame
 *      	wT =    width of threshold image in pixels including frame
 *      	hT =    height of threshold image in pixels including frame
 *      	tBpl =  bytes/line of thresh image including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		invertOutput = if cTrue, output should be inverted. This input
 *			       is only valid of no binarization tables are
 *			       supplied.
 *		pTabBinarizeByte0Inv,
 *		pTabBinarizeByte1Inv,
 *		pTabBinarizeByte2Inv,
 *		pTabBinarizeByte3Inv = four tables used for the operation.
 *				    If these args are NULL, NULL, NULL and NULL,
 *				    the tables will be allocated and freed
 *				    internally.  These tables can be obtained
 *				    from i_makeInvBinarizationTables.  NOTE
 *				    that the inverse tables are needed to
 *				    implement the gray->binary inversion.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_thresholdByPixel (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	wS,
    Int32	hS,
    Int32	sBpl,
    UInt8      *pT,
    Int32	wT,
    Int32	hT,
    Int32	tBpl,
    Bool	invertOutput,
    UInt8      *pD,
    Int32	dBpl,
    UInt32     *pTabBinarizeByte0Inv,
    UInt32     *pTabBinarizeByte1Inv,
    UInt32     *pTabBinarizeByte2Inv,
    UInt32     *pTabBinarizeByte3Inv)
{
    PROLOGUE;
    Int32	 srcLimit;		/* max bytes in source line */
    Int32	 srcInitOffset;		/* initial source line offset */
    Int32	 dstInitOffset;		/* initial dst line offset */
    
    UInt8	*pDOrig;		/* used to clear right frame */
    Int32	 hOrig;			/* ditto */
    UInt32	 tableShouldBeFreed;

    /* Variables used when scaling the threshold image to match
     * the gray source image. */
    UInt8       *pThreshLine = NULL;
    Int32       *xIndex, *yOffset;
    Int32	 threshYOffset;

    Int32	 scaleTableSize;
    NonInterpScalingInfo *pScaleTable = NULL;

	/* skip the pad */;
    pS += (sBpl << cLogFrameBits) + 8*cFrameBytes;
    pT += (tBpl << cLogFrameBits) + 8*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + cFrameBytes;
    wS -= 2*cFrameBits;
    hS -= 2*cFrameBits;
    wT -= 2*cFrameBits;
    hT -= 2*cFrameBits;
    pDOrig = pD;			/* for  below */
    hOrig = hS;

	/* if wS == 0, just return instead of blowing up below */
    if ((wS <= 0)  || (hS <= 0))
	NORMAL_RETURN;

	/* if the threshold image size is weird, complain */
    ERR_GEN(((wT <= 0)  || (hT <= 0)), ia_invalidParm);

	/* allocate binarization table if necessary */
    if (pTabBinarizeByte0Inv == NULL)
    {
	tableShouldBeFreed = cTrue;
	if (invertOutput)
	{
	    ERR_PASS( i_makeBinarizationTables (pIPState,
						&pTabBinarizeByte0Inv,
						&pTabBinarizeByte1Inv,
						&pTabBinarizeByte2Inv,
						&pTabBinarizeByte3Inv));
	}
	else
	{
	    ERR_PASS( i_makeInvBinarizationTables (pIPState,
						   &pTabBinarizeByte0Inv,
						   &pTabBinarizeByte1Inv,
						   &pTabBinarizeByte2Inv,
						   &pTabBinarizeByte3Inv));
	}
    }
    else
	tableShouldBeFreed = cFalse;

	/* this points to the last dest word */
    dstInitOffset = ((wS - 1)/32) * 4 ;
	/* it takes 8 source words to make one destination word */
    srcInitOffset = dstInitOffset * 8;	

	/* we want a pointer to the last word in the source line since
	 * this might not be a multiple of 4.
	 */
    srcLimit = ((wS - 1)/4) * 4;

	/* If the source and threshold are already the same size,
	 * do things the easy way.
	 */
    if ((wS == wT) && (hS == hT))
    {
	    /* process each line in the destination image */
	for (; hS > 0; hS--, pS += sBpl, pT += tBpl, pD += dBpl)
	{
	    ip_toBinaryByPixelLine(pIPState, pS, pT, pD,
				   dstInitOffset, srcInitOffset, srcLimit,
				   pTabBinarizeByte0Inv, pTabBinarizeByte1Inv,
				   pTabBinarizeByte2Inv, pTabBinarizeByte3Inv);
	}
    }
    else	/* need to scale threshold to src size. */
    {
	    /* Allocate a buffer that will hold one scan line of
	     * threshold values.  Note that scan lines are supposed
	     * to be rounded up to word boundaries.  The lower level
	     * code reads things a word at a time.
	     */
	pThreshLine = (UInt8 *) CALLOC (((wS + 3)/4)*4, sizeof (UInt8));
	ERR_GEN_CLEANUP (pThreshLine == NULL, ia_nomem);

	    /* This case will scale the threshold image up or down,
	     * but we generally assume that we need to scale up.
	     * We optimize things by scaling one line of the threshold
	     * image at a time, then reusing that line as needed.
	     */
	ERR_PASS_CLEANUP( i_getScaleToSizeTableSize (pIPState, 8, 
						     wS,
						     hS,
						     &scaleTableSize));
	/* Allocate the index tables */
	/* each entry in the array indicates which source pixel (offset
	 * from the left side of the line) to get that destination
	 * pixel.
	 */
	pScaleTable = (NonInterpScalingInfo *)MALLOC (scaleTableSize);
	ERR_GEN_CLEANUP (pScaleTable == NULL, ia_nomem);

	/* So far, so good.  Initialize the scaling tables */
	ERR_PASS_CLEANUP( i_initScaleToSizeTables (pIPState,
						   wT, hT,
						   tBpl, 8,
						   wS, hS,
						   scaleTableSize,
						   (void *)pScaleTable));

	/* For a little performance improvement, XOR the LS two bits
	 * of xIndex table with 3 if this is a litte endian
	 * machine.  This means we don't have to use UCHAR_ACCESS
	 * to read the data.
	 */
	xIndex = (Int32 *) ((UInt8 *)pScaleTable +
			sizeof(NonInterpScalingInfo));
	yOffset = (Int32 *) ((UInt8 *)xIndex + pScaleTable->xIndexSize);

#if _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
	for (threshYOffset = 0; threshYOffset < wS; threshYOffset++)
	    xIndex[threshYOffset] ^= 0x3;
#endif

	/* Do the real work.  For each scan line in the gray source,
	 * produced a scaled scan line of threshold pixel if necessary,
	 * then use the two scan lines to produced a threshholded
	 * output line.
	 */
	threshYOffset = -1;
	for (; hS > 0; hS--, pS += sBpl, pD += dBpl)
	{
	    /* if we need a new scaled threshold scan lines, make one. */
	    if (threshYOffset != *yOffset)
	    {
		Int32	*pXIndex = &(xIndex[wS-1]);
		UInt8	*dPtr = pThreshLine + wS-1;
		UInt8	*sPtr = pT + *yOffset;
		threshYOffset = *yOffset;

		    /* Do the partial word at the right end of the line */
		while (((UInt32) dPtr & 0x3) != 0x3)
		{
		    UCHAR_ACCESS(dPtr) = sPtr[*pXIndex];
		    pXIndex--;
		    dPtr--;
		}

		    /* Do the rest of the words in the line */
		dPtr -= 3;			/* point to beginning of
						 * last word */
		pXIndex -= 3;			/* stay consistent */
		while (dPtr >= pThreshLine)
		{
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
			/* This is the easy case, bytes are addressed from
			 * left to right. */
		    dPtr[0] = sPtr[pXIndex[0]];
		    dPtr[1] = sPtr[pXIndex[1]];
		    dPtr[2] = sPtr[pXIndex[2]];
		    dPtr[3] = sPtr[pXIndex[3]];
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt
			/* This is tricky.  Think about always moving from
			 * left to right within the word to be written.
			 * We've already XORed the LS two
			 * bit of the xIndex table entries with 3, so
			 * we don't need to do a UCHAR_ACCESS there.
			 * There is some debate about whether it's better
			 * to gather the four bytes into one word and
			 * write that, or to do four byte writes.  On
			 * a P6, gathering takes a clear of a reg, reading
			 * a byte into the reg, shifting an accumulator
			 * reg and ORing the first reg into the second.
			 * On the other hand, it has a write-back cache,
			 * so the writes are really fast.  I'll go with
			 * the four writes.
			 */
		    dPtr[3] = sPtr[pXIndex[0]];
		    dPtr[2] = sPtr[pXIndex[1]];
		    dPtr[1] = sPtr[pXIndex[2]];
		    dPtr[0] = sPtr[pXIndex[3]];
#endif
		    dPtr -= 4;
		    pXIndex -= 4;
		}
	    }
	    yOffset++;	/* go down to next scan line */

		/* go produce the next output line from the scaled
		 * threshold line and the gray source line.
		 */
	    ip_toBinaryByPixelLine(pIPState, pS, pThreshLine, pD,
				   dstInitOffset, srcInitOffset, srcLimit,
				   pTabBinarizeByte0Inv, pTabBinarizeByte1Inv,
				   pTabBinarizeByte2Inv, pTabBinarizeByte3Inv);

	}

	    /* We're done with the scaling case.  Free the tables
	     * and buffers.
	     */
	FREE(pThreshLine);
	pThreshLine = NULL;
	FREE(pScaleTable);
	pScaleTable = NULL;
    }
    ip_clearRightFrame (pIPState, pDOrig, wS, hOrig, dBpl, 1);

ERR_CLEANUP:
    if (pThreshLine != NULL)
	FREE(pThreshLine);
    if (pScaleTable != NULL)
	FREE(pScaleTable);
    if (tableShouldBeFreed)
    {
	FREE(pTabBinarizeByte0Inv);
	FREE(pTabBinarizeByte1Inv);
	FREE(pTabBinarizeByte2Inv);
	FREE(pTabBinarizeByte3Inv);
    }

    NORMAL_RETURN;
}



/* PUBLIC */
/*************************************************************************** 
 * i_depth4To8bpp
 *		Scale an 4-bit gray image to an 8-bit gray image.
 *		The bits are replicated.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		pTabExp2 = ptr to table used to do expansion.  If passed
 *			   in as NULL, the table is created and freed
 *			   internally.  The table can be created by calling
 *			   i_makeExp2LUT.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/
Int32  CDECL 
i_depth4To8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt16     *pTabExp2)
{
    PROLOGUE;
    Int32	j, wOver2;
    UInt8      *psByte;
    UInt16     *pdWord;
    Int32	i;
    UInt32	tableShouldBeFreed;

    pS += (sBpl << cLogFrameBits) + 4*cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 8*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
   
	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* allocate expansion table if necessary */
    if (pTabExp2 == NULL)
    {
	tableShouldBeFreed = cTrue;
	ERR_PASS( i_makeExp2LUT (pIPState, 4, &pTabExp2));
    }
    else
	tableShouldBeFreed = cFalse;


    wOver2 = w / 2 + (w & 1);

    for (i = 0; i < h; i++)
    {
	psByte = pS;
	pdWord = (UInt16 *) pD;
	for (j = 0; j < wOver2; j++)
	{
	    /* Look up each byte in pTabExp2, which expands the two nibbles
	     * into two bytes
	     */
	    USHORT_ACCESS(pdWord + j) =
				pTabExp2[(Int32)UCHAR_ACCESS(psByte + j)];
	}
	pS += sBpl;
	pD += dBpl;
    }

    if (tableShouldBeFreed)
    {
	FREE(pTabExp2);
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*************************************************************************** 
 * i_depth1To8bpp
 *		Scale a 1-bit gray image to an 8-bit gray image.
 *		0 becomes 0, 1 becomes ff (or inverted). The best strategies
 *		to do this are probably different on the Sparc and 486.
 *		We're going to make a byte lookup table (resulting in
 *		two longwords). On the 486, we're better off reading
 *		the source byte-at-a-time, but on the Sparc, we'd be
 *		better off reading the source long-word-at-a-time. We're
 *		going to do byte-at-a-time, assuming that if one needs
 *		fast Sparc code, one would be better off writing it in
 *		assembly code anyway.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		invert=	If true, invert the result, so a 1 in the source
 *			image yields a 0 in the result, and a 0 yeilds ff.
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/


static UInt32 resultBits[] = 
{   0x00000000, 0x00000000, 0x00000000, 0x000000FF, 0x00000000, 0x0000FF00, 
    0x00000000, 0x0000FFFF, 0x00000000, 0x00FF0000, 0x00000000, 0x00FF00FF, 
    0x00000000, 0x00FFFF00, 0x00000000, 0x00FFFFFF, 0x00000000, 0xFF000000, 
    0x00000000, 0xFF0000FF, 0x00000000, 0xFF00FF00, 0x00000000, 0xFF00FFFF, 
    0x00000000, 0xFFFF0000, 0x00000000, 0xFFFF00FF, 0x00000000, 0xFFFFFF00, 
    0x00000000, 0xFFFFFFFF, 0x000000FF, 0x00000000, 0x000000FF, 0x000000FF, 
    0x000000FF, 0x0000FF00, 0x000000FF, 0x0000FFFF, 0x000000FF, 0x00FF0000, 
    0x000000FF, 0x00FF00FF, 0x000000FF, 0x00FFFF00, 0x000000FF, 0x00FFFFFF, 
    0x000000FF, 0xFF000000, 0x000000FF, 0xFF0000FF, 0x000000FF, 0xFF00FF00, 
    0x000000FF, 0xFF00FFFF, 0x000000FF, 0xFFFF0000, 0x000000FF, 0xFFFF00FF, 
    0x000000FF, 0xFFFFFF00, 0x000000FF, 0xFFFFFFFF, 0x0000FF00, 0x00000000, 
    0x0000FF00, 0x000000FF, 0x0000FF00, 0x0000FF00, 0x0000FF00, 0x0000FFFF, 
    0x0000FF00, 0x00FF0000, 0x0000FF00, 0x00FF00FF, 0x0000FF00, 0x00FFFF00, 
    0x0000FF00, 0x00FFFFFF, 0x0000FF00, 0xFF000000, 0x0000FF00, 0xFF0000FF, 
    0x0000FF00, 0xFF00FF00, 0x0000FF00, 0xFF00FFFF, 0x0000FF00, 0xFFFF0000, 
    0x0000FF00, 0xFFFF00FF, 0x0000FF00, 0xFFFFFF00, 0x0000FF00, 0xFFFFFFFF, 
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0x000000FF, 0x0000FFFF, 0x0000FF00, 
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0x00FF0000, 0x0000FFFF, 0x00FF00FF, 
    0x0000FFFF, 0x00FFFF00, 0x0000FFFF, 0x00FFFFFF, 0x0000FFFF, 0xFF000000, 
    0x0000FFFF, 0xFF0000FF, 0x0000FFFF, 0xFF00FF00, 0x0000FFFF, 0xFF00FFFF, 
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFF00FF, 0x0000FFFF, 0xFFFFFF00, 
    0x0000FFFF, 0xFFFFFFFF, 0x00FF0000, 0x00000000, 0x00FF0000, 0x000000FF, 
    0x00FF0000, 0x0000FF00, 0x00FF0000, 0x0000FFFF, 0x00FF0000, 0x00FF0000, 
    0x00FF0000, 0x00FF00FF, 0x00FF0000, 0x00FFFF00, 0x00FF0000, 0x00FFFFFF, 
    0x00FF0000, 0xFF000000, 0x00FF0000, 0xFF0000FF, 0x00FF0000, 0xFF00FF00, 
    0x00FF0000, 0xFF00FFFF, 0x00FF0000, 0xFFFF0000, 0x00FF0000, 0xFFFF00FF, 
    0x00FF0000, 0xFFFFFF00, 0x00FF0000, 0xFFFFFFFF, 0x00FF00FF, 0x00000000, 
    0x00FF00FF, 0x000000FF, 0x00FF00FF, 0x0000FF00, 0x00FF00FF, 0x0000FFFF, 
    0x00FF00FF, 0x00FF0000, 0x00FF00FF, 0x00FF00FF, 0x00FF00FF, 0x00FFFF00, 
    0x00FF00FF, 0x00FFFFFF, 0x00FF00FF, 0xFF000000, 0x00FF00FF, 0xFF0000FF, 
    0x00FF00FF, 0xFF00FF00, 0x00FF00FF, 0xFF00FFFF, 0x00FF00FF, 0xFFFF0000, 
    0x00FF00FF, 0xFFFF00FF, 0x00FF00FF, 0xFFFFFF00, 0x00FF00FF, 0xFFFFFFFF, 
    0x00FFFF00, 0x00000000, 0x00FFFF00, 0x000000FF, 0x00FFFF00, 0x0000FF00, 
    0x00FFFF00, 0x0000FFFF, 0x00FFFF00, 0x00FF0000, 0x00FFFF00, 0x00FF00FF, 
    0x00FFFF00, 0x00FFFF00, 0x00FFFF00, 0x00FFFFFF, 0x00FFFF00, 0xFF000000, 
    0x00FFFF00, 0xFF0000FF, 0x00FFFF00, 0xFF00FF00, 0x00FFFF00, 0xFF00FFFF, 
    0x00FFFF00, 0xFFFF0000, 0x00FFFF00, 0xFFFF00FF, 0x00FFFF00, 0xFFFFFF00, 
    0x00FFFF00, 0xFFFFFFFF, 0x00FFFFFF, 0x00000000, 0x00FFFFFF, 0x000000FF, 
    0x00FFFFFF, 0x0000FF00, 0x00FFFFFF, 0x0000FFFF, 0x00FFFFFF, 0x00FF0000, 
    0x00FFFFFF, 0x00FF00FF, 0x00FFFFFF, 0x00FFFF00, 0x00FFFFFF, 0x00FFFFFF, 
    0x00FFFFFF, 0xFF000000, 0x00FFFFFF, 0xFF0000FF, 0x00FFFFFF, 0xFF00FF00, 
    0x00FFFFFF, 0xFF00FFFF, 0x00FFFFFF, 0xFFFF0000, 0x00FFFFFF, 0xFFFF00FF, 
    0x00FFFFFF, 0xFFFFFF00, 0x00FFFFFF, 0xFFFFFFFF, 0xFF000000, 0x00000000, 
    0xFF000000, 0x000000FF, 0xFF000000, 0x0000FF00, 0xFF000000, 0x0000FFFF, 
    0xFF000000, 0x00FF0000, 0xFF000000, 0x00FF00FF, 0xFF000000, 0x00FFFF00, 
    0xFF000000, 0x00FFFFFF, 0xFF000000, 0xFF000000, 0xFF000000, 0xFF0000FF, 
    0xFF000000, 0xFF00FF00, 0xFF000000, 0xFF00FFFF, 0xFF000000, 0xFFFF0000, 
    0xFF000000, 0xFFFF00FF, 0xFF000000, 0xFFFFFF00, 0xFF000000, 0xFFFFFFFF, 
    0xFF0000FF, 0x00000000, 0xFF0000FF, 0x000000FF, 0xFF0000FF, 0x0000FF00, 
    0xFF0000FF, 0x0000FFFF, 0xFF0000FF, 0x00FF0000, 0xFF0000FF, 0x00FF00FF, 
    0xFF0000FF, 0x00FFFF00, 0xFF0000FF, 0x00FFFFFF, 0xFF0000FF, 0xFF000000, 
    0xFF0000FF, 0xFF0000FF, 0xFF0000FF, 0xFF00FF00, 0xFF0000FF, 0xFF00FFFF, 
    0xFF0000FF, 0xFFFF0000, 0xFF0000FF, 0xFFFF00FF, 0xFF0000FF, 0xFFFFFF00, 
    0xFF0000FF, 0xFFFFFFFF, 0xFF00FF00, 0x00000000, 0xFF00FF00, 0x000000FF, 
    0xFF00FF00, 0x0000FF00, 0xFF00FF00, 0x0000FFFF, 0xFF00FF00, 0x00FF0000, 
    0xFF00FF00, 0x00FF00FF, 0xFF00FF00, 0x00FFFF00, 0xFF00FF00, 0x00FFFFFF, 
    0xFF00FF00, 0xFF000000, 0xFF00FF00, 0xFF0000FF, 0xFF00FF00, 0xFF00FF00, 
    0xFF00FF00, 0xFF00FFFF, 0xFF00FF00, 0xFFFF0000, 0xFF00FF00, 0xFFFF00FF, 
    0xFF00FF00, 0xFFFFFF00, 0xFF00FF00, 0xFFFFFFFF, 0xFF00FFFF, 0x00000000, 
    0xFF00FFFF, 0x000000FF, 0xFF00FFFF, 0x0000FF00, 0xFF00FFFF, 0x0000FFFF, 
    0xFF00FFFF, 0x00FF0000, 0xFF00FFFF, 0x00FF00FF, 0xFF00FFFF, 0x00FFFF00, 
    0xFF00FFFF, 0x00FFFFFF, 0xFF00FFFF, 0xFF000000, 0xFF00FFFF, 0xFF0000FF, 
    0xFF00FFFF, 0xFF00FF00, 0xFF00FFFF, 0xFF00FFFF, 0xFF00FFFF, 0xFFFF0000, 
    0xFF00FFFF, 0xFFFF00FF, 0xFF00FFFF, 0xFFFFFF00, 0xFF00FFFF, 0xFFFFFFFF, 
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0x000000FF, 0xFFFF0000, 0x0000FF00, 
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0x00FF0000, 0xFFFF0000, 0x00FF00FF, 
    0xFFFF0000, 0x00FFFF00, 0xFFFF0000, 0x00FFFFFF, 0xFFFF0000, 0xFF000000, 
    0xFFFF0000, 0xFF0000FF, 0xFFFF0000, 0xFF00FF00, 0xFFFF0000, 0xFF00FFFF, 
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0xFFFF00FF, 0xFFFF0000, 0xFFFFFF00, 
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF00FF, 0x00000000, 0xFFFF00FF, 0x000000FF, 
    0xFFFF00FF, 0x0000FF00, 0xFFFF00FF, 0x0000FFFF, 0xFFFF00FF, 0x00FF0000, 
    0xFFFF00FF, 0x00FF00FF, 0xFFFF00FF, 0x00FFFF00, 0xFFFF00FF, 0x00FFFFFF, 
    0xFFFF00FF, 0xFF000000, 0xFFFF00FF, 0xFF0000FF, 0xFFFF00FF, 0xFF00FF00, 
    0xFFFF00FF, 0xFF00FFFF, 0xFFFF00FF, 0xFFFF0000, 0xFFFF00FF, 0xFFFF00FF, 
    0xFFFF00FF, 0xFFFFFF00, 0xFFFF00FF, 0xFFFFFFFF, 0xFFFFFF00, 0x00000000, 
    0xFFFFFF00, 0x000000FF, 0xFFFFFF00, 0x0000FF00, 0xFFFFFF00, 0x0000FFFF, 
    0xFFFFFF00, 0x00FF0000, 0xFFFFFF00, 0x00FF00FF, 0xFFFFFF00, 0x00FFFF00, 
    0xFFFFFF00, 0x00FFFFFF, 0xFFFFFF00, 0xFF000000, 0xFFFFFF00, 0xFF0000FF, 
    0xFFFFFF00, 0xFF00FF00, 0xFFFFFF00, 0xFF00FFFF, 0xFFFFFF00, 0xFFFF0000, 
    0xFFFFFF00, 0xFFFF00FF, 0xFFFFFF00, 0xFFFFFF00, 0xFFFFFF00, 0xFFFFFFFF, 
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x000000FF, 0xFFFFFFFF, 0x0000FF00, 
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0x00FF0000, 0xFFFFFFFF, 0x00FF00FF, 
    0xFFFFFFFF, 0x00FFFF00, 0xFFFFFFFF, 0x00FFFFFF, 0xFFFFFFFF, 0xFF000000, 
    0xFFFFFFFF, 0xFF0000FF, 0xFFFFFFFF, 0xFF00FF00, 0xFFFFFFFF, 0xFF00FFFF, 
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF00FF, 0xFFFFFFFF, 0xFFFFFF00, 
    0xFFFFFFFF, 0xFFFFFFFF};

static UInt32 invertResultBits[] = 
{   0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFF00, 0xFFFFFFFF, 0xFFFF00FF, 
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFF00FFFF, 0xFFFFFFFF, 0xFF00FF00, 
    0xFFFFFFFF, 0xFF0000FF, 0xFFFFFFFF, 0xFF000000, 0xFFFFFFFF, 0x00FFFFFF, 
    0xFFFFFFFF, 0x00FFFF00, 0xFFFFFFFF, 0x00FF00FF, 0xFFFFFFFF, 0x00FF0000, 
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0x0000FF00, 0xFFFFFFFF, 0x000000FF, 
    0xFFFFFFFF, 0x00000000, 0xFFFFFF00, 0xFFFFFFFF, 0xFFFFFF00, 0xFFFFFF00, 
    0xFFFFFF00, 0xFFFF00FF, 0xFFFFFF00, 0xFFFF0000, 0xFFFFFF00, 0xFF00FFFF, 
    0xFFFFFF00, 0xFF00FF00, 0xFFFFFF00, 0xFF0000FF, 0xFFFFFF00, 0xFF000000, 
    0xFFFFFF00, 0x00FFFFFF, 0xFFFFFF00, 0x00FFFF00, 0xFFFFFF00, 0x00FF00FF, 
    0xFFFFFF00, 0x00FF0000, 0xFFFFFF00, 0x0000FFFF, 0xFFFFFF00, 0x0000FF00, 
    0xFFFFFF00, 0x000000FF, 0xFFFFFF00, 0x00000000, 0xFFFF00FF, 0xFFFFFFFF, 
    0xFFFF00FF, 0xFFFFFF00, 0xFFFF00FF, 0xFFFF00FF, 0xFFFF00FF, 0xFFFF0000, 
    0xFFFF00FF, 0xFF00FFFF, 0xFFFF00FF, 0xFF00FF00, 0xFFFF00FF, 0xFF0000FF, 
    0xFFFF00FF, 0xFF000000, 0xFFFF00FF, 0x00FFFFFF, 0xFFFF00FF, 0x00FFFF00, 
    0xFFFF00FF, 0x00FF00FF, 0xFFFF00FF, 0x00FF0000, 0xFFFF00FF, 0x0000FFFF, 
    0xFFFF00FF, 0x0000FF00, 0xFFFF00FF, 0x000000FF, 0xFFFF00FF, 0x00000000, 
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFF00, 0xFFFF0000, 0xFFFF00FF, 
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0xFF00FFFF, 0xFFFF0000, 0xFF00FF00, 
    0xFFFF0000, 0xFF0000FF, 0xFFFF0000, 0xFF000000, 0xFFFF0000, 0x00FFFFFF, 
    0xFFFF0000, 0x00FFFF00, 0xFFFF0000, 0x00FF00FF, 0xFFFF0000, 0x00FF0000, 
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0x0000FF00, 0xFFFF0000, 0x000000FF, 
    0xFFFF0000, 0x00000000, 0xFF00FFFF, 0xFFFFFFFF, 0xFF00FFFF, 0xFFFFFF00, 
    0xFF00FFFF, 0xFFFF00FF, 0xFF00FFFF, 0xFFFF0000, 0xFF00FFFF, 0xFF00FFFF, 
    0xFF00FFFF, 0xFF00FF00, 0xFF00FFFF, 0xFF0000FF, 0xFF00FFFF, 0xFF000000, 
    0xFF00FFFF, 0x00FFFFFF, 0xFF00FFFF, 0x00FFFF00, 0xFF00FFFF, 0x00FF00FF, 
    0xFF00FFFF, 0x00FF0000, 0xFF00FFFF, 0x0000FFFF, 0xFF00FFFF, 0x0000FF00, 
    0xFF00FFFF, 0x000000FF, 0xFF00FFFF, 0x00000000, 0xFF00FF00, 0xFFFFFFFF, 
    0xFF00FF00, 0xFFFFFF00, 0xFF00FF00, 0xFFFF00FF, 0xFF00FF00, 0xFFFF0000, 
    0xFF00FF00, 0xFF00FFFF, 0xFF00FF00, 0xFF00FF00, 0xFF00FF00, 0xFF0000FF, 
    0xFF00FF00, 0xFF000000, 0xFF00FF00, 0x00FFFFFF, 0xFF00FF00, 0x00FFFF00, 
    0xFF00FF00, 0x00FF00FF, 0xFF00FF00, 0x00FF0000, 0xFF00FF00, 0x0000FFFF, 
    0xFF00FF00, 0x0000FF00, 0xFF00FF00, 0x000000FF, 0xFF00FF00, 0x00000000, 
    0xFF0000FF, 0xFFFFFFFF, 0xFF0000FF, 0xFFFFFF00, 0xFF0000FF, 0xFFFF00FF, 
    0xFF0000FF, 0xFFFF0000, 0xFF0000FF, 0xFF00FFFF, 0xFF0000FF, 0xFF00FF00, 
    0xFF0000FF, 0xFF0000FF, 0xFF0000FF, 0xFF000000, 0xFF0000FF, 0x00FFFFFF, 
    0xFF0000FF, 0x00FFFF00, 0xFF0000FF, 0x00FF00FF, 0xFF0000FF, 0x00FF0000, 
    0xFF0000FF, 0x0000FFFF, 0xFF0000FF, 0x0000FF00, 0xFF0000FF, 0x000000FF, 
    0xFF0000FF, 0x00000000, 0xFF000000, 0xFFFFFFFF, 0xFF000000, 0xFFFFFF00, 
    0xFF000000, 0xFFFF00FF, 0xFF000000, 0xFFFF0000, 0xFF000000, 0xFF00FFFF, 
    0xFF000000, 0xFF00FF00, 0xFF000000, 0xFF0000FF, 0xFF000000, 0xFF000000, 
    0xFF000000, 0x00FFFFFF, 0xFF000000, 0x00FFFF00, 0xFF000000, 0x00FF00FF, 
    0xFF000000, 0x00FF0000, 0xFF000000, 0x0000FFFF, 0xFF000000, 0x0000FF00, 
    0xFF000000, 0x000000FF, 0xFF000000, 0x00000000, 0x00FFFFFF, 0xFFFFFFFF, 
    0x00FFFFFF, 0xFFFFFF00, 0x00FFFFFF, 0xFFFF00FF, 0x00FFFFFF, 0xFFFF0000, 
    0x00FFFFFF, 0xFF00FFFF, 0x00FFFFFF, 0xFF00FF00, 0x00FFFFFF, 0xFF0000FF, 
    0x00FFFFFF, 0xFF000000, 0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF, 0x00FFFF00, 
    0x00FFFFFF, 0x00FF00FF, 0x00FFFFFF, 0x00FF0000, 0x00FFFFFF, 0x0000FFFF, 
    0x00FFFFFF, 0x0000FF00, 0x00FFFFFF, 0x000000FF, 0x00FFFFFF, 0x00000000, 
    0x00FFFF00, 0xFFFFFFFF, 0x00FFFF00, 0xFFFFFF00, 0x00FFFF00, 0xFFFF00FF, 
    0x00FFFF00, 0xFFFF0000, 0x00FFFF00, 0xFF00FFFF, 0x00FFFF00, 0xFF00FF00, 
    0x00FFFF00, 0xFF0000FF, 0x00FFFF00, 0xFF000000, 0x00FFFF00, 0x00FFFFFF, 
    0x00FFFF00, 0x00FFFF00, 0x00FFFF00, 0x00FF00FF, 0x00FFFF00, 0x00FF0000, 
    0x00FFFF00, 0x0000FFFF, 0x00FFFF00, 0x0000FF00, 0x00FFFF00, 0x000000FF, 
    0x00FFFF00, 0x00000000, 0x00FF00FF, 0xFFFFFFFF, 0x00FF00FF, 0xFFFFFF00, 
    0x00FF00FF, 0xFFFF00FF, 0x00FF00FF, 0xFFFF0000, 0x00FF00FF, 0xFF00FFFF, 
    0x00FF00FF, 0xFF00FF00, 0x00FF00FF, 0xFF0000FF, 0x00FF00FF, 0xFF000000, 
    0x00FF00FF, 0x00FFFFFF, 0x00FF00FF, 0x00FFFF00, 0x00FF00FF, 0x00FF00FF, 
    0x00FF00FF, 0x00FF0000, 0x00FF00FF, 0x0000FFFF, 0x00FF00FF, 0x0000FF00, 
    0x00FF00FF, 0x000000FF, 0x00FF00FF, 0x00000000, 0x00FF0000, 0xFFFFFFFF, 
    0x00FF0000, 0xFFFFFF00, 0x00FF0000, 0xFFFF00FF, 0x00FF0000, 0xFFFF0000, 
    0x00FF0000, 0xFF00FFFF, 0x00FF0000, 0xFF00FF00, 0x00FF0000, 0xFF0000FF, 
    0x00FF0000, 0xFF000000, 0x00FF0000, 0x00FFFFFF, 0x00FF0000, 0x00FFFF00, 
    0x00FF0000, 0x00FF00FF, 0x00FF0000, 0x00FF0000, 0x00FF0000, 0x0000FFFF, 
    0x00FF0000, 0x0000FF00, 0x00FF0000, 0x000000FF, 0x00FF0000, 0x00000000, 
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFF00, 0x0000FFFF, 0xFFFF00FF, 
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0xFF00FFFF, 0x0000FFFF, 0xFF00FF00, 
    0x0000FFFF, 0xFF0000FF, 0x0000FFFF, 0xFF000000, 0x0000FFFF, 0x00FFFFFF, 
    0x0000FFFF, 0x00FFFF00, 0x0000FFFF, 0x00FF00FF, 0x0000FFFF, 0x00FF0000, 
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0x0000FF00, 0x0000FFFF, 0x000000FF, 
    0x0000FFFF, 0x00000000, 0x0000FF00, 0xFFFFFFFF, 0x0000FF00, 0xFFFFFF00, 
    0x0000FF00, 0xFFFF00FF, 0x0000FF00, 0xFFFF0000, 0x0000FF00, 0xFF00FFFF, 
    0x0000FF00, 0xFF00FF00, 0x0000FF00, 0xFF0000FF, 0x0000FF00, 0xFF000000, 
    0x0000FF00, 0x00FFFFFF, 0x0000FF00, 0x00FFFF00, 0x0000FF00, 0x00FF00FF, 
    0x0000FF00, 0x00FF0000, 0x0000FF00, 0x0000FFFF, 0x0000FF00, 0x0000FF00, 
    0x0000FF00, 0x000000FF, 0x0000FF00, 0x00000000, 0x000000FF, 0xFFFFFFFF, 
    0x000000FF, 0xFFFFFF00, 0x000000FF, 0xFFFF00FF, 0x000000FF, 0xFFFF0000, 
    0x000000FF, 0xFF00FFFF, 0x000000FF, 0xFF00FF00, 0x000000FF, 0xFF0000FF, 
    0x000000FF, 0xFF000000, 0x000000FF, 0x00FFFFFF, 0x000000FF, 0x00FFFF00, 
    0x000000FF, 0x00FF00FF, 0x000000FF, 0x00FF0000, 0x000000FF, 0x0000FFFF, 
    0x000000FF, 0x0000FF00, 0x000000FF, 0x000000FF, 0x000000FF, 0x00000000, 
    0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFFFF00, 0x00000000, 0xFFFF00FF, 
    0x00000000, 0xFFFF0000, 0x00000000, 0xFF00FFFF, 0x00000000, 0xFF00FF00, 
    0x00000000, 0xFF0000FF, 0x00000000, 0xFF000000, 0x00000000, 0x00FFFFFF, 
    0x00000000, 0x00FFFF00, 0x00000000, 0x00FF00FF, 0x00000000, 0x00FF0000, 
    0x00000000, 0x0000FFFF, 0x00000000, 0x0000FF00, 0x00000000, 0x000000FF, 
    0x00000000, 0x00000000};    

Int32  CDECL 
i_depth1To8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    Bool	invert)
{
    PROLOGUE;
    UInt32     *pDLong;
    UInt8	edgeBits;
    Int32	i, j, bytes, index, bitsLastByte;

    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 8*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
   
	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

    bytes = w / 8;
    if (w % 8)
	bytes++;

    bitsLastByte = w % 8;
    if (bitsLastByte == 0)
	bitsLastByte = 8;

    if (invert == cTrue)
    {
	/* Make a mask for the last byte on a line--keep the bits actually
	 * in the image, and set the rest to 0.  Then invert the
	 * whole thing so that we can OR the 1's on the end into the
	 * index, which will result in 0's off the edge of the image.
	 */
	edgeBits = ~(((1 << bitsLastByte) - 1) << (8 - bitsLastByte)) ;

	/* Crank through. We handle the last byte separately */
	for (i = 0; i < h; i++)
        {
	    pDLong = (UInt32 *) pD;
	    for (j = 0; j < bytes - 1; j++)
	    {
		index = (UCHAR_ACCESS (pS + j)) << 1;
		ULONG_ACCESS(pDLong++) = invertResultBits[index];
		ULONG_ACCESS(pDLong++) = invertResultBits[index + 1];
	    }

	    /* Handle the last 8 bits.  Remember that we're
	     * inverting stuff here, so 0's in the index turn
	     * into 1's in the image, or in the *frame*.
	     * So, we use the inverted edge mask to OR 1's
	     * into the index on the end, which produces
	     * 0's off the edge of the image as it should.
	     */
	    index = (UCHAR_ACCESS(pS + j) | edgeBits) << 1;
	    
	    ULONG_ACCESS(pDLong++) = invertResultBits[index];
	    /* Don't write that final word if it's not there */
	    if (bitsLastByte > 4)
	        ULONG_ACCESS(pDLong++) = invertResultBits[index + 1];
	
	    pS += sBpl;
	    pD += dBpl;
	}
    }
    else
    {
	/* Make a mask for the last byte on a line--keep the bits actually
	   in the image, and set the rest to 1 */
	edgeBits = (((1 << bitsLastByte) - 1) << (8 - bitsLastByte));

	/* Crank through. We handle the last byte separately */
	for (i = 0; i < h; i++)
        {
	    pDLong = (UInt32 *) pD;
	    for (j = 0; j < bytes - 1; j++)
	    {
		index = (UCHAR_ACCESS (pS + j)) << 1;
		ULONG_ACCESS(pDLong++) = resultBits[index];
		ULONG_ACCESS(pDLong++) = resultBits[index + 1];
	    }

	    /* Handle the last 8 bits */
	    index = (UCHAR_ACCESS(pS + j) & edgeBits) << 1;

	    ULONG_ACCESS(pDLong++) = resultBits[index];
	    /* Don't write that final word if it's not there */
	    if (bitsLastByte > 4)
	        ULONG_ACCESS(pDLong++) = resultBits[index + 1];
	
	    pS += sBpl;
	    pD += dBpl;
	}
    }

    NORMAL_RETURN;
}

/* PUBLIC */
/*************************************************************************** 
 * i_depth1To16bpp
 *		Scale a 1-bit gray image to an 16-bit gray image.
 *		0 becomes 0, 1 becomes 0xffff (or inverted). The
 *		best strategies to do this are probably different
 *		on the Sparc and x86.
 *		We're going to make a byte lookup table (resulting in
 *		four longwords). On the 486, we're better off reading
 *		the source byte-at-a-time, but on the Sparc, we'd be
 *		better off reading the source long-word-at-a-time. We're
 *		going to do byte-at-a-time, assuming that if one needs
 *		fast Sparc code, one would be better off writing it in
 *		assembly code anyway.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *		invert=	If true, invert the result, so a 1 in the source
 *			image yields a 0 in the result, and a 0 yeilds ff.
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/


static UInt32 resultBits16[] =
{   0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x0000FFFF,
    0x00000000, 0x00000000, 0x00000000, 0xFFFF0000,
    0x00000000, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0x0000FFFF, 0x00000000,
    0x00000000, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0xFFFF0000, 0x00000000,
    0x00000000, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,

    0x00000000, 0x0000FFFF, 0x00000000, 0x00000000,
    0x00000000, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0x00000000, 0xFFFF0000, 0x00000000, 0x00000000,
    0x00000000, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,

    0x00000000, 0xFFFFFFFF, 0x00000000, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0x0000FFFF, 0x00000000, 0x00000000, 0x00000000,
    0x0000FFFF, 0x00000000, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,

    0x0000FFFF, 0x0000FFFF, 0x00000000, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0x0000FFFF, 0xFFFF0000, 0x00000000, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,

    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFF0000, 0x00000000, 0x00000000, 0x00000000,
    0xFFFF0000, 0x00000000, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFF0000, 0x0000FFFF, 0x00000000, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFF0000, 0xFFFF0000, 0x00000000, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFFFFFF, 0x00000000, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,

    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF
};

static UInt32 invertResultBits16[] = 
{   0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x00000000,

    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0xFFFF0000, 0x00000000, 0x00000000,

    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0x0000FFFF, 0x00000000, 0x00000000,

    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0xFFFF0000, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0x0000FFFF, 0x00000000,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0xFFFF0000,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0x0000FFFF,
    0xFFFFFFFF, 0x00000000, 0x00000000, 0x00000000,

    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0xFFFFFFFF, 0x00000000, 0x00000000,

    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0xFFFF0000, 0x00000000, 0x00000000,

    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0x0000FFFF, 0x00000000, 0x00000000,

    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0xFFFF0000, 0x00000000,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0x0000FFFF, 0x00000000,
    0xFFFF0000, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0xFFFF0000, 0x00000000, 0x00000000, 0xFFFF0000,
    0xFFFF0000, 0x00000000, 0x00000000, 0x0000FFFF,
    0xFFFF0000, 0x00000000, 0x00000000, 0x00000000,

    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0xFFFFFFFF, 0x00000000, 0x00000000,

    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0xFFFF0000, 0x00000000, 0x00000000,

    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0x0000FFFF, 0x00000000, 0x00000000,

    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0xFFFF0000, 0x00000000,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0x0000FFFF, 0x00000000,
    0x0000FFFF, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0x0000FFFF, 0x00000000, 0x00000000, 0xFFFF0000,
    0x0000FFFF, 0x00000000, 0x00000000, 0x0000FFFF,
    0x0000FFFF, 0x00000000, 0x00000000, 0x00000000,

    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0xFFFF0000, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0x0000FFFF, 0x00000000,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFF0000,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0x0000FFFF,
    0x00000000, 0xFFFFFFFF, 0x00000000, 0x00000000,

    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0xFFFF0000, 0x00000000,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0x0000FFFF, 0x00000000,
    0x00000000, 0xFFFF0000, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0xFFFF0000, 0x00000000, 0xFFFF0000,
    0x00000000, 0xFFFF0000, 0x00000000, 0x0000FFFF,
    0x00000000, 0xFFFF0000, 0x00000000, 0x00000000,

    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0xFFFF0000, 0x00000000,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0x0000FFFF, 0x00000000,
    0x00000000, 0x0000FFFF, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0x0000FFFF, 0x00000000, 0xFFFF0000,
    0x00000000, 0x0000FFFF, 0x00000000, 0x0000FFFF,
    0x00000000, 0x0000FFFF, 0x00000000, 0x00000000,

    0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFF0000,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0x0000FFFF,
    0x00000000, 0x00000000, 0xFFFFFFFF, 0x00000000,
    0x00000000, 0x00000000, 0xFFFF0000, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0xFFFF0000, 0xFFFF0000,
    0x00000000, 0x00000000, 0xFFFF0000, 0x0000FFFF,
    0x00000000, 0x00000000, 0xFFFF0000, 0x00000000,
    0x00000000, 0x00000000, 0x0000FFFF, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0x0000FFFF, 0xFFFF0000,
    0x00000000, 0x00000000, 0x0000FFFF, 0x0000FFFF,
    0x00000000, 0x00000000, 0x0000FFFF, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0xFFFFFFFF,
    0x00000000, 0x00000000, 0x00000000, 0xFFFF0000,
    0x00000000, 0x00000000, 0x00000000, 0x0000FFFF,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
};    

Int32  CDECL 
i_depth1To16bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    Bool	invert)
{
    PROLOGUE;
    UInt32     *pDLong;
    UInt8	edgeBits;
    Int32	i, j, bytes, index, bitsLastByte;

    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 16*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;
   
	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

	/* compute the number of whole bytes in each scanline */
    bytes = (w + 7) / 8;

	/* Compute number of bits in last byte */
    bitsLastByte = ((w-1) % 8) + 1;

    if (invert == cTrue)
    {
	/* Make a mask for the last byte on a line--keep the bits actually
	 * in the image, and set the rest to 0.  Then invert the
	 * whole thing so that we can OR the 1's on the end into the
	 * index, which will result in 0's off the edge of the image.
	 */
	edgeBits = ~(((1 << bitsLastByte) - 1) << (8 - bitsLastByte)) ;

	/* Crank through. We handle the last byte separately */
	for (i = 0; i < h; i++)
        {
	    pDLong = (UInt32 *) pD;
	    for (j = 0; j < bytes - 1; j++)
	    {
		index = (UCHAR_ACCESS (pS + j)) << 2;
		*pDLong++ = invertResultBits16[index];
		*pDLong++ = invertResultBits16[index + 1];
		*pDLong++ = invertResultBits16[index + 2];
		*pDLong++ = invertResultBits16[index + 3];
	    }

	    /* Handle the last 8 bits.  Remember that we're
	     * inverting stuff here, so 0's in the index turn
	     * into 1's in the image, or in the *frame*.
	     * So, we use the inverted edge mask to OR 1's
	     * into the index on the end, which produces
	     * 0's off the edge of the image as it should.
	     */
	    index = (UCHAR_ACCESS(pS + j) | edgeBits) << 2;

	    *pDLong++ = invertResultBits16[index];
	    /* Don't write that final word if it's not there */
	    if (bitsLastByte > 2)
	        *pDLong++ = invertResultBits16[index + 1];
	    if (bitsLastByte > 4)
	        *pDLong++ = invertResultBits16[index + 2];
	    if (bitsLastByte > 6)
	        *pDLong++ = invertResultBits16[index + 3];
	
	    pS += sBpl;
	    pD += dBpl;
	}
    }
    else
    {
	/* Make a mask for the last byte on a line--keep the bits actually
	   in the image, and set the rest to 0 */
	edgeBits = ((1 << bitsLastByte) - 1) << (8 - bitsLastByte) ;

	/* Crank through. We handle the last byte separately */
	for (i = 0; i < h; i++)
        {
	    pDLong = (UInt32 *) pD;
	    for (j = 0; j < bytes - 1; j++)
	    {
		index = (UCHAR_ACCESS (pS + j)) << 2;
		*pDLong++ = resultBits16[index];
		*pDLong++ = resultBits16[index + 1];
		*pDLong++ = resultBits16[index + 2];
		*pDLong++ = resultBits16[index + 3];
	    }

	    /* Handle the last 8 bits */
	    index = (UCHAR_ACCESS(pS + j) & edgeBits) << 2;

	    *pDLong++ = resultBits16[index];
	    /* Don't write that final word if it's not there */
	    if (bitsLastByte > 2)
	        *pDLong++ = resultBits16[index + 1];
	    if (bitsLastByte > 4)
	        *pDLong++ = resultBits16[index + 2];
	    if (bitsLastByte > 6)
	        *pDLong++ = resultBits16[index + 3];

	    pS += sBpl;
	    pD += dBpl;
	}
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************** 
 * i_map1To8bpp
 *		Scale a 1-bit gray image to an 8-bit gray image.
 *              The caller specifies the gray values to which 0 and 1
 *              will be mapped.
 *
 *	Input:
 *		pS =    Pointer to upper left corner of source image's frame
 *      	w =     width of image in pixels including frame
 *      	h =     height of image in pixels including frame
 *      	sBpl =  bytes/line of source including frame.
 *		pD =    Pointer to upper left corner of destination image's
 *			frame
 *      	dBpl =  bytes/line of destination including frame.
 *              val0 =  Value to paint when 0 is found in the binary image
 *              val1 =  Value to paint when 1 is found in the binary image.
 *
 *	Return code:
 *		ia_successful	good completion
 *		ia_nomem	unable to allocate memory
 *
 *	Assumes destination already includes cleared frame.
 **************************************************************************/

Int32 CDECL 
i_map1To8bpp (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    UInt8      *pD,
    Int32	dBpl,
    UInt8	val0,
    UInt8	val1)
{
    PROLOGUE;
    UInt32     *lut, *pDLong;
    UInt8      *bLut;
    Int32	i, j, bytes, bitsLastByte, index;
    UInt32	lastMask[9] = {0x00000000,	/* unused */
			       0xFF000000,	/* 1 bit in last byte */
			       0xFFFF0000,	/* 2 bits */
			       0xFFFFFF00,	/* 3 bits */
			       0xFFFFFFFF,	/* 4 bits */
			       0xFF000000,	/* 5 bits, applied to next wd */
			       0xFFFF0000,	/* 6 bits */
			       0xFFFFFF00,	/* 7 bits */
			       0xFFFFFFFF};	/* 8 bits */

    pS += (sBpl << cLogFrameBits) + cFrameBytes;
    pD += (dBpl << cLogFrameBits) + 8*cFrameBytes;
    w -= 2*cFrameBits;
    h -= 2*cFrameBits;

	/* if w == 0, just return instead of blowing up below */
    if (w <= 0)
	NORMAL_RETURN;

    lut = (UInt32 *) MALLOC (512 * sizeof (UInt32));
    if (!lut)
	return ia_nomem;

    /* Fill the look-up table */
    for (i = 0; i < 256; i++)
    {
	bLut = (UInt8 *) &lut[i << 1];
	for (j = 0; j < 8; j++)
	{
	    UCHAR_ACCESS (bLut + j) = (i & (1 << (7 - j))) ? val1 : val0;
	}
    }

    bytes = w / 8;
    if (w % 8)
	bytes++;
    
    bitsLastByte = w % 8;
    if (bitsLastByte == 0)
	bitsLastByte = 8;
    
    /* Crank through. We handle the last byte separately */
    for (i = 0; i < h; i++)
    {
	pDLong = (UInt32 *) pD;
	for (j = 0; j < bytes - 1; j++)
	{
	    index = (UCHAR_ACCESS (pS + j)) << 1;
	    ULONG_ACCESS(pDLong++) = lut[index];
	    ULONG_ACCESS(pDLong++) = lut[index + 1];
	}

	/* Handle the last 8 bits */
	index = (UCHAR_ACCESS(pS + j)) << 1;

	if (bitsLastByte <= 4)
	    ULONG_ACCESS(pDLong++) = lut[index] & lastMask[bitsLastByte];
	else
	    ULONG_ACCESS(pDLong++) = lut[index];

	/* Don't write that final word if it's not there */
	if (bitsLastByte > 4)
	    ULONG_ACCESS(pDLong++) = lut[index + 1] & lastMask[bitsLastByte];
	
	pS += sBpl;
	pD += dBpl;
    }

    FREE (lut);
    NORMAL_RETURN;
}

    
/*

$Log: i_depth.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.21  1998/02/24 03:42:57  ddavies
ddavies on Mon Feb 23 19:39:01 PST 1998

i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v

Add new function i_thresholdByPixel.  This is a new, improved version of
the old w_thresholdByPixel that was in ipwrap.  This function is faster
(2x?) and also fixes a bug that was in the old function.  The old implementation
did not prevent borrows between pixels, so it would either subtract
threshold or (threshold+1) depending on the threshold computed for the
pixel to the right.
This implementation also implements the feature that scales the threshold
image up before using it.  This feature is used by binarize2, which is
our best adaptive binarization heuristic.
This is another one of those cases where the profiling code said that lots
of time was spent in this function, but speeding it up didn't do as much
to the total time as one would expect.

Revision 1.20  1998/02/10 19:46:16  ddavies
ddavies on Tue Feb 10 11:41:03 PST 1998

i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v

Fix indentation of automatic variables.
Fix bug in i_depth1to8 in which it had been writing bits into the right
frame when the gray output was supposed to be inverted relative to the
binary input.
Add new function i_depth1To16bpp which converts binary inputs to 16 bit/pix
output images.  Like i_depth1To8bpp, the output can be inverted.
Fix two giant bugs in i_map1To8bpp (apparently never used before?).
It wasn't adjusting the pointers and dimensions to account for the frame
and it was writing trash into the right frame.

Revision 1.19  1998/01/21 21:23:09  ddavies
i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v

Remove cast in FREE invocation.

Revision 1.18  1998/01/13 23:47:31  ddavies
i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.17  1995/12/13 02:43:12  ejaquith
i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.16  1995/10/18  03:52:35  ddavies
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.15  1995/05/17  01:35:17  macleod
 * macleod on Tue May 16 18:34:47 PDT 1995
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Bits were getting reversed in i_depth1To8bpp on the PC.
 * The regression test didn't notice, because it was counting bits.
 *
 * Revision 1.14  1995/04/27  21:32:09  macleod
 * macleod on Thu Apr 27 14:31:45 PDT 1995
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Added i_map1To8bpp, which does an upconvert from binary to 8 bits/pixel,
 * with the values for "0" and "1" passed in.
 *
 * Revision 1.13  1994/10/28  01:50:12  lperry
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Added an IP_YIELD call to i_map8To8. Its called
 * once per 1024 lines, or about 0.3 seconds.
 *
 * Revision 1.12  1994/09/17  17:50:28  ddavies
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Change the depth conversion procedures so they can either use tables
 * passed in by the client or create and destroy their own.  These procedures
 * are: i_threshold, i_thresholdI, i_threshold2To1bpp, i_thresholdI2To1bpp,
 * i_threshold4To1bpp, i_thresholdI4To1bpp, i_thresholdTo2bpp,
 * i_threshold4To2bpp, i_highpassfilt, i_highpassfiltI,
 * i_highpassfilt4To1bpp, i_highpassfiltI4To1bpp, i_passfilt, i_passfiltI,
 * i_passfilt4To1bpp, i_pasfiltI4To1bpp and i_depth4To8bpp.
 *
 * Revision 1.11  1994/08/01  21:05:57  ddavies
 * ddavies on Mon Aug 1 14:05:35 PDT 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Change "static const" to "static" to work around buggy gcc/Purify interaction.
 *
 * Revision 1.10  1994/07/30  00:03:26  dferrell
 * dferrell on Fri Jul 29 17:02:27 PDT 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Wrote new i_threshold4To2bpp () function.
 *
 * Revision 1.9  1994/07/27  15:28:36  ejaquith
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv".  Old globals are now
 * local pointers.  Changed i_depth1To8bpp function by having all tables
 * expanded (because reentrant code would have bashed the bits tables)
 * and removed globals.
 *
 * Revision 1.8  1994/07/21  00:13:08  ddavies
 * ddavies on Wed Jul 20 17:12:18 PDT 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Fix bug in i_map8To8 - tell ip_clearRightFrame that this is an 8 bit image,
 * not a 4 bit image.  The symptom of this bug was a vertical black stripe
 * down the middle of the image.
 *
 * Revision 1.7  1994/07/11  23:02:01  ddavies
 * ddavies on Mon Jul 11 15:59:14 PDT 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Fix a systematic error in clearing the right frame after operations that
 * could invert input frame bits.  We now use hOrig (the original value of
 * the image height) instead of h as the arg to ip_clearRightFrame.  At
 * the point ip_clearRightFrame is called, h is 0.
 *
 * Revision 1.6  1994/05/10  22:43:26  macleod
 * macleod on Tue May 10 15:42:57 PDT 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Fixed an "endian" problem on PC.
 * (i_depth1to8bpp)
 *
 * Revision 1.5  1994/05/06  02:11:59  macleod
 * macleod on Thu May 5 19:11:52 PDT 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Added depth1To8bpp
 *
 * Revision 1.4  1994/03/24  20:35:41  ddavies
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Add implementations of i_passfilt4To1bpp and i_passfiltI4To1bpp.  These
 * routines convert a 4 bit/pixel image to a 1 bit/pixel image using bandpass
 * filtering.  i_passfiltI4To1bpp inverts the output pixels.
 *
 * Revision 1.3  1994/02/09  22:30:31  macleod
 * macleod on Wed Feb 9 14:30:05 PST 1994
 *
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Added i_depth4To8, which upconverts 4-bit-per-pixel image data into
 * 8bpp.
 *
 * Revision 1.2  1994/02/04  20:14:20  dferrell
 * i_depth.c -> /product/ipcore/ipshared/src/RCS/i_depth.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/01/28  04:09:12  ddavies
 * Initial revision
 *
 * Revision 1.24  1993/12/02  19:21:37  ddavies
 * fprxform.c -> /product/ipcore/ipshared/src/RCS/fprxform.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.23  1993/11/24  20:15:30  macleod
 * fprxform.c -> /product/ipcore/ipshared/src/RCS/fprxform.c,v
 *
 * added i_map8To8
 *
 * Revision 1.22  1993/11/22  22:33:22  jcraig
 * jcraig on Mon Nov 22 14:31:14 PST 1993
 *
 * fprxform.c -> /product/ipcore/ipshared/src/RCS/fprxform.c,v
 *
 * in i_map4To4 keep the pointer to the top of the destination image
 * around so that it can be given to ip_clearRightFrame, instead of the
 * destination pointer that has gone down the image and points to the
 * end of the destination.
 *
 * Revision 1.21  1993/11/17  20:28:51  ddavies
 * ddavies on Wed Nov 17 12:23:36 PST 1993
 *
 * fprxform.c -> /product/ipcore/ipshared/src/RCS/fprxform.c,v
 *
 * Change _SPARC_ARCH_ to "_ALPACA_IMAGE_FMT_ == cBigEndianFmt" in the code
 * that maps 4 bit images to 4 bit or 8 bit images using tables.
 *
 * Revision 1.20  1993/11/16  02:52:37  ddavies
 * fprxform.c -> /product/ipcore/ipshared/src/RCS/fprxform.c,v
 *
 * All of the routines here were writing into the right frame.  They had been
 * using the source and dest bytes/line to find the right ends of the source
 * and dest images.  That doesn't work because Bytes/Line - 2*frame =
 * image + pad; not just image.
 * So, all of these routines now use the image width to calculate the
 * rightmost words of the source and destination images.  In general, the
 * widths have the frames removed from them once at the beginning instead of
 * every time they're used now.
 *
 * Revision 1.19  1993/11/05  01:25:28  ddavies
 * fprxform.c -> /product/ipcore/ipshared/src/RCS/fprxform.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

