/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  pcgenerm.c -- generates MMX assembly routines implementing erode ops
 */


#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#ifdef __WATCOMC__
#include <string.h>
#endif
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "defines.pub"
#include "frames.pub"
#include "memory.pub"

    /* prototypes */
#include "shrmorph.pub"
#include "shrmorph.prv"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: pcgenerm.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

/* This code has the formidible job of generating (nearly) optimal Pentium
 * and Pentium Pro MMX assembly code for a wide range of structure
 * elements to be used for erode (as opposed to dilates).  It's not so
 * bad as it sounds, however.  Code for all the SE's corresponds to a
 * template.  We use the file mortemp3.asm (morphological template)
 * as a souce of assembly code.
 *
 * There are extensive comments in mortemp3.asm.  They explain the algorithm
 * used to do the actual processing.  Seems a waste to duplicate those
 * comments here, but we'll do it anyway.
 *
 * Actually, we'll extract the interesting sections of comments.
 *
 ******************  COMMENTS FROM mortemp3.asm  ********************
 **
 ** This is primarily a template for fast MMX erosion code.  We'll add an OR
 ** operation so it can be used for dilation as well.  This code isn't really
 ** executable, but it has lots of nice fragments that C routines compose into
 ** real erosion and dilation routines.
 **
 ** In subsequent comments, I'll refer to 8 bit things as bytes, 16 bit
 ** things as halfwords, 32 bit things as words and 64 bit things as
 ** doublewords, or dwords.  Intel thinks in terms of bytes, words, doublewords
 ** and quadwords, but tough.
 **
 ** Note that we use no local or global data references, all variables are
 ** referenced as offsets from the ebp register.  The C code that generates
 ** real assembly need not worry about relocable variables.
 **
 ** The method we use for MMX code is very different from the SPARC
 ** method and somewhat different from the method used in regular 386 code.
 **
 ** Before I start talking about details, let me give an overall view of
 ** erosion and dilation.
 **
 ** In binary morphology, we have an input image, an output image and
 ** a structure element, or SE.  The SE is itself an array which contains
 ** a HIT, a MISS or a don't-care term at each location.  Each location
 ** in an SE is called an "SE element".  An active SE element is one with
 ** a HIT or MISS, as opposed to a don't-care term.  An active SE column is
 ** one that contain at least one active SE element.
 **
 ** One of the SE elements is also chosen as the "center".  It need not be
 ** the element that's actually at the midpoint of the SE.
 **
 ** In the simplest (and least efficient) implementation of erosion, one
 ** moves the SE around on the input image.  At every spot where the pattern
 ** is matched on the input image, a 1 is written on the output image at the
 ** location corresponding to the current location of the SE's center.
 ** You can think of the output as a record of all the places that the pattern
 ** specified by the SE was matched.  The pattern may be quite specific and
 ** used to detect particular blobs in the input.  It may contain a lot of
 ** don't-care terms, so only matches loosely defined shapes or textures.
 ** It may also be quite small.  In this case, it matches portions
 ** of all shapes.  For example, one might detect all left edges of
 ** things this way.
 **
 ** Dilation is almost the opposite.  At all places where the center is over
 ** a 1 in the input image, the HITs of the SE are ORed into the output image.
 ** This generally extends the boundaries of things (often things that were
 ** eroded with the same SE) in a structured way.
 **
 ** Moving the SE around to every position in the input is not very efficient.
 ** However, notice that results do not depend on each other so they can be
 ** calculated in parallel.  The cononical way to implement this is to combine
 ** shifted versions of the entire original input image to produce the output.
 ** For example, say the strel contained only two elements, a HIT to the left
 ** of a MISS, with the center over the MISS.  We could AND the image shifted
 ** right (corresponds to the HIT one position to the left of center) with
 ** the inverse of the original image (corresponds to the MISS at the center)
 ** to produce the output.  An output pixel will be set if and only if it
 ** was a 0 in the input image and there was a 1 immediately to the left
 ** of it.  In this view, the SE is really a little program that dictates
 ** a series of shifts and logic operations.  We have C code that compiles
 ** these little programs into machine code.  This file contains code
 ** fragments that act as building blocks when on an MMX machine.
 **
 **				*	*	*
 **
 ** The SPARC algorithms broke the entire image into short blocks, processing
 ** each block vertically.  We moved horizontally from one block to the next.
 ** This was a good compromise between efficient processing (processing
 ** the image vertically minimizes memory writes of partial results) and
 ** cache utilization (minimized by horizontal image reads).  In the first
 ** version of regular 386 algorithm, all inputs read for one dest word
 ** were read again as necessary for the next dest word.  It didn't
 ** matter whether we went horizontally or vertically, so we went
 ** horizontally because it was easier.
 **
 ** In that first version, we also shifted each source word once
 ** for each active SE element.  These are really double shifts because
 ** each shifted word really contains bits from two source words.
 ** That's a lot of slow shift operations.  Double shifts are still
 ** slow in MMX.  We thought we had to recompute every input value
 ** because the 386 doesn't have enough registers to hold lots
 ** of partial outputs.  MMX adds 8 registers, but that's still
 ** not a lot of values.
 **
 ** The second version of 386 morphology saved the shifted versions of each
 ** source word in a temporary buffer.  We saved a shifted source word
 ** for each SE column that had a HIT or MISS in it.  While we still
 ** calculated one destination word at a time, we could usually use
 ** preshifted input words, cutting the 5 (Pentium) cycles used to
 ** do double shifts out of the inner loop.
 **
 ** Huh?  Don't you have to do the shifts anyway?  Well, yes, but we only
 ** have to shift a word once for all the SE elements in a given column
 ** instead of repeating the shift for each SE element in that column.
 ** This optimization clearly works best if there are lots of active
 ** elements in an SE column.
 **
 ** The MMX code follows this idea, with a twist.  Since we have more
 ** MMX registers, we compute 7 destination words at a time, again using
 ** preshifted inputs.  The eighth MMX register holds one of the inputs.
 ** Having lots of partial results available at once allows us to be
 ** more efficient.  It usually allows us to do two AND ops per cycle
 ** in a Pentium instead of only one.  The total number of loads and
 ** stores and ANDs is the same whether we compute 1 result at a time or
 ** multiple results.  Computing multiple results lets us do a better job
 ** of filling clock cycles with 2 operations instead of 1.
 **
 ** Given that we want to use pre-shifted inputs, do we move through the
 ** image horizontally or vertically?  If we moved horizontally (and computed
 ** one output word at a time), each horizontal move would require us to read
 ** a new input word for *each row* of the SE.  That's because each word is
 ** only 1 bit tall.
 **
 ** If we move through the image vertically, we always read one dword and
 ** two words for each move down.  That's because the words are wide enough
 ** to cover the maximum legal shift done in our accelerated code (32 bits).
 ** So, we need to read the dword directly below the previous one and the
 ** words on either side of that.  This seems preferable to the extra
 ** reads horizontal movement would require.
 **
 ** All this means we'll go back to the SPARC method of breaking the
 ** image into relatively short blocks.  Each block will be processed
 ** vertically.  To avoid trashing the cache, we'll move horizontally from
 ** block to block.
 **
 ** How large is the buffer of preshifted inputs?  It will be at
 ** least as wide as the SE, except each active SE column will correspond
 ** to a (wordswapped, shifted) 64-bit input dword.  Because inverting
 ** words to process MISS elements is a pain when doing the ANDs, we'll
 ** actually store both positive and inverted versions of the shifted
 ** inputs as needed.
 **
 ** Editorial Comment: Can you believe the idiocy of the MMX designers in
 ** designing the pandn opcode?  Instead of inverting the source, then
 ** ANDing with the dest, it inverts the dest, then ANDs with the source.
 ** Gag me.  This makes it useless for combining a dest with a bunch of
 ** inverted and non-inverted masks.  End of Editorial Comment
 **
 ** We'll process the image in blocks that are one dword wide.  That means
 ** we'll make a vertical pass through the rows of the shifted input buffer
 ** for each output dword.  After processing one output dword, there's
 ** a row of shifted inputs we'll never use again.  We'll also need a
 ** new row.  We could move each shifted input word up one row in the buffer
 ** and insert one new row of shifted dwords at the bottom.  That
 ** generates a lot of writes.  It would be nicer to leave the buffer
 ** dwords in place and use a moving pointer.
 **
 ** We can't use one pre-allocated global array for the shifted inputs
 ** needed in all morphology operations because the code wouldn't be reentrant.
 ** Therefore, we must use a register to point to an array allocated by the
 ** calling procedure.  We know the width of the array, so can use constant
 ** displacements to address shifted inputs within a row.  It would be very
 ** nice to use just register+displacment addressing, it adds no cycles.
 ** However, we need some way to detect when the buffer wraps around.
 **
 ** Or do we?  We'll process a limited number of output dwords vertically before
 ** starting the next block.  This is done to make reasonable use of the
 ** processor's cache.  The buffer of preshifted inputs is reused for each
 ** block.  Say the calling routine allocates a buffer that's
 ** so large that we'll never overflow it.  If we process n lines vertically
 ** and have an H by V structure element, we'll need a buffer that can hold
 ** H * (V+n) * 8 bytes.  Typical values of H and V are under 10.  Typical
 ** processing block height is 64.  That's a 10 * 74 * 8 = 6 KB block.
 ** Not too bad, especially when compared to the 1MB required for the image.
 ** Say the SE were a maximum 64 x 64.  That requires a 64 * 128 * 8 = 64 KB
 ** block.  Seems ok on an MMX machine.  This is a ridiculously large strel.
 **
 ** In fact, we can address the entire array of shifted inputs with
 ** displacements relative to a single pointer.  The pointer points to
 ** the shifted dword in the upper left corner of the buffer.
 ** Well, if the generated code processes the output words in a loop,
 ** it can adjust the register to point to first preshifted input needed.
 ** Zowee.  We can do AND or OR ops in 1 cycle.  Well, actually,
 ** if we process two output lines at a time, we can do two ANDs or two ORs in
 ** a cycle.  Since there's no AND-with-inverse-of-source op, we'll store
 ** the inverse of the shifted words in the buffer as needed.  Very nice.
 **
 ** There's a somewhat more subtle benefit from eliminating extra buffer address
 ** arithmetic.  The 486 and Pentium have pipeline stalls when addresses
 ** are used immediately after being modified.  By addressing the entire
 ** input array relative to a single unchanging pointer, we eliminate
 ** these stalls.
 **
 ** In the plain 386 code, we combined the code that did the shifting with
 ** the code that did the ANDing.  We would start creating a new output
 ** word by ANDing a bunch of preshifted words together.  These words
 ** represented all the SE elements except for those on the last SE row.
 ** We'd then create one new preshifted input row, put it into the buffer
 ** and combine any of these new elements with the current output word
 ** as necessary.  The result would be a new output word and one new row
 ** of preshifted input words.
 **
 ** In MMX code, instruction pairing is important.  It's really a good idea
 ** to do two things at once so you can keep both execution units occupied
 ** on every clock cycle.  It's easier to design code that does this
 ** if you're basically doing the same thing on two data streams.  That's
 ** what we'll do here, we'll compute two rows of preshifted inputs in
 ** parallel.
 **
 ** We're really doing double shifts.  Each MMX double shift requires two input
 ** registers plus a scratch register.  If you figure out a fast scheme that
 ** uses fewer registers, let me know.  Anyway, doing two of these at
 ** a time uses up 2 * 3 = 6 MMX registers.  There are only 8 MMX registers.
 **
 ** Well, we could still integrate the shifting with the output word production.
 ** We'd produce two output dwords at a time in the remaining two MMX registers.
 ** We don't want to do that.  The reason is subtle, as we'll try to explain
 ** next.
 **
 ** We want to do two AND or OR operations per clock cycle.  Since you can't do
 ** two MMX operations with memory in the same cycle (!), at least one of
 ** the source operands has to be in a register.  That means we have to
 ** do an MMX read to fetch that operand.  If there are only two output
 ** words being produced at a time, the word fetched can be used at most
 ** twice, often only once.  If we produce more output dwords in parallel,
 ** we can almost always amortize the cost of that extra fetch over more
 ** AND ops.  Since there are fewer reads per useful op, the code runs faster.
 **
 ** Therefore, in the MMX implementation, we'll separate the code that
 ** produces shifted inputs from the code that does the ANDing and ORing.
 ** This lets us use lots of registers for each task.  For the sake of
 ** simplicity, all of the shifted inputs needed by the block are
 ** produced first, then all of the outputs are produced.  Note
 ** that we still have two spare registers during the shifting stage.
 ** We'll fill one of them with all 1's, using that to invert the shifted
 ** inputs as needed.  As stated above, both positive and inverted inputs
 ** are stored in the shifted input buffer.  We can produce the inverted
 ** input in one operation and write it with one more.  If we wait until
 ** we need the inverted inputs to invert them, we need to reserve two MMX
 ** registers for inputs (if we're going to do two ANDs in parallel
 ** eventually) instead of one and need to read both of them instead of
 ** just 1 and need to XOR both with a word of all 1's.  All told, a bummer.
 **
 **				*	*	*
 **
 ** We need to say something about wordswapping.  When a 64-bit dword is
 ** read from an image, the 32 bits on the left ends up in the *LS* end
 ** of the word since it has the lower address.  If we're going to shift
 ** the inputs, we'd really like to swap the two 32-bit halves of the
 ** input first.  Otherwise, a double shift requires us to insert bits
 ** into the ends of the two 32 bit halves independently (requiring at least two
 ** operations) instead of into the end of one 64-bit thing.  So, we
 ** wordswap the input words before shifting.
 **
 ** Sometime before the output word is written out, its bits must be wordswapped
 ** back.  We could do this by re-wordswapping all the preshifted input dwords.
 ** That's a lot of shifting.  We'll just wordswap each the output word before
 ** writing it.
 **
 ** What do we do in the degenerate cases of purely horizontal and purely
 ** vertical strels?  For purely vertical strels, no shifting is needed.  Since
 ** no shifting is done, we can skip the wordswapping, so we do.  Since
 ** there's no shifting, do we need the buffer of preshifted inputs?
 ** Well, yeah, we do.  We buffer them for two reasons.  First, it provides
 ** a place to store both inverted and non-inverted versions of the input
 ** words if needed.  Second, the preshifted input buffer is more easily
 ** addressed than the original image during the AND/OR phase.  Its geometry
 ** is known by the code that generates the strel, so we can use displacement
 ** addressing instead of having to repeatedly add the input images's 
 ** bytes/line to an address register.  Also, using the buffer in this
 ** case just makes all the cases more uniform.
 **
 ** Purely horizontal strels really are a special case.  Since there's no
 ** vertical component, each shifted input dword is used exactly once.  It
 ** would be nicely uniform to store each shifted dword into the buffer,
 ** then read it out before using it that one time, but it's too ugly.
 ** Remember that we calculate two rows of shifted inputs in parallel.
 ** We have two MMX registers left over.  In the purely horizontal case, we
 ** use the two MMX registers to hold the outputs.  As each shifted input
 ** is created, it's either ANDed or inverted and ANDed with the corresponding
 ** output word.  Since we can't use one of the registers to hold the all
 ** 1's mask used to do inverting, we kicked the mask out to memory.  There's
 ** a word of 1's kept in the data segment.  Since this piece of template
 ** code gets loaded, the loader will fix up the pointer of the template
 ** opcode that points to that mask word.  Therefore, we can just copy that
 ** opcode, address and all, to each piece of new machine code.
 **
 ** Since the purely horizontal case doesn't use the preshifted input buffer,
 ** it could process the image in vertical blocks or could just process
 ** entire scan lines at a time.  We choose to process the image in vertical
 ** blocks purely for the sake of uniformity.  If nothing else, it makes
 ** writing the upper level code easier.
 **
 **				*	*	*
 **
 ** We've seen that processing a block is done in two major steps - the
 ** preshifted input buffer is filled, then the outputs are created.  Let's
 ** say a little more about each step.
 **
 ** To efficiently use two ALUs in each clock cycle, there are really two
 ** independent trains of execution taking place in parallel.  Since
 ** they're basically the same, we'll just describe one of them.
 **
 ** When calculating the preshifted inputs, we dedicate one MMX register to
 ** holding the input dword that's centered on the column of outputs we're
 ** going to calculate.  A second MMX register holds 32 bits (read using
 ** the movd opcode so we don't have to worry about the other 32 bits) to the
 ** right or left of the column.  It holds the bits to the left when we're
 ** shifting right and vice versa.
 **
 ** This is a good place to distinguish between what the C code that generates
 ** machine code does and what the machine code does.  The C code parses the
 ** SE and puts out machine code.  The reason we go to all of this trouble
 ** is to eliminate virtually all of the conditional branches that would
 ** otherwise soak up time in a general morphology routine.  The C code
 ** generates the code that would be executed once all the branches have been
 ** decided.  Therefore, the C code does all the tests.  The steps listed
 ** below show what the C code does, and by inference, what the machine
 ** code does.
 **  1.	Generate code to load and wordswap the central register.
 **  2.	If there are no active columns to the left of the center's column,
 **	generate code to byteswap in the central dword and jump to step 12.
 **	This is not a jumps that are executed at run time, they're
 **	jumps that are executed by the C code that generates machine
 **	code.
 **  3.	Generate code to load the 32 bits to the left of the central dword.
 **  4. If the central dword is not already wordswapped, generate code to
 **	wordswap it.  It will already be wordswapped if we are executing this
 **	step as a result of step 14.
 **  5.	Generate code to copy the central dword to the scratch register.
 **  6.	Generate code to shift the central dword as far right as
 **	necessary to account for the distance from the leftmost active
 **	column of the strel to the center of the strel.  This is
 **   	Cx bit positions, where the center of the strel is a position
 **   	(Cx,Cy) inside the strel (starting from (0,0) in the upper left corner).
 **  7.	Generate code to shift the outboard register left so that Cx
 **	bits from the word on the left are visible in the MS end of 
 **	the register.  Initially, this is a shift of 32 + (32 - Cx) bits
 **	since we must move the bits read in step 3 to the upper word of
 **	the dword register in addition to moving them into position in that
 **	upper word.
 **  8.	Generate code to combine the outboard register with the scratch
 **	register, forming the shifted input.
 **  9.	If the leftmost active SE column has any HITs, generate code to
 **	write this value to the buffer of shifted inputs.
 ** 10.	If the leftmost active SE column has any MISSes, generate code to
 **	invert this value and write it to the next dword location in
 **   	the buffer of shifted inputs.  The dword is inverted by
 **	pxor-ing it with one of the two remaining MMX registers which
 **	is filled with all 1's.  The 1's register was filled in the
 **	prologue, which is not shown here.
 ** 11.	If there are any more active columns (contain HITs and/or MISSes) of
 **   	the SE to the left of the center column, repeat steps 5
 **	through 10 for each active column, working from the leftmost column
 **	of the strel towards the center.  Modify the shift distances so
 **	that the central bits in the scratch register are always shifted right
 **	by the distance from the column in question to the central
 **	strel column.  The outboard register is shifted left by the
 **   	distance betweeen the last column and the current column.
 **	It has progresively fewer bits in its top end as we compute
 **   	dwords with smaller and smaller shifts.
 ** 12.	If the central column of the strel has HITs, generate code to 
 **	write the central dword to the shifted input buffer.
 ** 13.   If the central column has MISSes, generate code to copy the
 **	central dword to the scratch register, invert it, and write it.
 ** 14.	If any of the columns of the strel to the right of the central
 **	column have HITs and/or MISSes, repeat steps 3 through 11
 **	where the outboard register holds bits from the 32 bits to the
 **	right of the central dword and the direction of all shifts is reversed
 **	(left <-> right).  This time, work from the rightmost column of the
 **	strel back towards the center.
 **
 ** To increase efficiency, we actually do two rows at once, interleaving their
 ** opcodes so that the shifts and memory references don't collide.  This
 ** code is hand written out below.  All the C code does is assembly little
 ** modular units of the template code.
 **
 **
 ** On to the section of code that does ANDs (erodes) and ORs (dilation).  Ok,
 ** dilation is really more complicated than just replacing ANDs with ORs.
 ** In dilation we need to project the center pixel out to all the HIT
 ** locations in the strel instead of carrying pixels at all the HIT/MISS
 ** locations back to the center.  This introduces an interesting set
 ** of reversals, but tough.
 **
 ** To maximize efficiency, we want to concurrently calculate as many
 ** partial results as possible.  We also want to do two ANDs per cycle
 ** when possible.  Since one can't do two MMX memory operations in the
 ** same cycle, one of the AND source operands must be in an MMX register.
 ** Therefore, we must reserve an MMX register for a source operand.
 ** Since there are 8 MMX registers and one must be used for a source
 ** operand, we can only hold 7 partial results at once.  Therefore,
 ** we break the block into subblocks of height 7.  To make this
 ** efficient, we'll tell the upper level code that the height of
 ** the standard block is some multiple of 7.
 **
 ** I fretted that things would be most efficient if *both* of the source
 ** operands were in registers.  It's not true.  In MMX land, operating with
 ** a memory operand costs nothing in time, though it does entail awkward
 ** pairing restrictions.  However, loading a memory operand into a
 ** register takes an extra cycle.  So, in every clock cycle, we'll do
 ** two ANDs, one from memory and one from a register that was preloaded.
 ** As noted above, the real reason we hold as many output results as
 ** possible is to minimize the number of times we need to do the preloading.
 **
 ** Yeah, but what about strels that have 7 or fewer rows?  Couldn't we
 ** process them more efficiently if we constructed the loop so that it
 ** retired two fully formed outputs and started two new outputs on each pass?
 ** That's how the SPARC code works, though it does one at a time since it
 ** doesn't have these pesky pairing restrictions.  Well, maybe.
 **
 ** The total number of ANDs done in both cases is the same.  The number
 ** of times a preshifted source word is preloaded goes up in our scheme
 ** since source dwords that lie in rows above and below the current
 ** output block are read twice.  The amount of redundant reading depends on
 ** the height and width of the strel.  The strel having the most redundant
 ** reads has a height of 7.  Remember that we're only concerned about
 ** strels whose height is less than or equal to 7 now.  Say we have
 ** a 7x7 strel that's fully populated with active elements.  This
 ** means we need to do a total of (3 + 7 + 3)x7 reads per subblock
 ** of which (3+3)x7 will be repeated in the subblocks above and below.
 ** Well, we really only do an extra (6x7)/2 reads since only every row
 ** is preloaded.
 **
 ** So say we're operating on a 2560 x 3300 image.  That image can be
 ** divided into 40 columns by 471.43 rows of 64x7 subblocks.  We need
 ** an extra (6x7)/2 * 40 * 471.43 = 396,002 operations to process
 ** the entire image.  What percentage is that of the total operations?
 ** Well, we need to produce 7 shifted dwords for every source dword and
 ** it takes 5 cycles to produce 2 dwords.  That's ((7 * 5)/2) * 40 * 3300
 ** or 2,310,000 cycles.  Then we need roughly 7*7 operations on each dword,
 ** but can do 2 per cycle, so add another (7*7)/2 * 40 * 3300 = 3,234,000
 ** cycles.  Thats a total of 5,544,000 cycles.  If we eliminate those
 ** extra loads, we increase performance by
 ** (5,544,000 + 396,002) / 5,544,000 = 1.07x.  Whoopee.  This is about the
 ** largest performance improvement we could expect to see.  Strels with
 ** fewer rows will have less overlap outside the 7 row block.  Strels with
 ** more rows will be processed 7 rows at a time anyway.
 **
 ** Yeah, but what if we wanted to do this anyway?  There are at least
 ** two ugly problems you'll have to overcome.  First, you'll have to
 ** figure out how to keep track of the moving window of outputs.  Second,
 ** you'll have to efficiently wordswap the two output registers and
 ** write them out before you need them again.
 **
 ** What's the deal with the revolving window of outputs?  You have
 ** an ugly choice of either moving the valid registers up as the
 ** output registers become vacant (an extra 5 ops per pair of input rows - or
 ** 5 * (3300/2) * 40 columns = 330,000 ops) or changing the "names" of the
 ** registers each time an output pair is written.  In the second option, the
 ** contents of the registers don't move, but we simply declare that if mm0
 ** held the top row of outputs before the 2 writes, mm2 holds them after the
 ** writes.  Not only is this confusing when trying to write and debug
 ** the code, it means that any loop must be unrolled so that it only
 ** repeats when the register names come around again, yet we must check
 ** to see if we're done after *every* register pair (2 ops * (3300/2) *
 ** 40 cols = 132,000 ops). 
 **
 ** Ok, but how hard can it be to wordswap and write registers?  You need
 ** two unpack ops to do the wordswap, a dest address modification
 ** (not paired with an MMX mem ref, since you can't do that) and an
 ** MMX write to vacate a register.  We need to vacate two registers,
 ** so we need to do all this twice.  This is particularly challenging when
 ** strels have a small number of elements since you may need the
 ** registers before they're written, which means you need some way of
 ** detecting this case and completing the write even if its inefficient.
 ** To make the writing process efficient in the normal case,
 ** you need a fancy code generator that can combine the write
 ** preparation and write ops with subsequent processing ops for the
 ** rest of the strel, whatever they may be.  This is all breaking my head.
 **
 ** So we'll do it the simpler way.  We process the columns in sub-blocks of
 ** height 7, then write all 7 out in one fell swoop.  This can be coded
 ** by hand and just copied by the template code.  It can be a tiny bit
 ** slower than the complicated way, but too bad.
 **
 ** Like the SPARC algorithms, we will have restrictions on the size of
 ** structure elements we can handle.  The distance from the SE's center to
 ** the edge of the SE (dimensions cx, cy, (nx-cx) and (ny-cy)) must all be
 ** less than 32.  This is the width and height of the pad all around
 ** the image.  You might think that allowing the distance to be equal to
 ** 32 would also be ok, but there's one case where it isn't.  As
 ** explained below, we fill the input cache buffer 2 lines at a time.
 ** We might start a 2-row strip on the last line of the image.
 ** Since the distance from the strel center to the strel bottom is at most
 ** 31, we will read down at most 31 lines from the line below the bottom of
 ** the image.  Since there are 32 lines of pad, we're still ok.
 **
 **				*	*	*
 **
 ** This is a good place to map out the generic sequence for filling the
 ** input cache.  The stuff shown is executed once for each row of the
 ** input cache.  We process two lines at a time, that's what "upper" and
 ** "lower" mean.  The upper address is addressed using rSrcAddr.  The
 ** lower address is addressed using rSrcAddr + rSrcPitch.
 **
 ** First, examine the columns of the strel in the following order:
 ** center column, leftmost col, next-to-leftmost col, ... col left of
 ** center, rightmost col, next-to-rightmost col, col to right of center.
 ** This peculiar order allows us to keep the pipeline flowing reasonably
 ** well.  How?  Well, processing from the outside in towards the middle
 ** allows us to keep throwing away sideword bits.  This means we can use
 ** a single register for the sideword instead of keeping the sideword in
 ** a register and copying it to a scratch reg.  This, in turn, means we don't
 ** have to reload the central dword (also) when moving from the left cols
 ** to the right cols since we never change it.
 **
 ** How else does our method keep things flowing?  There's obviously no
 ** preparation for the initial load.  This is generated when we find the
 ** first active strel column when investigating them in the order
 ** specified above.  The only preparation we need when writing the center
 ** column's data is done by the initial load, so we might as well get that
 ** out of the way immediately.  The left and right shifting code can always
 ** interleave some writes with preparation for the next writes.  This
 ** preparation is normally shifting.   However, when coming to the end
 ** of the row, there's no more shifting to do, but there is some general
 ** updating of pointers and counts.  Sadly, one cannot pair the opcodes
 ** that do this with MMX writes, so there are bunch of empty cycles on the
 ** last column.
 **
 ** Here is the order in which we decide which type of code to generate.
 ** The Preload code may come as a prelude to any other block.  Below that,
 ** only one type of block will be generated at a time:
 **
 ** Preload (loads registers and starts the shifts)
 ** See if col is center col, if so generate code for it.
 **    HITs only
 **    MISSes only
 **    Both HITs and MISSes
 ** See if col is last column to be processed (will be to left or right of
 ** center).  If so, generate code for it.
 **    HITs only
 **    MISSes only
 **    Both HITs and MISSes
 ** See if EITHER the col is to the left of center and is followed by
 ** another column that's also to the left of center OR the col is to the
 ** right of center and is followed by another col that's also to the right
 ** of center.  If so, generate code for it.  We think this is the common
 ** case.
 **    HITs only
 **    MISSes only
 **    Both HITs and MISSes
 ** If none of the cases above is true, this must be the last left col to
 ** be processed.  It must be followed by a right col.  Generate code for
 ** this case.
 **    HITs only
 **    MISSes only
 **    Both HITs and MISSes
 **
 **
 ** There's only one type of preload code.  It is:
 **START:	load upper dword
 **	load 1's mask
 **
 **	load left/right upper side word
 **	step 1 of upper wordswap
 **
 **	load lower dword
 **	step 2 of upper wordswap
 **
 **	load left/right lower side word
 **	step 1 of lower wordswap
 **
 **	shift upper sideword left/right
 **	step 2 of lower wordswap
 **
 **	copy upper dword to upper scratch
 **	shift lower sideword left/right
 **
 **	shift upper dword in scratch reg right/left
 **	copy lower dword to lower scratch
 **
 **
 ** Here are the three types of code used to process the center block.
 ** This always happens immediately after a preload if it happens at all.
 **
 ** Here's what happens if we want to write only the positive dword for
 ** the central columns:
 **LOOP:
 **	write unshifted upper dword to input cache
 **	; empty
 **
 **	write unshifted lower dword to input cache + bytes/row
 **	; empty
 **
 ** Go back to LOOP above.  This doesn't mean that the MMX code loops.
 ** It means that the C code that generates MMX code loops around to generate
 ** the code to process another column.  "LOOP" signifies the place where the
 ** code for each column starts/stops.
 **
 ** If this is a purely horizontal strel, these become:
 **LOOP:
 **	copy unshifted upper dword to mm6 temp
 **	copy unshifted lower dword to mm7 temp
 **
** Here's what happens if we want to write only the negative dword for
 ** the central columns:
 **LOOP:
 **	invert upper unshifted dword
 **	invert lower unshifted dword
 **
 **	write inverted, unshifted upper dword to input cache
 **	re-invert upper unshifted dword
 **
 **	write inverted, unshifted lower dword to input cache + bytes/row
 **	re-invert lower unshifted dword
 **
 **LOOP: ...
 **
 ** If this is a purely horizontal strel, this becomes:
 **LOOP:
 **	Invert upper unshifted dword, leaving result in mm6.
 **	load mm7 with all 1's.
 **
 **	invert lower unshifted dword, leaving result in mm7.
 **	; empty
 **
 ** Here's what happens if we want to write both the positive and the
 ** negative dword for the central columns:
 **LOOP:
 **	write positive unshifted upper dword to input cache
 **	invert upper unshifted dword
 **
 **	write positive unshifted lower dword to input cache + bytes/row
 **	invert lower unshifted dword
 **
 **	write inverted, unshifted upper dword to input cache + 8
 **	re-invert upper unshifted dword
 **
 **	write inverted, unshifted lower dword to input cache + 8 + bytes/row
 **	re-invert lower unshifted dword
 **
 **LOOP: ...
 **
 ** Note that we can't have a column with both hits and misses in a
 ** purely horizontal strel (there's only one thing in a column in this case)
 ** so we don't have to make a hit-and-miss horizontal strel case.
 **
 ** Here's how we process the last column to the left or right of center.
 ** The basic difference between this case and the normal case is that we
 ** don't get ready for a new write, but we do do the end-of-loop code.
 ** Here's the code for HITs only (shifted words are coming in):
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	increment inputCache address by 2 * number of input cache bytes/row 
 **	combine (OR) lower dword and sideword
 **
 **	add 2 * source bytes/line to source address	; wait for AGI
 **	; empty
 **
 **	write last shifted upper dword to inputCache - 2 * bytes/row. 
 **	; empty
 **
 **	write last shifted lower dword to inputCache - bytes/row
 **	; empty
 **
 **	decrement number of rows left by 2 in inputCache
 **	branch back to START if number of rows left > 0.
 **
 **
 ** If this is a purely horizontal strel, the code becomes:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	AND upper shifted dword with mm6 temp, result to mm2
 **	combine (OR) lower dword and sideword
 **
 **	wordswap step 1 for mm2, mm6 as scratch
 **	AND lower shifted dword with mm7 temp, result to mm5
 **
 **	add 2 * source bytes/line to source address
 **	wordswap step 2 for mm2, mm6 as scratch
 **
 **	wordswap step 1 for mm5, mm7 as scratch
 **	decrement number of rows left by 2
 **
 **	write mm2 to dstAddress		; note 1 cycle delay
 **	wordswap step 2 for mm5, mm7 as scratch
 **
 **	; empty
 **	; empty
 **
 **	write mm5 to dstAddress + dest bytes/row
 **	; empty
 **
 **	add 2 * dest bytes/line to dest address using lea (no cc's affected)
 **	branch back to START if number of rows left > 0.
 **
 **
 ** Here's the last col code for MISSes only:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	invert (pxor) upper dword in scratch reg
 **	combine (OR) lower dword and sideword
 **
 **	increment inputCache address by 2 * number of input cache bytes/row
 **	invert (pxor) lower dword in scratch reg
 **
 **	add 2 * source bytes/line to source address	; wait for AGI
 **	; empty
 **
 **	write last inverted, shifted upper dword to inputCache - 2 * bytes/row. 
 **	; empty
 **
 **	write last inverted, shifted lower dword to inputCache - bytes/row
 **	; empty
 **
 **	decrement number of rows left by 2 in inputCache
 **	branch back to START if number of rows left > 0.
 **
 ** If this is a purely horizontal strel, the code becomes:
 ** Here's the last col code for MISSes only:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	invert (pandn) upper dword in scratch reg, result to scratch reg (mm2)
 **	combine (OR) lower dword and sideword
 **
 **	wordswap step 1 for mm2, mm6 as scratch
 **	invert (pandn) lower dword in scratch reg, result to scratch reg (mm5)
 **
 **	add 2 * source bytes/line to source address
 **	wordswap step 2 for mm2, mm6 as scratch
 **
 **	wordswap step 1 for mm5, mm7 as scratch
 **	decrement number of rows left by 2
 **
 **	write mm2 to dstAddress		; note 1 cycle delay
 **	wordswap step 2 for mm5, mm7 as scratch
 **
 **	; empty
 **	; empty
 **
 **	write mm5 to dstAddress + dest bytes/row
 **	; empty
 **
 **	decrement number of rows left by 2 in inputCache
 **	branch back to START if number of rows left > 0.
 **
 **
 ** Here's the last col code for both HITs and MISSes:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	increment inputCache address by 2 * number of input cache bytes/row
 **	combine (OR) lower dword and sideword
 **
 **	add 2 * source bytes/line to source address	; wait for AGI
 **	; empty
 **
 **	write last shifted upper dword to inputCache - 2 * bytes/row. 
 **	invert (pxor) upper shifted dword
 **
 **	write last shifted lower dword to inputCache - bytes/row
 **	invert (pxor) lower shifted dword
 **
 **	write last inverted, shifted upper dword to inputCache+8-2*(bytes/row). 
 **	; empty
 **
 **	write last inverted, shifted lower dword to inputCache+8-(bytes/row)
 **	; empty
 **
 **	decrement number of rows left by 2 in inputCache
 **	branch back to START if number of rows left > 0.
 **
 **
 **
 ** Here's the code for processing a left or right col that's followed by
 ** a col on the same side of center. In other words, this isn't the last
 ** col in the row and it isn't a left col followed by a right col.
 ** As usual, we start with the code for HITs only:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	shift upper sideword left/right
 **	combine (OR) lower dword and sideword
 **
 **	write upper dword to input cache	; note 1 cycle delay from OR
 **	shift lower sideword left/right
 **
 **	write lower dword to input cache + bytes/row
 **	copy upper dword to upper scratch
 **
 **	shift upper dword in scratch reg right/left
 **	copy lower dword to lower scratch
 **
 ** If this is a purely horizontal strel, the code becomes:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	shift upper sideword left/right
 **	combine (OR) lower dword and sideword
 **
 **	pand upper dword to mm6
 **	shift lower sideword left/right
 **
 **	pand lower dword to mm7
 **	copy upper dword to upper scratch
 **
 **	shift upper dword in scratch reg right/left
 **	copy lower dword to lower scratch
 **
 **
 ** Here's the code for MISSes only:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	invert (pxor) upper dword in scratch reg
 **	combine (OR) lower dword and sideword
 **
 **	shift upper sideword left/right
 **	invert (pxor) lower dword in scratch reg
 **
 **	write upper dword to input cache	; note 1 cycle delay from XOR
 **	shift lower sideword left/right
 **
 **	write lower dword to input cache + bytes/row
 **	copy upper dword to upper scratch
 **
 **	shift upper dword in scratch reg right/left
 **	copy lower dword to lower scratch
 **	
 ** Go back to LOOP above.
 **
 ** If this is a purely horizontal strel, the code becomes:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	invert (pandn) upper dword in scratch reg, result in mm2
 **	combine (OR) lower dword and sideword
 **
 **	shift upper sideword left/right
 **	invert (pandn) lower dword in scratch reg, result in mm5
 **
 **	copy upper dword in mm2 back to mm6
 **	shift lower sideword left/right
 **
 **	copy lower dword in mm5 back to mm7
 **	copy upper dword to upper scratch
 **
 **	shift upper dword in scratch reg right/left
 **	copy lower dword to lower scratch
 **
 **
 ** If we need both positive and inverted versions of a dword, we get:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	shift upper sideword left/right
 **	combine (OR) lower dword and sideword
 **
 **	write upper dword to input cache	; note 1 cycle delay from OR
 **	invert upper dword in scratch reg
 **
 **	write lower dword to input cache + bytes/row
 **	invert lower dword in scratch reg
 **
 **	write inverted upper dword to input cache + 8
 **	shift lower sideword left/right
 **
 **	write inverted lower dword to input cache + 8 + bytes/row
 **	copy upper dword to upper scratch
 **
 **	shift upper dword in scratch reg right/left
 **	copy lower dword to lower scratch
 **
 ** Go back to LOOP above.
 **
 **
 **
 ** Finally, we need to show the code used for a left col that's followed
 ** by a right col.  If there was a center col, it's already been processed,
 ** so we don't need to worry about it here.  First, the code for HITs only:.
 **
 ** Here's how things go when we write no dwords for the central column
 ** (it's not active).
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	read right upper sideword (32 bits).
 **	combine (OR) lower dword and sideword
 **
 **	write last shifted upper dword to input cache
 **	copy upper dword to upper scratch
 **
 **	write last shifted lower dword to input cache + bytes/row
 **	shift upper dword in scratch reg left
 **
 **	load right lower sideword (32 bits).
 **	shift upper sideword right
 **
 **	copy lower dword to lower scratch
 **	shift lower sideword right
 **
 **LOOP: ...
 **
 ** If this is a purely horizontal strel, the code becomes:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	read right upper sideword (32 bits).
 **	combine (OR) lower dword and sideword
 **
 **	pand upper dword, result to mm6
 **	copy upper dword to upper scratch
 **
 **	pand lower dword, result to mm7
 **	shift upper dword in scratch reg left
 **
 **	load right lower sideword (32 bits).
 **	shift upper sideword right
 **
 **	copy lower dword to lower scratch
 **	shift lower sideword right
 **
 **
 ** Here's the code that processes MISSes only:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	invert (pxor) upper dword in scratch reg
 **	combine (OR) lower dword and sideword
 **
 **	read right upper sideword (32 bits).
 **	invert (pxor) lower dword in scratch reg
 **
 **	write last shifted upper dword to input cache
 **	copy upper dword to upper scratch
 **
 **	write last shifted lower dword to input cache + bytes/row
 **	shift upper dword in scratch reg left
 **
 **	load right lower sideword (32 bits).
 **	shift upper sideword right
 **
 **	copy lower dword to lower scratch
 **	shift lower sideword right
 **
 **LOOP: ...
 **
 ** If this is a purely horizontal strel, the code becomes:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	invert (pandn) upper dword, result in mm2
 **	combine (OR) lower dword and sideword
 **
 **	read right upper sideword (32 bits).
 **	invert (pandn) lower dword, result in mm5
 **
 **	copy upper dword in mm2 back to mm6
 **	copy upper dword to upper scratch
 **
 **	copy lower dword in mm5 back to mm7
 **	shift upper dword in scratch reg left
 **
 **	load right lower sideword (32 bits).
 **	shift upper sideword right
 **
 **	copy lower dword to lower scratch
 **	shift lower sideword right
 **
 **
 ** Here's the version that processes both HITs and MISSes:
 **LOOP:
 **	combine (OR) upper dword and sideword
 **	shift lower dword in scratch reg right/left
 **
 **	read right upper sideword (32 bits).
 **	combine (OR) lower dword and sideword
 **
 **	write last shifted upper dword to input cache
 **	invert (pxor) upper dword in scratch reg
 **
 **	write last shifted lower dword to input cache + bytes/row
 **	invert (pxor) lower dword in scratch reg
 **
 **	write last inverted, shifted upper dword to input cache + 8
 **	copy upper dword to upper scratch
 **
 **	write last inverted, shifted lower dword to input cache + 8 + bytes/row
 **	shift upper dword in scratch reg left
 **
 **	load right lower sideword (32 bits).
 **	shift upper sideword right
 **
 **	copy lower dword to lower scratch
 **	shift lower sideword right
 **
 **LOOP: ...
 **
 **
 ** We should also say how the input cache is filled for purely vertical
 ** strels.  For purely vertical strels containg only hits.  There is no
 ** LOOP section, only a START:
 **
 **START: read upper dword
 **	; empty
 **
 **	read lower dword
 **	; empty
 **
 **	add 2 * source bytes/line to source address
 **	increment input cache address by 2 * number of bytes/row of input cache
 **
 **	write upper dword to input cache - 2 * bytes/row
 **	; empty
 **
 **	write lower dword to input cache - bytes/row
 **	; empty
 **
 **	decrement number of rows left by 2 in input cache
 **	branch back to START if number of rows left > 0.
 **
 ** Don't all those empties make you sick?  The dumb thing just can't pair
 ** integer operations with MMX memory operations.
 **
 ** If we have a vertical strel containing only misses:
 **
 **START: read upper dword
 **	(re)load mm6 with all 1's
 **
 **	read lower dword
 **	invert upper dword using mm6
 **
 **	invert lower dword using mm6
 **	; empty
 **
 **	add 2 * source bytes/line to source address
 **	increment input cache address by 2 * number of bytes/row of input cache
 **
 **	write inverted upper dword to input cache - 2 * bytes/row
 **	; empty
 **
 **	write inverted lower dword to input cache - bytes/row
 **	; empty
 **
 **	decrement number of rows left by 2 in input cache
 **	branch back to START if number of rows left > 0.
 **
 ** If we have a vertical strel containing both hits and misses:
 **
 **START: read upper dword
 **	(re)load mm6 with all 1's
 **
 **	read lower dword
 **	; empty	
 **
 **	write upper dword to input cache
 **	invert upper dword using mm6
 **
 **	write lower dword to input cache + bytes/row
 **	invert lower dword using mm6
 **
 **	add 2 * source bytes/line to source address
 **	increment input cache address by 2 * number of bytes/row of input cache
 **
 **	write inverted upper dword to input cache - 2 * bytes/row
 **	; empty
 **
 **	write inverted lower dword to input cache - bytes/row
 **	; empty
 **
 **	decrement number of rows left by 2 in input cache
 **	branch back to START if number of rows left > 0.
 **
 **
 ** That's it for the shift loop.  Well, almost.  What happens if the block
 ** height is odd?  We fill two rows at a time.  Doesn't that mean that we
 ** might try to fill one row too far and extend off the bottom of the image?
 ** It would, if we allowed the distance from the strel's center to the bottom
 ** of the strel to be 32.  We only accelerate strels where the distance from
 ** the center to the top, bottom or sides is less than 32.  Therefore, even
 ** though we may be trying to calculate the data for the row below the last
 ** row in the block, we'll read at most 31 rows below that, so will still be
 ** within the 32 row frame.
 **
 **
 **
 ** We're ready for the ANDs and ORs with the cached, shifted inputs.
 ** We process the block in subblocks of height 7.  The C code will create
 ** a loop to do the processing of each subblock.
 **
 ** The basic loop structure has us working across the input cache,
 ** reading data for two rows at a time.  The two shifted, wordswapped,
 ** dwords are applied to two output dwords.  This is where things get
 ** confusing (as if they weren't already).
 **
 ** Each cached dword is associated with one column of the strel.  Each
 ** cached dword also has a polarity - it corresponds to a HIT or a MISS.
 ** Let's number the columns of the strel from 0 to nx-1 with the center
 ** of the strel at column cx.  The C code will maintain a table that
 ** produces a strel column and polarity given a horizontal position in the
 ** input cache.
 **
 ** We also need to worry about input cache rows.  Say we want to produce
 ** a block of height blockHeight which starts on row blockStart.  Say
 ** the strel has ny rows with a center on row cy.  Then the first input
 ** cache row will correspond to row blockStart - cy of the input (and output).
 ** The last cache row will correspond to row blockStart + (blockHeight-1) +
 ** ((ny-1) - cy).
 **
 ** A block is produced in units of subblocks.  Each subblock has 7 rows
 ** and it 1 dword in width.  The first row of the subblock
 ** is on row subBlockStart.  We process the subblock by looping through
 ** the rows of the input cache starting from row subBlockStart-cy and
 ** finishing with row subBlockStart + 7 - 1 + (ny - 1 - cy).
 **
 ** For each entry in the input cache on row R, we look up its
 ** corresponding strel column, C, and polarity P.  We then inspect all
 ** output dwords in the subblock.  Each output dword is in a row, say Ro.
 ** Ro = subBlockStart for the top dword, Ro = subBlockStart+1 for the
 ** next output word and so on.  For each output word and cache entry,
 ** we see if there's a strel element matching the input cache dword's
 ** polarity P at strel row (R - Ro + cy) and strel column C.  If so,
 ** we generate an AND opcode.  If not, we don't.
 **
 ** For a particular pair of cache words (one from row R and one from row
 ** R+1, both from col C and having polarity P), we cycle through all the
 ** output dwords asking the question above.  This means we generally get
 ** to use a given input word more than once before getting the next input
 ** pair.  We move across the input cache two rows at a time, then move down
 ** two rows so each input cache word is reused as many times as possible
 ** before moving on.
 **
 ** It's almost that easy.  First, there are two types of AND opcodes and
 ** a load opcode.  The load just loads an input cache entry into the
 ** MMX scratch register.  It's used the first time an output dword is
 ** touched, since we can't AND with the trash that was there previously.
 ** One type of AND opcode combines a register directly with an input
 ** cache dword.  The other type combines with the scratch reg.
 ** To do two ANDs in a single cycle, we use both types at once.
 **
 ** Can we always use both types in a cycle?  Usually, but not always.
 ** Usually, we can compare words from input cache rows R and R+1 with
 ** output dwords from rows Ro and Ro+1.  However, at the very top and bottom
 ** of the range of input cache rows, Ro or Ro+1 might not be in the range
 ** of output dwords.  In those cases, we can only do one AND per cycle.
 **
 ** Let's try an example.  Say the strel looks like this:
 **
 **	H  _  M
 **	_  c  _		<- "c" is the center
 **	H  _  M
 **	H  _  H
 **
 ** For this strel, nx = 3, ny = 4, cx = 1, cy = 1.  The input cache will
 ** have 3 dwords per row.  The first dword in a row will correspond to
 ** the input shifted right by 1.  That's for the HITs in the leftmost strel
 ** column (col 0).  That dword will not be inverted.  There is no dword
 ** corresponding to the center column (column 1) since there are no
 ** HITs or MISSes there.  The second input cache column will correspond
 ** to the input shifted left since there is a HIT in column 2.  The
 ** third dword in an input cache row is the inverse of the input
 ** shifted left by 1 since there is also a MISS in column 2.
 **
 ** The number of rows in a block will be some multiple of 7.  If there
 ** are 49 (7 * 7) rows in a block, there will be 49 + (4-1) = 52 rows
 ** in the input cache.  Each row will have 3 dwords * 8 bytes/dword =
 ** 24 bytes.
 **
 ** Say we want to process a block starting at row 49 of the input image.
 ** The input cache will have data from row number (49 - 1) to row number
 ** (49 - 1) + 52 - 1 = 99.
 **
 ** When processing the first subblock in the block, we start on the 
 ** input cache row corresponding to row 49 - 1 = 48 of the image.  The
 ** output dwords of this subblock have row numbers (Ro) from 49 through
 ** 55.  We read the first pair of input cache dwords.  Their rows, R, are 48
 ** and 49.  They both have column, C, equal to 0.  Both have a polarity of HIT.
 **
 ** Start looking at pairs of output dword rows (48, 49), (49, 50), (50, 51),...
 ** (54, 55), (55, 56).  Hey, but there are no dwords with output rows 48 or
 ** 56!  Yeah, but we have to deal with these special cases somehow, and this
 ** seems as rational as any.  The first input cache pair is
 ** (R, R+1) = (48, 49), C = 0, P = HIT.  Well, golly, at strel row
 ** (48 - 48 + 1), there's nothing.  In particular, there's no HIT,
 ** so we do issue no opcodes.  Had there been a HIT there, we would
 ** have generated one AND since there's really no output dword for
 ** row 48, only for row 49.  Let's go on to the next pair of
 ** outputs.  At strel row (49 - 48 + 1) = 1, col 0, there's a HIT.  Hooray!
 ** We generate two AND opcodes.  Well, not quite.  First, since we notice that
 ** the upper input cache word hasn't been loaded yet, we really issue a load
 ** opcode first.  This puts the cached input word into mm7, the scratch
 ** register.  Second, since we notice that the output dwords in question
 ** haven't been touched yet, we convert the ANDs into simple moves to
 ** initialize them.  This moves the remaining cached dword and the dword
 ** that we just loaded into mm7 into the output's registers.
 ** You shouldn't AND with trash.
 **
 ** We continue in this vein through output pair rows with the following
 ** results:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (48, 49)	((48, 49), 0, HIT)	(1, 0) = _	nothing
 ** (49, 50)	((48, 49), 0, HIT)	(0, 0) = HIT	Two loads to init
 **							first two dwords
 ** (51, 52)	((48, 49), 0, HIT)	(-1, 0) = ?	nothing
 **
 ** and in fact, we can stop looking at output pairs at this point since
 ** (R - Ro + cy) corresponds to negative strel rows (above top of strel)
 ** for Ro > 49.
 **
 ** So we move onto the next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (48, 49)	((48, 49), 2, HIT)	(1, 2) = _	nothing
 ** (49, 50)	((48, 49), 2, HIT)	(0, 2) = MISS	nothing 
 **
 ** and onto the next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (48, 49)	((48, 49), 2, MISS)	(1, 2) = _	nothing
 ** (49, 50)	((48, 49), 2, MISS)	(0, 2) = MISS	two ANDS
 **
 ** So much for the first two input cache rows.  Let's do the next two for
 ** practice.  Note that we can go farther down into the outputs this
 ** time.
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (48, 49)	((50, 51), 0, HIT)	(3, 0) = HIT	one AND (no row 48)
 ** (49, 50)	((50, 51), 0, HIT)	(2, 0) = HIT	two ANDS
 ** (50, 51)	((50, 51), 0, HIT)	(1, 0) = _	nothing
 ** (51, 52)	((50, 51), 0, HIT)	(0, 0) = HIT	two loads
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (48, 49)	((50, 51), 2, HIT)	(3, 2) = HIT	two ANDs
 ** (49, 50)	((50, 51), 2, HIT)	(2, 2) = MISS	nothing 
 ** (50, 51)	((50, 51), 2, HIT)	(1, 2) = _	nothing
 ** (51, 52)	((50, 51), 2, HIT)	(0, 2) = MISS	nothing 
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (48, 49)	((50, 51), 2, MISS)	(3, 2) = HIT	nothing
 ** (49, 50)	((50, 51), 2, MISS)	(2, 2) = MISS	two ANDs 
 ** (50, 51)	((50, 51), 2, MISS)	(1, 2) = _	nothing
 ** (51, 52)	((50, 51), 2, MISS)	(0, 2) = MISS	two ANDs 
 **
 ** Let's do one more row for fun.  Note that the set of interesting
 ** output pairs shifts down by 2.
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (50, 51)	((52, 53), 0, HIT)	(3, 0) = HIT	two ANDs
 ** (51, 52)	((52, 53), 0, HIT)	(2, 0) = HIT	two ANDS
 ** (52, 53)	((52, 53), 0, HIT)	(1, 0) = _	nothing
 ** (53, 54)	((52, 53), 0, HIT)	(0, 0) = HIT	two loads
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (50, 51)	((52, 53), 2, HIT)	(3, 2) = HIT	two ANDs
 ** (51, 52)	((52, 53), 2, HIT)	(2, 2) = MISS	nothing 
 ** (52, 53)	((52, 53), 2, HIT)	(1, 2) = _	nothing
 ** (53, 54)	((52, 53), 2, HIT)	(0, 2) = MISS	nothing 
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (50, 51)	((52, 53), 2, MISS)	(3, 2) = HIT	nothing
 ** (51, 52)	((52, 53), 2, MISS)	(2, 2) = MISS	two ANDs 
 ** (52, 53)	((52, 53), 2, MISS)	(1, 2) = _	nothing
 ** (53, 54)	((52, 53), 2, MISS)	(0, 2) = MISS	two ANDs 
 **
 ** What the heck. let's finish it. 
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (52, 53)	((54, 55), 0, HIT)	(3, 0) = HIT	two ANDs
 ** (53, 54)	((54, 55), 0, HIT)	(2, 0) = HIT	two ANDS
 ** (54, 55)	((54, 55), 0, HIT)	(1, 0) = _	nothing
 ** (55, 56)	((54, 55), 0, HIT)	(0, 0) = HIT	one load (no row 56)
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (52, 53)	((54, 55), 2, HIT)	(3, 2) = HIT	two ANDs
 ** (53, 54)	((54, 55), 2, HIT)	(2, 2) = MISS	nothing 
 ** (54, 55)	((54, 55), 2, HIT)	(1, 2) = _	nothing
 ** (55, 56)	((54, 55), 2, HIT)	(0, 2) = MISS	nothing 
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (52, 53)	((54, 55), 2, MISS)	(3, 2) = HIT	nothing
 ** (53, 54)	((54, 55), 2, MISS)	(2, 2) = MISS	two ANDs 
 ** (54, 55)	((54, 55), 2, MISS)	(1, 2) = _	nothing
 ** (55, 56)	((54, 55), 2, MISS)	(0, 2) = MISS	one AND (no row 56)
 **
 ** and finally:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (54, 55)	((56, 57), 0, HIT)	(3, 0) = HIT	two ANDs
 ** (55, 56)	((56, 57), 0, HIT)	(2, 0) = HIT	one AND (no row 56)
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (54, 55)	((56, 57), 2, HIT)	(3, 2) = HIT	two ANDs
 ** (55, 56)	((56, 57), 2, HIT)	(2, 2) = MISS	nothing 
 **
 ** next input cache dword:
 ** output pair	input pair		strel (Rs,Cs)	Action
 ** (54, 55)	((56, 57), 2, MISS)	(3, 2) = HIT	nothing
 ** (55, 56)	((56, 57), 2, MISS)	(2, 2) = MISS	one AND (no row 56) 
 **
 **
 **
 ** We're getting closer all the time.  We've dealt with purely vertical
 ** strels, purely horizontal strels and all other strels.  The thing we
 ** haven't dealt with is subblocks that are less than 7 rows tall.  This
 ** happens at the bottom of the image.  This routine will be called
 ** with either a multiple of the subblock height, or with exactly
 ** one subblock that's less than the subblock height.  For example, if
 ** the subblock height is 7, this routine will be called with heights of
 ** 7, 14, 21, 28, etc, or with heights of 1, 2, 3, 4, 5 or 6.
 ** The C code will put in one test after the code that fills the input
 ** cache to see if the block height is less than 7.  If so, it jumps
 ** down to a routine down that's after all the normal subblock processing.
 **
 ** This routine looks like the normal subblock processing except that
 ** it only does one output row before looping back.  This is less efficient,
 ** but we only execute it in a tiny strip at the bottom of the image, so
 ** no big deal.
 **
 **
 *******************  END OF COMMENTS FROM mortemp3.asm  *********************
 *
 *
 * So much for selected readings from the morphology template file.
 *
 * So how does the C code generate this stuff?  First, it decides whether
 * to do anything at all.  If the SE is too big, it just returns without
 * generating any code.
 *
 * We then figure out how big the code will be.  The size is needed for
 * buffer allocation.  It's ok to ask for a little more space than required,
 * so we'll just do a worst-case computation.  Well, actually, two worst
 * case computations:
 *	Purely Horizontal strel:
 *	Size of code used to load dword and sidewords for every strel
 *	element plus two ANDs for every strel element plus one write per
 *	output dword.
 *
 *	All other strels (we lump vertical strels into this too)
 *	Size of code to load preshifted input cache (one loop does all rows,
 *	size of loop depends on number of active strel columns and whether
 *	each col has a HIT, a MISS or both) plus 7 (number of rows in output
 *	block) times the number of active elements in the strel times the
 *	length of the length of the AND-with-memory opcode plus the length
 *	of the loop trailer used to write 7 dwords to memory.  To this,
 *	we must add the code that processes blocks one row at a time.  The
 *	code to fill the input cache is the same in this case.  The code
 *	that does the ANDs is different since it uses an output block of
 *	height 1 and has a different loop trailer.
 *
 * 
 * In the SPARC code, we replicated the basic processing loop many times with
 * different registers.  It made sense to figure out what the basic
 * opcodes were once, then plug them in as needed.  Here, there is only one
 * basic loop.  We'll just figure out what x86 opcodes are required as
 * needed.
 *
 * One word on the source of assembly opcodes: if I were very industrious,
 * I'd write a complete compiler producing 386 opcodes.  I'm not.  Instead,
 * the file mortemp3.asm contains little pieces of assembly code that can
 * be combined to produce our routines.  These are little building blocks.
 * This file is the successor to mortemp2.asm, the template for the 386 version
 * of assembly code.
 *
 * Each of the blocks in mortemp3 are labelled.  We simply copy blocks of
 * code from mortemp3 to the output code buffer as required.  Note that
 * for the most part, the particular routines being generated only
 * reference variables kept in the local frame (on the stack), or
 * passed in as parameters (also on the stack).  In particular, there are
 * is only one reference to globals or other remote variables or labels.
 * This allows us to use most addresses (which are really offsets from
 * EBP) from mortemp3 directly.  We must supply address offsets into
 * the cache of addressed inputs (since they depend on the SE) and jump
 * displacements ourselves.
 *
 * What's the one exception, the one global?  There's a dword filled with
 * all 1's.  It is loaded and initialized as part of mortemp3 being loaded and
 * initialized.  The displacement portion of the opcode that references
 * this mask is also updated when mortemp3.asm is loaded.  Therefore, we
 * can just use this opcode, displacement an all.
 *
 * The code to process a strel is broken into two main pieces (except for purely
 * horizontal strels, we'll look at them below): the part that fills the
 * input cache and the part that does the ANDs.  Let's look at each in
 * a little more detail.
 *
 * The code that fills the input cache is contained in one loop.  The
 * beginning of the loop loads the central dword and wordswaps it while
 * loading one sideword or the other.  In purely vertical strels, the
 * wordswapping and sideword loading is left out.  In the middle of the
 * loop, the strel is processed from left to center, then from right to
 * center+1.  At each active strel column, we either write out the positive
 * shifted word, the inverted shifted word or both.  Finally, there is
 * a little code at the end to decrement the number of input cache rows
 * and loop back.
 *
 * The next small section of code prepare for the ANDing section.  It
 * reloads pointers and checks the size of the block.  If the size is
 * less than 7, it jumps down to the code that processes the block one row
 * at a time.
 *
 * The normal ANDing and ORing code also has one loop.  The loop processes
 * one sub-block of height 7.  The top section of the loop does the loading,
 * and ANDing required to do the real processing.  The loop trailer writes
 * out the result words, decrements counters and loops.
 *
 * To generate the top code we read a pair of vertically adjacent input
 * cache dwords and look up the dwords' common strel column (C) and
 * polarity (P).  We also get the input image rows that produced that
 * cache dword.  As specified (far) above, we then go through the rows
 * of the output dwords seeing if strel row
 * (input cache row - output row + cy), strel col C has an active element
 * of polarity P.  If so, the final code want's to do an AND op.  Rather
 * than issue the opcode right away, we count the number of times each
 * input cache dword is used.  There are four cases: 1) neither cache dword
 * is used (This happens if the input cache word affects only a subset of
 * the output dwords and the strel positions for that subset have no
 * HITs or MISSes); 2) the upper cache dword is used only once and the lower
 * cache dword is only used once; 3) the upper cache dword is used more than
 * once and the lower word is only used once; 4) both are used more than once.
 * In the first case, we don't issue any opcodes.  In the second and
 * last cases, we generate an opcode to preload the lower word.  In the
 * third case, we preload the upper word.  We also remember which word
 * was loaded.  We then go through the output dwords again.  This time,
 * we generate AND opcodes.  If the source dword we want to AND with was
 * loaded into a register, we use a reg/reg opcode.  Otherwise, we use a
 * reg/mem opcode.
 *
 * The ANDing loop ends with a trailer that writes all the dwords to memory,
 * increments pointers, decrements the number of rows left (by 7) and
 * loops back.  Following this is a jump to the return opcodes.
 *
 * After the loop that processes rows 7 at a time, there's the code that
 * processes them one at a time.  This is basically the same as the code
 * above except it only looks at one output row.  Since we know that each
 * input cache dword can be used at most once in this case, the loop can
 * omit the preloads and will always use AND reg/mem opcode for every AND.
 * This code also has its own loop trailer that increments pointers
 * and decrements counts and loops back.  At the bottom of this code
 * is the return code for the entire routine.
 *
 * We use some subroutines here.  The most basic just copies a short block
 * of code from the template to the output buffer.  We'll use ip_copyCode
 * from pcgenero.c for this.  Another basic routine copies a block of code
 * that ends in a signed 8-bit or a signed 32-bit displacement.  The
 * number of bytes required for the displacement is calculated from
 * the value of the displacement.  This routine is used to copy
 * loads, stores and jumps.  There's also a routine that sets relative
 * jump displacements.  It takes a pointer to the *end* of the jump
 * and the new displacement value.  It's assumed that the number of bytes
 * reserved for the displacement is always 4.  This routine is used to
 * fix up forward jumps.
 * pcgenero.c has a routine that figures out which kind of AND is needed
 * to implement an SE element.  Since this only happens in one place, we
 * won't bother.
 *
 * The more advanced routines are built on these.  One of them puts out the
 * code to shift and write two dwords to the input cache.  It essentially
 * puts out the code described in the sections labelled "LOOP" in the
 * mortemp3.asm comments above.  This routine has to deal with the following
 * cases:
 *   1.	This is a purely vertical strel.
 *   2. We're to the left of center;
 *	This is the first dword in the row;
 *	there are more dwords in this row (assumed).
 *   3.	We're to the left of center;
 *	this is not the first dword in the row;
 *	there are more dwords in the row,
 *   4.	We're to the left of center;
 *	this is not the first dword in the row;
 *	This is the last dword in the row.
 *   5. We're in the center row;
 *	This is the first dword in the row;
 *	there are more dwords in this row (assumed).
 *   6.	We're in the center row;
 *	this is not the first dword in the row;
 *	there are more dwords in the row,
 *   7.	We're in the center row;
 *	this is not the first dword in the row;
 *	This is the last dword in the row.
 *   8. We're to the right of center;
 *	This is the first dword in the row;
 *	there are more dwords in this row (assumed).
 *   9.	We're to the right of center;
 *	this is not the first dword in the row;
 *	there are more dwords in the row,
 *  10.	We're to the right of center;
 *	this is not the first dword in the row;
 *	This is the last dword in the row.
 *
 * In all cases, the routine takes the pointers to the input cache,
 * the previous, current and next active strel columns (-1 => no col),
 * the strel's center column and whether the column has HITs, MISSes or both.
 * Note that we process the columns from far left to center, then from
 * far right to center + 1.  When processing from left to right, the
 * previous col is to the left and the next col is the to the right.
 * When processing from right to left, the previous column is to the *right*
 * and the next column is to the left.  The displacements in the input
 * cache increase from the left end of the row to the center, then from
 * the right to center + 1.  There are are two arrays, dispToCol and
 * dispToPolarity that are filled in with the strel column and polarity
 * corresponding to a given horizontal displacement in the input cache.
 * These arrays are used to decide how to use the contents of the input
 * cache.
 *
 * There's another higher level subroutine used to generate all the
 * loads and ANDs needed to retire one pair of vertically adjacent
 * input cache dwords.  It takes pointers to the two dwords, their
 * corresponding input image rows and their strel columns and
 * polarities.  It also takes the first row number and height of the output
 * subblock.  It decides whether to issue a preload or not, then generates
 * all the AND opcodes needed.
 *
 * The highest level routines figure out the estimated block size and
 * generate all the code.  These are called from ip_getErodeAsmSize and
 * ipGenErodeAsm respectively in pcgenero.c.  The routines in this module
 * are called when MMX support is present on the current platform.
 *
 * Leigh Klotz found out that if we give this code a
 * strel that's entirely DONT_CARE terms, it generates random code.
 * Check for this case.  If it occurs, don't allocate any assembly code
 * at all.  i_fastErode will fix things up.
 */


    /* dimensions of the largest structure element we can handle */
#define cMaxShiftDist	31

    /* Labels in the template assembly code */
extern	UInt8	mphT3Start, mphT3PreLoadSrcAddrPitch;
extern	UInt8	mphT3PreLoadCacheHeight;
extern	UInt8	mphT3PreLoadCacheAddr, mphT3LoadUpperSrcDWord;
extern	UInt8	mphT3PreLoad1sMask, mphT3PreLoad1sMaskToMM7;
extern	UInt8	mphT3PreLoadUpperLeftSideWord;
extern	UInt8	mphT3PreLoadUpperRightSideWord, mphT3WordSwapUpperStep1;
extern	UInt8	mphT3LoadLowerSrcDWord, mphT3WordSwapUpperStep2;
extern	UInt8	mphT3WordSwapUpperMM2Step1, mphT3WordSwapUpperMM2Step2;
extern	UInt8	mphT3WordSwapLowerMM5Step1, mphT3WordSwapLowerMM5Step2;
extern	UInt8	mphT3PreLoadLowerLeftSideWord, mphT3PreLoadLowerRightSideWord;
extern	UInt8	mphT3WordSwapLowerStep1, mphT3LeftShiftUpperSideWord;
extern	UInt8	mphT3RightShiftUpperSideWord, mphT3WordSwapLowerStep2;
extern	UInt8	mphT3CopyUpperDWord, mphT3LeftShiftLowerSideWord;
extern	UInt8	mphT3RightShiftLowerSideWord, mphT3LeftShiftUpperScratch;
extern	UInt8	mphT3RightShiftUpperScratch, mphT3CopyLowerDWord;
extern	UInt8	mphT3CombineShiftedUpperDWord, mphT3LeftShiftLowerScratch;
extern	UInt8	mphT3RightShiftLowerScratch, mphT3CombineShiftedLowerDWord;
extern	UInt8	mphT3WriteUpperScratchToCacheDisp8;
extern	UInt8	mphT3WriteUpperScratchToCacheDisp32;
extern	UInt8	mphT3WriteLowerScratchToCacheDisp8;
extern	UInt8	mphT3WriteLowerScratchToCacheDisp32;
extern	UInt8	mphT3WriteUpperScratchToDest;
extern	UInt8	mphT3WriteLowerScratchToDest;
extern	UInt8	mphT3InvertUpperShiftedDWord, mphT3InvertLowerShiftedDWord;
extern	UInt8	mphT3InvertAndUpperShiftedDWord;
extern	UInt8	mphT3InvertAndLowerShiftedDWord;
extern	UInt8	mphT3InvertUpperOrigDWord, mphT3InvertLowerOrigDWord;
extern	UInt8	mphT3InvertUpperOrigDWordToMM6, mphT3InvertLowerOrigDWordToMM7;
extern	UInt8	mphT3WriteUpperOrigToCacheDisp8;
extern	UInt8	mphT3WriteUpperOrigToCacheDisp32;
extern	UInt8	mphT3WriteLowerOrigToCacheDisp8;
extern	UInt8	mphT3WriteLowerOrigToCacheDisp32;
extern	UInt8	mphT3AdjustSrcAddrBy2Lines;
extern	UInt8	mphT3AdjustDstAddrBy2Lines;
extern	UInt8	mphT3AdjustCachePtr;
extern	UInt8	mphT3DecrRowsBy2AndBranch;
extern	UInt8	mphT3DecrRowsBy2;
extern	UInt8	mphT3JABranch;
extern	UInt8	mphT3LoadUpperShiftedDWord, mphT3LoadLowerShiftedDWord;
extern	UInt8	mphT3LoadUpperOrigDWord, mphT3LoadLowerOrigDWord;
extern	UInt8	mphT3AndUpperShiftedDWord, mphT3AndLowerShiftedDWord;
extern	UInt8	mphT3AndUpperShiftedDWordToMM2, mphT3AndLowerShiftedDWordToMM5;
extern	UInt8	mphT3OrUpperShiftedDWord, mphT3OrLowerShiftedDWord;
extern	UInt8	mphT3LoadBlockHeight, mphT3LoadDstAddr, mphT3LoadDstPitch;
extern	UInt8	mphT3TestBlockHeight, mphT3LoadMM0Cache8;
extern	UInt8	mphT3LoadMM1Cache8, mphT3LoadMM2Cache8;
extern	UInt8	mphT3LoadMM3Cache8, mphT3LoadMM4Cache8;
extern	UInt8	mphT3LoadMM5Cache8, mphT3LoadMM6Cache8;
extern	UInt8	mphT3LoadMM7Cache8, mphT3LoadMM0Cache32;
extern	UInt8	mphT3LoadMM1Cache32, mphT3LoadMM2Cache32;
extern	UInt8	mphT3LoadMM3Cache32, mphT3LoadMM4Cache32;
extern	UInt8	mphT3LoadMM5Cache32, mphT3LoadMM6Cache32;
extern	UInt8	mphT3LoadMM7Cache32, mphT3LoadMM0Scratch;
extern	UInt8	mphT3LoadMM1Scratch, mphT3LoadMM2Scratch;
extern	UInt8	mphT3LoadMM3Scratch, mphT3LoadMM4Scratch;
extern	UInt8	mphT3LoadMM5Scratch, mphT3LoadMM6Scratch;
extern	UInt8	mphT3AndMM0Cache8;
extern	UInt8	mphT3AndMM1Cache8, mphT3AndMM2Cache8;
extern	UInt8	mphT3AndMM3Cache8, mphT3AndMM4Cache8;
extern	UInt8	mphT3AndMM5Cache8, mphT3AndMM6Cache8;
extern	UInt8	mphT3AndMM0Cache32, mphT3AndMM1Cache32;
extern	UInt8	mphT3AndMM2Cache32, mphT3AndMM3Cache32;
extern	UInt8	mphT3AndMM4Cache32, mphT3AndMM5Cache32;
extern	UInt8	mphT3AndMM6Cache32, mphT3AndMM0Scratch;
extern	UInt8	mphT3AndMM1Scratch, mphT3AndMM2Scratch;
extern	UInt8	mphT3AndMM3Scratch, mphT3AndMM4Scratch;
extern	UInt8	mphT3AndMM5Scratch, mphT3AndMM6Scratch;
extern	UInt8	mphT3OrMM0Cache8, mphT3OrMM1Cache8, mphT3OrMM2Cache8;
extern	UInt8	mphT3OrMM3Cache8, mphT3OrMM4Cache8, mphT3OrMM5Cache8;
extern	UInt8	mphT3OrMM6Cache8, mphT3OrMM0Cache32, mphT3OrMM1Cache32;
extern	UInt8	mphT3OrMM2Cache32, mphT3OrMM3Cache32, mphT3OrMM4Cache32;
extern	UInt8	mphT3OrMM5Cache32, mphT3OrMM6Cache32, mphT3OrMM0Scratch;
extern	UInt8	mphT3OrMM1Scratch, mphT3OrMM2Scratch, mphT3OrMM3Scratch;
extern	UInt8	mphT3OrMM4Scratch, mphT3OrMM5Scratch, mphT3OrMM6Scratch;
extern	UInt8	mphT3WriteSubBlockStart, mphT3WriteSubBlockIncCachePtr;
extern	UInt8	mphT3WriteSubBlockRemainder, mphT3WriteVertSubBlockStart;
extern	UInt8	mphT3WriteVertSubBlockRemainder, mphT3WriteTinySubBlockStart;
extern	UInt8	mphT3WriteTinySubBlockRemainder;
extern	UInt8	mphT3WriteTinyVertSubBlockStart;
extern	UInt8	mphT3WriteTinyVertSubBlockRemainder, mphT3Return;

/* block lengths */
extern	UInt32	mphT3StartBlkLen, mphT3PreLoadSrcAddrPitchBlkLen;
extern	UInt32	mphT3PreLoadCacheHeightBlkLen;
extern	UInt32	mphT3PreLoadCacheAddrBlkLen, mphT3LoadUpperSrcDWordBlkLen;
extern	UInt32	mphT3PreLoad1sMaskBlkLen, mphT3PreLoad1sMaskToMM7BlkLen;
extern	UInt32	mphT3PreLoadUpperLeftSideWordBlkLen;
extern	UInt32	mphT3PreLoadUpperRightSideWordBlkLen;
extern	UInt32	mphT3WordSwapUpperStep1BlkLen;
extern	UInt32	mphT3LoadLowerSrcDWordBlkLen, mphT3WordSwapUpperStep2BlkLen;
extern	UInt32	mphT3WordSwapUpperMM2Step1BlkLen;
extern	UInt32	mphT3WordSwapUpperMM2Step2BlkLen;
extern	UInt32	mphT3WordSwapLowerMM5Step1BlkLen;
extern	UInt32	mphT3WordSwapLowerMM5Step2BlkLen;
extern	UInt32	mphT3PreLoadLowerLeftSideWordBlkLen;
extern	UInt32	mphT3PreLoadLowerRightSideWordBlkLen;
extern	UInt32	mphT3WordSwapLowerStep1BlkLen;
extern	UInt32	mphT3LeftShiftUpperSideWordBlkLen;
extern	UInt32	mphT3RightShiftUpperSideWordBlkLen;
extern	UInt32	mphT3WordSwapLowerStep2BlkLen;
extern	UInt32	mphT3CopyUpperDWordBlkLen, mphT3LeftShiftLowerSideWordBlkLen;
extern	UInt32	mphT3RightShiftLowerSideWordBlkLen;
extern	UInt32	mphT3LeftShiftUpperScratchBlkLen;
extern	UInt32	mphT3RightShiftUpperScratchBlkLen, mphT3CopyLowerDWordBlkLen;
extern	UInt32	mphT3CombineShiftedUpperDWordBlkLen;
extern	UInt32	mphT3LeftShiftLowerScratchBlkLen;
extern	UInt32	mphT3RightShiftLowerScratchBlkLen;
extern	UInt32	mphT3CombineShiftedLowerDWordBlkLen;
extern	UInt32	mphT3WriteUpperScratchToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteUpperScratchToCacheDisp32BlkLen;
extern	UInt32	mphT3WriteLowerScratchToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteLowerScratchToCacheDisp32BlkLen;
extern	UInt32	mphT3WriteUpperScratchToDestBlkLen;
extern	UInt32	mphT3WriteLowerScratchToDestBlkLen;
extern	UInt32	mphT3InvertUpperShiftedDWordBlkLen;
extern	UInt32	mphT3InvertLowerShiftedDWordBlkLen;
extern	UInt32	mphT3InvertAndUpperShiftedDWordBlkLen;
extern	UInt32	mphT3InvertAndLowerShiftedDWordBlkLen;
extern	UInt32	mphT3InvertUpperOrigDWordBlkLen;
extern	UInt32	mphT3InvertLowerOrigDWordBlkLen;
extern	UInt32	mphT3InvertUpperOrigDWordToMM6BlkLen;
extern	UInt32	mphT3InvertLowerOrigDWordToMM7BlkLen;
extern	UInt32	mphT3WriteUpperOrigToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteUpperOrigToCacheDisp32BlkLen;
extern	UInt32	mphT3WriteLowerOrigToCacheDisp8BlkLen;
extern	UInt32	mphT3WriteLowerOrigToCacheDisp32BlkLen;
extern	UInt32	mphT3AdjustSrcAddrBy2LinesBlkLen;
extern	UInt32	mphT3AdjustDstAddrBy2LinesBlkLen;
extern	UInt32	mphT3AdjustCachePtrBlkLen;
extern	UInt32	mphT3DecrRowsBy2AndBranchBlkLen;
extern	UInt32	mphT3DecrRowsBy2BlkLen;
extern	UInt32	mphT3JABranchBlkLen;
extern	UInt32	mphT3LoadUpperShiftedDWordBlkLen;
extern	UInt32	mphT3LoadLowerShiftedDWordBlkLen;
extern	UInt32	mphT3LoadUpperOrigDWordBlkLen;
extern	UInt32	mphT3LoadLowerOrigDWordBlkLen;
extern	UInt32	mphT3AndUpperShiftedDWordBlkLen;
extern	UInt32	mphT3AndLowerShiftedDWordBlkLen;
extern	UInt32	mphT3AndUpperShiftedDWordToMM2BlkLen;
extern	UInt32	mphT3AndLowerShiftedDWordToMM5BlkLen;
extern	UInt32	mphT3OrUpperShiftedDWordBlkLen, mphT3OrLowerShiftedDWordBlkLen;
extern	UInt32	mphT3LoadBlockHeightBlkLen, mphT3LoadDstAddrBlkLen;
extern	UInt32	mphT3LoadDstPitchBlkLen;
extern	UInt32	mphT3TestBlockHeightBlkLen, mphT3LoadMM0Cache8BlkLen;
extern	UInt32	mphT3LoadMM1Cache8BlkLen, mphT3LoadMM2Cache8BlkLen;
extern	UInt32	mphT3LoadMM3Cache8BlkLen, mphT3LoadMM4Cache8BlkLen;
extern	UInt32	mphT3LoadMM5Cache8BlkLen, mphT3LoadMM6Cache8BlkLen;
extern	UInt32	mphT3LoadMM7Cache8BlkLen, mphT3LoadMM0Cache32BlkLen;
extern	UInt32	mphT3LoadMM1Cache32BlkLen, mphT3LoadMM2Cache32BlkLen;
extern	UInt32	mphT3LoadMM3Cache32BlkLen, mphT3LoadMM4Cache32BlkLen;
extern	UInt32	mphT3LoadMM5Cache32BlkLen, mphT3LoadMM6Cache32BlkLen;
extern	UInt32	mphT3LoadMM7Cache32BlkLen, mphT3LoadMM0ScratchBlkLen;
extern	UInt32	mphT3LoadMM1ScratchBlkLen, mphT3LoadMM2ScratchBlkLen;
extern	UInt32	mphT3LoadMM3ScratchBlkLen, mphT3LoadMM4ScratchBlkLen;
extern	UInt32	mphT3LoadMM5ScratchBlkLen, mphT3LoadMM6ScratchBlkLen;
extern	UInt32	mphT3AndMM0Cache8BlkLen;
extern	UInt32	mphT3AndMM1Cache8BlkLen, mphT3AndMM2Cache8BlkLen;
extern	UInt32	mphT3AndMM3Cache8BlkLen, mphT3AndMM4Cache8BlkLen;
extern	UInt32	mphT3AndMM5Cache8BlkLen, mphT3AndMM6Cache8BlkLen;
extern	UInt32	mphT3AndMM0Cache32BlkLen, mphT3AndMM1Cache32BlkLen;
extern	UInt32	mphT3AndMM2Cache32BlkLen, mphT3AndMM3Cache32BlkLen;
extern	UInt32	mphT3AndMM4Cache32BlkLen, mphT3AndMM5Cache32BlkLen;
extern	UInt32	mphT3AndMM6Cache32BlkLen, mphT3AndMM0ScratchBlkLen;
extern	UInt32	mphT3AndMM1ScratchBlkLen, mphT3AndMM2ScratchBlkLen;
extern	UInt32	mphT3AndMM3ScratchBlkLen, mphT3AndMM4ScratchBlkLen;
extern	UInt32	mphT3AndMM5ScratchBlkLen, mphT3AndMM6ScratchBlkLen;
extern	UInt32	mphT3OrMM0Cache8BlkLen, mphT3OrMM1Cache8BlkLen;
extern	UInt32	mphT3OrMM2Cache8BlkLen;
extern	UInt32	mphT3OrMM3Cache8BlkLen, mphT3OrMM4Cache8BlkLen;
extern	UInt32	mphT3OrMM5Cache8BlkLen;
extern	UInt32	mphT3OrMM6Cache8BlkLen, mphT3OrMM0Cache32BlkLen;
extern	UInt32	mphT3OrMM1Cache32BlkLen;
extern	UInt32	mphT3OrMM2Cache32BlkLen, mphT3OrMM3Cache32BlkLen;
extern	UInt32	mphT3OrMM4Cache32BlkLen;
extern	UInt32	mphT3OrMM5Cache32BlkLen, mphT3OrMM6Cache32BlkLen;
extern	UInt32	mphT3OrMM0ScratchBlkLen;
extern	UInt32	mphT3OrMM1ScratchBlkLen, mphT3OrMM2ScratchBlkLen;
extern	UInt32	mphT3OrMM3ScratchBlkLen;
extern	UInt32	mphT3OrMM4ScratchBlkLen, mphT3OrMM5ScratchBlkLen;
extern	UInt32	mphT3OrMM6ScratchBlkLen;
extern	UInt32	mphT3WriteSubBlockStartBlkLen;
extern	UInt32	mphT3WriteSubBlockIncCachePtrBlkLen;
extern	UInt32	mphT3WriteSubBlockRemainderBlkLen;
extern	UInt32	mphT3WriteVertSubBlockStartBlkLen;
extern	UInt32	mphT3WriteVertSubBlockRemainderBlkLen;
extern	UInt32	mphT3WriteTinySubBlockStartBlkLen;
extern	UInt32	mphT3WriteTinySubBlockRemainderBlkLen;
extern	UInt32	mphT3WriteTinyVertSubBlockStartBlkLen;
extern	UInt32	mphT3WriteTinyVertSubBlockRemainderBlkLen, mphT3ReturnBlkLen;


    /* used to write upper scratch reg to input cache */
OpcodeBlock writeUpperScratch = {
    {&mphT3WriteUpperScratchToCacheDisp8,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteUpperScratchToCacheDisp8BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteUpperScratchToCacheDisp32,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteUpperScratchToCacheDisp32BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL}};
    /* used to write lower scratch reg to input cache */
OpcodeBlock writeLowerScratch = {
    {&mphT3WriteLowerScratchToCacheDisp8,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteLowerScratchToCacheDisp8BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteLowerScratchToCacheDisp32,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteLowerScratchToCacheDisp32BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL}};
    /* used to write upper original reg to input cache */
OpcodeBlock writeUpperOrig = {
    {&mphT3WriteUpperOrigToCacheDisp8,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteUpperOrigToCacheDisp8BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteUpperOrigToCacheDisp32,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteUpperOrigToCacheDisp32BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL}};
    /* used to write lower original reg to input cache */
OpcodeBlock writeLowerOrig = {
    {&mphT3WriteLowerOrigToCacheDisp8,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteLowerOrigToCacheDisp8BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteLowerOrigToCacheDisp32,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {&mphT3WriteLowerOrigToCacheDisp32BlkLen,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL}};
OpcodeBlock loadFromCache = {
    {&mphT3LoadMM0Cache8, &mphT3LoadMM1Cache8,
     &mphT3LoadMM2Cache8, &mphT3LoadMM3Cache8,
     &mphT3LoadMM4Cache8, &mphT3LoadMM5Cache8,
     &mphT3LoadMM6Cache8, &mphT3LoadMM7Cache8},
    {&mphT3LoadMM0Cache8BlkLen, &mphT3LoadMM1Cache8BlkLen,
     &mphT3LoadMM2Cache8BlkLen, &mphT3LoadMM3Cache8BlkLen,
     &mphT3LoadMM4Cache8BlkLen, &mphT3LoadMM5Cache8BlkLen,
     &mphT3LoadMM6Cache8BlkLen, &mphT3LoadMM7Cache8BlkLen},
    {&mphT3LoadMM0Cache32, &mphT3LoadMM1Cache32,
     &mphT3LoadMM2Cache32, &mphT3LoadMM3Cache32,
     &mphT3LoadMM4Cache32, &mphT3LoadMM5Cache32,
     &mphT3LoadMM6Cache32, &mphT3LoadMM7Cache32},
    {&mphT3LoadMM0Cache32BlkLen, &mphT3LoadMM1Cache32BlkLen,
     &mphT3LoadMM2Cache32BlkLen, &mphT3LoadMM3Cache32BlkLen,
     &mphT3LoadMM4Cache32BlkLen, &mphT3LoadMM5Cache32BlkLen,
     &mphT3LoadMM6Cache32BlkLen, &mphT3LoadMM7Cache32BlkLen}};
OpcodeBlock andFromCache = {
    {&mphT3AndMM0Cache8, &mphT3AndMM1Cache8,
     &mphT3AndMM2Cache8, &mphT3AndMM3Cache8,
     &mphT3AndMM4Cache8, &mphT3AndMM5Cache8,
     &mphT3AndMM6Cache8, NULL},
    {&mphT3AndMM0Cache8BlkLen, &mphT3AndMM1Cache8BlkLen,
     &mphT3AndMM2Cache8BlkLen, &mphT3AndMM3Cache8BlkLen,
     &mphT3AndMM4Cache8BlkLen, &mphT3AndMM5Cache8BlkLen,
     &mphT3AndMM6Cache8BlkLen, NULL},
    {&mphT3AndMM0Cache32, &mphT3AndMM1Cache32,
     &mphT3AndMM2Cache32, &mphT3AndMM3Cache32,
     &mphT3AndMM4Cache32, &mphT3AndMM5Cache32,
     &mphT3AndMM6Cache32, NULL},
    {&mphT3AndMM0Cache32BlkLen, &mphT3AndMM1Cache32BlkLen,
     &mphT3AndMM2Cache32BlkLen, &mphT3AndMM3Cache32BlkLen,
     &mphT3AndMM4Cache32BlkLen, &mphT3AndMM5Cache32BlkLen,
     &mphT3AndMM6Cache32BlkLen, NULL}};
OpcodeBlock loadFromScratch = {
    {&mphT3LoadMM0Scratch, &mphT3LoadMM1Scratch,
     &mphT3LoadMM2Scratch, &mphT3LoadMM3Scratch,
     &mphT3LoadMM4Scratch, &mphT3LoadMM5Scratch,
     &mphT3LoadMM6Scratch, NULL},
    {&mphT3LoadMM0ScratchBlkLen, &mphT3LoadMM1ScratchBlkLen,
     &mphT3LoadMM2ScratchBlkLen, &mphT3LoadMM3ScratchBlkLen,
     &mphT3LoadMM4ScratchBlkLen, &mphT3LoadMM5ScratchBlkLen,
     &mphT3LoadMM6ScratchBlkLen, NULL},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}};
OpcodeBlock andWithScratch = {
    {&mphT3AndMM0Scratch, &mphT3AndMM1Scratch,
     &mphT3AndMM2Scratch, &mphT3AndMM3Scratch,
     &mphT3AndMM4Scratch, &mphT3AndMM5Scratch,
     &mphT3AndMM6Scratch, NULL},
    {&mphT3AndMM0ScratchBlkLen, &mphT3AndMM1ScratchBlkLen,
     &mphT3AndMM2ScratchBlkLen, &mphT3AndMM3ScratchBlkLen,
     &mphT3AndMM4ScratchBlkLen, &mphT3AndMM5ScratchBlkLen,
     &mphT3AndMM6ScratchBlkLen, NULL},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}};

/* Define macros for common subroutine calls with fancy abort sequences.
 * This should clean up the code a lot since most of the information in
 * the calls does not change.
 */
#define	M_COPY_CODE(blkAddr) \
	    if ((__error__ = ip_copyCode (pIPState, &(blkAddr), \
					(blkAddr##BlkLen), \
				        pCodeBuf, &bufIndex, bufSize)) \
		!= ia_successful) \
		goto genCleanup

#define	M_COPY_DISP(opcode, reg, disp) \
	    if ((__error__ = ip_longCopyDisp (pIPState, &(opcode), \
					    (reg), (disp), \
					    pCodeBuf, &bufIndex, bufSize)) \
		!= ia_successful) \
		goto genCleanup

/* Define scratch register used when doing ANDing */
#define	cScratchReg 7


/* PRIVATE */
/*************************************************************************
 *  ip_longOpIPLibLinked()
 *		There are things we'd really not do if routines capable
 *		of long (64 bit or longer) are linked, like avoid
 *		making superflous tables.  This routine returns a
 *		boolean which is set to cTrue if the
 *		long memory access libraries are linked.
 *
 *	Input:
 *		None
 *	Output:
 *		cTrue if fast libraries are linked, cFalse otherwise.
 *************************************************************************/
Bool  CDECL 
ip_longOpIPLibLinked (
    IPState    *pIPState)
{
    PROLOGUE;
	/* Keep compiler happy. */
    pIPState = pIPState;
    __error__ = __error__;

	/* Clearly the fast libs are here since we're here. */
    return cTrue;
}


/* PRIVATE */
/************************************************************************
 *  i_getLongErodeAsmSize():  return number of bytes needed for assembly code
 *			used for erosion with this strel using Long (>= 64 bit)
 *			opcodes.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	pNumAsmBytes:	Pointer to the number of bytes needed for the assembly
 *			buffer.  The real assembly code will seldom fill the
 *			buffer, clients may want to resize the buffer
 *			after generating the code.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
i_getLongErodeAsmSize (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    Int32Ptr	pNumAsmBytes)
{
    PROLOGUE;
    Int32	blockLen;
    Int32	numActiveCols;
    Int32	numActiveElements;
    Int32	longestAccess;	/* longest data type supported on
				 * current platform in bytes.  Must
				 * be 8 for this routine to work */

	/* Reference cx to keep the compiler happy */
    cx = cx;
    cy = cy;

	/* See if long accesses are allowed on this platform.  If
	 * not, return ia_noImpl.
	 */
    ip_getLongestAccess (pIPState, &longestAccess);
    if (longestAccess != 8)
	return ia_noImpl;

	/* Since we're making worst case assumptions, and don't have
	 * the strel to inspect anyway.  Assume every strel column is
	 * active and has both HITs and misses.  Assume every possible
	 * strel element is active.
	 */
    numActiveCols = nx;
    numActiveElements = nx * ny;

	/* let's do this a little at a time.  Start with the length
	 * of the prologue code.
	 */
    blockLen = mphT3StartBlkLen + mphT3PreLoadSrcAddrPitchBlkLen +
	       mphT3PreLoadCacheHeightBlkLen +
	       mphT3PreLoadCacheAddrBlkLen;

	/* Now we need to generate code to fill the input cache.  If
	 * This is a purely horizontal strel, we don't actually do
	 * that, but the code isn't all that much smaller, so we
	 * pretend we do.  There's some start up code, then code to
	 * process all the elements to the left of center, then code
	 * to process the center, then code to process all the elements
	 * to the right of center.
	 * We start with the cononical preload.  Preloading for vertical
	 * strels is shorter, but that's ok.
	 */
    blockLen += mphT3LoadUpperSrcDWordBlkLen +
		mphT3PreLoad1sMaskBlkLen +
		mphT3PreLoadUpperLeftSideWordBlkLen +
		mphT3WordSwapUpperStep1BlkLen +
		mphT3LoadLowerSrcDWordBlkLen +
		mphT3WordSwapUpperStep2BlkLen +
		mphT3PreLoadLowerLeftSideWordBlkLen +
		mphT3WordSwapLowerStep1BlkLen +
		mphT3LeftShiftUpperSideWordBlkLen +
		mphT3WordSwapLowerStep2BlkLen +
		mphT3CopyUpperDWordBlkLen +
		mphT3LeftShiftLowerSideWordBlkLen +
		mphT3LeftShiftUpperScratchBlkLen +
		mphT3CopyLowerDWordBlkLen;


	/* For all the SE columns to the left and right of center (yeah
	 * the code for the last col is a little shorter, but so what),
	 * we add the standard loop processing that stores both positive
	 * and negative shifted words.
	 */
    blockLen += (numActiveCols - 1) *
		 (mphT3CombineShiftedUpperDWordBlkLen +
		  mphT3LeftShiftLowerScratchBlkLen +
		  mphT3LeftShiftUpperSideWordBlkLen +
		  mphT3CombineShiftedLowerDWordBlkLen +
		  mphT3WriteUpperScratchToCacheDisp32BlkLen +
		  mphT3InvertUpperShiftedDWordBlkLen +
		  mphT3WriteLowerScratchToCacheDisp32BlkLen +
		  mphT3InvertLowerShiftedDWord +
		  mphT3WriteUpperScratchToCacheDisp32BlkLen +
		  mphT3LeftShiftLowerSideWordBlkLen +
		  mphT3WriteLowerScratchToCacheDisp32BlkLen +
		  mphT3CopyUpperDWordBlkLen +
		  mphT3LeftShiftUpperScratchBlkLen +
		  mphT3CopyLowerDWordBlkLen);

	/* Add in cycles to deal with both polarities of the central column */
    blockLen += mphT3WriteUpperOrigToCacheDisp32BlkLen +
		mphT3InvertUpperOrigDWordBlkLen +
		mphT3WriteLowerOrigToCacheDisp32BlkLen +
		mphT3InvertLowerOrigDWordBlkLen +
		mphT3WriteUpperOrigToCacheDisp32BlkLen +
		mphT3InvertUpperOrigDWordBlkLen +
		mphT3WriteLowerOrigToCacheDisp32BlkLen +
		mphT3InvertLowerOrigDWordBlkLen;

	/* At the end of the loop, we test and branch back */
    blockLen += mphT3DecrRowsBy2AndBranchBlkLen;

	/* Now, we reinitialize things for the ANDs part of the loop. */
    blockLen += mphT3LoadBlockHeightBlkLen +
		mphT3LoadDstAddrBlkLen +
		mphT3LoadDstPitchBlkLen +
		mphT3PreLoadCacheAddrBlkLen +
		mphT3TestBlockHeightBlkLen;

	/* Now for the ANDs and stuff for the subblock.  Since this
	 * is a worst-case analysis, we assume that every AND is with
	 * memory.
	 */
    blockLen += numActiveElements * cLongSubBlockHeight *
		mphT3AndMM0Cache32BlkLen;

	/* Now we write all this stuff out to memory */
    blockLen += mphT3WriteSubBlockStartBlkLen +
		mphT3WriteSubBlockIncCachePtrBlkLen +
		mphT3WriteSubBlockRemainderBlkLen;

	/* Here's the first copy of the return code.  We return instead
	 * of jumping to the return below the 1-line-at-a-time code.
	 */
    blockLen += mphT3ReturnBlkLen;

	/* Here's the one-line-at-a-time code.  It's a lot like the regular
	 * code except that we don't multiply by cLongSubBlockHeight since it's
	 * equal to 1.
	 */
    blockLen += numActiveElements *
		mphT3AndMM0Cache32BlkLen;

	/* Now we write all this stuff out to memory */
    blockLen += mphT3WriteTinySubBlockStartBlkLen +
		mphT3WriteSubBlockIncCachePtrBlkLen +
		mphT3WriteTinySubBlockRemainderBlkLen;

	/* Here's the last copy of the return code.  This is how we
	 * return if we execute the 1-line-at-a-time code.
	 */
    blockLen += mphT3ReturnBlkLen;

	/* return the number we calculated */
    *pNumAsmBytes = blockLen;

    NORMAL_RETURN;
}


/* PUBLIC */
/************************************************************************
 *  i_genLongErodeAsm():  generate assembly code to do erode defined by strel.
 *			This assembly code makes use of long memory accesses
 *			and operations (long >= 64 bits)
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	cx, cy:		Structure element's "center" is at location (cx,cy)
 *			relative to upper left corner.  (0,0) is the upper
 *			left element of the structure element.
 *	data:		Pointer to an array of nx * ny bytes specifying
 *			the structure element.  Array is arranged so
 *			that the first ny elements contain the leftmost
 *			column of the SE from top to bottom.  The
 *			next ny elements have the next column and so on.
 *			Byte encodings are cDontCare=0, cHit=1, cMiss=2. 
 *	blockWidth:	Pointer to word that this routine will fill with
 *			width of block to be processed by assembly code.
 *			This parameter should be passed to i_fastErode as
 *			longBlockWidthInBytes.
 *	generated_code:	Pointer to pointer to final assembly code.  If
 *			*generated_code is not NULL on entry, we assume space
 *			has already been allocated for the assembly code.
 *			Otherwise, this routine will allocate it using MALLOC.
 *			If the buffer was passed in, we assume it has at
 *			least the number of bytes that i_getLongErodeAsmSize
 *			would calculate.
 *	code_length:	Pointer to a word containing the final length of the
 *			assembly code in units of BYTES (change from previous
 *			version).  May be used disassemblers during debugging.
 *
 * This function returns one of several completion codes defined in iaerror.h.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_SETooBig:		Structure element was too big to
 *				generate accelerated code OR was composed
 *				entirely of DONT_CARE terms (which is silly)
 *				OR was a 1x1 strel.
 *	ia_noImpl:		The current platform doesn't support long ops.
 *	ia_nomem:		Not enough memory to allocate the code buffer.
 *	ia_bufferExceeded:	The routine tried to generate too much
 *				assembly code.  This is a bug in the routine.
 *				If it happens, please let us know.
 ************************************************************************/
Int32  CDECL 
i_genLongErodeAsm (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cx,
    Int32	cy,
    const UInt8 *data,
    Int32      *blockWidth,
    LongErodeProc *generated_code,
    Int32      *code_length)
{
    PROLOGUE;
    UInt8      	 *pCodeBuf;		/* output code buffer */
    Bool	  bufAllocatedHere;	/* set if this proc allocated the
					 * asm code buffer. */
    Int32         bufIndex;		/* index of code buffer */
    Int32         bufSize;		/* size of mem allocated for code
					 * buffer */
    Int32	  row, col;		/* index through the SE */
    Int32	  cacheWidthInBytes;	/* width of cache in units of
					 * bytes. */
    Int32	  cacheHeight;		/* total height of input cache */
    Int32	  numActiveCols;	/* number of active SE columns */
    UInt8	 *seColActive = NULL;	/* array of booleans telling whether
					 * each SE col has any HITs or
					 * MISSes above the bottom row. */
    Int32	  prevCol;		/* the previous active column, used
					 * when filling input cache. */
    Int32	  curCol;		/* the current active column, used
					 * when filling input cache. */
    Int32	  nextCol;		/* next active column, used when
					 * filling input cache. */
    Int32	 *pWordDisplacement;	/* point to word address offset */
    UInt8	 *polIDArray = NULL;	/* array of polarities.  Used to
					 * get the polarity of a dword
					 * in an input cache row. */
    UInt8	 *colIDArray = NULL;	/* array of column numbers.  used to
					 * get the SE column of a dword
					 * in an input cache row. */
    Int32	  colIDIndex;		/* index into polIDArray and
					 * colIDArray. */
    Int32	  topOfLoopIndex;	/* Offset of top of shift loop from
					 * beginning of buffer. */
    Int32	  subBlockJumpIndex;	/* remembers location of jump that
					 * decides where to process an entire
					 * subblock or one line at a time.
					 */
    Int32	  topOfLowerLoopIndex;	/* location of top of loop that does
					 * ANDS. */
    Bool	  outInitArray[cLongSubBlockHeight];  /* used to decide whether
					 * an output reg has been touched */
    Int32	  longestAccess;	/* longest data type supported on
					 * current platform in bytes.  Must
					 * be 8 for this routine to work */
    


	/* First, decide if the SE is too big. */
    if ((cx > cMaxShiftDist) ||
	(cy > cMaxShiftDist) ||
	((nx - cx - 1) > cMaxShiftDist) ||
	((ny - cy - 1) > cMaxShiftDist) ||
	((nx == 1) && (ny == 1)) )
    {
        return ia_SETooBig;
    }

	/* See if long accesses are allowed on this platform.  If
	 * not, return ia_noImpl.
	 */
    ip_getLongestAccess (pIPState, &longestAccess);
    if (longestAccess != 8)
	return ia_noImpl;

	/* Go create an array of active columns.  First, allocate
	 * it.
	 */
    seColActive = (UInt8 *)CALLOC(nx, sizeof(UInt8));
    if (seColActive == NULL)
	return ia_nomem;

    __error__ = ip_getLongErodeSEParms (pIPState, nx, ny,
				      data, &cacheWidthInBytes,
				      &cacheHeight, &numActiveCols,
				      &(seColActive[0]));
    if (__error__ != ia_successful)
	goto genCleanup;

	/* If the SE has no active components, complain */
    if (numActiveCols == 0)
    {
	__error__ = ia_SETooBig;
	goto genCleanup;
    }

	/* our long blockWidth is 8, so just set that now. */
    *blockWidth = 8;

	/* If we don't already have an assembly buffer, figure out how
	 * big one should be and allocate it.
	 * Even if we were given a buffer, assume it's at least this big.
	 * If the real code doesn't fit into a buffer of this
	 * size, i_getLongErodeAsmSize should be fixed.
	 */
    i_getLongErodeAsmSize (pIPState, nx, ny, cx, cy, &bufSize);

	/* In Visual C++, turn off the warning that says we're about to
	 * cast a pointer to a function to a pointer to data, since
	 * that's exactly what we intend to do.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4054 )
#endif
    pCodeBuf = (UInt8 *) *generated_code;
#ifdef _MSC_VER
#pragma warning ( default : 4054 )
#endif

    if (pCodeBuf == NULL)
    {
	pCodeBuf = (UInt8 *) MALLOC(bufSize);
	bufAllocatedHere = cTrue;
    }
    else
	bufAllocatedHere = cFalse;

	/* if the pointer is null, punt.  Otherwise, return the address
	 * of the new buffer.
	 */
    if (pCodeBuf == NULL)
        return ia_nomem;
    else
	/* In Visual C++, turn off the warning that says we're casting
	 * a pointer to data to a pointer to a function.  We're going
	 * to do exactly that.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
	*generated_code = (LongErodeProc) pCodeBuf;
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif

	/* allocate the arrays used to convert a displacement into
	 * an input cache row into a polarity and SE column for
	 * the dword recorded there.
	 */
    polIDArray = (UInt8 *) CALLOC((cacheWidthInBytes / 8), sizeof(UInt8));
    if (polIDArray == NULL)
    {
	__error__ = ia_nomem;
	goto genCleanup;
    }
    colIDArray = (UInt8 *) CALLOC((cacheWidthInBytes / 8), sizeof(UInt8));
    if (colIDArray == NULL)
    {
	__error__ = ia_nomem;
	goto genCleanup;
    }
    colIDIndex = 0;		/* start with nothing in here */

	/* Initialize the index into the code buffer.  We use an index
	 * because we usually need to compute things using the *offset*
	 * into the buffer, as opposed to using the absolute buffer
	 * pointer itself.
	 */
    bufIndex = 0;


/* *************** START OF CODE GENERATION ********************** */


	/* Start with the initialization code */
    M_COPY_CODE(mphT3Start);

	/* preload the source address and source address pitch */
    M_COPY_CODE(mphT3PreLoadSrcAddrPitch);

	/* If this is not a purely horizontal strel, we need code
	 * that will load the input cache.  If this is a horizontal
	 * strel (height == 1), we don't load the input cache, but
	 * do a bunch of shifting and writing to the destination.
	 * So, if this is NOT a purely horizontal strel, generate
	 * code to load the height of the input cache and a pointer
	 * to the input cache.  If this is a purely horizontal strel,
	 * generate code to load the dest address and dest image
	 * pitch and the height of the entire block.  Purely horizontal
	 * strels are computed an entire block at a time.
	 */
    if (ny > 1)
    {
	M_COPY_CODE(mphT3PreLoadCacheAddr);
	M_COPY_CODE(mphT3PreLoadCacheHeight);
    }
    else	/* purely horizontal strel */
    {
	M_COPY_CODE(mphT3LoadDstAddr);
	M_COPY_CODE(mphT3LoadDstPitch);
	M_COPY_CODE(mphT3LoadBlockHeight);
    }
	/* Here's where we either fill the input cache (not a purely
	 * horizontal strel) or process the entire block (purely
	 * horizontal strel).  In both cases, we go through the columns
	 * of the strel starting at the center column, then working
	 * from the left edge towards column center - 1, then working from
	 * the right edge towards column center + 1.  ip_mmxCreateErodeShiftCode
	 * puts out the right code for each column.
	 */
    prevCol = curCol = nextCol = -1;	/* no cols seen yet */
    topOfLoopIndex = 0xFFFFFFFF;	/* not initialized yet */

	/* If the center's column is active, it will be the first
	 * column.  Otherwise, go look for the first column.
	 */
    if (seColActive[cx] != cDontCare)
    {
	nextCol = cx;
    }
    else	/* go look for first col.  Start at left end of strel.
		 * Since we know that there's at least one active
		 * column (we would have given up if there were
		 * no active columns) we can just assume this
		 * search is going to work.
		 */
    {
	for (nextCol = 0; nextCol < cx; nextCol++)
	{
	    if (seColActive[nextCol] != cDontCare)
		break;
	}
	if (nextCol >= cx)	/* haven't found the first active col yet,
				 * look on the right */
	{
	    for (nextCol = nx-1; nextCol > cx; nextCol--)
	    {
		if (seColActive[nextCol] != cDontCare)
		    break;
	    }
	}
    }

	/* Here's the loop where we generate columns.  First, update
	 * the current values of curCol and prevCol.  Then go find
	 * the next column to be active.
	 */
    while (nextCol != -1)
    {
	prevCol = curCol;	/* pass the columns down the chain */
	curCol = nextCol;

	    /* if we enter with nextCol == cx, start looking on
	     * the left for the next column.  If we were already
	     * on the left, move right by 1.  If we were already
	     * on the right, move left by 1.
	     */
	if (nextCol == cx)
	    nextCol = 0;
	else
	if (nextCol < cx)
	{
	    nextCol++;
	    if (nextCol == cx)
		nextCol = nx-1;
	}
	else
	if (nextCol > cx)
	{
	    nextCol--;
	    if (nextCol == cx)
		nextCol = -1;	/* there is no next col, so just finish
				 * the current column */
	}

	if (nextCol != -1)
	{
	    if (nextCol < cx)	/* look for next col on left */
	    {
		for (; nextCol < cx; nextCol++)
		{
		    if (seColActive[nextCol] != cDontCare)
			break;
		}
	    }
    
	    if (nextCol == cx)	/* haven't found one yet */
		nextCol = nx-1;	/* so start on right edge */
    
	    if (nextCol > cx)	/* keep looking on the right */
	    {
		for (; nextCol > cx; nextCol--)
		{
		    if (seColActive[nextCol] != cDontCare)
			    break;
		}
	    }
    
	    if (nextCol == cx)	/* there are no more columns */
		nextCol = -1;	/* do the last one */
	}

	__error__ = ip_mmxCreateErodeShiftCode (pIPState,
					      seColActive[curCol], /* polarity*/
					      prevCol, curCol, nextCol,
					      cx,		/* centerCol */
					      ny,		/* numRows */
					      nx,		/* numCols */
					      cacheWidthInBytes, /* cacheBpl */
					      colIDArray,
					      polIDArray,
					      &colIDIndex,
					      &topOfLoopIndex,
					      pCodeBuf,
					      &bufIndex,
					      bufSize);
	if (__error__ != ia_successful)
	    goto genCleanup;
    }

	/* Swell, we've generated the code to fill the input cache
	 * (not purely horizontal strel) or have generated the code
	 * to process the entire subblock (purely horizontal strel).
	 * If this isn't a purely horizontal strel, generate the
	 * code to do all the ANDs.
	 */
    if (ny > 1)
    {
	    /* First, load the stuff we'll need for the lower code. */
	M_COPY_CODE(mphT3LoadBlockHeight);
	M_COPY_CODE(mphT3LoadDstAddr);
	M_COPY_CODE(mphT3LoadDstPitch);
	M_COPY_CODE(mphT3PreLoadCacheAddr);

	    /* We'll loop back to here after processing one
	     * subblock.  This way, we can test to see if the
	     * next subblock is full height.  If not, we'll
	     * process it one line at a time.
	     */
	topOfLowerLoopIndex = bufIndex;

	    /* Insert the test/jump that used to decide if we're
	     * processing a full subblock.  If not, we process
	     * one row at a time.
	     */
	M_COPY_CODE(mphT3TestBlockHeight);

	    /* Remember the jump`s location so we can fix up the
	     * target address later.
	     */
	subBlockJumpIndex = bufIndex;
    
	    /* Cycle through the rows of the input cache.  We really
	     * only cycle through the rows needed for one subblock.
	     * The code we generate here processes the block in
	     * units of subblocks.  At the bottom, there's a loop
	     * that does additional subblocks.  Note that we
	     * process the rows two at a time, so increment the
	     * row number by 2.  We won't generate any opcodes
	     * for portions of the input cache that aren't used
	     * by a particular subblock.  First, initialize
	     * the array used to decide whether an output
	     * element has been touched yet.
	     */
	for (row = 0; row < cLongSubBlockHeight; row++)
	    outInitArray[row] = cFalse;

	for (row = -cy; row < cLongSubBlockHeight + (ny - cy); row += 2)
	{
		/* Loop through all the dwords in this input cache
		 * row.  Generate the right sequence of loads and
		 * ANDs for each dword.
		 */
	    for (col = 0; col < (cacheWidthInBytes/8); col++)
	    {
		__error__ = ip_mmxCreateANDCode (pIPState, nx, ny, cy, data, row,
					       (row+cy)*cacheWidthInBytes +
							(col * 8),
					       colIDArray[col],
					       polIDArray[col],
					       cacheWidthInBytes,
					       cLongSubBlockHeight,
					       &(outInitArray[0]),
					       pCodeBuf,
					       &bufIndex,
					       bufSize);
		if (__error__ != ia_successful)
		    goto genCleanup;
	    }
	}

	    /* We've generated all the ANDs for one subblock.  Generate
	     * the code that writes the subblock to memory.  There are
	     * two such blocks, one for vertical strels (no byteswapping)
	     * and one for strels that need some shifting.
	     */
	if ((numActiveCols > 1) || (seColActive[cx] == cDontCare))
	{
		/* Strel needs some shifting */
	    M_COPY_CODE(mphT3WriteSubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's cLongSubBlockHeight rows in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = cLongSubBlockHeight * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteSubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lower loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}
	else
	{
		/* Strel is purely vertical */
	    M_COPY_CODE(mphT3WriteVertSubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's cLongSubBlockHeight rows in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = cLongSubBlockHeight * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteVertSubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lower loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}

	    /* If we exit that code, just return now instead of
	     * jumping around the one-row-at-at-time code. */
	M_COPY_CODE(mphT3Return);

	    /* That was fine.  Now we need to generate the code used to
	     * create one output row at a time.  This looks a lot like
	     * the code above, but we use 1 instead of cLongSubBlockHeight.
	     */
	pWordDisplacement = (Int32 *) &(pCodeBuf[subBlockJumpIndex-4]);
	*pWordDisplacement = bufIndex - subBlockJumpIndex;

	    /* This will also be the top of the lowest loop.  We jump
	     * back here after processing one line.
	     */
	topOfLowerLoopIndex = bufIndex;

	    /* initialize the single output register to be untouched */
	outInitArray[0] = cFalse;

	    /* go through the rows and columns of the input cache that
	     * apply to one output row.
	     */
	for (row = -cy; row < 1 + (ny - cy); row += 2)
	{
		/* Loop through all the dwords in this input cache
		 * row.  Generate the right sequence of loads and
		 * ANDs for each dword.
		 */
	    for (col = 0; col < (cacheWidthInBytes/8); col++)
	    {
		__error__ = ip_mmxCreateANDCode (pIPState, nx, ny, cy, data, row,
					       (row+cy)*cacheWidthInBytes +
							(col * 8),
					       colIDArray[col],
					       polIDArray[col],
					       cacheWidthInBytes,
					       1,
					       &(outInitArray[0]),
					       pCodeBuf,
					       &bufIndex,
					       bufSize);
		if (__error__ != ia_successful)
		    goto genCleanup;
	    }
	}

	    /* We've generated all the ANDs for one subblock of height 1.
	     * Generate the code that writes the single dword reg to memory.
	     * There are two such code blocks, one for vertical strels
	     * (no byteswapping) and one for strels that need some shifting.
	     */
	if ((numActiveCols > 1) || (seColActive[cx] == cDontCare))
	{
		/* Strel needs some shifting */
	    M_COPY_CODE(mphT3WriteTinySubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's 1 row in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = 1 * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteTinySubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lowest loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}
	else
	{
		/* Strel is purely vertical */
	    M_COPY_CODE(mphT3WriteTinyVertSubBlockStart);
	    M_COPY_CODE(mphT3WriteSubBlockIncCachePtr);
		/* We increment the rCache register by the difference
		 * between the starting position of this subblock and
		 * the starting position of the next subblock.
		 * That's 1 row in the input cache.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = 1 * cacheWidthInBytes;

	    M_COPY_CODE(mphT3WriteTinyVertSubBlockRemainder);
		/* Set the negative displacement back to the top of the
		 * lowest loop.
		 */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[bufIndex-4]);
	    *pWordDisplacement = topOfLowerLoopIndex - bufIndex;
	}
    }	/* end of generating code to do ANDs in non-horizontal strel */

	/* We're almost done.  Generate the code to return. */
    M_COPY_CODE(mphT3Return);

	/* Now for a little error checking.  If we've generated more code
	   than fits in the buffer, complain */
    if (bufIndex >= bufSize)
    {
#if 0
	fprintf(stderr, " Buffer overflow, memory has been *smashed* in %s\n %d bytes written into buffer of size %d\n",
		i_genLongErodeAsm, bufIndex, bufSize);
#endif
	__error__ = ia_bufferExceeded;
	goto genCleanup;
    }

    *code_length = bufIndex;
    __error__ = ia_successful;

genCleanup:
    if (seColActive != NULL)
	FREE(seColActive);
    if (polIDArray != NULL)
	FREE(polIDArray);
    if (colIDArray != NULL)
	FREE(colIDArray);

    NORMAL_RETURN;
}



/* PRIVATE */
/************************************************************************
 *  ip_getLongErodeSEParms():
 *			return dimensions of input cache needed for long
 *			erode ops on a strel along with an array of the
 *			active columns in the SE.
 *
 *	nx, ny:		Structure element has nx columns and ny rows
 *	pSE:		Pointer to structure element's data.
 *	pCacheWidth:	Pointer to returned width of cache in bytes.
 *			This may be NULL.
 *	pCacheHeight:	Pointer to returned height of input cache.
 *			This may be NULL.
 *	pNumActiveCols:	Pointer to returned number of active columns.
 *			This may be NULL.
 *	pActiveCols:	Pointer to returned array of nx bytes.  This
 *			array should be allocated (and freed) by the caller.
 *			This routine will fill each element in the array
 *			with 0 if there are no SE elements in the column,
 *			cHit if the SE column contains only HITs and don't
 *			care terms, cMiss if the column contains only MISSes
 *			and don't care terms and (cHit | cMiss) if the
 *			column contains both HITs and MISSes (and possibly
 *			don't care terms).  This may be NULL.
 *
 * This function returns one of several completion codes defined in iaerror.pub.
 * They are:
 *	ia_successful:		The operation worked.
 *	ia_invalidParm:		parameters were illegal.
 ************************************************************************/
Int32  CDECL 
ip_getLongErodeSEParms (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    const UInt8 *pSE,
    Int32      *pCacheWidth,
    Int32      *pCacheHeight,
    Int32      *pNumActiveCols,
    UInt8      *pActiveCols)
{
    PROLOGUE;
    Int32	 numActiveCols;
    Int32	 cacheWidth;
    Int32	 currentRow, currentCol;
    Bool	 allocatedHere;

	/* See if the parameters are ok. */
    if (pSE == NULL)
	return ia_invalidParm;

	/* If we weren't passed a pActiveCols, create one.
	 * We'll need it to calculate stuff.
	 */
    if (pActiveCols == NULL)
    {
	pActiveCols = (UInt8 *)CALLOC(nx, sizeof(UInt8));
	if (pActiveCols == NULL)
	    return ia_nomem;
	allocatedHere = cTrue;
    }
    else
    {
	    /* Clear all the entries in pActiveCols */
	for (currentCol = nx-1; currentCol >= 0; currentCol--)
	    pActiveCols[currentCol] = 0;
	allocatedHere = cFalse;
    }

	/* Go through the SE looking for active elements. */
    for (currentRow = ny; currentRow > 0; currentRow--)
    {
	for (currentCol = 0; currentCol < nx; currentCol++, pSE++)
	{
	    pActiveCols[currentCol] |= *pSE;	/* add indicator for current
						 * element */
	}
    }

	/* Count the number of active columns in the SE.  Also reserve
	 * one spot for each dword in the input cache rows.
	 */
    for (numActiveCols = 0, cacheWidth = 0, currentCol = nx-1;
	 currentCol >= 0;
	 currentCol--)
    {
	if (pActiveCols[currentCol] != 0)
	{
	    numActiveCols++;
	    cacheWidth++;
	    if (pActiveCols[currentCol] == (cHit | cMiss))
		cacheWidth++;	/* both a hit and a miss, so reserve another
				 * spot */
	}
    }

	/* Return the calculated values.  Note that the height of the
	 * input cache is equal to the height of the strel plus the
	 * height of the blocks we'll be processing.  Since we always
	 * process two rows at a time, we round this up to the next
	 * multiple of 2.
	 */
    if (pCacheWidth != NULL)
	*pCacheWidth = cacheWidth * 8;		/* 8 bytes/dword */
    if (pCacheHeight != NULL)
	    /* should really be ((cMaxLongBlockHeight + ny-1)+1) & ~0x1 */
	*pCacheHeight = (cMaxLongBlockHeight + ny) & ~0x1;
    if (pNumActiveCols != NULL)
	*pNumActiveCols = numActiveCols;
    if (allocatedHere)
	FREE(pActiveCols);	/* otherwise, return it */

    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_longCopyDisp():  copy an opcode ending an a displacement into the
 *			output buffer.  Adjust the size of the displacement
 *			correctly.  Choose the opcode from a list of
 *			similar opcodes according to the register parm.
 *
 *	pOpcode:	Pointer to OpcodeBlock that contains ptrs to
 *			starting locations and lengths of possible
 *			source opcodes to be copied.
 *	reg:		Number of source or dest reg.  Used to select
 *			opcode.
 *	displacement;	the displacement to be added to the end of the opcode.
 *			If this is in [-128, 127], use and 8-bit displacement.
 *			Otherwise use a 32-bit displacement.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 * Return codes:
 *	ia_successful:		The operation worked.
 *	ia_bufferExceeded:	opcode didn't fit in buffer.
 ************************************************************************/
Int32  CDECL 
ip_longCopyDisp (
    IPState    *pIPState,
    OpcodeBlock *pOpcode,
    Int32	reg,
    Int32	displacement,
    UInt8      *pCodeBuf,
    Int32      *pBufIndex,
    Int32	bufSize)
{
    PROLOGUE;
    
    Int32	*pWordDisplacement;	/* pts to displacements */

	/* figure out whether to use short or long displacement */
    if ((displacement >= -128) && (displacement <= 127))
    {
	__error__ = ip_copyCode (pIPState, pOpcode->disp8[reg],
			       *(pOpcode->disp8Len[reg]),
			       pCodeBuf, pBufIndex, bufSize);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;

	pCodeBuf[(*pBufIndex)-1] = (Int8) (displacement);
    }
    else	/* use 32-bit displacement */
    {
	__error__ = ip_copyCode (pIPState, pOpcode->disp32[reg],
			       *(pOpcode->disp32Len[reg]),
			       pCodeBuf, pBufIndex, bufSize);
	if (__error__ != ia_successful)
	    NORMAL_RETURN;

	pWordDisplacement = (Int32 *) &(pCodeBuf[(*pBufIndex)-4]);
	*pWordDisplacement = displacement;
    }

    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_mmxCreateErodeShiftCode():
 *			Create a block of MMX code that puts out the shifted
 *			64-bit dword(s) for one Structure Element column.
 *
 *	polarity:	Indicates polarities of active elements in current
 *			SE column.  May equal cHit, cMiss or (cHit | cMiss).
 *	prevCol:	Number of the previous SE column for which shifted
 *			dwords were put out.  This is equal to -1 if curCol is
 *			the first column.
 *	curCol:		Number of the current SE column for which shifted
 *			dwords will be put out.
 *	nextCol;	Number of the next SE column for which shifted dwords
 *			will be put out.
 *	centerCol:	The number of the SE's center column.
 *	numRows:	The total number of rows in the strel.
 *	numCols:	The total number of columns in the strel.
 *	cacheBpl:	The bytes/line in the input cache.
 *	colIDArray	Pointer to array which maps horizontal displacements
 *			in the input cache to SE column numbers.  One or
 *			two entries will be added to this array here.
 *	polIDArrray:	Pointer to array which maps horizontal displacements
 *			in an input cache row to cHit or cMiss polarities.
 *			When used in conjunction with colIDArray, this array
 *			can be used to identify a particular element in the
 *			input cache.
 *	pCurDisp:	Pointer to current displacement in colIDArray and
 *			polIDArray.  The displacement will be updated upon
 *			returning from this routine.
 *	pTopOfLoopIndex: Pointer to the offset into pCodeBuf of the first
 *			byte of the shift loop.  The offset is set by
 *			this routine when the loop starts.  It is referenced
 *			by a later invocation of this routine when the
 *			code for the bottom of the loop is generated.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 * Return codes:
 *	ia_successful:		The operation worked.
 *	ia_bufferExceeded:	opcode didn't fit in buffer.
 ************************************************************************/
Int32  CDECL 
ip_mmxCreateErodeShiftCode (
    IPState    *pIPState,
    Int32	polarity,
    Int32	prevCol,
    Int32	curCol,
    Int32	nextCol,
    Int32	centerCol,
    Int32	numRows,
    Int32	numCols,
    Int32	cacheBpl,
    UInt8      *colIDArray,
    UInt8      *polIDArray,
    Int32      *pCurDisp,
    Int32      *pTopOfLoopIndex,
    UInt8      *pCodeBuf,
    Int32      *pBufIndex,
    Int32	bufSize)
{
    PROLOGUE;
    Int32	 bufIndex = *pBufIndex;	/* needed for macros */
    Int32	 curDisp = *pCurDisp;	/* current displacement in colIDArray
					 * and polIDArray. */
    Int32	*pWordDisplacement;	/* pts to displacements */


	/* keep compiler happy */
    numCols = numCols;

	/* First, see if this is a vertical strel.
	 * We have to use this weird test since some weirdo
	 * might pass in a strel with lots of columns in which only
	 * the center column has hits.
	 */
    if ((curCol == centerCol) && (prevCol == -1) && (nextCol == -1))
    {
	    /* This is a vertical strel.  There is special code for
	     * this case.  It puts out the entire cache-filling loop
	     * in one step.
	     * Decide whether the central column has hits, misses
	     * or both.
	     */
	switch (polarity)
	{

	    /* VERTICAL STREL, HITS ONLY */
	case cHit:
		/* the vertical strel contains only hits.
		 * The sequence is:
		 **START: read upper dword
		 **	; empty
		 **
		 **	read lower dword
		 **	; empty
		 **
		 **	add 2 * source bytes/line to source address
		 **	increment input cache address by
		 **			2 * number of bytes/row of input cache
		 **
		 **	write upper dword to input cache - 2 * bytes/row
		 **	; empty
		 **
		 **	write lower dword to input cache - bytes/row
		 **	; empty
		 **
		 **	decrement number of rows left by 2 in input cache
		 **	branch back to START if number of rows left > 0.
		 **
		 */
	    *pTopOfLoopIndex = bufIndex;	/* remember where we start */
	    M_COPY_CODE(mphT3LoadUpperSrcDWord);
	    /* ; empty */

	    M_COPY_CODE(mphT3LoadLowerSrcDWord);
	    /* ; empty */

	    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
	    M_COPY_CODE(mphT3AdjustCachePtr);
		/* Set the existing 32-bit displacment to be 2 * cache Bpl. */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	    *pWordDisplacement = 2 * cacheBpl;

		/* Put out first write opcode */
	    M_COPY_DISP(writeUpperOrig, 0, (-2 * cacheBpl));
		/* ; empty */

	    M_COPY_DISP(writeLowerOrig, 0, (-cacheBpl));
		/* ; empty */

		/* Remember what was written here */
	    colIDArray[curDisp] = (UInt8) curCol;
	    polIDArray[curDisp] = cHit;
	    curDisp++;

	    M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
	        /* Set the jump's relative offset correctly */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);

	    break;	/* That's it for the vertical strel, hits-only case */


	    /* VERTICAL STREL, MISSES ONLY */
	case cMiss:
		/* the vertical strel contains only misses.
		 * The sequence is:
		 **START: read upper dword
		 **	(re)load mm6 with all 1's
		 **
		 **	read lower dword
		 **	invert upper dword using mm6
		 **
		 **	invert lower dword using mm6
		 **	; empty
		 **
		 **	add 2 * source bytes/line to source address
		 **	increment input cache address by
		 **			2 * number of bytes/row of input cache
		 **
		 **	write inverted upper dword to
		 **			input cache - 2 * bytes/row
		 **	; empty
		 **
		 **	write inverted lower dword to
		 **			input cache - bytes/row
		 **	; empty
		 **
		 **	decrement number of rows left by 2 in input cache
		 **	branch back to START if number of rows left > 0.
		 **
		 */
	    *pTopOfLoopIndex = bufIndex;	/* remember where we start */
	    M_COPY_CODE(mphT3LoadUpperSrcDWord);
	    M_COPY_CODE(mphT3PreLoad1sMask);

	    M_COPY_CODE(mphT3LoadLowerSrcDWord);
	    M_COPY_CODE(mphT3InvertUpperOrigDWord);

	    M_COPY_CODE(mphT3InvertLowerOrigDWord);
		/* ; empty */

	    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
	    M_COPY_CODE(mphT3AdjustCachePtr);
		/* Set the existing 32-bit displacment to be 2 * cache Bpl. */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	    *pWordDisplacement = 2 * cacheBpl;

		/* Put out first write opcode */
	    M_COPY_DISP(writeUpperOrig, 0, (-2 * cacheBpl));
		/* ; empty */

	    M_COPY_DISP(writeLowerOrig, 0, (-cacheBpl));
		/* ; empty */

		/* Remember what was written here */
	    colIDArray[curDisp] = (UInt8) curCol;
	    polIDArray[curDisp] = cMiss;
	    curDisp++;

	    M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
		/* Set the jump's relative offset correctly */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);

	    break;	/* That's it for the vertical strel, misses-only case */


	    /* VERTICAL STREL, HITS and MISSES */
	case (cHit | cMiss):
		/* the vertical strel contains both hits and misses.
		 * The sequence is:
		 **START: read upper dword
		 **	(re)load mm6 with all 1's
		 **
		 **	read lower dword
		 **	; empty	
		 **
		 **	write upper dword to input cache
		 **	invert upper dword using mm6
		 **
		 **	write lower dword to input cache + bytes/row
		 **	invert lower dword using mm6
		 **
		 **	add 2 * source bytes/line to source address
		 **	increment input cache address by
		 **			2 * number of bytes/row of input cache
		 **
		 **	write inverted upper dword to
		 **			input cache - 2 * bytes/row
		 **	; empty
		 **
		 **	write inverted lower dword to
		 **			input cache - bytes/row
		 **	; empty
		 **
		 **	decrement number of rows left by 2 in input cache
		 **	branch back to START if number of rows left > 0.
		 */
	    *pTopOfLoopIndex = bufIndex;	/* remember where we start */
	    M_COPY_CODE(mphT3LoadUpperSrcDWord);
	    M_COPY_CODE(mphT3PreLoad1sMask);

	    M_COPY_CODE(mphT3LoadLowerSrcDWord);
		/* ; empty */

	    M_COPY_DISP(writeUpperOrig, 0, 0);
	    M_COPY_CODE(mphT3InvertUpperOrigDWord);

	    M_COPY_DISP(writeLowerOrig, 0, (cacheBpl));
	    M_COPY_CODE(mphT3InvertLowerOrigDWord);

		/* Remember what was written here */
	    colIDArray[curDisp] = (UInt8) curCol;
	    polIDArray[curDisp] = cHit;
	    curDisp++;

	    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
	    M_COPY_CODE(mphT3AdjustCachePtr);
		/* Set the existing 32-bit displacment to be 2 * cache Bpl. */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	    *pWordDisplacement = 2 * cacheBpl;

		/* Put out code to write inverted inputs */
	    M_COPY_DISP(writeUpperOrig, 0, (-2 * cacheBpl + 8));
		/* ; empty */

	    M_COPY_DISP(writeLowerOrig, 0, (-cacheBpl + 8));
		/* ; empty */

		/* Remember what was written here */
	    colIDArray[curDisp] = (UInt8) curCol;
	    polIDArray[curDisp] = cMiss;
	    curDisp++;

	    M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
		/* Set the jump's relative offset correctly */
	    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
	    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);

	    break;	/* That's it for the vertical strel, hits + misses */

	default:
	    __error__ = ia_internal;	/* shouldn't happen */
	    goto genCleanup;
	}	/* end of switch for vertical strels */
    }	/* end of vertical strel case */
    else	/* this is not a vertical strel */
    {
	    /* There are several factors that determine what kind of code
	     * will be generated:
	     *  Column Position (left of center, in center col, right of center)
	     *  Is First Column?
	     *  Polarities involved (cHit, cMiss, both)
	     *  Is Last Column?
	     *
	     * Since we dealt with purely vertical strels above, we
	     * know that no column we see here will be both the first
	     * and last column.
	     *
	     * If this is the first column, we need to generate some
	     * preamble code.  Then there's always some middle code to
	     * process the column.  If this is the last column, there's
	     * some code at the end of finish writing the last words
	     * and loop.  We'll deal with the three cases in this order.
	     *
	     * Note that the preamble and finish-up code are basically
	     * independent of the column position (left/center/right).
	     * The preamble only varies with whether the right or
	     * left sideword is loaded.  This makes things fairly
	     * straightforward.
	     *
	     * First, test to see if we need to insert preamble.
	     * The preamble code is:
	     *
	     **START: load upper dword
	     **	load 1's mask
	     **
	     **	load left/right upper side word
	     **	step 1 of upper wordswap
	     **
	     **	load lower dword
	     **	step 2 of upper wordswap
	     **
	     **	load left/right lower side word
	     **	step 1 of lower wordswap
	     **
	     **	shift upper sideword left/right
	     **	step 2 of lower wordswap
	     **
	     **	copy upper dword to upper scratch
	     **	shift lower sideword left/right
	     **
	     **	shift upper dword in scratch reg right/left
	     **	copy lower dword to lower scratch
	     **
	     */
	if (prevCol == -1)	/* preamble needed */
	{
	    *pTopOfLoopIndex = bufIndex;	/* remember where we start */
	    M_COPY_CODE(mphT3LoadUpperSrcDWord);
	    M_COPY_CODE(mphT3PreLoad1sMask);

		/* If the first off-center column is to the left of center, load
		 * the upper left sideword.  Otherwise, load the upper right
		 * sideword.
		 */
	    if ((curCol < centerCol) ||
		((curCol == centerCol) && (nextCol < centerCol)))
	    {
		M_COPY_CODE(mphT3PreLoadUpperLeftSideWord);
	    }
	    else
	    {
		M_COPY_CODE(mphT3PreLoadUpperRightSideWord);
	    }
	    M_COPY_CODE(mphT3WordSwapUpperStep1);

	    M_COPY_CODE(mphT3LoadLowerSrcDWord);
	    M_COPY_CODE(mphT3WordSwapUpperStep2);


		/* Same deal for loading lower sideword */
	    if ((curCol < centerCol) ||
		((curCol == centerCol) && (nextCol < centerCol)))
	    {
		M_COPY_CODE(mphT3PreLoadLowerLeftSideWord);
	    }
	    else
	    {
		M_COPY_CODE(mphT3PreLoadLowerRightSideWord);
	    }
	    M_COPY_CODE(mphT3WordSwapLowerStep1);

		/* The amount and direction we shift the sideword depends on
		 * whether we're to the left or right of center.
		 * If we're to the left of center, we want to shift
		 * the left sideword left by 32 + (32 - (centerCol - curCol)).
		 * That's because we've just loaded the sideword
		 * into the lower 32 bits of a reg and we want
		 * to start with (centerCol - curCol) bits
		 * at the top (left) end of the sideword.
		 * If we're in the center column or on the right side,
		 * then we just read 32 bits from the word on the
		 * right into a reg.  We want (curCol - centerCol) bits
		 * to be left there, so we shift it right by
		 * 32 - (curCol - centerCol).
		 */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftUpperSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 - 
						  (centerCol - curCol));
	    }
	    else	/* see if we start with left or right shift
			 * for nextCol */
	    if ((curCol == centerCol) && (nextCol < centerCol))
	    {
		M_COPY_CODE(mphT3LeftShiftUpperSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 -
						  (centerCol - nextCol));
	    }
	    else	/* start with right shift */
	    if ((curCol == centerCol) && (nextCol > centerCol))
	    {
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (curCol - centerCol));
	    }
	    M_COPY_CODE(mphT3WordSwapLowerStep2);


	    M_COPY_CODE(mphT3CopyUpperDWord);
		/* Insert the left/right shift for the lower word */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3LeftShiftLowerSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 - 
						  (centerCol - curCol));
	    }
	    else	/* see if we start with left or right shift
			 * for nextCol */
	    if ((curCol == centerCol) && (nextCol < centerCol))
	    {
		M_COPY_CODE(mphT3LeftShiftLowerSideWord);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 + 32 -
						  (centerCol - nextCol));
	    }
	    else	/* start with right shift */
	    if ((curCol == centerCol) && (nextCol > centerCol))
	    {
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (curCol - centerCol));
	    }


		/* Now we get ready to shift the dwords.  If we
		 * are to the left of center, we want to shift the
		 * dwords right to bring them to the center.  If
		 * we're to the right of center (or will be soon),
		 * we want to shift the dwords left.
		 */
	    if (curCol < centerCol)
	    {
		M_COPY_CODE(mphT3RightShiftUpperScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
	    }
	    else	/* see if we start with left or right shift
			 * for nextCol */
	    if ((curCol == centerCol) && (nextCol < centerCol))
	    {
		M_COPY_CODE(mphT3RightShiftUpperScratch);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - nextCol);
	    }
	    else	/* start with left shift */
	    if ((curCol == centerCol) && (nextCol > centerCol))
	    {
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);
	    }
	    else	/* curCol > centerCol */
	    {
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
	    }
	    M_COPY_CODE(mphT3CopyLowerDWord);

	}	/* end of case where preamble is needed */

	    /* Now we decide which kind of column block to generate.  Here's
	     * the basic search order.  Note that purely horizontal strels
	     * can't have both hits and misses in the same column since
	     * there's only one entry per column (only 1 row in the strel).
	     ** Preload (loads registers and starts the shifts)
	     ** See if col is center col, if so generate code for it.
	     **    HITs only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    MISSes only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    Both HITs and MISSes
	     ** See if col is last column to be processed (will be to left or
	     ** right of center).  If so, generate code for it.
	     **    HITs only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    MISSes only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    Both HITs and MISSes
	     ** See if EITHER the col is to the left of center and is
	     ** followed by another column that's also to the left of
	     ** center OR the colis to the right of center and is
	     **  followed by another col that's also to the right of
	     ** center.  If so, generate code for it.  We think this
	     ** is the common case.
	     **    HITs only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    MISSes only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    Both HITs and MISSes
	     ** If none of the cases above is true, this must be the last left
	     ** col to be processed.  It must be followed by a right col.
	     ** Generate code for this case.
	     **    HITs only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    MISSes only
	     **		non-horizontal strels
	     **		pure horizontal strels
	     **    Both HITs and MISSes
	     */
	if (curCol == centerCol)
	{
	    switch (polarity)
	    {
		/* CENTER COLUMN, HITS ONLY */
	    case cHit:
	        if (numRows > 1)	/* not a purely horizontal strel */
		{
			/*
			 **LOOP:
			 **	write unshifted upper dword to input cache
			 **	; empty
			 **
			 **	write unshifted lower dword to
			 **			input cache + bytes/row
			 **	; empty
			 **
			 */
			/* Put out code to write inverted inputs */
		    M_COPY_DISP(writeUpperOrig, 0, 0);
			/* ; empty */
	
		    M_COPY_DISP(writeLowerOrig, 0, +cacheBpl);
			/* ; empty */
    
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cHit;
		    curDisp++;
		}
		else	/* is purely horizontal strel */
		{
			/*
			 **LOOP:
			 **	copy unshifted upper dword to mm6 temp
			 **	copy unshifted lower dword to mm7 temp
			 **
			 */
		    M_COPY_CODE(mphT3LoadUpperOrigDWord);
		    M_COPY_CODE(mphT3LoadLowerOrigDWord);
		}

		break;

		/* CENTER COLUMN, MISSES ONLY */
	    case cMiss:
	        if (numRows > 1)	/* not a purely horizontal strel */
		{
			/*
			 **	invert upper unshifted dword
			 **	invert lower unshifted dword
			 **
			 **	write inverted, unshifted upper dword to
			 **				input cache
			 **	re-invert upper unshifted dword
			 **
			 **	write inverted, unshifted lower dword to
			 **				input cache + bytes/row
			 **	re-invert lower unshifted dword
			 **
			 */
		    M_COPY_CODE(mphT3InvertUpperOrigDWord);
		    M_COPY_CODE(mphT3InvertLowerOrigDWord);
    
		    M_COPY_DISP(writeUpperOrig, 0, 0);
		    M_COPY_CODE(mphT3InvertUpperOrigDWord);
	
		    M_COPY_DISP(writeLowerOrig, 0, (cacheBpl));
		    M_COPY_CODE(mphT3InvertLowerOrigDWord);
    
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cMiss;
		    curDisp++;
		}
		else	/* is purely horizontal strel */
		{
			/*
			 **LOOP:
			 **	Invert upper unshifted dword, leaving
			 **					result in mm6.
			 **	load mm7 with all 1's.
			 **
			 **	invert lower unshifted dword, leaving
			 **					result in mm7.
			 **	; empty
			 **
			 */
		    M_COPY_CODE(mphT3InvertUpperOrigDWordToMM6);
		    M_COPY_CODE(mphT3PreLoad1sMaskToMM7);

		    M_COPY_CODE(mphT3InvertLowerOrigDWordToMM7);
		    /* ; empty */
		}

		break;


		/* CENTER COLUMN, HITS AND MISSES */
	    case (cHit | cMiss):
		    /*
		     **	write positive unshifted upper dword to
		     **				input cache
		     **	invert upper unshifted dword
		     **
		     **	write positive unshifted lower dword to
		     **				input cache + bytes/row
		     **	invert lower unshifted dword
		     **
		     **	write inverted, unshifted upper dword to
		     **				input cache + 8
		     **	re-invert upper unshifted dword
		     **
		     **	write inverted, unshifted lower dword to
		     **				input cache + 8 + bytes/row
		     **	re-invert lower unshifted dword
		     */
		M_COPY_DISP(writeUpperOrig, 0, 0);
		M_COPY_CODE(mphT3InvertUpperOrigDWord);
    
		M_COPY_DISP(writeLowerOrig, 0, (cacheBpl));
		M_COPY_CODE(mphT3InvertLowerOrigDWord);

		M_COPY_DISP(writeUpperOrig, 0, 8);
		M_COPY_CODE(mphT3InvertUpperOrigDWord);
    
		M_COPY_DISP(writeLowerOrig, 0, (cacheBpl + 8));
		M_COPY_CODE(mphT3InvertLowerOrigDWord);

		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cHit;
		curDisp++;
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cMiss;
		curDisp++;

		break;

	    default:
		__error__ = ia_internal;	/* shouldn't happen */
		goto genCleanup;
	    }	/* end of switch for center column */
	}	/* end of center column case */

	else
	    /* see if this is the last column we'll process in the row */
	if (nextCol == -1)
	{
	    switch (polarity)
	    {
		/* LAST COLUMN, HITS ONLY */
	    case cHit:
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	increment inputCache address by
		     **			2 * number of input cache bytes/row 
		     **	combine (OR) lower dword and sideword
		     **
		     ** ; wait for AGI
		     **	add 2 * source bytes/line to source address
		     **	; empty
		     **
		     **	write last shifted upper dword to
		     **			inputCache - 2 * bytes/row. 
		     **	; empty
		     **
		     **	write last shifted lower dword to
		     **			inputCache - bytes/row
		     **	; empty
		     **
		     **	decrement number of rows left by 2 in inputCache
		     **	branch back to START if number of rows left > 0.
		     **
		     ** If this is a purely horizontal strel, the code becomes:
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	AND upper shifted dword with mm6 temp, result to mm2
		     **	combine (OR) lower dword and sideword
		     **
		     **	wordswap step 1 for mm2, mm6 as scratch
		     **	AND lower shifted dword with mm7 temp, result to mm5
		     **
		     **	add 2 * source bytes/line to source address
		     **	wordswap step 2 for mm2, mm6 as scratch
		     **
		     **	wordswap step 1 for mm5, mm7 as scratch
		     **	decrement number of rows left by 2
		     **
		     **	write mm2 to dstAddress		; note 1 cycle delay
		     **	wordswap step 2 for mm5, mm7 as scratch
		     **
		     **	; empty
		     **	; empty
		     **
		     **	write mm5 to dstAddress + dest bytes/row
		     **	; empty
		     **
		     **	add 2 * dest bytes/line to dest address using
		     **					lea (no cc's affected)
		     **	branch back to START if number of rows left > 0.
		     **
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		if (numRows > 1 )
		{
		    M_COPY_CODE(mphT3AdjustCachePtr);
			/* Set the existing 32-bit displacment
			 * to be 2*cacheBpl. */
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		    *pWordDisplacement = 2 * cacheBpl;
		}
		else	/* purely horizontal strel */
		{
		    M_COPY_CODE(mphT3AndUpperShiftedDWordToMM2);
		}
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);


		if (numRows > 1 )
		{
		    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		    /* ; empty */

		    M_COPY_DISP(writeUpperScratch, 0, 8*curDisp - 2*cacheBpl);
		    /* ; empty */

		    M_COPY_DISP(writeLowerScratch, 0, 8*curDisp - cacheBpl);
		    /* ; empty */
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cHit;
		    curDisp++;

		    M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
			/* Set the jump's relative offset correctly */
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);
		}
		else	/* purely horizontal strel */
		{
		    M_COPY_CODE(mphT3WordSwapUpperMM2Step1);
		    M_COPY_CODE(mphT3AndLowerShiftedDWordToMM5);

		    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		    M_COPY_CODE(mphT3WordSwapUpperMM2Step2);

		    M_COPY_CODE(mphT3WordSwapLowerMM5Step1);
		    M_COPY_CODE(mphT3DecrRowsBy2);

		    M_COPY_CODE(mphT3WriteUpperScratchToDest);
		    M_COPY_CODE(mphT3WordSwapLowerMM5Step2);

		    /* ; empty */
		    /* ; empty */

		    M_COPY_CODE(mphT3WriteLowerScratchToDest);
		    /* ; empty */

		    M_COPY_CODE(mphT3AdjustDstAddrBy2Lines);
		    M_COPY_CODE(mphT3JABranch);
			/* Set the jump's relative offset correctly */
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);
		}

		break;

		/* LAST COLUMN, MISSES ONLY */
	    case cMiss:
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	invert (pxor) upper dword in scratch reg
		     **	combine (OR) lower dword and sideword
		     **
		     **	increment inputCache address by
		     **			2 * number of input cache bytes/row
		     **	invert (pxor) lower dword in scratch reg
		     **
		     ** ; wait for AGI
		     **	add 2 * source bytes/line to source address
		     **	; empty
		     **
		     **	write last inverted, shifted upper dword to
		     **				inputCache - 2 * bytes/row. 
		     **	; empty
		     **
		     **	write last inverted, shifted lower dword to
		     **				inputCache - bytes/row
		     **	; empty
		     **
		     **	decrement number of rows left by 2 in inputCache
		     **	branch back to START if number of rows left > 0.
		     **
		     ** If this is a purely horizontal strel, the code becomes:
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	invert (pandn) upper dword in scratch reg,
		     **				result to scratch reg (mm2)
		     **	combine (OR) lower dword and sideword
		     **
		     **	wordswap step 1 for mm2, mm6 as scratch
		     **	invert (pandn) lower dword in scratch reg,
		     **				result to scratch reg (mm5)
		     **
		     **	add 2 * source bytes/line to source address
		     **	wordswap step 2 for mm2, mm6 as scratch
		     **
		     **	wordswap step 1 for mm5, mm7 as scratch
		     **	decrement number of rows left by 2
		     **
		     **	write mm2 to dstAddress		; note 1 cycle delay
		     **	wordswap step 2 for mm5, mm7 as scratch
		     **
		     **	; empty
		     **	; empty
		     **
		     **	write mm5 to dstAddress + dest bytes/row
		     **	; empty
		     **
		     **	decrement number of rows left by 2 in inputCache
		     **	branch back to START if number of rows left > 0.
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		if (numRows > 1)
		{
		    M_COPY_CODE(mphT3InvertUpperShiftedDWord);
		}
		else	/* purely horizontal strel */
		{
		    M_COPY_CODE(mphT3InvertAndUpperShiftedDWord);
		}
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		if (numRows > 1)
		{
		    M_COPY_CODE(mphT3AdjustCachePtr);
			/* Set the 32-bit displacment to be 2*cacheBpl. */
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		    *pWordDisplacement = 2 * cacheBpl;
		    M_COPY_CODE(mphT3InvertLowerShiftedDWord);
    
		    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		    /* ; empty */
    
		    M_COPY_DISP(writeUpperScratch, 0, 8*curDisp - 2*cacheBpl);
		    /* ; empty */
    
		    M_COPY_DISP(writeLowerScratch, 0, 8*curDisp - cacheBpl);
		    /* ; empty */
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cMiss;
		    curDisp++;
    
		    M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
			/* Set the jump's relative offset correctly */
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);
		}
		else	/* purely horizontal strel */
		{
		    M_COPY_CODE(mphT3WordSwapUpperMM2Step1);
		    M_COPY_CODE(mphT3InvertAndLowerShiftedDWord);

		    M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		    M_COPY_CODE(mphT3WordSwapUpperMM2Step2);

		    M_COPY_CODE(mphT3WordSwapLowerMM5Step1);
		    M_COPY_CODE(mphT3DecrRowsBy2);

		    M_COPY_CODE(mphT3WriteUpperScratchToDest);
		    M_COPY_CODE(mphT3WordSwapLowerMM5Step2);

		    /* ; empty */
		    /* ; empty */

		    M_COPY_CODE(mphT3WriteLowerScratchToDest);
		    /* ; empty */

		    M_COPY_CODE(mphT3AdjustDstAddrBy2Lines);
		    M_COPY_CODE(mphT3JABranch);
			/* Set the jump's relative offset correctly */
		    pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		    *pWordDisplacement = (*pTopOfLoopIndex - bufIndex);
		}

		break;


		/* LAST COLUMN, HITS AND MISSES */
	    case (cHit | cMiss):
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	increment inputCache address by
		     **			2 * number of input cache bytes/row
		     **	combine (OR) lower dword and sideword
		     **
		     ** ; wait for AGI
		     **	add 2 * source bytes/line to source address
		     **	; empty
		     **
		     **	write last shifted upper dword to
		     **				inputCache - 2 * bytes/row. 
		     **	invert (pxor) upper shifted dword
		     **
		     **	write last shifted lower dword to
		     **				inputCache - bytes/row
		     **	invert (pxor) lower shifted dword
		     **
		     **	write last inverted, shifted upper dword to
		     **				inputCache+8-2*(bytes/row). 
		     **	; empty
		     **
		     **	write last inverted, shifted lower dword to
		     **				inputCache+8-(bytes/row)
		     **	; empty
		     **
		     **	decrement number of rows left by 2 in inputCache
		     **	branch back to START if number of rows left > 0.
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		M_COPY_CODE(mphT3AdjustCachePtr);
		    /* Set the existing 32-bit displacment to be 2*cacheBpl. */
		pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		*pWordDisplacement = 2 * cacheBpl;
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		M_COPY_CODE(mphT3AdjustSrcAddrBy2Lines);
		/* ; empty */

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp - 2*cacheBpl);
		M_COPY_CODE(mphT3InvertUpperShiftedDWord);

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp - cacheBpl);
		M_COPY_CODE(mphT3InvertLowerShiftedDWord);
		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cHit;
		curDisp++;

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp - 2*cacheBpl);
		/* ; empty */

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp - cacheBpl);
		/* ; empty */
		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cMiss;
		curDisp++;

		M_COPY_CODE(mphT3DecrRowsBy2AndBranch);
		    /* Set the jump's relative offset correctly */
		pWordDisplacement = (Int32 *) &(pCodeBuf[(bufIndex)-4]);
		*pWordDisplacement = (*pTopOfLoopIndex - bufIndex);

		break;

	    default:
		__error__ = ia_internal;	/* shouldn't happen */
		goto genCleanup;
	    }	/* end of switch for last column */
	}	/* end of last column case */

	else
	    /* see if this is a left col followed by a left col or a
	     * right col followed by a right col.
	     */
	if ( ((curCol < centerCol) && (nextCol < centerCol)) ||
	     ((curCol > centerCol) && (nextCol > centerCol)) )
	{
	    switch (polarity)
	    {
	    case cHit:
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	shift upper sideword left/right
		     **	combine (OR) lower dword and sideword
		     **
		     ** ; note 1 cycle delay from OR
		     **	write upper dword to input cache
		     **	shift lower sideword left/right
		     **
		     **	write lower dword to input cache + bytes/row
		     **	copy upper dword to upper scratch
		     **
		     **	shift upper dword in scratch reg right/left
		     **	copy lower dword to lower scratch
		     **
		     ** If this is a purely horizontal strel, the code becomes:
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	shift upper sideword left/right
		     **	combine (OR) lower dword and sideword
		     **
		     **	pand upper dword to mm6
		     **	shift lower sideword left/right
		     **
		     **	pand lower dword to mm7
		     **	copy upper dword to upper scratch
		     **
		     **	shift upper dword in scratch reg right/left
		     **	copy lower dword to lower scratch
		     **
		     */

		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		    /* Note that this shift is for the next time
		     * through the loop, so we shift by the difference
		     * between nextCol and curCol.
		     */
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3LeftShiftUpperSideWord);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3RightShiftUpperSideWord);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
		}
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		if (numRows > 1)
		{
		    M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		}
		else
		{
			/* This works because we loaded mm6 with
			 * 1's in the prologue.
			 */
		    M_COPY_CODE(mphT3AndUpperShiftedDWord);
		}
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3LeftShiftLowerSideWord);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3RightShiftLowerSideWord);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
		}

		if (numRows > 1)
		{
		    M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cHit;
		    curDisp++;
		}
		else	/* purely horizontal strel */
		{
		    if (prevCol != -1)
		    {
			M_COPY_CODE(mphT3AndLowerShiftedDWord);
		    }
		    else
		    {
			M_COPY_CODE(mphT3LoadLowerShiftedDWord);
		    }
		}
		M_COPY_CODE(mphT3CopyUpperDWord);


		if (nextCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftUpperScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - nextCol);
		}
		else	/* nextCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftUpperScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);
		}
		M_COPY_CODE(mphT3CopyLowerDWord);

		break;

		/* CONTINUATION COLUMN, MISSES ONLY */
	    case cMiss:
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	invert (pxor) upper dword in scratch reg
		     **	combine (OR) lower dword and sideword
		     **
		     **	shift upper sideword left/right
		     **	invert (pxor) lower dword in scratch reg
		     **
		     ** ; note 1 cycle delay from OR
		     **	write upper dword to input cache
		     **	shift lower sideword left/right
		     **
		     **	write lower dword to input cache + bytes/row
		     **	copy upper dword to upper scratch
		     **
		     **	shift upper dword in scratch reg right/left
		     **	copy lower dword to lower scratch
		     **
		     **
		     ** If this is a purely horizontal strel, the code becomes:
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	invert (pandn) upper dword in scratch reg, result in mm2
		     **	combine (OR) lower dword and sideword
		     **
		     **	shift upper sideword left/right
		     **	invert (pandn) lower dword in scratch reg, result in mm5
		     **
		     **	copy upper dword in mm2 back to mm6
		     **	shift lower sideword left/right
		     **
		     **	copy lower dword in mm5 back to mm7
		     **	copy upper dword to upper scratch
		     **
		     **	shift upper dword in scratch reg right/left
		     **	copy lower dword to lower scratch
		     **
		     */

		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		if (numRows > 1)
		{
		    M_COPY_CODE(mphT3InvertUpperShiftedDWord);
		}
		else
		{
			/* This only works because we already
			 * loaded mm6 with 1's in the prologue.
			 */ 
		    M_COPY_CODE(mphT3InvertAndUpperShiftedDWord);
		}
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		    /* Note that this shift is for the next time
		     * through the loop, so we shift by the difference
		     * between nextCol and curCol.
		     */
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3LeftShiftUpperSideWord);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3RightShiftUpperSideWord);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
		}
		if (numRows > 1)
		{
		    M_COPY_CODE(mphT3InvertLowerShiftedDWord);
		}
		else
		{
		    if (prevCol == -1)
		    {
			    /* This is ugly, but I haven't found a
			     * better way.  We spend an extra cycle
			     * loading mm7 with 1's, then do the
			     * pandn.  We need to do the pandn to
			     * invert the shifted word.
			     */
			M_COPY_CODE(mphT3PreLoad1sMaskToMM7);
		    }
		    M_COPY_CODE(mphT3InvertAndLowerShiftedDWord);
		}

		if (numRows > 1)
		{
		    M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		}
		else
		{
		    M_COPY_CODE(mphT3LoadUpperShiftedDWord);
		}
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3LeftShiftLowerSideWord);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3RightShiftLowerSideWord);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
		}

		if (numRows > 1)
		{
		    M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cMiss;
		    curDisp++;
		}
		else
		    M_COPY_CODE(mphT3LoadLowerShiftedDWord);
		M_COPY_CODE(mphT3CopyUpperDWord);


		if (nextCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftUpperScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - nextCol);
		}
		else	/* nextCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftUpperScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);
		}
		M_COPY_CODE(mphT3CopyLowerDWord);

		break;

		/* CONTINUATION COLUMN, HITS AND MISSES */
	    case (cHit | cMiss):
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	shift upper sideword left/right
		     **	combine (OR) lower dword and sideword
		     **
		     ** ; note 1 cycle delay from OR
		     **	write upper dword to input cache
		     **	invert (pxor) upper dword in scratch reg
		     **
		     **	write lower dword to input cache
		     **	invert (pxor) lower dword in scratch reg
		     **
		     **	write inverted upper dword to
		     **				input cache + 8
		     **	shift lower sideword left/right
		     **
		     **	write inverted lower dword to
		     **				input cache + 8 + bytes/row
		     **	copy upper dword to upper scratch
		     **
		     **	shift upper dword in scratch reg right/left
		     **	copy lower dword to lower scratch
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		    /* Note that this shift is for the next time
		     * through the loop, so we shift by the difference
		     * between nextCol and curCol.
		     */
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3LeftShiftUpperSideWord);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3RightShiftUpperSideWord);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
		}
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		M_COPY_CODE(mphT3InvertUpperShiftedDWord);

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
		M_COPY_CODE(mphT3InvertLowerShiftedDWord);

		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cHit;
		curDisp++;

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3LeftShiftLowerSideWord);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3RightShiftLowerSideWord);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - nextCol);
		}

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
		M_COPY_CODE(mphT3CopyUpperDWord);

		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cMiss;
		curDisp++;

		if (nextCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftUpperScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - nextCol);
		}
		else	/* nextCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftUpperScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);
		}
		M_COPY_CODE(mphT3CopyLowerDWord);

		break;

	    default:
		__error__ = ia_internal;	/* shouldn't happen */
		goto genCleanup;
	    }	/* End of switch for continuation column */
	}	/* End of continuation column case */

	else	/* Well, it's not the center column, it's not the last
		 * column, it's not a left followed by left or right
		 * followed by right; so it must be a left followed by
		 * right.
		 */
	{
	    switch (polarity)
	    {
		/* LAST LEFT COLUMN, HITS ONLY */
	    case cHit:
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	read right upper sideword (32 bits).
		     **	combine (OR) lower dword and sideword
		     **
		     **	write last shifted upper dword to input cache
		     **	copy upper dword to upper scratch
		     **
		     **	write last shifted lower dword to
		     **				input cache + bytes/row
		     **	shift upper dword in scratch reg left
		     **
		     **	load right lower sideword (32 bits).
		     **	shift upper sideword right
		     **
		     **	copy lower dword to lower scratch
		     **	shift lower sideword right
		     **
		     **
		     ** If this is a purely horizontal strel, the code becomes:
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	read right upper sideword (32 bits).
		     **	combine (OR) lower dword and sideword
		     **
		     **	pand upper dword, result to mm6
		     **	copy upper dword to upper scratch
		     **
		     **	pand lower dword, result to mm7
		     **	shift upper dword in scratch reg left
		     **
		     **	load right lower sideword (32 bits).
		     **	shift upper sideword right
		     **
		     **	copy lower dword to lower scratch
		     **	shift lower sideword right
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		M_COPY_CODE(mphT3PreLoadUpperRightSideWord);
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		if (numRows > 1)
		{
		    M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		}
		else	/* purely horizontal strel */
		{
			/* This works because we loaded mm6 with
			 * 1's in the prologue.
			 */
		    M_COPY_CODE(mphT3AndUpperShiftedDWord);
		}
		M_COPY_CODE(mphT3CopyUpperDWord);

		if (numRows > 1)
		{
		    M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cHit;
		    curDisp++;
		}
		else
		{
		    if (prevCol != -1)
		    {
			M_COPY_CODE(mphT3AndLowerShiftedDWord);
		    }
		    else
		    {
			M_COPY_CODE(mphT3LoadLowerShiftedDWord);
		    }
		}
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);

		M_COPY_CODE(mphT3PreLoadLowerRightSideWord);
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));

		M_COPY_CODE(mphT3CopyLowerDWord);
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));

		break;


		/* LAST LEFT COLUMN, MISSES ONLY */
	    case cMiss:
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	invert (pxor) upper dword in scratch reg
		     **	combine (OR) lower dword and sideword
		     **
		     **	read right upper sideword (32 bits).
		     **	invert (pxor) lower dword in scratch reg
		     **
		     **	write last inverted, shifted upper dword to input cache
		     **	copy upper dword to upper scratch
		     **
		     **	write last inverted, shifted lower dword to
		     **				input cache + bytes/row
		     **	shift upper dword in scratch reg left
		     **
		     **	load right lower sideword (32 bits).
		     **	shift upper sideword right
		     **
		     **	copy lower dword to lower scratch
		     **	shift lower sideword right
		     **
		     **
		     ** If this is a purely horizontal strel, the code becomes:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	invert (pandn) upper dword, result in mm2
		     **	combine (OR) lower dword and sideword
		     **
		     **	read right upper sideword (32 bits).
		     **	invert (pandn) lower dword, result in mm5
		     **
		     **	copy upper dword in mm2 back to mm6
		     **	copy upper dword to upper scratch
		     **
		     **	copy lower dword in mm5 back to mm7
		     **	shift upper dword in scratch reg left
		     **
		     **	load right lower sideword (32 bits).
		     **	shift upper sideword right
		     **
		     **	copy lower dword to lower scratch
		     **	shift lower sideword right
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		if (numRows > 1)
		{
		    M_COPY_CODE(mphT3InvertUpperShiftedDWord);
		}
		else	/* purely horizontal strel */
		{
			/* This only works because we already
			 * loaded mm6 with 1's in the prologue.
			 */ 
		    M_COPY_CODE(mphT3InvertAndUpperShiftedDWord);
		}
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		M_COPY_CODE(mphT3PreLoadUpperRightSideWord);
		if (numRows > 1)
		{
		    M_COPY_CODE(mphT3InvertLowerShiftedDWord);
		}
		else
		{
		    if (prevCol == -1)
		    {
			    /* This is ugly, but I haven't found a
			     * better way.  We spend an extra cycle
			     * loading mm7 with 1's, then do the
			     * pandn.  We need to do the pandn to
			     * invert the shifted word.
			     */
			M_COPY_CODE(mphT3PreLoad1sMaskToMM7);
		    }
		    M_COPY_CODE(mphT3InvertAndLowerShiftedDWord);
		}

		if (numRows > 1)
		{
		    M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		}
		else
		{
		    M_COPY_CODE(mphT3LoadUpperShiftedDWord);
		}
		M_COPY_CODE(mphT3CopyUpperDWord);

		if (numRows > 1)
		{
		    M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
			/* Remember what was written here */
		    colIDArray[curDisp] = (UInt8) curCol;
		    polIDArray[curDisp] = cMiss;
		    curDisp++;
		}
		else
		    M_COPY_CODE(mphT3LoadLowerShiftedDWord);
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);


		M_COPY_CODE(mphT3PreLoadLowerRightSideWord);
		M_COPY_CODE(mphT3RightShiftUpperSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));

		M_COPY_CODE(mphT3CopyLowerDWord);
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));

		break;

		/* LAST LEFT COLUMN, HITS AND MISSES */
	    case (cHit | cMiss):
		    /*
		     **LOOP:
		     **	combine (OR) upper dword and sideword
		     **	shift lower dword in scratch reg right/left
		     **
		     **	read right upper sideword (32 bits).
		     **	combine (OR) lower dword and sideword
		     **
		     **	write last shifted upper dword to input cache
		     **	invert (pxor) upper dword in scratch reg
		     **
		     **	write last shifted lower dword to
		     **				input cache + bytes/row
		     **	invert (pxor) lower dword in scratch reg
		     **
		     **	write last shifted, inverted upper dword to input cache
		     **	copy upper dword to upper scratch
		     **
		     **	write last shifted, inverted lower dword to
		     **				input cache + bytes/row
		     **	shift upper dword in scratch reg left
		     **
		     **	load right lower sideword (32 bits).
		     **	shift upper sideword right
		     **
		     **	copy lower dword to lower scratch
		     **	shift lower sideword right
		     **
		     */
		M_COPY_CODE(mphT3CombineShiftedUpperDWord);
		if (curCol < centerCol)
		{
		    M_COPY_CODE(mphT3RightShiftLowerScratch);
			/* Set the right shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (centerCol - curCol);
		}
		else	/* curCol > centerCol */
		{
		    M_COPY_CODE(mphT3LeftShiftLowerScratch);
			/* Set the left shift distance */
		    pCodeBuf[bufIndex - 1] = (UInt8) (curCol - centerCol);
		}

		M_COPY_CODE(mphT3PreLoadUpperRightSideWord);
		M_COPY_CODE(mphT3CombineShiftedLowerDWord);

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		M_COPY_CODE(mphT3InvertUpperShiftedDWord);

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
		M_COPY_CODE(mphT3InvertLowerShiftedDWord);

		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cHit;
		curDisp++;

		M_COPY_DISP(writeUpperScratch, 0, 8*curDisp);
		M_COPY_CODE(mphT3CopyUpperDWord);

		M_COPY_DISP(writeLowerScratch, 0, 8*curDisp + cacheBpl);
		M_COPY_CODE(mphT3LeftShiftUpperScratch);
		    /* Set the left shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (nextCol - centerCol);

		    /* Remember what was written here */
		colIDArray[curDisp] = (UInt8) curCol;
		polIDArray[curDisp] = cMiss;
		curDisp++;

		M_COPY_CODE(mphT3PreLoadLowerRightSideWord);
		M_COPY_CODE(mphT3RightShiftUpperSideWord);

		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));

		M_COPY_CODE(mphT3CopyLowerDWord);
		M_COPY_CODE(mphT3RightShiftLowerSideWord);
		    /* Set the right shift distance */
		pCodeBuf[bufIndex - 1] = (UInt8) (32 - (nextCol - centerCol));

		break;

	    default:
		__error__ = ia_internal;	/* shouldn't happen */
		goto genCleanup;
	    }	/* end of last left switch statement */
	}	/* end of last left case */
    }	/* end of non-vertical strel case */


genCleanup:
    if (__error__ == ia_successful)
    {
	*pCurDisp = curDisp;
	*pBufIndex = bufIndex;
    }
    NORMAL_RETURN;
}


/* PRIVATE */
/************************************************************************
 *  ip_mmxCreateANDCode():
 *			Create a block of MMX code that combines two shifted
 *			dwords from the input cache with the numOutputRows
 *			output registers.
 *			The two shifted dwords are vertically adjacent in the
 *			input cache.  Therefore, they share the same
 *			SE column and polarity.
 *			
 *			When this routine is called, the code used to
 *			update the input cache pointer so that rCache plus
 *			topInputDisp points to the upper cached dword.
 *			This routine does not generate code to update
 *			rCache.
 *
 *	nx:		Number of columns in Structure Element (SE)
 *	ny:		Number of rows in Structure Element.
 *	cy:		Number of the SE's "center" row.
 *	pSE:		Pointer to byte array containing Structure Elements.
 *	topInputRow:	Each cache dword is formed from a dword (and usually
 *			a sideword) from one row of the input image. This
 *			parameter is equal to that input image row number.
 *			It is relative to the number of the top output row.
 *	topInputDisp:	The displacement from the "current" (in the sense
 *			of "the value that's the result of the code
 *			that's been generated so far") value of rCache to
 *			the upper cached input word.
 *	inputSECol:	Number of SE column for both input cache dwords.
 *	inputSEPolarity: Indicates polarity of both input dwords.  May equal
 *			cHit or cMiss.  Both input words have same polarity.
 *	cacheBpl:	Input cache bytes/line.
 *	numOutputRows:	The number of output rows.
 *	pOutInitArray:	Pointer to an array of Booleans that records whether
 *			each of the output words has been initialized.  Used
 *			to convert the first operation touching an output
 *			word from an AND to a load.  This routine updates
 *			this array as each output word is initialized.  The
 *			calling routine should set all components of this array,
 *			there are nunOutputRows of them, to cFalse before
 *			calling this routine for the first time for a given
 *			output subblock.
 *	pCodeBuf:	Pointer to beginning of buffer containing
 *			assembly code.
 *	pBufIndex:	Pointer to index into pCodeBuf.  Next code
 *			bytes will be written to pCodeBuf[*pBufIndex].
 *			If no buffer overflow, *pBufIndex will be incremented
 *			to point to next available byte in code buffer.
 *	bufSize:	Max size of the code buffer.  This routine
 *			will check for buffer overflow before doing
 *			the write.
 * Return codes:
 *	ia_successful:		The operation worked.
 *	ia_bufferExceeded:	opcode didn't fit in buffer.
 ************************************************************************/
Int32  CDECL 
ip_mmxCreateANDCode (
    IPState    *pIPState,
    Int32	nx,
    Int32	ny,
    Int32	cy,
    const UInt8 *pSE,
    Int32	topInputRow,
    Int32	topInputDisp,
    Int32	inputSECol,
    Int32	inputSEPolarity,
    Int32	cacheBpl,
    Int32	numOutputRows,
    Bool       *pOutInitArray,
    UInt8      *pCodeBuf,
    Int32      *pBufIndex,
    Int32	bufSize)
{
    PROLOGUE;
    Int32	 bufIndex = *pBufIndex;	/* needed for macros */
    Int32	 upperUses;		/* number of times upper dword
					 * is used */
    Bool	 upperLoaded;		/* if cTrue, upper input dword
					 * has been loaded into the scratch
					 * reg. */
    Int32	 lowerUses;		/* number of times lower dword
					 * is used */
    Bool	 lowerLoaded;		/* if cTrue, lower input dword
					 * has been loaded into the scratch
					 * reg. */
    Int32	 outRow;		/* temp number of output row */
    Int32	 seIndex, seEnd;	/* indices for pSE. */
    

	/* Initialize local vars */
    seEnd = nx * ny;

	/* First, we count the number of times the upper and lower
	 * dwords are used.  Whichever one is used the most will
	 * be loaded into a register before being used.
	 *
	 * The counting process takes a little thought.  Since
	 * we'll use basically the same method to generate the
	 * AND code, we'd better explain it.
	 *
	 * Since we're doing erosion, we want to bring all of the
	 * image dwords correspoinding to active SE elements
	 * into the center of the strel.  This is the opposite of
	 * dilation, in which we want to replicate the center's
	 * dword at all of the active element locations.
	 *
	 * Therefore, we imagine putting the strel's center on top
	 * of each output dword.  To see if a particular input dword
	 * (one of the two we're given) is used, we see if there's
	 * an active SE element at the offset corresponding to the
	 * input dword.  Furthermore, we see if that active element
	 * has the same polarity as the polarity of the stored input
	 * cache dword.  If so, we need to generate an AND opcode.
	 * We perform this process for the upper input dword, incrementing
	 * upperUses if we get a match.  We repeat it for the lower
	 * input dword, incrementing lowerUses if that gives a match.
	 * Note that since the the offsets from the output dword to the
	 * two cached input dwords are different, we need to look at
	 * two different SE elements.
	 *
	 * That's the obvious way to do it.  However, there's a little
	 * problem organizing things cleanly.  Another way to do the
	 * same thing is to start with the input cache dwords and
	 * iterate through the rows of the SE.  For each SE row, we
	 * see where the SE's center would end up if we put that row
	 * (and the given input col) on the input dword.  If it ends
	 * up on top of one of the output dwords (and the polarity of
	 * the SE element matches the polarity of the input cache dword)
	 * we increment the corresponding use counter (upperUses if we're
	 * looking at the upper dword, lowerUses otherwise).  This
	 * method has a couple of nice features.  First, it's a little
	 * easier to iterate through the rows of the SE, reading each
	 * one once, and checking output row numbers than to interate
	 * through output row numbers and check for array bounds problems
	 * on the SE, then read the proper SE element.  That's a
	 * minor thing.  A slightly nicer point is that we will want
	 * to generate ANDs in pairs based on two input dwords being
	 * combined in parallel with two output dwords based on the
	 * same SE element.  Iterating through the SE elements one at
	 * a time seems to make this flow more nicely.
	 * There's also the problem of what we do at the top of and
	 * bottom of the subblock.  We only generate one AND of the
	 * pair there.  Iterating through the SE elements makes this
	 * special case a bit more natural.
	 *
	 * Ok, so we iterate through the SE rows from top to bottom.
	 * This means we initially place the SE center cy rows
	 * below the input row.  Each time we move down one row in the
	 * SE, we move the output row up one.  This is a consequence of
	 * keeping the input location fixed.  Since we're placing each
	 * row of the SE in turn over the input and we're working
	 * through the SE from top to bottom, we must be moving the
	 * SE up in the image, so the output rows must decrease.
	 */
    upperUses = lowerUses = 0;
    for (seIndex = inputSECol, outRow = topInputRow + cy;
	 seIndex < seEnd;
	 seIndex += nx, outRow--)
    {
	if (pSE[seIndex] == inputSEPolarity)
	{
	    if ((outRow >= 0) && (outRow < numOutputRows))
		upperUses++;		/* upper input is used again */
	    if ((outRow + 1 >= 0) && (outRow + 1 < numOutputRows))
		lowerUses++;		/* lower input is used again */
	}
    }

	/* If the strel is pretty tall and/or there aren't many
	 * output rows, we might not use the inputs at all.
	 * If this is the case, we're done.
	 */
    if ((upperUses + lowerUses) == 0)
	NORMAL_RETURN;

	/* See which we should load into a register.  We load the
	 * one with more uses.  If there's a tie, and both counts
	 * are greater than 1, we load the lower word.
	 */
    upperLoaded = lowerLoaded = cFalse;
    if ((upperUses + lowerUses) > 2)
    {
	if (upperUses > lowerUses)
	{
	    M_COPY_DISP(loadFromCache, cScratchReg, topInputDisp);
	    upperLoaded = cTrue;
	}
	else
	{
	    M_COPY_DISP(loadFromCache, cScratchReg, topInputDisp + cacheBpl);
	    lowerLoaded = cTrue;
	}
    }

	/* Now go through the strel rows again, generating AND ops as
	 * we find proper places where the input dwords in question
	 * should be combined with the current output dwords.  Note
	 * that if we want to combine both the upper and lower dwords
	 * with output dwords, one of the combinations will be done
	 * using a AND-with-memory and the other will be AND-with-
	 * register.  The pairing rules dictate that the AND-with-memory
	 * be first.  This means we must look to see which input
	 * has been loaded before generating the opcodes.
	 *
	 * Also, if one of the output registers hasn't been touched
	 * yet, we must generate a load (from memory or a reg) instead
	 * of an AND opcode and must mark the corresponding Boolean
	 * in the initialization array.
	 *
	 * Normally, we operate on the upper word, then on the lower
	 * word.  The only case where we reverse the order is when
	 * a) we will operate on both, and b) the upper input word has
	 * been loaded into the scratch register.  In this case, we
	 * do the memory operation with the lower word in the first
	 * opcode and the scratch op with the upper dword in the second.
	 * We do this because memory ops must be in the first opcode of
	 * a pair.
	 */
    for (seIndex = inputSECol, outRow = topInputRow + cy;
	 seIndex < seEnd;
	 seIndex += nx, outRow--)
    {
	if (pSE[seIndex] == inputSEPolarity)	/* there's something to do */
	{
	    if ((upperLoaded) &&
		(outRow >= 0) && (outRow < numOutputRows) &&
		(outRow + 1 >= 0) && (outRow + 1 < numOutputRows))
	    {
		/* This is the weird case in which we process the
		 * lower input word first.
		 */
		    /* lower must not be loaded, so operate from
		     * memory.  Do we want a load or an AND?
		     */
		if (pOutInitArray[outRow+1])
		{
		    M_COPY_DISP(andFromCache, outRow + 1,
				topInputDisp + cacheBpl);
		}
		else
		{
		    M_COPY_DISP(loadFromCache, outRow + 1,
				topInputDisp + cacheBpl);
		    pOutInitArray[outRow+1] = cTrue;
		}

		    /* In second op, operate on upper word */
		if (pOutInitArray[outRow])
		{
			/* put out an AND opcode with the loaded
			 * scratch reg.  We don't use M_COPY_DISP
			 * because there's no displacement.
			 */
		    __error__ =
			ip_copyCode (pIPState, andWithScratch.disp8[outRow],
				     *(andWithScratch.disp8Len[outRow]),
				     pCodeBuf, &bufIndex, bufSize);
		    if (__error__ != ia_successful)
			goto genCleanup;
		}
		else	/* this is the first time we've touched
			     * this output register */
		{
			/* put out a LOAD opcode from the loaded
			 * scratch reg.  We don't use M_COPY_DISP
			 * because there's no displacement.
			 */
		    __error__ =
			ip_copyCode (pIPState, loadFromScratch.disp8[outRow],
				     *(loadFromScratch.disp8Len[outRow]),
				     pCodeBuf, &bufIndex, bufSize);
		    if (__error__ != ia_successful)
			goto genCleanup;
		    pOutInitArray[outRow] = cTrue;
		}
	    }
	    else	/* process in the normal order.  First,
			 * look at the upper word.
			 */
	    {
		if ((outRow >= 0) && (outRow < numOutputRows))
		{
			/* We need to process the upper word */
		    if (upperLoaded)
		    {
			    /* upper input dword is in scratch register */
			if (pOutInitArray[outRow])
			{
				/* put out an AND opcode with the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
				ip_copyCode (pIPState,
					     andWithScratch.disp8[outRow],
					     *(andWithScratch.disp8Len[outRow]),
					     pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			}
			else	/* this is the first time we've touched
				     * this output register */
			{
				/* put out a LOAD opcode from the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
			      ip_copyCode (pIPState,
					   loadFromScratch.disp8[outRow],
					   *(loadFromScratch.disp8Len[outRow]),
					   pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			    pOutInitArray[outRow] = cTrue;
			}
		    }
		    else	/* upper input dword is in memory */
		    {
			if (pOutInitArray[outRow])
			{
			    M_COPY_DISP(andFromCache, outRow,
					topInputDisp);
			}
			else
			{
			    M_COPY_DISP(loadFromCache, outRow,
					topInputDisp);
			    pOutInitArray[outRow] = cTrue;
			}
		    }
		}	/* end of processing the upper input dword */

		    /* See if there's anything to do for the lower dword. */
		if ((outRow + 1 >= 0) && (outRow + 1 < numOutputRows))
		{
			/* Now process the lower word.  See if it's
			 * already in the scratch register.
			 */
		    if (lowerLoaded)
		    {
			    /* lower dword is in scratch reg */
			if (pOutInitArray[outRow + 1])
			{
				/* put out an AND opcode with the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
			      ip_copyCode (pIPState,
					   andWithScratch.disp8[outRow+1],
					   *(andWithScratch.disp8Len[outRow+1]),
					   pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			}
			else	/* this is the first time we've touched
				     * this output register */
			{
				/* put out a LOAD opcode from the loaded
				 * scratch reg.  We don't use M_COPY_DISP
				 * because there's no displacement.
				 */
			    __error__ =
			     ip_copyCode (pIPState,
					  loadFromScratch.disp8[outRow+1],
					  *(loadFromScratch.disp8Len[outRow+1]),
					  pCodeBuf, &bufIndex, bufSize);
			    if (__error__ != ia_successful)
				goto genCleanup;
			    pOutInitArray[outRow+1] = cTrue;
			}
		    }
		    else	/* lower dword is in memory */
		    {
			if (pOutInitArray[outRow+1])
			{
			    M_COPY_DISP(andFromCache, outRow + 1,
					topInputDisp + cacheBpl);
			}
			else
			{
			    M_COPY_DISP(loadFromCache, outRow + 1,
					topInputDisp + cacheBpl);
			    pOutInitArray[outRow+1] = cTrue;
			}
		    }
		}	/* end of processing the lower input dword */
	    }	/* end of case where upper dword is processed before
		 * lower dword.
		 */
	}	/* end of case where SE polarity is correct */
    }	/* end of loop that goes through SE rows, dropping out code
	 * to process each SE element in a column.
	 */

genCleanup:
    if (__error__ == ia_successful)
    {
	*pBufIndex = bufIndex;
    }
    NORMAL_RETURN;
}


/*

$Log: pcgenerm.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.6  1998/01/15 01:46:52  ddavies
pcgenerm.c -> /product/ipcore/ipshared/src/RCS/pcgenerm.c,v

CommonCheck: Fix WATCOM complaints about unused procedure args.

Revision 1.5  1998/01/13 23:54:03  ddavies
pcgenerm.c -> /product/ipcore/ipshared/src/RCS/pcgenerm.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/08/22 01:29:35  ddavies
pcgenerm.c -> /product/ipcore/ipshared/src/RCS/pcgenerm.c,v

Make the MMX code generators generate ia_SETooBig errors for 1x1 strels.
The strels weren't going to be used anyway, but the bogus strels confused
Richard.

Revision 1.3  1997/01/31 00:47:43  ddavies
ddavies on Thu Jan 30 16:46:47 PST 1997

pcgenerm.c -> /product/ipcore/ipshared/src/RCS/pcgenerm.c,v

Include shrip.prv to get prototype for ip_longOpIPLibLinked.  Why didn't
MSVC complain about that?  It's supposed to.

Revision 1.2  1996/12/19 23:05:35  ddavies
pcgenerm.c -> /product/ipcore/ipshared/src/RCS/pcgenerm.c,v

Have i_getLongErodeAsmSize both use the cy parameter and check to see if
it's on an MMX machine.  If not, return ia_noImpl.  Also have
ip_mmxCreateErodeShiftCode use the numCols parameter.

Revision 1.1  1996/12/18 22:50:56  ddavies
Initial revision


 */

