/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * autoscan.c
 *	      
 *   Contains procedures to support the "autoscan" process. This doesn't
 *   actually do any scanning--it is intended to take the output of
 *   a scanner in a one- or two-pass mode, and perform a whole bunch
 *   of image processing tasks, including segmentation, text thresholding,
 *   image enhancement, text plane drycleaning, etc. 
 * 
 *   The particulars of the pipeline are as required by the PerfectScan
 *   API. It is hoped that it will be useful for other products.
 *   It is supposed to run in relatively little memory, so banded
 *   processing is used where possible, sometimes with an impact
 *   on performance.
 * 
 * Exports:
 *   w_autoScanTwoPassStart () - Initialize a scan, double-pass
 *   w_autoScanFirstPassData () - Shovel some data through
 *   w_autoScanFirstPassComplete () - Pass completion marker
 *   w_autoScanSecondPassData () - Shovel some data through
 *   w_autoScanSecondPassComplete () - Pass completion marker
 *
 *   w_autoScanOnePassStart () - Initialize a scan, single-pass
 *   w_autoScanOnePassData () - One-pass shovel data
 *   w_autoScanOnePassComplete () - 
 *
 *   w_autoScanGetProcessingState () - Get current state of processing
 *
 *   w_autoScanGetResult () - Get resulting page image, possibly composite
 *   w_autoScanFinish () - Clean up
 * 
 */

#include <stdio.h>
#include <math.h>
#include "ipwcore.prv"
#include "types.pub"
#include "defines.pub"
#include "geomadt.h"
#include "morph.h"
#include "morphg.h"
#include "pixr.h"
#include "thrshold.h"
#include "frames.pub"
#include "iaerror.pub"
#include "appsinc.h"
#include "logicops.h"
#include "ansiprot.h"
#include "arithops.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "thrshold.pub"
#include "fthresh.pub"
#include "shrscal.pub"
#include "page.pub"
#include "prxform.pub"
#include "shrip.pub"
#include "shrcc.pub"
#include "skew.pub"
#include "crop.pub"
#include "reduce.pub"
#include "expand.pub"
#include "rectops.pub"
#include "cut.pub"
#include "rotatop.pub"
#include "rescale.pub"
#include "fill.pub"
#include "halftone.pub"
#include "binarize.pub"
#include "coords.pub"
#include "imginfo.pub"
#include "clean.pub"
#include "imginfo.pub"
#include "iw_scale.pub"
#include "aie.pub"
#include "clean.pub"
#include "resconv.pub"
#include "morph.pub"
#include "morphg.pub"
#include "strel.pub"
#include "grbinop.pub"
#include "logicops.pub"
#include "appsutil.pub"
#include "cntr.pub"
#include "imageref.pub"
#include "arithops.pub"
#include "autoproc.pub"
#include "autoscan.pub"
#include "autoscan.prv"

#undef DEBUG
#ifdef DEBUG
#endif
#include "imagerep.h"
#include "io.pub"
#include "intops.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: autoscan.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:14 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Initialize things for a two-pass "autoscan" processing pipeline.
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function., and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "two-pass" scan involves sending two separate passes of data
 *   through the pipeline: a lower-resolution color image, and a 
 *   higher-resolution gray or binary image.  
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution gray or binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   Why do two-pass scanning? There are at least two reasons for
 *   doing two-pass scanning. First, many scanners are much slower
 *   at scanning high-resolution color than high-resolution gray
 *   (i.e. 3x slower), and so one gets a significant speed gain, since
 *   the low-resolution color is usually anywhere from 1/4 to 1/9
 *   the size of the high-resolution pass (depending on the relative
 *   resolutions, of course.) Second, one might be able to eliminate
 *   the second pass if one notices that the first pass contained no text. 
 *   <BR>
 *   The PerfectScan API uses a fairly simple data structure to represent
 *   composite pages, which is called the "segment list." The segment list
 *   is a list of images, with the convention that the first image is
 *   the "text plane" and subsequent images are "pictures." The subsequent
 *   images are placed into the composite image space using the text plane
 *   image as a coordinate space.
 *   <BR>
 *   Unlike in the PerfectScan API, I use a slightly more complicated
 *   set of data structures called the "Container", as defined in
 *   "page.c". See w_newPage() and associated functions for details,
 *   but the upshot is that it provides a composite page container structure
 *   with accessor and mutator functions, which allows more fancy page
 *   descriptions than is allowed in the PS API. This incarnation of the
 *   two-pass autoscan pipeline will always put out something which is
 *   easily translated into a PS API data structure--the text plane will
 *   always be first. However, I use an absolute coordinate space in units
 *   of Micas (1/2540th of an inch), which I think is a bit cleaner.
 *
 *   The "thumbnail" image, if provided by the caller (i.e. non-NULL)
 *   will be updated with a current view of the image as the scan
 *   and processing progresses. Note that if the thumbnail passed is 
 *   large, this could take a significant amount of time.
 *
 * PrivateDescription:::
 *   There are, I suppose, trade offs between the two page representation
 *   approaches, but since I want to be able to use this set of functions 
 *   to produce "DIR" images, where the binary image selects between 
 *   reduced-resolution fg/bg ink maps, I need some additional flexibility.
 *   It also may be that eventually, we don't want a single text plane,
 *   but want separate text block/paragraph objects. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  width1::         Width of pass-one scan
 *  height1::        Total height of pass-one scan
 *  xRes1::          X-resolution, in DPI of pass-one input image 
 *  yRes1::          Y-resolution, in DPI, of pass-one input image
 *  depth1::         Depth, bits/pixel, of pass-one input image
 *  pageType::       Requested page type, i.e. cTextOnly, cPictureOnly,
 *                   cTextPicture
 *  textXRes::       Requested X-resolution, in DPI, of text in output image
 *  textYRes::       Requested Y-resolution, in DPI, of text in output image
 *  pictXRes::       Requested X-resolution, DPI, of pictures in output image
 *  pictYRes::       Requested Y-resolution, DPI, of pictures in output image
 *  pictChannels::   Requested # of channels in picture images
 *  pictType::       Requested color type of picture images
 *  pictMap::        If picture images are to be color-mapped, use this map
 *  fgXRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the foreground plane at this X-resolution
 *  fgYRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the foreground plane at this Y-resolution
 *  bgXRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the background plane at this X-resolution
 *  bgYRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the background plane at this X-resolution
 *  processingOptions:: Processing flags, see autoscan.pub
 *  prThumbnail::    If non-NULL, this is the thumbnail picture which
 *                   is to be updated periodically. The thumbnail must
 *                   either be 24-bit RGB or 8-bit gray.
 *  
 * Output:::
 *  ppScanState::    Autoscan state handle
 *  pStripHeight1::  Strips passed to the first pass should be this high,
 *                   unless on the last band. 
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanTwoPassStart (
    IPWState          *pIPWState,
    Int32               width1,
    Int32               height1,
    ScaledInt1616       xRes1,
    ScaledInt1616       yRes1,
    Int32               depth1,
    Int32               pageType,
    ScaledInt1616       textXRes,
    ScaledInt1616       textYRes,
    ScaledInt1616       pictXRes,
    ScaledInt1616       pictYRes,
    Int32               pictChannels,
    ColorType           pictType,
    ColorMap           *pictMap,
    ScaledInt1616       fgXRes,
    ScaledInt1616       fgYRes,
    ScaledInt1616       bgXRes,
    ScaledInt1616       bgYRes,
    UInt32              processingOptions,
    PIXR               *prThumbnail,
    void              **ppScanState,
    Int32              *pStripHeight1)
{
    PROLOGUE;
    AutoScanState      *pState;

    /* Check parameters */
    ERR_GEN (*ppScanState != NULL, cErrNullPtr);

    /* All resolutions must be specified */
    ERR_GEN (xRes1 <= 0, cErrApps_invalidResolution);
    ERR_GEN (yRes1 <= 0, cErrApps_invalidResolution);
    ERR_GEN (textXRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (textYRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (pictXRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (pictYRes <= 0, cErrApps_invalidResolution);

    /* Depth is either 4 or 8 on the first-pass scan */
    ERR_GEN (depth1 != 4 && depth1 != 8, cErrPixr_depth);

    if (prThumbnail)
    {
	ERR_GEN_PIXR (prThumbnail, cByteOrderBig, cDepth8, 
		      cChannels1 | cChannels3, cFrame32);
    }

    /* Create a state structure */
    pState = (AutoScanState *) IPW_MALLOC (sizeof (AutoScanState),
					   cAllocatorData);
    ERR_GEN (!pState, cErrNomem);
    MEMSET((void *)pState, 0, sizeof(AutoScanState));

    pState->prThumbnail = prThumbnail;
    SET_STATE (cAutoState_unknown);
    
    /* Fill it in */
    pState->width1 = width1;
    pState->height1 = height1;
    pState->xRes1 = xRes1;
    pState->yRes1 = yRes1;
    pState->depth1 = depth1;
    pState->pageType = pageType;
    pState->orientation = 0;
    pState->skew = 0.0F;

    /* Set bounding boxes so that if we don't crop, it still works */
    pState->initialBoundingBox.x = 0;
    pState->initialBoundingBox.y = 0;
    pState->initialBoundingBox.w = width1;
    pState->initialBoundingBox.h = height1;
    pState->boundingBox.x = 0;
    pState->boundingBox.y = 0;
    pState->boundingBox.w = width1;
    pState->boundingBox.h = height1;

    /* Create the image plane. NOTE that the image plane is the
     * pass-1 scale, not scaled-up. We don't scale up until we
     * copy out the segments.
     */
    pState->pictXScale = (ScaledInt1616_TO_Float32(pictXRes) / 
			  ScaledInt1616_TO_Float32(xRes1));
    pState->pictYScale = (ScaledInt1616_TO_Float32(pictYRes) /
			  ScaledInt1616_TO_Float32(yRes1));
    pState->pictXRes = pictXRes;
    pState->pictYRes = pictYRes;
    pState->prImagePlane = NULL;
    pState->prTextPlane = NULL;
    ERR_PASS_CLEANUP (w_createPixrExtended (pIPWState, width1, height1, 
					    depth1, xRes1, yRes1, cFrameBits,
					    pictChannels, pictType, pictMap, 
					    &(pState->prImagePlane)));

    /* We're not going to deal with these at all yet */
    pState->prForegroundMap = pState->prBackgroundMap = NULL;
    pState->processingFlags = processingOptions;
    if (ENABLED(cAutoOption_mapInkAndPage))
    {
	/* Must use binarize2 */
	ERR_GEN (!ENABLED (cAutoOption_binarize2), cErrTypeNotSupported);
	/* Spiggot dose not do map-ink,  API want it for Pagis X */
	/* ERR_GEN (pageType == cSpiggotOnly, cErrTypeNotSupported); */

	/* If set both no-foreground and no-background, it is equal to
	   no map-ink */
	ERR_GEN (ENABLED (cAutoOption_noForeground) &&
		 ENABLED (cAutoOption_noBackground), cErrTypeNotSupported);
	/* Foreground resolution must be less or equal text resolution */
	ERR_GEN (textXRes < fgXRes, cErrApps_invalidResolution);
	ERR_GEN (textYRes < fgYRes, cErrApps_invalidResolution);
	ERR_GEN (depth1 != 8, cErrPixr_depth);

	if (ENABLED (cAutoOption_pastePictures))
	{
	    /* Must use the same resolution for BG and pictures in this case */
	    if ((ABS(pictXRes - bgXRes) > 32768L) || 
		(ABS(pictYRes - bgYRes) > 32768L))
	    {
		ERR_GEN (1, cErrApps_invalidResolution);
	    }
	}

	pState->fgXScale = (ScaledInt1616_TO_Float32(fgXRes) /
			      ScaledInt1616_TO_Float32(xRes1));
	pState->fgYScale = (ScaledInt1616_TO_Float32(fgYRes) /
			      ScaledInt1616_TO_Float32(yRes1));
	pState->fgXRes = fgXRes;
	pState->fgYRes = fgYRes;

	pState->bgXScale = (ScaledInt1616_TO_Float32(bgXRes) /
			      ScaledInt1616_TO_Float32(xRes1));
	pState->bgYScale = (ScaledInt1616_TO_Float32(bgYRes) /
			      ScaledInt1616_TO_Float32(yRes1));
	pState->bgXRes = bgXRes;
	pState->bgYRes = bgYRes;

	/*
	 * If possible, we want to do max-reduction as many times as 
	 * possible, then do a scaling. 
	 */
	if (FEQ (pState->bgXScale, pState->bgYScale) && 
	    pState->bgXScale >0)
	{
	    /* Ansi didn't bother defining log2.. :-p */
	    pState->bgReductions = (Int32) (log (1.0 / pState->bgXScale)/
		                            log (2.0));
	    if (pState->bgReductions > 4)
		      pState->bgReductions = 4;
	    if (pState->bgReductions < 0)
		      pState->bgReductions = 0;
	    pState->bgScaleAfterReduce = pState->bgXScale *
		(1 << pState->bgReductions);
	}
	else
	{
	    pState->bgReductions = 0;
	}  
	pState->bgY = 0;
   }

    /* Set stuff up for the first pass */
    pState->pass = 0;
    pState->passY = 0;
    pState->imageY = 0;
    pState->pRotateState = NULL;
    pState->prRotatedStrip = NULL;
    pState->textXRes = textXRes;
    pState->textYRes = textYRes;
    pState->skew = 0.0F;
    pState->pImages = NULL;
    pState->pMasks = NULL;
    *pStripHeight1 = cBandHeight;
    *ppScanState = (void *) pState;
    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prImagePlane));
    SET_STATE (cAutoState_initialized);
    SET_STATE (cAutoState_processPass1);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pState) 
	IPW_FREE (pState, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Process a hunk of "pass 1" data through the "autoscan" pipeline.
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function, and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "two-pass" scan involves sending two separate passes of data
 *   through the pipeline: a lower-resolution color image, and a 
 *   higher-resolution gray or binary image.  
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution gray or binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   This function takes a hunk of pass-1 data, expands if necessary, converts
 *   type if necessary, and stuffs it into the "picture" image at the right
 *   spot.
 *   <BR>
 *   Not all options are implemented. Basically, we expect that the data
 *   passed is the same depth and color type as the requested picture
 *   type.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanTwoPassStart()
 *  prImageStrip::    Strip of image data to process   
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanFirstPassData (
    IPWState           *pIPWState,
    void               *pScanState,
    PIXR               *prImageStrip)
{
    PROLOGUE;
    Int32          w, h;
    AutoScanState *pState = (AutoScanState *) pScanState;

    ERR_GEN (!pState, cErrNullPtr);
    ERR_GEN (!prImageStrip, cErrNullPtr);
    ERR_PASS (w_pixrGetWidth (pIPWState, prImageStrip, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImageStrip, &h));
    ERR_GEN (w != pState->width1, cErrApps_differentDimensions);
    ERR_GEN (h + pState->passY > pState->height1, cErrApps_tooMuchData);

    /* Never scaling in two-pass mode.. Just copy it in */
    ERR_PASS (w_rasterop (pIPWState, 0, 0, prImageStrip, 
			  PIX_SRC, 0, pState->imageY, w, h, 
			  pState->prImagePlane));
    pState->passY += h;
    pState->imageY += h;
    /* Limit the number to mod8 */
    if (((pState->imageY / h) % 2) == 0)
    {
	ERR_PASS (wp_updateThumbnail (pIPWState, pState, 
				      pState->prImagePlane));
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Process whole "pass 1" image through the "autoscan" pipeline.
 *
 * LongDescription:::
 *   This function takes whole image data into the image plane.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanTwoPassStart()
 *  prImage::        Image data to process   
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanFirstTransData (
    IPWState           *pIPWState,
    void               *pScanState,
    PIXR               *prImage)
{
    PROLOGUE;
    Int32          w, h;
    AutoScanState *pState = (AutoScanState *) pScanState;

    ERR_GEN (!pState, cErrNullPtr);
    ERR_GEN (!prImage, cErrNullPtr);
    ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));
    ERR_GEN (w != pState->width1, cErrApps_differentDimensions);
    ERR_GEN (h > pState->height1, cErrApps_tooMuchData);

    /* Never scaling in two-pass mode.. Just copy it in */
    ERR_PASS (w_rasterop (pIPWState, 0, 0, prImage, 
			  PIX_SRC, 0, 0, w, h, 
			  pState->prImagePlane));

    pState->passY = h;
    pState->imageY = h;

    ERR_PASS (wp_updateThumbnail (pIPWState, pState, 
				  pState->prImagePlane));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Signal that the first pass of data has all been sent
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function, and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "two-pass" scan involves sending two separate passes of data
 *   through the pipeline: a lower-resolution color image, and a 
 *   higher-resolution gray or binary image.  
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution gray or binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   This function should be called after all of the strips of data
 *   have been passed in via w_autoScanFirstPassData(). It actually
 *   does most of the image analysis as requested.
 *
 * PrivateDescription:::
 *   The processing steps are as follows:
 *<BR>
 *     1. Color detect. If not color, force the result to gray. (?)<BR>
 *     2. Threshold to binary with possible upconvert, run findSkewSample,
 *        use the result of that to determine whether there was text on 
 *        the page.<BR>
 *     3. If there was no text on the page, make a single-image container
 *        and call it a day.<BR>
 *     4. find document bounding box and orientation, if requested, and
 *        crop and orient.<BR>
 *     5. Find skew angle if requested, and deskew. <BR>
 *     6. If cropping was requested, find bounding box and crop again.<BR>
 *     7. call w_segmentPagePixrs() to get all of the images. Make a
 *        bunch of little PIXRs (expanded as needed), and get rid of the
 *        first-pass image.
 *  
 * Input:::
 *  pIPWState::              IPWrap State structure
 *  pScanState::             Scan state structure, as created by
 *                           w_autoScanTwoPassStart()
 *  xRes2::                  X-Resolution (dpi) of second-pass scan
 *  yRes2::                  Y-Resolution (dpi) of second-pass scan
 *  depth2::                 Depth (bits/pixel) of second-pass scan
 *
 * Output:::
 *  ppSecondPassCrop::       This is the crop rectangle that should be used
 *                           on the second-pass crop, specified in pixels
 *                           at the second-pass resolution. If this is NULL,
 *                           then there was no text, so don't do a 
 *                           second-pass scan.
 *  pSecondPassStripHeight:: Number of lines that the second pass
 *                           expects to receive in all strips but the
 *                           last one.
 * 
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanFirstPassComplete (
    IPWState  *pIPWState,
    void      *pScanState,
    Int32      xRes2,
    Int32      yRes2,
    Int32      depth2,
    RECTANGL **ppSecondPassCrop,
    Int32     *pSecondPassStripHeight)
{
    PROLOGUE;
    AutoScanState *pState = (AutoScanState *) pScanState;
    Int32          dpi, expansion = 0, confidence, reduction;
    PIXR          *prLowResText = NULL, *prRotated = NULL, *prTmp = NULL, 
	          *prTransfer = NULL, *prPrimary = NULL;
    Float32        skew, expandX, expandY;
    RECTANGL       *pRect = NULL;
    ScaledInt1616  xRes, yRes;
    PixrRef       *pMasks = NULL, *pImages = NULL;
    Bool           isBlank = cFalse;

    /* Check parameters */
    ERR_GEN (!pState, cErrNullPtr);
    ERR_GEN (!ppSecondPassCrop, cErrNullPtr);
    ERR_GEN (*ppSecondPassCrop, cErrPtrNotNull);
    ERR_GEN (depth2 != 1 && depth2 != 4 && depth2 != 8, cErrPixr_depth);

    pState->depth2 = depth2;
    pState->xRes2 = xRes2;
    pState->yRes2 = yRes2;

    SET_STATE (cAutoState_binarize);
    /* Threshold adaptively. Use binarize2 for now... */
    dpi = MAX(ScaledInt1616_TO_Int32 (pState->xRes1), 
	      ScaledInt1616_TO_Int32 (pState->yRes1));
    
    if (dpi <= 75)
	expansion = 2;
    else if (dpi <= 100)
	expansion = 1;
    else
	expansion = 0;

    ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, pState->prImagePlane, 
				       &prPrimary));

    /* Make a binary version */
    if (pState->depth1 == 8)
    {
	ERR_PASS (w_binarize2 (pIPWState, prPrimary, expansion,
			       cLowResGradientThreshold, &prLowResText));
    }
    else
    {
	ERR_PASS (w_binarize (pIPWState, prPrimary,
			      expansion, NULL, 1, 0, cFalse,
			      &prLowResText));
    }

    prPrimary = NULL;

    /* 
     * The state is "isText" even though this is a skew-detect; that
     * is what we're using it for
     */

    SET_STATE (cAutoState_isText);

    /* 4-17-98 check whether text is all white */
    ERR_PASS_CLEANUP (wp_checkBlankPage(pIPWState, prLowResText, &isBlank));

    /* Do a skew detect */
    if( isBlank == cFalse )
    {
	reduction = 2;
	ERR_PASS_CLEANUP (w_findSkewSample (pIPWState, prLowResText, 
					    reduction, &confidence, &skew));
	/* Double check text */
	if( (confidence == 0) && (skew == 0.0f))
	{
	    ERR_PASS_CLEANUP( wp_autoSkewDetection (pIPWState, 
						    prLowResText,
						    pState, &skew, 
						    &confidence));
	}
    }
    else
    {
	skew = 0.0f;
	confidence = 0;
    }

    if( (skew == 0.0f) && (confidence == 0) )
	pState->IsPhoto = cTrue;
    else
	pState->IsPhoto = cFalse;

    if (((skew == 0.0f) && (confidence == 0) && 
	 pState->pageType != cSpiggotOnly) || (isBlank == cTrue) )
    {
	/* There was no text. What this means is that we should,
	 * if requested, do a photo crop and deskew, and set a flag
	 * such that there will only be a photo on the page.
	 */
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prLowResText));

	/* This routine does: deskew, crop, AIE (and who knows what else): */
	if(isBlank == cFalse)
	    ERR_PASS_CLEANUP (wp_autoPhotoCrop (pIPWState, pState));

	/* We don't want a second pass */
	*ppSecondPassCrop = NULL;
	*pSecondPassStripHeight = 0;
    }
    else
    {
	/* Deskew reported some text, so we should act accordingly */
	ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, 
						  prLowResText,
						  &xRes, &yRes));

	if( pState->IsPhoto != cTrue )
	    ERR_PASS_CLEANUP(w_autoProcDetections (pIPWState, prLowResText,
						   pState->processingFlags,
						   &pState->initialBoundingBox,
						   &pState->orientation,
						   &pState->skew,
						   &confidence,
						   &pState->boundingBox));
	/* Get rid of the low-res text plane at this point */
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prLowResText));
		
	if (ENABLED (cAutoOption_crop) &&  (pState->IsPhoto != cTrue) &&
	    (ENABLED(cAutoOption_deskew) || ENABLED(cAutoOption_orient)))
	{

	    SET_STATE (cAutoState_crop);

	    /* Crush the box if needed, and cut the image */
	    if (expansion == 2)
	    {
		pState->initialBoundingBox.x /= 2;
		pState->initialBoundingBox.y /= 2;
		pState->initialBoundingBox.w /= 2;
		pState->initialBoundingBox.h /= 2;
	    }
	    else if (expansion == 1)
	    {
		pState->initialBoundingBox.x = 
		    pState->initialBoundingBox.x * 2 / 3;
		pState->initialBoundingBox.y = 
		    pState->initialBoundingBox.y * 2 / 3;
		pState->initialBoundingBox.w = 
		    pState->initialBoundingBox.w * 2 / 3;
		pState->initialBoundingBox.h = 
		    pState->initialBoundingBox.h * 2 / 3;
	    }
	
	    ERR_PASS_CLEANUP (
		w_cutRectIP (pIPWState,
			     pState->initialBoundingBox.x,
			     pState->initialBoundingBox.y,
			     pState->initialBoundingBox.w,
			     pState->initialBoundingBox.h,
			     pState->prImagePlane));
	}

	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prImagePlane));
	/* Check orientation */
	if (ENABLED (cAutoOption_orient))
	{
	    Bool  canInPlace;
	    if (pState->orientation != 0)
	    {    
		canInPlace = ENABLED(cAutoOption_moreMemory) ? cFalse : cTrue;

		SET_STATE (cAutoState_orientCorrect);
		ERR_PASS (w_autoProcOrientImage (pIPWState, 
						 pState->orientation, 
						 canInPlace, 
						 &pState->prImagePlane));
	    }
	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prImagePlane));
	}

	/* Save image , it may be used for re-cropping.
	   If enable cAutoOption_mapInkAndPage, it will be used
	   to do color segmentation */

	pState->prBackgroundMap = NULL;
	ERR_PASS_CLEANUP (w_copy (pIPWState, pState->prImagePlane, 
				  &pState->prBackgroundMap));

	if (ENABLED (cAutoOption_deskew) && (pState->IsPhoto != cTrue))
	{

	    /* 
	     * NOTE that we don't inform the caller if there wasn't
	     * enough information to find a skew, i.e. skew = 0.0f and
	     * confidence = 0. It might be useful to have some 
	     * way of notifying the caller, rather than just not 
	     * deskewing at all.
	     */

	    if (fabs(pState->skew) > cSmallestRotatingAngle)
	    {
		SET_STATE (cAutoState_deskew);

		if (fabs(pState->skew) > cLargestSmallAngle)
		{
		    /* Well, if you give THAT big a skew, you deserve to
		       use a lot of memory */
		    prRotated = NULL;
		    if( fabs(pState->skew) < cLargestGrayAngle )
			ERR_PASS_CLEANUP (w_rotateGray (pIPWState,
							pState->prImagePlane,
							pState->skew, 
							&prRotated)); 
		    else
			ERR_PASS_CLEANUP (w_rotateArb (pIPWState, 
						       pState->prImagePlane, 
						       pState->skew, cTrue, 
						       cTrue, &prRotated));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
						     &pState->prImagePlane));
		    pState->prImagePlane = prRotated;
		    prRotated = NULL;
		}
		else
		{
		    ERR_PASS_CLEANUP (w_rotateGrayIP (pIPWState, pState->skew, 
						      pState->prImagePlane));
		}

	    }

	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prImagePlane));
	}

	if (ENABLED (cAutoOption_crop) && (pState->IsPhoto != cTrue))
	{

	    /* Deskew the text plane, if needed */
	    /* Crush the box if needed, and cut the image */
	    if (expansion == 2)
	    {
		pState->boundingBox.x /= 2;
		pState->boundingBox.y /= 2;
		pState->boundingBox.w /= 2;
		pState->boundingBox.h /= 2;
	    }
	    else if (expansion == 1)
	    {
		pState->boundingBox.x = pState->boundingBox.x * 2 / 3;
		pState->boundingBox.y = pState->boundingBox.y * 2 / 3;
		pState->boundingBox.w = pState->boundingBox.w * 2 / 3;
		pState->boundingBox.h = pState->boundingBox.h * 2 / 3;
	    }

	    /* Crop the image in-place */
	    ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
					   pState->boundingBox.x,
					   pState->boundingBox.y,
					   pState->boundingBox.w,
					   pState->boundingBox.h,
					   pState->prImagePlane));
	    pState->boxXRes = xRes;
	    pState->boxYRes = yRes;
	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prImagePlane));
	}

	/* Do the segmentation */
	if( pState->pageType != cSpiggotOnly)
	    ERR_PASS_CLEANUP (wp_autoSegmentPage (pIPWState, pState));

	/* 
	 * Compute the bounding box for the "second-pass" scan, since
	 * we don't have to scan the whole platen if we did some
	 * cropping.
	 */
	
	pState->textXScale = (Float32) 
	    (ScaledInt1616_TO_Float32 (pState->textXRes) / 
	     ScaledInt1616_TO_Float32 (xRes2));
	pState->textYScale = (Float32)
	    (ScaledInt1616_TO_Float32 (pState->textYRes) /
	     ScaledInt1616_TO_Float32 (yRes2));
	expandX = (Float32) (ScaledInt1616_TO_Float32(xRes2) / 
			     ScaledInt1616_TO_Float32(pState->xRes1));
	expandY = (Float32) (ScaledInt1616_TO_Float32(yRes2) / 
			     ScaledInt1616_TO_Float32(pState->yRes1));

	if (ENABLED (cAutoOption_crop) && (pState->IsPhoto != cTrue))
	{
	    if (ENABLED (cAutoOption_deskew) ||
		ENABLED (cAutoOption_orient))
	    {

		ERR_PASS_CLEANUP (w_createRect (pIPWState, 
						pState->initialBoundingBox.x, 
						pState->initialBoundingBox.y,
						pState->initialBoundingBox.w, 
						pState->initialBoundingBox.h, 
						ppSecondPassCrop));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_createRect (pIPWState, 
						pState->boundingBox.x, 
						pState->boundingBox.y,
						pState->boundingBox.w, 
						pState->boundingBox.h, 
						ppSecondPassCrop));
	    }
		
	    /* Scale up to second-pass resolution */
	    pRect = *ppSecondPassCrop;
	    pRect->x = SCALE32_NR (pRect->x, expandX);
	    pRect->y = SCALE32_NR (pRect->y, expandY);
	    pRect->w = SCALE32_NR (pRect->w, expandX);
	    pRect->h = SCALE32_NR (pRect->h, expandY);
	}
	else
	{
	    ERR_PASS_CLEANUP (w_createRect (pIPWState, 0, 0, 0, 0, 
					    ppSecondPassCrop));
	    pRect = *ppSecondPassCrop;
	    pRect->x = 0;
	    pRect->y = 0;
	    pRect->w = SCALE32_NR (pState->width1, expandX);
	    pRect->h = SCALE32_NR (pState->height1, expandY);
	}

	pState->width2 = SCALE32_NR (pRect->w, pState->textXScale);
	pState->height2 = SCALE32_NR (pRect->h, pState->textYScale);
	pState->prTextPlane = NULL;

    /* Set second pass image cropping default flag. Default is no cropping 
       in autoScanSecondPassData, scanner will do cropping. */
	pState->secondPassCropFlag = 0;
	pState->secondPassImageY =0;
	pState->secondPassRect.x = pRect->x;
	pState->secondPassRect.y = pRect->y;
	pState->secondPassRect.w = pRect->w;
	pState->secondPassRect.h = pRect->h;

	/* This has to be initialized because we OR the text into it.
	 */
	ERR_PASS_CLEANUP (w_createPixr (pIPWState,
					pState->width2, pState->height2,
					1,  pState->textXRes,
					pState->textYRes, cFrameBits,
					&(pState->prTextPlane)));

	

	/* Set up banded binarize for the second pass */
	if (depth2 != 1)
	{
	    if (ENABLED (cAutoOption_binarize2))
	    {
		expansion = 0;
		if( pState->textXScale < 0.95 || pState->textXScale > 1.05 )
		    expansion = cIsScaledImage;
		ERR_PASS_CLEANUP (w_binarize2Init (pIPWState, 
						   pState->width2, 
						   pState->height2, 
						   8, MAX(pState->textXRes, 
							  pState->textYRes), 
						   expansion, 32,
						   &pState->prTextPlane,
						   &pState->pBinState));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (
		w_binarizeInit (pIPWState, pState->width2, pState->height2, 
				depth2, MAX(pState->textXRes,pState->textYRes),
				0, NULL, cTrue, cFalse,
				ENABLED(cAutoOption_pepperRemove) ? 
				cTrue:cFalse,
				&pState->prTextPlane, &pState->pBinState));
	    }
    
	    /* Set up the gray rotate stuff, if needed */
	    /* Actually, I'm going to blow it off for now,
	     * figuring that since we're going to deskew the binary
	     * anyway in the one-pass case, doing gray rotate in the
	     * two-pass case is pleasant but not as important as getting
	     * everything else done.
	     */
	    pState->passY = 0;
	}
	else
	{
#if 0
	    pState->prTextPlane = NULL;
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, pState->width2, 
						  pState->height2, 1, 
						  xRes2, yRes2, cFrameBits, 
						  &pState->prTextPlane));
#endif
	    pState->passY = 0;
	    pState->textY = 0;
	}

	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->prTextPlane));
		   
	*pSecondPassStripHeight = cBandHeight;

    }

    SET_STATE (cAutoState_processPass2);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prLowResText)
	w_destroyPixr (pIPWState, &prLowResText);
    if (prRotated)
	w_destroyPixr (pIPWState, &prRotated);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prTransfer)
	w_destroyPixr (pIPWState, &prTransfer);
    if (pState->prRotatedStrip)
	w_destroyPixr (pIPWState, &(pState->prRotatedStrip));

    if (pMasks)
	IPW_FREE (pMasks, cAllocatorData);
    if (pImages)
	IPW_FREE (pImages, cAllocatorData);
    if (pState->pRotateState)
	IPW_FREE (pState->pRotateState, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Set status of second pass image cropping.
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function, and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "two-pass" scan involves sending two separate passes of data
 *   through the pipeline: a lower-resolution color image, and a 
 *   higher-resolution gray or binary image.  
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution gray or binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   This function set second pass cropping rectangle if scanner can't
 *   scan part of image within a rectangle. It must be called before 
 *   w_autoScanSecondPassData().
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::         IPWrap State structure
 *  pScanState::        Scan state structure, as created by
 *                      w_autoScanTwoPassStart()
 *  width::             second pass image width
 *  height::            second pass image height
 *  pSecondPassCrop::   Point to rectangle structure  
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  x. zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanSetSecondPassRect (
    IPWState          *pIPWState,
    void              *pScanState,
    Int32              width,
    Int32              height,
    RECTANGL          *pSecondPassCrop)
{
    PROLOGUE;
    AutoScanState *pState = (AutoScanState *) pScanState;
    RECTANGL *tmpRect = NULL;
    Int32     w, h, x, y, expansion;
    Float32   scaleX, scaleY;
    Int32     imgX, imgY, imgW, imgH, i;

    /* Check parameters */
    ERR_GEN (!pState, cErrNullPtr);

    if( pSecondPassCrop == NULL )
    {
	pState->secondPassCropFlag = cNoSecondPassRectangle;
    }
    else
    {
	ERR_PASS (w_clipRect (pIPWState, pSecondPassCrop, width, height, 
			      &tmpRect));

	ERR_GEN (( pState->secondPassRect.x > tmpRect->x || 
		   pState->secondPassRect.y > tmpRect->y ), cErrOutOfBounds);

	pState->secondPassCropFlag = cSecondPassRectangle;
	pState->secondPassImageY =0;

	x = pState->secondPassRect.x + pState->secondPassRect.w;
	y = pState->secondPassRect.y + pState->secondPassRect.h;
	if( (tmpRect->x + tmpRect->w) > x )
	    tmpRect->w = x - tmpRect->x;
	if( (tmpRect->y + tmpRect->h) > y )
	    tmpRect->h = y - tmpRect->y;

	x = tmpRect->x - pState->secondPassRect.x;
	y = tmpRect->y - pState->secondPassRect.y;

	w = SCALE32_NR (tmpRect->w, pState->textXScale);
	h = SCALE32_NR (tmpRect->h, pState->textYScale);

	if( w != pState->width2 || h != pState->height2 )
	{
	    pState->height2 = h;
	    pState->width2 = w;
	    ERR_PASS (w_cutRectIP (pIPWState,
				   0, 0,
				   pState->width2,
				   pState->height2,
				   pState->prTextPlane));
	    if (pState->depth2 != 1)
	    {
		if (ENABLED (cAutoOption_binarize2))
		{
		    expansion = 0;
		    if(pState->textXScale < 0.95 || pState->textXScale > 1.05)
			expansion = cIsScaledImage;
		    ERR_PASS(w_binarize2Finish(pIPWState, &pState->pBinState));
		    ERR_PASS(w_binarize2Init (pIPWState, 
					      pState->width2, 
					      pState->height2, 
					      8, MAX(pState->textXRes, 
						     pState->textYRes), 
					      expansion, 32,
					      &pState->prTextPlane,
					      &pState->pBinState));
		}
		else
		{
		    ERR_PASS(w_binarizeFinish (pIPWState, &pState->pBinState));
		    ERR_PASS(w_binarizeInit (pIPWState, pState->width2, 
					pState->height2, pState->depth2, 
					MAX(pState->textXRes,pState->textYRes),
					0, NULL, cTrue, cFalse,
					ENABLED(cAutoOption_pepperRemove) ? 
					cTrue:cFalse,
					&pState->prTextPlane, 
					&pState->pBinState));
		}
    
		pState->passY = 0;
	    }
	}
	pState->secondPassRect.x = tmpRect->x;
	pState->secondPassRect.y = tmpRect->y;
	pState->secondPassRect.w = tmpRect->w;
	pState->secondPassRect.h = tmpRect->h;
	ERR_PASS (w_destroyRect (pIPWState, &tmpRect));

	scaleX = ScaledInt1616_TO_Float32(pState->xRes1)/
	    ScaledInt1616_TO_Float32(pState->xRes2);
	scaleY = ScaledInt1616_TO_Float32(pState->yRes1)/
	    ScaledInt1616_TO_Float32(pState->yRes2);
	if( pState->orientation == 90 || pState->orientation == 270 )
	{
	    imgH = SCALE32_NR( pState->secondPassRect.w, scaleX);
	    imgW = SCALE32_NR( pState->secondPassRect.h, scaleY);
	    imgY = SCALE32_NR( x, scaleX);
	    imgX = SCALE32_NR( y, scaleY);
	}
	else
	{
	    imgW = SCALE32_NR( pState->secondPassRect.w, scaleX);
	    imgH = SCALE32_NR( pState->secondPassRect.h, scaleY);
	    imgX = SCALE32_NR( x, scaleX);
	    imgY = SCALE32_NR( y, scaleY);
	}
	ERR_PASS ( w_pixrGetWidth (pIPWState, 
				   pState->prBackgroundMap, &w));
	ERR_PASS ( w_pixrGetHeight (pIPWState, 
				    pState->prBackgroundMap, &h));
	if( imgX != 0 || imgY != 0 || w > imgW || h > imgH )
	{
	    ERR_PASS (w_cutRectIP (pIPWState, imgX, imgY, imgW, imgH,
				   pState->prBackgroundMap));
	    if( pState->pageType == cSpiggotOnly)
		ERR_PASS (w_cutRectIP (pIPWState, imgX, imgY, imgW, imgH,
				       pState->prImagePlane)); 
	    else
	    { 
		/* Free old segmented images */
		for (i = 0; i < pState->imageCount; i++)
		{
		    /* Get rid of the images */
		    ERR_PASS ( w_destroyPixr (pIPWState, 
					      &(pState->pImages[i].pPixr)));
		    ERR_PASS ( w_destroyPixr (pIPWState, 
					      &(pState->pMasks[i].pPixr)));
		}
		pState->imageCount = 0;
		IPW_FREE (pState->pImages, cAllocatorData);
		IPW_FREE (pState->pMasks, cAllocatorData);

		pState->prImagePlane  = NULL;
		ERR_PASS (w_copy (pIPWState, pState->prBackgroundMap, 
				  &pState->prImagePlane));
	    }
	    ERR_PASS (wp_updateThumbnail (pIPWState, pState, 
					  pState->prImagePlane));


	    if (ENABLED (cAutoOption_deskew))
	    {
		PIXR  *prRotated;
	    /* 
	     * Deskew. If we oriented, do a skew-detect again.
	     * This might not be necessary.
	     */

		if (fabs(pState->skew) > cSmallestRotatingAngle)
		{
		    SET_STATE (cAutoState_deskew);

		    if (fabs(pState->skew) > cLargestSmallAngle)
		    {
		    /* Well, if you give THAT big a skew, you deserve to
		       use a lot of memory */
			prRotated = NULL;
			if( fabs(pState->skew) < cLargestGrayAngle )
			    ERR_PASS (w_rotateGray (pIPWState,
						    pState->prImagePlane,
						    pState->skew, 
						    &prRotated)); 
			else
			    ERR_PASS (w_rotateArb (pIPWState, 
						   pState->prImagePlane, 
						   pState->skew, 
						   cTrue, cTrue, 
						   &prRotated));
			ERR_PASS (w_destroyPixr (pIPWState, 
						 &pState->prImagePlane));
			pState->prImagePlane = prRotated;
			prRotated = NULL;
		    }
		    else
		    {
			ERR_PASS (w_rotateGrayIP (pIPWState, 
						  pState->skew, 
						  pState->prImagePlane));
		    }
		    
		}

		ERR_PASS (wp_updateThumbnail (pIPWState, pState, 
					      pState->prImagePlane));
	    }

	    if (ENABLED (cAutoOption_crop))
	    {
		if( pState->boundingBox.x < imgX )
		{
		    pState->boundingBox.w -= (imgX - pState->boundingBox.x);
		    if( pState->boundingBox.w > imgW)
			pState->boundingBox.w = imgW;
		    pState->boundingBox.x = 0;
		}
		else
		{
		    pState->boundingBox.x -= imgX;
		    if( (pState->boundingBox.x + pState->boundingBox.w) > imgW)
			pState->boundingBox.w= imgW - pState->boundingBox.x;
		}
		if( pState->boundingBox.y < imgY )
		{
		    pState->boundingBox.h -= (imgY - pState->boundingBox.y);
		    if( pState->boundingBox.h > imgH )
			pState->boundingBox.h = imgH;
		    pState->boundingBox.y = 0;
		}
		else
		{
		    pState->boundingBox.y -= imgY;
		    if( (pState->boundingBox.y + pState->boundingBox.h) > imgH)
			pState->boundingBox.h=imgH - pState->boundingBox.y;
		}
	    /* Crop the image in-place */
		ERR_PASS (w_cutRectIP (pIPWState,
				       pState->boundingBox.x,
				       pState->boundingBox.y,
				       pState->boundingBox.w,
				       pState->boundingBox.h,
				       pState->prImagePlane));

		ERR_PASS (wp_updateThumbnail (pIPWState, pState, 
						      pState->prImagePlane));
	    }

	    /* Do the segmentation */
	    if( pState->pageType != cSpiggotOnly)
		ERR_PASS (wp_autoSegmentPage (pIPWState, pState));
	
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Process a hunk of "pass 2" data through the "autoscan" pipeline.
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function, and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "two-pass" scan involves sending two separate passes of data
 *   through the pipeline: a lower-resolution color image, and a 
 *   higher-resolution gray or binary image.  
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution gray or binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   This function takes a hunk of pass-2 data, maybe resizes if needed,
 *   binarizes, and stuffs it into the "text" plane at the right spot.
 *   It should at some point use Dan's nifty deskewing code, but 
 *   figuring out exactly how to set up the buffering so that the client 
 *   can always send the same strip size made my head hurt.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanTwoPassStart()
 *  prImageStrip::   Strip of image data to process   
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanSecondPassData (
    IPWState          *pIPWState,
    void               *pScanState,
    PIXR               *prImageStrip)
{
    PROLOGUE;
    Int32 w, h, depth, diff;
    PIXR *prPrimary = NULL, *prCleaned = NULL, *prTmp = NULL;
    PIXR *prImage = NULL;
    AutoScanState *pState = (AutoScanState *) pScanState;

    /* Check the strip */
    ERR_GEN_PIXR (prImageStrip, cByteOrderBig, pState->depth2, 
		  cChannels1, cFrame32);

    if(ENABLESECONDPASSRECT( pState->secondPassCropFlag ))
    {
    /* The second pass image was not cropped */
	ERR_PASS (w_pixrGetWidth (pIPWState, prImageStrip, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImageStrip, &h));
	pState->secondPassImageY += h;
	diff = pState->secondPassImageY - pState->secondPassRect.y;
	if( diff <= 0 || (diff -h) >= pState->secondPassRect.h )
	    NORMAL_RETURN;

	ERR_PASS (w_copy (pIPWState, prImageStrip, &prImage));
	if( diff < h )
	{
	    ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
					   pState->secondPassRect.x,
					   (h - diff),
					   pState->secondPassRect.w,
					   diff,
					   prImage));
	}
 	else if( diff > pState->secondPassRect.h )
	{
	    ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
					   pState->secondPassRect.x,
					   0,
					   pState->secondPassRect.w,
					   (h +pState->secondPassRect.h -diff),
					   prImage));
	}
	else
	{
	    ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
					   pState->secondPassRect.x,
					   0,
					   pState->secondPassRect.w,
					   h,
					   prImage));
	}
    }
    else
	prImage = prImageStrip;

    /* 02-09-1998 We need scaling in second pass image. */  
 
    if ( !FEQ(pState->textXScale, 1.0) || !FEQ(pState->textYScale, 1.0))
    {
	Int32   scaledW, scaledH;
	PIXR   *prScaledImage = NULL;

	ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));

	/* Rescale it to the requested text resolution and then binarize */
	scaledW = SCALE32_NR (w, pState->textXScale);
	scaledH = SCALE32_NR (h, pState->textYScale);

	/* 
	 * The "automatic" scale function appears to be too stupid
	 * to use "resConv" in cases where it applies, so check here.
	 */
	if (FEQ(pState->textXScale, 1.5) && 
	    FEQ(pState->textYScale, 1.5) && 
	    pState->depth1 != 1)
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImage, 200, 300, 
					 &prScaledImage));
	}
	else if (FEQ(pState->textXScale, 2.0) &&
		 FEQ(pState->textYScale, 2.0) &&
		 pState->depth1 != 1)
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImage, 100, 200, 
					 &prScaledImage));
	}
	else
	{
	    ERR_PASS (w_scale (pIPWState, prImage, scaledW, scaledH, cFalse, 
			   IP_HI_QUALITY, &prScaledImage));
	}
	if(prImage != prImageStrip)
	    ERR_PASS(w_destroyPixr (pIPWState, &prImage));
	prImage = prScaledImage;
    }

    if (pState->depth2 != 1)
    {
	/* Always take gray for thresholding */
	ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, prImage,
					   &prPrimary));

	/* Binarize it */
	if (ENABLED(cAutoOption_binarize2))
	{
	    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prPrimary, &depth));
	    if (depth == 4)
	    {
		ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prPrimary, &prTmp));
		
		ERR_PASS_CLEANUP (w_binarize2Band (pIPWState,
						   pState->pBinState, 
						   prTmp, NULL));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_binarize2Band (pIPWState, 
						   pState->pBinState, 
						   prPrimary, NULL));
	    }
	}
	else
	{
	    ERR_PASS (w_binarizeBand (pIPWState, pState->pBinState, 
				      prPrimary));
	}
	ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));
	pState->passY += h;
    }
    else
    {
	/* Copy it in */
	ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));

	ERR_GEN (w != pState->width2, cErrApps_differentDimensions);
	ERR_GEN (h + pState->passY > pState->height2, cErrApps_tooMuchData);

	if (ENABLED(cAutoOption_pepperRemove))
	{
	    /*
	     * If we're not binarizing, and one wants to remove pepper,
	     * do it here.
	     */
	    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prImage, 4, 
					      &prCleaned));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prCleaned, PIX_SRC, 
					  0, pState->textY, w, h, 
					  pState->prTextPlane));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	}
	else
	{
	    /* Just copy it */
	    ERR_PASS (w_rasterop (pIPWState, 0, 0, prImage, PIX_SRC, 
				  0, pState->textY, w, h, 
				  pState->prTextPlane));
	}
	pState->passY += h;
	pState->textY += h;
    }

    if (((pState->passY / h) % (pState->depth2 == 1 ? 2 : 6)) == 1)
    {
	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->prTextPlane));
    }

    if(prImage != NULL && prImage != prImageStrip)
	ERR_PASS(w_destroyPixr (pIPWState, &prImage));

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (prCleaned)
	w_destroyPixr (pIPWState, &prCleaned);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if( prImage && prImage != prImageStrip )
	w_destroyPixr (pIPWState, &prImage);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Process whole "pass 2" data through the "autoscan" pipeline.
 *
 * LongDescription:::
 *   This function takes whole pass-2 data, maybe resizes if needed,
 *   binarizes, and stuffs it into the "text" plane at the right spot.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanTwoPassStart()
 *  prImage::        Whole image data to process   
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanSecondTransData (
    IPWState          *pIPWState,
    void               *pScanState,
    PIXR               *prImage)
{
    PROLOGUE;
    Int32 w, h, depth;
    PIXR *prPrimary = NULL, *prCleaned = NULL, *prTmp = NULL;
    PIXR *prImagePlane = NULL;
    AutoScanState *pState = (AutoScanState *) pScanState;

    /* Check the strip */
    ERR_GEN_PIXR (prImage, cByteOrderBig, pState->depth2, 
		  cChannels1, cFrame32);

    if(ENABLESECONDPASSRECT( pState->secondPassCropFlag ))
    {
    /* The second pass image was not cropped */
	ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));
	pState->secondPassImageY = h;

	ERR_PASS (w_copy (pIPWState, prImage, &prImagePlane));

	ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
				       pState->secondPassRect.x,
				       pState->secondPassRect.y,
				       pState->secondPassRect.w,
				       pState->secondPassRect.h,
				       prImagePlane));
    }
    else
	prImagePlane = prImage;

    /* 02-09-1998 We need scaling in second pass image. */  
 
    if ( !FEQ(pState->textXScale, 1.0) || !FEQ(pState->textYScale, 1.0))
    {
	Int32   scaledW, scaledH;
	PIXR   *prScaledImage = NULL;

	ERR_PASS (w_pixrGetWidth (pIPWState, prImagePlane, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImagePlane, &h));

	/* Rescale it to the requested text resolution and then binarize */
	scaledW = SCALE32_NR (w, pState->textXScale);
	scaledH = SCALE32_NR (h, pState->textYScale);

	/* 
	 * The "automatic" scale function appears to be too stupid
	 * to use "resConv" in cases where it applies, so check here.
	 */
	if (FEQ(pState->textXScale, 1.5) && 
	    FEQ(pState->textYScale, 1.5) && 
	    pState->depth1 != 1)
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImagePlane, 200, 300, 
					 &prScaledImage));
	}
	else if (FEQ(pState->textXScale, 2.0) &&
		 FEQ(pState->textYScale, 2.0) &&
		 pState->depth1 != 1)
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImagePlane, 100, 200, 
					 &prScaledImage));
	}
	else
	{
	    ERR_PASS (w_scale (pIPWState, prImagePlane, scaledW, scaledH, 
			       cFalse, IP_HI_QUALITY, &prScaledImage));
	}
	if(prImagePlane != prImage)
	    ERR_PASS(w_destroyPixr (pIPWState, &prImagePlane));
	prImagePlane = prScaledImage;
    }

    if (pState->depth2 != 1)
    {
	/* Always take gray for thresholding */
	ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, prImagePlane,
					   &prPrimary));

	prCleaned = NULL;
	ERR_PASS (w_pixrGetWidth (pIPWState, prImagePlane, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImagePlane, &h));

	/* Binarize it */
	if (ENABLED(cAutoOption_binarize2))
	{
	    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prPrimary, &depth));
	    if (depth == 4)
	    {
		ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prPrimary, &prTmp));

		ERR_PASS_CLEANUP (w_binarize2 (pIPWState, 
					       prTmp, 0,
					       cLowResGradientThreshold,
					       &prCleaned));		

		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		/* Copy it in */
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
					      prCleaned, PIX_SRC, 
					      0, 0, w, h,  
					      pState->prTextPlane));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_binarize2 (pIPWState, 
					       prPrimary, 0,
					       cLowResGradientThreshold,
					       &prCleaned));
		/* Copy it in */
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
					      prCleaned, PIX_SRC, 
					      0, 0, w, h,  
					      pState->prTextPlane));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	    }
	}
	else
	{
	    ERR_PASS (w_binarize (pIPWState, prPrimary,
				  0, NULL, cTrue, cFalse,
				  ENABLED(cAutoOption_pepperRemove) ? 
				  cTrue:cFalse,
				  &prCleaned));
	    /* Copy it in */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
					  prCleaned, PIX_SRC, 
					  0, 0, w, h,  
					  pState->prTextPlane));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	}

	pState->passY = h;
    }
    else
    {
	/* Copy it in */
	ERR_PASS (w_pixrGetWidth (pIPWState, prImagePlane, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImagePlane, &h));

	ERR_GEN (w != pState->width2, cErrApps_differentDimensions);
	ERR_GEN (h  > pState->height2, cErrApps_tooMuchData);

	if (ENABLED(cAutoOption_pepperRemove))
	{
	    /*
	     * If we're not binarizing, and one wants to remove pepper,
	     * do it here.
	     */
	    ERR_PASS_CLEANUP (w_removePepper (pIPWState, prImagePlane, 4, 
					      &prCleaned));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prCleaned, PIX_SRC, 
					  0, pState->textY, w, h, 
					  pState->prTextPlane));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	}
	else
	{
	    /* Just copy it */
	    ERR_PASS (w_rasterop (pIPWState, 0, 0, prImagePlane, PIX_SRC, 
				  0, pState->textY, w, h, 
				  pState->prTextPlane));
	}
	pState->passY = h;
	pState->textY = h;
    }

    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prTextPlane));

    if(prImagePlane != NULL && prImagePlane != prImage)
	ERR_PASS(w_destroyPixr (pIPWState, &prImagePlane));

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (prCleaned)
	w_destroyPixr (pIPWState, &prCleaned);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if( prImagePlane && prImagePlane != prImage )
	w_destroyPixr (pIPWState, &prImagePlane);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Pass 2 end-of-data marker
 *
 * LongDescription:::
 *   This marks the end of pass-two data. We should have gotten
 *   as much as the caller said we would. Now, if needed, we must
 *   do a few things to the binarized image. Pepper removal allegedly already
 *   happened as the data was binarized. We need to, if requested,
 *   deskew, remove stipple, and remove rules.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanTwoPassStart()
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanSecondPassComplete (
    IPWState          *pIPWState,
    void               *pScanState)
{
    PROLOGUE;
    PIXR *prTmp = NULL;
    AutoScanState *pState = (AutoScanState *) pScanState;
    Int32   strelhID = 0, strelvID = 0;
    STREL   *pTheStrel = NULL;
    PIXR    *prImage=NULL, *prImageStrip=NULL, *prBinStrip =NULL;
    PIXR    *prText=NULL, *prTextStrip=NULL;
    Int32   stripHeight, imgW, imgH, textW, textH;
    Int32   type, channels, depth, w, h, scaledHeight, i;
    Float32 scaleX, scaleY;
    Int32   tsY, tsX, psY, psX;

    /* Finish off the binarization */
    if (pState->depth2 != 1)
    {
	if (ENABLED (cAutoOption_binarize2))
	{
	    ERR_PASS (w_binarize2Finish (pIPWState, &pState->pBinState));
	}
	else
	{
	    ERR_PASS (w_binarizeFinish (pIPWState, &pState->pBinState));
	}
    }
    
    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prTextPlane));

    if (! ENABLED(cAutoOption_mapInkAndPage) && 
	pState->prBackgroundMap)
	ERR_PASS(w_destroyPixr (pIPWState, &pState->prBackgroundMap));

    /* If scanner dose not do cropping, we do cropping strip by strip.
       To be safe, here return to no cropping status */
    ERR_PASS (w_autoScanSetSecondPassRect(pIPWState, pState, 0, 0, NULL));

    /* Might have to auto-orient */
    /* We won't do it in place, reasoning that 1MB of memory
       isn't bad compared to the amount we used during segmentation...*/
    SET_STATE (cAutoState_orientCorrect);
    switch (pState->orientation)
    {
    case 270:
    case 90:
    case 180:
	ERR_PASS (w_autoProcOrientImage (pIPWState, pState->orientation, 
				  cFalse, &pState->prTextPlane));
	break;

    case 0:
	/* Nothing happens */
	break;

    default:
	ERR_GEN_CLEANUP (1, ia_internal);
    }

    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prTextPlane));

    /* Deskew if asked */
    if (ENABLED (cAutoOption_deskew))
    {
	SET_STATE (cAutoState_deskew);
	/* Do it in-place */
	if (fabs(pState->skew) > cSmallestRotatingAngle)
	{
	    ERR_PASS_CLEANUP (w_rotateArb (pIPWState, pState->prTextPlane, 
					   pState->skew, cTrue, cTrue, 
					   &pState->prTextPlane));
	    if (ENABLED (cAutoOption_mapInkAndPage))
		ERR_PASS_CLEANUP (w_rotateArb (pIPWState, 
					       pState->prBackgroundMap, 
					       pState->skew, cTrue, cTrue, 
					       &pState->prBackgroundMap));
	}
	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->prTextPlane));

    }
    if (ENABLED (cAutoOption_crop))
    {
	if( ENABLED (cAutoOption_deskew) || pState->orientation != 0)
	{
	    Float32 expX, expY;

	    if (ENABLED (cAutoOption_mapInkAndPage))
		ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
					       pState->boundingBox.x,
					       pState->boundingBox.y,
					       pState->boundingBox.w,
					       pState->boundingBox.h,
					       pState->prBackgroundMap));

	    expX = (Float32) (ScaledInt1616_TO_Float32(pState->textXRes) / 
			      ScaledInt1616_TO_Float32(pState->xRes1));
	    expY = (Float32) (ScaledInt1616_TO_Float32(pState->textYRes) / 
			      ScaledInt1616_TO_Float32(pState->yRes1));
	    ERR_PASS_CLEANUP ( w_autoProcScaledCrop (pIPWState,
						     &pState->boundingBox,
						     expX, expY,
						     pState->prTextPlane));
	}
    }


    /* Do the foreground/background mode */
    if (ENABLED (cAutoOption_mapInkAndPage))
    {
   	SET_STATE (cAutoState_makeInkPlanes);

	prImage = pState->prBackgroundMap;
	pState->prBackgroundMap = NULL;
   	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prImage, &imgW));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prImage, &imgH));
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pState->prTextPlane, 
					  &textW));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pState->prTextPlane, 
					   &textH));
	
	ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prImage, &depth));
	ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prImage, 
						 &channels));
	ERR_PASS_CLEANUP (w_pixrGetType (pIPWState, prImage, &type));

	ERR_PASS_CLEANUP (wp_getAlignment (pIPWState, pState, 
					   pState->prTextPlane,
					prImage, &tsY, &tsX, &psY, &psX));
	if( psY )
	{
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      0, psY, prImage, PIX_SRC, 
				      0, 0, imgW, imgH-psY, 
				      prImage));
	}
	else if (psX )
	{
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      psX, 0, prImage, PIX_SRC, 
				      0, 0, imgW-psX, imgH, 
				      prImage));
	}
	if (ENABLED (cAutoOption_noForeground))
	    pState->prForegroundMap = NULL;
	else
	{
	    w = SCALE32_NR (imgW, pState->fgXScale);
	    h = SCALE32_NR (imgH, pState->fgYScale);
	    ERR_PASS_CLEANUP (w_createPixrExtended (pIPWState, w, h, 
			     depth, pState->fgXRes, pState->fgYRes, 
			     cFrameBits, channels, type, NULL, 
			     &(pState->prForegroundMap)));
	}
	if (ENABLED (cAutoOption_noBackground))
	{
	    prText = NULL;
	    pState->prBackgroundMap = NULL;
	}
	else
	{
	    ERR_PASS_CLEANUP (w_textMask (pIPWState, pState->prTextPlane, 
				      &prText, 0));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pState->prTextPlane, 
				     AND, prText));
	    w = SCALE32_NR (imgW, pState->bgXScale);
	    h = SCALE32_NR (imgH, pState->bgYScale);
	    ERR_PASS_CLEANUP (w_createPixrExtended (pIPWState, w, h, 
				   depth, pState->bgXRes, pState->bgYRes, 
				   cFrameBits, channels, type, NULL, 
				   &(pState->prBackgroundMap)));
	}

   	stripHeight = (cBandHeight +2)*2;
	scaleX = ScaledInt1616_TO_Float32(pState->pictXRes) /
                     ScaledInt1616_TO_Int32(pState->textXRes);
	scaleY = ScaledInt1616_TO_Float32(pState->pictYRes) /
                     ScaledInt1616_TO_Int32(pState->textYRes);

	scaledHeight = SCALE32_NR (stripHeight, scaleY);
	pState->textY = tsY;
	pState->passY = 0; 
	pState->bgY =0;
	pState->realBGY = (Float32)0;
   	prBinStrip = prImageStrip = prTextStrip = NULL;
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, textW, stripHeight,
					1, pState->textXRes, pState->textYRes,
					cFrameBits, &prBinStrip));
	ERR_PASS_CLEANUP (w_createPixrExtendedNoInit (pIPWState,
						      imgW, scaledHeight, 
						      depth, pState->pictXRes,
						      pState->pictYRes, 
						      cFrameBits, channels,
						      type, NULL, 
						      &prImageStrip));
	if( prText )
	{
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, textW, 
						  stripHeight,
					      1, pState->textXRes,
					      pState->textYRes,
					      cFrameBits, &prTextStrip));
	}
	while( pState->textY < textH )
	{
	    if( (textH - pState->textY) < stripHeight )
	    {
         	stripHeight = textH - pState->textY;
		scaledHeight = SCALE32_NR (stripHeight, scaleY);
/* 3-20-98 stripHeight >0 not guarantee scaleHeight >0, it will cause crash */ 
		if( scaledHeight < 1) break;

		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImageStrip));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBinStrip)); 
		if( prTextStrip )
		    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &prTextStrip)); 

		prBinStrip = prImageStrip = prTextStrip =NULL;
		ERR_PASS_CLEANUP (w_createPixr (pIPWState,
					textW, stripHeight,
					1, pState->textXRes, pState->textYRes,
					cFrameBits, &prBinStrip));
		ERR_PASS_CLEANUP (w_createPixrExtendedNoInit (pIPWState, imgW, 
					scaledHeight, depth, pState->pictXRes,
					pState->pictYRes, cFrameBits, channels,
					type, NULL, &prImageStrip));
		if( prText )
		{
		    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					textW, stripHeight,
					1, pState->textXRes, pState->textYRes,
					cFrameBits, &prTextStrip));
		}
	    }
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      tsX, pState->textY, 
				      pState->prTextPlane, PIX_SRC, 
				      0, 0, (textW-tsX), stripHeight, 
				      prBinStrip));

	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
                  0, pState->passY, prImage, PIX_SRC, 
					  0, 0, imgW, scaledHeight, 
					  prImageStrip));

	    if( prText )
	    {
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
					  0, pState->textY, prText, PIX_SRC, 
					  0, 0, textW, stripHeight, 
					  prTextStrip));
	    }

	    ERR_PASS_CLEANUP (wp_twoPassMapink(pIPWState, pState,
                           prBinStrip, prImageStrip, prTextStrip));

	    pState->textY += stripHeight;
	    pState->passY = SCALE32_NR (pState->textY, scaleY);
	    
	}
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBinStrip));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImageStrip));
	if( prTextStrip )
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextStrip));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImage));
	if( prText )
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));
	/* Propagate the bits in the foreground plane */
	if( pState->prForegroundMap )
	{
	    ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel(pIPWState, 
						      pState->prForegroundMap,
						      &prTmp));
	    prBinStrip = NULL;
	    ERR_PASS_CLEANUP (w_threshold (pIPWState, prTmp,
					   2, 255, THRESH_NORM, &prBinStrip));
	    prTmp = NULL;

		/* propagate the black stuff out */
	    ERR_PASS_CLEANUP (w_grayPropagate4Mask (pIPWState,
						    prBinStrip,
						    cTrue,
						    255,
						    pState->prForegroundMap));
	/* open foreground */
	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strel7h));
	    ERR_PASS (w_getStrel(pIPWState, strel7h, &pTheStrel));	
	    ERR_PASS_CLEANUP (w_openGray (pIPWState, pState->prForegroundMap, 
					pTheStrel, &prTmp));
	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strel7v));
	    ERR_PASS (w_getStrel(pIPWState, strel7v, &pTheStrel));
	    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, prTmp, 
					pTheStrel, &pState->prForegroundMap));

	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strel5v));
	    ERR_PASS (w_getStrel(pIPWState, strel5v, &pTheStrel));
	    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, 
					    pState->prForegroundMap, 
					    pTheStrel, &prTmp));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
					     &pState->prForegroundMap));
	    pState->prForegroundMap = prTmp;
	    prTmp = NULL;
#if 1
	    prImageStrip = NULL;
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prBinStrip,
					1, 1, 1, 0, &prImageStrip));
	    pTheStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel3));
	    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel3, &pTheStrel));
	    ERR_PASS_CLEANUP (w_dilate (pIPWState, prImageStrip, pTheStrel,
				        &prTmp));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImageStrip));

	    ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prTmp));
	    ERR_PASS_CLEANUP (w_expand (pIPWState, prTmp, 8, &prImageStrip));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

	    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState,
						     pState->prForegroundMap, 
						     &channels));
		/* I supposed someone should fix w_opGrayBin sometime
		 * so it does all channels of the input image.
		 */
	    for (i = 0; i < channels; i++)
	    {
	        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState,
						    pState->prForegroundMap,
						    i, &prTmp));

	        ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prTmp, prImageStrip, 
					       PIX_SRC | PIX_DST,
					       &prTmp));
	    }
	    prTmp = NULL;	/* just to make sure */
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImageStrip));
#endif
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBinStrip));
	}
	if( pState->prBackgroundMap )
	{
	    ERR_PASS (w_pixrGetHeight(pIPWState, pState->prBackgroundMap, &h));
	    ERR_PASS (w_pixrGetWidth (pIPWState, pState->prBackgroundMap, &w));
	    while (h > pState->bgY)
	    {
		ERR_PASS_CLEANUP (w_rasterop (pIPWState,
				     0, pState->bgY - 1, 
				     pState->prBackgroundMap,
				     PIX_SRC,
				     0, pState->bgY, w, 1,
				     pState->prBackgroundMap));
		pState->bgY++;
	    }
	    ERR_PASS_CLEANUP (w_grayPropagate4 (pIPWState, 
					    pState->prBackgroundMap));
	}
	/* Dilate the background */
	if( scaleX < 0.5 && scaleY < 0.5 )
	{
	    strelhID = strel7h;
	    strelvID = strel7v;
	}
	else if( scaleX < 0.95 && scaleY < 0.95 )
	{
	    strelhID = strel6h;
	    strelvID = strel5v;
	}
	else
	{
	    strelhID = strel4h; 
	    strelvID = strel2v;
	}
	if (pState->prBackgroundMap)
	{
	    UInt8    map[256];

	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strel2h));
	    ERR_PASS (w_getStrel(pIPWState, strel2h, &pTheStrel));
	    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, pState->prBackgroundMap, 
					pTheStrel, &prTmp));

	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strelhID));
	    ERR_PASS (w_getStrel(pIPWState, strelhID, &pTheStrel));
	    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, 
					    prTmp, pTheStrel, 
					    &pState->prBackgroundMap));
	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strel3h));
	    ERR_PASS (w_getStrel(pIPWState, strel3h, &pTheStrel));
	    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, pState->prBackgroundMap, 
					pTheStrel, &prTmp));

	    pTheStrel = NULL;
	    ERR_PASS (w_initStrelByName(pIPWState, strelvID));
	    ERR_PASS (w_getStrel(pIPWState, strelvID, &pTheStrel));
	    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, 
					    prTmp, pTheStrel, 
					    &pState->prBackgroundMap));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    prTmp = NULL;
/* 4-15-98  map 255 to 254, left 255 for image masks */
	    for(i=0; i<256; ++i)
		map[i] = (UInt8) i;
	    map[255] = (UInt8) 254;
	    ERR_PASS_CLEANUP (w_map(pIPWState, pState->prBackgroundMap, 
				    &(map[0]), &pState->prBackgroundMap));
	}

	if( psX )
	{
	    for (i = 0; i < pState->imageCount; i++)
	    {
		if( pState->pImages[i].xOffset >= psX )
		{
		    pState->pImages[i].xOffset -= psX; 
		    pState->pMasks[i].xOffset -= psX;
		}
	    }
	}
	else if( psY )
	{
	    for (i = 0; i < pState->imageCount; i++)
	    {
		if( pState->pImages[i].yOffset >= psY )
		{
		    pState->pImages[i].yOffset -= psY;
		    pState->pMasks[i].yOffset -= psY;
		}
	    }
	}
    }

    /* Blast the masks */
    if (ENABLED (cAutoOption_blastMasks))
	ERR_PASS_CLEANUP (wp_autoBlastMasks (pIPWState, pState));
    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prTextPlane));

    /* Paste image to background image */
    if (ENABLED (cAutoOption_mapInkAndPage) &&
	ENABLED (cAutoOption_pastePictures) &&
	pState->prBackgroundMap)
    {
	/* Paste the segmented images onto the background plane.*/
	/* Notes: Perfectscan API need these segmented images to
	   get TRC for background image AIE. So we have to keep
	   them and let the API delet them. Dec. 2, 1997 */
	ERR_PASS_CLEANUP ( w_autoProcPastePict( pIPWState,
						 pState->prBackgroundMap,
						 pState->pImages,
						 pState->pMasks,
						 pState->imageCount));
    }

    /* Document drycleaning */
    ERR_PASS_CLEANUP (wp_autoCleanText (pIPWState, pState));
    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					  pState->prTextPlane));

    if( (pState->pageType == cSpiggotOnly) &&
	ENABLED(cAutoOption_paintTextplane))
	ERR_PASS_CLEANUP (w_autoProcTextplaneForOCRSeg(
	    pIPWState, pState->prTextPlane, pState->prImagePlane,
	    cAutoProc_FullBlacken, &pState->prTextPlane, NULL));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    if (prImage) w_destroyPixr (pIPWState, &prImage);
    if (prText) w_destroyPixr (pIPWState, &prText);
    if (prImageStrip) w_destroyPixr (pIPWState, &prImageStrip);
    if (prTextStrip) w_destroyPixr (pIPWState, &prTextStrip);
    if (prBinStrip) w_destroyPixr (pIPWState, &prBinStrip);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Initialize things for a one-pass "autoscan" processing pipeline.
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function., and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "one-pass" scan involves sending a single image
 *   through the pipeline: a higher-resolution color or gray image, and
 *   peeling off a low-resolution color image and a high-resolution
 *   binarized image as the bands come through.
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   <BR>
 *   The PerfectScan API uses a fairly simple data structure to represent
 *   composite pages, which is called the "segment list." The segment list
 *   is a list of images, with the convention that the first image is
 *   the "text plane" and subsequent images are "pictures." The subsequent
 *   images are placed into the composite image space using the text plane
 *   image as a coordinate space.
 *   <BR>
 *   Unlike in the PerfectScan API, I use a slightly more complicated
 *   set of data structures called the "Container", as defined in
 *   "page.c". See w_newPage() and associated functions for details,
 *   but the upshot is that it provides a composite page container structure
 *   with accessor and mutator functions, which allows more fancy page
 *   descriptions than is allowed in the PS API. This incarnation of the
 *   two-pass autoscan pipeline will always put out something which is
 *   easily translated into a PS API data structure--the text plane will
 *   always be first. However, I use an absolute coordinate space in units
 *   of Micas (1/2540th of an inch), which I think is a bit cleaner.
 *   <BR>
 *   The "thumbnail" image, if provided by the caller (i.e. non-NULL)
 *   will be updated with a current view of the image as the scan
 *   and processing progresses. Note that if the thumbnail passed is 
 *   large, this could take a significant amount of time.
 *   <BR>
 *   If the "mapInkAndPage" option is set in the options word,
 *   then we create an "image sandwich", which is a map of the colors
 *   to use at each position in the image for text color and background
 *   color. In other words, the resulting (high resolution) binary
 *   image becomes a high resolution selector between these two low-resolution
 *   color images. This allows us to make a compact representation of
 *   a full color page, including colored text and background. NOTE that
 *   currently, we only do this option in ONE-PASS scanning mode. 
 *   This is because in two-pass mode, we don't have enough information
 *   to get the foreground text color. 
 *
 * PrivateDescription:::
 *   There are, I suppose, trade offs between the two page representation
 *   approaches, but since I want to be able to use this set of functions 
 *   to produce "DIR" images, where the binary image selects between 
 *   reduced-resolution fg/bg ink maps, I need some additional flexibility.
 *   It also may be that eventually, we don't want a single text plane,
 *   but want separate text block/paragraph objects. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  width::          Width of incoming scan
 *  height::         Total height of incoming scan
 *  xRes::           X-resolution, in DPI of input image 
 *  yRes::           Y-resolution, in DPI, of input image
 *  depth::          Depth, in bits/pixel, of input image
 *  pageType::       Requested page type, i.e. cTextOnly, cPictureOnly,
 *                   cTextPicture
 *  textXRes::       Requested X-resolution, in DPI, of text in output image
 *  textYRes::       Requested Y-resolution, in DPI, of text in output image
 *  pictXRes::       Requested X-resolution, DPI, of pictures in output image
 *  pictYRes::       Requested Y-resolution, DPI, of pictures in output image
 *  pictChannels::   Requested # of channels in picture images
 *  pictType::       Requested color type of picture images
 *  pictMap::        If picture images are to be color-mapped, use this map
 *  fgXRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the foreground plane at this X-resolution
 *  fgYRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the foreground plane at this Y-resolution
 *  bgXRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the background plane at this X-resolution
 *  bgYRes::         If making "image sandwich" (cAutoOption_mapInkAndPage)
 *                   then make the background plane at this Y-resolution
 *  processingOptions:: Processing flags, see autoscan.pub
 *  prThumbnail::    If non-NULL, this is the thumbnail picture which
 *                   is to be updated periodically. The thumbnail must
 *                   either be 24-bit RGB or 8-bit gray.
 *  
 * Output:::
 *  ppScanState::    Autoscan state handle
 *  pStripHeight::   Strips passed  should be this high, unless on the last 
 *                   band, in which case they can be smaller
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanOnePassStart (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    ScaledInt1616       xRes,
    ScaledInt1616       yRes,
    Int32               depth,
    Int32               pageType,
    ScaledInt1616       textXRes,
    ScaledInt1616       textYRes,
    ScaledInt1616       pictXRes,
    ScaledInt1616       pictYRes,
    Int32               pictChannels,
    ColorType           pictType,
    ColorMap           *pictMap,
    ScaledInt1616       fgXRes,
    ScaledInt1616       fgYRes,
    ScaledInt1616       bgXRes,
    ScaledInt1616       bgYRes,
    UInt32              processingOptions,
    PIXR               *prThumbnail,
    void              **ppScanState,
    Int32              *pStripHeight)
{
    PROLOGUE;
    AutoScanState      *pState = NULL;
    Int32               w, h, expansion;

    /* Check parameters */
    ERR_GEN (*ppScanState != NULL, cErrNullPtr);

    /* All resolutions must be specified */
    ERR_GEN (xRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (yRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (textXRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (textYRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (pictXRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (pictYRes <= 0, cErrApps_invalidResolution);
    ERR_GEN (depth != 1 && depth != 4 && depth != 8, cErrPixr_depth);


    /* Create a state structure */
    pState = (AutoScanState *) IPW_MALLOC (sizeof (AutoScanState),
					   cAllocatorData);
    ERR_GEN (!pState, cErrNomem);
    MEMSET((void *)pState, 0, sizeof(AutoScanState));
    pState->prThumbnail = prThumbnail;
    
    /* Fill it in */
    pState->width1 = width;
    pState->height1 = height;
    pState->xRes1 = xRes;
    pState->yRes1 = yRes;
    pState->pageType = pageType;
    pState->depth1 = depth;
    pState->orientation = 0;
    pState->skew = 0.0F;
    pState->prImagePlane = NULL;

    /* Set bounding boxes so that if we don't crop, it still works */
    pState->initialBoundingBox.x = 0;
    pState->initialBoundingBox.y = 0;
    pState->initialBoundingBox.w = width;
    pState->initialBoundingBox.h = height;
    pState->boundingBox.x = 0;
    pState->boundingBox.y = 0;
    pState->boundingBox.w = width;
    pState->boundingBox.h = height;

    /* Create the text plane */
    pState->textXScale = ScaledInt1616_TO_Float32(textXRes) / 
	                 ScaledInt1616_TO_Float32(xRes);
    pState->textYScale = ScaledInt1616_TO_Float32(textYRes) /
	                 ScaledInt1616_TO_Float32(yRes);
    SET_STATE (cAutoState_unknown);

    w = SCALE32_NR (width, pState->textXScale);
    h = SCALE32_NR (height, pState->textYScale);
    pState->textXRes = textXRes;
    pState->textYRes = textYRes;
    pState->prTextPlane = NULL;
    pState->processingFlags = processingOptions;
    if (ENABLED (cAutoOption_mapInkAndPage))
    {
	/* Must use binarize2 */
	ERR_GEN (!ENABLED (cAutoOption_binarize2), cErrTypeNotSupported);
	/* If set both no-foreground and no-background, it is equal to
	   no map-ink */
	ERR_GEN (ENABLED (cAutoOption_noForeground) &&
		 ENABLED (cAutoOption_noBackground), cErrTypeNotSupported);
	/* Page type can't be cPictureOnly or cTextOnly */
	ERR_GEN (pageType == cPictureOnly, cErrTypeNotSupported);
	ERR_GEN (pageType == cTextOnly, cErrTypeNotSupported);
	/*
	ERR_GEN (pageType == cSpiggotOnly, cErrTypeNotSupported);
	*/
	/* Foreground resolution must be less or equal text resolution */
	ERR_GEN (textXRes < fgXRes, cErrApps_invalidResolution);
	ERR_GEN (textYRes < fgYRes, cErrApps_invalidResolution);
	ERR_GEN (depth != 8, cErrPixr_depth);

	if (ENABLED (cAutoOption_pastePictures))
	{
	    /* Must use the same resolution for BG and pictures in this case */
	    if ((ABS(pictXRes - bgXRes) > 32768L) || 
		(ABS(pictYRes - bgYRes) > 32768L))
	    {
		ERR_GEN (1, cErrApps_invalidResolution);
	    }
	}
    }

    /* There might not be any text, so clear the text plane when we
     * allocate it.  It might not be touched again.
     */
    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 1, 
				    textXRes, textYRes,
				    cFrameBits, &(pState->prTextPlane)));

    /* 
     * Set up banded binarize and the image plane.
     * We binarize at the target image's resolution.
     * Note that we only make an image plane if the image passed
     * is not binary. I suppose we could do it anyway, and
     * scale-to-gray the binary as it came in.
     */
    if (depth != 1)
    {
	pState->pBinState = NULL;
	if ( pState->pageType != cPictureOnly )
	{
	    if (ENABLED (cAutoOption_binarize2))
	    {
		expansion = 0;
		if( pState->textXScale < 0.95 || pState->textXScale > 1.05 )
		    expansion = cIsScaledImage;
		ERR_PASS_CLEANUP (
		    w_binarize2Init (pIPWState, w, h, 8, 
				     MAX(pState->textXRes, pState->textYRes),
				     expansion, 32, &pState->prTextPlane, 
				     &pState->pBinState));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (
		    w_binarizeInit (pIPWState, w, h, depth, 
				    MAX(pState->textXRes, pState->textYRes), 
				    0, NULL,  cTrue, cFalse,
				    ENABLED(cAutoOption_pepperRemove) ? 
				    cTrue:cFalse,
				    &pState->prTextPlane, &pState->pBinState));
	    }
	}

	/* Create the image plane */
	pState->pictXScale = (ScaledInt1616_TO_Float32(pictXRes) /
			      ScaledInt1616_TO_Float32(xRes));
	pState->pictYScale = (ScaledInt1616_TO_Float32(pictYRes) /
			      ScaledInt1616_TO_Float32(yRes));
	pState->pictXRes = pictXRes;
	pState->pictYRes = pictYRes;
	pState->prImagePlane = NULL;
	if ( pState->pageType != cTextOnly )
	{
	    w = SCALE32_NR (width, pState->pictXScale);
	    h = SCALE32_NR (height, pState->pictYScale);
	    ERR_PASS_CLEANUP (w_createPixrExtended (pIPWState, w, h, 
						    depth, pictXRes, pictYRes, 
						    cFrameBits,
						    pictChannels, pictType,
						    pictMap, 
						    &(pState->prImagePlane)));
	}
    }
    else
    {
	pState->pictXScale = pState->pictYScale = 1.0f;
	pState->pictXRes = pictXRes;
	pState->pictYRes = pictYRes;
    }

    /*
     * Initialize for ink/text sandwich mode, if requested
     */
    pState->prForegroundMap = pState->prBackgroundMap = NULL;
    if (ENABLED(cAutoOption_mapInkAndPage))
    {
	pState->fgXScale = (ScaledInt1616_TO_Float32(fgXRes) /
			      ScaledInt1616_TO_Float32(xRes));
	pState->fgYScale = (ScaledInt1616_TO_Float32(fgYRes) /
			      ScaledInt1616_TO_Float32(yRes));
	pState->fgXRes = fgXRes;
	pState->fgYRes = fgYRes;

	w = SCALE32_NR (width, pState->fgXScale);
	h = SCALE32_NR (height, pState->fgYScale);

	/* 
	 * We create it with the same number of channels, type, and color
	 * map as the picture plane. It is conceivable that if we 
	 * get better at coalescing color classes, we might want
	 * the picture plane to be 24-bit, and the ink maps to
	 * be color mapped.
	 */
	if (ENABLED (cAutoOption_noForeground))
	    pState->prForegroundMap = NULL;
	else
	{
	    ERR_PASS_CLEANUP (w_createPixrExtended (pIPWState, w, h, 
						depth, fgXRes, fgYRes, 
						cFrameBits,
						pictChannels, pictType,
						pictMap, 
						&(pState->prForegroundMap)));
	}
	pState->bgXScale = (ScaledInt1616_TO_Float32(bgXRes) /
			      ScaledInt1616_TO_Float32(xRes));
	pState->bgYScale = (ScaledInt1616_TO_Float32(bgYRes) /
			      ScaledInt1616_TO_Float32(yRes));
	pState->bgXRes = bgXRes;
	pState->bgYRes = bgYRes;
	w = SCALE32_NR (width, pState->bgXScale);
	h = SCALE32_NR (height, pState->bgYScale);

	if (ENABLED (cAutoOption_noBackground))
	    pState->prBackgroundMap = NULL;
	else
	{
	    ERR_PASS_CLEANUP (w_createPixrExtendedNoInit (
					pIPWState, w, h, 
					depth, bgXRes, bgYRes, 
					cFrameBits,
					pictChannels, pictType,
					pictMap, 
					&(pState->prBackgroundMap)));

	}
	/*
	 * If possible, we want to do max-reduction as many times as 
	 * possible, then do a scaling. 
	 */
	if (FEQ (pState->bgXScale, pState->bgYScale) &&  
	    pState->bgXScale >0)
	{
	    /* Ansi didn't bother defining log2.. :-p */
	    pState->bgReductions = (Int32) (log (1.0 / pState->bgXScale)/
		                            log (2.0));
	    if (pState->bgReductions > 4)
		pState->bgReductions = 4;
	    if (pState->bgReductions < 0)
		pState->bgReductions = 0;
	    pState->bgScaleAfterReduce = pState->bgXScale *
		(1 << pState->bgReductions);
	}
	else
	{
	    pState->bgReductions = 0;
	}
	pState->bgY = 0;
    }


    /* Set stuff up for the incoming data */
    pState->pass = 0;
    pState->passY = 0;
    pState->imageY = 0;
    pState->textY = 0;
    pState->realImageY = 0.0F;
    pState->realBGY = 0.0F;
    pState->skew = 0.0F;
    pState->pRotateState = NULL;
    pState->prRotatedStrip = NULL;
    pState->imageCount = 0;
    pState->pImages = pState->pMasks = NULL;

    *ppScanState = (void *) pState;

    /* How much data do we want in a strip? We want the amount that
     * will give us cBandHeight lines in the text plane.
     * NOTE that this is actually quite bogus, and we should write
     * a real banded expand, maybe rolled in with the binarize, maybe not.
     * (Fractional output lines will and do cause artifacts; you can see
     * them in the old PerfectScan, though surprisingly nobody but
     * me noticed them...)
     */
    if (ENABLED(cAutoOption_binarize2))
    {
	/* Actually, if doing binarize2, we can take more, and it's
	 * better to do so.
	 */
	*pStripHeight = (Int32) ((Float32) (cBandHeight * 2) / 
				 pState->textYScale + 0.5);
    }
    else
    {
	*pStripHeight = (Int32) ((Float32) cBandHeight / 
				 pState->textYScale + 0.5);
    }
    SET_STATE (cAutoState_initialized);
    SET_STATE (cAutoState_processPass1);

    if( pState->pageType == cTextOnly )
	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->prTextPlane));
    else
	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->depth1 == 1 ?
					      pState->prTextPlane :
					      pState->prImagePlane));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pState) 
	IPW_FREE (pState, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Process a hunk of data through the single-pass "autoscan" pipeline.
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function, and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "one-pass" scan involves sending a single high-resolution image
 *   through the pipeline, and converting it to a lower-resolution color
 *   or gray image, and a higher-resolution binary image on the way in.
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   Not all options are implemented. Basically, we expect that the data
 *   passed is the same depth and color type as the requested picture
 *   type.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanOnePassStart()
 *  prImageStrip::    Strip of image data to process   
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanOnePassData (
    IPWState           *pIPWState,
    void               *pScanState,
    PIXR               *prImageStrip)
{
    PROLOGUE;
    Int32                w, h, scaledW, scaledH, y, depth, height;
    Float32              realH;
    AutoScanState       *pState = (AutoScanState *) pScanState;
    PIXR                *prScaledStrip = NULL, *prPrimary = NULL, 
	                *prCleaned = NULL, *prTmp = NULL;
    RECTANGL             binarized;

    ERR_GEN (!pState, cErrNullPtr);
    ERR_GEN (!prImageStrip, cErrNullPtr);

    ERR_PASS (w_pixrGetWidth (pIPWState, prImageStrip, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImageStrip, &h));

	/* First, save a copy of the strip at the picture
	 * resolution.  We'll extract the pictures at
	 * the end of scanning.
	 */
    if (pState->depth1 != 1 && pState->pageType != cTextOnly)
    {
	scaledW = SCALE32_NR (w, pState->pictXScale);
	if (w && (scaledW == 0))
	    scaledW = 1;

	scaledH = SCALE32_NR (h, pState->pictYScale);
	if (h && (scaledH == 0))
	    scaledH = 1;
	realH = (Float32) h * pState->pictYScale;

	/* Squish it down for the low-res version. */

	/* Check for ratios that we know about */
	if (FEQ(pState->pictXScale, (1.0/3.0)) && 
	    FEQ(pState->pictYScale, (1.0/3.0)))
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImageStrip, 300, 100, 
					 &prScaledStrip));
	}
	else if (FEQ(pState->pictXScale, (1.0/2.0)) && 
		 FEQ(pState->pictYScale, (1.0/2.0)))
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImageStrip, 200, 100, 
					 &prScaledStrip));
	}
	else
	{
	    ERR_PASS (w_scale (pIPWState, prImageStrip, scaledW, scaledH,
			       cFalse, IP_HI_QUALITY, &prScaledStrip));
	}

	/* Copy it in */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prScaledStrip, PIX_SRC, 
				      0, pState->imageY, 
				      scaledW, scaledH,  
				      pState->prImagePlane));
	pState->imageY += scaledH;
	pState->realImageY += realH;

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip));
    }

    /*
     * Always take gray for the high-res version. This could
     * be done by converting to luminance, which could get us
     * better results on things like yellow text, but generating
     * luminance is very expensive.
     */

    /* If no scaling is required to produce the text plane
     * directly from the green plane of the input image strip.
     * For mysterious reasons, if no scaling is required,
     * we can take in a 1bpp, 4 bpp or 8 bpp image.  If it's
     * a 1bpp image, we can remove pepper from it.  Otherwise,
     * we refuse.  We assume our thresholding doesn't produce
     * pepper??
     * If scaling is required, we demand an 8 bpp image.
     */
    if (FEQ(pState->textXScale, 1.0) && FEQ(pState->textYScale, 1.0))
    {
	if (pState->depth1 != 1)
	{
	    if ( pState->pageType != cPictureOnly )
	    {
		    /* If we're going to be making a foreground map,
		     * make prScaledStrip point to the (already scaled)
		     * input image.
		     */
		if ((ENABLED (cAutoOption_mapInkAndPage)) &&
		    (pState->prForegroundMap != NULL))
		    prScaledStrip = prImageStrip;

		/* Just binarize */
		ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel (pIPWState,
							   prImageStrip,
							   &prPrimary));
		if (ENABLED (cAutoOption_binarize2))
		{
		    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prPrimary,
						      &depth));
		    if (depth == 4)
		    {
			ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prPrimary,
						       &prTmp));
		
			ERR_PASS_CLEANUP (w_binarize2Band (pIPWState,
							   pState->pBinState, 
							   prTmp, &binarized));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		    }
		    else
		    {
			ERR_PASS_CLEANUP (w_binarize2Band (pIPWState, 
							   pState->pBinState,
							   prPrimary, 
							   &binarized));
		    }
		}
		else
		{
		    ERR_PASS_CLEANUP (w_binarizeBand (pIPWState, 
						      pState->pBinState,
						      prPrimary));
		}
	    }
	}
	else
	{
	    if (ENABLED(cAutoOption_pepperRemove))
	    {
		/*
		 * If we're not binarizing, and one wants to remove pepper,
		 * do it here.
		 */
		ERR_PASS_CLEANUP (w_removePepper (pIPWState, prImageStrip, 4, 
						  &prCleaned));
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prCleaned, 
					      PIX_SRC, 
					      0, pState->textY, w, h, 
					      pState->prTextPlane));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	    }
	    else
	    {
		/* Copy it in */
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prImageStrip,
					      PIX_SRC,
					      0, pState->textY, w, h, 
					      pState->prTextPlane));
	    }
	    pState->textY += h;
	}
    }
    else if ((pState->pageType != cPictureOnly) ||
	(pState->depth1 == 1))
    {
	/* Rescale the input strip to the requested text
	 * resolution and then binarize (unless the input
	 * is already binary).  If we're scanning in PictureOnly
	 * mode, don't bother with all this.
	 * If we're going to generate a foreground
	 * plane later, we might as well scale the entire image up
	 * to high-resolution now.  Otherwise, just scale up the
	 * primary plane.
	 */
	scaledW = SCALE32_NR (w, pState->textXScale);
	scaledH = SCALE32_NR (h, pState->textYScale);

	if ((ENABLED (cAutoOption_mapInkAndPage)) &&
	    (pState->prForegroundMap != NULL))
	{
	    prTmp = prImageStrip;
	}
	else	/* no foreground processing, so just scale the primary plane */
	{
	    ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel (pIPWState, prImageStrip,
						       &prTmp));
	}

	    /* w_scale doesn't seem to use the resolution conversion
	     * functions, which are both fast and do a nice job of
	     * smoothing the output.  Choose them explicitly here.
	     */
	if (FEQ(pState->textXScale, 1.5) && 
	    FEQ(pState->textYScale, 1.5) && 
	    pState->depth1 != 1 )
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prTmp, 200, 300, 
					 &prScaledStrip));
	}
	else
	if (FEQ(pState->textXScale, 2.0) &&
	    FEQ(pState->textYScale, 2.0) && 
	    pState->depth1 != 1 )
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prTmp, 100, 200, 
					 &prScaledStrip));
	}
	else
	if (!FEQ(pState->textXScale, 1.0) ||
	    !FEQ(pState->textYScale, 1.0) )
	{
		/* The size of the text plane differs from the
		 * size of the input by some factor other than
		 * 1.0, 1.5 or 2.0 or the input is binary.
		 * Do scaling the hard way.
		 */
	    scaledW = SCALE32_NR (w, pState->textXScale);
	    scaledH = SCALE32_NR (h, pState->textYScale);
	    ERR_PASS_CLEANUP (w_scale (pIPWState, prTmp, scaledW, 
				       scaledH, cFalse,
				       IP_HI_QUALITY, &prScaledStrip));
	}
	else	/* input and output are the same size.  They may be
		 * gray or binary.
		 */
	{
	    prScaledStrip = prTmp;	/* a lot faster than copying all
					 * the pixels.
					 */
	}
	prTmp = NULL;	/* this pointed to either the entire input image
			 * or one plane of it, so there's nothing to
			 * destroy.
			 */
	    /* Point to the primary plane of the scaled image.  This
	     * is the one from which the selector is computed.
	     */
	ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel (pIPWState, prScaledStrip,
						   &prPrimary));

	    /* Do the real thresholding of the gray data.  We know
	     * that we're not in PictureOnly mode at this point
	     * so don't have to check for it.
	     */
	if (pState->depth1 != 1)
	{
	    if (ENABLED(cAutoOption_binarize2))
	    {
		    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prPrimary,
						      &depth));
		    if (depth == 4)
		    {
			ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prPrimary,
						       &prTmp));
			ERR_PASS_CLEANUP (w_binarize2Band (pIPWState,
							   pState->pBinState, 
							   prTmp, &binarized));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		    }
		    else
		    {
			ERR_PASS_CLEANUP (w_binarize2Band (pIPWState, 
						   pState->pBinState, 
						   prPrimary, 
						   &binarized));
		    }
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_binarizeBand (pIPWState, 
						  pState->pBinState, 
						  prPrimary));
	    }
	}
	else
	{
	    /* Copy it in */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prPrimary,
					  PIX_SRC,
					  0, pState->textY,
					  scaledW, scaledH, 
					  pState->prTextPlane));
	    pState->textY += scaledH;
	}

	    /* If we're not doing foreground processing later,
	     * and prScaledStrip is not pointing to the
	     * input image strip, delete it now.
	     */
	if (!((ENABLED (cAutoOption_mapInkAndPage)) &&
	      (pState->prForegroundMap != NULL)) &&
	    (prScaledStrip != prImageStrip))
	{
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip));
	}
	prPrimary = NULL;	/* true in any case */
    }

	/* We use the position of the selector as the one truth about
	 * where things should be.  If the selector extends down beyond
	 * where the current picture extends, copy the last line of
	 * the image to the line below to make things line up.
	 * The basic assumption here is that the text is always higher
	 * resolution than anything else, so has the most exact
	 * placement.
	 */
    if (pState->depth1 != 1 && pState->pageType != cTextOnly)
    {
	if ( pState->pageType != cPictureOnly )
	{
	    realH = ScaledInt1616_TO_Float32(pState->pictYRes)/
		ScaledInt1616_TO_Float32(pState->textYRes);
	    if ( !ENABLED (cAutoOption_binarize2) )
	    {
		ERR_PASS( w_binarizeGetCurImgH(pIPWState, pState->pBinState,
						&binarized.h) );
		binarized.y =0;
	    }
	    y = (Int32)((Float32)(binarized.y+binarized.h) * realH);
	}
	else y = (Int32) pState->realImageY;
	if (y > pState->imageY)
	{
	    /*
	     * The most disgusting hack you can imagine...replicate lines.
	     * This should be fixed with a real banded interpolator
	     * before something ships to real people. I suppose the
	     * fact that nobody's ever noticed could be a mitigating
	     * factor.
	     */
	    ERR_PASS (w_pixrGetHeight (pIPWState, pState->prImagePlane, 
				       &height));
	    ERR_PASS (w_pixrGetWidth (pIPWState, prImageStrip, &w));
	    scaledW = SCALE32_NR (w, pState->pictXScale);
	    if (w && (scaledW == 0))
		scaledW = 1;
	    if (pState->imageY < height)
	    {
		ERR_PASS_CLEANUP (w_rasterop (pIPWState,
					      0, pState->imageY - 1, 
					      pState->prImagePlane,
					      PIX_SRC,
					      0, pState->imageY, scaledW, 1,
					      pState->prImagePlane));
	    }
	    pState->imageY++;
	}
    }
 
    if (ENABLED (cAutoOption_mapInkAndPage))
    {
	    /* We're doing foreground/background processing. */

	    /* We pass in the input image strip as is.  That's
	     * used to make the background image since it involves
	     * less scaling up and scaling down (which tends to
	     * degrade the image?).  It's also a wee bit faster
	     * to scale from a small image to a tiny image than
	     * it is to scale from a big image to a tiny image.
	     *
	     * We tried to use the same image for the foreground
	     * and background and be very careful about which pixels were
	     * propagated in the foreground, but it didn't work.
	     * Text strokes are just too thin.  If you allow any
	     * light background pixel into the foreground, you
	     * end up propagating them.  You really do
	     * have to pass in the exact image from which the selector
	     * was made (prScaledStrip) and scale exactly the pixels
	     * that are indentified by the selector as being
	     * in the foreground into the foreground plane.
	     *
	     * Copy out the band that we just binarized, so that
	     * we can hack it. That's a bit inefficient, since
	     * the thresholder just did the opposite, but exposing
	     * the actual PIXR seems a bit messy. I don't think
	     * that it will constitute too much of a performance hit.
	     *
	     * Note that at this point, pState->pageType cannot equal
	     * cPictureOnly.  The initialization routine guarantees
	     * that cAutoOption_mapInkAndPage and cPictureOnly do
	     * not both occure.  Therefore, a selector plane exists.
	     * Also, prScaledStrip will point to whatever image was used
	     * to make the selector at this point.
	     */
	prTmp = NULL;
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      binarized.w, binarized.h,
					      1, pState->textXRes,
					      pState->textYRes,
					      cFrameBits, &prTmp));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      binarized.x, binarized.y, 
				      pState->prTextPlane, PIX_SRC, 
				      0, 0, binarized.w, binarized.h, 
				      prTmp));

	y = pState->textY;
	pState->textY = binarized.y; 
	ERR_PASS_CLEANUP (wp_onePassMapink(pIPWState, pState, 
					   prImageStrip, prScaledStrip,
					   prTmp));
	pState->textY = y;

	    /* If prScaledStrip does not point to the input, delete it now.
	     * We can't delete the input image. */
	if (prScaledStrip != prImageStrip)
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    }


    if (((pState->passY / h) % 2) == 0)
    {
	if( pState->pageType == cTextOnly )
	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prTextPlane));
	else
	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->depth1 == 1 ?
						  pState->prTextPlane :
						  pState->prImagePlane));
    }
    pState->passY += h;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prCleaned)
	w_destroyPixr (pIPWState, &prCleaned);
    if (prScaledStrip)
	w_destroyPixr (pIPWState, &prScaledStrip);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Process whole data through the single-pass "autoscan" pipeline.
 *
 * LongDescription:::
 *  This function convert whole data into low resolution image plane
 *  and high resolution text plane. 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap State structure
 *  pScanState::     Scan state structure, as created by
 *                   w_autoScanOnePassStart()
 *  prImage::        Whole image data to process   
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanOneTransData (
    IPWState           *pIPWState,
    void               *pScanState,
    PIXR               *prImage)
{
    PROLOGUE;
    Int32                w, h, scaledW, scaledH, y, depth;
    Float32              realH;
    AutoScanState       *pState = (AutoScanState *) pScanState;
    PIXR                *prScaledStrip = NULL, *prPrimary = NULL, 
	                *prCleaned = NULL, *prTmp = NULL;
    PIXR                *prTmp1 = NULL, *prStrip = NULL;
    RECTANGL             binarized;
    Int32                stripHeight, channels, type, scaledHeight;
    Int32                textH;

    ERR_GEN (!pState, cErrNullPtr);
    ERR_GEN (!prImage, cErrNullPtr);

    ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prImage, &channels));
    ERR_PASS (w_pixrGetType (pIPWState, prImage, &type));

	/* First, save a copy of the strip at the picture
	 * resolution.  We'll extract the pictures at
	 * the end of scanning.
	 */
    if (pState->depth1 != 1 && pState->pageType != cTextOnly)
    {
	scaledW = SCALE32_NR (w, pState->pictXScale);
	if (w && (scaledW == 0))
	    scaledW = 1;

	scaledH = SCALE32_NR (h, pState->pictYScale);
	if (h && (scaledH == 0))
	    scaledH = 1;
	realH = (Float32) h * pState->pictYScale;

	/* Squish it down for the low-res version. */

	/* Check for ratios that we know about */
	if (FEQ(pState->pictXScale, (1.0/3.0)) && 
	    FEQ(pState->pictYScale, (1.0/3.0)))
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImage, 300, 100, 
					 &prScaledStrip));
	}
	else if (FEQ(pState->pictXScale, (1.0/2.0)) && 
		 FEQ(pState->pictYScale, (1.0/2.0)))
	{
	    ERR_PASS_CLEANUP (w_resConv (pIPWState, prImage, 200, 100, 
					 &prScaledStrip));
	}
	else
	{
	    ERR_PASS (w_scale (pIPWState, prImage, scaledW, scaledH,
			       cFalse, IP_HI_QUALITY, &prScaledStrip));
	}

	/* Copy it in */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prScaledStrip, PIX_SRC, 
				      0, 0, scaledW, scaledH,  
				      pState->prImagePlane));
	pState->imageY = scaledH;
	pState->realImageY = realH;

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip));
    }

    /*
     * Always take gray for the high-res version. This could
     * be done by converting to luminance, which could get us
     * better results on things like yellow text, but generating
     * luminance is very expensive.
     */

    /* If no scaling is required to produce the text plane
     * directly from the green plane of the input image strip.
     * For mysterious reasons, if no scaling is required,
     * we can take in a 1bpp, 4 bpp or 8 bpp image.  If it's
     * a 1bpp image, we can remove pepper from it.  Otherwise,
     * we refuse.  We assume our thresholding doesn't produce
     * pepper??
     * If scaling is required, we demand an 8 bpp image.
     */
    if (FEQ(pState->textXScale, 1.0) && FEQ(pState->textYScale, 1.0))
    {
	if (pState->depth1 != 1)
	{
	    if ( pState->pageType != cPictureOnly )
	    {
		/* Just binarize */
		ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel (pIPWState,
							   prImage,
							   &prPrimary));
		prCleaned = NULL;
		if (ENABLED (cAutoOption_binarize2))
		{
		    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prPrimary,
						      &depth));
		    if (depth == 4)
		    {
			ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prPrimary,
						       &prTmp));
			ERR_PASS_CLEANUP (w_binarize2 (pIPWState,
						       prTmp, 0,
						     cLowResGradientThreshold, 
						       &prCleaned));
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
			/* Copy it in */
			ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						      prCleaned, PIX_SRC, 
						      0, 0, w, h,  
						      pState->prTextPlane));
			ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prCleaned));
		    }
		    else
		    {
			ERR_PASS_CLEANUP (w_binarize2 (pIPWState, 
						       prPrimary, 0,
						     cLowResGradientThreshold,
						       &prCleaned));
			/* Copy it in */
			ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						      prCleaned, PIX_SRC, 
						      0, 0, w, h,  
						      pState->prTextPlane));
			ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prCleaned));
		    }
		}
		else
		{
		    ERR_PASS (w_binarize (pIPWState, prPrimary,
					  0, NULL, cTrue, cFalse,
					  ENABLED(cAutoOption_pepperRemove) ? 
					  cTrue:cFalse,
					  &prCleaned));
		    /* Copy it in */
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
						  prCleaned, PIX_SRC, 
						  0, 0, w, h,  
						  pState->prTextPlane));
		    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prCleaned));
		}
	    }
	}
	else
	{
	    if (ENABLED(cAutoOption_pepperRemove))
	    {
		/*
		 * If we're not binarizing, and one wants to remove pepper,
		 * do it here.
		 */
		ERR_PASS_CLEANUP (w_removePepper (pIPWState, prImage, 4, 
						  &prCleaned));
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prCleaned, 
					      PIX_SRC, 
					      0, pState->textY, w, h, 
					      pState->prTextPlane));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prCleaned));
	    }
	    else
	    {
		/* Copy it in */
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prImage,
					      PIX_SRC,
					      0, pState->textY, w, h, 
					      pState->prTextPlane));
	    }
	    pState->textY = h;
	}
	if (ENABLED (cAutoOption_mapInkAndPage))
	{
	    y = pState->textY;
	    pState->textY = 0; 
	    ERR_PASS_CLEANUP (wp_onePassMapink(pIPWState, pState, 
					       prImage, prImage,
					       pState->prTextPlane));
	    pState->textY = y;
	}
    }
    else if ( pState->depth1 == 1 )
    {
	scaledW = SCALE32_NR (w, pState->textXScale);
	scaledH = SCALE32_NR (h, pState->textYScale);

	if (!FEQ(pState->textXScale, 1.0) ||
	    !FEQ(pState->textYScale, 1.0) )
	{
	    /* The size of the text plane differs from the
	     * size of the input by some factor other than
	     * 1.0, 1.5 or 2.0 or the input is binary.
	     * Do scaling the hard way.
	     */
	    ERR_PASS_CLEANUP (w_scale (pIPWState, prImage, 
				       scaledW, 
				       scaledH, cFalse,
				       IP_HI_QUALITY, 
				       &prScaledStrip));
	}
	else
	    prScaledStrip = prImage;
	/* Copy it in */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, 
				      prScaledStrip, PIX_SRC,
				      0, 0, scaledW, scaledH, 
				      pState->prTextPlane));
	pState->textY = scaledH;
	if( prScaledStrip != prImage )
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip));

    }
    else if ( pState->pageType != cPictureOnly )
    {
	/* Rescale the input strip to the requested text
	 * resolution and then binarize (unless the input
	 * is already binary).  If we're scanning in PictureOnly
	 * mode, don't bother with all this.
	 * If we're going to generate a foreground
	 * plane later, we might as well scale the entire image up
	 * to high-resolution now.  Otherwise, just scale up the
	 * primary plane.
	 */
	scaledW = SCALE32_NR (w, pState->textXScale);
	scaledH = SCALE32_NR (h, pState->textYScale);

	if (ENABLED(cAutoOption_binarize2))
	{
	    /* Actually, if doing binarize2, we can take more, and it's
	     * better to do so.
	     */
	    stripHeight = (Int32) ((Float32) (cBandHeight * 2) / 
				     pState->textYScale + 0.5);
	}
	else
	{
	    stripHeight = (Int32) ((Float32) cBandHeight / 
				     pState->textYScale + 0.5);
	}
	scaledHeight = SCALE32_NR (stripHeight, pState->textYScale);

	ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prImage,
					  &depth));
	if ((ENABLED (cAutoOption_mapInkAndPage)) &&
	    (pState->prForegroundMap != NULL))
	{
	    prTmp = prImage;
	    ERR_PASS_CLEANUP (w_createPixrExtendedNoInit (pIPWState,
						      w, stripHeight, 
						      depth, pState->xRes1,
						      pState->yRes1, 
						      cFrameBits, channels,
						      type, NULL, 
						      &prStrip));
	}
	else	/* no foreground processing, so just scale the primary plane */
	{
	    ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel (pIPWState, prImage,
						       &prTmp));
	    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, stripHeight,
					    depth, pState->xRes1, 
					    pState->yRes1,
					    cFrameBits, &prStrip));
	}


	/* Do strip by strip to reduce memory size */

	for(y=0; y<h; y+= stripHeight )
	{
	    if( (h-y) < stripHeight )
	    {
		stripHeight = h-y;
		scaledHeight = SCALE32_NR (stripHeight, pState->textYScale);
		if( scaledHeight < 1) break;

		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prStrip));
		if ((ENABLED (cAutoOption_mapInkAndPage)) &&
		    (pState->prForegroundMap != NULL))
		{
		    ERR_PASS_CLEANUP (w_createPixrExtendedNoInit (pIPWState,
						      w, stripHeight, 
						      depth, pState->xRes1,
						      pState->yRes1, 
						      cFrameBits, channels,
						      type, NULL, 
						      &prStrip));
		}
		else	
		{
		    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, stripHeight,
					    depth, pState->xRes1, 
					    pState->yRes1,
					    cFrameBits, &prStrip));
		}
	    }

	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      0, y, prTmp, PIX_SRC, 
				      0, 0, w, stripHeight, 
				      prStrip));

	    /* w_scale doesn't seem to use the resolution conversion
	     * functions, which are both fast and do a nice job of
	     * smoothing the output.  Choose them explicitly here.
	     */
	    if (FEQ(pState->textXScale, 1.5) && 
		FEQ(pState->textYScale, 1.5) )
	    {
		ERR_PASS_CLEANUP (w_resConv (pIPWState, prStrip, 200, 300, 
					 &prScaledStrip));
	    }
	    else if (FEQ(pState->textXScale, 2.0) &&
		    FEQ(pState->textYScale, 2.0) )
	    {
		ERR_PASS_CLEANUP (w_resConv (pIPWState, prStrip, 100, 200, 
					     &prScaledStrip));
	    }
	    else if (!FEQ(pState->textXScale, 1.0) ||
		     !FEQ(pState->textYScale, 1.0) )
	    {
		/* The size of the text plane differs from the
		 * size of the input by some factor other than
		 * 1.0, 1.5 or 2.0 or the input is binary.
		 * Do scaling the hard way.
		 */
		scaledW = SCALE32_NR (w, pState->textXScale);
		scaledH = SCALE32_NR (h, pState->textYScale);
		ERR_PASS_CLEANUP (w_scale (pIPWState, prStrip, 
					   scaledW, 
					   scaledHeight, cFalse,
					   IP_HI_QUALITY, 
					   &prScaledStrip));
	    }
	    else	
	    {
		prScaledStrip = prStrip;
	    }

	    /* Point to the primary plane of the scaled image.  This
	     * is the one from which the selector is computed.
	     */
	    ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel (pIPWState,
						       prScaledStrip,
						       &prPrimary));

	    /* Do the real thresholding of the gray data.  We know
	     * that we're not in PictureOnly mode at this point
	     * so don't have to check for it.
	     */
	    prCleaned = NULL;
	    if (ENABLED(cAutoOption_binarize2))
	    {
		if (depth == 4)
		{
		    ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prPrimary,
						   &prTmp1));
		    ERR_PASS_CLEANUP (w_binarize2Band (pIPWState,
						       pState->pBinState, 
						       prTmp1, &binarized));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
		}
		else
		{
		    ERR_PASS_CLEANUP (w_binarize2Band (pIPWState,
						       pState->pBinState, 
						       prPrimary, &binarized));
		}
	    }
	    else
	    {
		ERR_PASS (w_binarizeBand (pIPWState, 
					  pState->pBinState, 
					  prPrimary));
	    }

	    if (ENABLED (cAutoOption_mapInkAndPage))
	    {
		prTmp1 = NULL;
		ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      binarized.w, binarized.h,
					      1, pState->textXRes,
					      pState->textYRes,
					      cFrameBits, &prTmp1));

		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
					      binarized.x, binarized.y, 
					      pState->prTextPlane, PIX_SRC, 
					      0, 0, binarized.w, binarized.h, 
					      prTmp1));

		textH = pState->textY;
		pState->textY = binarized.y; 
		ERR_PASS_CLEANUP (wp_onePassMapink(pIPWState, pState, 
						   prImage, prScaledStrip,
						   prTmp1));
		pState->textY = textH;
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

		/* If prScaledStrip does not point to the input, delete it now.
		 * We can't delete the input image. */
	    }
	    if (prScaledStrip != prStrip)
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip));
	    prPrimary = NULL;	/* true in any case */
	}
	if( prStrip )
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prStrip));

	prTmp = NULL;	/* this pointed to either the entire input image
			 * or one plane of it, so there's nothing to
			 * destroy.
			 */
    }

    if( pState->pageType == cTextOnly )
	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->prTextPlane));
    else
	ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
					      pState->depth1 == 1 ?
					      pState->prTextPlane :
					      pState->prImagePlane));

    pState->passY = h;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prCleaned)
	w_destroyPixr (pIPWState, &prCleaned);
    if (prScaledStrip)
	w_destroyPixr (pIPWState, &prScaledStrip);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prTmp1)
	w_destroyPixr (pIPWState, &prTmp1);
    if (prStrip)
	w_destroyPixr (pIPWState, &prStrip);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Signal that the data has all been sent for a one-pass pipeline
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function, and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "one-pass" scan involves sending a single high-resolution image
 *   through the pipeline, and converting it to a lower-resolution color
 *   or gray image, and a higher-resolution binary image on the way in.
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   Not all options are implemented. Basically, we expect that the data
 *   passed is the same depth and color type as the requested picture
 *   type.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::              IPWrap State structure
 *  pScanState::             Scan state structure, as created by
 *                           w_autoScanTwoPassStart()
 *  
 * Output:::
 * 
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanOnePassComplete (
    IPWState  *pIPWState,
    void      *pScanState)
{
    PROLOGUE;
    AutoScanState *pState = (AutoScanState *) pScanState;
    Int32          confidence, wB, hB;
    Int32	   i, res = 0, strelID = 0;
    Int32          reduction, channels;
    PIXR          *prRotated = NULL, *prTmp = NULL, 
	          *prTransfer = NULL, *prPlaneMask = NULL;
    Float32        skew, pictureTextRatio, xScale, yScale;
    PixrRef       *pImages = NULL;
    STREL         *pTheStrel = NULL;
    Bool           isBlank = cFalse;
   
    /* Check parameters */
    ERR_GEN (!pState, cErrNullPtr);

    /* Compute some useful constants */
    pictureTextRatio = ScaledInt1616_TO_Float32 (pState->pictXRes) / 
	               ScaledInt1616_TO_Float32 (pState->textXRes);

    /* Finish off the binarization */
    if (pState->depth1 != 1)
    {
	if ( pState->pageType != cPictureOnly )
	{
	    if (ENABLED(cAutoOption_binarize2))
	    {
		ERR_PASS (w_binarize2Finish (pIPWState, &pState->pBinState));
	    }
	    else
	    {
		ERR_PASS (w_binarizeFinish (pIPWState, &pState->pBinState));
	    }
	}
    }

    /*  4- 17 -98 , check textPlane. */
    /*  5- 6-98 Add condition for pictureonly, since no textPlane for it */
    if ( pState->pageType != cPictureOnly )
	ERR_PASS_CLEANUP (wp_checkBlankPage(pIPWState, pState->prTextPlane, 
				       &isBlank));

    /* Do the finish-up for foreground/background mode */
    if (ENABLED (cAutoOption_mapInkAndPage) && (isBlank == cFalse))
    {
	SET_STATE (cAutoState_makeInkPlanes);

	/* We want to propagate the bits in the foreground plane.
	 * To do that, we need the text mask scaled to the foreground
	 * plane's resolution.
	 */
	if (pState->prForegroundMap)
	{
	    ERR_PASS_CLEANUP (w_pixrGetPrimaryChannel(pIPWState, 
						      pState->prForegroundMap,
						      &prTmp));
	    prPlaneMask = NULL;
	    ERR_PASS_CLEANUP (w_threshold (pIPWState, prTmp,
					   2, 255, THRESH_NORM, &prPlaneMask));
	    prTmp = NULL;

		/* propagate the black stuff out */
	    ERR_PASS_CLEANUP (w_grayPropagate4Mask (pIPWState,
						    prPlaneMask,
						    cTrue,
						    255,
						    pState->prForegroundMap));

#if 1
		/* An interesting fraction of a XIF file is taken up by the
		 * foreground plane.  We'll attempt to make them smaller.
		 * This is a temporary hack in that foreground planes may
		 * well disappear eventually anyway.
		 * We will often (always?) use JPEG to compress the
		 * foreground plane.  JPEG compresses the image in units of
		 * 8x8 pixel tiles.  We'll reduce prPlaneMask by 8x,
		 * invert it and expand it back up, so that each 8x8
		 * tile will be white if there are some valid foreground
		 * bits in it and black if not.  When then "paint" the
		 * mask into the foreground, which turns all non-used
		 * tiles to black.
		 * Well, this works, but doesn't have the gigantic effect
		 * that you'd like.  It might reduce the total XIF file size
		 * by 5-15%.
		 */
	    ERR_PASS_CLEANUP (w_reduce (pIPWState, prPlaneMask,
					1, 1, 1, 0, &prTmp));
	    pTheStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel2));
	    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel2, &pTheStrel));
	    prTransfer = NULL;
	    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmp, pTheStrel,
				        &prTransfer));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    prTmp = prTransfer;
	    prTransfer = NULL;
	    ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prTmp));
	    ERR_PASS_CLEANUP (w_expand (pIPWState, prTmp, 8, &prTransfer));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState,
						     pState->prForegroundMap, 
						     &channels));
		/* I supposed someone should fix w_opGrayBin sometime
		 * so it does all channels of the input image.
		 */
	    for (i = 0; i < channels; i++)
	    {
	        ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState,
						    pState->prForegroundMap,
						    i, &prTmp));

	        ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prTmp, prTransfer, 
					       PIX_SRC | PIX_DST,
					       &prTmp));
	    }
	    prTmp = NULL;	/* just to make sure */
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTransfer));
#endif
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prPlaneMask));
	}
	if (pState->prBackgroundMap)
	{
	    UInt8   map[256];

	    ERR_PASS_CLEANUP (w_pixrGetHeight(pIPWState,
					      pState->prBackgroundMap, &hB));
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState,
					      pState->prBackgroundMap, &wB));

		/* If the background mask isn't quite tall enough,
		 * copy the last line down to the remaining lines.
		 * We'll use the same rasterop trick we used in the
		 * foreground above.
		 */
	    if (hB > pState->bgY)
	    {
		ERR_PASS_CLEANUP (w_rasterop (pIPWState,
					      0, pState->bgY - 1, 
					      pState->prBackgroundMap,
					      PIX_SRC,
					      0, pState->bgY,
					      wB, (hB - pState->bgY),
					      pState->prBackgroundMap));
		pState->bgY = hB;
	    }
	    /* Take text to do dilation and leave others alone. Otherwise
             the dilation will close reverse text gaps, then propogation
             spread out wrong color to those places. */
	    prTmp = NULL;
	    ERR_PASS_CLEANUP (w_textMask (pIPWState, 
					  pState->prTextPlane, 
					  &prTmp,0));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pState->prTextPlane, 
					 AND, prTmp));
	    prPlaneMask = NULL;
	    ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, prTmp, 
					     wB, hB,
					     &prPlaneMask));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

		/* Just to make sure we're away from the edges of the
		 * foreground, dilate the foreground mask a tad.  This
		 * will cause the background pixels to come from farther
		 * away, so will be less tainted by foreground.
		 */
	    pTheStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel3));
	    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel3, &pTheStrel));

	    ERR_PASS_CLEANUP (w_dilate (pIPWState, prPlaneMask, pTheStrel,
				        &prTmp));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prPlaneMask));
	    ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, pState->prTextPlane, 
					     wB, hB,
					     &prPlaneMask));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prPlaneMask, 
					 OR, prTmp));
		/* release the strel as soon as we're done with it.
		 * "releasing" means setting the ptr to NULL.  The
		 * strel manager will take care of it.
		 * We're also done with the prPlaneMask.
		 */
	    pTheStrel = NULL;
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prPlaneMask));

	    ERR_PASS_CLEANUP (w_grayPropagate4Mask (pIPWState, 
						    prTmp,
						    cFalse,
						    255,
						    pState->prBackgroundMap));

		/* One way or another, we're done with the plane mask. */
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

	    /* Dilate the background */
	    if( pState->bgReductions > 0)
	    {
		res = ScaledInt1616_TO_Int32(pState->bgXRes);
		if (res <= 75)
		{
		    strelID = strel3;
		}
		else
		{
		    strelID = strel5;
		}
	    }
	    else strelID = strel2;
	    
	    ERR_PASS (w_initStrelByName(pIPWState, strelID));
	    ERR_PASS (w_getStrel(pIPWState, strelID, &pTheStrel));
	    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, 
					    pState->prBackgroundMap, 
					    pTheStrel, &prTmp));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
					     &pState->prBackgroundMap));
	    pState->prBackgroundMap = prTmp;
/* 4-15-98  map 255 to 254, left 255 for image masks */
	    prTmp = NULL;
	    for(i=0; i<256; ++i)
		map[i] = (UInt8) i;
	    map[255] = (UInt8) 254;
	    ERR_PASS_CLEANUP (w_map(pIPWState, pState->prBackgroundMap, 
				    &(map[0]), &pState->prBackgroundMap));
	}
    }

    SET_STATE (cAutoState_isText);

    /* Do photo or text & photo detection by use of skew detection */
    if( pState->pageType != cPictureOnly && (isBlank == cFalse))
    {
	reduction =4;
	ERR_PASS (w_findSkewSample (pIPWState, pState->prTextPlane, reduction, 
				&confidence, &skew));

    /* Double check text */
	if( (confidence == 0) && (skew == 0.0f))
	{
	    ERR_PASS_CLEANUP( wp_autoSkewDetection (pIPWState, 
						    pState->prTextPlane,
						    pState, &skew, 
						    &confidence));
	}
    }
    else 
    {
	confidence =0;
	skew =0.0f;
    }

/*
 *  w_write (pIPWState, "/tmp/image.tif", pState->prImagePlane, TIFF_UNCOMP);
 *  w_write (pIPWState, "/tmp/text.tif", pState->prTextPlane, TIFF_UNCOMP);
 */
    if ( (skew == 0.0f) && (confidence == 0) )
	pState->IsPhoto = cTrue;
    else
	pState->IsPhoto = cFalse;

    if ( ((skew == 0.0f) && (confidence == 0) && 
	(pState->pageType != cSpiggotOnly)) || (isBlank == cTrue) )
    {
	/* There was no text. What this means is that we should,
	 * if requested, do a photo crop and deskew, and set a flag
	 * such that there will only be a photo on the page.
	 */
	if (pState->depth1 != 1 && pState->pageType != cSpiggotOnly)
	{
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pState->prTextPlane));
	    if (ENABLED (cAutoOption_mapInkAndPage))
	    {
		/* free the foreground and background */
		if (pState->prForegroundMap)
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
						 &pState->prForegroundMap));
		if (pState->prBackgroundMap)
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
						 &pState->prBackgroundMap));
	    }
	}

	/* This routine does: deskew, crop, AIE (and who knows what else): */
	if( isBlank == cFalse )
	    ERR_PASS_CLEANUP (wp_autoPhotoCrop (pIPWState, pState));
    }
    else
    {
	/* Deskew reported some text, so we should act accordingly */
	if( pState->IsPhoto != cTrue )
	    ERR_PASS_CLEANUP (w_autoProcDetections(pIPWState, 
						   pState->prTextPlane,
						   pState->processingFlags,
						   &pState->initialBoundingBox,
						   &pState->orientation,
						   &pState->skew,
						   &confidence,
						   &pState->boundingBox));	
	if (ENABLED (cAutoOption_crop) && (pState->IsPhoto != cTrue ) &&
	    (ENABLED(cAutoOption_deskew) || ENABLED(cAutoOption_orient)))
	{
	    SET_STATE (cAutoState_crop);
	    ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
					     &pState->initialBoundingBox, 
						    1.0F, 1.0F,
						    pState->prTextPlane));
			   

	    if (pState->prImagePlane)
	    {
		ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						 &pState->initialBoundingBox, 
						 pictureTextRatio, 
						 pictureTextRatio,
						 pState->prImagePlane));
	    }
	

	    if (ENABLED (cAutoOption_mapInkAndPage))
	    {
		/* crop the foreground and background */
		if (pState->prForegroundMap)
		{
		    xScale = pState->fgXScale / pState->textXScale;
		    yScale = pState->fgYScale / pState->textYScale;
		
		    ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						 &pState->initialBoundingBox,
						 xScale, yScale, 
						 pState->prForegroundMap));
		}
		if (pState->prBackgroundMap)
		{
		    xScale = pState->bgXScale / pState->textXScale;
		    yScale = pState->bgYScale / pState->textYScale;
		
		    ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						 &pState->initialBoundingBox,
						 xScale, yScale, 
						 pState->prBackgroundMap));
		}
	    }
		
	}

	if( pState->pageType == cTextOnly )
	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prTextPlane));
	else
	    ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->depth1 == 1 ?
						  pState->prTextPlane :
						  pState->prImagePlane));

	/* Check orientation */
	if (ENABLED (cAutoOption_orient) && (pState->IsPhoto != cTrue ))
	{
	    Bool  canInPlace;
	    if (pState->orientation != 0)
	    {    
		canInPlace = ENABLED(cAutoOption_moreMemory) ? cFalse : cTrue;

		SET_STATE (cAutoState_orientCorrect);
		if( pState->prImagePlane )
		    ERR_PASS (w_autoProcOrientImage (pIPWState, 
						 pState->orientation, 
					  canInPlace, &pState->prImagePlane));
		ERR_PASS (w_autoProcOrientImage (pIPWState, 
						 pState->orientation, 
					  canInPlace, &pState->prTextPlane));
		if (ENABLED (cAutoOption_mapInkAndPage))
		{
		    ERR_PASS (w_autoProcOrientImage (pIPWState, 
						     pState->orientation, 
					      canInPlace, 
					      &pState->prForegroundMap));
		    ERR_PASS (w_autoProcOrientImage (pIPWState, 
						     pState->orientation, 
					      canInPlace, 
					      &pState->prBackgroundMap));
		}
	    }
	    if( pState->pageType == cTextOnly )
		ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prTextPlane));
	    else
		ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->depth1 == 1 ?
						  pState->prTextPlane :
						  pState->prImagePlane));
	}

	if (ENABLED (cAutoOption_deskew) && (pState->IsPhoto != cTrue ))
	{

	    /* 
	     * NOTE that we don't inform the caller if there wasn't
	     * enough information to find a skew, i.e. skew = 0.0f and
	     * confidence = 0. It might be useful to have some 
	     * way of notifying the caller, rather than just not 
	     * deskewing at all.
	     */
	    if( fabs(pState->skew) > cSmallestRotatingAngle )
	    {
		if( pState->prImagePlane )
		    ERR_PASS_CLEANUP (w_autoProcDeskewImage (pIPWState, 
						      pState->skew, 1, 
						      &pState->prImagePlane));
		ERR_PASS_CLEANUP(w_autoProcDeskewImage (pIPWState,
							pState->skew, 1, 
							&pState->prTextPlane));

		if (ENABLED (cAutoOption_mapInkAndPage))
		{
		    if (pState->prForegroundMap)
			ERR_PASS_CLEANUP (w_autoProcDeskewImage (pIPWState, 
							  pState->skew, 0, 
						 &pState->prForegroundMap));
		    if (pState->prBackgroundMap)
			ERR_PASS_CLEANUP (w_autoProcDeskewImage (pIPWState, 
							  pState->skew, 0, 
						 &pState->prBackgroundMap));
		}
	    }
	    if( pState->pageType == cTextOnly )
		ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prTextPlane));
	    else
		ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->depth1 == 1 ?
						  pState->prTextPlane :
						  pState->prImagePlane));
	}

	if (ENABLED (cAutoOption_crop) && (pState->IsPhoto != cTrue ))
	{

	    ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						    &pState->boundingBox,
						    1.0F, 1.0F,
						    pState->prTextPlane));
	    
	    if (pState->prImagePlane)
	    {
		ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						 &pState->boundingBox,
						 pictureTextRatio,
						 pictureTextRatio,
						 pState->prImagePlane));
	    }

	    if (ENABLED (cAutoOption_mapInkAndPage))
	    {
		/* crop the foreground and background */
		if (pState->prForegroundMap)
		{
		    xScale = pState->fgXScale / pState->textXScale;
		    yScale = pState->fgYScale / pState->textYScale;
		
		    ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						 &pState->boundingBox,
						 xScale, yScale, 
						 pState->prForegroundMap));
		}
		if (pState->prBackgroundMap)
		{
		    xScale = pState->bgXScale / pState->textXScale;
		    yScale = pState->bgYScale / pState->textYScale;
		
		    ERR_PASS_CLEANUP (w_autoProcScaledCrop (pIPWState, 
						 &pState->boundingBox,
						 xScale, yScale, 
						 pState->prBackgroundMap));
		}
	    }
		
	    if( pState->pageType == cTextOnly )
		ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->prTextPlane));
	    else		    
		ERR_PASS_CLEANUP (wp_updateThumbnail (pIPWState, pState, 
						  pState->depth1 == 1 ?
						  pState->prTextPlane :
						  pState->prImagePlane));
	}
	
	if (pState->prImagePlane)
	{
	    if ( ENABLED(cAutoOption_noSegmentation) ||
		pState->pageType == cSpiggotOnly )	
	    {
		if( pState->pageType != cSpiggotOnly )
		{
		    if (pState->depth1 != 1)
			ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, 
							&pState->prTextPlane));
		    if (ENABLED(cAutoOption_imageEnhance))
			ERR_PASS_CLEANUP (w_autoProcAIE(pIPWState, 0,
							&pState->prImagePlane,
							&pImages));
		}
		else
		{
		/* Document drycleaning */
		    ERR_PASS_CLEANUP (wp_autoCleanText (pIPWState, pState));
		    if (ENABLED(cAutoOption_paintTextplane))
			ERR_PASS_CLEANUP (w_autoProcTextplaneForOCRSeg(
			    pIPWState, pState->prTextPlane,
			    pState->prImagePlane,
			    cAutoProc_FullBlacken,
			    &pState->prTextPlane, NULL));
		}
	    }
	    else
	    {
	    /* Do the segmentation. */
		ERR_PASS_CLEANUP (wp_autoSegmentPage (pIPWState, pState));

	    /* Blast away the masks in the text plane */
		if (ENABLED (cAutoOption_blastMasks))
		    ERR_PASS_CLEANUP (wp_autoBlastMasks (pIPWState, pState));

		if (ENABLED (cAutoOption_mapInkAndPage) &&
		    ENABLED (cAutoOption_pastePictures) &&
		    pState->prBackgroundMap)
		{
		    /* Paste the segmented images onto the background plane.*/
		    /* Notes: Perfectscan API need these segmented images to
		       get TRC for background image AIE. So we have to keep
		       them and let the API delet them. Dec. 2, 1997 */
		    ERR_PASS_CLEANUP ( w_autoProcPastePict( pIPWState,
						 pState->prBackgroundMap,
						 pState->pImages,
						 pState->pMasks,
						 pState->imageCount));
		}
		/* Document drycleaning */
		ERR_PASS_CLEANUP (wp_autoCleanText (pIPWState, pState));
	    }
	}
	else if( pState->prTextPlane )
	{
	/* Document drycleaning */
	    ERR_PASS_CLEANUP (wp_autoCleanText (pIPWState, pState));
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prPlaneMask)
	w_destroyPixr (pIPWState, &prPlaneMask);
    if (prRotated)
	w_destroyPixr (pIPWState, &prRotated);
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prTransfer)
	w_destroyPixr (pIPWState, &prTransfer);
    if (pState->prRotatedStrip)
	w_destroyPixr (pIPWState, &(pState->prRotatedStrip));

    if (pImages)
	IPW_FREE (pImages, cAllocatorData);
    if (pState->pRotateState)
	IPW_FREE (pState->pRotateState, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Done with autoscan pipeline; get the resulting page
 *
 * LongDescription:::
 *   The "autoscan" processing pipeline is a "directive" on top of ipwrap
 *   heuristics which does a whole slew of processing automatically, with
 *   the goal of pouring raw (but scanner-corrected, i.e. color and gamma
 *   corrected) data from a scanner in, and getting a segmented page out.
 *   The PerfectScan API defines such a function., and this series of
 *   functions is the ipwrap manifestation of such a function.
 *   <BR>
 *   The "two-pass" scan involves sending two separate passes of data
 *   through the pipeline: a lower-resolution color image, and a 
 *   higher-resolution gray or binary image.  
 *   <BR> 
 *   The low-resolution color image is used to perform most image analysis,
 *   and the bits are also used to make the final segmented pictures.
 *   The high-resolution gray or binary image is used to extract the
 *   text plane of the image. 
 *   <BR>
 *   Why do two-pass scanning? There are at least two reasons for
 *   doing two-pass scanning. First, many scanners are much slower
 *   at scanning high-resolution color than high-resolution gray
 *   (i.e. 3x slower), and so one gets a significant speed gain, since
 *   the low-resolution color is usually anywhere from 1/4 to 1/9
 *   the size of the high-resolution pass (depending on the relative
 *   resolutions, of course.) Second, one might be able to eliminate
 *   the second pass if one notices that the first pass contained no text. 
 *   <BR>
 *   The PerfectScan API uses a fairly simple data structure to represent
 *   composite pages, which is called the "segment list." The segment list
 *   is a list of images, with the convention that the first image is
 *   the "text plane" and subsequent images are "pictures." The subsequent
 *   images are placed into the composite image space using the text plane
 *   image as a coordinate space.
 *   <BR>
 *   Unlike in the PerfectScan API, I use a slightly more complicated
 *   set of data structures called the "Container", as defined in
 *   "page.c". See w_newPage() and associated functions for details,
 *   but the upshot is that it provides a composite page container structure
 *   with accessor and mutator functions, which allows more fancy page
 *   descriptions than is allowed in the PS API. This incarnation of the
 *   two-pass autoscan pipeline will always put out something which is
 *   easily translated into a PS API data structure--the text plane will
 *   always be first. However, I use an absolute coordinate space in units
 *   of Micas (1/2540th of an inch), which I think is a bit cleaner.
 *   <BR>
 *   This function gets the resulting page container.
 * 
 * PrivateDescription:::
 *   There are, I suppose, trade offs between the two page representation
 *   approaches, but since I want to be able to use this set of functions 
 *   to produce "DIR" images, where the binary image selects between 
 *   reduced-resolution fg/bg ink maps, I need some additional flexibility.
 *   It also may be that eventually, we don't want a single text plane,
 *   but want separate text block/paragraph objects. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  pScanState::     Scan state structure
 *  
 * Output:::
 *  ppPage::         Container object which contains the page
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanGetResult (
    IPWState   *pIPWState,
    void       *pScanState,
    Container **ppPage)
{
    PROLOGUE;
    Container *pPage = NULL, *pImage = NULL, *pSandwich = NULL, 
	      *pFG = NULL, *pBG = NULL;
    AutoScanState *pState = (AutoScanState *) pScanState;
    Int32      w, h, wM, hM, xM, yM, i;
    ScaledInt1616 xDPI, yDPI;

    SET_STATE (cAutoState_buildPage);

    if (pState->prTextPlane)
    {
	/* How big is this page, anyway? */
	ERR_PASS (w_pixrGetWidth (pIPWState, pState->prTextPlane, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, pState->prTextPlane, &h));
	
	/* Create the page container */
	wM = PIXELS_TO_MICAS (w, pState->textXRes);
	hM = PIXELS_TO_MICAS (h, pState->textYRes);

	ERR_PASS_CLEANUP (w_newPage (pIPWState, wM, hM, &pPage));

	if (ENABLED (cAutoOption_mapInkAndPage))
	{
	    if (pState->prForegroundMap)
	    {
		ERR_PASS_CLEANUP (w_newImage (pIPWState, 0, 0, 
					  pState->prForegroundMap, NULL, 
					  cContainerContone, &pFG));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_newNullImage (pIPWState, 0, 0, 
					  pState->prForegroundMap, NULL, 
					  cContainerContone, &pFG));
	    }
	    if (pState->prBackgroundMap)
	    {
		ERR_PASS_CLEANUP (w_newImage (pIPWState, 0, 0, 
					  pState->prBackgroundMap, NULL, 
					  cContainerContone, &pBG));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_newNullImage (pIPWState, 0, 0, 
					  pState->prBackgroundMap, NULL, 
					  cContainerContone, &pBG));
 	    }
	    /* It's an "image sandwich" rather than just a "text plane" */
	    ERR_PASS_CLEANUP (w_newSandwich (pIPWState, 0, 0, 
					     pState->prTextPlane, 
					     pFG, pBG,
					     &pSandwich));
	    ERR_PASS_CLEANUP (w_addContainer (pIPWState, pPage, pSandwich, 
					      cFalse));

/*
w_write (pIPWState, "/tmp/fg.tif", pState->prForegroundMap, TIFF_UNCOMP);
w_write (pIPWState, "/tmp/bg.tif", pState->prBackgroundMap, TIFF_UNCOMP);
*/
	}
	else
	{
	    /* Add the text plane */
	    ERR_PASS_CLEANUP (w_newImage (pIPWState, 0, 0, pState->prTextPlane,
					  NULL, cContainerText, &pImage));
	    
	    ERR_PASS_CLEANUP (w_addContainer (pIPWState, pPage, pImage, 
					      cFalse));
	}

	if ( pState->prImagePlane && 
	     pState->pageType == cSpiggotOnly)
	{
	    pImage = NULL;
	    ERR_PASS_CLEANUP (w_newImage (pIPWState, 0, 0, 
					  pState->prImagePlane,
					  NULL, cContainerContone, &pImage));

	    ERR_PASS_CLEANUP (w_addContainer (pIPWState, pPage, 
					      pImage, cTrue));
	}
	else
	{
	/* Make containers for all of the images and add them to the page*/
	    for (i = 0; i < pState->imageCount; i++)
	    {
		ERR_PASS_CLEANUP (
		    w_pixrGetResolutionDPI (pIPWState, 
					    pState->pImages[i].pPixr, 
					    &xDPI, &yDPI));

		xM = PIXELS_TO_MICAS (pState->pImages[i].xOffset,
				      xDPI);
		yM = PIXELS_TO_MICAS (pState->pImages[i].yOffset, 
				      yDPI);
		pImage = NULL;

		ERR_PASS_CLEANUP (w_newImage (pIPWState, xM, yM, 
					      pState->pImages[i].pPixr, 
					      pState->pMasks[i].pPixr, 
					      cContainerContone, 
					      &pImage));
    
		ERR_PASS_CLEANUP (w_addContainer(pIPWState, pPage, 
						 pImage, cTrue));
	    }
	}
    }
    else
    {
	/* It's just a photo, so make a container with just that */
	ERR_PASS (w_pixrGetWidth (pIPWState, pState->prImagePlane, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, pState->prImagePlane, &h));
	wM = PIXELS_TO_MICAS (w, pState->pictXRes);
	hM = PIXELS_TO_MICAS (h, pState->pictYRes);

	ERR_PASS_CLEANUP (w_newPage (pIPWState, wM, hM, &pPage));

	ERR_PASS_CLEANUP (w_newImage (pIPWState, 0, 0, pState->prImagePlane,
				      NULL, cContainerContone, &pImage));

	ERR_PASS_CLEANUP (w_addContainer (pIPWState, pPage, pImage, cTrue));
    }

    *ppPage = pPage;

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Clean some stuff up. Gotta add w_deleteContainer to page.c */
    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Output segmentation (w_segmentPage()) results to a rdiff file.
 *
 * LongDescription:::
 *   Autoscan save w_segmentation results in state.pMaskes. Now we render it
 *   as rdiff file format and output it to a file. Then cluzo can do banchmark 
 *   on this rdiff file. If auto-crop is selected, please set flag:
 *   no-margin and fill-edges in crop processing.
 *
 * Input:::
 *   pIPWState::     Ptr to State structure
 *  pScanState::     Ptr to Scan state structure, NULL on exit
 *    fileName::     Ptr to output file name string
 *     passOne::     cTrue: one pass processing, cFalse: two pass processing
 *      inName::     Ptr to input file name string
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * Make sure to call w_autoScanFinish() after calling this function.
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoScanOutputRdiff (
    IPWState   *pIPWState,
    void       *pScanState,
    char       *fileName,
    Bool       passOne,
    char       *inName)
{
    PROLOGUE;
    AutoScanState *pState = (AutoScanState *) pScanState;
    ipPoint       *polygon;
    PixrRef       *pixrList;
    Int32         w, h, x, y;
    Int32         pixrCount, polySize, i, j;
    Int32         boundX, boundY;
    Int32         px, py, ph, pw;
    Float32       pictTextRatio, redX, redY;
    FILE          *f;

    redX = ScaledInt1616_TO_Float32 (pState->xRes1) / 
	               ScaledInt1616_TO_Float32 (pState->pictXRes);
    redY = ScaledInt1616_TO_Float32 (pState->yRes1) / 
	               ScaledInt1616_TO_Float32 (pState->pictYRes);
    if( passOne == cTrue )
	pictTextRatio = ScaledInt1616_TO_Float32 (pState->pictXRes)/
	             ScaledInt1616_TO_Float32 (pState->textXRes);
    else
    {
	Int32 dpi;
	dpi = MAX( ScaledInt1616_TO_Int32 (pState->xRes1),
		   ScaledInt1616_TO_Int32 (pState->yRes1) );
	pictTextRatio = (Float32) 1.0;
	if(dpi <= 75) pictTextRatio = (Float32) 0.5;
	else if(dpi <= 100) pictTextRatio = (Float32) 2.0/ (Float32)3.0;
    }

    pixrCount = pState->imageCount;
    w = pState->width1;
    h = pState->height1;
    pixrList = pState->pMasks;
    /* Check maskes */
    ERR_GEN ((!pixrList && pixrCount >0), cErrNullPtr);
    x = pState->boundingBox.x + pState->initialBoundingBox.x;
    y = pState->boundingBox.y + pState->initialBoundingBox.y;
    boundX = SCALE32_NR(x, pictTextRatio);
    boundY = SCALE32_NR(y, pictTextRatio);

    f = fopen(fileName, "w");
    ERR_GEN (!f, cErrNomem);

    /* Write the header */
    fprintf(f, "BEGIN_IMAGE\n");
    fprintf(f, "FILENAME %s\n", inName);
    fprintf(f, "IMAGE_WIDTH %ld\n", w);
    fprintf(f, "IMAGE_HEIGHT %ld\n", h);
    fprintf(f, "IMAGE_XRES %ld\n", ScaledInt1616_TO_Int32(pState->xRes1));
    fprintf(f, "IMAGE_YRES %ld\n", ScaledInt1616_TO_Int32(pState->yRes1));
    fprintf(f, "END_IMAGE\n");
    fprintf(f, "BEGIN_SUMMARY\n");
    fprintf(f, "TOTAL_REGIONS %ld\n", pixrCount);
    fprintf(f, "TEXT_REGIONS 0\nIMAGE_REGIONS %ld\n", pixrCount);
    fprintf(f, "VRULE_REGION 0\nHRULE_REGIONS 0\n");
    fprintf(f, "ORDER_FLAGS\nEND_SUMMARY\n");
    
    fprintf(f, "BEGIN_REGIONS\n\n");
    for (i = 0; i < pixrCount; i++)
    {
	fprintf(f, "R_TYPE IMAGE\n");
	fprintf(f, "R_NUMBER %ld\n", i + 1);
	fprintf(f, "R_OUT_ORDER %ld\n", i + 1);
	polygon = NULL;

	/* Convert that region to a polygon */
	ERR_PASS(w_contourRegionBinary (pIPWState, pixrList[i].pPixr,
				   NULL, 1, 1, &polySize, &polygon) );

	ERR_PASS(w_pixrGetWidth (pIPWState, pixrList[i].pPixr, &w));
	ERR_PASS(w_pixrGetHeight (pIPWState, pixrList[i].pPixr, &h));
	x = pixrList[i].xOffset;
	y = pixrList[i].yOffset;

	/* Adjust for the crop */
	x += boundX;
	y += boundY;

	/* Write out the polygon */
	py = SCALE32_NR(y, redY);
	px = SCALE32_NR(x, redX);
	ph = SCALE32_NR((y + h-1), redY);
	pw = SCALE32_NR((x + w-1), redX);
	fprintf(f, "REGION_POLYGON %ld %ld %ld %ld\n", py, ph, px, pw);
 
	fprintf(f, "COUNT %ld\n", polySize);
	for (j = 0; j < polySize; j++)
	{
	    py = SCALE32_NR((polygon[j].y + y), redY);
	    px = SCALE32_NR((polygon[j].x + x), redX);
	    fprintf(f, "%ld %ld\n",  px, py);
	}
	fprintf(f, "\n");
    }

    fclose(f);
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Get the current processing state
 *
 * LongDescription:::
 *   "autoscan" maintains a current processing state, useful for giving
 *   feedback to a user while the machine is processing an image.
 *   This call returns that state. This is usually called from a UI
 *   thread.<BR>
 *   <BR>
 *   The processing states are defined in autoscan.pub, as cAutoState_*
 *
 * Input:::
 *  pIPWState::      State structure
 *  pScanState::     Scan state structure
 *  
 * Output:::
 *  pProcessingState:: Current processing state
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanGetProcessingState (
    IPWState   *pIPWState,
    void       *pScanState,
    Int32      *pProcessingState)
{
    PROLOGUE;
    AutoScanState *pState = (AutoScanState *) pScanState;
    ERR_GEN (!pState, cErrNullPtr);

    *pProcessingState = pState->currentState;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Create spiggot state, copy information from autoscan state.
 *
 * LongDescription:::
 *   This call create spiggot state and copy information from autoscan
 *   state. This call should be called before w_autoScanFinish. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  pScanState::     Scan state structure
 *  
 * Output:::
 *  ppSpiggotState:: Spiggot state structure
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * If one fails to call this function beore calling w_autoScanFinish,
 * this function will do error return.
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanCreateSpiggotInfo (
    IPWState      *pIPWState,
    void          *pScanState,
    SpiggotState **ppSpiggotState)
{
    PROLOGUE;
    AutoScanState *pState = (AutoScanState *) pScanState;
    SpiggotState *pSpiggot;

    ERR_GEN (pState == NULL, cErrNullPtr);
    ERR_GEN (ppSpiggotState == NULL, cErrNullPtr);
    pSpiggot = *ppSpiggotState;
    ERR_GEN (pSpiggot != NULL, cErrNullPtr);

    pSpiggot = (SpiggotState *) IPW_MALLOC (sizeof (SpiggotState),
					   cAllocatorData);
    ERR_GEN (!pSpiggot, cErrNomem);
    MEMSET((void *)pSpiggot, 0, sizeof(SpiggotState));

    /* Copy orignal image info from autoscan state */
    pSpiggot->width1 = pState->width1;
    pSpiggot->height1 = pState->height1;
    pSpiggot->xRes1 = pState->xRes1;
    pSpiggot->yRes1 = pState->yRes1;
    pSpiggot->depth1 = pState->depth1;

    pSpiggot->width2 = pState->width2;
    pSpiggot->height2 = pState->height2;
    pSpiggot->xRes2 = pState->xRes2;
    pSpiggot->yRes2 = pState->yRes2;
    pSpiggot->depth2 = pState->depth2;

    pSpiggot->pageType = pState->pageType;
    pSpiggot->processingFlags = pState->processingFlags;
    pSpiggot->orientation = pState->orientation;
    pSpiggot->skew = pState->skew;

    *ppSpiggotState = pSpiggot;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Done with processing; free spiggot state
 *
 * LongDescription:::
 *   This call frees resources associated with processing
 *
 * Input:::
 *  pIPWState::      State structure
 *  
 * Output:::
 *  ppSpiggotState::    Spiggot state structure, NULL on exit
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanFreeSpiggotInfo (
    IPWState       *pIPWState,
    SpiggotState  **ppSpiggotState)
{
    PROLOGUE;
    SpiggotState *pSpiggot;
    
    ERR_GEN (!*(ppSpiggotState), cErrNullPtr);
    pSpiggot = *ppSpiggotState;

    IPW_FREE (pSpiggot, cAllocatorData);
    *ppSpiggotState = NULL;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Done with autoscan pipeline; free resources
 *
 * LongDescription:::
 *   This call frees resources associated with an "autoscan" pipeline
 *
 * Input:::
 *  pIPWState::      State structure
 *  
 * Output:::
 *  ppScanState::    Scan state structure, NULL on exit
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * If one fails to call w_AutoScanGetResult() beore calling this function,
 * then the PIXRs to which pImages and pMasks refer will be leaked!
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoScanFinish (
    IPWState   *pIPWState,
    void      **ppScanState)
{
    PROLOGUE;
    AutoScanState **ppState = (AutoScanState **) ppScanState, *pState;
    

    ERR_GEN (!*(ppState), cErrNullPtr);
    pState = *ppState;

/*   Free this not! It is owned by the page at this point. */
/*	ERR_PASS (w_destroyPixr (pIPWState, &(pState->prTextPlane))); */

    
    if (pState->pageType != cSpiggotOnly && 
	pState->prTextPlane && pState->prImagePlane) 
    {
	ERR_PASS (w_destroyPixr (pIPWState, &(pState->prImagePlane))); 
    }

    if (pState->pImages)
	IPW_FREE (pState->pImages, cAllocatorData);
    if (pState->pMasks)
	IPW_FREE (pState->pMasks, cAllocatorData);
	
    IPW_FREE (pState, cAllocatorData);
    *ppState = NULL;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Crop out a photo in the image plane
 *
 * LongDescription:::
 *   Internal routine: we've determined that there is no text on the
 *   platen, but just an image, so we just crop it out using w_cropPhoto().
 *
 * Input:::
 *  pIPWState::      State structure
 *  
 * Output:::
 *  pState::         Scan state structure
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_autoPhotoCrop (
    IPWState            *pIPWState,
    AutoScanState       *pState)
{
    PROLOGUE;
    COORDS      *pCropPoints = NULL;
    Float64      error;
    RECTANGL     rect;
    PIXR        *prRotated = NULL, *prImageToCrop = NULL;
    PIXR	*prCroppedImage = NULL;
    ScaledInt1616 xRes, yRes;
    PixrRef     *pImages = NULL;

    /* We didn't make an image plane if a binary source was given */
    if (pState->depth1 == 1)
	prImageToCrop = pState->prTextPlane;
    else
        prImageToCrop = pState->prImagePlane;

    /* We need to do a "photo" crop whether we're cropping 
	       or deskewing */
    if (ENABLED(cAutoOption_deskew) || ENABLED(cAutoOption_crop))
    {
	SET_STATE (cAutoState_photoCropSkewDetect);

	/* Force a reasonable resolution */
	ERR_PASS (w_pixrGetResolutionDPI (pIPWState, 
					  prImageToCrop,
					  &xRes, &yRes));
	if (xRes <= 0)
	{
	    if (yRes > 0) xRes = yRes;

	    /* Assume anything reasonable for photos */
	    else xRes = Int32_TO_ScaledInt1616(100);
	}
	if (yRes <= 0)
	{
	    yRes = xRes;
	}
	ERR_PASS_CLEANUP(w_pixrSetResolutionDPI (pIPWState,
						 prImageToCrop,
						 xRes, yRes));

	/* Find photo and boundary markers, use watershedding to locate
	 * the photo boundary, convert boundary to polygon, find best
	 * orthogonal direction, crop to first edge, rotate, crop to
	 * final edge.
	 * If that doesn't work, use old method.
	 */
	SET_STATE (cAutoState_photoCropSkewDetect);
	ERR_PASS_CLEANUP (w_cropPhoto(pIPWState, prImageToCrop, 2,
				      ENABLED(cAutoOption_deskew),
				      ENABLED(cAutoOption_crop),
				      &prCroppedImage));
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prImageToCrop));
	if (pState->depth1 == 1)
	    prImageToCrop = pState->prTextPlane = prCroppedImage;
	else
	    prImageToCrop = pState->prImagePlane = prCroppedImage;
	prCroppedImage = NULL;
    }
    else if (ENABLED(cAutoOption_crop))
    {
	/* Just crop it */
	SET_STATE (cAutoState_photoCropDetect);
	ERR_PASS_CLEANUP (w_crop (pIPWState, prImageToCrop, NULL,
				  &error, &rect));
	SET_STATE (cAutoState_photoCrop);
	ERR_PASS_CLEANUP (w_cutRectIP (pIPWState, 
				       rect.x, rect.y, rect.w, rect.h,
				       prImageToCrop));
    } 
    
    /* AIE it if requested */
    if (ENABLED(cAutoOption_imageEnhance))
    {
	ERR_PASS_CLEANUP (w_autoProcAIE (pIPWState, 0,
					 &prImageToCrop,
					 &pImages));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prCroppedImage)
	w_destroyPixr (pIPWState, &prCroppedImage);
    if (prRotated)
	w_destroyPixr (pIPWState, &prRotated);
    if (pCropPoints)
	w_destroyCoords (pIPWState, &pCropPoints);
    ERR_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Skew detection.
 *
 * LongDescription:::
 *   Internal routine: use w_findSkewSample to detect skew. If failed,
 *   use documentBounding and textMask to get ride of image and empty
 *   part, then do findSkewSample again.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pTextImage::     Pointer to binary image for skew detection
 *  pState::         Scan state structure
 *  
 * Output:::
 *  pSkew::          Pointer to skew angle
 *  pConfidence::    Pointer to confidence of detection
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  X. Zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_autoSkewDetection (
    IPWState            *pIPWState,
    PIXR                *pTextImage,
    AutoScanState       *pState,
    Float32             *pSkew,
    Int32               *pConfidence)
{
    PROLOGUE;
    PIXR           *prImage = NULL, *prTmp = NULL;
    Bool            fillEdges = cFalse;
    ScaledInt1616   xRes, yRes;
    Int32           margin, dpi, reduction;
    RECTANGL        textRect, *tmpRect = NULL;

    ERR_GEN (pTextImage == NULL, cErrNullPtr);

    ERR_PASS_CLEANUP (w_copy (pIPWState, pTextImage, &prImage));
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, 
					      prImage,
					      &xRes, &yRes));
	    
    margin = MAX (ScaledInt1616_TO_Int32 (xRes),
		  ScaledInt1616_TO_Int32 (yRes)) / 2;
    fillEdges = ENABLED(cAutoOption_fillEdges) ? cTrue : cFalse;
    
    ERR_PASS_CLEANUP ( w_documentBoundingBox (pIPWState, prImage,
					      NULL, margin, fillEdges,
					      &textRect));

    ERR_PASS_CLEANUP (w_clipRectToPixr (pIPWState, prImage,
					&textRect, &tmpRect));

	    /* Crop the text plane in-place */
    ERR_PASS_CLEANUP (w_cutRectIP(pIPWState, 
				  tmpRect->x, tmpRect->y,
				  tmpRect->w, tmpRect->h, 
				  prImage));
    ERR_PASS_CLEANUP (w_destroyRect (pIPWState, &tmpRect));

    ERR_PASS_CLEANUP (w_textMask (pIPWState, prImage, &prTmp, cTrue));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prImage, AND, prTmp)); 

    dpi = MAX(ScaledInt1616_TO_Int32 (xRes), ScaledInt1616_TO_Int32 (yRes));
    reduction = 2;
    if( dpi >= 300 ) reduction =4;
    else if(dpi < 200 ) reduction =1;

    do
    {
	ERR_PASS_CLEANUP (w_findSkewSample (pIPWState, prTmp, reduction, 
				    pConfidence, pSkew));
	reduction /= 2;
    } while( reduction >0 && *pConfidence == 0);
    ERR_PASS (w_destroyPixr (pIPWState, &prTmp)); 

    if( *pConfidence == 0 && *pSkew == (Float32) 0)
    {
	reduction = 2;
	if( dpi >= 300 ) reduction =4;
	else if(dpi < 200 ) reduction =1;
	ERR_PASS_CLEANUP (w_rotate90 (pIPWState, prImage, &prTmp));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prImage));
	prImage = prTmp;
	prTmp = NULL;
	ERR_PASS_CLEANUP (w_textMask (pIPWState, prImage, &prTmp, cTrue));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prImage, AND, prTmp)); 

	do
	{
	    ERR_PASS_CLEANUP (w_findSkewSample (pIPWState, prTmp, reduction, 
						pConfidence, pSkew));
	    reduction /= 2;
	} while( reduction >0 && *pConfidence == 0);
	ERR_PASS (w_destroyPixr (pIPWState, &prTmp)); 
    }

    ERR_PASS (w_destroyPixr (pIPWState, &prImage));

    NORMAL_RETURN;

ERR_CLEANUP:

    if(prImage) w_destroyPixr (pIPWState, &prImage);
    if(tmpRect) w_destroyRect (pIPWState, &tmpRect);
    if(prTmp) w_destroyPixr (pIPWState, &prTmp);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform segmentation on a page
 *
 * LongDescription:::
 *   Internal routine: Find the images in the image plane by calling
 *   w_segmentPagePixrs, and make the appropriate data structures inside the
 *   state structure
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_autoSegmentPage(
    IPWState       *pIPWState,
    AutoScanState  *pState)
{
    PROLOGUE;
    Int32          maskCount;
    PIXR          *pImage = NULL;
    PixrRef       *pMasks = NULL, *pImages = NULL;

    SET_STATE (cAutoState_segmentDetect);
    ERR_PASS (w_autoProcSegmentPage (pIPWState, pState->prImagePlane,
				     &pMasks, &maskCount));

    SET_STATE (cAutoState_segment);
    ERR_PASS (w_autoProcCuttingPage (pIPWState, pState->prImagePlane,
				     pMasks, maskCount, pState->skew,
				     &pImages));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pState->prImagePlane));
    
    pState->pMasks = pMasks;
    pState->pImages = pImages;
    pState->imageCount = maskCount;

    if (ENABLED(cAutoOption_imageEnhance))
    {
	/*
	 * Run "AIE" on all of the images, if requested. We do each
	 * one individually.
	 * We should also do something about white-point here, though
	 * AIE is supposed to do something already.
	 */
	if( maskCount > 0)
	    ERR_PASS_CLEANUP (w_autoProcAIE (pIPWState, maskCount,
					     &pImage, &pImages));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pMasks)
	IPW_FREE (pMasks, cAllocatorData);
    if (pImages)
	IPW_FREE (pImages, cAllocatorData);

    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Blast away the image masks in the text plane
 *
 * LongDescription:::
 *   Internal routine: Given that we have a bunch of masks for images
 *   on the page, expand the masks and erase what's under them in the
 *   text plane.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_autoBlastMasks (
    IPWState        *pIPWState,
    AutoScanState   *pState)
{
    PROLOGUE;

    SET_STATE (cAutoState_cleanMasks);
    ERR_PASS ( w_autoProcBlastMasks( pIPWState,
				     pState->prTextPlane,
				     pState->prForegroundMap,
				     pState->prBackgroundMap,
				     pState->pImages,
				     pState->pMasks,
				     pState->imageCount));


    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform "document drycleaning" on the text plane
 *
 * LongDescription:::
 *   Internal routine: Given that we have a text plane, perform requested
 *   "document drycleaning" functions on it. At the moment, we assume
 *   that we did, if requested, do pepper removal during binarization,
 *   so what we need to do here is stipple (shading) and rule removal, unless
 *   we got a binary input, in which case we didn't do any binarization.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_autoCleanText (
    IPWState      *pIPWState,
    AutoScanState *pState)
{
    PROLOGUE;

    ERR_PASS ( w_autoProcCleanText( pIPWState, &pState->prTextPlane,
				    pState->processingFlags));

    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Update the "thumbnail image" for the current state of autoscan
 *
 * LongDescription:::
 *   Internal routine: update the "thumbnail" image, if appropriate, for
 *   the current state of autoscan, as embodied by the image passed. This
 *   involves subsampling the image into an image the size of the 
 *   thumbnail, and then casting it to the correct depth for the thumbnail.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  prImage::        Image representing the current state
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_updateThumbnail (
    IPWState      *pIPWState,
    AutoScanState *pState,
    PIXR          *prImage)
{
    PROLOGUE;
    PIXR *prTmp = NULL, *prTmp1 = NULL, *prChannel = NULL, *prTmp2 = NULL;
    Int32 wT, hT, wI, hI, wC, hC, channels, imageDepth, imageChannels, i,
	  thumbDepth;
    Float32 aI, aT, ratio;
#ifdef DEBUG
    static Int32 invocation = 0;
    char str[128];
#endif

    if (pState->prThumbnail != NULL)
    {
	
	ERR_PASS (w_pixrGetWidth (pIPWState, pState->prThumbnail, &wT));
	ERR_PASS (w_pixrGetHeight (pIPWState, pState->prThumbnail, &hT));
	ERR_PASS (w_pixrGetChannelCount (pIPWState, pState->prThumbnail,
					 &channels));

	ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &wI));
	ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &hI));

	/* Make it fit */
	aI = (Float32) wI / (Float32) hI;
	aT = (Float32) wT / (Float32) hT;
	ratio = aI / aT;
	if (ratio < 1.0f)
	{
	    hC = hT;
	    wC = (Int32) (ratio * wT);
	}
	else
	{
	    wC = wT;
	    hC = (Int32) (hT / ratio);
	}

	/* Clear the thumbnail if the sizes don't match */
	if (wC < wT)
	{
	    ERR_PASS (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET, wC, 0, 
				  wT - wC + 1, hC, pState->prThumbnail));
	}

	if (hC < hT)
	{
	    ERR_PASS (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET, 0, hC, 
				  wT, hT - hC + 1, pState->prThumbnail));
	}

	/* Subsample it */
	ERR_PASS (w_scaleToSize (pIPWState, prImage, wC, hC, &prTmp));
    
	ERR_PASS_CLEANUP (w_pixrGetDepth(pIPWState, prImage, &imageDepth));
	ERR_PASS_CLEANUP (w_pixrGetDepth(pIPWState, pState->prThumbnail,
					 &thumbDepth));
	ERR_PASS_CLEANUP (w_pixrGetChannelCount(pIPWState, prImage, 
						&imageChannels));

	/* Change the depth, if necessary */
	if (imageDepth != thumbDepth)
	{
	    switch (thumbDepth)
	    {
	    case 4:
		switch (imageDepth)
		{
		case 1:
		    ERR_PASS_CLEANUP (w_depth1To8 (pIPWState, prTmp, cTrue, 
						   &prTmp2));
		    ERR_PASS_CLEANUP (w_scale8To4 (pIPWState, prTmp2, 
						   &prTmp1));
		    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));
		    break;
		    
		case 8:
		    ERR_PASS_CLEANUP (w_scale8To4 (pIPWState, prTmp, &prTmp1));
		    break;
		    
		default:
		    ERR_GEN_CLEANUP (pIPWState, cErrTypeNotSupported);

		    break;
		}

		break;

	    case 8:
		switch (imageDepth)
		{
		case 1:
		    ERR_PASS_CLEANUP (w_depth1To8 (pIPWState, prTmp, cTrue, 
						   &prTmp1));
		    break;
		    
		case 4:
		    ERR_PASS_CLEANUP (w_depth4To8 (pIPWState, prTmp, &prTmp1));
		    break;
		    
		default:
		case 2:
		    ERR_GEN_CLEANUP (pIPWState, cErrTypeNotSupported);
		}
		break;

	    default:
		    ERR_GEN_CLEANUP (pIPWState, cErrTypeNotSupported);
	    }


	}
	else
	{
	    prTmp1 = prTmp;
	}


	if (channels > imageChannels)
	{
	    /* 
	     * If the image was gray/bin and the dest is color, copy each
	     * channel
	     */
	    for (i = 0; i < channels; i++)
	    {
		ERR_PASS_CLEANUP (
		    w_pixrGetChannel (pIPWState, pState->prThumbnail, 
				      i, &prChannel));
		
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prTmp1, PIX_SRC,
					      0, 0, wC, hC, prChannel));
		prChannel = NULL;
	    }
	}
	else if (channels == imageChannels)
	{
	    /* Otherwise, just copy it into the thumbnail */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prTmp1, PIX_SRC,
					  0, 0, wC, hC, pState->prThumbnail));
	}
	else
	{
	    ERR_GEN_CLEANUP (pIPWState, cErrTypeNotSupported);
	}

	if (imageDepth != thumbDepth)
	{
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
	}
	else
	{
	    prTmp1 = NULL;
	}

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

#ifdef DEBUG
	sprintf(str, "/tmp/thumb%04ld.tif", invocation++);
	ERR_PASS_CLEANUP (w_write (pIPWState, str, pState->prThumbnail, TIFF_UNCOMP));
#endif

    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (imageDepth != 8 && prTmp1)
	w_destroyPixr (pIPWState, &prTmp1);

    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prTmp2)
	w_destroyPixr (pIPWState, &prTmp2);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Private function to erode masked gray pixles. 
 *
 * LongDescription:::
 *    To save time, this function erode masked gray scale pixels, unmasked
 *  pixels will be unchanged. The structure element is  
 *                                                     *
 *                                                   * * *
 *                                                     *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure 
 *  pBinImage::      Binary image ( thresholded from color image )
 *  pImage::         Gray or color image, depth is 8 
 *
 * Output:::
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  colored text
 *
 * Keywords:::
 *  colored text
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_erodeMaskedGray (
    IPWState               *pIPWState,
    PIXR                   *pBinImage,
    PIXR                   *pImage)
{
    PROLOGUE;
    Int32        width, height, bpl, bbpl, channels, cha, w, h;
    PIXR        *prTmp = NULL, *pEdge = NULL;
    PIXR        *subL;
    UInt8       *pl, *pMask, *pm;
    STREL       *pTheStrel = NULL;
    Int32        row, col, bits, byte, x;
    UInt8        left, right, up, down, cent;

    ERR_GEN_PIXR (pBinImage, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR (pImage, cByteOrderBig, cDepth8, cChannels1|cChannels3, 
		  cFrame32);

/* Get edge */
    pTheStrel = NULL;
    ERR_PASS (w_initStrelByName (pIPWState, strel3));
    ERR_PASS (w_getStrel (pIPWState, strel3, &pTheStrel));
    ERR_PASS_CLEANUP (w_erode (pIPWState, pBinImage, pTheStrel, &prTmp));
    ERR_PASS_CLEANUP (w_copy (pIPWState, pBinImage, &pEdge));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp, SUBTRACT, pEdge));    
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pEdge, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pEdge, &height));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, pEdge, &bbpl)); 
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, pEdge, &pMask));


    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, pImage, &channels));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, pImage, &bpl)); 

    w = width>>3;
    h = height - 1;
    for(cha=0; cha<channels; ++cha)
    {
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, pImage, 
					    cha, &subL));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, subL, &pl));

	pm = pMask + bbpl;
	pl += bpl;
	for(row=1; row<h; ++row, pm+=bbpl, pl+=bpl)
	{
	    for(bits=1, col=8; bits<w; ++bits, col+=8)
	    {
		byte = UCHAR_ACCESS(pm+bits);
		if(byte == 0) continue;
		for(x = col; x<(col+8) && byte; ++x)
		{
		    if( byte & 0x80 )
		    {
			cent = UCHAR_ACCESS(pl+x);
			left = UCHAR_ACCESS(pl+x-1);
			right = UCHAR_ACCESS(pl+x+1);
			up = UCHAR_ACCESS(pl+x-bpl);
			down = UCHAR_ACCESS(pl+x+bpl);
			if( cent > left ) cent = left;
			if( cent > right ) cent = right;
			if( cent > up ) cent = up;
			if( cent > down ) cent = down;
			UCHAR_ACCESS(pl+x) = cent;
		    }
		    byte <<= 1;
		}
	    }
	}
    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pEdge));

    NORMAL_RETURN;

ERR_CLEANUP:

    if( prTmp ) w_destroyPixr(pIPWState, &prTmp);
    if( pEdge ) w_destroyPixr(pIPWState, &pEdge);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Generate background and foreground map.
 *
 * LongDescription:::
 *   Generate background and foreground map.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  prImageStrip::   Color or gray image at input resolution.
 *  prScaledStrip::  Color or gray image at text resolution.  This is
 *		     only needed to generate a foreground plane.  If
 *		     pState->foregroundMap == NULL, this may be supplied
 *		     as NULL.
 *  prBinStrip::     Binary selector image. Like prScaledStrip, this is
 *		     only needed to generate a foreground plane.  If
 *		     pState->foregroundMap == NULL, this may be supplied
 *		     as NULL.
 *  
 * Output:::
 *  
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_onePassMapink (
   IPWState      *pIPWState,
   AutoScanState *pState,
   PIXR          *prImageStrip,
   PIXR		 *prScaledStrip,
   PIXR		 *prBinStrip)
{
    PROLOGUE; 
    Float32	  scaleX, scaleY;
    PIXR	 *prReduced = NULL, *prPreReduced = NULL;
    Int32	  w, h;
    Int32	  scaledW = 0, scaledH = 0, yOffset = 0;
    Float32	  realH;

    ERR_GEN (!pState, cErrNullPtr);
    ERR_GEN (!prImageStrip, cErrNullPtr);
    
    ERR_PASS (w_pixrGetWidth (pIPWState, prImageStrip, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, prImageStrip, &h));
    
	/* We're doing foreground/background processing.  Start
	 * with the foreground.
	 * */

    if (pState->prForegroundMap)
    {
	    /* These are only needed for generating a
	     * foreground plane.
	     */
	ERR_GEN (!prScaledStrip, cErrNullPtr);
	ERR_GEN (!prBinStrip, cErrNullPtr);
	   /* we scale the incoming prScaledStrip from
	    * the resolution of the text plane to the resolution
	    * of the foreground plane.
	    */
	scaleX = pState->fgXScale / pState->textXScale;
	scaleY = pState->fgYScale / pState->textYScale;

	if( prScaledStrip == prImageStrip ) 
	    ERR_PASS_CLEANUP (w_copy (pIPWState, prImageStrip, &prReduced));
	else 
	    prReduced = prScaledStrip;

	ERR_PASS_CLEANUP (wp_erodeMaskedGray (pIPWState, prBinStrip,
						  prReduced));

	   /* This is a very subtle operation.  On the surface,
	    * it looks like a performance hack to produce a
	    * foreground plane by only inspecting the image pixels
	    * that are under the selector.  However, by only
	    * looking at the pixels that the selector indicates are
	    * "dark", we end up with only dark pixels in the
	    * foreground.  In particular, say that a 3x3 square
	    * of pixels could be scaled into one foreground pixel.
	    * Furthermore, say that only one of these pixels is
	    * dark.  Using almost any ordinary scaling operation,
	    * we'd get a light pixel in the foreground.  Using
	    * w_scaleMaskedArea, we're guaranteed to get the
	    * one dark pixel.
	    * There are a couple of other cute features in this
	    * operation.  First, the area covered by the selector
	    * is likely to be light around the edges and dark in
	    * the middle.  w_scaledMaskedArea chooses the MIN of
	    * all pixels that map to a given foreground pixel, so
	    * gets the dark center value, which is good.
	    * Ok, but why do we spend the time and memory to scale
	    * the input image up to the size of the text instead
	    * of scaling the selector down to the size of the
	    * input before selecting?  You have to remember that
	    * even at 300 dpi, many character strokes are only 1
	    * or 2 pixels wide.  When scaled down to the input
	    * image resolution, they'd either vanish entirely or
	    * overlap the light stuff around the outsides of thin
	    * strokes.  By expanding the input up to the text
	    * resolution, you make all the features bigger, giving
	    * you a better chance to select exactly the right
	    * pixel to scale.
	    */
	ERR_PASS_CLEANUP (w_scaleMaskedArea (pIPWState, prBinStrip,
					     prReduced, scaleX, scaleY,
					     pState->textY, 
					     pState->prForegroundMap));
	if( prReduced != prScaledStrip )
	   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));
	prReduced = NULL;
    }


	/* If there's a background map, scale the incoming
	 * strip to the background resolution.
	 */
    if (pState->prBackgroundMap)
    {
	    /* Prepare to scale the incoming image strip to the background
	     * resolutions.  This may be different from the foreground
	     * resolution.
	     */
	scaleX = pState->bgXScale;
	scaleY = pState->bgYScale;
	scaledW = SCALE32_NR (w, scaleX);
	scaledH = SCALE32_NR (h, scaleY);
	realH = (Float32) h * scaleY;

	/* Pure linear scaling causes moire effects in the background
	 * plane, so do a first scaling pass using resolution
	 * conversion if possible.
	 */
	if (scaleX < 0.34f && scaleY < 0.34f )
	{
	   ERR_PASS_CLEANUP (w_resConv (pIPWState, prImageStrip, 300, 100, 
					&prPreReduced));
	}
	else if (scaleX <= 0.5f && scaleY <= 0.5f )
	{
	   ERR_PASS_CLEANUP (w_resConv (pIPWState, prImageStrip, 200, 100, 
					&prPreReduced));
	}

	    /* Do the final background scaling step with exact results.
	     * If prPreReduced == NULL, this is also the first step.
	     */
	if (prPreReduced == NULL)
	   ERR_PASS_CLEANUP (w_scaleLinearSize (pIPWState, prImageStrip,
						scaledW, scaledH, 
						&prReduced));
	else
	{
	   ERR_PASS_CLEANUP (w_scaleLinearSize (pIPWState, prPreReduced,
						scaledW, scaledH, 
						&prReduced));
	   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prPreReduced));
	   prPreReduced = NULL;
	}
	    /* Sometimes we reduce really tiny strips (like 8 lines to
	     * 2 lines instead of 100 lines to 33 lines) and get weird
	     * resolutions.  Set the final resolution to be the
	     * right value.
	     */
	ERR_PASS (w_pixrSetResolutionDPI(pIPWState, prReduced,
					 pState->bgXRes, pState->bgYRes));
	
	    /* If there would be a gap between the last existing line of
	     * background and the first line of this strip, duplicate the
	     * last existing background line to fill in.
	     */
	yOffset = (Int32) ( ((Float32) pState->textY *
			     (pState->bgYScale / pState->textYScale)) + 0.5);
	if ((Int32) yOffset > pState->bgY)
	{
	   ERR_PASS_CLEANUP (w_rasterop (pIPWState,
					 0, pState->bgY - 1, 
					 pState->prBackgroundMap,
					 PIX_SRC,
					 0, pState->bgY, scaledW, 1,
					 pState->prBackgroundMap));
	   pState->bgY++;
	}  

	    /* At last.  Do the real work of copying the scaled image
	     * strip into the background image.
	     */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0,  prReduced, PIX_SRC, 
					  0, pState->bgY, scaledW, scaledH, 
					  pState->prBackgroundMap));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));
	prReduced = NULL;
	pState->bgY += scaledH;
	pState->realBGY += realH;
    }

   NORMAL_RETURN;

ERR_CLEANUP:
   if(prReduced != NULL) w_destroyPixr (pIPWState, &prReduced);
   if(prPreReduced != NULL) w_destroyPixr (pIPWState, &prPreReduced);

   ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Generate background and foreground map.
 *
 * LongDescription:::
 *   Generate background and foreground map. Binary image is high resolution
 *   image and gray or color image is low resolution image.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  prBinStrip::      Binary image -- high resolution
 *  prImageStrip::    Color or gray image
 *  prTextStrip::     Binary image -- text
 *  
 * Output:::
 *  
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_twoPassMapink (
   IPWState      *pIPWState,
   AutoScanState *pState,
   PIXR          *prBinStrip,
   PIXR          *prImageStrip,
   PIXR          *prTextStrip)
{
   PROLOGUE;   
   Float32   scaleX, scaleY;
   PIXR      *prScaledStrip = NULL;
   PIXR      *prGroundStrip = NULL, *prReduced = NULL;
   PIXR      *prPrimary ;
   Int32     channels, y, w, h, textW, textH;
   Int32     scaledW, scaledH, yOffset;
   Float32   realH;
   STREL     *pTheStrel = NULL;

   ERR_GEN (!pState, cErrNullPtr);
   ERR_GEN (!prImageStrip, cErrNullPtr);
   ERR_GEN (!prBinStrip, cErrNullPtr);

   ERR_PASS (w_pixrGetWidth (pIPWState, prImageStrip, &w));
   ERR_PASS (w_pixrGetHeight (pIPWState, prImageStrip, &h));
   ERR_PASS (w_pixrGetWidth (pIPWState, prBinStrip, &textW));
   ERR_PASS (w_pixrGetHeight (pIPWState, prBinStrip, &textH));

   /* We're doing foreground/background processing. */
   prScaledStrip = NULL;
   if (w != textW || h != textH )
   {
       ERR_PASS (w_scaleToSize (pIPWState, prImageStrip, textW, textH,
			       &prScaledStrip));
       ERR_PASS (w_pixrSetResolutionDPI(pIPWState, prScaledStrip,
					pState->textXRes, pState->textYRes));
   }
   else
       prScaledStrip = prImageStrip;

   if( pState->prForegroundMap )
   {
       scaleX = ScaledInt1616_TO_Float32(pState->fgXRes)/
	   ScaledInt1616_TO_Float32(pState->textXRes);
       scaleY = ScaledInt1616_TO_Float32(pState->fgYRes)/
	   ScaledInt1616_TO_Float32(pState->textYRes);

       ERR_PASS_CLEANUP (w_scaleMaskedArea (pIPWState, prBinStrip,
					prScaledStrip, scaleX, scaleY,
					pState->textY,
					pState->prForegroundMap));
   }

   if( pState->prBackgroundMap )
   {
       prGroundStrip = NULL;
       ERR_PASS_CLEANUP (w_copy (pIPWState, prScaledStrip, &prGroundStrip));
   }

   if( prScaledStrip != prImageStrip )
      ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledStrip)); 

   if( !pState->prBackgroundMap )
       NORMAL_RETURN;

   scaleX = pState->bgXScale;
   scaleY = pState->bgYScale;
   scaledW = SCALE32_NR (w, scaleX);
   scaledH = SCALE32_NR (h, scaleY);
   realH = (Float32) h * scaleY;

   scaleX = ScaledInt1616_TO_Float32(pState->bgXRes)/
            ScaledInt1616_TO_Float32(pState->textXRes);
   scaleY = ScaledInt1616_TO_Float32(pState->bgYRes)/
            ScaledInt1616_TO_Float32(pState->textYRes);

   /* scaling to bg resolution */

   if (scaleX < 0.34f && scaleY < 0.34f )
   {
       ERR_PASS_CLEANUP (w_resConv (pIPWState, prGroundStrip, 300, 100, 
					 &prReduced));
       ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prGroundStrip));
       prGroundStrip = prReduced;
       prReduced = NULL;
   }
   else if (scaleX <= 0.5f && scaleY <= 0.5f )
   {
       ERR_PASS_CLEANUP (w_resConv (pIPWState, prGroundStrip, 200, 100, 
				    &prReduced));
       ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prGroundStrip));
       prGroundStrip = prReduced;
       prReduced = NULL;
   }

   /*  linear scaling to bg size */

   ERR_PASS_CLEANUP (w_scaleLinearSize (pIPWState, prGroundStrip,
					scaledW, scaledH, &prReduced));
   ERR_PASS_CLEANUP (w_pixrSetResolutionDPI(pIPWState, prReduced,
				    pState->bgXRes, pState->bgYRes));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prGroundStrip));
   prGroundStrip = NULL;

   ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel2));
   ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel2, &pTheStrel));
   ERR_PASS_CLEANUP (w_dilate (pIPWState, prTextStrip, 
			       pTheStrel, &prTextStrip));   
   ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, prTextStrip, scaledW, scaledH,
				    &prGroundStrip));
   ERR_PASS_CLEANUP (w_pixrSetResolutionDPI(pIPWState, prGroundStrip,
				    pState->bgXRes, pState->bgYRes));
   ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prReduced, 
					    &channels));

   ERR_PASS_CLEANUP (wp_setPixelLowestBit (pIPWState, prReduced)); 

   for (y = 0; y < channels; y++)
   {
       ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prReduced,
					   y, &prPrimary));

       ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prPrimary, prGroundStrip, 
				      PIX_SRC | PIX_DST, &prPrimary));
   }
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prGroundStrip));
   prGroundStrip = NULL;

	/* Copy that reduced strip in */
   yOffset = (Int32) ( (Float32) pState->textY * scaleY + 0.5);
   if ((Int32) yOffset > pState->bgY)
   {
       ERR_PASS_CLEANUP (w_rasterop (pIPWState,
				     0, pState->bgY - 1, 
				     pState->prBackgroundMap,
				     PIX_SRC,
				     0, pState->bgY, scaledW, 1,
				     pState->prBackgroundMap));
       pState->bgY++;
   }
   else if((Int32) yOffset < pState->bgY)
      pState->bgY = yOffset;


   ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0,  prReduced, PIX_SRC, 
				 0, yOffset, scaledW, scaledH, 
				 pState->prBackgroundMap));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));
   prReduced = NULL;

   pState->bgY += scaledH;
   pState->realBGY += realH;

   NORMAL_RETURN;

ERR_CLEANUP:

   if( prScaledStrip != prImageStrip && prScaledStrip != NULL )
       w_destroyPixr (pIPWState, &prScaledStrip); 
   if(prGroundStrip != NULL) w_destroyPixr (pIPWState, &prGroundStrip);
   if(prReduced != NULL) w_destroyPixr (pIPWState, &prReduced);

   ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Test binary and color or gray image alignment (within 2 pixels).
 *
 * LongDescription:::
 *  Test alignment between binary image and color or gray image.
 *  First color or gray image is binarized, then matching two images 
 *  which shifting each others to find best matching position.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pState::         Scan state structure
 *  prBin::          Binary image -- high resolution
 *  prImage::        Color or gray image
 *  tsY::            Binary image Y begining alignment
 *  tsX::            Binary image X begining alignment
 *  psY::            Color/gray image Y begining alignment
 *  psX::            Color/gray image X begining alignment
 *  
 * Output:::
 *  
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_getAlignment (
   IPWState      *pIPWState,
   AutoScanState *pState,
   PIXR          *prBin,
   PIXR          *prImage,
   Int32         *tsY,
   Int32         *tsX,
   Int32         *psY,
   Int32         *psX)
{
   PROLOGUE;   
   PIXR      *prScaledBin =NULL, *prOpBin =NULL, *prResult =NULL;
   PIXR      *prPrimary;
   Int32     w, h, textW, textH;
   Int32     bits, maxBits, orgBits;

   ERR_GEN (!pState, cErrNullPtr);
   ERR_GEN (!prImage, cErrNullPtr);
   ERR_GEN (!prBin, cErrNullPtr);

   *tsY = *tsX = *psY = *psX =0;

   ERR_PASS (w_pixrGetWidth (pIPWState, prImage, &w));
   ERR_PASS (w_pixrGetHeight (pIPWState, prImage, &h));
   ERR_PASS (w_pixrGetWidth (pIPWState, prBin, &textW));
   ERR_PASS (w_pixrGetHeight (pIPWState, prBin, &textH));

   ERR_PASS (w_scaleToSize (pIPWState, prBin, w, h, &prScaledBin));
   ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, prImage, &prPrimary));

    /* Make a binary version */
   ERR_PASS (w_binarize2 (pIPWState, prPrimary, 0,
			  cLowResGradientThreshold, &prOpBin));
   maxBits  =0;
   /* find max bits number position */
   prResult = NULL;
   ERR_PASS_CLEANUP (w_copy (pIPWState, prScaledBin, &prResult));
   ERR_PASS (w_logicOp (pIPWState, prOpBin, AND, prResult)); 
   ERR_PASS_CLEANUP (w_numberBits(pIPWState, prResult, &maxBits));
   orgBits = maxBits;
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prResult));
   /* text shifts left */
   prResult = NULL;
   ERR_PASS_CLEANUP (w_copy (pIPWState, prScaledBin, &prResult));
   ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0,  prOpBin, 
				 PIX_SRC&PIX_DST, 2, 0, w-2, h, 
				 prResult));
   ERR_PASS_CLEANUP (w_numberBits(pIPWState, prResult, &bits));
   if( bits > maxBits )
   {
      maxBits= bits;
      if( bits > (orgBits + 10) )
      {
	  *tsX = 2;
	  *psY = *psX = *tsY =0;
      }
   }
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prResult));
   /* text down */
   prResult = NULL;
   ERR_PASS_CLEANUP (w_copy (pIPWState, prScaledBin, &prResult));
   ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0,  prOpBin, 
				 PIX_SRC&PIX_DST, 0, 2, w, h-2, 
				 prResult));
   ERR_PASS_CLEANUP (w_numberBits(pIPWState, prResult, &bits));
   if( bits > maxBits )
   {
      maxBits= bits;
      if( bits > (orgBits + 10) )
      {
	  *tsY = 2;
	  *psY = *tsX = *psX =0;
      }
   }
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prResult));
   /* image shifts left */
   prResult = NULL;
   ERR_PASS_CLEANUP (w_copy (pIPWState, prOpBin, &prResult));
   ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0,  prScaledBin, 
               PIX_SRC&PIX_DST, 2, 0, w-2, h, 
				   prResult));
   ERR_PASS_CLEANUP (w_numberBits(pIPWState, prResult, &bits));
   if( bits > maxBits )
   {
      maxBits= bits;
      if( bits > (orgBits + 10) )
      {
	  *psX = 2;
	  *psY = *tsX = *tsY =0;
      }
   }
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prResult));
   /* image down */
   prResult = NULL;
   ERR_PASS_CLEANUP (w_copy (pIPWState, prOpBin, &prResult));
   ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0,  prScaledBin, 
				 PIX_SRC&PIX_DST, 0, 2, w, h-2, 
				 prResult));
   ERR_PASS_CLEANUP (w_numberBits(pIPWState, prResult, &bits));
   if( bits > maxBits )
   {
       maxBits= bits;
       if( bits > (orgBits + 10) )
       {
	  *psY = 2;
	  *psX = *tsX = *tsY =0;
       }
   }
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prResult));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prOpBin));
   ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prScaledBin));

   NORMAL_RETURN;

ERR_CLEANUP:
   if(prScaledBin != NULL) w_destroyPixr (pIPWState, &prScaledBin);
   if(prOpBin != NULL) w_destroyPixr (pIPWState, &prOpBin);
   if(prResult != NULL) w_destroyPixr (pIPWState, &prResult);

   ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Set image last bit to one.
 *
 * LongDescription:::
 *   We set all of background image pixels' last bit to one, then set all 
 *   of text pixel to zero. When we do propgation, we just wash out
 *   text pixel (value == 0) in background.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImage::         Background image
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_setPixelLowestBit (
    IPWState        *pIPWState,
    PIXR            *pImage)
{
    PROLOGUE;
    Int32    row, col, chan;
    Int32    w, h, channels, bpl, w4, w1;
    UInt8    *prImg;
    UInt32   *prLine; 
    PIXR     *prTmp;

    ERR_GEN (!pImage, cErrNullPtr);
    ERR_GEN_PIXR (pImage, cByteOrderBig, cDepth8, 
		      cChannels1 | cChannels3, cFrame32);

    ERR_PASS (w_pixrGetChannelCount (pIPWState, pImage, &channels));
    ERR_PASS (w_pixrGetBpl (pIPWState, pImage, &bpl));
    ERR_PASS (w_pixrGetWidth (pIPWState, pImage, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, pImage, &h));

    w4 = w / 4;
    w1 = w4 * 4;
    for(chan=0; chan<channels; ++chan)
    {
	ERR_PASS (w_pixrGetChannel (pIPWState, pImage, chan, &prTmp));
	ERR_PASS (w_pixrGetImage (pIPWState, prTmp, &prImg));

	for(row=0; row<h; ++row, prImg += bpl)
	{
	    prLine = (UInt32 *) prImg;
	    for(col=0; col<w4; ++col, ++prLine)
		*prLine |= 0x01010101;
	    for(col=w1; col<w; ++col)
		UCHAR_ACCESS( prImg + col) |= 1;
	}
    }

    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Check white page.
 *
 * LongDescription:::
 *   Check whether it is white page before deskew, cropping and oritation.
 *   If it is white page, return cTrue and skip all of processings.
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImage::         Binary image
 *  
 * Output:::
 *  pIsBlank::       Pointer to bool.       
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

static ErrCode CDECL
wp_checkBlankPage (
    IPWState        *pIPWState,
    PIXR            *pImage,
    Bool            *pIsBlank)
{
    PROLOGUE;
    PIXR            *reduced = NULL;
    Int32            Bits, dpi;
    ScaledInt1616    xDPI, yDPI;

    *pIsBlank = cFalse;
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, pImage, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    if (dpi <= 100)
    {
	ERR_PASS (w_reduce4 (pIPWState, pImage, 1, 4, &reduced));
    }
    else if (dpi <= 200)
    {
	/* 8x reduction */
	ERR_PASS (w_reduce8 (pIPWState, pImage, 1, 1, 4, &reduced));
    }
    else
    {
	/* 16x reduction */
	ERR_PASS (w_reduce16 (pIPWState, pImage, 1, 1, 1, 4, &reduced));
    }

    ERR_PASS_CLEANUP (w_numberBits(pIPWState, reduced, &Bits));
    if( Bits <= 2 ) *pIsBlank = cTrue;

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &reduced));

    NORMAL_RETURN;

ERR_CLEANUP:

    if (reduced)
	w_destroyPixr (pIPWState, &reduced);

    ERR_RETURN;
}


/*
$Log: autoscan.c,v $
Revision 1.1.1.1  1999/09/20 18:19:14  jlin
Imported sources

Revision 1.112  1998/11/02 19:33:48  xzhu
xzhu on Mon Nov 2 11:31:49 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

To fix blank page bug reported by Andrew, add checkBlankPage to check blank page. This function do reducing first (same as BoundingBox doing) then get numberBits.

Revision 1.111  1998/10/14 00:26:02  xzhu
xzhu on Tue Oct 13 17:25:45 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Follow the change of autoproc.pub.

Revision 1.110  1998/10/09 23:49:22  xzhu
xzhu on Fri Oct 9 16:47:35 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Do strip by strip in OnePassTransData. Output binary image for Spiggot pageType.

Revision 1.109  1998/10/02 23:17:19  xzhu
xzhu on Fri Oct 2 16:16:44 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Check new option in Spiggot pagetype.

Revision 1.108  1998/08/31 21:53:31  xzhu
xzhu on Mon Aug 31 14:51:49 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Just erode masked pixels with fixed structure element. It is much faster than general erodeGray. He

Revision 1.107  1998/07/28 23:29:47  xzhu
xzhu on Tue Jul 28 16:28:42 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Moved private functions (crop & deskew) to autoproc as public functions. It can be shared with API.

Revision 1.106  1998/07/24 20:07:44  xzhu
xzhu on Fri Jul 24 13:06:17 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Spiggot pagetype can output fore-back-ground if do ink-map. Use w_autoProcAIE to do image enhance.

Revision 1.105  1998/06/25 00:59:19  xzhu
xzhu on Wed Jun 24 17:55:28 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

If failed to find text, call rotate90, then call w_findSkewSample again to fixing failing to detecting text in the orientation 90 degree page. PS API already did this way. Now autoscan do the same way. The failing case reported by Allen and Stan is fixed.

Revision 1.104  1998/06/18 23:09:21  xzhu
xzhu on Thu Jun 18 16:08:33 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added TransData functions.

Revision 1.103  1998/06/12 22:30:41  xzhu
xzhu on Fri Jun 12 15:30:01 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Use paste & blast functions in autoproc.

Revision 1.102  1998/06/11 18:29:04  xzhu
xzhu on Thu Jun 11 11:27:37 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Spiggot will not work with map-ink. Added error report if options and pageType not consistent.

Revision 1.101  1998/06/10 19:36:47  xzhu
xzhu on Wed Jun 10 12:36:21 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed problem in w_autoScanCreatSpiggotInfo.

Revision 1.100  1998/06/09 21:01:37  xzhu
xzhu on Tue Jun 9 14:01:13 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Changed test and image container order in spiggot pagetype.

Revision 1.99  1998/06/05 22:07:53  xzhu
xzhu on Fri Jun 5 15:07:44 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added new functions.

Revision 1.98  1998/06/03 22:54:29  xzhu
xzhu on Wed Jun 3 15:53:27 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Implemented text only page type in one pass. Input gray, color or binaray, output binary.

Revision 1.97  1998/05/29 17:30:15  xzhu
xzhu on Fri May 29 10:29:09 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

W_createPixrExtendedNoInit caused purify complained.

Revision 1.96  1998/05/27 19:39:47  xzhu
xzhu on Wed May 27 12:32:51 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Second pass image can be inside the secondPassRect by calling  w_autoScanSetSecondPassRect. The rectangle beginning (x, y), can be changed. But changed (x, y) must be >= orignal (x, y). Implemented page type -- cSpiggotOnly, output low resolution color iamge and high resolution binaray image and save them in container.

Revision 1.95  1998/05/26 17:21:24  xzhu
xzhu on Tue May 26 10:20:21 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Using w_autoProcdetections to get croping rectangle, skew angle and orientation.

Revision 1.94  1998/05/19 21:07:31  xzhu
xzhu on Tue May 19 14:07:16 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Use functions in autoproc.

Revision 1.93  1998/05/06 18:48:25  xzhu
xzhu on Wed May 6 11:45:47 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed problem in one pass with pictureOnly. If set pictureOnly, there is no textPlane in one pass, checking the white paper by bits counting is not correct.

Revision 1.92  1998/04/17 22:41:03  xzhu
xzhu on Fri Apr 17 15:37:52 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added w_numberBits to check the textPlane to avoid error return from documentBoundBox or other function. If there is nothing, free pixrs and return, otherwise continue.

Revision 1.91  1998/04/16 22:41:04  xzhu
xzhu on Thu Apr 16 15:40:59 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added codes to tell binarize2 whether the input gray image is scaled or  not. This is because that scaling reduced noise, there is no nessary to do extra processing in binarize2.

Revision 1.90  1998/04/16 00:17:57  xzhu
xzhu on Wed Apr 15 17:17:10 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Mapped 255 to 254 in background image, left 255 just for blasted image area.

Revision 1.89  1998/04/07 22:25:58  xzhu
xzhu on Tue Apr 7 15:24:30 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Changed offset calculation in wp_autoBlastMasks to follow rounding down rule.

Revision 1.88  1998/04/07 00:16:48  xzhu
xzhu on Mon Apr 6 17:15:47 PDT 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Put erode 5*5 back.

Revision 1.87  1998/03/23 22:56:32  xzhu
xzhu on Mon Mar 23 14:56:24 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed the bug to rotate 270 with 180 degree. Added error message for that orientation was not zero and 2nd pass was smaller than expanded 1st pass image. When we do rotation 90, 270 or 180, the orignal offsets are also changed. For example, before rotation you crop right and bottom,  this is the change of width and height. After rotation, you may crop left and bottom, right and top or left and top according to what rotation you did. They are both offsets x or y and width or height change.

Revision 1.86  1998/03/21 02:42:45  xzhu
xzhu on Fri Mar 20 18:42:36 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed a crash bug.

Revision 1.85  1998/03/21 02:19:37  xzhu
xzhu on Fri Mar 20 18:12:36 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

To help fixing aligment problem, changes back to strle3 for erosion image mask before blasting. Moved check whether 2nd pass image include 1st pass image to w_autoScaneSecondPassComplet from w_autoScanSetSecondPassRect. If image reduced from 2nd pass image is smaller than 1 st pass image, do crop to every segmented image. If map-ink is selected, do crop 1st image also. This just handle image width and height changed, not handle orignal offset changed. Therefore I put temp comments there to wait for whether we support offset changing.

Revision 1.84  1998/03/18 20:48:57  xzhu
xzhu on Wed Mar 18 12:48:00 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added oritation check before crop first pass image and segmented images in w_autoScanSecondPassRectSet().

Revision 1.83  1998/03/18 02:23:54  xzhu
xzhu on Tue Mar 17 18:22:09 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Reduced balsting area in foreground, background and selector to make up some aligment error.

Revision 1.82  1998/03/17 20:38:22  xzhu
xzhu on Tue Mar 17 12:38:14 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed a bug.

Revision 1.81  1998/03/17 03:03:03  xzhu
xzhu on Mon Mar 16 19:00:50 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Modified w_autoScanSetSecondPassRect. When rectangle calculated from first pass image is out of the recond pass image (0,0, width, height). We modified rectangle, text plane, saved first pass image and segmented images.

Revision 1.80  1998/03/14 02:18:07  xzhu
xzhu on Fri Mar 13 18:16:21 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Dilated foreground in vertical direction.

Revision 1.79  1998/03/11 20:24:11  xzhu
xzhu on Wed Mar 11 12:17:03 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Refine foreground for 2 pass. Color Character will not be mixed.

Revision 1.78  1998/03/04 01:31:24  xzhu
xzhu on Tue Mar 3 17:30:30 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Did some changes in 1 pass map-ink.

Revision 1.77  1998/02/27 19:12:38  xzhu
xzhu on Fri Feb 27 11:11:35 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added depth check for map-ink. Support 4-bits scaled image in 1 pass.

Revision 1.76  1998/02/17 03:04:58  ddavies
ddavies on Mon Feb 16 19:03:21 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Back out of cheating way of computing foreground map for one pass scanning.
We now use w_scaleMaskedArea again.  Also don't scale the primary plane
of the image up to text resolution, then throw it away, then scale all three
planes up again.  Just scale what we need.

Revision 1.75  1998/02/13 01:33:26  ddavies
ddavies on Thu Feb 12 17:31:31 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Use w_grayPropagate4Mask in one-pass processing for the time being.
This checkin gives everyone time to evaluate speed and image quality.
Also rearrange things in wp_onePassMapink.  This code is about 50%
faster and the resulting XIF file size is 5% - 10% smaller.
Also added comments to the one pass code.

Revision 1.74  1998/02/10 18:54:35  xzhu
xzhu on Tue Feb 10 10:54:06 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed problem in scaling second pass image. This problem was found by Andrew.

Revision 1.73  1998/02/07 00:25:53  xzhu
xzhu on Fri Feb 6 16:19:27 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added documentBoundingBox in skew detection to avoid zero confidence for small images, which was found by Andrew. Enabled option shadingRemove and lineRemove for one pass binary input.

Revision 1.72  1998/02/05 22:47:26  xzhu
xzhu on Thu Feb 5 14:45:17 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Now in one pass textPicture mode, we can select noSegmentation. If noSegmentation is set, we can do deskew, crop and AIE, then output gray or color image without binary and segmented images.

Revision 1.71  1998/01/31 23:23:11  aber
aber on Sat Jan 31 15:22:46 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Removed unused variables, they've permanently gone away.

Revision 1.70  1998/01/31 02:05:06  ddavies
ddavies on Fri Jan 30 18:04:45 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Comment out unused variables in wp_autoPhotoCrop.

Revision 1.69  1998/01/31 00:58:07  aber
aber on Fri Jan 30 16:56:28 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Moved code to do crop and rotate into a separate routine (w_cropAndRotate()).
Bug fix, if cropping but no deskew, still call photo crop detection.

Revision 1.68  1998/01/31 00:34:07  xzhu
xzhu on Fri Jan 30 16:30:29 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

If last strip of image is very small, the resolution of color and binary may be diffferent after scaling them to same size. w_opGrayBin will return Err. So we set resolutionDPI before w_opGrayBin.

Revision 1.67  1998/01/20 18:47:15  aber
aber on Tue Jan 20 10:46:50 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Change ERR_CHECK to ERR_CHECK_NO_MSG, to avoid assertion error.

Revision 1.66  1998/01/15 01:54:11  xzhu
xzhu on Wed Jan 14 17:53:19 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Use w_binarizeGetCurImgH to get binary and gray image aligment.

Revision 1.65  1998/01/10 01:31:01  xzhu
xzhu on Fri Jan 9 17:28:14 PST 1998

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

To speed up, in one pass, if page type is cPictureOnly and input image is not binarary, we don't do binarization during passing data by checking page type.

Revision 1.64  1997/12/04 01:11:19  xzhu
xzhu on Wed Dec 3 17:06:00 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added codes to update second pass image width and height in w_autoScanSetSecondPassRect. If second pass image is less than rectangle, we crop rectangle and second pass image pixr. But the rectangle orignal x, y can't be changed. Changed control in pastePicture, we keep segmented image for API AIE analysis, thwy will be deleted by API.

Revision 1.63  1997/11/19 01:16:46  xzhu
xzhu on Tue Nov 18 17:16:06 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Speeded up background segmentation in two pass model.

Revision 1.62  1997/11/15 01:24:55  xzhu
autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

CommonCheck: Use ScaledInt1616 resolution to calculate Micas.

Revision 1.61  1997/10/29 23:59:59  xzhu
xzhu on Wed Oct 29 15:58:27 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed uninitialized memory problem.

Revision 1.60  1997/10/28 19:57:03  xzhu
xzhu on Tue Oct 28 11:55:36 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added new function w_autoScanSetSecondPassRect(). If scanner can't do crop to second pass image, autoscan will do crop in w_autoScanSecondPassData strip by strip.

Revision 1.59  1997/10/28 18:46:59  aber
aber on Tue Oct 28 10:46:42 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Don't do what we were told not to (crop etc.).

Revision 1.58  1997/08/28 16:42:13  xzhu
xzhu on Thu Aug 28 09:41:28 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed memory leaking.

Revision 1.57  1997/08/27 23:31:21  xzhu
xzhu on Wed Aug 27 16:29:14 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added alignment check algorithm in 2 pass. It improves image quality.

Revision 1.56  1997/08/27 00:06:00  aber
aber on Tue Aug 26 17:01:08 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Add new photo crop function (see crop3.c). If new photo crop fails,
run old method (currently only happens for extreme cases, like when
a white business card is placed on a white platen, and we can't
locate the edge of the card; in this case, the polygon of the edge
is zero points, and we fail gracefully).

Revision 1.55  1997/08/18 16:27:11  xzhu
xzhu on Mon Aug 18 09:25:27 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed NT compiler complain.

Revision 1.54  1997/08/16 00:44:12  xzhu
xzhu on Fri Aug 15 17:42:12 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

In 1 pass, if pageType == cPictureOnly, do photo crop and free text plane and others.

Revision 1.53  1997/08/12 22:27:17  xzhu
xzhu on Tue Aug 12 15:25:41 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Improved 1 pass text and image aligment and set smallest deskew angle (0.3 degree).

Revision 1.52  1997/08/12 00:37:14  xzhu
xzhu on Mon Aug 11 17:36:00 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Enable crop and deskew in 2 pass model. If skew angle is less than 0.3 degree, we do not deskew it.

Revision 1.51  1997/07/29 19:26:18  xzhu
xzhu on Tue Jul 29 12:20:56 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed problem in one pass image copy. If picture resolution is higher than orignal image, the repeating line copying used wrong width and height limitation. Also added condition that foreground resolution must be lower or equal text resolution since we just do reducing not expansion. We may release this condition later by adding expanding function.

Revision 1.50  1997/07/28 19:26:31  xzhu
xzhu on Mon Jul 28 12:25:37 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added pastePictures in two pass model.

Revision 1.49  1997/07/19 00:03:39  xzhu
xzhu on Fri Jul 18 17:02:49 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added resolution setting in autoBlastMasks to avoid error message.

Revision 1.48  1997/07/08 17:13:54  xzhu
xzhu on Tue Jul 8 10:13:29 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed bug #128 reported by Andrew.

Revision 1.47  1997/06/27 01:09:23  xzhu
xzhu on Thu Jun 26 18:08:27 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Improve one pass background generation.

Revision 1.46  1997/06/26 20:00:11  ddavies
autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

CommonCheck: Add inclusion of new file rast.pub.

Revision 1.45  1997/05/23 20:38:29  xzhu
xzhu on Fri May 23 13:36:13 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

In orientation detection, adde remove lines before second orentation check.

Revision 1.44  1997/05/22 19:45:38  xzhu
xzhu on Thu May 22 12:44:37 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added control for options no-foreground and no-background.

Revision 1.43  1997/05/21 00:52:55  xzhu
xzhu on Tue May 20 17:51:16 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Blasting image segments' masks is optional.

Revision 1.42  1997/05/16 18:09:57  ddavies
ddavies on Fri May 16 11:09:38 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Rename w_scaleMapedArea to w_scaleMaskedArea.
This routine is called here, but not implemented here.

Revision 1.41  1997/05/13 00:08:02  xzhu
xzhu on Mon May 12 17:03:09 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Did double check " text " to find image is photo or not.

Revision 1.40  1997/05/02 01:23:48  xzhu
xzhu on Thu May 1 18:22:17 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed crash reported by Andrew. When angle is large, use rotateArb.

Revision 1.39  1997/04/18 21:50:45  xzhu
xzhu on Fri Apr 18 14:48:57 PDT 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Made background map more smooth and output image with better quality.

Revision 1.38  1997/04/15 20:09:39  ddavies
autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Substitute NoInit (uninitialized) version of pixr creation routines where
possible.

Revision 1.37  1997/03/29 01:26:04  xzhu
autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

CommonCheck: Add a input parameter IsTextSeg in w_textMask(). If want keep more text, turn on this parameter, otherwise set it 0.

Revision 1.36  1997/03/19 00:42:58  xzhu
xzhu on Tue Mar 18 16:39:55 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Moved out variables defined in if () loop. If we don't move them out, when error happens in loop, it will cause memory leak.

Revision 1.35  1997/03/11 22:44:04  xzhu
xzhu on Tue Mar 11 14:43:01 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

When we output rdiff, if there is no image, we should do normal return not error return.

Revision 1.34  1997/03/06 23:52:46  xzhu
xzhu on Thu Mar 6 15:51:41 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Now we can set picture, fg and bg different resolution to save more space.

Revision 1.33  1997/03/05 21:11:16  xzhu
xzhu on Wed Mar 5 13:08:56 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Put foreground and background in two pass model.

Revision 1.32  1997/02/28 01:10:57  xzhu
xzhu on Thu Feb 27 17:07:40 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Modified generating foreground and background map.

Revision 1.31  1997/01/02 16:20:10  macleod
macleod on Thu Jan 2 08:18:31 PST 1997

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

When using the "binarize2" algorithm, upconvert each band's
primary channel to 8 bits/pixel if it's only 4, because "binarize2"
doesn't work at 4bpp, even though it advertises that it does. This
adds a little time for the allocation and depth conversion.

Revision 1.30  1996/12/30 15:13:40  macleod
autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

When doing "DIR"-type segmentation, use a larger STREL for dilating
the background when the resolution is above 75dpi. This should be
more general than it is.

Revision 1.29  1996/11/21 01:25:40  xzhu
xzhu on Wed Nov 20 17:24:08 PST 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added new function w_autoScanOutputRdiff().

Revision 1.28  1996/11/20 22:12:57  macleod
macleod on Wed Nov 20 14:11:22 PST 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added a hack of a fix which is incomplete: sometimes on very small
strips (i.e. one line high on the last strip) the scale factor gets
truncated down to zero, which causes everything to blow up when
it would be better if it did nothing.

Revision 1.27  1996/10/23 21:57:52  macleod
macleod on Wed Oct 23 14:57:27 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed a bug in two-pass mode where if deskewing was off but cropping
was on, the text and pictures didn't line up.

Revision 1.26  1996/10/02 05:29:19  macleod
macleod on Tue Oct 1 22:28:50 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Initialize bounding box so that if cropping didn't happen, there's
still a valid bounding box.

Revision 1.25  1996/09/04 22:55:35  macleod
macleod on Wed Sep 4 15:55:23 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Changed the computation of target image sizes from rounding to truncating,
because all of the ipcore scaling routines truncate, and the rounding
was causing 1-pixel-off errors.

Revision 1.24  1996/08/07 23:07:18  macleod
macleod on Wed Aug 7 16:06:07 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added "pastePictures" option to "DIR"-style autoscan. This
shoves all of the pictures into the background plane in the segmented
result, rather than keeping around separate images and masks for each
 contone region. This does require that the background and picture
resolutions be the same.

Revision 1.23  1996/07/30 14:14:49  macleod
macleod on Tue Jul 30 07:14:42 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed "nag" script warnings

Revision 1.22  1996/07/23 15:29:52  macleod
macleod on Tue Jul 23 08:29:30 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Took out a comment about a debugging routine that was left in,
confusing the "nag" script

Revision 1.21  1996/07/22 22:02:56  macleod
macleod on Mon Jul 22 15:01:01 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed a bug in which binary images without discernable text caused
everything to blow up.

Revision 1.20  1996/07/22 18:35:39  macleod
macleod on Mon Jul 22 11:35:15 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

If an image was recognized as a "photo", the text plane pointer was
left uninitialized, causing things to fail randomly.

Revision 1.19  1996/07/22 14:12:04  macleod
macleod on Mon Jul 22 07:11:17 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed "nag script" warnings--some missing colons, missing ERR_CLEANUP
wrappers, misspelled arguments in comment, etc.

Revision 1.18  1996/07/19 22:30:13  macleod
macleod on Fri Jul 19 15:29:52 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Forgot to cast floating constants to single precision, resulting in
a Microsoft compiler warning

Revision 1.17  1996/07/19 15:25:24  macleod
macleod on Fri Jul 19 08:23:01 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added support to autoscan for the second adaptive thresholding algorithm,
known as "binarize2." The algorithm can be used by invoking the
"cAutoOption_binarize2" flag. Also added support for "DIR"-style
segmentation, where we compute a foreground image, a background image,
a selector (binary) image, and several smaller contone images representing
the pictures on the page. At this point, these two features should not
be considered "released", because they haven't been tested enough,
but I'm checking them in so that they can be version controlled and
so that others can try them.

Revision 1.16  1996/05/31 17:14:53  macleod
macleod on Fri May 31 10:14:36 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added "f" after character constant to make VC++ happy

Revision 1.15  1996/05/30 22:21:52  macleod
macleod on Thu May 30 15:21:10 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed a bug in which uninitialized memory was sometimes used
as a scale factor in certain modes

Revision 1.14  1996/05/29 18:08:01  macleod
macleod on Wed May 29 11:07:20 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Oops--left a debugging flag on, which caused writing of partial
thumbnail images to "/tmp/thumbXXXX.tif".

Revision 1.13  1996/05/22 16:36:01  macleod
macleod on Wed May 22 09:35:40 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Changed references to "BEAST_ANGLE" with checks for conf == 0 and skew == 0

Revision 1.12  1996/05/21 23:33:13  macleod
macleod on Tue May 21 16:32:56 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed several bugs which caused some modes to fail

Revision 1.11  1996/05/20 21:14:04  macleod
macleod on Mon May 20 14:12:41 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed two bugs: First, 1-bit one-pass scanning blew up randomly due
to not initializing the image pointer to NULL, which was used to
check whether an image pointer was used. Second, 4-bit one-pass scanning
didn't work if one was making thumbnails, due to the fact that it
scaled the depth up to 8 bits/pixel and then complained about the
fact that the thumbnail wasn't 8 bits/pixel.

Revision 1.10  1996/05/15 01:07:00  macleod
autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fix nag script warnings

Revision 1.9  1996/05/09 20:43:47  macleod
macleod on Thu May 9 13:42:49 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Added the ability to generate a "thumbnail" image as the scan progresses.
The thumbnail image is passed into the init routine, and should have been
allocated by the caller. It gets updated when significant things happen.

Revision 1.8  1996/04/30 21:27:32  macleod
macleod on Tue Apr 30 14:26:40 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Make a couple of floating-point constants floats instead of doubles
in order to silence VC4 compiler warnings

Revision 1.7  1996/04/29 15:03:41  macleod
macleod on Mon Apr 29 08:02:14 PDT 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed a couple of bugs having to do with not initializing things
in the state structure. Changed segmentation call to add additional
parameter.
Modified calls to a couple of Container routines.

Revision 1.6  1996/03/13 00:59:59  macleod
macleod on Tue Mar 12 16:59:44 PST 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

Fixed two memory leaks and an egregious bug

Revision 1.5  1996/03/07 23:21:07  macleod
macleod on Thu Mar 7 15:19:13 PST 1996

autoscan.c -> /product/ipcore/ipwrap/src/RCS/autoscan.c,v

one-pass now accepts 1-, 4- or 8-bit data
second-pass now accepts 1-, 4-, or 8-bit data
first-pass now accepts 4- or 8-bit data
Added w_autoScanFinish, to release resources
Added w_autoScanGetProcessingState, a call which gets the current
processing step on which the pipeline is working (see autoscan.pub for
a list of codes) -- this can be used to give feedback to the user
 
*/
